/*
 * LySwingConstraint.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import leon.view.*;
import leon.view.graphic.*;


/**
 * This class implements the internal description of a Constraint object as described in the dtd.
 *
 * <p>A constraint allows to specify for a given graphical component the way it will be placed
 * inside its parent (to be specific, the graphical component that contains it).</p>
 *
 * @version $Id: LySwingConstraint.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     LyComponent
 */
public class LySwingConstraint extends LySwingObject
	implements LyGraphicalConstraint
{
	public static final String __VERSION =
		"$Id: LySwingConstraint.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The constraint description. */
	protected LyConstraint     _constraint;

	/**
	 * Class constructor of a constraint object.
	 *
	 * @param constraint a constraint object that is applied to a swing component.
	 */
	public LySwingConstraint(LyConstraint constraint)
	{
		super(constraint);
		_constraint = constraint;
	}

	/**
	 * Builds a contraint object for the layout of a label or component. The constraints are the
	 * same for every data node component. The only change that appears is on the <code>
	 * requestedConstraint</code> parameter which applies another constraint to the original. If no
	 * requested constraint is specified (its value is <code>null</code>), then the default
	 * GridBagConstaint is returned.
	 *
	 * @param  dataNode            the component to build its constraints, not used in this
	 *                             implementation
	 * @param  requestedConstraint the constraints to apply on the dataNode component
	 * @param  defaultGB           a default GridBagConstraint to use if no other constraint is
	 *                             usable
	 * @return the constraints of the specified component
	 */
	public static GridBagConstraints buildConstraint(LyComponent dataNode,
		LyConstraint requestedConstraint, GridBagConstraints defaultGB)
	{
		if (requestedConstraint == null)
			return defaultGB;

		GridBagConstraints c = new GridBagConstraints();

		// ### using specified user filter ###

		// index of current cell (x coordinate, or RELATIVE)
		if (defaultGB != null)
			c.gridx = defaultGB.gridx; // keep default
		else
			c.gridx = GridBagConstraints.RELATIVE;

		// index of current cell (y coordinate, or RELATIVE)
		if (defaultGB != null)
			c.gridy = defaultGB.gridy;
		else
			c.gridy = GridBagConstraints.RELATIVE;

		// width of current cell (cell count or REMAINDER)
		if (requestedConstraint.isRemainder())
			c.gridwidth = GridBagConstraints.REMAINDER;
		else if ((defaultGB != null)
				&& (requestedConstraint.getColSpan() == LyConstraint.DEFAULT_VALUE)) // not found
			c.gridwidth = defaultGB.gridwidth;
		else if (requestedConstraint.getColSpan() < 0) // Kept for former compatibility
			c.gridwidth = GridBagConstraints.REMAINDER;
		else
			c.gridwidth = requestedConstraint.getColSpan();

		// height of current cell (cell count or REMAINDER)
		if ((defaultGB != null) && (requestedConstraint.getRowSpan() == LyConstraint.DEFAULT_VALUE)) // not
			// found
			c.gridheight = defaultGB.gridheight;
		else if (requestedConstraint.getRowSpan() <= 0)
			c.gridheight = defaultGB.gridheight;
		else
			c.gridheight = requestedConstraint.getRowSpan();

		// Shall the component fill the space given for the cell
		switch (requestedConstraint.getFill())
		{
		case HORIZONTAL:
			c.fill = GridBagConstraints.HORIZONTAL;
			break;
		case VERTICAL:
			c.fill = GridBagConstraints.VERTICAL;
			break;
		case BOTH:
			c.fill = GridBagConstraints.BOTH;
			break;

		case DEFAULT:
			if (defaultGB != null)
			{
				c.fill = defaultGB.fill;
				break;
			}

		case NONE:
		default:
			c.fill = GridBagConstraints.NONE;
			break;
		}

		// if container is bigger than the preferred size, how much
		// of the free space is occupied by current cell (barycentric)
		if ((defaultGB != null)
				&& (requestedConstraint.getColWeight() == LyConstraint.DEFAULT_VALUE))
			c.weightx = defaultGB.weightx;
		else
			c.weightx = requestedConstraint.getColWeight();
			// if container is bigger than the preferred size, how much
			// of the free space is occupied by current cell (barycentric)

		if ((defaultGB != null)
				&& (requestedConstraint.getRowWeight() == LyConstraint.DEFAULT_VALUE))
			c.weighty = defaultGB.weighty;
		else
			c.weighty = requestedConstraint.getRowWeight();

		// internal padding of the component in its cell
		if ((defaultGB != null)
				&& (requestedConstraint.getCellPadding() == LyConstraint.DEFAULT_VALUE))
		{
			c.ipadx = defaultGB.ipadx;
			c.ipady = defaultGB.ipady;
		}
		else
		{
			// If negative, keep 0 as ipad
			c.ipadx = Math.max(requestedConstraint.getCellPadding(), 0);
			c.ipady = Math.max(requestedConstraint.getCellPadding(), 0); // same as x
		}

		// external padding of the component between cells
		if ((defaultGB != null)
				&& (requestedConstraint.getCellSpacing() == LyConstraint.DEFAULT_VALUE))
			c.insets = defaultGB.insets;
		else
		{
			int sp = requestedConstraint.getCellSpacing();

			// If negative, keep 0 as insets
			if (sp < 0)
				sp = 0;

			c.insets = new Insets(sp, sp, sp, sp);
		}

		if ((defaultGB != null)
				&& (requestedConstraint.getHAlign() == LyConstraintBase.HAlign.DEFAULT)
				&& (requestedConstraint.getVAlign() == LyConstraintBase.VAlign.DEFAULT))
			c.anchor = defaultGB.anchor;
		else
		{
			if (requestedConstraint.getHAlign() == LyConstraintBase.HAlign.LEFT)
			{
				switch (requestedConstraint.getVAlign())
				{
				case MIDDLE:
					c.anchor = GridBagConstraints.WEST;
					break;

				case BOTTOM:
					c.anchor = GridBagConstraints.SOUTHWEST;
					break;

				case TOP:
				default:
					c.anchor = GridBagConstraints.NORTHWEST;
					break;
				}
			}

			if (requestedConstraint.getHAlign() == LyConstraintBase.HAlign.CENTER)
			{
				switch (requestedConstraint.getVAlign())
				{
				case MIDDLE:
					c.anchor = GridBagConstraints.NORTH;
					break;

				case BOTTOM:
					c.anchor = GridBagConstraints.CENTER;
					break;

				case TOP:
				default:
					c.anchor = GridBagConstraints.SOUTH;
					break;
				}
			}

			if (requestedConstraint.getHAlign() == LyConstraintBase.HAlign.RIGHT)
			{
				switch (requestedConstraint.getVAlign())
				{
				case MIDDLE:
					c.anchor = GridBagConstraints.EAST;
					break;

				case BOTTOM:
					c.anchor = GridBagConstraints.SOUTHEAST;
					break;

				case TOP:
				default:
					c.anchor = GridBagConstraints.NORTHEAST;
					break;
				}
			}

			if (((defaultGB != null)
						&& (requestedConstraint.getHAlign() == LyConstraintBase.HAlign.DEFAULT)))
			{
				boolean eastSide = ((defaultGB.anchor == GridBagConstraints.NORTHEAST)
							|| (defaultGB.anchor == GridBagConstraints.EAST)
							|| (defaultGB.anchor == GridBagConstraints.SOUTHEAST));

				switch (requestedConstraint.getVAlign())
				{
				case MIDDLE:
					c.anchor = (eastSide ? GridBagConstraints.EAST : GridBagConstraints.WEST);
					break;

				case BOTTOM:
					c.anchor = (eastSide ? GridBagConstraints.SOUTHEAST
										 : GridBagConstraints.SOUTHWEST);
					break;

				case TOP:
				default:
					c.anchor = (eastSide ? GridBagConstraints.NORTHEAST
										 : GridBagConstraints.NORTHWEST);
					break;
				}
			}
		}

		return c;
	} // end method buildConstraint

	/**
	 * Prints the content of the given GridBagConstraints (useful for debug).
	 *
	 * @param c the GridBagConstraints to dump
	 */
	public static void dump(GridBagConstraints c)
	{
		if (c == null)
		{
			System.out.println("  NULL ");
			return;
		}

		System.out.println("  c.gridx       = " + c.gridx);
		System.out.println("  c.gridy       = " + c.gridy);
		System.out.println("  c.gridwidth   = " + c.gridwidth);
		System.out.println("  c.gridheight  = " + c.gridheight);
		System.out.println("  c.fill        = " + c.fill);
		System.out.println("  c.weightx     = " + c.weightx);
		System.out.println("  c.weighty     = " + c.weighty);
		System.out.println("  c.ipadx       = " + c.ipadx);
		System.out.println("  c.ipady       = " + c.ipady);
		System.out.println("  c.insets      = " + c.insets);
		System.out.println("  c.anchor      = " + c.anchor);
	}

} // end class LySwingConstraint
