/*
 * LySwingDate.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.event.*;
import java.text.*;
import java.util.*;
import javax.swing.*;

import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;
import leon.view.win.swing.component.LyJFormattedInput._LyFormattedField;


/**
 * This class implements the Date selector object (as described in the dtd) for swing viewer.
 *
 * <p>An input field of Date type allows to display a graphical component allowing to input or to
 * consult date and/or hour. It does not comprise any element. It is possible to specify the type of
 * the date as well as the display format.</p>
 *
 * @version $Id: LySwingDate.java 34861 2012-03-19 14:38:00Z jlsaouli $
 * @author  Lyria
 * @see     LyRelation
 */
public class LySwingDate extends LySwingFieldItem
	implements LySwingComponent,
		LySetLabelInterface,
		FocusListener,
		LyGraphicalDate,
		LyActionListener,
		LyGetLabelInterface,
		LySetEnabledInterface,
		LySetEditableInterface,
		KeyListener,
		ItemListener
{
	public static final String __VERSION =
		"$Id: LySwingDate.java 34861 2012-03-19 14:38:00Z jlsaouli $";

	/** Width of the selector's button image. */
	public static int		   IMAGE_W_OFFSET = 6;

	/** Height of the selector's button image. */
	public static int		   IMAGE_H_OFFSET = 15;

	/** The view independent date description. */
	protected LyDate		   _date;

	/** The swing component of this date object. */
	protected Component		   _rootComponent;

	/** Date selector component. */
	protected Object		   _choice;

	/** Time selector component. */
	protected Component		   _timeSelector;

	/** AM/PM selector component. */
	protected JComboBox		   _amPmSelector;

	/** Keeps the last valid date on error. */
	protected String		   _previousValidStringDate;

	/** Keeps the last valid time on error. */
	protected String		   _previousValidStringTime;

	/** Global format of the date (date + time) in component. */
	protected DateFormat	   _dateFormat;

	/** Date format for time selector. */
	protected DateFormat	   _timeDateFormat;

	/** Date format for date selector. */
	protected DateFormat	   _choiceDateFormat;

	/** Date format of the parent component. */
	protected DateFormat	   _parentDateFormat;

	/** Idle mode. If true, this component ignores notifications. */
	protected boolean		   _idle;

	/**
	 * Class constructor of a Date object for the Swing target.
	 *
	 * @param date a DATE component defined for every type of view targets. This date is initialized
	 *             specificaly for the Swing view.
	 */
	public LySwingDate(LyDate date)
	{
		super(date);
		_date = date;

		initialize();
		setContent();
	}

	/**
	 * Initializes the current date component. Depending on the type of the <code>_date</code>
	 * description, the component will be initialized to let the choice only of date, time or both,
	 * with different formats.
	 */
	public void initialize()
	{
		// create a multiple component : Combo (Choice + Date selector)
		// bound with a popup relation.
		_idle = false;
		_previousValidStringDate = "";
		_previousValidStringTime = "";
		// XXX FORMAT date : move this in config files

		LyEnvironment     environment = getEnvironment();
		String			  format;
		LyJSelector		  choice;
		LyJFormattedInput timeSelector;

		boolean	isTimeHourMin = (_date.getType()== LyDate.Type.DATE_HOUR_MIN) || (_date.getType()== LyDate.Type.HOUR_MIN);
		String	timeFormat = environment.getStringDateFormat(isTimeHourMin ? LyEnvironment.TYPE_HOUR_MIN : LyEnvironment.TYPE_TIME);
		boolean	addAmPm = false;

		if (timeFormat.endsWith(" a"))
		{
			addAmPm = true;
			timeFormat = timeFormat.substring(0, timeFormat.lastIndexOf(" a"));
		}

		if (!timeFormat.startsWith("hh") && timeFormat.startsWith("h"))
			timeFormat = "h" + timeFormat;

		String	timeParseFormat = (addAmPm ? "[1-12]:[0-59]" : "[0-23]:[0-59]");
		
		if (!isTimeHourMin)
			timeParseFormat += ":[0-59]";
		
		if (_date.getType()== LyDate.Type.DATE)
        {
			format = environment.getStringDateFormat(LyEnvironment.TYPE_DATE);
			choice = buildDateSelector(format);
			_choice = choice;
			_rootComponent = choice;
			_dateFormat = new SimpleDateFormat(format);
			_choiceDateFormat = _dateFormat;
        }
        else if (_date.getType()== LyDate.Type.TIME)
        {
			timeSelector = buildTimeSelector(timeParseFormat, timeFormat);
			_timeSelector = timeSelector;

			if (addAmPm)
			{
				JPanel		timePanel = new JPanel();
				
				_amPmSelector = buildAmPmSelector();
				timePanel.setLayout(new BorderLayout(3, 0));
				timePanel.setOpaque(false);
				timePanel.add(timeSelector, BorderLayout.CENTER);
				timePanel.add(_amPmSelector, BorderLayout.EAST);
				_rootComponent = timePanel;
			}
			else
				_rootComponent = timeSelector;
			
			_dateFormat = new SimpleDateFormat(timeFormat);
			_timeDateFormat = _dateFormat;
        }
        else if (_date.getType()== LyDate.Type.HOUR_MIN)
        {
			timeSelector = buildTimeSelector(timeParseFormat, timeFormat);
			_timeSelector = timeSelector;

			if (addAmPm)
			{
				JPanel		timePanel = new JPanel();
				
				_amPmSelector = buildAmPmSelector();
				timePanel.setLayout(new BorderLayout(3, 0));
				timePanel.setOpaque(false);
				timePanel.add(timeSelector, BorderLayout.CENTER);
				timePanel.add(_amPmSelector, BorderLayout.EAST);
				_rootComponent = timePanel;
			}
			else
				_rootComponent = timeSelector;
			
			_dateFormat = new SimpleDateFormat(timeFormat);
			_timeDateFormat = _dateFormat;
        }
        else if (_date.getType()== LyDate.Type.DATE_HOUR_MIN)
		{
			JPanel p = new JPanel();
			// Flandres / FFT00039
			p.setLayout(new BorderLayout(3, 0));
			p.setOpaque(false);

			// Date
			format = environment.getStringDateFormat(LyEnvironment.TYPE_DATE);
			choice = buildDateSelector(format); // new LyJSpecialSelector("",environment);

			_choiceDateFormat = new SimpleDateFormat(format);
			p.add(choice, BorderLayout.WEST);

			// time
			timeSelector = buildTimeSelector(timeParseFormat, timeFormat);
			
			if (addAmPm)
			{
				JPanel		timePanel = new JPanel();
				
				_amPmSelector = buildAmPmSelector();
				timePanel.setLayout(new BorderLayout(3, 0));
				timePanel.setOpaque(false);
				timePanel.add(timeSelector, BorderLayout.CENTER);
				timePanel.add(_amPmSelector, BorderLayout.EAST);
				p.add(timePanel, BorderLayout.EAST);
			}
			else
				p.add(timeSelector, BorderLayout.CENTER);
			
			format = environment.getStringDateFormat(LyEnvironment.TYPE_DATE_HOUR_MIN);
			_dateFormat = new SimpleDateFormat(format);
			_timeDateFormat = new SimpleDateFormat(timeFormat);
			_choice = choice;
			_timeSelector = timeSelector;
			_rootComponent = p;
			
			p.addFocusListener(this);
		}
        else if (_date.getType()== LyDate.Type.DATE_TIME)
		{
			JPanel p = new JPanel();
			// Flandres / FFT00039
			p.setLayout(new BorderLayout(3, 0));
			p.setOpaque(false);

			format = environment.getStringDateFormat(LyEnvironment.TYPE_DATE);
			String parseFormat = LyString.replace(format, "MM", "[1-12]");

			parseFormat = LyString.replace(parseFormat, "dd", "[1-31]");
			parseFormat = LyString.replace(parseFormat, "yyyy", "[1-9999]");
			// Date
			choice = buildDateSelector(format);
			p.add(choice, BorderLayout.CENTER);
			_choiceDateFormat = new SimpleDateFormat(format);

			// time
			timeSelector = buildTimeSelector(timeParseFormat, timeFormat);
			
			if (addAmPm)
			{
				JPanel		timePanel = new JPanel();
				
				_amPmSelector = buildAmPmSelector();
				timePanel.setLayout(new BorderLayout(3, 0));
				timePanel.setOpaque(false);
				timePanel.add(timeSelector, BorderLayout.CENTER);
				timePanel.add(_amPmSelector, BorderLayout.EAST);
				p.add(timePanel, BorderLayout.EAST);
			}
			else
				p.add(timeSelector, BorderLayout.EAST);
			
			format = environment.getStringDateFormat(LyEnvironment.TYPE_DATE_TIME);
			_dateFormat = new SimpleDateFormat(format);
			_timeDateFormat = new SimpleDateFormat(timeFormat);
			_choice = choice;
			_timeSelector = timeSelector;
			_rootComponent = p;
			
			p.addFocusListener(this);
		}

		if (_date.getFormat() != null)
			_parentDateFormat = new SimpleDateFormat(_date.getFormat());
		else
		{
			format = environment.getStringDateFormat(LyDate.getType(_date.getType()));
			_parentDateFormat = new SimpleDateFormat(format);
		}

		setLabel(_date.getLabel());

		// This handler is used to validate the form using ENTER key or close using ESCAPE key
		LySwingFormKeyHandler keyHandler = new LySwingFormKeyHandler(getActionTransmitter());

		keyHandler.setNotifyValidate(false); // already default in swing

		// Escape means current time !
		if (_choice instanceof Component)
			((Component)_choice).addKeyListener(keyHandler);
		if (_timeSelector != null)
			_timeSelector.addKeyListener(keyHandler);
		if (_amPmSelector != null)
			_amPmSelector.addItemListener(this);
	} // end method initialize

	/**
	 * Builds a date selector for the given format.
	 *
	 * @param  format the date format
	 * @return a formatted input that let anyone enter a date conforming to the specified date
	 *         format
	 */
	protected LyJFormattedInput buildDateSelector(String format)
	{
		LyEnvironment   environment = getEnvironment();
		Locale		    locale = environment.getLocale();
		Color		    enabledColor = getColor("LY_ENABLED_BACKGROUND");
		Color		    disabledColor = getColor("LY_DISABLED_BACKGROUND");
		Font		    font = getFont("LY_DEFAULT_FONT");

		LyJDateSelector sel = new LyJDateSelector(locale);

		String		    parseFormat = LyString.replace(format, "MM", "[1-12]");

		parseFormat = LyString.replace(parseFormat, "dd", "[1-31]");
		parseFormat = LyString.replace(parseFormat, "yyyy", "[1-9999]");
		// Date
		LyJFormattedDateInput choice = new LyJFormattedDateInput("", parseFormat,
				getImage("LY_ARROW_DOWN_BIG"), environment, true);

		choice.setEnabledColor(enabledColor);
		choice.setDisabledColor(disabledColor);
		choice.setFont(font);
		sel.setDateFormat(format);
		//if (_date.isShowArrows())
		new LyJPopupDateSelector(choice, sel, environment);
		choice.setMaxChars(10);
		// choice.setNCols(6);

		if (sel != null)
		{
			sel.setFont(font);
			LyActionTransmitter transmitter = sel.getActionTransmitter();

			if (transmitter != null)
				transmitter.addActionListener(this);
		}
		// Register for Escape event
		choice.addKeyListener(this);
		return choice;
	} // end method buildDateSelector

	/**
	 * Builds a date selector for the given format.
	 *
	 * @param  format     the date format
	 * @param  dateFormat the date format
	 * @return a formatted input that let anyone enter a date and time conforming to the specified
	 *         formats
	 */
	protected LyJFormattedInput buildTimeSelector(String format, String dateFormat)
	{
		LyEnvironment		  environment = getEnvironment();
		Image				  up = getImage("LY_ARROW_UP");
		Image				  down = getImage("LY_ARROW_DOWN");
		Color				  enabledColor = getColor("LY_ENABLED_BACKGROUND");
		Color				  disabledColor = getColor("LY_DISABLED_BACKGROUND");
		Font				  font = getFont("LY_DEFAULT_FONT");

		LyJFormattedTimeInput timeSelector = new LyJFormattedTimeInput(null, format, up, down,
				environment, _date.isShowArrows());

		timeSelector.setFormat(dateFormat);
		timeSelector.setEnabledColor(enabledColor);
		timeSelector.setDisabledColor(disabledColor);
		timeSelector.setFont(font);

		return timeSelector;
	}
	
	/**
	 * Builds an AM/PM selector.
	 */
	protected JComboBox buildAmPmSelector()
	{
		JComboBox	amPmSelector = new JComboBox();
		
		amPmSelector.setFont(getFont("LY_DEFAULT_FONT"));
		amPmSelector.addItem("AM".intern());
		amPmSelector.addItem("PM".intern());

		return amPmSelector;
	}

	/**
	 * Sets the date of the date selector component.
	 *
	 * @param date the new date value to use
	 * @see   #getDate()
	 */
	protected void setDate(String date)
	{
		setDate(date, true);
	}

	/**
	 * Sets the date of the date selector component.
	 *
	 * @param date the new date value to use
	 * @param notify indicates if new date must be notified to control layer.
	 * @see   #getDate()
	 */
	protected void setDate(String date, boolean notify)
	{
		((LyJSelector)_choice).setValue(date, notify);
	}

	/**
	 * Sets the time of the time selector component.
	 *
	 * @param time the new time value to use
	 * @see   #getTime()
	 */
	protected void setTime(String time)
	{
		setTime(time, true);
	}

	/**
	 * Sets the time of the time selector component.
	 *
	 * @param time the new time value to use
	 * @param notify indicates if new time must be notified to control layer.
	 * @see   #getTime()
	 */
	protected void setTime(String time, boolean notify)
	{
		((LyJFormattedInput)_timeSelector).setValue(time, notify);
	}

	/**
	 * Gets the date of the date selector.
	 *
	 * @return the formatted date
	 * @see    #setDate(String)
	 */
	protected String getDate()
	{
		return ((LyJSelector)_choice).getValue();
	}

	/**
	 * Gets the time of the time selector.
	 *
	 * @return the formatted time
	 * @see    #setTime(String)
	 */
	protected String getTime()
	{
		return ((LyJFormattedInput)_timeSelector).getValue();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetEditableInterface#setEditable(boolean)
	 */
	public void setEditable(boolean editable)
	{
		if (_choice != null)
			((LyJSelector)_choice).setEditable(editable);

		if (_timeSelector != null)
			((LyJFormattedInput)_timeSelector).setEditable(editable);

		if (_amPmSelector != null)
		{
			Color	enabledColor = getColor("LY_ENABLED_BACKGROUND");
			Color	disabledColor = getColor("LY_DISABLED_BACKGROUND");
			
			_amPmSelector.setBackground(editable ? enabledColor : disabledColor);
			_amPmSelector.setEnabled(editable);
		}		
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetEnabledInterface#setEnabled(boolean)
	 */
	public void setEnabled(boolean status)
	{
		if (_choice != null)
			((LyJSelector)_choice).setEnabled(status);

		if (_timeSelector != null)
			((LyJFormattedInput)_timeSelector).setEnabled(status);

		if (_amPmSelector != null)
		{
			Color	enabledColor = getColor("LY_ENABLED_BACKGROUND");
			Color	disabledColor = getColor("LY_DISABLED_BACKGROUND");
			
			_amPmSelector.setBackground(status ? enabledColor : disabledColor);
			_amPmSelector.setEnabled(status);
		}		
	}

	/**
	 * Adds action listeners to the time and date selectors.
	 */
	protected void setContent()
	{
		LyActionTransmitter transmitter;

		if (_choice instanceof LyEventTransmitter)
		{
			transmitter = ((LyEventTransmitter)_choice).getActionTransmitter();

			if (transmitter != null)
				transmitter.addActionListener(this);
		}

		if (_timeSelector instanceof LyEventTransmitter)
		{
			transmitter = ((LyEventTransmitter)_timeSelector).getActionTransmitter();

			if (transmitter != null)
				transmitter.addActionListener(this);
		}
	}

	/**
	 * Checks the values and corrects them if an event is generated.
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		if (actionEvent.getValue() == null)
		{
			LyComponent parentField = _date.getParent();

			if (parentField == null)
				return; // May occur when focus is lost because of closing window
			// if is optional : set null
			if (((LyFormField)parentField).isOptional())
				setLabel(null, true);
			else if ((_previousValidStringDate.length() > 0)
					|| (_previousValidStringTime.length() > 0))
				setLabel((_previousValidStringDate + " " + _previousValidStringTime).trim(), true);
			else
				setLabel(null, true);

			_date.setLabel(getLabel());

			// may notify any listeners
			if (_actionTransmitter != null)
				_actionTransmitter.processActionEvent(actionEvent);
		}
		else if (actionEvent.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
		{
			// Update Parent info
			_idle = true;

			checkLabel();

			_date.setLabel(getLabel());
			_idle = false;

			actionEvent.setValue(getLabel());
			
			// may notify any listeners
			if (_actionTransmitter != null)
				_actionTransmitter.processActionEvent(actionEvent);
		}
	} // end method actionPerformed

	/**
	 * Checks the date and time label content. If the labels don't correspond to their selectors,
	 * they are reset and reinitialized with the correct values.
	 */
	protected void checkLabel()
	{
		String  labelDate, labelTime;
		boolean choiceSetToNull = false;

		// Get each part of the label
		labelDate = ((_choice == null) ? null : getDate());
		labelTime = ((_timeSelector == null) ? null : getTime());

		// Check Choice
		if (_choice != null)
		{
			// String val = getDate();
			if ((labelDate == null) || (labelDate.length() == 0))
			{
				LyComponent parentField = _date.getParent();

				if ((parentField instanceof LyFormField)
						&& (((LyFormField)parentField).isOptional()))
				{
					if (_date.getLabel() != null)
						choiceSetToNull = true; // <- reset to null
				}
				else if ((_previousValidStringDate != null)
						&& (_previousValidStringDate.length() > 0))
					setDate(_previousValidStringDate);
				else
				{
					if (_date.getLabel() != null)
						choiceSetToNull = true; // <- reset to null
				}
			}
			else
			{
				// checkDate returns the date in the good locale format (US, FRENCH or UNKNOWN)
				String newDate = LyDateFormatter.checkDate(labelDate, _choiceDateFormat);

				// if the checking has failed, raise a format error
				// if (newDate == null)
				// raiseFormatError(val);

				if (newDate == null)
					newDate = labelDate; // Validation is achieved in LyDateFieldInfo

				boolean isIdle = _idle;

				if ((_timeSelector != null) && (getTime() == null))
				{
					String     newTime = null;
					DateFormat timeFormat = null;

					// If a time format is given, use it to set the time label
					if (_timeDateFormat != null)
						timeFormat = _timeDateFormat;
					else
					{
						if (_date.getType() == LyDate.Type.DATE_HOUR_MIN)
							timeFormat = new SimpleDateFormat("HH:mm");
                        else 
							timeFormat = new SimpleDateFormat("HH:mm:ss");
					}

					LyEnvironment environment = getEnvironment();
					Calendar	  cal = environment.calendarFactory();

					cal.set(Calendar.AM_PM, Calendar.AM);
					cal.setTime(new Date());

					newTime = timeFormat.format(cal.getTime());

					setTime(newTime, false);
					_previousValidStringTime = newTime;
				}

				if (!isIdle)
					_idle = true;

				setDate(newDate, false);
				_previousValidStringDate = getDate();

				if (!isIdle)
					_idle = false;
			}
		}

		if ((_timeSelector != null) && (labelTime != null))
		{
			// also reset time
			boolean isIdle = _idle; // is it already idle ?

			if (!isIdle)
				_idle = true;
			if (choiceSetToNull)
				setTime(null, false);
			else
			{
				setTime(labelTime, false);
				_previousValidStringTime = labelTime;

				if ((_choice != null) && (labelDate == null))
				{
					String     newDate = null;
					DateFormat dateFormat = _dateFormat;

					// Use current date (see WIN)
					newDate = dateFormat.format(new Date());

					setDate(newDate, false);
					_previousValidStringDate = newDate;
				}
			}
			if (!isIdle)
				_idle = false;
		}
	} // end method checkLabel

	/**
	 * Gets the label for the time and date this object represents, without validations.
	 *
	 * @see leon.view.getinterface.LyGetLabelInterface#getLabel()
	 * @see #setLabel(String)
	 */
	public String getLabel()
	{
		StringBuffer sb = new StringBuffer();
		String		 val;

		if (_choice != null)
		{
			val = getDate();

			if ((val != null) && (val.length() > 0))
				sb.append(val);
			else
				return null; // <- Stop here
		}

		if (_timeSelector != null)
		{
			val = getTime();

			if ((val != null) && (val.length() > 0))
			{
				if (sb.length() > 0)
					sb.append(" ");

				val = removeBlanks(val);

				sb.append(val);
			}
		}

		if (sb.length() == 0)
			return null;

		if (_amPmSelector != null)
			sb.append(" ").append(_amPmSelector.getSelectedItem());
		
		return sb.toString(); // Keep as is : no validation
	} // end method getLabel

	/**
	 * Removes blank spaces in the specified string.
	 *
	 * @param  str the time string containing the blanks to remove
	 * @return the time string without blank spaces
	 */
	protected String removeBlanks(String str)
	{
		if (str == null)
			return null;

		StringTokenizer st = new StringTokenizer(str, " ");
		StringBuffer    sb = new StringBuffer(str.length());

		while (st.hasMoreTokens())
			sb.append(st.nextToken());

		return sb.toString();
	}

	/**
	 * Sets the time and date components from the specified label.
	 *
	 * @param notify indicates if new label must be notified to control layer.
	 * @see leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
	 * @see #getLabel()
	 */
	public void setLabel(String label)
	{
		setLabel(label, false);
	}

	/**
	 * Sets the time and date components from the specified label.
	 *
	 * @param label the new label for date and time.
	 * @param notify indicates if new label must be notified to control layer.
	 * @see #getLabel()
	 */
	public void setLabel(String label, boolean notify)
	{
		if (_idle)
			return;

		if (_timeSelector != null)
		{
			String timeLabel = label;
			// if the label equals previous date : the time is set to current time
			if ((label != null) && (label.equals(_previousValidStringDate)))
			{
				Date date = new Date();

				if ((_parentDateFormat == null) || (_timeDateFormat == null))
				{
					String	timeFormat = getEnvironment().getStringDateFormat(LyEnvironment.TYPE_TIME);

					if ((timeFormat != null) && !timeFormat.startsWith("hh") && timeFormat.startsWith("h"))
						timeFormat = "h" + timeFormat;
					
					timeLabel = new SimpleDateFormat(timeFormat).format(date);
				}
				else
					timeLabel = _parentDateFormat.format(date);
			}

			if ((_parentDateFormat == null) || (_timeDateFormat == null))
			{
				setTime(timeLabel, notify);
				_previousValidStringTime = timeLabel;
			}
			else
			{
				Date d = parseDate(timeLabel, _parentDateFormat);

				if (d == null)
					setTime(null, notify);
				else
				{
					String s = _timeDateFormat.format(d);

					setTime(s, notify);
					_previousValidStringTime = s;
				}
				
				if ((_amPmSelector != null) && (d != null))
				{
					String	amPm = new SimpleDateFormat("a").format(d);

					_amPmSelector.setSelectedItem(amPm.intern());
				}
			}
		}

		if ((_choice != null) && (label != _previousValidStringDate)) // no need to change
		{
			if ((_parentDateFormat == null) || (_choiceDateFormat == null))
			{
				_previousValidStringDate = (label == null) ? "" : label;
				setDate((label == null) ? "" : label, notify);
			}
			else
			{
				Date d = parseDate(label, _parentDateFormat);

				if (d == null)
				{
					_previousValidStringDate = "";
					setDate("", notify);
				}
				else
				{
					String newValue = _choiceDateFormat.format(d);

					_previousValidStringDate = newValue;
					setDate(newValue, notify);
				}
			}
		}
	} // end method setLabel

	/**
	 * Shows an error to the user when the selected date does not match the required format. Keeps
	 * consistency with the web view.
	 *
	 * @param value the invalid value selected by the user.
	 */
	protected void raiseFormatError(String value)
	{
		LyFormField formField = (LyFormField)_date.getParent();
		LyCaption   caption = formField.getCaption();
		String	    name;

		if (caption != null)
			name = translate(caption.getLabel());
		else
			name = "";

		// create the error message
		String errorMessage = getMessage("LY_ER_MATCH_FORMAT", new String[] { value, name });

		// display the error
		getViewManager().showError(_date.getParentView(), errorMessage);
	}

	/**
	 * Parses a date String for the given date format and raises a warning message if needed. If
	 * both the dateString and dateFormat are <code>null</code>, then <code>null</code> is returned.
	 *
	 * @param  dateString the date String we want to parse
	 * @param  dateFormat the DateFormat instance used for parsing
	 * @return the date if the parsing succeeded, <code>null</code> otherwise
	 */
	protected Date parseDate(String dateString, DateFormat dateFormat)
	{
		// WARNING : Should be kept consistent with method in LyDateFieldInfo
		if ((dateString == null) || (dateFormat == null))
			return null;

		try
		{
			return (dateFormat.parse(dateString));
		}
		catch (ParseException e)
		{
			// Raise a Warning message
			LyEnvironment environment = getEnvironment();

			environment.logWarning("Invalid date format in LySwingDate " + " (display format) :"
				+ dateString + " : " + e.getMessage());
		}

		return null;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
	 */
	public void keyTyped(KeyEvent evt)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
	 */
	public void keyPressed(KeyEvent evt)
	{
		if (evt.getSource() == _choice)
		{
			int key = evt.getKeyCode();

			if (key == KeyEvent.VK_ESCAPE)
			{
				// Set current date / time
				Date d = new Date();

				if (_parentDateFormat != null)
					setLabel(_parentDateFormat.format(d), true);
				else if (_dateFormat != null)
					setLabel(_dateFormat.format(d), true);
				else
					setLabel(d.toString(), true);

				actionPerformed(new LyActionEvent(this, null, getLabel(),
						LyActionDemand.TYPE_VALUE_CHANGED));
			}
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
	 */
	public void keyReleased(KeyEvent evt)
	{
		// MAZ : SITA FFT00036 & FFT00037 
		LyComponent	parent = _date.getParent();
		boolean		isInTable = false;
		
		while (parent != null)
		{
			if (parent instanceof LyRow)
			{
				isInTable = true;
				break;
			}
			
			parent = parent.getParent();
		}
		
		if (isInTable && (evt.getKeyCode() == KeyEvent.VK_ESCAPE))
		{
			if (_choice instanceof LyJFormattedInput)
			{
				_LyFormattedField formattedText = (_LyFormattedField)((LyJFormattedInput)_choice).getTextField();
				
				((LyJFormattedInput)_choice).focusLost();
				formattedText.focusLost(null);
			}
		}
	}

	/**
	 * Don't do anything if the focus is gained.
	 *
	 * @see java.awt.event.FocusListener#focusGained(java.awt.event.FocusEvent)
	 */
	public void focusGained(FocusEvent e)
	{
		if ((_choice != null) && (_choice instanceof LyJSelector))
			((LyJSelector)_choice).getTextField().requestFocus();
	}

	/**
	 * If the focus is lost, a notify event must be sent.
	 *
	 * @see java.awt.event.FocusListener#focusLost(java.awt.event.FocusEvent)
	 * @see _LyJSpecialSelector#notifyController()
	 */
	public void focusLost(FocusEvent e)
	{
	}

	/**
	 * When AM or PM is selected, a notify event must be sent.
	 */
	public void itemStateChanged(ItemEvent e)
	{
		if (e.getSource() == _amPmSelector)
		{
			_idle = true;

			checkLabel();

			String	label = getLabel();
			
			_date.setLabel(label);
			_idle = false;
			
			if (_actionTransmitter != null)
				_actionTransmitter.processActionEvent(new LyActionEvent(this, null, label, LyActionDemand.TYPE_VALUE_CHANGED));
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _rootComponent;
	}

	/* Internal Class */

	/**
	 * Special Swing selector for the date elements.
	 */
	public class _LyJSpecialSelector extends LyJSelector
	{
		private static final long serialVersionUID = -2625393714591357232L;

		/**
		 * Creates and initializes a new special selector for the given text value and environment.
		 *
		 * @param value selector text value
		 * @param env   current environment
		 */
		public _LyJSpecialSelector(String value, LyEnvironment env)
		{
			super(value, env, _date.isShowArrows());

			// FFT00559 Handle the Escape key
			InputMap  inputMap = _text.getInputMap();
			ActionMap actionMap = _text.getActionMap();
			KeyStroke stroke = KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0);
			String    key = "setCurrentDate";

			inputMap.put(stroke, key);
			Action action = new AbstractAction(key)
				{
					private static final long serialVersionUID = -3258279455939518565L;

					public void actionPerformed(ActionEvent e)
					{
						// Set current date / time
						Date d = new Date();

						if (_parentDateFormat != null)
							LySwingDate.this.setLabel(_parentDateFormat.format(d));
						else if (_dateFormat != null)
							LySwingDate.this.setLabel(_dateFormat.format(d));
						else
							LySwingDate.this.setLabel(d.toString());

						LySwingDate.this.actionPerformed(new LyActionEvent(this, null, getLabel(),
								LyActionDemand.TYPE_VALUE_CHANGED));
					}
				};
			actionMap.put(key, action);
		}

		/*
		 * (non-Javadoc)
		 * @see leon.view.win.swing.component.LyJSelector#addButton(java.awt.Image,
		 * java.lang.String)
		 */
		public LyJButton addButton(Image image, String label)
		{
			LyJButton button = new LyJButton(image, label, null, "selector_button");

			button.setRollover(false);
			Icon icon = new ImageIcon(image);

			button.setIcon(icon);
			button.setPreferredSize(new Dimension(icon.getIconWidth() + IMAGE_W_OFFSET,
					icon.getIconHeight()
					+ IMAGE_H_OFFSET));
			button.setHorizontalAlignment(SwingConstants.CENTER);
			button.setVerticalAlignment(SwingConstants.CENTER);
			button.setMargin(new Insets(0, 0, 0, 1));

			LyActionTransmitter transmit = button.getActionTransmitter();

			if (transmit != null)
				transmit.addActionListener(_LyJSpecialSelector.this);

			_buttons.add(button);
			return button;
		}

		/**
		 * Checks the buttons of this component and enables or disables them to propagate this
		 * component enabled property.
		 *
		 * @see leon.view.win.swing.component.LyJSelector#checkButtons()
		 */
		public void checkButtons()
		{
			if (!isEnabled() || !_buttons.isEnabled())
			{
				int i, n = _buttons.getComponentCount();

				for (i = 0; i < n; i++)
					_buttons.getComponent(i).setEnabled(false);
			}
			else
			{
				int i, n = _buttons.getComponentCount();

				for (i = 0; i < n; i++)
					_buttons.getComponent(i).setEnabled(true);
			}
		}

		// FFT00413
		/**
		 * Sends notification events on changed values. This also saves the date value if the focus
		 * is lost.
		 *
		 * @see LySwingDate#focusLost(FocusEvent)
		 */
		public void notifyController()
		{
			String value = getValue();
			// check if something has changed

			if ((!this._idle)
					&& (((this._lastValue == null) && (value != null))
						|| ((this._lastValue != null) && (!this._lastValue.equals(value)))))
			{
				if (this._actionTransmitter != null)
					this._actionTransmitter.processActionEvent(new LyActionEvent(this, null, value,
							LyActionDemand.TYPE_VALUE_CHANGED));
			}
		}

	} // end class _LyJSpecialSelector

} // end class LySwingDate
