/*
 * LySwingDialog.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import leon.misc.*;
import leon.view.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;


/**
 * This class implements the Dialog object (as described in the dtd) for swing viewer.
 *
 * <p>A dialog window is, generally, linked to a main window of an application. Unlike the main
 * window, it contains neither a menu, nor a message bar. Its display can be modal (blocking) or
 * not. This type of window is generally used to allow a user to interact with an application: input
 * request, presentation of the results, warning messages, questions, etc.</p>
 *
 * @version $Id: LySwingDialog.java 35461 2012-05-10 14:17:48Z nhamon $
 * @author  Lyria
 * @see     LyDialog
 */
public class LySwingDialog extends LySwingView
	implements LySwingComponent,
		LyGraphicalDialog
{
	public static final String __VERSION =
		"$Id: LySwingDialog.java 35461 2012-05-10 14:17:48Z nhamon $";

	/**
	 * Maximum ratio, in percent, of the window between its size (with / height) and the screen.
	 * Default is 90%.
	 */
	public static int		   MAX_SIZE_RATIO = 90;

	/**
	 * Default reserved width for scrollbars, in pixels. Defaut value is 25. If scrollbars are
	 * needed for height or width, we reserve the size in the other direction.
	 */
	public static int		   SCROLLBAR_WIDTH = 25;

	/** The view independent dialog description. */
	protected LyDialog		   _dialog;

	/** Graphical component used to represent this dialog. */
	protected Dialog		   _dialogComp;

	/** The minimum size of the dialog if any specified. */
	protected Dimension		   _minimumSize;

	/** The Swing Component for this dialog window. */
	protected JDialog		   _dialogWindow;

	/** Parent view. */
	protected LyDialog		   _parentView;

	/**
	 * Class constructor of a Dialog window object for the Swing target.
	 *
	 * @param dialog a DIALOG component defined for every type of view targets. This dialog is
	 *               initialized and configured specificaly for the Swing view.
	 */
	public LySwingDialog(LyDialog dialog)
	{
		super(dialog);
		_dialog = dialog;

		initialize();
		setContent();
		configureDefaultButton();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingView#initialize()
	 */
	public void initialize()
	{
		super.initialize();

		// Initialize component
		_dialogComp = createDialog();
		_dialogComp.setUndecorated(!_dialog.isDecorated());
		setLabel(_dialog.getLabel());

		if ((_dialog.getWidth() > 0) && (_dialog.getHeight() > 0))
			setSize(_dialog.getWidth(), _dialog.getHeight());

		_dialogComp.setResizable(_dialog.isResizable());

		if ((_dialog.getX() >= 0) && (_dialog.getY() >= 0))
			setLocation(_dialog.getX(), _dialog.getY());

		_winAdapter = new _LyWindowAdapter();
		_dialogComp.addWindowListener(_winAdapter);
		_compAdapter = new _LyComponentAdapter();
		_dialogComp.addComponentListener(_compAdapter);
		
		if (_dialog.isResizable())
		{
			int minWidth = getIntResource("LY_WINDOW_MIN_WIDTH");
			int minHeight = getIntResource("LY_WINDOW_MIN_HEIGHT");

			if ((minWidth > 0) || (minHeight > 0))
				updateMinimumSize(_dialog, new Dimension(minWidth, minHeight));
		}
	} // end method initialize

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyFreeInterface#free()
	 */
	public void free()
	{
		super.free();

		if (_dialogComp != null)
			_dialogComp.dispose();
		_dialogComp = null;
		_dialog = null;
	}

	/**
	 * Sets the default button. At this time no default button is defined in Leonardi, so the policy
	 * is to set the first command button as the default button.
	 */
	public void configureDefaultButton()
	{
		LyArea area = _dialog.getArea();

		if (area != null)
		{
			// Retrieve all the command buttons of this area and give the focus to the
			// first one.
			Vector<LyCommand> commands = new Vector<LyCommand>(2);

			area.getComponents(LyCommand.class, commands);

			if (commands.size() > 0)
			{
				LyCommand command = commands.elementAt(0);
				JButton   button = ((JButton)((LySwingComponent)command.getInternalView())
							.getComponent());

				((JDialog)_dialogComp).getRootPane().setDefaultButton(button);
			}
		}
	}

	/**
	 * Shows dialog component and gives focus to its first editable field.
	 */
	protected void showDialog()
	{
		if (!_dialogComp.isModal())
		{
			_dialogComp.setVisible(true);
			_dialogComp.toFront();
			_dialogComp.requestFocus();

			handleFocus();
		}
		else
			// Showing a dialog window from AWT-Event-Queue crashes the JVM
			// so we display it starting from another thread
			new Thread()
				{

					/* (non-Javadoc)
					 * @see java.lang.Runnable#run()
					 */
					public void run()
					{
						handleFocus();
						_dialogComp.setVisible(true);
					}
				}.start();
	}

	/**
	 * Gives the focus to the first editable field in the form.
	 */
	protected void handleFocus()
	{
	    if (_dialog == null)
	        return;
	    
		// Give focus to the first editable field of the dialog
		Vector<LyFormField> result = new Vector<LyFormField>();

		_dialog.getComponents(LyFormField.class, result);
        
		int i, n = result.size();

		for (i = 0; i < n; i++)
		{
			LyFormField field = result.elementAt(i);

			if (field.isEditable() && field.isEnabled())
			{
				LyFieldItem		 item = field.getFieldItem();
				LySwingComponent view = (LySwingComponent)item.getInternalView();

				view.getComponent().requestFocus();
				break;
			}
		}
	}

	/**
	 * Creates a new Dialog instance for the view and initializes it. May be overriden if needed.
	 *
	 * @return a dialog instance
	 */
	public Dialog createDialog()
	{
		LyView			   parent;
		JFrame			   frame = null;
		LySwingViewManager manager = (LySwingViewManager)getViewManager();
		boolean			   noModal = manager.hasProperty("NO_MODAL");
		boolean			   modal = !noModal && _dialog.isModal();

		// Seek parent view
		parent = manager.getView(_dialog.getParentId());

		if (parent == _dialog)
			parent = null; // Prevent all case of same id parent % dialog

		// If we have found the parent, try to get its frame in the
		// WinComponent

		JDialog parentDialog = null;

		// FFT00314 : if parent is a dialog, build the new dialog
		// with this dialog as owner.
		if (parent instanceof LyDialog)
		{
			LySwingComponent internal = (LySwingComponent)parent.getInternalView();

			// At the top, we have either a Frame or dialog ...
			if ((internal != null) && (internal.getComponent() instanceof JDialog))
				parentDialog = (JDialog)internal.getComponent();
		}

		if (parentDialog != null)
		{
			_dialogWindow = new JDialog(parentDialog, null, modal);
			_parentView = (LyDialog)parent;

			((LySwingDialog)_parentView.getInternalView()).setWaiting(true, false);

			_dialogWindow.addWindowListener(new _LyDialogAdapter());
		}
		else
		{
			frame = manager.seekParentJFrame(parent);
			_dialogWindow = new JDialog(frame, modal);
		}

		_dialogWindow.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);

		if (LyEnvironment.isOperatingSystem(LyEnvironment.OS_SOLARIS))
			_dialogWindow.addWindowListener(new _LyWindowAdapter3());

		return _dialogWindow;
	} // end method createDialog

	/* (non-Javadoc)
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _dialogComp;
	}

	/* (non-Javadoc)
	 * @see leon.view.win.swing.LySwingView#setArea(java.awt.Component)
	 */
	protected void setArea(Component area)
	{
		if (_dialogComp != null)
		{
			Container	contentPane = ((JDialog)_dialogComp).getContentPane();
			
			if (contentPane.getComponentCount() > 0)
			{
				contentPane.remove(0);
				contentPane.add(area, "Center");
				_dialogComp.validate();
				_dialogComp.doLayout();
				contentPane.repaint();
				
				// If dialog is non resizable with a default width or height, do not change size
				if (!isResizable()
						&& ((_dialog.getHeight() <= 0) || (_dialog.getWidth() <= 0)))
					resetContent();
			}
			else
				contentPane.add(area, "Center");
		}
	}

	/* (non-Javadoc)
	 * @see leon.view.setinterface.LySetWaitingInterface#setWaiting(boolean)
	 */
	public void setWaiting(boolean status)
	{
		setWaiting(status, true);
	}

	/**
	 * Sets the waiting status of this component. and the enabled state of the component.
	 *
	 * @param status    sets the waiting property of the given component, false unsets it.
	 * @param setCursor sets the waiting image of the cursor
	 *                  (<code>LySwingViewManager.WAIT_CURSOR</code> or <code>
	 *                  LySwingViewManager.DEFAULT_CURSOR</code>)
	 */
	public void setWaiting(boolean status, boolean setCursor)
	{
		if (_dialogComp != null)
		{
			if (setCursor)
			{
				if (status)
					_dialogComp.setCursor(LySwingViewManager.WAIT_CURSOR);
				else
					_dialogComp.setCursor(LySwingViewManager.DEFAULT_CURSOR);
			}

			// XXX Windows Bug on dialog/Frame comp
			// if not enabled, cursor is unchanged or returns
			// to default
			// _dialogComp.setEnabled(!status);

			// Make our own recursion ...
			LyArea area = _dialog.getArea();
			Object internal = area.getInternalView();

			if (internal instanceof LySetWaitingInterface)
				((LySetWaitingInterface)internal).setWaiting(status);
		}
	}

	/**
	 * Sets the title bar of the dialog.
	 *
	 * @param label the new title bar label
	 * @see   LyGraphicalDialog#setLabel(String)
	 */
	public void setLabel(String label)
	{
		if ((_dialogComp != null) && (label != null))
        {
            label = translate(label);
            label = label.replace('\n', ' ');
			_dialogComp.setTitle(translate(label));
        }
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingView#isResizable()
	 */
	public boolean isResizable()
	{
		return _dialogComp.isResizable();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetVisibleInterface#setVisible(boolean)
	 */
	public void setVisible(boolean status)
	{
		if (_dialogComp == null)
			return;

		if (_dialogComp.isShowing() == status)
		{
			if (status)
				_dialogComp.toFront();
			return;
		}

		if (status)
		{
			if (!_dialogComp.isVisible())
				doLayout();

			if (!_resized)
			{
				pack();
				checkSize();
			}
		}

		if (status)
			showDialog();
		else
			_dialogComp.setVisible(false);

		_requiredSizeDate = new Date();
	} // end method setVisible

	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalDialog#setArea(leon.view.LyArea)
	 */
	public void setArea(LyArea area)
	{
		LySwingComponent	internal = (LySwingComponent)area.getInternalView();
		Component			areaComp = internal.getComponent();
		
		setArea(areaComp);
		configureDefaultButton();
		handleFocus();
	}

	/**
	 * Updates the minimum size occupied by this component.
	 *
	 * @param child child component in this container. This parameter is not used in this
	 *              implementation.
	 * @param d     minimum size to use
	 */
	public void updateMinimumSize(LyComponent child, Dimension d)
	{
		if (!getEnvironment().testResource("LY_CHECK_WINDOW_MINIMUM_SIZE"))
			return;

		if ((_minimumSize == null) && (d != null) && (_dialogComp.isResizable()))
			_dialogComp.addComponentListener(new _LyWindowSizeListener());

		_minimumSize = d;

		if (_minimumSize != null)
		{
			int minWidth = getIntResource("LY_WINDOW_MIN_WIDTH");
			int minHeight = getIntResource("LY_WINDOW_MIN_HEIGHT");

			if (minWidth > _minimumSize.width)
				_minimumSize.width = minWidth;
			if (minHeight > _minimumSize.height)
				_minimumSize.height = minHeight;
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingView#checkSize()
	 */
	protected void checkSize()
	{
		Dimension d = _dialogComp.getPreferredSize();
		int[]     dscreen = getEnvironment().getScreenSize(_dialogComp.getParent());
		int		  w, h, off_w, off_h, ratio_w, ratio_h;

		w = d.width;
		h = d.height;
		ratio_w = (dscreen[0] * MAX_SIZE_RATIO) / 100;
		ratio_h = (dscreen[1] * MAX_SIZE_RATIO) / 100;
		off_w = 0;
		off_h = 0;

		if (w > ratio_w)
		{
			w = ratio_w;
			off_h = SCROLLBAR_WIDTH; // Reserve space for scrollbar
		}
		if (h > ratio_h)
		{
			h = ratio_h;
			if (w != d.width)
				off_h = 0; // Abord reservation : does not fit.
			else
				off_w = SCROLLBAR_WIDTH; // Reserve space for scrollbar
		}

		setSize(w + off_w, h + off_h);
	}

    /**
     * Accessor to the location.
     *
     * @see LySetBoundsInterface#setLocation
     */
    public void setLocation(int x, int y)
    {
        if (_dialogComp == null)
            return;

        // If location is set while view is not shown, check the size of the view before placing it
        if (!_dialog.isDecorated()
            && ((x < 0) && (y < 0))
            && (_dialogComp.getParent() != null))
        {
            Point       location = _dialogComp.getParent().getLocationOnScreen();
            Point       mouseLocation = _dialogComp.getParent().getMousePosition();
            
            if (location != null && mouseLocation != null)
            {
                x = location.x + mouseLocation.x;
                y = location.y + mouseLocation.y;
            }
        }

        super.setLocation(x, y);
    }

	/**
	 * Maximum size ratio of the view on the screen in %.
	 *
	 * @return 90 by default
	 * @see    LySwingView#getMaxSizeRatio()
	 */
	public int getMaxSizeRatio()
	{
		return MAX_SIZE_RATIO;
	}

	/**
	 * Sets all the components of this dialog to their preferred size and readjust the size of the
	 * dialog.
	 */
	public void repack()
	{
		if (_dialog == null)
			return;
		int w, h;

		if ((_dialog.getWidth() >= 0) && (_dialog.getHeight() >= 0))
		{
			w = _dialog.getWidth();
			h = _dialog.getHeight();
		}
		else
		{
			pack();

			// Check screen size and make it consistent
			Dimension d = _dialogComp.getPreferredSize();
			int[]     dscreen = getEnvironment().getScreenSize(_dialogComp);

			w = Math.min(d.width, (dscreen[0] * MAX_SIZE_RATIO) / 100);
			h = Math.min(d.height, (dscreen[1] * MAX_SIZE_RATIO) / 100);
		}
		setSize(w, h);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingView#snapshot(java.lang.String)
	 */
	public void snapshot(String file)
	{
		LySwingUtils.snapshot(_dialogComp, file);
	}

	/* Internal Class */

	/**
	 * Handle the closing events of this dialog.
	 */
	public class _LyDialogAdapter extends WindowAdapter
	{

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.WindowListener#windowClosing(java.awt.event.WindowEvent)
		 */
		public void windowClosing(WindowEvent e)
		{
			if (e.getSource() == _dialogWindow)
				_parentView.setWaiting(false);
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.WindowListener#windowClosed(java.awt.event.WindowEvent)
		 */
		public void windowClosed(WindowEvent e)
		{
			if (e.getSource() == _dialogWindow)
				_parentView.setWaiting(false);
		}
	}


	/**
	 * Handles the opening events of this dialog.
	 */
	protected class _LyWindowAdapter3 extends WindowAdapter
	{

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.WindowListener#windowOpened(java.awt.event.WindowEvent)
		 */
		public void windowOpened(WindowEvent e)
		{
			if (LyEnvironment.isOperatingSystem(LyEnvironment.OS_SOLARIS))
				SwingUtilities.invokeLater(new Runnable()
					{

						/*
						 * (non-Javadoc)
						 * @see java.lang.Runnable#run()
						 */
						public void run()
						{
							repack();
						}
					});
		}
	}


	/**
	 * Handle the resizing of this dialog and force its minimum size.
	 */
	public class _LyWindowSizeListener extends ComponentAdapter
	{

		/** Minimum size (in pixels) for the borders of this dialog. */
		public int OFFSET_BORDER = 10;

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.ComponentListener#componentResized(java.awt.event.ComponentEvent)
		 */
		public void componentResized(ComponentEvent e)
		{
			if (_minimumSize == null)
				return;

			Dimension d = e.getComponent().getSize();
			boolean   resize = false;

			if (d.height < (_minimumSize.height + OFFSET_BORDER))
			{
				resize = true;
				d.height = _minimumSize.height + OFFSET_BORDER;
			}

			if (d.width < (_minimumSize.width + OFFSET_BORDER))
			{
				resize = true;
				d.width = _minimumSize.width + OFFSET_BORDER;
			}

			if (resize)
				e.getComponent().setSize(d);
		}
	} // end class _LyWindowSizeListener

} // end class LySwingDialog
