/*
 * LySwingDictionary.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.util.*;
import javax.swing.*;
import javax.swing.tree.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Dictionary object (as described in the dtd) for swing viewer.
 *
 * <p>An input field of dictionary type allows to display a graphical component containing a combo
 * with a tree.</p>
 *
 * @version $Id: LySwingDictionary.java 40917 2012-09-25 11:49:07Z smaupin $
 * @author  Lyria
 * @see     LyDictionary
 */
public class LySwingDictionary extends LySwingFieldItem
	implements LyGraphicalDictionary,
		LySelectionListener,
		LyActionListener,
		LySetEntryInterface,
		LySetEnabledInterface,
		LySetEditableInterface,
		LyGetValueInterface
{
	public static final String	    __VERSION =
		"$Id: LySwingDictionary.java 40917 2012-09-25 11:49:07Z smaupin $";

	/** The view independent dictionary description. */
	protected LyDictionary		    _dictionary;

	/** The Swing tree component for this dictionary. */
	protected LyJTreeComponent	    _treeComp;

	/** Scroll panel that contains the tree component. */
	protected Container			    _panel;

	/** Popup that contains the (scrollable) tree component. */
	protected LyJDictionarySelector _selector;

	/**
	 * Menu component for this dictionary (this menu is not used in this implementation).
	 *
	 * @deprecated
	 */
	protected Menu				    _menuComp;

	/**
	 * Window component for this dictionary (this window is not used in this implementation).
	 *
	 * @deprecated
	 */
	protected Window			    _dicoWindow;

	/**
	 * Class constructor of a Dictionary object for the Swing target.
	 *
	 * @param dico a DICTIONARY component defined for every type of view targets. This component is
	 *             initialized specificaly for the Swing view.
	 */
	public LySwingDictionary(LyDictionary dico)
	{
		super(dico);
		_dictionary = dico;

		initialize();
		setContent();
	}

	/**
	 * Initializes this component. The tree component is created and populated with the dictionary
	 * entries. Then it is placed inside a popup selector.
	 */
	public void initialize()
	{
		// We need to show that the component has the focus
		// we overide the paint of the composite to draw the border
		// In the same class, we become focus listener of the canvas
		// it allows us to know when we should draw the focus border
		Color color = getColor("LY_TREE_BACKGROUND");
		Color enabledColor = getColor("LY_ENABLED_BACKGROUND");

		if (color == null)
			color = getColor("LY_DEFAULT_BACKGROUND");
		if (color == null)
			color = Color.LIGHT_GRAY;

		// Build the tree canvas
		LyEntry		 entry = _dictionary.getEntry();
		LySwingEntry internal = ((entry == null) ? new LySwingEntry(null)
												 : (LySwingEntry)entry.getInternalView());

		_treeComp = new LyJTreeComponent(internal.getTreeElement());

		_treeComp.setNRows(_dictionary.getNRows());

		if (_dictionary.getSelectionType() == LyDictionaryBase.SelectionType.ALL)
			_treeComp.setSelectLeavesOnly(false);
		else
			_treeComp.setSelectLeavesOnly(true);

		TreeSelectionModel selectionModel = _treeComp.getSelectionModel();

		if (_dictionary.isMultiple())
		{
			selectionModel.setSelectionMode(TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
			_treeComp.setCheckboxMode(true);
		}
		else
			selectionModel.setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);

        if (testResource("LY_USE_TREE_FOR_DICTIONARY"))
        {
            _panel = new JScrollPane(_treeComp);

        }
        else
        {
            _panel = new JScrollPane(_treeComp);
            _selector = new LyJDictionarySelector("", getImage("LY_ARROW_DOWN_BIG"), getEnvironment());

            Font f = getFont("LY_DEFAULT_FONT");

            _selector.setFont(f);
            _selector.setEnabledColor(enabledColor);

            int nCols = _dictionary.getNCols();

            if (nCols <= 0)
                nCols = computeNCols(_dictionary.getEntry());

            if (nCols <= 0)
            {
                int minWidth = getIntResource("LY_COMBO_MIN_WIDTH");

                _selector.setWidth(minWidth);
            }
            else
                _selector.setNCols(nCols);

            LyJDictionaryPopupSelector popupSelector = new LyJDictionaryPopupSelector(_selector, _panel,
                    _dictionary.getEnvironment());

            if (color != null)
                popupSelector.setBackground(color);
        }
		if (_selector != null)
		{
			Dimension selectorSize = _selector.getPreferredSize();

			_treeComp.setMinWidth(selectorSize.width);
			_treeComp.setPopupMode(true);
		}

		if (color != null)
			_treeComp.setBackground(color);

		_treeComp.setVisible(true);
		_treeComp.setEnabled(true);

		// register for events
		LySelectionTransmitter selectionTransmitter = _treeComp.getSelectionTransmitter();

		if (selectionTransmitter != null)
			selectionTransmitter.addSelectionListener(this);

		// Create another tree just to compute the width of the tree when fully expanded
		LyJTreeComponent treeComp = new LyJTreeComponent(internal.getTreeElement(), false, true);

		_treeComp.setFixedWidth(treeComp.getPreferredScrollableViewportSize().width);
	} // end method initialize

	/**
	 * Computes the maximum length of all string labels in this dictionary.
	 *
	 * @param  entry a dictionary entry
	 * @return the maximum length of all dictionary labels or -1 if there is none
	 */
	protected int computeNCols(LyEntry entry)
	{
		// Search the maximum string length in all children.
		if (entry == null)
			return -1;

		String  label = translate(entry.getLabel());
		int     nCols = ((label == null) ? -1 : label.length());
		int     n = entry.getEntryCount();
		LyEntry subEntry;

		for (int i = 0; i < n; i++)
		{
			subEntry = entry.getEntry(i);
			nCols = Math.max(nCols, computeNCols(subEntry));
		}
		return nCols;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyFreeInterface#free()
	 */
	public void free()
	{
	}

	/**
	 * Fills the tree with the dictionary entries.
	 */
	public void setContent()
	{
		setEntry(_dictionary.getEntry());

		LyActionTransmitter transmitter;

		if (_selector != null)
		{
			transmitter = _selector.getActionTransmitter();

			if (transmitter != null)
				transmitter.addActionListener(this);
		}
	}

	/**
	 * Updates the current component.
	 */
	public void update()
	{
		if (_selector != null)
			_selector.setValue(getDisplayedValue());
	}

	/**
	 * Sets the root entry of the tree.
	 *
	 * @param entry the new root entry
	 * @see   LyGraphicalDictionary#setEntry(LyEntry)
	 */
	public void setEntry(LyEntry entry)
	{
		if (entry == null)
			return;
		LySwingEntry internal = (LySwingEntry)entry.getInternalView();

		((DefaultTreeModel)_treeComp.getModel()).setRoot(internal.getMutableTreeNode());
		_treeComp.updateTree(internal.getTreeElement());

		// Compute tree validity status
		if (getEnvironment().testResource("LY_TREE_VALIDITY") && _dictionary.isMultiple())
			internal.computeTreeNodeValidity(entry);

		if (_selector != null)
		{
			_treeComp.doLayout();

			String  val = getValue();
			String  displayedValue = getDisplayedValue();

			LyEntry entry2 = _dictionary.getEntry(val);

			if (entry2 == null)
				_selector.setValue(displayedValue);
			else
				_selector.setValue(translate(entry2.getLabel()));
		}
	} // end method setEntry

	/**
	 * Gets the current ids for all the dictionary entries, separated with {@link LyString#SEP}.
	 *
	 * @return the value as a string
	 */
	public String getValue()
	{
		TreePath[]			   paths = _treeComp.getSelectionPaths();
		Vector<LyJTreeElement> selection = new Vector<LyJTreeElement>();

		if ((paths != null) && (paths.length > 0))
		{
			for (int i = 0; i < paths.length; i++)
			{
				TreePath			   path = paths[i];
				DefaultMutableTreeNode mutableTreeNode = (DefaultMutableTreeNode)path
						.getLastPathComponent();
				LyJTreeElement		   node = ((LyJTreeElement)mutableTreeNode.getUserObject());

				selection.addElement(node);
			}
		}

		int			   i;
		int			   n = ((selection == null) ? 0 : selection.size());
		StringBuffer   sb = new StringBuffer();
		LyJTreeElement element;
		char		   sep = getEnvironment().getString().SEP;

		for (i = 0; i < n; i++)
		{
			element = selection.elementAt(i);
			if (sb.length() > 0)
				sb.append(sep);
			sb.append(element.getId());
		}

		if (sb.length() == 0)
			return null;

		return sb.toString();
	} // end method getValue

	/**
	 * Gets the current values for all the dictionary entries, separated with {@link
	 * LyString#VISIBLE_SEP2}.
	 *
	 * @return the value as a string
	 */
	public String getDisplayedValue()
	{
		TreePath[]			   paths = _treeComp.getSelectionPaths();
		Vector<LyJTreeElement> selection = new Vector<LyJTreeElement>();

		if ((paths != null) && (paths.length > 0))
		{
			for (int i = 0; i < paths.length; i++)
			{
				TreePath			   path = paths[i];
				DefaultMutableTreeNode mutableTreeNode = (DefaultMutableTreeNode)path
						.getLastPathComponent();
				LyJTreeElement		   node = ((LyJTreeElement)mutableTreeNode.getUserObject());

				selection.addElement(node);
			}
		}

		int			   i;
		int			   n = ((selection == null) ? 0 : selection.size());
		StringBuffer   sb = new StringBuffer();

		LyJTreeElement element;
		char		   sep = getEnvironment().getString().VISIBLE_SEP2;

		for (i = 0; i < n; i++)
		{
			element = selection.elementAt(i);
			if (i > 0)
				sb.append(sep);
			sb.append(translate(element.getLabel()));
		}

		if (sb.length() == 0)
			return null;

		return sb.toString();
	} // end method getDisplayedValue

	/**
	 * Checks or unchecks the selected element in this dictionary. This method checks every
	 * dictionary entry to find the selected one, thus the <code>selectionEvent</code> parameter is
	 * never used and may be <code>null</code>.
	 *
	 * @see leon.view.event.LySelectionListener#selectionPerformed(leon.view.event.LySelectionEvent)
	 */
	public void selectionPerformed(LySelectionEvent selectionEvent)
	{
        if (!_treeComp.isShowing())
            return;

		String		    val = getValue();
		String		    displayedValue = getDisplayedValue();

		// Update entries selection
		char		    sep = getEnvironment().getString().SEP;
		String[]	    ids = LyString.stringToList(val, sep);
		Vector<LyEntry> entries = new Vector<LyEntry>();

		if (ids == null)
			ids = new String[0];

		_dictionary.getComponents(LyEntry.class, entries);

		int     i, j, n = entries.size();
		LyEntry entry;
		String  entryId;

		for (i = 0; i < n; i++)
		{
			entry = entries.elementAt(i);
			entryId = entry.getId();

			for (j = 0; j < ids.length; j++)
			{
				if (entryId.equals(ids[j]))
				{
					entry.setChecked(true);
					break;
				}
			}

			if (j == ids.length)
				entry.setChecked(false);
		}

		if (_selector != null)
		{
			entry = _dictionary.getEntry(val);

			if (entry == null)
				_selector.setValue(displayedValue, false);
			else
				_selector.setValue(translate(entry.getLabel()), false);
		}

		String		  value = ((val == null) ? null : val);

		LyActionEvent event = new LyActionEvent(this, null, value,
				LyActionDemand.TYPE_VALUE_CHANGED);

		if (_actionTransmitter != null)
			_actionTransmitter.processActionEvent(event);
	} // end method selectionPerformed

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		if (actionEvent.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
		{
			// This part is invoked when we change the value in a selector like LyDictionarySelector
			// with del for example
			if ((actionEvent.getSource() == _selector) && (actionEvent.getValue() == null)
					&& (_actionTransmitter != null))
				_treeComp.setSelectionPaths(null);
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetEnabledInterface#setEnabled(boolean)
	 */
	public void setEnabled(boolean status)
	{
		if (_selector != null)
		{
			_selector.setButtonEnabled(status);
			_selector.setEnabled(status);
		}
		else
			_treeComp.setEnabled(status);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetEditableInterface#setEditable(boolean)
	 */
	public void setEditable(boolean status)
	{
		if (_selector != null)
		{
			_selector.setButtonEnabled(status && _selector.isEnabled());
			_selector.setEditable(status && _selector.isEnabled());
		}
		else
			_treeComp.setEnabled(status);
	}

	/**
	 * Sets the background color of the tree.
	 *
	 * @param color background color to use
	 */
	protected void setBackground(Color color)
	{
		if (_selector != null)
			_selector.setBackground(color);
		_treeComp.setBackground(color);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		if (_selector != null)
			return _selector;
		return _panel;
	}

} // end class LySwingDictionary
