/*
 * LySwingEntry.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import javax.swing.tree.*;
import leon.view.*;
import leon.view.graphic.*;
import leon.view.win.swing.component.*;
import leon.view.win.swing.component.LyJTreeElementImpl.*;


/**
 * This class implements the dictionary Entry object (as described in the dtd) for swing viewer.
 *
 * <p>An entry represents a node or a leaf of a dictionary. An entry may contain several
 * sub-entries.</p>
 *
 * @version $Id: LySwingEntry.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     LyDictionary
 */
public class LySwingEntry extends LySwingObject
	implements LyGraphicalEntry
{
	public static final String		  __VERSION =
		"$Id: LySwingEntry.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The Swing tree element for this entry. */
	protected LyJTreeElementEntryImpl _treeElement;

	/** The view independent entry description. */
	protected LyEntry				  _entry;

	/** The dictionary which this entry belongs. */
	protected LyDictionary			  _dictionary;

	/**
	 * Class constructor of an Entry object for the Swing target.
	 *
	 * @param entry an ENTRY component defined for every type of view targets. This component is
	 *              initialized specificaly for the Swing view.
	 */
	public LySwingEntry(LyEntry entry)
	{
		super(entry);
		_entry = entry;
		initialize();
		setContent();
	}

	/**
	 * Initialises the component. Retrieves the parent dictionary of this entry then configures the
	 * swing implementation of this tree element.
	 */
	public void initialize()
	{
		// find the parent tree
		LyComponent parent = _entry.getParent();

		while ((parent != null) && !(parent instanceof LyDictionary))
			parent = parent.getParent();

		boolean hasCheckbox = false;

		if (parent != null)
		{
			_dictionary = (LyDictionary)parent;
			hasCheckbox = _dictionary.isMultiple();

			if ((_dictionary.getSelectionType() == LyDictionaryBase.SelectionType.LEAVES_ONLY)
					&& (_entry.getChildCount() > 0))
				hasCheckbox = false;
		}

		_treeElement = new LyJTreeElementEntryImpl(_entry);

		_LyMutableTreeNode mutableTreeNode = new _LyMutableTreeNode();

		mutableTreeNode.setUserObject(_treeElement);

		_treeElement.setMutableTreeNode(mutableTreeNode);

		Image image;

		if (_entry.getImage() != null)
			image = (Image)getEnvironment().getImage(_entry.getImage());
		else
			image = null;

		mutableTreeNode.setNodeContent(new _LyNodeContent(image, _entry.getLabel(),
				_entry.isChecked(), _entry.isOpen()));

		if (hasCheckbox)
		{
			_treeElement.setEmptyCheckboxImage(getImage("LY_TREENODE_EMPTY_CHECKBOX"));
			_treeElement.setSelectedCheckboxImage(getImage("LY_TREENODE_CHECKED_CHECKBOX"));
		}
		else if (_dictionary.isMultiple() && getEnvironment().testResource("LY_TREE_VALIDITY")
				&& (_entry.getImage() == null))
			setValidity(_entry.getValidity());
	} // end method initialize

	/**
	 * Updates the associated dictionary.
	 */
	protected void updateParent()
	{
		LySwingDictionary dico = (LySwingDictionary)_dictionary.getInternalView();

		dico.update();
	}

	/**
	 * Returns <code>null</code>.
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return null;
	}

	/**
	 * Fills this component with its associated entry values.
	 */
	public void setContent()
	{
		int				   i, n = _entry.getEntryCount();
		_LyMutableTreeNode mutableTreeNode = (_LyMutableTreeNode)getMutableTreeNode();
		_LyMutableTreeNode subMutableTreeNode;

		// Add all the children nodes
		for (i = 0; i < n; i++)
		{
			LyEntry		 entry = _entry.getEntry(i);
			LySwingEntry internal = (LySwingEntry)entry.getInternalView();

			subMutableTreeNode = (_LyMutableTreeNode)internal.getMutableTreeNode();
			mutableTreeNode.add(subMutableTreeNode);
		}
	}

	/**
	 * Doesn't do anything.
	 *
	 * @see leon.view.graphic.LyGraphicalEntry#setChecked(boolean)
	 */
	public void setChecked(boolean status)
	{
	}

	/**
	 * Doesn't do anything.
	 *
	 * @see leon.view.graphic.LyGraphicalEntry#setOpen(boolean)
	 */
	public void setOpen(boolean open)
	{
	}

	/**
	 * Doesn't do anything.
	 *
	 * @see leon.view.setinterface.LySetImageInterface#setImage(java.lang.String)
	 */
	public void setImage(String image)
	{
	}

	/**
	 * Gets the associated tree element.
	 *
	 * @return the tree element of this entry
	 */
	public LyJTreeElement getTreeElement()
	{
		return _treeElement;
	}

	/**
	 * Gets the internal mutable tree node of this entry.
	 *
	 * @return the internal tree node
	 */
	public DefaultMutableTreeNode getMutableTreeNode()
	{
		return _treeElement.getMutableTreeNode();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetValidityInterface#setValidity(java.lang.String)
	 */
	public void setValidity(LyTreeNode.Validity validity)
	{
		if ((_treeElement != null) && (validity != null))
		{
			if (!_dictionary.isMultiple())
				return;
			if (!_entry.getEnvironment().testResource("LY_TREE_VALIDITY"))
				return;
            
			_treeElement.setValidityImage(getImage("LY_ENTRY_VALIDITY_" + validity.toString()));

			if (_entry.getEntryCount() == 0)
			{
				if (validity == LyTreeNode.Validity.INVALID)
					_entry.setChecked(false);
				else
					_entry.setChecked(true);
			}
		}
	}

	/**
	 * Computes the subtree validity.
	 *
	 * @param node current node which may be a LyTreeNode or a LyEntry (dictionnary)
	 * @see   LyJTreeElementEntryImpl#computeTreeNodeValidity(LyComponent)
	 */
	public void computeTreeNodeValidity(LyComponent node)
	{
		_treeElement.computeTreeNodeValidity(node);
	}

} // end class LySwingEntry
