/*
 * LySwingFile.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.event.*;
import java.io.*;

import javax.swing.*;

import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the File selector object (as described in the dtd) for swing viewer.
 *
 * <p>An input field of File type allows to display a graphical component allowing to select a file.
 * The type of file can be specified (text, image, audio, video). According to the form in which
 * this field is present, the graphical component allows to choose the location of the file and/or
 * to view the access path to the file.</p>
 *
 * @version $Id: LySwingFile.java 35686 2012-06-01 14:27:50Z nhamon $
 * @author  Lyria
 * @see     LyFile
 */
public class LySwingFile extends LySwingFieldItem
	implements LySwingComponent,
		LyActionListener,
		LySetEditableInterface,
		LySetEnabledInterface,
		LySetLabelInterface,
		LyGraphicalFile,
		ActionListener
{
	public static final String __VERSION =
		"$Id: LySwingFile.java 35686 2012-06-01 14:27:50Z nhamon $";

	/** The swing component for this object. */
	protected Component		   _choice;

	/** The view independent file description. */
	protected LyFile		   _file;

	/** The file dialog to show to select a file. */
	protected LyJFileDialog    _dialog;

	/**
	 * Class constructor of a File object for the Swing target.
	 *
	 * @param file a FILE component defined for every type of view targets. This component is
	 *             initialized specificaly for the Swing view.
	 */
	public LySwingFile(LyFile file)
	{
		super(file);
		_file = file;
		_dialog = null;

		initialize();
		setContent();
	}

	/**
	 * Initialize this object. Builds the graphic component according to its type and registers its
	 * events.
	 */
	public void initialize()
	{
        LyFile.Type type = LyFile.Type.TEXT;

		if (_file.getType() == LyFile.Type.IMAGE)
			type = LyFile.Type.IMAGE;
		else if (_file.getType() == LyFile.Type.AUDIO)
			type = LyFile.Type.AUDIO;
		else if (_file.getType() == LyFile.Type.VIDEO)
			type = LyFile.Type.VIDEO;
		else if (_file.getType() == LyFile.Type.DIRECTORY)
			type = LyFile.Type.DIRECTORY;

		LyEnvironment environment = getEnvironment();
		String		  button, toolTip;

		if (type == LyFile.Type.DIRECTORY)
			button = translate("LY_CHOOSE_DIRECTORY");
		else
			button = translate("LY_CHOOSE_FILE");

		toolTip = button;

		String directory = _file.getDirectory();

		if (directory != null)
			directory = expand(directory);

		if (type == LyFile.Type.IMAGE)
		{
			String	label = _file.getLabel();
			String	path = (directory != null ? directory : "") + (label != null ? label : "");
			Image	image = null;
			
			if (label == null)
				path = "LY_DEFAULT_FILE_IMAGE";
			//  First try with the label only
			if (label != null)
			{
				// Check if the label is directly a valid file
				File test = new File(label);
				if (test.exists())
					path = label;
			}
			if (path.startsWith("file://"))
			{
				path = path.substring(7);
				image = (Image)getEnvironment().getImages().loadImage(null, path, true);
			}
			else
				image = getImage(path);
			
			int		nCols = _file.getNCols();
			int		nRows = _file.getNRows();
			
			if ((image != null)
					&& ((nCols > 0 && nRows <= 0) || (nRows > 0 && nCols <= 0)))
			{
				int		imageWidth = image.getWidth(null);
				int		imageHeight = image.getHeight(null);

				// Keep image ratio
				if ((nRows > 0) && (imageHeight != 0))
					nCols = (imageWidth * nRows) / imageHeight;
				else if ((nCols > 0) && (imageWidth != 0))
					nRows = (imageHeight * nCols) / imageWidth;
			}

			_choice = new LyJIconJLabel(image, _file.getLabel(), nCols, nRows);
			
			LyIcon.Border	border = LyIcon.Border.NONE;
			
			if (_file.getBorder() == LyFile.Border.SINGLE)
				border = LyIcon.Border.SINGLE;
			else if (_file.getBorder() == LyFile.Border.DOUBLE)
				border = LyIcon.Border.DOUBLE;
			else if (_file.getBorder() == LyFile.Border.LOWERED)
				border = LyIcon.Border.LOWERED;
			else if (_file.getBorder() == LyFile.Border.RAISED)
				border = LyIcon.Border.RAISED;

			((LyJIconJLabel)_choice).setBorderType(border);
			((LyJIconJLabel)_choice).setSensitive(true);
			
			LyActionTransmitter transmitter = ((LyJIconJLabel)_choice).getActionTransmitter();

			if (transmitter != null)
				transmitter.addActionListener(this);
		}
		else
		{
			Image  openImage = getImage("LY_OPEN_FILE");
			String filter = _file.getFilter();

			_choice = new LyJFileSelector(type, button, toolTip, openImage, directory, filter, _file.getNCols(), environment);

			Color enabledColor = getColor("LY_ENABLED_BACKGROUND");
			Color disabledColor = getColor("LY_DISABLED_BACKGROUND");
			Font  font = getFont("LY_DEFAULT_FONT");

			((LyJFileSelector)_choice).setEnabledColor(enabledColor);
			((LyJFileSelector)_choice).setDisabledColor(disabledColor);
			_choice.setFont(font);

			LyActionTransmitter transmitter = ((LyJFileSelector)_choice).getActionTransmitter();

			if (transmitter != null)
				transmitter.addActionListener(this);

			// This handler is used to validate the form using ENTER key or close using ESCAPE key
			LySwingFormKeyHandler keyHandler = new LySwingFormKeyHandler(getActionTransmitter());

			keyHandler.setNotifyValidate(false); // already default in swing
			_choice.addKeyListener(keyHandler);
		}
	} // end method initialize
	
	/**
	 * Sets the inital value of this object.
	 */
	public void setContent()
	{
		if (_choice instanceof LyJFileSelector)
			((LyJFileSelector)_choice).setValue(_file.getLabel());
	}

	/* (non-Javadoc)
	 * @see leon.view.setinterface.LySetEditableInterface#setEditable(boolean)
	 */
	public void setEditable(boolean editable)
	{
		if (_choice instanceof LyJFileSelector)
			((LyJFileSelector)_choice).setEditable(editable);
		else if (_choice instanceof LyJIconJLabel)
			((LyJIconJLabel)_choice).setSensitive(editable);
	}

	/**
	 * Indicates the parent enablen status.
	 *
	 * @parameter enabled new status of the component
	 * @see       LySetEnabledInterface#setEnabled(boolean)
	 */
	public void setEnabled(boolean enabled)
	{
		if (_choice instanceof LyJFileSelector)
			_choice.setEnabled(enabled);
		else if (_choice instanceof LyJIconJLabel)
			((LyJIconJLabel)_choice).setSensitive(enabled);
	}

	/* (non-Javadoc)
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		if (actionEvent.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
		{
			// Update Parent info
			_file.setLabel(getLabel());

			actionEvent.setValue(getLabel());

			// may notify any listeners
			if (_actionTransmitter != null)
				_actionTransmitter.processActionEvent(actionEvent);
		}
		else if (_choice instanceof LyJIconJLabel)
		{
			if (actionEvent.getType() == LyMenuDemand.TYPE_POPUP)
			{
				int[]			point = (int[])actionEvent.getParameter();
				LyJPopupMenu	menu = new LyJPopupMenu();
				JMenuItem		modifyItem = new JMenuItem(translate("LY_MODIFY_IMAGE"));
				JMenuItem		deleteItem = new JMenuItem(translate("LY_DELETE_IMAGE"));
				Font			font = getFont("LY_MENU_FONT");

				modifyItem.setFont(font);
				deleteItem.setFont(font);
				modifyItem.enableInputMethods(false);
				deleteItem.enableInputMethods(false);
				modifyItem.addActionListener(this);
				deleteItem.addActionListener(this);
				menu.add(modifyItem);
				
				if (_file.getLabel() != null)
					menu.add(deleteItem);
				
				menu.show(_choice, point[0], point[1]);
				
				return;
			}
			
			openFileDialog();
		}
	}
	
    /**
     * Invoked when an action occurs.
     */
    public void actionPerformed(ActionEvent e)
    {
    	if (e.getActionCommand().equals(translate("LY_MODIFY_IMAGE")))
    		openFileDialog();
    	else if (e.getActionCommand().equals(translate("LY_DELETE_IMAGE")))
    	{
			// Update Parent info
			_file.setLabel(null);

			LyActionEvent	actionEvent = new LyActionEvent(this);
			
			// may notify any listeners
			if (_actionTransmitter != null)
				_actionTransmitter.processActionEvent(actionEvent);
    	}
    }

    /**
     * Opens the file dialog used to select an image.
     */
    protected void openFileDialog()
    {
		String	directory = _file.getDirectory();

		if (directory != null)
			directory = expand(directory);
		
		if (_dialog == null)
			_dialog = new LyJFileDialog(LySwingViewManager.getRootFrame(_choice), translate("LY_CHOOSE_FILE"),
					LyFile.Type.IMAGE, directory, _file.getFilter());

		String value = _file.getLabel();

		if ((value != null) && (value.length() > 0))
		{
			if ((directory != null)
					&& !value.toUpperCase().startsWith(directory.toUpperCase()))
				value = directory + value;

			File existFile = new File(value);

			if (existFile.exists())
				_dialog.setFile(existFile);
			else if (existFile.isDirectory())
				_dialog.setDirectory(existFile);
		}

		if (_dialog.show() == JFileChooser.APPROVE_OPTION)
		{
			String	newValue = _dialog.getFile();
			
			// Update Parent info
			_file.setLabel(newValue);

			LyActionEvent	actionEvent = new LyActionEvent(this);
			
			actionEvent.setValue(newValue);

			// may notify any listeners
			if (_actionTransmitter != null)
				_actionTransmitter.processActionEvent(actionEvent);
		}
    }
    
	/**
	 * Gets the label of the file component.
	 *
	 * @return the title of the choice selector
	 * @see    #setLabel(String)
	 */
	public String getLabel()
	{
		if (_choice instanceof LyJFileSelector)
			return ((LyJFileSelector)_choice).getValue();
		else if (_choice instanceof LyJIconJLabel)
			return ((LyJIconJLabel)_choice).getText();
		
		return null;
	}

	/**
	 * Sets the label of the file component.
	 *
	 * @param label new label
	 * @see   #getLabel()
	 * @see   LySetLabelInterface#setLabel(String)
	 */
	public void setLabel(String label)
	{
		if (_choice instanceof LyJFileSelector)
			((LyJFileSelector)_choice).setValue(label);
		else if (_choice instanceof LyJIconJLabel)
		{
			Image	image = null;
			String	directory = _file.getDirectory();

			if (directory != null)
				directory = expand(directory);

			String	path = (directory != null ? directory : "") + (label != null ? label : "");
			
			if ((label != null) && (directory != null) && label.startsWith(directory))
				path = label;
			else if (label != null)
			{
				File test = new File(label);
				if (test.exists())
					path = label;
			}
			if (label == null)
				path = "LY_DEFAULT_FILE_IMAGE";
			
			if ((path != null) && path.startsWith("file://"))
			{
				path = path.substring(7);
				image = (Image)getEnvironment().getImages().loadImage(null, path, true);
			}
			else
				image = getImage(path);
			
			int		nCols = _file.getNCols();
			int		nRows = _file.getNRows();
			
			if ((image != null)
					&& ((nCols > 0 && nRows <= 0) || (nRows > 0 && nCols <= 0)))
			{
				int		imageWidth = image.getWidth(null);
				int		imageHeight = image.getHeight(null);

				// Keep image ratio
				if ((nRows > 0) && (imageHeight != 0))
					nCols = (imageWidth * nRows) / imageHeight;
				else if ((nCols > 0) && (imageWidth != 0))
					nRows = (imageHeight * nCols) / imageWidth;
			}
			
			((LyJIconJLabel)_choice).setDimension(nCols, nRows);
			((LyJIconJLabel)_choice).setImage(image);
			((LyJIconJLabel)_choice).setToolTipText(label);
		}
	}

	/* (non-Javadoc)
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _choice;
	}

} // end class LySwingFile
