/*
 * LySwingForm.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.border.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;


/**
 * This class implements the Form object (as described in the dtd) for swing viewer.
 *
 * <p>A form allows to show a set of fields. This type of component is used to allow a user of an
 * application to consult application data (non editable form) or else to input data (editable
 * form).</p>
 *
 * <p>The default arrangement of the fields inside a form is based on a grid of two columns
 * comprising as many lines as there are fields. The first column contains the labels of the fields
 * and the second one contains the input fields.</p>
 *
 * @version $Id: LySwingForm.java 34692 2012-03-15 08:25:29Z jlsaouli $
 * @author  Lyria
 * @see     LyForm
 */
public class LySwingForm extends LySwingPanel
	implements LySwingComponent,
		LyGraphicalForm,
		LyActionListener,
		LySetEnabledInterface,
		LyResetContentInterface
{
	public static final String __VERSION =
		"$Id: LySwingForm.java 34692 2012-03-15 08:25:29Z jlsaouli $";

	/** The view independent form description. */
	protected LyForm		   _form;

	/** The swing form panel. */
	protected Container		   _internalPanel;

	/** The swing form root panel. */
	protected Container		   _rootPanel;

	/** The layout of this component. */
	protected GridBagLayout    _layout;

	/** The background color of this form. */
	protected Color			   _color;

	/**
	 * Class constructor of a Footer object for the Swing target.
	 *
	 * @param form a FORM component defined for every type of view targets. This component is
	 *             initialized specificaly for the Swing view.
	 */
	public LySwingForm(LyForm form)
	{
		super(form);
		_form = form;

		initialize();
		setContent();
	}

	/**
	 * Creates the panel that contains the form fields.
	 *
	 * @param  layout the layoutManager of the returned panel
	 * @return the new panel with the given layout
	 */
	public Container createFormPanel(LayoutManager layout)
	{
		LayoutManager rootLayout = new GridLayout(1, 1, 0, 0);

		_rootPanel = new JPanel(rootLayout);
		_internalPanel = new JPanel(layout)
			{
				private static final long serialVersionUID = -3378138478744313522L;

				/*
				 * (non-Javadoc)
				 * @see javax.swing.JComponent#getMinimumSize()
				 */
				public Dimension getMinimumSize()
				{
					return new Dimension(10, getPreferredSize().height);
				}
			};

		((JPanel)_rootPanel).setOpaque(false);
		((JPanel)_internalPanel).setOpaque(false);

		if (_form.isBorder())
		{
			Border internalBorder = BorderFactory.createCompoundBorder(BorderFactory
						.createEtchedBorder(EtchedBorder.LOWERED),
					BorderFactory.createEmptyBorder(3, 3, 3, 3));

			((JPanel)_rootPanel).setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createEmptyBorder(1, 1, 1, 1), internalBorder));
		}

		_rootPanel.add(_internalPanel);

		return _rootPanel;
	} // end method createFormPanel

	/**
	 * Creates the panel for this form's toolbar.
	 *
	 * @return a new default toolbar panel
	 */
	public Container createToolBarPanel()
	{
		return new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));
	}

	/**
	 * Initializes the current component with the default values.
	 */
	public void initialize()
	{
		_layout = new GridBagLayout();
		_rootPanel = createFormPanel(_layout);

		_color = getColor("LY_FORM_BACKGROUND");

		if (_color != null)
		{
			_internalPanel.setBackground(_color);
			((JComponent)_internalPanel).setOpaque(true);
		}

		if (_form.getEnvironment().testResource("LY_CHECK_WINDOW_MINIMUM_SIZE"))
			_internalPanel.addComponentListener(new _LyJFormComponentAdapter());
	}

	/**
	 * Builds a contraint object for the layout of the specified label or component. The component
	 * object is not used, thus the constraints are generic from one component to another.
	 *
	 * @param  comp      a component, not used in this implementation
	 * @param  isCaption Applies constraints specific to the captions.
	 * @return the constraints of a generic component
	 */
	public GridBagConstraints buildDefaultConstraint(LyComponent comp, boolean isCaption)
	{
		GridBagConstraints c = new GridBagConstraints();

		c.gridx = GridBagConstraints.RELATIVE;
		c.gridy = GridBagConstraints.RELATIVE;
		// c.gridwidth defined as labeldependant
		c.gridheight = 1;

		c.fill = GridBagConstraints.NONE;
		c.weighty = 0.0;
		// NB : !! Seen endparse. If no item has a weighty, last item
		// will use all remaining space
		c.ipadx = 0;
		c.ipady = 0;

		if (isCaption)
		{
			c.insets = new Insets(0, 1, 1, 1);
			c.anchor = GridBagConstraints.NORTHEAST;
			c.gridwidth = 1; // same line
			c.weightx = 0.0; // space left for components
		}
		else
		{
			c.insets = new Insets(1, 1, 1, 1);
			c.anchor = GridBagConstraints.NORTHWEST;
			c.gridwidth = GridBagConstraints.REMAINDER;
			c.weightx = 1.0;
		}

		return c;
	} // end method buildDefaultConstraint

	/**
	 * Builds a contraint object for the layout for the specified label or component.
	 *
	 * @param  comp                component to build the constraints
	 * @param  requestedConstraint constraint object to apply on the component
	 * @param  isCaption           Applies constraints specific to the captions.
	 * @return the new constraints for the given component
	 */
	public GridBagConstraints buildConstraint(LyComponent comp, LyConstraint requestedConstraint,
		boolean isCaption)
	{
		GridBagConstraints c = LySwingConstraint.buildConstraint(comp, requestedConstraint,
				buildDefaultConstraint(comp, isCaption));

		return c;
	}

	/**
	 * Adds the label/item couple to the layout and sets its display contraints.
	 *
	 * @param field   label to add
	 * @param content component list
	 */
	public void addField(LyFormField field, Vector<_LyFormContent> content)
	{
		GridBagConstraints  c = null;

		LyCaption		    caption = field.getCaption();
		LyRelation		    relation = field.getRelation();
		LyComponent		    item = field.getFieldItem();
		LyActionTransmitter transmitter;
		LySwingFormField    internal = (LySwingFormField)field.getInternalView();

		// first add label into the layout of the container
		if (caption != null)
		{
			LyConstraint captionConstraint = caption.getConstraint();

			// set contraints on component
			c = buildConstraint(field, captionConstraint, true);

			content.addElement(new _LyFormContent(internal.getCaptionComponent(), c));
		}

		// add the type relation if necessary
		if (relation != null)
		{
			LySwingRelation winRelation = (LySwingRelation)relation.getInternalView();
			LyConstraint    relationConstraint = new LyConstraint();

			relationConstraint.setColSpan(-1);
			relationConstraint.setRemainder(true);
			relationConstraint.setHAlign(LyConstraintBase.HAlign.LEFT);
			c = buildConstraint(field, relationConstraint, false);
			content.addElement(new _LyFormContent(winRelation.getComponent(), c));
		}

		// first add main component item
		if (item != null)
		{
			LyConstraint     itemConstraint = field.getConstraint();
			LyFieldUnit		 unit = field.getFieldUnit();
			LyFieldItem		 unitItem;
			LyToolBar		 toolBar = field.getToolBar();
			LySwingComponent winRelation, winLabel;

			// set contraints on component
			c = buildConstraint(field, itemConstraint, false);

			if ((toolBar != null) || (unit != null))
			{
				// Flandres / FFT00040
				Container itemToolPanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));
				Container toolPanel = new JPanel(new GridBagLayout());

				itemToolPanel.add(internal.getItemComponent());
				itemToolPanel.add(toolPanel);

				if (unit != null)
				{
					unitItem = unit.getFieldItem();
					
					GridBagConstraints	gbc = new GridBagConstraints();
					
					gbc.insets = new Insets(0, 4, 0, 0);
					
					if (unitItem instanceof LyRelation)
					{
						winRelation = (LySwingComponent)unitItem.getInternalView();
						internal.resizeCombo(winRelation);
						toolPanel.add(winRelation.getComponent(), gbc);
					}
					else
					{
						winLabel = (LySwingComponent)unitItem.getInternalView();
						Component labelComponent = winLabel.getComponent();

						labelComponent.setBackground(_internalPanel.getBackground());
						toolPanel.add(labelComponent, gbc);
					}
				}
				if (field.getToolBar() != null)
				{
					LySwingComponent winToolBar = (LySwingComponent)field.getToolBar()
							.getInternalView();

					toolPanel.add(winToolBar.getComponent());
				}

				((JPanel)toolPanel).setOpaque(false);
				((JPanel)itemToolPanel).setOpaque(false);
				content.addElement(new _LyFormContent(itemToolPanel, c));
			}
			else
				content.addElement(new _LyFormContent(internal.getItemComponent(), c));
		}

		transmitter = ((LyEventTransmitter)internal).getActionTransmitter();

		if (transmitter != null)
		{
			transmitter.removeActionListener(this); // Avoid duplicate listener
			transmitter.addActionListener(this);
		}
	} // end method addField

	/**
	 * Sets the content of this component.
	 */
	public void setContent()
	{
		int					   i, n = _form.getFormFieldCount();
		Vector<_LyFormContent> content = new Vector<_LyFormContent>((2 * n) + 1);
		LyFormField			   field;
		_LyFormContent		   o;

		for (i = 0; i < n; i++)
		{
			field = _form.getFormField(i);
			if (field != null)
				addField(field, content);
		}

		// Let's make a first pass just to compute the grid size (width, height)
		int			 y = 1, x = 1, rowHeightRemaining = 0, max_x = 0, max_y = 0, k; // first line
		_LyFormTable layout = new _LyFormTable();

		for (i = 0, n = content.size(); i < n; i++)
		{
			o = content.elementAt(i);
			layout.add(y, ((o._c.gridwidth == GridBagConstraints.REMAINDER) ? 1 : o._c.gridwidth),
				o._c.gridheight);
			if (o._c.gridwidth == GridBagConstraints.REMAINDER)
				y++;
		}

		max_x = layout.getWidth();
		max_y = layout.getHeight();
		// now create an array of available cells
		// default : all cells available.
		boolean[][] cellAvailable = new boolean[max_x + 1][max_y + 1];

		for (y = 0; y < (max_y + 1); y++)
			for (x = 0; x < (max_x + 1); x++)
				cellAvailable[x][y] = true;

		// ****** now add content ******
		y = 1;
		x = 1;
		rowHeightRemaining = 0; // first line

		for (i = 0; i < n; i++)
		{
			o = content.elementAt(i);

			// is it posible to add a component at (x, y) ?
			while (!cellAvailable[x][y])
			{
				if (x > max_x)
				{
					y++;
					x = 1;
				}
				else
					x++; // skip already used cell
			}
			o._c.gridy = y; // do not use RELATIVE positioning
			o._c.gridx = x; // do not use RELATIVE positioning

			// Check if rowHeight exists : REMAINDER AND RELATIVE should be computed
			if (o._c.gridheight > (rowHeightRemaining + 1))
				rowHeightRemaining = o._c.gridheight;

			// check unavailable vertical cells
			for (k = y; k < (y + o._c.gridheight); k++)
				cellAvailable[x][k] = false;

			if (o._c.gridwidth == GridBagConstraints.REMAINDER)
			{
				y++; // end row

				x = 1;
				// When one cell is spread ver several rows, REMAINDER and RELATIVE do not
				// behave properly
				if (rowHeightRemaining > 0)
				{
					// o._c.gridwidth = 1;
					o._c.gridwidth = 0;
					rowHeightRemaining--;
				}
			}
			else
				x += o._c.gridwidth;
		}

		// add empty panel with weighty
		Component		   panel = createToolBarPanel();
		GridBagConstraints c = new GridBagConstraints();

		c.weighty = 0.01;
		content.addElement(new _LyFormContent(panel, c));

		// add everything (build view)
		for (i = 0, n = content.size(); i < n; i++)
		{
			o = content.elementAt(i);
			if (o._comp == null)
			    continue;
			
			_layout.setConstraints(o._comp, o._c);
			_internalPanel.add(o._comp);
		}

		setEditable(_form.isEditable(), true);
	} // end method setContent

	
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyResetContentInterface#resetContent()
	 */
	public void resetContent()
	{
		resetContent(false);
	}
	
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyResetContentInterface#resetContent(boolean)
	 */
	public void resetContent(boolean full)
	{
		_internalPanel.removeAll();
		_internalPanel.repaint();
		setContent();
		// The size of the form may have changed !
		// Check in the hierachy if another component may accept
		// the resetContent action.
		LyComponent parent = _form.getParent();
		Object	    internal;
		boolean     resetContent;

		while (parent != null)
		{
			internal = parent.getInternalView();

			if (internal instanceof LyResetContentInterface)
			{
				resetContent = true;

				// Reset content only if parent is not a resizable frame (otherwise the frame
				// may flash too much)
				if ((internal instanceof LySwingFrame) && (((LySwingFrame)internal).isResizable()))
					resetContent = false;

				// If parent is a non resizable dialog which has a default width or height,
				// do not change size
				if ((internal instanceof LySwingDialog)
						&& !(((LyDialog)parent).isResizable()))
				{
					LyDialog dialog = (LyDialog)parent;

					resetContent = ((dialog.getHeight() <= 0) || (dialog.getWidth() <= 0));
				}

				if (resetContent)
					((LyResetContentInterface)internal).resetContent();

				// break;
				return;
			}

			parent = parent.getParent();
		}

		// Try to do our best !
		_internalPanel.validate();
	} // end method resetContent

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		// Notify controller if any
		if (_form.getController() != null)
		{
			LyDemand demand = null;

			if (actionEvent.getType() == LyMenuDemand.TYPE_POPUP)
			{
				int[] position = (int[])actionEvent.getParameter();

				demand = new LyMenuDemand(actionEvent.getType(), actionEvent.getSourceId());
				((LyMenuDemand)demand).setItem(actionEvent.getValue());
				((LyMenuDemand)demand).setSubIndex(actionEvent.getSubIndex());
				((LyMenuDemand)demand).setPosition(position);
				demand.addIndex(actionEvent.getIndex(0));
			}
			else
			// Notify controller
				demand = LySwingArea.buildActionDemand(actionEvent);

			// Set info :
			demand.setParameterValues(new LySwingParameterValues(_form));

			processDemand(demand);
			actionEvent.consume();
		}

		if (_actionTransmitter != null)
			_actionTransmitter.processActionEvent(actionEvent);
	} // end method actionPerformed

	/**
	 * Sets the editable status of this form.
	 *
	 * @see #setEditable(boolean, boolean)
	 * @see leon.view.setinterface.LySetEditableInterface#setEditable(boolean)
	 */
	public void setEditable(boolean editable)
	{
		setEditable(editable, false);
	}

	/**
	 * Sets the editable status of this form.
	 *
	 * @param editable new status
	 * @param init     indicates if called by initialisation of the component
	 */
	public void setEditable(boolean editable, boolean init)
	{
		int				 i, n = _form.getFormFieldCount();
		LyFormField		 field;
		LySwingFormField internal;

		for (i = 0; i < n; i++)
		{
			field = _form.getFormField(i);
			if ((field != null)
				&& (field.getInternalView() instanceof LySwingFormField))
			{
				// we force a check on the editable status : reset the SAME status !
				internal = (LySwingFormField)field.getInternalView();
                if (editable && field.isEditable() && field.isEnabled())
                    internal.setEditable(true, init);
                else if (!editable)
                    internal.setEditable(false, init);
			}
		}
	}

	/**
	 * Sets the enabled status of this form.
	 *
	 * @see #setEnabled(boolean, boolean)
	 * @see leon.view.setinterface.LySetEnabledInterface#setEnabled(boolean)
	 */
	public void setEnabled(boolean enabled)
	{
		setEnabled(enabled, false);
	}

	/**
	 * Sets the enabled status of this form.
	 *
	 * @param enabled new status
	 * @param init    indicates if called by initialisation of the component
	 */
	public void setEnabled(boolean enabled, boolean init)
	{
		int				 i, n = _form.getFormFieldCount();
		LyFormField		 field;
		LySwingFormField internal;

		for (i = 0; i < n; i++)
		{
			field = _form.getFormField(i);
			if (field.getInternalView() instanceof LySwingFormField)
			{
				internal = (LySwingFormField)field.getInternalView();
				internal.setEnabled(enabled, init);
			}
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _rootPanel;
	}

	/**
	 * Returns the component used to display popup menus.
	 *
	 * @return the parent component of popup menus.
	 */
	protected JComponent getPopupMenuParent()
	{
		return (JComponent)_rootPanel;
	}

	/**
	 * Gets the background color of this component.
	 *
	 * @return the component's color
	 */
	public Color getColor()
	{
		return _color;
	}

	/* Internal classes */

	/**
	 * Couple of component and constraints elements.
	 */
	protected class _LyFormContent
	{

		/** Component element. */
		public Component		  _comp;

		/** Contraint element. */
		public GridBagConstraints _c;

		/**
		 * Creates a new pair of component/contraints.
		 *
		 * @param comp component element
		 * @param c    contraint element
		 */
		public _LyFormContent(Component comp, GridBagConstraints c)
		{
			_comp = comp;
			_c = c;
		}
	}


	/**
	 * Table representation of a form to allow easy computations of its grid size.
	 */
	protected class _LyFormTable
	{
		/** Lines of the form. */
		public Vector<Integer> _lines;

		/**
		 * Creates a new content form.
		 */
		public _LyFormTable()
		{
			_lines = new Vector<Integer>();
		}

		/**
		 * Adds lines to the form.
		 *
		 * @param currentLine number of new lines to add to the form
		 * @param width       width of the new lines
		 * @param height      height of the component
		 */
		public void add(int currentLine, int width, int height)
		{
			int     size = _lines.size();
			Integer line;

			// First create all needed lines
			if (currentLine > size)
			{
				for (int i = 0; i < (currentLine - size); i++)
				{
					line = new Integer(0);
					_lines.addElement(line);
				}
			}
			size = _lines.size();
			// If the height of the component implies more lines, add them;
			if (height > 1)
			{
				if ((currentLine + height - 1) > size)
				{
					for (int i = 0; i < (currentLine + height - size - 1); i++)
					{
						line = new Integer(0);
						_lines.addElement(line);
					}
				}
			}
			int value;
			// for each line, add the width
			for (int i = 0; i < height; i++)
			{
				line = _lines.elementAt(currentLine + i - 1);
				value = line.intValue() + width;

				line = new Integer(value);
				_lines.setElementAt(line, currentLine + i - 1);
			}
		} // end method add

		/**
		 * Returns the height of the form.
		 *
		 * @return the total number of lines
		 */
		public int getHeight()
		{
			return _lines.size();
		}

		/**
		 * Returns the width of the form.
		 *
		 * @return the maxmimum width of the form, based on its content' size
		 */
		public int getWidth()
		{
			Integer line;
			int     value;
			int     max = 0;

			for (int i = 0; i < _lines.size(); i++)
			{
				line = _lines.elementAt(i);
				value = line.intValue();
				if (value > max)
					max = value;
			}

			return max;
		}
	} // end class _LyFormTable


	/**
	 * Handles component resize and checks if minimum sizes are valid.
	 */
	public class _LyJFormComponentAdapter extends ComponentAdapter
	{
		/** Last dimensions of the form panel. */
		protected Dimension _last;

		/*
		 * (non-Javadoc)
		 *
		 * @see java.awt.event.ComponentListener#componentResized(java.awt.event.ComponentEvent)
		 */
		public void componentResized(ComponentEvent event)
		{
			// GO BACK TO THE TOP and specify a MINIMUM size
			Dimension d = _internalPanel.getPreferredSize();

			// Do we have anything to do ?
			if ((_last != null) && (d.equals(_last)))
				return;

			_last = d;

			// Notify hierachy
			if (_form.getParent().getInternalView() instanceof LySwingContainer)
				((LySwingContainer)_form.getParent().getInternalView()).updateMinimumSize(_form, d);
		}
	}

} // end class LySwingForm
