/*
 * LySwingFormField.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;

import javax.swing.*;

import leon.view.*;
import leon.view.event.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the internal description of a Field object as described in the dtd.
 *
 * <p>A field is contained in a form and allows to show or to input application data. It is
 * principally made up of a label (optional) and an input field depending of the type of the data
 * (text, number, date, color...). It is possible to associate a tool bar and a help entry to this
 * field.</p>
 *
 * @version $Id: LySwingFormField.java 34782 2012-03-16 15:32:05Z jlsaouli $
 * @author  Lyria
 * @see     LyFormField
 */
public class LySwingFormField extends LySwingObject
	implements LySwingComponent,
		LyGraphicalFormField,
		LyActionListener,
		LySelectionListener,
		LyEventTransmitter,
		LyResetContentInterface,
		LySetFocusInterface
{
	public static final String		 __VERSION =
		"$Id: LySwingFormField.java 34782 2012-03-16 15:32:05Z jlsaouli $";

	/** The view independent field description. */
	protected LyFormField			 _formField;

	/** The view independent field toolbar description. */
	protected LyToolBar				 _toolbar;

	/** Swing component of this field's caption, if any. */
	protected Component				 _captionComponent;

	/** Swing component of this field's caption label, if any. */
	protected Component				 _captionLabel;

	/** Swing component of the item associated to this field. */
	protected Component				 _itemComponent;

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter    _actionTransmitter;

	/** Provides notification events on component's selections. */
	protected LySelectionTransmitter _selectionTransmitter;

	/** Last field unit processed. */
	protected String				 _lastUnit;

	/**
	 * Class constructor of a Field object for the Swing target.
	 *
	 * @param formField a FIELD component defined for every type of view targets. This component is
	 *                  filled with the default values.
	 */
	public LySwingFormField(LyFormField formField)
	{
		super(formField);
		_formField = formField;

		setContent();
	}

	/**
	 * Sets the default content of the current field component.
	 */
	public void setContent()
	{
		_toolbar = _formField.getToolBar();

		LyActionTransmitter    actionTransmitter;
		LySelectionTransmitter selectionTransmitter;

		// ** set up caption **
		LyCaption			   caption = _formField.getCaption();

		if (caption != null)
		{
			LySwingCaption internal = (LySwingCaption)caption.getInternalView();

			_captionComponent = internal.getComponent();
			_captionLabel = internal.getCaptionLabel();

			/* Set the same background color that their parent */
			_captionComponent.setBackground(null);
			_captionLabel.setBackground(null);
		}

		// ** set up type relation **
		LyRelation relation = _formField.getRelation();

		if (relation != null)
		{
			LySwingRelation internal = (LySwingRelation)relation.getInternalView();

			internal.getComponent().setBackground(null);

			resizeCombo(internal);

			// attempt to register for notifications
			actionTransmitter = ((LyEventTransmitter)internal).getActionTransmitter();

			if (actionTransmitter != null)
				actionTransmitter.addActionListener(this);

			selectionTransmitter = ((LyEventTransmitter)internal).getSelectionTransmitter();

			if (selectionTransmitter != null)
				selectionTransmitter.addSelectionListener(this);
		}

		// ** Set up component **
		LyComponent item = _formField.getFieldItem();

		if (item != null)
		{
			LySwingComponent internal = (LySwingComponent)item.getInternalView();
			LyFieldUnit		 unit = _formField.getFieldUnit();
			LyRelation		 unitRelation;
			LySwingComponent winRelation;

			if ((unit != null) && (unit.getFieldItem() instanceof LyRelation))
			{
				unitRelation = (LyRelation)unit.getFieldItem();
				winRelation = (LySwingComponent)unitRelation.getInternalView();

				// attempt to register for notifications
				if (winRelation instanceof LyEventTransmitter)
				{
					actionTransmitter = ((LyEventTransmitter)winRelation).getActionTransmitter();

					if (actionTransmitter != null)
						actionTransmitter.addActionListener(this);

					selectionTransmitter = ((LyEventTransmitter)winRelation)
							.getSelectionTransmitter();

					if (selectionTransmitter != null)
						selectionTransmitter.addSelectionListener(this);
				}
				_lastUnit = unitRelation.getValue();
			}

			if (_itemComponent != null)
			{
				Container parent = _itemComponent.getParent();

				/* Set the same background color as its parent */
				_itemComponent.setBackground(null);
				
				if (parent != null)
					parent.remove(_itemComponent);
				
				_itemComponent = internal.getComponent();

				if (parent != null)
				{
					// If parent is a form, try to rebuild constraint with the method used during init process.
					if (_formField.getParent() instanceof LyForm)
					{
						LySwingForm			swingForm = (LySwingForm)_formField.getParent().getInternalView();
						GridBagConstraints	gbc = LySwingConstraint.buildConstraint(_formField, _formField.getConstraint(),
								swingForm.buildDefaultConstraint(_formField, false));
		
						parent.add(_itemComponent, gbc);
					}
					else
						parent.add(_itemComponent);
				}
			}
			else if (internal != null)
				_itemComponent = internal.getComponent();

			// attempt to register for notifications
			if (internal instanceof LyEventTransmitter)
			{
				actionTransmitter = ((LyEventTransmitter)internal).getActionTransmitter();

				if (actionTransmitter != null)
					actionTransmitter.addActionListener(this);

				selectionTransmitter = ((LyEventTransmitter)internal).getSelectionTransmitter();

				if (selectionTransmitter != null)
					selectionTransmitter.addSelectionListener(this);
			}
		}

		setOptional(_formField.isOptional());
		setEditable(_formField.isEditable(), true);
		setEnabled(_formField.isEnabled(), true);
		setHidden(_formField.isHidden());
		setState(_formField.getState());
	} // end method setContent

	
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyResetContentInterface#resetContent()
	 */
	public void resetContent()
	{
		resetContent(false);
	}
	
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyResetContentInterface#resetContent(boolean)
	 */
	public void resetContent(boolean full)
	{
		setContent();
	}

	/**
	 * Set the width of the specified combo to the minimum.
	 *
	 * @param winCombo Combo to resize.
	 */
	protected void resizeCombo(LySwingComponent winCombo)
	{
		JComboBox comboBox = (JComboBox)winCombo.getComponent();
		int		  preferredHeight = comboBox.getPreferredSize().height;

		comboBox.setPreferredSize(new Dimension(comboBox.getUI().getPreferredSize(comboBox).width,
				preferredHeight));
	}

	/**
	 * Sets the editable status of this field item.
	 *
	 * @param editable new status
	 * @param init     indicates if called by initialisation of the component. If true, the field
	 *                 toolbar is uneditable, if false, the toolbar follows the field status.
	 */
	public void setEditable(boolean editable, boolean init)
	{
		LyComponent item = _formField.getFieldItem();
		LyFieldUnit unit = _formField.getFieldUnit();

		if (item != null)
		{
			LySwingComponent internal = (LySwingComponent)item.getInternalView();

			// If item is a form, transit the init state
			if (internal instanceof LySwingForm)
				((LySwingForm)internal).setEditable(editable, init);
			else if (internal instanceof LySetEditableInterface)
				// the field_item is editable if the form and
				// the field are simultaneously editable.
				((LySetEditableInterface)internal).setEditable(editable);
		}

		if (unit != null)
		{
			LySwingComponent internal = (LySwingComponent)unit.getFieldItem().getInternalView();

			if (internal instanceof LySetEditableInterface)
				// the field_item is editable if the form and
				// the field are simultaneously editable.
				((LySetEditableInterface)internal).setEditable(editable);
		}

		LyRelation relation = _formField.getRelation();

		if (relation != null)
		{
			LySwingComponent internal = (LySwingComponent)relation.getInternalView();

			if (internal instanceof LySetEnabledInterface)
				((LySetEditableInterface)internal).setEditable(editable);
		}

		updateBackgroundColor();
		updateParentRow();
	}
	
	/**
	 * Updates the row containing this form field if needed.<br />
	 * If field is not displayed in a table, this method does nothing.
	 */
	protected void updateParentRow()
	{
		// If field is in a table, update the row containing the table
		LyComponent parent = _formField;

		while (parent != null)
		{
			parent = parent.getParent();

			if (parent instanceof LyRow)
				break;
		}

		if (parent instanceof LyRow)
		{
			LyRow	   row = (LyRow)parent;
			LySwingRow swingRow = (LySwingRow)row.getInternalView();

			swingRow.updateRow();
		}
	} // end method setEditable

	/**
	 * Creates a panel.
	 *
	 * @param  layout the LayoutManager to use
	 * @return a new panel with the specified layout manager
	 */
	public Container createPanel(LayoutManager layout)
	{
		JPanel result = new JPanel(layout);

		result.setOpaque(false);

		return result;
	}

	/**
	 * Returns the current caption component.
	 *
	 * @return the caption component
	 */
	public Component getCaptionComponent()
	{
		return _captionComponent;
	}

	/**
	 * Returns the current item component.
	 *
	 * @return the item component
	 */
	public Component getItemComponent()
	{
		return _itemComponent;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return getItemComponent();
	}

	/**
	 * Returns all the components of the current form. If this form field has a toolbar, the
	 * components are added to a new panel.
	 *
	 * @return the full form component
	 */
	public Component getFullComponent()
	{
	    if (_formField.isFreed())
	        return null;
	    
		LyFieldUnit unit = _formField.getFieldUnit();
		LyFieldItem unitItem = null;
		LyToolBar   toolBar = _formField.getToolBar();

		if ((toolBar != null) || (unit != null))
		{
			// Flandres / FFT00040
			GridBagLayout		layout = new GridBagLayout();
			GridBagConstraints	gbc = new GridBagConstraints();
			Container			itemPanel = createPanel(layout);
			Container			itemToolPanel = createPanel(new BorderLayout());
			Container			toolPanel = createPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));
			Component			item = getItemComponent();
			
			gbc.fill = GridBagConstraints.HORIZONTAL;
			gbc.weightx = 1;
			layout.setConstraints(item, gbc);
			itemPanel.add(item);
			itemToolPanel.add(itemPanel, BorderLayout.CENTER);
			itemToolPanel.add(toolPanel, BorderLayout.EAST);
			
			if (unit != null)
			{
				// Label used to add an extra space between item and unit
				JLabel label = new JLabel(" ");

				label.setBackground(null);
				toolPanel.add(label);

				unitItem = unit.getFieldItem();

				if (unitItem instanceof LyRelation)
				{
					LySwingComponent winRelation = (LySwingComponent)unitItem.getInternalView();

					resizeCombo(winRelation);
					toolPanel.add(winRelation.getComponent());
				}
				else
				{
					LySwingComponent winLabel = (LySwingComponent)unitItem.getInternalView();
					Component		 labelComponent = winLabel.getComponent();

					toolPanel.add(labelComponent);
				}
			}

			if (_formField.getToolBar() != null)
			{
				LySwingToolBar winToolBar = (LySwingToolBar)_formField.getToolBar()
						.getInternalView();

				toolPanel.add(winToolBar.getComponent());
			}

			return itemToolPanel;
		}

		return getComponent();
	} // end method getFullComponent

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{

		if (actionEvent.getSource() instanceof LySwingRelation)
		{
			LySwingRelation relation = (LySwingRelation)actionEvent.getSource();
			LyFieldUnit     unit = _formField.getFieldUnit();

			if ((unit != null) && (relation == unit.getFieldItem().getInternalView()))
			{
				String temp = actionEvent.getValue();

				actionEvent = new LyActionEvent(relation, null, temp,
						LyActionDemand.TYPE_UNIT_CHANGED);
				actionEvent.setAdditionalValue(_lastUnit);
				_lastUnit = temp;
			}

			LyRelation typeRelation = _formField.getRelation();

			if ((typeRelation != null) && (relation == typeRelation.getInternalView()))
				actionEvent = new LyActionEvent(relation, null, actionEvent.getValue(),
						LyActionDemand.TYPE_TYPE_CHANGED);
		}

		// Notify controller if any
		if (_formField.getController() != null)
		{
			LyDemand demand;

			if (actionEvent.getType() == LyMenuDemand.TYPE_POPUP)
			{
				Object	source = actionEvent.getSource();
				int[]	position = (int[])actionEvent.getParameter();
				
				if (source instanceof Component)
				{
					Component	component = (Component)source;
					Point		positionInForm = getPositionInForm(component);
					
					if (positionInForm != null)
					{
						position[0] += positionInForm.x;
						position[1] += positionInForm.y;
					}
				}
				
				demand = new LyMenuDemand(LyMenuDemand.TYPE_POPUP, _formField.getId());
				((LyMenuDemand)demand).setPosition(position);
			}
			else
			{
				// Notify controller
				// Use field id as event identifier
				// should be actionEvent.getSourceId() == null but be sure to keep compatibility
				if ((actionEvent.getSourceId() != LySwingFormKeyHandler.DEMAND_VALIDATE)
						&& (actionEvent.getSourceId() != LySwingFormKeyHandler.DEMAND_CANCEL)
						&& (actionEvent.getSourceId() != LySwingFormKeyHandler.DEMAND_APPLY))
					actionEvent.setSourceId(_formField.getId());

				demand = LySwingArea.buildActionDemand(actionEvent);

				// Set info :
				demand.setParameterValues(new LySwingParameterValues(_formField));

				processDemand(demand);
				actionEvent.consume();
			}
		}

		// may notify any listeners
		if ((actionEvent != null) && (_actionTransmitter != null))
		{
			// Use field id as event identifier
			// should be actionEvent.getSourceId() == null but be sure to keep compatibility
			if ((actionEvent.getSourceId() != LySwingFormKeyHandler.DEMAND_VALIDATE)
					&& (actionEvent.getSourceId() != LySwingFormKeyHandler.DEMAND_CANCEL)
					&& (actionEvent.getSourceId() != LySwingFormKeyHandler.DEMAND_APPLY))
				actionEvent.setSourceId(_formField.getId());

			_actionTransmitter.processActionEvent(actionEvent);
		}
	} // end method actionPerformed

	/**
	 * Return the position of current form field in its parent form.
	 * 
	 * @param subComponent the component whose relative position must be computed.
	 * @return the position of current form field in its parent form.
	 */
	protected Point getPositionInForm(Component subComponent)
	{
		LyForm		parentForm = _formField.getParentForm();
		
		if (parentForm == null)
			return new Point(0, 0);

		int[]	position = parentForm.getBounds(null);
		Point	subPosition = subComponent.getLocationOnScreen();
		
		if ((position == null) || (subPosition == null))
			return new Point(0, 0);
		
		return new Point(subPosition.x - position[0], subPosition.y - position[1]);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LySelectionListener#selectionPerformed(leon.view.event.LySelectionEvent)
	 */
	public void selectionPerformed(LySelectionEvent selectionEvent)
	{
		// may notify any listeners
		if ((selectionEvent != null) && (_selectionTransmitter != null))
		{
			// Use field id as event identifier
			if (!(selectionEvent.getItem() instanceof String))
				selectionEvent.setItem(_formField.getId());

			_selectionTransmitter.processSelectionEvent(selectionEvent);
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetEditableInterface#setEditable(boolean)
	 */
	public void setEditable(boolean editable)
	{
		setEditable(editable, false);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetOptionalInterface#setOptional(boolean)
	 */
	public void setOptional(boolean optional)
	{
		Font f;

		if (_formField.getParent() instanceof LySelectBar)
			f = getFont("LY_SELECT_BAR_TITLE");
		else if (optional)
			f = getFont("LY_DEFAULT_FONT");
		else
			f = getFont("LY_DEFAULT_BOLD_FONT");

		if ((f != null) && (_captionLabel != null))
			_captionLabel.setFont(f);
		
		String	asterisk = getString("LY_CAPTION_ASTERISK", "");
		
		if ((asterisk != null) && (asterisk.trim().length() > 0) && (_captionLabel != null))
		{
			LyCaption	caption = _formField.getCaption();
			String      label = caption.getLabel();

			if ((label != null) && (label.length() > 0))
			{
				label = translate(label);
				if (label != null)
				{
					label = label.trim();
					
					if (label.length() > 0)
					{			
					    if (!_formField.isOptional())
					        label += getString("LY_CAPTION_ASTERISK", "");
	                    if (caption.isColon())
	                        label += getString("LY_CAPTION_COLON", "");
					}
				}
				else
					label = "";
			}
			else
				label = "";
			
			((LyJMultiLineLabel)_captionLabel).setLabel(label);
		}

		LyComponent item = _formField.getFieldItem();

		if (item != null)
		{
			LySwingComponent internal = (LySwingComponent)item.getInternalView();

			if (internal instanceof LySetOptionalInterface)
				((LySetOptionalInterface)internal).setOptional(optional);
		}
		
		updateBackgroundColor();
	}

	/**
	 * Sets the state of this field.
	 *
	 * @param state the state value that defines this field's foreground color. This is one of the
	 *              following value:
	 *
	 *              <ul>
	 *                <li>{@link LyFormFieldBase#STATE_NORMAL}</li>
	 *                <li>{@link LyFormFieldBase#STATE_ERROR}</li>
	 *                <li>{@link LyFormFieldBase#STATE_WARNING}</li>
	 *              </ul>
	 * @see   leon.view.setinterface.LySetStateInterface#setState(short)
	 */
	public void setState(short state)
	{
		// ** state **
		if (_captionLabel != null)
		{
			Color c;

			if (state == LyFormFieldBase.STATE_NORMAL)
				c = getColor("LY_LABEL_NORMAL_STATE");
			else if (state == LyFormFieldBase.STATE_ERROR)
				c = getColor("LY_LABEL_ERROR_STATE");
			else if (state == LyFormFieldBase.STATE_WARNING)
				c = getColor("LY_LABEL_WARNING_STATE");
			else
				c = getColor("LY_LABEL_STATE_" + state);

			_captionLabel.setForeground(c);
		}

		updateBackgroundColor();
	}

	/**
     * Updates attribute message to the given value.
     * @param message New value of message.
     */
    public void setMessage(String message)
    {
    }
    
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetEnabledInterface#setEnabled(boolean)
	 */
	public void setEnabled(boolean enabled)
	{
		setEnabled(enabled, false);
	}

	/**
	 * Sets the enabled status of this form field.
	 *
	 * @param enabled new status
	 * @param init    indicates if called by initialisation of the component
	 */
	public void setEnabled(boolean enabled, boolean init)
	{
		LyComponent item = _formField.getFieldItem();
		LyFieldUnit unit = _formField.getFieldUnit();

		if (item != null)
		{
			LySwingComponent internal = (LySwingComponent)item.getInternalView();

			// If item is a form, transit the init state
			if (internal instanceof LySwingForm)
				((LySwingForm)internal).setEnabled(enabled, init);
			else if (internal instanceof LySetEnabledInterface)
				((LySetEnabledInterface)internal).setEnabled(enabled);
		}

		if (unit != null)
		{
			LySwingComponent internal = (LySwingComponent)unit.getFieldItem().getInternalView();

			if (internal instanceof LySetEnabledInterface)
				((LySetEnabledInterface)internal).setEnabled(enabled);
		}

		LyRelation relation = _formField.getRelation();

		if (relation != null)
		{
			LySwingComponent internal = (LySwingComponent)relation.getInternalView();

			if (internal instanceof LySetEnabledInterface)
				((LySetEnabledInterface)internal).setEnabled(enabled);
		}

		updateBackgroundColor();
		updateParentRow();
	} // end method setEnabled

	/**
	 * (non-Javadoc)
	 * 
	 * @see leon.view.graphic.LyGraphicalFormField#setChecked(boolean)
	 */
	public void setChecked(boolean checked)
	{
		if ((_captionComponent instanceof JCheckBox)
				&& (((JCheckBox)_captionComponent).isSelected() != checked))
			((JCheckBox)_captionComponent).setSelected(checked);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalFormField#setHidden(boolean)
	 */
	public void setHidden(boolean hidden)
	{
		// set components not visible
		if (_captionComponent != null)
			_captionComponent.setVisible(!hidden);

		if (_captionLabel != null)
			_captionLabel.setVisible(!hidden);

		if (_itemComponent != null)
			_itemComponent.setVisible(!hidden);

		// if there is a toolbar : set not visible
		LyToolBar toolBar = _formField.getToolBar();

		if (toolBar != null)
		{
			LySwingComponent internal = (LySwingComponent)toolBar.getInternalView();

			if (internal != null)
			{
				Component component = internal.getComponent();

				component.setVisible(!hidden);
			}
		}

		LyFieldUnit unit = _formField.getFieldUnit();

		if (unit != null)
		{
			LyFieldItem unitItem = unit.getFieldItem();

			if (unitItem != null)
			{
				LySwingFieldItem internal = (LySwingFieldItem)unitItem.getInternalView();

				internal.getComponent().setVisible(!hidden);
			}
		}
	}

	/**
	 * Set the toolbar for the form field.
	 *
	 * @parem toolBar the new toolbar
	 * @see   LyGraphicalFormField#setToolBar(LyToolBar)
	 */
	public void setToolBar(LyToolBar toolBar)
	{
		Container parent = null;

		// Remove old toolbar
		if (_toolbar != null)
		{
			LySwingComponent internal = (LySwingComponent)_toolbar.getInternalView();

			if (internal != null)
			{
				Component component = internal.getComponent();

				parent = component.getParent();
				parent.remove(component);
			}
		}

		_toolbar = toolBar;

		// Create new toolbar component and add it to parent panel if existing
		if ((parent != null) && (toolBar != null))
		{
			LySwingComponent internal = (LySwingComponent)toolBar.getInternalView();

			if (internal != null)
			{
				Component component = internal.getComponent();

				parent.add(component);
				parent.getParent().validate();
			}
		}
		else if ((parent == null) && (toolBar != null))
		{
			LyForm form = ((_formField.getParent() instanceof LyForm)
					? (LyForm)_formField.getParent() : null);

			if (form != null)
			{
				LySwingForm winForm = (LySwingForm)form.getInternalView();

				winForm.resetContent();
			}
		}
	} // end method setToolBar

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return (_actionTransmitter);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		if (_selectionTransmitter == null)
			_selectionTransmitter = new LySimpleSelectionTransmitter();

		return (_selectionTransmitter);
	}

	/**
	 * Updates the background color depending on attributes of the field. The priority of the
	 * searching properties is the following:
	 *
	 * <ol>
	 *   <li>the state of the field is LyFormFieldBase.STATE_ERROR</li>
	 *   <li>the state of the field is LyFormFieldBase.STATE_WARNING</li>
	 *   <li>the field is disabled</li>
	 *   <li>the field is not editable</li>
	 *   <li>the field is mandatory</li>
	 *   <li>the field is not editable</li>
	 * </ol>
	 *
	 * In any of these cases (in the given order), the background color of this field is changed
	 * accordingly.
	 */
	protected void updateBackgroundColor()
	{
		if (_itemComponent == null)
			return;

		Color	    color = null;
		String	    colorKey = null;
		LyFieldItem fieldItem = _formField.getFieldItem();

		if ((fieldItem instanceof LyChoice) || (fieldItem instanceof LyLabel)
				|| (fieldItem instanceof LyHRef) || (fieldItem instanceof LyHRefs))
			colorKey = null;
		else if (_formField.getState() == LyFormFieldBase.STATE_ERROR)
			colorKey = "LY_INVALID_FIELD_BACKGROUND";
		else if (_formField.getState() == LyFormFieldBase.STATE_WARNING)
			colorKey = "LY_WARNING_FIELD_BACKGROUND";
		else if (!_formField.isEnabled())
			colorKey = "LY_DISABLED_FIELD_BACKGROUND";
		else if (!_formField.isEditable())
			colorKey = "LY_NOT_EDITABLE_FIELD_BACKGROUND";
		else if (!_formField.isOptional())
			colorKey = "LY_MANDATORY_FIELD_BACKGROUND";
		else if (_formField.isOptional())
			colorKey = "LY_OPTIONAL_FIELD_BACKGROUND";

		if (colorKey != null)
		{
			color = getColor(colorKey);
			if (color != null)
			{
				_itemComponent.setBackground(color);

				LyFieldUnit unit = _formField.getFieldUnit();

				if (unit != null)
				{
					LyFieldItem unitItem = unit.getFieldItem();

					if (unitItem != null)
					{
						LySwingFieldItem internal = (LySwingFieldItem)unitItem.getInternalView();

						internal.getComponent().setBackground(color);
					}
				}
			}
		}
	} // end method updateBackgroundColor

	/**
	 * Gives the focus to this formField.
	 *
	 * @see LySetFocusInterface#setFocus()
	 */
	public void setFocus()
	{
		LyFieldItem		 item = _formField.getFieldItem();
		LySwingComponent view = (LySwingComponent)item.getInternalView();

		if (view != null)
		{
			if (view instanceof LySetFocusInterface)
				((LySetFocusInterface)view).setFocus();
			else
			{
				(view.getComponent()).requestFocus();
				ensureVisible();
			}
		}
	}

	/**
	 * Force the form to scroll to show this component...
	 */
	private void ensureVisible()
	{
       Component component         = getComponent();
        
        if (!component.isShowing())
            return;
	        
		// get the parent form;
		LyComponent parent = _formField;
		
		while ((parent != null) && ! (parent instanceof LyForm))
			parent = parent.getParent();
		if (parent == null)
			return;
		
		LyForm form					= (LyForm)parent;
		LySwingForm sForm			= (LySwingForm)form.getInternalView();
		Component parentComponent 	= sForm.getComponent();
		
		if (!parentComponent.isShowing())
		    return;
		
		Rectangle rect 				= component.getBounds();
		Point compOnScreen 			= component.getLocationOnScreen();
		Point formOnScreen			= parentComponent.getLocationOnScreen();
		Rectangle view				= new Rectangle(compOnScreen.x - formOnScreen.x , compOnScreen.y
				- formOnScreen.y, rect.width, rect.height);
		
		((JComponent)parentComponent).scrollRectToVisible(view);
	}
} // end class LySwingFormField
