/*
 * LySwingFormKeyHandler.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.event.*;

import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.win.swing.component.*;


/**
 * This class implments a key listener for swing form components. When enter is pressed, the form is
 * validated. When escape is pressed, the form is canceled.
 *
 * @version $Id: LySwingFormKeyHandler.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class LySwingFormKeyHandler extends KeyAdapter
{
	public static final String    __VERSION =
		"$Id: LySwingFormKeyHandler.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Demand identifiers for validation. Should be the same as {@link
	 * leon.control.LyFormController#DEMAND_VALIDATE}.
	 */
	public static String		  DEMAND_VALIDATE = "_validate".intern();

	/**
	 * Demand identifiers for application. Should be the same as {@link
	 * leon.control.LyFormController#DEMAND_APPLY}.
	 */
	public static String		  DEMAND_APPLY = "_apply".intern();

	/**
	 * Demand identifiers for cancelling. Should be the same as {@link
	 * leon.control.LyFormController#DEMAND_CANCEL}.
	 */
	public static String		  DEMAND_CANCEL = "_close".intern();

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter _actionTransmitter;

	/** Indicates if the VALIDATE event should be notified (default is true). */
	protected boolean			  _notifyValidate;

	/** Indicates if the CANCEL event should be notified (default is true). */
	protected boolean			  _notifyCancel;

	/** Indicates if the TAB event should be notified (default is true). */
	protected boolean			  _notifyTab;
	
	/** Instance of LyGetValueInterface used to send tab events with value. */
	protected LyGetValueInterface _valueInterface;
	
	protected boolean			  _ignoreTabWithCtrl;
	
	/**
	 * Class constructor of a handler object for the keys in a Swing form.
	 *
	 * @param actionTransmitter an action transmitter that processes the key events of the form
	 */
	public LySwingFormKeyHandler(LyActionTransmitter actionTransmitter)
	{
		this(actionTransmitter, null);
	}

	/**
	 * Class constructor of a handler object for the keys in a Swing form.
	 *
	 * @param actionTransmitter an action transmitter that processes the key events of the form
	 */
	public LySwingFormKeyHandler(LyActionTransmitter actionTransmitter, LyGetValueInterface valueInterface)
	{
		super();
		_actionTransmitter = actionTransmitter;
		_notifyValidate = true;
		_notifyCancel = true;
		_notifyTab = true;
		_valueInterface = valueInterface;
		_ignoreTabWithCtrl = false;
	}

	/**
	 * Apply the correct action when a key event is generated. The validation action is executed
	 * when the released key is Enter. The cancellation action is executed when the released key is
	 * Escape. The tab next action is executed when the released key is Tab.
	 *
	 * @param e the key event with the event informations
	 * @see   java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
	 */
	public void keyReleased(KeyEvent e)
	{
		if ((_notifyValidate) && (e.getKeyCode() == KeyEvent.VK_ENTER))
			actionPerformed(DEMAND_VALIDATE);
		else if ((_notifyCancel) && (e.getKeyCode() == KeyEvent.VK_ESCAPE))
			actionPerformed(DEMAND_CANCEL);
		else if ((_notifyTab) && (e.getKeyCode() == KeyEvent.VK_TAB))
		{
			if (_actionTransmitter != null)
			{
				if (!e.isControlDown() ||  !_ignoreTabWithCtrl)
				{
				short	type = (e.isShiftDown() ? LyActionDemand.TYPE_TAB_PREVIOUS : LyActionDemand.TYPE_TAB_NEXT);
				
				LyActionEvent	evt = new LyActionEvent(this, null, (_valueInterface != null ? _valueInterface.getValue() : null),
						type);
				_actionTransmitter.processActionEvent(evt);
			}
		}
		}
		else if (!_notifyTab
					&& (e.getKeyCode() == KeyEvent.VK_TAB)
					&& (e.isControlDown() || e.isShiftDown()))
		{
			if (_actionTransmitter != null)
			{
				short	type = (e.isShiftDown() ? LyActionDemand.TYPE_TAB_PREVIOUS : LyActionDemand.TYPE_TAB_NEXT);
				
				LyActionEvent	evt = new LyActionEvent(this, null, (_valueInterface != null ? _valueInterface.getValue() : null),
						type);
				_actionTransmitter.processActionEvent(evt);
			}
		}
	}

	/**
	 * Apply the correct action when a key event is generated. The tab previous action is executed
	 * when the pressed key is Tab and the shift key is down.
	 *
	 * @param e the key event with the event informations
	 * @see   java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
	 */
	public void keyPressed(KeyEvent e)
	{
    	if (e.getKeyCode() == KeyEvent.VK_TAB)
		{
			if (_actionTransmitter != null)
			{
				boolean		generateAction = true;
				Component	comp = e.getComponent();
				Container	parent = comp.getParent();
				
				while ((parent != null) && !(parent instanceof LyJTableComponent))
					parent = parent.getParent();
				
				if (parent instanceof LyJTableComponent)
				{
					LyJTableComponent	table = (LyJTableComponent)parent;
					
					if (!table.hasEditableCells())
						generateAction = false;
				}

				if (generateAction && e.isShiftDown())
				{
					// A SHIFT+TAB action is managed by keyReleased method.
					e.consume();
				}
				else if (_notifyTab && !generateAction)
				{
					((LyJTableComponent)parent).editNextCell(!e.isShiftDown());
				}
				else if (!_notifyTab && !generateAction && e.isControlDown())
				{
					((LyJTableComponent)parent).editNextCell(!e.isShiftDown());
				}
				
				if (!_notifyTab && e.isControlDown())
					e.consume();
			}
		}
	}

	/**
	 * Invoked when an the button is pressed.
	 *
	 * @param id source event id ({@link #DEMAND_VALIDATE} or {@link #DEMAND_CANCEL})
	 * @see   LyActionListener#actionPerformed(LyActionEvent)
	 */
	public void actionPerformed(String id)
	{
		if (_actionTransmitter != null)
		{
			LyActionEvent evt = new LyActionEvent(this);

			evt.setSourceId(id);

			_actionTransmitter.processActionEvent(evt);
		}
	}

	/**
	 * Indicates if the VALIDATE event should be notified.
	 *
	 * @param status new validation status
	 */
	public void setNotifyValidate(boolean status)
	{
		_notifyValidate = status;
	}

	/**
	 * Indicates if the CANCEL event should be notified.
	 *
	 * @param status new validation status
	 */
	public void setNotifyCancel(boolean status)
	{
		_notifyCancel = status;
	}


	/**
	 * Indicates if the TAB event should be notified.
	 *
	 * @param status new validation status
	 */
	public void setNotifyTab(boolean status)
	{
		_notifyTab = status;
	}
	
	/**
	 * Indicates if the CTRL-TAB event should be notified.
	 *
	 * @param status new validation status
	 */
	public void setIgnoreCtrlTab(boolean status)
	{
		_ignoreTabWithCtrl = status;
	}
} // end class LySwingFormKeyHandler
