/*
 * LySwingFrame.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;


/**
 * This class implements the Frame object (as described in the dtd) for swing viewer.
 *
 * <p>The root of an <code>XML</code> file describing a complete Leonardi view is an element of main
 * window or dialog window type.</p>
 *
 * @version $Id: LySwingFrame.java 34781 2012-03-16 15:30:36Z jlsaouli $
 * @author  Lyria
 * @see     LyFrame
 */
public class LySwingFrame extends LySwingView
	implements LySwingComponent,
		LyGraphicalFrame
{
	public static final String  __VERSION =
		"$Id: LySwingFrame.java 34781 2012-03-16 15:30:36Z jlsaouli $";

	private static final String CENTER = "Center";
	private static final String SOUTH = "South";

	/**
	 * Maximum ratio in percent of the window between its size (width/height) and the screen.
	 * Default is 100%.
	 */
	public static int		    MAX_SIZE_RATIO = 100;

	/** The view independent frame description. */
	protected LyFrame		    _frame;

	/** The swing frame component. */
	protected Frame			    _frameComp;

	/** The minimum size of the frame if specified. */
	protected Dimension		    _minimumSize;

	/** Indicates that frame component must be maximized when showing. */
	protected boolean		    _maximizeFrame;

	/**
	 * Class constructor of a frame object for the Swing target.
	 *
	 * @param frame a FRAME component defined for every type of view targets. The frame component is
	 *              initialized and configured.
	 */
	public LySwingFrame(LyFrame frame)
	{
		super(frame);
		_frame = frame;
		_maximizeFrame = false;

		initialize();
		setContent();

		configureDefaultButton();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingView#initialize()
	 */
	public void initialize()
	{
		super.initialize();

		_frameComp = createFrame();

		if (!_frame.isDecorated())
			_frameComp.setUndecorated(true);

		setLabel(_frame.getLabel());

		if ((_frame.getWidth() >= 0) && (_frame.getHeight() >= 0))
			setSize(_frame.getWidth(), _frame.getHeight());

		registerWindowEvents();
		_compAdapter = new _LyComponentAdapter();
		_frameComp.addComponentListener(_compAdapter);

		if ((_frame.getX() > 0) && (_frame.getY() >= 0))
			setLocation(_frame.getX(), _frame.getY());
	}

	/**
	 * Sets the content of this frame. This comprises the frame's menubar, a window icon, an area
	 * content and messages components.
	 *
	 * @see leon.view.win.swing.LySwingView#setContent()
	 */
	public void setContent()
	{
		// ** Add menubar **
		LyMenuBar menubar = _frame.getMenuBar();

		if (menubar != null)
		{
			LySwingMenuBar internal = (LySwingMenuBar)menubar.getInternalView();

			setMenuBar(internal.getMenuBar());

			// * register events *
			LyActionTransmitter transmitter = ((LyEventTransmitter)internal).getActionTransmitter();

			transmitter.addActionListener(this);
		}

		if (getEnvironment().isDebugLevel(LyEnvironment.DEBUG_WARNING))
			checkShortcuts();

		// ** Add window icon **
		LyIcon icon = _frame.getIcon();

		if (icon != null)
		{
			LySwingIcon wicon = (LySwingIcon)icon.getInternalView();

			setIcon(wicon.getImage());
		}

		// ** add content **
		LyArea area = _frame.getArea();

		if (area != null)
		{
			LySwingComponent internal = (LySwingComponent)area.getInternalView();

			setArea(internal.getComponent());

			// register events
			if (internal instanceof LyEventTransmitter)
			{
				LyActionTransmitter transmitter = ((LyEventTransmitter)internal)
						.getActionTransmitter();

				transmitter.addActionListener(this);
			}
		}

		// ** add messages **
		setMessages(_frame.getMessages());
	} // end method setContent

	/**
	 * Sets sub-component messages.
	 *
	 * @param messages New value of messages
	 */
	public void setMessages(LyMessages messages)
	{
		if (messages != null)
		{
			LySwingComponent internal = (LySwingComponent)messages.getInternalView();

			setMessages(internal.getComponent());

			/* register events */
			if (internal instanceof LyEventTransmitter)
			{
				LyActionTransmitter transmitter = ((LyEventTransmitter)internal)
						.getActionTransmitter();

				transmitter.addActionListener(this);
			}
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _frameComp;
	}

	/**
	 * Sets the default button. At this time no default button is defined in Leonardi, so the policy
	 * is to set the first command button as the default button.
	 */
	public void configureDefaultButton()
	{
		LyArea area = _frame.getArea();

		if (area != null)
		{
			LyCommands commands = area.getCommands();

			if (commands != null)
			{
				LyCommand command = commands.getCommand(0);

				if (command != null)
				{
					JButton button = ((JButton)((LySwingComponent)command.getInternalView())
								.getComponent());

					((JFrame)_frameComp).getRootPane().setDefaultButton(button);
				}
			}
		}
	}

	/**
	 * Registers the opening and closing window events.
	 */
	public void registerWindowEvents()
	{
		_winAdapter = new _LyWindowAdapter2();
		_frameComp.addWindowListener(_winAdapter);
	}

	/**
	 * Creates the frame.
	 *
	 * @return a new frame
	 */
	public Frame createFrame()
	{
		JFrame frameComp = new JFrame();

		frameComp.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
		
		return frameComp;
	}

	/**
	 * Checks that menus' shortcuts are unique. If they are already used in the environment, a
	 * warning is printed for the user.
	 */
	protected void checkShortcuts()
	{
		Hashtable<Integer, String> dico = new Hashtable<Integer, String>();
		JMenuBar				   menuBar = ((JFrame)_frameComp).getJMenuBar();

		if (menuBar == null)
			return;

		int menuCount = menuBar.getMenuCount();

		for (int i = 0; i < menuCount; i++)
		{
			JMenu menu = menuBar.getMenu(i);
			int   itemCount = menu.getItemCount();

			for (int j = 0; j < itemCount; j++)
			{
				JMenuItem item = menu.getItem(j);
				KeyStroke shortcut;

				if (item != null)
					shortcut = item.getAccelerator();
				else
					// item is a separator
					shortcut = null;

				if (null != shortcut)
				{
					Integer key = new Integer(shortcut.getKeyCode());

					if (null == dico.get(key))
						dico.put(key, item.getText());
					else
						getEnvironment().logWarning("The shortcut for menu item " + item.getText()
							+ " is already used by menu item " + dico.get(key));
				}
			}
		}
	} // end method checkShortcuts

	/**
	 * Sets the menu bar in the frame.
	 *
	 * @param menuBar menu bar of the frame
	 */
	public void setMenuBar(Object menuBar)
	{
		((JFrame)_frameComp).setJMenuBar((JMenuBar)menuBar);
	}

	/**
	 * Sets the message bar in the frame.
	 *
	 * @param messages message bar of the frame
	 */
	public void setMessages(Component messages)
	{
		((JFrame)_frameComp).getContentPane().add(messages, SOUTH);
	}

	/**
	 * Sets the client area in the frame.
	 *
	 * @see LySwingView#setArea(Component)
	 */
	public void setArea(Component area)
	{
		((JFrame)_frameComp).getContentPane().add(area, CENTER);
	}

	/**
	 * Sets the waiting status of this frame. This method changes the cursor image and enables or
	 * disables the frame's menubar.
	 *
	 * @see LySwingView#setWaiting(boolean)
	 */
	public void setWaiting(boolean status)
	{
		if (_frameComp != null)
		{
			if (status)
				_frameComp.setCursor(LySwingViewManager.WAIT_CURSOR);
			else
				_frameComp.setCursor(LySwingViewManager.DEFAULT_CURSOR);

			// XXX Windows Bug on dialog/Frame comp
			// if not enabled, cursor is unchanged or returns
			// to default
			// _frameComp.setEnabled(!status);
			// Workaround
			LyArea area = _frame.getArea();
			Object internal = area.getInternalView();

			if (internal instanceof LySetWaitingInterface)
				((LySetWaitingInterface)internal).setWaiting(status);

			// XX Workaround for same bug as above : we shall
			// also ebale/disable the menu items since frame is not
			// disabled yet
			JMenuBar menuBar = ((JFrame)_frameComp).getJMenuBar();

			if (menuBar != null)
			{
				synchronized (menuBar.getTreeLock())
				{
					// (de)activate main menus
					int   i, n = menuBar.getMenuCount();
					JMenu menu;

					for (i = 0; i < n; i++)
					{
						menu = menuBar.getMenu(i);
						menu.setEnabled(!status);
					}
					menuBar.setEnabled(!status);
				}
			}
		}
	} // end method setWaiting

	/**
	 * Sets all the components of this frame to their preferred size and readjust the size of the
	 * frame.
	 */
	public void repack()
	{
		if (_frame == null)
			return;

		int w, h;

		if ((_frame.getWidth() >= 0) && (_frame.getHeight() >= 0))
		{
			w = _frame.getWidth();
			h = _frame.getHeight();
		}
		else
		{
			pack();

			// Check screen size and make it consistent
			Dimension d = _frameComp.getPreferredSize();
			int[]     dscreen = getEnvironment().getScreenSize(_frameComp);

			w = Math.min(d.width, (dscreen[0] * MAX_SIZE_RATIO) / 100);
			h = Math.min(d.height, (dscreen[1] * MAX_SIZE_RATIO) / 100);
		}

		setSize(w, h);
	}

	/**
	 * Sets the title bar of the frame.
	 *
	 * @see LySetLabelInterface#setLabel(String)
	 */
	public void setLabel(String label)
	{
		if ((_frameComp != null) && (label != null))
        {
            label = translate(label);
            label = label.replace('\n', ' ');
			_frameComp.setTitle(translate(label));
        }
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingView#isResizable()
	 */
	public boolean isResizable()
	{
		return _frameComp.isResizable();
	}

	/**
	 * Allows the frame to be resizable or not.
	 *
	 * @param resizable set the resizable property of the frame
	 */
	public void setResizable(boolean resizable)
	{
		_frameComp.setResizable(resizable);
	}

	/**
	 * Sets the icon of the frame.
	 *
	 * @param icon the icon to be displayed in minimized state
	 */
	public void setIcon(Image icon)
	{
		if (icon != null)
			_frameComp.setIconImage(icon);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetVisibleInterface#setVisible(boolean)
	 */
	public void setVisible(boolean status)
	{
		if (_frameComp == null)
			return;

		if (_frameComp.isShowing() == status)
		{
			if (status)
			{
				if (_frameComp.getState() == Frame.ICONIFIED)
					_frameComp.setState(Frame.NORMAL);
				_frameComp.toFront();
			}
			return;
		}

		if (!status)
		{
			_frameComp.setVisible(false);
			return;
		}

		if (!_frameComp.isVisible())
			doLayout();

		if (!_resized)
		{
			pack();
			checkSize();
			// XXX To compare to minimumSize,
			// See remark in LySwingDialog
		}

		setResizable(_frame.isResizable());

		if (_maximizeFrame)
		{
			_frameComp.setExtendedState(Frame.MAXIMIZED_BOTH);
			_maximizeFrame = false;
		}
		else
			_frameComp.setState(Frame.NORMAL);
		
		_frameComp.setVisible(true);

		_frameComp.toFront();

		_requiredSizeDate = new Date();

		if (_frame.isResizable())
		{
			int minWidth = getIntResource("LY_WINDOW_MIN_WIDTH");
			int minHeight = getIntResource("LY_WINDOW_MIN_HEIGHT");

			if ((minWidth > 0) || (minHeight > 0))
				updateMinimumSize(_frame, new Dimension(minWidth, minHeight));
		}

	} // end method setVisible
	
	/**
	 * Sets the width and height of this frame.
	 *
	 * @see leon.view.setinterface.LySetBoundsInterface#setSize(int, int)
	 */
	public void setSize(int width, int height)
	{
		Window window = getWindow();

		if (window == null)
			return;

		// FFT00557: Pb de taille des fentres en bi-ecran
		int[] screenSize = getViewManager().getMaxScreenSize();

		if ((width <= 0) || (width > screenSize[0]))
			width = screenSize[0];
		if ((height <= 0) || (height > screenSize[1]))
			height = screenSize[1];

		if ((width == screenSize[0]) && (height == screenSize[1]))
			_maximizeFrame = true;
		else
			super.setSize(width, height);
	} // end method setSize

	/**
	 * Free this frame and all its sub-components.
	 *
	 * @see LySwingView#free()
	 */
	public void free()
	{
		super.free();

		if (_frameComp != null)
			_frameComp.dispose();

		_frameComp = null;
		_frame = null;
	}

	/**
	 * Updates the minimum size occupied by this component.
	 *
	 * @param child child component in this container. This parameter is not used in this
	 *              implementation.
	 * @param d     minimum size to use
	 */
	public void updateMinimumSize(LyComponent child, Dimension d)
	{
		if (!getEnvironment().testResource("LY_CHECK_WINDOW_MINIMUM_SIZE"))
			return;

		if ((_minimumSize == null) && (d != null) && (_frameComp.isResizable()))
			_frameComp.addComponentListener(new _LyWindowSizeListener());

		_minimumSize = d;

		if (_minimumSize != null)
		{
			int minWidth = getIntResource("LY_WINDOW_MIN_WIDTH");
			int minHeight = getIntResource("LY_WINDOW_MIN_HEIGHT");

			if (minWidth > _minimumSize.width)
				_minimumSize.width = minWidth;
			if (minHeight > _minimumSize.height)
				_minimumSize.height = minHeight;
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingView#checkSize()
	 */
	protected void checkSize()
	{
		// Check screen size and make it consistent
		Dimension d = _frameComp.getPreferredSize();
		int[]     dscreen = getEnvironment().getScreenSize(_frameComp);

		int		  w, h;

		w = Math.min(d.width, (dscreen[0] * MAX_SIZE_RATIO) / 100);
		h = Math.min(d.height, (dscreen[1] * MAX_SIZE_RATIO) / 100);

		setSize(w, h);
	}

	/**
	 * Maximum size ratio of this frame on the screen (in %).
	 *
	 * @return MAX_SIZE_RATIO (100% by default)
	 * @see    LySwingView#getMaxSizeRatio()
	 */
	public int getMaxSizeRatio()
	{
		return MAX_SIZE_RATIO;
	}

	/**
	 * This class ensures that the minimum size of the dialog is correct.
	 */
	public class _LyWindowSizeListener extends ComponentAdapter
	{

		/** Minimum size border of the dialog. */
		public int OFFSET_BORDER = 10;

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.ComponentListener#componentResized(java.awt.event.ComponentEvent)
		 */
		public void componentResized(ComponentEvent e)
		{
			if (_minimumSize == null)
				return;

			Dimension d = e.getComponent().getSize();
			boolean   resize = false;

			if (d.height < (_minimumSize.height + OFFSET_BORDER))
			{
				resize = true;
				d.height = _minimumSize.height + OFFSET_BORDER;
			}
			if (d.width < (_minimumSize.width + OFFSET_BORDER))
			{
				resize = true;
				d.width = _minimumSize.width + OFFSET_BORDER;
			}
			if (resize)
				e.getComponent().setSize(d);
		}
	} // end class _LyWindowSizeListener

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingView#snapshot(java.lang.String)
	 */
	public void snapshot(String file)
	{
		LySwingUtils.snapshot(_frameComp, file);
	}

	/**
	 * Handles the windows events of this frame (opening and closing events).
	 */
	protected class _LyWindowAdapter2 extends _LyWindowAdapter
	{

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.WindowListener#windowOpened(java.awt.event.WindowEvent)
		 */
		public void windowOpened(WindowEvent e)
		{
			// FFT Studio 000185
			if (LyEnvironment.isOperatingSystem(LyEnvironment.OS_SOLARIS))
				SwingUtilities.invokeLater(new Runnable()
					{

						/*
						 * (non-Javadoc)
						 * @see java.lang.Runnable#run()
						 */
						public void run()
						{
							repack();
						}
					});
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.WindowListener#windowClosing(java.awt.event.WindowEvent)
		 */
		public void windowClosing(WindowEvent e)
		{
			// ** WORKAROUND for hide showing menu when closing and reopening frame **
			MenuSelectionManager.defaultManager().clearSelectedPath();
			super.windowClosing(e);
		}
	} // end class _LyWindowAdapter2

} // end class LySwingFrame
