/*
 * LySwingFreeChart.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import org.jfree.chart.*;
import java.awt.*;
import java.awt.PageAttributes.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.text.*;
import java.util.*;

import javax.swing.*;

import leon.misc.*;
import leon.view.*;
import leon.view.common.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.print.*;
import leon.view.setinterface.*;


/**
 * This class implements the Chart object (as described in the dtd) for swing viewer using
 * JFreeChart components.
 *
 * @version $Id: LySwingFreeChart.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     LyChartConverter
 * @see     JFreeChart
 */
public class LySwingFreeChart extends LySwingPanel
	implements LySwingComponent,
		LyGraphicalChart,
		LyPrintInterface,
		LyGetChartConverterInterface,
		ImageObserver
{
	public static final String			 __VERSION =
		"$Id: LySwingFreeChart.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The datas which are used to build the chart. */
	protected LyChart					 _chart;

	/** The swing component of the chart. */
	protected JComponent				 _component;

	/** The chart converter that format the chart datas for {@link JFreeChart}. */
	protected LyCommonFreeChartConverter _chartConv;

	/**
	 * Class constructor of a chart object for the Swing target.
	 *
	 * @param chart a CHART component defined for every type of view targets created by {@link
	 *              leon.control.LyChartController}. The chart component is initialized and created
	 *              from a chart converter for JFreeChart.
	 * @see   leon.control.LyChartController
	 */
	public LySwingFreeChart(LyChart chart)
	{
		super(chart);
        
        LyEnvironment   environment = chart.getEnvironment();
        int             width = environment.getIntResource("LY_CHART_WIDTH");
        int             height = environment.getIntResource("LY_CHART_HEIGHT");
        
        if (width < 0)
            width = LyCommonFreeChartConverter.DEFAULT_CHART_WIDTH;

        if (height < 0)
            height = LyCommonFreeChartConverter.DEFAULT_CHART_HEIGHT;
        
		_chart = chart;
		_component = null;
		_chartConv = new LyCommonFreeChartConverter(_chart);
		_component = new ChartPanel(_chartConv.getFreeChart());
		_component.setPreferredSize(new Dimension(width, height));
	}

	/**
	 * Prints the content of the chart on a printer output.
	 *
	 * @param listener the listener of this chart
	 * @see   LyPrintInterface#print(LyPrintListenerInterface)
	 */
	public void print(LyPrintListenerInterface listener)
	{
		PrintJob		   job = null;
		LyEnvironment	   environment = _chart.getEnvironment();
		LySwingViewManager viewManager = (LySwingViewManager)environment.getViewManager();
		PageAttributes	   pageAttributes = new PageAttributes();
		
		pageAttributes.setOrientationRequested(OrientationRequestedType.LANDSCAPE);
		job = viewManager.getPrintJob(null, _chart.getId(), pageAttributes);

		if (job == null)
			return;

		printEnd(job, _chart, listener);

		job.end();
	}

	/**
	 * Prints the content of the given chart on the specified print job. The resulting image is
	 * printed on one page if a printer listener is given, then the JFreeChart is modified with the
	 * corresponding image.
	 *
	 * @param job      printer handler
	 * @param canvas   the component to print
	 * @param listener the listener of this chart
	 */
	public void printEnd(PrintJob job, LyChart canvas, LyPrintListenerInterface listener)
	{
		LyEnvironment environment = canvas.getEnvironment();
		Graphics	  g = job.getGraphics();
		Dimension     jd = job.getPageDimension();
		String		  title = null, logo = null;
		String		  client_logo = "LY_COMPANY_LOGO";
		String		  date = DateFormat.getDateTimeInstance(DateFormat.FULL, DateFormat.MEDIUM,
				environment.getLocale()).format(new Date());
		Dimension	  dim = _component.getSize();
		int			  w = jd.width - LyPrintManager.LEFT_MARGIN - LyPrintManager.RIGHT_MARGIN;
		int			  hh, fh;

		LyComponent   parent = canvas.getParent();

		if (parent instanceof LyArea)
		{
			LyTitleBar titleBar = ((LyArea)parent).getTitleBar();

			if (titleBar != null)
			{
				title = titleBar.getLeftLabel();

				if ((title == null) || (title.length() <= 0))
					title = titleBar.getCenterLabel();

				if ((title == null) || (title.length() <= 0))
					title = titleBar.getRightLabel();

				title = environment.translate(title);
				logo = titleBar.getImage();
			}
		}

		if (listener != null)
			listener.print(1);

		// Draw header
		hh = LySwingPrintManager.printHeader(_component, g, jd, logo, title, client_logo, environment);

		// Draw footer
		fh = LySwingPrintManager.printFooter(_component, g, jd, date, null, "1", environment);

		// Set the background to white
		Image image = _component.createImage(dim.width, dim.height);

		// Calculate the dimension of the printed graphic : dilate with the same rate for height and
		// width
		float height = new Integer(image.getHeight(this)).floatValue();
		float width = new Integer(image.getWidth(this)).floatValue();
		float heightMax = new Integer(jd.height - hh - fh - 30).floatValue();
		float widthMax = new Integer(w - 15).floatValue();
		float rate = 1;

		if ((height != 0) && (width != 0))
		{
			if (((heightMax / height) > (widthMax / width)) && ((widthMax / width) > 0))
				rate = widthMax / width;
			else if (((widthMax / width) > (heightMax / height)) && ((heightMax / height) > 0))
				rate = heightMax / height;
		}
		int   newHeight = new Float(height * rate).intValue();
		int   newWidth = new Float(width * rate).intValue();

		// calculate the position of the chart, at the center of the page
		float topMin = new Integer(hh).floatValue();
		float leftMin = new Integer(LyPrintManager.LEFT_MARGIN).floatValue();
		int   topNew = new Float(topMin + ((heightMax - newHeight) / 2)).intValue();
		int   leftNew = new Float(leftMin + ((widthMax - newWidth) / 2)).intValue();

		Paint oldColor = _chartConv.getFreeChart().getBackgroundPaint();

		_chartConv.getFreeChart().setBackgroundPaint(Color.white);
		_chartConv.getFreeChart().draw((Graphics2D)image.getGraphics(),
			new Rectangle2D.Double(0, 0, dim.width, dim.height), null);
		_chartConv.getFreeChart().setBackgroundPaint(oldColor);

		g.drawImage(image, leftNew, topNew, newWidth, newHeight, this);
		g.dispose();
	} // end method printEnd

	/**
	 * Always updates the current chart image if requested. None of the given parameters are used.
	 *
	 * @param  img       the image being observed
	 * @param  infoflags the bitwise inclusive OR of the following flags:
	 *
	 *                   <ul>
	 *                     <li>{@link ImageObserver#WIDTH}</li>
	 *                     <li>{@link ImageObserver#HEIGHT}</li>
	 *                     <li>{@link ImageObserver#PROPERTIES}</li>
	 *                     <li>{@link ImageObserver#SOMEBITS}</li>
	 *                     <li>{@link ImageObserver#FRAMEBITS}</li>
	 *                     <li>{@link ImageObserver#ALLBITS}</li>
	 *                     <li>{@link ImageObserver#ERROR}</li>
	 *                     <li>{@link ImageObserver#ABORT}</li>
	 *                   </ul>
	 * @param  x         the x coordinate
	 * @param  y         the y coordinate
	 * @param  width     the width
	 * @param  height    the height
	 * @return true
	 * @see    ImageObserver#imageUpdate(java.awt.Image, int, int, int, int, int)
	 */
	public boolean imageUpdate(Image img, int infoflags, int x, int y, int width, int height)
	{
		return true;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyFreeInterface#free()
	 */
	public void free()
	{
		_chart = null;

		if (_component != null)
		{
			if (_component instanceof LyFreeInterface)
				((LyFreeInterface)_component).free();
		}

		_chartConv = null;
	}

	/**
	 * Inserts a sub-component column at given position.
	 *
	 * <p>This method is not implemented and don't do anything.</p>
	 *
	 * @param column the column to add to columns list
	 * @param pos    the position where the sub-component is inserted
	 */
	public void addColumn(LyColumn column, int pos)
	{
		// XXX Not implemented
	}

	/**
	 * Removes a sub-component column.
	 *
	 * <p>This method is not implemented and don't do anything.</p>
	 *
	 * @param column the column to remove to columns list
	 */
	public void removeColumn(LyColumn column)
	{
		// XXX Not implemented
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
        LyEnvironment   environment = getEnvironment();
        int             width = environment.getIntResource("LY_CHART_WIDTH");
        int             height = environment.getIntResource("LY_CHART_HEIGHT");
        
        if (width < 0)
            width = LyCommonFreeChartConverter.DEFAULT_CHART_WIDTH;

        if (height < 0)
            height = LyCommonFreeChartConverter.DEFAULT_CHART_HEIGHT;
        
		_component = new ChartPanel(_chartConv.getFreeChart());
		_component.setPreferredSize(new Dimension(width, height));
		
		return _component;
	}

	/**
	 * Retreives the chart Converter.
	 *
	 * @return the chart Converter
	 * @see    LyGetChartConverterInterface#getChartConverter()
	 */
	public LyChartConverter getChartConverter()
	{
		return _chartConv;
	}

	/**
	 * Get the dimension of the component used to represent the chart.
	 * 
	 * @return an array of two integers (width and height of the component).
	 */
	public int[] getComponentDimension()
	{
		Dimension	dim = _component.getSize();
		
		return new int[] {dim.width, dim.height};
	}
} // end class LySwingFreeChart
