/*
 * LySwingHRef.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.graphic.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the HRef object (as described in the dtd) for swing viewer.
 *
 * <p>An input field of Hyperlink type allows to display a graphical component containing an
 * hyperlink. This link allows to launch an action by a simple click. It does not contain any
 * element.</p>
 *
 * @version $Id: LySwingHRef.java 34691 2012-03-15 08:25:10Z jlsaouli $
 * @author  Lyria
 * @see     LyLabel
 */
public class LySwingHRef extends LySwingFieldItem
	implements LySwingComponent,
		LyGraphicalHRef,
		LyActionListener
{
	public static final String    __VERSION =
		"$Id: LySwingHRef.java 34691 2012-03-15 08:25:10Z jlsaouli $";

	/** The view independent HRef description. */
	protected LyHRef			  _hRef;

	/** The swing component used to display this HRef. */
	protected Component			  _comp;


	/**
	 * Class constructor of a HRef object for the Swing target.
	 *
	 * @param hRef a HREF component defined for every type of view targets. This component is
	 *             initialized with the href requesites.
	 */
	public LySwingHRef(LyHRef hRef)
	{
		super(hRef);
		_hRef = hRef;

		setContent();
	}

	/**
	 * Creates and configures a Hyper Text Reference component with the specified parameters.
	 *
	 * @param  label       the text displayed by this reference
	 * @param  image       the image displayed by this reference
	 * @param  font        the font used to display the label
	 * @param  isVisited   true if the link has already been visited
	 * @param  isSensitive the sensitive property of the link
	 * @param  hRef        the underlying data used to properly configure this reference
	 * @param  disable     the disable color
	 * @param  color       the normal color
	 * @param  highlight   the color used to display text when cursor is over the reference
	 * @param  visited     the visited color
	 * @return the new hypertext component
	 */
	public Component createHRefComponent(String label, Image image, Font font, boolean isVisited,
		boolean isSensitive, LyHRef hRef, Color disable, Color color, Color highlight,
		Color visited)
	{
		LyJHRefComponent comp = new LyJHRefComponent(label, image, font, isVisited, isSensitive,
				hRef);

		comp.setDisableColor(disable);
		comp.setDefaultColor(color);
		comp.setHighlightColor(highlight);
		comp.setVisitedColor(visited);
		comp.setOpaque(false);

		return comp;
	}

	/**
	 * Builds the component and registers its events.
	 */
	public void setContent()
	{
		Image img = null;

		if (_hRef.getImage() != null)
			img = getImage(_hRef.getImage());

		String label = translate(_hRef.getLabel());
		Font   font = getFont("LY_HREF_FONT");

		_comp = createHRefComponent(label, img, font, _hRef.isVisited(), _hRef.isSensitive(), _hRef,
				getColor("LY_HREF_DISABLE"), getColor("LY_HREF_COLOR"),
				getColor("LY_HREF_HIGHLIGHT"), getColor("LY_HREF_VISITED"));

		LyActionTransmitter transmitter = ((LyEventTransmitter)_comp).getActionTransmitter();

		if (transmitter != null)
			transmitter.addActionListener(this);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		if (_actionTransmitter != null)
		{
			// LyActionEvent event = new LyActionEvent(this, null);
			// _actionTransmitter.processActionEvent(event);

			actionEvent.setValue(_hRef.getId());

			_actionTransmitter.processActionEvent(actionEvent);
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return (_actionTransmitter);
	}

	/**
	 * This HRef link doesn't handle selection events, so it has no selection transmitter.
	 *
	 * @return <code>null</code>
	 * @see    leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		return (null);
	}

	/**
	 * Sets the label of the HRef link.
	 *
	 * @param label new label
	 * @see   leon.view.setinterface.LySetLabelInterface#setLabel(String)
	 */
	public void setLabel(String label)
	{
		if (_comp != null)
		{
			label = translate(label);

			// Keep the old width to avoid layout issues when label is displayed in a border panel.
			int	width = ((LyJHRefComponent)_comp).getWidth();
			
			((LyJHRefComponent)_comp).setMinimumWidth(width);
			((LyJHRefComponent)_comp).setLabel(label);
		}
	}

	/**
	 * Sets the new image.
	 *
	 * @param image the new image
	 * @see   leon.view.setinterface.LySetImageInterface#setImage(String)
	 */
	public void setImage(String image)
	{
		if (_comp != null)
		{
			Image img = null;

			if (image != null)
				img = getImage(image);

			((LyJHRefComponent)_comp).setImage(img);
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetSensitiveInterface#setSensitive(boolean)
	 */
	public void setSensitive(boolean sensitive)
	{
		((LyJHRefComponent)_comp).setSensitive(sensitive);
	}

	/**
	 * (Modification) Accessor (LySetVisitedInterface interface).
	 */
	public void setVisited(boolean visited)
	{
		((LyJHRefComponent)_comp).setVisited(visited);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _comp;
	}

} // end class LySwingHRef
