/*
 * LySwingHRefs.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.graphic.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the HRefs object (as described in the dtd) for swing viewer.
 *
 * <p>An input field of Hyperlinks type allows to display a graphical component containing a set of
 * hyperlinks. Such a component is a simple list of hyperlinks. It does not contain any attribute.
 * </p>
 *
 * @version $Id: LySwingHRefs.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     LyLabel
 */
public class LySwingHRefs extends LySwingFieldItem
	implements LySwingComponent,
		LyGraphicalHRefs,
		LyActionListener
{
	public static final String __VERSION =
		"$Id: LySwingHRefs.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The view independent header list description. */
	protected LyHRefs		   _hRefs;

	/** The swing component used to display the header list. */
	protected LyJHRefContainer _hRefCont;

	/**
	 * Class constructor of a HRefs object for the Swing target.
	 *
	 * @param hRefs a HREFS component defined for every type of view targets. This hyperlink
	 *              component list is filled with the hRefs it contains.
	 */
	public LySwingHRefs(LyHRefs hRefs)
	{
		super(hRefs);
		_hRefs = hRefs;

		setContent();
	}

	/**
	 * Creates the container used to place all the HRefComponents.
	 *
	 * @param  background the background color of the container
	 * @return a new component for the HRefComponents
	 */
	public LyJHRefContainer createHRefContainer(Color background)
	{
		LyEnvironment    environment = getEnvironment();
		LyJHRefContainer hRefCont = new LyJHRefContainer(environment);
		int				 height = _hRefs.getNRows();

		if (height > 0)
			hRefCont.setContentHeight(height);

		if (!_hRefs.isBorder())
			hRefCont.setBorder(null);
		
		hRefCont.setBackgroundColor(background);
		return hRefCont;
	}

	/**
	 * Gets the underlying data used to build hRefComponent.
	 *
	 * @param  hRefComp the HRef component
	 * @return the HRef used by the specified component
	 */
	public LyHRef getUserData(Component hRefComp)
	{
		return (LyHRef)((LyJHRefComponent)hRefComp).getUserData();
	}

	/**
	 * Gets the current count of HRefComponents.
	 *
	 * @param  hRefCont the HRef container
	 * @return the number of HRef component contained on the specified container
	 */
	public int getHRefComponentCount(LyJHRefContainer hRefCont)
	{
		return hRefCont.getHRefComponentCount();
	}

	/**
	 * Gets the HRefComponent at position index.
	 *
	 * @param  index    the position of the HRef component in the container
	 * @param  hRefCont the HRef container
	 * @return the HRef component at the given position in the container
	 */
	public Component getHRefComponent(int index, LyJHRefContainer hRefCont)
	{
		return hRefCont.getHRefComponent(index);
	}

	/**
	 * Adds a HRefComponent in this container.
	 *
	 * @param cont  the place to add the component
	 * @param comp  the component to add
	 * @param index the position where to add the component in the container
	 */
	public void addHRefComponent(Container cont, Component comp, int index)
	{
		((LyJHRefContainer)cont).add((LyJHRefComponent)comp, index);
	}

	/**
	 * Removes a HRefComponent from its container.
	 *
	 * @param cont HRef container of all the HRef components
	 * @param comp HRef component to remove
	 */
	public void removeHRefComponent(Container cont, Component comp)
	{
		((LyJHRefContainer)cont).remove((LyJHRefComponent)comp);
	}

	/**
	 * Fills with HRefs.
	 */
	public void setContent()
	{
		Color background = getColor("LY_DEFAULT_BACKGROUND");

		_hRefCont = createHRefContainer(background);

		int    i, n = _hRefs.getHRefCount();
		LyHRef hRef;

		for (i = 0; i < n; i++)
		{
			hRef = _hRefs.getHRef(i);

			addHRef(hRef, -1);
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		if (_actionTransmitter != null)
		{
			Component hRefComponent = (Component)actionEvent.getSource();
			LyHRef    hRef = getUserData(hRefComponent);

			if (hRef != null)
				actionEvent.setValue(hRef.getId());

			_actionTransmitter.processActionEvent(actionEvent);
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _hRefCont;
	}

	/* Implementation of LySetHRefsInterface */

	/**
	 * Adds an hRef to this hRefs list.
	 *
	 * @see LyGraphicalHRefs#addHRef(LyHRef, int)
	 */
	public void addHRef(LyHRef hRef, int pos)
	{
		// Create the Href component and add it to the list
		if (_hRefCont != null)
		{
			Component		    hRefComp = ((LySwingHRef)hRef.getInternalView()).getComponent();

			// Register for user events
			LyActionTransmitter transmitter = ((LyEventTransmitter)hRefComp).getActionTransmitter();

			if (transmitter != null)
				transmitter.addActionListener(this);

			addHRefComponent(_hRefCont, hRefComp, pos);
		}
	}

	/**
	 * Removes the specified hRef from the hRefs list.
	 *
	 * @see LyGraphicalHRefs#removeHRef(LyHRef)
	 */
	public void removeHRef(LyHRef hRef)
	{
		// Find hRefCanvas corresponding to hRef
		if (_hRefCont != null)
		{
			Component hRefComp;
			int		  i, n = getHRefComponentCount(_hRefCont);

			for (i = 0; i < n; i++)
			{
				// Remove hRefCanvas
				hRefComp = getHRefComponent(i, _hRefCont);
				if (getUserData(hRefComp) == hRef)
				{
					removeHRefComponent(_hRefCont, hRefComp);
					return;
				}
			}
		}
	}

	/**
	 * Removes all the hRefs from the hRefs list.
	 */
	public void removeAll()
	{
		if (_hRefCont != null)
			_hRefCont.removeAll();
	}

} // end class LySwingHRefs
