/*
 * LySwingIcon.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import leon.misc.*;
import leon.view.*;
import leon.view.graphic.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Icon object (as described in the dtd) for swing viewer.
 *
 * <p>An icon allows to define an image to which a text can be associated.</p>
 *
 * @version $Id: LySwingIcon.java 34785 2012-03-16 15:36:16Z jlsaouli $
 * @author  Lyria
 * @see     LyIcon
 */
public class LySwingIcon extends LySwingFieldItem
	implements LySwingComponent,
		LyGraphicalIcon
{
	public static final String __VERSION =
		"$Id: LySwingIcon.java 34785 2012-03-16 15:36:16Z jlsaouli $";

	/** The view independent icon description. */
	protected LyIcon		   _icon;

	/** The swing component of this icon image. */
	protected Component		   _component;

	/** The image of this icon. */
	protected Image			   _image;

	/** Width of this icon. -1 for auto. */
	protected int			   _width;

	/** Height of this icon. -1 for auto. */
	protected int			   _height;

	/**
	 * Border of the icon.
	 */
	protected LyIcon.Border   _borderType;

	/**
	 * Class constructor of an Icon object for the Swing target.
	 *
	 * @param icon a ICON component defined for every type of view targets. This icon is initialized
	 *             for the Swing view specificaly.
	 */
	public LySwingIcon(LyIcon icon)
	{
		super(icon);
		_icon = icon;

		initialize();
	}

	/**
	 * Initializes this icon component depending on its image and the properties of this class.
	 */
	protected void initialize()
	{
		String image = _icon.getImage();

		_width = _icon.getWidth();
		_height = _icon.getHeight();
		_borderType = _icon.getBorder();

		if (((_width < 0) || (_height < 0)) && ((image == null) || (image.length() == 0)))
			_image = null;
		else if ((image != null)
		     && image.startsWith("file://"))
		{
			image = image.substring(7);
			_image = (Image)getEnvironment().getImages().loadImage(null, image, true);
		}
		else
			_image = getImage(_icon.getImage());

		if (_image != null)
		{
			if ((_width < 0) && (_height < 0))
			{
				_width = _image.getWidth(null);
				_height = _image.getHeight(null);
			}
			else if (_width < 0)
			{
				int w = _image.getWidth(null);
				int h = _image.getHeight(null);

				// Keep aspect ratio
				if ((h > 0) && (w > 0))
					_width = w * _height / h;
			}
			else if (_height < 0)
			{
				int w = _image.getWidth(null);
				int h = _image.getHeight(null);

				// Keep aspect ratio
				if ((h > 0) && (w > 0))
					_height = h * _width / w;
			}
		}
	} // end method initialize

	/**
	 * Creates the swing icon.
	 *
	 * @return the default icon for this component
	 */
	public Component createIcon()
	{
		LyJIconJLabel c = new LyJIconJLabel(_image, _width, _height, _borderType);
		Color		  col = getColor("LY_FORM_BACKGROUND");

		if (col == null)
			col = getColor("LY_DEFAULT_BACKGROUND");

		if (col != null)
			c.setBackground(col);

		if ((_icon.getEnvironment() != null) && (_icon.getLabel() != null))
		{
			String	tooltip = _icon.getEnvironment().translate(_icon.getLabel());
			
			c.setToolTipText(tooltip);
		}
		
		return c;
	}

	/**
	 * Sets the image displayed by this icon.
	 *
	 * @param image the image to use for the icon
	 * @see   #getImage()
	 */
	public void setImage(Image image)
	{
		((LyJIconJLabel)_component).setImage(image);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		if (_component == null)
			_component = createIcon();

		return _component;
	}

	/**
	 * Gets the image displayed by this icon.
	 *
	 * @return the image of the icon
	 * @see    #setImage(Image)
	 */
	public Image getImage()
	{
		return (_image);
	}

	/**
	 * Sets the new image of this icon from an image on the filesystem or in the resources.
	 *
	 * @param image the new image
	 * @see   #getImage()
	 * @see   leon.view.setinterface.LySetImageInterface#setImage(String)
	 */
	public void setImage(String image)
	{
		LyEnvironment environment = getEnvironment();

		if ((image != null)
			&& image.startsWith("file://"))
		{
			image = image.substring(7);
			_image = (Image)environment.getImages().loadImage(null, image, true);
		}
		else
		{
			if (((_width < 0) || (_height < 0)) && ((image == null) || (image.length() == 0)))
				_image = null;
			else
				_image = getImage(_icon.getImage());
		}

		setImage(_image);
	}

} // end class LySwingIcon
