/*
 * LySwingInterval.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.text.*;
import java.util.*;
import leon.misc.*;
import leon.view.*;
import leon.view.graphic.*;
import leon.view.win.swing.component.*;
import leon.view.win.swing.component.LyGanttElementImpl.*;


/**
 * This class implements the Interval object (as described in the dtd) for swing viewer.
 *
 * <p>An interval is defined by a begin date and an end date. If the end date is not specified, the
 * interval is replaced by a point (an image representing a date). An interval has a label and a
 * color. It may be resizable (it is possible to change its begin and end dates), movable on its
 * parent line and may have vertical lines on its extremities. It may also have dependencies to
 * other intervals represented by arrows linking this interval with the pointed intervals. An
 * interval may also contain periods allowing to split it into several sub-intervals. These periods
 * are described by a begin date and an end date.</p>
 *
 * @version $Id: LySwingInterval.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     leon.view.LyInterval
 */
public class LySwingInterval extends LySwingObject
	implements LyGraphicalInterval
{
	public static final String   __VERSION =
		"$Id: LySwingInterval.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The view independent interval description. */
	protected LyInterval		 _interval;

	/** Gantt element implementation. */
	protected LyGanttElementImpl _ganttElement;
	
	/** Begin date cache **/
	protected Date _beginDate;
	
	/** End date cache **/
	protected Date _endDate;

	/**
	 * Class constructor of an Interval object for the Swing target.
	 *
	 * @param interval an INTERVAL component defined for every type of view targets. This interval
	 *                 is initialized for the Swing view specificaly and filled with the default
	 *                 values.
	 */
	public LySwingInterval(LyInterval interval)
	{
		super(interval);
		_interval = interval;
		_ganttElement = null;
		initialize();
		setContent();
	}

	/**
	 * Initializes the component by creating a new Gantt element associated with this interval.
	 */
	public void initialize()
	{
		_ganttElement = new LyGanttElementImpl(_interval);

		// Do not use an image : Gantt element draws itself if represented by a point
		// _ganttElement.setGanttImage(getImage("LY_GANTT_DIAMOND"));
		_ganttElement.setDependencyColor(getColor("LY_GANTT_DEPENDENCY_COLOR"));
		_ganttElement.setMovable(false);
		_ganttElement.setLeftResizable(false);
		_ganttElement.setRightResizable(false);
		_ganttElement.setDrawVerticalLines(false);
	}

	/**
	 * Sets the Gantt element default properties.
	 */
	public void setContent()
	{
		String label = _interval.getLabel();
		String centerLabel = _interval.getCenterLabel();
		String tooltip = _interval.getTooltip();
		String color = _interval.getColor();
		String font = _interval.getFont();

		if ((color != null) && (color.length() > 0))
			_ganttElement.setColor(getColor(color));
		else
			_ganttElement.setColor(getColor("LY_GANTT_INTERVAL_COLOR"));

		if ((font != null) && (font.length() > 0))
			_ganttElement.setFont(getFont(font));
		else
			_ganttElement.setFont(getFont("LY_GANTT_FONT"));

		color = _interval.getFontColor();

		if ((color != null) && (color.length() > 0))
			_ganttElement.setLabelColor(getColor(color));
		else
			_ganttElement.setLabelColor(getColor("LY_GANTT_FONT_COLOR"));

		color = _interval.getVerticalLinesColor();

		if ((color != null) && (color.length() > 0))
			_ganttElement.setVerticalLinesColor(getColor(color));

		_ganttElement.setComponent(_interval);
		_ganttElement.setId(_interval.getId());
		_ganttElement.setLabel(translate(label));
		_ganttElement.setCenterLabel(translate(centerLabel));
		_ganttElement.setTooltip(translate(tooltip));
		_ganttElement.setMovable(_interval.isMovable());
		_ganttElement.setLeftResizable(_interval.isLeftResizable());
		_ganttElement.setRightResizable(_interval.isRightResizable());
		_ganttElement.setDrawVerticalLines(_interval.isDrawVerticalLines());

		resetDates();
		resetPeriods(null);
	} // end method setContent

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
	 */
	public void setLabel(String label)
	{
		_ganttElement.setLabel(translate(label));
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalInterval#setCenterLabel(java.lang.String)
	 */
	public void setCenterLabel(String label)
	{
		_ganttElement.setCenterLabel(translate(label));
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalInterval#setTooltip(java.lang.String)
	 */
	public void setTooltip(String tooltip)
	{
		_ganttElement.setTooltip(translate(tooltip));
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetColorInterface#setColor(java.lang.String)
	 */
	public void setColor(String color)
	{
		_ganttElement.setColor(getColor(color));
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetFontInterface#setFont(java.lang.String)
	 */
	public void setFont(String font)
	{
		_ganttElement.setFont(getFont(font));
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalInterval#setForegroundColor(java.lang.String)
	 */
	public void setForegroundColor(String color)
	{
		_ganttElement.setLabelColor(getColor(color));
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalInterval#setVerticalLinesColor(java.lang.String)
	 */
	public void setVerticalLinesColor(String color)
	{
		_ganttElement.setVerticalLinesColor(getColor(color));
	}

	/**
	 * Resets the dates of the Gantt element with those of the interval then sets the type of this
	 * element: {@link LyGanttElement#REAL_INTERVAL} if it has an end date or {@link
	 * LyGanttElement#POINT_INTERVAL} if not.
	 */
	public void resetDates()
	{
		// clear cache
		_beginDate = null;
		_endDate = null;
		// If interval has no parent Gantt, date format may not be found
		if (getParentGantt() == null)
			return;

		Date	   beginDate = getBeginDate();
		Date	   endDate = getEndDate();
		DateFormat dateFormat = getDateTimeFormat();

		_ganttElement.setBeginDate(beginDate);
		_ganttElement.setEndDate(endDate);
		_ganttElement.setDateFormat(dateFormat);

		// Gantt element is a point only if it has no end date
		if (endDate == null)
			_ganttElement.setType(LyGanttElement.POINT_INTERVAL);
		else
			_ganttElement.setType(LyGanttElement.REAL_INTERVAL);
	}

	/**
	 * Returns the date and time format of the Gantt object associated with this interval.
	 *
	 * @return the date and time format
	 */
	protected DateFormat getDateTimeFormat()
	{
		// Get the date format from the Gantt component
		LyGantt		 gantt = getParentGantt();
		LySwingGantt winGantt = (LySwingGantt)gantt.getInternalView();
		DateFormat   dateFormat = winGantt.getDateTimeFormat();

		return dateFormat;
	}

	/**
	 * Returns the date format of the Gantt object associated with this interval.
	 *
	 * @return the date format
	 */
	protected DateFormat getDateFormat()
	{
		// Get the date format from the Gantt component
		LyGantt		 gantt = getParentGantt();
		LySwingGantt winGantt = (LySwingGantt)gantt.getInternalView();
		DateFormat   dateFormat = winGantt.getDateFormat();

		return dateFormat;
	}

	/**
	 * Returns the begin date of the interval using the date format.
	 *
	 * @return the begin date
	 * @see    #setBeginDate(String)
	 */
	public Date getBeginDate()
	{
		if (_beginDate == null)
			_beginDate = parseDate(_interval.getBeginDate());
		return _beginDate;
	}

	/**
	 * Returns the end date of the interval using the date format.
	 *
	 * @return the end date
	 * @see    #setEndDate(String)
	 */
	public Date getEndDate()
	{
		if (_endDate == null)
			_endDate = parseDate(_interval.getEndDate());
		return _endDate;
	}

	/**
	 * Parses a date value using date time format and date format. If the date value cannot be
	 * parsed, the returned date is <code>null</code>.
	 *
	 * @param  value the date value
	 * @return the parsed date
	 */
	public Date parseDate(String value)
	{
        if (value == null)
            return null;
        
		DateFormat dateFormat = getDateTimeFormat();
		Date	   date = null;

		try
		{
			date = dateFormat.parse(value);
		}
		catch (Exception e)
		{
			try
			{
				// If date could not be parsed with date and time format, try with date only
				dateFormat = getDateFormat();
				date = dateFormat.parse(value);
			}
			catch (Exception e2)
			{
			}
		}

		return date;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalInterval#setBeginDate(java.lang.String)
	 */
	public void setBeginDate(String beginDate)
	{
		resetDates();

		LyLine	    parent = (LyLine)_interval.getParent();
		LySwingLine internal = (LySwingLine)parent.getInternalView();

		internal.resetDates();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalInterval#setEndDate(java.lang.String)
	 */
	public void setEndDate(String endDate)
	{
		resetDates();

		LyLine	    parent = (LyLine)_interval.getParent();
		LySwingLine internal = (LySwingLine)parent.getInternalView();

		internal.resetDates();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalInterval#setDrawVerticalLines(boolean)
	 */
	public void setDrawVerticalLines(boolean drawVerticalLines)
	{
		_ganttElement.setDrawVerticalLines(drawVerticalLines);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalInterval#setLeftResizable(boolean)
	 */
	public void setLeftResizable(boolean leftResizable)
	{
		_ganttElement.setLeftResizable(leftResizable);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalInterval#setRightResizable(boolean)
	 */
	public void setRightResizable(boolean rightResizable)
	{
		_ganttElement.setRightResizable(rightResizable);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalInterval#setMovable(boolean)
	 */
	public void setMovable(boolean movable)
	{
		_ganttElement.setMovable(movable);
	}

	/**
	 * Resets the dependencies of this Gantt element's parent component. The specified dependency is
	 * already set when this method is called and is only used here to refresh the parent Swing
	 * component.
	 *
	 * @param dependencies value of dependencies, not used in this implementation
	 * @see   leon.view.graphic.LyGraphicalInterval#setDependencies(java.lang.String)
	 */
	public void setDependencies(String dependencies)
	{
		LyGantt		 gantt = getParentGantt();
		LySwingGantt winGantt = (LySwingGantt)gantt.getInternalView();

		winGantt.resetDependencies();
	}

	/**
	 * Reset the dependencies of the Gantt element with those of the interval. This method must be
	 * called at initialization by the Gantt component.
	 */
	protected void resetDependencies()
	{
		String				   dependencies = _interval.getDependencies();
		Vector<LyGanttElement> result = new Vector<LyGanttElement>();

		if ((dependencies == null) || (dependencies.length() == 0))
		{
			_ganttElement.setDependencies(null);
			return;
		}

		LyGantt  gantt = getParentGantt();
		String[] dependenciesArray = LyString.stringToList(dependencies, ' ');

		if (dependenciesArray != null)
		{
			// For each dependency, retrieve the pointed interval and its Gantt element
			// and add it to the list of dependencies
			for (int j = 0; j < dependenciesArray.length; j++)
			{
				String     intervalId = dependenciesArray[j];
				LyInterval intervalTmp = (LyInterval)gantt.getComponent(intervalId);

				if (intervalTmp != null)
				{
					LySwingInterval winInterval = (LySwingInterval)intervalTmp.getInternalView(
							true);

					result.addElement(winInterval.getGanttElement());
				}
			}
		}

		if (!result.isEmpty())
			_ganttElement.setDependencies(result);
	} // end method resetDependencies

	/**
	 * Recomputes all the periods of this Gantt element. The specified period is already added when
	 * this method is called and is only used here to refresh the Swing component.
	 *
	 * @param period a period to add to this Gantt element, not used in this implementation
	 * @param i      the position to add the specified period, not used in this implementation
	 * @see   #resetPeriods(LyPeriod)
	 * @see   leon.view.graphic.LyGraphicalInterval#addPeriod(LyPeriod, int)
	 */
	public void addPeriod(LyPeriod period, int i)
	{
		resetPeriods(null);
	}

	/**
	 * @see leon.view.graphic.LyGraphicalInterval#removePeriod(LyPeriod)
	 */
	/**
	 * Recomputes all the periods of this Gantt element. The specified period is already removed
	 * when this method is called and is only used here to refresh the Swing component.
	 *
	 * @param period a period to remove to this Gantt element, not used in this implementation
	 * @see   #resetPeriods(LyPeriod)
	 * @see   leon.view.graphic.LyGraphicalInterval#removePeriod(LyPeriod)
	 */
	public void removePeriod(LyPeriod period)
	{
		resetPeriods(period);
	}

	/**
	 * Recomputes periods for the Gantt element without the specified period.
	 *
	 * @param removedPeriod period to remove during the recalculation
	 */
	protected void resetPeriods(LyPeriod removedPeriod)
	{
		int				  i, n = _interval.getPeriodCount();
		LyPeriod		  period;
		Vector<_LyPeriod> periods = ((n == 0) ? null : new Vector<_LyPeriod>(n));

		for (i = 0; i < n; i++)
		{
			period = _interval.getPeriod(i);
			if (period != removedPeriod)
				periods.addElement(new _LyPeriod(parseDate(period.getBeginDate()),
						parseDate(period.getEndDate())));
		}

		_ganttElement.setPeriods(periods);
	}

	/**
	 * Returns the parent Gantt component of the interval.
	 *
	 * @return the parent Gantt component
	 */
	protected LyGantt getParentGantt()
	{
		LyComponent parent = _interval.getParent();

		if (parent instanceof LyLine)
		{
			LyLine	    line = (LyLine)parent;
			LySwingLine winLine = (LySwingLine)line.getInternalView();

			return winLine.getParentGantt();
		}

		return null;
	}

	/**
	 * Returns the parent super line of the given component. A super line is a line with sub-lines
	 * only.
	 *
	 * @param  component the component to get its super line
	 * @return the parent super line of the specified component
	 */
	protected LyLine getSuperLine(LyComponent component)
	{
		if (component == null)
			return null;

		LyComponent parent = component.getParent();

		if (parent instanceof LyLine)
		{
			LyLine line = (LyLine)parent;
			int    i, n = line.getComponentCount();

			for (i = 0; i < n; i++)
			{
				LyComponent child = line.getComponent(i);

				// If line has an interval child, recurse
				if (child instanceof LyInterval)
					return line;
			}

			return getSuperLine(line);
		}

		return null;
	}

	/**
	 * This interval component belongs to a Gantt planning. It is not meant to be accessed directly.
	 *
	 * @return <code>null</code>
	 * @see    leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return null;
	}

	/**
	 * Accessor to the gantt element.
	 *
	 * @return the gantt element
	 */
	public LyGanttElement getGanttElement()
	{
		return _ganttElement;
	}

} // end class LySwingInterval
