/*
 * LySwingItem.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.event.*;
import java.lang.reflect.*;

import javax.swing.*;

import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.graphic.*;


/**
 * This class implements the Item object (as described in the dtd) for swing viewer.
 *
 * <p>A menu entry allows to execute an action like, as when using an action button. The entry can
 * be of check box type (checkbox attribute) and can open a dialog window (dialog attribute).</p>
 *
 * @version $Id: LySwingItem.java 30546 2011-04-15 07:18:03Z jlsaouli $
 * @author  Lyria
 * @see     LySwingMenu
 * @see     LyItem
 */
public class LySwingItem extends LySwingObject
	implements LyGraphicalItem,
		ActionListener,
		ItemListener,
		LyEventTransmitter
{
	public static final String    __VERSION =
		"$Id: LySwingItem.java 30546 2011-04-15 07:18:03Z jlsaouli $";

	/** The view independent item description. */
	protected LyItem			  _item;

	/** The swing component of this item. */
	protected Object			  _itemComp;

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter _actionTransmitter;

	/**
	 * Class constructor of an Item object for the Swing target.
	 *
	 * @param item an ITEM component defined for every type of view targets. This item is
	 *             initialized for the Swing view specificaly and filled with the default values.
	 */
	public LySwingItem(LyItem item)
	{
		super(item);
		_item = item;

		initialize();
		setContent();
	}

	/**
	 * Initializes the item and sets its action listeners. If this item has a checkbox, the menu
	 * item is created accordingly.
	 */
	public void initialize()
	{
		boolean isCheckBox = _item.isCheckBox();

		_itemComp = createItem(isCheckBox);
		registerListeners();
	}

	/**
	 * Initializes the item content info.
	 */
	public void setContent()
	{
		LyEnvironment environment = getEnvironment();
		String		  label = translate(_item.getLabel());
		String		  shortcut = translate(_item.getShortcut());
		Character     mnemonic = environment.getMnemonic(_item.getLabel());

		// Add "..." if dialog item
		if (_item.isDialog()
		         && (label != null)
		         && !label.endsWith("..."))
			label += "...";

		setLabel(label);
		setEnabled(_item.isEnabled());
		setShortCut(shortcut);
		setImage(_item.getImage());

		if (mnemonic != null)
			setMnemonic(mnemonic.charValue());
	}

	/**
	 * Creates a menu item with a checkbox if the item needs one.
	 *
	 * @param  isCheckBox specify if this item has a checkBox or no
	 * @return the menu item
	 */
	public Object createItem(boolean isCheckBox)
	{
		Object itemComp;

		if (isCheckBox)
		{
			itemComp = new JCheckBoxMenuItem("");
			((JCheckBoxMenuItem)itemComp).setState(_item.isChecked());
		}
		else
			itemComp = new JMenuItem("");

		Font font = getFont("LY_MENU_FONT");

		((JMenuItem)itemComp).setFont(font);

		/* Workaround for the menus bug */
		((JMenuItem)itemComp).enableInputMethods(false);

		return itemComp;
	}

	/**
	 * Gets the item component.
	 *
	 * @return the item of the component
	 */
	public Object getItem()
	{
		return _itemComp;
	}

	/**
	 * Invoked when a checkboxed menu item is selected or unselected.
	 */
	public void itemStateChanged(ItemEvent event)
	{
		boolean checked = (event.getStateChange() == ItemEvent.SELECTED);

		// Refresh parent info
		_item.setChecked(checked);

		// Notify controller if any
		if (_item.getController() != null)
		{
			// Notify controller
			LyActionDemand demand = new LyActionDemand(LyActionDemand.TYPE_ACTION, _item.getId());

			processDemand(demand);
		}
		else if (_actionTransmitter != null)
		{
			LyActionEvent actionEvent = new LyActionEvent(event.getSource());

			actionEvent.setSourceId(_item.getId());
			_actionTransmitter.processActionEvent(actionEvent);
		}
	}

	/**
	 * Registers for item events.
	 */
	public void registerListeners()
	{
		if (_itemComp instanceof JCheckBoxMenuItem)
			((JCheckBoxMenuItem)_itemComp).addItemListener(this);
		else if (_itemComp instanceof JMenuItem)
			((JMenuItem)_itemComp).addActionListener(this);
	}

	/**
	 * Sets the keyboard shortcut for this item.
	 *
	 * @param shortcut a String representing the sequence of keys to press
	 */
	public void setShortCut(String shortcut)
	{
		if ((shortcut == null) || (shortcut.length() == 0))
			return;

		String    result = null;
		KeyStroke menushortcut = null;

		if (shortcut.length() == 1)
		{
			result = shortcut.toUpperCase();

			if (result.equals(" "))
				result = "SPACE";

			result = "control pressed " + result;
			result = result.trim();
			menushortcut = KeyStroke.getKeyStroke(result);
		}
		else if (shortcut.startsWith("F") && ((shortcut.length() == 2) || (shortcut.length() == 3)))
		{
			// Function key as shortcut
			try
			{
				Field   field = KeyEvent.class.getField("VK_" + shortcut);
				Integer key = (Integer)field.get(null);

				menushortcut = KeyStroke.getKeyStroke(key.intValue(), 0);
			}
			catch (Exception e)
			{
				_item.getEnvironment().logWarning("[LySwingItem] Invalid shortcut : " + shortcut);
			}
		}
		else if (shortcut.startsWith("SHIFT+F")
				&& ((shortcut.length() == 8) || (shortcut.length() == 9)))
		{
			// Shift + function key as shortcut
			try
			{
				String	newShortcut = "VK_" + shortcut.substring(("SHIFT+F").length() - 1);
				Field  	field = KeyEvent.class.getField(newShortcut);
				Integer key = (Integer)field.get(null);
				
				menushortcut = KeyStroke.getKeyStroke(key.intValue(), ActionEvent.SHIFT_MASK);
			}
			catch (Exception e)
			{
				_item.getEnvironment().logWarning("[LySwingItem] Invalid shortcut : " + shortcut);
			}
		}
		else if (shortcut.equals("DEL"))
		{
			menushortcut = KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0);
		}

		if (menushortcut != null)
			((JMenuItem)_itemComp).setAccelerator(menushortcut);
	} // end method setShortCut

	/**
	 * Sets the keyboard mnemonic for this item.
	 *
	 * @param mnemonic a character used as mnemonic for the item
	 */
	public void setMnemonic(char mnemonic)
	{
		((JMenuItem)_itemComp).setMnemonic(mnemonic);
	}

	/**
	 * Sets the text label of this item.
	 *
	 * @see LyGraphicalItem#setLabel(String)
	 */
	public void setLabel(String label)
	{
		label = translate(label);
		((JMenuItem)_itemComp).setText(label);
	}

	/**
	 * Sets the checked state of the checkbox menu item without triggering the action listener. If
	 * not a checkbox menu item, the method does nothing.
	 *
	 * @param checked new state
	 * @see   LyGraphicalItem#setChecked(boolean)
	 */
	public void setChecked(boolean checked)
	{
		if (_itemComp instanceof JCheckBoxMenuItem)
		{
			((JCheckBoxMenuItem)_itemComp).removeItemListener(this);
			((JCheckBoxMenuItem)_itemComp).setState(checked);
			((JCheckBoxMenuItem)_itemComp).addItemListener(this);
		}
	}

	/**
	 * Sets the status of the menu. If disabled, the menu is grayed and inactive.
	 *
	 * @param enabled new status
	 * @see   leon.view.setinterface.LySetEnabledInterface#setEnabled(boolean)
	 */
	public void setEnabled(boolean enabled)
	{
		if (_itemComp != null)
			((JMenuItem)_itemComp).setEnabled(enabled);
	}

	/**
	 * Sets the image icon for the menu item.
	 *
	 * @param image Image of the item.
	 * @see   LyGraphicalItem#setImage(String)
	 */
	public void setImage(String image)
	{
		if (image != null)
		{
			Image img = (Image)getEnvironment().getImage(image);
			Icon  icon = new ImageIcon(img);

			((JMenuItem)_itemComp).setIcon(icon);
		}
	}

	/**
	 * Invoked when a menu item is activated.
	 *
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent event)
	{
		// Notify controller if any
		if (_item.getController() != null)
		{
			// Notify controller
			LyActionDemand demand = new LyActionDemand(LyActionDemand.TYPE_ACTION, _item.getId());

			processDemand(demand);
		}
		else if (_actionTransmitter != null)
		{
			LyActionEvent actionEvent = new LyActionEvent(event.getSource());

			actionEvent.setSourceId(_item.getId());
			_actionTransmitter.processActionEvent(actionEvent);
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return (_actionTransmitter);
	}

	/**
	 * This menu item doesn't handle selection events, so it has no selection transmitter.
	 *
	 * @return <code>null</code>
	 * @see    leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		return null;
	}

} // end class LySwingItem
