/*
 * LySwingLabel.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import javax.swing.*;
import leon.view.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;


/**
 * This class implements a text shown as a label (as described in the dtd) for swing viewer.
 *
 * <p>An input field of label type allows to display a simple, non-editable label inside a form. It
 * is used in the consultation forms to show different types of fields (date, number, choice,...).
 * It can be made up of one or several lines. It does not comprise any element.</p>
 *
 * @version $Id: LySwingLabel.java 33540 2011-12-13 09:23:18Z nhamon $
 * @author  Lyria
 * @see     LyLabel
 */
public class LySwingLabel extends LySwingFieldItem
	implements LySwingComponent,
		LySetFontInterface,
		LyGraphicalLabel
{
	public static final String __VERSION =
		"$Id: LySwingLabel.java 33540 2011-12-13 09:23:18Z nhamon $";

	/** Number of pixels of the border. */
	public static int		   LABEL_BORDER = 3;

	/** The view independent label description. */
	protected LyLabel		   _label;

	/** The swing component that display the content of this label. */
	protected Component		   _labelComp;

	/**
	 * Class constructor of a Label object for the Swing target.
	 *
	 * @param label a LABEL component defined for every type of view targets. This label is
	 *              initialized with the default values.
	 */
	public LySwingLabel(LyLabel label)
	{
		super(label);
		_label = label;

		setContent();
	}

	/**
	 * Creates the swing label component.
	 *
	 * @param  label the text displayed by this label
	 * @param  align an int specifying the alignment of this label ({@link LyLabel#H_Align.LEFT},
	 *               {@link LyLabel#H_Align.RIGHT} or {@link LyLabel#H_Align.CENTER})
	 * @return the new label
	 */
	public Component createLabel(String label, LyLabel.HAlign align)
	{
		int compAlign = LyJMultiLineLabel.LEFT;

		if (align == LyLabel.HAlign.LEFT)
			compAlign = LyJMultiLineLabel.LEFT;
		else if (align == LyLabel.HAlign.CENTER)
			compAlign = LyJMultiLineLabel.CENTER;
		else if (align == LyLabel.HAlign.RIGHT)
			compAlign = LyJMultiLineLabel.RIGHT;

		LyJMultiLineLabel result = new LyJMultiLineLabel(label, -1, LABEL_BORDER, LABEL_BORDER,
				compAlign);

		result.setOpaque(false);

		return result;
	}

	/**
	 * Builds the component with the default properties of a label.
	 */
	public void setContent()
	{
        LyLabel.HAlign    align = _label.getHAlign();
		String            label = _label.getLabel();

		if (label != null)
			label = translate(label);
		else
			label = "";

		_labelComp = createLabel(label, align);

		Font font;

		if ((_label.getFont() != null) && !_label.isDefaultFont())
			font = getFont(_label.getFont());
		else
			font = getFont("LY_DEFAULT_FONT");

		_labelComp.setFont(font);

		((LyJMultiLineLabel)_labelComp).setOpaque(false);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _labelComp;
	}

	/**
	 * Sets the text label.
	 *
	 * @param label new label
	 * @see   LySetLabelInterface#setLabel(String)
	 */
	public void setLabel(String label)
	{
		if (_labelComp != null)
		{
			label = translate(label);

			// Keep the old width to avoid layout issues when label is displayed in a border panel.
			int	width = ((LyJMultiLineLabel)_labelComp).getWidth();
			
			((LyJMultiLineLabel)_labelComp).setMinimumWidth(width);
			((LyJMultiLineLabel)_labelComp).setLabel(label);
			
			// We resize the component all the time
			if (_labelComp.getParent() != null)
			{
				_labelComp.invalidate();
				((JComponent)_labelComp.getParent()).revalidate();
			}

			int	newWidth = ((LyJMultiLineLabel)_labelComp).getMinimumSize().width;
			
			if (newWidth > width)
			{
				LyComponent parent = _label.getParent();

				while (parent != null)
				{
					if (parent instanceof LyView)
					{
						// Reset content only if parent is not a resizable frame (otherwise the frame
						// may flash too much) or not a dialog with a given width or height
						if ((parent instanceof LyFrame) && (((LyFrame)parent).isResizable()))
							return;

						if ((parent instanceof LyDialog)
								&& ((((LyDialog)parent).getWidth() > 0) || ((((LyDialog)parent).getHeight() > 0))))
							return;

						if (((LyView)parent).isVisible())
						{
							Object internal = parent.getInternalView();

							((LyResetContentInterface)internal).resetContent();
						}
						return;
					}
					else
						parent = parent.getParent();
				}
			}
		}
	}

	/**
	 * Sets the font of the label component.
	 *
	 * @param fontValue new font value
	 * @see   LySetFontInterface#setFont(String)
	 */
	public void setFont(String fontValue)
	{
		if (_labelComp != null)
		{
			Font f = getFont(fontValue);

			_labelComp.setFont(f);

			// We resize the component all the time
			if (_labelComp.getParent() != null)
			{
				_labelComp.invalidate();
				((JComponent)_labelComp.getParent()).revalidate();
			}
		}
	}

	/**
     * Sets the image of the component.
     *
     * @param key new image key
     * @see   LySetImageInterface#setImage(String)
     */
    public void setImage(String key)
    {
    }
} // end class LySwingLabel
