/*
 * LySwingLine.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.text.*;
import java.util.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import leon.view.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.win.swing.component.*;
import leon.view.win.swing.component.LyJTreeElementImpl.*;


/**
 * This class implements the line object (as described in the dtd) for swing viewer.
 *
 * @version $Id: LySwingLine.java 34777 2012-03-16 15:25:13Z jlsaouli $
 * @author  Lyria
 * @see     LyLine
 */
public class LySwingLine extends LySwingObject
	implements LySwingComponent,
		LyJTreeElement,
		LyGetIdInterface,
		LyGraphicalLine
{
	public static final String		 __VERSION =
		"$Id: LySwingLine.java 34777 2012-03-16 15:25:13Z jlsaouli $";

	/** The object contained by this node. */
	protected _LyNodeContent		 _treeNodeContent;

	/** Used to notify the displayed tree that a node has changed. */
	protected LyJTreeComponent		 _treeComp;

	/** Mutable tree node object related to the current node. */
	protected _LyMutableTreeNode     _mutableTreeNode;

	/** Height factor of the current node. */
	protected int					 _heightFactor;

	/** The view independent line description. */
	protected LyLine				 _line;

	/** Indicates if this line is a super line (for example a line that contains sub-lines). */
	protected boolean				 _isSuperLine;

	/** List of Gantt elements representing the line. */
	protected Vector<LyGanttElement> _ganttElements;

	/** Reference to the interval that is being removed from the line. */
	protected LyInterval			 _removedInterval;

	/**
	 * Class constructor of a Line object for the Swing target.
	 *
	 * @param line a LINE component defined for every type of view targets. This line is initialized
	 *             with the default values.
	 */
	public LySwingLine(LyLine line)
	{
		super(line);
		_line = line;
		_isSuperLine = false;
		_removedInterval = null;
		initialize();
		setContent();
	}

	/**
	 * Initializes this line component.
	 */
	public void initialize()
	{
		_mutableTreeNode = new _LyMutableTreeNode();
		_mutableTreeNode.setUserObject(this);

		Image image;

		if (_line.getImage() != null)
			image = (Image)getEnvironment().getImage(_line.getImage());
		else if (getEnvironment() != null)
			image = (Image)getEnvironment().getImage("LY_GANTT_LINE_NODE");
		else
			image = null;

		_mutableTreeNode.setNodeContent(new _LyNodeContent(image, _line.getLabel(),
				_line.getTooltip(), _line.isSelected(), _line.isOpen()));
		_mutableTreeNode.setTreeElement(this);
	}

	/**
	 * Fills the line. Adds the lines to the {@link #_ganttElements} list if they are intervals or
	 * creates the super lines if not. Then update the Gantt tree elements swing component with the
	 * new content that just have been set.
	 *
	 * @see #getContent()
	 */
	public void setContent()
	{
		int				   i, n = _line.getComponentCount();
		boolean			   setSuperLine = false;
		_LyMutableTreeNode mutableTreeNode;

		_ganttElements = new Vector<LyGanttElement>();

		for (i = 0; i < n; i++)
		{
			LyComponent component = _line.getComponent(i);

			if (component instanceof LyLine)
			{
				LyLine	    line = (LyLine)component;
				LySwingLine internal = (LySwingLine)line.getInternalView();

				mutableTreeNode = (_LyMutableTreeNode)internal.getMutableTreeNode();
				_mutableTreeNode.add(mutableTreeNode);

				// If the line contains other line, it means that this one is a super-line
				setSuperLine = true;
			}
			else if (component instanceof LyInterval)
			{
				LyInterval	    interval = (LyInterval)component;
				LySwingInterval internal = (LySwingInterval)interval.getInternalView(true);
				LyGanttElement  ganttElement = internal.getGanttElement();

				_ganttElements.addElement(ganttElement);

				// If the line contains intervals, it means that this one is not a super-line
				_isSuperLine = false;
			}
		}

		if (setSuperLine)
			setSuperLine();

		updateGanttElements();
	} // end method setContent

	/**
	 * Sets the line as a super line. This line will be represented in the Gantt canvas by a black
	 * line with two arrows at its extremities.
	 *
	 * @see #getSuperLine(LyComponent)
	 */
	protected void setSuperLine()
	{
		_isSuperLine = true;

		LyGanttElementImpl ganttElement = new LyGanttElementImpl(_line);

		ganttElement.setId(_line.getId());
		ganttElement.setType(LyGanttElement.SUPER_LINE);

		String superLineColor = _line.getSuperLineColor();

		if ((getEnvironment() != null) && (superLineColor != null) && (superLineColor.length() > 0))
		{
			ganttElement.setColor(getColor(superLineColor));
			ganttElement.setLabelColor(getColor(superLineColor));
		}
		else if (getEnvironment() != null)
		{
			ganttElement.setColor(getColor("LY_GANTT_SUPERINTERVAL_COLOR"));
			ganttElement.setLabelColor(getColor("LY_GANTT_SUPERINTERVAL_COLOR"));
		}

		ganttElement.setLabel(_line.getSuperLineLabel());
		ganttElement.setHidden(_line.isSuperLineHidden());
		ganttElement.setTooltip(_line.getTooltip());
		ganttElement.setMovable(false);
		ganttElement.setLeftResizable(false);
		ganttElement.setRightResizable(false);
		ganttElement.setDrawVerticalLines(false);

		if (getEnvironment() != null)
			ganttElement.setFont(getFont("LY_GANTT_FONT"));

		_ganttElements.clear();
		_ganttElements.addElement(ganttElement);

		resetDates();
		setMovable(isMovable());
	} // end method setSuperLine

	/**
	 * Sets the movable state of the Gantt element.
	 *
	 * @param movable true if movable, false otherwise
	 */
	public void setMovable(boolean movable)
	{
		if (_isSuperLine)
		{
			LyGanttElementImpl ganttElement = (LyGanttElementImpl)_ganttElements.elementAt(0);

			ganttElement.setMovable(movable);
		}
	}

	/**
	 * Indicates if this line is a super line (for example, a line that contains sub-lines).
	 *
	 * @return true if this line is a super line, false otherwise
	 */
	public boolean isSuperLine()
	{
		return _isSuperLine;
	}

	/**
	 * Indicates if this line has at least one super line as a child.
	 *
	 * @return true if this line has at least one super line as a child, false otherwise
	 */
	public boolean hasSuperLineChildren()
	{
		int		    i, n = _line.getComponentCount();
		LyComponent component;

		for (i = 0; i < n; i++)
		{
			component = _line.getComponent(i);

			if (component instanceof LyLine)
			{
				LyLine	    line = (LyLine)component;
				LySwingLine internalLine = (LySwingLine)line.getInternalView();

				if (internalLine.isSuperLine())
					return true;
			}
		}

		return false;
	}

	/**
	 * Updates the tree element and Gantt elements association in the Gantt component.
	 */
	protected void updateGanttElements()
	{
		LyGantt gantt = getParentGantt();

		if (gantt != null)
		{
			LySwingGantt winGantt = (LySwingGantt)gantt.getInternalView();

			winGantt.setGanttElements(this, _ganttElements);
		}
	}

	/**
	 * Resets the dates of the line and updates the Gantt component.
	 */
	public void resetDates()
	{
		LyGantt gantt = getParentGantt();

		// If line has no parent Gantt, date format may not be found
		if (gantt == null)
			return;

		LySwingGantt winGantt = (LySwingGantt)gantt.getInternalView();
		DateFormat   dateFormat = winGantt.getDateFormat();
		Date		 beginDate = getLimitDate(false);
		Date		 endDate = getLimitDate(true);

		// If line is a super line, update the Gantt element dates
		if (_isSuperLine)
		{
			LyGanttElementImpl ganttElement = (LyGanttElementImpl)_ganttElements.elementAt(0);

			ganttElement.setBeginDate(beginDate);
			ganttElement.setEndDate(endDate);
			ganttElement.setDateFormat(dateFormat);
		}

		// Reset super line if necessary
		LyLine parentNoInterval = getSuperLine(_line);

		if ((parentNoInterval != null) && (parentNoInterval != _line))
		{
			LySwingLine internal = (LySwingLine)parentNoInterval.getInternalView();

			internal.resetDates();
		}

		// Reset Gantt if necessary (min or max date has changed)
		winGantt.fitSize(false);
	} // end method resetDates

	/**
	 * Resets the dependencies for the entire Gantt component.
	 */
	public void resetDependencies()
	{
		LyGantt		 gantt = getParentGantt();
		LySwingGantt winGantt = (LySwingGantt)gantt.getInternalView();

		winGantt.resetDependencies();
	}

	/**
	 * Returns the parent super line of the given component. If no parent superline is found,
	 * returns <code>null</code>.
	 *
	 * @param  component the component to get its super line
	 * @return the parent super line of the specified component
	 * @see    #setSuperLine()
	 */
	protected LyLine getSuperLine(LyComponent component)
	{
		if (component == null)
			return null;

		LyComponent parent = component.getParent();

		if (parent instanceof LyLine)
		{
			LyLine line = (LyLine)parent;
			int    i, n = line.getComponentCount();

			for (i = 0; i < n; i++)
			{
				LyComponent child = line.getComponent(i);

				// If line has an interval child, recurse
				if (child instanceof LyLine)
					return line;
			}

			return getSuperLine(line);
		}

		return null;
	}

	/**
	 * Returns the parent Gantt component of the line. If no parent is found, returns <code>
	 * null</code>.
	 *
	 * @return the parent Gantt component
	 */
	protected LyGantt getParentGantt()
	{
		LyComponent parent = _line.getParent();

		if (parent instanceof LyLine)
		{
			LyLine	    line = (LyLine)parent;
			LySwingLine winLine = (LySwingLine)line.getInternalView();

			return winLine.getParentGantt();
		}
		else if (parent instanceof LyGantt)
			return (LyGantt)parent;

		return null;
	}

	/**
	 * Returns the limit date for this line (minimum or maximum). Browses all the sub-components of
	 * the line and retrieves their minimum (or maximum) dates.
	 *
	 * @param  max Indicates if maximum or minimum date must be returned. If end date is <code>
	 *             null</code>, begin date is considered instead.
	 * @return the minimum date if max is true, the maximum date otherwise
	 */
	public Date getLimitDate(boolean max)
	{
		Date date = null;
		Date dateTmp = null;
		int  i, n = _line.getComponentCount();

		// Browse all the sub-component
		for (i = 0; i < n; i++)
		{
			LyComponent component = _line.getComponent(i);

			dateTmp = null;

			if (component instanceof LyLine)
			{
				LyLine	    line = (LyLine)component;
				LySwingLine internal = (LySwingLine)line.getInternalView();

				dateTmp = internal.getLimitDate(max);
			}
			else if (component instanceof LyInterval)
			{
				LyInterval	    interval = (LyInterval)component;
				LySwingInterval internal = (LySwingInterval)interval.getInternalView();

				// If the interval is being removed (removedcomponent()), do not consider it
				if (_removedInterval == interval)
					continue;

				if (max)
				{
					dateTmp = internal.getEndDate();

					if (dateTmp == null)
						// If end date is null, we consider the begin date
						dateTmp = internal.getBeginDate();
				}
				else
					dateTmp = internal.getBeginDate();
			}

			if (dateTmp == null)
				continue;

			if (date == null)
				date = dateTmp;
			else
			{
				if (max && dateTmp.after(date))
					date = dateTmp;

				if (!max && dateTmp.before(date))
					date = dateTmp;
			}

		}

		return date;
	} // end method getLimitDate

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return null;
	}

	/**
	 * Accessor to the gantt element list.
	 *
	 * @return the gantt element
	 */
	public Vector<LyGanttElement> getGanttElements()
	{
		return _ganttElements;
	}

	/**
	 * Indicates if the element is movable.
	 *
	 * @return true if the element is movable, false otherwise
	 */
	public boolean isMovable()
	{
		if (_line.isSuperLineHidden())
			return false;

		int i, n = _line.getComponentCount();

		for (i = 0; i < n; i++)
		{
			LyComponent component = _line.getComponent(i);

			if (component instanceof LyLine)
			{
				LyLine	    line = (LyLine)component;
				LySwingLine internal = (LySwingLine)line.getInternalView();

				if (!internal.isMovable())
					return false;
			}
			else if (component instanceof LyInterval)
			{
				LyInterval interval = (LyInterval)component;

				if (!interval.isMovable() && (interval.getBeginDate() != null))
					return false;
			}
		}

		return true;
	} // end method isMovable

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
	 */
	public void setLabel(String label)
	{
		_mutableTreeNode.getNodeContent().setLabel(label);

		if ((_treeComp == null)
				&& (_mutableTreeNode.getTreeElement() != null))
			_treeComp = _mutableTreeNode.getTreeElement().getTree();

		if (_treeComp != null)
			((DefaultTreeModel)_treeComp.getModel()).nodeChanged(_mutableTreeNode);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalLine#setTooltip(java.lang.String)
	 */
	public void setTooltip(String tooltip)
	{
		_mutableTreeNode.getNodeContent().setTooltip(tooltip);

		if ((_treeComp == null)
				&& (_mutableTreeNode.getTreeElement() != null))
			_treeComp = _mutableTreeNode.getTreeElement().getTree();

		if (_treeComp != null)
			((DefaultTreeModel)_treeComp.getModel()).nodeChanged(_mutableTreeNode);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetSelectedInterface#setSelected(boolean)
	 */
	public void setSelected(boolean status)
	{
		if ((_treeComp == null) && (_mutableTreeNode != null)
				&& (_mutableTreeNode.getTreeElement() != null))
			_treeComp = _mutableTreeNode.getTreeElement().getTree();

		if (_treeComp == null)
			return;

		_mutableTreeNode.getNodeContent().setSelected(status);

		TreePath   path = getTreePath();
		TreePath[] paths = _treeComp.getSelectionPaths();
		boolean    found = false;
		int		   pathsCount = (paths != null) ? paths.length : 0;

		for (int i = 0; ((i < pathsCount) && !found); i++)
		{
			if (path.equals(paths[i]))
				found = true;
		}

		if ((status && found) || (!status && !found))
			return;

		// this method is called from above (LyView), no need to notify
		_treeComp.setNotifySelections(false);
		if (status)
			_treeComp.addSelectionPath(path);
		else
			_treeComp.removeSelectionPath(getTreePath());
		_treeComp.setNotifySelections(true);
	} // end method setSelected

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalLine#setOpen(boolean)
	 */
	public void setOpen(boolean status)
	{
		_mutableTreeNode.getNodeContent().setOpen(status);

		if ((_ganttElements != null) && (_ganttElements.size() == 1))
		{
			LyGanttElementImpl ganttElement = (LyGanttElementImpl)_ganttElements.elementAt(0);

			ganttElement.refreshCanvas();
		}
	}

	/**
	 * Refreshes the tree component when it is expanded, no matter what event is triggered.
	 *
	 * @param event Not used in this implementation.
	 * @see   leon.view.win.swing.component.LyJTreeElement#treeExpanded(javax.swing.event.TreeExpansionEvent)
	 */
	public void treeExpanded(TreeExpansionEvent event)
	{
		if ((_ganttElements != null) && (_ganttElements.size() == 1))
		{
			LyGanttElementImpl ganttElement = (LyGanttElementImpl)_ganttElements.elementAt(0);

			ganttElement.refreshCanvas();
		}
	}

	/**
	 * Refreshes the tree component when it is collapsed, no matter what event is triggered.
	 *
	 * @param event Not used in this implementation.
	 * @see   leon.view.win.swing.component.LyJTreeElement#treeCollapsed(javax.swing.event.TreeExpansionEvent)
	 */
	public void treeCollapsed(TreeExpansionEvent event)
	{
		if ((_ganttElements != null) && (_ganttElements.size() == 1))
		{
			LyGanttElementImpl ganttElement = (LyGanttElementImpl)_ganttElements.elementAt(0);

			ganttElement.refreshCanvas();
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.getinterface.LyGetIdInterface#getId()
	 */
	public String getId()
	{
		return _line.getId();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getLabel()
	 */
	public String getLabel()
	{
		return _line.getLabel();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#isSelected()
	 */
	public boolean isSelected()
	{
		return _line.isSelected();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#isOpen()
	 */
	public boolean isOpen()
	{
		return _line.isOpen();
	}

	/**
	 * Returns a TreePath for this node.
	 *
	 * @return this node's tree path
	 * @see    leon.view.win.swing.component.LyJTreeElement#getTreePath()
	 */
	public TreePath getTreePath()
	{
		TreeNode[] nodesPath = _mutableTreeNode.getPath();
		TreePath   path = new TreePath(nodesPath);

		return path;
	}

	/**
	 * Accessor to the internal mutable tree node.
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getMutableTreeNode()
	 */
	public DefaultMutableTreeNode getMutableTreeNode()
	{
		return _mutableTreeNode;
	}

	/**
	 * Gets the content of this node.
	 *
	 * @return the content of the node
	 * @see    leon.view.win.swing.component.LyJTreeElement#getContent()
	 * @see    #setContent()
	 */
	public _LyNodeContent getContent()
	{
		return _mutableTreeNode.getNodeContent();
	}

	/**
	 * Sets the treeComponent that display this node.
	 *
	 * @param tree the new tree component
	 * @see   #getTree()
	 * @see   leon.view.win.swing.component.LyJTreeElement#setTree(leon.view.win.swing.component.LyJTreeComponent)
	 */
	public void setTree(LyJTreeComponent tree)
	{
		_treeComp = tree;
	}

	/**
	 * Gets the tree component that display this node.
	 *
	 * @return the current tree component
	 * @see    #setTree(LyJTreeComponent)
	 * @see    leon.view.win.swing.component.LyJTreeElement#getTree()
	 */
	public LyJTreeComponent getTree()
	{
		return _treeComp;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetImageInterface#setImage(java.lang.String)
	 */
	public void setImage(String image)
	{
		if (image != null)
			_mutableTreeNode.getNodeContent().setImage(getImage(_line.getImage()));

		if ((_treeComp == null) && (_mutableTreeNode != null)
				&& (_mutableTreeNode.getTreeElement() != null))
			_treeComp = _mutableTreeNode.getTreeElement().getTree();

		if (_treeComp != null)
			((DefaultTreeModel)_treeComp.getModel()).nodeChanged(_mutableTreeNode);
	}

	/**
	 * The specified color is already set for this line when this method is called and is only used
	 * here to refresh the Swing tree component.
	 *
	 * @param color Not used in this implementation.
	 * @see   #getColor()
	 * @see   leon.view.setinterface.LySetColorInterface#setColor(java.lang.String)
	 */
	public void setColor(String color)
	{
		if ((_treeComp == null) && (_mutableTreeNode != null)
				&& (_mutableTreeNode.getTreeElement() != null))
			_treeComp = _mutableTreeNode.getTreeElement().getTree();

		if (_treeComp != null)
			((DefaultTreeModel)_treeComp.getModel()).nodeChanged(_mutableTreeNode);
	}

	/**
	 * Returns the color of this line.
	 *
	 * @return the color of this line, or <code>null</code> if _line is <code>null</code>
	 * @see    #setColor(String)
	 * @see    leon.view.win.swing.component.LyJTreeElement#getColor()
	 */
	public Color getColor()
	{
		if (_line.getColor() != null)
			return getColor(_line.getColor());
		return null;
	}

	/**
	 * Returns the color of this line when selected.
	 *
	 * @return the color of this line when selected, or <code>null</code> if _line is <code>
	 *         null</code>
	 * @see    #getColor()
	 * @see    leon.view.win.swing.component.LyJTreeElement#getSelectedColor()
	 */
	public Color getSelectedColor()
	{
		if (getEnvironment() != null)
			return (Color)getEnvironment().getColor("LY_TREENODE_SELECTED_BG_COLOR");
		else
			return null;
	}

	/**
	 * Returns the font color of this line.
	 *
	 * @return the font color of this line, or <code>null</code> if _line is <code>null</code>
	 * @see    #setFontColor(String)
	 * @see    leon.view.win.swing.component.LyJTreeElement#getFontColor()
	 */
	public Color getFontColor()
	{
		if (_line.getFontColor() != null)
			return getColor(_line.getFontColor());
		return null;
	}

	/**
	 * Returns the font color of this line when selected.
	 *
	 * @return the font color of this line when selected, or <code>null</code> if _line is
	 *         <code>null</code>
	 * @see    #getFontColor()
	 * @see    leon.view.win.swing.component.LyJTreeElement#getSelectedFontColor()
	 */
	public Color getSelectedFontColor()
	{
		if (getEnvironment() != null)
			return (Color)getEnvironment().getColor("LY_TREENODE_SELECTED_FONT_COLOR");
		else
			return null;
	}

	/**
	 * Returns the font for this line.
	 *
	 * @return the font for this line, or <code>null</code> if _line is <code>null</code>
	 * @see    #setFont(String)
	 * @see    leon.view.win.swing.component.LyJTreeElement#getFont()
	 */
	public Font getFont()
	{
		if (_line.getFont() != null)
			return getFont(_line.getFont());
		return null;
	}

	/**
	 * The specified font is already set for this line when this method is called and is only used
	 * here to refresh the Swing tree component.
	 *
	 * @param font Not used in this implementation.
	 * @see   #getFont()
	 * @see   leon.view.setinterface.LySetFontInterface#setFont(java.lang.String)
	 */
	public void setFont(String font)
	{
		if ((_treeComp == null) && (_mutableTreeNode != null)
				&& (_mutableTreeNode.getTreeElement() != null))
			_treeComp = _mutableTreeNode.getTreeElement().getTree();

		if (_treeComp != null)
			((DefaultTreeModel)_treeComp.getModel()).nodeChanged(_mutableTreeNode);
	}

	/**
	 * The specified color is already set for this line when this method is called and is only used
	 * here to refresh the Swing tree component.
	 *
	 * @param color Not used in this implementation.
	 * @see   #getForegroundColor()
	 * @see   leon.view.graphic.LyGraphicalLine#setForegroundColor(java.lang.String)
	 */
	public void setForegroundColor(String color)
	{
		if ((_treeComp == null) && (_mutableTreeNode != null)
				&& (_mutableTreeNode.getTreeElement() != null))
			_treeComp = _mutableTreeNode.getTreeElement().getTree();

		if (_treeComp != null)
			((DefaultTreeModel)_treeComp.getModel()).nodeChanged(_mutableTreeNode);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalLine#setSuperLineColor(java.lang.String)
	 */
	public void setSuperLineColor(String color)
	{
		if (_isSuperLine && (color != null))
		{
			LyGanttElementImpl ganttElement = (LyGanttElementImpl)_ganttElements.elementAt(0);

			ganttElement.setColor(getColor(color));
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalLine#setSuperLineLabel(java.lang.String)
	 */
	public void setSuperLineLabel(String label)
	{
		if (_isSuperLine)
		{
			LyGanttElementImpl ganttElement = (LyGanttElementImpl)_ganttElements.elementAt(0);

			ganttElement.setLabel(translate(label));
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalLine#setSuperLineHidden(boolean)
	 */
	public void setSuperLineHidden(boolean hidden)
	{
		if (_isSuperLine)
		{
			LyGanttElementImpl ganttElement = (LyGanttElementImpl)_ganttElements.elementAt(0);

			ganttElement.setHidden(hidden);
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalLine#addComponent(leon.view.LyComponent, int)
	 */
	public void addComponent(LyComponent component, int pos)
	{
		_LyMutableTreeNode mutableTreeNode;

		if (component instanceof LyLine)
		{
			LyLine	    line = (LyLine)component;
			LySwingLine internal = (LySwingLine)line.getInternalView();

			mutableTreeNode = (_LyMutableTreeNode)internal.getMutableTreeNode();

			if (pos < 0)
				pos = _mutableTreeNode.getChildCount();

			int[] indice = new int[1];

			indice[0] = pos;

			_mutableTreeNode.insert(mutableTreeNode, pos);

			if ((_treeComp == null)
					&& (_mutableTreeNode.getTreeElement() != null))
				_treeComp = _mutableTreeNode.getTreeElement().getTree();

			if (_treeComp != null)
			{
				((DefaultTreeModel)_treeComp.getModel()).nodesWereInserted(_mutableTreeNode,
					indice);
				_treeComp.updateTree(this);
			}

			setSuperLine();
		}
		else if (component instanceof LyInterval)
		{
			LyInterval	    interval = (LyInterval)component;
			LySwingInterval internal = (LySwingInterval)interval.getInternalView(true);
			LyGanttElement  ganttElement = internal.getGanttElement();

			_ganttElements.addElement(ganttElement);

			_isSuperLine = false;
		}

		resetDates();
		updateGanttElements();
		setMovable(isMovable());
	} // end method addComponent

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalLine#removeComponent(leon.view.LyComponent)
	 */
	public void removeComponent(LyComponent component)
	{
		if (component instanceof LyLine)
		{
			_LyMutableTreeNode   mutableTreeNode;
			LyLine				 line = (LyLine)component;
			LySwingLine			 internal = (LySwingLine)line.getInternalView();

			int[]				 indice = new int[1];
			_LyMutableTreeNode[] removed = new _LyMutableTreeNode[1];

			removed[0] = (_LyMutableTreeNode)internal.getMutableTreeNode();
			indice[0] = _mutableTreeNode.getIndex(removed[0]);

			mutableTreeNode = (_LyMutableTreeNode)internal.getMutableTreeNode();
			_mutableTreeNode.remove(mutableTreeNode);

			if ((_treeComp == null)
					&& (_mutableTreeNode.getTreeElement() != null))
				_treeComp = _mutableTreeNode.getTreeElement().getTree();

			if (_treeComp != null)
			{
				((DefaultTreeModel)_treeComp.getModel()).nodesWereRemoved(_mutableTreeNode, indice,
					removed);
				_treeComp.updateTree(this);
			}
		}
		else if (component instanceof LyInterval)
		{
			LyInterval	    interval = (LyInterval)component;
			LySwingInterval internal = (LySwingInterval)interval.getInternalView(true);
			LyGanttElement  ganttElement = internal.getGanttElement();

			_ganttElements.removeElement(ganttElement);

			// Register the interval that is being removed, because the view has not been updated
			// yet.
			// This reference will be used to compute the new limits of the line.
			_removedInterval = interval;
		}

		// If last child is being removed, clear the gantt elements list
		int n = _line.getComponentCount();

		if (n == 1)
		{
			_ganttElements.clear();
			_isSuperLine = false;
		}

		resetDates();
		updateGanttElements();
		setMovable(isMovable());

		_removedInterval = null;
	} // end method removeComponent

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetHiddenInterface#setHidden(boolean)
	 */
	public void setHidden(boolean hidden)
	{
		LyComponent parent = _line.getParent();

		if (parent instanceof LyLine)
		{
			LyLine	    parentLine = (LyLine)parent;
			LySwingLine internal = (LySwingLine)parentLine.getInternalView();

			if (internal != null)
			{
				if (hidden)
					internal.removeComponent(_line);
				else
				{
					int		    pos = 0, i, n = parentLine.getComponentCount();
					LyComponent child;

					for (i = 0; i < n; i++)
					{
						child = parentLine.getComponent(i);

						if (_line.equals(child))
							break;

						if ((child instanceof LyLine) && !((LyLine)child).isHidden())
							pos++;
					}

					internal.addComponent(_line, pos);
				}
			}
		}

		if ((_ganttElements != null) && (_ganttElements.size() > 0))
		{
			LyGanttElementImpl ganttElement = (LyGanttElementImpl)_ganttElements.elementAt(0);

			ganttElement.refreshCanvas();
		}
	} // end method setHidden

	/**
	 * Sets the height factor for this node.
	 *
	 * @param factor the new height factor
	 * @see   #getHeightFactor()
	 */
	public void setHeightFactor(int factor)
	{
		_heightFactor = factor;
	}

	/**
	 * Returns the height factor for this node.
	 *
	 * @return the current height factor
	 * @see    #setHeightFactor(int)
	 */
	public int getHeightFactor()
	{
		return _heightFactor;
	}

	/**
	 * Returns <code>null</code> as the current line don't have a selected checkbox image.
	 *
	 * @return <code>null</code>
	 * @see    leon.view.win.swing.component.LyJTreeElement#getSelectedCheckboxImage()
	 */
	public Image getSelectedCheckboxImage()
	{
		return null;
	}

	/**
	 * Returns <code>null</code> as the current line don't have an empty checkbox image.
	 *
	 * @return <code>null</code>
	 * @see    leon.view.win.swing.component.LyJTreeElement#getEmptyCheckboxImage()
	 */
	public Image getEmptyCheckboxImage()
	{
		return null;
	}

	/**
	 * Returns <code>null</code> as the current line don't have an additional images.
	 *
	 * @return <code>null</code>
	 * @see    leon.view.win.swing.component.LyJTreeElement#getAdditionalImages()
	 */
	public Vector<Image> getAdditionalImages()
	{
		return null;
	}

	/**
	 * Returns -1 as the current line don't have validity mechanisms.
	 *
	 * @return -1
	 * @see    leon.view.win.swing.component.LyJTreeElement#getValidityImageIndex()
	 */
	public int getValidityImageIndex()
	{
		return -1;
	}

	/**
	 * Don't do anything. This current line don't have validity mechanisms.
	 *
	 * @param rect Not used in this implementation.
	 * @see   #getValidityRectangle()
	 * @see   leon.view.win.swing.component.LyJTreeElement#setValidityRectangle(java.awt.Rectangle)
	 */
	public void setValidityRectangle(Rectangle rect)
	{
	}

	/**
	 * Returns <code>null</code> as this line don't implement validity mechanisms.
	 *
	 * @return <code>null</code>
	 * @see    #setValidityRectangle(Rectangle)
	 * @see    leon.view.win.swing.component.LyJTreeElement#getValidityRectangle()
	 */
	public Rectangle getValidityRectangle()
	{
		return null;
	}

	/**
	 * Don't do anything. This current line don't implement validity mechanisms.
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#changeValidity()
	 */
	public void changeValidity()
	{
	}

} // end class LySwingLine
