/*
 * LySwingList.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import javax.swing.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the List object (as described in the dtd) for swing viewer.
 *
 * <p>An input field of List type allows to choose one or more values among a discrete list of
 * values. It is possible to specify if the selection is multiple or not as well as the height of
 * the list.</p>
 *
 * @version $Id: LySwingList.java 29817 2011-02-23 18:00:31Z jlsaouli $
 * @author  Lyria
 * @see     LyList
 */
public class LySwingList extends LySwingFieldItem
	implements LySwingComponent,
		LyGraphicalList,
		LySelectionListener,
		LyActionListener,
		LySetEnabledInterface,
		LySetEditableInterface,
		LyGetValueInterface
{
	public static final String __VERSION =
		"$Id: LySwingList.java 29817 2011-02-23 18:00:31Z jlsaouli $";

	/** The view independent list description. */
	protected LyList		   _list;

	/** The swing component used to display this list. */
	protected Component		   _listComp;

	/** The swing scroll pane used to display this list. */
	protected JScrollPane	   _scrollPane;

	/**
	 * Class constructor of a List object for the Swing target.
	 *
	 * @param list a LIST component defined for every type of view targets. This list is initialized
	 *             for the swing view and filled with the default values.
	 */
	public LySwingList(LyList list)
	{
		super(list);
		_list = list;
		_scrollPane = null;

		initialize();
		setContent();
	}

	/**
	 * Builds the component and registers events.
	 */
	public void initialize()
	{
		boolean monoselection = (_list.getSelectionMode() == LyList.SelectionMode.SINGLE);
		Color   color = getColor("LY_TABLE_SELECT_COLOR");
		Color   inactiveColor = getColor("LY_INACTIVE_SELECT_COLOR");

		_listComp = createListComponent(monoselection, color, inactiveColor);

		LySelectionTransmitter selectionTransmitter = ((LyEventTransmitter)_listComp)
				.getSelectionTransmitter();

		if (selectionTransmitter != null)
			selectionTransmitter.addSelectionListener(this);

		LyActionTransmitter actionTransmitter = ((LyEventTransmitter)_listComp)
				.getActionTransmitter();

		if (actionTransmitter != null)
			actionTransmitter.addActionListener(this);
	}

	/**
	 * Creates the component used to display this list.
	 *
	 * @param  monoselection          a boolean that specifies if this list allow mono or multi
	 *                                selection
	 * @param  selectionColor         the background color of selected options
	 * @param  inactiveSelectionColor the color of the inactive selected options. This color is not
	 *                                used in this implementation.
	 * @return the component for this list
	 */
	public Component createListComponent(boolean monoselection, Color selectionColor,
		Color inactiveSelectionColor)
	{
		LyJListComponent listComp = new LyJListComponent();

		listComp.setCellRenderer(new _LyJCellRenderer());
		listComp.setFixedCellWidth(400);

		int lines = _list.getNRows();
		int minLines = 6;

		if (lines < minLines)
			listComp.setVisibleRowCount(minLines);
		else
			listComp.setVisibleRowCount(lines);

		listComp.setBackground(Color.white);
		listComp.setFont(getFont("LY_DEFAULT_FONT"));

		if (monoselection)
			listComp.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		else
			listComp.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);

		if (selectionColor != null)
		{
			listComp.setSelectionBackground(selectionColor);

			// if selection background too dark, selection foreground is white
			int sumColor = selectionColor.getRed() + selectionColor.getGreen()
				+ selectionColor.getBlue();

			if (sumColor < 255)
				listComp.setSelectionForeground(Color.white);
		}

		return listComp;
	} // end method createListComponent

	/**
	 * Fills this list component with the elements in the description.
	 */
	public void fillList()
	{
		LyListOption listOption;
		int			 i, n = _list.getListOptionCount();

		for (i = 0; i < n; i++)
		{
			listOption = _list.getListOption(i);

			((LyJListComponent)_listComp).insertItemAt(listOption, i);

			// Check selection list
			if (listOption.isSelected())
				((LyJListComponent)_listComp).selectItem(listOption, true);
		}
	}

	/**
	 * Selects the specified list elements in this list.
	 *
	 * @param selectedOptions an array containing the LyListOption elements to set selected
	 * @see   #getSelectedObjects(Component)
	 */
	public void selectListOption(Object[] selectedOptions)
	{
		LyListOption listOption;

		int			 i, n = selectedOptions.length;

		for (i = 0; i < n; i++)
		{
			listOption = (LyListOption)selectedOptions[i];
			listOption.setSelected(true);
		}
	}

	/**
	 * Gets the list of selected elements in this list.
	 *
	 * @param  listComponent the list component to get its selected objects
	 * @return an array containing the selected LyListOption
	 * @see    #selectListOption(Object[])
	 * @see    JList#getSelectedValues()
	 */
	public Object[] getSelectedObjects(Component listComponent)
	{
		return ((LyJListComponent)listComponent).getSelectedValues();
	}

	/**
	 * Sets the content of this list.
	 */
	public void setContent()
	{
		fillList();
	}

	/**
	 * Updates view object from the current instance.
	 */
	protected void updateListOptionSelection()
	{
		Object[]     selection = getSelectedObjects(_listComp);
		LyListOption listOption;

		int			 i, n = _list.getListOptionCount();

		// First unselect all the list options
		for (i = 0; i < n; i++)
		{
			listOption = _list.getListOption(i);
			listOption.setSelected(false);
		}

		selectListOption(selection);
	}
	
	/**
	 * Gets the swing component of this list. This component is a {@link JScrollPane} which
	 * preferred size is the same as its parent.
	 *
	 * @return a scrollPane that contain this component's list.
	 */
	public Component getComponent()
	{
		if (_scrollPane == null)
		{
			_scrollPane = new JScrollPane(_listComp,
					JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER)
			{
				private static final long serialVersionUID = 1901136082009981887L;

				@Override
				public Dimension getMinimumSize()
				{
					return new Dimension((int)_listComp.getMinimumSize().getWidth(),
							((LyJListComponent)_listComp).getPreferredScrollableViewportSize().height);
				}

				/*
				 * (non-Javadoc)
				 * @see java.awt.Component#getPreferredSize()
				 */
				public Dimension getPreferredSize()
				{
					return ((LyJListComponent)_listComp).getPreferredScrollableViewportSize();
				}
			};
		}
		
		return _scrollPane;
	}

	/**
	 * Adds a list option at the given position.
	 *
	 * @param item the list option to add
	 * @param pos  the position of the option in this list
	 * @see   #removeListOption(LyListOption)
	 * @see   LyGraphicalList#addListOption(LyListOption, int)
	 */
	public void addListOption(LyListOption item, int pos)
	{
		LyJListComponent comp = (LyJListComponent)_listComp;

		if (pos < 0)
			comp.addItem(item);
		else
			comp.insertItemAt(item, pos);
		
        if (item.isSelected())
              comp.selectItem(item, true);

		comp.invalidate();
		comp.validate();
	}

	/**
	 * Removes the given list option from this list.
	 *
	 * @param listOption the list option to remove
	 * @see   #addListOption(LyListOption, int)
	 * @see   LyGraphicalList#removeListOption(LyListOption)
	 */
	public void removeListOption(LyListOption listOption)
	{
		LyJListComponent comp = (LyJListComponent)_listComp;

		comp.removeItem(listOption);
		comp.doLayout();

	}

	/* LySetEditableInterface */

	/**
	 * Disabling is not implemented on Swing lists.
	 *
	 * @param editable the new status, not used in this implementation
	 * @see   LySetEditableInterface#setEditable(boolean)
	 */
	public void setEditable(boolean editable)
	{
		// XXX : TO DO implement disabling on LyListCanvas
	}

	/* LySetEnabledInterface */

	/**
	 * Sets the enable status of the field item.
	 *
	 * @param enabled The new status.
	 * @see   LySetEnabledInterface#setEnabled(boolean)
	 */
	public void setEnabled(boolean enabled)
	{
		_listComp.setEnabled(enabled);
	}

	/* LySetValueInterface */

	/**
	 * Returns the ids of the elements in this list. These ids are concatenated in a String
	 * separated with {@link LyJListComponent#ITEM_SEP}.
	 *
	 * @return the value of this list component, or <code>null</code> if there is none
	 * @see    leon.view.getinterface.LyGetValueInterface#getValue()
	 */
	public String getValue()
	{
		StringBuffer sb = null;
		Object[]	 selection = getSelectedObjects(_listComp);
		int			 i, n = (selection == null ? 0 : selection.length);
		LyListOption listOption;

		for (i = 0; i < n; i++)
		{
			if (sb == null)
				sb = new StringBuffer();
			else
				sb.append(LyJListComponent.ITEM_SEP);
			
			listOption = (LyListOption)selection[i];
			sb.append(listOption.getId());
		}
		return ((sb == null) ? null : sb.toString());
	}

	/* LySelectionListener */

	/**
	 * Invoked when an selection occurs: the list component is refreshed.
	 *
	 * @see LySelectionListener#selectionPerformed(LySelectionEvent)
	 */
	public void selectionPerformed(LySelectionEvent selectionEvent)
	{
		LyActionEvent event = null;
		String		  val = getValue();

		updateListOptionSelection();

		event = new LyActionEvent(this, null, val, LyActionDemand.TYPE_VALUE_CHANGED);

		if (event != null)
			_actionTransmitter.processActionEvent(event);
	}

	/* LyActionListener */

	/**
	 * (non-Javadoc).
	 *
	 * @see LyActionListener#actionPerformed(LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		if (actionEvent.getType() == LyMenuDemand.TYPE_POPUP)
			return;

		LyActionEvent event = null;
		String		  val = getValue();

		event = new LyActionEvent(this, _list.getParentFormFieldId(), val, actionEvent.getType());

		if (event != null)
			_actionTransmitter.processActionEvent(event);
	}

	/**
	 * This class implements a customized renderer for list cell.
	 */
	public class _LyJCellRenderer extends DefaultListCellRenderer
	{
		private static final long serialVersionUID = -3395577130051019302L;

		/** Image of this cell. */
		Icon					  icon;

		/** Label of this cell. */
		String					  label;

		/**
		 * Constructs a customized renderer for a list cell.
		 */
		public _LyJCellRenderer()
		{
			super();
		}

		/*
		 * (non-Javadoc)
		 * @see javax.swing.ListCellRenderer#getListCellRendererComponent(javax.swing.JList,
		 * java.lang.Object, int, boolean, boolean)
		 */
		public Component getListCellRendererComponent(JList list, Object value, int index,
			boolean isSelected, boolean hasFocus)
		{
			String		 image;
			LyListOption listOption = (LyListOption)value;

			JLabel		 comp = (JLabel)super.getListCellRendererComponent(list, value, index,
					isSelected, hasFocus);

			label = translate(listOption.getLabel());
			comp.setText(label);

			image = listOption.getImage();
			if (image != null)
			{
				icon = new ImageIcon(getImage(image));
				comp.setIcon(icon);
			}
			return comp;
		}
	} // end class _LyJCellRenderer

    /**
     * Selects given list option.
     */
    protected void selectOption(LyListOption listOption, boolean selected)
    {
        ((LyJListComponent)_listComp).selectItem(listOption, selected);
    }
} // end class LySwingList
