/*
 * LySwingMap.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.PageAttributes.*;
import java.text.*;
import java.util.*;
import javax.swing.*;
import leon.misc.*;
import leon.view.*;
import leon.view.LyMapBase.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.print.*;
import leon.view.setinterface.*;
import leon.view.win.swing.map.*;


/**
 * This class implements the Map object (as described in the dtd) for swing viewer.
 *
 * <p>A map allows to display any type of map view: maps representing geographic or synoptic areas,
 * displaying of elements with different types of positioning, displaying of nodes, links, images,
 * etc.</p>
 *
 * <p>There are four types of elements that can be placed on a map: the nodes, the links, the
 * regions and the images. The elements placed on a map can be sensitive or not (except the images
 * that can never be sensitive). It is possible to specify for the nodes and the links the layer on
 * which they are. This layer allows to place the elements of a higher layer over the elements of a
 * lower layer (for example, the elements of the layer 2 are placed over the elements of the layer
 * 1).</p>
 *
 * <p>The nodes allow to define elements that can be placed at any location on the map and that can
 * be moved. By using a special positioning manager, it is possible to place these nodes according
 * to a predefined geometry (matrix, circle, ellipse, line,...). When zooming on the map, a node
 * keeps always the same aspect unlike an element of image type.</p>
 *
 * <p>The links allow to define segments between two nodes of the map. If one of the two nodes
 * defining the link is moved, the link follows the moved node so that it always starts from its
 * source node and that it always arrives to its destination node.</p>
 *
 * <p>The regions allow to represent geometric areas defined by a discrete set of points. These
 * areas can be sensitive. It is notably possible to lead to other map views when launching an
 * action over one of these regions.</p>
 *
 * <p>It is possible to define the background that will be used for the map (defined by an image),
 * the selection mode (none, simple or multiple), the visualization mode (through a viewer or with
 * scroll bars) and to specify if the map is in two or three dimensions.</p>
 *
 * <p>The possible movements on a map are translations, zooms (forward/backward) and rotations. A
 * map in three dimensions allows spatial rotations enabling to display the regions in three
 * dimensions. A map in two dimensions allows the axial rotations around the center of the map.</p>
 *
 * @version $Id: LySwingMap.java 31811 2011-07-28 12:41:24Z jlsaouli $
 * @author  Lyria
 * @see     LyMap
 */
public class LySwingMap extends LySwingPanel
	implements LySwingComponent,
		LyGraphicalMap,
		LyActionListener,
		LySelectionListener,
		LySetIdleInterface,
		LyResetContentInterface,
		LyPrintInterface,
		LySetVisibleInterface,
		LySetViewerInterface,
		LyGetLockInterface
{
	public static final String __VERSION =
		"$Id: LySwingMap.java 31811 2011-07-28 12:41:24Z jlsaouli $";
	
	/** The class used as the panel component. */
	public static Class<?>		 PANEL_CLASS = null;
	
	/** The class used as the map manager. */
	public static Class<?>		 MANAGER_CLASS = null;

	/** The view independent map description. */
	protected LyMap			   _map;

	/** The view component that contain the elements of this map. */
	protected LyMapContainer   _mapbox;

	/** The layout manager to arrange the node positions in this map. */
	protected LyLayout		   _layout;

    /** The constant used to represent the key IS_SHIFT_KEY_DOWN **/
    public final static String IS_SHIFT_KEY_DOWN = "IS_SHIFT_KEY_DOWN";
    
    /** The constant used to represent the key IS_CONTROL_KEY_DOWN **/
    public final static String IS_CONTROL_KEY_DOWN = "IS_CONTROL_KEY_DOWN";

	/**
	 * Class constructor of a Map object for the Swing target.
	 *
	 * @param map a MAP component defined for every type of view targets. This map is initialized
	 *            for the swing view and filled with the default values.
	 */
	public LySwingMap(LyMap map)
	{
		super(map);
		_map = map;

		initialize();
		setContent();
	}

	/**
	 * Initializes the default swing values of this component and registers its listeners.
	 */
	public void initialize()
	{
		_mapbox = createMapContainer();

		// register for events
		LyActionTransmitter    actTransmitter = _mapbox.getManager().getActionTransmitter();
		LySelectionTransmitter selTransmitter = _mapbox.getManager().getSelectionTransmitter();

		if (actTransmitter != null)
			actTransmitter.addActionListener(this);

		if (selTransmitter != null)
			selTransmitter.addSelectionListener(this);

		LyMapComponent canvas = _mapbox.getMapComponent();

		if (_map.getBackground() != null)
			canvas.getManager().setBackground(getImage(_map.getBackground()));
		else
		{
			Color col = getColor("LY_MAP_BACKGROUND");

			if (col == null)
				col = getColor("LY_DEFAULT_BACKGROUND");

			if (col != null)
				((Component)canvas).setBackground(col);
		}

		if (_map.getLabel() != null)
			canvas.getManager().setTitle(translate(_map.getLabel()));

		canvas.getManager().setSelectionColor(getColor("LY_MAP_SELECTED_COLOR"));
	} // end method initialize

	/**
	 * Sets sub-component mapContent.
	 *
	 * @param mapContent New value of mapContent
	 */
	public void setMapContent(LyMapContent mapContent)
	{
		setContent();
	}

	/**
	 * Sets the content of this component. Creates the layout manager then adds the map elements to
	 * the map container.
	 */
	public void setContent()
	{
		synchronized (getLock())
		{
			// Building the layout manager.
			// We do not change current idle status if already set
			boolean wasIdle = _mapbox.getManager().isIdle();
	
			if (!wasIdle)
				setIdle(true);
	
			String center = null;
	
			if (_map.getLayout() != null)
			{
				if (_layout == null)
					setLayout(_map.getLayout());
	
				center = _map.getLayout().getCenter();
	
			}
	
			// Reset the elements of the mapbox
			_mapbox.getManager().resetElements();
	
			// Add the elements to the map
			LyComponent  component;
			LyMapContent content = _map.getMapContent();
			// Retrieve the mapContent view (in order to instanciate it)
			Vector<LyComponent> components = new Vector<LyComponent>();
			int			 i, n;
	
			if (content != null)
			{
				n = content.getComponentCount();
				for (i = 0; i < n; i++)
					components.addElement(content.getComponent(i));
			}
	
			n = _map.getComponentCount();
			for (i = 0; i < n; i++)
				components.addElement(_map.getComponent(i));
	
			n = components.size();
			for (i = 0; i < n; i++)
			{
				component = components.elementAt(i);
				addComponent(component, -1, center);
			}
	
			if (_map.getBackground() != null)
				_mapbox.getMapComponent().getManager().setBackground(getImage(
						_map.getBackground()));
	
			if (!wasIdle)
				setIdle(false);
		}
	} // end method setContent

	/**
	 * Creates a generic container for this map. The default container is a {@link LyMapScrollPane}
	 * object.
	 *
	 * @return the new map container
	 */
	public LyMapContainer createMapContainer()
	{
		BitSet bs = new BitSet();

		if (_map.isSpatial())
			bs.set(LyAbstractMapComponentManager.SPATIAL);
		if (_map.isViewer())
			bs.set(LyAbstractMapComponentManager.VIEWER);
		if (LyMapBase.ViewerStatus.SHOW.equals(_map.getViewerStatus()))
			bs.set(LyAbstractMapComponentManager.VIEWER_STATUS);

		/*
		 * if ("NONE".equalsIgnoreCase(_selectionMode)) selectionMode = LyMapCanvas.NO_SELECTION;
		 * else if ("SINGLE".equalsIgnoreCase(_selectionMode)) selectionMode =
		 * LyMapCanvas.SINGLE_SELECTION; else selectionMode = LyMapCanvas.MULTIPLE_SELECTION;
		 */
		LyMapScrollPane	mapContainer = new LyMapScrollPane(bs, _map.getSelectionMode(), this);
		
		if (!_map.hasBorder())
			mapContainer.setBorder(null);
		return mapContainer;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyFreeInterface#free()
	 */
	public void free()
	{
		if (_mapbox != null)
			_mapbox.getManager().free();

		_mapbox = null;
		_map = null;
		_layout = null;
	}

	/**
	 * Adds a map component in this map at the specified position.
	 *
	 * @param component the new component to add
	 * @param pos       the position of the component in this map
	 * @see   #addComponent(LyComponent, int, String)
	 * @see   leon.view.graphic.LyGraphicalMap#addComponent(leon.view.LyComponent, int)
	 */
	public void addComponent(LyComponent component, int pos)
	{
		addComponent(component, pos, null);
	}

	/**
	 * Adds the specified component at the given position in this box.
	 *
	 * @param component the component to add
	 * @param pos       the position of the component in this box
	 * @param center    center property of the box
	 */
	protected void addComponent(LyComponent component, int pos, String center)
	{
		// XXXX pos
		String		 name, id;
		LyMapElement element;

		if (component == null)
			return;

		element = (LyMapElement)component.getInternalView();
		name = null;
		if ((center != null) && (component instanceof LyGetIdInterface))
		{
			id = ((LyGetIdInterface)component).getId();
			if ((id != null) && (id.equals(center)))
				name = "Center";
		}

		if (element instanceof LySwingMapLink)
		{
			// We must search for extremities.
			String		   nodeId;
			LySwingMapNode node;

			nodeId = ((LyMapLink)component).getStartNode();
			node = getNodeById(nodeId);
			((LySwingMapLink)element).setStartNode(node);

			nodeId = ((LyMapLink)component).getEndNode();
			node = getNodeById(nodeId);
			((LySwingMapLink)element).setEndNode(node);
		}
		else if (element instanceof LySwingMapNode)
		{
			// Retrieve the parent node (if any)
			String		   nodeId;
			LySwingMapNode node;

			nodeId = ((LyMapNode)component).getParentNode();
			node = getNodeById(nodeId);
			((LySwingMapNode)element).setParentNode(node);
			// affect the attribute 'child' of his parent node
			if (node != null)
				node.setChild((LySwingMapNode)element);

		}

		_mapbox.getManager().addElement(name, element, pos);
	} // end method addComponent

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalMap#removeComponent(leon.view.LyComponent)
	 */
	public void removeComponent(LyComponent component)
	{
		if (component == null)
			return;

		LyMapElement element = (LyMapElement)component.getInternalView();

		_mapbox.getManager().removeElement(element);
	}

	/**
	 * Sets the spatial mode.
	 *
	 * @param spatial Indicates if the map manager is spatial or not.
	 * @see   LyAbstractMapComponentManager#SPATIAL
	 * @see   LyGraphicalMap#setSpatial(boolean)
	 */
	public void setSpatial(boolean spatial)
	{
		_mapbox.getManager().setSpatial(spatial);
	}

	/**
	 * Sets the viewer mode.
	 *
	 * @param viewer Indicates if the map manager must show the viewer or not.
	 * @see   LyAbstractMapComponentManager#VIEWER
	 * @see   LyGraphicalMap#setViewer(boolean)
	 */
	public void setViewer(boolean viewer)
	{
		_mapbox.getManager().setViewer(viewer);
	}

	/**
	 * (non-Javadoc).
	 * 
	 * @see leon.view.graphic.LyGraphicalMap#setViewerMode(leon.view.LyMapBase.ViewerMode)
	 */
	public void setViewerStatus(LyMapBase.ViewerStatus viewerStatus)
	{
		_mapbox.getManager().setViewerStatus(viewerStatus);
	}

	/**
	 * Get a lock on the table to prevent concurrent access on rows.
	 * 
	 * @return a valid lock for synchronized()
	 */
	public Object getLock()
	{
		Object lock;
	
		if (_mapbox instanceof Component)
			return ((Component)_mapbox).getTreeLock();
		else
			lock = this;
			
		return lock;
	}

	/**
	 * Sets the Layout of the map.
	 *
	 * @param layout the new layout to use
	 * @see   LyGraphicalMap#setLayout(LyLayout)
	 */
	public void setLayout(LyLayout layout)
	{
		if ((layout != null) && (_layout != layout))
		{
			_layout = layout;

			LySwingLayout    internal;
			LyLayoutManager  layoutMgr;
			LyExternalLayout externalLayout;

			internal = (LySwingLayout)layout.getInternalView();

			// if there is an external layout, the class java is recovered
			if (layout.getJavaName() != null)
			{
				try
				{
					Class<?> beanClass;

					beanClass = Class.forName(layout.getJavaName());
					externalLayout = (LyExternalLayout)beanClass.newInstance();
					_mapbox.getMapComponent().getManager().setExternalLayout(externalLayout);
					if (layout.getStyle() != LyLayout.Style.COORDINATES)
						layout.setStyle(LyLayout.Style.GEOGRAPHIC);
				}
				catch (Exception e)
				{
					getEnvironment().writeException(e);
				}
			}
			boolean isModelisation = (layout.getStyle() == LyLayout.Style.LINEAR);

			layoutMgr = internal.layoutFactory(this);
			_mapbox.getMapComponent().getManager().setLayoutManager(layoutMgr);

			if (isModelisation)
				_map.setSelectionDraw(LyMapBase.SelectionDraw.BORDER);

			resetContent();
		}
	} // end method setLayout

	/**
	 * Sets the title of this map.
	 *
	 * @param label the new title of the map
	 * @see   LyAbstractMapContainerManager#setTitle(String)
	 * @see   leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
	 */
	public void setLabel(String label)
	{
		_mapbox.getManager().setTitle(translate(label));
	}

	/**
	 * Invoked when an action occurs. The action demands that trigger the processing of the event
	 * are the following:
	 *
	 * <ul>
	 *   <li>{@link LyActionDemand#TYPE_ACTION}</li>
	 *   <li>{@link LyMenuDemand#TYPE_POPUP}</li>
	 *   <li>{@link LyActionDemand#TYPE_MOVE}</li>
	 *   <li>{@link LyActionDemand#TYPE_DELETE}</li>
	 *   <li>and any actions from {@link LyDragDropDemand#TYPE_DRAG_START} to {@link
	 *     LyDragDropDemand#TYPE_DRAG_OVER}</li>
	 * </ul>
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		actionEvent.setSourceId(_map.getId());

		String controller = getController();

		if (controller != null)
		{
			// Notify controller
			LyDemand demand = null;

			if (actionEvent.getType() == LyActionDemand.TYPE_ACTION)
			{
				demand = LySwingArea.buildActionDemand(actionEvent);

                int[] params = (int[])actionEvent.getParameter();
                //Set 0 and 1 indexes to cursor position values.
                demand.addIndex(params[0]);
                demand.addIndex(params[1]);
                //Other parameters could be added if they exist.
                if(params.length > 2) {
                  //Verify that the 3rd parameter exists in parameter list
                  if(params.length >=2) {
                    demand.setAttribute(IS_SHIFT_KEY_DOWN, new Integer(params[2]));
                  }
                  //Verify that the 3rd parameter exists in parameter list
                  if(params.length >=3) {
                    demand.setAttribute(IS_CONTROL_KEY_DOWN, new Integer(params[3]));
                  }
                }
			}
			else if (actionEvent.getType() == LyMenuDemand.TYPE_POPUP)
			{
				demand = new LyMenuDemand(LyMenuDemand.TYPE_POPUP, _map.getId());

				((LyMenuDemand)demand).setItem(actionEvent.getValue());
				((LyMenuDemand)demand).setPosition((int[])actionEvent.getParameter());
			}
			else if (actionEvent.getType() == LyActionDemand.TYPE_MOVE)
				demand = new LyActionDemand(LyActionDemand.TYPE_MOVE, _map.getId());
			// We do not need to indicated the translated nodes : translated
			// nodes are selected nodes.
			else if (actionEvent.getType() == LyActionDemand.TYPE_DELETE)
				demand = new LyActionDemand(LyActionDemand.TYPE_DELETE, _map.getId());
			// We do not need to indicated the deleted nodes : deleted
			// nodes are selected nodes.
			else if ((actionEvent.getType() >= LyDragDropDemand.TYPE_DRAG_START)
					&& (actionEvent.getType() <= LyDragDropDemand.TYPE_DRAG_OVER))
			{
				String elementId = actionEvent.getValue();

				demand = new LyDragDropDemand(actionEvent.getType(), elementId);
				((LyDragDropDemand)demand).setValue(actionEvent.getValue());
				((LyDragDropDemand)demand).setPosition((int[])actionEvent.getParameter());
			}
			else if(actionEvent.getType() == LyActionDemand.TYPE_EXPAND)
			{
			  demand = new LyActionDemand(LyActionDemand.TYPE_EXPAND, _map.getId());
			}
			processDemand(demand);
			actionEvent.consume();
		}
	} // end method actionPerformed

	/**
	 * Invoked when a selection occurs.
	 *
	 * @see LySelectionListener#selectionPerformed(LySelectionEvent)
	 */
	public void selectionPerformed(LySelectionEvent selectionEvent)
	{
		if (getController() != null)
		{
			int    type_sel = selectionEvent.isSelected() ? LySelectionDemand.TYPE_SELECT
														  : LySelectionDemand.TYPE_UNSELECT;

			Object source = selectionEvent.getSource();
			String source_id;

			if (source instanceof LyGetIdInterface)
				source_id = ((LyGetIdInterface)source).getId();
			else
				source_id = source.toString();

			// Notify controller
			LySelectionDemand demand = new LySelectionDemand(type_sel, source_id);
			Object			  item = selectionEvent.getItem();

			if (item instanceof LyGetIdInterface)
				demand.addSelectionId(((LyGetIdInterface)item).getId());
			else if (item instanceof Vector<?>)
			{
				Vector<?> list = (Vector<?>)item;

				for (int i = 0; i < list.size(); i++)
				{
					Object o = list.elementAt(i);

					if (o instanceof LyGetIdInterface)
						demand.addSelectionId(((LyGetIdInterface)o).getId());
				}
			}

			demand.setExtend(selectionEvent.isExtend());

			processDemand(demand);
		}
	} // end method selectionPerformed

	/**
	 * Invoked when a demand occurs.
	 *
	 * @param demand the demand to process
	 * @see   leon.view.event.LyDemandListener#demandPerformed(LyDemand)
	 */
	public void demandPerformed(LyDemand demand)
	{
		processDemand(demand);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetIdleInterface#setIdle(boolean)
	 */
	public void setIdle(boolean status)
	{
		_mapbox.getManager().setIdle(status);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetVisibleInterface#setVisible(boolean)
	 */
	public void setVisible(boolean status)
	{
		if (_mapbox != null)
			_mapbox.getManager().setVisible(status);
	}

	
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyResetContentInterface#resetContent()
	 */
	public void resetContent()
	{
		resetContent(false);
	}
	
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyResetContentInterface#resetContent()
	 */
	public void resetContent(boolean full)
	{
		_mapbox.getManager().enableXRotation(true);
		_mapbox.getManager().enableYRotation(true);
		_mapbox.getManager().enableZRotation(true);

		setContent();
		_mapbox.getManager().resetView();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return (Component)_mapbox;
	}

	/**
	 * Returns the component used to display popup menus.
	 *
	 * @return the parent component of popup menus.
	 */
	protected JComponent getPopupMenuParent()
	{
		return (JComponent)_mapbox;
	}

	/**
	 * Gets the map node with the specified Id. Returns <code>null</code> if the node is not found.
	 *
	 * @param  nodeId the Id of the node to retreive
	 * @return the map node with the given Id
	 */
	public LySwingMapNode getNodeById(String nodeId)
	{
		int		    i, n = _map.getComponentCount();
		LyComponent element;

		for (i = 0; i < n; i++)
		{
			element = _map.getComponent(i);

			if ((element instanceof LyMapNode) && (((LyMapNode)element).getId().equals(nodeId)))
				return (LySwingMapNode)element.getInternalView();
		}

		// If node not found in the map, check if in the map content
		LyMapContent content = _map.getMapContent();

		if (content != null)
		{
			n = content.getComponentCount();

			for (i = 0; i < n; i++)
			{
				element = content.getComponent(i);

				if ((element instanceof LyMapNode)
						&& (((LyMapNode)element).getId().equals(nodeId)))
					return (LySwingMapNode)element.getInternalView();
			}
		}

		return null;
	} // end method getNodeById

	/**
	 * Gets the controller associated with this map. If not found, try to get parent controller.
	 *
	 * @return the map controller id
	 * @see    leon.view.graphic.LyGraphicalObject#getController()
	 */
	public String getController()
	{
		String controller = _map.getController();

		if ((controller == null) && (_map.getParent() instanceof LyGetControllerInterface))
			controller = ((LyGetControllerInterface)_map.getParent()).getController();

		return controller;
	}

	/**
	 * Initializes the printing process and starts printing the content of the map on the default
	 * printer.
	 *
	 * @param listener the printing listener
	 * @see   LyPrintInterface#print(LyPrintListenerInterface)
	 */
	public void print(LyPrintListenerInterface listener)
	{
		PrintJob		   job = null;
		LyEnvironment	   environment = getEnvironment();
		LySwingViewManager viewManager = (LySwingViewManager)environment.getViewManager();
		PageAttributes	   pageAttributes = new PageAttributes();
		
		pageAttributes.setOrientationRequested(OrientationRequestedType.LANDSCAPE);
		job = viewManager.getPrintJob(null, _map.getLabel(), pageAttributes);

		if (job == null)
			return;

		printEnd(job, _mapbox.getMapComponent(), listener);

		job.end();
	}

	/**
	 * Prints the content of the map with the specified printer job.
	 *
	 * @param job      the printer handler
	 * @param canvas   the drawing area
	 * @param listener the print listener, not used in this implementation
	 */
	public void printEnd(PrintJob job, LyMapComponent canvas, LyPrintListenerInterface listener)
	{
		LyEnvironment environment = getEnvironment();
		Graphics	  g = job.getGraphics();
		Graphics	  pg;
		Dimension     jd = job.getPageDimension();
		String		  title = _map.getLabel();
		String		  win_logo = null; // XXX window.getLogoPath();
		String		  client_logo = "LY_COMPANY_LOGO";
		String		  date = DateFormat.getDateTimeInstance(DateFormat.FULL, DateFormat.MEDIUM,
				environment.getLocale()).format(new Date());
		int			  w = jd.width - LyPrintManager.LEFT_MARGIN - LyPrintManager.RIGHT_MARGIN;
		int			  hh, fh;
		boolean		  hasViewer = _mapbox.getManager().isViewer();
		boolean		  isDrawViewer = canvas.getManager().isDrawViewer();
		Dimension     oldDim = ((Component)canvas).getSize();

		// XX Attempt to find additional information : IMPROVE !
		if (win_logo == null)
		{
			LyComponent parent = _map.getParent();

			if (parent instanceof LyArea)
			{
				LyTitleBar titleBar = ((LyArea)parent).getTitleBar();

				if (titleBar != null)
				{
					win_logo = titleBar.getImage();
					title = titleBar.getLeftLabel();

					if ((title == null) || (title.length() <= 0))
						title = titleBar.getCenterLabel();

					if ((title == null) || (title.length() <= 0))
						title = titleBar.getRightLabel();

					title = translate(title);
				}
			}
		}

		// Draw header
		hh = LySwingPrintManager.printHeader((Component)_mapbox, g, jd, win_logo, title,
				client_logo, environment);

		// Draw footer
		fh = LySwingPrintManager.printFooter((Component)_mapbox, g, jd, date, null, null,
				environment);

		pg = g.create(LyPrintManager.LEFT_MARGIN, hh, w, jd.height - hh - fh);

		float heightMax = new Integer(jd.height - hh - fh).floatValue();
		
		// Don't draw the viewer
		canvas.getManager().setIdle(true);
		_mapbox.getManager().setViewer(false);
		canvas.getManager().drawViewer(false);

		if ((_map.getLayout() != null)
				&& (_map.getLayout().getStyle() == LyLayout.Style.COORDINATES))
		{
			Rectangle	bounds = canvas.getManager().getBoundingRectangle();
			Image		image = canvas.getManager().getComponent().createImage(bounds.width, bounds.height);
			Graphics	gImage = image.getGraphics();
			float		xRate = bounds.width / w;
			float		yRate = bounds.height / heightMax;
			
			if ((xRate > 1) && (xRate > yRate))
			{
				bounds.width = (int)(bounds.width / xRate);
				bounds.height = (int)(bounds.height / xRate);
			}
			else if ((yRate > 1) && (yRate > xRate))
			{
				bounds.width = (int)(bounds.width / yRate);
				bounds.height = (int)(bounds.height / yRate);
			}
			
			canvas.getManager().paint(gImage, true);
			pg.drawImage(image, (w - bounds.width) / 2, (int)((heightMax - bounds.height) / 2), bounds.width, bounds.height, canvas.getManager().getComponent());
			gImage.dispose();
		}
		else
		{
			// Calculate the dimension of the printed graphic : dilate with the same rate for height and
			// width
			float height = new Integer(jd.height).floatValue();
			float width = new Integer(jd.width).floatValue();
			float widthMax = new Integer(w ).floatValue();
			float rate = 1;

			if ((height != 0) && (width != 0))
			{
				if (((heightMax / height) > (widthMax / width)) && ((widthMax / width) > 0))
					rate = widthMax / width;
				else if (((widthMax / width) > (heightMax / height)) && ((heightMax / height) > 0))
					rate = heightMax / height;
			}
			
			int   		newHeight = new Float(height * rate).intValue();
			Dimension 	oldViewDim = canvas.getManager().getViewSize();

			((Component)canvas).setSize(new Dimension(w, newHeight));
			canvas.getManager().setViewSize(new Dimension(w, newHeight));
			canvas.getManager().paint(pg, true);

			((Component)canvas).setSize(oldDim);
			canvas.getManager().setViewSize(oldViewDim);
		}
		
		// Reset the viewer mode
		_mapbox.getManager().setViewer(hasViewer);
		canvas.getManager().drawViewer(isDrawViewer);
		canvas.getManager().setIdle(false);
		
		g.dispose();
	} // end method printEnd

	public void setTranslation(boolean translation)
	{
	}

	public void setTranslationStatus(TranslationStatus translationStatus)
	{
	}

	public void setZoom(boolean zoom)
	{
	}

	public void setZoomFactor(float zoomFactor)
	{
	}

	public void setZoomStatus(ZoomStatus zoomStatus)
	{
	}
} // end class LySwingMap
