/*
 * LySwingMapLink.java
 * Copyright (c) 2008, Lyria.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;

import javax.swing.*;

import leon.misc.*;
import leon.view.*;
import leon.view.LyMapLink.*;
import leon.view.LyMapNode.Style;
import leon.view.common.*;
import leon.view.graphic.*;
import leon.view.win.swing.map.*;


/**
 * This class implements the Map link object (as described in the dtd) for win viewer.
 *
 * @version $Id: LySwingMapLink.java 36510 2012-08-17 15:43:06Z jlsaouli $
 * @author  Lyria
 * @see     leon.view.LyMapContent
 */
public class LySwingMapLink extends LySwingMapObject
implements LyGraphicalMapLink,
LyMapElement
{
	public static final String __VERSION =
			"$Id: LySwingMapLink.java 36510 2012-08-17 15:43:06Z jlsaouli $";

	/** Offset between the link and the selection area. */
	public static final int    SELECTION_OFFSET = 6;

	/** Enum with label alignment **/
	private enum LabelAlignment { START, CENTER, END }

	/** Step value used to calculate each point of loop link  */
	private static final int LOOP_STEP = 50;

	/** Index of a line starting point in the result of the
	 *  computeDrawingLineExtremePoints method call.
	 */
	private static final int START_POINT_IDX = 0;

	/** Index of a line ending point in the result of the
	 *  computeDrawingLineExtremePoints method call.
	 */
	private static final int END_POINT_IDX = 1;

	private enum RectangleSide {EAST, SOUTH, WEST, NORTH}; 

	/** Value used to displace multiple links on same node */
	public static int		   DISPLACEMENT_VALUE = 2;

	/** Value of the angle drawn in an auto link */
	public static int		   AUTO_LINK_ANGLE = 240;

	/** Default size of the image node. */
	private static int        NOT_INITIALIZED = -1;

	/** The view independent map link description. */
	protected LyMapLink		   _mapLink;

	/** Start node. */
	protected LySwingMapNode   _startNode;

	/** End node. */
	protected LySwingMapNode   _endNode;

	/** Local cache for image. */
	/** Cached image for the start node. */
	protected Image			   _imageStart;

	/** Cached image name for the start node. */
	protected String		   _imageStartString;

	/** Cached image for the node. */
	protected Image			   _imageMiddle;

	/** Cached image name for the node. */
	protected String		   _imageMiddleString;

	/** Cached image for the end node. */
	protected Image			   _imageEnd;

	/** Cached image name for the end node. */
	protected String		   _imageEndString;

	/** Local cache for colors. */
	/** Cached color for the start node. */
	protected Color			   _colorStart;

	/** Cached color name for the start node. */
	protected String		   _colorStartString;

	/** Cached color for the node. */
	protected Color			   _color;

	/** Cached color name for the node. */
	protected String		   _colorString;

	/** Cached color for the end node. */
	protected Color			   _colorEnd;

	/** Cached color name for the end node. */
	protected String		   _colorEndString;

	/** Cached selected color for the node. */
	protected Color			   _selectedColor;

	/** Cached inactive selected color for the node. */
	protected Color			   _inactiveSelectedColor;

	/** Cached font color for the node. */
	protected Color			   _fontColor;

	/** Cached outline color for the node. */
	protected Color			   _outlineColor;

	/** Font of the node. */
	protected Font			   _graphicalFont;

	/** Cached translated label that can be accessed many times. */
	protected String		   _labelCache;

	/** Width of the label (for optimization). */
	protected int			   _labelWidth = NOT_INITIALIZED;

	/** Height of the label (for optimization). */
	protected int			   _labelHeight = NOT_INITIALIZED;

	/** Cached translated label that can be accessed many times. */
	protected String		   _endLabelCache;

	/** Width of the label (for optimization). */
	protected int			   _endLabelWidth = NOT_INITIALIZED;

	/** Height of the label (for optimization). */
	protected int			   _endLabelHeight = NOT_INITIALIZED;

	/** Cached translated label that can be accessed many times. */
	protected String		   _startLabelCache;

	/** Width of the label (for optimization). */
	protected int			   _startLabelWidth = NOT_INITIALIZED;

	/** Height of the label (for optimization). */
	protected int			   _startLabelHeight = NOT_INITIALIZED;

	/** Positions of the images of the nodes in the link. */
	protected Point[]		   _imagePos;

	/** Cache the list of links that has the same starting and ending node as the current link */
	private java.util.List<LySwingMapLink> _sameLinks = null;

	/**
	 * Class constructor of a Map link object for the Swing target.
	 *
	 * @param mapLink a MAP_LINK component defined for every type of view targets. This map link is
	 *                initialized for the swing view.
	 */
	public LySwingMapLink(LyMapLink mapLink)
	{
		super(mapLink);
		_mapLink = mapLink;

		initialize();
	}

	/**
	 * Initializes the class fields with default values.
	 */
	public void initialize()
	{
		// Get images String
		_imageStartString = null;
		_imageMiddleString = null;
		_imageEndString = null;
		_imageStart = null;
		_imageMiddle = null;
		_imageEnd = null;

		_colorStart = null;
		_colorStartString = null;
		_color = null;
		_colorString = null;
		_colorEnd = null;
		_colorEndString = null;

		_selectedColor = null;
		_inactiveSelectedColor = null;
		_fontColor = null;

		_labelCache = null;
		_labelWidth =  NOT_INITIALIZED;
		_labelHeight = NOT_INITIALIZED;

		_startLabelCache = null;
		_startLabelWidth = NOT_INITIALIZED;
		_startLabelHeight = NOT_INITIALIZED;

		_endLabelCache = null;
		_endLabelWidth  = NOT_INITIALIZED;
		_endLabelHeight  = NOT_INITIALIZED;
		getImages();

		_graphicalFont = getFont("LY_MAPNODE_FONT");
	} // end method initialize

	/**
	 * Returns the images of the link.
	 *
	 * @return the list of images used in the link
	 * @see    LyMapElement#getImages()
	 */
	public Image[] getImages()
	{
		Image imageStart = null;
		Image imageMiddle = null;
		Image imageEnd = null;

		if (_mapLink.getImageStart() != null)
		{
			if (_imageStartString != null) {
				if (_imageStartString.equals(_mapLink.getImageStart())) {
					imageStart = _imageStart;
				}
				else
				{
					_imageStartString = _mapLink.getImageStart();
					_imageStart = getImage(_mapLink.getImageStart());
					imageStart = _imageStart;
				}
			}
			else
			{
				_imageStartString = _mapLink.getImageStart();
				_imageStart = getImage(_mapLink.getImageStart());
				imageStart = _imageStart;
			}
		}

		if (_mapLink.getImageMiddle() != null)
		{
			if (_imageMiddleString != null) {
				if (_imageMiddleString.equals(_mapLink.getImageMiddle()))
					imageMiddle = _imageMiddle;
				else
				{
					_imageMiddleString = _mapLink.getImageMiddle();
					_imageMiddle = getImage(_mapLink.getImageMiddle());
					imageMiddle = _imageMiddle;
				}
			}
			else
			{
				_imageMiddleString = _mapLink.getImageMiddle();
				_imageMiddle = getImage(_mapLink.getImageMiddle());
				imageMiddle = _imageMiddle;
			}
		}

		if (_mapLink.getImageEnd() != null)
		{
			if (_imageEndString != null) {
				if (_imageEndString.equals(_mapLink.getImageEnd()))
					imageEnd = _imageEnd;
				else
				{
					_imageEndString = _mapLink.getImageEnd();
					_imageEnd = getImage(_mapLink.getImageEnd());
					imageEnd = _imageEnd;
				}
			}
			else
			{
				_imageEndString = _mapLink.getImageEnd();
				_imageEnd = getImage(_mapLink.getImageEnd());
				imageEnd = _imageEnd;
			}
		}

		return (new Image[] { imageStart, imageMiddle, imageEnd });
	} // end method getImages

	/**
	 * Gets the start color of this map link.
	 *
	 * @param  environment the environment
	 * @return the starting color of the link
	 */
	public Color getColorStart(LyEnvironment environment)
	{
		String mapLinkColorStart = _mapLink.getColorStart();

		if (mapLinkColorStart != null)
		{
			if (_colorStartString != null)
			{
				if (_colorStartString.equals(mapLinkColorStart))
					return _colorStart;
				_colorStartString = mapLinkColorStart;
				_colorStart = getColor(mapLinkColorStart);
			}
			else
			{
				_colorStartString = mapLinkColorStart;
				_colorStart = getColor(mapLinkColorStart);
			}
		}
		else
		{
			if (_colorStartString != null)
				return _colorStart;
			_colorStart = getColor(environment);
			_colorStartString = _colorString;
		}
		return _colorStart;
	} // end method getColorStart

	/**
	 * Gets the color of this map link.
	 *
	 * @param  environment the environment, not used in this implementation
	 * @return the color of the link
	 */
	public Color getColor(LyEnvironment environment)
	{
		String colorParam = _mapLink.getColor();

		if (colorParam != null)
		{
			if (_colorString != null)
			{
				if (_colorString.equals(colorParam))
					return _color;
				_colorString = colorParam;
				_color = getColor(colorParam);
			}
			else
			{
				_colorString = colorParam;
				_color = getColor(colorParam);
			}
		}
		else
		{
			if (_colorString != null)
				return _color;
			_colorString = "LY_MAP_DEFAULT_LINK_COLOR";
			_color = getColor("LY_MAP_DEFAULT_LINK_COLOR");
		}

		// Use default.
		if (_color == null)
			_color = getColor("LY_MAP_DEFAULT_LINK_COLOR");

		return _color;
	} // end method getColor

	/**
	 * Gets the end color of this map link.
	 *
	 * @param  environment the environment
	 * @return the ending color of the link
	 */
	public Color getColorEnd(LyEnvironment environment)
	{
		String mapLinkColorEnd = _mapLink.getColorEnd();

		if (mapLinkColorEnd != null)
		{
			if (_colorEndString != null)
			{
				if (_colorEndString.equals(mapLinkColorEnd))
					return _colorEnd;
				_colorEndString = mapLinkColorEnd;
				_colorEnd = getColor(mapLinkColorEnd);
			}
			else
			{
				_colorEndString = mapLinkColorEnd;
				_colorEnd = getColor(mapLinkColorEnd);
			}
		}
		else
		{
			if (_colorEndString != null)
				return _colorEnd;
			_colorEnd = getColor(environment);
			_colorEndString = _colorString;
		}
		return _colorEnd;
	} // end method getColorEnd

	/**
	 * Gets the selected color of this map link.
	 *
	 * @param  environment the environment, not used in this implementation
	 * @return the selected color of the link
	 */
	public Color getSelectedColor(LyEnvironment environment)
	{
		if (_selectedColor != null)
			return _selectedColor;
		_selectedColor = getColor("LY_MAP_SELECTED_COLOR");
		return _selectedColor;
	}

	/**
	 * Gets the inactive selected color of this map link.
	 *
	 * @param  environment the environment, not used in this implementation
	 * @return the inactive selected color of the link
	 */
	public Color getInactiveSelectedColor(LyEnvironment environment)
	{
		if (_inactiveSelectedColor != null)
			return _inactiveSelectedColor;
		_inactiveSelectedColor = getColor("LY_INACTIVE_SELECT_COLOR");
		return _inactiveSelectedColor;
	}

	/**
	 * Gets the font color of this map link.
	 *
	 * @param  environment the environment, not used in this implementation
	 * @return the font color of the link
	 */
	public Color getFontColor(LyEnvironment environment)
	{
		if (_fontColor != null)
			return _fontColor;
		_fontColor = getColor("LY_MAP_FONT_COLOR");
		return _fontColor;
	}

	/**
	 * Gets the outline color of this map link.
	 *
	 * @param  environment the environment, not used in this implementation
	 * @return the outline color of the link
	 */
	public Color getOutlineColor(LyEnvironment environment)
	{
		if (_outlineColor != null)
			return _outlineColor;
		_outlineColor = getColor("LY_MAP_OUTLINE_COLOR");
		return _outlineColor;
	}

	/**
	 * Returns the dimension of the label part of the link. If the {@link #_labelWidth} and {@link
	 * #_labelHeight} values of this link are not set (equal -1), the width and height are computed
	 * to the correct values for the label to be displayed.
	 *
	 * @param map  the map containing the label to draw
	 * @param  g the graphics context
	 * @return the label size of the current node
	 */
	public Dimension getLabelSize(LyAbstractMapComponentManager map, Graphics g)
	{
		if ((_labelCache == null) || (_labelWidth < 0) || (_labelHeight < 0))
		{
			String label = getLabel();

			if (label == null)
				return new Dimension(0, 0);

			FontMetrics fm = getCurrentFontMetrics(map, g);			

			StringTokenizer tkn = new StringTokenizer(label,"\n");

			int num_lines = tkn.countTokens();
			String[] lines       = new String[num_lines];
			int[] line_widths = new int   [num_lines];

			for (int i=0;i<num_lines;i++)
				lines[i] = tkn.nextToken();

			_labelHeight = fm.getHeight();
			_labelWidth = fm.stringWidth(label);

			int line_height=fm.getHeight();

			int max_width=0;

			for (int i=0;i<num_lines;i++)
			{
				line_widths[i] = fm.stringWidth(lines[ i ]);
				max_width = Math.max(max_width, line_widths[i]);
			}

			int btnMarginWidth =1;
			max_width += 2*btnMarginWidth;
			int text_height=  num_lines*line_height;
			if(_labelWidth!= max_width || _labelHeight!= text_height)
			{
				_labelWidth = max_width;
				_labelHeight= text_height;
			}

		}

		return (new Dimension(_labelWidth, _labelHeight));
	}

	/**
	 * Returns the dimension of the label part of the link. If the {@link #_labelWidth} and {@link
	 * #_labelHeight} values of this link are not set (equal -1), the width and height are computed
	 * to the correct values for the label to be displayed.
	 *
	 * @param map  the map containing the label to draw
	 * @param  g the graphics context
	 * @return the label size of the current node
	 */
	public Dimension getStartLabelSize(LyAbstractMapComponentManager map, Graphics g)
	{
		if ((_startLabelCache == null) || (_startLabelWidth < 0) || (_startLabelHeight < 0))
		{
			String label = getStartLabel();

			if (label == null)
				return new Dimension(0, 0);

			FontMetrics fm = getCurrentFontMetrics(map, g);

			/****/
			StringTokenizer tkn = new StringTokenizer(label,"\n");

			int num_lines = tkn.countTokens();
			String[] lines       = new String[num_lines];
			int[] line_widths = new int   [num_lines];

			for (int i=0;i<num_lines;i++)
				lines[i] = tkn.nextToken();

			/****/

			_startLabelHeight = fm.getHeight();
			_startLabelWidth = fm.stringWidth(label);

			/*****/
			int line_height=fm.getHeight();

			int max_width=0;

			for (int i=0;i<num_lines;i++)
			{
				line_widths[i] = fm.stringWidth(lines[ i ]);
				max_width = Math.max(max_width, line_widths[i]);
			}

			int btnMarginWidth =1;
			max_width += 2*btnMarginWidth;
			int text_height=  num_lines*line_height;
			if(_startLabelWidth!= max_width || _startLabelHeight!= text_height)
			{
				_startLabelWidth = max_width;
				_startLabelHeight= text_height;
			}
			/*****/
		}

		return (new Dimension(_startLabelWidth, _startLabelHeight));
	}

	/**
	 * Returns the dimension of the label part of the link. If the {@link #_labelWidth} and {@link
	 * #_labelHeight} values of this link are not set (equal -1), the width and height are computed
	 * to the correct values for the label to be displayed.
	 *
	 * @param map  the map containing the label to draw
	 * @param  g the graphics context
	 * @return the label size of the current node
	 */
	public Dimension getEndLabelSize(LyAbstractMapComponentManager map, Graphics g)
	{
		if ((_endLabelCache == null) || (_endLabelWidth < 0) || (_endLabelHeight < 0))
		{
			String label = getEndLabel();

			if (label == null)
				return new Dimension(0, 0);

			FontMetrics fm = getCurrentFontMetrics(map, g);

			/****/
			StringTokenizer tkn = new StringTokenizer(label,"\n");

			int num_lines = tkn.countTokens();
			String[] lines       = new String[num_lines];
			int[] line_widths = new int   [num_lines];

			for (int i=0;i<num_lines;i++)
				lines[i] = tkn.nextToken();

			/****/

			_endLabelHeight = fm.getHeight();
			_endLabelWidth = fm.stringWidth(label);

			/*****/
			int line_height=fm.getHeight();

			int max_width=0;

			for (int i=0;i<num_lines;i++)
			{
				line_widths[i] = fm.stringWidth(lines[ i ]);
				max_width = Math.max(max_width, line_widths[i]);
			}

			int btnMarginWidth =1;
			max_width += 2*btnMarginWidth;
			int text_height=  num_lines*line_height;
			if(_endLabelWidth!= max_width || _endLabelHeight!= text_height)
			{
				_endLabelWidth = max_width;
				_endLabelHeight= text_height;
			}
			/*****/
		}

		return (new Dimension(_endLabelWidth, _endLabelHeight));
	}

	/**
	 * Draws the link on the map, then prints its label at its center.
	 *
	 * @see leon.view.win.swing.map.LyMapElement#draw(leon.view.win.swing.map.LyAbstractMapComponentManager,
	 *      java.awt.Graphics, int)
	 */
	public void draw(LyAbstractMapComponentManager map, Graphics g, int mode)
	{
		if ((_startNode == null)
				|| (_endNode == null)
				|| (getLinkWidth() == 0)
				|| isHidden())
			return;

		drawLink(map, g, mode);
		drawLabels(map, g, mode);
	}

	/**
	 * Draws the label of this link on the specified map.
	 *
	 * @param map  the map containing the label to draw
	 * @param g    the graphics context
	 * @param mode the mode of the drawing, not used in this implementation
	 */
	public void drawLabels(LyAbstractMapComponentManager map, Graphics g, int mode)
	{
		Dimension   labelDim = getLabelSize(map, g);
		Dimension   startLabelDim = getStartLabelSize(map, g);
		Dimension   endLabelDim = getEndLabelSize(map, g);
		String	    label = getLabel();
		String		  startLabel = getStartLabel();
		String		  endLabel = getEndLabel();

		LyComponent parent = _mapLink.getParent();
		boolean     fill = true;

		if (parent instanceof LyMapContent)
			parent = parent.getParent();

		if (((LyMap)parent).getSelectionDraw() == LyMapBase.SelectionDraw.BORDER)
			fill = false;

		drawLabel(label, labelDim, map, g, mode, fill, LabelAlignment.CENTER);
		drawLabel(startLabel, startLabelDim, map, g, mode, fill, LabelAlignment.START);
		drawLabel(endLabel, endLabelDim, map, g, mode, fill, LabelAlignment.END);
	} // end method drawLabels

	/**
	 * Draw a label
	 * @param label
	 * @param labelDim
	 * @param map
	 * @param g
	 * @param mode
	 * @param fill
	 * @param align
	 */
	private void drawLabel(String label, Dimension labelDim, LyAbstractMapComponentManager map, Graphics g, 
			int mode, boolean fill, LabelAlignment align)
	{
		if ((label == null) || (label.trim().length() == 0))
			return;

		Font currentFont = getCurrentFont( map );
		Component component = map.getComponent();
		FontMetrics fm = component.getFontMetrics(currentFont);

		// Split the current label on the end line char 
		StringTokenizer tkn = new StringTokenizer(label,"\n");

		int num_lines = tkn.countTokens();
		String[] lines       = new String[num_lines];
		int[] line_widths = new int[num_lines];

		for (int i=0;i<num_lines;i++)
			lines[i] = tkn.nextToken();
		// Caculate the line widh
		for (int i=0;i<num_lines;i++)
		{
			line_widths[i] = fm.stringWidth(lines[ i ]);
		}
		//Get the current line heigth
		int line_height=fm.getHeight();
		// Get the current line ascent
		int line_ascent=fm.getAscent();

		LyPoint startNode = _startNode.getCenter();
		LyPoint endNode = _endNode.getCenter();

		LyPoint moved = getDisplacedValues(map, startNode, endNode);

		LyPoint newStart = new LyPoint(startNode.getX() + moved.getX(),
				startNode.getY() + moved.getY(),
				startNode.getZ());
		LyPoint newEnd = new LyPoint(endNode.getX() + moved.getX(),
				endNode.getY() + moved.getY(),
				endNode.getZ());

		Point		  startPoint = map.convertForDrawing(newStart);
		Point		  endPoint = map.convertForDrawing(newEnd);

		int			  labelX = 0;
		int			  labelY = 0;

		Rectangle endBoundingBox = _endNode.getBoundingRectangle(map);
		Rectangle startBoundingBox = _startNode.getBoundingRectangle(map);
		//Get the step
		int step = (int)(LOOP_STEP*map.getZoom());

		if ((align == null) || align.equals(LabelAlignment.CENTER))
		{
			if (_mapLink.getPointCount() >= 1)
			{
				int     i, np = _mapLink.getPointCount();
				int     middleX = (startPoint.x + endPoint.x) / 2;
				Point   start = startPoint;
				Point   end;

				for (i = 0; i < np; i++)
				{
					end = map.convertForDrawing(_mapLink.getPoint(i));
					if ((middleX >= start.x) && (middleX <= end.x))
					{
						startPoint = start;
						endPoint = end;
					}
					else
						start = end;
				}
			}

			if ( endBoundingBox.contains(map.convert(_startNode.getCenter())) ||  
					startBoundingBox.contains(map.convert(_endNode.getCenter())) ) 
			{
				//Calculate x and y position of the label
				labelX= startPoint.x - step - labelDim.width;
				labelY = startPoint.y + step + 4 * getLinkWidth() ;
			}
			else
			{
				labelX = (startPoint.x + endPoint.x) / 2 - (labelDim.width / 2);
				labelY = (startPoint.y + endPoint.y) / 2;
			}
		}
		else if (align.equals(LabelAlignment.START))
		{
			Dimension dimStart = _startNode.getImageSize(map.getComponent());
			// if the source and the target of the link are on the same position
			if ( endBoundingBox.contains(map.convert(_startNode.getCenter())) ||  
					startBoundingBox.contains(map.convert(_endNode.getCenter())) ) 
			{
				//Calculate x and y position of the label
				labelX = startPoint.x + 4*getLinkWidth();
				labelY = startPoint.y + step + 4*getLinkWidth() ;
			}
			else {
				if (startPoint.x < endPoint.x) {
					labelX = startPoint.x + dimStart.width;
				}
				else {
					labelX = startPoint.x - dimStart.width - labelDim.width; 
				}
				if(startPoint.y < endPoint.y) {
					labelY = startPoint.y + dimStart.height;
				}
				else {
					labelY = startPoint.y - dimStart.height - lines.length*line_height;
				}
			}
		}
		else if (align.equals(LabelAlignment.END))
		{
			Dimension dimEnd = _endNode.getImageSize(map.getComponent());

			// if the source and the target of the link are on the same position
			if ( endBoundingBox.contains(map.convert(_startNode.getCenter())) ||  
					startBoundingBox.contains(map.convert(_endNode.getCenter())) ) 
			{
				//Calculate x and y position of the label
				labelX= startPoint.x - step - labelDim.width;
				labelY = startPoint.y - labelDim.height - 4*getLinkWidth() ;
			}
			else {
				if (startPoint.x < endPoint.x) {
					labelX = endPoint.x - dimEnd.width - labelDim.width; 
				}
				else {
					labelX = endPoint.x + dimEnd.width;
				}
				if(startPoint.y < endPoint.y) {
					labelY = endPoint.y - dimEnd.height - lines.length*line_height;
				}
				else {
					labelY = endPoint.y + dimEnd.height;
				}
			}
		}

		LyEnvironment environment = getEnvironment();

		// If selected, draw a rectangle around the label
		if (_mapLink.isSelected())
		{
			if (map.isActivated())
				g.setColor(getSelectedColor(environment));
			else
				g.setColor(getInactiveSelectedColor(environment));
		}
		else
			g.setColor(map.getComponent().getBackground());

		int offset = (((getLinkWidth() % 2) == 0) ? 0 : 1);

		if (_mapLink.isSelected()) {
			g.drawRect(labelX - 2,
					labelY + ((getLinkWidth() + SELECTION_OFFSET) / 2)
					+ offset, labelDim.width + 2, labelDim.height);
			if (fill)
				g.fillRect(labelX-2,
						labelY + ((getLinkWidth() + SELECTION_OFFSET) / 2)
						+ offset, labelDim.width + 2, labelDim.height);
		}

		// Draw the label
		String labelColor = _mapLink.getFontColor();
		if (labelColor == null)
			g.setColor(getFontColor(environment));
		else
			g.setColor(getColor(labelColor));
		g.setFont(currentFont);

		// Get y values of the label
		int y = labelY + line_ascent + ((getLinkWidth() + SELECTION_OFFSET) / 2) + offset;

		Insets insets = ((JComponent)component).getInsets();

		for (int i=0; i<num_lines; i++)
		{ 
			int x = labelX + insets.left + (labelDim.width - line_widths[i])/2;
			g.drawString(lines[i],x,y);
			y+=line_height;
		}
	}

	/**
	 * Computes the value the links need to be displaced in order to draw it if other links has the
	 * same start and end points.
	 * 
	 * @param map the map that is drawn
	 * @param start the starting node
	 * @param end the ending node
	 * @return a LyPoint that has the X displacement in its X value and the Y displacement in its Y value.
	 */
	protected LyPoint getDisplacedValues(LyAbstractMapComponentManager map, LyPoint start, LyPoint end)
	{		
		// numLinks holds the number of links that has the same starting and ending nodes
		_sameLinks = new ArrayList<LySwingMapLink>();

		for (int i = 0; i < map.getElementCount(); i++)
		{
			LyMapElement element = map.getElementAt(i);

			if (element instanceof LySwingMapLink && !_sameLinks.contains(element))
			{
				LySwingMapLink link = (LySwingMapLink)element;

				// Tests that the links has the same endings as the current one, in both ways
				if ((_startNode != null)
						&& (_endNode != null)
						&& ((_startNode.equals(link._startNode)
								&& _endNode.equals(link._endNode))
								|| (_startNode.equals(link._endNode)
										&& _endNode.equals(link._startNode))))
					_sameLinks.add(link);
			}
		}

		double pos = _sameLinks.indexOf(this) - (_sameLinks.size() - 1) / 2d;

		if (start.equals(end))
		{
			return new LyPoint((float)(DISPLACEMENT_VALUE * pos), (float)(DISPLACEMENT_VALUE * pos), 0);
		}
		else
		{
			double x1 = start.getX();
			double y1 = start.getY();
			double x2 = end.getX();
			double y2 = end.getY();

			double hyp = Math.hypot(x2 - x1, y2 - y1);
			if (hyp != 0)
			{
				double cosA, sinA;

				if (y2 > y1)
				{
					cosA = (x2 - x1) / hyp;
					sinA = (y2 - y1) / hyp;
				}
				else
				{
					cosA = (x1 - x2) / hyp;
					sinA = (y1 - y2) / hyp;
				}

				double cstX = DISPLACEMENT_VALUE * pos * sinA;
				double cstY = -DISPLACEMENT_VALUE * pos * cosA;

				return new LyPoint((float)cstX, (float)cstY, 0);
			}
		}
		return new LyPoint(0, 0, 0);
	}

	/**
	 * Computes the point that is the intersection between <code>rectangle</code>
	 * and the line that links <code>internalPoint</code> with
	 * <code>externalPoint</code>.
	 * <code>internalPoint is contained in <code>rectangle</code> while
	 * we don't know where is<code>externalPoint</code>.
	 * 
	 * @param rectangle
	 * @param internalPoint
	 * @param externalPoint
	 * @return the <code>Point</code> representing the intersection between
	 * <code>rectangle</code> and the line that links <code>internalPoint</code> with
	 * <code>externalPoint</code> if there is any, otherwise <code>null</code>
	 */
	private Point getIntersection(Rectangle rectangle, RectangleSide rectSide, Point internalPoint, Point externalPoint) {
		// The (internalPoint, externalPoint) line is completely contained in
		// rectangle, there is no intersection
		if (rectangle.contains(externalPoint))   {
			return null;
		}
		// There is an intersection
		double iX = internalPoint.getX();    // internalPoint coordinates
		double iY = internalPoint.getY();
		double eX = externalPoint.getX();    // externalPoint coordinates
		double eY = externalPoint.getY();

		// (internalPoint, externalPoint) line equation
		// Y = ((iY - eY) * X - iY * eX + eY * iX) / (iX - eX)

		// Compute the intersection
		double x;
		double y;

		if (rectSide == RectangleSide.EAST)   {
			// The intersection is on the east side
			x = rectangle.getX() + rectangle.getWidth(); 
			y = ((iY - eY) * x - iY * eX + eY * iX) / (iX - eX);        
		} else if (rectSide == RectangleSide.SOUTH)   {
			// The intersection is on the south side
			y = rectangle.getY() + rectangle.getHeight(); 
			x = ((iX - eX) * y + iY * eX - iX * eY) / (iY - eY);               
		} else if (rectSide == RectangleSide.WEST)   {       
			// The intersection is on the west side
			x = rectangle.getX(); 
			y = ((iY - eY) * x - iY * eX + eY * iX) / (iX - eX); 
		} else { // rectSide == RectangleSide.NORTH
			// The intersection is on the north side
			y = rectangle.getY(); 
			x = ((iX - eX) * y + iY * eX - iX * eY) / (iY - eY);                       
		}
		return new Point((int)x, (int)y);
	}	

	/**
	 * Draws the link.
	 *
	 * @param map  the link to draw
	 * @param g    the graphics context
	 * @param mode the mode of the drawing
	 */
	public void drawLink(LyAbstractMapComponentManager map, Graphics g, int mode) {

		if ((_startNode == null) || (_endNode == null))
			return;

		boolean isNormalMode = (mode == LyMapElement.NORMAL_MODE);

		if (!isNormalMode)
			return;

		// Get environment.
		LyEnvironment environment = getEnvironment();
		// Get default colors.
		Color colorStart = getColorStart(environment);
		Color colorMiddle = getColor(environment);
		Color colorEnd = getColorEnd(environment);
		// Set basic graphics properties.
		Graphics2D g2d = (Graphics2D) g;
		g2d.setStroke(new BasicStroke(1, BasicStroke.CAP_SQUARE,
				BasicStroke.JOIN_ROUND));

		// Compute the link start and endPoint. 
		// The link will be drawn to link those two points
		Point startLinkPoint = getStartLinkPoint(map, true);
		Point endLinkPoint = getEndLinkPoint(map, true);


		// Define coordinates of the start point and the end point.
		int x1 = startLinkPoint.x;
		int y1 = startLinkPoint.y;
		int x2 = endLinkPoint.x;
		int y2 = endLinkPoint.y;

		// Divide the link into three parts
		int firstSectionX = ((2 * x1) + x2) / 3;
		int firstSectionY = ((2 * y1) + y2) / 3;
		int secondSectionX = (x1 + (2 * x2)) / 3;
		int secondSectionY = (y1 + (2 * y2)) / 3;

		// Modify sections if there are intermediate points (sub-component)
		if (_mapLink.getPointCount() >= 1) {
			firstSectionX = map.convert(_mapLink.getPoint(0)).x;
			firstSectionY = map.convert(_mapLink.getPoint(0)).y;

			if (_mapLink.getPointCount() >= 2) {
				int lastPoint = _mapLink.getPointCount() - 1;

				secondSectionX = map.convert(_mapLink.getPoint(lastPoint)).x;
				secondSectionY = map.convert(_mapLink.getPoint(lastPoint)).y;
			}
			else {
				secondSectionX = ((2 * firstSectionX) + x2) / 3;
				secondSectionY = ((2 * firstSectionY) + y2) / 3;
			}
		}

		// Check whether three different parts must be generated or only the last
		// one
		boolean onlyDrawLastPart = shallDrawOnlyOneSection();

		if (onlyDrawLastPart) {
			secondSectionX = x1;
			secondSectionY = y1;
			colorStart = colorMiddle;
			colorEnd = colorMiddle;
		}

		LyComponent parent = _mapLink.getParent();
		boolean fill = true;

		if (parent instanceof LyMapContent)
			parent = parent.getParent();

		if (((LyMap) parent).getSelectionDraw() == LyMapBase.SelectionDraw.BORDER)
			fill = false;

		// Get bounding boxes for the start node and the end node of the link.
		Rectangle endBoundingBox = _endNode.getBoundingRectangle(map);
		Rectangle startBoundingBox = _startNode.getBoundingRectangle(map);
		Point tmp = new Point(0, 0);

		if (_mapLink.getMotif() != null) {
			// There is a motif, don't use the different sections ?!
			drawLinkWithMotif(map, g, environment, x1, y1, x2, y2, fill);
		}
		else {
			// If the start node style is image, then based the link start point on
			// real coordinates.
			if (_startNode.getMapNodeStyle().equals(Style.IMAGE)) {
				x1 = startLinkPoint.x;
				y1 = startLinkPoint.y;
			}
			// If the end node style is image, then based the link end point on real
			// coordinates.
			if (_endNode.getMapNodeStyle().equals(Style.IMAGE)) {
				x2 = endLinkPoint.x;
				y2 = endLinkPoint.y;
			}

			if ((endBoundingBox.contains(map.convert(_startNode.getCenter())) || startBoundingBox
					.contains(map.convert(_endNode.getCenter())))) { // link on the same
				// node

				if (_mapLink.isSelected()) {
					if (map.isActivated())
						g.setColor(getSelectedColor(environment));
					else
						g.setColor(getInactiveSelectedColor(environment));

					drawAutoLink(map, g, x1, y1, startBoundingBox.y + startBoundingBox.height, x2, y2,
							endBoundingBox.x, (int) (LOOP_STEP * map.getZoom()), getLinkWidth()
							+ SELECTION_OFFSET, leon.view.LyMapLink.Style.LINE, _mapLink
							.isDoubleLink(), false);
				}
				// Force color of the link
				g.setColor(colorEnd);

				drawAutoLink(map, g, x1, y1, startBoundingBox.y + startBoundingBox.height, x2, y2,
						endBoundingBox.x, (int) (LOOP_STEP * map.getZoom()),
						getLinkWidth(), _mapLink.getStyle(), _mapLink.isDoubleLink(),
						true);
			}

			if (!onlyDrawLastPart) {
				// Draw the first part
				if ((_mapLink.getArrow() == LyMapLink.Arrow.START)
						|| (_mapLink.getArrow() == LyMapLink.Arrow.BOTH)) {
					boolean fillArrow = (_mapLink.getStartArrowFillMode() == LyMapLink.ArrowFill.FULL || _mapLink.getStartArrowFillMode() == null);
					double distance = Math.hypot(firstSectionX - x1, firstSectionY - y1);

					int xArrow = (int) startLinkPoint.getX();
					int yArrow = (int) startLinkPoint.getY();

					if (endBoundingBox.contains(map.convert(_startNode.getCenter()))
							|| startBoundingBox.contains(map.convert(_endNode.getCenter()))) {
						firstSectionX = x1;
						firstSectionY = y1;
					}

					// If the points are superposed, draw auto links on node
					if ((distance > 0)
							|| endBoundingBox.contains(map.convert(_startNode.getCenter()))
							|| startBoundingBox.contains(map.convert(_endNode.getCenter()))) {
						// If selected, draw a polygon under the middle part
						if (_mapLink.isSelected()) {

							if (map.isActivated())
								g.setColor(getSelectedColor(environment));
							else
								g.setColor(getInactiveSelectedColor(environment));

							// Draw a link that loops between (firstSectionX, firstSectionY) and
							// (xArrow, yArrow)
							drawLinkDecorator(
									g,
									_mapLink.getStartArrowType(),
									firstSectionX,
									firstSectionY,
									xArrow,
									yArrow,
									getLinkWidth() + SELECTION_OFFSET,
									!fill,
									fillArrow,
									false,
									LyMapLink.Style.LINE
									);
						}

						g.setColor(colorStart);

						drawLinkDecorator(
								g,
								_mapLink.getStartArrowType(),
								firstSectionX,
								firstSectionY,
								xArrow,
								yArrow,
								getLinkWidth(),
								_mapLink.isDoubleLink(),
								fillArrow,
								true,
								_mapLink.getStyle()
								);

					}
				}
				else if (!endBoundingBox.contains(map.convert(_startNode.getCenter()))
						|| !startBoundingBox.contains(map.convert(_endNode.getCenter()))) {
					// If selected, draw a polygon under the start part
					if (_mapLink.isSelected()) {
						if (map.isActivated())
							g.setColor(getSelectedColor(environment));
						else
							g.setColor(getInactiveSelectedColor(environment));

						drawLine(g, x1, y1, firstSectionX, firstSectionY, getLinkWidth()
								+ SELECTION_OFFSET, LyMapLink.Style.LINE, !fill, false);
					}

					g.setColor(colorStart);

					if (_mapLink.getStyle() == LyMapLink.Style.TUBE)    {
						LySwingUtils.drawTube(g, x1, y1, firstSectionX, firstSectionY,
								getLinkWidth(), colorStart);
					}
					else    {
						drawLine(g, x1, y1, firstSectionX, firstSectionY, getLinkWidth(),
								_mapLink.getStyle(), _mapLink.isDoubleLink(), true);
					}
				}
			}

			// Draw the third part
			g.setColor(colorEnd);

			if ((_mapLink.getArrow() == LyMapLink.Arrow.END)
					|| (_mapLink.getArrow() == LyMapLink.Arrow.BOTH)) {

				double distance = Math.hypot(x2 - secondSectionX, y2 - secondSectionY);

				int xArrow = (int) endLinkPoint.getX();
				int yArrow = (int) endLinkPoint.getY();

				if (endBoundingBox.contains(map.convert(_startNode.getCenter()))
						|| startBoundingBox.contains(map.convert(_endNode.getCenter()))) {
					xArrow = endBoundingBox.x;  // The loop link arrow ends on the endNode bounding rectangle west side.
					secondSectionX = endLinkPoint.x - (int)(LOOP_STEP * map.getZoom()/ 2);
					secondSectionY = endLinkPoint.y;
				}

				// If the points are superposed, draw auto links on node
				if ((distance > 0)
						|| endBoundingBox.contains(map.convert(_startNode.getCenter()))
						|| startBoundingBox.contains(map.convert(_endNode.getCenter()))) {
					boolean fillArrow = (_mapLink.getEndArrowFillMode() == LyMapLink.ArrowFill.FULL || _mapLink.getEndArrowFillMode() == null);
					// If selected, draw a polygon under the third part
					if (_mapLink.isSelected()) {
						if (map.isActivated())
							g.setColor(getSelectedColor(environment));
						else
							g.setColor(getInactiveSelectedColor(environment));

						drawLinkDecorator(
								g,
								_mapLink.getEndArrowType(),
								secondSectionX,
								secondSectionY,
								xArrow,
								yArrow,
								getLinkWidth() + SELECTION_OFFSET,
								!fill,
								fillArrow,
								false,
								LyMapLink.Style.LINE
								);           
					}
					// force the color setting
					g.setColor(colorEnd);
					drawLinkDecorator(
							g,
							_mapLink.getEndArrowType(),
							secondSectionX,
							secondSectionY,
							xArrow,
							yArrow,
							getLinkWidth(),
							_mapLink.isDoubleLink(),
							fillArrow,
							true,
							_mapLink.getStyle()
							);           
				}
			}
			else if (!endBoundingBox.contains(map.convert(_startNode.getCenter()))
					&& !startBoundingBox.contains(map.convert(_endNode.getCenter()))) {

				// If selected, draw a polygon under the end part
				if (_mapLink.isSelected()) {

					if (map.isActivated())
						g.setColor(getSelectedColor(environment));
					else
						g.setColor(getInactiveSelectedColor(environment));

					drawLine(g, secondSectionX, secondSectionY, x2, y2, getLinkWidth()
							+ SELECTION_OFFSET, LyMapLink.Style.LINE, !fill, false);
				}
				// Force color setting
				g.setColor(colorEnd);

				if (_mapLink.getStyle() == LyMapLink.Style.TUBE)  {
					LySwingUtils.drawTube(g, secondSectionX, secondSectionY, x2, y2,
							getLinkWidth(), colorEnd);
				}
				else {
					drawLine(g, secondSectionX, secondSectionY, x2, y2, getLinkWidth(),
							_mapLink.getStyle(), _mapLink.isDoubleLink(), true);
				}
			}

			if (!onlyDrawLastPart) {
				if (!endBoundingBox.contains(map.convert(_startNode.getCenter()))
						&& !startBoundingBox.contains(map.convert(_endNode.getCenter()))) {
					// If selected, draw a polygon under the middle part
					if (_mapLink.isSelected()) {
						if (map.isActivated())
							g.setColor(getSelectedColor(environment));
						else
							g.setColor(getInactiveSelectedColor(environment));

						tmp.x = firstSectionX;
						tmp.y = firstSectionY;
						// If there are intermediate points
						for (int i = 1; i < (_mapLink.getPointCount() - 1); i++) {
							Point aPoint = map.convertForDrawing(_mapLink.getPoint(i));

							drawLine(g, tmp.x, tmp.y, aPoint.x, aPoint.y, getLinkWidth()
									+ SELECTION_OFFSET, LyMapLink.Style.LINE, !fill, false);
							tmp.x = aPoint.x;
							tmp.y = aPoint.y;
						}
						drawLine(g, tmp.x, tmp.y, secondSectionX, secondSectionY, getLinkWidth()
								+ SELECTION_OFFSET, LyMapLink.Style.LINE, !fill, false);
					}

					// Draw the middle part
					g.setColor(colorMiddle);

					tmp.x = firstSectionX;
					tmp.y = firstSectionY;
					// If there are intermediate points
					for (int i = 1; i < (_mapLink.getPointCount() - 1); i++) {
						LyPoint tmpEnd = _mapLink.getPoint(i);
						LyPoint moved = getDisplacedValues(map, _mapLink.getPoint(i - 1), tmpEnd);
						LyPoint newEnd = new LyPoint(tmpEnd.getX() + moved.getX(), tmpEnd.getY()
								- moved.getY(), tmpEnd.getZ());
						Point aPoint = map.convertForDrawing(newEnd);

						drawLine(g, tmp.x, tmp.y, aPoint.x, aPoint.y, getLinkWidth(),
								_mapLink.getStyle(), _mapLink.isDoubleLink(), true);
						tmp.x = aPoint.x;
						tmp.y = aPoint.y;
					}

					if (_mapLink.getStyle() == LyMapLink.Style.TUBE)    {
						LySwingUtils.drawTube(g, tmp.x, tmp.y, secondSectionX,
								secondSectionY, getLinkWidth(), colorMiddle);
					}
					else    {
						drawLine(g, tmp.x, tmp.y, secondSectionX, secondSectionY, getLinkWidth(),
								_mapLink.getStyle(), _mapLink.isDoubleLink(), true);
					}
				}
			}
		}

		// Draw the images
		Image[] images = getImages();

		Point imageStartPos = null;
		Point imageMiddlePos = null;
		Point imageEndPos = null;

		if (images != null) {
			Image imageStart = images[0];
			Image imageMiddle = images[1];
			Image imageEnd = images[2];

			boolean overlap = (endBoundingBox.contains(map.convert(_startNode.getCenter()))
					|| startBoundingBox.contains(map.convert(_endNode.getCenter())));

			if (imageStart != null) {

				// Draw the start link image at the appropriate position
				imageStartPos = drawLinkImage(
						map, g, environment, 
						(x1 + firstSectionX) / 2,
						(y1 + firstSectionY) / 2,
						overlap,
						startLinkPoint.x + 4 * getLinkWidth(),
						startLinkPoint.y + (int) (LOOP_STEP * map.getZoom()) + 4 * getLinkWidth(),
						imageStart,
						_imageStartString);

			}

			if (imageMiddle != null) {

				imageMiddlePos = drawLinkImage(
						map, g, environment, 
						(secondSectionX + firstSectionX) / 2,
						(secondSectionY + firstSectionY) / 2,
						overlap,
						startLinkPoint.x - (int) (LOOP_STEP * map.getZoom()),
						startLinkPoint.y + (int) (LOOP_STEP * map.getZoom()),
						imageMiddle,
						_imageMiddleString);
			}

			if (imageEnd != null) {

				imageEndPos = drawLinkImage(
						map, g, environment, 
						(secondSectionX + x2) / 2,
						(secondSectionY + y2) / 2,
						overlap,
						startLinkPoint.x - (int) (LOOP_STEP * map.getZoom()),
						startLinkPoint.y,
						imageEnd,
						_imageEndString);
			}

			_imagePos = new Point[] { imageStartPos, imageMiddlePos, imageEndPos };
		}

	} // end method drawLink

	private Point drawLinkImage(
			LyAbstractMapComponentManager map,
			Graphics g,
			LyEnvironment environment,
			int xImage,
			int yImage,
			boolean overlap,
			int overlapXImage,
			int overlapYImage,
			Image image,
			String imagePath) {

		int width = image.getWidth(map.getComponent());
		int height = image.getHeight(map.getComponent());

		// Check if image is the same one as an unknown image. If so, try to
		// load image
		// as a full path
		Image unknownImage = (Image) environment.getImage("_unknownImage");
		if (image == unknownImage) {
			// Given image resource may be a full path : try to load image as a
			// full path
			Image pathImage = (Image) environment.getImages().loadImage(null,
					imagePath, false);
			if (pathImage != null)
				image = pathImage;
		}

		// if the source and the target of the link are on the same position
		if (overlap) {

			// Calculate x and y position of the label
			xImage = overlapXImage;
			yImage = overlapYImage;
		}

		g.drawImage(image, xImage - (width / 2), yImage - (height / 2),
				width, height, map.getComponent());

		return new Point(xImage, yImage);
	}

	/**
	 * Draws the link decorator depending on the arrowType.
	 * In the case of an Arrow, draws the triangle that figures the link way.
	 * @param g
	 * @param arrowType
	 * @param x
	 * @param y
	 * @param xArrow
	 * @param yArrow
	 * @param width
	 * @param dble
	 * @param fillArrow
	 * @param outline
	 * @param style
	 */private void drawLinkDecorator(Graphics g, ArrowType arrowType, int x, int y,
			 int xArrow, int yArrow, int width, boolean dble, boolean fillArrow, boolean outline, LyMapLink.Style style) {

		 if (arrowType == LyMapLink.ArrowType.CIRCLE) {
			 drawCircle(g, xArrow, yArrow, x, y,
					 width,
					 style, dble, outline, fillArrow);
		 } else if (arrowType == LyMapLink.ArrowType.SQUARE)  {
			 drawSquare(g, xArrow, yArrow, x, y,
					 width,
					 style, dble, outline, fillArrow);
		 } else  {
			 drawArrow(g, xArrow, yArrow, x, y,
					 width,
					 style, dble, outline, fillArrow);
		 }
	 }

	 /**
	  * Computes the radius of a circle that encompasses the specified node
	  * @param map
	  * @param node
	  * @return
	  */
	 private final double computeNodeRadius(LyAbstractMapComponentManager map, LySwingMapNode node) {

		 double radius;
		 // Calculate the radius for the start node of the current link
		 Dimension dimNode = node.getImageSize(map.getComponent());
		 if (((LyMapNode) node.getViewComponent()).getStyle() == LyMapNode.Style.CIRCLE) {
			 radius = dimNode.width / 2;
		 }
		 else {
			 radius = Math.hypot(dimNode.width, dimNode.height) / 2;
		 }
		 // Apply zoom value on the calculated radius if is this node is resizable
		 if (node._mapNode.isResizable()) {
			 radius *= map.getZoom();
		 }
		 return radius;
	 }

	 /**
	  * Checks whether only the last section of the link shall be drawn or not
	  * @return
	  */
	 private boolean shallDrawOnlyOneSection() {

		 boolean onlyDrawLastPart = true;

		 onlyDrawLastPart &= (_mapLink.getPointCount() == 0);
		 onlyDrawLastPart &= (_mapLink.getColorStart() == null)
				 && (_mapLink.getColorEnd() == null);
		 onlyDrawLastPart &= (_mapLink.getArrow() != LyMapLink.Arrow.START)
				 && (_mapLink.getArrow() != LyMapLink.Arrow.BOTH);
		 onlyDrawLastPart &= (_mapLink.getImageEnd() == null)
				 && (_mapLink.getImageMiddle() == null)
				 && (_mapLink.getImageStart() == null);
		 onlyDrawLastPart &= (!_startNode.getMapNodeStyle().equals(Style.IMAGE));
		 return onlyDrawLastPart;
	 }

	 private void drawLinkWithMotif(LyAbstractMapComponentManager map, Graphics g,
			 LyEnvironment environment, int x1, int y1, int x2, int y2, boolean fill) {

		 Point tmp = new Point();
		 // If selected, draw a polygon under the tube
		 if (_mapLink.isSelected()) {
			 if (map.isActivated())
				 g.setColor(getSelectedColor(environment));
			 else
				 g.setColor(getInactiveSelectedColor(environment));

			 tmp.x = x1;
			 tmp.y = y1;
			 // If there are intermediate points
			 for (int i = 0; i < _mapLink.getPointCount(); i++) {
				 Point aPoint = map.convertForDrawing(_mapLink.getPoint(i));

				 drawLine(g, tmp.x, tmp.y, aPoint.x, aPoint.y, getLinkWidth()
						 + SELECTION_OFFSET, LyMapLink.Style.LINE, !fill, false);
				 tmp.x = aPoint.x;
				 tmp.y = aPoint.y;
			 }
			 drawLine(g, tmp.x, tmp.y, x2, y2, getLinkWidth()
					 + SELECTION_OFFSET, LyMapLink.Style.LINE, !fill, false);
		 }

		 Image image = getImage(_mapLink.getMotif());
		 int length = Math.max(image.getWidth(null), image.getHeight(null));

		 tmp.x = x1;
		 tmp.y = y1;
		 // If there are intermediate points
		 for (int i = 0; i < _mapLink.getPointCount(); i++) {
			 Point aPoint = map.convertForDrawing(_mapLink.getPoint(i));

			 drawDashedLine(g, tmp.x, tmp.y, aPoint.x, aPoint.y, length / 2, image);
			 tmp.x = aPoint.x;
			 tmp.y = aPoint.y;
		 }
		 drawDashedLine(g, tmp.x, tmp.y, x2, y2, length / 2, image);
	 }

	 /**
	  * Adapt the coordinates of the center of the <code>node</code> parameter.
	  * If the node is a child node, its center is translated in regards to its
	  * parent node.
	  * 
	  * @param map
	  * @param node
	  */
	 private final void adaptChildNodeCoordinates(LyAbstractMapComponentManager map, LySwingMapNode node) {

		 if (node.getParentNode() != null) {
			 Point parent = map.convert(node.getParentNode().getCenter());
			 Point child = map.convert(node.getCenter());

			 if (node.getOffset() == null) {
				 Point newPoint = new Point(parent.x - child.x, parent.y - child.y);

				 node.setOffset(newPoint);
			 }
			 else {
				 child.x = parent.x - node.getOffset().x;
				 child.y = parent.y - node.getOffset().y;
				 node._center = map.revert(child);
			 }
		 }
	 }

	 /**
	  * Computes the screen starting and end points of the line between the
	  * specified map nodes.
	  * Those points correspond to the node center eventually moved from an offset
	  * if there are several links between the two nodes to avoid that the lines
	  * overlap.
	  * 
	  * @param map
	  * @param startNode
	  * @param endNode
	  * @return
	  */
	 private final Point[] computeDrawingLineExtremePoints(
			 LyAbstractMapComponentManager map, 
			 LySwingMapNode startNode,
			 LySwingMapNode endNode, boolean computingStartNode)   {
		 // Dans le cas ou le lien est affich avant le noeud, on vrifie s'il s'agit
		 // d'un noeud fils
		 // et on applique son dcalage vis  vis du noeud parent
		 adaptChildNodeCoordinates(map, startNode);
		 adaptChildNodeCoordinates(map, endNode);

		 // Get the point of the center of end node and start node.
		 LyPoint start = startNode.getCenter();
		 LyPoint end = endNode.getCenter();

		 if (_mapLink.getPointCount() > 0)
		 {
			 if (computingStartNode)
				 end = _mapLink.getPoint(0);
			 else
				 start = _mapLink.getPoint(_mapLink.getPointCount() - 1);
		 }
		 // Get the displaced point to each other links between the start node and
		 // the end node.
		 LyPoint moved = getDisplacedValues(map, start, end);
		 // Now calculate the start and the end of the link compute with the
		 // displaced values.
		 LyPoint newStart = new LyPoint(start.getX() + moved.getX(), start.getY()
				 + moved.getY(), start.getZ());
		 LyPoint newEnd = new LyPoint(end.getX() + moved.getX(), end.getY()
				 + moved.getY(), end.getZ());

		 Point[] result = new Point[2];

		 // And convert them to draw and put them into the result array
		 result[START_POINT_IDX] = map.convertForDrawing(newStart);
		 result[END_POINT_IDX] = map.convertForDrawing(newEnd);

		 return result;
	 }

	 /**
	  * Draws a line in the given graphics context from the point (x1,y1) to (x2,y2) using the
	  * default color and the specified attributes.
	  * 
	  * @param g       the graphics context
	  * @param x1      the x coordinate of the starting point
	  * @param y1      the y coordinate of the starting point
	  * @param x2      the x coordinate of the ending point
	  * @param y2      the y coordinate of the ending point
	  * @param width   the width of the line
	  * @param style   the style of the line, is one of the following:
	  *
	  *                <ul>
	  *                  <li>{@link LyMapLink.Style#LINE}</li>
	  *                  <li>{@link LyMapLink.Style#DOT}</li>
	  *                  <li>{@link LyMapLink.Style#DASHED}</li>
	  *                  <li>{@link LyMapLink.Style#TUBE}</li>
	  *                </ul>
	  * @param dble    fill the line or not
	  * @param outline outline the link with the outline color
	  */
	 protected void drawLine(Graphics g, int x1, int y1, int x2, int y2, int width,
			 LyMapLink.Style style, boolean dble, boolean outline)
	 {
		 // Keep the color in memory
		 Color color = g.getColor();
		 Graphics2D g2d = (Graphics2D) g;
		 LyEnvironment environment = getEnvironment();

		 if (style == LyMapLink.Style.LINE)
		 {

			 // For a linear wide line
			 if (width > 1)
			 {
				 // Calculate the polygon containing the link
				 Point[] polygon = LySwingUtils.getPolygon(x1, y1, x2, y2, width);

				 if (dble)
				 {
					 g2d.drawLine(polygon[0].x, polygon[0].y, polygon[1].x, polygon[1].y);
					 g2d.drawLine(polygon[2].x, polygon[2].y, polygon[3].x, polygon[3].y);
				 }
				 else
				 {
					 int[] xPoints = new int[]
							 {
							 polygon[0].x,
							 polygon[1].x,
							 polygon[2].x,
							 polygon[3].x
							 };
					 int[] yPoints = new int[]
							 {
							 polygon[0].y,
							 polygon[1].y,
							 polygon[2].y,
							 polygon[3].y
							 };

					 g2d.fillPolygon(xPoints, yPoints, 4);

					 if (outline)
					 {

						 g2d.setColor(getOutlineColor(environment));
						 g2d.drawLine(polygon[0].x, polygon[0].y, polygon[1].x, polygon[1].y);
						 g2d.drawLine(polygon[2].x, polygon[2].y, polygon[3].x, polygon[3].y);
						 g2d.setColor(color);
					 }
				 }
			 }
			 else
				 g2d.drawLine(x1, y1, x2, y2);
		 }
		 // If dot or dashed line, draw a dashed line with the corresponding
		 // length for segments.
		 else
		 {
			 int length = 1;		

			 if (style == LyMapLink.Style.DOT) {
				 length = 1;
				 float[] styleLine = {1,8}; /** les pointills seront gros que les blancs */
				 g2d.setStroke( new BasicStroke( 
						 width,
						 BasicStroke.CAP_ROUND,
						 BasicStroke.JOIN_MITER,
						 10.0f,
						 styleLine,
						 0
						 ));

			 }
			 else if (style == LyMapLink.Style.DASHED) {
				 length = 5;
				 float[] styleLine = {10,5}; /** les tirets seront 2 fois plus long que les blancs */
				 g2d.setStroke( new BasicStroke( 
						 width,
						 BasicStroke.CAP_BUTT,
						 BasicStroke.JOIN_BEVEL,
						 10.0f,
						 styleLine,
						 0
						 ));
			 }


			 if (dble)
			 {
				 Point[] polygon = LySwingUtils.getPolygon(x1, y1, x2, y2, width);

				 drawDashedLine(g2d, polygon[0].x, polygon[0].y, polygon[1].x, polygon[1].y, length);
				 drawDashedLine(g2d, polygon[2].x, polygon[2].y, polygon[3].x, polygon[3].y, length);
			 }
			 else {
				 g2d.drawLine(x1, y1, x2, y2);
			 }

		 }
		 // Reset graphics properties
		 g2d.setColor(color);
		 g2d.setStroke(new BasicStroke(1,
				 BasicStroke.CAP_SQUARE,BasicStroke.JOIN_ROUND));

	 } // end method drawLine

	 /**
	  * (non-Javadoc).
	  *
	  * @see leon.view.win.swing.map.LyMapElement#drawInViewer(leon.view.win.swing.map.LyAbstractMapComponentManager,
	  *      java.awt.Graphics)
	  */
	 public void drawInViewer(LyAbstractMapComponentManager map, Graphics g)
	 {
		 if ((_startNode == null)
				 || (_endNode == null)
				 || (getLinkWidth() == 0))
			 return;
		 // Set the Graphic2D object
		 Graphics2D g2d = (Graphics2D) g;
		 g2d.setStroke(new BasicStroke(1,
				 BasicStroke.CAP_SQUARE,BasicStroke.JOIN_ROUND));
		 LyPoint start = _startNode.getCenter();
		 LyPoint end = _endNode.getCenter();

		 LyPoint moved = getDisplacedValues(map, start, end);

		 LyPoint newStart = new LyPoint(start.getX() + moved.getX(),
				 start.getY() + moved.getY(),
				 start.getZ());
		 LyPoint newEnd = new LyPoint(end.getX() + moved.getX(),
				 end.getY() + moved.getY(),
				 end.getZ());

		 Point startPoint = map.convert(newStart);

		 startPoint = map.convertToViewer(startPoint);

		 Point endPoint = map.convert(newEnd);

		 endPoint = map.convertToViewer(endPoint);

		 LyEnvironment environment = getEnvironment();

		 g.setColor(getColor(environment));

		 Rectangle endBoundingBox = _endNode.getBoundingRectangle(map);
		 Rectangle startBoundingBox = _startNode.getBoundingRectangle(map);

		 if ( (endBoundingBox.contains(map.convert(_startNode.getCenter())) ||  
				 startBoundingBox.contains(map.convert(_endNode.getCenter()))) &&
				 _startNode.getId().equals(_endNode.getId()))
		 {
			 //Get the step
			 int step = LOOP_STEP;//y1 - y2;

			 //Calculate point to draw the square
			 Point aPoint = map.convert(new LyPoint(newStart.getX(),newStart.getY()+step,0));
			 aPoint = map.convertToViewer(aPoint);
			 Point bPoint = map.convert(new LyPoint(newStart.getX()-step,newStart.getY()+step,0));
			 bPoint = map.convertToViewer(bPoint);
			 Point cPoint = map.convert(new LyPoint(newStart.getX()-step,newStart.getY(),0));
			 cPoint = map.convertToViewer(cPoint);

			 // Create initial arrow shape.
			 GeneralPath path = new GeneralPath();
			 path.moveTo(startPoint.x, startPoint.y);
			 path.lineTo(aPoint.x, aPoint.y);
			 path.lineTo(bPoint.x, bPoint.y);
			 path.lineTo(cPoint.x, cPoint.y);
			 path.lineTo(startPoint.x, startPoint.y);
			 g2d.draw(path);
		 }
		 else {
			 g.drawLine(startPoint.x, startPoint.y, endPoint.x, endPoint.y);
		 }
	 }

	 /**
	  * Draws an arrow from the (x1,y1) coordinate to the (x2,y2) coordinate with the specified
	  * properties.
	  *
	  * @param g       the graphics context
	  * @param x1      starting x coordinate of the arrow
	  * @param y1      starting y coordinate of the arrow
	  * @param x2      ending x coordinate of the arrow
	  * @param y2      ending y coordinate of the arrow
	  * @param width   width of the arrow
	  * @param style   style of the arrow
	  * @param dble    indicates a double arrow
	  * @param outline outlines the arrow
	  * @param fill	  fills the arrow
	  */
	 protected void drawArrow(Graphics g, int x1, int y1, int x2, int y2, int width,
			 LyMapLink.Style style, boolean dble, boolean outline, boolean fill)
	 {
		 // Keep the color in memory
		 Color		  color = g.getColor();
		 LyEnvironment environment = getEnvironment();
		 Point[]		  polygon = LySwingUtils.getPolygon(x1, y1, x2, y2, width + 5);
		 double		  lineLength = Math.hypot(x2 - x1, y2 - y1);
		 double		  ratio = lineLength / 10;
		 // Set the Graphic2D object
		 Graphics2D g2d = (Graphics2D) g;
		 g2d.setStroke(new BasicStroke(1,
				 BasicStroke.CAP_SQUARE,BasicStroke.JOIN_ROUND));

		 // Calculate the base of the pointer
		 int			  xc = (x1 == x2 ? x1 : (int)((ratio * x1 + x2) / (ratio + 1)));
		 int 		  yc = (y1 == y2 ? y1 : (int)((ratio * y1 + y2) / (ratio + 1)));

		 // Calculate the 2 other extremities of the pointer
		 int 		  xa = xc + (int)((polygon[0].x - x1) * 1.5);
		 int 		  ya = yc + (int)((polygon[0].y - y1) * 1.5);
		 int 		  xb = xc + (int)((polygon[3].x - x1) * 1.5);
		 int 		  yb = yc + (int)((polygon[3].y - y1) * 1.5);

		 // Get the polygon that represents the body of the arrow
		 Point[]		polygon2 = LySwingUtils.getPolygon(xc, yc, x2, y2, width);
		 int[]		  xPoints;
		 int[]		  yPoints;

		 if (style == LyMapLink.Style.LINE)
		 {
			 if (dble)
			 {
				 // Draw the pointer
				 g.drawLine(x1, y1, xa, ya);
				 g.drawLine(x1, y1, xb, yb);

				 g.drawLine(polygon2[0].x, polygon2[0].y, xa, ya);
				 g.drawLine(polygon2[3].x, polygon2[3].y, xb, yb);

				 // Draw the body
				 g.drawLine(polygon2[0].x, polygon2[0].y, polygon2[1].x, polygon2[1].y);
				 g.drawLine(polygon2[2].x, polygon2[2].y, polygon2[3].x, polygon2[3].y);
			 }
			 else
			 {
				 // Draw the pointer
				 xPoints = new int[] { x1, xa, xb };
				 yPoints = new int[] { y1, ya, yb };

				 if (fill)
					 g.fillPolygon(xPoints, yPoints, 3);

				 // Draw the body
				 if (width > 1)
				 {
					 if (outline)
					 {
						 g.setColor(getOutlineColor(environment));

						 // Draw the pointer
						 g.drawLine(x1, y1, xa, ya);
						 g.drawLine(x1, y1, xb, yb);

						 g.drawLine(polygon2[0].x, polygon2[0].y, xa, ya);
						 g.drawLine(polygon2[3].x, polygon2[3].y, xb, yb);

						 // Draw the body
						 g.drawLine(polygon2[0].x, polygon2[0].y, polygon2[1].x, polygon2[1].y);
						 g.drawLine(polygon2[2].x, polygon2[2].y, polygon2[3].x, polygon2[3].y);

						 g.setColor(color);
					 }

					 xPoints = new int[]
							 {
							 polygon2[0].x,
							 polygon2[1].x,
							 polygon2[2].x,
							 polygon2[3].x
							 };
					 yPoints = new int[]
							 {
							 polygon2[0].y,
							 polygon2[1].y,
							 polygon2[2].y,
							 polygon2[3].y
							 };

					 if (fill)
						 g.fillPolygon(xPoints, yPoints, 4);

					 if (outline)
					 {
						 g.setColor(getOutlineColor(environment));

						 // Draw the pointer
						 g.drawLine(x1, y1, xa, ya);
						 g.drawLine(x1, y1, xb, yb);

						 g.drawLine(polygon2[0].x, polygon2[0].y, xa, ya);
						 g.drawLine(polygon2[3].x, polygon2[3].y, xb, yb);

						 // Draw the body
						 g.drawLine(polygon2[0].x, polygon2[0].y, polygon2[1].x, polygon2[1].y);
						 g.drawLine(polygon2[2].x, polygon2[2].y, polygon2[3].x, polygon2[3].y);

						 g.setColor(color);
					 }
				 }
			 }
		 }
		 else
		 {
			 int length = 1;

			 if (style == LyMapLink.Style.DOT) {
				 length = 1;
				 float[] styleLine = {1,8}; /** les pointills seront gros que les blancs */
				 g2d.setStroke( new BasicStroke( 
						 width,
						 BasicStroke.CAP_ROUND,
						 BasicStroke.JOIN_MITER,
						 10.0f,
						 styleLine,
						 0
						 ));

			 }
			 else if (style == LyMapLink.Style.DASHED) {
				 length = 5;
				 float[] styleLine = {10,5}; /** les tirets seront 2 fois plus long que les blancs */
				 g2d.setStroke( new BasicStroke( 
						 width,
						 BasicStroke.CAP_BUTT,
						 BasicStroke.JOIN_BEVEL,
						 10.0f,
						 styleLine,
						 0
						 ));
			 }

			 if (dble)
			 {
				 // Draw the pointer
				 drawDashedLine(g, x1, y1, xa, ya, length);
				 drawDashedLine(g, x1, y1, xb, yb, length);

				 drawDashedLine(g, polygon2[0].x, polygon2[0].y, xa, ya, length);
				 drawDashedLine(g, polygon2[3].x, polygon2[3].y, xb, yb, length);

				 // Draw the body
				 drawDashedLine(g, polygon2[0].x, polygon2[0].y, polygon2[1].x, polygon2[1].y,
						 length);
				 drawDashedLine(g, polygon2[2].x, polygon2[2].y, polygon2[3].x, polygon2[3].y,
						 length);
			 }
			 else
			 {
				 xPoints = new int[] { x1, xa, xb };
				 yPoints = new int[] { y1, ya, yb };

				 // Draw the pointer
				 if (fill)
					 g.fillPolygon(xPoints, yPoints, 3);								
			 }

		 }


		 g.drawLine(xc, yc, x2, y2);

		 // Reset g properties
		 g2d = (Graphics2D)g;
		 g2d.setStroke(new BasicStroke(1,
				 BasicStroke.CAP_SQUARE,BasicStroke.JOIN_ROUND));
	 } // end method drawArrow

	 /**
	  * Draws a circle from the (x1,y1) coordinate to the (x2,y2) coordinate with the specified
	  * properties.
	  *
	  * @param g       the graphics context
	  * @param x1      starting x coordinate of the circle
	  * @param y1      starting y coordinate of the circle
	  * @param x2      ending x coordinate of the circle
	  * @param y2      ending y coordinate of the circle
	  * @param width   width of the circle
	  * @param style   style of the circle
	  * @param dble    indicates a double circle
	  * @param outline outlines the circle
	  * @param fill	  fills the circle
	  */
	 protected void drawCircle(Graphics g, int x1, int y1, int x2, int y2, int width,
			 LyMapLink.Style style, boolean dble, boolean outline, boolean fill)
	 {
		 // Keep the color in memory
		 Color		  color = g.getColor();
		 double		  lineLength = Math.hypot(x2 - x1, y2 - y1);
		 double		  ratio = lineLength / 10;

		 // Calculate the base of the pointer
		 int			  xc = (int)((ratio * x1 + x2) / (ratio + 1));
		 int 		  yc = (int)((ratio * y1 + y2) / (ratio + 1));
		 // Get the polygon that represents the body of the arrow
		 Point[]		  polygon2 = LySwingUtils.getPolygon(xc, yc, x2, y2, width);

		 LyEnvironment environment = getEnvironment();

		 int x = (x1 + polygon2[0].x) / 2;
		 int y = (y1 + polygon2[0].y) / 2;

		 // compute center
		 int xWidth = Math.abs((x1 - polygon2[0].x) / 2);
		 int yWidth = Math.abs((y1 - polygon2[0].y) / 2);
		 int rayon = (int)Math.hypot(xWidth, yWidth) + 1;

		 x = x - rayon;
		 y = y - rayon;
		 if (fill)
			 g.fillOval( x , y, 2 * rayon, 2 * rayon);
		 if (outline)
		 {
			 g.setColor(getOutlineColor(environment));
			 g.drawOval( x , y, 2 * rayon, 2 * rayon);
			 g.setColor(color);
		 }
		 g.drawLine(xc, yc, x2, y2);

	 }

	 /**
	  * Draws a square from the (x1,y1) coordinate to the (x2,y2) coordinate with the specified
	  * properties.
	  *
	  * @param g       the graphics context
	  * @param x1      starting x coordinate of the square
	  * @param y1      starting y coordinate of the square
	  * @param x2      ending x coordinate of the square
	  * @param y2      ending y coordinate of the square
	  * @param width   width of the square
	  * @param style   style of the square
	  * @param dble    indicates a double square
	  * @param outline outlines the square
	  * @param fill	  fills the square
	  */
	 protected void drawSquare(Graphics g, int x1, int y1, int x2, int y2, int width,
			 LyMapLink.Style style, boolean dble, boolean outline, boolean fill)
	 {
		 Color		  color = g.getColor();
		 Point[]		  polygon = LySwingUtils.getPolygon(x1, y1, x2, y2, width + 5);
		 double		  lineLength = Math.hypot(x2 - x1, y2 - y1);
		 double		  ratio = lineLength / 10;

		 // Calculate the base of the pointer
		 int			  xc = (int)((ratio * x1 + x2) / (ratio + 1));
		 int 		  yc = (int)((ratio * y1 + y2) / (ratio + 1));

		 // Calculate the 2 other extremities of the pointer
		 int 		  xa = xc + (int)((polygon[0].x - x1) * 1.5);
		 int 		  ya = yc + (int)((polygon[0].y - y1) * 1.5);
		 int 		  xb = xc + (int)((polygon[3].x - x1) * 1.5);
		 int 		  yb = yc + (int)((polygon[3].y - y1) * 1.5);

		 // Get the polygon that represents the body of the arrow
		 Point[]		  polygon2 = LySwingUtils.getPolygon(xc, yc, x2, y2, width);

		 LyEnvironment environment = getEnvironment();

		 // compute center
		 int xWidth = (x1 - polygon2[0].x) / 2;
		 int yWidth = (y1 - polygon2[0].y) / 2;

		 Polygon p = new Polygon();
		 p.addPoint(x1, y1);
		 p.addPoint(xa + xWidth, ya + yWidth);
		 p.addPoint(xc, yc);
		 p.addPoint(xb + xWidth, yb + yWidth);

		 if (fill)
			 g.fillPolygon(p);
		 if (outline)
		 {
			 g.setColor(getOutlineColor(environment));
			 g.drawPolygon(p);
			 g.setColor(color);
		 }

		 g.drawLine(xc, yc, x2, y2);
	 }

	 /**
	  * Draws a link that starts and ends on the given points.
	  * Both points belong to the same node, therefore the link must implement a 
	  * loop
	  * 
	  * @param g       the graphics context
	  * @param x1      starting x coordinate of the arc
	  * @param y1      starting y coordinate of the arc
	  * @param x2      passing point x coordinate
	  * @param y2      passing point y coordinate
	  * @param width   width of the square
	  * @param step    step used to draw the square
	  * @param style   style of the square
	  * @param dble    indicates a double square
	  * @param outline outlines the square
	  * @param fill	  fills the square
	  */
	 protected void drawAutoLink(LyAbstractMapComponentManager map, Graphics g,
			 int x1, int y1, int maxY1, int x2, int y2, int maxX2, int step, int width, LyMapLink.Style style,
			 boolean dble, boolean outline)
	 {
		 Graphics2D g2d = (Graphics2D) g;
		 g2d.setStroke(new BasicStroke(1,BasicStroke.CAP_SQUARE,BasicStroke.JOIN_ROUND));

		 //Calculate point to draw the square
		 int xa = x1;
		 int ya = y1 + step ;
		 int xb = x2 - step ;
		 int yb = ya ;
		 int xc = x2 - step ;
		 int yc = y2 ;

		 // Create initial arrow shape.
		 GeneralPath path = new GeneralPath();
		 path.moveTo(x1, maxY1);
		 path.lineTo(xa, ya);
		 path.lineTo(xb, yb);
		 path.lineTo(xc, yc);
		 path.lineTo(maxX2, y2);

		 if(dble)
		 {
			 width *= 2;      
		 }

		 // set the line style
		 if (style == LyMapLink.Style.LINE) {
			 if(width > 1) {
				 g2d.setStroke(new BasicStroke(width,BasicStroke.CAP_SQUARE,BasicStroke.JOIN_ROUND));
			 }        
		 }
		 else if (style == LyMapLink.Style.DOT) {

			 float[] styleLine = {1,8}; /** les pointills seront gros que les blancs */
			 g2d.setStroke( new BasicStroke( 
					 width,
					 BasicStroke.CAP_ROUND,
					 BasicStroke.JOIN_MITER,
					 10.0f,
					 styleLine,
					 0
					 ));

		 }
		 else if (style == LyMapLink.Style.DASHED) {

			 float[] styleLine = {10,5}; /** les tirets seront 2 fois plus long que les blancs */
			 g2d.setStroke( new BasicStroke( 
					 width,
					 BasicStroke.CAP_BUTT,
					 BasicStroke.JOIN_BEVEL,
					 10.0f,
					 styleLine,
					 0
					 ));
		 }

		 g2d.draw(path);
	 }

	 /**
	  * The specified point is added in the view layer.
	  *
	  * @see LyGraphicalMapLink#addPoint(leon.view.LyPoint, int)
	  */
	 public void addPoint(LyPoint point, int pos)
	 {
		 // The point was added in the view layer (notified by leon.view package)
	 }

	 /**
	  * The specified point is removed in the view layer.
	  *
	  * @see LyGraphicalMapLink#removePoint(LyPoint)
	  */
	 public void removePoint(LyPoint point)
	 {
		 // The point was added in the view layer (notified by leon.view package)
	 }

	 /**
	  * Adds a new point in the view at the given index position.
	  *
	  * @param point new point inserted
	  * @param idx   index of the point
	  */
	 public void addNewPoint(LyPoint point, int idx)
	 {
		 _mapLink.insertPoint(point, idx);
	 }

	 /**
	  * Clears all points in the view elements.
	  */
	 public void removeAllPoints()
	 {
		 int i, n = _mapLink.getPointCount();

		 for (i = n - 1; i >= 0; i--)
			 _mapLink.removePoint(_mapLink.getPoint(i));
	 }

	 /*
	  * (non-Javadoc).
	  *
	  * @see leon.view.getinterface.LyGetIdInterface#getId()
	  */
	 public String getId()
	 {
		 return _mapLink.getId();
	 }

	 /**
	  * Gets the bounding box of this link in the 3D space. The bounding box is already in 3D, so the
	  * link points doesn't need to be converted to the 2D space with the specified map manager.
	  *
	  * @param  map the map manager of this link. Unused in this method.
	  * @return the bounding box of this map link in the 3D space
	  * @see    leon.view.win.swing.map.LyMapElement#getBoundingBox(leon.view.win.swing.map.LyAbstractMapComponentManager)
	  */
	 public Ly3DBox getBoundingBox(LyAbstractMapComponentManager map)
	 {
		 LyPoint start3D = new LyPoint();
		 LyPoint end3D = new LyPoint();

		 if (_startNode != null)
			 start3D = _startNode.getCenter();
		 if (_endNode != null)
			 end3D = _endNode.getCenter();

		 LyPoint moved = getDisplacedValues(map, start3D, end3D);

		 LyPoint newStart = new LyPoint(start3D.getX(), start3D.getY(), start3D.getZ());
		 LyPoint newEnd = new LyPoint(end3D.getX(), end3D.getY(), end3D.getZ());
		 newStart.translate(moved.getX(), moved.getY(), 0);
		 newEnd.translate(moved.getX(), moved.getY(), 0);

		 return new Ly3DBox(newStart, newEnd);
	 }

	 /*
	  * (non-Javadoc).
	  *
	  * @see leon.view.win.swing.map.LyMapElement#getBoundingRectangle(leon.view.win.swing.map.LyAbstractMapComponentManager)
	  */
	 public Rectangle getBoundingRectangle(LyAbstractMapComponentManager map)
	 {
		 LyPoint start3D = new LyPoint();
		 LyPoint end3D = new LyPoint();

		 if (_startNode != null)
			 start3D = _startNode.getCenter();
		 if (_endNode != null)
			 end3D = _endNode.getCenter();

		 LyPoint moved = getDisplacedValues(map, start3D, end3D);

		 LyPoint newStart = new LyPoint(start3D.getX(), start3D.getY(), start3D.getZ());
		 LyPoint newEnd = new LyPoint(end3D.getX(), end3D.getY(), end3D.getZ());
		 newStart.translate(moved.getX(), moved.getY(), 0);
		 newEnd.translate(moved.getX(), moved.getY(), 0);

		 Point start = map.convert(newStart);
		 Point end = map.convert(newEnd);
		 int   x1, y1;

		 x1 = ((end.x < start.x) ? end.x : start.x);
		 y1 = ((end.y < start.y) ? end.y : start.y);

		 return new Rectangle(x1, y1, Math.abs(end.x - start.x), Math.abs(end.y - start.y));
	 }

	 /*
	  * (non-Javadoc).
	  *
	  * @see leon.view.win.swing.map.LyMapElement#getClippingArea(leon.view.win.swing.map.LyAbstractMapComponentManager,
	  *      java.awt.Image)
	  */
	 public Shape getClippingArea(LyAbstractMapComponentManager map, Image image)
	 {
		 Image[] images = getImages();

		 if ((image != null) && (images != null))
		 {

			 int   xImage;
			 int   yImage;
			 int   width;
			 int   height;

			 Image imageStart = images[0];
			 Image imageMiddle = images[1];
			 Image imageEnd = images[2];
			 Point imageStartPos = _imagePos[0];
			 Point imageMiddlePos = _imagePos[1];
			 Point imageEndPos = _imagePos[2];

			 if ((imageStart != null) && (image == imageStart) && (imageStartPos != null))
			 {
				 width = imageStart.getWidth(map.getComponent());
				 height = imageStart.getHeight(map.getComponent());
				 xImage = _imagePos[0].x; // (x1 + firstSectionX) / 2 - width /2;
				 yImage = _imagePos[0].y; // (y1 + firstSectionY) / 2 - height / 2;

				 return new Rectangle(xImage - 1 - (width / 2), yImage - 1 - (height / 2), width + 2,
						 height + 2);
			 }

			 if ((imageMiddle != null) && (image == imageMiddle) && (imageMiddlePos != null))
			 {
				 width = imageMiddle.getWidth(map.getComponent());
				 height = imageMiddle.getHeight(map.getComponent());
				 xImage = _imagePos[1].x; // (secondSectionX + firstSectionX) / 2 - width /2;
				 yImage = _imagePos[1].y; // (secondSectionY + firstSectionY) / 2 - height / 2;

				 return new Rectangle(xImage - 1 - (width / 2), yImage - 1 - (height / 2), width + 2,
						 height + 2);
			 }

			 if ((imageEnd != null) && (image == imageEnd) && (imageEndPos != null))
			 {
				 width = imageEnd.getWidth(map.getComponent());
				 height = imageEnd.getHeight(map.getComponent());
				 xImage = _imagePos[2].x; // (secondSectionX + x2) / 2 - width /2;
				 yImage = _imagePos[2].y; // (secondSectionY + y2) / 2 - height / 2;

				 return new Rectangle(xImage - 1 - (width / 2), yImage - 1 - (height / 2), width + 2,
						 height + 2);
			 }
		 }
		 else
			 return getBoundingRectangle(map);

		 return null;
	 } // end method getClippingArea

	 /*
	  * (non-Javadoc).
	  *
	  * @see leon.view.win.swing.map.LyMapElement#getLayer()
	  */
	 public int getLayer()
	 {
		 int layer = _mapLink.getLayer();

		 return ((layer < 0) ? LyAbstractMapComponentManager.LINK_LAYER : layer);
	 }

	 /*
	  * (non-Javadoc).
	  *
	  * @see leon.view.win.swing.map.LyMapElement#getLabel()
	  */
	 public String getLabel()
	 {
		 if (_labelCache == null)
			 _labelCache = translate(_mapLink.getLabel());

		 return _labelCache;
	 }

	 /**
	  * (non-Javadoc).
	  *
	  * @see leon.view.win.swing.map.LyMapElement#getLabel()
	  */
	 public String getStartLabel()
	 {
		 if (_startLabelCache == null)
			 _startLabelCache = translate(_mapLink.getStartLabel());

		 return _startLabelCache;
	 }

	 /**
	  * (non-Javadoc).
	  *
	  * @see leon.view.win.swing.map.LyMapElement#getLabel()
	  */
	 public String getEndLabel()
	 {
		 if (_endLabelCache == null)
			 _endLabelCache = translate(_mapLink.getEndLabel());

		 return _endLabelCache;
	 }

	 /**
	  * Sets the label attribute to <code>null</code> in order to get rid of cache.
	  *
	  * @param label This attribute is not used in this implementation.
	  * @see   #getLabel()
	  */
	 public void setLabel(String label)
	 {
		 // Get rid of cache
		 _labelCache = null;
	 }

	 /*
	  * (non-Javadoc).
	  *
	  * @see leon.view.win.swing.map.LyMapElement#getTooltip()
	  */
	 public String getTooltip()
	 {
		 return translate(_mapLink.getTooltip());
	 }

	 /*
	  * (non-Javadoc).
	  *
	  * @see leon.view.win.swing.map.LyMapElement#isAtPosition(leon.view.win.swing.map.LyAbstractMapComponentManager,
	  *      java.awt.Graphics, int, int)
	  */
	 public boolean isAtPosition(LyAbstractMapComponentManager map, Graphics g,
			 int x, int y) {

		 if ((_startNode == null) || (_endNode == null)
				 || (getLinkWidth() == 0) || _mapLink.isHidden())
			 return false;

		 Rectangle	scrollRectangle = map.getScrollRectangle();

		 x -= scrollRectangle.x;
		 y -= scrollRectangle.y;

		 // Compute the link start and endPoint. 
		 // The link will be drawn to link those two points
		 Point startPoint = getStartLinkPoint(map, true);
		 Point endPoint = getEndLinkPoint(map, true);

		 Rectangle endBoundingBox = _endNode.getBoundingRectangle(map);
		 Rectangle startBoundingBox = _startNode.getBoundingRectangle(map);

		 if (endBoundingBox.contains(map.convert(_startNode.getCenter()))
				 || startBoundingBox.contains(map.convert(_endNode.getCenter()))) {
			 // There is an overlap between the start and end nodes. That means that
			 // the link has been drawn through the DrawAutoLink method.
			 boolean contains = false;
			 // TODO Here we suppose the point is located over the node, depending on
			 // the node radius

			 int x1 = startPoint.x;
			 int y1 = startBoundingBox.y + startBoundingBox.height;
			 int x2 = endBoundingBox.x;
			 int y2 = endPoint.y;

			 // Get the step
			 int step = (int) (LOOP_STEP * map.getZoom());

			 // Calculate point to draw the square
			 int xa = x1;
			 int ya = startPoint.y + step;
			 int xb = endPoint.x - step;
			 int yb = ya;
			 int xc = endPoint.x - step;
			 int yc = y2;

			 final int polygonWidth = getLinkWidth() + SELECTION_OFFSET;

			 // Build a polygon that encloses the link, it has following form
			 //
			 //    P5---------P6
			 //    | Pc-------P2
			 //    | | P8-----P7
			 //    | | |
			 //    | | |           P11P1P2
			 //    | | |            | | |
			 //    | | P9----------P10| |
			 //    | Pb---------------Pa|
			 //    P4-------------------P3
			 //
			 // The link (P1, Pa, Pb, Pc, P2) is enclosed by the 
			 // (P2, P3, P4, P5, P6, P7, P8, P9, P10, P11) polygon.

			 int[] polygonXPoints = new int[] {
					 x1 + polygonWidth, // P2
					 x1 + polygonWidth, // P3
					 xb - polygonWidth, // P4
					 xc - polygonWidth, // P5
					 x2,                // P6
					 x2,                // P7
					 xc + polygonWidth, // P8
					 xb + polygonWidth, // P9
					 xa - polygonWidth, // P10
					 x1 - polygonWidth  // P11
			 };
			 int[] polygonYPoints = new int[] {
					 y1               , // P2
					 ya + polygonWidth, // P3
					 yb + polygonWidth, // P4
					 yc - polygonWidth, // P5
					 y2 - polygonWidth, // P6
					 y2 + polygonWidth, // P7
					 yc + polygonWidth, // P8
					 yb - polygonWidth, // P9
					 ya - polygonWidth, // P10
					 y1                 // P11
			 };
			 Polygon enclosingPolygon = new Polygon(polygonXPoints,
					 polygonYPoints, polygonXPoints.length);

			 // Now test if the point(x,y) is in the selection area
			 // (ie: in the external rectangle and not into the internal rectangle)
			 if (enclosingPolygon.contains(x, y)) {
				 contains = true;
			 } else {
				 // Test also if the point(x,y) is in the label rectangles or not
				 String label = getLabel();
				 if (label != null && !contains) {
					 Dimension labelDim = getLabelSize(map, g);

					 x1 = xb - labelDim.width;
					 y1 = yb + 4 * getLinkWidth();
					 Rectangle rectLabel = new Rectangle(x1, y1, labelDim.width,
							 labelDim.height);
					 contains = rectLabel.contains(x, y);
				 }
				 if (!contains)  {
					 // Test also if the point(x,y) is in the label rectangles or not
					 String startLabel = getStartLabel();
					 if (startLabel != null && !contains) {
						 Dimension labelDim = getStartLabelSize(map, g);

						 x1 = xa + 4 * getLinkWidth();
						 y1 = ya + 4 * getLinkWidth();
						 Rectangle rectLabel = new Rectangle(x1, y1, labelDim.width,
								 labelDim.height);
						 contains = rectLabel.contains(x, y);
					 }
				 }

				 if (!contains)  {
					 // Test also if the point(x,y) is in the label rectangles or not
					 String endLabel = getEndLabel();
					 if (endLabel != null && !contains) {
						 Dimension labelDim = getEndLabelSize(map, g);

						 x1 = xc - labelDim.width;
						 y1 = yc - labelDim.height - 4 * getLinkWidth();
						 Rectangle rectLabel = new Rectangle(x1, y1, labelDim.width,
								 labelDim.height);
						 contains = rectLabel.contains(x, y);
					 }
				 }
			 }
			 return contains;
		 }
		 else {
			 Point tmp = new Point(0, 0);
			 boolean contains = false;

			 int x1 = startPoint.x;
			 int y1 = startPoint.y;
			 int x2 = endPoint.x;
			 int y2 = endPoint.y;

			 tmp.x = x1;
			 tmp.y = y1;

			 Polygon poly;
			 Point[] polygon;
			 int[] xPoints, yPoints;

			 // If there are intermediate points, polygon is modified
			 for (int i = 0; i < _mapLink.getPointCount(); i++) {
				 Point aPoint = map.convert(_mapLink.getPoint(i));

				 polygon = LySwingUtils.getPolygon(tmp.x, tmp.y, aPoint.x, aPoint.y,
						 getLinkWidth() + SELECTION_OFFSET);
				 xPoints = new int[] { polygon[0].x, polygon[1].x, polygon[2].x,
						 polygon[3].x };
				 yPoints = new int[] { polygon[0].y, polygon[1].y, polygon[2].y,
						 polygon[3].y };
				 poly = new Polygon(xPoints, yPoints, 4);
				 contains = poly.contains(x, y) ? true : contains;
				 tmp.x = aPoint.x;
				 tmp.y = aPoint.y;
			 }
			 polygon = LySwingUtils.getPolygon(tmp.x, tmp.y, x2, y2, getLinkWidth()
					 + SELECTION_OFFSET);
			 xPoints = new int[] { polygon[0].x, polygon[1].x, polygon[2].x,
					 polygon[3].x };
			 yPoints = new int[] { polygon[0].y, polygon[1].y, polygon[2].y,
					 polygon[3].y };
			 poly = new Polygon(xPoints, yPoints, 4);
			 contains = poly.contains(x, y) ? true : contains;

			 String label = getLabel();
			 Dimension labelDim = getLabelSize(map, g);
			 Rectangle rectLabel = new Rectangle();

			 if (label != null)
				 rectLabel = new Rectangle((x1 + x2 - labelDim.width) / 2, (y1 + y2
						 + getLinkWidth() + SELECTION_OFFSET) / 2, labelDim.width,
						 labelDim.height);

			 return (contains || rectLabel.contains(x, y));
		 }
	 } // end method isAtPosition

	 /*
	  * (non-Javadoc).
	  * 
	  * @see leon.view.setinterface.LySetSelectedInterface#setSelected(boolean)
	  */
	 public void setSelected(boolean selected)
	 {
		 if (_mapLink.isSensitive())
			 _mapLink.setSelected(selected);
	 }

	 /*
	  * (non-Javadoc).
	  *
	  * @see leon.view.win.swing.map.LyMapElement#isSelected()
	  */
	 public boolean isSelected()
	 {
		 return _mapLink.isSelected();
	 }

	 /**
	  * Don't move this object.
	  *
	  * @param dx the x delta, not used in this implementation
	  * @param dy the y delta, not used in this implementation
	  * @param dz the z delta, not used in this implementation
	  * @see   leon.view.win.swing.map.LyMapElement#translate(float, float, float)
	  */
	 public void translate(float dx, float dy, float dz)
	 {
	 }

	 /*
	  * (non-Javadoc).
	  *
	  * @see leon.view.win.swing.map.LyMapElement#isOverlay()
	  */
	 public boolean isOverlay()
	 {
		 if ((_startNode == null) || (_endNode == null))
			 return false;

		 if (_mapLink.isSelected() || _startNode.isOverlay() || _endNode.isOverlay())
			 return true;
		 return false;
	 }


	 /*
	  * (non-Javadoc).
	  *
	  * @see leon.view.win.swing.map.LyMapElement#isHidden()
	  */
	 public boolean isHidden()
	 {
		 return _mapLink.isHidden();
	 }

	 /*
	  * (non-Javadoc).
	  *
	  * @see leon.view.setinterface.LySetHiddenInterface#setHidden(boolean)
	  */
	 public void setHidden(boolean hidden)
	 {
		 if (_mapLink.isHidden() != hidden)
			 _mapLink.setHidden(hidden);
	 }

	 /*
	  * (non-Javadoc).
	  *
	  * @see leon.view.win.swing.map.LyMapElement#isSensitive()
	  */
	 public boolean isSensitive()
	 {
		 return _mapLink.isSensitive();
	 }

	 /**
	  * Draws a dashed line in the given graphics using the given coordinates.
	  *
	  * @param g      the graphics context
	  * @param x1     the x coordinate of the starting point
	  * @param y1     the y coordinate of the starting point
	  * @param x2     the x coordinate of the ending point
	  * @param y2     the y coordinate of the ending point
	  * @param length size of a dash
	  * @see   #drawDashedLine(Graphics, int, int, int, int, int, Image)
	  */
	 protected void drawDashedLine(Graphics g, int x1, int y1, int x2, int y2, int length)
	 {
		 drawDashedLine(g, x1, y1, x2, y2, length, null);
	 }

	 /**
	  * Draws a dashed line in the given graphics using the given coordinates.
	  *
	  * @param g      the graphics context
	  * @param x1     the x coordinate of the starting point
	  * @param y1     the y coordinate of the starting point
	  * @param x2     the x coordinate of the ending point
	  * @param y2     the y coordinate of the ending point
	  * @param length size of a dash
	  * @param image  image to use on the dash, if null, a line is drawn instead
	  */
	 protected void drawDashedLine(Graphics g, int x1, int y1, int x2, int y2, int length,
			 Image image)
	 {
		 int     width = (x2 - x1);
		 int     height = (y2 - y1);
		 int     invert = 1;

		 int     maxDim = Math.max(Math.abs(width), Math.abs(height));
		 boolean horizontalParsing = (maxDim == Math.abs(width));
		 int     i, coord, coord2;
		 int     draw = 0;


		 // May be factorized
		 if (horizontalParsing)
		 {
			 if (x2 < x1)
				 invert = -1;

			 for (i = 0; i < maxDim; i += length)
			 {
				 coord = (int)((invert * i * ((float)height / width)) + y1);
				 coord2 = (int)((invert * (i + (length / 2)) * ((float)height / width)) + y1);

				 if ((draw % 3) == 0)
				 {
					 if (image != null)
					 {
						 int imageWidth = image.getWidth(null);
						 int imageHeight = image.getHeight(null);

						 g.drawImage(image, x1 + (invert * i) - (imageWidth / 2),
								 coord
								 - (imageHeight
										 / 2), imageWidth, imageHeight, null);
					 }
					 else
						 g.drawLine(x1 + (invert * i), coord, x1 + (invert * (i + (length / 2))),
								 coord2);
				 }

				 draw++;
			 }
		 }
		 else
		 {
			 if (y2 < y1)
				 invert = -1;

			 for (i = 0; i < maxDim; i += length)
			 {
				 coord = (int)((invert * i * ((float)width / height)) + x1);
				 coord2 = (int)((invert * (i + (length / 2)) * ((float)width / height)) + x1);

				 if ((draw % 3) == 0)
				 {
					 if (image != null)
					 {
						 int imageWidth = image.getWidth(null);
						 int imageHeight = image.getHeight(null);

						 g.drawImage(image, coord - (imageWidth / 2),
								 y1 + (invert * i)
								 - (imageHeight
										 / 2), imageWidth, imageHeight, null);
					 }
					 else
						 g.drawLine(coord, y1 + (invert * i), coord2,
								 y1 + (invert * (i + (length / 2))));
				 }

				 draw++;
			 }
		 }
	 } // end method drawDashedLine

	 /**
	  * Sets the start node of this link.
	  *
	  * @param node value of the starting node
	  * @see   #getStartNode()
	  */
	 public void setStartNode(LySwingMapNode node)
	 {
		 _startNode = node;
	 }

	 /**
	  * Sets the end node of this link.
	  *
	  * @param node value of the ending node
	  * @see   #getEndNode()
	  */
	 public void setEndNode(LySwingMapNode node)
	 {
		 _endNode = node;
	 }

	 /**
	  * Gets the start node of this link.
	  *
	  * @return the starting node
	  * @see    #setStartNode(LySwingMapNode)
	  */
	 public LySwingMapNode getStartNode()
	 {
		 return _startNode;
	 }

	 /**
	  * Gets the end node of this link.
	  *
	  * @return the ending node
	  * @see    #setEndNode(LySwingMapNode)
	  */
	 public LySwingMapNode getEndNode()
	 {
		 return _endNode;
	 }

	 /*
	  * (non-Javadoc)
	  * @see leon.view.graphic.LyGraphicalMapLink#setColorStart(java.lang.String)
	  */
	 public void setColorStart(String colorStart)
	 {
		 _mapLink.setColorStart(colorStart);
	 }

	 /*
	  * (non-Javadoc)
	  * @see leon.view.graphic.LyGraphicalMapLink#setColor(java.lang.String)
	  */
	 public void setColor(String color)
	 {
		 _mapLink.setColor(color);
	 }

	 /*
	  * (non-Javadoc)
	  * @see leon.view.graphic.LyGraphicalMapLink#setColorEnd(java.lang.String)
	  */
	 public void setColorEnd(String colorEnd)
	 {
		 _mapLink.setColorEnd(colorEnd);
	 }

	 /*
	  * (non-Javadoc)
	  * @see leon.view.graphic.LyGraphicalMapLink#setFontColor(java.lang.String)
	  */
	 public void setFontColor(String labelColor)
	 {
		 _mapLink.setFontColor(labelColor);
	 }

	 /**
	  * Set the left label and clear the label cache.
	  * @see leon.view.graphic.LyGraphicalMapLink#setEndLabel(java.lang.String)
	  */
	 public void setEndLabel(String leftLabel)
	 {
		 _endLabelCache = null;

	 }

	 /**
	  * Set the right label and clear the label cache.
	  * @see leon.view.graphic.LyGraphicalMapLink#setStartLabel(java.lang.String)
	  */
	 public void setStartLabel(String rightLabel)
	 {
		 _startLabelCache = null;

	 }

	 /**
	  * Finds the position of the starting arrow of the link.
	  * 
	  * @param map the map that is drawn
	  * @param onBoundingRectangle if <code>onBoundingRectangle</code>, the 
	  * starting point is the intersection of the (startNode, endNode) line and
	  * the startNode boundingRectangle, otherwise it is the intersection of the
	  * (startNode, endNode) line and the circle around the startNode.
	  * 
	  * @return a LyPoint containing the point where the arrow starts, or <code>null</code> if one of
	  * the node is <code>null</code>.
	  */
	 public Point getStartLinkPoint(LyAbstractMapComponentManager map, boolean onBoundingRectangle) {

		 if ((_startNode == null) || (_endNode == null))
			 return null;

		 Point[] linePoints = computeDrawingLineExtremePoints(map, _startNode, _endNode, true);
		 Point startPoint = linePoints[START_POINT_IDX];
		 Point endPoint = linePoints[END_POINT_IDX];

		 double x1 = startPoint.getX();
		 double y1 = startPoint.getY();
		 double x2 = endPoint.getX();
		 double y2 = endPoint.getY();

		 // Divide the link into three parts
		 double firstSectionX = ((2 * x1) + x2) / 3;
		 double firstSectionY = ((2 * y1) + y2) / 3;

		 // Modify sections if there are intermediate points (sub-component)
		 if (_mapLink.getPointCount() >= 1) {
			 firstSectionX = map.convert(_mapLink.getPoint(0)).getX();
			 firstSectionY = map.convert(_mapLink.getPoint(0)).getY();
		 }

		 LyComponent parent = _mapLink.getParent();

		 if (parent instanceof LyMapContent)
			 parent = parent.getParent();

		 if (_mapLink.getMotif() != null) {
			 return new Point((int) x1, (int) y1);
		 }
		 else {
			 if (onBoundingRectangle)  {
				 Rectangle startBoundingBox = _startNode.getBoundingRectangle(map);
				 Rectangle scrollRect = map.getScrollRectangle();

				 if (!map.getStyle().get(LyAbstractMapComponentManager.VIEWER))
				 {
					 startBoundingBox.x = startBoundingBox.x - scrollRect.x;
					 startBoundingBox.y = startBoundingBox.y - scrollRect.y;
				 }

				 RectangleSide rectSide = getRectangleIntersectionSide(map, startBoundingBox, startPoint, endPoint);
				 Point startLinkPoint = getIntersection(startBoundingBox, rectSide, startPoint, endPoint);
				 if (startLinkPoint == null) { 
					 // There is no intersection, it means that endPoint is in the bounding rectangle
					 // Set the startLinkPoint to the center of the south border of the bounding rectangle
					 startLinkPoint = new Point(startPoint.x, startPoint.y + startBoundingBox.height/2);
				 }
				 return startLinkPoint;
			 } else {
				 double startRadius = computeNodeRadius(map, _startNode);

				 if ((_mapLink.getArrow() == LyMapLink.Arrow.START)
						 || (_mapLink.getArrow() == LyMapLink.Arrow.BOTH)
						 || _startNode.getMapNodeStyle().equals(Style.IMAGE)) {
					 double distance = Math.hypot(firstSectionX - x1, firstSectionY - y1);

					 double xArrow;
					 double yArrow;
					 Rectangle startBoundingBox = _startNode.getBoundingRectangle(map);
					 Rectangle endBoundingBox = _endNode.getBoundingRectangle(map);
					 if (endBoundingBox.contains(map.convert(_startNode.getCenter()))
							 || startBoundingBox.contains(map.convert(_endNode.getCenter()))) {
						 firstSectionX = x1;
						 firstSectionY = y1 + startRadius;

						 xArrow = firstSectionX;
						 yArrow = firstSectionY;
					 }
					 else {
						 xArrow = (int) ((((distance - startRadius) * x1) + (startRadius * firstSectionX)) / distance);
						 yArrow = (int) ((((distance - startRadius) * y1) + (startRadius * firstSectionY)) / distance);
					 }
					 return new Point((int) xArrow, (int) yArrow);
				 }
			 }
		 }

		 return map.convert(_startNode.getCenter());
	 }

	 /**
	  * Determines on which side of the <code>node1BoundingRect</code> rectangle, the
	  * link between node1 and node2 should be attached.
	  * 
	  * The side is either :
	  * <ul>
	  * <li>NORTH : above side
	  * <li>SOUTH : below side
	  * <li>WEST  : left side
	  * <li>EAST  : rigth side
	  * </ul>
	  * 
	  * @param map
	  * @param node1BoundingBox a rectangle that encapsulates <code>node1</code> and
	  *    which sides are parallel to the screen borders.
	  * @param node1 starting point of the link
	  * @param node2 ending point of the link
	  * @return
	  */
	 private RectangleSide getRectangleIntersectionSide(
			 LyAbstractMapComponentManager map,
			 Rectangle node1BoundingRect,
			 Point point1, 
			 Point point2){

		 double x1 = point1.getX();
		 double y1 = point1.getY();
		 double x2 = point2.getX();
		 double y2 = point2.getY();

		 Line2D line = new Line2D.Double(x1, y1, x2, y2);    

		 if (x2 > x1)  {
			 // The intersection can not be on the West side of the rectangle
			 // It is either on the north, east or south side
			 if (y2 > y1)    {

				 // The intersection can not be on the North side of the rectangle
				 // It is either on the east or south side

				 // To decide whether it is on the east or the south side, we check
				 // the position of the (south, east) corner of the rectangle compared
				 // to the (point1, point2) line:
				 // if the corner is above the line, the intersection side is SOUTH
				 // otherwise, it is EAST
				 double rectPointX = node1BoundingRect.getX() + node1BoundingRect.getWidth();
				 double rectPointY = node1BoundingRect.getY() + node1BoundingRect.getHeight();

				 int direction = line.relativeCCW(rectPointX, rectPointY);
				 if (direction == 1)    {
					 return RectangleSide.SOUTH;
				 } else {
					 return RectangleSide.EAST;
				 }
			 } else {  // y2 <= y1
				 // The intersection can not be on the South side of the rectangle
				 // It is either on the east or north side

				 // To decide whether it is on the east or the north side, we check
				 // the position of the (north, east) corner of the rectangle compared
				 // to the (point1, point2) line:
				 // if the corner is above the line, the intersection side is EAST
				 // otherwise, it is NORTH
				 double rectPointX = node1BoundingRect.getX() + node1BoundingRect.getWidth();
				 double rectPointY = node1BoundingRect.getY();

				 int direction = line.relativeCCW(rectPointX, rectPointY);
				 if (direction == 1)    {
					 return RectangleSide.EAST;
				 } else {
					 return RectangleSide.NORTH;
				 }     
			 }
		 } else { // x2 <= x1
			 // The intersection can not be on the east side of the rectangle
			 // It is either on the north, west or south side
			 if (y2 > y1)    {

				 // The intersection can not be on the North side of the rectangle
				 // It is either on the west or south side

				 // To decide whether it is on the west or the south side, we check
				 // the position of the (south, west) corner of the rectangle compared
				 // to the (point1, point2) line:
				 // if the corner is above the line, the intersection side is WEST
				 // otherwise, it is SOUTH
				 double rectPointX = node1BoundingRect.getX();
				 double rectPointY = node1BoundingRect.getY() + node1BoundingRect.getHeight();

				 int direction = line.relativeCCW(rectPointX, rectPointY);
				 if (direction == 1)    {
					 return RectangleSide.WEST;
				 } else {
					 return RectangleSide.SOUTH;
				 }
			 } else {  // y2 <= y1
				 // The intersection can not be on the South side of the rectangle
				 // It is either on the west or north side

				 // To decide whether it is on the west or the north side, we check
				 // the position of the (north, west) corner of the rectangle compared
				 // to the (point1, point2) line:
				 // if the corner is above the line, the intersection side is WEST
				 // otherwise, it is NORTH
				 double rectPointX = node1BoundingRect.getX();
				 double rectPointY = node1BoundingRect.getY();

				 int direction = line.relativeCCW(rectPointX, rectPointY);
				 if (direction == 1)    {
					 return RectangleSide.NORTH;
				 } else {
					 return RectangleSide.WEST;
				 }

			 }
		 }
	 }

	 /**
	  * Finds the position of the ending arrow of the link.
	  * 
	  * @param map the map that is drawn
	  * @return a LyPoint containing the point where the arrow ends, or <code>null</code> if one of
	  * the node is <code>null</code>.
	  */
	 public Point getEndLinkPoint(LyAbstractMapComponentManager map, boolean onBoundingRectangle) {

		 if ((_startNode == null) || (_endNode == null))
			 return null;

		 Point[] linePoints = computeDrawingLineExtremePoints(map, _startNode, _endNode, false);
		 Point startPoint = linePoints[START_POINT_IDX];
		 Point endPoint = linePoints[END_POINT_IDX];

		 double x1 = startPoint.getX();
		 double y1 = startPoint.getY();
		 double x2 = endPoint.getX();
		 double y2 = endPoint.getY();

		 // Divide the link into three parts
		 double secondSectionX = (x1 + (2 * x2)) / 3;
		 double secondSectionY = (y1 + (2 * y2)) / 3;

		 // Modify sections if there are intermediate points (sub-component)
		 if (_mapLink.getPointCount() >= 1) {
			 if (_mapLink.getPointCount() >= 2) {
				 int lastPoint = _mapLink.getPointCount() - 1;

				 secondSectionX = map.convert(_mapLink.getPoint(lastPoint)).getX();
				 secondSectionY = map.convert(_mapLink.getPoint(lastPoint)).getY();
			 }
			 else {
				 secondSectionX = ((2 * map.convert(_mapLink.getPoint(0)).getX()) + x2) / 3;
				 secondSectionY = ((2 * map.convert(_mapLink.getPoint(0)).getY()) + y2) / 3;
			 }
		 }

		 LyComponent parent = _mapLink.getParent();

		 if (parent instanceof LyMapContent)
			 parent = parent.getParent();

		 if (_mapLink.getMotif() != null) {
			 return new Point((int) x2, (int) y2);
		 }
		 else {
			 if (onBoundingRectangle)  {
				 Rectangle endBoundingRectangle = _endNode.getBoundingRectangle(map);
				 Rectangle scrollRect = map.getScrollRectangle();

				 if (!map.getStyle().get(LyAbstractMapComponentManager.VIEWER))
				 {
					 endBoundingRectangle.x = endBoundingRectangle.x - scrollRect.x;
					 endBoundingRectangle.y = endBoundingRectangle.y - scrollRect.y;
				 }

				 RectangleSide rectSide = getRectangleIntersectionSide(map, endBoundingRectangle, endPoint, startPoint);
				 Point endLinkPoint = getIntersection(endBoundingRectangle, rectSide, endPoint, startPoint);
				 if (endLinkPoint == null)   {
					 // There is no intersection, it means that startdPoint is in the bounding rectangle
					 // Set the endLinkPoint to the center of the west border of the bounding rectangle
					 endLinkPoint = new Point(endPoint.x - endBoundingRectangle.width/2, endPoint.y);
				 }
				 return endLinkPoint;
			 } 
			 else {

				 // Draw the third part  
				 if ((_mapLink.getArrow() == LyMapLink.Arrow.END)
						 || (_mapLink.getArrow() == LyMapLink.Arrow.BOTH)
						 || _endNode.getMapNodeStyle().equals(Style.IMAGE)) {
					 double endRadius = computeNodeRadius(map, _endNode);  
					 double distance = Math.hypot(x2 - secondSectionX, y2 - secondSectionY);

					 double xArrow;
					 double yArrow;

					 Rectangle startBoundingBox = _startNode.getBoundingRectangle(map);
					 Rectangle endBoundingBox = _endNode.getBoundingRectangle(map);
					 if (endBoundingBox.contains(map.convert(_startNode.getCenter()))
							 || startBoundingBox.contains(map.convert(_endNode.getCenter()))) {

						 secondSectionX = x2 - endRadius;
						 secondSectionY = y2;

						 xArrow = secondSectionX;
						 yArrow = secondSectionY;

					 }
					 else {
						 xArrow = (int) ((((distance - endRadius) * x2) + (endRadius * secondSectionX)) / distance);
						 yArrow = (int) ((((distance - endRadius) * y2) + (endRadius * secondSectionY)) / distance);
					 }
					 return new Point((int) xArrow, (int) yArrow);
				 }
			 }
		 }

		 return map.convert(_endNode.getCenter());
	 }

	 @Override
	 public String toString()
	 {
		 StringBuffer sb = new StringBuffer();
		 sb.append(_startNode);
		 sb.append(' ');

		 Arrow arrow = _mapLink.getArrow();
		 if (arrow.equals(Arrow.BOTH) || arrow.equals(Arrow.START))
			 sb.append('<');
		 sb.append("--");
		 if (arrow.equals(Arrow.BOTH) || arrow.equals(Arrow.END))
			 sb.append('>');
		 sb.append(' ');
		 sb.append(_endNode);
		 return sb.toString();
	 }

	 /**
	  * Get the current font metrics according to the current zoom level of the map
	  * @param map      the map representing the node
	  * @param  g           the graphics context
	  * @return
	  */
	 @SuppressWarnings("deprecation")
	 private FontMetrics getCurrentFontMetrics(LyAbstractMapComponentManager map, Graphics g)
	 {
		 FontMetrics fm;    
		 Font currentFont = getCurrentFont( map );    
		 if (g == null)
		 {
			 Toolkit toolkit = Toolkit.getDefaultToolkit();
			 fm = toolkit.getFontMetrics(currentFont);
		 }
		 else
		 {
			 fm = g.getFontMetrics(currentFont);
		 }
		 return fm;
	 }

	 /**
	  * Get the current font accordingly to the current zoom level of the map
	  * @param map the map representing the node
	  * @return current font
	  */
	 private Font getCurrentFont(LyAbstractMapComponentManager map)
	 {
		 //Check if the Font value has changed. 
		 float defFontSize = getFont("LY_MAPNODE_FONT").getSize2D();     
		 float expectedFontSize = isLabelZoomable() ? defFontSize*map.getZoom() : defFontSize;
		 if(_graphicalFont.getSize2D() != expectedFontSize) 
		 {
			 // Update the current font size according to current zoom level
			 _graphicalFont = _graphicalFont.deriveFont(expectedFontSize);

			 // Box will to be recalculated because of font change
			 _labelWidth = NOT_INITIALIZED;
			 _labelHeight = NOT_INITIALIZED;
			 _startLabelWidth = NOT_INITIALIZED;
			 _startLabelHeight = NOT_INITIALIZED;
			 _endLabelWidth = NOT_INITIALIZED;
			 _endLabelHeight = NOT_INITIALIZED;
		 }    

		 return _graphicalFont;
	 }

	 /**
	  * Return true is displayed label should be zoom dependant.
	  * @return true is label are zoomable
	  */
	 public boolean isLabelZoomable() {

		 // Supposed that a label is zoomable if both source and target 
		 // nodes labels are zoomable.
		 return _startNode.isLabelZoomable() && _endNode.isLabelZoomable();
	 } 	

	 /**
	  * Return the width of the link. It handle the default value of -1 (it returns 1)
	  * @return the width of the linkg
	  */
	 public int getLinkWidth()
	 {
		 int width = _mapLink.getWidth();
		 if (width < 0)
			 width = 1;

		 return width;
	 }

} // end class LySwingMapLink
