/*
 * LySwingMapNode.java
 * Copyright (c) 2008, Lyria.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.Toolkit;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.swing.JComponent;

import leon.misc.LyEnvironment;
import leon.view.LyComponent;
import leon.view.LyMap;
import leon.view.LyMapBase;
import leon.view.LyMapContent;
import leon.view.LyMapNode;
import leon.view.LyPoint;
import leon.view.common.Ly3DBox;
import leon.view.filters.LyReplaceColorImageFilter;
import leon.view.graphic.LyGraphicalMapNode;
import leon.view.win.swing.map.LyAbstractMapComponentManager;
import leon.view.win.swing.map.LyLayElement;
import leon.view.win.swing.map.LyMapElement;
import leon.view.win.swing.map.LyMapNodeBeanInterface;

/**
 * This class implements the Map node object (as described in the dtd) for swing viewer.
 *
 * @version $Id: LySwingMapNode.java 41709 2012-12-05 08:01:31Z nhamon $
 * @author  Lyria
 * @see     leon.view.LyMapNode
 */
public class LySwingMapNode extends LySwingMapObject implements
													LyGraphicalMapNode,
													LyLayElement
{
	public static final String			__VERSION			= "$Id: LySwingMapNode.java 41709 2012-12-05 08:01:31Z nhamon $";

	/** Default size of the image node. */
	public static int					DEFAULT_SIZE		= 15;

	/** Default size of the image node. */
	private static int					NOT_INITIALIZED		= -1;

	/** The view independent map node description. */
	protected LyMapNode					_mapNode;

	/** Center of the node. */
	protected LyPoint					_center;

	/** Font of the node. 
	 * Should not be accessed directly but through getCurrentFont() 
	 * to be properly set accordingly to current zoom level  */
	protected Font						_graphicalFont;

	/** Width of the image (for optimization). */
	protected int						_imageWidth			= NOT_INITIALIZED;

	/** Height of the image (for optimization). */
	protected int						_imageHeight		= NOT_INITIALIZED;

	/** Width of the label (for optimization). */
	protected int						_labelWidth			= NOT_INITIALIZED;

	/** Height of the label (for optimization). */
	protected int						_labelHeight		= NOT_INITIALIZED;

	/** Width of the left label (for optimization). */
	protected int						_leftLabelWidth		= NOT_INITIALIZED;

	/** Height of the left label (for optimization). */
	protected int						_leftLabelHeight	= NOT_INITIALIZED;

	/** Width of the right label (for optimization). */
	protected int						_rightLabelWidth	= NOT_INITIALIZED;

	/** Height of the right label (for optimization). */
	protected int						_rightLabelHeight	= NOT_INITIALIZED;

	/** Width of the top label (for optimization). */
	protected int						_topLabelWidth		= NOT_INITIALIZED;

	/** Height of the top label (for optimization). */
	protected int						_topLabelHeight		= NOT_INITIALIZED;

	/** Width of the bottom label (for optimization). */
	protected int						_bottomLabelWidth	= NOT_INITIALIZED;

	/** Height of the label (for optimization). */
	protected int						_bottomLabelHeight	= NOT_INITIALIZED;

	/** Label offset (for optimization). */
	protected Point						_labelOffset;

	/** Left label offset (for optimization). */
	protected Point						_leftLabelOffset;

	/** Right label offset (for optimization). */
	protected Point						_rightLabelOffset;

	/** Top label offset (for optimization). */
	protected Point						_topLabelOffset;

	/** Bottom label offset (for optimization). */
	protected Point						_bottomLabelOffset;

	/** Parent node (optional). */
	protected LySwingMapNode			_parentNode;

	/** Children of this node (optional). */
	protected Vector<LySwingMapNode>	_children			= new Vector<LySwingMapNode>();

	/** Offset between this node and its parent (if parent exists). */
	protected Point						_offset;

	/** Cached translated label that can be accessed many times. */
	protected String					_labelCache;

	/** Cache to translate left label only once (may access very often). */
	protected String					_leftLabelCache;

	/** Cache to translate right label only once (may access very often). */
	protected String					_rightLabelCache;

	/** Cache to translate top label only once (may access very often). */
	protected String					_topLabelCache;

	/** Cache to translate bottom label only once (may access very often). */
	protected String					_bottomLabelCache;

	/** Cached image for this node. */
	protected Image						_image;

	/** Cached image name for this node. */
	protected String					_imageString;

	/** Cached small image for this node. */
	protected Image						_smallImage;

	/** Cached small image name for this node. */
	protected String					_smallImageString;

	/** Cache node color, internal Frame color and external frame color. */
	/** Cached color for this node. */
	protected Color						_nodeColor;

	/** Cached color name for this node. */
	protected String					_nodeColorString;

	/** Cached color for the internal frame node. */
	protected Color						_internalFrameColor;

	/** Cached color name for the internal frame node. */
	protected String					_internalFrameColorString;

	/** Cached color for the external frame node. */
	protected Color						_externalFrameColor;

	/** Cached color name for the external frame node. */
	protected String					_externalFrameColorString;

	/** Cached cross color for this node. */
	protected Color						_crossColor;

	/** Cached cross color name for this node. */
	protected String					_crossColorString;

	/** Cached outline color for this node. */
	protected Color						_outlineColor;

	/** Cached selected color for this node. */
	protected Color						_selectedColor;

	/** Cached inactive selected color for this node. */
	protected Color						_inactiveSelectedColor;

	/** Cached font color for this node. */
	protected Color						_fontColor;

	/** Instance of a class that will draw the node.* */
	protected LyMapNodeBeanInterface	_nodeBean;

	/**
	 * Class constructor of a Map node object for the Swing target.
	 *
	 * @param mapNode a MAP_NODE component defined for every type of view targets. This map node is
	 *                initialized for the swing view.
	 */
	public LySwingMapNode(LyMapNode mapNode)
	{
		super(mapNode);
		_mapNode = mapNode;

		_parentNode = null;

		_image = null;
		_imageString = null;
		_smallImage = null;
		_smallImageString = null;

		_nodeColor = null;
		_nodeColorString = null;
		_internalFrameColor = null;
		_internalFrameColorString = null;
		_externalFrameColor = null;
		_externalFrameColorString = null;
		_crossColor = null;
		_crossColorString = null;

		_outlineColor = null;
		_selectedColor = null;
		_inactiveSelectedColor = null;
		_nodeBean = null;

		initialize();
	}

	/**
	 * Initializes the map node center attribute, its font and the image and label sizes to use.
	 */
	protected void initialize()
	{
		_center = new LyPoint(_mapNode.getX(), _mapNode.getY(), _mapNode.getZ());

		String	nodeFont = _mapNode.getFont();
		
		if (nodeFont == null)
			nodeFont = "LY_MAPNODE_FONT";
		
		_graphicalFont = getFont(nodeFont);

		_imageWidth = NOT_INITIALIZED;
		_imageHeight = NOT_INITIALIZED;
		_labelWidth = NOT_INITIALIZED;
		_labelHeight = NOT_INITIALIZED;
		_leftLabelWidth = NOT_INITIALIZED;
		_leftLabelHeight = NOT_INITIALIZED;
		_rightLabelWidth = NOT_INITIALIZED;
		_rightLabelHeight = NOT_INITIALIZED;
		_topLabelWidth = NOT_INITIALIZED;
		_topLabelHeight = NOT_INITIALIZED;
		_bottomLabelWidth = NOT_INITIALIZED;
		_bottomLabelHeight = NOT_INITIALIZED;
		_labelOffset = null;
		_leftLabelOffset = null;
		_rightLabelOffset = null;
		_topLabelOffset = null;
		_bottomLabelOffset = null;
		checkLabels();
	}

	/**
	 * Returns the image and small image of this node.
	 *
	 * @return the list of images used in the node
	 * @see    #setImage(String)
	 * @see    LyMapElement#getImages()
	 */
	public Image[] getImages()
	{
		Image image = null;
		Image smallImage = null;

		if (_mapNode.getStyle() == LyMapNode.Style.IMAGE)
		{
			if (_mapNode.getImage() != null)
				image = getImage(_mapNode.getImage());
			else
				image = getImage("LY_DEFAULT_MAP_NODE");
		}

		if (_mapNode.getSmallImage() != null)
			smallImage = getImage(_mapNode.getSmallImage());

		return (new Image[] { image, smallImage });
	}

	/**
	 * Returns the image of the node in the specified environment.
	 *
	 * @param  environment the environment
	 * @return the image of the node
	 * @see    #setImage(String)
	 */
	public Image getImage(LyEnvironment environment)
	{
		if (_mapNode.getStyle() == LyMapNode.Style.IMAGE)
		{
			if (_mapNode.getImage() != null)
			{
				if ((_imageString == null) || !_imageString.equals(_mapNode.getImage()))
				{
					_imageString = _mapNode.getImage();
					_image = (Image) environment.getImage(_mapNode.getImage());
					// Check if image is the same one as an unknown image. If so, try to load image
					// as a full path
					Image unknownImage = (Image) environment.getImage("_unknownImage");
					if (_image == unknownImage)
					{
						// Given image resource may be a full path : try to load image as a full path
						Image pathImage = (Image) environment.getImages().loadImage(null,
								_mapNode.getImage(), false);
						if (pathImage != null)
							_image = pathImage;
					}
				}
			}
			else
			{
				_imageString = "LY_DEFAULT_MAP_NODE";
				_image = getImage("LY_DEFAULT_MAP_NODE");
			}
		}
		return _image;
	} // end method getImage

	/**
	 * Returns the small image of the node or <code>null</code> if there is none.
	 *
	 * @param  environment the environment is not needed to find the small image. This parameter is
	 *                     not used in this implementation.
	 * @return the small image of the node
	 */
	public Image getSmallImage(LyEnvironment environment)
	{
		if (_mapNode.getSmallImage() != null)
		{
			if (_smallImageString != null)
			{
				if (!_smallImageString.equals(_mapNode.getSmallImage()))
				{
					_smallImageString = _mapNode.getSmallImage();
					_smallImage = getImage(_smallImageString);
				}
			}
			else
			{
				_smallImageString = _mapNode.getSmallImage();
				_smallImage = getImage(_smallImageString);
			}
		}
		else
		{
			_smallImageString = null;
			_smallImage = null;
		}

		return _smallImage;
	}

	/**
	 * Returns the current node color or <code>null</code> if there is none.
	 *
	 * @param  environment the environment is not needed to find this node color. This parameter is
	 *                     not used in this implementation.
	 * @return the color of the current node
	 */
	public Color getNodeColor(LyEnvironment environment)
	{
		String mapNodeColor = _mapNode.getColor();

		if (mapNodeColor != null)
		{
			if (_nodeColorString != null)
			{
				if (_nodeColorString.equals(mapNodeColor))
					return _nodeColor;
				_nodeColorString = mapNodeColor;
				_nodeColor = getColor(_nodeColorString);
			}
			else
			{
				_nodeColorString = mapNodeColor;
				_nodeColor = getColor(_nodeColorString);
			}
		}
		else
		{
			if (_nodeColor == null)
				_nodeColor = getColor("LY_MAP_NODE_COLOR");
		}
		return _nodeColor;
	}

	/**
	 * Returns the cross color of the node or <code>null</code> if there is none.
	 *
	 * @param  environment the environment is not needed to find this node cross color. This
	 *                     parameter is not used in this implementation.
	 * @return the cross color of the node
	 */
	public Color getCrossColor(LyEnvironment environment)
	{
		String mapNodeColorCross = _mapNode.getColorCross();

		if (mapNodeColorCross != null)
		{
			if (_crossColorString != null)
			{
				if (_crossColorString.equals(mapNodeColorCross))
					return _crossColor;
				_crossColorString = mapNodeColorCross;
				_crossColor = getColor(_crossColorString);
			}
			else
			{
				_crossColorString = mapNodeColorCross;
				_crossColor = getColor(_crossColorString);
			}
		}
		else
			return null;
		return _crossColor;
	}

	/**
	 * Returns the internal frame color of the node or <code>null</code> if there is none.
	 *
	 * @param  environment the environment is not needed to find this node's internal frame color.
	 *                     This parameter is not used in this implementation.
	 * @return the internal frame color
	 */
	public Color getInternalFrameColor(LyEnvironment environment)
	{
		String mapNodeColorFrameInt = _mapNode.getColorFrameInt();

		if (mapNodeColorFrameInt != null)
		{
			if (_internalFrameColorString != null)
			{
				if (_internalFrameColorString.equals(mapNodeColorFrameInt))
					return _internalFrameColor;
				_internalFrameColorString = mapNodeColorFrameInt;
				_internalFrameColor = getColor(_internalFrameColorString);
			}
			else
			{
				_internalFrameColorString = mapNodeColorFrameInt;
				_internalFrameColor = getColor(_internalFrameColorString);
			}
		}
		else
			return null;
		return _internalFrameColor;
	}

	/**
	 * Returns the external frame color of the node or <code>null</code> if there is none.
	 *
	 * @param  environment the environment is not needed to find this node's external frame color.
	 *                     This parameter is not used in this implementation.
	 * @return the external frame color
	 */
	public Color getExternalFrameColor(LyEnvironment environment)
	{
		String mapNodeColorFrameExt = _mapNode.getColorFrameExt();

		if (mapNodeColorFrameExt != null)
		{
			if (_externalFrameColorString != null)
			{
				if (_externalFrameColorString.equals(mapNodeColorFrameExt))
					return _externalFrameColor;
				_externalFrameColorString = _mapNode.getColorFrameExt();
				_externalFrameColor = getColor(_externalFrameColorString);
			}
			else
			{
				_externalFrameColorString = mapNodeColorFrameExt;
				_externalFrameColor = getColor(_externalFrameColorString);
			}
		}
		else
			return null;
		return _externalFrameColor;
	}

	/**
	 * Returns the selected color of the node.
	 *
	 * @param  environment the environment is not needed to find the selected color of this node.
	 *                     This parameter is not used in this implementation.
	 * @return the selected color
	 */
	public Color getSelectedColor(LyEnvironment environment)
	{
		if (_selectedColor != null)
			return _selectedColor;
		_selectedColor = getColor("LY_MAP_SELECTED_COLOR");
		return _selectedColor;
	}

	/**
	 * Returns the inactive selected color of the node.
	 *
	 * @param  environment the environment is not needed to find the inactive selected color of this
	 *                     node. This parameter is not used in this implementation.
	 * @return the inactive selected color
	 */
	public Color getInactiveSelectedColor(LyEnvironment environment)
	{
		if (_inactiveSelectedColor != null)
			return _inactiveSelectedColor;
		_inactiveSelectedColor = getColor("LY_INACTIVE_SELECT_COLOR");
		return _inactiveSelectedColor;
	}

	/**
	 * Returns the outline color of the node.
	 *
	 * @param  environment the environment is not needed to find the outline color of this node.
	 *                     This parameter is not used in this implementation.
	 * @return the outline color
	 */
	public Color getOutlineColor(LyEnvironment environment)
	{
		if (_outlineColor != null)
			return _outlineColor;
		_outlineColor = getColor("LY_MAP_OUTLINE_COLOR");
		return _outlineColor;
	}

	/**
	 * Returns the font color of the node.
	 *
	 * @param  environment the environment is not needed to find the font color of this node. This
	 *                     parameter is not used in this implementation.
	 * @return the font color
	 */
	public Color getFontColor(LyEnvironment environment)
	{
		if (_fontColor != null)
			return _fontColor;
		_fontColor = getColor("LY_MAP_FONT_COLOR");
		return _fontColor;
	}

	/**
	 * Returns the dimension of the image part of the node.
	 *
	 * @param  comp the map
	 * @return the dimension of the image part of the node
	 * @see    #getImageSize(Component, boolean)
	 */
	public Dimension getImageSize(Component comp)
	{
		return getImageSize(comp, false);
	}

	/**
	 * Returns the dimension of the image part of the node (without the frame).
	 *
	 * @param  comp the map
	 * @return the dimension of the image part of the node (without the frame)
	 */
	public Dimension getSymbolSize(Component comp)
	{
		int width = DEFAULT_SIZE;
		int height = width;

		switch (_mapNode.getStyle())
		{
		case IMAGE:
			Image graphicalImage = null;
			LyEnvironment environment = getEnvironment();

			graphicalImage = getImage(environment);
			if (graphicalImage != null)
			{    
    			width = graphicalImage.getWidth(comp);
    			height = graphicalImage.getHeight(comp);
			}
			break;

		case CIRCLE:
		case SQUARE:
		case TRIANGLE:
			width = _mapNode.getSize();
			if (width < 0)
				width = DEFAULT_SIZE;
			height = width;
			break;
			
		default:
			break;
		}

		Dimension frameSize = new Dimension(width, height);

		return frameSize;
	} // end method getSymbolSize

	/**
	 * Returns the dimension of the image part of the node.
	 *
	 * @param  comp        the map
	 * @param  recalculate Indicates if the size must be recalculated.
	 * @return the dimension of the image part of the node
	 */
	public Dimension getImageSize(Component comp, boolean recalculate)
	{
		if (recalculate || (_imageWidth < 0) || (_imageHeight < 0))
		{
			int dx = 0;
			int dy = 0;

			// Enlarge the size if there is a border
			if (_mapNode.getColorFrameExt() != null)
			{
				dx = 16;
				dy = 16;
			}
			else if (_mapNode.getColorFrameInt() != null)
			{
				dx = 10;
				dy = 10;
			}

			switch (_mapNode.getStyle())
			{
			case IMAGE:
				LyEnvironment environment = getEnvironment();
				Image graphicalImage = getImage(environment);

				if (graphicalImage != null)
				{
					_imageWidth = graphicalImage.getWidth(comp);
					_imageHeight = graphicalImage.getHeight(comp);
				}
				break;

			case CIRCLE:
			case SQUARE:
			case TRIANGLE:
				_imageWidth = _mapNode.getSize();
				if (_imageWidth < 0)
					_imageWidth = DEFAULT_SIZE;
				_imageHeight = _imageWidth;
				
			default:
				break;
			}

			_imageWidth += dx;
			_imageHeight += dy;
		}
		Dimension result = new Dimension(_imageWidth, _imageHeight);

		return result;
	} // end method getImageSize

	/**
	 * Returns the dimension of the label part of the node.
	 *
	 * @param map      the map representing the node
	 * @param  g the graphics context
	 * @return the dimension of the label part of the node
	 * @see    #getLabelSize(Graphics, boolean)
	 */
	public Dimension getLabelSize(LyAbstractMapComponentManager map, Graphics g)
	{
		return getLabelSize(map, g, false);
	}

	/**
	 * Returns the dimension of the label part of the node.
	 *
	 * @param map      the map representing the node
	 * @param  g           the graphics context
	 * @param  recalculate indicates if the size must be recalculated
	 * @return the dimension of the label part of the node
	 */
	public Dimension getLabelSize(LyAbstractMapComponentManager map, Graphics g, boolean recalculate)
	{
		if ((recalculate) || (_labelWidth < 0) || (_labelHeight < 0) || (_labelCache == null))
		{
			String label = getLabel();

			if (label == null)
				return new Dimension(0, 0);

			StringTokenizer tkn = new StringTokenizer(label, "\n");

			int num_lines = tkn.countTokens();
			String[] lines = new String[num_lines];
			int[] line_widths = new int[num_lines];

			for (int i = 0; i < num_lines; i++)
				lines[i] = tkn.nextToken();

			FontMetrics fm = getCurrentFontMetrics(map, g);

			_labelHeight = fm.getHeight();
			_labelWidth = fm.stringWidth(label);

			int line_height = fm.getHeight();

			int max_width = 0;

			for (int i = 0; i < num_lines; i++)
			{
				line_widths[i] = fm.stringWidth(lines[i]);
				max_width = Math.max(max_width, line_widths[i]);
			}

			int btnMarginWidth = 1;
			max_width += 2 * btnMarginWidth;
			int text_height = num_lines * line_height;
			if (_labelWidth != max_width || _labelHeight != text_height)
			{
				_labelWidth = max_width;
				_labelHeight = text_height;
			}

		}

		return (new Dimension(_labelWidth, _labelHeight));
	}

	/**
	 * Returns the dimension of the left label part of the node.
	 *
	 * @param map      the map representing the node
	 * @param g the graphics context
	 */
	public Dimension getLeftLabelSize(LyAbstractMapComponentManager map, Graphics g)
	{
		return getLeftLabelSize(map, g, false);
	}

	/**
	 * Returns the dimension of the left label part of the node.
	 *
	 * @param map      the map representing the node
	 * @param g           the graphics context
	 * @param recalculate indicates if the size must be recalculated
	 */
	public Dimension getLeftLabelSize(LyAbstractMapComponentManager map, Graphics g,
			boolean recalculate)
	{
		if ((recalculate) || (_leftLabelWidth < 0) || (_leftLabelHeight < 0)
				|| (_leftLabelCache == null))
		{
			String label = getLeftLabel();

			if (label == null)
				return new Dimension(0, 0);

			FontMetrics fm = getCurrentFontMetrics(map, g);

			StringTokenizer tkn = new StringTokenizer(label, "\n");

			int num_lines = tkn.countTokens();
			String[] lines = new String[num_lines];
			int[] line_widths = new int[num_lines];

			for (int i = 0; i < num_lines; i++)
				lines[i] = tkn.nextToken();

			_leftLabelHeight = fm.getHeight();
			_leftLabelWidth = fm.stringWidth(label);

			int line_height = fm.getHeight();

			int max_width = 0;

			for (int i = 0; i < num_lines; i++)
			{
				line_widths[i] = fm.stringWidth(lines[i]);
				max_width = Math.max(max_width, line_widths[i]);
			}

			int btnMarginWidth = 1;
			max_width += 2 * btnMarginWidth;
			int text_height = num_lines * line_height;
			if (_leftLabelWidth != max_width || _leftLabelHeight != text_height)
			{
				_leftLabelWidth = max_width;
				_leftLabelHeight = text_height;
			}

		}

		return (new Dimension(_leftLabelWidth, _leftLabelHeight));
	}

	/**
	 * Returns the dimension of the right label part of the node.
	 *
	 * @param map      the map representing the node
	 * @param g the graphics context
	 */
	public Dimension getRightLabelSize(LyAbstractMapComponentManager map, Graphics g)
	{
		return getRightLabelSize(map, g, false);
	}

	/**
	 * Returns the dimension of the right label part of the node.
	 *
	 * @param map      the map representing the node
	 * @param g           the graphics context
	 * @param recalculate indicates if the size must be recalculated
	 */
	public Dimension getRightLabelSize(LyAbstractMapComponentManager map, Graphics g,
			boolean recalculate)
	{
		if ((recalculate) || (_rightLabelWidth < 0) || (_rightLabelHeight < 0)
				|| (_rightLabelCache == null))
		{
			String label = getRightLabel();

			if (label == null)
				return new Dimension(0, 0);

			FontMetrics fm = getCurrentFontMetrics(map, g);

			StringTokenizer tkn = new StringTokenizer(label, "\n");

			int num_lines = tkn.countTokens();
			String[] lines = new String[num_lines];
			int[] line_widths = new int[num_lines];

			for (int i = 0; i < num_lines; i++)
				lines[i] = tkn.nextToken();

			_rightLabelHeight = fm.getHeight();
			_rightLabelWidth = fm.stringWidth(label);

			int line_height = fm.getHeight();

			int max_width = 0;

			for (int i = 0; i < num_lines; i++)
			{
				line_widths[i] = fm.stringWidth(lines[i]);
				max_width = Math.max(max_width, line_widths[i]);
			}

			int btnMarginWidth = 1;
			max_width += 2 * btnMarginWidth;
			int text_height = num_lines * line_height;
			if (_rightLabelWidth != max_width || _rightLabelHeight != text_height)
			{
				_rightLabelWidth = max_width;
				_rightLabelHeight = text_height;
			}
		}

		return (new Dimension(_rightLabelWidth, _rightLabelHeight));
	}

	/**
	 * Returns the dimension of the top label part of the node.
	 *
	 * @param map      the map representing the node
	 * @param g the graphics context
	 */
	public Dimension getTopLabelSize(LyAbstractMapComponentManager map, Graphics g)
	{
		return getTopLabelSize(map, g, false);
	}

	/**
	 * Returns the dimension of the top label part of the node.
	 *
	 * @param map      the map representing the node
	 * @param g           the graphics context
	 * @param recalculate indicates if the size must be recalculated
	 */
	public Dimension getTopLabelSize(LyAbstractMapComponentManager map, Graphics g,
			boolean recalculate)
	{
		if ((recalculate) || (_topLabelWidth < 0) || (_topLabelHeight < 0)
				|| (_topLabelCache == null))
		{
			String label = getTopLabel();

			if (label == null)
				return new Dimension(0, 0);

			FontMetrics fm = getCurrentFontMetrics(map, g);

			StringTokenizer tkn = new StringTokenizer(label, "\n");

			int num_lines = tkn.countTokens();
			String[] lines = new String[num_lines];
			int[] line_widths = new int[num_lines];

			for (int i = 0; i < num_lines; i++)
				lines[i] = tkn.nextToken();

			_topLabelHeight = fm.getHeight();
			_topLabelWidth = fm.stringWidth(label);

			int line_height = fm.getHeight();

			int max_width = 0;

			for (int i = 0; i < num_lines; i++)
			{
				line_widths[i] = fm.stringWidth(lines[i]);
				max_width = Math.max(max_width, line_widths[i]);
			}

			int btnMarginWidth = 1;
			max_width += 2 * btnMarginWidth;
			int text_height = num_lines * line_height;
			if (_topLabelWidth != max_width || _topLabelHeight != text_height)
			{
				_topLabelWidth = max_width;
				_topLabelHeight = text_height;
			}
		}

		return (new Dimension(_topLabelWidth, _topLabelHeight));
	}

	/**
	 * Returns the dimension of the bottom label part of the node.
	 *
	 * @param map      the map representing the node
	 * @param g the graphics context
	 */
	public Dimension getBottomLabelSize(LyAbstractMapComponentManager map, Graphics g)
	{
		return getBottomLabelSize(map, g, false);
	}

	/**
	 * Returns the dimension of the bottom label part of the node.
	 *
	 * @param map      the map representing the node
	 * @param g           the graphics context
	 * @param recalculate indicates if the size must be recalculated
	 */
	public Dimension getBottomLabelSize(LyAbstractMapComponentManager map, Graphics g,
			boolean recalculate)
	{
		if ((recalculate) || (_bottomLabelWidth < 0) || (_bottomLabelHeight < 0)
				|| (_bottomLabelCache == null))
		{
			String label = getBottomLabel();

			if (label == null)
				return new Dimension(0, 0);

			FontMetrics fm = getCurrentFontMetrics(map, g);

			StringTokenizer tkn = new StringTokenizer(label, "\n");

			int num_lines = tkn.countTokens();
			String[] lines = new String[num_lines];
			int[] line_widths = new int[num_lines];

			for (int i = 0; i < num_lines; i++)
				lines[i] = tkn.nextToken();

			_bottomLabelHeight = fm.getHeight();
			_bottomLabelWidth = fm.stringWidth(label);

			int line_height = fm.getHeight();

			int max_width = 0;

			for (int i = 0; i < num_lines; i++)
			{
				line_widths[i] = fm.stringWidth(lines[i]);
				max_width = Math.max(max_width, line_widths[i]);
			}

			int btnMarginWidth = 1;
			max_width += 2 * btnMarginWidth;
			int text_height = num_lines * line_height;
			if (_bottomLabelWidth != max_width || _bottomLabelHeight != text_height)
			{
				_bottomLabelWidth = max_width;
				_bottomLabelHeight = text_height;
			}
		}

		return (new Dimension(_bottomLabelWidth, _bottomLabelHeight));
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#draw(leon.view.win.swing.map.LyAbstractMapComponentManager,
	 *      java.awt.Graphics, int)
	 */
	public void draw(LyAbstractMapComponentManager map, Graphics g, int mode)
	{
		if (getBean() != null)
		{
			getBean().draw(map, g, mode);
			return;
		}

		if (isHidden())
			return;

		if (_mapNode instanceof LyMapElement)
		{
			((LyMapElement) _mapNode).draw(map, g, mode);
			return;
		}
		if (_parentNode != null)
		{
			// When a child node is draw
			Point parent = map.convert(_parentNode.getCenter());
			Point child = map.convert(_center);

			if (_offset == null)
			{
				// Recovery of the position compared to parent and update of the attribute '_offset'
				Point newPoint = new Point(parent.x - child.x, parent.y - child.y);

				_offset = newPoint;
			}
			else
			{
				// On applique ce dcalage sur les coordonnes 3D (aprs un zoom, translation...)
				child.x = parent.x - _offset.x;
				child.y = parent.y - _offset.y;
				_center = map.revert(child);
			}
		}

		Point point = map.convertForDrawing(_center);
		int x = point.x, y = point.y;

		drawImage(map, g, x, y, mode);
		drawLabel(map, g, x, y);
		drawLeftLabel(map, g, x, y);
		drawRightLabel(map, g, x, y);
		drawTopLabel(map, g, x, y);
		drawBottomLabel(map, g, x, y);
	} // end method draw

	/**
	 * Draws the image of this node depending on its style at the given (x,y) position.
	 *
	 * @param map  the map representing the node
	 * @param g    the current graphics
	 * @param x    the x coordinate
	 * @param y    the y coordinate
	 * @param mode mode of drawing
	 */
	protected void drawImage(LyAbstractMapComponentManager map, Graphics g, int x, int y, int mode)
	{
		boolean isNormalMode = (mode == LyMapElement.NORMAL_MODE);

		if (!isNormalMode)
			return;

		Dimension dim = getImageSize(map.getComponent());
		Dimension symbolSize = getSymbolSize(map.getComponent());
		float zoom = map.getZoom();

		if (_mapNode.isResizable())
		{
			dim = new Dimension((int) (dim.width * zoom), (int) (dim.height * zoom));
			symbolSize = new Dimension((int) (symbolSize.width * zoom),
					(int) (symbolSize.height * zoom));
		}

		Color internalFrameColor;
		Color externalFrameColor;
		Color color;
		LyEnvironment environment = getEnvironment();

		internalFrameColor = getInternalFrameColor(environment);
		externalFrameColor = getExternalFrameColor(environment);
		color = getNodeColor(environment);
		// Draw an area around the node with a color depending on the
		// state of the node (selected or not)
		if (_mapNode.isSelected())
		{
			LyComponent parent = _mapNode.getParent();
			boolean fill = true;

			if (parent instanceof LyMapContent)
				parent = parent.getParent();

			if (((LyMap) parent).getSelectionDraw() == LyMapBase.SelectionDraw.BORDER)
				fill = false;

			if (map.isActivated())
				g.setColor(getSelectedColor(environment));
			else
				g.setColor(getInactiveSelectedColor(environment));

			if ((internalFrameColor != null) || (externalFrameColor != null))
			{
				if (!fill)
					g.drawRect(x - (dim.width / 2) - 3, y - (dim.height / 2) - 3, dim.width + 7,
							dim.height + 7);
				else
					g.fillRect(x - (dim.width / 2) - 3, y - (dim.height / 2) - 3, dim.width + 7,
							dim.height + 7);
			}
			else
			{
				switch (_mapNode.getStyle())
				{
				case IMAGE:
					// Draw a rectangle
					/*
					 * Too large for big icons... g.fillRect(x - (5 * dim.width / 8), y - (5 *
					 * dim.height / 8), 10 * dim.width / 8, 10 * dim.height / 8);
					 */
					if (!fill)
						g.drawRect(x - (symbolSize.width / 2) - 3, y - (symbolSize.height / 2) - 3,
								symbolSize.width + 7, symbolSize.height + 6);
					else
						g.fillRect(x - (symbolSize.width / 2) - 3, y - (symbolSize.height / 2) - 3,
								symbolSize.width + 7, symbolSize.height + 7);
					break;

				case CIRCLE:
					// Draw a circle
					if (!fill)
						g.drawOval(x - (symbolSize.width / 2) - 3, y - (symbolSize.width / 2) - 3,
								symbolSize.width + 5, symbolSize.height + 5);
					else
						g.fillOval(x - (symbolSize.width / 2) - 3, y - (symbolSize.width / 2) - 3,
								symbolSize.width + 5, symbolSize.height + 5);
					break;

				case SQUARE:
					// Draw a rectangle
					if (!fill)
						g.drawRect(x - (symbolSize.width / 2) - 3, y - (symbolSize.width / 2) - 3,
								symbolSize.width + 5, symbolSize.height + 5);
					else
						g.fillRect(x - (symbolSize.width / 2) - 3, y - (symbolSize.width / 2) - 3,
								symbolSize.width + 5, symbolSize.height + 5);
					break;

				case TRIANGLE:
					// Draw a triangle
					if (!fill)
						g.drawPolygon(new int[] { x - (2 * symbolSize.width / 3), x,
								x + (2 * symbolSize.width / 3) }, new int[] {
								y + (2 * symbolSize.height / 3), y - (2 * symbolSize.height / 3),
								y + (2 * symbolSize.height / 3) }, 3);
					else
						g.fillPolygon(new int[] { x - (2 * symbolSize.width / 3), x,
								x + (2 * symbolSize.width / 3) }, new int[] {
								y + (2 * symbolSize.height / 3), y - (2 * symbolSize.height / 3),
								y + (2 * symbolSize.height / 3) }, 3);
					break;
				
				default:
					break;
				}
			}
		}

		// Draw a single frame
		if ((externalFrameColor != null) || (internalFrameColor != null))
		{
			Color outLineColor = getOutlineColor(environment);

			// Draw the external frame
			if (externalFrameColor != null)
			{
				g.setColor(externalFrameColor);
				g.fillRect(x - (dim.width / 2), y - (dim.height / 2), dim.width, dim.height);

				// Draw the external border
				g.setColor(outLineColor);
				g.drawRect(x - (dim.width / 2), y - (dim.height / 2), dim.width, dim.height);

				// Draw the internal frame
				if (internalFrameColor != null)
				{
					g.setColor(internalFrameColor);
					g.fillRect(x - (dim.width / 2) + 4, y - (dim.height / 2) + 4, dim.width - 8,
							dim.height - 8);

					// Draw the internal border
					g.setColor(outLineColor);
					g.drawRect(x - (dim.width / 2) + 4, y - (dim.height / 2) + 4, dim.width - 8,
							dim.height - 8);
				}
			}
			else if (internalFrameColor != null)
			{
				// Draw the single frame
				g.setColor(internalFrameColor);
				g.fillRect(x - (dim.width / 2), y - (dim.height / 2), dim.width, dim.height);

				// Draw the external border
				g.setColor(outLineColor);
				g.drawRect(x - (dim.width / 2), y - (dim.height / 2), dim.width, dim.height);
			}
		}

		// Draw the image representing the node
		switch (_mapNode.getStyle())
		{
		case IMAGE:
			// Draw the image
			Image graphicalImage = getImage(environment);

			int width = DEFAULT_SIZE;
			int height = DEFAULT_SIZE;
			
			if (graphicalImage != null)
			{
    			// Replace the black color with the node color
    			if (_mapNode.getColor() != null)
    				graphicalImage = LyReplaceColorImageFilter.getFilteredImage(graphicalImage,
    						Color.black, color);
    
    			width = graphicalImage.getWidth(map.getComponent());
    			height = graphicalImage.getHeight(map.getComponent());
			}
			if (_mapNode.isResizable())
			{
				width = (int) (width * zoom);
				height = (int) (height * zoom);
			}

			g.drawImage(graphicalImage, x - (width / 2), y - (height / 2), width, height, map
					.getComponent());
			break;

		case CIRCLE:
			// Draw a circle
			g.setColor(color);
			g.fillOval(x - (symbolSize.width / 2), y - (symbolSize.height / 2), symbolSize.width,
					symbolSize.height);
			g.setColor(getOutlineColor(environment));
			g.drawOval(x - (symbolSize.width / 2), y - (symbolSize.height / 2),
					symbolSize.width - 1, symbolSize.height - 1);
			break;

		case SQUARE:
			g.setColor(color);
			g.fillRect(x - (symbolSize.width / 2), y - (symbolSize.height / 2), symbolSize.width,
					symbolSize.height);
			g.setColor(getOutlineColor(environment));
			g.drawRect(x - (symbolSize.width / 2), y - (symbolSize.height / 2),
					symbolSize.width - 1, symbolSize.height - 1);
			break;

		case TRIANGLE:
			g.setColor(color);
			g.fillPolygon(new int[] { x - (symbolSize.width / 2), x, x + (symbolSize.width / 2) },
					new int[] { y + (symbolSize.height / 2), y - (symbolSize.height / 2),
							y + (symbolSize.height / 2) }, 3);
			g.setColor(getOutlineColor(environment));
			g.drawPolygon(new int[] { x - (symbolSize.width / 2) + 1, x,
					x + (symbolSize.width / 2) - 1 }, new int[] { y + (symbolSize.height / 2) - 1,
					y - (symbolSize.height / 2) + 1, y + (symbolSize.height / 2) - 1 }, 3);
			break;
		
		default:
			break;
		}

		Color crossColor = getCrossColor(environment);

		if (crossColor != null)
		{
			// Draw a cross on the node
			g.setColor(Color.red);

			Dimension crossDim = new Dimension(symbolSize.width - 2, symbolSize.height - 2);

			g.drawLine(x - (crossDim.width / 2), y - (crossDim.height / 2), x
					+ (crossDim.width / 2), y + (crossDim.height / 2));
			g.drawLine(x - (crossDim.width / 2) + 1, y - (crossDim.height / 2), x
					+ (crossDim.width / 2), y + (crossDim.height / 2) - 1);
			g.drawLine(x - (crossDim.width / 2) + 2, y - (crossDim.height / 2), x
					+ (crossDim.width / 2), y + (crossDim.height / 2) - 2);
			g.drawLine(x - (crossDim.width / 2), y - (crossDim.height / 2) + 1, x
					+ (crossDim.width / 2) - 1, y + (crossDim.height / 2));
			g.drawLine(x - (crossDim.width / 2), y - (crossDim.height / 2) + 2, x
					+ (crossDim.width / 2) - 2, y + (crossDim.height / 2));

			g.drawLine(x - (crossDim.width / 2), y + (crossDim.height / 2), x
					+ (crossDim.width / 2), y - (crossDim.height / 2));
			g.drawLine(x - (crossDim.width / 2) + 1, y + (crossDim.height / 2), x
					+ (crossDim.width / 2), y - (crossDim.height / 2) + 1);
			g.drawLine(x - (crossDim.width / 2) + 2, y + (crossDim.height / 2), x
					+ (crossDim.width / 2), y - (crossDim.height / 2) + 2);
			g.drawLine(x - (crossDim.width / 2), y + (crossDim.height / 2) - 1, x
					+ (crossDim.width / 2) - 1, y - (crossDim.height / 2));
			g.drawLine(x - (crossDim.width / 2), y + (crossDim.height / 2) - 2, x
					+ (crossDim.width / 2) - 2, y - (crossDim.height / 2));
		}

		// Draw the small image
		Image smallImage = getSmallImage(environment);

		if (smallImage != null)
		{
			int width = smallImage.getWidth(map.getComponent());
			int height = smallImage.getHeight(map.getComponent());

			g.drawImage(smallImage, x + (dim.width / 2) - width, y + (dim.height / 2) - height, map
					.getComponent());
		}
	} // end method drawImage

	/**
	 * Returns the offset between the center of the node and the left upper corner of the label.
	 *
	 * @param  map the map representing the node
	 * @param  g   the current graphics
	 * @param  x   the x coordinate
	 * @param  y   the y coordinate
	 * @return the offset between this box and a label
	 * @see    #getLabelOffset(LyAbstractMapComponentManager, Graphics, int, int, boolean)
	 */
	protected Point getLabelOffset(LyAbstractMapComponentManager map, Graphics g, int x, int y)
	{
		return getLabelOffset(map, g, x, y, false);
	}

	/**
	 * Returns the offset between the center of the node and the left upper corner of the label.
	 *
	 * @param  map         the map representing the node
	 * @param  g           the current graphics
	 * @param  x           the x coordinate
	 * @param  y           the y coordinate
	 * @param  recalculate indicates if the offset must be recalculated
	 * @return the offset between this box and a label
	 */
	protected Point getLabelOffset(LyAbstractMapComponentManager map, Graphics g, int x, int y,
			boolean recalculate)
	{
		if ((_labelOffset == null) || recalculate || _mapNode.isResizable())
		{
			FontMetrics fm = getCurrentFontMetrics(map, g);

			// Recovery of the rectangle of the largest child node.(in order to modify the
			// '_labelOffset')
			Rectangle childRectangle = getBoundingRectangleOfChild(map, 0);

			Dimension labelDim = getLabelSize(map, g);
			Dimension imageDim = getImageSize(map.getComponent());
			int offx = 0, offy = 0;

			if (_mapNode.isResizable())
			{
				float zoom = map.getZoom();

				imageDim = new Dimension((int) (imageDim.width * zoom),
						(int) (imageDim.height * zoom));
			}

			LyMapNode.LabelAlign align = _mapNode.getLabelAlign();

			if (align == null)
				align = LyMapNode.LabelAlign.CENTER;

			switch (align)
			{
			case BOTTOM:
				offx = -labelDim.width / 2;
				offy = (imageDim.height / 2) + 3;
				offy = (childRectangle == null) ? offy : (offy + childRectangle.height);
				break;
			case TOP:
				offx = -labelDim.width / 2;
				offy = (-imageDim.height / 2) - fm.getHeight();
				offy = (childRectangle == null) ? offy : (offy - childRectangle.height);
				break;
			case LEFT:
				offx = (-imageDim.width / 2) - labelDim.width - 3;
				offy = -fm.getHeight() / 2;
				offx = (childRectangle == null) ? offx : (offx - childRectangle.width);
				break;
			case RIGHT:
				offx = (imageDim.width / 2) + 4;
				offy = -fm.getHeight() / 2;
				offx = (childRectangle == null) ? offx : (offx + childRectangle.width);
				break;
			case CENTER:
				offx = -labelDim.width / 2;
				offy = -fm.getHeight() / 2;
				break;
			case EXTERN:
				Dimension dim = map.getComponent().getSize();
				Point mapCenter = new Point(dim.width / 2, dim.height / 2);
				double alpha;

				// If the center of the node and the center of the map have the
				// same x coordonate, alpha equals PI/2 or -PI/2
				if (mapCenter.x == x)
				{
					if (y > mapCenter.y)
						alpha = -Math.PI / 2;
					else
						alpha = Math.PI / 2;
				}
				// Else calculate alpha in common case
				else
					alpha = -Math.atan(((float) (y - mapCenter.y)) / (x - mapCenter.x));

				// Give alpha a value between -PI and PI
				if (x < mapCenter.x)
				{
					if (y > mapCenter.y)
						alpha -= Math.PI;
					else
						alpha += Math.PI;
				}

				// Recovery of the rectangle of the largest child node.(in order to modify the
				// '_labelOffset')
				childRectangle = getBoundingRectangleOfChild(map, alpha);

				// Treat each of the four sectors
				if ((alpha >= (-Math.PI / 4)) && (alpha < (Math.PI / 4)))
				{
					offx = (imageDim.width / 2) + 4;
					offy = (int) ((-fm.getHeight() / 2) - ((fm.getHeight() + imageDim.height)
							* Math.tan(alpha) / 2));
					offx = (childRectangle == null) ? offx : (offx + childRectangle.width);
				}
				else if ((alpha >= (Math.PI / 4)) && (alpha < (3 * Math.PI / 4)))
				{
					offx = (int) ((-labelDim.width / 2) - ((labelDim.width + imageDim.width)
							* Math.tan(alpha + (Math.PI / 2)) / 2));
					offy = (-imageDim.height / 2) - fm.getHeight();
					offy = (childRectangle == null) ? offy : (offy - childRectangle.height);
				}
				else if (((alpha >= (3 * Math.PI / 4)) && (alpha <= Math.PI))
						|| ((alpha >= -Math.PI) && (alpha < (-3 * Math.PI / 4))))
				{
					offx = (-imageDim.width / 2) - labelDim.width - 3;
					offy = (int) ((-fm.getHeight() / 2) + ((fm.getHeight() + imageDim.height)
							* Math.tan(alpha) / 2));
					offx = (childRectangle == null) ? offx : (offx - childRectangle.width);
				}
				else
				{
					offx = (int) ((-labelDim.width / 2) + ((labelDim.width + imageDim.width)
							* Math.tan(alpha + (Math.PI / 2)) / 2));
					offy = (imageDim.height / 2) + 1 + 3;
					offy = (childRectangle == null) ? offy : (offy + childRectangle.height);
				}
				break;
			}

			_labelOffset = new Point(offx, offy);
		}

		return _labelOffset;
	} // end method getLabelOffset

	/**
	 * Returns the offset between the center of the node and the left upper corner of the left
	 * label.
	 *
	 * @param  map the map representing the node
	 * @param  g   the current graphics
	 * @return the offset between this box and a label
	 * @see    #getLeftLabelOffset(LyAbstractMapComponentManager, Graphics, boolean)
	 */
	protected Point getLeftLabelOffset(LyAbstractMapComponentManager map, Graphics g)
	{
		return getLeftLabelOffset(map, g, false);
	}

	/**
	 * Returns the offset between the center of the node and the left upper corner of the left
	 * label.
	 *
	 * @param  map the map representing the node
	 * @param  g   the current graphics
	 * @param  x   the x coordinate
	 * @param  y   the y coordinate
	 * @return the offset between this box and a label
	 */
	protected Point getLeftLabelOffset(LyAbstractMapComponentManager map, Graphics g,
			boolean recalculate)
	{
		if ((_leftLabelOffset == null) || recalculate || _mapNode.isResizable())
		{
			FontMetrics fm = getCurrentFontMetrics(map, g);

			//Recovery of the rectangle of the largest child node.(in order to modify the '_leftLabelOffset')
			Rectangle childRectangle = getBoundingRectangleOfChild(map, 0);
			Dimension labelDim = getLeftLabelSize(map, g);
			Dimension imageDim = getImageSize(map.getComponent());
			int offx = 0, offy = 0;

			if (_mapNode.isResizable())
			{
				float zoom = map.getZoom();

				imageDim = new Dimension((int) (imageDim.height * zoom),
						(int) (imageDim.height * zoom));
			}

			offx = (-imageDim.width / 2) - labelDim.width - 3;
			offy = -fm.getHeight() / 2;
			offx = (childRectangle == null) ? offx : (offx - childRectangle.width);
			_leftLabelOffset = new Point(offx, offy);
		}

		return _leftLabelOffset;
	} // end method getLeftLabelOffset

	/**
	 * Returns the offset between the center of the node and the left upper corner of the right
	 * label.
	 *
	 * @param  map the map representing the node
	 * @param  g   the current graphics
	 * @return the offset between this box and a label
	 * @see    #getRightLabelOffset(LyAbstractMapComponentManager, Graphics, boolean)
	 */
	protected Point getRightLabelOffset(LyAbstractMapComponentManager map, Graphics g)
	{
		return getRightLabelOffset(map, g, false);
	}

	/**
	 * Returns the offset between the center of the node and the left upper corner of the right
	 * label.
	 *
	 * @param  map the map representing the node
	 * @param  g   the current graphics
	 * @param  x   the x coordinate
	 * @param  y   the y coordinate
	 * @return the offset between this box and a label
	 */
	protected Point getRightLabelOffset(LyAbstractMapComponentManager map, Graphics g,
			boolean recalculate)
	{
		if ((_rightLabelOffset == null) || recalculate || _mapNode.isResizable())
		{
			FontMetrics fm = getCurrentFontMetrics(map, g);

			//Recovery of the rectangle of the largest child node.(in order to modify the '_rightLabelOffset')
			Rectangle childRectangle = getBoundingRectangleOfChild(map, 0);
			Dimension imageDim = getImageSize(map.getComponent());
			int offx = 0, offy = 0;

			if (_mapNode.isResizable())
			{
				float zoom = map.getZoom();

				imageDim = new Dimension((int) (imageDim.width * zoom),
						(int) (imageDim.height * zoom));
			}

			offx = (imageDim.width / 2) + 4;
			offy = -fm.getHeight() / 2;
			offx = (childRectangle == null) ? offx : (offx + childRectangle.width);
			_rightLabelOffset = new Point(offx, offy);
		}

		return _rightLabelOffset;
	} // end method getRightLabelOffset

	/**
	 * Returns the offset between the center of the node and the left upper corner of the top label.
	 *
	 * @param  map the map representing the node
	 * @param  g   the current graphics
	 * @return the offset between this box and a label
	 * @see    #getTopLabelOffset(LyAbstractMapComponentManager, Graphics, boolean)
	 */
	protected Point getTopLabelOffset(LyAbstractMapComponentManager map, Graphics g)
	{
		return getTopLabelOffset(map, g, false);
	}

	/**
	 * Returns the offset between the center of the node and the left upper corner of the top label.
	 *
	 * @param  map the map representing the node
	 * @param  g   the current graphics
	 * @param  x   the x coordinate
	 * @param  y   the y coordinate
	 * @return the offset between this box and a label
	 */
	protected Point getTopLabelOffset(LyAbstractMapComponentManager map, Graphics g,
			boolean recalculate)
	{
		if ((_topLabelOffset == null) || recalculate || _mapNode.isResizable())
		{
			//Recovery of the rectangle of the largest child node.(in order to modify the '_topLabelOffset')
			Rectangle childRectangle = getBoundingRectangleOfChild(map, 0);

			Dimension labelDim = getTopLabelSize(map, g);
			Dimension imageDim = getImageSize(map.getComponent());
			int offx = 0, offy = 0;

			if (_mapNode.isResizable())
			{
				float zoom = map.getZoom();

				imageDim = new Dimension((int) (imageDim.width * zoom),
						(int) (imageDim.height * zoom));
			}

			offx = -labelDim.width / 2;
			offy = (-imageDim.height / 2) - labelDim.height - 2;
			offy = (childRectangle == null) ? offy : (offy - childRectangle.height);
			_topLabelOffset = new Point(offx, offy);
		}

		return _topLabelOffset;
	} // end method getTopLabelOffset

	/**
	 * Returns the offset between the center of the node and the left upper corner of the bottom
	 * label.
	 *
	 * @param  map the map representing the node
	 * @param  g   the current graphics
	 * @return the offset between this box and a label
	 * @see    #getBottomLabelOffset(LyAbstractMapComponentManager, Graphics, boolean)
	 */
	protected Point getBottomLabelOffset(LyAbstractMapComponentManager map, Graphics g)
	{
		return getBottomLabelOffset(map, g, false);
	}

	/**
	 * Returns the offset between the center of the node and the left upper corner of the bottom
	 * label.
	 *
	 * @param  map the map representing the node
	 * @param  g   the current graphics
	 * @param  x   the x coordinate
	 * @param  y   the y coordinate
	 * @return the offset between this box and a label
	 */
	protected Point getBottomLabelOffset(LyAbstractMapComponentManager map, Graphics g,
			boolean recalculate)
	{
		if ((_bottomLabelOffset == null) || recalculate || _mapNode.isResizable())
		{
			//Recovery of the rectangle of the largest child node.(in order to modify the '_bottomLabelOffset')
			Rectangle childRectangle = getBoundingRectangleOfChild(map, 0);

			Dimension labelDim = getBottomLabelSize(map, g);
			Dimension imageDim = getImageSize(map.getComponent());
			int offx = 0, offy = 0;

			if (_mapNode.isResizable())
			{
				float zoom = map.getZoom();

				imageDim = new Dimension((int) (imageDim.width * zoom),
						(int) (imageDim.height * zoom));
			}

			offx = -labelDim.width / 2;
			offy = (imageDim.height / 2) + 3;
			offy = (childRectangle == null) ? offy : (offy + childRectangle.height);
			_bottomLabelOffset = new Point(offx, offy);
		}

		return _bottomLabelOffset;
	}

	/**
	 * Draws the label of the node at the given (x,y) position.
	 *
	 * @param map the map representing the node
	 * @param g   the current graphics
	 * @param x   the x coordinate
	 * @param y   the y coordinate
	 */
	protected void drawLabel(LyAbstractMapComponentManager map, Graphics g, int x, int y)
	{
		// Recalculate the offset and size if label is extern or if labelCache is null
		boolean recalculate = (_mapNode.getLabelAlign() == LyMapNode.LabelAlign.EXTERN);

		if (_labelCache == null)
			recalculate = true;

		String label = getLabel();

		if ((label == null) || (label.length() == 0))
			return;

		Dimension labelDim = getLabelSize(map, g, recalculate);
		Point labelOffset = getLabelOffset(map, g, x, y, recalculate);
		int xstart, ystart;

		// Calculate the label position
		xstart = x + labelOffset.x;
		ystart = y + labelOffset.y;

		drawLabelAtPosition(map, g, label, xstart, ystart, labelDim);
	}

	/**
	 * Draw the left label of the node.
	 */
	protected void drawLeftLabel(LyAbstractMapComponentManager map, Graphics g, int x, int y)
	{
		// Recalculate the offset and size if label cache is null
		boolean recalculate = (_leftLabelCache == null);
		String label = getLeftLabel();

		if ((label == null) || (label.length() == 0))
			return;

		Dimension labelDim = getLeftLabelSize(map, g, recalculate);
		Point labelOffset = getLeftLabelOffset(map, g, recalculate);
		int xstart, ystart;

		// Calculate the label position
		xstart = x + labelOffset.x;
		ystart = y + labelOffset.y;

		drawLabelAtPosition(map, g, label, xstart, ystart, labelDim);
	}

	/**
	 * Draw the right label of the node.
	 */
	protected void drawRightLabel(LyAbstractMapComponentManager map, Graphics g, int x, int y)
	{
		// Recalculate the offset and size if label cache is null
		boolean recalculate = (_rightLabelCache == null);
		String label = getRightLabel();

		if ((label == null) || (label.length() == 0))
			return;

		Dimension labelDim = getRightLabelSize(map, g, recalculate);
		Point labelOffset = getRightLabelOffset(map, g, recalculate);
		int xstart, ystart;

		// Calculate the label position
		xstart = x + labelOffset.x;
		ystart = y + labelOffset.y;

		drawLabelAtPosition(map, g, label, xstart, ystart, labelDim);
	}

	/**
	 * Draw the top label of the node.
	 */
	protected void drawTopLabel(LyAbstractMapComponentManager map, Graphics g, int x, int y)
	{
		// Recalculate the offset and size if label cache is null
		boolean recalculate = (_topLabelCache == null);
		String label = getTopLabel();

		if ((label == null) || (label.length() == 0))
			return;

		Dimension labelDim = getTopLabelSize(map, g, recalculate);
		Point labelOffset = getTopLabelOffset(map, g, recalculate);
		int xstart, ystart;

		// Calculate the label position
		xstart = x + labelOffset.x;
		ystart = y + labelOffset.y;

		drawLabelAtPosition(map, g, label, xstart, ystart, labelDim);
	}

	/**
	 * Draw the bottom label of the node.
	 */
	protected void drawBottomLabel(LyAbstractMapComponentManager map, Graphics g, int x, int y)
	{
		// Recalculate the offset and size if label cache is null
		boolean recalculate = (_bottomLabelCache == null);
		String label = getBottomLabel();

		if ((label == null) || (label.length() == 0))
			return;

		Dimension labelDim = getBottomLabelSize(map, g, recalculate);
		Point labelOffset = getBottomLabelOffset(map, g, recalculate);
		int xstart, ystart;

		// Calculate the label position
		xstart = x + labelOffset.x;
		ystart = y + labelOffset.y;

		drawLabelAtPosition(map, g, label, xstart, ystart, labelDim);
	}

	/**
	 * Draws a label of the node at the given (x,y) position.
	 *
	 * @param map      the map representing the node
	 * @param g        the current graphics
	 * @param label    the label that must be drawn
	 * @param xstart   the x coordinate
	 * @param ystart   the y coordinate
	 * @param labelDim the dimension of the label
	 */
	protected void drawLabelAtPosition(LyAbstractMapComponentManager map, Graphics g, String label,
			int xstart, int ystart, Dimension labelDim)
	{
		boolean fillLabel = testResource("LY_FILL_MAP_NODE_LABEL");
		LyComponent parent = _mapNode.getParent();
		boolean drawBorder = false;

		if (parent instanceof LyMapContent)
			parent = parent.getParent();

		if (((LyMap) parent).getSelectionDraw() == LyMapBase.SelectionDraw.BORDER)
			drawBorder = true;

		Font currentFont = getCurrentFont(map);
		Component component = map.getComponent();
		FontMetrics fm = component.getFontMetrics(currentFont);

		StringTokenizer tkn = new StringTokenizer(label, "\n");

		int num_lines = tkn.countTokens();
		String[] lines = new String[num_lines];
		int[] line_widths = new int[num_lines];
		// Set the text to draw with the current label
		for (int i = 0; i < num_lines; i++)
			lines[i] = tkn.nextToken();
		// Caculate the line widh
		for (int i = 0; i < num_lines; i++)
		{
			line_widths[i] = fm.stringWidth(lines[i]);
		}

		int line_height = fm.getHeight();
		int line_ascent = fm.getAscent();
		LyEnvironment environment = getEnvironment();

		if (fillLabel)
		{
			// Draw a rectangle around the label
			if (_mapNode.isSelected() && !drawBorder && map.isActivated())
				g.setColor(getSelectedColor(environment));
			else if (_mapNode.isSelected() && !drawBorder && !map.isActivated())
				g.setColor(getInactiveSelectedColor(environment));
			else
				g.setColor(map.getComponent().getBackground());

			g.fillRect(xstart, ystart, labelDim.width, labelDim.height);
		}

		if (_mapNode.isSelected())
		{
			if (map.isActivated())
				g.setColor(getSelectedColor(environment));
			else
				g.setColor(getInactiveSelectedColor(environment));

			if (drawBorder)
				g.drawRect(xstart - 2, ystart, labelDim.width + 3, labelDim.height - 1);
			else
				g.fillRect(xstart, ystart, labelDim.width, labelDim.height);
		}

		String labelColor = _mapNode.getFontColor();
		if (labelColor == null)
			g.setColor(getFontColor(environment));
		else
			g.setColor(getColor(labelColor));
		g.setFont(currentFont);

		Insets insets = ((JComponent) component).getInsets();

		int y = ystart + line_ascent;

		for (int i = 0; i < num_lines; i++)
		{
			int x = xstart + insets.left + (labelDim.width - line_widths[i]) / 2;
			g.drawString(lines[i], x, y);
			y += line_height;
		}

	} // end method drawLabelAtPosition

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#drawInViewer(leon.view.win.swing.map.LyAbstractMapComponentManager,
	 *      java.awt.Graphics)
	 */
	public void drawInViewer(LyAbstractMapComponentManager map, Graphics g)
	{
		if (getBean() != null)
		{
			getBean().drawInViewer(map, g);
			return;
		}
		Point point = map.convert(_center);
		LyEnvironment environment = getEnvironment();

		point = map.convertToViewer(point);

		g.setColor(getNodeColor(environment));

		switch (_mapNode.getStyle())
		{
		case IMAGE:
			// Draw the image
			Image graphicalImage = getImage(environment);

			if (graphicalImage != null)
			{
				if (_mapNode.isResizable())
				{
					Dimension dim = getImageSize(map.getComponent());
					int width = dim.width;
					int height = dim.height;

					LyPoint point3D1 = new LyPoint(_center.getX() - (width / 2), _center.getY()
							- (height / 2), _center.getZ());
					LyPoint point3D2 = new LyPoint(_center.getX() + (width / 2), _center.getY()
							+ (height / 2), _center.getZ());

					Point point1 = map.convert(point3D1);
					Point point2 = map.convert(point3D2);

					point1 = map.convertToViewer(point1);
					point2 = map.convertToViewer(point2);

					g.drawImage(graphicalImage, point1.x, point1.y, point2.x - point1.x, point2.y
							- point1.y, map.getComponent());
				}
				else
					g.drawImage(graphicalImage, point.x - 3, point.y - 3, 6, 6, map.getComponent());
			}
			break;

		case CIRCLE:
			// Draw a circle
			g.fillOval(point.x - 2, point.y - 2, 4, 4);
			break;

		case SQUARE:
			g.fillRect(point.x - 2, point.y - 2, 4, 4);
			break;

		case TRIANGLE:
			g.fillPolygon(new int[] { point.x - 3, point.x, point.x + 3 }, new int[] { point.y + 3,
					point.y - 3, point.y + 3 }, 3);
			break;
		
		default:
			break;
		}
	} // end method drawInViewer

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#getBoundingBox(leon.view.win.swing.map.LyAbstractMapComponentManager)
	 */
	public Ly3DBox getBoundingBox(LyAbstractMapComponentManager map)
	{
		Dimension dim = getImageSize(map.getComponent());

		int width = dim.width;
		int height = dim.height;

		return new Ly3DBox(_center.getX() - (width / 2), _center.getY() - (height / 2), _center
				.getZ(), _center.getX() + (width / 2), _center.getY() + (height / 2), _center
				.getZ());
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#getBoundingRectangle(leon.view.win.swing.map.LyAbstractMapComponentManager)
	 */
	public Rectangle getBoundingRectangle(LyAbstractMapComponentManager map)
	{
		if (getBean() != null)
			return getBean().getBoundingRectangle(map);
		Point center = map.convert(_center);
		Dimension dimImage = getImageSize(map.getComponent());
		Point labelOffset = getLabelOffset(map, null, center.x, center.y);
		Point leftLabelOffset = getLeftLabelOffset(map, null);
		Point rightLabelOffset = getRightLabelOffset(map, null);
		Point topLabelOffset = getTopLabelOffset(map, null);
		Point bottomLabelOffset = getBottomLabelOffset(map, null);
		
		// Recompute the label if needed
		getRightLabelSize(map, null);
		getLeftLabelSize(map, null);
		getTopLabelSize(map, null);
		getBottomLabelSize(map, null);
		
		float zoom = map.getZoom();

		if (_mapNode.isResizable())
			dimImage = new Dimension((int) (dimImage.width * zoom), (int) (dimImage.height * zoom));

		int xLabelOffset = ((labelOffset == null) ? 0 : _labelOffset.x);
		int yLabelOffset = ((labelOffset == null) ? 0 : _labelOffset.y);
		int xLeftLabelOffset = ((leftLabelOffset == null) ? 0 : _leftLabelOffset.x);
		int yLeftLabelOffset = ((leftLabelOffset == null) ? 0 : _leftLabelOffset.y);
		int xRightLabelOffset = ((rightLabelOffset == null) ? 0 : _rightLabelOffset.x);
		int yRightLabelOffset = ((rightLabelOffset == null) ? 0 : _rightLabelOffset.y);
		int xTopLabelOffset = ((topLabelOffset == null) ? 0 : _topLabelOffset.x);
		int yTopLabelOffset = ((topLabelOffset == null) ? 0 : _topLabelOffset.y);
		int xBottomLabelOffset = ((bottomLabelOffset == null) ? 0 : _bottomLabelOffset.x);
		int yBottomLabelOffset = ((bottomLabelOffset == null) ? 0 : _bottomLabelOffset.y);

		Rectangle imageRect = new Rectangle(center.x - (dimImage.width / 2), center.y
				- (dimImage.height / 2), dimImage.width, dimImage.height);
		Rectangle labelRect = new Rectangle(center.x + xLabelOffset, center.y + yLabelOffset,
				_labelWidth, _labelHeight);
		Rectangle leftLabelRect = new Rectangle(center.x + xLeftLabelOffset, center.y
				+ yLeftLabelOffset, _leftLabelWidth, _leftLabelHeight);
		Rectangle rightLabelRect = new Rectangle(center.x + xRightLabelOffset, center.y
				+ yRightLabelOffset, _rightLabelWidth, _rightLabelHeight);
		Rectangle topLabelRect = new Rectangle(center.x + xTopLabelOffset, center.y
				+ yTopLabelOffset, _topLabelWidth, _topLabelHeight);
		Rectangle bottomLabelRect = new Rectangle(center.x + xBottomLabelOffset, center.y
				+ yBottomLabelOffset, _bottomLabelWidth, _bottomLabelHeight);

		Rectangle result = imageRect;

		if ((_labelWidth >= 0) && (_labelHeight >= 0))
			result = result.union(labelRect);

		if ((_leftLabelWidth >= 0) && (_leftLabelHeight >= 0))
			result = result.union(leftLabelRect);

		if ((_rightLabelWidth >= 0) && (_rightLabelHeight >= 0))
			result = result.union(rightLabelRect);

		if ((_topLabelWidth >= 0) && (_topLabelHeight >= 0))
			result = result.union(topLabelRect);

		if ((_bottomLabelWidth >= 0) && (_bottomLabelHeight >= 0))
			result = result.union(bottomLabelRect);

		return result;
	} // end method getBoundingRectangle

	/**
	 * Gets the clipping area of this node.
	 *
	 * @param  map   the map manager
	 * @param  image the image to clip. This image is not used in this implementation.
	 * @return the bounding rectangle of the map manager.
	 * @see    #getBoundingRectangle(LyAbstractMapComponentManager)
	 * @see    LyMapElement#getClippingArea(LyAbstractMapComponentManager, Image)
	 */
	public Shape getClippingArea(LyAbstractMapComponentManager map, Image image)
	{
		return getBoundingRectangle(map);
	}

	/**
	 * Returns the rectangle of the largest child node.
	 *
	 * @param  map   the map representing the node
	 * @param  alpha angle of node points alignment
	 * @return the rectangle of the largest child node
	 */
	public Rectangle getBoundingRectangleOfChild(LyAbstractMapComponentManager map, double alpha)
	{
		Rectangle r = null;
		Rectangle tmp = null;

		for (int i = 0; i < getNumberOfChildren(); i++)
		{
			LySwingMapNode aChild = getChildren().elementAt(i);

			switch (_mapNode.getLabelAlign())
			{
			case BOTTOM:

				if (aChild.getLayoutInfo().equalsIgnoreCase("extern_southeast")
						|| aChild.getLayoutInfo().equalsIgnoreCase("extern_southwest")
						|| aChild.getLayoutInfo().equalsIgnoreCase("extern_south"))
				{
					tmp = aChild.getBoundingRectangle(map);
					if (aChild._mapNode.getLabelAlign() == LyMapNode.LabelAlign.TOP)
						tmp.height -= aChild._labelHeight;
					if ((r == null) || (tmp.height > r.height))
						r = tmp;
				}
				if (aChild.getLayoutInfo().equalsIgnoreCase("southeast")
						|| aChild.getLayoutInfo().equalsIgnoreCase("southwest")
						|| aChild.getLayoutInfo().equalsIgnoreCase("south"))
				{
					tmp = aChild.getBoundingRectangle(map);
					tmp.height -= aChild.getImageSize(map.getComponent()).height;
					if (aChild._mapNode.getLabelAlign() == LyMapNode.LabelAlign.TOP)
						tmp.height -= aChild._labelHeight;
					if ((r == null) || (tmp.height > r.height))
						r = tmp;
				}
				break;

			case TOP:

				if (aChild.getLayoutInfo().equalsIgnoreCase("extern_northeast")
						|| aChild.getLayoutInfo().equalsIgnoreCase("extern_northwest")
						|| aChild.getLayoutInfo().equalsIgnoreCase("extern_north"))
				{
					tmp = aChild.getBoundingRectangle(map);
					if (aChild._mapNode.getLabelAlign() == LyMapNode.LabelAlign.BOTTOM)
						tmp.height -= aChild._labelHeight;
					if ((r == null) || (tmp.height > r.height))
						r = tmp;
				}
				if (aChild.getLayoutInfo().equalsIgnoreCase("northeast")
						|| aChild.getLayoutInfo().equalsIgnoreCase("northwest")
						|| aChild.getLayoutInfo().equalsIgnoreCase("north"))
				{
					tmp = aChild.getBoundingRectangle(map);
					tmp.height -= aChild.getImageSize(map.getComponent()).height;
					if (aChild._mapNode.getLabelAlign() == LyMapNode.LabelAlign.BOTTOM)
						tmp.height -= aChild._labelHeight;
					if ((r == null) || (tmp.height > r.height))
						r = tmp;
				}
				break;
			case LEFT:

				if (aChild.getLayoutInfo().equalsIgnoreCase("extern_west")
						|| aChild.getLayoutInfo().equalsIgnoreCase("extern_westnorth")
						|| aChild.getLayoutInfo().equalsIgnoreCase("extern_westsouth"))
				{
					tmp = aChild.getBoundingRectangle(map);
					if (aChild._mapNode.getLabelAlign() == LyMapNode.LabelAlign.RIGHT)
						tmp.width -= aChild._labelWidth;
					if ((r == null) || (tmp.width > r.width))
						r = tmp;
				}
				if (aChild.getLayoutInfo().equalsIgnoreCase("westnorth")
						|| aChild.getLayoutInfo().equalsIgnoreCase("westsouth")
						|| aChild.getLayoutInfo().equalsIgnoreCase("west"))
				{
					tmp = aChild.getBoundingRectangle(map);
					tmp.width -= aChild.getImageSize(map.getComponent()).width;
					if (aChild._mapNode.getLabelAlign() == LyMapNode.LabelAlign.RIGHT)
						tmp.width -= aChild._labelWidth;
					if ((r == null) || (tmp.width > r.width))
						r = tmp;
				}
				break;
			case RIGHT:

				if (aChild.getLayoutInfo().equalsIgnoreCase("extern_east")
						|| aChild.getLayoutInfo().equalsIgnoreCase("extern_eastnorth")
						|| aChild.getLayoutInfo().equalsIgnoreCase("extern_eastsouth"))
				{
					tmp = aChild.getBoundingRectangle(map);
					if (aChild._mapNode.getLabelAlign() == LyMapNode.LabelAlign.LEFT)
						tmp.width -= aChild._labelWidth;
					if ((r == null) || (tmp.width > r.width))
						r = tmp;
				}
				if (aChild.getLayoutInfo().equalsIgnoreCase("eastnorth")
						|| aChild.getLayoutInfo().equalsIgnoreCase("eastsouth")
						|| aChild.getLayoutInfo().equalsIgnoreCase("east"))
				{
					tmp = aChild.getBoundingRectangle(map);
					tmp.width -= aChild.getImageSize(map.getComponent()).width;
					if (aChild._mapNode.getLabelAlign() == LyMapNode.LabelAlign.LEFT)
						tmp.width -= aChild._labelWidth;
					if ((r == null) || (tmp.width > r.width))
						r = tmp;
				}
				break;
			case CENTER:

				break;
			case EXTERN:

				if ((alpha >= (-Math.PI / 4)) && (alpha < (Math.PI / 4)))
				{
					if (aChild.getLayoutInfo().equalsIgnoreCase("extern_east")
							|| aChild.getLayoutInfo().equalsIgnoreCase("extern_eastnorth")
							|| aChild.getLayoutInfo().equalsIgnoreCase("extern_eastsouth"))
					{
						tmp = aChild.getBoundingRectangle(map);
						if (aChild._mapNode.getLabelAlign() == LyMapNode.LabelAlign.LEFT)
							tmp.width -= aChild._labelWidth;
						if ((r == null) || (tmp.width > r.width))
							r = tmp;
					}
					if (aChild.getLayoutInfo().equalsIgnoreCase("eastnorth")
							|| aChild.getLayoutInfo().equalsIgnoreCase("eastsouth")
							|| aChild.getLayoutInfo().equalsIgnoreCase("east"))
					{
						tmp = aChild.getBoundingRectangle(map);
						tmp.width -= aChild.getImageSize(map.getComponent()).width;
						if (aChild._mapNode.getLabelAlign() == LyMapNode.LabelAlign.LEFT)
							tmp.width -= aChild._labelWidth;
						if ((r == null) || (tmp.width > r.width))
							r = tmp;
					}
				}
				else if ((alpha >= (Math.PI / 4)) && (alpha < (3 * Math.PI / 4)))
				{
					if (aChild.getLayoutInfo().equalsIgnoreCase("extern_northeast")
							|| aChild.getLayoutInfo().equalsIgnoreCase("extern_northwest")
							|| aChild.getLayoutInfo().equalsIgnoreCase("extern_north"))
					{
						tmp = aChild.getBoundingRectangle(map);
						if (aChild._mapNode.getLabelAlign() == LyMapNode.LabelAlign.BOTTOM)
							tmp.height -= aChild._labelHeight;
						if ((r == null) || (tmp.height > r.height))
							r = tmp;
					}
					if (aChild.getLayoutInfo().equalsIgnoreCase("northeast")
							|| aChild.getLayoutInfo().equalsIgnoreCase("northwest")
							|| aChild.getLayoutInfo().equalsIgnoreCase("north"))
					{
						tmp = aChild.getBoundingRectangle(map);
						tmp.height -= aChild.getImageSize(map.getComponent()).height;
						if (aChild._mapNode.getLabelAlign() == LyMapNode.LabelAlign.BOTTOM)
							tmp.height -= aChild._labelHeight;
						if ((r == null) || (tmp.height > r.height))
							r = tmp;
					}
				}
				else if (((alpha >= (3 * Math.PI / 4)) && (alpha <= Math.PI))
						|| ((alpha >= -Math.PI) && (alpha < (-3 * Math.PI / 4))))
				{
					if (aChild.getLayoutInfo().equalsIgnoreCase("extern_west")
							|| aChild.getLayoutInfo().equalsIgnoreCase("extern_westnorth")
							|| aChild.getLayoutInfo().equalsIgnoreCase("extern_westsouth"))
					{
						tmp = aChild.getBoundingRectangle(map);
						if (aChild._mapNode.getLabelAlign() == LyMapNode.LabelAlign.RIGHT)
							tmp.width -= aChild._labelWidth;
						if ((r == null) || (tmp.width > r.width))
							r = tmp;
					}
					if (aChild.getLayoutInfo().equalsIgnoreCase("westnorth")
							|| aChild.getLayoutInfo().equalsIgnoreCase("westsouth")
							|| aChild.getLayoutInfo().equalsIgnoreCase("west"))
					{
						tmp = aChild.getBoundingRectangle(map);
						tmp.width -= aChild.getImageSize(map.getComponent()).width;
						if (aChild._mapNode.getLabelAlign() == LyMapNode.LabelAlign.RIGHT)
							tmp.width -= aChild._labelWidth;
						if ((r == null) || (tmp.width > r.width))
							r = tmp;
					}
				}
				else
				{
					if (aChild.getLayoutInfo().equalsIgnoreCase("extern_southeast")
							|| aChild.getLayoutInfo().equalsIgnoreCase("extern_southwest")
							|| aChild.getLayoutInfo().equalsIgnoreCase("extern_south"))
					{
						tmp = aChild.getBoundingRectangle(map);
						if (aChild._mapNode.getLabelAlign() == LyMapNode.LabelAlign.TOP)
							tmp.height -= aChild._labelHeight;
						if ((r == null) || (tmp.height > r.height))
							r = tmp;
					}
					if (aChild.getLayoutInfo().equalsIgnoreCase("southeast")
							|| aChild.getLayoutInfo().equalsIgnoreCase("southwest")
							|| aChild.getLayoutInfo().equalsIgnoreCase("south"))
					{
						tmp = aChild.getBoundingRectangle(map);
						tmp.height -= aChild.getImageSize(map.getComponent()).height;
						if (aChild._mapNode.getLabelAlign() == LyMapNode.LabelAlign.TOP)
							tmp.height -= aChild._labelHeight;
						if ((r == null) || (tmp.height > r.height))
							r = tmp;
					}
				}
				break;
			}
		}
		return r;
	} // end method getBoundingRectangleOfChild

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#getLayer()
	 */
	public int getLayer()
	{
		int layer = _mapNode.getLayer();

		return ((layer < 0) ? LyAbstractMapComponentManager.NODE_LAYER : layer);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#getLabel()
	 */
	public String getLabel()
	{
		if (_labelCache == null)
			_labelCache = translate(_mapNode.getLabel());

		return _labelCache;
	}

	/**
	 * Implementation of LyMapElement.
	 */
	public String getLeftLabel()
	{
		if (_leftLabelCache == null)
			_leftLabelCache = translate(_mapNode.getLeftLabel());

		return _leftLabelCache;
	}

	/**
	 * Implementation of LyMapElement.
	 */
	public String getRightLabel()
	{
		if (_rightLabelCache == null)
			_rightLabelCache = translate(_mapNode.getRightLabel());

		return _rightLabelCache;
	}

	/**
	 * Implementation of LyMapElement.
	 */
	public String getTopLabel()
	{
		if (_topLabelCache == null)
			_topLabelCache = translate(_mapNode.getTopLabel());

		return _topLabelCache;
	}

	/**
	 * Implementation of LyMapElement.
	 */
	public String getBottomLabel()
	{
		if (_bottomLabelCache == null)
			_bottomLabelCache = translate(_mapNode.getBottomLabel());

		return _bottomLabelCache;
	}

	/**
	 * Sets the label attribute and its offset to <code>null</code> in order to get rid of cache.
	 *
	 * @param label This attribute is not used in this implementation.
	 * @see   #getLabel()
	 */
	public void setLabel(String label)
	{
		// Get rid of cache
		_labelCache = null;
		_labelOffset = null;
		checkLabels();
	}

	/**
	 * Sets attribute left label to the given value.
	 *
	 * @param label New value of left label
	 */
	public void setLeftLabel(String label)
	{
		// Get rid of cache
		_leftLabelCache = null;
		_leftLabelOffset = null;
		checkLabels();
	}

	/**
	 * Sets attribute right label to the given value.
	 *
	 * @param label New value of right label
	 */
	public void setRightLabel(String label)
	{
		// Get rid of cache
		_rightLabelCache = null;
		_rightLabelOffset = null;
		checkLabels();
	}

	/**
	 * Sets attribute top label to the given value.
	 *
	 * @param label New value of top label
	 */
	public void setTopLabel(String label)
	{
		// Get rid of cache
		_topLabelCache = null;
		_topLabelOffset = null;
		checkLabels();
	}

	/**
	 * Sets attribute bottom label to the given value.
	 *
	 * @param label New value of bottom label
	 */
	public void setBottomLabel(String label)
	{
		// Get rid of cache
		_bottomLabelCache = null;
		_bottomLabelOffset = null;
		checkLabels();
	}

	/**
	 * Checks the consistency of the label values (label, leftLabel, rightLabel, topLabel and
	 * bottomLabel). For example, if both label with a left alignement and leftLabel are set, a
	 * warning will be displayed to indicate that labels may overlap.
	 */
	protected void checkLabels()
	{
		String warning = null;

		if ((_mapNode.getLabel() != null)
				&& (_mapNode.getLabelAlign() == LyMapNode.LabelAlign.LEFT)
				&& (_mapNode.getLeftLabel() != null))
			warning = "Both left label and label with left align are set on map node "
					+ _mapNode.getId();
		else if ((_mapNode.getLabel() != null)
				&& (_mapNode.getLabelAlign() == LyMapNode.LabelAlign.RIGHT)
				&& (_mapNode.getRightLabel() != null))
			warning = "Both right label and label with right align are set on map node "
					+ _mapNode.getId();
		else if ((_mapNode.getLabel() != null)
				&& (_mapNode.getLabelAlign() == LyMapNode.LabelAlign.TOP)
				&& (_mapNode.getTopLabel() != null))
			warning = "Both top label and label with top align are set on map node "
					+ _mapNode.getId();
		else if ((_mapNode.getLabel() != null)
				&& (_mapNode.getLabelAlign() == LyMapNode.LabelAlign.BOTTOM)
				&& (_mapNode.getBottomLabel() != null))
			warning = "Both bottom label and label with bottom align are set on map node "
					+ _mapNode.getId();

		if (warning != null)
		{
			LyEnvironment environment = getEnvironment();

			environment.logWarning("[LyWinMapNode] " + warning
					+ ". Label will not be displayed correctly.");
		}
	} // end method checkLabels

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#getTooltip()
	 */
	public String getTooltip()
	{
		return translate(_mapNode.getTooltip());
	}

	/**
	 * Do not set tooltips of a map node as it is not used.
	 *
	 * @param tooltip This attribute is not used in this implementation.
	 * @see   #getTooltip()
	 */
	public void setTooltip(String tooltip)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#isAtPosition(leon.view.win.swing.map.LyAbstractMapComponentManager,
	 *      java.awt.Graphics, int, int)
	 */
	public boolean isAtPosition(LyAbstractMapComponentManager map, Graphics g, int x, int y)
	{
		if (getBean() != null)
			return getBean().isAtPosition(map, g, x, y);

		if (isHidden())
			return false;

		Dimension dimImage = getImageSize(map.getComponent());

		if (_mapNode.isResizable())
		{
			float zoom = map.getZoom();

			dimImage = new Dimension((int) (dimImage.width * zoom), (int) (dimImage.height * zoom));
		}

		Dimension dimLabel = getLabelSize(map, g);
		Dimension dimLeftLabel = getLeftLabelSize(map, g);
		Dimension dimRightLabel = getRightLabelSize(map, g);
		Dimension dimTopLabel = getTopLabelSize(map, g);
		Dimension dimBottomLabel = getBottomLabelSize(map, g);
		Point point = map.convert(_center);
		Point labelOffset = getLabelOffset(map, g, x, y);
		Point leftLabelOffset = getLeftLabelOffset(map, g);
		Point rightLabelOffset = getRightLabelOffset(map, g);
		Point topLabelOffset = getTopLabelOffset(map, g);
		Point bottomLabelOffset = getBottomLabelOffset(map, g);
		int xnode = point.x, ynode = point.y;
		int xlabel = (point.x + labelOffset.x);
		int ylabel = (point.y + labelOffset.y);
		int xleftlabel = (point.x + leftLabelOffset.x);
		int yleftlabel = (point.y + leftLabelOffset.y);
		int xrightlabel = (point.x + rightLabelOffset.x);
		int yrightlabel = (point.y + rightLabelOffset.y);
		int xtoplabel = (point.x + topLabelOffset.x);
		int ytoplabel = (point.y + topLabelOffset.y);
		int xbottomlabel = (point.x + bottomLabelOffset.x);
		int ybottomlabel = (point.y + bottomLabelOffset.y);

		// Check if (x, y) in the image
		if ((x >= (xnode - (dimImage.width / 2))) && (x <= (xnode + (dimImage.width / 2)))
				&& (y >= (ynode - (dimImage.height / 2))) && (y <= (ynode + (dimImage.height / 2))))
			return true;

		// Check if (x,y) in the label
		if ((x >= xlabel) && (x <= (xlabel + dimLabel.width)) && (y >= ylabel)
				&& (y <= (ylabel + dimLabel.height)))
			return true;

		// Check if (x,y) in the left label
		if ((x >= xleftlabel) && (x <= (xleftlabel + dimLeftLabel.width)) && (y >= yleftlabel)
				&& (y <= (yleftlabel + dimLeftLabel.height)))
			return true;

		// Check if (x,y) in the right label
		if ((x >= xrightlabel) && (x <= (xrightlabel + dimRightLabel.width)) && (y >= yrightlabel)
				&& (y <= (yrightlabel + dimRightLabel.height)))
			return true;

		// Check if (x,y) in the top label
		if ((x >= xtoplabel) && (x <= (xtoplabel + dimTopLabel.width)) && (y >= ytoplabel)
				&& (y <= (ytoplabel + dimTopLabel.height)))
			return true;

		// Check if (x,y) in the bottom label
		if ((x >= xbottomlabel) && (x <= (xbottomlabel + dimBottomLabel.width))
				&& (y >= ybottomlabel) && (y <= (ybottomlabel + dimBottomLabel.height)))
			return true;

		return false;
	} // end method isAtPosition

	/*
	 * (non-Javadoc).
	 *
	 * @see leon.view.getinterface.LyGetIdInterface#getId()
	 */
	public String getId()
	{
		return _mapNode.getId();
	}

	/*
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#setSelected(boolean)
	 */
	public void setSelected(boolean selected)
	{
		if ((_mapNode.isSensitive()) && (_mapNode.isSelected() != selected))
			_mapNode.setSelected(selected);
	}

	/*
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#isHidden()
	 */
	public boolean isHidden()
	{
		return _mapNode.isHidden();
	}

	/*
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetHiddenInterface#setHidden(boolean)
	 */
	public void setHidden(boolean hidden)
	{
		if (_mapNode.isHidden() != hidden)
			_mapNode.setHidden(hidden);
	}

	/**
	 * Checks if value of attribute movable is set.
	 * @return true if movable is set, false otherwise.
	 */
	public boolean isMovable()
	{
		return _mapNode.isMovable();
	}

	/**
	 * Initializes attribute movable to the given value.
	 * @param movable Boolean indicating if attribute movable is set or unset.
	 */
	public void setMovable(boolean movable)
	{
		if (_mapNode.isMovable() != movable)
			_mapNode.setMovable(movable);
	}

	/**
	 * Sets this node image name to <code>null</code> to get rid of cache.
	 *
	 * @param image This attribute is not used in this implementation.
	 * @see   #getImage(LyEnvironment)
	 * @see   leon.view.setinterface.LySetImageInterface#setImage(java.lang.String)
	 */
	public void setImage(String image)
	{
		// Get rid of cache
		_imageString = null;
	}

	/*
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalMapNode#setX(float)
	 */
	public void setX(float x)
	{
		_center.setX(x);
	}

	/*
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalMapNode#setY(float)
	 */
	public void setY(float y)
	{
		_center.setY(y);
	}

	/*
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalMapNode#setZ(float)
	 */
	public void setZ(float z)
	{
		_center.setZ(z);
	}
	
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalMapNode#setLayer(int)
	 */
	public void setLayer(int layer)
	{
	}

	/*
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalMapNode#setColorFrameInt(java.lang.String)
	 */
	public void setColorFrameInt(String colorFrameInt)
	{
		_mapNode.setColorFrameInt(colorFrameInt);
	}

	/*
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalMapNode#setColorFrameExt(java.lang.String)
	 */
	public void setColorFrameExt(String colorFrameExt)
	{
		_mapNode.setColorFrameExt(colorFrameExt);
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalMapNode#setFontColor(java.lang.String)
	 */
	public void setFontColor(String labelColor)
	{
		_mapNode.setFontColor(labelColor);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#isSelected()
	 */
	public boolean isSelected()
	{
		return _mapNode.isSelected();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#translate(float, float, float)
	 */
	public void translate(float dx, float dy, float dz)
	{
		if (!_mapNode.isMovable())
			return;

		_center.translate(dx, dy, dz);

		_mapNode.setX(_center.getX());
		_mapNode.setY(_center.getY());
		_mapNode.setZ(_center.getZ());
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyLayElement#setPosition(leon.view.LyPoint)
	 */
	public void setPosition(LyPoint point3D)
	{
		_center = point3D;

		_mapNode.setX(_center.getX());
		_mapNode.setY(_center.getY());
		_mapNode.setZ(_center.getZ());
	}

	/**
	 * A map node has always the overlay property.
	 *
	 * @return <code>true</code>
	 * @see    leon.view.win.swing.map.LyMapElement#isOverlay()
	 */
	public boolean isOverlay()
	{
        return isSelected();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#isSensitive()
	 */
	public boolean isSensitive()
	{
		return _mapNode.isSensitive();
	}

	/**
	 * Sets the offset between this component and its parent.
	 *
	 * @param offset the new value of the offset
	 * @see   #getOffset()
	 */
	public void setOffset(Point offset)
	{
		_offset = offset;
	}

	/**
	 * Sets the parent node of this node.
	 *
	 * @param node the parent node
	 * @see   #getParentNode()
	 */
	public void setParentNode(LySwingMapNode node)
	{
		_parentNode = node;
	}

	/**
	 * Sets the child node of this node if the child element is not already in the children list.
	 *
	 * @param node the child node
	 * @see   #getChildren()
	 */
	public void setChild(LySwingMapNode node)
	{
		if (!_children.contains(node))
			_children.addElement(node);
	}

	/**
	 * Returns the center of the node.
	 *
	 * @return the center of the node
	 */
	public LyPoint getCenter()
	{
		if (getBean() != null)
		{
			LyPoint center = getBean().getCenter();

			if (center != null)
				_center = center;
		}
		return _center;
	}

	/**
	 * Returns the shift between the node and his parent.
	 *
	 * @return the offset between the node and its parent
	 * @see    #setOffset(Point)
	 */
	public Point getOffset()
	{
		return _offset;
	}

	/**
	 * Returns the parent node of the node.
	 *
	 * @return the parent node
	 * @see    #setParentNode(LySwingMapNode)
	 */
	public LySwingMapNode getParentNode()
	{
		return _parentNode;
	}

	/**
	 * Returns all children of the node.
	 *
	 * @return the children of the node
	 * @see    #setChild(LySwingMapNode)
	 */
	public Vector<LySwingMapNode> getChildren()
	{
		return _children;
	}

	/**
	 * Returns the children count.
	 *
	 * @return the number of children of the node
	 */
	public int getNumberOfChildren()
	{
		return _children.size();
	}

	/**
	 * Returns the layout info of the node.
	 *
	 * @return the layout info
	 */
	public String getLayoutInfo()
	{
		return _mapNode.getLayoutInfo();
	}

	/**
	 * Build the bean associated to this node.
	 *
	 * @return the bean or null if none
	 */
	public LyMapNodeBeanInterface getBean()
	{
		if (_nodeBean != null)
			return _nodeBean;

		String beanJavaName = _mapNode.getJavaName();

		if (beanJavaName == null)
			return null;

		Class<?> beanClass = null;
		ClassLoader cl = getEnvironment().getClassLoader();

		try
		{
			if (cl != null)
				beanClass = cl.loadClass(beanJavaName);
			else
				beanClass = Class.forName(beanJavaName);

			if (beanClass == null)
				return null;

			_nodeBean = (LyMapNodeBeanInterface) beanClass.newInstance();
		}
		catch (ClassNotFoundException e)
		{
			getEnvironment().writeException(e);
			_nodeBean = null;
		}
		catch (InstantiationException e)
		{
			getEnvironment().writeException(e);
			_nodeBean = null;
		}
		catch (IllegalAccessException e)
		{
			getEnvironment().writeException(e);
			_nodeBean = null;
		}
		if (_nodeBean == null)
			return null;

		_nodeBean.setMapNode(_mapNode);
		return _nodeBean;
	} // end method getBean

	/**
	 * Free the map node.
	 */
	public void free()
	{
		if (getBean() != null)
			getBean().free();
		super.free();
	}

	/**
	 * getMapNodeStyle
	 * 
	 * @return
	 */
	public leon.view.LyMapNode.Style getMapNodeStyle()
	{
		return _mapNode.getStyle();
	}

	@Override
	public String toString()
	{
		StringBuffer sb = new StringBuffer();
		sb.append(getId());
		sb.append(": ");
		sb.append(getCenter());

		return sb.toString();
	}

	/**
	 * Get the current font metrics according to the current zoom level of the map
	 * @param map      the map representing the node
	 * @param  g           the graphics context
	 * @return
	 */
	@SuppressWarnings("deprecation")
	private FontMetrics getCurrentFontMetrics(LyAbstractMapComponentManager map, Graphics g)
	{
		FontMetrics fm;
		Font currentFont = getCurrentFont(map);
		if (g == null)
		{
			Toolkit toolkit = Toolkit.getDefaultToolkit();
			fm = toolkit.getFontMetrics(currentFont);
		}
		else
		{
			fm = g.getFontMetrics(currentFont);
		}
		return fm;
	}

	/**
	 * Get the current font accordingly to the current zoom level of the map
	 * @param map the map representing the node
	 * @return current font
	 */
	protected final Font getCurrentFont(LyAbstractMapComponentManager map)
	{
		String	nodeFont = _mapNode.getFont();
		
		if (nodeFont == null)
			nodeFont = "LY_MAPNODE_FONT";
		
		float defFontSize = getFont(nodeFont).getSize2D();
		float expectedFontSize = isLabelZoomable() ? defFontSize * map.getZoom() : defFontSize;
		if (_graphicalFont.getSize2D() != expectedFontSize)
		{
			// Update the current font size according to current zoom level
			_graphicalFont = _graphicalFont.deriveFont(expectedFontSize);

			// Box will to be recalculated because of font change
			_labelWidth = NOT_INITIALIZED;
			_labelHeight = NOT_INITIALIZED;
			_leftLabelWidth = NOT_INITIALIZED;
			_leftLabelHeight = NOT_INITIALIZED;
			_rightLabelWidth = NOT_INITIALIZED;
			_rightLabelHeight = NOT_INITIALIZED;
			_topLabelWidth = NOT_INITIALIZED;
			_topLabelHeight = NOT_INITIALIZED;
			_bottomLabelWidth = NOT_INITIALIZED;
			_bottomLabelHeight = NOT_INITIALIZED;
			_labelOffset = null;
			_leftLabelOffset = null;
			_rightLabelOffset = null;
			_topLabelOffset = null;
			_bottomLabelOffset = null;
		}

		return _graphicalFont;
	}

	/**
	 * Return true is displayed label should be zoom dependant.
	 * @return true is label are zoomable
	 */
	public boolean isLabelZoomable()
	{
		// Supposed that a label is zoomable if map node is resizable only.
		return _mapNode.isResizable();
	}

	public void setViewerImage(String viewerImage)
	{
	}
} // end class LySwingMapNode
