/*
 * LySwingMapRegion.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.util.*;
import leon.misc.*;
import leon.view.*;
import leon.view.common.*;
import leon.view.graphic.*;
import leon.view.win.swing.map.*;


/**
 * This class implements the Map region object (as described in the dtd) for swing viewer.
 *
 * <p>A region allows to display on a map a graphical element made up of a finite set of points.
 * These points define polygons that are displayed on the map with a given color. These elements
 * follow the zooms made on the map (their size increases or decreases when zooming in or out).</p>
 *
 * @version $Id: LySwingMapRegion.java 28997 2010-12-23 10:17:15Z bmichel $
 * @author  Lyria
 * @see     leon.view.LyMapRegion
 */
public class LySwingMapRegion extends LySwingMapObject
	implements LyGraphicalMapRegion,
		LyMapElement
{
	public static final String __VERSION =
		"$Id: LySwingMapRegion.java 28997 2010-12-23 10:17:15Z bmichel $";

	/** Offset for region shadow. */
	public static final int    SHADOW_OFFSET = 3;

	/** The view independent map region description. */
	protected LyMapRegion	   _mapRegion;

	/**
	 * List of 2D polygons representing the region as another polygon in the map. Used to determine
	 * if a click happened in the region.
	 */
	protected Vector<Shape>    _drawedPolygons;

	/** Font of the region. */
	protected Font			   _graphicalFont;

	/** Font of the tooltip when mouse over. */
	protected Font			   _tooltipFont;

	/** Color of the font of the region. */
	protected Color			   _fontColor;

	/** Frame used for the outside label. */
	protected Rectangle		   _labelFrame;

	/** Bounds of the small image. */
	protected Rectangle		   _smallImageBounds;

	/** Indicates if this mouse is over the map region. */
	protected boolean		   _mouseOver;

	/** The x coordonate of the mouse when mouse over. */
	protected int			   _mouseOverX;

	/** The y coordonate of the mouse when mouse over. */
	protected int			   _mouseOverY;

	/** Cached region label (for optimisation). */
	protected String		   _labelCache;

	/** Cached translated region label (for optimisation). */
	protected String		   _translatedLabelCache;

	/**
	 * Class constructor of a Map region object for the Swing target.
	 *
	 * @param mapRegion a MAP_REGION component defined for every type of view targets. This map
	 *                  region is initialized for the swing view and its attributes are given
	 *                  default values.
	 */
	public LySwingMapRegion(LyMapRegion mapRegion)
	{
		super(mapRegion);
		_mapRegion = mapRegion;
		_labelFrame = null;
		_smallImageBounds = null;
		_mouseOver = false;
		_mouseOverX = -1;
		_mouseOverY = -1;
		_labelCache = null;
		_translatedLabelCache = null;

		setContent();
	}

	/**
	 * Sets the content of this component and initializes its default colors and fonts.
	 */
	public void setContent()
	{
		LyEnvironment environment = getEnvironment();

		// Build the polygons of the region (if given in an extra file)
		_mapRegion.loadPolygonsFile(environment);

		_graphicalFont = getFont("LY_MAPNODE_FONT");
		_tooltipFont = getFont("LY_TOOL_TIP_FONT");
		_fontColor = getColor("LY_MAP_FONT_COLOR");
	}

	/**
	 * Sets the mouse over mode.
	 *
	 * @param mouseOver true if map region must be active, false otherwise
	 * @param x         the x coordinate of the mouse
	 * @param y         the y coordinate of the mouse
	 */
	public void setMouseOver(boolean mouseOver, int x, int y)
	{
		_mouseOver = mouseOver;
		_mouseOverX = x;
		_mouseOverY = y;
	}

	/**
	 * This map region component does not use any images.
	 *
	 * @return <code>null</code>
	 */
	public Image[] getImages()
	{
		return null;
	}

	/**
	 * Draws the map region.
	 *
	 * @see #draw(LyAbstractMapComponentManager, Graphics, int, boolean, int)
	 * @see leon.view.win.swing.map.LyMapElement#draw(leon.view.win.swing.map.LyAbstractMapComponentManager,
	 *      java.awt.Graphics, int)
	 */
	public void draw(LyAbstractMapComponentManager map, Graphics g, int mode)
	{
		draw(map, g, mode, false, 0);
	}

	/**
	 * Draws the element in the map.
	 *
	 * @param map       the map
	 * @param g         the graphical context
	 * @param mode      drawing mode
	 * @param forceDraw force the drawing of its sub-elements
	 * @param level     level of the element to draw
	 */
	protected void draw(LyAbstractMapComponentManager map, Graphics g, int mode, boolean forceDraw,
		int level)
	{
		if (isHidden())
			return;

		// If map region has a parent map region, parent will draw its sub elements itself
		if (!forceDraw && (_mapRegion.getParent() instanceof LyMapRegion))
			return;

		LyLayoutManager layout = map.getLayoutManager();
		boolean		    isTreemapLayout = (layout instanceof LyTreemapLayout);

		_smallImageBounds = null;

		// If region has no weight, do not display it
		if (isTreemapLayout && (_mapRegion.getWeight() <= 0))
			return;

		_drawedPolygons = new Vector<Shape>();

		int			  i, n = _mapRegion.getPolygonCount();
		LyPolygon     polygon;
		LyEnvironment env = (isFreed() ? null : getEnvironment());

		// A paint may be called after freeing component. Check it to avoid an exception.
		if (isFreed() || (env == null))
			return;
		synchronized (_mapRegion)
		{
			// If a map region has children, draw its children, otherwise draw its polygon
			if (_mapRegion.getMapRegionCount() == 0)
			{
				for (i = 0; i < n; i++)
				{
					polygon = _mapRegion.getPolygon(i);

					drawPolygon(map, g, polygon, false, mode);
				}
			}
			else
			{
				LyMapRegion		 region;
				LySwingMapRegion winRegion;

				n = _mapRegion.getMapRegionCount();

				for (i = 0; i < n; i++)
				{
					region = _mapRegion.getMapRegion(i);
					env = region.getEnvironment();

					// A paint may be called after freeing component. Check it to avoid an
					// exception.
					if (env == null)
						continue;

					winRegion = (LySwingMapRegion)region.getInternalView();
					winRegion.draw(map, g, mode, true, level + 1);
				}

				// Draw selection even if region has children
				n = _mapRegion.getPolygonCount();

				if (_mapRegion.isSelected())
				{
					for (i = 0; i < n; i++)
					{
						polygon = _mapRegion.getPolygon(i);

						drawPolygon(map, g, polygon, false, LyMapElement.REDUCED_MODE);
					}
				}
			}

			if (isTreemapLayout)
			{
				// For treemap layout draw the label of the region
				LyTreemapLayout treemapLayout = (LyTreemapLayout)layout;
				short		    labelStyle = treemapLayout.getLabelStyle(level);
				Rectangle	    bounds = getBoundingRectangle(map);
				String		    label = getLabel();
				int			    smallImageWidth = 0;
				int			    smallImageHeight = 0;
				boolean		    displaySmallImage = false;

				// A paint may be called after freeing component. Check it to avoid an exception.
				if (isFreed())
					return;

				// Draw the small image
				Image smallImage = getImage(_mapRegion.getSmallImage());
				int   xImage = -1, yImage = -1;

				if (smallImage != null)
				{
					smallImageWidth = smallImage.getWidth(map.getComponent()) + 1;
					smallImageHeight = smallImage.getHeight(map.getComponent());
					displaySmallImage = ((bounds.height > (smallImageHeight + 1))
								&& (bounds.width > (smallImageWidth + 1)));
				}

				if ((label != null) && (label.length() > 0))
				{
					FontMetrics fm = g.getFontMetrics(_graphicalFont);
					int		    xLabel = 0, yLabel = 0;
					int		    labelWidth = fm.stringWidth(label), l = label.length();
					int		    width = bounds.width;

					if (labelStyle == LyTreemapLayout.LABEL_OUTSIDE)
						bounds = _labelFrame;

					if (_labelFrame != null)
					{
						Rectangle border = new Rectangle(_labelFrame.x, _labelFrame.y,
								_labelFrame.width, _labelFrame.height).intersection(bounds);

						g.setColor(getColor("LY_TREEMAP_OUTSIDE_BACKGROUND_COLOR"));
						g.fillRect(border.x, border.y, border.width, border.height);
						g.setColor(getColor("LY_TREEMAP_OUTSIDE_BORDER_COLOR"));
						g.drawRect(border.x, border.y, border.width, border.height);
					}

					// Do not display label if not enough height
					if ((bounds != null) && (bounds.height >= fm.getHeight()))
					{
						// Do not display entire label if not enough space
						if (((labelWidth + smallImageWidth) > width)
								&& (labelStyle != LyTreemapLayout.LABEL_NONE))
						{
							while ((labelWidth + smallImageWidth) > width)
							{
								if (label.length() <= 1)
									break;

								l--;

								if (l == 0)
								{
									label = "";
									break;
								}

								label = label.substring(0, l) + "...";
								labelWidth = fm.stringWidth(label);
							}
						}

						if (labelStyle == LyTreemapLayout.LABEL_CENTER)
						{
							xImage = bounds.x
								+ ((bounds.width - fm.stringWidth(label) - smallImageWidth) / 2);
							yImage = bounds.y + ((bounds.height - smallImageHeight) / 2);
							xLabel = xImage + smallImageWidth + 1;
							yLabel = bounds.y + ((bounds.height - fm.getHeight()) / 2)
								+ fm.getAscent();
						}
						else if (labelStyle == LyTreemapLayout.LABEL_TOPLEFT)
						{
							xImage = bounds.x + 1;
							xLabel = xImage + smallImageWidth + 1;

							if (smallImageHeight > fm.getHeight())
							{
								yImage = bounds.y + 1;
								yLabel = bounds.y + ((smallImageHeight + 1 - fm.getHeight()) / 2)
									+ fm.getAscent();
							}
							else
							{
								yImage = bounds.y + ((fm.getHeight() - smallImageHeight) / 2);
								yLabel = bounds.y + fm.getAscent();
							}
						}
						else if (labelStyle == LyTreemapLayout.LABEL_OUTSIDE)
						{
							xImage = bounds.x + 1;
							yImage = bounds.y + bounds.height + 2;
							xLabel = xImage + ((bounds.width - fm.stringWidth(label)) / 2);
							yLabel = bounds.y + bounds.height - fm.getHeight() + fm.getAscent();
						}

						if (labelStyle != LyTreemapLayout.LABEL_NONE)
						{
							g.setFont(_graphicalFont);
							g.setColor(_fontColor);

							if (_labelFrame != null)
								g.setColor(getColor("LY_TREEMAP_OUTSIDE_LABEL_COLOR"));

							g.drawString(label, xLabel, yLabel);
						}
					}
				}

				if (displaySmallImage)
				{
					// If image position has not been computed (no label for example), put the image
					// in the top left corner
					if ((xImage < 0) && (yImage < 0))
					{
						xImage = bounds.x + 1;
						yImage = bounds.y + 1;
					}

					_smallImageBounds = new Rectangle(xImage, yImage, smallImageWidth,
							smallImageHeight);
					g.drawImage(smallImage, xImage, yImage, map.getComponent());
				}
			}
		}
	} // end method draw

	/**
	 * Draws this map region's mouse over tooltip (for a treemap layout only).
	 *
	 * @param map the treemap layout to draw its tooltip
	 * @param g   the graphics context
	 */
	public void drawMouseOverTooltip(LyAbstractMapComponentManager map, Graphics g)
	{
		LyEnvironment env = getEnvironment();
		Color		  tooltipColor = getColor("LY_TOOLTIP_COLOR");
		FontMetrics   fm = g.getFontMetrics(_tooltipFont);
		Rectangle     bounds = getBoundingRectangle(map);
		Rectangle     mapBounds = map.getComponent().getBounds();
		int			  x = 0, y = 0, compWidth, width = 0, height = 0;
		int			  gap = 5;
		String		  tooltip = null;

		if ((_smallImageBounds != null) && (_smallImageBounds.contains(_mouseOverX, _mouseOverY)))
		{
			x = _smallImageBounds.x + (_smallImageBounds.width / 2);
			y = _smallImageBounds.y + (_smallImageBounds.height / 2);
			compWidth = _smallImageBounds.width;
			tooltip = _mapRegion.getSmallImageTooltip();
		}
		else
		{
			x = bounds.x + (bounds.width / 2);
			y = bounds.y + (bounds.height / 2);
			compWidth = bounds.width;
			tooltip = _mapRegion.getTooltip();
		}

		if ((tooltip != null) && (tooltip.length() > 0))
		{
			tooltip = env.translate(tooltip);

			StringTokenizer t = new StringTokenizer(tooltip, "\n");
			String		    line;
			Vector<String>  lines = new Vector<String>();
			int			    i, n = 0;

			while (t.hasMoreElements())
			{
				line = t.nextToken();
				lines.addElement(line);
				width = Math.max(width, fm.stringWidth(line));
				height += fm.getHeight();
				n++;
			}

			width += 2 * gap;
			height += 2 * gap;

			if (x > (mapBounds.width / 2))
				// Draw on the left
				x = Math.max(gap, x - (compWidth / 2) - width - gap);
			else
				// Draw on the right
				x = Math.min(mapBounds.width - width - gap, x + (compWidth / 2) + gap);

			y = y - (height / 2);
			y = Math.max(gap, y);
			y = Math.min(mapBounds.height - height - gap, y);

			// Check if tooltip is over region
			if (bounds.intersects(x, y, width, height))
			{
				int upMargin = bounds.y;
				int bottomMargin = mapBounds.height - (bounds.y + bounds.height);

				if (upMargin >= bottomMargin)
				{
					y = bounds.y - gap - height;
					y = Math.max(gap, y);
				}
				else
				{
					y = bounds.y + bounds.height + gap;
					y = Math.min(mapBounds.height - height - gap, y);
				}
			}

			g.setColor(tooltipColor);
			g.fillRect(x, y, width, height);
			g.setColor(Color.BLACK);
			g.drawRect(x, y, width, height);
			g.setFont(_tooltipFont);

			x += gap;
			y += fm.getAscent() + gap;

			for (i = 0; i < n; i++)
			{
				line = lines.elementAt(i);
				g.drawString(line, x, y);
				y += fm.getHeight();
			}
		}
	} // end method drawMouseOverTooltip

	/**
	 * Draws a polygon on the map canvas.
	 *
	 * @param map      the map canvas
	 * @param g        the graphic context
	 * @param polygon  the polygon to draw
	 * @param inViewer indicates whether polygon is drawn in the viewer or not
	 * @param mode     the drawing mode ({@link LyMapElement#NORMAL_MODE} or {@link
	 *                 LyMapElement#REDUCED_MODE})
	 */
	protected void drawPolygon(LyAbstractMapComponentManager map, Graphics g, LyPolygon polygon,
		boolean inViewer, int mode)
	{
		LyLayoutManager layout = map.getLayoutManager();
		boolean		    isTreemapLayout = (layout instanceof LyTreemapLayout);
		boolean		    isNormalMode = (mode == LyMapElement.NORMAL_MODE);

		// Optimization for treemap layout : polygons are rectangles
		if (isTreemapLayout)
		{
			LyPoint   point1 = polygon.getPoint(0);
			LyPoint   point3 = polygon.getPoint(2);
			Rectangle r = new Rectangle((int)point1.getX(), (int)point1.getY(),
					(int)point3.getX()
					- (int)point1.getX() + 1, (int)point3.getY() - (int)point1.getY() + 1);

			if (isNormalMode)
			{
				String color = polygon.getColor();
				Color  graphicalColor = null;

				if ((color != null) && (color.length() != 0))
					graphicalColor = getColor(color);
				else
					graphicalColor = getColor("LY_DEFAULT_MAP_REGION_COLOR");

				g.setColor(graphicalColor);
				g.fillRect(r.x, r.y, r.width, r.height);
			}

			_drawedPolygons.addElement(r);

			if (_mapRegion.isSelected() || _mouseOver)
			{
				g.setColor(getColor("LY_MAP_SELECTED_COLOR"));

				// If select border size is greater than one, draw others polygons
				int		  j, borderWidth = ((LyTreemapLayout)layout).getSelectBorderSize();
				Rectangle bounds = getBoundingRectangle(map);

				borderWidth = Math.max(borderWidth, 1);
				borderWidth = Math.min(borderWidth, bounds.width);
				borderWidth = Math.min(borderWidth, bounds.height);

				for (j = 0; j < borderWidth; j++)
					g.drawRect(r.x + j, r.y + j, r.width - (2 * j) - 1, r.height - (2 * j) - 1);
			}

			return;
		}

		LyPoint point3D;
		Point   point, prevPoint = null;
		Polygon poly = new Polygon();
		int     j, nPoints = polygon.getPointCount();
		int     pitch = 1;

		if (!isNormalMode)
			pitch = Math.max((nPoints / 100), 1);

		for (j = 0; j < nPoints; j += pitch)
		{
			point3D = polygon.getPoint(j);
			point = map.convertForDrawing(point3D);

			if (inViewer)
				point = map.convertToViewer(point);

			// Optimization : don't draw the point if same as previous
			if ((prevPoint == null) || (prevPoint.x != point.x) || (prevPoint.y != point.y))
				poly.addPoint(point.x, point.y);

			prevPoint = point;
		}

		// Draw a shadow of the region if background style and not in viewer
		if (isNormalMode)
		{
			if (!inViewer && _mapRegion.isBackground())
			{
				poly.translate(SHADOW_OFFSET, SHADOW_OFFSET);
				g.setColor(getColor("LY_MAP_OUTLINE_COLOR"));
				g.fillPolygon(poly);
				poly.translate(-SHADOW_OFFSET, -SHADOW_OFFSET);
			}

			// Retrieve the polygon color
			String color = polygon.getColor();
			Color  graphicalColor = null;

			if ((color != null) && (color.length() != 0))
				graphicalColor = getColor(color);
			else
				graphicalColor = getColor("LY_DEFAULT_MAP_REGION_COLOR");

			g.setColor(graphicalColor);
			g.fillPolygon(poly);
		}

		if (!inViewer)
			_drawedPolygons.addElement(poly);

		if ((!_mapRegion.isBackground()
					&& !inViewer)
				|| (!isNormalMode)
				|| (nPoints <= 2))
		{
			if (_mapRegion.isSelected() || _mouseOver)
			{
				if (map.isActivated())
					g.setColor(getColor("LY_MAP_SELECTED_COLOR"));
				else
					g.setColor(getColor("LY_INACTIVE_SELECT_COLOR"));
			}
			else
				g.setColor(getColor("LY_MAP_OUTLINE_COLOR"));

			g.drawPolygon(poly);
		}
	} // end method drawPolygon

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#drawInViewer(leon.view.win.swing.map.LyAbstractMapComponentManager,
	 *      java.awt.Graphics)
	 */
	public void drawInViewer(LyAbstractMapComponentManager map, Graphics g)
	{
		if (isHidden())
			return;

		int		  i, n = _mapRegion.getPolygonCount();
		LyPolygon polygon;

		for (i = 0; i < n; i++)
		{
			polygon = _mapRegion.getPolygon(i);

			drawPolygon(map, g, polygon, true, LyMapElement.NORMAL_MODE);
		}
	}

	/**
	 * Gets the bounding box of this map region. If the region has no point, the resulting bounding
	 * box occupies the total space available.
	 *
	 * @param  map the map manager
	 * @return the bounding box of this region
	 * @see    leon.view.win.swing.map.LyMapElement#getBoundingBox(leon.view.win.swing.map.LyAbstractMapComponentManager)
	 */
	public Ly3DBox getBoundingBox(LyAbstractMapComponentManager map)
	{
		Vector<LyPoint>  points = getPoints();
		LyPoint          point3D;
		int              i, n = points.size();

		// If there is no point, return an empty bouding box
		if (n == 0)
			return new Ly3DBox(0, 0, 0, 0, 0, 0);

		float xmin = Float.MAX_VALUE;
		float xmax = -Float.MAX_VALUE;
		float ymin = Float.MAX_VALUE;
		float ymax = -Float.MAX_VALUE;
		float zmin = Float.MAX_VALUE;
		float zmax = -Float.MAX_VALUE;

		for (i = 0; i < n; i++)
		{
			point3D = points.elementAt(i);

			if (point3D.getX() > xmax)
				xmax = point3D.getX();
			if (point3D.getX() < xmin)
				xmin = point3D.getX();
			if (point3D.getY() > ymax)
				ymax = point3D.getY();
			if (point3D.getY() < ymin)
				ymin = point3D.getY();
			if (point3D.getZ() > zmax)
				zmax = point3D.getZ();
			if (point3D.getZ() < zmin)
				zmin = point3D.getZ();
		}

		return new Ly3DBox(xmin, ymin, zmin, xmax, ymax, zmax);
	} // end method getBoundingBox

	/**
	 * Gets the bounding rectangle of this map region. If the region has no point, the resulting
	 * bounding rectangle occupies the total space available.
	 *
	 * @param  map the map manager
	 * @return the bounding rectangle of this region
	 * @see    leon.view.win.swing.map.LyMapElement#getBoundingRectangle(leon.view.win.swing.map.LyAbstractMapComponentManager)
	 */
	public Rectangle getBoundingRectangle(LyAbstractMapComponentManager map)
	{
		LyLayoutManager layout = map.getLayoutManager();
		boolean		    isTreemapLayout = (layout instanceof LyTreemapLayout);

		Vector<LyPoint>	points = getPoints();
		LyPoint		    point3D;
		Point		    point;
		int			    i, n = points.size();

		int			    xmin = Integer.MAX_VALUE;
		int			    xmax = Integer.MIN_VALUE;
		int			    ymin = Integer.MAX_VALUE;
		int			    ymax = Integer.MIN_VALUE;
		int			    x, y;

		for (i = 0; i < n; i++)
		{
			point3D = (LyPoint)points.elementAt(i);

			if (!isTreemapLayout)
			{
				point = map.convert(point3D);
				x = point.x;
				y = point.y;
			}
			else
			{
				x = (int)point3D.getX();
				y = (int)point3D.getY();
			}

			if (x > xmax)
				xmax = x;
			if (x < xmin)
				xmin = x;
			if (y > ymax)
				ymax = y;
			if (y < ymin)
				ymin = y;
		}

		return new Rectangle(xmin, ymin, (xmax - xmin), (ymax - ymin));
	} // end method getBoundingRectangle

	/**
	 * Gets the clipping area of this region.
	 *
	 * @param  map   the map manager
	 * @param  image the image to clip. This image is not used in this implementation.
	 * @return the bounding rectangle of the map manager.
	 * @see    #getBoundingRectangle(LyAbstractMapComponentManager)
	 * @see    LyMapElement#getClippingArea(LyAbstractMapComponentManager, Image)
	 */
	public Shape getClippingArea(LyAbstractMapComponentManager map, Image image)
	{
		return getBoundingRectangle(map);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#getLayer()
	 */
	public int getLayer()
	{
		if (_mapRegion.getLayer() > 0)
			return _mapRegion.getLayer();
		
		if (_mapRegion.isBackground())
			return LyAbstractMapComponentManager.REGION_BACKGROUND_LAYER;
		return LyAbstractMapComponentManager.REGION_LAYER;
	}

	/**
	 * Gets the label of this map region. Use a cache to avoid many calls to the translate method.
	 *
	 * @see leon.view.win.swing.map.LyMapElement#getLabel()
	 */
	public String getLabel()
	{
		String label = _mapRegion.getLabel();

		if ((_labelCache == null) || !_labelCache.equals(label))
		{
			LyEnvironment env = getEnvironment();

			_labelCache = label;
			_translatedLabelCache = env.translate(label);
		}

		return _translatedLabelCache;
	}

	/**
	 * Gets the tooltip of this map region. If the mouse is over this map region, the returned value
	 * is <code>null</code> and it is managed in the draw method.
	 *
	 * @see leon.view.win.swing.map.LyMapElement#getTooltip()
	 */
	public String getTooltip()
	{
		// If mouse over mode, do not display tooltip : managed in draw method
		if (_mouseOver)
			return null;

		return _mapRegion.getTooltip();
	}

	/**
	 * Indicates whether the moue is over this region or not.
	 *
	 * @return true if the mouse is over this region
	 */
	public boolean isMouseOver()
	{
		return _mouseOver;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#isAtPosition(leon.view.win.swing.map.LyAbstractMapComponentManager,
	 *      java.awt.Graphics, int, int)
	 */
	public boolean isAtPosition(LyAbstractMapComponentManager map, Graphics g, int x, int y)
	{
		if (isHidden())
			return false;
		
		if (_drawedPolygons == null)
			return false;

		Rectangle	scrollRectangle = map.getScrollRectangle();
		int			i, n = _drawedPolygons.size();
		Shape		polygon;

		x -= scrollRectangle.x;
		y -= scrollRectangle.y;
		
		for (i = 0; i < n; i++)
		{
			polygon = _drawedPolygons.elementAt(i);

			// If polygon is a square, right and bottom borders must be taken into account
			if (polygon.contains(x, y) || polygon.contains(x - 1, y) || polygon.contains(x, y - 1)
					|| polygon.contains(x - 1, y - 1))
				return true;
		}

		return false;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.getinterface.LyGetIdInterface#getId()
	 */
	public String getId()
	{
		return _mapRegion.getId();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetSelectedInterface#setSelected(boolean)
	 */
	public void setSelected(boolean selected)
	{
		if (!_mapRegion.isBackground())
			_mapRegion.setSelected(selected);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#isSelected()
	 */
	public boolean isSelected()
	{
		return _mapRegion.isSelected();
	}

	/**
	 * A map region doesn't handle tooltips for its small images. This method don't do anything with
	 * the given tooltip text.
	 *
	 * @param tooltip Not used in this implementation.
	 * @see   leon.view.graphic.LyGraphicalMapRegion#setSmallImageTooltip(java.lang.String)
	 */
	public void setSmallImageTooltip(String tooltip)
	{
	}

	/**
	 * A map region doesn't handle weight. This method don't do anything with the given weight
	 * value.
	 *
	 * @param weight Not used in this implementation.
	 * @see   LyGraphicalMapRegion#setWeight(float)
	 */
	public void setWeight(float weight)
	{
	}

	/**
	 * A map region doesn't handle small images. This method don't do anything with the given small
	 * image name.
	 *
	 * @param smallImage Not used in this implementation.
	 * @see   LyGraphicalMapRegion#setSmallImage(String)
	 */
	public void setSmallImage(String smallImage)
	{
	}

	/**
	 * A map region doesn't handle translations. This method don't do anything with the given
	 * displacement values.
	 *
	 * @param dx Not used in this implementation.
	 * @param dy Not used in this implementation.
	 * @param dz Not used in this implementation.
	 * @see   LyMapElement#translate(float, float, float)
	 */
	public void translate(float dx, float dy, float dz)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#isHidden()
	 */
	public boolean isHidden()
	{
		return _mapRegion.isHidden();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetHiddenInterface#setHidden(boolean)
	 */
	public void setHidden(boolean hidden)
	{
		if (_mapRegion.isHidden() != hidden)
			_mapRegion.setHidden(hidden);
	}

	/**
	 * This map region has the overlay property if it is selected.
	 *
	 * @return true if it is overlay, false otherwise
	 * @see    leon.view.win.swing.map.LyMapElement#isOverlay()
	 */
	public boolean isOverlay()
	{
        return isSelected();
	}

	/**
	 * Don't do anything in this implementation.
	 *
	 * @param mapRegion a map region, not used
	 * @param pos       the position of the added map region, not used
	 * @see   LyGraphicalMapRegion#addMapRegion(LyMapRegion, int)
	 */
	public void addMapRegion(LyMapRegion mapRegion, int pos)
	{
	}

	/**
	 * Don't do anything in this implementation.
	 *
	 * @param mapRegion a map region, not used
	 * @see   LyGraphicalMapRegion#removeMapRegion(LyMapRegion)
	 */
	public void removeMapRegion(LyMapRegion mapRegion)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#isSensitive()
	 */
	public boolean isSensitive()
	{
		return (_mapRegion.isSensitive() && (!_mapRegion.isBackground()));
	}

	/**
	 * Sets the position of the label frame.
	 *
	 * @param rect a rectangle that define the borders of the label frame
	 */
	public void setLabelFrame(Rectangle rect)
	{
		_labelFrame = rect;
	}

	/**
	 * Gets the points of all the polygon elements in the map.
	 *
	 * @return the points of all the elements of the map
	 */
	public Vector<LyPoint> getPoints()
	{
		Vector<LyPoint> points = new Vector<LyPoint>();
		int			    i, n = _mapRegion.getPolygonCount();
		LyPolygon	    polygon;
		int			    j, nPoints;

		for (i = 0; i < n; i++)
		{
			polygon = _mapRegion.getPolygon(i);
			nPoints = polygon.getPointCount();

			for (j = 0; j < nPoints; j++)
				points.addElement(polygon.getPoint(j));
		}

		return points;
	}

} // end class LySwingMapRegion
