/*
 * LySwingMenu.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import javax.swing.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.graphic.*;


/**
 * This class implements the Menu object (as described in the dtd) for swing viewer.
 *
 * <p>A menu contains a set of menu, sub-menus entries and separators. The menus are available only
 * with the heavy clients (Swing and SWT) viewers.</p>
 *
 * @version $Id: LySwingMenu.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     LySwingMenuBar
 * @see     LyMenu
 */
public class LySwingMenu extends LySwingObject
	implements LyGraphicalMenu,
		LyActionListener,
		LyEventTransmitter
{
	public static final String    __VERSION =
		"$Id: LySwingMenu.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The view independent menu description. */
	protected LyMenu			  _menu;

	/** The swing component for this menu. */
	protected Object			  _menuComp;

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter _actionTransmitter;

	/**
	 * Class constructor of a Menu object for the Swing target.
	 *
	 * @param menu a MENU component defined for every type of view targets. This menu is initialized
	 *             for the swing view and its attributes are configured with the default values.
	 */
	public LySwingMenu(LyMenu menu)
	{
		super(menu);
		_menu = menu;

		initialize();
		setContent();
	}

	/**
	 * Initializes the current menu.
	 */
	public void initialize()
	{
		_menuComp = createMenu();
	}

	/**
	 * Sets the menu items of this menu.
	 */
	public void setContent()
	{
		String		  label = translate(_menu.getLabel());
		LyEnvironment environment = getEnvironment();
		Character     mnemonic = environment.getMnemonic(_menu.getLabel());

		setLabel(label);
		if (mnemonic != null)
			setMnemonic(mnemonic.charValue());
		setEnabled(_menu.isEnabled());

		// ** add menu items **/
		int           n = _menu.getComponentCount();
		LyComponent   comp;

		for (int i = 0; i < n; i++)
		{
			comp = _menu.getComponent(i);

			if ((comp == null)
			        || LyComponent.SEP_ID.equals(comp.getId()))
				addSeparator();
			else if (comp instanceof LyItem)
			{
				LyItem	    item = (LyItem)comp;
				LySwingItem internal = (LySwingItem)item.getInternalView();

				// register events
				if (internal != null)
				{
					addItem(internal.getItem());

					LyActionTransmitter transmitter = internal.getActionTransmitter();

					transmitter.addActionListener(this);
				}
			}
			else if (comp instanceof LyMenu)
			{
				LyMenu	    menu = (LyMenu)comp;
				LySwingMenu internal = (LySwingMenu)menu.getInternalView();

				// register events
				if (internal != null)
				{
					addMenu(internal.getMenu());

					LyActionTransmitter transmitter = internal.getActionTransmitter();

					transmitter.addActionListener(this);
				}
			}
		}
	} // end method setContent

	/**
	 * Creates and configures the component that display this menu.
	 *
	 * @return the new menu component
	 */
	public Object createMenu()
	{
		JMenu menuComp = new JMenu();
		Font  font = getFont("LY_MENU_FONT");

		menuComp.enableInputMethods(false);
		menuComp.setFont(font);

		return menuComp;
	}

	/**
	 * Gets the menu component.
	 *
	 * @return the component for the menu
	 */
	public Object getMenu()
	{
		return _menuComp;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		// Notify controller if any
		if (_menu.getController() != null)
		{
			// Notify controller
			LyActionDemand demand = new LyActionDemand(LyActionDemand.TYPE_ACTION,
					actionEvent.getSourceId());

			processDemand(demand);
			actionEvent.consume();
		}

		if (_actionTransmitter != null)
			_actionTransmitter.processActionEvent(actionEvent);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return (_actionTransmitter);
	}

	/**
	 * This menu doesn't handle selection events, so it has no selection transmitter.
	 *
	 * @return <code>null</code>
	 * @see    leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		return null;
	}

	/**
	 * Adds a new menu item at the specified position.
	 *
	 * @param component a menu item to add to this menu. <code>null</code> adds a separator.
	 * @param pos       the position of this menu item in the menu
	 * @see   leon.view.graphic.LyGraphicalMenu#addComponent(leon.view.LyComponent, int)
	 */
	public void addComponent(LyComponent component, int pos)
	{
		if (component == null)
		{
			if (pos < 0)
				addSeparator();
			else
				insertSeparator(pos);
		}
		else if (component instanceof LyItem)
		{
			LySwingItem internal = (LySwingItem)component.getInternalView();
			if (internal != null)
			{
				if (pos < 0)
					addItem(internal.getItem());
				else
					insertItem(internal.getItem(), pos);
				
				// Register events
				LyActionTransmitter transmitter = internal.getActionTransmitter();

				transmitter.addActionListener(this);
			}
		}
		else if (component instanceof LyMenu)
		{
			LySwingMenu internal = (LySwingMenu)component.getInternalView();
			if (internal != null)
			{

			if (pos < 0)
				addMenu(internal.getMenu());
			else
				insertMenu(internal.getMenu(), pos);
			
			// Register events
			LyActionTransmitter transmitter = internal.getActionTransmitter();

			transmitter.addActionListener(this);
			}
		}
	} // end method addComponent

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalMenu#removeComponent(leon.view.LyComponent)
	 */
	public void removeComponent(LyComponent component)
	{
		if (component == null)
			return;
		else if (component instanceof LyItem)
		{
			LySwingItem internal = (LySwingItem)component.getInternalView();
			if ((internal != null) && (internal.getItem() != null))
				removeItem(internal.getItem());
		}
		else if (component instanceof LyMenu)
		{
			LySwingMenu internal = (LySwingMenu)component.getInternalView();
			if ((internal != null) && (internal.getMenu() != null))
				removeMenu(internal.getMenu());
		}
	}

	/**
	 * Enables or disables this menu.
	 *
	 * @param enabled Indicates if the menu is enabled or not
	 * @see   LyGraphicalMenu#setEnabled(boolean)
	 * @see   JMenuItem#setEnabled(boolean)
	 */
	public void setEnabled(boolean enabled)
	{
		((JMenu)_menuComp).setEnabled(enabled);
	}

	/**
	 * Sets the menu label of this menu.
	 *
	 * @param label the new value of the menu label
	 * @see   AbstractButton#setText(String)
	 */
	public void setLabel(String label)
	{
		((JMenu)_menuComp).setText(label);
	}

	/**
	 * Sets the mnemonic of this menu.
	 *
	 * @param mnemonic the new value of the mnemonic
	 * @see   AbstractButton#setMnemonic(char)
	 */
	public void setMnemonic(char mnemonic)
	{
		((JMenu)_menuComp).setMnemonic(mnemonic);
	}

	/**
	 * Adds an item at the end of this menu.
	 *
	 * @param item the menu item to add to this menu
	 * @see   JMenu#add(JMenuItem)
	 */
	public void addItem(Object item)
	{
		((JMenu)_menuComp).add((JMenuItem)item);
	}

	/**
	 * Adds an item at the specified position in this menu.
	 *
	 * @param item the menu item to add to this menu
	 * @param pos  the position where the sub-component is inserted
	 * @see   JMenu#insert(javax.swing.JMenuItem, int)
	 */
	public void insertItem(Object item, int pos)
	{
		((JMenu)_menuComp).insert((JMenuItem)item, pos);
	}

	/**
	 * Adds a sub-menu at the end of this menu.
	 *
	 * @param menu the sub-menu to add in this menu
	 * @see   JMenu#add(JMenuItem)
	 */
	public void addMenu(Object menu)
	{
		((JMenu)_menuComp).add((JMenu)menu);
	}

	/**
	 * Adds a sub-menu at the specified position in this menu.
	 *
	 * @param menu the sub-menu to add in this menu
	 * @param pos  the position where the sub-component is inserted
	 * @see   JMenu#insert(javax.swing.JMenuItem, int)
	 */
	public void insertMenu(Object menu, int pos)
	{
		((JMenu)_menuComp).insert((JMenu)menu, pos);
	}

	/**
	 * Adds a separator at the end of this menu.
	 *
	 * @see JMenu#addSeparator()
	 */
	public void addSeparator()
	{
		((JMenu)_menuComp).addSeparator();
	}

	/**
	 * Inserts a separator at the specified position in this menu.
	 *
	 * @param pos the position where the separator will be inserted
	 * @see   JMenu#insertSeparator(int)
	 */
	public void insertSeparator(int pos)
	{
		((JMenu)_menuComp).insertSeparator(pos);
	}

	/**
	 * Removes an item from this menu.
	 *
	 * @param item the menu item to remove
	 * @see   JMenu#remove(javax.swing.JMenuItem)
	 */
	public void removeItem(Object item)
	{
		((JMenu)_menuComp).remove((JMenuItem)item);
	}

	/**
	 * Removes a menu from this menu.
	 *
	 * @param menu the menu to remove
	 * @see   JMenu#remove(javax.swing.JMenuItem)
	 */
	public void removeMenu(Object menu)
	{
		((JMenu)_menuComp).remove((JMenu)menu);
	}

} // end class LySwingMenu
