/*
 * LySwingMenuBar.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import javax.swing.*;
import javax.swing.border.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;


/**
 * This class implements the MenuBar object (as described in the dtd) for swing viewer.
 *
 * <p>A menu bar allows to display a set of menus. It can be displayed in a main window of a
 * Leonardi application. The menu bars are available only with the heavy client (Swing and SWT)
 * viewers.</p>
 *
 * @version $Id: LySwingMenuBar.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     LyMenuBar
 */
public class LySwingMenuBar extends LySwingObject
	implements LyGraphicalMenuBar,
		LyActionListener,
		LyEventTransmitter,
		LySetIdleInterface,
		LyIsIdleInterface
{
	public static final String    __VERSION =
		"$Id: LySwingMenuBar.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The view independent menu bar description. */
	protected LyMenuBar			  _menuBar;

	/** The menu bar swing component. */
	protected Object			  _menuBarComp;

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter _actionTransmitter;

	/** Idle mode. */
	protected boolean			  _idle;

	/**
	 * Class constructor of a Menu bar for the Swing target.
	 *
	 * @param menuBar a MENU_BAR component defined for every type of view targets. This menu bar is
	 *                initialized for the swing view.
	 */
	public LySwingMenuBar(LyMenuBar menuBar)
	{
		super(menuBar);
		_menuBar = menuBar;

		initialize();
		setContent();
	}

	/**
	 * Initializes the menuBar.
	 */
	public void initialize()
	{
		_menuBarComp = createMenuBar();
		_idle = false;
	}

	/**
	 * Adds the menu items to this menu bar and registers their action listeners.
	 */
	public void setContent()
	{
		int i, n = _menuBar.getMenuCount();

		for (i = 0; i < n; i++)
		{
			LyMenu	    menu = _menuBar.getMenu(i);
			LySwingMenu internal = (LySwingMenu)menu.getInternalView();

			// register events
			if (internal != null)
			{
				addMenu(internal.getMenu());

				LyActionTransmitter transmitter = ((LyEventTransmitter)internal)
						.getActionTransmitter();

				transmitter.addActionListener(this);
			}
		}
	}

	/**
	 * Creates the swing component used to display this menu bar and initializes its font and
	 * border.
	 *
	 * @return the new menu bar component
	 */
	public Object createMenuBar()
	{
		JMenuBar bar = new JMenuBar();
		Font     font = getFont("LY_MENU_FONT");

		bar.setBorder(new MenuBarBorder());
		bar.setFont(font);

		return bar;
	}

	/**
	 * Adds the specified menu at the end of this menu bar.
	 *
	 * @param menu the menu to add
	 * @param pos  this position is not used in this implementation
	 * @see   #addMenu(Object)
	 * @see   leon.view.graphic.LyGraphicalMenuBar#addMenu(leon.view.LyMenu, int)
	 */
	public void addMenu(LyMenu menu, int pos)
	{
		LySwingMenu internal = (LySwingMenu)menu.getInternalView();

		addMenu(internal.getMenu());
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalMenuBar#removeComponent(leon.view.LyComponent)
	 */
	public void removeMenu(LyMenu menu)
	{
		LySwingMenu internal = (LySwingMenu)menu.getInternalView();

		removeMenu(internal.getMenu());
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		if (_actionTransmitter != null)
			_actionTransmitter.processActionEvent(actionEvent);
	}

	/**
	 * Gets the menu bar component.
	 *
	 * @return the component for the menu bar
	 */
	public Object getMenuBar()
	{
		return _menuBarComp;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return (_actionTransmitter);
	}

	/**
	 * This menu bar doesn't handle selection events, so it has no selection transmitter.
	 *
	 * @return <code>null</code>
	 * @see    leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		return null;
	}

	/**
	 * Indictates whether this menu bar is idle or not. The idled menu bar does not dynamically
	 * refresh its menu contents in views.
	 *
	 * @return idle status
	 * @see    #setIdle(boolean)
	 * @see    LyIsIdleInterface#isIdle()
	 */
	public boolean isIdle()
	{
		return _idle;
	}

	/**
	 * Sets the idle status of this menu bar. If the menu bar is not idle, the view is refreshed.
	 *
	 * @param idle the new idle status
	 * @see   #isIdle()
	 * @see   LySetIdleInterface#setIdle(boolean)
	 */
	public void setIdle(boolean idle)
	{
		if (_idle == idle)
			return;

		_idle = idle;

		if (!_idle)
			// reset content
			resetContent();
	}

	/**
	 * Adds a menu in this menu bar.
	 *
	 * @param menu the menu to add in this menu bar
	 * @see   JMenuBar#add(javax.swing.JMenu)
	 */
	public void addMenu(Object menu)
	{
		synchronized (_menuBarComp)
		{
			((JMenuBar)_menuBarComp).add((JMenu)menu);
		}
	}

	/**
	 * Removes the specified menu from the menu bar.
	 *
	 * @param menu the menu to remove
	 * @see   java.awt.Container#remove(java.awt.Component)
	 */
	public void removeMenu(Object menu)
	{
		synchronized (_menuBarComp)
		{
			((JMenuBar)_menuBarComp).remove((JMenu)menu);
		}
	}

	/**
	 * .
	 */

	/**
	 * This class defines MS-Windows style menu bar borders.
	 */
	public class MenuBarBorder extends EmptyBorder
	{
		private static final long serialVersionUID = -2118931941137989122L;

		/**
		 * Creates a new MenuBarBorder object that mimics the MS-Wnidows style menu bar borders.
		 */
		public MenuBarBorder()
		{
			super(0, 0, 2, 0);
		}

		/*
		 * (non-Javadoc)
		 * @see javax.swing.border.Border#paintBorder(java.awt.Component, java.awt.Graphics, int,
		 * int, int, int)
		 */
		public void paintBorder(Component c, Graphics g, int x, int y, int width, int height)
		{
			Color background = c.getBackground();

			g.setColor(background.darker());
			g.drawLine(x, y + height - 2, x + width - 1, y + height - 2);
			g.setColor(background.brighter());
			g.drawLine(x, y + height - 1, x + width - 1, y + height - 1);
		}

	} // end class MenuBarBorder

	/**
	 * Resets the content of the menu bar and refreshes its view. This method is usually called
	 * after changing the menu bar idle status.
	 */
	public void resetContent()
	{
		((JMenuBar)_menuBarComp).removeAll();

		setContent();

		((JMenuBar)_menuBarComp).invalidate();
		((JMenuBar)_menuBarComp).validate();
		((JMenuBar)_menuBarComp).doLayout();
		((JMenuBar)_menuBarComp).repaint();
	}

} // end class LySwingMenuBar
