/*
 * LySwingMessage.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;

import javax.swing.*;

import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Message object (as described in the dtd) for swing viewer.
 *
 * <p>A message is placed inside a message bar. It allows to display a text and an image.</p>
 *
 * @version $Id: LySwingMessage.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     LySwingMessage
 */
public class LySwingMessage extends LySwingObject
	implements LySwingComponent,
		LyGraphicalMessage,
		LyActionListener
{
	public static final String __VERSION =
		"$Id: LySwingMessage.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The view independent message description. */
	protected LyMessage		   _message;

	/** The panel component that contain the message text and icon. */
	protected Container		   _borderPanel;

	/** The message icon. */
	protected Component		   _messageIcon;

	/**
	 * Class constructor of a Message object for the Swing target.
	 *
	 * @param message a MESSAGE component defined for every type of view targets. This message
	 *                component is initialized specificaly for the Swing view.
	 */
	public LySwingMessage(LyMessage message)
	{
		super(message);
		_message = message;

		setContent();
	}

	/**
	 * Creates a bordered panel used to contain this message.
	 *
	 * @param  border the type of the border. One of the following:
	 *
	 *                <ul>
	 *                  <li>{@link LyMessages#Border.NONE}</li>
	 *                  <li>{@link LyMessages#Border.SOLID}</li>
	 *                  <li>{@link LyMessages#Border.IN}</li>
	 *                  <li>{@link LyMessages#Border.OUT}</li>
	 *                  <li>{@link LyMessages#Border.RAISED}</li>
	 *                  <li>{@link LyMessages#Border.LOWERED}</li>
	 *                </ul>
	 * @return the new border panel
	 */
	public Container createBorderPanel(LyMessages.Border border)
	{
		LyJBorderPanel borderPanel = new LyJBorderPanel(LyArea.Border.NONE, 1);

		borderPanel.setGap(2);
		borderPanel.setOpaque(false);

		return borderPanel;
	}

	/**
	 * Creates the icon associated with this message.
	 *
	 * @param  env      the current environment
	 * @param  label    message of the component
	 * @param  image    image name of the icon
	 * @param  align    alignment of the icon
	 * @param  listener action listener to use on the component
	 * @return the message icon component
	 */
	public Component createMessageIcon(LyEnvironment env, String label, String image,
        LyMessage.Align align, LyActionListener listener)
	{
		Component comp = new LyJMessageIcon(env, label, image, align, listener);

		return comp;
	}

	/**
	 * Sets the content of this message component. Creates the message icon panel and adds it to
	 * this border panel.
	 */
	public void setContent()
	{
        Color color = getColor(_message.getColor());
        
	    if (_message.getId() == LyComponent.SEP_ID)
	    {
	        if (color == null)
	            color = Color.darkGray;
	        else
	            color = color.darker().darker();

	        _borderPanel = new JLabel();// new LyJSeparator(SwingConstants.VERTICAL,
	                                    // LyJSeparator.TOOLBAR_STYLE, color, 4);
	        return;
	    }
	    
		// Retrieve the border value and create a BorderPanel with the
		// specified border
		LyMessages messages = (LyMessages)_message.getParent();
		LyMessages.Border border = messages.getBorder();

		_borderPanel = createBorderPanel(border);
		_borderPanel.setLayout(new BorderLayout());

		// Create a new LyMessageIcon
		_messageIcon = createMessageIcon(getEnvironment(), _message.getLabel(), _message.getImage(),
				_message.getAlign(), this);

		_borderPanel.add(_messageIcon, "Center");
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _borderPanel;
	}

	/**
	 * Sets the text of the this message icon panel.
	 *
	 * @param label new label
	 * @see   leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
	 */
	public void setLabel(String label)
	{
		((LyAbstractMessageIconManager)((LyGetManagerInterface)_messageIcon).getManager()).setText(
			label);
	}

	/**
	 * Sets the new message icon.
	 *
	 * @param image the new image
	 * @see   LyGraphicalMessage#setImage(String)
	 */
	public void setImage(String image)
	{
		((LyAbstractMessageIconManager)((LyGetManagerInterface)_messageIcon).getManager()).setImage(
			image);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		// Notify controller if any
		if (_message.getController() != null)
		{
			// Notify controller
			LyActionDemand demand = new LyActionDemand(LyActionDemand.TYPE_ACTION,
					_message.getId());

			processDemand(demand);
			actionEvent.consume();
		}
	}

	/**
	 * Returns message component.
	 *
	 * @return the current message component
	 */
	public Component getMessage()
	{
		return _borderPanel;
	}

} // end class LySwingMessage
