/*
 * LySwingMessageWindow.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Message window (as described in the dtd) for swing viewer.
 *
 * @version $Id: LySwingMessageWindow.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     LyViewManager
 * @see     LyMessageWindowInterface
 */
public class LySwingMessageWindow
	implements ActionListener,
		LyMessageWindowInterface
{
	public static final String		   __VERSION =
		"$Id: LySwingMessageWindow.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The message displayed in the message box. */
	protected String				   _message;

	/** The title of the message box. */
	protected String				   _title;

	/** The icon of the message box. */
	protected String				   _icon;

	/** The modal state of the message box. */
	protected boolean				   _modal;

	/** The parent of the message box. */
	protected LyComponent			   _source;

	/** Client's data stored by the message box. */
	protected Object				   _clientData;

	/** The commands present in the message box. */
	protected BitSet				   _commands;

	/** The background color. */
	protected String				   _bgColor;

	/** The environment. */
	protected LyEnvironment			   _environment;

	/** The view manager. */
	protected LyViewManager			   _viewManager;

	/** The listener for this message window. */
	protected Vector<LyActionListener> _listeners;

	/** The dialog of this message window. */
	protected LyJMessageWindow		   _dialog;

	/**
	 * Class constructor for a new swing message window.
	 *
	 * @param source      the source component
	 * @param viewManager the view manager
	 */
	public LySwingMessageWindow(LyComponent source, LyViewManager viewManager)
	{
		_viewManager = viewManager;
		_environment = viewManager.getEnvironment();
		_source = source;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyMessageWindowInterface#initialize()
	 */
	public void initialize()
	{
		LyView parent;
		// from parent get frame or Dialog
		if (_source != null)
		{
			if (_source instanceof LyView)
				parent = (LyView)_source;
			else
				parent = _source.getParentView();
		}
		else
			parent = null;

		// If we have found the parent, try to get its frame in the
		// WinComponent
		Dialog parentDialog = null;
		JFrame parentFrame = null;

		// FFT00314 : if parent is a dialog, build the new dialog
		// with this dialog as owner.
		if (parent instanceof LyDialog)
		{
			LySwingComponent internal = (LySwingComponent)parent.getInternalView();

			// At the top, we have either a Frame or dialog ...
			if (internal.getComponent() instanceof Dialog)
				parentDialog = (Dialog)internal.getComponent();
		}

		if (parentDialog != null)
			_dialog = new LyJMessageWindow(parentDialog, _environment.translate(_title), _commands,
					_modal, (Color)_environment.getColor(_bgColor), _environment);
		else
		{
			parentFrame = ((LySwingViewManager)_viewManager).seekParentJFrame(_source);
			_dialog = new LyJMessageWindow(parentFrame, _environment.translate(_title), _commands,
					_modal, (Color)_environment.getColor(_bgColor), _environment);
		}

		_dialog.addActionListener(this);
		_dialog.setMessage(_environment.translate(_message));
		_dialog.setIcon(_icon);
		_dialog.setClientData(_clientData);
	} // end method initialize

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyMessageWindowInterface#free()
	 */
	public void free()
	{
		_dialog.free();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyMessageWindowInterface#addActionListener(leon.view.event.LyActionListener)
	 */
	public void addActionListener(LyActionListener listener)
	{
		if (_listeners == null)
			_listeners = new Vector<LyActionListener>();

		_listeners.addElement(listener);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyMessageWindowInterface#getCommandResult()
	 */
	public int getCommandResult()
	{
		return _dialog.getCommandResult();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyMessageWindowInterface#setVisible(boolean, boolean)
	 */
	public void setVisible(boolean status, boolean newThread)
	{
		_dialog.setVisible(status, newThread);

		if (!_modal)
			_dialog.toFront();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent actionEvent)
	{
		if (_listeners == null)
			return;

		LyActionEvent e = new LyActionEvent(this, null, actionEvent.getActionCommand(), 0);

		for (int i = 0; i < _listeners.size(); i++)
			_listeners.elementAt(i).actionPerformed(e);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyMessageWindowInterface#getClientData()
	 */
	public Object getClientData()
	{
		return _clientData;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyMessageWindowInterface#setClientData(java.lang.Object)
	 */
	public void setClientData(Object clientData)
	{
		_clientData = clientData;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyMessageWindowInterface#setIcon(java.lang.String)
	 */
	public void setIcon(String icon)
	{
		_icon = icon;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyMessageWindowInterface#setMessage(java.lang.String)
	 */
	public void setMessage(String message)
	{
		_message = message;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyMessageWindowInterface#setTitle(java.lang.String)
	 */
	public void setTitle(String title)
	{
		_title = title;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyMessageWindowInterface#setCommands(java.util.BitSet)
	 */
	public void setCommands(BitSet commands)
	{
		_commands = commands;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyMessageWindowInterface#setBackgroundColor(java.lang.String)
	 */
	public void setBackgroundColor(String bgColor)
	{
		_bgColor = bgColor;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyMessageWindowInterface#setModal(boolean)
	 */
	public void setModal(boolean modal)
	{
		_modal = modal;
	}

} // end class LySwingMessageWindow
