/*
 * LySwingMessages.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.util.*;

import javax.swing.*;

import leon.view.*;
import leon.view.graphic.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Messages object (as described in the dtd) for swing viewer.
 *
 * <p>A message bar can be used in a main window to provide the user with information about the
 * displayed data (number of objects, printing in progress...).</p>
 *
 * @version $Id: LySwingMessages.java 32334 2011-09-14 12:22:31Z jlsaouli $
 * @author  Lyria
 * @see     LyMessages
 */
public class LySwingMessages extends LySwingObject
	implements LySwingComponent,
		LyGraphicalMessages
{
	public static final String __VERSION =
		"$Id: LySwingMessages.java 32334 2011-09-14 12:22:31Z jlsaouli $";

	/** The view independent message bar description. */
	protected LyMessages	   _messages;

	/** The panel that contains the messages. */
	protected Container		   _panel;
	
	/** The list of separator components. */
	protected Hashtable<Component, Vector<Component>>  _separators;

	/**
	 * Class constructor of a Message bar object for the Swing target.
	 *
	 * @param messages a MESSAGE component defined for every type of view targets. This message bar
	 *                 component is initialized specificaly for the Swing view.
	 */
	public LySwingMessages(LyMessages messages)
	{
		super(messages);
		_messages = messages;

		initialize();
		setContent();
	}

	/**
	 * Creates the panel used to place all the messages.
	 *
	 * @return the panel for the messages
	 */
	public Container createPanel()
	{
		// return new JPanel(new LySplitterLayout(LySplitterLayout.HORIZONTAL));

		JPanel		  panel = new JPanel();
		GridBagLayout layout = new GridBagLayout();

		panel.setLayout(layout);
		return panel;
	}

	/**
	 * Initializes the graphic component and sets its background color.
	 */
	public void initialize()
	{
		_panel = createPanel();

		Color color = getColor("LY_MESSAGES_BACKGROUND");

		if (color == null)
			color = getColor("LY_DEFAULT_BACKGROUND");
		if (color == null)
			color = Color.lightGray;

		_panel.setBackground(color);
	}

	/**
	 * Fills the panel with its messages.
	 */
	public void setContent()
	{
		int i, n = _messages.getMessageCount();

		/* Add all the messages */
		for (i = 0; i < n; i++)
		{
			LyMessage		 message = _messages.getMessage(i);
			LySwingComponent internal = (LySwingComponent)message.getInternalView();
	        int              size = message.getSize();
	            
            if (message.getId() == LyComponent.SEP_ID)
                size = 0;
            else if (size <= 0)
                size = 1;
	            
			addMessage(internal.getComponent(), size);
		}
	}

	/**
	 * Removes a message from panel.
	 *
	 * @param message the message to remove
	 */
	public void removeMessage(Component message)
	{
		if ((_separators != null)
			&& (_separators.get(message) != null))
		{
			Vector<Component>	separators = _separators.remove(message);
			int					i, n = separators.size();
			
			for (i = 0; i < n; i++)
			{
				Component	separator = separators.elementAt(i);
				
				_panel.remove(separator);
			}
		}
		
		_panel.remove(message);

		if (_panel.isShowing() && (_panel.getComponentCount() == 0))
		{
			_panel.getParent().invalidate();
			_panel.getParent().validate();
		}
	}

	/**
	 * Adds a message to this message bar at its end.
	 *
	 * @param message the message to add
	 * @param pos     the position is not used
	 * @see   leon.view.graphic.LyGraphicalMessages#addMessage(leon.view.LyMessage, int)
	 */
	public void addMessage(LyMessage message, int pos)
	{
		LySwingMessage internal = (LySwingMessage)(message.getInternalView());
        int            size = message.getSize();
        
        if (message.getId() == LyComponent.SEP_ID)
            size = 0;
        else if (size <= 0)
            size = 1;

        addMessage(internal.getMessage(), size);

		if (_panel.isShowing())
		{
			_panel.invalidate();
			_panel.validate();
		}
	}

	/**
	 * Removes the specified message from this message bar.
	 *
	 * @param message the message to remove
	 * @see   leon.view.graphic.LyGraphicalMessages#removeMessage(leon.view.LyMessage)
	 */
	public void removeMessage(LyMessage message)
	{
		LySwingMessage internal = (LySwingMessage)(message.getInternalView());

		removeMessage(internal.getMessage());

		if (_panel.isShowing())
		{
			_panel.invalidate();
			_panel.validate();
		}
	}

	/**
	 * Adds a message at the end of the panel with the specified size.
	 *
	 * @param message the message to add
	 * @param size    the size of the message on the panel
	 */
	public void addMessage(Component message, int size)
	{
		if (size <= 0)
			size = 1;

		Color color = _panel.getBackground();

		if (color == null)
			color = Color.darkGray;
		else
			color = color.darker().darker();

		LyJSeparator		sep = new LyJSeparator(SwingConstants.VERTICAL, LyJSeparator.TOOLBAR_STYLE, color, 4);
		Vector<Component>	separators = new Vector<Component>(3);

		_panel.add(sep.getBeforeGap());
		separators.add(sep.getBeforeGap());

		if (_panel.getComponentCount() > 1)
		{
			_panel.add(sep);
			_panel.add(sep.getAfterGap());
			separators.add(sep);
			separators.add(sep.getAfterGap());
		}
		
		if (_separators == null)
			_separators = new Hashtable<Component, Vector<Component>>();
		
		_separators.put(message, separators);
		
		GridBagConstraints c = new GridBagConstraints();

		c.weightx = size;
		c.fill = GridBagConstraints.BOTH;
		((JComponent)message).setPreferredSize(new Dimension(0, message.getPreferredSize().height));
		((GridBagLayout)_panel.getLayout()).setConstraints(message, c);
		_panel.add(message);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _panel;
	}

} // end class LySwingMessages
