/*
 * LySwingNotice.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import javax.swing.*;
import javax.swing.border.*;
import leon.view.*;
import leon.view.graphic.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Notice object (as described in the dtd) for swing viewer.
 *
 * <p>An information panel allows to display an insert having an informative purpose. It is possible
 * to assign an icon, a label and a progress bar to an information panel.</p>
 *
 * @version $Id: LySwingNotice.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     LyForm
 */
public class LySwingNotice extends LySwingPanel
	implements LySwingComponent,
		LyGraphicalNotice
{
	public static final String __VERSION =
		"$Id: LySwingNotice.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Number of pixels of the border. */
	public int				   BORDER_SIZE = 15;

	/** The view independent information panel description. */
	protected LyNotice		   _notice;

	/** The bordered panel used to display this notice. */
	protected Container		   _internalPanel;

	/**
	 * Class constructor of an Information panel object for the Swing target.
	 *
	 * @param notice a NOTICE component defined for every type of view targets. This information
	 *               panel is initialized specificaly for the Swing view.
	 */
	public LySwingNotice(LyNotice notice)
	{
		super(notice);
		_notice = notice;

		initialize();
		setContent();
	}

	/**
	 * Builds the panel and sets its background color as well as the layout to use.
	 */
	public void initialize()
	{
		_internalPanel = createInternalPanel();
		_internalPanel.setLayout(new GridBagLayout());
	}

	/**
	 * Fills the panel with default properties.
	 */
	public void setContent()
	{
		// get possible content
		Component		 comp;
		LySwingComponent sComp;

		LyIcon			 icon = _notice.getIcon();
		LyLabel			 label = _notice.getLabel();
		LyGauge			 gauge = _notice.getGauge();

		// ** add everything **
		JPanel			 p = createContentPanel();

		GridBagLayout    layout = new GridBagLayout();

		p.setLayout(layout);
		GridBagConstraints c = new GridBagConstraints();

		p.setOpaque(false);

		// - icon to the left
		if (icon != null)
		{
			sComp = (LySwingComponent)icon.getInternalView();
			comp = sComp.getComponent();
			comp.setBackground(getColor("LY_FORM_BACKGROUND"));
			c.weightx = -1.0;
			c.fill = GridBagConstraints.NONE;
			c.anchor = GridBagConstraints.WEST;
			layout.setConstraints(comp, c);
			p.add(comp);
		}

		// - label in the middle-right
		if (label != null)
		{
			sComp = (LySwingComponent)label.getInternalView();
			comp = sComp.getComponent();

			checkMultiLineLabel(comp);

			comp.setBackground(getColor("LY_FORM_BACKGROUND"));
			c.weightx = 1.0;
			c.fill = GridBagConstraints.HORIZONTAL;
			c.gridwidth = GridBagConstraints.REMAINDER;
			c.anchor = GridBagConstraints.WEST;
			layout.setConstraints(comp, c);
			p.add(comp);
		}

		// - gauge at the bottom
		if (gauge != null)
		{
			sComp = (LySwingComponent)gauge.getInternalView();
			comp = sComp.getComponent();

			c.weightx = 1.0;
			c.gridwidth = GridBagConstraints.REMAINDER;
			c.fill = GridBagConstraints.HORIZONTAL;
			layout.setConstraints(comp, c);
			p.add(comp);
		}

		_internalPanel.add(p);
	} // end method setContent

	/**
	 * Creates and configures the internal panel. Initializes its display properties as well as the
	 * layout to use.
	 *
	 * @return the configured internal panel
	 */
	public Container createInternalPanel()
	{
		LyArea			   parentArea = _notice.getParentArea();
		Image			   image = (Image)getEnvironment().getImage("LY_AREA_BACKGROUND_IMAGE",
				null, false);
		
		// If notice is placed in an area, background image is managed by the area itself.
		if (parentArea != null)
			image = null;
		
		LyJBackgroundPanel internalPanel = new LyJBackgroundPanel(image)
			{
				private static final long serialVersionUID = -6096517233300872709L;

				/*
				 * (non-Javadoc)
				 * @see java.awt.Component#getPreferredSize()
				 */
				public Dimension getPreferredSize()
				{
					Dimension size = super.getPreferredSize();

					return new Dimension(size.width + 15, size.height + 10);
				}

			};

		internalPanel.setBackground(getColor("LY_FORM_BACKGROUND"));
		internalPanel.setBorder(new EmptyBorder(BORDER_SIZE, BORDER_SIZE, BORDER_SIZE,
				BORDER_SIZE));
		internalPanel.setOpaque((image == null) && (parentArea == null));

		return internalPanel;
	} // end method createInternalPanel

	/**
	 * Creates and configures the internal panel's container.
	 *
	 * @return an empty panel
	 */
	public JPanel createContentPanel()
	{
		return new JPanel();
	}

	/**
	 * Sets up correctly the label rows and columns if it's multilined.
	 *
	 * @param comp the label component
	 */
	public void checkMultiLineLabel(Component comp)
	{
		if (comp instanceof LyJMultiLineLabel)
		{
			((LyJMultiLineLabel)comp).setNRows(_notice.getNRows());
			((LyJMultiLineLabel)comp).setNCols(_notice.getNCols());
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return (_internalPanel);
	}

} // end class LySwingNotice
