/*
 * LySwingNumber.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.event.*;

import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Number object (as described in the dtd) for swing viewer.
 *
 * <p>An input field of Number type allows to display a graphical component allowing to input or to
 * consult a number value. It does not comprise any element. It is possible to specify the minimum
 * and the maximum values, the incrementation step, the number of decimals and the unit.</p>
 *
 * @version $Id: LySwingNumber.java 35079 2012-03-29 09:33:49Z nhamon $
 * @author  Lyria
 * @see     LyNumber
 */
public class LySwingNumber extends LySwingFieldItem
	implements LySwingComponent,
		LyActionListener,
		LySetEnabledInterface,
		LySetEditableInterface,
		LySetLabelInterface,
		LySetNumberInterface,
		LyGraphicalNumber,
		LyGetLabelInterface,
		KeyListener
{
	public static final String  __VERSION =
		"$Id: LySwingNumber.java 35079 2012-03-29 09:33:49Z nhamon $";

	/** The view independent number description. */
	protected LyNumber		    _number;

	/** The swing number selector component. */
	protected LyJNumberSelector _selector;
	
	/** Boolean indicating when a key is being released on the component. */
	protected boolean			_keyReleased;

	/**
	 * Class constructor of a Number object for the Swing target.
	 *
	 * @param number a NUMBER component defined for every type of view targets. This number
	 *               component is initialized specificaly for the Swing view.
	 */
	public LySwingNumber(LyNumber number)
	{
		super(number);
		_number = number;
		_keyReleased = false;
		setContent();
	}

	/**
	 * Sets the content of this component. Creates the {@link LyJNumberSelector} associated with
	 * this number description, sets its default values and registers the event listeners.
	 */
	public void setContent()
	{
		boolean initLabel = (_number.getLabel() != null);

		_selector = new LyJNumberSelector(LyTools.doubleFromString(_number.getLabel()),
				_number.getVMin(), _number.getVMax(), _number.getPitch(), _number.getDecimals(),
				_number.getUnit(), initLabel, getEnvironment(), _number.isShowArrows());

		if (initLabel)
			setLabel(_number.getLabel());

		LyActionTransmitter transmitter = _selector.getActionTransmitter();

		if (transmitter != null)
			transmitter.addActionListener(this);

		// This handler is used to validate the form using ENTER key or close using ESCAPE key
		LySwingFormKeyHandler keyHandler = new LySwingFormKeyHandler(getActionTransmitter());

		keyHandler.setNotifyValidate(false); // already default in swing
		_selector.addKeyListener(keyHandler);

		Color enabledColor = getColor("LY_ENABLED_BACKGROUND");
		Color disabledColor = getColor("LY_DISABLED_BACKGROUND");
		Font  font = getFont("LY_DEFAULT_FONT");

		_selector.setEnabledColor(enabledColor);
		_selector.setDisabledColor(disabledColor);
		_selector.setFont(font);
		_selector.setOpaque(false);
		
		_selector.getTextField().addKeyListener(this);
	} // end method setContent
	
	/*
	 * (non-Javadoc)
	 * @see java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
	 */
	public void keyTyped(KeyEvent e)
	{
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
	 */
	public void keyPressed(KeyEvent e)
	{
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
	 */
	public void keyReleased(KeyEvent e)
	{
		// Update Parent info
		_keyReleased = true;
		_number.setLabel(_selector.getTextField().getText());
		_keyReleased = false;
	}
	
	/**
	 * Checks the right number of decimals for the given label.
	 *
	 * @param  label the label to test
	 * @return the label with additional '0' if needed
	 */
	protected String checkDecimals(String label)
	{
		int decimals = _number.getDecimals();

		if ((decimals > 0) && (label != null))
		{
			int pos = label.indexOf('.');
			int decimalsNb = label.length() - pos - 1;

			if ((pos > 0) && (decimalsNb < decimals))
			{
				int i, n = decimals - decimalsNb;

				for (i = 0; i < n; i++)
					label = label + '0';
			}
		}

		return label;
	}

	/**
	 * Sets the editable status of the field item.
	 *
	 * @param editable the new editable status
	 * @see   LySetEditableInterface#setEditable(boolean)
	 */
	public void setEditable(boolean editable)
	{
		_selector.setEditable(editable);
	}

	/**
	 * Sets the enabled status of this number component.
	 *
	 * @parameter status the new enables status
	 * @see       LySetEnabledInterface#setEnabled(boolean)
	 */
	public void setEnabled(boolean status)
	{
		if (getComponent() != null)
			getComponent().setEnabled(status);
	}

	/**
	 * Sets the label of this component, with the correct number of decimals.
	 *
	 * @param label the new label
	 * @see   LySetLabelInterface#setLabel(String)
	 */
	public void setLabel(String label)
	{
		if ((_selector != null) && !_keyReleased)
		{
			// Keep previous idle status
			boolean wasIdle = _selector.isIdle();

			_selector.setIdle(true);
			label = checkDecimals(label);
			// Set value but do not notify
			_selector.setValue(label);
			_selector.setIdle(wasIdle);
		}
	}
	
	/*
	 * (non-Javadoc)
	 * @see leon.view.getinterface.LyGetLabelInterface#getLabel()
	 */
	public String getLabel()
	{
		return (_selector == null ? null : _selector.getValue());
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _selector;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{

		// Update Parent info
		if (actionEvent.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
			_number.setLabel(actionEvent.getValue());

		if (_actionTransmitter != null)
			_actionTransmitter.processActionEvent(actionEvent);
	}

	/* LySetNumberInterface implementation */

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetNumberInterface#setFormat(java.lang.String)
	 */
	public void setFormat(String format)
	{
		if (_selector != null)
			_selector.setFormat(getEnvironment(), format);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetNumberInterface#setVMin(double)
	 */
	public void setVMin(double vMin)
	{
		if (_selector != null)
			_selector.setMin(getEnvironment(), vMin);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetNumberInterface#setVMax(double)
	 */
	public void setVMax(double vMax)
	{
		if (_selector != null)
			_selector.setMax(getEnvironment(), vMax);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetNumberInterface#setPitch(double)
	 */
	public void setPitch(double pitch)
	{
		if (_selector != null)
			_selector.setPitch(getEnvironment(), pitch);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetNumberInterface#setDecimals(int)
	 */
	public void setDecimals(int decimals)
	{
		if (_selector != null)
			_selector.setDecimals(getEnvironment(), decimals);
	}

} // end class LySwingNumber
