/*
 * LySwingOption.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import javax.swing.*;
import leon.view.*;
import leon.view.graphic.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Option object (as described in the dtd) for swing viewer.
 *
 * <p>An option represents a possible value in an input field of Choice type. It can be represented
 * by a check box (case of a multiple choice) or else by a radio button (case of a simple choice).
 * </p>
 *
 * @version $Id: LySwingOption.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     LyOption
 */
public class LySwingOption extends LySwingObject
	implements LySwingComponent,
		LyGraphicalOption
{
	public static final String __VERSION =
		"$Id: LySwingOption.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The view independent number description. */
	protected LyOption		   _option;

	/**
	 * The swing option component. May be a check box or a radio button depending on the multiple
	 * selection of the Choice parent component.
	 */
	protected Component		   _checkbox;

	/** The choice parent component that contains this option. */
	LySwingChoice			   _parent;

	/**
	 * Class constructor of an Option object for the Swing target.
	 *
	 * @param option an OPTION component defined for every type of view targets. This number
	 *               component is initialized specificaly for the Swing view.
	 */
	public LySwingOption(LyOption option)
	{
		super(option);
		_parent = null;
		_option = option;

		setContent();
	}

	/**
	 * Sets the content of this component and enables it if necessary.
	 */
	public void setContent()
	{
		// !!! CheckBoxGroup is set by parent
		_checkbox = createCheckBox(translate(_option.getLabel()), _option.isChecked());
		setEnabled(_option.isEnabled());
	}

	/**
	 * Creates an option component. Depending on the parent multiple selection property, this option
	 * will be a check box or a radio button.
	 *
	 * @param  label     the label of this option
	 * @param  isChecked the checked status of this option
	 * @return the new option component with the specified label and status
	 */
	public Component createCheckBox(String label, boolean isChecked)
	{
		// FFT00999
		LyChoice  parent = (LyChoice)_option.getParent();

		Component check;

		if (parent.isMultiple())
			check = new JCheckBox(label, isChecked);
		else
			check = new JRadioButton(label, isChecked);

		check.setFont(getFont("LY_DEFAULT_FONT"));
		return check;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalOption#setChecked(boolean)
	 */
	public void setChecked(boolean status)
	{
		if (_parent == null)
			_parent = (LySwingChoice)_option.getParent().getInternalView();

		// FFT00999
		if ((_checkbox != null) && (((JToggleButton)_checkbox).isSelected() != status))
		{
			_parent.setChecked(_checkbox, status);
			updateParentTable();
		}
	}
	
	/**
	 * Graphically updates the parent table of this option if needed.
	 */
	protected void updateParentTable()
	{
		// If option is displayed in a table, table may need to be repaint.
		LyComponent	parent = _option.getParent();
		LyRow		parentRow = null;
		
		while ((parent != null) && !(parent instanceof LyTable))
		{
			if (parent instanceof LyRow)
				parentRow = (LyRow)parent;
			
			parent = parent.getParent();
		}
		
		if ((parent instanceof LyTable) && (parentRow != null))
		{
			LySwingTable		swingTable = (LySwingTable)parent.getInternalView();
			LyJTableComponent	tableComponent = swingTable.getTableComponent();
			
			tableComponent.repaint();
		}
	}

	/**
	 * Sets the parent of the option.
	 *
	 * @param parent the new parent value
	 */
	protected void setParent(LySwingChoice parent)
	{
		_parent = parent;
	}

	/**
	 * Sets the enabled status of this number component.
	 *
	 * @parameter status the new enables status
	 * @see       LyGraphicalOption#setEnabled(boolean)
	 */
	public void setEnabled(boolean status)
	{
		if (getComponent() != null)
			getComponent().setEnabled(status);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _checkbox;
	}

} // end class LySwingOption
