/*
 * LySwingPopupMenu.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import javax.swing.*;

import leon.view.*;
import leon.view.event.*;
import leon.view.graphic.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Popup Menu object (as described in the dtd) for swing viewer.
 *
 * <p>A context-sensitive menu is a menu which positioning and content depends on the context that
 * made possible to display it. This context depends, most of the time, of an element belonging to a
 * high-level graphical object (tree node, element of a map, line of a list,...). If need be, this
 * context is used as context for the launched action. The context-sensitive menus are available
 * only with the heavy clients viewers.</p>
 *
 * @version $Id: LySwingPopupMenu.java 33062 2011-11-04 08:36:25Z nhamon $
 * @author  Lyria
 * @see     LySwingMenuBar
 * @see     LyMenu
 */
public class LySwingPopupMenu extends LySwingObject
	implements LyGraphicalPopupMenu,
		LyActionListener,
		LyEventTransmitter
{
	public static final String    __VERSION =
		"$Id: LySwingPopupMenu.java 33062 2011-11-04 08:36:25Z nhamon $";

	/** The view independent popup menu description. */
	protected LyPopupMenu		  _popupMenu;

	/** The swing component of this popup menu. */
	protected Object			  _menuComp;

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter _actionTransmitter;

	/**
	 * Class constructor of a Popup menu object for the Swing target.
	 *
	 * @param popupMenu a POPUP_MENU component defined for every type of view targets. This popup
	 *                  menu component is initialized specificaly for the Swing view.
	 */
	public LySwingPopupMenu(LyPopupMenu popupMenu)
	{
		super(popupMenu);
		_popupMenu = popupMenu;

		initialize();
		setContent();
	}

	/**
	 * Initializes the popup menu component.
	 */
	public void initialize()
	{
		_menuComp = createPopupMenu();
	}

	/**
	 * Sets the popup menu content by adding the current menu items and the associated event
	 * listeners.
	 */
	public void setContent()
	{
		// ** add menu items **/
		int           i, n = _popupMenu.getComponentCount();
		LyComponent   comp;

		for (i = 0; i < n; i++)
		{
			comp = _popupMenu.getComponent(i);

			if (_popupMenu.getStyle() == LyPopupMenu.Style.LIGHT)
			{
		        if (comp instanceof LyItem && ((LyItem)comp).isEnabled())
				{
					LyItem	    item = (LyItem)comp;
					LySwingItem internal = (LySwingItem)item.getInternalView();

					// register events
					if (internal != null)
					{
						JMenuItem	menuItem = (JMenuItem)internal.getItem();
						
						menuItem.setToolTipText(menuItem.getText());
						menuItem.setText(null);
						addItem(menuItem);

						LyActionTransmitter transmitter = ((LyEventTransmitter)internal)
								.getActionTransmitter();

						transmitter.addActionListener(this);
					}
				}
			}
			
	        if ((comp == null)
                    || LyComponent.SEP_ID.equals(comp.getId()))
                addSeparator();
	        else if (comp instanceof LyItem)
			{
				LyItem	    item = (LyItem)comp;
				LySwingItem internal = (LySwingItem)item.getInternalView();

				// register events
				if (internal != null)
				{
					addItem(internal.getItem());

					LyActionTransmitter transmitter = ((LyEventTransmitter)internal)
							.getActionTransmitter();

					transmitter.addActionListener(this);
				}
			}
			else if (comp instanceof LyMenu)
			{
				LyMenu	    menu = (LyMenu)comp;
				LySwingMenu internal = (LySwingMenu)menu.getInternalView();

				// register events
				if (internal != null)
				{
					addMenu(internal.getMenu());

					LyActionTransmitter transmitter = ((LyEventTransmitter)internal)
							.getActionTransmitter();

					transmitter.addActionListener(this);
				}
			}
		}
	} // end method setContent

	/**
	 * Create a new popup menu.
	 *
	 * @return the new popup menu
	 */
	public Object createPopupMenu()
	{
		return new LyJPopupMenu(_popupMenu.getEnvironment(), _popupMenu.getStyle() == LyPopupMenu.Style.LIGHT);
	}

	/**
	 * Enables or disables this popup menu.
	 *
	 * @param enabled the enabled status
	 * @see   leon.view.setinterface.LySetEnabledInterface#setEnabled(boolean)
	 */
	public void setEnabled(boolean enabled)
	{
		((JPopupMenu)_menuComp).setEnabled(enabled);
	}

	/**
	 * Adds an item at the end of this menu.
	 *
	 * @param item the new item to add
	 */
	public void addItem(Object item)
	{
		((JPopupMenu)_menuComp).add((JMenuItem)item);
	}

	/**
	 * Adds a menu at the end of the menu.
	 *
	 * @param menu the new menu to add
	 */
	public void addMenu(Object menu)
	{
		((JPopupMenu)_menuComp).add((JMenuItem)menu);
	}

	/**
	 * Adds a separator at the end of the menu.
	 */
	public void addSeparator()
	{
		((JPopupMenu)_menuComp).addSeparator();
	}

	/**
	 * Inserts a separator at the specified position in this menu.
	 *
	 * @param pos the position where the separator will be inserted
	 * @see   JMenu#insertSeparator(int)
	 */
	public void insertSeparator(int pos)
	{
		((JPopupMenu)_menuComp).insert(new JPopupMenu.Separator() ,pos);
	}
	
	/**
	 * Gets the menu component.
	 *
	 * @return the popup menu component
	 */
	public Object getMenu()
	{
		return _menuComp;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		// Notify controller if any
		if (_popupMenu.getController() != null)
		{
			// Notify controller
			LyActionDemand demand = new LyActionDemand(LyActionDemand.TYPE_ACTION,
					actionEvent.getSourceId());

			processDemand(demand);
			actionEvent.consume();
		}

		if (_actionTransmitter != null)
			_actionTransmitter.processActionEvent(actionEvent);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return (_actionTransmitter);
	}

	/**
	 * This popup menu doesn't handle selection events, so it has no selection transmitter.
	 *
	 * @return <code>null</code>
	 * @see    leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		return null;
	}

	public void addComponent(LyComponent component, int pos)
	{
		if (component == null)
		{
			if (pos < 0)
				addSeparator();
			else
				insertSeparator(pos);
		}
		else if (component instanceof LyItem)
		{
			LySwingItem internal = (LySwingItem)component.getInternalView();

			if (pos < 0)
				addItem(internal.getItem());
			else
				insertItem(internal.getItem(), pos);
		}
		else if (component instanceof LyMenu)
		{
			LySwingMenu internal = (LySwingMenu)component.getInternalView();

			if (pos < 0)
				addMenu(internal.getMenu());
			else
				insertMenu(internal.getMenu(), pos);
		}
		
	}

	/**
	 * Adds a sub-menu at the specified position in this menu.
	 *
	 * @param menu the sub-menu to add in this menu
	 * @param pos  the position where the sub-component is inserted
	 * @see   JMenu#insert(javax.swing.JMenuItem, int)
	 */
	public void insertMenu(Object menu, int pos)
	{
		((JPopupMenu)_menuComp).insert((JMenu)menu, pos);
	}
	
	/**
	 * Adds an item at the specified position in this menu.
	 *
	 * @param item the menu item to add to this menu
	 * @param pos  the position where the sub-component is inserted
	 * @see   JMenu#insert(javax.swing.JMenuItem, int)
	 */
	public void insertItem(Object item, int pos)
	{
		((JPopupMenu)_menuComp).insert((JMenuItem)item, pos);
	}
	
	public void removeComponent(LyComponent component)
	{
		if (component == null)
			return;
		else if (component instanceof LyItem)
		{
			LySwingItem internal = (LySwingItem)component.getInternalView();

			removeItem(internal.getItem());
		}
		else if (component instanceof LyMenu)
		{
			LySwingMenu internal = (LySwingMenu)component.getInternalView();

			removeMenu(internal.getMenu());
		}
		
	}
	
	/**
	 * Removes an item from this menu.
	 *
	 * @param item the menu item to remove
	 * @see   JMenu#remove(javax.swing.JMenuItem)
	 */
	public void removeItem(Object item)
	{
		((JPopupMenu)_menuComp).remove((JMenuItem)item);
	}

	/**
	 * Removes a menu from this menu.
	 *
	 * @param menu the menu to remove
	 * @see   JMenu#remove(javax.swing.JMenuItem)
	 */
	public void removeMenu(Object menu)
	{
		((JPopupMenu)_menuComp).remove((JMenu)menu);
	}
} // end class LySwingPopupMenu
