/*
 * LyAwtPrintManager.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.PageAttributes.*;
import java.text.*;
import java.util.*;

import leon.data.*;
import leon.misc.*;
import leon.view.getinterface.*;
import leon.view.print.*;


/**
 * Interface for AWT and Swing printing.
 */
public class LySwingPrintManager extends LyPrintManager
{
	public static final String __VERSION =
		"$Id: LySwingPrintManager.java 28997 2010-12-23 10:17:15Z bmichel $";

	/** Background color. */
	protected static Color     BACKGROUND_COLOR = Color.white;

	/** Font color. */
	protected static Color     FONT_COLOR = Color.black;

	/** Line color. */
	protected static Color     LINE_COLOR = Color.gray;

	/** Table header color. */
	protected static Color     HEADER_COLOR = Color.lightGray;

	/** Sort arrows color. */
	protected static Color     SORT_COLOR = new Color(150, 150, 255);

	/** The print job. */
	protected PrintJob		   _job;

	/**
	 * Draws the page header with a logo to the left, title in the middle and logo to the right.
	 *
	 * @param  component   client component that handle images
	 * @param  g           printer handler
	 * @param  size        usable print size in the view
	 * @param  left_logo   the left logo image name
	 * @param  title       centered title
	 * @param  right_logo  the right logo image name
	 * @param  environment the current environment
	 * @return effective drawing size (with margins)
	 */
	public static int printHeader(Component component, Graphics g, Dimension size,
		String left_logo, String title, String right_logo, LyEnvironment environment)
	{
		if ((component == null) || (g == null) || (size == null))
			return 0;

		Image	    llogo = ((left_logo == null) ? null : (Image)environment.getImage(left_logo));
		Image	    rlogo = ((right_logo == null) ? null : (Image)environment.getImage(right_logo));
		Font	    font = g.getFont();
		FontMetrics fm = g.getFontMetrics(font);
		int		    llw = ((llogo == null) ? 0 : llogo.getWidth(component));
		int		    llh = ((llogo == null) ? 0 : llogo.getHeight(component));
		int		    rlw = ((rlogo == null) ? 0 : rlogo.getWidth(component));
		int		    rlh = ((rlogo == null) ? 0 : rlogo.getHeight(component));
		int		    w = size.width - LEFT_MARGIN - RIGHT_MARGIN;
		int		    th, sh;
		int		    hh;
		int		    y;
		String[]    hstr = null;

		if (title != null)
		{
			// estimate available width for the string and cut if it does not fit
			int    w_dispo = w - llw - rlw;
			int    max_w = fm.charWidth('X');
			int    nbchar = w_dispo / max_w;
			String format_title = LyString.format(title, nbchar, false, '\0', false);

			hstr = LyString.stringToList(format_title, '\n');
			sh = fm.getHeight() - 2;
			th = sh * hstr.length;
		}
		else
		{
			sh = 0;
			th = 0;
		}
		hh = Math.max(th, Math.max(llh, rlh));
		y = TOP_MARGIN;

		// Printing header
		if (llogo != null)
			g.drawImage(llogo, LEFT_MARGIN, y + ((hh - llh) / 2), component);

		if (rlogo != null)
			g.drawImage(rlogo, LEFT_MARGIN + w - rlw, y + ((hh - rlh) / 2), component);

		if (title != null)
		{
			int tw;
			int top = y + ((hh - th) / 2) + sh;

			for (int i = 0; i < hstr.length; i++)
			{
				tw = fm.stringWidth(hstr[i]);
				g.drawString(hstr[i], LEFT_MARGIN + ((w + llw - rlw - tw) / 2), top + (i * sh));
			}
			// put down title if it does not fit
			// if (tw > (w-llw-rlw-6))
			// hh += 4 + fm.getAscent();
		}
		g.drawLine(LEFT_MARGIN, y + hh + 5, LEFT_MARGIN + w, y + hh + 5);
		hh += y + 15;

		return hh;
	} // end method printHeader

	/**
	 * Draws the footer of the view.
	 *
	 * @param  component   client component
	 * @param  g           printer handler
	 * @param  size        usable print size in the view
	 * @param  left_str    the left message
	 * @param  middle_str  the center message
	 * @param  right_str   the right message
	 * @param  environment the current environment, not used in this implementation
	 * @return effective drawing size (with margins)
	 */
	public static int printFooter(Component component, Graphics g, Dimension size, String left_str,
		String middle_str, String right_str, LyEnvironment environment)
	{
		if ((component == null) || (g == null) || (size == null))
			return 0;

		Font	    font = g.getFont();
		FontMetrics fm = component.getFontMetrics(font);
		int		    th = (((left_str == null) && (right_str == null)) ? 0 : fm.getHeight());
		int		    y = size.height - BOTTOM_MARGIN - 2;
		int		    fh = BOTTOM_MARGIN;
		int		    w = size.width - LEFT_MARGIN - RIGHT_MARGIN;

		// Printing footer
		if (left_str != null)
			g.drawString(left_str, LEFT_MARGIN, y);
		if (middle_str != null)
			g.drawString(middle_str, LEFT_MARGIN + ((w - fm.stringWidth(middle_str)) / 2), y);
		if (right_str != null)
			g.drawString(right_str, LEFT_MARGIN + w - fm.stringWidth(right_str), y);
		fh += th;
		g.drawLine(LEFT_MARGIN, y - th - 5, LEFT_MARGIN + w, y - th - 5);
		fh += 15;

		return fh;
	}

	/**
	 * Gets the size of a String, in pixels, for the specified font. This string may have several
	 * lines separated with \n.
	 *
	 * @param  str the original String
	 * @param  fm  the font metrics used to compute the string size
	 * @return the width (in pixels) of the String
	 */
	public static Dimension getSize(String str, FontMetrics fm)
	{
		if ((str == null) || (fm == null))
			return null;

		String[] lines = LyString.stringToList(str, '\n');
		int		 l, nl = lines.length;
		int		 w = 0, h = nl * fm.getHeight();

		for (l = 0; l < nl; l++)
			w = Math.max(w, fm.stringWidth(lines[l]));

		return new Dimension(w, h);
	}

	/**
	 * Formats the specified string before drawing it to the output.
	 *
	 * @param  str   the string to print
	 * @param  fm    the font metrics of the current font
	 * @param  width the maximum width of the drawing
	 * @param  sep   character separator used to aggregate the result
	 * @return the formatted string with the specified separator
	 */
	public static String formatString(String str, FontMetrics fm, int width, char sep)
	{
		if ((str == null) || (fm == null) || (width <= 0))
			return null;

		int tw = fm.stringWidth(str);

		if (tw <= width)
			return str;

		String[]     lines = LyString.stringToList(str, '\n');
		int			 l, nl = lines.length;
		String[]     words;
		int			 w, nw;
		StringBuffer buffer = new StringBuffer();
		int			 cw = 0;
		int			 c, nc;
		char		 car;

		for (l = 0; l < nl; l++)
		{
			if ((lines[l] == null) || (lines[l].length() == 0))
			{
				buffer.append(sep);
				cw = 0;
				continue;
			}

			lines[l] = lines[l] + ' ';
			tw = fm.stringWidth(lines[l]);

			if ((cw + tw) <= width)
			{
				// Write the line
				buffer.append(lines[l]);
				buffer.append(sep);
				cw = 0;
				continue;
			}

			words = LyString.stringToList(lines[l], ' ');
			nw = words.length;

			for (w = 0; w < nw; w++)
			{
				words[w] = words[w] + ' ';
				tw = fm.stringWidth(words[w]);
				if ((cw > 0) && ((cw + tw) > width))
				{
					buffer.append(sep);
					cw = 0;
				}
				if (tw > width)
				{
					// write char per char if the word is longer than width
					nc = words[w].length();
					for (c = 0; c < nc; c++)
					{
						car = words[w].charAt(c);
						tw = fm.charWidth(car);
						if ((cw + tw) > width)
						{
							buffer.append(sep);
							cw = 0;
						}
						buffer.append(car);
						cw += tw;
					}
				}
				else
				{
					// write current word
					buffer.append(words[w]);
					cw += tw;
				}
			}
		}

		return buffer.toString();
	} // end method formatString

	/**
	 * Class constructor of a print manager object.
	 *
	 * @param environment        the current environment
	 * @param printDataInterface the interface to use for printing
	 * @param properties         the properties of the printing manager
	 */
	public LySwingPrintManager(LyEnvironment environment, LyPrintDataInterface printDataInterface,
		Properties properties)
	{
		super(environment, printDataInterface, properties);

		Toolkit		   toolkit = Toolkit.getDefaultToolkit();
		Frame		   frame = LySwingViewManager.getDefaultFrame();

		// Convert properties
		String		   orientation = (String)_properties.get("orientation");
		PageAttributes pageAttributes = new PageAttributes();

		if (orientation != null)
		{
			if ("portrait".equals(orientation))
				pageAttributes.setOrientationRequested(OrientationRequestedType.PORTRAIT);
			else
				pageAttributes.setOrientationRequested(OrientationRequestedType.LANDSCAPE);
		}

		_job = toolkit.getPrintJob(frame, "BF AppEngine Print Job", null, pageAttributes);
		if (_job == null) // Canceled or refused
			return;

		_inFile = ("file".equals(_properties.get("awt.print.destination")) ? true : false);
		_supportName = (_inFile ? (String)_properties.get("awt.print.fileName")
								: (String)_properties.get("awt.print.printer"));

		if ((_format == null) && !_inFile)
		{
			String format = (String)_properties.get("awt.print.options");

			_format = ((format != null) ? format.intern() : _format);
			_properties.put("awt.print.options", "");
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.print.LyPrintManager#end()
	 */
	public void end()
	{
		if (_job != null)
		{
			_job.end();
			_job.finalize();
		}
	}

	/**
	 * This print job is not used in Pdf mode.
	 *
	 * @return <code>false</code>
	 */
	protected boolean isPdfMode()
	{
		return false;
	}

	/**
	 * Gets graphics for current print job.
	 *
	 * @param  landscape the landscape orientation. This attribute is not used in this
	 *                   implementation.
	 * @return the graphics context for the current print job
	 */
	protected Graphics getGraphics(boolean landscape)
	{
		return _job.getGraphics();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.print.LyPrintManager#wasCancelled()
	 */
	public boolean wasCancelled()
	{
		return (!DEBUG_TO_SCREEN && (_job == null));
	}

	/**
	 * Prints a LyObject instance.
	 *
	 * @param  object        the LyObject to print
	 * @param  fields        the list of fields to print
	 * @param  fontReduction font reduction factor
	 * @param  startPage     the page number to start printing
	 * @return the number of pages drawn for the given object
	 * @see    LyPrintManager#print(Object, Vector, int, int)
	 */
	@SuppressWarnings("deprecation")
	public int print(Object object, Vector<Object> fields, int fontReduction, int startPage)
	{
		Frame  f_debug = null;
		Canvas canvas = null;

		if (DEBUG_TO_SCREEN)
		{
			// / *** DEBUG : DRAW TO SCREEN
			f_debug = new Frame();
			canvas = new Canvas();

			f_debug.setLayout(new BorderLayout());
			f_debug.add(canvas, "Center");
			f_debug.setSize(PAGE_WIDTH, PAGE_HEIGHT);
			f_debug.show();
		}
		else
		{
			if ((object == null) || (_job == null))
				return 0;
		}

		LyEnvironment environment = getEnvironment();
		Font		  bfont = getFont(getBoldFontName());
		Font		  font = getFont(getFontName());

		if (fontReduction > 0)
		{
			font = new Font(font.getName(), font.getStyle(), font.getSize() - fontReduction);
			bfont = new Font(bfont.getName(), bfont.getStyle(), bfont.getSize() - fontReduction);
		}

		Font	  ifont = new Font(bfont.getName(), bfont.getStyle() | Font.ITALIC,
				bfont.getSize());
		Graphics  g, pg;
		Dimension jd;

		// First call to getGraphics : mandatory for pdf mode
		if (_job != null)
			g = _job.getGraphics();
		else
			g = null;

		if (DEBUG_TO_SCREEN)
			jd = canvas.getSize();
		else
			jd = _job.getPageDimension();

		String		   title = environment.translate(_printDataInterface.getTitle(object));
		String		   parentImage = "LY_COMPANY_LOGO";
		String		   classImage = _printDataInterface.getImage(object);
		String		   date = DateFormat.getDateTimeInstance(DateFormat.FULL, DateFormat.MEDIUM,
				environment.getLocale()).format(new Date());

		int			   ph, pw = jd.width - LyPrintManager.LEFT_MARGIN - LyPrintManager.RIGHT_MARGIN;
		int			   x, y;
		int			   p, h, hh, fh, mlw, mh, nl;
		FontMetrics    fm = null, bfm = null, ifm = null;
		int			   th = 0, tw, bth = 0, ith = 0;

		int			   i, j, l, na = fields.size();
		Vector<String> tabs = new Vector<String>(10);
		String		   tab, null_tab = "null".intern();
		String[]	   labels = new String[na], lines = null;
		Object		   field;
		int			   t, nt;
		String		   field_tab, str;
		boolean		   new_tab, next_page;
		Dimension	   dim;
		Frame		   defaultFrame = LySwingViewManager.getDefaultFrame();
		String		   fieldLabel, tabString;

		// Check fontmetrics if needed
		if (fm == null)
		{
			if (g == null)
				g = defaultFrame.getGraphics();

			fm = g.getFontMetrics(font);
			th = fm.getHeight() + 2;
			bfm = g.getFontMetrics(bfont);
			bth = bfm.getHeight() + 2;
			ifm = g.getFontMetrics(ifont);
			ith = ifm.getHeight() + 2;
		}

		// Get all tabs and format labels
		for (i = 0, mlw = 0; i < na; i++)
		{
			field = fields.elementAt(i);
			tab = _printDataInterface.getTab(field);

			if ((tab == null) && !tabs.contains(null_tab))
				tabs.insertElementAt(null_tab, 0);
			else if ((tab != null) && !tabs.contains(tab))
				tabs.addElement(tab);

			fieldLabel = _printDataInterface.getLabel(field);
			labels[i] = LyString.format(environment.translate(fieldLabel) + " :", 30, false, '\0',
					false);

			dim = getSize(labels[i], bfm);
			if (dim.width > mlw)
				mlw = dim.width;
		}

		// Draw content
		for (p = startPage, t = 0, i = 0, l = 0, nl = 0, nt = tabs.size(), new_tab = true; t < nt;
				p++)
		{
			// DEBUG ...
			if (DEBUG_TO_SCREEN)
			{
				g = canvas.getGraphics();
				if (p != 1)
				{
					try
					{
						Thread.sleep(5000);
					}
					catch (InterruptedException e)
					{
					}
				}
			}
			else if (p != startPage)
				g = _job.getGraphics();

			g.setColor(BACKGROUND_COLOR);
			g.fillRect(0, 0, 420, 580); // XXX ????
			g.setColor(FONT_COLOR);

			g.setFont(bfont);
			next_page = false;

			// Draw Header
			hh = printHeader(defaultFrame, g, jd, classImage, title, parentImage, environment);

			// Draw page footer
			fh = printFooter(defaultFrame, g, jd, date, null, String.valueOf(p), environment);

			// Draw page content
			x = getXStart();
			y = getYStart(hh);
			ph = getPageHeight(jd.height, hh, fh);

			pg = g.create(LyPrintManager.LEFT_MARGIN, hh - y, pw, ph);

			for (; t < nt; t++)
			{
				tab = tabs.elementAt(t);
				if (tab == null_tab)
					tab = null;

				for (; i < na; i++)
				{
					field = fields.elementAt(i);
					field_tab = _printDataInterface.getTab(field);
					if (field_tab != tab)
						continue;

					if (new_tab)
					{
						// Draw tab content
						if ((y + ith + 10) > ph)
						{
							next_page = true;
							break;
						}

						y += ith + 5;
						if (tab != null)
						{
							y += 10;
							tabString = environment.translate(tab);
							pg.setFont(ifont);

							tw = ifm.stringWidth(tab);
							tw = ifm.stringWidth(tabString);

							pg.drawLine(x, y - 2, x + (mlw / 4), y - 2);
							pg.drawLine(x + (mlw / 4) + tw + 10, y - 2, x + pw, y - 2);

							pg.drawString(tabString, x + (mlw / 4) + 5, y);
						}

						y += 5;
						new_tab = false;
					}

					if (lines == null)
					{
						h = LyString.getNbLines(labels[i]) * bth;

						if ((y + h) > ph) // If the label does not fit
						{
							next_page = true;
							break;
						}

						str = _printDataInterface.getLabel(object, field);
						str = formatString(str, fm, pw - mlw - 10, '\n');
						mh = Math.max(h, LyString.getNbLines(str) * th);

						// Draw label
						pg.setFont(bfont);
						lines = LyString.stringToList(labels[i], '\n');
						nl = ((lines == null) ? 0 : lines.length);

						for (j = 0, h = 0; j < nl; j++)
						{
							h += bth;
							pg.drawString(lines[j], x, y + h);
						}

						lines = LyString.stringToList(str, '\n');
						nl = ((lines == null) ? 0 : lines.length);
						l = 0;
					}
					else
						mh = (nl - l) * th;

					// Draw the value
					pg.setFont(font);
					pg.setColor(FONT_COLOR);

					for (h = 0; l < nl; l++)
					{
						h += th;
						if ((y + h) > ph)
						{
							// The value does not fit
							next_page = true;
							break;
						}
						pg.drawString(lines[l], x + mlw + 10, y + h);
					}

					if (next_page)
						break;
					lines = null;
					y += mh;
				}
				if (next_page)
					break;
				i = 0;
				new_tab = true;
			}

			g.dispose();
		}

		return p - startPage;
	} // end method print

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.print.LyPrintManager#print(java.util.Vector, java.util.Vector, int)
	 */
	@SuppressWarnings("deprecation")
	public void print(Vector<LyObject> objects, Vector<Object> fields, int fontReduction)
	{
		LyEnvironment environment = getEnvironment();
		Dimension     jd, dd;
		Frame		  f_debug = null;
		Canvas		  canvas = null;
		int			  pv, ph, npv = 0, nph = 0;
		Graphics	  g, pg;
		boolean		  isPDF = isPdfMode();

		// Set landscape orientation for pdf
		g = getGraphics(true);

		String title = environment.translate(_printDataInterface.getTitle(null));
		String parentImage = "LY_COMPANY_LOGO";
		String classImage = _printDataInterface.getImage(null);

		if (DEBUG_TO_SCREEN)
		{
			// / *** DEBUG : DRAW TO SCREEN
			f_debug = new Frame();
			canvas = new Canvas();

			f_debug.setLayout(new BorderLayout());
			f_debug.add(canvas, "Center");
			f_debug.setSize(PAGE_HEIGHT, PAGE_WIDTH);
			f_debug.show();
		}
		else
		{
			if ((_job == null) || (objects == null))
				return;
		}

		if (DEBUG_TO_SCREEN)
			jd = canvas.getSize();
		else
			jd = _job.getPageDimension();

		String date = DateFormat.getDateTimeInstance(DateFormat.FULL, DateFormat.MEDIUM,
				environment.getLocale()).format(new Date());
		String page;
		int    w = jd.width - LyPrintManager.LEFT_MARGIN - LyPrintManager.RIGHT_MARGIN;
		int    hh, fh;
		Font   font = getFont("LY_PRINT_CAPTION");
		Frame  defaultFrame = LySwingViewManager.getDefaultFrame();

		ph = 0;

		do
		{
			ph++;
			pv = 0;

			do
			{
				pv++;

				if (_printDataInterface instanceof LyPrintListenerInterface)
					((LyPrintListenerInterface)_printDataInterface).print(((ph - 1) * npv) + pv);

				//page = environment.translate("LY_PAGE") + " " + pv + " - " + ph;
				//environment.logInfo("Printing : " + page);

				if (DEBUG_TO_SCREEN)
				{
					g = canvas.getGraphics();
					g.setColor(BACKGROUND_COLOR);
					g.fillRect(0, 0, jd.width, jd.height);
					g.setColor(FONT_COLOR);
				}
				else if ((ph != 1) || (pv != 1))
					g = getGraphics(true);

				g.setFont(font);

				// Draw Header
				hh = printHeader(defaultFrame, g, jd, classImage, title, parentImage, environment);

				// Draw footer a first time to know its height
				fh = printFooter(defaultFrame, g, jd, date, null, "", environment);

				int x0 = 0, y0 = 0;

				// Draw page content
				if (isPDF)
				{
					pg = g.create();
					pg.setClip(LyPrintManager.LEFT_MARGIN, hh, w, jd.height - hh - fh);
					x0 = LyPrintManager.LEFT_MARGIN;
					y0 = hh;
				}
				else
					pg = g.create(LyPrintManager.LEFT_MARGIN, hh, w, jd.height - hh - fh);

				dd = printPage(pg, ph, pv, objects, fields, fontReduction, x0, y0, jd);

				if (DEBUG_TO_SCREEN)
				{
					try
					{
						Thread.sleep(3000);
					}
					catch (InterruptedException e)
					{
					}
				}
				else
					g.dispose();

				if ((npv == 0) && (nph == 0))
				{
					nph = dd.width;
					npv = dd.height;
				}
				page = environment.translate("LY_PAGE") + " " + pv + " / " + npv;
				environment.logInfo("Printing : " + page);
				// At this point the max number of pages is know -> redraw footer
				printFooter(defaultFrame, g, jd, date, null, page, environment);
			}
			while (pv < npv);
		}
		while (ph < nph);
	} // end method print

	/**
	 * Prints a page of LyObjects.
	 *
	 * @param  g             graphic context
	 * @param  hpage         page width
	 * @param  vpage         page height
	 * @param  objects       list of objects that should be printed
	 * @param  fields        list of fields to print
	 * @param  fontReduction font reduction factor
	 * @param  x0            horizontal offset
	 * @param  y0            vertical offset
	 * @param  size          usable print size in the view
	 * @return the dimension of the printed page
	 */
	protected Dimension printPage(Graphics g, int hpage, int vpage, Vector<LyObject> objects, Vector<Object> fields,
		int fontReduction, int x0, int y0, Dimension size)
	{
		if ((g == null) || (objects == null) || (objects.size() == 0) || (fields == null)
				|| (fields.size() == 0))
			return null;

		Rectangle bounds = g.getClipBounds();

		if (bounds == null)
			return null;

		LyEnvironment environment = getEnvironment();
		Font		  font;
		Font		  hfont;

		font = getFont(getFontName());
		hfont = getFont(getBoldFontName());

		// Reduce font Size
		if (fontReduction > 0)
		{
			font = new Font(font.getName(), font.getStyle(), font.getSize() - fontReduction);
			hfont = new Font(hfont.getName(), hfont.getStyle(), hfont.getSize() - fontReduction);
		}

		FontMetrics fm = g.getFontMetrics(font);
		FontMetrics hfm = g.getFontMetrics(hfont);

		// Check if we have footers
		boolean     hasFooters = _printDataInterface.hasFooters();

		// Create drawing parameters
		int		    c, nc = fields.size();
		Object	    field;
		int		    lh = fm.getHeight() + 6;
		int		    hh = hfm.getHeight() + 8;
		int		    fh = (hasFooters ? lh : 0);
		int		    hu = bounds.height - hh; // available height
		int		    wu = bounds.width; // available width
		int		    nl = objects.size(); // number of lines
		int		    nlp = hu / lh; // lines / page
		int		    nlt = ((fh != 0) ? (nl + ((fh + lh - 1) / lh)) : nl);

		if (nlp == 0)
			return null;

		int    nph = 1, npv = (nlt + nlp - 1) / nlp; // page count
		int    x, y, p, wp;
		int    cmin, cmax;

		// Compute column width
		Object object;
		int    o, no = objects.size();
		int    w;
		int[]  col_width = new int[nc];
		String str;
		int    sortIndex = -1;
		int    sortIconSize = (1 * hh / 4);
		BitSet compactedCols = new BitSet();

		for (c = 0; c < nc; c++)
		{
			field = fields.elementAt(c);
			str = _printDataInterface.getLabel(field);
			str = environment.translate(str);

			col_width[c] = hfm.stringWidth(str) + 2;

			// Draw indication on sort
			sortIndex = _printDataInterface.getSortIndex(field);
			if (sortIndex >= 0)
			{
				col_width[c] += sortIconSize;
				col_width[c] += fm.stringWidth(String.valueOf(sortIndex)) + 2;
			}

			for (o = 0; o < no; o++)
			{
				object = objects.elementAt(o);
				str = _printDataInterface.getShortLabel(object, field);
				if (str != null)
				{
					w = fm.stringWidth(str) + 6;

					if (w > col_width[c])
						col_width[c] = w;
				}

				str = _printDataInterface.getFooterLabel(field);
				if (str != null)
				{
					w = fm.stringWidth(str) + 2;

					if (w > col_width[c])
						col_width[c] = w;
				}
			}
			
			// If column is wider than page, reduce its width
			if (col_width[c] > size.width - LEFT_MARGIN - RIGHT_MARGIN)
			{
				col_width[c] = size.width - LEFT_MARGIN - RIGHT_MARGIN;
				compactedCols.set(c);
			}
		}

		// calculate column width and number of horizontal pages
		for (c = 0, x = x0; c < nc; c++)
		{
			if ((x + col_width[c]) > wu)
			{
				x = col_width[c];
				nph++;
			}
			else
				x += col_width[c];
		}

		if ((hpage < 1) || (vpage < 1) || (hpage > nph) || (vpage > npv))
			return new Dimension(nph, npv);

		// find columns to draw
		for (c = 0, x = x0, cmin = 0, p = 1; c < nc; c++)
		{
			if ((x + col_width[c]) > wu)
			{
				if (p == hpage)
					break;
				p++;
				x = col_width[c];
				cmin = c;
			}
			else
				x += col_width[c];
		}

		cmax = c;

		// Add column footers
		String[] footers = new String[cmax];

		for (c = cmin; c < cmax; c++)
		{
			field = fields.elementAt(c);
			str = _printDataInterface.getFooterLabel(field);
			footers[c] = str;
		}

		// Draw requested page
		int     l, lmin, lmax;
		boolean left_alignment;
		boolean last_page = (vpage == npv);
		int     offset = 0, invertSort;
		Polygon poly;

		g = g.create();

		lmin = (vpage - 1) * nlp;
		lmax = (((lmin + nlp) < nl) ? (lmin + nlp) : nl);
		wp = 0;
		x = x0;

		for (c = cmin; c < cmax; c++)
		{
			// Draw label
			g.setFont(hfont);
			hfm = g.getFontMetrics(hfont);

			y = y0;
			field = fields.elementAt(c);
			left_alignment = _printDataInterface.isLeftAligned(field);
			str = _printDataInterface.getLabel(field);
			str = environment.translate(str);

			g.setColor(HEADER_COLOR);
			g.fillRect(x, y, col_width[c], hh - 1);

			g.setColor(LINE_COLOR);
			g.drawRect(x, y, col_width[c], hh - 1);

			g.setColor(FONT_COLOR);
			offset = 0;

			// Draw indication on sort
			sortIndex = _printDataInterface.getSortIndex(field);
			if (sortIndex >= 0)
			{
				sortIndex++;
				invertSort = 1;

				poly = new Polygon();

				if (_printDataInterface.isIncreasing(field))
					invertSort = -1;

				poly.addPoint(x + col_width[c] - fm.stringWidth(String.valueOf(sortIndex)) - 2
					- (sortIconSize / 2), y + ((hh - (sortIconSize * invertSort)) / 2) + 2);
				poly.addPoint(x + col_width[c] - fm.stringWidth(String.valueOf(sortIndex)) - 2
					- sortIconSize, y + ((hh + (sortIconSize * invertSort)) / 2) + 2);
				poly.addPoint(x + col_width[c] - fm.stringWidth(String.valueOf(sortIndex)) - 2,
					y
					+ ((hh + (sortIconSize * invertSort)) / 2) + 2);

				g.setColor(SORT_COLOR);
				g.fillPolygon(poly);
				g.setColor(FONT_COLOR);
				g.drawPolygon(poly);

				g.setFont(font);
				g.drawString("" + sortIndex,
					x
					+ (col_width[c] - fm.stringWidth(String.valueOf(sortIndex)) - 1), y + hh
					- 7);

				offset = fm.stringWidth(String.valueOf(sortIndex)) + sortIconSize + 1;
			}

			g.setFont(hfont);
			g.drawString(str, x + ((col_width[c] - offset - hfm.stringWidth(str)) / 2), y + hh - 7);

			y += hh - 1;

			// Draw values
			g.setFont(font);
			g.setColor(FONT_COLOR);
			fm = g.getFontMetrics(font);

			for (l = lmin; l < lmax; l++)
			{
				object = objects.elementAt(l);
				str = _printDataInterface.getShortLabel(object, field);

				// is the column for real, and some of it visible?
				if ((str != null) && (col_width[c] > 0) && (x < wu))
				{
					String	fullLabel = str;
					
					if (compactedCols.get(c))
					{
						while ((fm.stringWidth(fullLabel) >= col_width[c] - 6) && (str.length() > 0))
						{
							str = str.substring(0, str.length() - 1);
							fullLabel = str + "...";
						}
					}
					
					if (left_alignment)
						g.drawString(fullLabel, x + 2, y + lh - 4);
					else
						g.drawString(fullLabel, x + col_width[c] - fm.stringWidth(fullLabel) - 2, y + lh - 4);
				}
				
				y += lh;
			}

			// Draw column bottom
			if (last_page && (footers != null))
			{
				y = y0 + hh + ((nlp + nl - nlt) * lh);
				g.setFont(font);
				hfm = g.getFontMetrics(font);

				g.setColor(HEADER_COLOR);
				g.fillRect(x, y, col_width[c], lh - 1);

				g.setColor(LINE_COLOR);
				g.drawRect(x, y, col_width[c], lh - 1);

				str = footers[c];

				if (str != null)
				{
					g.setColor(FONT_COLOR);

					if (left_alignment)
						g.drawString(str, x + 2, y + lh - 5);
					else
						g.drawString(str, x + col_width[c] - hfm.stringWidth(str) - 2, y + lh - 5);
				}
			}

			x += col_width[c];
			wp += col_width[c];

		}

		wp = ((wp > wu) ? wu : wp);

		if ((vpage == npv) && hasFooters)
			nlp = nlp + nl - nlt;

		// Draw lines
		x = x0;
		y = y0 + hh;

		for (l = 0; l < nlp; l++)
		{
			y += lh;
			g.setColor(LINE_COLOR);
			g.drawLine(x, y, x + wp, y);
		}

		// Draw columns
		hu = nlp * lh;
		y = y0 + hh - 1;
		g.setColor(LINE_COLOR);
		g.drawLine(x, y, x, y + hu);

		for (c = cmin; c < cmax; c++)
		{
			x += col_width[c];

			g.setColor(LINE_COLOR);
			g.drawLine(x, y, x, y + hu);
		}

		g.setColor(LINE_COLOR);
		g.drawLine(x, y, x, y + hu);

		return new Dimension(nph, npv);
	} // end method printPage

	/**
	 * Gets the standard font name.
	 *
	 * @return the font name.
	 */
	protected String getFontName()
	{
		return "LY_PRINT_TEXT";
	}

	/**
	 * Gets the bold font name.
	 *
	 * @return the font name.
	 */
	protected String getBoldFontName()
	{
		return "LY_PRINT_CAPTION";
	}

	/**
	 * Gets horizontal start position. It is always 0 in this implementation.
	 *
	 * @return 0
	 */
	protected int getXStart()
	{
		return 0;
	}

	/**
	 * Gets vertical start position. It is always 0 in this implementation.
	 *
	 * @param  headerHeight the height of the header, not used here
	 * @return 0
	 */
	protected int getYStart(int headerHeight)
	{
		return 0;
	}

	/**
	 * Gets the page height.
	 *
	 * @param  jobHeight    height of the job
	 * @param  headerHeight height of the header
	 * @param  footerHeight height of the footer
	 * @return the height of the page. It is the height job minus the header and footer heights.
	 */
	protected int getPageHeight(int jobHeight, int headerHeight, int footerHeight)
	{
		return (jobHeight - headerHeight - footerHeight);
	}

	/**
	 * Gets the font with the specified name.
	 *
	 * @param  key the name of the font to retreive
	 * @return the <code>Font</code> associated with the given name
	 */
	// XXX TO REMOVE
	final Font getFont(String key)
	{
		LyFont font = getEnvironment().getLyFont(key);
		int    style = 0;

		if (font.isBold())
			style |= Font.BOLD;
		if (font.isItalic())
			style |= Font.ITALIC;

		return new Font(font.getFontName(), style, font.getSize());
	}
} // end class LySwingPrintManager
