/*
 * LySwingRelation.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.plaf.basic.*;
import javax.swing.plaf.metal.*;

import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Relation object (as described in the dtd) for swing viewer.
 *
 * <p>An input field of Relation type allows to choose one or more values among a list of values.
 * This list offers generally application objects belonging to a same class and is used to establish
 * a relation between an object and one or more objects of this class. It is possible to specify
 * whether the relation is multiple or not.</p>
 *
 * @version  $Id: LySwingRelation.java 30713 2011-05-03 07:09:31Z jlsaouli $
 * @author   Lyria
 * @see      LyRelation
 */
public class LySwingRelation extends LySwingFieldItem
    implements LySwingComponent,
        LyGraphicalRelation,
        ItemListener,
        LyActionListener,
        LySetEnabledInterface,
        LySetEditableInterface,
        LyGetValueInterface,
        LySetIdleInterface
{
    public static final String __VERSION = "$Id: LySwingRelation.java 30713 2011-05-03 07:09:31Z jlsaouli $";

    /** Separator between relation items. */
    public static final char ITEM_SEP = '\1';

    /**
     * Separator between values ids.
     *
     * @deprecated
     */
    public static final char VALUE_SEP = '\2';

    /**
     * Empty link. A space filled <code>String</code> is used to pack correctly its container
     * components.
     */
    public static final String DEFAULT_EMPTY_LINK = "                         ";

    /** The view independent relation description. */
    protected LyRelation _relation;

    /** The swing component for this relation element. */
    protected Component _component;

    /** Idle mode. */
    protected boolean _idle;

    /** Indicates if this relation is editable. */
    protected boolean _editable;

    /** Indicates if this relation is enabled. */
    protected boolean _enabled;

    /** Viewer used for auto-completion. */
    protected LyJAutoCompletionViewer _autoCompletionViewer;

    /** Number of pixels of the border, -1 for auto size (2 pixels by default). */
    public static int RELATION_BORDER = 2;

    /**
     * Class constructor of a Relation object for the Swing target.
     *
     * @param  relation  a RELATION component defined for every type of view targets. This relation
     *                   is initialized for the swing view and filled with its default values.
     */
    public LySwingRelation(LyRelation relation)
    {
        super(relation);
        _relation = relation;
        _autoCompletionViewer = null;

        initialize();
        setContent();
    }

    /**
     * Initializes the relation component.
     */
    public void initialize()
    {
        LyComponent parent = _relation.getParent();
        LyForm      grandParent = null;

        while (!(parent instanceof LyFormField))
            parent = parent.getParent();

        if (((LyFormField)parent).getParent() instanceof LyForm)
            grandParent = (LyForm)((LyFormField)parent).getParent();

        Font font = getFont("LY_DEFAULT_FONT");

        if (!_relation.isMultiple())
        {
            if (_relation.getAutoCompletionMax() != 0)
                _component = createAutoCompletionCombo();
            else
            {
                _component = createCombo();
                ((ItemSelectable)_component).addItemListener(this);
            }
        }
        else
        {
            Color color = getColor("LY_TABLE_SELECT_COLOR");

            if (_relation.getAutoCompletionMax() != 0)
                _component = createAutoCompletionMultiSelector(getEnvironment(), color);
            else
                _component = createMultiSelector(getEnvironment(), color);

            LyActionTransmitter actionTransmitter = ((LyEventTransmitter)_component).getActionTransmitter();

            if (actionTransmitter != null)
                actionTransmitter.addActionListener(this);
        }

        _component.setFont(font);

        _editable = ((LyFormField)parent).isEditable()
                && ((grandParent == null) || grandParent.isEditable());
        _enabled = ((LyFormField)parent).isEnabled();

        _idle = false;
    } // end method initialize

    /**
     * Refreshes the content of the relation component.
     *
     * @see  #setContent(LyLink, boolean)
     */
    public void setContent()
    {
        setContent(null, true);
    }

    /**
     * Sets the content of the view. Adds or removes the specified link.
     *
     * @param  paramLink  a link that may be or may not be in the parent list <code>null</code> to
     *                    don't add or remove a link and only trigger a refresh.
     * @param  addLink    Indicates whether the link should be added or removed.
     */
    protected void setContent(LyLink paramLink, boolean addLink)
    {
        // FFT00176 : avoid clipping while changing selection in the
        // multi-selector
        if (_idle)
            return;

        LyLink link;
        int    n = _relation.getLinkCount();

        if (!_relation.isMultiple())
        {
            if (_relation.getAutoCompletionMax() != 0)
            {
                String[] newInternalValues = new String[n];
                String[] newDisplayedValues = new String[n];
                String[] newDisplayedShortValues = new String[n];
                int      selectedIndex = -1;

                for (int i = 0; i < n; i++)
                {
                    link = _relation.getLink(i);

                    // make sure that the internal view is ready
                    link.getInternalView();

                    newInternalValues[i] = link.getId();
                    newDisplayedValues[i] = translate(link.getLabel());
                    newDisplayedShortValues[i] = translate(link.getShortLabel());

                    if (newDisplayedShortValues[i] == null)
                        newDisplayedShortValues[i] = newDisplayedValues[i];

                    if (link.isSelected())
                        selectedIndex = i;
                }

                _autoCompletionViewer.setValues(newInternalValues, newDisplayedValues, newDisplayedShortValues);

                if (selectedIndex >= 0)
                    _autoCompletionViewer.setTextValue(selectedIndex, false);

                updateAutoCompletionWidth();
                updateComboToolTip();
            }
            else
            {
                // Clear
                clearCombo(_component);

                int selectedIdx = -1;

                for (int i = 0; i < n; i++)
                {
                    link = _relation.getLink(i);

                    // make sure that the internal view is ready
                    link.getInternalView();

                    addLink(link, -1);

                    if (link.isSelected())
                        selectedIdx = i;
                }

                if (selectedIdx >= 0)
                    setComboSelectedIndex(_component, selectedIdx);
            }
        }
        else
        {
            if (_relation.getAutoCompletionMax() == 0)
                fillMultiSelector(_component, paramLink, addLink);
            else
            {
                fillAutoCompletionMultiSelector(_component, paramLink, addLink);
                updateAutoCompletionWidth();
            }
        }
    } // end method setContent

    /**
     * Creates a new ComboBox for the relations elements and registers the key listeners.
     *
     * @return  a new empty ComboBox
     */
    public Component createCombo()
    {
        JComboBox combo = new JComboBox()
            {
                private static final long serialVersionUID = -4181933161210777534L;
                Dimension                 _preferredSize;

                /*
                 * (non-Javadoc)
                 *
                 * @see java.awt.Component#getPreferredSize()
                 */
                public Dimension getPreferredSize()
                {
                    if (_preferredSize != null)
                        return _preferredSize;

                    Dimension dim = super.getPreferredSize();

                    // May be null when removed in simple filters XXXX Clean up
                    if (getEnvironment() == null)
                        return dim;

                    int         minWidth = _relation.getMinWidth();
                    int         maxWidth = _relation.getMaxWidth();
                    int         nCols = _relation.getNCols();
                    int         width;
                    FontMetrics fm = this.getFontMetrics(this.getFont());

                    // If we don't have font metrics yet, just return.
                    int line_height = dim.height;

                    if ((RELATION_BORDER >= 0) && (fm != null))
                        line_height = fm.getHeight() + (2 * RELATION_BORDER);

                    width = Math.max(minWidth, dim.width);

                    if (maxWidth > 0)
                        width = Math.min(width, maxWidth);

                    if (nCols > 0)
                        width = (fm.stringWidth("0123456789") * nCols) / 10;

                    _preferredSize = new Dimension(width, line_height);

                    if (_relation.getListWidth() > 0)
                        ((_LyComboBoxRenderer)getRenderer()).setWidth(_relation.getListWidth());
                    else
                        ((_LyComboBoxRenderer)getRenderer()).setWidth(_preferredSize.width);

                    return _preferredSize;
                }

                /*
                 * (non-Javadoc)
                 *
                 * @see javax.swing.JComponent#setPreferredSize(java.awt.Dimension)
                 */
                public void setPreferredSize(Dimension dimension)
                {
                    _preferredSize = dimension;

                    if (_relation.getListWidth() > 0)
                        ((_LyComboBoxRenderer)getRenderer()).setWidth(_relation.getListWidth());
                    else
                        ((_LyComboBoxRenderer)getRenderer()).setWidth(_preferredSize.width);
                }
            };

        if (_relation.getListWidth() > 0)
            combo.setUI(new _LyComboBoxUI(_relation.getListWidth()));

        combo.setRenderer(new _LyComboBoxRenderer());

        // Compute width
        int         maxWidth = _relation.getMaxWidth();
        int         nCols = _relation.getNCols();
        int         width = _relation.getMinWidth();
        FontMetrics fm = combo.getFontMetrics(combo.getFont());

        if (maxWidth > 0)
            width = Math.min(width, maxWidth);

        if (nCols > 0)
            width = (fm.stringWidth("0123456789") * nCols) / 10;

        if (_relation.getListWidth() > 0)
            ((_LyComboBoxRenderer)combo.getRenderer()).setWidth(_relation.getListWidth());
        else
            ((_LyComboBoxRenderer)combo.getRenderer()).setWidth(width);

        // Compute height
        int height = _relation.getNRows();

        if (height > 0)
            combo.setMaximumRowCount(height);

        // May be null when removed in simple filters XXXX Clean up
        if (getEnvironment() != null)
            combo.setFont(getFont("LY_DEFAULT_FONT"));

        // This handler is used to validate the form using ENTER key or close using ESCAPE key
        LySwingFormKeyHandler keyHandler = new LySwingFormKeyHandler(getActionTransmitter());

        keyHandler.setNotifyValidate(false); // already default in swing
        combo.addKeyListener(keyHandler);

        return combo;
    } // end method createCombo

    /**
     * Creates a new editable comboBox for the relations elements.
     *
     * @return  a new empty editable ComboBox
     */
    public Component createAutoCompletionCombo()
    {
        LyJSelector selector = new LyJSelector(null, _relation.getEnvironment(), true);
        LyJButton   button = selector.addButton(getImage("LY_ARROW_DOWN_BIG"), "");

        _autoCompletionViewer = new LyJAutoCompletionViewer(selector.getTextField(), button, new String[0],
                new String[0], new String[0], true, _relation.getAutoCompletionMax(), getEnvironment());

        _autoCompletionViewer.setActionListener(this);

        // This handler is used to validate the form using ENTER key or close using ESCAPE key
        LySwingFormKeyHandler keyHandler = new LySwingFormKeyHandler(getActionTransmitter(), this);

        keyHandler.setNotifyValidate(false); // already default in swing
        selector.addKeyListener(keyHandler);

        return selector;
    }

    /**
     * Creates a new multiselector component to represent a multiselection relation.
     *
     * @param   env    the environment needed to find mutliselector images
     * @param   color  the selection color, not used in this implementation
     *
     * @return  a new empty multiselector
     */
    public Component createMultiSelector(LyEnvironment env, Color color)
    {
        int              height = _relation.getNRows();
        LyJMultiSelector component = new LyJMultiSelector(env, height, _relation.isOrdered());

        // This handler is used to validate the form using ENTER key or close using ESCAPE key
        LySwingFormKeyHandler keyHandler = new LySwingFormKeyHandler(getActionTransmitter());

        keyHandler.setNotifyValidate(false); // already default in swing
        component.addKeyListener(keyHandler);

        return component;
    }

    /**
     * Creates a new editable multiselector component to represent a multiselection relation.
     *
     * @param   env    the environment needed to find mutliselector images
     * @param   color  the selection color, not used in this implementation
     *
     * @return  a new empty editable multiselector
     */
    public Component createAutoCompletionMultiSelector(LyEnvironment env, Color color)
    {
        int                            height = _relation.getNRows();
        LyJAutoCompletionMultiSelector component = new LyJAutoCompletionMultiSelector(env, height,
                _relation.getAutoCompletionMax(), _relation.isOrdered());

        // This handler is used to validate the form using ENTER key or close using ESCAPE key
        LySwingFormKeyHandler keyHandler = new LySwingFormKeyHandler(getActionTransmitter());

        keyHandler.setNotifyValidate(false); // already default in swing
        component.addKeyListener(keyHandler);
        _autoCompletionViewer = component.getAutoCompletionViewer();

        return component;
    }

    /**
     * Removes all the items of a combo box.
     *
     * @param  combo  the combo box to unfill
     */
    public void clearCombo(Component combo)
    {
        ((ItemSelectable)combo).removeItemListener(this);

        ((JComboBox)combo).removeAllItems();
        updateComboToolTip();
        ((ItemSelectable)combo).addItemListener(this);
    }

    /**
     * Sets the selected index of the combo.
     *
     * @param  combo          select the items of this combo box
     * @param  selectedIndex  index of the item to select in the combo box
     *
     * @see    #getComboSelectedIndex(Component)
     */
    public void setComboSelectedIndex(Component combo, int selectedIndex)
    {
        if (_autoCompletionViewer != null)
        {
            _autoCompletionViewer.setTextValue(selectedIndex, false);
            updateComboToolTip();
        }
        else
        {
            ((ItemSelectable)combo).removeItemListener(this);
            ((JComboBox)combo).setSelectedIndex(selectedIndex);
            updateComboToolTip();
            ((ItemSelectable)combo).addItemListener(this);
        }
    }

    /**
     * Selects/unselects an object id in the multiselector.
     *
     * @param  multiSelector   the source multiselector
     * @param  selectedObject  sets or unsets the selected flag for this given object
     * @param  selected        the selected flag for the given object
     */
    public void setMultiSelectorSelectedObject(Component multiSelector, Object selectedObject, boolean selected)
    {
        LyLink selectedLink = (LyLink)selectedObject;

        if (_relation.getAutoCompletionMax() == 0)
        {
            if (selected)
                ((LyJMultiSelector)multiSelector).select(selectedLink);
            else
                ((LyJMultiSelector)_component).unselect(selectedLink);
        }
        else
            ((LyJAutoCompletionMultiSelector)multiSelector).selectValue(selectedLink.getId(), selected);
    }

    /**
     * Sets the content of the multiselector depending on the relation's links.
     *
     * @param  multiSelector  the multiselector to fill
     * @param  paramLink      a link that may be or may not be in the parent list
     * @param  addLink        specifies if we should adds the links or discard them
     */
    public void fillMultiSelector(Component multiSelector, LyLink paramLink, boolean addLink)
    {
        Vector<LyLink> unSelectedLinks = new Vector<LyLink>();
        Vector<LyLink> selectedLinks = new Vector<LyLink>();
        int            i, n = _relation.getLinkCount();
        LyLink         link;

        for (i = 0; i < n; i++)
        {
            link = _relation.getLink(i);

            // Make sure that the internal view is ready
            link.getInternalView();

            // Check if we should ignore this link
            if ((link == paramLink) && (!addLink))
                continue;

            if (!link.isSelected())
                unSelectedLinks.addElement(link);
            else
                selectedLinks.addElement(link);
        }

        ((LyJMultiSelector)multiSelector).set(unSelectedLinks, selectedLinks);
    }

    /**
     * Sets the content of the auto-completion multiselector depending on the relation's links.
     *
     * @param  multiSelector  the multiselector to fill
     * @param  paramLink      a link that may be or may not be in the parent list
     * @param  addLink        specifies if we should adds the links or discard them
     */
    public void fillAutoCompletionMultiSelector(Component multiSelector, LyLink paramLink, boolean addLink)
    {
        LyLink         link;
        int            i, n = _relation.getLinkCount();
        Vector<String> newInternalValuesList = new Vector<String>(n);
        Vector<String> newDisplayedValuesList = new Vector<String>(n);
        Vector<String> newDisplayedShortValuesList = new Vector<String>(n);
        Vector<String> selectedValues = new Vector<String>();

        for (i = 0; i < n; i++)
        {
            link = _relation.getLink(i);

            // Make sure that the internal view is ready
            link.getInternalView();

            // Check if we should ignore this link
            if ((link == paramLink) && (!addLink))
                continue;

            newInternalValuesList.addElement(link.getId());
            newDisplayedValuesList.addElement(translate(link.getLabel()));

            if (translate(link.getShortLabel()) != null)
                newDisplayedShortValuesList.addElement(translate(link.getShortLabel()));
            else
                newDisplayedShortValuesList.addElement(translate(link.getLabel()));

            if (link.isSelected())
                selectedValues.addElement(link.getId());
        }

        n = newInternalValuesList.size();

        String[] newInternalValues = new String[n];
        String[] newDisplayedValues = new String[n];
        String[] newDisplayedShortValues = new String[n];

        for (i = 0; i < n; i++)
        {
            newInternalValues[i] = newInternalValuesList.elementAt(i);
            newDisplayedValues[i] = newDisplayedValuesList.elementAt(i);
            newDisplayedShortValues[i] = newDisplayedShortValuesList.elementAt(i);
        }

        _autoCompletionViewer.setValues(newInternalValues, newDisplayedValues, newDisplayedShortValues);
        ((LyJAutoCompletionMultiSelector)multiSelector).selectValues(selectedValues);

        // Select first item of the list
        if ((paramLink == null) || addLink)
            _autoCompletionViewer.setTextValue(0, false);
    }

    /**
     * Gets the index of the specified combo box in the current combo.
     *
     * @param   combo  the combo that is searched in the current combo box
     *
     * @return  the selected index of the combo
     *
     * @see     #setComboSelectedIndex(Component, int)
     */
    public int getComboSelectedIndex(Component combo)
    {
        if (_relation.getAutoCompletionMax() == 0)
            return ((JComboBox)combo).getSelectedIndex();
        else
            return _autoCompletionViewer.getSelectedIndex();
    }

    /**
     * Gets the multiselector value of the specified multiselector.
     *
     * @param   multiSelector  the multiselector component that is searched for its selection
     *
     * @return  the multiselector selection as a String
     */
    public String getMultiSelectorValue(Component multiSelector)
    {
        if (_relation.getAutoCompletionMax() == 0)
            return ((LyJMultiSelector)multiSelector).getValue();
        else
            return ((LyJAutoCompletionMultiSelector)multiSelector).getValue();
    }

    /**
     * Adds an item in the specified combo at the specified position.
     *
     * @param  combo  the combo box component
     * @param  item   the item to add
     * @param  pos    the position of the item in the combo
     */
    public void addItemToComboAt(Component combo, Object item, int pos)
    {
        // FFT00075
        // if idle mode, don't update component
        if (_idle)
            return;

        ((ItemSelectable)combo).removeItemListener(this);

        if (pos == -1) // add at the end
            ((JComboBox)combo).addItem(item);
        else
            ((JComboBox)combo).insertItemAt(item, pos);

        ((ItemSelectable)combo).addItemListener(this);
    }

    /**
     * Removes the item from the given combo at the specified position.
     *
     * @param  combo  the combo containing the item to remove
     * @param  pos    the position of the item to remove
     */
    public void removeItemFromComboAt(Component combo, int pos)
    {
        // FFT00075
        // if idle mode, don't update component
        if (_idle)
            return;

        ((ItemSelectable)combo).removeItemListener(this);

        ((JComboBox)combo).removeItemAt(pos);

        ((ItemSelectable)combo).addItemListener(this);
    }

    /**
     * Adds an item in the specified multiselector at the specified position.
     *
     * @param  multiSelector  the multiselector component
     * @param  item           the item to add
     * @param  pos            the position of the item in the combo
     */
    public void addItemToMultiSelectorAt(Component multiSelector, Object item, int pos)
    {
        // FFT00075
        // if idle mode, don't update component
        if (_idle)
            return;

        ((LyJMultiSelector)multiSelector).addItemAt((LyLink)item, pos);
    }

    /**
     * Gets the number of combo items in the specified combo box.
     *
     * @param   combo  the combo box
     *
     * @return  the number of elements in the combo
     */
    public int getComboItemCount(Component combo)
    {
        return ((JComboBox)combo).getItemCount();
    }

    /**
     * Gets the combo item at the specified index in the given combo box.
     *
     * @param   combo  the combo box
     * @param   index  the index of the item
     *
     * @return  the item at the specified index from the combo
     */
    public Object getComboItemAt(Component combo, int index)
    {
        return ((JComboBox)combo).getItemAt(index);
    }

    // FFT00075
    /**
     * Sets the idle mode on this component. Implements idle mode on relations.
     *
     * @see  leon.view.setinterface.LySetIdleInterface#setIdle(boolean)
     */
    public void setIdle(boolean status)
    {
        _idle = status;
        if (!status)
            setContent();

    }

    /**
     * Processes changed state event only if relation is not multiple.
     *
     * @see  java.awt.event.ItemListener#itemStateChanged(java.awt.event.ItemEvent)
     */
    public void itemStateChanged(ItemEvent e)
    {
        boolean selected = (e.getStateChange() == ItemEvent.SELECTED);

        // FFT00461
        if (!selected)
            return;

        // Retrieve the item on which the event occured
        if (_actionTransmitter != null)
        {
            int           selectedIndex = getComboSelectedIndex(_component);
            LyActionEvent event = null;

            // retrieve the associated link
            if (event == null)
            {
                LyLink link = _relation.getLink(selectedIndex);

                if (link != null)
                {
                    updateLinkSelection();

                    event = new LyActionEvent(this, null, link.getId(), LyActionDemand.TYPE_VALUE_CHANGED);
                }
            }

            if (event != null)
                _actionTransmitter.processActionEvent(event);
        }

        updateComboToolTip();
    } // end method itemStateChanged

    /**
     * Updates the tooltip for this combo.
     */
    protected void updateComboToolTip()
    {
        if (_component == null)
            return;

        if (_component instanceof LyJSelector)
        {
            String internalValue = _autoCompletionViewer.getSelectedValue();
            String tooltip = _autoCompletionViewer.getDisplayedValue(internalValue);

            if ((tooltip != null) && (tooltip.length() > 0))
                ((LyJSelector)_component).getTextField().setToolTipText(tooltip);
            else
                ((LyJSelector)_component).getTextField().setToolTipText(null);

            return;
        }

        if (!(_component instanceof JComboBox))
            return;

        JComboBox combo = (JComboBox)_component;
        Object    selectedItem = combo.getSelectedItem();

        if (selectedItem != null)
        {
            String value = selectedItem.toString();
            String tmp = value.trim();

            if (tmp.length() > 0)
                combo.setToolTipText(value);
            else
                combo.setToolTipText(null);
        }
        else
            combo.setToolTipText(null);
    }

    /**
     * Updates the width of the text field used for auto-completion.<br />
     * Depending on the values contained in the auto-completion mechanism, text component must have
     * its width changed.
     */
    protected void updateAutoCompletionWidth()
    {
        Font        font = getFont("LY_DEFAULT_FONT");
        FontMetrics fm = _component.getFontMetrics(font);
        String[]    values = _autoCompletionViewer.getDisplayedShortValues();
        int         i, n = ((values == null) ? 0 : values.length);
        int         width = _relation.getMinWidth() - 15; // Remove button width
        int         maxWidth = _relation.getMaxWidth() - 15; // Remove button width
        int         nCols = _relation.getNCols();
        int         stringWidth;

        for (i = 0; i < n; i++)
        {
            stringWidth = fm.stringWidth(values[i]);

            if ((stringWidth + 8) > width)
                width = stringWidth + 8;
        }

        int height = _autoCompletionViewer.getTextField().getPreferredSize().height;

        if (maxWidth > 0)
            width = Math.min(width, maxWidth);

        if (nCols > 0)
            width = (fm.stringWidth("0123456789") * nCols) / 10;

        _autoCompletionViewer.getTextField().setPreferredSize(new Dimension(width, height));
        _autoCompletionViewer.getTextField().setSize(new Dimension(width, height));

        if (_relation.getListWidth() > 0)
            _autoCompletionViewer.setPopupWidth(_relation.getListWidth());

        if ((_component.getParent() != null) && (_component.getParent().getParent() != null))
        {
            _component.getParent().getParent().validate();
            _component.getParent().getParent().doLayout();
        }
    }

    /**
     * Retrieves the index of the specified link in this combo.
     *
     * @param   combo  a combo box component. The link is tested against the class combo box, this
     *                 component is not used here.
     * @param   link   the object to retrieve the index
     *
     * @return  the index of link in the combo
     *
     * @throws  IllegalArgumentException  if the specified link is not found in this combo box
     */
    protected int getIndex(LyLink link)
    {
        int i, n = _relation.getLinkCount();

        for (i = 0; i < n; i++)
        {
            if (_relation.getLink(i) == link)
                return i;
        }

        // Not found
        throw new IllegalArgumentException("BF AppEngine : link " + link.getLabel()
            + " doesn't belong to this relation");
    }

    /**
     * Sets the selected flag for the given link in this component.
     *
     * @param  link      the link to set
     * @param  selected  the selected flag
     */
    public void setSelected(LyLink link, boolean selected)
    {
        if (_idle)
            return;

        if (_relation.isMultiple())
        {
            if (link == null)
                setMultiSelectorSelectedObject(_component, null, selected);
            else if (selected)
            {
                _relation.removeLink(link);
                _relation.addLink(link);
            }
            else
                setMultiSelectorSelectedObject(_component, link, false); // XXX
        }
        else
        {
            if (selected)
            {
                // Deselect previous one
                int selectedIdx = getComboSelectedIndex(_component);

                if (selectedIdx >= 0)
                {
                    LyLink selectedLink = _relation.getLink(selectedIdx);
                    // MAZ SITA FFT00041
                    if (link != selectedLink)
                        selectedLink.setSelected(false);
                }

                // Select new one
                int idx = getIndex(link);

                setComboSelectedIndex(_component, idx);
            }
            else
                setComboSelectedIndex(_component, 0);
        }
    } // end method setSelected

    /**
     * (non-Javadoc).
     *
     * @see  leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
     */
    public void actionPerformed(LyActionEvent actionEvent)
    {
        if (actionEvent.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
            updateLinkSelection();

        // may notify any listeners
        if (_actionTransmitter != null)
            _actionTransmitter.processActionEvent(actionEvent);
    }

    /**
     * (non-Javadoc).
     *
     * @see  leon.view.getinterface.LyGetValueInterface#getValue()
     */
    public String getValue()
    {
        if (!_relation.isMultiple()
                && (_relation.getAutoCompletionMax() != 0))
        {
            int index = _autoCompletionViewer.getSelectedIndex();

            if (index < 0)
                return _autoCompletionViewer.getTextField().getText();
        }

        // XXX
        if (_relation.getLinkCount() <= 0)
            return null;

        if (!_relation.isMultiple())
        {
            int idx = getComboSelectedIndex(_component);

            if (idx >= 0)
            {
                LyLink link = _relation.getLink(idx);

                if (link != null)
                    return link.getId();
            }
        }
        else
            return getMultiSelectorValue(_component);

        return null;
    }

    /**
     * Sets current value.
     *
     * @param  value  the new value
     *
     * @see    #getValue()
     */
    public void setValue(String value)
    {
        // XXX
        if (value != null)
        {
            if (!_relation.isMultiple())
            {
                boolean found = false;
                int     i, n = _relation.getLinkCount();
                LyLink  link;
                int     offset = 0;

                for (i = 0; (i < n) && (!found); i++)
                {
                    link = _relation.getLink(i);
                    if (value.equals(link.getId()))
                    {
                        setComboSelectedIndex(_component, i + offset);
                        found = true;
                    }
                }

                // if not found, try to select by displyed value
                for (i = 0; (i < n) && (!found); i++)
                {
                    link = _relation.getLink(i);

                    if (value.equals(link.getLabel()))
                    {
                        setComboSelectedIndex(_component, i + offset);
                        found = true;
                    }
                }
            }
            else
                // XXX
                // reset
                setContent();
        }
    } // end method setValue

    /**
     * Sets the editable status of the field item.
     *
     * @param  editable  new status
     *
     * @see    LySetEditableInterface#setEditable(boolean)
     */
    public void setEditable(boolean editable)
    {
        _editable = editable;

        setComponentState();
    }

    /**
     * Indicates the parent editable status. The component is editable if its editable state is true
     * and if its parent edition status is true.
     *
     * @param  status  new enabled status
     */
    public void setEnabled(boolean status)
    {
        _enabled = status;

        setComponentState();

        // If component must be enabled, update link selection
        updateLinkSelection();
    }

    /**
     * Sets the component state background color depending on its enabled and editable states.
     */
    protected void setComponentState()
    {
        if (getComponent() != null)
        {
            _component.setEnabled(_enabled && _editable);

            Color bgColor;

            if (_enabled && _editable)
                bgColor = getColor("LY_ENABLED_BACKGROUND");
            else
                bgColor = getColor("LY_DISABLED_BACKGROUND");

            _component.setBackground(bgColor);
        }
    }

    /**
     * Updates parent selection status.
     */
    protected void updateLinkSelection()
    {
        LyLink link = null;
        int    i, n = _relation.getLinkCount();

        if (!_relation.isMultiple())
        {
            if (_relation.getAutoCompletionMax() == 0)
            {
                int idx = getComboSelectedIndex(_component);

                _idle = true;
                for (i = 0; i < n; i++)
                {
                    link = _relation.getLink(i);
                    link.setSelected((i == idx));
                }

                _idle = false;
            }
            else
            {
                String selectedLinkId = _autoCompletionViewer.getSelectedValue();

                _idle = true;
                for (i = 0; i < n; i++)
                {
                    link = _relation.getLink(i);
                    link.setSelected(link.getId().equals(selectedLinkId));
                }

                _idle = false;
            }
        }
        else
        {
            // Refresh parent links status
            boolean wasIdle = _idle;

            _idle = true;

            for (i = 0; i < n; i++)
            {
                link = _relation.getLink(i);
                link.setSelected(false);
            }

            String value = getValue();

            if (value != null)
            {
                // Put selected links at the end of the list to keep selection order
                String[] ids = LyString.stringToList(value, ITEM_SEP);
                int      j, ns = ids.length;

                for (j = 0; j < ns; j++)
                {
                    for (i = 0; i < n; i++)
                    {
                        link = _relation.getLink(i);
                        if (link.getId().equals(ids[j]))
                            break;
                    }

                    link.setSelected(true);
                    _relation.removeLink(link);
                    _relation.addLink(link);
                }
            }

            _idle = wasIdle;
        }
    } // end method updateLinkSelection

    /**
     * (non-Javadoc).
     *
     * @see  leon.view.win.swing.LySwingComponent#getComponent()
     */
    public Component getComponent()
    {
        return _component;
    }

    /* Implementation of LySetLinksInterface */

    /**
     * Adds the specified link at the specified position.
     *
     * @param  link  the link to add to this relation
     * @param  pos   the position of the link in this relation component
     *
     * @see    LySetLinksInterface#addLink
     */
    public void addLink(LyLink link, int pos)
    {
        if (!_relation.isMultiple())
        {
            if (_relation.getAutoCompletionMax() == 0)
            {
                String shortLabel = link.getShortLabel();
                String label = link.getLabel();

                // Make sure that the internal view is ready
                link.getInternalView();

                shortLabel = translate(shortLabel);
                label = translate(label);

                if ((label == null) || (label.length() == 0))
                    label = DEFAULT_EMPTY_LINK;

                // Create a structure with both label and id to ensure that two items with the same
                // labels will send ItemEvents when selected.
                _LyComboItem item = new _LyComboItem(shortLabel, label, link.getId());

                if (pos >= 0)
                {
                    int index = getComboSelectedIndex(_component);

                    addItemToComboAt(_component, item, pos);

                    // PATCH : Selection is lost if pos <= selection, we reselect index
                    if (!_idle && (pos <= index))
                        setComboSelectedIndex(_component, index + 1);
                }
                else
                    addItemToComboAt(_component, item, -1);

                if (link.isSelected())
                    setSelected(link, true);
            }
            else
            {
                String label = link.getLabel();
                String shortLabel = link.getShortLabel();

                // make sure that the internal view is ready
                link.getInternalView();

                label = translate(label);
                shortLabel = translate(shortLabel);

                if (shortLabel == null)
                    shortLabel = label;

                _autoCompletionViewer.addValue(link.getId(), label, shortLabel, pos);

                if (link.isSelected())
                {
                    if (pos < 0)
                        pos = _autoCompletionViewer.getDisplayedValues().length - 1;

                    _autoCompletionViewer.setTextValue(pos, true);
                }
            }
        }
        else
        {
            if (_relation.getAutoCompletionMax() == 0)
            {
                // make sure that the internal view is ready
                link.getInternalView();

                addItemToMultiSelectorAt(_component, link, pos);
            }
            else
            {
                // if idle mode, don't update component
                if (_idle)
                    return;

                String label = link.getLabel();
                String shortLabel = link.getShortLabel();

                // make sure that the internal view is ready
                link.getInternalView();

                label = translate(label);
                shortLabel = translate(shortLabel);

                if (shortLabel == null)
                    shortLabel = label;

                _autoCompletionViewer.addValue(link.getId(), label, shortLabel, pos);

                if (link.isSelected())
                {
                    Vector<String> selectedValues = new Vector<String>(1);

                    selectedValues.addElement(link.getId());
                    ((LyJAutoCompletionMultiSelector)_component).selectValues(selectedValues);
                }
            }
        }
    } // end method addLink

    /**
     * Removes the specified link of the relation.
     *
     * @see  LySetLinksInterface#removeLink
     */
    public void removeLink(LyLink link)
    {
        if (!_relation.isMultiple())
        {
            int idx = getIndex(link);

            if (_relation.getAutoCompletionMax() == 0)
            {
                removeItemFromComboAt(_component, idx);

                // if (getComboItemCount(_component) == 0)
                // addLink(null, -1); // Avoid empty combo
            }
            else
                _autoCompletionViewer.removeValueAt(idx);
        }
        else
            setContent(link, false);
    }

    /**
     * Internal class used to represent an item in a combo.
     */
    class _LyComboItem
    {
        /** The short label of the item. */
        String _shortLabel;

        /** The label of the item. */
        String _label;

        /** The identifier of the item. */
        String _id;

        /**
         * Constructor with both label and identifier.
         *
         * @param  label  the label of the item.
         * @param  id     the identifier of the item.
         */
        _LyComboItem(String shortLabel, String label, String id)
        {
            _shortLabel = shortLabel;
            _label = label;
            _id = id;
        }

        /*
         * (non-Javadoc)
         * @see java.lang.Object#toString()
         */
        public String toString()
        {
            return _label;
        }
    }

    class _LyComboBoxUI extends MetalComboBoxUI
    {
        int _listWidth;

        public _LyComboBoxUI(int listWidth)
        {
            _listWidth = listWidth;
        }

        protected ComboPopup createPopup()
        {
            return new _LyComboPopup(comboBox, _listWidth);
        }
    }

    class _LyComboPopup extends BasicComboPopup
    {
        /**  */
        private static final long serialVersionUID = 457695724143742935L;
        int                       _listWidth;

        _LyComboPopup(JComboBox c, int listWidth)
        {
            super(c);
            _listWidth = listWidth;
        }

        protected void configurePopup()
        {
            setLayout(new BorderLayout(0, 0));
            setBorderPainted(true);
            setBorder(new LineBorder(Color.BLACK, 1));
            setOpaque(false);
            add(scroller);
            setDoubleBuffered(true);
            setFocusable(false);
        }

        public Dimension getPreferredSize()
        {
            return new Dimension(_listWidth, super.getPreferredSize().height);
        }
    }

    class _LyComboBoxRenderer extends BasicComboBoxRenderer
    {
        /**  */
        private static final long serialVersionUID = -1727548802556754372L;
        protected int             _width;

        public _LyComboBoxRenderer()
        {
            super();
            _width = -1;
        }

        public void setWidth(int width)
        {
            _width = width;
        }

        public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected,
            boolean cellHasFocus)
        {
            super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);

            _LyComboItem item = ((value instanceof _LyComboItem) ? (_LyComboItem)value : null);
            FontMetrics  fm = list.getFontMetrics(list.getFont());

            if ((_width > 0)
                    && (item != null)
                    && (item._label != null)
                    && (item._label.trim().length() > 0)
                    && (fm.stringWidth(item._label) > (_width - 18)))
            {
                if (index >= 0)
                {
                    LyJMultiLineLabel l = new LyJMultiLineLabel(item._label, -1, 1, 1);

                    l.setMaxWidth(_width - 18);
                    l.setBackground(getBackground());
                    l.setForeground(getForeground());
                    l.setFont(list.getFont());

                    return l;
                }
            }

            if ((index < 0) && (item != null))
                setText(((item._shortLabel != null) ? item._shortLabel : item._label));

            return this;
        }
    }
} // end class LySwingRelation
