/*
 * LySwingRow.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.util.*;
import leon.view.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Row object (as described in the dtd) for swing viewer.
 *
 * <p>A row of a list is a set of non-empty cells (or input fields). It is possible to indicate if
 * the row is selected, if an image is assigned to it and if a tooltip is associated to it.</p>
 *
 * @version $Id: LySwingRow.java 35311 2012-04-20 09:36:57Z jlsaouli $
 * @author  Lyria
 * @see     LyRow
 */
public class LySwingRow extends LySwingObject
	implements LyGetIdInterface,
		LySetColorInterface,
		LySetImageInterface,
		LyGraphicalRow
{
	public static final String __VERSION =
		"$Id: LySwingRow.java 35311 2012-04-20 09:36:57Z jlsaouli $";

	/** The view independent row description. */
	protected LyRow			   _row;

	/** The image displayed by this row. */
	protected Image			   _image;

	/** The color of the row. */
	protected Color			   _color;

	/** Image for an open row. */
	protected Image			   _openRowImage;

	/** Image for a closed row. */
	protected Image			   _closedRowImage;

	/** The cell list contained by this row. */
	protected Vector<Object>   _cells;
	
	/** Indicates if a reset of the row may be performed. */
	protected boolean		   _resetRowComponents;

	/**
	 * Class constructor of a Row object for the Swing target.
	 *
	 * @param row a ROW component defined for every type of view targets. This row is initialized
	 *            for the swing view and filled with its default values.
	 */
	public LySwingRow(LyRow row)
	{
		super(row);
		_row = row;
		_resetRowComponents = true;

		initialize();
		setContent();
	}

	/**
	 * Sets the initial image and color parameters for this row.
	 */
	public void initialize()
	{
		setImage(_row.getImage(), false);
		setColor(_row.getColor(), false);
	}

	/**
	 * Sets the content of this row. By default, a row is empty and this method doesn't do anything.
	 */
	public void setContent()
	{
	}

	/**
	 * Gets the row color.
	 *
	 * @return the color of the row
	 * @see    #setColor(String)
	 */
	public Color getColor()
	{
		return _color;
	}

	/**
	 * Gets the row image.
	 *
	 * @return the image displayed by this row
	 * @see    #setImage(String)
	 */
	public Image getImage()
	{
		return _image;
	}

	/**
	 * Sets the image displayed by this row. The row is updated after the image is set.
	 *
	 * @param image a string that identifies the image (a key or a path)
	 * @see   #getImage()
	 * @see   #setImage(String, boolean)
	 * @see   leon.view.setinterface.LySetImageInterface#setImage(java.lang.String)
	 */
	public void setImage(String image)
	{
		setImage(image, true);
	}

	/**
	 * Sets the image displayed by this row.
	 *
	 * @param image  a string that identifies the image (a key or a path)
	 * @param update Indicates if the row must be updated (by calling the updateRow method).
	 * @see   #getImage()
	 */
	protected void setImage(String image, boolean update)
	{
		if (image == null)
			return;

		Image img = getImage(image);

		if (img != null)
			_image = img;

		if (update)
			updateRow();
	}

	/**
	 * Updates the label for current row.
	 *
	 * @param label the label for this row
	 * @see   leon.view.graphic.LyGraphicalRow#setTooltip(java.lang.String)
	 */
	public void setLabel(String label)
	{
		updateRow();
	}
	
	/**
	 * @see leon.view.setinterface.LyGraphicalRow#setIndex(java.lang.String)
	 */
	public void setIndex(String index)
	{
		// Nothing to do
	}
	
	/**
	 * The tooltip is already set when this method is called and only the swing component of the row
	 * is updated.
	 *
	 * @param tooltip the tooltip for this row, not used in this implementation
	 * @see   leon.view.graphic.LyGraphicalRow#setTooltip(java.lang.String)
	 */
	public void setTooltip(String tooltip)
	{
		updateRow();
	}
	
	/**
	 * Indicates whether this row is sensitive or not.
	 */
	public boolean isSensitive()
	{
		return _row.isSensitive();
	}

	/**
	 * Updates the current row in the table component by firing a rows update event.
	 */
	protected void updateRow()
	{
		LyTable table = getParentTable();

		if (table != null)
		{
			LySwingTable	  swingTable = (LySwingTable)table.getInternalView();
			LyJTableComponent comp = swingTable.getTableComponent();

			if (comp != null)
			{
				LyRows rows = table.getRows();
				int    idx = rows.getRowIndex(_row);

				if (idx >= 0)
					comp.getTableModel().fireTableRowsUpdated(idx, idx);
			}
		}
	}

	/**
	 * Gets the translated tooltip of this row.
	 *
	 * @return the tooltip of this row
	 */
	public String getToolTip()
	{
		return translate(_row.getTooltip());
	}

	/**
	 * Sets the color of this row.
	 *
	 * @param color a string that identifies the color (a key or a path)
	 * @see   #getColor()
	 * @see   LySetColorInterface#setColor(String)
	 */
	public void setColor(String color)
	{
		setColor(color, true);
	}
    
	/**
	 * Sets the color of this row.
	 *
	 * @param color a string that identifies the color (a key or a path)
	 * @param update indicates if the table must be updated (by calling the updateTable method)
	 * @see   #getColor()
	 * @see   LySetColorInterface#setColor(String)
	 */
	public void setColor(String color, boolean update)
	{
		Color col = getColor(color);

		_color = col;
		updateTable();
	}

	/**
	 * The open property is already set when this method is called and only the swing table is
	 * updated.
	 *
	 * @param open the open property, not used in this implementation
	 * @see   LyGraphicalRow#setOpen(boolean)
	 */
	public void setOpen(boolean open)
	{
		LyTable table = getParentTable();

		if (table != null)
		{
			LySwingTable	  swingTable = (LySwingTable)table.getInternalView();

			swingTable.checkComputeHeight();
			swingTable.computeRowHeights();
			swingTable.cacheHeight();

			if (swingTable._statusColumn != null)
				swingTable._statusColumn.updateContent();

			updateTable();
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetHiddenInterface#setHidden(boolean)
	 */
	public void setHidden(boolean hidden)
	{
		LyRows			    rows = (LyRows)_row.getParent();
		LySetTableInterface table = rows.getTableInterface();

		table.hideRow(-1, hidden);
	}
	
	/**
	 * (non-Javadoc)
	 *
	 * @see leon.view.graphic.LyGraphicalRow#setBorderHeight(int)
	 */
	public void setBorderHeight(int height)
	{
	}
	
	/**
	 * (non-Javadoc)
	 *
	 * @see leon.view.graphic.LyGraphicalRow#setBorderColor(String)
	 */
	public void setBorderColor(String color)
	{
	}

	/**
	 * Inserts a sub-component component at given position.
	 * 
	 * @param component The component to add to components list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void addComponent(LyComponent component, int pos)
	{
		resetRowComponents(component);
		
		LyTable		table = getParentTable();

		if (component instanceof LyFormField)
		{
			LyFormField			formField = (LyFormField)component;
			LyFieldItem			fieldItem = formField.getFieldItem();
			LySwingComponent	swingComponent = (LySwingComponent)fieldItem.getInternalView();
			_LyFocusRequest		_focusRequest = new _LyFocusRequest(swingComponent, pos);
            LyViewManager		viewManager = getViewManager();
            
			viewManager.execGui(_focusRequest);

			if (table != null)
			{
				LySwingTable	swingTable = (LySwingTable) table.getInternalView();
				int				rowIndex = swingTable.getIndex(_row);
			
				swingTable._needResizeColumns = false;
				swingTable.ensureColumnsWidth(rowIndex);
				swingTable.getTableComponent().scrollToRow(rowIndex);

				int		i, n = _row.getComponentCount();
				
				for (i = 0; i < n; i++)
				{
					if (_row.getComponent(i) == formField)
					{
						swingTable.getTableComponent().scrollToColumn(i);
						break;
					}
				}
			}
		}
		else
		{
			if (table != null)
			{
				LySwingTable	swingTable = (LySwingTable)table.getInternalView();
				int				rowIndex = swingTable.getIndex(_row);
				
				swingTable.getTableComponent().editCellAt(rowIndex, pos);
			}
		}
	}

	/**
	 * Removes a sub-component component.
	 * 
	 * @param component The component to remove to components list.
	 */
	public void removeComponent(LyComponent component)
	{
		resetRowComponents(component);
	}
	
	/**
	 * Performs a reset of the row on the Swing component.
	 * 
	 * @param removedComponent component that is being removed from the table. 
	 */
	protected void resetRowComponents(LyComponent removedComponent)
	{
		if (!_resetRowComponents)
			return;
		
		LyTable		table = getParentTable();

		if (table != null)
		{
			LySwingTable	swingTable = (LySwingTable) table.getInternalView();
			int				rowIndex = swingTable.getIndex(_row);
			boolean			wasIdle = swingTable.isIdle();
			
			// Rebuilt the components of the row to ensure the new components
			// are taken into account by the swing table.
			getCells(true);
			swingTable.setIdle(true, removedComponent);
			swingTable.removeRow(_row);
			swingTable.addRow(_row, rowIndex);
			swingTable.setIdle(wasIdle, removedComponent);
		}
	}
	
	/**
	 * Indicates if a reset of the row may be performed.
	 * 
	 * @param reset true if a reset may be performed, false otherwise.
	 */
	public void setResetRowComponents(boolean reset)
	{
		_resetRowComponents = reset;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see leon.view.getinterface.LyGetIdInterface#getId()
	 */
	public String getId()
	{
		return _row.getId();
	}

	/**
	 * Gets the selected state of this row.
	 *
	 * @return true if the row is selected, false otherwise
	 */
	public boolean isSelected()
	{
		return _row.isSelected();
	}

	/**
	 * Gets the cell element at the specified index.
	 *
	 * @param  i the position of the cell in this row
	 * @return the cell at position i. Builds the vector cells if needed. The content is usually a
	 *         LySwingCell but might be a LySwingFormField for editable tables.
	 */
	public Object getCell(int i)
	{
		if (_cells == null)
			getCells();

		return (_cells.get(i));
	}

	/**
	 * Gets the vector of the LySwingCell objects. Builds the vector if needed.
	 *
	 * @return the list of cells of this row
	 */
	public Vector<Object> getCells()
	{
		return getCells(false);
	}

	/**
	 * Gets the vector of the LySwingCell objects. Builds the vector if needed.
	 * @param reset indicates if the list of cells must be rebuilt.
	 * @return the list of cells of this row
	 */
	public Vector<Object> getCells(boolean reset)
	{
		if ((_cells == null) || reset)
		{
			_cells = new Vector<Object>();

			int i, n = _row.getComponentCount();

			for (i = 0; i < n; i++)
				_cells.add(_row.getComponent(i).getInternalView());
		}
		return _cells;
	}

	/**
	 * Returns the depth of the row in the table when hierarchical.
	 *
	 * @return the depth of the row
	 */
	public int getDepth()
	{
		int		    depth = 0;
		LyComponent parent = _row.getParent();

		while ((parent != null) && !(parent instanceof LyTable))
		{
			parent = parent.getParent();

			if (parent instanceof LyRow)
				depth++;
		}

		return depth;
	}

	/**
	 * Returns the image for an open row.
	 *
	 * @return the open row image
	 */
	public Image getOpenRowImage()
	{
		if (_openRowImage == null)
			_openRowImage = getImage("LY_OPEN_ROW");

		return _openRowImage;
	}

	/**
	 * Returns the image for a closed row.
	 *
	 * @return the closed row image
	 */
	public Image getClosedRowImage()
	{
		if (_closedRowImage == null)
			_closedRowImage = getImage("LY_CLOSED_ROW");

		return _closedRowImage;
	}

	/**
	 * Returns the parent table component.
	 *
	 * @return the parent table component
	 */
	public LyTable getParentTable()
	{
		if (_row == null)
			return null;

		LyRows rows;

		rows = (LyRows)_row.getParent();

		return ((rows == null) ? null : rows.getParentTable());
	}

	/**
	 * Updates the table component.
	 */
	protected void updateTable()
	{
		LyTable table = getParentTable();

		if (table != null)
		{
			LySwingTable	  swingTable = (LySwingTable)table.getInternalView();
			LyJTableComponent comp = swingTable.getTableComponent();

			if (comp != null)
				comp.repaint();
		}
	}
    
    class _LyFocusRequest
        implements Runnable
    {
        /** Current Control instance. */
        LySwingComponent        _focusControl;
        
        /** Index of the column of focusable component. */
        int                     _columnIndex;
        
        public _LyFocusRequest(LySwingComponent focusControl, int columnIndex)
        {
            _focusControl = focusControl;
            _columnIndex = columnIndex;   
        }
        
        public void run()
        {
            Component   component = _focusControl.getComponent();
            
            if (component == null)
                return;

            LyTable     table = getParentTable();

            if (table != null)
            {
                LySwingTable    swingTable = (LySwingTable) table.getInternalView();
                int             rowIndex = swingTable.getIndex(_row);
                
                swingTable.getTableComponent().editCellAt(rowIndex, _columnIndex);
            }
            
            component.requestFocus();
        }
    }
} // end class LySwingRow
