/*
 * LySwingSelectBar.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;

import javax.swing.*;

import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the selection bar object (as described in the dtd) for swing viewer.
 *
 * <p>A navigation bar allows to display a set of page numbers (or page names) in order to go
 * directly to the selected page. This bar must be associated to an area presenting paginated data
 * (a long list for example).</p>
 *
 * @version $Id: LySwingSelectBar.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     LySelectBar
 */
public class LySwingSelectBar extends LySwingPanel
	implements LySwingComponent,
		LyGraphicalSelectBar,
		LyActionListener,
		LySetIdleInterface,
		LyIsIdleInterface
{
	public static final String   __VERSION =
		"$Id: LySwingSelectBar.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The view independent select bar description. */
	protected LySelectBar		 _selectBar;

	/** The panel that contains the selectable items. */
	protected Container			 _panel;

	/** Idle mode. */
	protected boolean			 _idle;

	/** Title of the select bar. */
	protected LyJHRefComponent   _title;

	/** List of navigation tools for this select bar. */
	protected LyJHRefComponent[] _nav_tool;

	/** List of selected items in this select bar. */
	protected LyJHRefComponent[] _select_item;

	/**
	 * Class constructor of a Select bar object for the Swing target.
	 *
	 * @param selectBar a SELECTBAR component defined for every type of view targets. This select
	 *                  bar is initialized specificaly for the Swing view.
	 */
	public LySwingSelectBar(LySelectBar selectBar)
	{
		super(selectBar);
		_selectBar = selectBar;

		initialize();
		setContent();
	}

	/**
	 * Initializes the graphic component. Creates the associated panel for the selectable items, and
	 * sets its background color to the default select bar color.
	 */
	public void initialize()
	{
		_idle = false;

		_panel = createPanel();

		Color col = getColor("LY_SELECTBAR_BACKGROUND");

		if (col == null)
			col = getColor("LY_DEFAULT_BACKGROUND");

		if (col != null)
			_panel.setBackground(col);
	}

	/**
	 * Fills the panel with selectable items and enables or disables the navigation tools depending
	 * on the elemnts.
	 */
	public void setContent()
	{
		int			 i, n = _selectBar.getComponentCount();
		// Retreive selected index
		int			 selectedIndex = -1;
		LySelectItem item;
		String		 label, tooltip;

		for (i = 0; (i < n) && (selectedIndex < 0); i++)
		{
			item = (LySelectItem)_selectBar.getComponent(i);
			
			if ((item != null) && item.isSelected())
				selectedIndex = i; // Exits loop we have the answer
		}

		LyEnvironment environment = getEnvironment();

		label = environment.getMessage("LY_PAGE_INDEX", Integer.toString(selectedIndex + 1));

		setSelectBarTitle(label);

		// Now refresh the list of items
		int start = _selectBar.getStartItem();
		int last = Math.min(_selectBar.getComponentCount(), start + _selectBar.getItemCount());

		n = _selectBar.getItemCount();

		if (start > 0)
		{
			enableNavTool(0, true);
			enableNavTool(1, true);
		}
		else
		{
			enableNavTool(0, false);
			enableNavTool(1, false);
		}

		/* refresh the selectable items */
		for (i = 0; i < n; i++)
		{
			if ((start + i) < last)
			{
				item = (LySelectItem)_selectBar.getComponent(start + i);
				label = (item == null ? null : translate(item.getLabel()));
				tooltip = (item == null ? null : translate(item.getTooltip()));
				updateSelectItem(i, label, tooltip, (item == null ? null : item.getId()),
						(item == null ? false : item.isSelected()), true,
						(item == null) || (item.getId() == LyComponent.SEP_ID));
			}
			else
				// out of range : disable select items
				updateSelectItem(i, null, null, null, false, false, false);
		}

		// check next/last tools state
		if (last < _selectBar.getComponentCount())
		{
			enableNavTool(4, true);
			enableNavTool(5, true);
		}
		else
		{
			enableNavTool(4, false);
			enableNavTool(5, false);
		}

		// check show next/previous tools state
		enableNavTool(2, (_selectBar.getComponentCount() > 1));
		enableNavTool(3, (_selectBar.getComponentCount() > 1));

		_panel.invalidate();
		_panel.validate();
	} // end method setContent

	/**
	 * Creates the panel used to place all the messages. Creates the associated navigation tools and
	 * adds them at the end of the panel.
	 *
	 * @return the new panel for the messages
	 */
	public Container createPanel()
	{
		LayoutManager	rootLayout = new GridLayout(1, 1, 0, 0);
		LyJBorderPanel	rootPanel = new LyJBorderPanel(LyArea.Border.IN, 1);
		JPanel			p = new JPanel();
		GridBagLayout	layout = new GridBagLayout();
		
		rootPanel.add(p);
		rootPanel.setLayout(rootLayout);
		rootPanel.setThickness(1);
		rootPanel.setGap(1);

		p.setLayout(layout);
		// Selectbar may be reduced to a minimum size
		p.setMinimumSize(new Dimension(10, 10));

		Font font = getFont("LY_SELECT_BAR");
		Font titleFont = getFont("LY_SELECT_BAR_TITLE");

		// *** Add the label title
		_title = createHRefComponent(null, null, titleFont, false, false, "title", Color.black,
				Color.black, Color.black, Color.black);

		p.add(_title);

		// *** Now add the first/previous navigation tools
		Image			    image1, image2, image3;
		LyActionTransmitter transmitter;

		_nav_tool = new LyJHRefComponent[6];

		image1 = getImage("LY_PAGE_FIRST");
		image2 = getImage("LY_PAGE_PREVIOUS");
		image3 = getImage("LY_PAGE_SHOW_PREVIOUS");

		_nav_tool[0] = createHRefComponent(null, image1, font, false, false, "_select_page_first",
				null, null, null, null);

		p.add(_nav_tool[0]);

		// Register events
		transmitter = _nav_tool[0].getActionTransmitter();

		if (transmitter != null)
			transmitter.addActionListener(this);

		_nav_tool[1] = createHRefComponent(null, image2, font, false, false,
				"_select_page_previous", null, null, null, null);

		p.add(_nav_tool[1]);

		// Register events
		transmitter = _nav_tool[1].getActionTransmitter();

		if (transmitter != null)
			transmitter.addActionListener(this);

		_nav_tool[2] = createHRefComponent(null, image3, font, false, false,
				"_select_page_showprevious", null, null, null, null);

		p.add(_nav_tool[2]);

		// Register events
		transmitter = _nav_tool[2].getActionTransmitter();

		if (transmitter != null)
			transmitter.addActionListener(this);

		// *** Now add the selectable items
		int start = _selectBar.getStartItem();
		int n = _selectBar.getItemCount();

		_select_item = new LyJHRefComponent[n];

		Color disableColor = getColor("LY_HREF_DISABLE");
		Color hrefColor = getColor("LY_HREF_COLOR");
		Color highlightColor = getColor("LY_HREF_HIGHLIGHT");
		Color visitedColor = getColor("LY_HREF_VISITED");

		for (int i = start; i < (start + n); i++)
		{
			_select_item[i] = createHRefComponent(null, null, font, false, true, null, disableColor,
					hrefColor, highlightColor, visitedColor);

			p.add(_select_item[i]);

			transmitter = _select_item[i].getActionTransmitter();

			if (transmitter != null)
				transmitter.addActionListener(this);
		}

		// *** Now add the next/last navigation tools
		image1 = getImage("LY_PAGE_SHOW_NEXT");
		image2 = getImage("LY_PAGE_NEXT");
		image3 = getImage("LY_PAGE_LAST");

		_nav_tool[3] = createHRefComponent(null, image1, font, false, false,
				"_select_page_shownext", null, null, null, null);
		transmitter = _nav_tool[3].getActionTransmitter();

		p.add(_nav_tool[3]);

		// Register events
		transmitter = _nav_tool[3].getActionTransmitter();

		if (transmitter != null)
			transmitter.addActionListener(this);

		_nav_tool[4] = createHRefComponent(null, image2, font, false, false, "_select_page_next",
				null, null, null, null);
		transmitter = _nav_tool[4].getActionTransmitter();

		p.add(_nav_tool[4]);

		// Register events
		transmitter = _nav_tool[4].getActionTransmitter();

		if (transmitter != null)
			transmitter.addActionListener(this);

		_nav_tool[5] = createHRefComponent(null, image3, font, false, false, "_select_page_last",
				null, null, null, null);
		transmitter = _nav_tool[5].getActionTransmitter();

		p.add(_nav_tool[5]);

		// Register events
		transmitter = _nav_tool[5].getActionTransmitter();

		if (transmitter != null)
			transmitter.addActionListener(this);
		
		// Retrieve form field if set
		LyFormField		formField = _selectBar.getFormField();
		
		if (formField != null)
		{
			LySwingFormField	swingField = (LySwingFormField)formField.getInternalView();
			JPanel				panel = new JPanel(new BorderLayout());
			
			panel.setBorder(BorderFactory.createEmptyBorder(0, 0, 0, 3));
			panel.setOpaque(false);
			
			if (swingField.getCaptionComponent() != null)
				panel.add(swingField.getCaptionComponent(), BorderLayout.WEST);
			
			panel.add(swingField.getFullComponent(), BorderLayout.CENTER);
			
			JLabel	separator = new JLabel("  ");
			
			separator.setOpaque(false);
			p.add(separator);
			p.add(panel);
			
			// Register events on parent swing area
			LyArea				area = (LyArea)_selectBar.getParent();
			LySwingArea			swingArea = (LySwingArea)area.getInternalView();
			
			transmitter = ((LyEventTransmitter)swingField).getActionTransmitter();
			transmitter.addActionListener(swingArea);
		}
		
		return rootPanel;
	} // end method createPanel

	/**
	 * Update the selection bar title.
	 *
	 * @param title title to set
	 * @see   LyJHRefComponent#setLabel(String)
	 */
	public void setSelectBarTitle(String title)
	{
		_title.setLabel(title);
	}

	/**
	 * Enables/disables the specified nav tool.
	 *
	 * @param idx       index of the nav tool in the view
	 * @param sensitive indicates whether the component is activated or not
	 */
	public void enableNavTool(int idx, boolean sensitive)
	{
		_nav_tool[idx].setSensitive(sensitive);
		_nav_tool[idx].setIdle(!sensitive);
		_nav_tool[idx].repaint();
	}

	/**
	 * Updates the selectable item content.
	 *
	 * @param idx         index of the select item in the view
	 * @param label       title of the selectable item
	 * @param tooltip     tooltip of the selectable item. This attribute is not used in this
	 *                    implementation.
	 * @param id          id of the selectable item
	 * @param selected    indicates whether the component is selected or not
	 * @param visible     indicates whether the component is activated or not
	 * @param isSeparator indicates whether the component is a separator or not
	 */
	public void updateSelectItem(int idx, String label, String tooltip, String id,
		boolean selected, boolean visible, boolean isSeparator)
	{
		synchronized(_select_item[idx].getTreeLock())
		{
			_select_item[idx].setUserData(id);
			// FFT 813 : descativate clic on unactive fields
			_select_item[idx].setSensitive((!selected) && (label != null));
			_select_item[idx].setIdle(!visible);
			// FF812 : center items
			// _select_item[idx].setLabel((label == null ? " " : label));
	
			_select_item[idx].setLabel(label);
			
			if (isSeparator)
			{
				Image	image = getImage("LY_VERTICAL_SEPARATOR");
				
				_select_item[idx].setImage(image);
			}
			
			if (visible)
				_select_item[idx].setBorder(BorderFactory.createEmptyBorder(0, 5, 0, 5));
			else
				_select_item[idx].setBorder(BorderFactory.createEmptyBorder(0, 0, 0, 0));
	
			/*
			 * if (label == null) { // _select_item[idx].setXGap(0); } else { //
			 * _select_item[idx].setXGap(2); }
			 */
			_select_item[idx].invalidate();
		}
	}

	/**
	 * Creates and configures a Hyper Text Reference component with the specified parameters.
	 *
	 * @param  label       the text displayed by this reference
	 * @param  image       the image displayed by this reference
	 * @param  font        the font used to display the label
	 * @param  isVisited   specifies the visited property of a link
	 * @param  isSensitive specifies the sensitive property of a link
	 * @param  clientData  the underlying data used to properly configure this reference
	 * @param  disable     the disable color
	 * @param  color       the normal color
	 * @param  highlight   the color used to display text when cursor is over the reference
	 * @param  visited     the visited color
	 * @return a new HRef component for the specified references
	 */
	public LyJHRefComponent createHRefComponent(String label, Image image, Font font,
		boolean isVisited, boolean isSensitive, Object clientData, Color disable, Color color,
		Color highlight, Color visited)
	{
		LyJHRefComponent comp;

		comp = new LyJHRefComponent(label, image, font, isVisited, isSensitive, clientData);

		comp.setDisableColor(disable);
		comp.setDefaultColor(color);
		comp.setHighlightColor(highlight);
		comp.setVisitedColor(visited);

		return comp;
	}

	/**
	 * Retreives the source identifier of the object.
	 *
	 * @param  source the source object
	 * @return source identifier
	 */
	public String getSourceId(Object source)
	{
		// FFT 813 : desactiver le clic sur les champs inactifs
		if ((source instanceof LyJHRefComponent)
				&& (((LyJHRefComponent)source).getUserData() != null))
			return ((LyJHRefComponent)source).getUserData().toString();
		return source.toString();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		// Notify controller if any
		if (_selectBar.getController() != null)
		{
			// Retreive source id
			String		   id = getSourceId(actionEvent.getSource());

			// Notify controller
			LyActionDemand demand = new LyActionDemand(LyActionDemand.TYPE_ACTION, id);

			processDemand(demand);
			actionEvent.consume();
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetIdleInterface#setIdle(boolean)
	 */
	public void setIdle(boolean status)
	{
		// Invoked when the view has been refreshed by the controller
		// Must reset if false
		if ((_idle) && (!status))
			setContent();

		_idle = status;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.getinterface.LyIsIdleInterface#isIdle()
	 */
	public boolean isIdle()
	{
		return _idle;
	}

	/**
	 * Nothing is implemented yet.
	 *
	 * @param label the new value of label, not used
	 * @see   LyGraphicalSelectBar#setLabel(String)
	 */
	public void setLabel(String label)
	{
		// XXX Diff win/web, to implement
	}

	/**
	 * Nothing is implemented yet.
	 *
	 * @param startItem the new value of startItem, not used
	 * @see   LyGraphicalSelectBar#setStartItem(int)
	 */
	public void setStartItem(int startItem)
	{
		// XXX Diff win/web, to implement
	}

	/**
	 * Nothing is implemented yet.
	 *
	 * @param component the component to add, not used
	 * @param pos       the position where the component is added, not used
	 * @see   LyGraphicalSelectBar#addComponent(LyComponent, int)
	 */
	public void addComponent(LyComponent component, int pos)
	{
		// XXX Diff win/web, to implement
	}

	/**
	 * Nothing is implemented yet.
	 *
	 * @param component the component to remove, not used
	 * @see   LyGraphicalSelectBar#removeComponent(LyComponent)
	 */
	public void removeComponent(LyComponent component)
	{
		// XXX Diff win/web, to implement
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _panel;
	}

} // end class LySwingSelectBar
