/*
 * LySwingSplitter.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;

import leon.misc.*;
import leon.view.*;
import leon.view.LySplitter.ActivePart;
import leon.view.graphic.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Splitter object (as described in the dtd) for swing viewer.
 *
 * <p>A container of panel type with splitter allows to display different panels that are splitted
 * by vertical or horizontal separation bars. These bars can be moved so as to allocate more or less
 * place to each of the panels. The minimum number of panels that is required is two panels. There
 * is no constraint on the maximum number of panels allowed in such a container.</p>
 *
 * <p>Including such containers in containers of same type allows to get windows combining vertical
 * and horizontal separators.</p>
 *
 * @version $Id: LySwingSplitter.java 35434 2012-05-02 13:16:51Z bangiole $
 * @author  Lyria
 * @see     LySplitter
 */
public class LySwingSplitter extends LySwingContainer
	implements LySwingComponent,
		LyGraphicalSplitter
{
	public static final String __VERSION =
		"$Id: LySwingSplitter.java 35434 2012-05-02 13:16:51Z bangiole $";

	/** The view independent splitter description. */
	protected LySplitter	   _splitter;

	/** Indicates if a border layout is used to lay the areas. */
	protected boolean		   _useBorderLayout;

	/** The component that is displayed. */
	protected Container		   _panel;

	/** Minimum dimension of the left panel. */
	protected Dimension		   _leftMinimumSize;

	/** Minimum dimension of the right panel. */
	protected Dimension		   _rightMinimumSize;

	/**
	 * The orientation of the component: {@link JSplitPane#VERTICAL_SPLIT} or {@link
	 * JSplitPane#HORIZONTAL_SPLIT}.
	 */
	protected int			   _orientation;

	/**
	 * Class constructor of a Splitter object for the Swing target.
	 *
	 * @param splitter a SPLITTER component defined for every type of view targets. This splitter is
	 *                 initialized specificaly for the Swing view.
	 */
	public LySwingSplitter(LySplitter splitter)
	{
		super(splitter);
		_splitter = splitter;

		initialize();
		setContent();
	}

	/**
	 * Retrieves the initial orientation of this splitter and initializes the class attributes.
	 */
	public void initialize()
	{
		if (_splitter.getStyle() == LySplitter.Style.VERTICAL)
			_orientation = JSplitPane.VERTICAL_SPLIT;
		else
			_orientation = JSplitPane.HORIZONTAL_SPLIT;

		_useBorderLayout = false;

		// If only two areas and one of these has a negative size
		// we use a border layout instead of a splitter one.
		int		 i, n = _splitter.getPanelCount();
		String   sizeInfo;
		String   sizes = _splitter.getSizes();
		String[] sizesArray = LyString.stringToList(sizes, ' ');
		int		 l = ((sizesArray == null) ? 0 : sizesArray.length);

		if (n == 2)
		{
			for (i = 0; i < n; i++)
			{
				if (i < l)
				{
					sizeInfo = sizesArray[i];

					if (sizeInfo.indexOf("px") > 0)
						sizeInfo = "-1";

					if (LyTools.intFromString(sizeInfo) < 0)
					{
						_useBorderLayout = true;
						break;
					}
				}
			}
		}

		if (_useBorderLayout)
			_panel = new JPanel(new BorderLayout(0, 0));
	} // end method initialize

	/**
	 * Fills this splitter with all its panels.
	 */
	public void setContent()
	{
		if (!_useBorderLayout)
		{
			int		    i, n = _splitter.getPanelCount();
			Component[] components = new Component[n];

			// Set the minimum size of the components to allow the LySplitPane
			// to reduce any component to a non visible size
			for (i = 0; i < n; i++)
			{
				components[i] = ((LySwingComponent)_splitter.getPanel(i).getInternalView())
						.getComponent();
				((JComponent)components[i]).setMinimumSize(new Dimension(0, 0));
			}

			// Create the Splitter component with all the componenets
			_panel = new LyJSplitterComponent(_orientation, components,
					getIntSizesArray(_splitter.getSizes(), n));

			if (!_splitter.isResizable())
				((LyJSplitterComponent)_panel).setResizable(false);
			else if (getEnvironment().testResource("LY_CHECK_WINDOW_MINIMUM_SIZE"))
				_panel.addComponentListener(new _LyJSplitterComponentAdapter());
		}
		else
		{
			int		 i, n = _splitter.getPanelCount();
			String   sizes = _splitter.getSizes();
			String[] sizesArray = LyString.stringToList(sizes, ' ');
			int		 ns = ((sizesArray == null) ? 0 : sizesArray.length);
			String   sizeInfo;
			boolean	 borderPanelAdded = false;

			for (i = 0; i < n; i++)
			{
				LyPanel			 item = _splitter.getPanel(i);
				LySwingComponent internal = (LySwingComponent)item.getInternalView();

				if (i < ns)
				{
					if (LyTools.intFromString(sizesArray[i]) > 0)
						sizeInfo = sizesArray[i];
					else
					{
						if (borderPanelAdded)
							sizeInfo = "10";
						else
						{
							sizeInfo = sizesArray[i];
							borderPanelAdded = true;
						}
					}
				}
				else
					sizeInfo = "10";

				addAreaBorderLayout(internal.getComponent(), sizeInfo);
			}
		}
	} // end method setContent

	/**
	 * Transforms a string representing a list of int separated with the space character in an array
	 * of int.
	 *
	 * @param  sizes      a string that contains a list of int
	 * @param  panelCount the size of the returned array
	 * @return the array of int representing the specified <code>String</code>
	 */
	public int[] getIntSizesArray(String sizes, int panelCount)
	{
		String[] sizes_array = LyString.stringToList(sizes, ' ');
		int[]    result = new int[panelCount];
		int		 i;

		if (sizes_array != null)
		{
			for (i = 0; i < Math.min(panelCount, sizes_array.length); i++)
				result[i] = Math.max(0, LyTools.intFromString(sizes_array[i]));

			// completion of the array with the value 10
			if (sizes_array.length < panelCount)
			{
				for (i = sizes_array.length; i < panelCount; i++)
					result[i] = 10;
			}
		}
		// completion of the array with the value 10
		else
		{
			for (i = 0; i < panelCount; i++)
				result[i] = 10;
		}

		return result;
	}

	/**
	 * Creates an horizontal separator.
	 *
	 * @param  color the color of the separator
	 * @return the created separator
	 * @see    LyJSeparator
	 */
	public Component createHorizontalSeparator(Color color)
	{
		return new LyJSeparator(SwingConstants.HORIZONTAL, LyJSeparator.TOOLBAR_STYLE,
			color);
	}

	/**
	 * Creates a vertical separator.
	 *
	 * @param  color the color of the separator
	 * @return the created separator
	 * @see    LyJSeparator
	 */
	public Component createVerticalSeparator(Color color)
	{
		return new LyJSeparator(SwingConstants.VERTICAL, LyJSeparator.TOOLBAR_STYLE,
			color);
	}

	/**
	 * Creates panel with a border layout.
	 *
	 * @return the created panel
	 */
	public Container createBorderLayoutPanel()
	{
		return new JPanel(new BorderLayout());
	}

	/**
	 * Gets the size of the splitter if available, else 0.
	 *
	 * @return the size of the splitter
	 */
	public int getSplitterSize()
	{
		if (_panel instanceof LyJSplitterComponent)
			return ((LyJSplitterComponent)_panel).getSplitterSize();
		return 0;
	}

	/**
	 * Handles component resize and checks if minimum sizes are valid.
	 */
	public class _LyJSplitterComponentAdapter extends ComponentAdapter
	{

		/**
		 * Invoked when the component's size changes.
		 *
		 * @param event describing cmponent size change
		 * @see   ComponentAdapter#componentResized(java.awt.event.ComponentEvent)
		 */
		public void componentResized(ComponentEvent event)
		{
			if (event.getComponent() instanceof LyJSplitterComponent)
				((LyJSplitterComponent)event.getComponent()).checkMinimumSize(_leftMinimumSize,
					_rightMinimumSize);
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _panel;
	}

	/**
	 * Adds the specified component to the displayed panel component, depending on the splitting
	 * orientation.
	 *
	 * @param component the component to add
	 * @param sizeInfo  the size of the component (if this size is less or equals to 0, the
	 *                  component is not added, as it is not visible)
	 */
	protected void addAreaBorderLayout(Component component, String sizeInfo)
	{
		int number = _panel.getComponentCount();

		if (LyTools.intFromString(sizeInfo) > 0)
			_panel.add(component, "Center");
		else
		{
			Container subPanel = createBorderLayoutPanel();
			String    pos, barPos;

			subPanel.add(component, "Center");

			if (_splitter.getStyle() == LySplitter.Style.VERTICAL)
			{
				if (number == 0)
				{
					pos = "North";
					barPos = "South";
				}
				else
				{
					pos = "South";
					barPos = "North";
				}
			}
			else
			{
				if (number == 0)
				{
					pos = "West";
					barPos = "East";
				}
				else
				{
					pos = "East";
					barPos = "West";
				}
			}
			String splitterColor =_splitter.getColor();
			Color color;
			if ((splitterColor != null) && !_splitter.isDefaultColor())
			    color = (Color)getEnvironment().getColor(splitterColor);
			else
			    color = Color.lightGray.darker().darker();
			// If -2 is the size, do not add border
			if (LyTools.intFromString(sizeInfo) != -2)
			{
				if (_splitter.getStyle() == LySplitter.Style.VERTICAL)
					subPanel.add(createHorizontalSeparator(color), barPos);
				else
					subPanel.add(createVerticalSeparator(color), barPos);
			}

			_panel.add(subPanel, pos);
		}
	} // end method addAreaBorderLayout

	/**
	 * Sets the minimumSize of the left area.
	 *
	 * @param minSize the dimension to use
	 */
	public void setLeftMinimumSize(Dimension minSize)
	{
		_leftMinimumSize = minSize;
	}

	/**
	 * Sets the minimumSize of the right area.
	 *
	 * @param minSize the dimension to use
	 */
	public void setRightMinimumSize(Dimension minSize)
	{
		_rightMinimumSize = minSize;
	}

	/**
	 * Updates the minimum size occupied by the specified child component. This splitter object can
	 * have only two children so the minimum size is updated only for the left/top or right/bottom
	 * areas.
	 *
	 * @param child child component in this container
	 * @param d     minimum size to use
	 * @see   LySwingContainer#updateMinimumSize(LyComponent, Dimension)
	 */
	public void updateMinimumSize(LyComponent child, Dimension d)
	{
		if (_splitter.getPanel(0) == child)
			setLeftMinimumSize(d);
		else
			setRightMinimumSize(d);

		Dimension d1 = new Dimension();

		if (_splitter.getStyle() == LySplitter.Style.VERTICAL)
		{
			d1.width = Math.max(((_leftMinimumSize == null) ? 0 : _leftMinimumSize.width),
					((_rightMinimumSize == null) ? 0 : _rightMinimumSize.width));
			d1.height = ((_leftMinimumSize == null) ? 0 : _leftMinimumSize.height)
				+ ((_rightMinimumSize == null) ? 0 : _rightMinimumSize.height)
				+ getSplitterSize();
		}
		else
		{
			d1.width = ((_leftMinimumSize == null) ? 0 : _leftMinimumSize.width)
				+ ((_rightMinimumSize == null) ? 0 : _rightMinimumSize.width) + getSplitterSize();
			d1.height = Math.max(((_leftMinimumSize == null) ? 0 : _leftMinimumSize.height),
					((_rightMinimumSize == null) ? 0 : _rightMinimumSize.height));
		}

		super.updateMinimumSize(_splitter, d1);
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalSplitter#setSizes(java.lang.String)
	 */
	public void setSizes(String sizes)
	{
	}

	public void setSelected(boolean status)
	{
		// TODO Auto-generated method stub

	}

	public void setActivePart(ActivePart activePart)
	{
		// TODO Auto-generated method stub

	}
} // end class LySwingSplitter
