/*
 * LySwingTable.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.event.*;
import java.awt.print.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.table.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.LySwingRow.*;
import leon.view.win.swing.component.*;
import leon.view.win.swing.component.LyJTableComponent.*;
import leon.view.win.swing.table.*;


/**
 * This class implements the Table object (as described in the dtd) for swing viewer.
 *
 * <p>A table allows to display a set of rows and columns containing a set of cells. It can contain
 * a header providing information about the content of the cells of the list. It is possible to
 * define its selection modes (simple, extended, multiple or none) and finally to indicate if
 * horizontal and/or vertical separation lines should be drawn.</p>
 *
 * @version $Id: LySwingTable.java 35415 2012-04-30 15:55:03Z nhamon $
 * @author  Lyria
 * @see     LyTable
 */
public class LySwingTable extends LySwingPanel
	implements LySwingComponent,
		LyIsIdleInterface,
		LySetIdleInterface,
		LySetTableInterface,
		LySetWaitingInterface,
		LyGetLabelInterface,
		LyGetValueInterface,
		LyPrintInterface,
		LyPdfInterface,
		LyGraphicalTable,
		LyActionListener,
		LySelectionListener,
		LyGetLockInterface
{
	public static final String			  __VERSION =
		"$Id: LySwingTable.java 35415 2012-04-30 15:55:03Z nhamon $";


	/** Minimum row height of the table, in pixels. */
	public static int					  MIN_ROW_HEIGHT = 16;

	/** Maximum limit of the width computation. */
	public static int					  MAX_COMPUTED_WIDTH_COUNT = 200;
	
	/** Min. width for arrays. */
	public static int				 	  ARRAY_MIN_WIDTH = 30;

	/** The headers of this table. */
	protected Vector<LySwingHeader>		  _headers;

	/** The footers of this table. */
	protected Vector<LySwingFooter>		  _footers;

	/** Alignments of this table. */
	protected int[]						  _aligns;

	/** Image of increasing sort. */
	protected Image						  _ascending;

	/** Image of decreasing sort. */
	protected Image						  _descending;

	/** The swing source model. */
	protected LyJTableModel				  _tableModel;

	/** Renderer of normal cells. */
	protected LyJCellRenderer			  _cellRenderer;

	/** Renderer of header cells. */
	protected LyJHeaderRenderer			  _headerRenderer;

	/** The scrollPane that contains this table. */
	protected Component					  _comp;

	/** The component used to display this table. */
	protected LyJTableComponent			  _tableComp;

	/** The component used to display the table's footer. */
	protected LyJTableFooter			  _footerComp; // the
	// JTableHeader used as a footer

	/** A special column for the status. It is unaffected by the horizontal slider. */
	protected LyJTableStatusColumn		  _statusColumn;

	/** Boolean indicating if a column status is present or not. */
	protected Boolean					  _hasStatusColumn;

	/** The name used for the status column. */
	public String						  STATUS_COLUMN_NAME;

	/** Idle mode. */
	protected boolean					  _isIdle;

	/** Workaround to prevents events when removing rows. */
	protected boolean					  _skipSelectionDemand;

	/** Automatically computes width when table is opened. */
	protected boolean					  _needResizeColumns;

	/** Indicates which rows have a cursor to open/close rows and their offset. */
	protected Hashtable<Integer, Integer> _rowCursors;

	/** Indicates if the table has at least one row with an image but no header for status. */
	protected boolean					  _imageAndNoStatusHeaderMode;

	/** The view independent table description. */
	protected LyTable					  _table;

	/** Provides notification events on component's selections. */
	protected LySelectionTransmitter	  _selectionTransmitter;

	/** Indicates if the table is in compact mode. */
	protected boolean					  _compactMode;

	/**
	 * Class constructor of a Table object for the Swing target.
	 *
	 * @param table a TABLE component defined for every type of view targets. This table is
	 *              initialized for the swing view and filled with the default table values.
	 */
	public LySwingTable(LyTable table)
	{
		super(table);
		_table = table;

		initialize();
		setContent();
	}

	/**
	 * Creates the table component by initializing the table model and its cell renderers, then
	 * registers the event listeners.
	 */
	public void initialize()
	{
		STATUS_COLUMN_NAME = translate("LY_STATUS_SORT");

		// init the tableModel
		_tableModel = new LyJTableModel(this);
		_aligns = null;
		_skipSelectionDemand = false;
		_imageAndNoStatusHeaderMode = false;

		// init the cellRenderers
		initRenderers();

		_ascending = getImage("LY_TABLE_SORT_ASC");
		_descending = getImage("LY_TABLE_SORT_DES");
		_compactMode = false;
		_tableComp = null;
		_footerComp = null;
		_statusColumn = null;
		_hasStatusColumn = null;

		_comp = createLyJTableComponent();

		/* register for events */
		LyActionTransmitter    actionTransmitter = getTableComponent().getActionTransmitter();
		LySelectionTransmitter selectionTransmitter = getTableComponent().getSelectionTransmitter();

		if (actionTransmitter != null)
			actionTransmitter.addActionListener(this);

		if (selectionTransmitter != null)
			selectionTransmitter.addSelectionListener(this);

		_isIdle = false;
		_needResizeColumns = true;
		_rowCursors = null;

		// InputMap inputMap =_tableComp.getInputMap();
		// ActionMap actionMap = _tableComp.getActionMap();
		// inputMap.clear();
		// actionMap.clear();

		setEnabled(_table.isEnabled());
	} // end method initialize

	/**
	 * Sets the content of the table.
	 */
	public void setContent()
	{
		_needResizeColumns = true; // Content fully updated
	}
	
	public void setNeedResizeColumns(boolean needResizeColumns)
	{
		_needResizeColumns = needResizeColumns;
	}

	/**
	 * Initialize the cell and header renderers. Sets also their colors and fonts.
	 */
	public void initRenderers()
	{
		Font	  headerAndFooterFont = getFont("LY_TABLE_LABEL");
		Color     headerAndFooterColor = getColor("LY_HEADER_COLOR");
		Color     headerForegroundColor = getColor("LY_TABLE_HEADER_FOREGROUND");
		Color     oddColor = getColor("LY_ROW_COLOR_ODD");
		Color     evenColor = getColor("LY_ROW_COLOR_EVEN");
		LyHeaders headers = _table.getHeaders();

		if (headers != null)
		{
			int		 i, align, n = headers.getHeaderCount();
			LyHeader header = null;

			_aligns = new int[n];

			for (i = 0; i < n; i++)
			{
				header = headers.getHeader(i);

				if (header.getAlign() == LyHeader.Align.CENTER)
					align = SwingConstants.CENTER;
				else if (header.getAlign() == LyHeader.Align.RIGHT)
					align = SwingConstants.RIGHT;
				else
					align = SwingConstants.LEFT;

				_aligns[i] = align;
			}
		}

		_cellRenderer = new LyJCellRenderer(oddColor, evenColor, _aligns);
		_headerRenderer = new LyJHeaderRenderer(headerAndFooterFont, headerAndFooterColor,
				headerForegroundColor, ((headers != null) ? headers.isHierarchical() : false));
	} // end method initRenderers

	/**
	 * Indicates if the table is hierarchical.
	 *
	 * @return true if this table's rows can contain sub-rows, false otherwise
	 */
	public boolean isHierarchical()
	{
		LyRows rows = _table.getRows();

		return ((rows != null) && (rows.isHierarchical()));
	}

	/**
	 * Gets the swing table component.
	 *
	 * @return the table component
	 */
	public LyJTableComponent getTableComponent()
	{
		return _tableComp;
	}

	/**
	 * Gets the table model.
	 *
	 * @return the source model of this table
	 */
	public LyJTableModel getTableModel()
	{
		return _tableModel;
	}

	/**
	 * Gets the swing table footer.
	 *
	 * @return the table's footer component
	 */
	public LyJTableFooter getTableFooter()
	{
		return _footerComp;
	}

	/**
	 * Gets the swing component for the status column.
	 *
	 * @return the component that display the status column
	 */
	public LyJTableStatusColumn getStatusColumn()
	{
		return _statusColumn;
	}

	/**
	 * Gets the swing component for the status column header.
	 *
	 * @return the component that display the status column header
	 */
	public LyJTableStatusHeader getStatusHeader()
	{
		return (_statusColumn != null) ? (_statusColumn.getHeader()) : (null);
	}

	/**
	 * Returns the offset of the headers component relative to this table component.
	 *
	 * @return the offset of the headers.
	 */
	public Dimension getTableComponentOffset()
	{
		Dimension dimension = new Dimension(0, 0);

		if (_statusColumn != null)
			dimension.width = _statusColumn.getPreferredSize().width;

		_LyJTableHeader tableHeader = (_LyJTableHeader)_tableComp.getTableHeader();

		if (tableHeader != null)
			dimension.height = tableHeader.getPreferredSize().height;

		return dimension;
	}

	/**
	 * Returns the component used to display popup menus.
	 *
	 * @return the parent component of popup menus.
	 */
	protected JComponent getPopupMenuParent()
	{
		return getTableComponent();
	}

	/**
	 * Return the visibility of headers.
	 * @return true if all the headers are hidden
	 */
	public boolean areAllHeadersHidden()
	{
		LyHeaders headers = _table.getHeaders();
		if (headers != null)
		{
			for (int i = 0; i < headers.getHeaderCount(); i++)
			{
				// if we find one header visible, show all
				LyHeader header = headers.getHeader(i);
				if (!header.isHidden())
				{
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * Creates the graphical table using models and renderers. Defines and registers the listeners
	 * for the table to behave like it should.
	 *
	 * @return a fully initialized graphical table
	 */
	public Component createLyJTableComponent()
	{
		Font		  tableFont = getFont("LY_TABLE");
		Font		  headerAndFooterFont = getFont("LY_TABLE_LABEL");
		Color		  selectColor = getColor("LY_TABLE_SELECT_COLOR");
		Color		  headerAndFooterColor = getColor("LY_HEADER_COLOR");
		Color		  gridColor = getColor("LY_TABLE_LINES");
		MouseListener mouseListener = new LyJTableMouseListener(this);
		boolean		  dragNDrop = getEnvironment().testResource("LY_DRAG_AND_DROP");

		// If at least one row has an image, check if there is a header for status column
		boolean		  hasRowImage = false;
		int			  maxCols = 1;
		LyRows		  rows = _table.getRows();
		int			  i, n = ((rows == null) ? 0 : rows.getRowCount());
		LyRow		  row;

		for (i = 0; i < n; i++)
		{
			row = rows.getRow(i);

			if ((row.getImage() != null) && (row.getImage().length() > 0))
				hasRowImage = true;

			if (row.getComponentCount() > maxCols)
				maxCols = row.getComponentCount();
		}

		if (_table.getHeaders() != null)
		{
			// Check if status header already exists. If not create it.
			LyHeaders headers = _table.getHeaders();
			LyHeader  firstHeader = (headers.getHeaderCount() == 0 ? null : headers.getHeader(0));

			if (hasRowImage
			    && ((firstHeader == null)
			        || !"LY_STATUS_SORT".equals(firstHeader.getLabel())))
			{
				_imageAndNoStatusHeaderMode = true;
				firstHeader = new LyHeader();
				firstHeader.setLabel("LY_STATUS_SORT");
				_table.getHeaders().insertHeader(firstHeader, 0);
				updateEmptyCells();
			}


			if (areAllHeadersHidden())
			{
				// SMA : create component with hidden headers for display list ...
				_tableComp = new LyJTableComponent(_descending, _ascending, mouseListener,
						new LyJTableSelectionListener(this))
				{
					private static final long serialVersionUID = 1756357133053629525L;

					/*
					 * (non-Javadoc)
					 * @see javax.swing.JTable#configureEnclosingScrollPane()
					 */
					protected void configureEnclosingScrollPane()
					{
						Container p = getParent();

						if (p instanceof JViewport)
						{
							Container gp = p.getParent();

							if (gp instanceof JScrollPane)
							{
								JScrollPane scrollPane = (JScrollPane)gp;

								scrollPane.setColumnHeader(null);
								scrollPane.setColumnHeaderView(null);
							}
						}
					}
				};
			}
			else
			{
				_tableComp = new LyJTableComponent(_descending, _ascending, mouseListener,
						new LyJTableSelectionListener(this));
			}
			_LyJTableHeader tableHeader = (_LyJTableHeader)_tableComp.getTableHeader();

			if (tableHeader != null)
				tableHeader.setReorderingAllowed(_table.getHeaders().isReorderingAllowed());
		}
		else
		{
			// If table has no headers build one but do not display it
			LyHeaders headers = new LyHeaders();
			LyHeader  header;

			_imageAndNoStatusHeaderMode = true;

			// If table has no header, always add a header for status column
			// and an empty cell for each row to display the images
			header = new LyHeader();
			header.setLabel("LY_STATUS_SORT");
			headers.addHeader(header);
			updateEmptyCells();

			for (i = 0; i < maxCols; i++)
			{
				header = new LyHeader();
				header.setLabel(" ");
				headers.addHeader(header);
				_table.setHeaders(headers);
			}

			_tableComp = new LyJTableComponent(_descending, _ascending, mouseListener,
					new LyJTableSelectionListener(this))
				{
					private static final long serialVersionUID = -8932335005227319361L;

					/*
					 * (non-Javadoc)
					 * @see javax.swing.JTable#configureEnclosingScrollPane()
					 */
					protected void configureEnclosingScrollPane()
					{
						Container p = getParent();

						if (p instanceof JViewport)
						{
							Container gp = p.getParent();

							if (gp instanceof JScrollPane)
							{
								JScrollPane scrollPane = (JScrollPane)gp;

								scrollPane.setColumnHeader(null);
								scrollPane.setColumnHeaderView(null);
							}
						}
					}
				};
		}

		// Set default min size
		if ((_table.getParentFormFieldId() != null)
				&& (ARRAY_MIN_WIDTH > 0))
			_tableComp.setMinChars(ARRAY_MIN_WIDTH);

		checkComputeHeight();

		// FFT00475 Escape key handling
		InputMap  inputMap = _tableComp.getInputMap();
		ActionMap actionMap = _tableComp.getActionMap();

		// Removing the selectAll action in the Swing table
		if (actionMap.getParent() != null)
		{
			// actionMap.getParent().clear();
			actionMap.getParent().remove("selectAll");
			actionMap.getParent().remove("clearSelection");
		}

		actionMap.remove("clearSelection");
		actionMap.remove("selectAll");

		KeyStroke stroke = KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0);
		String    key = "clearSelection";

		inputMap.put(stroke, key);
		Action action = new AbstractAction(key)
			{
				private static final long serialVersionUID = -258078603759807209L;

				/*
				 * (non-Javadoc)
				 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
				 */
				public void actionPerformed(ActionEvent e)
				{
					_tableComp.clearSelection();
				}
			};
		actionMap.put(key, action);

		// Remove CTRL-A input
		stroke = KeyStroke.getKeyStroke(KeyEvent.VK_A, java.awt.Event.CTRL_MASK);
		inputMap.remove(stroke);

		// Check if table is in edition before performing a VK_DOWN
		stroke = KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, 0);
		inputMap.remove(stroke);
		action = new AbstractAction("selectNextRow")
			{
				private static final long serialVersionUID = 1437010620856956055L;

				public void actionPerformed(ActionEvent e)
				{
					LyJTableComponent table = (LyJTableComponent)e.getSource();

					if (!table.isEditing())
					{
						// If a cell is being edited, do not change selection (VK_DOWN may be used
						// by cell component used for edition)
						int selectedRow = Math.min(table.getSelectedRow() + 1,
								getDataRowCount() - 1);

						table.setRowSelectionInterval(selectedRow, selectedRow);
						table.scrollToRow(selectedRow);
					}
				}
			};
		actionMap.put("selectNextRow", action);
		inputMap.put(stroke, "selectNextRow");

		// Check if table is in edition before performing a VK_UP
		stroke = KeyStroke.getKeyStroke(KeyEvent.VK_UP, 0);
		inputMap.remove(stroke);
		action = new AbstractAction("selectPreviousRow")
			{
				private static final long serialVersionUID = 3437340917856956056L;

				public void actionPerformed(ActionEvent e)
				{
					LyJTableComponent table = (LyJTableComponent)e.getSource();

					if (!table.isEditing())
					{
						// If a cell is being edited, do not change selection (VK_UP may be used
						// by cell component used for edition)
						int selectedRow = Math.max(table.getSelectedRow() - 1, 0);

						table.setRowSelectionInterval(selectedRow, selectedRow);
						table.scrollToRow(selectedRow);
					}
				}
			};
		actionMap.put("selectPreviousRow", action);
		inputMap.put(stroke, "selectPreviousRow");

		// stroke = KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, java.awt.Event.SHIFT_MASK);
		// inputMap.put(stroke,"selectNextRowExtendSelection");
		// stroke = KeyStroke.getKeyStroke(KeyEvent.VK_UP, java.awt.Event.SHIFT_MASK);
		// inputMap.put(stroke,"selectPreviousRowExtendSelection");

		// When the user press ENTER, it execute the default action on the selected item (if only
		// one selected)
		stroke = KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0);
		inputMap.remove(stroke);
		action = new AbstractAction("execute")
			{
				private static final long serialVersionUID = 8910138082777468830L;

				/*
				 * (non-Javadoc)
				 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
				 */
				public void actionPerformed(ActionEvent e)
				{
					if (_tableComp.getSelectedRowCount() != 1)
						return;

					int		   index = _tableComp.getSelectedRow();
					LySwingRow row = getRow(index);

					if (row == null)
						return;
					String		  item = row.getId();
					LyActionEvent actionEvent = new LyActionEvent(this);

					actionEvent.setValue(item);
					actionEvent.addIndex(index);
					_tableComp.actionPerformed(actionEvent);
				}
			};
		actionMap.put("execute", action);
		inputMap.put(stroke, "execute");

		// INSERT invoke a create process
		stroke = KeyStroke.getKeyStroke(KeyEvent.VK_INSERT, 0);
		inputMap.remove(stroke);
		action = new AbstractAction("insert")
			{
				private static final long serialVersionUID = 1913470576544040547L;

				/*
				 * (non-Javadoc)
				 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
				 */
				public void actionPerformed(ActionEvent e)
				{
					if (_tableComp.getActionTransmitter() != null)
					{
						LyActionEvent actionEvent = new LyActionEvent(this, null, null,
								LyActionDemand.TYPE_INSERT);

						_tableComp.getActionTransmitter().processActionEvent(actionEvent);
					}
				}

			};
		actionMap.put("insert", action);
		inputMap.put(stroke, "insert");

		// TAB give the focus to next component
		stroke = KeyStroke.getKeyStroke(KeyEvent.VK_TAB, 0);
		inputMap.remove(stroke);
		action = new AbstractAction("next_focus")
			{
				private static final long serialVersionUID = 2810138769377468830L;

				/*
				 * (non-Javadoc)
				 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
				 */
				public void actionPerformed(ActionEvent e)
				{
					Container			 cycleRoot = _tableComp.getFocusCycleRootAncestor();
					FocusTraversalPolicy policy = _tableComp.getFocusTraversalPolicy();

					if ((policy == null) && (cycleRoot != null))
						policy = cycleRoot.getFocusTraversalPolicy();

					if (policy != null)
					{
						Component target = policy.getComponentAfter(cycleRoot, _tableComp);

						if (target == null)
							target = policy.getComponentBefore(cycleRoot, _tableComp);

						if (target != null)
							target.requestFocusInWindow();
					}
				}
			};
		actionMap.put("next_focus", action);
		inputMap.put(stroke, "next_focus");

		// SHIFT+TAB give the focus to next component
		stroke = KeyStroke.getKeyStroke(KeyEvent.VK_TAB, java.awt.Event.SHIFT_MASK);
		inputMap.remove(stroke);
		action = new AbstractAction("previous_focus")
			{
				private static final long serialVersionUID = 4910183658277468830L;

				/*
				 * (non-Javadoc)
				 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
				 */
				public void actionPerformed(ActionEvent e)
				{
					Container			 cycleRoot = _tableComp.getFocusCycleRootAncestor();
					FocusTraversalPolicy policy = _tableComp.getFocusTraversalPolicy();

					if ((policy == null) && (cycleRoot != null))
						policy = cycleRoot.getFocusTraversalPolicy();

					if (policy != null)
					{
						Component target = policy.getComponentBefore(cycleRoot, _tableComp);

						if (target == null)
							target = policy.getComponentAfter(cycleRoot, _tableComp);

						if (target != null)
							target.requestFocusInWindow();
					}
				}
			};
		actionMap.put("previous_focus", action);
		inputMap.put(stroke, "previous_focus");

		// DELETE invoke a deletion process
		stroke = KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0);
		inputMap.remove(stroke);
		action = new AbstractAction("delete")
			{
				private static final long serialVersionUID = -1472404391944058373L;

				/*
				 * (non-Javadoc)
				 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
				 */
				public void actionPerformed(ActionEvent e)
				{
					if (_tableComp.getActionTransmitter() != null)
					{
						LyActionEvent actionEvent = new LyActionEvent(this, null, null,
								LyActionDemand.TYPE_DELETE);

						_tableComp.getActionTransmitter().processActionEvent(actionEvent);
					}
				}

			};

		actionMap.put("delete", action);
		inputMap.put(stroke, "delete");

		// Remove handling of selectAll by the table
		actionMap.put("selectAll", new AbstractAction()
			{
				private static final long serialVersionUID = 5862273273654059833L;

				/*
				 * (non-Javadoc)
				 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
				 */
				public void actionPerformed(ActionEvent e)
				{
				}
			});

		_tableComp.setInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW, null);
		_tableComp.setInputMap(JComponent.WHEN_FOCUSED, inputMap);
		_tableComp.setInputMap(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT, inputMap);

		if (tableFont != null)
			_tableComp.setFont(tableFont);

		if (selectColor != null)
		{
			_tableComp.setSelectionBackground(selectColor);

			// if selection background too dark, selection foreground is white
			int sumColor = selectColor.getRed() + selectColor.getGreen() + selectColor.getBlue();

			if (sumColor < 450)
				_tableComp.setSelectionForeground(Color.white);
		}

		// SET THE SELECTION ATTRIBUTE OF A TABLE
		LyTable.SelectionMode selectionMode = getSelectionMode();

		if (selectionMode == LyTable.SelectionMode.SINGLE)
			_tableComp.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		else if (selectionMode == LyTable.SelectionMode.MULTIPLE)
			_tableComp.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		else
		// LyTable.SelectionMode.NONE
		{
			_tableComp.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
			_tableComp.setRowSelectionAllowed(false);
		}

		// Set the visibility of the grid
		_tableComp.setShowHorizontalLines(isHLines());
		_tableComp.setShowVerticalLines(isVLines());

		if (gridColor != null)
			_tableComp.setGridColor(gridColor);

		// Set the column model and the data model
		_tableComp.setColumnModel(new LyJTableColumnModel(this));
		_tableComp.setModel(getTableModel());

		computeRowHeights();

		// Set the cell renderer
		_tableComp.setCellRenderer(_cellRenderer);

		// Configure properly the column model
		configureColumnModel(_tableComp);

		// set min height
		int height = _table.getNRows();

		if (height < 0)
			height = 10;
		else
			_tableComp.setDisplayedHeight(height);

		_tableComp.setMinHeight(height);

		// If table is in a form, remove right margin
		String formFieldId = _table.getParentFormFieldId();

		if (formFieldId != null)
		{
			LyView view = _table.getParentView();
			// Margin seems to be greater in a dialog window
			if (view instanceof LyFrame)
				_tableComp.setRightMargin(0);
			else
				_tableComp.setRightMargin(-15);
		}

		_tableComp.setAutoscrolls(false);

		// Create and configure the scrollpanel that will contain the table
		LyViewManager	   viewManager = getViewManager();
		LyJTableScrollPane scrollPane = new LyJTableScrollPane();
		Color			   bgColor = getColor("LY_TABLE_BACKGROUND");

		scrollPane.setScreenSize(viewManager.getScreenSize(_tableComp));
		scrollPane.setViewportView(_tableComp);
		scrollPane.getViewport().setBackground(bgColor);
		scrollPane.getViewport().addMouseListener(mouseListener);

		if (dragNDrop)
			_tableComp.setDragAndDrop();

		// Add table header
		if (hasHeaders())
		{
			scrollPane.setColumnHeaderView(_tableComp.getTableHeader());
			_tableComp.getTableHeader().setColumnModel(_tableComp.getColumnModel());

			LyHeaders headers = _table.getHeaders();

			if (headers.isHierarchical())
				addHeaderGroups();

			JLabel label = new JLabel();

			label.setBorder(new LyJSpecialCellBorder(false));
			label.setBackground(headerAndFooterColor);
			label.setOpaque(true);
			scrollPane.setCorner(ScrollPaneConstants.UPPER_RIGHT_CORNER, label);
		}

		// Create and add the table footer
		if (hasFooters())
		{
			Color headerForegroundColor = getColor("LY_TABLE_HEADER_FOREGROUND");

			// Create the footer component
			_footerComp = new LyJTableFooter(_tableComp);

			if (headerAndFooterFont != null)
				_footerComp.setFont(headerAndFooterFont);
			if (headerAndFooterColor != null)
				_footerComp.setBackground(headerAndFooterColor);
			if (headerAndFooterColor != null)
				_footerComp.setForeground(headerForegroundColor);

			scrollPane.setColumnFooterView(_footerComp);

			JLabel label = new JLabel();

			label.setBorder(new LyJFooterCellBorder(false));
			label.setBackground(headerAndFooterColor);
			label.setOpaque(true);
			scrollPane.setCorner(ScrollPaneConstants.LOWER_RIGHT_CORNER, label);
		}

		// Create and add status column component
		if (hasStatusColumn())
		{
			_statusColumn = new LyJTableStatusColumn(_tableComp, headerAndFooterFont,
					headerAndFooterColor);
			_statusColumn.getHeader().addMouseListener(mouseListener);

			if (!_imageAndNoStatusHeaderMode)
				_statusColumn.setHasImage();

			_statusColumn.setBackground(bgColor);
			scrollPane.setRowHeaderView(_statusColumn);
			scrollPane.setCorner(ScrollPaneConstants.UPPER_LEFT_CORNER, _statusColumn.getHeader());
			scrollPane.setCorner(ScrollPaneConstants.LOWER_LEFT_CORNER, _statusColumn.getFooter());
		}

		cacheHeight();
		updateSelection(0, n - 1);

		return scrollPane;
	} // end method createLyJTableComponent

	/**
	 * Checks if the row height must be computed for each row of the table.
	 */
	protected void checkComputeHeight()
	{
		LyRows	    rows = _table.getRows();
		int		    i, n = ((rows == null) ? 0 : rows.getRowCount());
		int		    j, m;
		LyRow	    row;
		boolean     computeHeight = false;
		LyComponent component;
		boolean		hasSubComp;

		// Height must be computed only if a row contains a form field or an image or sub-cells or if it has a label with no cell.
		for (i = 0; i < n && !computeHeight; i++)
		{
			row = rows.getRow(i);
			m = row.getComponentCount();
			hasSubComp = false;

			for (j = 0; j < m; j++)
			{
				component = row.getComponent(j);

				if ((component instanceof LyFormField) || (component instanceof LyToolBar))
				{
					computeHeight = true;
					hasSubComp = true;
					break;
				}
				else if (component instanceof LyCell)
				{
					LyCell cell = (LyCell)component;
					String label = cell.getLabel();

					hasSubComp = true;

					if ((cell.getImage() != null)
							|| (cell.getCellCount() > 0)
							|| ((label != null)
									&& (label.indexOf('\n') >= 0)))
					{
						computeHeight = true;
						break;
					}
				}
			}

			if (!hasSubComp && (row.getLabel() != null) && (row.getLabel().length() > 0))
				computeHeight = true;
		}

		if (computeHeight)
			_tableComp.setComputeHeight(true);
	} // end method checkComputeHeight

	/**
	 * Updates each row of the table in order to add an empty cell at first position. Must be used
	 * only when there is a status column and no header to correctly display the rows.
	 */
	protected void updateEmptyCells()
	{
		if (_table.getRows() == null)
			return;

		LyRows	    rows = _table.getRows();
		LyRow	    row;
		LySwingRow  swingRow;
		LyCell	    cell;
		LyComponent comp;
		int		    i, n = rows.getRowCount();

		for (i = 0; i < n; i++)
		{
			row = rows.getRow(i);
			comp = row.getChild(0);

			if (!(comp instanceof LyCell) || ((LyCell)comp).getLabel() != LyJTableModel.EMPTY_CELL_VALUE)
			{
				swingRow = (LySwingRow)row.getInternalView(false);
				cell = new LyCell();
				cell.setLabel(LyJTableModel.EMPTY_CELL_VALUE);
				if (swingRow != null)
					swingRow.setResetRowComponents(false);

				row.insertCell(cell, 0);

				if (swingRow != null)
					swingRow.setResetRowComponents(true);
			}
		}
	} // end method updateEmptyCells

	/**
	 * Creates and displays the components used to display headers when table header is
	 * hierarchical.
	 */
	protected void addHeaderGroups()
	{
		Font		    headerAndFooterFont = getFont("LY_TABLE_LABEL");
		Color		    headerAndFooterColor = getColor("LY_HEADER_COLOR");
		Color		    headerForegroundColor = getColor("LY_TABLE_HEADER_FOREGROUND");
		_LyJTableHeader tableHeader = (_LyJTableHeader)_tableComp.getTableHeader();
		LyHeaders	    headers = _table.getHeaders();
		int			    i, n = headers.getHeadersCount(), j, m;
		LyHeaders	    subHeaders;
		LyHeaders	    parent;
		LyHeader	    startHeader, subHeader;
		int			    depth;
		JLabel		    label;

		for (i = 0; i < n; i++)
		{
			subHeaders = headers.getHeaders(i);
			parent = (LyHeaders)subHeaders.getParent();
			depth = 1;
			m = headers.getHeaderCount();

			// Retrieve depth of this headers in the top headers
			while ((parent != null) && !(parent == headers))
			{
				parent = (LyHeaders)parent.getParent();
				depth++;
			}

			// Retrieve start column and end column of this headers
			startHeader = subHeaders.getHeader(0);

			for (j = 0; j < m; j++)
			{
				subHeader = headers.getHeader(j);

				if (subHeader == startHeader)
					break;
			}

			m = subHeaders.getHeaderCount();

			// Create the component used for header group
			label = new JLabel(translate(subHeaders.getLabel()));
			label.setOpaque(true);
			label.setFont(headerAndFooterFont);
			label.setBackground(headerAndFooterColor);
			label.setForeground(headerForegroundColor);
			label.setBorder(new LyJSpecialCellBorder(3, 3, 3, 3));

			if (subHeaders.getAlign() == LyHeader.Align.LEFT)
				label.setHorizontalAlignment(SwingConstants.LEFT);
			else if (subHeaders.getAlign() == LyHeader.Align.RIGHT)
				label.setHorizontalAlignment(SwingConstants.RIGHT);
			else
				label.setHorizontalAlignment(SwingConstants.CENTER);

			tableHeader.addHeaderGroup(label, j, j + m, depth, translate(subHeaders.getTooltip()));
		}

		// Reordering of columns is not allowed since columns are grouped
		tableHeader.setReorderingAllowed(false);
	} // end method addHeaderGroups

	/**
	 * Resets the list of row containing cursors.
	 */
	public void resetRowCursors()
	{
		if (_rowCursors != null)
			_rowCursors.clear();
	}

	/**
	 * Sets the row of given index to have a pointing cursor.
	 *
	 * @param row    the index of the row
	 * @param offset the offset for current row
	 */
	public void addRowCursor(int row, int offset)
	{
		if (_rowCursors == null)
			_rowCursors = new Hashtable<Integer, Integer>();

		_rowCursors.put(new Integer(row), new Integer(offset));

		// Reordering of columns is not allowed since cursors must stay in first column
		_LyJTableHeader tableHeader = (_LyJTableHeader)_tableComp.getTableHeader();

		tableHeader.setReorderingAllowed(false);
	}

	/**
	 * Returns the list of row indexes that have a cursor.
	 *
	 * @return the list of row indexes that have a cursor
	 */
	public Hashtable<Integer, Integer> getRowCursors()
	{
		if (_rowCursors != null)
		{
			@SuppressWarnings("unchecked")
			Hashtable<Integer, Integer> temp = (Hashtable<Integer, Integer>)_rowCursors.clone();
			return temp;
		}
		return null;
	}

	/**
	 * Computes the height for each cell of each row.
	 */
	protected void cacheHeight()
	{
		if (!_tableComp.isComputeHeight())
			return;

		_tableComp.cacheHeight();
	}

	/**
	 * Computes the height of each row.
	 */
	protected void computeRowHeights()
	{
		// Compute preferred row height with font dimension
		FontMetrics fm = _tableComp.getFontMetrics(_tableComp.getFont());
		LyRows	    rows = _table.getRows();
		int		    fmHeight = fm.getHeight() + 2, rowHeight;

		_tableComp.setRowHeight(Math.max(fmHeight, MIN_ROW_HEIGHT));

		if (!_tableComp.isComputeHeight())
			return;

		int n = getDataRowCount();

		if (rows == null)
			return;

		LyRow subRow;
		int   i;

		if (!rows.isHierarchical() && !rows.hasHiddenRow())
		{
			for (i = 0; i < n; i++)
			{
				subRow = rows.getRow(i);
				rowHeight = Math.max(fmHeight,
						_cellRenderer.computeDimension(subRow, _tableComp.getFont()));
				_tableComp.setRowHeight(i, Math.max(rowHeight, MIN_ROW_HEIGHT));
			}
		}
		else
		{
			Vector<LyRow> subRows = new Vector<LyRow>();

			rows.getDisplayedSubRows(subRows);

			for (i = 0; i < n; i++)
			{
				subRow = subRows.elementAt(i);
				rowHeight = computeRowHeight(subRow);
				_tableComp.setRowHeight(i, Math.max(rowHeight, MIN_ROW_HEIGHT));
			}
		}
	} // end method computeRowHeights

	/**
	 * Compute the height of a given row.
	 *
	 * @param  row the row whose height must be computed.
	 * @return the height of the row (in pixels).
	 */
	public int computeRowHeight(LyRow row)
	{
		FontMetrics fm = _tableComp.getFontMetrics(_tableComp.getFont());
		int		    fmHeight = fm.getHeight() + 2;
		int			result = Math.max(fmHeight, _cellRenderer.computeDimension(row, _tableComp.getFont()));

		if ((row.getFont() != null) && (row.getLabel() != null) && (row.getLabel().length() > 0))
		{
			fm = _tableComp.getFontMetrics(getFont(row.getFont()));
			fmHeight = fm.getHeight() + 2;

			result = Math.max(fmHeight, result);
		}

		return result;
	}

	/**
	 * Gets the preferred size of this table depending on its columns and rows as well as its status
	 * column and footer row, if any.
	 *
	 * @return the preferred size of this component
	 */
	public Dimension getPreferredSize()
	{
		int width = getTableComponent().getPreferredSize().width;
		int height = getTableComponent().getPreferredSize().height;

		if (_statusColumn != null)
			width += _statusColumn.getPreferredSize().width;

		if (_footerComp != null)
			height += _footerComp.getPreferredSize().height;

		return new Dimension(width, height);
	}

	/**
	 * Checks if the column at the specified index is the table status column.
	 *
	 * @param  columnIndex index of the column to test
	 * @return true if column at columnIndex is the status column
	 */
	public boolean isStatusColumn(int columnIndex)
	{
		return getTableModel().getColumnName(columnIndex).equals(STATUS_COLUMN_NAME);
	}

	/**
	 * Checks if this table has a status column.
	 *
	 * @return true if this table has a status column
	 */
	public boolean hasStatusColumn()
	{
		if (_hasStatusColumn == null)
			_hasStatusColumn = Boolean.valueOf(getTableModel().findColumn(STATUS_COLUMN_NAME) != -1);

		return _hasStatusColumn.booleanValue();
	}

	/**
	 * Configures the TableColumnModel of the table. Sets the header and footer values and the
	 * header renderer for each column, then configures these column to their preferred width, with
	 * a final unresizable column to fill the unused space of the table.
	 *
	 * @param tableComp the table component to configure
	 */
	public void configureColumnModel(LyJTableComponent tableComp)
	{
		TableColumnModel columnModel = tableComp.getColumnModel();
		LyJTableColumn   column;
		LyJTableModel    tableModel = getTableModel();
		int				 i, n = tableModel.getColumnCount();

		// for each column
		for (i = 0; i < n; i++)
		{
			column = (LyJTableColumn)columnModel.getColumn(i);
			column.setIdentifier(tableModel.getColumnName(i));

			// Set the header values and renderer
			if (hasHeaders())
			{
				column.setHeaderValue(tableModel.getColumnHeaderValue(i));
				column.setHeaderRenderer(_headerRenderer);
			}

			// Set the footer values
			if (hasFooters())
			{
				column.setFooterValue(tableModel.getColumnFooterValue(i));

				if ((_aligns != null) && (i < _aligns.length))
					column.setFooterAlign(_aligns[i]);
			}

			// Compute and set the preferred width of this column
			int width = computeColumnPreferredWidth(tableComp, i, tableComp.getFont(), -1, true);

			column.setPreferredWidth(width);

			// The last column (no data, just for nice seeing)
			if (i == (n - 1))
			{
				column.setMinWidth(Toolkit.getDefaultToolkit().getScreenSize().width);
				column.setResizable(false);
			}

			// SET THE RESIZABLE ATTRIBUTE OF HEADER
			if (hasHeaders() && (i < getHeaders().size()))
			{
				column.setResizable(getHeader(i).isResizable());
				column.setTooltip(getHeader(i).getTooltip());
			}
		}

		// This column is invisible since the status column is
		// managed by its own component.
		if (hasStatusColumn())
		{
			column = (LyJTableColumn)tableComp.getColumn(STATUS_COLUMN_NAME);
			column.setResizable(false);
			column.setMaxWidth(0);
			column.setMinWidth(0);
			column.setPreferredWidth(0);
		}
	} // end method configureColumnModel

	/**
	 * Computes the minimum width of a column used to display entirely its data.
	 *
	 * @param  tableComp   the table component
	 * @param  indexColumn index of the column
	 * @param  font        the font used in the given column
	 * @param  currentRow  current column
	 * @param  useColWidth indicates whether columnWidth must be taken into account
	 * @return the minimum width for a column where each data of it is displayed
	 */
	public int computeColumnPreferredWidth(LyJTableComponent tableComp, int indexColumn, Font font,
		int currentRow, boolean useColWidth)
	{
		String text = (_compactMode ? "" : tableComp.getColumnName(indexColumn));

		text = translate(text);

		FontMetrics fm = tableComp.getFontMetrics(font);
		int		    i, n = getTableModel().getDataRowCount();
		int		    width = 0;

		// Check if a specific column width has been set on the header
		if (hasHeaders() && (indexColumn < getHeaders().size()))
		{
			LyHeaders headers = _table.getHeaders();
			LyHeader  header = headers.getHeader(indexColumn);

			if (header.isSortable())
			{
				Font	    headerFont = getFont("LY_TABLE_LABEL_SORTABLE");
				FontMetrics headerFm = tableComp.getFontMetrics(headerFont);

				width = ((text == null) ? 0 : headerFm.stringWidth(text));

				if (!_compactMode || (width != 0))
					width += 40;
			}
			else
			{
				Font	    headerFont = getFont("LY_TABLE_LABEL");
				FontMetrics headerFm = tableComp.getFontMetrics(headerFont);

				width = ((text == null) ? 0 : headerFm.stringWidth(text));

				if (!_compactMode || (width != 0))
					width += 10;
			}

			// If header contains an image, add space for image
			if (header.getImage() != null)
			{
				Image image = getImage(header.getImage());

				width += image.getWidth(null);
			}
		}

		int			  startIdx = ((currentRow < 0) ? 0 : currentRow);
		int			  endIdx = ((currentRow < 0) ? Math.min(n, MAX_COMPUTED_WIDTH_COUNT)
												 : (currentRow + 1));

		// Check footer width
		LySwingFooter footer = getFooter(indexColumn);

		if (footer != null)
		{
			String label = translate(footer.getLabel());

			if (label != null)
				width = Math.max(width, fm.stringWidth(label) + 10);
		}

		for (i = startIdx; i < endIdx; i++)
		{
			Object value = tableComp.getValueAt(i, indexColumn);

			if (value instanceof LySwingCell)
				width = Math.max(width, ((LySwingCell)value).getWidth(fm));
			else if (value instanceof LySwingFormField)
			{
				LySwingFormField swingField = (LySwingFormField)value;
				Component		 c = swingField.getFullComponent();

				// For _imageAndNoStatusHeaderMode mode, columns are not resizable : use preferred width of each
				// component to ensure that there will be enough place to display them.
				int	fieldWidth = (_imageAndNoStatusHeaderMode ? c.getPreferredSize().width : c.getMinimumSize().width);

				width = Math.max(width, fieldWidth + 2);
			}
			else if (value instanceof LySwingComponent)
			{
				Component c = ((LySwingComponent)value).getComponent();

				width = Math.max(width, c.getPreferredSize().width + 2);
			}
		}

		if (_compactMode && (width == 0))
			return 15;

		// Check if a specific column width has been set on the header
		if (useColWidth && hasHeaders() && (indexColumn < getHeaders().size()))
		{
			LyHeaders headers = _table.getHeaders();
			LyHeader  header = headers.getHeader(indexColumn);

			if (header.getColumnWidth() >= 1)
				width = (int)(LySwingUtils.getFontCharWidth(tableComp, font)
						* (header.getColumnWidth() + 1));
		}

		return width;
	} // end method computeColumnPreferredWidth

	/**
	 * .
	 */

	/**
	 * Gets the row at the specified position.
	 *
	 * @param  i the row position
	 * @return the LySwingRow at position i
	 */
	public LySwingRow getRow(int i)
	{
		LyRows rows = _table.getRows();

		if (rows == null)
			return null;

		if (!rows.isHierarchical() && !rows.hasHiddenRow())
		{
			LyRow subRow = rows.getRow(i);

			if (subRow == null)
				return null;

			return ((LySwingRow)subRow.getInternalView());
		}
		else
		{
			Vector<LyRow> subRows = new Vector<LyRow>();

			rows.getDisplayedSubRows(subRows);

			if (i >= subRows.size())
				return null;

			LyRow subRow = subRows.elementAt(i);

			return ((LySwingRow)subRow.getInternalView());
		}
	}

	/**
	 * Gets the number of rows.
	 *
	 * @return the number of available rows (data)
	 */
	public int getDataRowCount()
	{
		LyRows rows = _table.getRows();

		if (rows == null)
			return 0;

		return (rows.getSubRowCount(true));
	}

	/**
	 * Returns the header of the column at the specified position.
	 *
	 * @param  i the position of the column
	 * @return the LySwingHeader at position i
	 */
	public LySwingHeader getHeader(int i)
	{
		if (_headers == null)
			getHeaders();

		return _headers.get(i);
	}

	/**
	 * Gets the list of headers for this table.
	 *
	 * @return a Vector that contains all the header objects
	 */
	public Vector<LySwingHeader> getHeaders()
	{
		if (_headers == null)
		{
			_headers = new Vector<LySwingHeader>();

			LyHeaders headers = _table.getHeaders();

			if (headers == null)
				return _headers;

			int i, n = headers.getHeaderCount();

			for (i = 0; i < n; i++)
				_headers.add((LySwingHeader)headers.getHeader(i).getInternalView());
		}

		return _headers;
	}

	/**
	 * Returns the footer of the column at the specified position.
	 *
	 * @param  i the position of the column
	 * @return the LySwingFooter at position i
	 */
	public LySwingFooter getFooter(int i)
	{
		if (_footers == null)
			getFooters();

		if ((i < 0) || (i >= _footers.size()))
			return null;

		return _footers.get(i);
	}

	/**
	 * Gets the list of footers for this table.
	 *
	 * @return a Vector that contains all the LySwingFooter objects
	 */
	public Vector<LySwingFooter> getFooters()
	{
		if (_footers == null)
		{
			_footers = new Vector<LySwingFooter>();

			LyFooters footers = _table.getFooters();

			if (footers == null)
				return _footers;

			int i, n = footers.getFooterCount();

			for (i = 0; i < n; i++)
				_footers.add((LySwingFooter)footers.getFooter(i).getInternalView());
		}
		return _footers;
	}

	/**
	 * Indicates if this table has headers.
	 *
	 * @return true if this table has some header
	 */
	public boolean hasHeaders()
	{
		LyHeaders headers = _table.getHeaders();

		return ((headers != null) && (headers.getHeaderCount() != 0));
	}

	/**
	 * Indicates if this table has footers.
	 *
	 * @return true if this table has some header
	 */
	public boolean hasFooters()
	{
		LyFooters footers = _table.getFooters();

		return ((footers != null) && (footers.getFooterCount() != 0));
	}

	/**
	 * Returns the header image for the ascending or descending properties.
	 *
	 * @param  ascending the ascending property if true, else the descending property
	 * @return the sort image
	 */
	public Image getSortImage(boolean ascending)
	{
		if (ascending)
			return _ascending;
		return _descending;
	}

	/**
	 * Gets this table selection mode. It may be one of the following:
	 *
	 * <ul>
	 *   <li>{@link LyTable#SelectionMode.NONE}</li>
	 *   <li>{@link LyTable#SelectionMode.SINGLE}</li>
	 *   <li>{@link LyTable#SelectionMode.MULTIPLE}</li>
	 * </ul>
	 *
	 * @return the selection mode
	 */
	public LyTable.SelectionMode getSelectionMode()
	{
		return _table.getSelectionMode();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.getinterface.LyGetLabelInterface#getLabel()
	 */
	public String getLabel()
	{
		return _table.getLabel();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.getinterface.LyGetValueInterface#getValue()
	 */
	public String getValue()
	{
		// The value returned by this method contains the identifiers of the headers separated by the
		// LyTools.OBJECT_SEP constant. It may be used by controllers to get information about
		// headers order that may be changed by the user in Swing tables.
		TableColumnModel model = getTableComponent().getColumnModel();
		int				 i, n = model.getColumnCount();
		StringBuffer     buffer = new StringBuffer();
		TableColumn		 column;
		LySwingHeader    swingHeader;

		for (i = 0; i < n; i++)
		{
			column = model.getColumn(i);

			if (column.getHeaderValue() instanceof LySwingHeader)
			{
				if (buffer.length() != 0)
					buffer.append(LyTools.OBJECT_SEP);

				swingHeader = (LySwingHeader)column.getHeaderValue();
				buffer.append(swingHeader.getId());
			}
		}

		return buffer.toString();
	}

	/**
	 * Indicates if the horizontal lines are drawn or not.
	 *
	 * @return the horizontal lines property
	 * @see    LyTable#isHLines()
	 */
	public boolean isHLines()
	{
		return _table.isHLines();
	}

	/**
	 * Indicates if the vertical lines are drawn or not.
	 *
	 * @return the vertical lines property
	 * @see    LyTable#isVLines()
	 */
	public boolean isVLines()
	{
		return _table.isVLines();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalObject#getController()
	 */
	public String getController()
	{
		return _table.getController();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _comp;
	}

	/**
	 * Gets the index of the specified row.
	 *
	 * @param  row the row
	 * @return the index of the specified row
	 */
	public int getIndex(LyRow row)
	{
		if (row == null)
			return -1;

		LyRows rows = _table.getRows();
		LyRow  subrow;
		int    i, count = 0, n;

		if (!rows.isHierarchical())
		{
			for (i = 0, n = rows.getRowCount(); i < n; i++)
			{
				subrow = rows.getRow(i);

				if (subrow.isHidden())
					continue;

				if (subrow == row)
					return count;

				count++;
			}
		}
		else
		{
			// If table is hierarchical, get the list of all rows and retrieve the right one
			Vector<LyRow> rowsList = new Vector<LyRow>();

			rows.getDisplayedSubRows(rowsList);

			return rowsList.indexOf(row);
		}

		return -1;
	} // end method getIndex

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.getinterface.LyIsIdleInterface#isIdle()
	 */
	public final boolean isIdle()
	{
		return _isIdle;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetIdleInterface#setIdle(boolean)
	 */
	public void setIdle(boolean status)
	{
		setIdle(status, null);
	}

	/**
	 * Sets the idle status of the table.
	 *
	 * @param status the idle status of the table.
	 * @param removedComponent component that is being removed from the table.
	 */
	protected void setIdle(boolean status, LyComponent removedComponent)
	{
		if (_isIdle == status)
			return;

		synchronized (getLock())
		{
			_isIdle = status;

			if (!_isIdle)
			{
				// Clear caches
				_headers = null;
				_footers = null;

				// If table has no header but a status column, add a header for status column
				// and an empty cell for each row to display the images
				if (_imageAndNoStatusHeaderMode)
					updateEmptyCells();

				if (_statusColumn != null)
					_statusColumn.updateContent();

				// If no header, update columns width to ensure that content will be correctly displayed
				// (since it is not possible to increase columns width with the graphical interface).
				if (_imageAndNoStatusHeaderMode || _needResizeColumns)
				{
					updateColumnWidth();
//					_needResizeColumns = false;
				}

				_skipSelectionDemand = true;
				getTableModel().fireTableDataChanged(); // Force redraw now

				// headers may be changed : we must redraw them too
				JTableHeader tableHeader = _tableComp.getTableHeader();

				tableHeader.repaint();

				// Reselect selected rows
				LyRows        rows = _table.getRows();
	            Vector<LyRow> rowsList = new Vector<LyRow>();
                LyRow         row;
                int           i, n;
	            if(rows != null)
	            {
	            	rows.getDisplayedSubRows(rowsList);
	            }

				for (i = 0, n = rowsList.size(); i < n; i++)
				{
					row = rowsList.elementAt(i);
					if (row.isSelected())
						selectRow(i, true);
				}

				_skipSelectionDemand = false;

				checkComputeHeight();
				computeRowHeights();
				cacheHeight();

				if (_statusColumn != null)
					_statusColumn.doLayout();

				// Check edition status
				checkEditionStatus(removedComponent);
			}
		}
	} // end method setIdle

	public void checkEditionStatus(LyComponent removedComponent)
	{
        LyRow  row;
		int    r = _tableComp.getEditingRow();
		int    c = _tableComp.getEditingColumn();

		if (!_tableComp.getTableModel().isCellEditable(r, c))
			_tableComp.editCellAt(-1, -1);

		Vector<LyFormField> fields = new Vector<LyFormField>();

		_table.getComponents(LyFormField.class, fields);

		if ((removedComponent != null) && fields.contains(removedComponent))
			fields.remove(removedComponent);

		if (fields.size() > 0)
		{
			LyFormField		 formField = fields.elementAt(0);
            LyComponent      fieldParent = formField.getParent();

            if (!(fieldParent instanceof LyRow))
                return;

			LyFieldItem		 fieldItem = formField.getFieldItem();
			LySwingComponent swingComponent = (LySwingComponent)fieldItem
					.getInternalView();

			row = (LyRow)fieldParent;

			int i, n = row.getComponentCount();

			for (i = 0; i < n; i++)
			{
				if (row.getComponent(i) == formField)
					break;
			}

			LySwingRow		   swingRow = (LySwingRow)row.getInternalView();
			_LyFocusRequest	   _focusRequest = swingRow.new _LyFocusRequest(swingComponent, i);
            LyViewManager	   viewManager = getViewManager();

			viewManager.execGui(_focusRequest);
		}
	}

	/**
	 * @see leon.view.setinterface.LySetEnabledInterface#setEnabled(boolean)
	 */
	public void setEnabled(boolean status)
	{
		_tableComp.setEnabled(status);

		if (_statusColumn != null)
			_statusColumn.setEnabled(status);
	}

	/**
	 * @see leon.view.graphic.LyGraphicalTable#setNRows(int)
	 */
	public void setNRows(int nRows)
	{
	}

	/**
	 * @see leon.view.setinterface.LySetTableInterface#setHeaders(LyHeaders)
	 */
	public void setHeaders(LyHeaders headers)
	{
		// Not implemented in Swing : must rebuild the table entirely
	}

	/**
	 * @see leon.view.setinterface.LySetTableInterface#LyFooters(LyFooters)
	 */
	public void setFooters(LyFooters footers)
	{
		// Not implemented in Swing : must rebuild the table entirely
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetTableInterface#addRow(leon.view.LyRow)
	 */
	public void addRow(LyRow row)
	{
		addRow(row, -1);
	}

	/**
	 * Adds the specified row at the given index in this table.
	 *
	 * @param row the row to add
	 * @param idx the position where the row must be inserted. If this position is -1, the row is
	 *            added at the end of the rows in the table.
	 * @see   leon.view.setinterface.LySetTableInterface#addRow(leon.view.LyRow, int)
	 */
	public void addRow(LyRow row, int idx)
	{
		if (isIdle()
				|| (getTableModel() == null))
			return;

		checkComputeHeight();

		if (idx == -1)
		{
			LyRows rows = _table.getRows();

			idx = rows.getRowCount() - 1;
		}

		if (_imageAndNoStatusHeaderMode)
			updateEmptyCells();

		synchronized (this)
		{
			_skipSelectionDemand = true;
			getTableModel().fireTableRowsInserted(idx, idx);
			_skipSelectionDemand = false;
		}

		selectRow(row, row.isSelected());

		if (!isIdle())
			checkComputeHeight();

		if (_tableComp.isComputeHeight())
		{
			computeRowHeights();
			cacheHeight();
		}

		// FFT00969 then removed by FFT 1147
		// updateColumnWidth();
		ensureColumnsWidth(idx);
	} // end method addRow

	/**
	 * Computes the column width based on the first elements in columns.
	 */
	public void updateColumnWidth()
	{
		if (isIdle())
			return;

		ensureColumnsWidth(-1);

		// This column is invisible since the status column is
		// managed by its own component.
		if (hasStatusColumn())
		{
			LyJTableComponent tableComp = getTableComponent();

			if (tableComp == null)
				return;

			LyJTableColumn column = (LyJTableColumn)tableComp.getColumn(STATUS_COLUMN_NAME);

			column.setPreferredWidth(0);
		}
	}

	/**
	 * Checks if the columns are wide enough for the row specified by its index.
	 *
	 * @param idx of the row we check
	 */
	public void ensureColumnsWidth(int idx)
	{
		if (isIdle())
			return;

		LyJTableComponent tableComp = getTableComponent();

		if (tableComp == null)
			return;

		TableColumnModel columnModel = tableComp.getColumnModel();
		LyJTableColumn   column;
		LyJTableModel    tableModel = getTableModel();
		int				 i, n = tableModel.getColumnCount();
		int				 preferredWidth;

		// for each column
		for (i = 0; i < n; i++)
		{
			column = (LyJTableColumn)columnModel.getColumn(i);

			// Compute and set the preferred width of this column
			preferredWidth = computeColumnPreferredWidth(tableComp, i, tableComp.getFont(), idx, true);
//			if ((idx >= 0) && (preferredWidth > column.getPreferredWidth()))
				column.setPreferredWidth(preferredWidth);
		}

		// If there is a status column, force its width to 0
		if (hasStatusColumn())
		{
			column = (LyJTableColumn)tableComp.getColumn(STATUS_COLUMN_NAME);
			column.setPreferredWidth(0);
		}
	} // end method ensureColumnsWidth

	/**
	 * Checks if a column is wide enough for the row and column specified by their index.
	 *
	 * @param rowIndex index of the row we check
	 * @param colIndex index of the column we check
	 */
	public void ensureColumnsWidth(int rowIndex, int colIndex)
	{
		if (isIdle())
			return;

		LyJTableComponent tableComp = getTableComponent();

		if (tableComp == null)
			return;

		TableColumnModel columnModel = tableComp.getColumnModel();
		LyJTableColumn   column;
		int				 preferredWidth;

		column = (LyJTableColumn)columnModel.getColumn(colIndex);

		// Compute and set the preferred width of this column
		preferredWidth = computeColumnPreferredWidth(tableComp, colIndex, tableComp.getFont(),
				rowIndex, true);

		if (preferredWidth > column.getPreferredWidth())
			column.setPreferredWidth(preferredWidth);
	} // end method ensureColumnsWidth

	/**
	 * Don't do anything.
	 *
	 * @param idx this attribute is not used in this implementation
	 * @see   leon.view.setinterface.LySetTableInterface#freezeColumn(int)
	 */
	public void freezeColumn(int idx)
	{
	}

	/**
	 * Updates the specified row. This method is not implemented yet.
	 *
	 * @see leon.view.setinterface.LySetTableInterface#updateRow(leon.view.LyRow)
	 */
	public void updateRow(LyRow row)
	{
		System.out.println(" update row on row " + row);
		// XXX TODO
	}

	/**
	 * Removes the row specified by its index in the table. If the idle mode is set, the row is not
	 * removed.
	 *
	 * @param idx the position of the row in this table
	 * @see   leon.view.setinterface.LySetTableInterface#removeRow(int)
	 */
	public void removeRow(int idx)
	{
		if (isIdle()
				|| (getTableModel() == null))
			return;

		synchronized (this)
		{
			_skipSelectionDemand = true;
			getTableModel().fireTableRowsDeleted(idx, idx);
			_skipSelectionDemand = false;

			// Check if row contains toolbars and hide them if needed.
			LyRows			  rows = _table.getRows();
			LyRow			  row = rows.getRow(idx);

			if (row == null)
				return;

			Vector<LyToolBar> toolbars = new Vector<LyToolBar>();

			// Cancel editing status if row was in edition
			int				  r = _tableComp.getEditingRow();

			if (r == idx)
				_tableComp.editCellAt(-1, -1);

			row.getComponents(LyToolBar.class, toolbars);

			int			   i, n = toolbars.size();
			LyToolBar	   toolbar;
			LySwingToolBar swingToolbar;

			for (i = 0; i < n; i++)
			{
				toolbar = toolbars.elementAt(i);
				swingToolbar = (LySwingToolBar)toolbar.getInternalView();
				swingToolbar.getComponent().setVisible(false);
			}
		}
	} // end method removeRow

	/**
	 * Removes the specified row in the table. If the idle mode is set, the row is not removed.
	 *
	 * @param row the row to remove
	 * @see   #removeRow(int)
	 * @see   leon.view.setinterface.LySetTableInterface#removeRow(leon.view.LyRow)
	 */
	public void removeRow(LyRow row)
	{
		if (isIdle())
			return;

		LyRows rows = _table.getRows();

		int    idx = rows.getRowIndex(row);

		removeRow(idx);
	}

	/**
	 * Triggers events when table cells are selected. Nothing is done if {@link
	 * #_skipSelectionDemand} is set.
	 *
	 * @see leon.view.event.LySelectionListener#selectionPerformed(leon.view.event.LySelectionEvent)
	 */
	public void selectionPerformed(LySelectionEvent selectionEvent)
	{
		if (_skipSelectionDemand || !_table.isEnabled())
			return;

		if (_table.getController() != null)
		{
			int    type_sel = selectionEvent.isSelected() ? LySelectionDemand.TYPE_SELECT
														  : LySelectionDemand.TYPE_UNSELECT;

			Object source = selectionEvent.getSource();
			String source_id;

			if (source instanceof LyGetIdInterface)
				source_id = ((LyGetIdInterface)source).getId();
			else
				source_id = source.toString();

			// Notify controller
			LySelectionDemand demand = new LySelectionDemand(type_sel, source_id);
			Object			  item = selectionEvent.getItem();

			// if (item instanceof LyComponent)
			// demand.addSelectionId(((LyComponent)item).getId());

			if (item instanceof Vector<?>)
			{
				Vector<?> list = (Vector<?>)item;

				for (int i = 0; i < list.size(); i++)
				{
					Object o = list.elementAt(i);

					if (o instanceof String)
						demand.addSelectionId((String)o);
						// else if (o instanceof LyGetIdInterface)
						// demand.addSelectionId(((LyGetIdInterface)o).getId());
				}
			}

			int[]	position = selectionEvent.getPosition();

			if ((position != null) && (position.length == 2))
			{
				demand.setAttribute("x", position[0]);
				demand.setAttribute("y", position[1]);
			}

			// check index
			int idx = 0;

			for (int i = 0; idx >= 0; i++)
			{
				idx = selectionEvent.getIndex(i);
				if (idx >= 0)
					demand.addIndex(idx);
			}
			if (selectionEvent.getSelectionSize() >= 0)
				demand.setSelectionSize(selectionEvent.getSelectionSize());

			demand.setExtend(selectionEvent.isExtend());
			demand.setSubIndex(selectionEvent.getSubIndex());
			processDemand(demand);
			selectionEvent.consume();
		}

		// may notify any listeners
		if (_selectionTransmitter != null)
			_selectionTransmitter.processSelectionEvent(selectionEvent);
	} // end method selectionPerformed

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		if (!_table.isEnabled())
			return;

		if (actionEvent.getSourceId() == null)
			actionEvent.setSourceId(_table.getId());

		if (_table.getController() != null)
		{
			// Notify controller
			LyDemand demand;

			if ((actionEvent.getType() == LyMenuDemand.TYPE_POPUP)
					|| (actionEvent.getType() == LyMenuDemand.TYPE_EXTEND_POPUP))
			{
				int[] position = (int[])actionEvent.getParameter();

				if ("_headers".equals(actionEvent.getSourceId()))
				{
					Point     componentPosition = getComponent().getLocationOnScreen();
					Dimension headersOffset = getTableComponentOffset();

					position[0] += componentPosition.x + headersOffset.width;
					position[1] += componentPosition.y;
				}
				else if (actionEvent.getType() == LyMenuDemand.TYPE_EXTEND_POPUP)
				{
					Point     componentPosition = getComponent().getLocationOnScreen();
					Dimension headersOffset = getTableComponentOffset();

					position[0] += componentPosition.x + headersOffset.width;
					position[1] += componentPosition.y + headersOffset.height;
				}

				if ("_headers".equals(actionEvent.getSourceId()))
					demand = new LyMenuDemand(actionEvent.getType(),
							_table.getId() + actionEvent.getSourceId());
				else
					demand = new LyMenuDemand(actionEvent.getType(), _table.getId());

				((LyMenuDemand)demand).setItem(actionEvent.getValue());
				((LyMenuDemand)demand).setPosition(position);
				((LyMenuDemand)demand).setSubIndex(actionEvent.getSubIndex());
				demand.addIndex(actionEvent.getIndex(0));
			}
			else if ((actionEvent.getType() >= LyDragDropDemand.TYPE_DRAG_START)
					&& (actionEvent.getType() <= LyDragDropDemand.TYPE_DRAG_OVER))
			{
				LyRows rows = _table.getRows();
				LyRow  row = null;
				int    index = actionEvent.getIndex(0);

				if (rows != null)
				{
					row = rows.getRow(index);

					if (row != null)
						actionEvent.setValue(row.getId());
				}

				demand = new LyDragDropDemand(actionEvent.getType(), _table.getId());
				((LyDragDropDemand)demand).setValue(actionEvent.getValue());
				demand.addIndex(actionEvent.getIndex(0));
			}
			else
				demand = LySwingArea.buildActionDemand(actionEvent);

			processDemand(demand);
			actionEvent.consume();
		}

		// may notify any listeners
		if (_actionTransmitter != null)
			_actionTransmitter.processActionEvent(actionEvent);
	} // end method actionPerformed

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		if (_selectionTransmitter == null)
			_selectionTransmitter = new LySimpleSelectionTransmitter();

		return (_selectionTransmitter);
	}

	/**
	 * Selects or deselects all rows, columns and cells in this table without triggering selection
	 * events.
	 *
	 * @param select     new selection status
	 * @param updateView indicates if the view must be updated. This attribute is not used in this
	 *                   implementation.
	 * @see   JTable#selectAll()
	 * @see   JTable#clearSelection()
	 */
	public void selectAll(boolean select, boolean updateView)
	{
		getTableComponent().removeLyJTableSelectionListener();

		if (select)
			getTableComponent().selectAll();
		else
			getTableComponent().clearSelection();

		getTableComponent().addLyJTableSelectionListener();
	}

	/**
	 * Selects or deselects the row specified by its index.
	 *
	 * @param idx    the position of the row in this table
	 * @param select the selection status
	 * @see   leon.view.setinterface.LySetTableInterface#selectRow(int, boolean)
	 */
	public void selectRow(int idx, boolean select)
	{
		getTableComponent().removeLyJTableSelectionListener();

		if (select)
			((DefaultListSelectionModel)getTableComponent().getSelectionModel())
				.addSelectionInterval(idx, idx);
		else
			((DefaultListSelectionModel)getTableComponent().getSelectionModel())
				.removeSelectionInterval(idx, idx);
		getTableComponent().addLyJTableSelectionListener();
	}

	/**
	 * Selects or deselects the specified row in the table. If the idle mode is set, the row is not
	 * (de)selected.
	 *
	 * @param row    the row in this table
	 * @param select the selection status
	 * @see   #selectRow(int, boolean)
	 * @see   LySetTableInterface#selectRow(LyRow, boolean)
	 */
	public void selectRow(LyRow row, boolean select)
	{
		if (isIdle())
			return;

		int idx = getIndex(row);

		if (idx != -1)
			selectRow(idx, select);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetTableInterface#showRow(int)
	 */
	public void showRow(int idx)
	{
		getTableComponent().scrollToRow(idx);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetTableInterface#hideRow(int, boolean)
	 */
	public void hideRow(int idx, boolean hide)
	{
		if (isIdle()
				|| (getTableModel() == null))
			return;

		getTableModel().fireTableDataChanged();

		if (_statusColumn != null)
			_statusColumn.updateContent();

		computeRowHeights();
	}

	/**
	 * Swaps the rows at the specified indexes. Do nothing if the idle mode is set.
	 *
	 * @param idx1 the position of the first row to swap
	 * @param idx2 the position of the second row to swap
	 * @see   leon.view.setinterface.LySetTableInterface#swapRows(int, int)
	 */
	public void swapRows(int idx1, int idx2)
	{
		if (isIdle()
				|| (getTableModel() == null))
			return;

		LyRows  rows = _table.getRows();

		LyRow   row1 = rows.getRow(idx1);
		LyRow   row2 = rows.getRow(idx2);

		boolean isSelected1 = row1.isSelected();
		boolean isSelected2 = row2.isSelected();

		getTableModel().fireTableRowsUpdated(idx1, idx1);
		getTableModel().fireTableRowsUpdated(idx2, idx2);

		selectRow(idx1, isSelected2);
		selectRow(idx2, isSelected1);
	}

	/**
	 * Updates the selection state of the rows from index1 to index2.
	 *
	 * @param index1 the beginning of the selection
	 * @param index2 the end of the selection
	 */
	public void updateSelection(int index1, int index2)
	{
		if (isIdle())
			return;

		LyRows rows = _table.getRows();

		for (int i = index1; i <= index2; i++)
		{
			LyRow row = rows.getRow(i);

			selectRow(i, row.isSelected());
		}
	}

	/**
	 * Nothing is done in this method.
	 *
	 * @param status is not used in this implementation
	 * @see   leon.view.setinterface.LySetWaitingInterface#setWaiting(boolean)
	 */
	public void setWaiting(boolean status)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyFreeInterface#free()
	 */
	public void free()
	{
		_headers = null;
		_footers = null;

		_aligns = null;
		_ascending = null;
		_descending = null;
		_tableModel = null;
		_cellRenderer = null;
		_headerRenderer = null;
		_comp = null;
		_tableComp = null;
		_footerComp = null;
		_statusColumn = null;
	}

	/**
	 * Nothing is done in this method.
	 *
	 * @see leon.view.getinterface.LyPdfInterface#generatePdf(java.lang.String)
	 */
	public void generatePdf(String fileName)
		throws IOException
	{
	}

	/**
	 * Invokes a print dialog.
	 *
	 * @param listener this attribut is not used in this implementation
	 * @see   _TablePrinter
	 * @see   leon.view.getinterface.LyPrintInterface#print(leon.view.getinterface.LyPrintListenerInterface)
	 */
	public void print(LyPrintListenerInterface listener)
	{
		new _TablePrinter();
	}

	/**
	 * Get a lock on the table to prevent concurrent access on rows.
	 *
	 * @return a valid lock for synchronized()
	 */
	public Object getLock()
	{
		Object lock;

		if (_tableComp != null)
			return _tableComp.getTreeLock();
		else
			lock = this;

		return lock;
	}

	/**
	 * *************************** Internal Classes.
	 */

	/**
	 * This class manages the print processus of the table.
	 */
	public class _TablePrinter
		implements Printable
	{

		/** The resize ratio between screen and print. */
		protected double		    _ratio = 0.7;

		/** The font used with ratio It's the JTable font resize with ratio. */
		protected Font			    _resizedFont;

		/** Flag of printjob cancel. */
		protected boolean		    _printJobCancelled = false;

		/**
		 * This array contains the first column of vertical page. Its size is the number of vertical
		 * pages.
		 */
		protected int[]			    _colStart;

		/**
		 * This array contains the last column of vertical page. Its size is the number of vertical
		 * pages.
		 */
		protected int[]			    _colEnd;

		/** The table to print. */
		protected LyJTableComponent _table;

		/** The table header to print. */
		protected JTableHeader	    _tableHeader;

		/** The table footer to print. */
		protected LyJTableFooter    _tableFooter;

		/**
		 * Flag of pre-printing processus. If pre-printing computation has been processed, it's
		 * <code>true</code>.
		 */
		protected boolean		    _pageinfoCalculated = false;

		/** The total number of pages (vertical pages * number of horizontal pages). */
		protected int			    _totalNumPages = 0;

		/** The imageable height to table printing. */
		protected double		    _tableHeightOnFullPage;

		/** The height of the table header. */
		protected double		    _headerHeight;

		/** The height of the table footer. */
		protected double		    _footerHeight;

		/** Dimension of the imageable width of the page. */
		protected double		    _pageWidth;

		/** Dimension of the imageable height of the page. */
		protected double		    _pageHeight;

		/** The height of the resized font. */
		protected int			    _fontHeight;

		/** The ascent of the resized font. */
		protected int			    _fontDescent;

		/** The total height of the table. */
		protected double		    _tableHeight;

		/** The height of a row. */
		protected double		    _rowHeight;

		/**
		 * Creates a new printer job for this table and invokes its print dialog.
		 */
		public _TablePrinter()
		{
			_table = _tableComp;
			_tableHeader = _table.getTableHeader();
			_tableFooter = _footerComp;

			PrinterJob pj = PrinterJob.getPrinterJob();

			pj.setPrintable(_TablePrinter.this);

			if (pj.printDialog())
			{
				try
				{
					pj.print();
				}
				catch (Exception PrintException)
				{
					_printJobCancelled = true;
					PrintException.printStackTrace();
				}
			}
		}

		/**
		 * Sets the ratio of the rescaling. The new ratio value must be comprised between 0.1 and 5.
		 *
		 * @param ratio the value of the ratio
		 * @see   #getRatio()
		 */
		public void setRatio(double ratio)
		{
			if ((ratio >= 0.1) && (ratio <= 5))
				_ratio = ratio;
		}

		/**
		 * Gets the rescaling ratio value.
		 *
		 * @return the ratio
		 * @see    #setRatio(double)
		 */
		public double getRatio()
		{
			return _ratio;
		}

		/**
		 * Prints a header cell depending on its dimensions. The text of the cell is truncated if it
		 * is too long for the printing job.
		 *
		 * @param  g      the target graphics
		 * @param  column the index of the cell
		 * @param  x      the x coordinate of the left border of the cell
		 * @param  y      the x coordinate of the top border of the cell
		 * @return the dimension of the cell
		 */
		public Dimension printHeaderCell(Graphics g, int column, int x, int y)
		{
			Color			 color = g.getColor();

			int				 viewColumn = _table.convertColumnIndexToModel(column);

			LyJTableModel    model = getTableModel();
			TableColumnModel columnModel = _table.getColumnModel();

			Object			 value = model.getColumnHeaderValue(viewColumn);
			String			 text = (value instanceof LySwingHeader)
				? (" " + ((LySwingHeader)value).getLabel()) : (" ");

			if ((text == null) || text.equals(""))
				text = " ";

			int width = (int)(columnModel.getColumn(column).getWidth() * _ratio);

			if (width > _pageWidth)
				width = (int)_pageWidth;

			int height = (int)_headerHeight;

			g.setColor(new Color(204, 204, 204));
			g.fillRect(x + 1, y + 1, width - 2, height - 2);
			g.setColor(Color.black);
			g.drawRect(x, y, width, height);
			g.setFont(_resizedFont);

			FontMetrics fm = g.getFontMetrics();
			int		    threePointsWidth = fm.stringWidth("...");
			int		    secureSpace = fm.stringWidth("WW");

			if (fm.stringWidth(text) >= (width - secureSpace))
			{
				while ((fm.stringWidth(text) + threePointsWidth) >= (width - 5))
					text = text.substring(0, text.length() - 1);
				text = text + "...";
			}

			g.drawString(text, x, y + ((height + g.getFontMetrics().getAscent()) / 2));
			g.setFont(_table.getFont());
			g.setColor(color);

			return new Dimension(width, height);
		} // end method printHeaderCell

		/**
		 * Prints a footer cell depending on its dimensions. The text of the cell is truncated if it
		 * is too long for the printing job.
		 *
		 * @param  g      the target graphics
		 * @param  column the index of the cell
		 * @param  x      the x coordinates of the left border of the cell
		 * @param  y      the y coordinates of the top border of the cell
		 * @return the dimension of the cell
		 */
		public Dimension printFooterCell(Graphics g, int column, int x, int y)
		{
			Color			 color = g.getColor();
			Font			 font = g.getFont();

			int				 viewColumn = _table.convertColumnIndexToModel(column);

			LyJTableModel    model = getTableModel();
			TableColumnModel columnModel = _table.getColumnModel();

			Object			 value = model.getColumnFooterValue(viewColumn);
			String			 text = (value instanceof LySwingFooter)
				? (" " + ((LySwingFooter)value).getLabel()) : (" ");

			if ((text == null) || text.equals(""))
				text = " ";

			int width = (int)(columnModel.getColumn(column).getWidth() * _ratio);

			if (width > _pageWidth)
				width = (int)_pageWidth;

			int height = (int)_footerHeight;

			g.setColor(new Color(204, 204, 204));
			g.fillRect(x + 1, y + 1, width - 2, height - 2);
			g.setColor(Color.black);
			g.drawRect(x, y, width, height);
			g.setFont(_resizedFont);

			FontMetrics fm = g.getFontMetrics();
			int		    threePointsWidth = fm.stringWidth("...");
			int		    secureSpace = fm.stringWidth("WW");

			if (fm.stringWidth(text) >= (width - secureSpace))
			{
				while ((fm.stringWidth(text) + threePointsWidth) >= (width - 5))
					text = text.substring(0, text.length() - 1);
				text = text + "...";
			}

			g.drawString(text, x, y + ((height + g.getFontMetrics().getAscent()) / 2));
			g.setColor(color);
			g.setFont(font);

			return new Dimension(width, height);
		} // end method printFooterCell

		/**
		 * Prints a table cell depending on its dimensions. The text of the cell is truncated if it
		 * is too long for the printing job.
		 *
		 * @param  g      the target graphics
		 * @param  row    the row index of the cell
		 * @param  column the column index of the cell
		 * @param  x      the x coordinates of the left border of the cell
		 * @param  y      the y coordinates of the top border of the cell
		 * @return the dimension of the cell
		 */
		public Dimension printTableCell(Graphics g, int row, int column, int x, int y)
		{
			Color			 color = g.getColor();
			Font			 font = g.getFont();

			int				 viewColumn = _table.convertColumnIndexToModel(column);

			LyJTableModel    model = getTableModel();
			TableColumnModel columnModel = _table.getColumnModel();

			Object			 value = model.getValueAt(row, viewColumn);
			String			 text = (value instanceof LySwingCell)
				? (" " + ((LySwingCell)value).getLabel()) : (" ");

			if ((text == null) || text.equals(" null"))
				text = " ";

			int width = (int)(columnModel.getColumn(column).getWidth() * _ratio);

			if (width > _pageWidth)
				width = (int)_pageWidth;
			int height = (int)_rowHeight;

			g.setColor(Color.black);
			g.drawRect(x, y, width, height);
			g.setFont(_resizedFont);

			FontMetrics fm = g.getFontMetrics();
			int		    threePointsWidth = fm.stringWidth("...");
			int		    secureSpace = fm.stringWidth("WW");

			if (fm.stringWidth(text) >= (width - 5))
			{
				while ((fm.stringWidth(text) + threePointsWidth) >= (width - secureSpace))
					text = text.substring(0, text.length() - 1);
				text = text + "...";
			}

			g.drawString(text, x, y + ((height + g.getFontMetrics().getAscent()) / 2));
			g.setColor(color);
			g.setFont(font);

			return new Dimension(width, height);
		} // end method printTableCell

		/**
		 * Prints a page. It comprises the page header message and the page's table cells
		 * sourrounded by the header and footer cells.
		 *
		 * @param g          the target graphics
		 * @param pageFormat the format of this page, this attribute is not used here
		 * @param hPageNum   the horizontal index of page
		 * @param vPageNum   the vertical index of page
		 */
		public void printTablePart(Graphics2D g, PageFormat pageFormat, int hPageNum, int vPageNum)
		{
			int		  cS = _colStart[vPageNum];
			int		  cE = _colEnd[vPageNum];

			int		  rS = (int)((_tableHeightOnFullPage / _rowHeight) * hPageNum);
			int		  rE = (int)(Math.min(_table.getTableModel().getDataRowCount() - 1,
						rS
						+ (_tableHeightOnFullPage / _rowHeight) - 1));

			Dimension dim;
			int		  xOffset = 0;
			int		  yOffset = 0;

			String    pageMessage = "Page " + (hPageNum + 1) + "." + (vPageNum + 1);

			g.drawString(pageMessage, ((int)_pageWidth / 2) - 35, (int)(_pageHeight - _fontHeight));

			int width = 0;

			for (int i = cS; i <= cE; i++)
				width += (int)(_table.getColumnModel().getColumn(i).getWidth() * _ratio);

			int offsetToCenter = (int)((_pageWidth - width) / 2);

			if (offsetToCenter >= 0)
				g.translate(offsetToCenter, 0);

			for (int c = cS; c <= cE; c++)
			{
				dim = printHeaderCell(g, c, xOffset, yOffset);

				yOffset += dim.height;

				for (int r = rS; r <= rE; r++)
				{
					dim = printTableCell(g, r, c, xOffset, yOffset);
					yOffset += dim.height;
				}

				if (_footerComp != null)
					printFooterCell(g, c, xOffset, yOffset);
				yOffset = 0;
				xOffset += dim.width;
			}
		} // end method printTablePart

		/*
		 * (non-Javadoc)
		 * @see java.awt.print.Printable#print(java.awt.Graphics, java.awt.print.PageFormat, int)
		 */
		public int print(Graphics g, PageFormat pageFormat, int pageIndex)
			throws PrinterException
		{

			if (!_pageinfoCalculated)
				getPageInfo(g, pageFormat);

			Graphics2D g2 = (Graphics2D)g;

			g2.setColor(Color.black);
			if ((pageIndex >= _totalNumPages) || _printJobCancelled)
			{
				_printJobCancelled = false;
				return NO_SUCH_PAGE;
			}

			g2.translate(pageFormat.getImageableX(), pageFormat.getImageableY());

			int hPageNum = pageIndex / (_colStart.length);
			int colIndex = pageIndex % (_colStart.length);

			printTablePart(g2, pageFormat, hPageNum, colIndex);

			return Printable.PAGE_EXISTS;
		}

		/**
		 * Computes all the parameters of the printing job.
		 *
		 * @param g          the target graphics
		 * @param pageFormat the format of the page
		 */
		public void getPageInfo(Graphics g, PageFormat pageFormat)
		{
			Font font = _table.getFont();

			_resizedFont = new Font(font.getName(), font.getStyle(),
					(int)(font.getSize() * _ratio));

			_fontHeight = g.getFontMetrics().getHeight();
			_fontDescent = g.getFontMetrics().getDescent();

			_headerHeight = (_tableHeader.getHeight() + _table.getRowMargin());
			_headerHeight *= _ratio;

			_footerHeight = (_tableFooter != null)
				? (_tableFooter.getSize().getHeight() + _table.getRowMargin()) : (0);
			_footerHeight *= _ratio;

			_pageHeight = pageFormat.getImageableHeight();
			_pageWidth = pageFormat.getImageableWidth();

			_rowHeight = _table.getRowHeight(); // + _table.getRowMargin();
			_rowHeight *= _ratio;

			_tableHeight = (_rowHeight) * getTableModel().getDataRowCount();

			_tableHeightOnFullPage = (int)(_pageHeight - _headerHeight - _footerHeight
					- (_fontHeight * 2));
			_tableHeightOnFullPage = _tableHeightOnFullPage / _rowHeight * _rowHeight;

			TableColumnModel tableColumnModel = _tableHeader.getColumnModel();
			int				 columns = tableColumnModel.getColumnCount() - 1;
			int				 columnMargin = 0; // tableColumnModel.getColumnMargin();

			int[]			 cS = new int[columns];
			int[]			 cE = new int[columns];
			int				 columnIndex = (hasStatusColumn()) ? (1) : (0);

			int				 columnWidth;

			int				 length = 0;
			int				 vPageCount = 0;

			cS[vPageCount] = columnIndex;

			while (columnIndex < columns)
			{

				columnWidth = tableColumnModel.getColumn(columnIndex).getWidth();
				columnWidth *= _ratio;

				if ((length + columnWidth + columnMargin) > _pageWidth)
				{

					if (cS[vPageCount] != columnIndex) // normal page jump
					{
						cE[vPageCount] = columnIndex - 1;
						cS[vPageCount + 1] = columnIndex;
						length = columnWidth;
					}
					else
					// this column is larger than the page
					{
						cE[vPageCount] = columnIndex;
						cS[vPageCount + 1] = columnIndex + 1;
						length = 0;
					}

					vPageCount++;
				}
				else
					length += columnWidth + columnMargin;
				columnIndex++;
			} // while

			if (length > 0)
			{ // if are more columns left, part page
				cE[vPageCount] = columnIndex - 1;
				vPageCount++;
			}

			_colStart = new int[vPageCount];
			_colEnd = new int[vPageCount];

			for (int i = 0; i < vPageCount; i++)
			{
				_colStart[i] = cS[i];
				_colEnd[i] = cE[i];
			}

			_totalNumPages = (int)(_tableHeight / _tableHeightOnFullPage);
			if ((_tableHeight % _tableHeightOnFullPage) >= _rowHeight)
			{ // at least 1 more row left
				_totalNumPages++;
			}

			_totalNumPages *= (vPageCount);

			_pageinfoCalculated = true;
		} // end method getPageInfo
	} // end class _TablePrinter

} // end class LySwingTable
