/*
 * LySwingTabs.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.plaf.*;

import leon.view.*;
import leon.view.event.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Tabs object (as described in the dtd) for swing viewer.
 *
 * <p>A container of tab type allows to display different panels in the same graphical area, only
 * one panel being visible at once. Each panel is actually contained in a tab, the container being
 * made up of a set of tabs. The displaying that corresponds to a container of tab type must present
 * two areas: a first one showing the names of the different tabs and a second one allowing to
 * display the panels.</p>
 *
 * @version $Id: LySwingTabs.java 35433 2012-05-02 12:19:54Z nhamon $
 * @author  Lyria
 * @see     LyTabs
 */
public class LySwingTabs extends LySwingContainer
	implements LySwingComponent,
		ChangeListener,
		LyGraphicalTabs,
		LySetTabInterface,
		MouseListener,
		MouseMotionListener
{
	public static final String __VERSION =
		"$Id: LySwingTabs.java 35433 2012-05-02 12:19:54Z nhamon $";

	/** Default darkening factor for uneselected tabs. */
	public static float		   DARKER_FACTOR = 0.9F;

	/** The view independent tabs description. */
	protected LyTabs		   _tabs;

	/** The swing component that display the tabs. */
	protected LyJTabbedPane	   _tabPanel;

	/** Keep darker color for optimization for unselected tabs. */
	protected Color			   _darkerColor;

	/**
	 * Class constructor of a Tabs object for the Swing target.
	 *
	 * @param tabs a TABS component defined for every type of view targets. This tabs container is
	 *             initialized specificaly for the Swing view.
	 */
	public LySwingTabs(LyTabs tabs)
	{
		super(tabs);
		_tabs = tabs;

		initialize();
		setContent();
		_tabPanel.addMouseListener(this);
		_tabPanel.addMouseMotionListener(this);
	}

	/**
	 * Initializes the tab panel component with the default colors.
	 */
	public void initialize()
	{
		Color col = getColor("LY_FORM_BACKGROUND");

		if (col == null)
			col = getColor("LY_DEFAULT_BACKGROUND");

		_tabPanel = createTabPanel(col);
	}

	/**
	 * Fills this tabs with their associated panels and registers their event listeners.
	 */
	public void setContent()
	{
		int    i, n = _tabs.getTabCount();
		Color  bgcol = getColor("LY_FORM_BACKGROUND");
		Color  fgcol;
		String selectedTab = _tabs.getSelectedTab();

		if (bgcol == null)
			bgcol = getColor("LY_DEFAULT_BACKGROUND");

		/* Add all the panels */
		for (i = 0; i < n; i++)
		{
			LyTab			 tab = _tabs.getTab(i);
			LyPanel			 panel = tab.getPanel();

			LySwingComponent internal = (LySwingComponent)panel.getInternalView();

			fgcol = getColor(tab.getColor());
			addPanel(internal, -1, tab.getLabel(), tab.getImage(), bgcol, fgcol);

			if (selectedTab != null)
			{
				// FFT00957 initial selection.
				if (selectedTab.equals(tab.getLabel()))
					selectTab(tab.getLabel());
			}

			if (!tab.isEnabled())
				setEnabled(tab.getLabel(), false);

			/* register for events */
			if (internal instanceof LyEventTransmitter)
			{
				LyActionTransmitter transmitter = ((LyEventTransmitter)internal)
						.getActionTransmitter();

				transmitter.addActionListener(this);
			}
		}
	} // end method setContent

	/**
	 * Creates the tabbed panel.
	 *
	 * @param  color the background color
	 * @return the new tab panel
	 */
	public LyJTabbedPane createTabPanel(Color color)
	{
		if (color != null)
		{
			UIManager.put("TabbedPane.selected", color);
			UIManager.put("TabbedPane.selectHighlight", color.darker());
			UIManager.put("TabbedPane.lightHighlight", color.darker());
			UIManager.put("TabbedPane.contentAreaColor", color);
		}

		LyJTabbedPane	tabPanel;
		Font			font = getFont("LY_TABS_FONT");

		tabPanel = new LyJTabbedPane();
		tabPanel.setBackground(color);

		if (color != null)
			_darkerColor = new Color((int)(color.getRed() * DARKER_FACTOR),
					(int)(color.getGreen() * DARKER_FACTOR),
					(int)(color.getBlue() * DARKER_FACTOR));

		if (font != null)
			tabPanel.setFont(font);

		if (_tabs.isClosable())
		{
			Image		closeImage = getImage("LY_CLOSE_TAB");
			Image		closeOverImage = getImage("LY_CLOSE_TAB_OVER");
			
			tabPanel.setCloseImage(closeImage, closeOverImage);
		}
		
		tabPanel.setBorder(BorderFactory.createEmptyBorder(0, 0, 1, 2));
		
		return tabPanel;
	}

	/**
	 * Selects the specified tab in this component.
	 *
	 * @param tabId the id or label of the tab that must be selected
	 * @see   #selectTab(String)
	 * @see   leon.view.graphic.LyGraphicalTabs#setSelectedTab(java.lang.String)
	 */
	public void setSelectedTab(String tabId)
	{
		selectTab(tabId);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalTabs#addTab(leon.view.LyTab, int)
	 */
	public void addTab(LyTab tab, int pos)
	{
		LyPanel			 panel = tab.getPanel();
		LySwingComponent internal = (LySwingComponent)panel.getInternalView();
		Color			 bgcol = getColor("LY_FORM_BACKGROUND");
		Color			 fgcol;

		if (bgcol == null)
			bgcol = getColor("LY_DEFAULT_BACKGROUND");
		fgcol = getColor(tab.getColor());
		addPanel(internal, pos, tab.getLabel(), tab.getImage(), bgcol, fgcol);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalTabs#removeTab(leon.view.LyTab)
	 */
	public void removeTab(LyTab tab)
	{
		LyPanel			 panel = tab.getPanel();
		LySwingComponent internal = (LySwingComponent)panel.getInternalView();

		removePanel(internal);
	}

	/**
	 * Associates a panel to the specified tab.
	 *
	 * @param component the panel to add
	 * @param label     the name of the panel tab
	 * @param bgcol     the background color of the panel
	 * @param fgcol     the foreground color of the panel
	 */
	public void addPanel(LySwingComponent component, String label, Color bgcol, Color fgcol)
	{
		addPanel(component, -1, label, null, bgcol, fgcol);
	}

	/**
	 * Associates a panel to the specified tab.
	 *
	 * @param component the panel to add
	 * @param label     the name of the panel tab
	 * @param bgcol     the background color of the panel
	 * @param fgcol     the foreground color of the panel
	 */
	public void addPanel(LySwingComponent component, int pos, String label, Color bgcol, Color fgcol)
	{
		addPanel(component, pos, label, null, bgcol, fgcol);
	}

	/**
	 * Associates a panel to the specified tab.
	 *
	 * @param component the panel to add
	 * @param label     the name of the panel tab
	 * @param bgcol     the background color of the panel
	 * @param fgcol     the foreground color of the panel
	 */
	public void addPanel(LySwingComponent component, int pos, String label, String image, Color bgcol, Color fgcol)
	{
		JTabbedPane tabPanel = _tabPanel;
		Component   panel;

		// FFT00557
		// FFT01391: The internal area must be scrollable (the form may not fit in the dialog)
		// if(component instanceof LySwingArea)
		// {
		// ((LySwingArea)component).setScrollingEnabled(false);
		// }

		panel = component.getComponent();

		label = translate(label);

		if ((label != null) && (label.endsWith("\\n")))
			label = label.substring(0, label.length() - 2);

		if ((pos < 0) || (pos >= tabPanel.getTabCount()))
			pos = tabPanel.getTabCount();

		ImageIcon	icon = null;
		
		if (image != null)
		{
			Image	im = getImage(image);

			icon = new ImageIcon(im);
		}
		
		tabPanel.insertTab(label, icon, panel, null, pos);
		
		_tabPanel.setAddingPanel(true);

		if (bgcol != null)
		{
			if (bgcol.equals(panel.getBackground()))
				bgcol = _darkerColor;
			
			tabPanel.setBackgroundAt(pos, bgcol);
		}
		else
			tabPanel.setBackgroundAt(pos, _darkerColor);

		_tabPanel.setAddingPanel(false);

		if (fgcol != null)
			tabPanel.setForegroundAt(pos, fgcol);

		if (tabPanel.getTabCount() == 1)
			tabPanel.addChangeListener(this);
	} // end method addPanel

	/**
	 * Removes a panel from the tabbed panel.
	 *
	 * @param component the panel to remove
	 */
	public void removePanel(LySwingComponent component)
	{
		JTabbedPane tabPanel = _tabPanel;
		Component   panel;

		panel = component.getComponent();
		tabPanel.remove(panel);
		if (tabPanel.getTabCount() == 0)
			tabPanel.removeChangeListener(this);
	}

	/**
	 * Selects the specified tab.
	 *
	 * @param tabId the id or label of the tab that must be selected
	 */
	protected void selectTab(String tabId)
	{
		if (tabId == null)
			return;

		JTabbedPane tabPanel = _tabPanel;
		boolean     found = false;
		int		    currentTab = tabPanel.getSelectedIndex();
		String	    currentTitle = _tabs.getTab(currentTab).getLabel();

		tabId = translate(tabId);
		currentTitle = translate(currentTitle);

		// If we try to select the currently selected tab, return
		if (tabId.equals(currentTitle))
			return;
		
		int    i, n = _tabs.getTabCount();
		LyTab  tab;
		String label, id;

		for (i = 0; i < n; i++)
		{
			tab = _tabs.getTab(i);
			label = translate(tab.getLabel());
			id = tab.getId();
			
			if (((id != null) && id.equals(tabId))
					|| ((label != null) && label.equals(tabId)))
			{
				found = true;
				break;
			}
		}

		if (found)
			tabPanel.setSelectedIndex(i);
	} // end method selectTab

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.swing.event.ChangeListener#stateChanged(javax.swing.event.ChangeEvent)
	 */
	public void stateChanged(ChangeEvent e)
	{
		LyComponent parent = _tabs.getParent();
		JTabbedPane tabPanel = (JTabbedPane)e.getSource();
		int		    index = tabPanel.getSelectedIndex();

		if ((parent instanceof LyArea)
            && (index >= 0)
            && (index < _tabs.getTabCount()))
		{
			LyActionDemand demand = new LyActionDemand(LyActionDemand.TYPE_ACTION, "_tabs");

			// attempt to get the value
			demand.setValue(_tabs.getTab(index).getId());

			// check index
			demand.addIndex(index);

			processDemand(((LyArea)parent).getController(), demand);
		}
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
	 */
	public void mouseDragged(MouseEvent e)
	{
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseMotionListener#mouseMoved(java.awt.event.MouseEvent)
	 */
	public void mouseMoved(MouseEvent e)
	{
		TabbedPaneUI	ui = _tabPanel.getUI();
		
		if (ui instanceof LyJMetalTabbedPaneUI)
		{
			int		index = ((LyJMetalTabbedPaneUI)ui).getCloseButtonAtPosition(e.getPoint());
			
			((LyJMetalTabbedPaneUI)ui).setMouseOverCloseButton(index);
		}
		else if (ui instanceof LyJTabbedPaneUI)
		{
			int		index = ((LyJTabbedPaneUI)ui).getCloseButtonAtPosition(e.getPoint());
			
			((LyJTabbedPaneUI)ui).setMouseOverCloseButton(index);
		}
		else if (ui instanceof LyJWindowsTabbedPaneUI)
		{
			int		index = ((LyJWindowsTabbedPaneUI)ui).getCloseButtonAtPosition(e.getPoint());
			
			((LyJWindowsTabbedPaneUI)ui).setMouseOverCloseButton(index);
		}
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
	 */
	public void mouseClicked(MouseEvent e)
	{
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
	 */
	public void mouseEntered(MouseEvent e)
	{
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
	 */
	public void mouseExited(MouseEvent e)
	{
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
	 */
	public void mousePressed(MouseEvent e)
	{
		TabbedPaneUI	ui = _tabPanel.getUI();
		int				index = -1;
		
		if (ui instanceof LyJMetalTabbedPaneUI)
			index = ((LyJMetalTabbedPaneUI)ui).getCloseButtonAtPosition(e.getPoint());
		else if (ui instanceof LyJTabbedPaneUI)
			index = ((LyJTabbedPaneUI)ui).getCloseButtonAtPosition(e.getPoint());
		else if (ui instanceof LyJWindowsTabbedPaneUI)
			index = ((LyJWindowsTabbedPaneUI)ui).getCloseButtonAtPosition(e.getPoint());
					
		if (index >= 0)
		{
			LyComponent parent = _tabs.getParent();

			if ((parent instanceof LyArea)
	            && (index >= 0)
	            && (index < _tabs.getTabCount()))
			{
				LyActionDemand demand = new LyActionDemand(LyActionDemand.TYPE_CLOSE_TAB, "_tabs");

				// attempt to get the value
				demand.setValue(_tabs.getTab(index).getId());

				// check index
				demand.addIndex(index);

				processDemand(((LyArea)parent).getController(), demand);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	public void mouseReleased(MouseEvent e)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetTabInterface#setEnabled(java.lang.String, boolean)
	 */
	public void setEnabled(String tabId, boolean enabled)
	{
		if (tabId == null)
			return;

		JTabbedPane tabPanel = _tabPanel;
		boolean     found = false;
		int		    i, n = _tabs.getTabCount();
		LyTab	    tab;
		String	    label, id;

		tabId = translate(tabId);

		for (i = 0; i < n; i++)
		{
			tab = _tabs.getTab(i);
			label = translate(tab.getLabel());
			id = tab.getId();
			
			if (((id != null) && id.equals(tabId))
					|| ((label != null) && label.equals(tabId)))
			{
				found = true;
				break;
			}
		}

		if (found)
			tabPanel.setEnabledAt(i, enabled);
	}
	
	/**
	 * This method allows to change the label of a specific tab.
	 *
	 * @param tabId the label or identifier of the tab to be changed
	 * @param newLabel the new label of the tab to be changed
	 */
	public void setLabel(String tabId, String newLabel)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _tabPanel;
	}
	
	/**
	 * (non-Javadoc)
	 * @see leon.view.win.swing.LySwingObject#getChildrenBounds(leon.view.LyComponent)
	 */
	public int[] getChildrenBounds(LyComponent children)
	{
		if (!(children instanceof LyTab))
			return super.getChildrenBounds(children);
		
		LyTab			tab = (LyTab)children;
		JTabbedPane		tabPanel = _tabPanel;
		TabbedPaneUI	ui = tabPanel.getUI();
		Rectangle		bounds = null;
		String          label = tab.getLabel();
		
		label = translate(label);
		
		if (ui instanceof LyJMetalTabbedPaneUI)
			bounds = ((LyJMetalTabbedPaneUI)ui).getLabelBounds(label);
		else if (ui instanceof LyJTabbedPaneUI)
			bounds = ((LyJTabbedPaneUI)ui).getLabelBounds(label);
		else if (ui instanceof LyJWindowsTabbedPaneUI)
			bounds = ((LyJWindowsTabbedPaneUI)ui).getLabelBounds(label);
		else
			return null;
		
		if (bounds == null)
			return null;
		
		return new int[] {bounds.x, bounds.y, bounds.width, bounds.height};
	}
} // end class LySwingTabs
