/*
 * LySwingText.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;
import leon.view.win.swing.text.*;

import java.awt.*;
import java.awt.event.*;

import java.util.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;


/**
 * This class implements the Text object (as described in the dtd) for swing viewer.
 *
 * <p>An input field of Text type allows to display a graphical input or text consultation
 * component. It does not comprise any element. It is possible to specify if the text should be
 * hidden, if it should respect a given format, if it can be divided into several lines and the
 * maximum number of characters that it can contain.</p>
 *
 * @version $Id: LySwingText.java 34980 2012-03-22 18:14:40Z jlsaouli $
 * @author  Lyria
 * @see     LyText
 */
public class LySwingText extends LySwingFieldItem implements LySwingComponent, FocusListener,
    KeyListener, DocumentListener, LyActionListener, LySetLabelInterface, LySetEnabledInterface,
    LySetEditableInterface, LyGraphicalText
{
    public static final String __VERSION =
        "$Id: LySwingText.java 34980 2012-03-22 18:14:40Z jlsaouli $";

    /**
     * Change the default key binding.
	 * When set to true: TAB allows to change the focus field, ENTER validate the form,
	 * SHIFT+TAB allows to insert a tab and SHIFT+ENTER to insert a line break.
	 * When set to ALTERNATE: TAB allows to change the focus field and CTRL+TAB
	 * add a tab character.
     */
    public static String	CHANGE_KEY_MAPPING = "false";
    
    /** The swing text component used to display this object. */
    protected JComponent _textComp;

    /** The scroll panel that contains the text component. */
    protected JComponent _scrollPane;

    /** Indicates if this text component gained the keyboard focus. */
    protected boolean _hasGainedFocus;

    /** Distinguishes between Leon setLabel and component setLabel. */
    protected boolean _lockSetLabel;

    /** Viewer used for auto-completion. */
    protected LyJAutoCompletionViewer _autoCompletionViewer;

    /** The view independent text description. */
    protected LyText _text;

    /**
     * Class constructor of a Text object for the Swing target.
     *
     * @param text a TEXT component defined for every type of view targets. This text is initialized
     *             specificaly for the Swing view.
     */
    public LySwingText(LyText text)
    {
        super(text);
        _text = text;

        _autoCompletionViewer = null;

        initialize();
        setContent();
    }

    /**
     * Builds the text component and places it in the scroll panel if needed.
     */
    public void initialize()
    {
        String label = _text.getLabel();

        _hasGainedFocus = false;
        _lockSetLabel = false;

        int nRows, nCols;

        if (label == null)
            label = "";

        LyEnvironment env = getEnvironment();

        if (_text.getNRows() > 1)
        {
            int widthOffset = 0;
            int heightOffset = 3;
            int hScroll = ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER;

            nCols = ((_text.getNCols() >= 0) ? _text.getNCols() : 0);
            nRows = _text.getNRows();

            if (!_text.isWrap())
            {
                _textComp = new LyJTextArea(new LyJTextAreaDocument(_text.getNChars()), label,
                        nRows, nCols, env)
                    {
                        private static final long serialVersionUID = 3189160543315873473L;

                        /*
                         * (non-Javadoc)
                         * @see javax.swing.JTextArea#getColumnWidth()
                         */
                        protected int getColumnWidth()
                        {
                            return 8;
                        }
                    };
                hScroll = ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED;
                widthOffset = 3;
                heightOffset = 18;
            }
            else
            {
                _textComp = new LyJTextArea(new LyJTextAreaDocument(_text.getNChars()), label,
                        nRows, nCols, env)
                    {
                        private static final long serialVersionUID = 731942073228406814L;

                        /*
                         * (non-Javadoc)
                         * @see javax.swing.JTextArea#getColumnWidth()
                         */
                        protected int getColumnWidth()
                        {
                            return 8;
                        }
                    };
                ((JTextArea) _textComp).setLineWrap(true);
                ((JTextArea) _textComp).setWrapStyleWord(true);
            }

            Font        f = getFont("LY_DEFAULT_FONT");
            FontMetrics metrics = _textComp.getFontMetrics(f);

            _scrollPane = new _LyJScrollPane(_textComp,
                    ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, hScroll);

            ((JScrollPane) _scrollPane).setMinimumSize(new Dimension((nCols * 8) + widthOffset,
                    (nRows * metrics.getHeight()) + heightOffset));

            ((LyJTextArea) _textComp).setMinimumSize(new Dimension((nCols * 8) + widthOffset,
                    (nRows * metrics.getHeight()) + heightOffset));
        }
        else
        {

            if ((_text.getFormat() != null) && (_text.getFormat().length() > 0))
            {
                Image up = getImage("LY_ARROW_UP");
                Image down = getImage("LY_ARROW_DOWN");

                _textComp = new LyJFormattedInput(label, _text.getFormat(), up, down, env,
                        _text.isShowArrows());
            }
            else
            {
                nCols = LyText.DEFAULT_NCOLS;

                if (label.length() > nCols)
                    nCols = label.length();

                if (_text.getNCols() > 0)
                    nCols = _text.getNCols();

                if (_text.isSecret())
                {
                    _textComp = new JPasswordField(new LyJTextFieldDocument(_text.getNChars()),
                            label, nCols)
                        {
                            private static final long serialVersionUID = -2474071998904678668L;

                            /*
                             * (non-Javadoc)
                             * @see javax.swing.JTextField#getColumnWidth()
                             */
                            public int getColumnWidth()
                            {
                                return 8;
                            }
                        };

                    Insets insets = ((JPasswordField) _textComp).getMargin();

                    insets = new Insets(insets.top, insets.left + 1, insets.bottom,
                            insets.right + 1);
                    ((JPasswordField) _textComp).setMargin(insets);

                    ((JPasswordField) _textComp).setBorder(BorderFactory.createCompoundBorder(
                            BorderFactory.createMatteBorder(1, 1, 1, 1,
                                UIManager.getColor("TextField.darkShadow")),
                            BorderFactory.createEmptyBorder(1, 2, 1, 2)));
                    ((JPasswordField) _textComp).setEchoChar('*');
                }
                else
                    _textComp = new LyJTextField(new LyJTextFieldDocument(_text.getNChars()), label,
                            nCols, env)
                        {
                            private static final long serialVersionUID = -1930187279601817420L;

                            /*
                             * (non-Javadoc)
                             * @see javax.swing.JTextField#getColumnWidth()
                             */
                            public int getColumnWidth()
                            {
                                return 8;
                            }
                        };
            }
        }

        if (_textComp != null)
        {
            if ("true".equals(CHANGE_KEY_MAPPING))
                changeKeyMapping();
            else if ("ALTERNATE".equals(CHANGE_KEY_MAPPING))
                alternateKeyMapping();

            // FST00035 FONT redefinition
            Font f = getFont("LY_DEFAULT_FONT");

            _textComp.setFont(f);

            if (_textComp instanceof JTextComponent)
            {
                JTextComponent textComponent = (JTextComponent) _textComp;
                String         text = textComponent.getText();

                if (text != null)
                    textComponent.setCaretPosition(text.length());
            }
        }

        // Manage auto-completion
        if ((_textComp instanceof JTextComponent) && (_text.getAutoCompletionValues() != null)
                && (_text.getAutoCompletionValues().length() > 0))
        {
            String autoCompletionString = _text.getAutoCompletionValues();
            int    autoCompletionMax = -1;
            int    pos = autoCompletionString.indexOf(LyTools.PIPE_SEP);
            String autoCompletionValue = autoCompletionString.substring(0, pos);

            // If a maximum number of values is given, remove it from list of values
            if (pos > 0)
            {
                autoCompletionMax = LyTools.intFromString(autoCompletionValue);
                autoCompletionString = autoCompletionString.substring(pos + 1);
            }

            String[] values = LyString.stringToList(autoCompletionString, LyTools.VALUE_SEP);

            // Create the auto-completion viewer for text component
            _autoCompletionViewer = new LyJAutoCompletionViewer((JTextComponent) _textComp, null,
                    values, values, values, false, autoCompletionMax, getEnvironment());
        }
    } // end method initialize

    /**
     * Registers keyboard events for this text component.
     */
    public void setContent()
    {

        if (_textComp instanceof LyEventTransmitter)
        {
            LyActionTransmitter actionTransmitter = ((LyEventTransmitter) _textComp)
                .getActionTransmitter();

            if (actionTransmitter != null)
                actionTransmitter.addActionListener(this);
        }
        else if (_textComp instanceof JTextComponent)
        {
            ((JTextComponent) _textComp).addFocusListener(this);
            ((JTextComponent) _textComp).addKeyListener(this);

            if (_text.isNotifyValueChanged())
                ((JTextComponent) _textComp).getDocument().addDocumentListener(this);

            initPatterns();
        }

        // This handler is used to validate the form using ENTER key or close using ESCAPE key
        if (_textComp != null)
        {
            LySwingFormKeyHandler keyHandler = new LySwingFormKeyHandler(getActionTransmitter());

            keyHandler.setNotifyValidate(false); // already default in swing

            if (_textComp instanceof JTextArea)
            {
                if (!"ALTERNATE".equals(CHANGE_KEY_MAPPING))
                    keyHandler.setNotifyTab(false);
                else
                    keyHandler.setIgnoreCtrlTab(true);

                // Check if we are in a table
                LyComponent parent = _text.getParent();

                while ((parent != null) && !(parent instanceof LyRow))
                    parent = parent.getParent();

                if (parent instanceof LyRow)
                {
                    Set<AWTKeyStroke> newForwardKeys = new HashSet<AWTKeyStroke>(1);
                    //newForwardKeys.add(AWTKeyStroke.getAWTKeyStroke(KeyEvent.VK_TAB,0));
                    _textComp.setFocusTraversalKeys(KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS,
                        Collections.unmodifiableSet(newForwardKeys));

                    // affecter nos touches alternatives (dplacement arrire)
                    Set<AWTKeyStroke> newBackwardKeys = new HashSet<AWTKeyStroke>(1);
                    //newBackwardKeys.add(AWTKeyStroke.getAWTKeyStroke(KeyEvent.VK_TAB,KeyEvent.SHIFT_MASK+KeyEvent.SHIFT_DOWN_MASK));
                    _textComp.setFocusTraversalKeys(KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS,
                        Collections.unmodifiableSet(newBackwardKeys));
                }
            }

            _textComp.addKeyListener(keyHandler);
        }
    } // end method setContent

    /**
     * Initializes patterns list.
     */
    protected void initPatterns()
    {

        // Initialize patterns list.
        if (_textComp instanceof JTextComponent)
        {
            JTextComponent      jTextComp = (JTextComponent) _textComp;
            LyJTextAreaDocument d = (LyJTextAreaDocument) jTextComp.getDocument();
            int                 n = _text.getCharTypeCount();

            if (n > 0)
            {
                LyEnvironment      environment = _text.getEnvironment();
                LyCharType         charType;
                String             value;
                Collection<String> patterns;

                patterns = new ArrayList<String>(n);

                for (int i = 0; i < n; i++)
                {
                    charType = _text.getCharType(i);
                    value = charType.getValue();
                    value = environment.getString(value);
                    patterns.add(value);
                }

                d.setPatterns(patterns);
                d.setEnvironment(environment);
            }
        }
    } // end method initPatterns


    /**
     * Changes the default key binding for a textArea.
     */
    private void alternateKeyMapping()
    {
        if (!(_textComp instanceof LyJTextArea))
            return;

        InputMap  map = _textComp.getInputMap();
        ActionMap actionMap = _textComp.getActionMap();
        KeyStroke stroke;
        InputMap  parent;

        Set<? extends AWTKeyStroke> newForwardKeys = new HashSet<AWTKeyStroke>(1);
        
        _textComp.setFocusTraversalKeys(KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS,
            Collections.unmodifiableSet(newForwardKeys));
        stroke = KeyStroke.getKeyStroke(KeyEvent.VK_TAB, java.awt.Event.CTRL_MASK);
        map.remove(stroke);
        parent = map.getParent();

        while (parent != null)
        {
            parent.remove(stroke);
            parent = parent.getParent();
        }

        map.put(stroke, "insert-tab");

        stroke = KeyStroke.getKeyStroke(KeyEvent.VK_TAB, 0);

        parent = map.getParent();

        while (parent != null)
        {
            parent.remove(stroke);
            parent = parent.getParent();
        }

        Action      action = new AbstractAction("navigation-out-forward")
            {
                private static final long serialVersionUID = -6465023027672382784L;

                /*
                 * (non-Javadoc)
                 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
                 */
                public void actionPerformed(ActionEvent e)
                {
                    javax.swing.FocusManager.getCurrentManager().focusNextComponent(_textComp);
                }
            };

        LyComponent compParent = _text.getParent();

        while ((compParent != null) && !(compParent instanceof LyRow))
            compParent = compParent.getParent();

        if (compParent instanceof LyRow)
            map.remove(stroke);
        else
        {
            actionMap.put("navigation-out-forward", action);
            map.put(stroke, "navigation-out-forward");
        }

    }

    /**
     * Changes the default key binding for a textArea.
     */
    private void changeKeyMapping()
    {

        if (!(_textComp instanceof LyJTextArea))
            return;

        InputMap  map = _textComp.getInputMap();
        ActionMap actionMap = _textComp.getActionMap();
        KeyStroke stroke;
        InputMap  parent;

        stroke = KeyStroke.getKeyStroke(KeyEvent.VK_TAB, 0);

        parent = map.getParent();

        while (parent != null)
        {
            parent.remove(stroke);
            parent = parent.getParent();
        }

        Action action = new AbstractAction("navigation-out-forward")
            {
                private static final long serialVersionUID = -6465023027672382784L;

                /*
                 * (non-Javadoc)
                 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
                 */
                public void actionPerformed(ActionEvent e)
                {
                    javax.swing.FocusManager.getCurrentManager().focusNextComponent(_textComp);
                }
            };

        actionMap.put("navigation-out-forward", action);
        map.put(stroke, "navigation-out-forward");
        //map.remove(stroke);

        stroke = KeyStroke.getKeyStroke(KeyEvent.VK_TAB, java.awt.Event.CTRL_MASK);

        parent = map.getParent();

        while (parent != null)
        {
            parent.remove(stroke);
            parent = parent.getParent();
        }

        map.put(stroke, "insert-tab");

//              stroke = KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0);
//
//              parent = map.getParent();
//              while (parent != null)
//              {
//                      parent.remove(stroke);
//                      parent = parent.getParent();
//              }
//
//              // map.put(stroke, "navigation-out ");
//              map.remove(stroke);
//
//              stroke = KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, java.awt.Event.SHIFT_MASK);
//
//              parent = map.getParent();
//              while (parent != null)
//              {
//                      parent.remove(stroke);
//                      parent = parent.getParent();
//              }
//
//              map.put(stroke, "insert-break");
    } // end method changeKeyMapping

    /**
     * (non-Javadoc).
     *
     * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
     */
    public void actionPerformed(LyActionEvent actionEvent)
    {

        if (actionEvent.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
        {
            String value = actionEvent.getValue();

            // Update parent
            _text.setLabel(value);

            // may notify any listeners
            if (_actionTransmitter != null)
                _actionTransmitter.processActionEvent(actionEvent);
        }

        if (actionEvent.getType() == LyActionDemand.TYPE_COMPLETE_VALUE)
        {

            // Transmit to parent
            if (_actionTransmitter != null)
                _actionTransmitter.processActionEvent(actionEvent);
        }
    }

    /**
     * Nothing is done when a key is typed.
     *
     * @param e discarded key event
     * @see   java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
     */
    public void keyTyped(KeyEvent e)
    {
    }

    /**
     * Keeps the focus when the user pressed the Enter key and it is necessary to keep it.
     *
     * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
     */
    public void keyPressed(KeyEvent e)
    {

        if (!(_textComp instanceof LyJTextArea) && (e.getKeyCode() == KeyEvent.VK_ENTER)
                && !e.isShiftDown())
            focusLost(null);
    }

    /**
     * Nothing is done when a key is released.
     *
     * @param e discarded key event
     * @see   java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
     */
    public void keyReleased(KeyEvent e)
    {
        LyComponent parent = _text.getParent();
        boolean     isInTable = false;

        while (parent != null)
        {

            if (parent instanceof LyRow)
            {
                isInTable = true;
                break;
            }

            parent = parent.getParent();
        }

        // If text in in a table, first send the new value to the control layer.
        if (isInTable && (e.getKeyCode() == KeyEvent.VK_TAB))
            focusLost(null);

        // MAZ : SITA FFT00036 & FFT00037
        if (isInTable && (e.getKeyCode() == KeyEvent.VK_ESCAPE))
            focusLost(null);
    }

    /**
     * Nothing is done when document events are triggered at a changed property.
     *
     * @param e discarded document event
     * @see   javax.swing.event.DocumentListener#changedUpdate(javax.swing.event.DocumentEvent)
     */
    public void changedUpdate(DocumentEvent e)
    {
    }

    /**
     * Restores the text component at every insertion update.
     *
     * @see javax.swing.event.DocumentListener#insertUpdate(javax.swing.event.DocumentEvent)
     */
    public void insertUpdate(DocumentEvent e)
    {
        _hasGainedFocus = true;
        _lockSetLabel = true;
        focusLost(null);
        _lockSetLabel = false;
    }

    /**
     * Restores the text component at every removing update.
     *
     * @see javax.swing.event.DocumentListener#removeUpdate(javax.swing.event.DocumentEvent)
     */
    public void removeUpdate(DocumentEvent e)
    {
        _hasGainedFocus = true;
        _lockSetLabel = true;
        focusLost(null);
        _lockSetLabel = false;
    }

    /**
     * Updates the text value in the given focus event.
     *
     * @param value the new value of the text component.
     * @param e     the focus event for the components that gain and lost focus.
     */
    protected void updateText(String value, FocusEvent e)
    {
        String label = _text.getLabel();

        if (value.length() == 0)
            value = null; // Avoid useless events

        if ((label == value) || ((label != null) && (value != null) && label.equals(value)))
            return;

        // Update text
        _text.setLabel(value);

        if ((_autoCompletionViewer == null) || (e == null)
                || !_autoCompletionViewer.componentIsMine(e.getOppositeComponent()))
        {

            if (_actionTransmitter != null)
                _actionTransmitter.processActionEvent(new LyActionEvent(this, null, value,
                        LyActionDemand.TYPE_VALUE_CHANGED));
        }
    }

    public LyJAutoCompletionViewer getAutoCompletionViewer()
    {
        return _autoCompletionViewer;
    }

    /**
     * (non-Javadoc).
     *
     * @see java.awt.event.FocusListener#focusGained(java.awt.event.FocusEvent)
     */
    public void focusGained(FocusEvent e)
    {
        _hasGainedFocus = true;

        if (_textComp instanceof JTextComponent)
        {
            String text = ((JTextComponent) _textComp).getText();

            updateText(text, e);
        }
    }

    /**
     * (non-Javadoc).
     *
     * @see java.awt.event.FocusListener#focusLost(java.awt.event.FocusEvent)
     */
    public void focusLost(FocusEvent e)
    {

        if (!_hasGainedFocus)
            return;

        _hasGainedFocus = false;

        String value = null;

        if (_textComp instanceof JTextComponent)
        {
            JTextComponent textComponent = (JTextComponent) _textComp;

            value = textComponent.getText();

            int start = textComponent.getSelectionStart();
            int end = textComponent.getSelectionEnd();
            int caret = textComponent.getCaretPosition();

            // check if something has changed
            updateText(value, e);

            // If there was a selection, just put it back
            String text = textComponent.getText();

            if (text != null)
                caret = Math.min(caret, text.length());
            else
                caret = 0;

            textComponent.setCaretPosition(caret);

            if (start != end)
            {
                textComponent.setSelectionStart(start);
                textComponent.setSelectionEnd(end);
            }
        }
    } // end method focusLost

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
     */
    public void setLabel(String labelNew)
    {

        if (_lockSetLabel)
            return;

        if (_textComp instanceof JTextComponent)
        {
            ((JTextComponent) _textComp).setText(labelNew);

            if (labelNew != null)
                ((JTextComponent) _textComp).setCaretPosition(labelNew.length());
            else
                ((JTextComponent) _textComp).setCaretPosition(0);
        }
        else if (_textComp instanceof LyJFormattedInput)
            ((LyJFormattedInput) _textComp).setValue(labelNew, false);
    }

    /**
     * Sets the editable status of the field item.
     *
     * @param editable new status
     * @see   LySetEditableInterface#setEditable(boolean)
     */
    public void setEditable(boolean editable)
    {

        if (_textComp != null)
        {

            if (_textComp instanceof JTextComponent)
                ((JTextComponent) _textComp).setEditable(editable);
            else if (_textComp instanceof LySetEditableInterface)
                ((LySetEditableInterface) _textComp).setEditable(editable);
            else
                setEnabled(_textComp.isEnabled() && editable);

            Color bgColor;

            if (editable && _textComp.isEnabled())
                bgColor = getColor("LY_ENABLED_BACKGROUND");
            else
                bgColor = getColor("LY_DISABLED_BACKGROUND");

            _textComp.setBackground(bgColor);
        }
    }

    /**
     * Indicates if the field item is editable or not.
     *
     * @return true if the field is editable, false otherwise
     */
    public boolean isEditable()
    {
        boolean editable = false;

        if (_textComp instanceof JTextComponent)
            editable = ((JTextComponent) _textComp).isEditable();
        else if (_textComp instanceof LyJFormattedInput)
            editable = ((LyJFormattedInput) _textComp).isEditable();

        return editable;
    }

    /**
     * Sets the text component editable status. The status changes the background color of the text
     * component whether it is enabled and editable at the same time, or not.
     *
     * @parame status the new enabled status
     */
    public void setEnabled(boolean status)
    {

        if (_textComp != null)
        {
            _textComp.setEnabled(status);

            Color bgColor;

            if (status && isEditable())
                bgColor = getColor("LY_ENABLED_BACKGROUND");
            else
                bgColor = getColor("LY_DISABLED_BACKGROUND");

            _textComp.setBackground(bgColor);
        }
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.win.swing.LySwingComponent#getComponent()
     */
    public Component getComponent()
    {

        if (_textComp instanceof JTextArea)
            return _scrollPane;

        return _textComp;
    }

    /**
     * *********************************** Internal Classes.
     */

    /**
     * Handles text areas scoll panels.
     */
    public class _LyJScrollPane extends JScrollPane
    {
        private static final long serialVersionUID = -4677927208477321398L;

        /** The view component that needs a scroll panel. */
        protected Component _comp;

        /**
         * Class constructor of a {@link JScrollPane}.
         *
         * @param view      the component to display in the scrollpanes viewport
         * @param vsbPolicy an integer that specifies the vertical scrollbar policy
         * @param hsbPolicy an integer that specifies the horizontal scrollbar policy
         */
        public _LyJScrollPane(Component view, int vsbPolicy, int hsbPolicy)
        {
            super(view, vsbPolicy, hsbPolicy);
            _comp = view;
        }

        /**
         * Overriden to give the focus to the component displayed in the scrollpanes viewport.
         *
         * @see Component#requestFocus()
         */
        public void requestFocus()
        {
            _comp.requestFocus();
        }

        /**
         * Overriden to keep preferred size of text areas in views. This method returns the
         * preferred minimum size.
         *
         * @return the preferred minimumSize property
         * @see    JComponent#getMinimumSize()
         */
        public Dimension getMinimumSize()
        {

            if (_comp == null)
                return super.getMinimumSize();

            Dimension d = _comp.getMinimumSize();
            Dimension d1 = super.getMinimumSize();

            // Keep 1px space to avoid scrollbars
            int w = Math.max(d.width + 1, d1.width);
            int h = Math.max(d.height + 1, d1.height);

            return new Dimension(w, h);
        }

        /*
         * (non-Javadoc)
         * @see java.awt.Component#setBackground(java.awt.Color)
         */
        public void setBackground(Color bg)
        {

            if (_comp != null)
                _comp.setBackground(bg);

            super.setBackground(bg);
        }

        public Component getTextComponent()
        {
            return _comp;
        }
    } // end class _LyJScrollPane
} // end class LySwingText
