/*
 * LySwingTitleBar.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import javax.swing.*;
import javax.swing.border.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the internal description of an Icon object as described in the dtd.
 *
 * <p>A title bar allows to display a colored horizontal bar containing images and texts. The text
 * comprises three independent parts and each one of them is optional: a left, a central and a right
 * part. The color of the title bar is defined in the resources of the application.</p>
 *
 * @version $Id: LySwingTitleBar.java 35005 2012-03-23 13:23:39Z jlsaouli $
 * @author  Lyria
 * @see     LyComponent
 */
public class LySwingTitleBar extends LySwingObject
	implements LySwingComponent,
		LyGraphicalTitleBar,
		LyActionListener,
		LySetPopupMenuInterface
{
	public static final String __VERSION =
		"$Id: LySwingTitleBar.java 35005 2012-03-23 13:23:39Z jlsaouli $";

	/** Popup menu associated to the title bar. */
	public JPopupMenu		   _popupMenu;

	/** The view independent title bar description. */
	protected LyTitleBar	   _titleBar;

	/** The swing component of this title bar. */
	protected Component		   _titlePanel;

	/**
	 * Class constructor of a Title Bar object for the Swing target.
	 *
	 * @param titleBar a TITLEBAR component defined for every type of view targets. This title bar
	 *                 is initialized specificaly for the Swing view.
	 */
	public LySwingTitleBar(LyTitleBar titleBar)
	{
		super(titleBar);
		_titleBar = titleBar;

		initialize();
	}

	/**
	 * Creates the title bar.
	 *
	 * @param  bgImage         the background image
	 * @param  img             the main image
	 * @param  imgToolTip      the main image tooltip
	 * @param  rightImg        the right image
	 * @param  rightImgToolTip the right image tooltip
	 * @param  leftLabel       the left label
	 * @param  centerLabel     the center label
	 * @param  rightLabel      the right label
	 * @param  listener        the action listener
	 * @return the new title bar with the specified properties and listeners
	 */
	public Component createTitleBar(Image bgImage, Image img, String imgToolTip, Image rightImg,
		String rightImgToolTip, String leftLabel, String centerLabel, String rightLabel,
		LyActionListener listener)
	{
		LyJTitlePanel panel = new LyJTitlePanel(bgImage, img, imgToolTip, rightImg, rightImgToolTip,
				leftLabel, centerLabel, rightLabel, listener);

		panel.setBorder(new TitleBarBorder());

		return panel;
	}

	/**
	 * Initializes this swing component. Sets its font and colors.
	 */
	public void initialize()
	{
		Image  img = null;
		Image  rightImg = null;
		String imgToolTip = null;
		String rightImgToolTip = null;

		// If there is no image, do not handle tooltip
		if (_titleBar.getImage() != null)
		{
			img = getImage(_titleBar.getImage());
			imgToolTip = translate(_titleBar.getImageToolTip());
		}

		if (_titleBar.getRightImage() != null)
		{
			rightImg = getImage(_titleBar.getRightImage());
			rightImgToolTip = translate(_titleBar.getRightImageToolTip());
		}

		String leftLabel = translate(_titleBar.getLeftLabel());
		String rightLabel = translate(_titleBar.getRightLabel());
		String centerLabel = translate(_titleBar.getCenterLabel());
		Image  bgImage = (Image)getEnvironment().getImage("LY_TITLEBAR_BACKGROUND_IMAGE", null,
				false);
		Color  color;
		Font   font;

		if ((leftLabel != null) && (leftLabel.endsWith("\\n")))
			leftLabel = leftLabel.substring(0, leftLabel.length() - 2);

		_titlePanel = createTitleBar(bgImage, img, imgToolTip, rightImg, rightImgToolTip, leftLabel,
				centerLabel, rightLabel, this);

		LyAbstractTitlePanelManager manager = (LyAbstractTitlePanelManager)
			((LyGetManagerInterface)_titlePanel).getManager();

		font = getFont("LY_TITLEBAR_TITLE");
		manager.setTitleFont(font);
		color = getColor("LY_TITLEBAR_BACKGROUND");

		if (color == null)
			color = getColor("LY_DEFAULT_BACKGROUND");

		manager.setTitleBackgroundColor(color);

		color = getColor("LY_TITLEBAR_IMAGE_BACKGROUND");
		manager.setImageBackgroundColor(color);
		color = getColor("LY_TITLEBAR_IMAGE_BORDER");
		manager.setImageBorderColor(color);
		color = getColor("LY_TITLEBAR_TITLE");
		manager.setTitleColor(color);
		manager.setFading(true);
		
		// Place the toolbar on the right if set
		LyToolBar	toolbar = _titleBar.getToolBar();
		
		if (toolbar != null)
		{
			LySwingComponent	component = (LySwingComponent)toolbar.getInternalView();
			JPanel				panel = new JPanel(new GridBagLayout());
			GridBagConstraints	c = new GridBagConstraints();
			
			c.anchor = GridBagConstraints.CENTER;
			panel.setOpaque(false);
			panel.add(component.getComponent(), c);
			((JPanel)_titlePanel).setLayout(new BorderLayout());
			((JPanel)_titlePanel).add(panel, BorderLayout.EAST);
			
			// Register events on parent swing area
			if (component instanceof LyEventTransmitter)
			{
				LyArea				area = (LyArea)_titleBar.getParent();
				LySwingArea			swingArea = (LySwingArea)area.getInternalView();
				LyActionTransmitter transmitter = ((LyEventTransmitter)component)
						.getActionTransmitter();

				transmitter.addActionListener(swingArea);
			}
		}
	} // end method initialize

	/**
	 * Sets the title bar popup menu then shows it.
	 *
	 * @param menu the new popup menu of this title bar
	 */
	public void setPopupMenu(LyPopupMenu menu)
	{
        if (_popupMenu != null)
        {
            _popupMenu.setVisible(false);
            ((LyJTitlePanel)_titlePanel).remove(_popupMenu);
            _popupMenu = null;
        }
        
		if (menu == null)
			return;

		LySwingPopupMenu internal = (LySwingPopupMenu)menu.getInternalView();

		_popupMenu = (JPopupMenu)internal.getMenu();
		((LyJTitlePanel)_titlePanel).add(_popupMenu);
		// Show popup
		_popupMenu.show(_titlePanel, menu.getX(), menu.getY());
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return (_titlePanel);
	}

	/**
	 * Sets the image of this title bar and refreshes it.
	 *
	 * @param image the new image
	 * @see   LyGraphicalTitleBar#setImage(String)
	 */
	public void setImage(String image)
	{
		if (_titlePanel != null)
		{
			Image img = null;

			if (image != null)
				img = getImage(image);

			((LyAbstractTitlePanelManager)((LyGetManagerInterface)_titlePanel).getManager())
				.setImage(img);
			_titlePanel.repaint();
		}
	}

	/**
	 * Sets the tooltip of this title bar image.
	 *
	 * @param toolTip the new ToolTip
	 * @see   leon.view.graphic.LyGraphicalTitleBar#setImageToolTip(java.lang.String)
	 */
	public void setImageToolTip(String toolTip)
	{
		if (_titlePanel != null)
		{
			toolTip = translate(toolTip);
			((LyAbstractTitlePanelManager)((LyGetManagerInterface)_titlePanel).getManager())
				.setImageToolTip(toolTip);
		}
	}

	/**
	 * Sets the right image of this title bar.
	 *
	 * @param image the new image
	 * @see   leon.view.graphic.LyGraphicalTitleBar#setRightImage(java.lang.String)
	 */
	public void setRightImage(String image)
	{
		if (_titlePanel != null)
		{
			Image img = null;

			if (image != null)
				img = getImage(image);

			((LyAbstractTitlePanelManager)((LyGetManagerInterface)_titlePanel).getManager())
				.setRightImage(img);
			_titlePanel.repaint();
		}
	}

	/**
	 * Sets the tooltip of the right image.
	 *
	 * @param toolTip new ToolTip
	 * @see   leon.view.graphic.LyGraphicalTitleBar#setRightImageToolTip(java.lang.String)
	 */
	public void setRightImageToolTip(String toolTip)
	{
		if (_titlePanel != null)
		{
			toolTip = translate(toolTip);
			((LyAbstractTitlePanelManager)((LyGetManagerInterface)_titlePanel).getManager())
				.setRightImageToolTip(toolTip);
		}
	}

	/**
	 * Sets the left label text.
	 *
	 * @param leftLabel the new label
	 * @see   leon.view.graphic.LyGraphicalTitleBar#setLeftLabel(java.lang.String)
	 */
	public void setLeftLabel(String leftLabel)
	{
		if (_titlePanel != null)
		{
			leftLabel = translate(leftLabel);
			((LyAbstractTitlePanelManager)((LyGetManagerInterface)_titlePanel).getManager())
				.setLeftLabel(leftLabel);
			_titlePanel.repaint();
		}
	}

	/**
	 * Sets the center label text.
	 *
	 * @param centerLabel the new label
	 * @see   leon.view.graphic.LyGraphicalTitleBar#setCenterLabel(java.lang.String)
	 */
	public void setCenterLabel(String centerLabel)
	{
		if (_titlePanel != null)
		{
			centerLabel = translate(centerLabel);
			((LyAbstractTitlePanelManager)((LyGetManagerInterface)_titlePanel).getManager())
				.setCenterLabel(centerLabel);
			_titlePanel.repaint();
		}
	}

	/**
	 * Sets the right label text.
	 *
	 * @param rightLabel the new label
	 * @see   leon.view.graphic.LyGraphicalTitleBar#setRightLabel(java.lang.String)
	 */
	public void setRightLabel(String rightLabel)
	{
		if (_titlePanel != null)
		{
			rightLabel = translate(rightLabel);
			((LyAbstractTitlePanelManager)((LyGetManagerInterface)_titlePanel).getManager())
				.setRightLabel(rightLabel);
			_titlePanel.repaint();
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		// Notify controller if any
		if (_titleBar.getRootController() != null)
		{
			String sourceId = actionEvent.getSourceId();

			if (actionEvent.getType() == LyActionDemand.TYPE_ACTION)
			{
				if ((_titleBar.isImageSensitive() && sourceId.equals("leftImage"))
						|| (_titleBar.isRightLabelSensitive() && sourceId.equals("rightLabel")))
				{
					// Notify controller
					String		   titleBarId = _titleBar.getId();
					String		   id = ((titleBarId == null) ? sourceId
															  : (titleBarId + "." + sourceId));
					LyActionDemand demand = new LyActionDemand(LyActionDemand.TYPE_ACTION, id);

					processDemand(_titleBar.getRootController(), demand);
					actionEvent.consume();
				}
			}
			else if (actionEvent.getType() == LyMenuDemand.TYPE_POPUP)
			{
				String		 titleBarId = _titleBar.getId();
				String		 id = ((titleBarId == null) ? sourceId : (titleBarId + "." + sourceId));
				LyMenuDemand demand = new LyMenuDemand(LyMenuDemand.TYPE_POPUP, titleBarId);

				demand.setItem(id);
				demand.setPosition((int[])actionEvent.getParameter());
				demand.addIndex(0);
				demand.setSubIndex(0);

				processDemand(_titleBar.getRootController(), demand);
				actionEvent.consume();
			}
		}
	} // end method actionPerformed

	/**
	 * .
	 */

	/**
	 * This class describes special borders for Titlebar.
	 */
	public class TitleBarBorder extends EmptyBorder
	{
		private static final long serialVersionUID = -2784172676790923743L;

		/**
		 * Class constructor of a new special title bar border.
		 */
		public TitleBarBorder()
		{
			super(0, 0, 0, 0);
		}

		/**
		 * Draws the title bar borders.
		 *
		 * @see javax.swing.border.Border#paintBorder(java.awt.Component, java.awt.Graphics, int,
		 *      int, int, int)
		 */
		public void paintBorder(Component c, Graphics g, int x, int y, int width, int height)
		{
			Color background = c.getBackground();

			g.setColor(background.darker());
			g.drawLine(x, y + height - 2, x + width - 1, y + height - 2);
			g.setColor(background.brighter());
			g.drawLine(x, y + height - 1, x + width - 1, y + height - 1);
		}
	} // end class TitleBarBorder

} // end class LySwingTitleBar
