/*
 * LySwingTool.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.net.*;

import javax.swing.*;
import javax.swing.border.*;

import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Action object (as described in the dtd) for swing viewer.
 *
 * <p>An action button allows to execute an action. In the case where the button contains an area
 * (AREA), this area is displayed in a new window when launching the action.</p>
 *
 * @version $Id: LySwingTool.java 35032 2012-03-23 17:17:16Z jlsaouli $
 * @author  Lyria
 * @see     LySwingToolBar
 * @see     LyTool
 */
public class LySwingTool extends LySwingObject
	implements LySwingComponent,
		LyFreeInterface,
		LyGraphicalTool,
		LyActionListener,
		LyEventTransmitter
{
	public static final String    __VERSION =
		"$Id: LySwingTool.java 35032 2012-03-23 17:17:16Z jlsaouli $";

	/** Width of the label tool. */
	public static int			  LY_LABELED_TOOL_WIDTH = 50;

	/** Auxiliary components. */
	protected Component			  _auxButton;

	/** The popup menu of this action button. */
	protected LyJPopupButton	  _popup;

	/** The view independent action button description. */
	protected LyTool			  _tool;

	/** The component used to display this button. */
	protected Component			  _button;

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter _actionTransmitter;

	/**
	 * Class constructor of a Tool object for the Swing target.
	 *
	 * @param tool a TOOL component defined for every type of view targets. This tool is initialized
	 *             specificaly for the Swing view.
	 */
	public LySwingTool(LyTool tool)
	{
		super(tool);
		_tool = tool;

		initialize();
		setContent();
	}

	/**
	 * Builds the tool component and registers its events.
	 */
	public void initialize()
	{
		_button = createTool();

		Color color = getColor("LY_TOOL_LABEL_COLOR");

		if (color != null)
			_button.setForeground(color);

		setEnabled(_tool.isEnabled());

		if (_tool.getStyle() == LyTool.Style.TOGGLE)
			setPressed(_tool.isPressed());
	}

	/**
	 * Fills the area if this tool has one.
	 */
	public void setContent()
	{
		LyComponent component = _tool.getComponent();

		if (component != null)
			fillToolComponent();
	}

	/**
	 * Creates the new tool components: button, menu item or panel.
	 *
	 * @return the new tool component
	 */
	public Component createTool()
	{
		Component   button;
		LyComponent parent = _tool.getParent();
		String	    label = translate(_tool.getLabel());
		String	    tooltip = translate(_tool.getTooltip());
		Image	    image = getImage(_tool.getImage());
		boolean     isMenu = false;
		
		// If tooltip is Null, set label FFT00996
		if (tooltip == null)
			tooltip = label;

		/* if the action opens a dialog, append "..." */
		if ((_tool.isDialog()) && (label != null) && (!label.endsWith("...")))
			tooltip += "...";

		if (parent instanceof LyToolBar)
			isMenu = (((LyToolBar)parent).getStyle() == LyToolBar.Style.MENU);

		if (isMenu) /* the tool is a menu item */
		{
			button = new JMenuItem(label, new ImageIcon(image));
			((JMenuItem)_button).setFont(getFont("LY_BUTTON_FONT"));
			button.setEnabled(_tool.isEnabled());
			((JMenuItem)button).setBorderPainted(false);
			((JMenuItem)button).setOpaque(false);
		}
		else
		/* the tool is a button */
		{
			// If image is animated, create the tool with the image path since using the image instance
			// does not work (image won't be animated).
			boolean		animated = isImageAnimated(_tool.getImage());
			
			if (animated)
			{
				ClassLoader	classLoader = LyEnvironment.class.getClassLoader();

				if (classLoader == null)
					classLoader = ClassLoader.getSystemClassLoader();
				
				String	imagePath = getEnvironment().getImagePath(_tool.getImage());
				
				imagePath = getEnvironment().getPathWithoutSize(imagePath);

				URL		imgURL = classLoader.getResource(imagePath);
				
				button = new LyJButton(imgURL, getImage(imagePath), "", tooltip, _tool.getId());
				
			}
			else
				button = new LyJButton(image, "", tooltip, _tool.getId());

			/* Add the label if needed */
			if ((image == null)
				|| testResource("LY_WRITE_TOOL_LABEL"))
			{
				((LyJButton)button).setText(label);

				if (testResource("LY_TOOL_LABEL_RIGHT_POSITION"))
				{
					((LyJButton)button).setVerticalTextPosition(SwingConstants.CENTER);
					((LyJButton)button).setHorizontalTextPosition(SwingConstants.RIGHT);
				}
				else
				{
					((LyJButton)button).setVerticalTextPosition(SwingConstants.BOTTOM);
					((LyJButton)button).setHorizontalTextPosition(SwingConstants.CENTER);
				}
			}
			
			((LyJButton)button).setBorder(BorderFactory.createMatteBorder(1, 1, 1, 1,
					(Icon)null));

			((LyJButton)button).setBorderPainted(true);
			((LyJButton)button).setFont(getFont("LY_BUTTON_FONT"));
			((LyJButton)button).setGrayProp(0);
			
			if (!animated)
				((LyJButton)button).setImage(image);
			
			((LyJButton)button).setOpaque(false);
			((LyJButton)button).setRolloverChangeBackground(_tool.getStyle() != LyTool.Style.TOGGLE);

			button.setEnabled(_tool.isEnabled());

			LyActionTransmitter transmitter = ((LyJButton)button).getActionTransmitter();

			if (transmitter != null)
				transmitter.addActionListener(this);
		}

		return button;
	} // end method createTool

	/**
	 * Fills the tool area.
	 */
	public void fillToolComponent()
	{
		if (_tool.getComponent() instanceof LyArea)
		{
			LyArea			 area = (LyArea)_tool.getComponent();
			LySwingComponent internalView = (LySwingComponent)area.getInternalView();
			Component		 internalComponent = internalView.getComponent();
	
			LyJBorderPanel   border = new LyJBorderPanel(LyArea.Border.SOLID);
	
			if (internalComponent instanceof LyJBorderPanel)
				((LyJBorderPanel)internalComponent).setGap(0);
	
			border.add(internalComponent);
			border.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
			border.setGap(0);
			border.setBorder((Border)UIManager.get("PopupMenu.border"));
	
			_popup = new LyJPopupButton(((LyJButton)_button), border, LyJPopupSelector.ALIGN_LEFT, getEnvironment());
			_button = _popup.getButton();
			_auxButton = _popup.getOpenerButton();
			((LyJButton)_button).setOpenerButton((LyJOpenerButton)_auxButton);
			setEnabled(_tool.isEnabled());
	
			if ((internalView instanceof LyEventTransmitter)
					&& (((LyEventTransmitter)internalView).getActionTransmitter() != null))
				((LyEventTransmitter)internalView).getActionTransmitter().addActionListener(_popup);
		}
		else if (_tool.getComponent() instanceof LyPopupMenu)
		{
			LyPopupMenu		popupMenu = (LyPopupMenu)_tool.getComponent();
			
			popupMenu.setEnvironment(_tool.getEnvironment());
			
			LySwingPopupMenu	swingPopupMenu = (LySwingPopupMenu)popupMenu.getInternalView();
			JPopupMenu		jPopupMenu = (JPopupMenu)swingPopupMenu.getMenu();
			
			// Check if the parent is a TOOLBAR tag and retrieve its fill value
			LyComponent      parent = _tool.getParent();
            LyToolBar.Fill	 fill = LyToolBar.Fill.HORIZONTAL;
			int		         align;
	
			if (parent instanceof LyToolBar)
				fill = ((LyToolBar)parent).getFill();
	
			if (fill == LyToolBar.Fill.VERTICAL)
				align = LyJPopupSelector.ALIGN_TOP;
			else
				align = LyJPopupSelector.ALIGN_LEFT;
	
			_popup = new LyJPopupButton(((LyJButton)_button), jPopupMenu, align, getEnvironment());
			_button = _popup.getButton();
			_auxButton = _popup.getOpenerButton();
			((LyJButton)_button).setOpenerButton((LyJOpenerButton)_auxButton);
			setEnabled(_tool.isEnabled());
		}
	} // end method fillArea

	/**
	 * Sets the new button image.
	 *
	 * @param image the new image name
	 * @see   LySetImageInterface#setImage(String)
	 */
	public void setImage(String image)
	{
		if (_button instanceof LyJButton)
		{
			// If image is animated, set the tool with the image path since using the image instance
			// does not work (image won't be animated).
			boolean	   animated = isImageAnimated(image);
			LyJButton  jbutton = (LyJButton)_button;

			if (animated)
			{
				ClassLoader	classLoader = LyEnvironment.class.getClassLoader();

				if (classLoader == null)
					classLoader = ClassLoader.getSystemClassLoader();
				
				String	imagePath = getEnvironment().getImagePath(image);
				
				imagePath = getEnvironment().getPathWithoutSize(imagePath);

				URL		imgURL = classLoader.getResource(imagePath);
				
				jbutton.setImage(imgURL, getImage(image));
			}
			else
				jbutton.setImage(getImage(image));
		}
		else if (_button instanceof JMenuItem)
			((JMenuItem)_button).setIcon(new ImageIcon(getImage(image)));
		
		updateParentTable();
	}

	/**
	 * Gets the auxiliary component of this component.
	 *
	 * @return the auxiliary button
	 */
	public Component getAuxComponent()
	{
		return _auxButton;
	}

	/**
	 * Sets the new button label.
	 *
	 * @param label the new label
	 * @see   LySetLabelInterface#setLabel(String)
	 */
	public void setLabel(String label)
	{
		if (_button instanceof LyJButton)
		{
			LyEnvironment environment = _tool.getEnvironment();
			String	      image = _tool.getImage();
			
			// Only propagate label change if we have to write it...
			if ((image == null)
				|| environment.testResource("LY_WRITE_TOOL_LABEL"))
			{
				label = environment.translate(label);

				((LyJButton)_button).setText(label);
			}
		}
		
		updateParentTable();
	}
	
	/**
	 * Graphically updates the parent table of this tool if needed.
	 * When the image or the label of a tool is changed, and if this tool
	 * is contained in a table, table must be updated to reset its columns
	 * width in order to correctly display the tool.
	 */
	protected void updateParentTable()
	{
		// If tool is displayed in a table, table may need to be updated to
		// show the new tool image and label.
		LyComponent	parent = _tool.getParent();
		LyRow		parentRow = null;
		
		while ((parent != null) && !(parent instanceof LyTable))
		{
			if (parent instanceof LyRow)
				parentRow = (LyRow)parent;
			
			parent = parent.getParent();
		}
		
		if ((parent instanceof LyTable) && (parentRow != null))
		{
			LyRows				rows = ((LyTable)parent).getRows();
			int					rowIndex = rows.getRowIndex(parentRow);
			LySwingTable		swingTable = (LySwingTable)parent.getInternalView();
			LyJTableComponent	tableComponent = swingTable.getTableComponent();
			int					column = -1, i, n = parentRow.getComponentCount();
			
			for (i = 0; i < n; i++)
			{
				if (parentRow.getComponent(i) == _tool.getParent())
					column = i;
			}

			if (column >= 0)
				swingTable.ensureColumnsWidth(rowIndex, column);
			else
				swingTable.ensureColumnsWidth(rowIndex);

			tableComponent.repaint();
		}
	}

	/**
	 * Updates attribute pressed to the given value.
	 * 
	 * @param pressed Boolean indicating if attribute pressed is set or unset.
	 */
	public void setPressed(boolean pressed)
	{
		((LyJButton)_button).setOpaque(pressed);
		
		if (pressed)
		{
			((LyJButton)_button).setBackground((Color)UIManager.get("Button.select"));
			((LyJButton)_button).setBorder(BorderFactory.createLineBorder(Color.BLACK, 1));
		}
		else
		{
			((LyJButton)_button).setBackground(null);
			((LyJButton)_button).setBorder(null);
		}
	}

	/**
	 * Sets the status of the menu. If disabled, the menu is grayed and inactive.
	 *
	 * @param enabled the new status
	 * @see   LySetEnabledInterface#setEnabled(boolean)
	 */
	public void setEnabled(boolean enabled)
	{
		if (_button != null)
			_button.setEnabled(enabled);

		if (_auxButton != null)
			_auxButton.setEnabled(enabled);
	}

	/**
	 * Sets the new tooltip.
	 *
	 * @param tooltip the new tooltip
	 */
	public void setTooltip(String tooltip)
	{
		if (_button instanceof LyJButton)
		{
			tooltip = translate(tooltip);
	
			if (tooltip == null)
				tooltip = translate(_tool.getLabel());
	
			// If the action opens a dialog, append "..."
			if ((_tool.isDialog()) && (tooltip != null) && !tooltip.endsWith("..."))
				tooltip += "...";
			
			((LyJButton)_button).setToolTipText(tooltip);
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyFreeInterface#free()
	 */
	public void free()
	{
		if (_popup != null)
			_popup.hideTipWindow();
		if (_button instanceof JButton)
		{
			Icon icon = ((JButton)_button).getRolloverIcon();
			if ((icon != null) && (icon instanceof ImageIcon))
				((ImageIcon)icon).getImage().flush();
			
			icon = ((JButton)_button).getDisabledIcon();
			if ((icon != null) && (icon instanceof ImageIcon))
				((ImageIcon)icon).getImage().flush();
		}
		super.free();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		if (_actionTransmitter != null)
		{
			if (_tool.getComponent() == null)
			{				
				if (_tool.getStyle() == LyTool.Style.TOGGLE)
					_tool.setPressed(!_tool.isPressed());
				
				LyActionEvent evt = new LyActionEvent(this);

				evt.setSourceId(_tool.getId());
				evt.setAdditionalValue(actionEvent.getAdditionalValue());

				_actionTransmitter.processActionEvent(evt);
			}
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _button;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return (_actionTransmitter);
	}

	/**
	 * This action button doesn't handle selection events, so it has no selection transmitter.
	 *
	 * @return <code>null</code>
	 * @see    leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		return null;
	}

} // end class LySwingTool
