/*
 * LySwingToolBar.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.util.*;
import javax.swing.*;

import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Toolbar object (as described in the dtd) for swing viewer.
 *
 * <p>A tool bar allows to display a set of action buttons. A tool bar can contain other tool bars,
 * which allows to group together the action buttons into logical groups and sub-groups.</p>
 *
 * @version $Id: LySwingToolBar.java 35375 2012-04-26 09:49:42Z nhamon $
 * @author  Lyria
 * @see     LySwingTool
 * @see     LyToolBar
 */
public class LySwingToolBar extends LySwingObject
	implements LySwingComponent,
		LySetColorInterface,
		LySetImageInterface,
		PropertyChangeListener,
		LyGraphicalToolBar,
		ActionListener,
		LyActionListener,
		LyEventTransmitter,
		LyGetLockInterface
{
	public static final String    __VERSION =
		"$Id: LySwingToolBar.java 35375 2012-04-26 09:49:42Z nhamon $";

	/** Keeps last popup menu to remove it. This is used to save memory. */
	protected JPopupMenu		  _popupMenu;

	/** Label that supports toolbar image when needed. */
	protected JLabel			  _logo;
	
	/** Panel used at top of expand toolbars. */
	protected JPanel			  _expandTopPanel;

	/** Indicates if this toolbar element is the first toolbar in its parent container. */
	protected boolean			  _first = true;

	/** The view independent toolbar description. */
	protected LyToolBar			  _toolBar;

	/** Component representing the actions: may be a panel or a button (with menu actions). */
	protected Component			  _comp;
	
	/** Scroll panel used in expand bar mode. */
	protected JScrollPane		  _expandScrollPanel;

	/** Preferred width of the scroll panel used in expand bar mode. */
	protected int				  _expandScrollPanelPreferredWidth;

	/** Popup menu used for menu style. */
	protected Object			  _popup;

	/** Association table for the popup menus and their associated actions. */
	protected Hashtable<Object, LyTool>			  _popupAssoc;

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter _actionTransmitter;

	/**
	 * Class constructor of a Tool bar object for the Swing target.
	 *
	 * @param toolBar a TOOLBAR component defined for every type of view targets. This tool bar is
	 *                initialized specificaly for the Swing view.
	 */
	public LySwingToolBar(LyToolBar toolBar)
	{
		super(toolBar);
		_toolBar = toolBar;
		_expandScrollPanel = null;
		_expandScrollPanelPreferredWidth = -1;

		initialize();
		setContent();
	}

	/**
	 * Creates the default toolbar component.
	 */
	public void initialize()
	{
		synchronized(getLock())
		{
			_comp = createToolBar();
		}
	}

	/**
	 * Fills the toolBar with its tools if it is not a toolbar menu.
	 */
	public void setContent()
	{
		if (_toolBar.getStyle() != LyToolBar.Style.MENU)
		{
			int i, n = _toolBar.getComponentCount();

			for (i = 0; i < n; i++)
			{
				LyComponent tool = _toolBar.getComponent(i);

                if ((tool == null)
                        || (tool.getId() == LyComponent.SEP_ID)) // SEP
                    // Add a separator between actions
                    addSeparator();
				else
				{
					// Retrieve the action graphical component
					LySwingComponent    internal = (LySwingComponent)tool.getInternalView();
					LyActionTransmitter transmitter = ((LyEventTransmitter)internal)
							.getActionTransmitter();

					// Register for events
					if (transmitter != null)
						transmitter.addActionListener(this);

					addTool(internal);
				}
			}
		}
	} // end method setContent

	/**
	 * Creates and initializes a new empty toolBar.
	 *
	 * @return the new tool bar
	 */
	public Component createToolBar()
	{
		Component comp;
		String    label = translate(_toolBar.getLabel());

		if (label == null)
			label = "";

		if (_toolBar.getStyle() == LyToolBar.Style.MENU)
		{
			comp = new LyJToggleButton(getImage("LY_ACTION_MENU"), label, label, null);
			((LyJToggleButton)comp).setFont(getFont("LY_BUTTON_FONT"));
			comp.setBackground(null);

			LyActionTransmitter transmitter = ((LyJToggleButton)comp).getActionTransmitter();

			if (transmitter != null)
				transmitter.addActionListener(this);
		}
		else if ((_toolBar.getStyle() == LyToolBar.Style.EXPAND)
				&& (_toolBar.getParent() instanceof LyArea))
		{
			// If toolbar has the expand style with an area as a parent, build an expand bar
			// and place it in a scroll pane.
			LyJExpandBar	bar = new LyJExpandBar(_toolBar.getEnvironment());
			GridBagLayout	layout = new GridBagLayout();
			LyEnvironment	env = _toolBar.getEnvironment();

			_expandScrollPanel = new JScrollPane(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
			_expandScrollPanel.setViewportView(bar);
			_expandScrollPanel.getVerticalScrollBar().setUnitIncrement(20);
			_expandScrollPanel.getViewport().setBackground((Color)env.getColor("LY_EXPANDBAR_CONTENT_COLOR"));
			_expandTopPanel = new JPanel(layout);
			_expandTopPanel.setOpaque(false);
			bar.add(_expandTopPanel);
			
			if ((_toolBar.getImage() != null) && (_toolBar.getImage().length() != 0))
			{
				GridBagConstraints	gbc = new GridBagConstraints();
				
				_logo = new JLabel(new ImageIcon(getImage(_toolBar.getImage())));
				_logo.setHorizontalAlignment(SwingConstants.LEFT);
				gbc.anchor = GridBagConstraints.WEST;
				gbc.fill = GridBagConstraints.BOTH;
				gbc.weightx = 1;
				layout.setConstraints(_logo, gbc);
				_expandTopPanel.add(_logo);
			}
			
			comp = new JPanel(new BorderLayout());
			((JPanel)comp).add(_expandScrollPanel, BorderLayout.CENTER);
			
			Image			image = (Image)getEnvironment().getImage("LY_AREA_BACKGROUND_IMAGE", null, false);
			JPanel			buttonPanel = new LyJBackgroundPanel(image);
			FlowLayout		panelLayout = new VFlowLayout(VFlowLayout.MIDDLE, 0, 0, true);
			LyJButton		button = new LyJButton(getImage("LY_EXPANDBAR_COLLAPSE"), null, null, "switchExpandBar", true)
			{
				/**
				 * 
				 */
				private static final long serialVersionUID = -5527695236097118694L;

				public Insets getInsets()
				{
					return new Insets(0, 0, 0, 1);
				}
			};

			button.setBorderPainted(false);
			button.setOpaque(false);
			button.setContentAreaFilled(false);
			button.setRolloverChangeBackground(false);
			button.setFocusable(false);
			button.addActionListener(this);
			buttonPanel.setLayout(panelLayout);
			buttonPanel.setOpaque(true);
			buttonPanel.add(button);
			((JPanel)comp).add(buttonPanel, BorderLayout.EAST);
		}
		// Retrieve the border value and create a BorderPanel with the
		// specified border
		else
		{
			Font    font = getFont("LY_TOOLBAR_TAB_FONT");
			Color   color = getColor("LY_GROUP_FONT_COLOR");
			int     maxWidth = getMaxWidth();
			int		align = LyFlowLayout.LEFT;

			// Check LY_GROUP_FIELDS_WITH_NO_GROUP resource. If not defined, we consider that default
			// value is true. Otherwise, we retrieve value defined on environment.
			String		   wrapResource = "LY_AUTOMATIC_TOOLBAR_WRAP";
			String		   wrapValue = getResource(wrapResource);
			boolean		   wrap;
			
			if (wrapValue == null)
				wrap = true;
			else
				wrap = testResource(wrapResource);
			
			if (_toolBar.getHAlign() == LyToolBar.HAlign.RIGHT)
				align = LyFlowLayout.RIGHT;

			if (_toolBar.getStyle() == LyToolBar.Style.BORDER)
				comp = new LyJToolBarComponent(label,
						LyJBorderPanel.createBorder(LyArea.Border.IN, 3,
							LyJBorderPanel.DEFAULT_GAP, color, font, LyJBorderPanel.LEFT, label),
						maxWidth, wrap, align);
			else
				comp = new LyJToolBarComponent(label,
						LyJBorderPanel.createBorder(LyArea.Border.SOLID, 0,
							LyJBorderPanel.DEFAULT_GAP, LyJBorderPanel.DEFAULT_COLOR),
						maxWidth, wrap, align);
			if (_toolBar.getFill() == LyToolBar.Fill.VERTICAL)
			{
				((JToolBar)comp).setOrientation(SwingConstants.VERTICAL);

				// FFT01418 - Centrage des toolbars
				LyEnvironment env = _toolBar.getEnvironment();

				if (env.testResource("LY_WRITE_TOOL_LABEL")
						&& !env.testResource("LY_TOOL_LABEL_RIGHT_POSITION"))
					((JToolBar)comp).setLayout(new VFlowLayout(VFlowLayout.TOP, 0, 0, true));
			}
			else
				((JToolBar)comp).setOrientation(SwingConstants.HORIZONTAL);

			// FFT1141 SWING - Toolbar et popup
			((LyJToolBarComponent)comp).setOrientationLock(true, getViewManager());
			((LyJToolBarComponent)comp).setOpaque(false);

			comp.setFont(font);
			comp.setForeground(color);

			if ((_toolBar.getImage() != null) && (_toolBar.getImage().length() != 0))
			{
				_logo = new JLabel(new ImageIcon(getImage(_toolBar.getImage())));
				((LyJToolBarComponent)comp).add(_logo);
			}

			((LyJToolBarComponent)comp).addPropertyChangeListener( /* "orientation", */this);
			comp.setBackground(null);
		}
		
		if ((_toolBar.getParent() instanceof LyToolBar)
				&& (_toolBar.getStyle() == LyToolBar.Style.EXPAND)
				&& (((LyToolBar)_toolBar.getParent()).getStyle() == LyToolBar.Style.EXPAND))
		{
			LyJExpandItem	panel = new LyJExpandItem(_toolBar.getEnvironment(), _toolBar.getId(), _toolBar.getLabel(), _toolBar.getImage());
			
			panel.setActionListener(this);
			panel.setOpen(_toolBar.isOpen());
			comp = panel;
		}

		if (_toolBar.getColor() != null)
		{
			Color col = getColor(_toolBar.getColor());

			if (col != null)
			{
				comp.setBackground(col);
				if (comp instanceof LyJToolBarComponent)
				    ((LyJToolBarComponent)comp).setOpaque(true);
			}
		}

		return comp;
	} // end method createToolBar

	/**
	 * Sets the {@link #_logo} label icon to the specified image.
	 *
	 * @param imageName the name of the image in the environment
	 * @see   leon.view.setinterface.LySetImageInterface#setImage(java.lang.String)
	 */
	public void setImage(String imageName)
	{
		if (_logo == null)
			return;

		Image image = getImage(imageName);

		_logo.setIcon(new ImageIcon(image));
	}

	/**
	 * Sets the background color of this toolbar component and refreshes it after the setting.
	 *
	 * @param color the new background color
	 * @see   leon.view.setinterface.LySetColorInterface#setColor(java.lang.String)
	 */
	public void setColor(String color)
	{
		if (color != null)
		{
			Color col = getColor(color);

			if (col != null)
			{
				_comp.setBackground(col);
				((LyJToolBarComponent)_comp).setOpaque(true);
				_comp.repaint();
				((LyJToolBarComponent)_comp).revalidate();
			}
		}
	}

	/**
	 * Add a separator at the end of the toolBar.
	 */
	public void addSeparator()
	{
		int align;

		if (_toolBar.getFill() == LyToolBar.Fill.VERTICAL)
			align = SwingConstants.HORIZONTAL;
		else
			align = SwingConstants.VERTICAL;

		Color c = _comp.getBackground();

		if (c == null)
			c = Color.darkGray;
		else
			c = c.darker().darker();

		LyJSeparator sep = new LyJSeparator(align, LyJSeparator.TOOLBAR_STYLE, c, 5);

		if (_comp instanceof LyJToolBarComponent)
			((LyJToolBarComponent)_comp).addSeparator(sep);
		else if (_expandScrollPanel != null)
		{
			LyJExpandBar	expandBar = (LyJExpandBar)_expandScrollPanel.getViewport().getView();
			GridLayout		layout = new GridLayout(1, 1, 0, 0);
			JPanel			panel = new JPanel();				
			
			panel.setOpaque(false);
			panel.setLayout(layout);
			panel.add(sep);
			expandBar.add(panel);
		}
	}

	/**
	 * Adds a tool at the end of the toolBar.
	 *
	 * @param tool the tool to add
	 */
	public void addTool(LySwingComponent tool)
	{
		Component	component = tool.getComponent();
		
		if (tool instanceof LySwingFormField)
		{
			JPanel	panel = new JPanel(new BorderLayout());
			
			panel.setOpaque(false);
			
			if (((LySwingFormField)tool).getCaptionComponent() != null)
				panel.add(((LySwingFormField)tool).getCaptionComponent(), BorderLayout.WEST);
			
			panel.add(((LySwingFormField)tool).getFullComponent(), BorderLayout.CENTER);
			component = panel;
		}

		if (_expandScrollPanel != null)
		{
			// If component is a scroll pane, we have an expand toolbar
			LyJExpandBar	expandBar = (LyJExpandBar)_expandScrollPanel.getViewport().getView();
			
			if (tool instanceof LySwingToolBar)
			{
				LyToolBar	toolbar = ((LySwingToolBar)tool)._toolBar;
				
				if ((toolbar.getVAlign() == LyToolBar.VAlign.TOP)
						&& (toolbar.getHAlign() == LyToolBar.HAlign.RIGHT))
				{
					if (_logo == null)
					{
						BorderLayout	layout = new BorderLayout();
						
						_expandTopPanel.setLayout(layout);
						_expandTopPanel.add(component, BorderLayout.EAST);
					}
					else
					{
						GridBagLayout		layout = (GridBagLayout)_expandTopPanel.getLayout();
						GridBagConstraints	gbc = new GridBagConstraints();
						
						gbc.anchor = GridBagConstraints.EAST;
						gbc.fill = GridBagConstraints.NONE;
						gbc.weightx = 0;
						
						layout.setConstraints(component, gbc);
						_expandTopPanel.add(component);
					}
					
					return;
				}
			}
			else if (tool instanceof LySwingTool)
			{
				Component			opener = ((LySwingTool)tool).getAuxComponent();
				Container 			parent = (Container)_comp;
				BorderLayout		layout = new BorderLayout(0, 0);
				LyJToolBarComponent	panel = new LyJToolBarComponent(null, null, -1, true);				
				
				panel.setOpaque(false);
				panel.setLayout(layout);
				parent = panel;
				parent.add(component, BorderLayout.CENTER);

				if (opener != null)
					parent.add(opener, BorderLayout.EAST);
				
				component = panel;
			}
			
			expandBar.add(component);
		}
		else if (tool instanceof LySwingToolBar)
			((Container)_comp).add(component);
		else
		{
			Container parent = (Container)_comp;

			// If component is an expand item, add a sub-toolbar to be sure that all the tools displayed in the
			// toolbar will have a correct width.
			if (_comp instanceof LyJExpandItem)
			{
				BorderLayout		layout = new BorderLayout(0, 0);
				LyJToolBarComponent	panel = new LyJToolBarComponent(null, null, -1, true);				
				
				panel.setOpaque(false);
				panel.setLayout(layout);
				parent.add(panel);
				parent = panel;
			}
			
			parent.add(component, BorderLayout.CENTER);
			
			if (tool instanceof LySwingTool)
			{
				Component opener = ((LySwingTool)tool).getAuxComponent();

				if (opener != null)
					parent.add(opener, BorderLayout.EAST);
			}
		}
	}

	/**
	 * Creates a popup menu (used if toolBar has menu style).
	 *
	 * @param  label the label of the menu
	 * @param  font  the font used to render the menu
	 * @return the popup menu for this tool bar
	 */
	public Object createPopupMenu(String label, Font font)
	{
		JPopupMenu popup = new JPopupMenu(label);

		popup.setFont(font);
		return popup;
	}

	/**
	 * Creates a menu item (used if toolBar has menu style).
	 *
	 * @param  label the label of the item
	 * @param  font  the font used to render the item
	 * @return the menu item for this tool bar
	 */
	public Object createMenuItem(String label, Font font)
	{
		JMenuItem item = new JMenuItem(label);

		item.setFont(font);
		return item;
	}

	/**
	 * Adds a menu item at the end of the specified menu.
	 *
	 * @param item the item to add
	 * @param menu the menu where the item is added
	 */
	public void addItemToMenu(Object item, Object menu)
	{
		((JMenu)menu).add((JMenuItem)item);
	}

	/**
	 * Registers action events for the specified item.
	 *
	 * @param item     the item where the listeners are registered
	 * @param listener the listener to add
	 */
	public void registerActionEvents(Object item, ActionListener listener)
	{
		((JMenuItem)item).addActionListener(listener);
	}

	/**
	 * Adds a popup menu to the specified toolBar.
	 *
	 * @param popup the popup menu to add
	 * @param comp  the toolbar component where the popup is attached
	 */
	public void addPopupToComponent(Object popup, Component comp)
	{
		if (_popupMenu != null)
			((JComponent)comp).remove(_popupMenu);

		_popupMenu = (JPopupMenu)popup;

		((JComponent)comp).add(_popupMenu);
	}

	/**
	 * Shows the popupmenu at the specified position on the given component.
	 *
	 * @param popup the popup menu to show
	 * @param comp  the invoker component
	 * @param x     the x coordinate in the comp space
	 * @param y     the y coordinate in the comp space
	 * @see   JPopupMenu#show(java.awt.Component, int, int)
	 */
	public void showPopup(Object popup, Component comp, int x, int y)
	{
		((JPopupMenu)popup).show(comp, x, y);
	}

	/**
	 * Gets the maximum width of the associated view manager.
	 *
	 * @return the width of the view manager screen size altered with the {@link
	 *         LySwingFrame#MAX_SIZE_RATIO}.
	 */
	protected int getMaxWidth()
	{
		LyViewManager viewManager = getViewManager();
		int[]		  screenSize = viewManager.getScreenSize(_comp);
		int			  maxWidth = screenSize[0];

		maxWidth = maxWidth * LySwingFrame.MAX_SIZE_RATIO / 100;
		return maxWidth;
	}

	/**
	 * Called when the orientation of the toolBar component has changed. Resizes the parent window
	 * if needed.
	 */
	public void propertyChange(PropertyChangeEvent e)
	{
		String name = e.getPropertyName();

		if (name.equals("orientation"))
		{
			int orientation = ((Integer)e.getNewValue()).intValue();
			int oldOrientation = ((Integer)e.getOldValue()).intValue();

			if (orientation == oldOrientation)
				return;

			LyView view = _toolBar.getParentView();

			if (view != null)
			{
				if (view instanceof LyFrame)
				{
					LyFrame frame = (LyFrame)view;

					if (frame.isResizable())
						return;
					SwingUtilities.invokeLater(new Runnable()
						{

							/*
							 * (non-Javadoc)
							 * @see java.lang.Runnable#run()
							 */
							public void run()
							{
								((LySwingFrame)_toolBar.getParentView().getInternalView()).repack();
							}
						});
				}
				else if (view instanceof LyDialog)
				{
					LyDialog dialog = (LyDialog)view;

					if (dialog.isResizable())
						return;
					SwingUtilities.invokeLater(new Runnable()
						{

							/*
							 * (non-Javadoc)
							 * @see java.lang.Runnable#run()
							 */
							public void run()
							{
								((LySwingDialog)_toolBar.getParentView().getInternalView())
									.repack();
							}
						});
				}
			}
		}
	} // end method propertyChange

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _comp;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent e)
	{
		// A menu item has been clicked
		if (_popupAssoc != null)
		{
			LyTool tool = _popupAssoc.get(e.getSource());

			if (tool != null)
			{
				LyActionEvent actionEvent = new LyActionEvent(this);

				actionEvent.setValue(tool.getId());
				actionPerformed(actionEvent);
			}
		}
		else if (_expandScrollPanel != null)
		{
			LyJButton		button = (LyJButton)e.getSource();

			if (_expandScrollPanelPreferredWidth <= 0)
				_expandScrollPanelPreferredWidth = _expandScrollPanel.getPreferredSize().width;
			
			Rectangle		bounds = _expandScrollPanel.getBounds();
			int				minWidth = 1;
			boolean			collapse = (_expandScrollPanel.getPreferredSize().width > minWidth);
			int				newWidth = (collapse ? minWidth : _expandScrollPanelPreferredWidth);

			_expandScrollPanel.setPreferredSize(new Dimension(newWidth, _expandScrollPanel.getPreferredSize().height));			
			_expandScrollPanel.setBounds(bounds.x, bounds.y, newWidth, bounds.height);
			_expandScrollPanel.getParent().getParent().getParent().validate();
			_expandScrollPanel.getParent().getParent().getParent().doLayout();
			button.setImage(getImage(collapse ? "LY_EXPANDBAR_EXPAND" : "LY_EXPANDBAR_COLLAPSE"));
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		if (actionEvent.getType() == LyActionDemand.TYPE_ACTION)
		{
			if ((actionEvent.getSource() == _comp) && (_toolBar.getStyle() == LyToolBar.Style.MENU))
			{
				if (_popup == null)
				{
					String label = translate(_toolBar.getLabel());
					Font   font = getFont("LY_MENU_FONT");

					_popup = createPopupMenu(label, font);

					Object toolItem = null;

					int    i, n = _toolBar.getComponentCount();
					LyTool tool = null;
					String toolLabel;

					_popupAssoc = new Hashtable<Object, LyTool>();

					for (i = 0; i < n; i++)
					{
						tool = (LyTool)_toolBar.getComponent(i);

						toolLabel = tool.getLabel();
						toolLabel = translate(toolLabel);

						if (tool.isDialog())
							toolLabel = toolLabel + "...";

						toolItem = createMenuItem(toolLabel, font);

						addItemToMenu(toolItem, _popup);
						// register for user action
						registerActionEvents(toolItem, this);
						// keep association between popup menu and the action
						_popupAssoc.put(toolItem, tool);
					}
				}

				Component parent = _comp.getParent();
				Dimension compDim = _comp.getSize();

				addPopupToComponent(_popup, parent);
				showPopup(_popup, _comp, 0, compDim.height);

				actionEvent.consume();
			}
			else
			{
				String controllerId = _toolBar.getController();

				if ((controllerId == null)
                    && (_actionTransmitter == null))
				{
					LyComponent parent = _toolBar.getParent();

					if (parent instanceof LyGetControllerInterface)
						controllerId = ((LyGetControllerInterface)parent).getController();
				}

				if (controllerId != null)
				{
					// notify controller
					LyActionDemand demand = new LyActionDemand(LyActionDemand.TYPE_ACTION,
							actionEvent.getSourceId());
					Object		   source = actionEvent.getSource();

					if (source != null)
					{
						int n = _toolBar.getComponentCount();
						int idx = -1;

						for (int i = 0; (i < n) && (idx < 0); i++)
						{
							LyComponent tool = _toolBar.getComponent(i);

							// Search for the event sourvce
							if ((tool != null) && (tool.getInternalView() == source))
								idx = i; // Ends loop : found index
						}

						if (idx >= 0)
							demand.addIndex(idx);
					}

					// set this id as value (Useful ?)
					demand.setValue(_toolBar.getId());
					demand.setAdditionalValue(actionEvent.getAdditionalValue());
					processDemand(controllerId, demand);
					actionEvent.consume();
				}
			}
		}

		if (!actionEvent.isConsumed())
		{
			// Notify controller if any
			if (_toolBar.getController() != null)
			{
				// Notify controller
				LyDemand demand = LySwingArea.buildActionDemand(actionEvent);

				processDemand(_toolBar.getController(), demand);
				actionEvent.consume();
			}
		}

		if (_actionTransmitter != null)
			_actionTransmitter.processActionEvent(actionEvent);
	} // end method actionPerformed

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return (_actionTransmitter);
	}

	/**
	 * This tool bar doesn't need to handle selections. This method doesn't do anything.
	 *
	 * @return <code>null</code>
	 * @see    LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		return null;
	}

	/**
	 * This method is not implemented yet.
	 *
	 * @param open not used here
	 * @see   leon.view.graphic.LyGraphicalToolBar#setOpen(boolean)
	 */
	public void setOpen(boolean open)
	{
		// XXX To implement
	}

	/**
	 * This method is not implemented yet.
	 *
	 * @param component not used here
	 * @param pos       not used here
	 * @see   leon.view.graphic.LyGraphicalToolBar#addComponent(leon.view.LyComponent, int)
	 */
	public void addComponent(LyComponent component, int pos)
	{
		// XXX To implement
	}

	/**
	 * Removes a sub-component component.
	 *
	 * @param component The component to remove to components list.
	 */
	/**
	 * This method is not implemented yet.
	 *
	 * @param component not used here
	 * @see   leon.view.graphic.LyGraphicalToolBar#removeComponent(leon.view.LyComponent)
	 */
	public void removeComponent(LyComponent component)
	{
		// XXX To implement
	}
	
	/**
	 * Get a lock on the table to prevent concurrent access on rows
	 * @return a valid lock for synchronized()
	 */
	public Object getLock()
	{
		if (_comp != null)
			return _comp.getTreeLock();
		else
			return this;
	}	
} // end class LySwingToolBar
