/*
 * LySwingTree.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.tree.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Tree object (as described in the dtd) for swing viewer.
 *
 * <p>A tree allows to display a set of elements under the form of a tree. It is possible to specify
 * its selection mode (none, simple, multiple or sensitive).</p>
 *
 * <p>Such a component is interesting to display a hierarchy of objects of an application. It allows
 * to not display some sub-sets of these objects.</p>
 *
 * @version $Id: LySwingTree.java 35031 2012-03-23 17:14:12Z jlsaouli $
 * @author  Lyria
 * @see     LySwingTree
 */
public class LySwingTree extends LySwingPanel
	implements LySwingComponent,
		LyActionListener,
		LySelectionListener,
		LySetWaitingInterface,
		LyGraphicalTree,
		LyIsIdleInterface,
		LySetIdleInterface, 
		LyGetLockInterface
{
	public static final String		 __VERSION =
		"$Id: LySwingTree.java 35031 2012-03-23 17:14:12Z jlsaouli $";

	/** Minimum size for the viewport. */
	public static Dimension			 MINIMUM_SIZE = new Dimension(300, 100);

	/** The swing component used to display this tree. */
	protected LyJTreeComponent		 _treeComp;

	/** The scrolling area of this tree. */
	protected JScrollPane			 _scrollingArea;

	/** The root node. */
	protected LyTreeNode			 _rootNode;

	/** The view independent tree description. */
	protected LyTree				 _tree;

	/** Provides notification events on component's selections. */
	protected LySelectionTransmitter _selectionTransmitter;

	/** Idle mode. If true, this component ignores notifications. */
	protected boolean				 _idle = false;

	/** Indicates if root node has changed */
	protected boolean				 _updateRootNode;

	/**
	 * Class constructor of a Tree object for the Swing target.
	 *
	 * @param tree a TREE component defined for every type of view targets. This tree is initialized
	 *             specificaly for the Swing view.
	 */
	public LySwingTree(LyTree tree)
	{
		super(tree);
		_tree = tree;

		initialize();
		setContent();
	}

	/**
	 * Initializes the tree component cell renderers, fonts and colors, then register its events
	 * listeners.
	 */
	public void initialize()
	{
		boolean    dragNDrop = getEnvironment().testResource("LY_DRAG_AND_DROP");
		
		_rootNode = _tree.getTreeNode();
		
		// If root node is null, create a temporary empty node
		if (_rootNode == null)
		{
			boolean wasIdle = _idle;

			_idle = true;
			_rootNode = new LyTreeNode();
			_tree.setTreeNode(_rootNode);
			_idle = wasIdle;
		}

		LySwingTreeNode internal = (LySwingTreeNode)_rootNode.getInternalView();

		_updateRootNode = false;

		// Compute tree validity status
		if (getEnvironment().testResource("LY_TREE_VALIDITY"))
			internal.computeTreeNodeValidity(_rootNode);

		_treeComp = new LyJTreeComponent(internal.getTreeElement(), dragNDrop)
			{
				private static final long serialVersionUID = -2722099791854814394L;

				/*
				 * (non-Javadoc)
				 * @see javax.swing.Scrollable#getPreferredScrollableViewportSize()
				 */
				public Dimension getPreferredScrollableViewportSize()
				{
					Dimension dim = getPreferredSize();
					int		  width, height;

					width = Math.max(LySwingTree.MINIMUM_SIZE.width, dim.width);
					height = Math.max(LySwingTree.MINIMUM_SIZE.height, dim.height);
					return new Dimension(width, height);
				}
			};

		((DefaultTreeCellRenderer)_treeComp.getCellRenderer()).setOpenIcon(new ImageIcon(
				getImage("LY_DIR_OPEN")));
		((DefaultTreeCellRenderer)_treeComp.getCellRenderer()).setClosedIcon(new ImageIcon(
				getImage("LY_DIR_CLOSED")));
		((DefaultTreeCellRenderer)_treeComp.getCellRenderer()).setLeafIcon(new ImageIcon(
				getImage("LY_DIR_CLOSED")));

		TreeSelectionModel selectionModel = _treeComp.getSelectionModel();

		if (_tree.getSelectionMode() == LyTree.SelectionMode.NONE)
			_treeComp.setSelectionModel(null);
		else if (_tree.getSelectionMode() == LyTree.SelectionMode.SINGLE)
			selectionModel.setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
		else if (_tree.getSelectionMode() == LyTree.SelectionMode.MULTIPLE)
			selectionModel.setSelectionMode(TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
		else if (_tree.getSelectionMode() == LyTree.SelectionMode.CHECKBOX)
		{
			selectionModel.setSelectionMode(TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
			_treeComp.setCheckboxMode(true);
		}
		else if (_tree.getSelectionMode() == LyTree.SelectionMode.SENSITIVE)
		{
			selectionModel.setSelectionMode(DefaultTreeSelectionModel.SINGLE_TREE_SELECTION);
			_treeComp.setSensitive(true);
		}
		
		if (_tree.getSelectionType() == LyTree.SelectionType.LEAVES_ONLY)
			_treeComp.setSelectLeavesOnly(true);

		Color color = null;

		if (_tree.getColor() != null)
			color = getColor(_tree.getColor());
		if (color == null)
			color = getColor("LY_TREE_BACKGROUND");
		if (color == null)
			color = getColor("LY_DEFAULT_BACKGROUND");
		if (color != null)
			_treeComp.setBackground(color);

		Color foregroundColor = null;

		if (_tree.getFontColor() != null)
			foregroundColor = getColor(_tree.getFontColor());
		if (foregroundColor != null)
			_treeComp.setForeground(color);

		Font font = null;

		if (_tree.getFont() != null)
			font = getFont(_tree.getFont());
		if (font == null)
			font = getFont("LY_TREENODE_FONT");
		if (font == null)
			font = getFont("LY_DEFAULT_FONT");
		if (font != null)
			_treeComp.setFont(font);

		/* register for events */
		LyActionTransmitter    actionTransmitter = _treeComp.getActionTransmitter();
		LySelectionTransmitter selectionTransmitter = _treeComp.getSelectionTransmitter();

		if (actionTransmitter != null)
			actionTransmitter.addActionListener(this);

		if (selectionTransmitter != null)
			selectionTransmitter.addSelectionListener(this);
	} // end method initialize

	/**
	 * The tree is empty, thus this content setting doesn't do anything.
	 */
	public void setContent()
	{
	}

	/**
	 * This implementation does not set the tree title label.
	 *
	 * @param label new label, not used in this implementation
	 * @see   LySetLabelInterface#setLabel
	 */
	public void setLabel(String label)
	{
		// / XXX Title bar ?
	}

	/**
	 * This implementation does not set the tree background color yet.
	 *
	 * @param color value of background color, not used
	 */
	public void setColor(String color)
	{
		// TO DO
	}

	/**
	 * This implementation does not set the font yet.
	 *
	 * @param font value of label font, not used
	 */
	public void setFont(String font)
	{
		// TO DO
	}

	/**
	 * This implementation does not set the tree foreground color yet.
	 *
	 * @param color value of foreground color, not used
	 */
	public void setForegroundColor(String color)
	{
		// TO DO
	}

	/**
	 * Sets the root tree node.
	 *
	 * @param treeNode new node at the top level of the tree
	 * @see   LyGraphicalTree#setTreeNode(LyTreeNode)
	 */
	public void setTreeNode(LyTreeNode treeNode)
	{
		if (treeNode == null)
			return;

		_rootNode = treeNode;

		if (!_idle)
		{
			LySwingTreeNode internal = (LySwingTreeNode)treeNode.getInternalView();

			((DefaultTreeModel)_treeComp.getModel()).setRoot(internal.getMutableTreeNode());
			_treeComp.updateTree(internal.getTreeElement());

			// Compute tree validity status
			if (getEnvironment().testResource("LY_TREE_VALIDITY"))
				internal.computeTreeNodeValidity(treeNode);
		}
		else
			_updateRootNode = true;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		if (_tree.getController() != null)
		{
			// Notify controller
			LyDemand demand;

			if (actionEvent.getType() == LyMenuDemand.TYPE_POPUP)
			{
				demand = new LyMenuDemand(LyMenuDemand.TYPE_POPUP, _tree.getId());

				((LyMenuDemand)demand).setItem(actionEvent.getValue());
				((LyMenuDemand)demand).setPosition((int[])actionEvent.getParameter());
			}
			else if ((actionEvent.getType() >= LyDragDropDemand.TYPE_DRAG_START)
					&& (actionEvent.getType() <= LyDragDropDemand.TYPE_DRAG_OVER))
			{
				String elementId = actionEvent.getValue();

				demand = new LyDragDropDemand(actionEvent.getType(), elementId);
				((LyDragDropDemand)demand).setValue(actionEvent.getValue());
			}
			else
			{
				demand = new LyActionDemand(actionEvent.getType(), _tree.getId());

				// attempt to get the value
				((LyActionDemand)demand).setValue(actionEvent.getValue());
			}

			processDemand(demand);
			actionEvent.consume();
		}

		// may notify any listeners
		if (_actionTransmitter != null)
			_actionTransmitter.processActionEvent(actionEvent);
	} // end method actionPerformed

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LySelectionListener#selectionPerformed(leon.view.event.LySelectionEvent)
	 */
	public void selectionPerformed(LySelectionEvent selectionEvent)
	{
		Vector<?> nodes = ((Vector<?>)selectionEvent.getItem());

		if ((_tree.getController() != null) && (!isIdle()))
		{
			int				  type_sel = selectionEvent.isSelected()
				? LySelectionDemand.TYPE_SELECT : LySelectionDemand.TYPE_UNSELECT;

			// Notify controller
			LySelectionDemand demand = new LySelectionDemand(type_sel, _tree.getId());

			for (int i = 0; i < nodes.size(); i++)
			{
				DefaultMutableTreeNode mutableTreeNode = (DefaultMutableTreeNode)nodes.elementAt(i);
				String				   id = ((LyJTreeElement)mutableTreeNode.getUserObject())
						.getId();
				// set item id
				if (id != null)
					demand.addSelectionId(id);
			}

			processDemand(demand);
			selectionEvent.consume();
		}

		// may notify any listeners
		if (_selectionTransmitter != null)
			_selectionTransmitter.processSelectionEvent(selectionEvent);
	} // end method selectionPerformed

	/**
	 * Sets the waiting status of the swing tree component. This waiting status changes the cursor
	 * on the component ({@link LySwingViewManager#DEFAULT_CURSOR} or {@link
	 * LySwingViewManager#WAIT_CURSOR}) as well as the enabled property of this tree.
	 *
	 * @param status the waiting status to use on this tree component
	 * @see   LySetWaitingInterface#setWaiting
	 */
	public void setWaiting(boolean status)
	{
		if (_treeComp != null)
		{
			synchronized (getLock()) 
			{
				if (status)
					_treeComp.setCursor(LySwingViewManager.WAIT_CURSOR);
				else
					_treeComp.setCursor(LySwingViewManager.DEFAULT_CURSOR);

				_treeComp.setEnabled(!status);
			}
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetIdleInterface#setIdle(boolean)
	 */
	public void setIdle(boolean status)
	{
		if (status != _idle)
		{
			synchronized (getLock()) 
			{
				_idle = status;
				_treeComp.setIdle(_idle);
		  		
				if (!_idle)
				{
					// Do not notify internal selections that may be triggered by the tree component
					// while tree is being rebuilt.
					boolean	oldNotify = _treeComp.isNotifySelections();
					
					_treeComp.setNotifySelections(false);
					
					if (_updateRootNode)
					{
						setTreeNode(_rootNode);
						_updateRootNode = false;
					}
					
			 		LySwingTreeNode 	internal	= (LySwingTreeNode)_rootNode.getInternalView();
	
			 		if (internal != null)
			 		{
				 		_treeComp.updateTree(internal.getTreeElement());
		
						// Compute tree validity status
						if (getEnvironment().testResource("LY_TREE_VALIDITY"))
							internal.computeTreeNodeValidity(_rootNode);
			 		}
					
					_treeComp.setNotifySelections(oldNotify);
				}
			}
		}
	}

	/**
	 * Adds the specified popup menu to the tree component and shows it.
	 *
	 * @param popupMenu the popup menu to show
	 * @see   LySetPopupMenuInterface#setPopupMenu(LyPopupMenu)
	 */
	public void setPopupMenu(LyPopupMenu popupMenu)
	{
		super.setPopupMenu(popupMenu);
		
        if (popupMenu == null)
        {
            _treeComp.setPopupMenu(null);
            return;
        }

		LySwingPopupMenu internal = (LySwingPopupMenu)popupMenu.getInternalView();
		JPopupMenu		 popup = (JPopupMenu)internal.getMenu();

		_treeComp.setPopupMenu(popup);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return (_actionTransmitter);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		if (_selectionTransmitter == null)
			_selectionTransmitter = new LySimpleSelectionTransmitter();

		return (_selectionTransmitter);
	}

	/**
	 * This tree doesn't have active nodes. This method don't do anything.
	 *
	 * @param activeNode node to activate. Not used in this implementation.
	 * @see   leon.view.graphic.LyGraphicalTree#setActiveNode(java.lang.String)
	 */
	public void setActiveNode(String activeNode)
	{
	}

	/**
	 * Returns the idle mode of this tree.
	 *
	 * @return current idle status
	 * @see    LyIsIdleInterface#isIdle
	 */
	public boolean isIdle()
	{
		return _idle;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		if (_scrollingArea == null)
		{
			_scrollingArea = new JScrollPane(_treeComp);

			if (_tree.getEnvironment().testResource("LY_CHECK_WINDOW_MINIMUM_SIZE"))
				_treeComp.addComponentListener(new _LyTreeComponentAdapter());
		}
		return _scrollingArea;
	}
	
	/**
	 * Returns the component used to display popup menus.
	 * 
	 * @return the parent component of popup menus.
	 */
	protected JComponent getPopupMenuParent()
	{
		return _treeComp;
	}
	
	/**
	 * Get a lock on the table to prevent concurrent access on rows
	 * @return a valid lock for synchronized()
	 */
	public Object getLock()
	{
		Object lock;
	
		if (_treeComp != null)
			return _treeComp.getTreeLock();
		else
			lock = this;
			
		return lock;
	}
	/**
	 * Handles component resize and checks if minimum sizes are valid.
	 */
	public class _LyTreeComponentAdapter extends ComponentAdapter
	{

		/** The last dimension of the component. */
		protected Dimension _last;

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.ComponentListener#componentResized(java.awt.event.ComponentEvent)
		 */
		public void componentResized(ComponentEvent event)
		{
			// GO BACK TO THE TOP and specify a MINIMUM size
			Dimension d = event.getComponent().getMinimumSize();

			if (d.width == 0)
				// at least few pixels
				d = new Dimension(100, d.height);
			if (d.height == 0)
				// at least few pixels
				d = new Dimension(d.width, 100);

			// Do we have anything to do ?
			if ((_last != null) && (d.equals(_last)))
				return;

			_last = d;

			// Notify hierachy
			if (_tree.getParent().getInternalView() instanceof LySwingContainer)
				((LySwingContainer)_tree.getParent().getInternalView()).updateMinimumSize(_tree, d);
		}
	} // end class _LyTreeComponentAdapter
	


} // end class LySwingTree
