/*
 * LySwingTreeNode.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;

import javax.swing.event.*;
import javax.swing.tree.*;

import leon.view.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;
import leon.view.win.swing.component.LyJTreeElementImpl.*;


/**
 * This class implements the Tree Node object (as described in the dtd) for swing viewer.
 *
 * <p>A tree node can contain a set of nodes (this node is then called parent node of these last
 * ones). It is possible to assign a label and several images to a tree node, and to indicate if it
 * is selected, valid or open. It is also possible to indicate the font and the different colors of
 * the node.</p>
 *
 * @version $Id: LySwingTreeNode.java 34778 2012-03-16 15:27:38Z jlsaouli $
 * @author  Lyria
 * @see     LyTreeNode
 */
public class LySwingTreeNode extends LySwingObject
	implements LySetLabelInterface,
		LySetSelectedInterface,
		LySetImageInterface,
		LyGraphicalTreeNode,
		TreeExpansionListener
{
	public static final String   __VERSION =
		"$Id: LySwingTreeNode.java 34778 2012-03-16 15:27:38Z jlsaouli $";

	/** The swing component for this tree node element. */
	protected LyJTreeComponent   _treeComp;

	/** The tree element associated with this tree node. */
	protected LyJTreeElementImpl _treeElement;

	/** Height factor of the current node. */
	protected int				 _heightFactor;

	/** Indicates if this node has already been initialized. */
	protected boolean			 _initialized;

	/** The view independent tree node description. */
	protected LyTreeNode		 _treeNode;

	/**
	 * Class constructor of a Tree node object for the Swing target.
	 *
	 * @param node a TREE_NODE component defined for every type of view targets. This tree node is
	 *             initialized for the swing view and filled with its default values.
	 */
	public LySwingTreeNode(LyTreeNode node)
	{
		super(node);
		_treeNode = node;
		_initialized = false;

		initialize();
		setContent();
	}

	/**
	 * Initializes the content of this node.
	 */
	public void initialize()
	{
		_treeElement = new LyJTreeElementImpl(_treeNode);

		_LyMutableTreeNode mutableTreeNode = new _LyMutableTreeNode();

		mutableTreeNode.setUserObject(_treeElement);

		_treeElement.setMutableTreeNode(mutableTreeNode);

		Image image;

		if (_treeNode.getImage() != null)
			image = (Image)getEnvironment().getImage(_treeNode.getImage());
		else
			image = null;

		mutableTreeNode.setNodeContent(new _LyNodeContent(image, _treeNode.getLabel(), _treeNode.getTooltip(),
				_treeNode.isSelected(), _treeNode.isOpen()));

		_treeElement.setExpandable(_treeNode.isExpandable());
		
		// find the parent tree
		LyComponent parent = _treeNode.getParent();

		while ((parent != null) && !(parent instanceof LyTree))
			parent = parent.getParent();

		boolean hasCheckbox = false;
		boolean leafMode = false;

		if (parent != null)
		{
			LyTree parentTree = (LyTree)parent;

			hasCheckbox = (parentTree.getSelectionMode() == LyTree.SelectionMode.CHECKBOX);
			leafMode = (parentTree.getSelectionType() == LyTree.SelectionType.LEAVES_ONLY);

			if (leafMode && (_treeNode.getTreeNodeCount() > 0))
				hasCheckbox = false;
		}

		if (hasCheckbox)
		{
			_treeElement.setEmptyCheckboxImage(getImage("LY_TREENODE_EMPTY_CHECKBOX"));
			_treeElement.setSelectedCheckboxImage(getImage("LY_TREENODE_CHECKED_CHECKBOX"));
		}

		int     idx = 0;
		LyImage additionalImage;

		do
		{
			additionalImage = _treeNode.getImage(idx);
			idx++;
			if (additionalImage != null)
				_treeElement.addAdditionalImage(getImage(additionalImage.getValue()));
		}
		while (additionalImage != null);

		// We may have a validity status (hierarchy will be computed)
		if ((_treeNode.getEnvironment() != null)
                && _treeNode.getEnvironment().testResource("LY_TREE_VALIDITY"))
			setValidity(_treeNode.getValidity());

		_initialized = true;
	} // end method initialize

	/**
	 * Builds the links to children nodes.
	 *
	 * @see #getContent()
	 */
	public void setContent()
	{
		int				   i, n = _treeNode.getTreeNodeCount();
		_LyMutableTreeNode mutableTreeNode = (_LyMutableTreeNode)getMutableTreeNode();
		_LyMutableTreeNode subMutableTreeNode;

		/* Add all the children nodes */
		for (i = 0; i < n; i++)
		{
			LyTreeNode	    node = _treeNode.getTreeNode(i);
			LySwingTreeNode internal = (LySwingTreeNode)node.getInternalView();

			subMutableTreeNode = (_LyMutableTreeNode)internal.getMutableTreeNode();
			mutableTreeNode.add(subMutableTreeNode);
		}
	}

	/**
	 * Gets the tree element of this tree node.
	 *
	 * @return the tree element
	 */
	public LyJTreeElement getTreeElement()
	{
		return _treeElement;
	}

	/**
	 * Adds the specified node at the specified index.
	 *
	 * @param treeNode the node to add
	 * @param pos      the position of the insertion
	 * @see   LyGraphicalTreeNode#addTreeNode(LyTreeNode, int)
	 */
	public void addTreeNode(LyTreeNode treeNode, int pos)
	{
		_LyMutableTreeNode mutableTreeNode = (_LyMutableTreeNode)getMutableTreeNode();

		if (pos < 0)
			pos = mutableTreeNode.getChildCount();

		int[] indice = new int[1];

		indice[0] = pos;

		LySwingTreeNode internal = (LySwingTreeNode)treeNode.getInternalView();

		mutableTreeNode.insert(internal.getMutableTreeNode(), pos);

		if ((_treeComp == null) && (_treeElement != null))
			_treeComp = _treeElement.getTree();

		if (_treeComp != null)
		{
			((DefaultTreeModel)_treeComp.getModel()).nodesWereInserted(mutableTreeNode, indice);
			_treeComp.updateTree(_treeElement);
		}
	}

	/**
	 * Removes the specified node.
	 *
	 * @param treeNode the node to remove
	 * @see   LyGraphicalTreeNode#removeTreeNode(LyTreeNode)
	 */
	public void removeTreeNode(LyTreeNode treeNode)
	{
		int[]				 indice = new int[1];
		_LyMutableTreeNode[] removed = new _LyMutableTreeNode[1];
		_LyMutableTreeNode   mutableTreeNode = (_LyMutableTreeNode)_treeElement.getMutableTreeNode();
		LySwingTreeNode		 internal = (LySwingTreeNode)treeNode.getInternalView();

		removed[0] = (_LyMutableTreeNode)internal.getMutableTreeNode();
		indice[0] = mutableTreeNode.getIndex(removed[0]);

		mutableTreeNode.remove(removed[0]);

		if (_treeComp == null)
			_treeComp = _treeElement.getTree();

		if (_treeComp != null)
		{
			((DefaultTreeModel)_treeComp.getModel()).nodesWereRemoved(mutableTreeNode, indice, removed);
			_treeComp.updateTree(_treeElement);
		}
	}

	/**
	 * Sets the treeComponent that display this node.
	 *
	 * @param tree the value of the tree component
	 */
	public void setTree(LyJTreeComponent tree)
	{
		_treeComp = tree;
	}

	/**
	 * Sets the label of the title bar.
	 *
	 * @param label new label
	 * @see   #getLabel()
	 * @see   LySetLabelInterface#setLabel(String)
	 */
	public void setLabel(String label)
	{
		_LyMutableTreeNode treeNode = (_LyMutableTreeNode)_treeElement.getMutableTreeNode();

		treeNode.getNodeContent().setLabel(label);

		if (_treeComp == null)
			_treeComp = _treeElement.getTree();

		if (_treeComp != null)
			((DefaultTreeModel)_treeComp.getModel()).nodeChanged(treeNode);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetSelectedInterface#setSelected(boolean)
	 */
	public void setSelected(boolean status)
	{
		if ((_treeComp == null) && (_treeElement != null))
			_treeComp = _treeElement.getTree();

		if (_treeComp == null)
			return;

		_LyMutableTreeNode treeNode = (_LyMutableTreeNode)_treeElement.getMutableTreeNode();

		treeNode.getNodeContent().setSelected(status);

		TreePath   path = getTreePath();
		TreePath[] paths = _treeComp.getSelectionPaths();
		boolean    found = false;
		int		   pathsCount = (paths != null) ? paths.length : 0;

		for (int i = 0; ((i < pathsCount) && !found); i++)
		{
			if (path.equals(paths[i]))
				found = true;
		}

		if ((status && found) || (!status && !found))
			return;

		if (status)
			_treeComp.addSelectionPath(path);
		else
			_treeComp.removeSelectionPath(getTreePath());
	} // end method setSelected

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalTreeNode#setOpen(boolean)
	 */
	public void setOpen(boolean status)
	{
		_LyMutableTreeNode treeNode = (_LyMutableTreeNode)_treeElement.getMutableTreeNode();

		treeNode.getNodeContent().setOpen(status);
		
		if (_treeComp == null)
			_treeComp = _treeElement.getTree();

		if (_treeComp == null)
			return;
		
		TreePath   path = getTreePath();
		
		if (status)
			_treeComp.expandRow(_treeComp.getRowForPath(path));
		else
			_treeComp.collapseRow(_treeComp.getRowForPath(path));
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetHiddenInterface#setHidden(boolean)
	 */
	public void setHidden(boolean hidden)
	{
		LyComponent parent = _treeNode.getParent();

		if (parent instanceof LyTreeNode)
		{
			LyTreeNode	    parentTreeNode = (LyTreeNode)parent;
			LySwingTreeNode internal = (LySwingTreeNode)parentTreeNode.getInternalView();

			if (internal != null)
			{
				if (hidden)
					internal.removeTreeNode(_treeNode);
				else
				{
					int		   pos = 0, i, n = parentTreeNode.getTreeNodeCount();
					LyTreeNode child;

					for (i = 0; i < n; i++)
					{
						child = parentTreeNode.getTreeNode(i);

						if (_treeNode.equals(child))
							break;

						if (!child.isHidden())
							pos++;
					}

					internal.addTreeNode(_treeNode, pos);
				}
			}
		}
	} // end method setHidden

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetImageInterface#setImage(java.lang.String)
	 */
	public void setImage(String key)
	{
		_LyMutableTreeNode treeNode = (_LyMutableTreeNode)_treeElement.getMutableTreeNode();

		if (key != null)
			treeNode.getNodeContent().setImage(getImage(_treeNode.getImage()));

		if ((_treeComp == null) && (_treeElement != null))
			_treeComp = _treeElement.getTree();

		if (_treeComp != null)
			((DefaultTreeModel)_treeComp.getModel()).nodeChanged(treeNode);
	}

	/**
	 * Nothing is done in this method. The color setting is not implemented yet.
	 *
	 * @param color value of background color, not used here
	 */
	public void setColor(String color)
	{
		_LyMutableTreeNode treeNode = (_LyMutableTreeNode)_treeElement.getMutableTreeNode();

		if (_treeComp != null)
			((DefaultTreeModel)_treeComp.getModel()).nodeChanged(treeNode);
	}

	/**
	 * Nothing is done in this method. The font setting is not implemented yet.
	 *
	 * @param font value of label font
	 */
	public void setFont(String font)
	{
		_LyMutableTreeNode treeNode = (_LyMutableTreeNode)_treeElement.getMutableTreeNode();

		if (_treeComp != null)
			((DefaultTreeModel)_treeComp.getModel()).nodeChanged(treeNode);
	}

	/**
	 * Nothing is done in this method. The foreground color setting is not implemented yet.
	 *
	 * @param color value of foreground color
	 */
	public void setForegroundColor(String color)
	{
		_LyMutableTreeNode treeNode = (_LyMutableTreeNode)_treeElement.getMutableTreeNode();

		if (_treeComp != null)
			((DefaultTreeModel)_treeComp.getModel()).nodeChanged(treeNode);
	}

	/**
	 * Gets this tree node description.
	 *
	 * @return the decription of the node
	 */
	public LyTreeNode getTreeNode()
	{
		return _treeNode;
	}

	/**
	 * Returns a TreePath for this node.
	 *
	 * @return the tree path for this node
	 */
	public TreePath getTreePath()
	{
		return _treeElement.getTreePath();
	}

	/**
	 * Accessor to the internal mutable tree node.
	 *
	 * @return the mutable tree node of this tree element
	 */
	public DefaultMutableTreeNode getMutableTreeNode()
	{
		return _treeElement.getMutableTreeNode();
	}

	/**
	 * Gets the content of this tree node tree element.
	 *
	 * @return the content of the node
	 * @see    #setContent()
	 */
	public _LyNodeContent getContent()
	{
		return _treeElement.getContent();
	}

	/**
	 * Gets the identifier of this tree node.
	 *
	 * @return the tree node Id
	 */
	public String getId()
	{
		return _treeNode.getId();
	}

	/**
	 * Gets this tree node label.
	 *
	 * @return the tree node label
	 * @see    #setLabel(String)
	 */
	public String getLabel()
	{
		return _treeNode.getLabel();
	}

	/**
	 * Indicates if this tree node is selected or not.
	 *
	 * @return true if the tree node is selected, false if not
	 */
	public boolean isSelected()
	{
		return _treeNode.isSelected();
	}

	/**
	 * Indicates if this tree node is opened or not.
	 *
	 * @return true if the tree node is opened, false if not
	 */
	public boolean isOpen()
	{
		return _treeNode.isOpen();
	}

	/**
	 * Sets the height factor of this current node.
	 *
	 * @param factor the new height factor
	 * @see   #getHeightFactor()
	 */
	public void setHeightFactor(int factor)
	{
		_heightFactor = factor;
	}

	/**
	 * Returns the height factor.
	 *
	 * @return the height factor of the node
	 * @see    #setHeightFactor(int)
	 */
	public int getHeightFactor()
	{
		return _heightFactor;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalTreeNode#addImage(leon.view.LyImage, int)
	 */
	public void addImage(LyImage image, int pos)
	{
		_treeElement.addAdditionalImage(getImage(image.getValue()));
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalTreeNode#removeImage(leon.view.LyImage)
	 */
	public void removeImage(LyImage image)
	{
		_treeElement.removeAdditionalImage(getImage(image.getValue()));
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetValidityInterface#setValidity(java.lang.String)
	 */
	public void setValidity(LyTreeNode.Validity validity)
	{
		if ((_treeElement != null) && (validity != null))
		{
			if (!_treeNode.getEnvironment().testResource("LY_TREE_VALIDITY"))
				return;

			_treeElement.setValidityImage(getImage("LY_TREE_VALIDITY_" + validity.toString()));

			if ((!_initialized) || (_treeNode.getTreeNodeCount() == 0))
				return;

			_treeElement.updateValidityFromView();

			_LyMutableTreeNode treeNode = (_LyMutableTreeNode)_treeElement.getMutableTreeNode();

			if (_treeComp == null)
				_treeComp = _treeElement.getTree();

			if (_treeComp != null)
				((DefaultTreeModel)_treeComp.getModel()).nodeChanged(treeNode);
		}
	}

	/**
	 * Computes the subtree validity.
	 *
	 * @param node the current node
	 * @see   leon.view.swt.widgets.LyTreeElement#computeTreeNodeValidity(leon.view.LyComponent)
	 */
	public void computeTreeNodeValidity(LyComponent node)
	{
		_treeElement.computeTreeNodeValidity(node);
	}

	/**
	 * Accesses to the top level tree node.
	 *
	 * @param  treeNode the tree node
	 * @return the top level root tree node
	 */
	public LyTreeNode getRootTreeNode(LyTreeNode treeNode)
	{
		while ((treeNode != null) && (treeNode.getParent() instanceof LyTreeNode))
			treeNode = (LyTreeNode)treeNode.getParent();

		return treeNode;
	}

	/**
	 * Nothing is done when the tree is expanded.
	 *
	 * @param event the <code>TreeExpansionEvent</code>, not used in this implementation
	 * @see   javax.swing.event.TreeExpansionListener#treeExpanded(javax.swing.event.TreeExpansionEvent)
	 */
	public void treeExpanded(TreeExpansionEvent event)
	{
	}

	/**
	 * Nothing is done when the tree is collapsed.
	 *
	 * @param event the <code>TreeExpansionEvent</code>, not used in this implementation
	 * @see   javax.swing.event.TreeExpansionListener#treeCollapsed(javax.swing.event.TreeExpansionEvent)
	 */
	public void treeCollapsed(TreeExpansionEvent event)
	{
	}

	/**
	 * Sets the expandable attribute. This method is not implemented yet.
	 *
	 * @param expandable not used in this implementation yet
	 * @see   leon.view.graphic.LyGraphicalTreeNode#setExpandable(boolean)
	 */
	public void setExpandable(boolean expandable)
	{
		_treeElement.setExpandable(expandable);
	}
	
	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalTreeNode#setTooltip(java.lang.String)
	 */
	public void setTooltip(String tooltip) 
	{
		_LyMutableTreeNode mutableTreeNode = (_LyMutableTreeNode)getMutableTreeNode();
		
		mutableTreeNode.getNodeContent().setTooltip(tooltip);
		
		if ((_treeComp == null) && (_treeElement != null))
			_treeComp = _treeElement.getTree();

		if (_treeComp != null)
			((DefaultTreeModel)_treeComp.getModel()).nodeChanged(mutableTreeNode);
	}

} // end class LySwingTreeNode
