/*
 * LySwingUtils.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.view.win.swing;

import java.awt.*;
import java.awt.image.*;
import java.io.*;
import javax.imageio.*;


/**
 * Utility class for Swing viewers.
 *
 * @author Lyria
 */
public class LySwingUtils
{
	public static final String __VERSION =
		"$Id: LySwingUtils.java 36613 2012-08-28 11:02:44Z nhamon $";

	/**
	 * Returns a new color that is value% darker.
	 *
	 * @param  color the current color
	 * @param  value the darkness percentage
	 * @return the new darker color
	 */
	public static Color getDarkerColor(Color color, int value)
	{
		int red = color.getRed();
		int blue = color.getBlue();
		int green = color.getGreen();

		red = red * value / 100;
		blue = blue * value / 100;
		green = green * value / 100;

		return new Color(red, green, blue);
	}

	/**
	 * Returns a new color that is value% lighter.
	 *
	 * @param  color the current color
	 * @param  value the lightness percentage
	 * @return the new lighter color
	 */
	public static Color getLighterColor(Color color, int value)
	{
		int red = color.getRed();
		int blue = color.getBlue();
		int green = color.getGreen();

		if (red == 0)
			red = value;

		if (blue == 0)
			blue = value;

		if (green == 0)
			green = value;

		red = red + (red * value / 100);
		blue = blue + (blue * value / 100);
		green = green + (green * value / 100);

		if (red > 255)
			red = 255;
		if (blue > 255)
			blue = 255;
		if (green > 255)
			green = 255;

		return new Color(red, green, blue);
	} // end method getLighterColor

	/**
	 * Draws a tube-style line in the given graphics context from the point (x1,y1) to (x2,y2) using
	 * the specified color with the specified width.
	 *
	 * @param g     the graphics context
	 * @param x1    the x coordinate of the starting point
	 * @param y1    the y coordinate of the starting point
	 * @param x2    the x coordinate of the ending point
	 * @param y2    the y coordinate of the ending point
	 * @param width the width of the line
	 * @param color color of the drawn tube
	 */
	public static void drawTube(Graphics g, int x1, int y1, int x2, int y2, int width, Color color)
	{
		if (width <= 0)
			return;

		int[] xPoints;
		int[] yPoints;
		int   red = color.getRed(), green = color.getGreen(), blue = color.getBlue();
		int   w;
		int   offset = 400 / width;

		for (w = width; w > 0; w = w - 2)
		{
			// Calculate the polygon containing the link
			Point[] polygon = getPolygon(x1, y1, x2, y2, w);

			xPoints = new int[] { polygon[0].x, polygon[1].x, polygon[2].x, polygon[3].x };
			yPoints = new int[] { polygon[0].y, polygon[1].y, polygon[2].y, polygon[3].y };

			g.setColor(color);
			g.fillPolygon(xPoints, yPoints, 4);

			red = ((red > (255 - offset)) ? 255 : (red + offset));
			green = ((green > (255 - offset)) ? 255 : (green + offset));
			blue = ((blue > (255 - offset)) ? 255 : (blue + offset));

			color = new Color(red, green, blue);
		}
	} // end method drawTube

	/**
	 * Returns the polygon where (x1,y1) and (x2,y2) are centers of two opposite sides of the
	 * specified width.
	 *
	 * @param  x1    the x coordinate of the first point
	 * @param  y1    the y coordinate of the first point
	 * @param  x2    the x coordinate of the second point
	 * @param  y2    the y coordinate of the second point
	 * @param  width the width of the polygon
	 * @return the list of points of the new polygon
	 */
	public static Point[] getPolygon(int x1, int y1, int x2, int y2, int width)
	{
		double alpha;

		if (y2 == y1)
			alpha = (Math.PI) / 2;
		else
			alpha = Math.atan(((float)(x2 - x1)) / (y2 - y1));

		width++;

		int   dx = (int)((float)(width * Math.cos(alpha))) / 2;
		int   dy = -(int)((float)(width * Math.sin(alpha))) / 2;

		Point point1 = new Point(x1 - dx, y1 - dy);
		Point point2 = new Point(x2 - dx, y2 - dy);
		Point point3 = new Point(x2 + dx, y2 + dy);
		Point point4 = new Point(x1 + dx, y1 + dy);

		return new Point[] { point1, point2, point3, point4 };
	}

	/**
	 * Draws a focus rectangle in the given graphics context using the given coordinates.
	 *
	 * @param g      the graphics context
	 * @param x1     the x coordinate of the left border of the point
	 * @param y1     the y coordinate of the right border of the point
	 * @param width  the width of the rectangle
	 * @param height the height of the rectangle
	 */
	public static void drawFocus(Graphics g, int x1, int y1, int width, int height)
	{
		g.setColor(Color.gray);
		drawDotLine(g, x1, y1, x1 + width, y1);
		drawDotLine(g, x1 + width, y1, x1 + width, y1 + height);
		drawDotLine(g, x1, y1 + height, x1 + width, y1 + height);
		drawDotLine(g, x1, y1, x1, y1 + height);
	}

	/**
	 * Draws a dotline in the given graphics context using the given coordinates.
	 *
	 * @param g  the graphics context
	 * @param x1 the x coordinate of the starting point
	 * @param y1 the y coordinate of the starting point
	 * @param x2 the x coordinate of the ending point
	 * @param y2 the y coordinate of the ending point
	 */
	public static void drawDotLine(Graphics g, int x1, int y1, int x2, int y2)
	{
		int     width = Math.abs(x2 - x1);
		int     height = Math.abs(y2 - y1);

		int     maxDim = Math.max(width, height);
		boolean horizontalParsing = (maxDim == width);
		int     i, coord;

		// May be factorized
		if (horizontalParsing)
		{
			for (i = x1; i < (x1 + maxDim); i++)
			{
				coord = getOddNumber((i * (height / width)) + y1);

				if ((i % 2) == (coord % 2))
					g.drawLine(i, coord, i, coord);
			}
		}
		else
		{
			for (i = y1; i < (y1 + maxDim); i++)
			{
				i = getOddNumber(i);

				if (width == 0)
				{
					coord = x1;
					
					if ((i % 2) == 1)
						g.drawLine(coord, i, coord, i);
				}
				else
				{
					coord = getOddNumber((i * (width / height)) + x1);

					if ((i % 2) == (coord % 2))
						g.drawLine(coord, i, coord, i);
				}
			}
		}
	} // end method drawDotLine

	/**
	 * Returns the odd number directly superior to the specified number.
	 *
	 * @param  n the number
	 * @return the first superior odd number
	 */
	public static int getOddNumber(int n)
	{
		return (((n / 2) * 2) + 1);
	}

	/**
	 * Draws a filled rectangle with a shadow aspect.
	 *
	 * @param gc     the Graphic Context
	 * @param x      position on the X axis
	 * @param y      position on the Y axis
	 * @param width  Width
	 * @param height Height
	 */
	public static void fillShadowRect(Graphics gc, int x, int y, int width, int height)
	{
		int     i, j;
		boolean draw = true;

		gc.setColor(Color.BLACK);

		for (i = x; i < (x + width); i++)
		{
			draw = ((i % 2) == 0);

			for (j = y; j < (y + height); j++)
			{
				if (draw)
					gc.drawLine(i, j, i, j);

				draw = !draw;
			}
		}
	}

	/**
	 * Draws a background using a texture image.
	 *
	 * @param gc      the Graphic Context
	 * @param width   width of the visible part of the graphics
	 * @param height  height of the visible part of the graphics
	 * @param bgImage the image to be used to draw the background
	 */
	public static void fillBackground(Graphics gc, int width, int height, Image bgImage)
	{
		int imWidth = bgImage.getWidth(null);
		int imHeight = bgImage.getHeight(null);
		int x, y;

		if ((imWidth > 0) && (imHeight > 0))
		{
			for (y = 0; y < height; y += imHeight)
			{
				for (x = 0; x < width; x += imWidth)
					gc.drawImage(bgImage, x, y, null);
			}
		}
	}

	/**
	 * Takes a snapshot of the given component.
	 *
	 * @param comp the component
	 * @param file the destination file
	 */
	public static void snapshot(Component comp, String file)
	{

		Point     p = comp.getLocationOnScreen();
		Rectangle region = comp.getBounds();

		region.x = p.x;
		region.y = p.y;
		BufferedImage image = null;

		try
		{
			image = new Robot().createScreenCapture(region);
		}
		catch (AWTException e)
		{
			e.printStackTrace();
		}
		if ((file == null) || (image == null))
			return;
		int    offset = file.lastIndexOf(".");
		String type = (offset == -1) ? "jpg" : file.substring(offset + 1);

		try
		{
			ImageIO.write(image, type, new File(file));
		}
		catch (IOException e)
		{
			e.printStackTrace();
		}
	} // end method snapshot

	/**
	 * Returns the average width of a character for a given font.
	 *
	 * @param  comp the component used to compute width
	 * @param  font the font
	 * @return the average width of a character for the given font
	 */
	public static int getFontCharWidth(Component comp, Font font)
	{
		FontMetrics metrics = comp.getFontMetrics(font);

		return (metrics.stringWidth("0123456789") / 10);
	}
} // end class LySwingUtils
