/*
 * LySwingView.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;


/**
 * This class implements the common behavior of the View object (as described in the dtd) for swing
 * viewer.
 *
 * @version $Id: LySwingView.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     LyView
 */
public abstract class LySwingView extends LySwingObject
	implements LyGraphicalView,
		LyActionListener,
		LySetVisibleInterface,
		LyIsVisibleInterface,
		LyResetContentInterface,
		LyGetBoundsInterface,
		LySetBoundsInterface,
		LySetWaitingInterface
{
	public static final String    __VERSION =
		"$Id: LySwingView.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Height of Windows task bar.
	 * @deprecated
	 */
	public static int			  WINDOWS_TASKBAR_HEIGHT = 27;

	/** Indicates if the size of the window has been set or automatically calculated. */
	protected boolean			  _resized;

	// workaround for setresizable bug (Unix)
	/** Required size, return to original size for the associated window of this view. */
	protected Dimension			  _requiredSize;

	// Workaround
	/** Handles the size changes randomly on linux. */
	protected Date				  _requiredSizeDate;

	/** Window adapter of this view. Used in removing listeners of the associated view window. */
	protected _LyWindowAdapter    _winAdapter;

	/** Component adapter of this view. Used in removing listeners of the associated view window. */
	protected _LyComponentAdapter _compAdapter;

	/**
	 * Class constructor of a View object for the Swing target.
	 *
	 * @param view a Leonardi view defined for every type of view targets.
	 */
	public LySwingView(LyView view)
	{
		super(view);
	}

	/**
	 * Initializes the view. This method is called by the methods that extend this one.
	 */
	public void initialize()
	{
		_resized = false;
	}

	/**
	 * Gets the view component.
	 *
	 * @return the view component
	 */
	public final LyView getView()
	{
		return (LyView)getViewComponent();
	}

	/**
	 * Gets the window component.
	 *
	 * @return the window component
	 */
	public final Window getWindow()
	{
		return (Window)getComponent();
	}

	/**
	 * Sets the content of this view and registers its event listeners.
	 */
	public void setContent()
	{
		LyView			 view = getView();
		LyArea			 area = view.getArea();
		LySwingComponent internal = (LySwingComponent)(area.getInternalView());

		setArea(internal.getComponent());

		// register events
		if (internal instanceof LyEventTransmitter)
		{
			LyActionTransmitter transmitter = ((LyEventTransmitter)internal).getActionTransmitter();

			transmitter.addActionListener(this);
		}
	}

	/**
	 * Sets the area content of this dialog.
	 *
	 * @param area the component to set
	 */
	protected abstract void setArea(Component area);

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetBoundsInterface#setLocation(int, int)
	 */
	public void setLocation(int x, int y)
	{
		Window window = getWindow();

		if (window == null)
			return;

		// FFT00435/Flandres: Screen locations can be negative but cannot be outside screen limits
		int minX = 0, minY = 0;

		GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
		if (ge != null)
		{
			GraphicsDevice[] devices = ge.getScreenDevices();
			GraphicsDevice device = null;

			Rectangle deviceBounds = null;
			for (int i = 0; i < devices.length; i++)
			{
				device = devices[i];
				deviceBounds = device.getDefaultConfiguration().getBounds();
				minX = Math.min(deviceBounds.x, minX);
				minY = Math.min(deviceBounds.y, minY);
			}
		}

		if (x < minX)
			x = minX;
		if (y < minY)
			y = minY;

		if (window != null)
			window.setLocation(x, y);
	} // end method setLocation

	/**
	 * Gets the view window position in its parent component.
	 *
	 * @return the position of the window, or <code>null</code> in none
	 * @see    #setLocation(int, int)
	 * @see    leon.view.getinterface.LyGetBoundsInterface#getLocation()
	 */
	public int[] getLocation()
	{
		Window window = getWindow();

		if (window == null)
			return null;

		Point location = window.getLocation();

		return new int[] { location.x, location.y };
	}

	/**
	 * Checks if window is resizable or not.
	 *
	 * @return true if the window is resizable, false otherwise
	 */
	public abstract boolean isResizable();

	/**
	 * Sets the size of the view window at the specified width and height. Sets the {@link
	 * #_resized} attribute.
	 *
	 * @param width  the new width of the window
	 * @param height the new height of the window
	 * @see   #getSize()
	 * @see   LySetBoundsInterface#setSize
	 */
	public void setSize(int width, int height)
	{
		Window window = getWindow();

		if (window != null)
		{
			window.setSize(width, height);
			_resized = true;

			if (isResizable())
				_requiredSize = null;
			else
				_requiredSize = new Dimension(width, height);
		}
	}

	/**
	 * Gets the current size of the view window.
	 *
	 * @return the size of the window, or <code>null</code> if none
	 * @see    #setSize(int, int)
	 * @see    LyGetBoundsInterface#getSize
	 */
	public int[] getSize()
	{
		Window window = getWindow();

		if (window == null)
			return null;

		Dimension size;

		if (!_resized)
		{
			pack();
			size = window.getPreferredSize();
		}
		else
			size = window.getSize();

		return new int[] { size.width, size.height };
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyResetContentInterface#resetContent()
	 */
	public void resetContent()
	{
		resetContent(false);
	}
	
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyResetContentInterface#resetContent()
	 */
	public void resetContent(boolean full)
	{
		// Only resize implementation so far (partial)
		Window window = getWindow();

		if (window != null)
			pack();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		LyView view = getView();

		// Notify controller if any
		if (view.getController() != null)
		{
			// Notify controller
			LyDemand demand = LySwingArea.buildActionDemand(actionEvent);

			// Set info :
			demand.setParameterValues(new LySwingParameterValues(view));

			processDemand(demand);
			actionEvent.consume();
		}
	}

	/**
	 * Checks the size of the view. If the size is higher than the max size, it will be resized.
	 */
	protected abstract void checkSize();

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetVisibleInterface#setVisible(boolean)
	 */
	public abstract void setVisible(boolean status);

	/**
	 * Checks if the view window component is visible or not.
	 *
	 * @return true if the component is showing, false if not or if there is no window
	 * @see    LySetVisibleInterface#setVisible
	 */
	public boolean isVisible()
	{
		Window window = getWindow();

		return ((window == null) ? false : window.isVisible());
	}

	/**
	 * Sets the waiting status of this component.
	 *
	 * @param status the new status for this component
	 * @see   LySetWaitingInterface#setWaiting
	 */
	public abstract void setWaiting(boolean status);

	/**
	 * Packs the current window content and repaints it.
	 */
	public void pack()
	{
		Window window = getWindow();

		if (window == null)
			return;

		_requiredSize = null;

		// if visible, it's a resetContent
		if (isVisible())
		{
			window.pack();
			checkSize();
			window.repaint();
		}
		else
		{
			window.pack();
			window.repaint();
		}

		if (!isResizable())
			_requiredSize = window.getSize();
	}

	/**
	 * Sets the default button. At this time no default button is defined in Leonardi, so the policy
	 * is to set the first command button as the default button.
	 */
	public abstract void configureDefaultButton();

	/**
	 * Lays out the window components.
	 */
	protected void doLayout()
	{
		Window window = getWindow();

		if (window == null)
			return;

		_requiredSize = null;
		window.doLayout();

		if (!isResizable())
			_requiredSize = window.getSize();
	}

	/**
	 * Free all resources used by this dialog view. Unshows the window then removes its listeners
	 * before disposing it.
	 *
	 * @see LyGraphicalObject#free()
	 */
	public void free()
	{
		Window window = getWindow();

		if (window != null)
		{
			// Post window freeing in AWT event queue to avoid deadlocks
			LyViewManager viewManager = getViewManager();
			
			viewManager.execGui(new _LyFreeWindow(window), false);
		}

		super.free();
	} // end method free

	/**
	 * Closes the window.
	 */
	protected void close()
	{
		Window window = getWindow();

		if (window == null)
			return;

		LyViewManager	    viewManager = getViewManager();
		LyDemandTransmitter transmitter = viewManager.getDemandTransmitter();
		String			    controllerId = getController();

		if ((controllerId != null) && (transmitter != null)
				&& (transmitter.getListenerById(controllerId) != null))
		{
			LyActionDemand demand = new LyActionDemand(LyActionDemand.TYPE_WINDOW_CLOSED,
					getView().getId());

			processDemand(demand);
		}
		else
		{
			LyView view = getView();

			if (view != null)
				view.free();
		}
	}

	/**
	 * Internal class to wait for closing events.
	 */
	protected class _LyWindowAdapter extends WindowAdapter
	    implements Runnable
	{
		/**
		 * Empty class constructor of a waiting events class.
		 */
		public _LyWindowAdapter()
		{
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.WindowListener#windowClosing(java.awt.event.WindowEvent)
		 */
		public void windowClosing(WindowEvent e)
		{
		    getWindow().requestFocus();
		    
			// Close window asynchronously
		    LyViewManager     viewManager = getViewManager();
		    
			viewManager.execGui(this, false);
		}

		/**
		 * Invoked when a window state is changed.
		 *
		 * @see java.awt.event.WindowListener#windowActivated(java.awt.event.WindowEvent)
		 */
		public void windowActivated(WindowEvent e)
		{
			if (e.getID() == WindowEvent.WINDOW_ACTIVATED)
			{
				LyView view = getView();

				// Notify controller if any
				if (view != null && view.getController() != null)
				{
					// Notify controller
					LyDemand demand = new LyActivateViewDemand(view.getId());

					// Set info :
					demand.setParameterValues(new LySwingParameterValues(view));
					processDemand(demand);
				}
			}
		}
		
		public void run()
		{
		    close();
		    
            // Workaround to stop waiting if something occurred :
            // each attempt to close a window stop the waiting status
            LyView          view = getView();

            if (view != null)
            {
                LyViewManager   viewManager = getViewManager();
                
                viewManager.setWaiting(false);
            }
		}
	} // end class _LyWindowAdapter


	/**
	 * Internal class to wait for resize events.
	 */
	protected class _LyComponentAdapter extends ComponentAdapter
	{

		/**
		 * Empty class constructor of a resizing events class.
		 */
		public _LyComponentAdapter()
		{
		}

		/**
		 * Invoked when the window has been resized.
		 *
		 * @see java.awt.event.ComponentListener#componentResized(java.awt.event.ComponentEvent)
		 */
		public void componentResized(ComponentEvent e)
		{
			Component comp = e.getComponent();
            
            if (!comp.isShowing())
                return;
            
			Dimension size = comp.getSize();

			if (_requiredSize != null)
			{
				if ((_requiredSizeDate == null)
						|| (((new Date()).getTime() - _requiredSizeDate.getTime()) < 200))
					_requiredSize = size;

				if (!_requiredSize.equals(size))
				{
					comp.setSize(_requiredSize);
					return;
				}
			}

			LyView view = getView();
			int    w = size.width, h = size.height;
			int    minWidth = view.getMinWidth();
			int    minHeight = view.getMinHeight();

			if (getEnvironment().testResource("LY_CHECK_WINDOW_MINIMUM_SIZE"))
			{
				if (minWidth < 0)
					minWidth = getIntResource("LY_WINDOW_MIN_WIDTH");
				if (minHeight < 0)
					minHeight = getIntResource("LY_WINDOW_MIN_HEIGHT");
			}

			if (w < minWidth)
				w = minWidth;
			if (h < minHeight)
				h = minHeight;

			// Never larger than the screen size : FFT001554
			int[] dscreen = getEnvironment().getMaxScreenSize();

			// When asking for max size on Windows, component width is 8 pixels
			// greater than screen width
			if (LyEnvironment.isOperatingSystem(LyEnvironment.OS_WINDOWS))
				dscreen[0] += 8;

			if (comp.isVisible())
			{
				Point location = comp.getLocationOnScreen();
				comp.setLocation(location);
			}

			if (comp instanceof Frame)
			{
				Frame frame = (Frame)comp;

				if (frame.getExtendedState() == Frame.MAXIMIZED_BOTH)
					return;
			}
			if (LyEnvironment.isOperatingSystem(LyEnvironment.OS_LINUX))
				dscreen[0] += 10;

			if (w > ((dscreen[0] * getMaxSizeRatio()) / 100))
				w = (dscreen[0] * getMaxSizeRatio()) / 100;

			if (h > ((dscreen[1] * getMaxSizeRatio()) / 100))
				h = (dscreen[1] * getMaxSizeRatio()) / 100;

			if ((w != size.width) || (h != size.height))
				setSize(w, h);
		} // end method componentResized
	} // end class _LyComponentAdapter

	/**
	 * Maximum size ratio of the view on the screen, in %.
	 *
	 * @return 100 (100%)
	 */
	public int getMaxSizeRatio()
	{
		return 100;
	}

	/**
	 * Takes a snapshot of the current view.
	 *
	 * @param file the file that will contain the snapshot
	 */
	public abstract void snapshot(String file);
	
	/**
	 * This class free the current window in a thread, typically the AWT Event Queue
	 */
	protected class _LyFreeWindow
		implements Runnable
	{
		/** The Window component that is freed. */
		Window _window;
		
		public _LyFreeWindow(Window window)
		{
			_window = window;
		}
		
		public void run()
		{
			setVisible(false);

			if (_winAdapter != null)
				_window.removeWindowListener(_winAdapter);

			if (_compAdapter != null)
				_window.removeComponentListener(_compAdapter);

			try
			{
				// FFT1346
				_window.enableInputMethods(false);

				_window.dispose();
			}
			catch (Exception e)
			{
				getEnvironment().writeException(e);
			}
		}
	}

} // end class LySwingView
