/*
 * LySwingViewManager.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.datatransfer.*;
import java.io.*;
import java.lang.reflect.*;
import java.util.*;

import javax.swing.*;
import leon.misc.*;
import leon.view.*;
import leon.view.common.*;
import leon.view.event.*;
import leon.view.print.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.LyJFileDialog;


/**
 * This class serves as a layer between the communication controller and the view. It implements
 * remote update notifications.
 *
 * @version $Id: LySwingViewManager.java 41416 2012-10-31 09:34:36Z jlsaouli $
 * @author  Lyria
 */
public class LySwingViewManager extends LyViewManager
	implements LyActionListener,
		ClipboardOwner
{
	public static final String			 __VERSION =
		"$Id: LySwingViewManager.java 41416 2012-10-31 09:34:36Z jlsaouli $";

	/**
	 *  Defines the style of the Synth look and feel using an XML description.
	 * 	If you use this resource you must feel the LY_SYNTH_XML_CLASS_FILE resource
	 */
	public static String				SYNTH_XML_STYLE_FILE = null;

	/**
	 * Defines a class file used to load the Synth look and feel.
	 *	This class must be in the same package than the xml file.
	 */
	public static String				SYNTH_XML_CLASS_FILE = null;

	/** Defines the style of the Synth look and feel using a factory. */
	public static String				SYNTH_STYLE_FACTORY = null;
		
	/** Default swing frame. */
	protected static JFrame				_defaultJFrame;

	/** Waiting cursor. */
	public static Cursor				WAIT_CURSOR = new Cursor(Cursor.WAIT_CURSOR);

	/** Default cursor. */
	public static Cursor				DEFAULT_CURSOR = new Cursor(Cursor.DEFAULT_CURSOR);

	/**
	 * Start waiting delay. Miminum elapsed time before using waiting cursor + inactivate
	 * components.
	 */
	public static long					START_WAITING_DELAY = 500;

	/**
	 * Default Frame used in some parts of the code when a default AWT component is needed (for
	 * example to create an image).
	 */
	public static Frame			 		_frame;

	/** Splash screen. */
	protected static LySwingSplashScreen _splashScreen;

	/** Current Thread waiting for START_WAITING_DELAY timeout. */
	protected _LyWaitingThread			 _currentWaitingThread;

	/** A simple Lock for synchronized access on {@link #setWaiting(boolean)}. */
	protected Object					 _waitingThreadLock = new Object();

	/** Extern events queue. */
	protected Vector<_LyWinEvent>		 _eventQueue;

	/** Number of events in this view manager event queue. */
	protected int						 _nEvent;


	/**
	 * Returns the default swing frame.
	 *
	 * @return the default frame
	 */
	public static JFrame getDefaultJFrame()
	{
		if (_defaultJFrame == null)
			_defaultJFrame = new JFrame();

		return _defaultJFrame;
	}

	/**
	 * Disposes and removes the default swing frame.
	 */
	private static void removeDefaultJFrame()
	{
		if (_defaultJFrame != null)
		{
			_defaultJFrame.dispose();
			_defaultJFrame = null;
		}
	}

	/**
	 * Class constructor of a view manager. This view manager takes care of the swing look and feel
	 * to use.
	 *
	 * @param environment the current context environment of this application
	 */
	public LySwingViewManager(LyEnvironment environment)
	{
		super(environment);

		// Set anti-aliasing for text.
		// Do not move this call after the first call to any UIManager method or
		// it won't work any more.
		System.setProperty("swing.aatext", "true");

		_beans = new LyResources(environment, "views_swing");

		String lookAndFeelClassName = getEnvironment().getResource("LY_LOOK_AND_FEEL");

		if ((lookAndFeelClassName != null)
			&& !lookAndFeelClassName.equals(""))
		{
			String lookAndFeelWarningMessage = null;
			
			if (lookAndFeelClassName.equals("LY_CROSS_PLATFORM"))
				lookAndFeelClassName = UIManager.getCrossPlatformLookAndFeelClassName();
			else if (lookAndFeelClassName.equals("LY_SYSTEM"))
				lookAndFeelClassName = UIManager.getSystemLookAndFeelClassName();

			if (!lookAndFeelClassName.equals(UIManager.getLookAndFeel().getClass().getName()))
			{
				if (lookAndFeelClassName.equals("javax.swing.plaf.synth.SynthLookAndFeel"))
					lookAndFeelWarningMessage = installSynthLookAndFeel();
				else
				{
					try
					{
						UIManager.setLookAndFeel(lookAndFeelClassName);
						lookAndFeelWarningMessage = "LookAndFeel " + lookAndFeelClassName + " successfully installed";
					}
					catch (ClassNotFoundException e)
					{
						lookAndFeelWarningMessage = "Class " + lookAndFeelClassName + " not found ";
						lookAndFeelWarningMessage += ". Using default LookAndFeel " + UIManager.getLookAndFeel().getName();
					}
					catch (InstantiationException e)
					{
						lookAndFeelWarningMessage = "Class " + lookAndFeelClassName + " cannot be instantiated ";
						lookAndFeelWarningMessage += ". Using default LookAndFeel " + UIManager.getLookAndFeel().getName();
					}
					catch (IllegalAccessException e)
					{
						lookAndFeelWarningMessage = "Class " + lookAndFeelClassName + " cannot be accessed ";
						lookAndFeelWarningMessage += ". Using default LookAndFeel " + UIManager.getLookAndFeel().getName();
					}
					catch (UnsupportedLookAndFeelException e)
					{
						lookAndFeelWarningMessage = "LookAndFeel " + lookAndFeelClassName + " is not supported on this platform ";
						lookAndFeelWarningMessage += ". Using default LookAndFeel " + UIManager.getLookAndFeel().getName();
					}
				}
			}
			
			if (lookAndFeelWarningMessage != null)
			    getEnvironment().logWarning(lookAndFeelWarningMessage);
		}
		else
			getEnvironment().logInfo("Using default LookAndFeel " + UIManager.getLookAndFeel().getName());

		// Set tooltip properties
		LyColor color = environment.getLyColor("LY_TOOLTIP_COLOR");
		LyFont  font = environment.getLyFont("LY_TOOL_TIP_FONT");

		UIManager.put("ToolTip.background", getColor(color));
		UIManager.put("ToolTip.font", getFont(font));
		UIManager.put("ToolTip.foregroundInactive", Color.black);
		UIManager.put("ToolTip.borderInactive", BorderFactory.createLineBorder(Color.black));
		UIManager.put("ToolTip.border", BorderFactory.createLineBorder(Color.black));
		UIManager.put("ToolTip.foreground", Color.black);

		// Add foreground for disabled combo box
		UIManager.put("ComboBox.disabledForeground", Color.black);

		// Add foreground and gradient for buttons
		color = environment.getLyColor("LY_HIGHLIGHTED_BACKGROUND");

		if (color != null)
		{
			UIManager.put("Button.select", getColor(color));
			UIManager.put("Button.darkShadow",
				LySwingUtils.getDarkerColor((Color)getColor(color), 80));
			UIManager.put("Button.gradient",
				Arrays.asList(
					new Object[]
					{
						new Float(.3f),
						new Float(0f),
						LySwingUtils.getLighterColor((Color)getColor(color), 20),
						LySwingUtils.getLighterColor((Color)getColor(color), 50),
						getColor(color)
					}));
		}
	}

	/**
	 * Install the Synth Look And Feel, based on the corresponding resource.
	 *
	 * @return a message to display to the user if needed.
	 */
	private String installSynthLookAndFeel()
	{
		String	    lookAndFeelWarningMessage = null;
		Class<?>    synthLookAndFeelClass = null;
		ClassLoader cl = getEnvironment().getClassLoader();

		if (cl == null)
			cl = getClass().getClassLoader();
		try
		{
			synthLookAndFeelClass = cl.loadClass("javax.swing.plaf.synth.SynthLookAndFeel");
		}
		catch (ClassNotFoundException e)
		{
			lookAndFeelWarningMessage = "Class javax.swing.plaf.synth.SynthLookAndFeel not found ";
			lookAndFeelWarningMessage += ". Using default LookAndFeel "
				+ UIManager.getLookAndFeel().getName();
			return lookAndFeelWarningMessage;
		}
		
		String	xmlLookFile = SYNTH_XML_STYLE_FILE;
		String	xmlClassFile = SYNTH_XML_CLASS_FILE;
		String	synthStyleFactory = SYNTH_STYLE_FACTORY;

		if ((synthStyleFactory == null)
			&& (xmlLookFile == null))
		{
			lookAndFeelWarningMessage =
				"When using Synth look and feel, you must provide a XML style file (using resource LY_SYNTH_XML_STYLE_FILE) or a Synth file factory (using resource LY_SYNTH_STYLE_FACTORY). Using default look and feel";
			return lookAndFeelWarningMessage;
		}

		if ((xmlLookFile != null) && (xmlClassFile == null))
		{
			lookAndFeelWarningMessage =
				"When using a Synth look and feel with an XML style file, you must provide a LY_SYNTH_XML_CLASS_FILE to allow image handling. Using default look and feel";
			return lookAndFeelWarningMessage;
		}

		Object synthLookAndFeel;

		try
		{
			synthLookAndFeel = synthLookAndFeelClass.newInstance();
			Class<?> xmlClass = cl.loadClass(xmlClassFile);

			if (synthStyleFactory != null)
			{
				Class<?>[] parameters = new Class[1];
				parameters[0] = cl.loadClass("javax.swing.plaf.synth.SynthStyleFactory");
				Method   setFactoryMethod = synthLookAndFeelClass.getDeclaredMethod(
						"setStyleFactory", parameters);

				Class<?>    userFactory = cl.loadClass(synthStyleFactory);
				Object[] values = new Object[2];

				values[0] = userFactory.newInstance();

				setFactoryMethod.invoke(synthLookAndFeel, values);
			}
			else
			{
				Class<?>[] parameters = new Class[2];
				parameters[0] = InputStream.class;
				parameters[1] = Class.class;
				Method   loadMethod = synthLookAndFeelClass.getDeclaredMethod("load",
						parameters);

				Object[] values = new Object[2];

				values[0] = getEnvironment().getResourceAsStream(xmlLookFile,
						getEnvironment().getClassLoader());
				values[1] = xmlClass;
				loadMethod.invoke(synthLookAndFeel, values);
			}
			UIManager.setLookAndFeel((LookAndFeel)synthLookAndFeel);
			lookAndFeelWarningMessage =
				"LookAndFeel javax.swing.plaf.synth.SynthLookAndFeel successfully installed";
		}
		catch (InstantiationException e)
		{
			lookAndFeelWarningMessage = "Cannot instantiate Synth Look and Feel";
			getEnvironment().writeException(e);
		}
		catch (IllegalAccessException e)
		{
			lookAndFeelWarningMessage = "Cannot instantiate Synth Look and Feel";
			getEnvironment().writeException(e);
		}
		catch (ClassNotFoundException e)
		{
			lookAndFeelWarningMessage = "Cannot find class " + xmlClassFile
				+ ". This file is needed in order to be able to load image resources.";

			getEnvironment().writeException(e);
		}
		catch (SecurityException e)
		{
			lookAndFeelWarningMessage = "Cannot instantiate Synth Look and Feel";
			getEnvironment().writeException(e);
		}
		catch (NoSuchMethodException e)
		{
			lookAndFeelWarningMessage = "Cannot instantiate Synth Look and Feel";
			getEnvironment().writeException(e);
		}
		catch (IllegalArgumentException e)
		{
			lookAndFeelWarningMessage = "Cannot instantiate Synth Look and Feel";
			getEnvironment().writeException(e);
		}
		catch (InvocationTargetException e)
		{
			lookAndFeelWarningMessage = "Cannot instantiate Synth Look and Feel";
			getEnvironment().writeException(e);
		}
		catch (UnsupportedLookAndFeelException e)
		{
			lookAndFeelWarningMessage = "Cannot instantiate Synth Look and Feel";
			getEnvironment().writeException(e);
		}
		return lookAndFeelWarningMessage;
	} // end method installSynthLookAndFeel

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.LyViewManager#buildInternalView(leon.view.LyComponent)
	 */
	public LyViewComponent buildInternalView(LyComponent component)
	{
		// Check if field item and if a bean must be used
		if (component instanceof LyFieldItem)
		{
			LyFieldItem item = (LyFieldItem)component;

			if (item.getBean() != null)
			{
				LyViewComponent bean = item.buildBean();

				if (bean instanceof LySwingComponent)
					return bean;
				else if (bean != null)
					_environment.logWarning("Can't use bean " + bean.getClass().getName()
						+ " because it's not a LySwingComponent");
			}
		}

		if (_beans != null)
		{
			String tag = LyViewNodeFactory.getTag(component.getClass().getName());
			String beanClassName = _beans.getString(tag);

			if (beanClassName != null)
			{
				if (beanClassName.length() == 0)
					return null;
				return buildInternalView(beanClassName, component);
			}
		}

		return buildInternalView("leon.view.win.swing", "LySwing", component);
	} // end method buildInternalView

	/**
	 * Returns the full name of the class that handles messages.
	 *
	 * @return the String <code>"leon.view.win.swing.LySwingMessageWindow"</code> which handle the
	 *         messages
	 */
	public String getMessageWindowClassName()
	{
		return "leon.view.win.swing.LySwingMessageWindow";
	}

	/**
	 * Recursive method to find the parent frame of a dialog.
	 *
	 * @param  parent the current parent may be <code>null</code> for default frame
	 * @return a root frame. This method never returns <code>null</code>, but may use the default
	 *         Frame.
	 */
	public JFrame seekParentJFrame(LyComponent parent)
	{
		if (parent == null)
			return LySwingViewManager.getDefaultJFrame();

		return (JFrame)seekParentFrame(parent);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.LyViewManager#exit(int)
	 */
	public void exit(int code)
	{
		LySwingViewManager.removeDefaultJFrame();

		if (code == 0)
		{
			// Normal exit, do some cleaning and awt queue will end by itself.
			LySwingViewManager.removeDefaultFrame();
			LyImageUtils.clean();

			Frame[] frames = Frame.getFrames();

			if (frames != null)
			{
				for (int i = 0; i < frames.length; i++)
					frames[i].dispose();
			}
		}
		else
			System.exit(code);
	}

	/**
	 * Returns the default AWT frame.
	 *
	 * @return the default frame
	 */
	public static Frame getDefaultFrame()
	{
		if (_frame == null)
		{
			_frame = new Frame();
			// ensure that the peer is ready.
			// a peer is required for some methods such as createImage()
			// see also : LyEnvironment.getImage()
			if (LyEnvironment.isOperatingSystem(LyEnvironment.OS_WINDOWS))
				_frame.addNotify();
		}

		return _frame;
	}

	/**
	 * Disposes and removes the default AWT frame.
	 */
	private static void removeDefaultFrame()
	{
		if (_frame != null)
		{
			_frame.dispose();
			_frame = null;
		}
	}

	/**
	 * Returns the top-level frame for the given AWT component.
	 *
	 * @param  component the AWT component
	 * @return the top-level frame of the given AWT component or the default frame if the component
	 *         is not in a frame
	 */
	public static Frame getRootFrame(Component component)
	{
		if (component == null)
			return getDefaultFrame();

		if (!(component instanceof Frame))
			return getRootFrame(component.getParent());
		return (Frame)component;
	}

	/**
	 * Returns the top-level window for the given AWT component.
	 *
	 * @param  component the AWT component
	 * @return the top-level window of the given AWT component or <code>null</code> if the component
	 *         is <code>null</code>
	 */
	public static Window getRootWindow(Component component)
	{
		if (component == null)
			return null;

		if (!(component instanceof Window))
			return getRootWindow(component.getParent());
		return (Window)component;
	}

	/**
	 * Sets properties common to all window view managers.
	 */
	protected void setProperties()
	{
		super.setProperties();

		setProperty("MENU");
		setProperty("NOTIFY_SELECT");
		setProperty("DOUBLE_CLICK");
		setProperty("LOCAL_FILE");
		setProperty("MULTI_FRAME_FORM");
		setProperty("MULTI_WINDOW");
		setProperty("UNDO_REDO");
		setProperty("LOCAL_MAP_SELECTION");
		setProperty("ADVANCED_GANTT_FEATURES");
		setProperty("NOTFORMOBILE");
		setProperty("MANAGE_IMAGE_FILE_SELECTOR");
	}

	/**
	 * Prepares the GUI to start by setting its splash screen.
	 *
	 * @see leon.misc.LyViewManagerInterface#prepareGui()
	 */
	public void prepareGui()
	{
		_splashScreen = new LySwingSplashScreen(_environment);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.misc.LyViewManagerInterface#startGui(leon.misc.LyGuiClientInterface)
	 */
	public void startGui(LyGuiClientInterface client)
	{
		super.startGui(client);

		if (_splashScreen != null)
		{
			_splashScreen.free();
			_splashScreen = null;
		}
	}

	/**
	 * Nothing is done at this level. There is no temporary resources for this swing view manager.
	 *
	 * @see LyViewManager#clearTemporaryResources()
	 */
	public void clearTemporaryResources()
	{
	}

	/**
	 * Returns the same directory as {@link #getTemporaryDir()}, <code>null</code> in this
	 * implementation.
	 *
	 * @return <code>null</code>
	 * @see    LyViewManager#getTemporaryLocation()
	 */
	public String getTemporaryLocation()
	{
		return getTemporaryDir();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.LyViewManager#setWaiting(boolean)
	 */
	public void setWaiting(boolean status)
	{
        if (status)
        {
            _waitingCounter++;
            if (_waitingCounter > 1)
                return; // Already waiting
        }
        else
        {
            if (_waitingCounter == 0)
                return;

            _waitingCounter--;
            if (_waitingCounter > 0)
                return; // Still waiting
        }

		// Do not use Multi Thread
		if (START_WAITING_DELAY == 0)
		{
			completeSetWaiting(status);
			return;
		}

		if (START_WAITING_DELAY < 0)
			return;

		// ** Check if we start or stop waiting **
		synchronized (_waitingThreadLock)
		{
			if (status)
			{
				// New Thread
				if (_currentWaitingThread == null)
				{
					_currentWaitingThread = new _LyWaitingThread(this);
					_currentWaitingThread.start();
				}
			}
			else
			{
				// Stop it if possible
				if (_currentWaitingThread != null)
				{
					_currentWaitingThread.endWait(); // Don't interrupt or stop Thread it may be
					// active.
					_currentWaitingThread = null;
				}
			}
		}
	} // end method setWaiting

	/**
	 * Executes effective setWaiting impact.
	 *
	 * @param status the waiting status
	 */
	protected void completeSetWaiting(boolean status)
	{
		if (_viewList == null)
			return;

		synchronized (_viewListLock) // <-- Do NOT remove this lock.
		// Concurrent access AWT + waitingThread
		{
			int    i, n = getViewCount();
			LyView view;

			for (i = 0; i < n; i++)
			{
				view = getView(i);
				view.setWaiting(status);
			}
		}
	}

	/**
	 * Copy a given text to the system clipboard.
	 *
	 * @param text the text to be copied to the clipboard.
	 */
	public void copyToClipboard(String text)
	{
		StringSelection stringSelection = new StringSelection(text);
		Clipboard	    clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();

		clipboard.setContents(stringSelection, this);
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.datatransfer.ClipboardOwner#lostOwnership(java.awt.datatransfer.Clipboard, java.awt.datatransfer.Transferable)
	 */
	public void lostOwnership(Clipboard clipboard, Transferable contents)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.LyViewManager#showWarning(leon.view.LyComponent, java.lang.String)
	 */
	public void showWarning(LyComponent source, String message)
	{
		showMessage(source, "LY_WARNING_TITLE", "LY_WARNING_ICON", message, null /* OK */, true,
			null);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.LyViewManager#showError(leon.view.LyComponent, java.lang.String)
	 */
	public void showError(LyComponent source, String message)
	{
		showMessage(source, "LY_ERROR_TITLE", "LY_ERROR_ICON", message, null /* OK */, true, null);
	}

	/**
	 * Asks a question to the user. The method is synchronous and will block untill validation is
	 * done. A timeout may also be implemented by subclasses.
	 *
	 * @param  source  view element which asks the question
	 * @param  message text of the question
	 * @return the user answer
	 * @see    LyViewManager#askQuestion
	 */
	public boolean askQuestion(LyComponent source, String message)
	{
		BitSet bs = new BitSet();

		bs.set(LyViewManager.COMMAND_YES);
		bs.set(LyViewManager.COMMAND_NO);

		return (showMessage(source, "LY_QUESTION_TITLE", "LY_QUESTION_ICON", message, bs, true,
				null) == LyViewManager.COMMAND_YES);
	}

	/**
	 * Asks a question to the user. The method is asynchronous and the result is returned to the
	 * client in the client attribute.
	 *
	 * @param source     view element which asks the question
	 * @param message    text of the question
	 * @param client     registered to receive the response
	 * @param clientData freely set data which will be returned with the response
	 */
	public void askQuestion(LyComponent source, String message, LyQuestionAnswerInterface client,
		Object clientData)
	{
		BitSet bs = new BitSet();

		bs.set(LyViewManager.COMMAND_YES);
		bs.set(LyViewManager.COMMAND_NO);

		_LyMessageAnswerInfo info = new _LyMessageAnswerInfo(client, clientData);

		// FFT00435/Flandres
		if (source == null)
			source = getView(0);

		showMessage(source, "LY_QUESTION_TITLE", "LY_QUESTION_ICON", message, bs, true, info);
	}

	/**
	 * Shows a message to the user. This method is asynchronous and the result is returned in the
	 * client attribute.
	 *
	 * @param source     the source component
	 * @param title      the title of the message
	 * @param icon       the icon of the message
	 * @param message    the message
	 * @param commands   the commands associated with the message
	 * @param client     registred to receive the response
	 * @param clientData freely set data which will be returned with the response
	 * @see   leon.view.LyViewManager#showMessage(leon.view.LyComponent, java.lang.String,
	 *        java.lang.String, java.lang.String, java.util.BitSet,
	 *        leon.view.setinterface.LyMessageAnswerInterface, java.lang.Object)
	 */
	public void showMessage(LyComponent source, String title, String icon, String message,
		BitSet commands, LyMessageAnswerInterface client, Object clientData)
	{
		_LyMessageAnswerInfo info = new _LyMessageAnswerInfo(client, clientData);

		showMessage(source, title, icon, message, commands, true, info);
	}

	/**
	 * Shows a message to the user. This method is synchronous and will block until validation is
	 * done.
	 *
	 * @param  source   the source component
	 * @param  title    the title of the message
	 * @param  icon     the icon of the message
	 * @param  message  the message
	 * @param  commands the commands associated with the message
	 * @return the user answer
	 * @see    LyViewManager#showMessage
	 */
	public int showMessage(LyComponent source, String title, String icon, String message,
		BitSet commands)
	{
		return (showMessage(source, title, icon, message, commands, true, null));
	}

	/**
	 * Shows a message window to the user.
	 *
	 * @param  source   the source component
	 * @param  title    the title of the message
	 * @param  icon     the icon of the message
	 * @param  message  the message
	 * @param  commands the commands associated with the message
	 * @param  modal    modality of the window message
	 * @param  info     freely set data which will be returned with the response
	 * @return the user answer
	 * @see    LyViewManager#showMessage
	 */
	public int showMessage(LyComponent source, String title, String icon, String message,
		BitSet commands, boolean modal, _LyMessageAnswerInfo info)
	{
		if (modal)
			modal = !hasProperty("NO_MODAL");

		String bgColor = null;
		String type = null;

		if ((icon == null) || (icon.equals("LY_WARNING_ICON")))
		{
			bgColor = "LY_WARNING_BACKGROUND";
			type = "INFO";
		}
		else if (icon.equals("LY_ERROR_ICON"))
		{
			bgColor = "LY_ERROR_BACKGROUND";
			type = "ERROR";
		}
		else if (icon.equals("LY_QUESTION_ICON"))
		{
			bgColor = "LY_QUESTION_BACKGROUND";
			type = "QUESTION";
		}
		else
		{
			if ((commands != null)
					&& ((commands.size() > 2)
						|| (commands.get(COMMAND_HELP) && (commands.size() > 1))))
				type = "QUESTION";
			else
				type = "INFO";
		}

		String messageWindowClassName = _beans.getString("MESSAGE_WINDOW");

		if ((messageWindowClassName == null) || (messageWindowClassName.length() == 0))
			messageWindowClassName = getMessageWindowClassName();

		Class<?>[]  types = new Class[2];
		Object[] parameters = new Object[2];

		types[0] = LyComponent.class;
		parameters[0] = source;

		types[1] = LyViewManager.class;
		parameters[1] = this;

		LyMessageWindowInterface messageWindow = null;

		try
		{
			Class<?>    	viewClass = Class.forName(messageWindowClassName);
			Constructor<?> 	constructor = viewClass.getConstructor(types);

			messageWindow = (LyMessageWindowInterface)constructor.newInstance(parameters);
		}
		catch (InvocationTargetException e1)
		{
			_environment.writeException(e1.getTargetException());
			_environment.writeException(e1);
			throw new IllegalArgumentException("BF AppEngine : can't build message window class : "
				+ messageWindowClassName);
		}
		catch (Exception e)
		{
			_environment.writeException(e);
			throw new IllegalArgumentException("BF AppEngine : can't build message window class : "
				+ messageWindowClassName);
		}

		// LySwingMessageWindow messageWindow = new LySwingMessageWindow(source, this);
		messageWindow.setBackgroundColor(bgColor);
		messageWindow.setIcon(icon);
		messageWindow.setTitle(title);
		messageWindow.setMessage(message);
		messageWindow.setCommands(commands);
		messageWindow.setClientData(info);
		messageWindow.setModal(modal);
		messageWindow.addActionListener(this);
		messageWindow.initialize();

		LyDemandTransmitter transmitter = getDemandTransmitter();

		if (transmitter instanceof LyMessageWindowRegister)
			((LyMessageWindowRegister)transmitter).registerMessageWindow(messageWindow, type,
				message);

		if ((icon == null) || icon.equals("LY_WARNING_ICON"))
			messageWindow.setVisible(true, false);
		else
			messageWindow.setVisible(true, true);

		if (modal)
			return messageWindow.getCommandResult();

		return -1; // Asynchronous
	} // end method showMessage

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		// Put environment as current one
		LyEnvironment oldEnvironment = LyEnvironment.setCurrentEnvironment(_environment);

		try
		{
			LyMessageWindowInterface messageWindow = (LyMessageWindowInterface)actionEvent.getSource();
			_LyMessageAnswerInfo     info = (_LyMessageAnswerInfo)messageWindow.getClientData();
			int						 result = COMMAND_CLOSED;

			if (info != null)
				result = messageWindow.getCommandResult();

			messageActionPerformed(messageWindow, result);
		}
		finally
		{
			LyEnvironment.setCurrentEnvironment(oldEnvironment);
		}
	}

	/**
	 * Shows a location to the user in a new window.
	 *
	 * @param source   view element which raises the location
	 * @param location location
	 * @see   #showLocation(LyComponent, String, boolean)
	 * @see   LyViewManager#showFile(LyComponent, String)
	 */
	public void showFile(LyComponent source, String location)
	{
		showLocation(source, location, true);
	}

	/**
	 * Shows a location to the user.
	 *
	 * @param source    view element which raises the location
	 * @param location  location
	 * @param newWindow This method shows modal windows that are always new. This attribute is not
	 *                  used.
	 * @see   LyViewManager#showLocation(LyComponent, String, boolean)
	 */
	public void showLocation(LyComponent source, String location, boolean newWindow)
	{
		showLocation(source, location, newWindow, null);
	}

	/**
	 * Shows a location to the user.
	 *
	 * @param source    view element which raises the location
	 * @param location  location
	 * @param newWindow This method shows modal windows that are always new. This attribute is not
	 *                  used.
	 * @param parameters list of parameters
	 * @see   LyViewManager#showLocation(LyComponent, String, boolean)
	 */
	public void showLocation(LyComponent source, String location, boolean newWindow, Hashtable<Short, String> parameters)
	{
		// case a file being consulted with an hyperlink.
		String FILE_REF = "file://";
		String HTTP_REF = "http://";
	    String HTTPS_REF = "https://";
		String externalViewer = _environment.getFile("LY_EXTERNAL_VIEWER");

		if (location.startsWith(FILE_REF))
			location = location.substring(FILE_REF.length());
		location = _environment.getString().expand(location);

		if ((externalViewer == null)
				|| ("LY_EXTERNAL_VIEWER".equals(externalViewer)))
		{
			String message = _environment.getMessage("LY_NO_EXTERNAL_VIEWER", location);

			showMessage(source, "LY_WARNING_TITLE", "LY_WARNING_ICON", message, null, true, null);
		}
		else
		{
			if (location.startsWith(HTTP_REF)
			    || location.startsWith(HTTPS_REF))
			{
				String command [] = new String [2];
				command [0] = externalViewer;
				command [1] = location;
				_environment.executeCommand(command, false);
			}
			else
			{
				String script = externalViewer + " \"" + location + "\"";
				_environment.executeCommand(script, false);
			}
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.LyViewManager#showPDFPreview(leon.view.LyComponent, java.lang.String)
	 */
	public void showPDFPreview(LyComponent source, String location)
	{
		// The file should be displayed with its path on the server
		String pdfViewer = _environment.getFile("LY_PDF_VIEWER");
		String script = "";
		LyProcessViewer process;

		// The LY_PDF_VIEWER element is not set in the ini File
		if ((pdfViewer == null) || (pdfViewer.trim().length() == 0)
				|| pdfViewer.equals("LY_PDF_VIEWER"))
		{
			// We try to launch AcroRd32.exe this command can work if AcroRd32.exe is set in the
			// windows PATH environement variable.
			script = "AcroRd32" + " \"" + location + "\"";
			process = new LyProcessViewer(_environment, script, false);
		}
		else
		{
			script = pdfViewer + " \"" + location + "\"";
			process = new LyProcessViewer(_environment, script, false);
		}


		if(process.getReturnCode() < 0)
		{
			script = "cmd.exe /c \""+location +"\"";
			process = new LyProcessViewer(_environment, new String[] { "cmd.exe", "/c", "\""+location+"\"" }, false);
		}

		// The command AcroRd32 doesn't work.
		if(process.getReturnCode() < 0)
		{
			LyJFileDialog dialog = new LyJFileDialog(null, _environment.translate("LY_SELECT_PDF_VIEWER"), LyFile.Type.TEXT, "", null);

			int choice = dialog.show();

			if(choice == JFileChooser.CANCEL_OPTION || choice == JFileChooser.ERROR_OPTION)
				return;
			else if(choice == JFileChooser.APPROVE_OPTION)
			{
				script = dialog.getFile() + " \"" +  location + "\"";
				process = new LyProcessViewer(_environment, script, false);

				if(process.getReturnCode() <= 0)
					_environment.raiseError("LY_ERROR_CANNOT_SHOW_PDF");
			}
		}
	}

	/**
	 * Finds recursively a parent frame for the specified dialog.
	 *
	 * @param  parent the current parent, may be <code>null</code> for default frame
	 * @return a root frame, never return <code>null</code>, but may use a default Frame
	 */
	public Frame seekParentFrame(LyComponent parent)
	{
		if (parent == null)
			return getDefaultFrame();

		Object internalObj = parent.getInternalView(false);

		if ((internalObj == null) && (parent.getParent() == null))
			// Top level : force creation of internal view
			internalObj = parent.getInternalView();

		if (internalObj instanceof LySwingComponent)
		{
			LySwingComponent internal = (LySwingComponent)internalObj;

			// At the top, we have either a Frame or dialog ...
			if (internal.getComponent() instanceof Frame)
				return ((Frame)internal.getComponent());

			if (internal.getComponent() instanceof Dialog)
			{
				// And the parent of dialog is a Frame
				Component parentWin = ((Dialog)internal.getComponent()).getParent();

				if (parentWin instanceof Frame)
					return (Frame)parentWin;
			}
		}

		return (seekParentFrame(parent.getParent()));
	} // end method seekParentFrame

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.misc.LyViewManagerInterface#getScreenSize(java.lang.Object)
	 */
	public int[] getScreenSize(Object component)
	{
		// FFT00435/Flandres
		Rectangle screenSize = getScreenBounds(component);

		return new int[] { screenSize.width, screenSize.height, screenSize.x, screenSize.y };
	}

	/**
	 * Retrieves the viewport size available on the computer.
	 * @return The addition of all screen sizes.
	 */
	public int[] getMaxScreenSize()
	{
		GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
		GraphicsDevice[] devices = ge.getScreenDevices();
		Rectangle viewPort = new Rectangle();

		for (int i = 0; i < devices.length; i++)
		{
			Rectangle deviceBounds = devices[i].getDefaultConfiguration().getBounds();
			viewPort = viewPort.createUnion(deviceBounds).getBounds();
		}

		return new int[] { viewPort.width, viewPort.height, viewPort.x, viewPort.y };
	}
	
	/**
	 * Retrieves the maximum size of a window on the computer.
	 */
	public int[] getMaxWindowSize()
	{
		Rectangle	maxBounds = GraphicsEnvironment.getLocalGraphicsEnvironment().getMaximumWindowBounds();

		if (maxBounds != null)
			return new int[] {maxBounds.width, maxBounds.height, maxBounds.x, maxBounds.y};
		
		return super.getMaxWindowSize();
	}

	/**
	 * Gets the screen boundaries for the given component.
	 *
	 * @param  component a graphic component defining a monitor screen
	 * @return a Rectangle containing the size and location of the screen associated with the given
	 *         component. If the component does not belong to a graphics device, or if it does not
	 *         represents a graphical object, and no default device where to be found, then the
	 *         screen size of the default toolkit is returned, or null if no toolkit is available.
	 */
	public Rectangle getScreenBounds(Object component)
	{
		// FFT00435/Flandres
		GraphicsDevice device = getGraphicsDevice(component);

		if (device == null)
		{
			Toolkit toolkit = Toolkit.getDefaultToolkit();

			if (toolkit != null)
			{
				Dimension screenDim = toolkit.getScreenSize();

				return new Rectangle(screenDim);
			}

			return null;
		}

		return device.getDefaultConfiguration().getBounds();
	}

	/**
	 * Gets the graphics device object of the given component.
	 *
	 * @param  component a graphic component defining a monitor screen. The monitor screen is the
	 *                   one that holds the center of the boundaries of the component.
	 * @return the graphic device corresponding to the given graphic component. If the component
	 *         does not represents a graphical object, then the default screen device of the
	 *         computer is used.
	 */
	protected GraphicsDevice getGraphicsDevice(Object component)
	{
		// FFT00435/Flandres
		GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();

		if (component instanceof LyView)
		{
			LyViewComponent viewComponent = ((LyView)component).getInternalView();

			if (viewComponent instanceof LySwingObject)
				component = ((LySwingObject)viewComponent).getComponent();
		}

		if (component instanceof Component)
		{
			Component cmp = ((Component)component);

			if (ge == null)
				return cmp.getGraphicsConfiguration().getDevice();

			Rectangle		 componentBounds = cmp.getBounds();
			GraphicsDevice[] devices = ge.getScreenDevices();
			GraphicsDevice   device = null;

			Rectangle		 deviceBounds = null;
			Rectangle		 intersection;
			int				 max = -1, area;
			GraphicsDevice   max_device = null;

			for (int i = 0; i < devices.length; i++)
			{
				device = devices[i];
				deviceBounds = device.getDefaultConfiguration().getBounds();
				intersection = deviceBounds.intersection(componentBounds);
				area = intersection.width * intersection.height;
				if (area > max)
				{
					max = area;
					max_device = device;
				}
			}
			if (max_device != null)
				return max_device;
		}

		if (ge != null)
			return ge.getDefaultScreenDevice();
		return null;
	} // end method getGraphicsDevice

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.LyViewManager#getPrintManager(leon.view.print.LyPrintDataInterface,
	 *      java.util.Properties)
	 */
	public LyPrintManager getPrintManager(LyPrintDataInterface printDataInterface,
		Properties properties)
	{
		LySwingPrintManager printManager = new LySwingPrintManager(_environment, printDataInterface,
				properties);

		return (printManager.wasCancelled() ? null : printManager);
	}

	/**
	 * Builds a printJob instance.
	 *
	 * @param  view       the view for the print job
	 * @param  jobtitle   the title of the print job
	 * @param  properties the properties of the printing
	 * @return the print job for the view with the given properties
	 */
	public PrintJob getPrintJob(LyView view, String jobtitle, Properties properties)
	{
		Toolkit toolkit;
		Frame   frame;

		if (view == null)
		{
			toolkit = Toolkit.getDefaultToolkit();
			frame = getDefaultFrame();
		}
		else
		{
			// XXX : Extract all from LyView when possible
			toolkit = Toolkit.getDefaultToolkit();
			frame = getDefaultFrame();
		}

		return toolkit.getPrintJob(frame, jobtitle, properties);
	}

	/**
	 * Builds a printJob instance.
	 *
	 * @param  view           the view for the print job
	 * @param  jobtitle       the title of the print job
	 * @param  pageAttributes the page attributes of the printing
	 * @return the print job for the view with the given properties
	 */
	public PrintJob getPrintJob(LyView view, String jobtitle, PageAttributes pageAttributes)
	{
		Toolkit toolkit;
		Frame   frame;

		if (view == null)
		{
			toolkit = Toolkit.getDefaultToolkit();
			frame = getDefaultFrame();
		}
		else
		{
			// XXX : Extract all from LyView when possible
			toolkit = Toolkit.getDefaultToolkit();
			frame = getDefaultFrame();
		}

		return toolkit.getPrintJob(frame, jobtitle, null, pageAttributes);
	}

	/**
	 * Indicates wether this view manager handles multiple user sessions or not. Current
	 * implementation returns <code>false</code>.
	 *
	 * @return <code>false</code>
	 * @see    leon.misc.LyViewManagerInterface#isMultiSession()
	 */
	public boolean isMultiSession()
	{
		return false;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.misc.LyViewManagerInterface#beep()
	 */
	public void beep()
	{
		Toolkit toolkit = Toolkit.getDefaultToolkit();

		toolkit.beep();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.misc.LyViewManagerInterface#getColor(leon.misc.LyColor)
	 */
	public Object getColor(LyColor color)
	{
		if (color == null)
			return null;

		return new Color(color.getRGB());
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.misc.LyViewManagerInterface#getFont(leon.misc.LyFont)
	 */
	public Object getFont(LyFont font)
	{
		if (font == null)
			return getDefaultFrame().getFont();

		int style = 0;

		if (font.isBold())
			style |= Font.BOLD;
		if (font.isItalic())
			style |= Font.ITALIC;

		return new Font(font.getFontName(), style, font.getSize());
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.misc.LyViewManagerInterface#getStringWidth(leon.misc.LyFont, java.lang.String)
	 */
	public int getStringWidth(LyFont font, String text)
	{
		Font	    f = (Font)getFont(font);
		FontMetrics fm = getDefaultFrame().getFontMetrics(f);

		return fm.stringWidth(text);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.misc.LyViewManagerInterface#getDefaultImage()
	 */
	public Object getDefaultImage()
	{
		return LyImageUtils.getDefaultImage();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.misc.LyViewManagerInterface#getSizes(java.lang.Object)
	 */
	public int[] getSizes(Object image)
	{
		return LyImageUtils.getSizes((Image)image);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.misc.LyViewManagerInterface#getScaledInstance(java.lang.Object, int, int)
	 */
	public Object getScaledInstance(Object image, int width, int height)
	{
		return LyImageUtils.getScaledInstance((Image)image, width, height);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.misc.LyViewManagerInterface#waitForLoading(java.lang.Object, java.lang.Object)
	 */
	public void waitForLoading(Object component, Object image)
		throws InterruptedException
	{
		LyImageUtils.waitForLoading((Component)component, (Image)image);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.misc.LyViewManagerInterface#loadImageFromStream(java.lang.Object,
	 *      java.io.InputStream)
	 */
	public Object loadImageFromStream(Object component, InputStream resource)
		throws IOException
	{
		return LyImageUtils.loadImageFromStream((Component)component, resource);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.LyViewManager#isSameImage(java.lang.Object, java.lang.Object)
	 */
	public boolean isSameImage(Object image1, Object image2)
		throws InterruptedException
	{
		return LyImageUtils.isSameImage((Image)image1, (Image)image2);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.misc.LyViewManagerInterface#freeImage(java.lang.Object)
	 */
	public void freeImage(Object image)
	{
		LyImageUtils.freeImage((Image)image);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.misc.LyViewManagerInterface#execGui(java.lang.Runnable, boolean)
	 */
	public void execGui(Runnable runnable, boolean synchronous)
	{
		if (synchronous)
			runnable.run();
		else
			new _LyWinEvent(runnable);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.LyViewManager#encodeImage(java.lang.Object, java.lang.String,
	 *      java.lang.String)
	 */
	public boolean encodeImage(Object image, String imageDir, String imageFile)
	{
		return LyImageUtils.encodeImage((Image)image, imageDir, imageFile);
	}

	/**
	 * Takes a snapshot of the specified view and stores it in the given file. The snapshot must be
	 * done in a new thread for the popup menu to disappear.
	 *
	 * @param view the view to take a snapshot
	 * @param file the file to save the snapshot
	 * @see   leon.view.LyViewManager#takeSnapshot(leon.view.LyView, java.lang.String)
	 */
	public void takeSnapshot(LyView view, String file)
	{
		// Take the snapshot into a display Image
		// we must do it in a new thread to let the popup menu disappear
		Thread thread = new TakePicture(view, file);

		thread.run();

	}

	/**
	 * Takes a snapshot of the specified area and stores it in the given file. The snapshot must be
	 * done in a new thread for the popup menu to disappear.
	 *
	 * @param area the area to take a snapshot
	 * @param file the file to save the snapshot
	 * @see   leon.view.LyViewManager#takeSnapshot(leon.view.LyArea, java.lang.String)
	 */
	public void takeSnapshot(LyArea area, String file)
	{
		// Take the snapshot into a display Image
		// we must do it in a new thread to let the popup menu disappear
		Thread thread = new TakePicture(area, file);

		thread.run();
	}

	/**
	 * This class handles the window events in a separated thread.
	 */
	class _LyWinEvent extends AWTEvent
		implements ActiveEvent
	{
		private static final long serialVersionUID = -4257652045285678950L;

		/** The object the window events thread runs. */
		Runnable				  _runnable;

		/**
		 * Initializes a new Window event class handler and configures its queue of events.
		 *
		 * @param runnable the {@link Runnable} object that is called by the associated thread
		 */
		_LyWinEvent(Runnable runnable)
		{
			super(runnable, AWTEvent.RESERVED_ID_MAX + 1);

			_runnable = runnable;
			_nEvent++;

			if (_eventQueue == null)
				_eventQueue = new Vector<_LyWinEvent>();

			synchronized (_eventQueue)
			{
				_eventQueue.add(this);

				if (_eventQueue.size() == 1)
					// Rearm events post
					post();
			}
		}

		/**
		 * Processes the event in the event queue if any.
		 */
		protected void post()
		{
			if (_eventQueue.size() == 0)
				return;

			_LyWinEvent   event = _eventQueue.elementAt(0);
			Toolkit		  toolkit = Toolkit.getDefaultToolkit();
			EventQueue    eventQueue = toolkit.getSystemEventQueue();
			LyEnvironment environment = getEnvironment();

			if (environment != null)
				environment.logTrace("Event " + _nEvent + " reposted in AWT-Event-Queue for "
					+ _runnable + " from thread " + Thread.currentThread() + ", "
					+ (_eventQueue.size() - 1) + " remaining events");
			eventQueue.postEvent(event);
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.ActiveEvent#dispatch()
		 */
		public void dispatch()
		{
           try
            {
                _runnable.run();
            }
            catch (Exception e)
            {
                e.printStackTrace();
            }
            finally
            {
                synchronized(_eventQueue)
                {
                    _eventQueue.remove(this);
                    if (_eventQueue.size() > 0)
                        post();
                }
            }
		}
	} // end class _LyWinEvent


	/**
	 * Handles the snapshots of a specified view.
	 */
	class TakePicture extends Thread
	{

		/** The view to take a snapshot. */
		protected LyView _view;

		/** The area to take a snapshot. */
		protected LyArea _area;

		/** The file the snapshot is taken. */
		protected String _file;

		/**
		 * Takes a snapshot of the specified view in the given file.
		 *
		 * @param view the view to capture
		 * @param file the output file
		 */
		public TakePicture(LyView view, String file)
		{
			_view = view;
			_file = file;
		}


		/**
		 * Takes a snapshot of the specified area in the given file.
		 *
		 * @param view the area to capture
		 * @param file the output file
		 */
		public TakePicture(LyArea area, String file)
		{
			_area = area;
			_file = file;
		}

		/**
		 * Takes snapshots of the view at constant time intervals. The view is not rebuilt when the
		 * snapshot is taken.
		 *
		 * @see java.lang.Runnable#run()
		 */
		public void run()
		{
			try
			{
				sleep(100);
			}
			catch (InterruptedException e)
			{
			}
			if(_view!=null)
			{
				LySwingView winView = (LySwingView)_view.getInternalView(false);
				winView.snapshot(_file);
			}
			else if(_area!=null)
			{

				LySwingArea winArea = (LySwingArea)_area.getInternalView(false);
				Component component=winArea.getComponent();
				LySwingUtils.snapshot(component, _file);
			}
		}
	} // end class TakePicture


	/**
	 * Waiting Thread for setWaiting implementation.
	 */
	protected class _LyWaitingThread extends Thread
	{

		/** Initialization status. */
		public final int			 STATUS_INIT = 0;

		/** Started status. */
		public final int			 STATUS_STARTED = 1;

		/** Waiting status. */
		public final int			 STATUS_WAITING = 2;

		/** Done waiting status. */
		public final int			 STATUS_WAITING_DONE = 3;

		/** Over status. */
		public final int			 STATUS_OVER = 4;

		/** Ending the wait request status. */
		public final int			 STATUS_END_WAIT_REQUEST = 5;

		/** The current view manager. */
		protected LySwingViewManager _viewManager;

		/** The current waiting status. */
		protected int				 _waitingStatus;

		/**
		 * Class constructor for a waiting thread. The waiting status is set to {@link
		 * #STATUS_INIT}.
		 *
		 * @param manager the view manager
		 */
		public _LyWaitingThread(LySwingViewManager manager)
		{
			_viewManager = manager;
			_waitingStatus = STATUS_INIT;

			setName("BF AppEngine : LySwingViewManager waiting thread");
		}

		/*
		 * (non-Javadoc)
		 * @see java.lang.Runnable#run()
		 */
		public void run()
		{
			synchronized (this)
			{
				if (_waitingStatus == STATUS_INIT)
					_waitingStatus = STATUS_STARTED;
			}
			try
			{
				if (_waitingStatus == STATUS_STARTED)
					sleep(START_WAITING_DELAY);
			}
			catch (InterruptedException e)
			{
				_waitingStatus = STATUS_OVER;
			}

			boolean needComplete;

			// Lock within status change
			synchronized (this)
			{
				if ((_waitingStatus == STATUS_STARTED) && (_viewManager != null))
				{
					_waitingStatus = STATUS_WAITING;
					needComplete = true;
				}
				else
				{
					_waitingStatus = STATUS_OVER;
					needComplete = false;
				}
			}

			if (needComplete)
			{
				_viewManager.completeSetWaiting(true);

				synchronized (this)
				{
					// Meanwhile, the AWT Thread may have requested stop waiting
					if (_waitingStatus == STATUS_END_WAIT_REQUEST)
						_viewManager.completeSetWaiting(false);
					_waitingStatus = STATUS_WAITING_DONE;
				}
			}
		} // end method run

		/**
		 * Stops waiting and do not execute completion.
		 */
		public void endWait()
		{
			// Lock within status change
			synchronized (this)
			{
				if (_waitingStatus == STATUS_INIT)
					// Thrtead has not even started yet !
					_waitingStatus = STATUS_OVER;
				if (_waitingStatus == STATUS_STARTED)
					// <- Stop request before any completion
					// Thread will end (This case should happen quite often)
					_waitingStatus = STATUS_OVER;
				else if (_waitingStatus == STATUS_WAITING)
					// When Threads ends : request endWait
					// completeSetWaiting is currently in progress !
					_waitingStatus = STATUS_END_WAIT_REQUEST;
				else if (_waitingStatus == STATUS_WAITING_DONE)
				{
					// completeSetWaiting has finished. Thread is over.
					_waitingStatus = STATUS_OVER;
					if (_viewManager != null)
						_viewManager.completeSetWaiting(false);
				}
				else if (_waitingStatus == STATUS_OVER)
				{
					// Nothing to do : Thread has ended
				}
			}
		} // end method endWait
	} // end class _LyWaitingThread

	/**
	 * Stops GUI.
	 */
	public void stopGui(LyGuiClientInterface client)
	{
		execGui(new _LySwingViewManagerClientStop(client));
	}

	/**
	 * Stop client.
	 */
	class _LySwingViewManagerClientStop
		implements Runnable
	{
		LyGuiClientInterface _client;

		_LySwingViewManagerClientStop(LyGuiClientInterface client)
		{
			_client = client;
		}

		public void run()
		{
			_client.guiStop();
		}
	}

	/**
	 * Accessor to current application splashScreen.
	 *
	 * @return the splashscreen which is an instance of a Swing JWindow
	 */
	public static LySwingSplashScreen getSplashScreen()
	{
		return _splashScreen;
	}

	/**
	 * Return the action that allow to download temporary resources but dependant from the
	 * viewer type (for ex: url / partial url for web).
	 */
	public String getDownloadAction()
	{
		return null;
	}
	
	/**
	 * Indicates whether given image may be cached.
	 * Default implementation returns true.
	 * 
	 * @param path  The path of image that may be cached.
	 * @param image The image that may be cached.
	 * @return true if image may be cached, false otherwise
	 */
	public final boolean useCacheForImage(String path, Object image)
	{
		return true;
	}

} // end class LySwingViewManager
