/*
 * LyAbstractBooleanPanelManager.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import leon.view.filters.*;
import leon.view.win.swing.*;


/**
 * This component handles a checkbox with two or three states.
 *
 * <p>When optional, the mode is three states: {@link #BOOLEAN_CHECKED}, {@link #BOOLEAN_UNCHECKED}
 * and {@link #BOOLEAN_EMPTY}.</p>
 *
 * <p>When not optional, there is only two states: CHECKED and UNCHECKED</p>
 *
 * @version $Id: LyAbstractBooleanPanelManager.java 34538 2012-03-13 13:19:02Z jlsaouli $
 * @author  Lyria
 */
public abstract class LyAbstractBooleanPanelManager
{
	public static final String __VERSION =
		"$Id: LyAbstractBooleanPanelManager.java 34538 2012-03-13 13:19:02Z jlsaouli $";

	/** Empty status. Nothin is selected. */
	public static final int    BOOLEAN_EMPTY = 0;

	/** Checked status. */
	public static final int    BOOLEAN_CHECKED = 1;

	/** Unchecked status. */
	public static final int    BOOLEAN_UNCHECKED = 2;

	/** Gap value for drawing focus rectangle (on both sides). */
	protected static final int GAP = 2;

	/** Background image of the component. */
	Image					   _background;

	/** Background image of the component in hoover mode. */
	Image					   _hoover;

	/** Image of the component when nothing is selected. */
	Image					   _selectEmpty;

	/** Image of the component when true is selected. */
	Image					   _selectChecked;

	/** Image of the component when false is selected. */
	Image					   _selectUnchecked;

	/** Background image of the component when it is disabled. */
	Image					   _disabledBackground;

	/** Disabled image of the component when nothing is selected. */
	Image					   _disabledEmpty;

	/** Disabled image of the component when true is selected. */
	Image					   _disabledChecked;

	/** Disabled image of the component when false is selected. */
	Image					   _disabledUnchecked;

	/** Offscreen image (for double buffering). */
	Image					   _offscreen;

	/**
	 * Status of the component. May be:
	 *
	 * <ul>
	 *   <li><code>null</code> for BOOLEAN_EMPTY</li>
	 *   <li>{@link Boolean#TRUE} for BOOLEAN_TRUE</li>
	 *   <li>{@link Boolean#FALSE} for BOOLEAN_FALSE</li>
	 * </ul>
	 */
	Boolean					   _status;

	/** Enabled status of the component. */
	boolean					   _isEnabled;

	/**
	 * Optional status of the component. If the component is not optional, the BOOLEAN_EMPTY status
	 * is not available.
	 */
	boolean					   _isOptional;

	/** Hoover status. */
	boolean					   _hooverStatus;

	/** Stores all action listeners. */
	Vector<ActionListener>	   _actionListeners;

	/** The empty selection label. */
	String					   _emptyLabel;

	/** The checked selection label. */
	String					   _checkedLabel;

	/** The unchecked selection label. */
	String					   _uncheckedLabel;

	/** The component linked to the manager. */
	protected Component		   _comp;

	/**
	 * Class constructor of a boolean panel manager.
	 *
	 * @param comp the panel component that is initialized
	 */
	public LyAbstractBooleanPanelManager(Component comp)
	{
		_comp = comp;

		init();
	}

	/**
	 * Initializes the boolean panel component. Registers event listeners and sets the parameters at
	 * their default values.
	 */
	protected void init()
	{
		// Register this component as a mouse Listener
		_comp.addMouseListener(new _LyBooleanCanvasMouseListener());

		// Register this component as a key Listener
		_comp.addKeyListener(new _LyBooleanCanvasKeyListener());

		// Register this component as a focus Listener
		_comp.addFocusListener(new _LyBooleanCanvasFocusListener());

		_isEnabled = true;
		_hooverStatus = false;
		_background = null;
		_hoover = null;
		_selectEmpty = null;
		_selectChecked = null;
		_selectUnchecked = null;
		_disabledBackground = null;
		_disabledEmpty = null;
		_disabledChecked = null;
		_disabledUnchecked = null;
		_emptyLabel = null;
		_checkedLabel = null;
		_uncheckedLabel = null;
	}

	/**
	 * Sets the images for this component.
	 *
	 * @param background  the background image of this component
	 * @param hoover      the hoover image used for this component
	 * @param selectEmpty the image used when BOOLEAN_EMPTY status is selected
	 * @param selectFalse the image used when BOOLEAN_FALSE status is selected
	 * @param selectTrue  the image used when BOOLEAN_TRUE status is selected
	 */
	public void setImages(Image background, Image hoover, Image selectEmpty, Image selectFalse,
		Image selectTrue)
	{
		_background = background;
		_hoover = hoover;
		_selectEmpty = selectEmpty;
		_selectChecked = selectTrue;
		_selectUnchecked = selectFalse;

		computeDisabledImages();
	}

	/**
	 * Sets the labels of this component.
	 *
	 * @param emptyLabel     label displayed when the selection is empty
	 * @param checkedLabel   label displayed when the selection is checked
	 * @param uncheckedLabel label displayed when the selection is unchecked
	 * @see   LyAbstractBooleanPanelManager#setLabels(String, String, String)
	 */
	public void setLabels(String emptyLabel, String checkedLabel, String uncheckedLabel)
	{
		_emptyLabel = emptyLabel;
		_checkedLabel = checkedLabel;
		_uncheckedLabel = uncheckedLabel;
	}

	/**
	 * Computes the disabled images for this component. Disposes the images if they are allocated,
	 * then computes the grayed disabled images if the enabled image already exists.
	 */
	protected void computeDisabledImages()
	{
		// If the images were previously allocated, dispose them
		if (_disabledBackground != null)
			_disabledBackground.flush();
		if (_disabledEmpty != null)
			_disabledEmpty.flush();
		if (_disabledChecked != null)
			_disabledChecked.flush();
		if (_disabledUnchecked != null)
			_disabledUnchecked.flush();

		// Now, generate disabled images (only if there is an equivalent enabled image)
		if (_background != null)
			_disabledBackground = LyDisableImageFilter.getFilteredImage(_background);
		else
			_disabledBackground = null;

		if (_selectEmpty != null)
			_disabledEmpty = LyDisableImageFilter.getFilteredImage(_selectEmpty);
		else
			_disabledEmpty = null;

		if (_selectChecked != null)
			_disabledChecked = LyDisableImageFilter.getFilteredImage(_selectChecked);
		else
			_disabledChecked = null;

		if (_selectUnchecked != null)
			_disabledUnchecked = LyDisableImageFilter.getFilteredImage(_selectUnchecked);
		else
			_disabledUnchecked = null;
	} // end method computeDisabledImages

	/**
	 * Redraws the component.
	 */
	public void redraw()
	{
		_comp.repaint();
	}

	/**
	 * Changes the hoover drawing mode.
	 *
	 * @param status Indicates that the hoover image must be drawn if this component is enabled.
	 */
	protected void drawHoover(boolean status)
	{
		if (_isEnabled)
		{
			_hooverStatus = status;
			redraw();
		}
	}

	/**
	 * Returns the status of the component.
	 *
	 * @return the status of the component
	 * @see    #setValue(Boolean)
	 */
	public Boolean getValue()
	{
		return _status;
	}

	/**
	 * Sets the current status of the component and redraws it.
	 *
	 * @param status the new status
	 * @see   #getValue()
	 */
	public void setValue(Boolean status)
	{
		_status = status;
		redraw();
	}

	/**
	 * Returns the enabled status of this component.
	 *
	 * @return true if the component is enabled, false otherwise
	 */
	public boolean isEnabled()
	{
		return _isEnabled;
	}

	/**
	 * Returns the optional status of the component.
	 *
	 * @return true if the component is optional, false otherwise
	 */
	public boolean isOptional()
	{
		return _isOptional;
	}

	/**
	 * Enables this component.
	 *
	 * @param enabled true to enable this component
	 */
	public void setEnabled(boolean enabled)
	{
		_isEnabled = enabled;
		redraw();
	}

	/**
	 * Sets the optional parameter of this component. When this component is not optional, the
	 * status BOOLEAN_EMPTY is not available.
	 *
	 * @param optional the optional status of the component
	 */
	public void setOptional(boolean optional)
	{
		_isOptional = optional;
	}

	/**
	 * The size of this component. This is computed as follow:
	 *
	 * <ul>
	 *   <li>the size of the biggest image between background and hoover, if both are specified</li>
	 *   <li>the size of the background image if the hoover image is <code>null</code></li>
	 *   <li>the size of the hoover image if the background image is <code>null</code></li>
	 *   <li>the size of the biggest image between {@link #_selectEmpty}, {@link #_selectChecked}
	 *     and {@link #_selectUnchecked}, if the background and hoover image are <code>
	 *     null</code></li>
	 * </ul>
	 *
	 * @return the preferred size of this component.
	 */
	public Dimension getPreferredSize()
	{
		int finalWidth, finalHeight;

		if ((_background != null) && (_hoover != null))
		{
			int backgroundHeight = _background.getHeight(null);
			int backgroundWidth = _background.getWidth(null);
			int hooverHeight = _background.getHeight(null);
			int hooverWidth = _background.getWidth(null);

			finalWidth = Math.max(backgroundWidth, hooverWidth);
			finalHeight = Math.max(backgroundHeight, hooverHeight);
		}
		else if (_background != null)
		{
			finalHeight = _background.getHeight(null);
			finalWidth = _background.getWidth(null);
		}
		else if (_hoover != null)
		{
			finalHeight = _hoover.getHeight(null);
			finalWidth = _hoover.getWidth(null);
		}
		else
		{
			int emptyWidth = ((_selectEmpty == null) ? -1 : _selectEmpty.getWidth(null));
			int emptyHeight = ((_selectEmpty == null) ? -1 : _selectEmpty.getHeight(null));
			int trueWidth = ((_selectEmpty == null) ? -1 : _selectChecked.getWidth(null));
			int trueHeight = ((_selectEmpty == null) ? -1 : _selectChecked.getHeight(null));
			int falseWidth = ((_selectEmpty == null) ? -1 : _selectUnchecked.getWidth(null));
			int falseHeight = ((_selectEmpty == null) ? -1 : _selectUnchecked.getHeight(null));

			int width = Math.max(emptyWidth, trueWidth);

			finalWidth = Math.max(width, falseWidth);

			int height = Math.max(emptyHeight, trueHeight);

			finalHeight = Math.max(height, falseHeight);
		}

		// compute label size
		int		    emptyLabelWidth = 0;
		int		    checkedLabelWidth = 0;
		int		    uncheckedLabelWidth = 0;

		FontMetrics fm = _comp.getFontMetrics(_comp.getFont());

		if (_emptyLabel != null)
		{
			emptyLabelWidth = fm.stringWidth(_emptyLabel);
			finalHeight = Math.max(finalHeight, fm.getHeight());
		}

		if (_checkedLabel != null)
		{
			checkedLabelWidth = fm.stringWidth(_checkedLabel);
			finalHeight = Math.max(finalHeight, fm.getHeight());
		}

		if (_uncheckedLabel != null)
		{
			uncheckedLabelWidth = fm.stringWidth(_uncheckedLabel);
			finalHeight = Math.max(finalHeight, fm.getHeight());
		}

		int maxWidth = Math.max(emptyLabelWidth, checkedLabelWidth);

		maxWidth = Math.max(maxWidth, uncheckedLabelWidth);

		return new Dimension(finalWidth + (LyAbstractBooleanPanelManager.GAP * 4) + maxWidth,
			finalHeight + (LyAbstractBooleanPanelManager.GAP * 2)); // add size for focus
		// rectangle and label
	} // end method getPreferredSize

	/**
	 * Registers a new listener on this component.
	 *
	 * @param listener a new listener
	 * @see   #removeActionListener(ActionListener)
	 */
	public void addActionListener(ActionListener listener)
	{
		if (listener == null)
			return;
		if (_actionListeners == null)
			_actionListeners = new Vector<ActionListener>();

		_actionListeners.addElement(listener);
	}

	/**
	 * Removes a previously registered listener on this component.
	 *
	 * @param listener the listener to remove
	 * @see   #addActionListener(ActionListener)
	 */
	public void removeActionListener(ActionListener listener)
	{
		if (listener == null)
			return;

		if (_actionListeners == null)
			return;

		if (_actionListeners.contains(listener))
			_actionListeners.removeElement(listener);
	}

	/**
	 * Cycles through the different status of this component.
	 */
	public void nextStatus()
	{
		if (_status == null)
			_status = Boolean.FALSE;
		else if (_status.booleanValue() && _isOptional)
			_status = null;
		else if (_status.booleanValue())
			_status = Boolean.FALSE;
		else
			_status = Boolean.TRUE;

		redraw();

		// Notify listener that something has changed
		if (_actionListeners != null)
		{
			ActionListener listener;

			for (int i = 0; i < _actionListeners.size(); i++)
			{
				listener = _actionListeners.elementAt(i);
				listener.actionPerformed(null);
			}
		}
	}

	/**
	 * Gets the component linked to this manager.
	 *
	 * @return the associated component
	 */
	public Component getComponent()
	{
		return _comp;
	}

	/**
	 * Returns the offscreen image if any.
	 *
	 * @param  offScreenSize size of the offscreen image
	 * @return the offscreen image
	 */
	public abstract Image getImageOffScreen(Dimension offScreenSize);

	/**
	 * Returns the graphics context of the specified offscreen image or the default graphics context
	 * if none is available.
	 *
	 * @param  offScreenImage  the image to get its graphics context
	 * @param  defaultGraphics the default graphics context if the image doesn't have one
	 * @return a graphics context to use
	 */
	public abstract Graphics getTargetGraphics(Image offScreenImage, Graphics defaultGraphics);

	/**
	 * Paints the specified offscreen image to the onscreen graphics context.
	 *
	 * @param offScreenImage   the offscreen image to paint on the onscreen context
	 * @param onScreenGraphics the graphics context to paint the offscreen image on
	 */
	public abstract void paintOnScreen(Image offScreenImage, Graphics onScreenGraphics);

	/**
	 * Draws the title panel in this graphics.
	 *
	 * @param g the graphics context to paint the images on
	 */
	public void paintCommon(Graphics g)
	{
		Image     background = null;
		Image     hoover = null;
		Image     selectEmpty = null;
		Image     selectFalse = null;
		Image     selectTrue = null;
		String    label = null;
		Image     middleImage = null;
		Dimension dim = getPreferredSize();

		// Prepare the correct images
		if (_isEnabled)
		{
			background = _background;
			hoover = _hoover;
			selectEmpty = _selectEmpty;
			selectFalse = _selectUnchecked;
			selectTrue = _selectChecked;
		}
		else
		{
			background = _disabledBackground;
			hoover = null;
			selectEmpty = _disabledEmpty;
			selectFalse = _disabledUnchecked;
			selectTrue = _disabledChecked;
		}

		Color c = g.getColor();

		if (_comp.isOpaque())
		{
			g.setColor(_comp.getBackground());
			g.fillRect(0, 0, dim.width, dim.height);
		}

		// If we have a background image, draw it
		if (background != null)
		{
			int posY = (dim.height / 2) - (background.getHeight(_comp) / 2);

			g.drawImage(background, LyAbstractBooleanPanelManager.GAP, posY,
				background.getWidth(_comp), background.getHeight(_comp), _comp);
		} // For focus

		// If we are in hoover mode and if we have a hoover image
		// draw it
		if (_hooverStatus && (hoover != null))
		{
			int posX = LyAbstractBooleanPanelManager.GAP;
			int posY = (dim.height / 2) - (hoover.getHeight(_comp) / 2);

			if (background != null)
			{
				int width = hoover.getWidth(_comp);

				int backgroundWidth = _background.getWidth(_comp);

				posX = posX + (backgroundWidth / 2) - (width / 2);
			}

			g.drawImage(hoover, posX, posY, hoover.getWidth(_comp), hoover.getHeight(_comp), _comp);
		}

		// Compute the correct image for the middle image
		// compute the correct label
		if (_status == null)
		{
			middleImage = selectEmpty;
			label = _emptyLabel;
		}
		else if (_status.booleanValue())
		{
			middleImage = selectTrue;
			label = _checkedLabel;
		}
		else
		{
			middleImage = selectFalse;
			label = _uncheckedLabel;
		}

		FontMetrics fm = _comp.getFontMetrics(_comp.getFont());
		Point	    labelSize;

		if (label != null)
			labelSize = new Point(fm.stringWidth(label), fm.getHeight());
		else
			labelSize = new Point(0, 0);

		// If the component is enabled and if it has the focus, show the focus box
		if (_isEnabled && _comp.hasFocus())
		{
			int max = 0;

			if (_background != null)
				max = _background.getWidth(_comp) + (2 * LyAbstractBooleanPanelManager.GAP);
			else if (middleImage != null)
				max = middleImage.getWidth(_comp) + (2 * LyAbstractBooleanPanelManager.GAP);

			if (label != null)
				max = max + labelSize.x + LyAbstractBooleanPanelManager.GAP;

			// if the focus has changed,
			// draw a rectangle around the component
			LySwingUtils.drawFocus(g, 0, 0, max - 2, _comp.getHeight() - 2);
		}

		g.setColor(c);

		// If we don't have a middle image, skip
		if (middleImage == null)
			return;

		// Compute the position of the middle image and draw it
		int posX = LyAbstractBooleanPanelManager.GAP;
		int posY = (dim.height / 2) - (middleImage.getHeight(_comp) / 2);

		if (_background != null)
		{
			int width = middleImage.getWidth(_comp);

			int backgroundWidth = _background.getWidth(_comp);

			posX = posX + (backgroundWidth / 2) - (width / 2);
		}

		g.drawImage(middleImage, posX, posY, middleImage.getWidth(_comp),
			middleImage.getHeight(_comp), _comp);

		// If we have a label to draw, compute its position and draw it
		if (label != null)
		{
			if (_background != null)
				posX = _background.getWidth(_comp) + (2 * LyAbstractBooleanPanelManager.GAP);
			else
				posX = middleImage.getWidth(_comp) + (2 * LyAbstractBooleanPanelManager.GAP);

			posY = (dim.height / 2) - (labelSize.y / 2) + fm.getAscent();

			g.drawString(label, posX, posY);
		}
	} // end method paintCommon

	/**
	 * This class handles mouse events on this component.
	 */
	class _LyBooleanCanvasMouseListener
		implements MouseListener
	{

		/**
		 * Nothing is done on a double click event.
		 *
		 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
		 */
		public void mouseClicked(MouseEvent event)
		{
			// Nothing to do on double click
		}

		/**
		 * Changes the status of the boolean component.
		 *
		 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
		 */
		public void mousePressed(MouseEvent event)
		{
			// Change Status
			if (!_isEnabled)
				return;

			nextStatus();
		}

		/**
		 * Nothing is done on the mouse release.
		 *
		 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
		 */
		public void mouseReleased(MouseEvent event)
		{
			// nothing to do
		}

		/**
		 * Draws the hoover image.
		 *
		 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
		 */
		public void mouseEntered(MouseEvent event)
		{
			drawHoover(true);
		}

		/**
		 * Stops drawing the hoover image.
		 *
		 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
		 */
		public void mouseExited(MouseEvent event)
		{
			drawHoover(false);
		}
	} // end class _LyBooleanCanvasMouseListener


	/**
	 * This class handles key events for this component.
	 */
	class _LyBooleanCanvasKeyListener
		implements KeyListener
	{

		/**
		 * Changes the status of the boolean component when the space bar is pressed.
		 *
		 * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
		 */
		public void keyPressed(KeyEvent e)
		{
			// If the component is not enabled, skip event
			if (!_isEnabled)
				return;

			// Change the status
			if (e.getKeyChar() == ' ') // SPACE BAR
				nextStatus();
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
		 */
		public void keyTyped(KeyEvent e)
		{
			// nothing to do
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
		 */
		public void keyReleased(KeyEvent e)
		{
			// nothing to do
		}
	} // end class _LyBooleanCanvasKeyListener


	/**
	 * This class handle focus for this component.
	 */
	class _LyBooleanCanvasFocusListener
		implements FocusListener
	{

		/**
		 * Redraws the component.
		 *
		 * @see java.awt.event.FocusListener#focusGained(java.awt.event.FocusEvent)
		 */
		public void focusGained(FocusEvent e)
		{
			redraw();
		}

		/**
		 * Redraws the component.
		 *
		 * @see java.awt.event.FocusListener#focusLost(java.awt.event.FocusEvent)
		 */
		public void focusLost(FocusEvent e)
		{
			redraw();
		}
	}
} // end class LyAbstractBooleanPanelManager
