/*
 * LyAbstractCanvasSplitterManager.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;
import leon.view.win.swing.*;


/**
 * Handles the splitting of an area in multiple coordinated areas.
 *
 * @version $Id: LyAbstractCanvasSplitterManager.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public abstract class LyAbstractCanvasSplitterManager
{
	public static final String   __VERSION =
		"$Id: LyAbstractCanvasSplitterManager.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** None of possible parts. */
	public final static short    NO_PART = 0;

	/** The main top left part. */
	public final static short    MAIN_TOP_LEFT_PART = 1;

	/** The main top right part. */
	public final static short    MAIN_TOP_RIGHT_PART = 2;

	/** The main bottom left part. */
	public final static short    MAIN_BOTTOM_LEFT_PART = 3;

	/** The main bottom right part. */
	public final static short    MAIN_BOTTOM_RIGHT_PART = 4;

	/** The top header left part. */
	public final static short    TOP_HEADER_LEFT_PART = 5;

	/** The top header right part. */
	public final static short    TOP_HEADER_RIGHT_PART = 6;

	/** The left header top part. */
	public final static short    LEFT_HEADER_TOP_PART = 7;

	/** The left header bottom part. */
	public final static short    LEFT_HEADER_BOTTOM_PART = 8;

	/** Thickness of the splitters. */
	public final static int		 SPLITTER_THICKNESS = 4;

	/** Thickness of the scrollbars. */
	public final static int		 SCROLLBAR_THICKNESS = 15;

	/** Default cursor. */
	protected Cursor			 _defaultCursor;

	/** Cursor for horizontal resize. */
	protected Cursor			 _hResizeCursor;

	/** Cursor for vertical resize. */
	protected Cursor			 _vResizeCursor;

	/** Cursor for horizontal and vertical resize. */
	protected Cursor			 _crossCursor;

	/** Main canvas of the splitter. */
	protected LySplittableCanvas _mainCanvas;

	/** Top header canvas of the splitter. */
	protected LySplittableCanvas _topHeaderCanvas;

	/** Left header canvas of the splitter. */
	protected LySplittableCanvas _leftHeaderCanvas;

	/** Height of the top header. */
	protected int				 _topHeaderHeight;

	/** Width of the left header. */
	protected int				 _leftHeaderWidth;

	/** Indicates which part of the splitter is being drawn. */
	protected short				 _drawnPart;

	/** Indicates which part of the splitter has the mouse over. */
	protected short				 _mouseOverPart;

	/** Ratio of the horizontal splitter. */
	protected int				 _horizontalSplitterRatio;

	/** Ratio of the vertical splitter. */
	protected int				 _verticalSplitterRatio;

	/** Position of the horizontal splitter. */
	protected int				 _horizontalSplitterPos;

	/** Position of the vertical splitter. */
	protected int				 _verticalSplitterPos;

	/** Indicates left splitter movement. */
	protected boolean			 _movingLeftSplitter;

	/** X position of the left splitter when moving. */
	protected int				 _movingLeftSplitterX;

	/** Indicates horizontal splitter movement. */
	protected boolean			 _movingHorizontalSplitter;

	/** X position of the horizontal splitter when moving. */
	protected int				 _movingHorizontalSplitterX;

	/** Indicates horizontal splitter movement. */
	protected boolean			 _movingVerticalSplitter;

	/** Y position of the vertical splitter when moving. */
	protected int				 _movingVerticalSplitterY;

	/** Left scrollbar. */
	protected Component			 _leftSplitter;

	/** Right scrollbar. */
	protected Component			 _rightSplitter;

	/** Top scrollbar. */
	protected Component			 _topSplitter;

	/** Bottom scrollbar. */
	protected Component			 _bottomSplitter;

	/** Left top splitter. */
	protected Component			 _leftTopSplitter;

	/** Left bottom splitter. */
	protected Component			 _leftBottomSplitter;

	/** Indicates the active canvas. */
	protected LySplittableCanvas _activatedCanvas;

	/** Offscreen image. */
	protected Image				 _offscreen = null;

	/** Dimension of offscreen image. */
	protected Dimension			 _offscreenDim = null;

	/** Last offscreen image. */
	protected Image				 _lastOffscreen = null;

	/** Last offscreen graphics. */
	protected Graphics			 _lastOffscreenGraphics;

	/** The component linked to the manager. */
	protected Container			 _comp;

	/**
	 * Class constructor of a canvas splitter manager.
	 *
	 * @param comp the splitted component
	 */
	public LyAbstractCanvasSplitterManager(Container comp)
	{
		super();

		_comp = comp;
		_mainCanvas = null;
		_drawnPart = NO_PART;
		_horizontalSplitterRatio = 100;
		_verticalSplitterRatio = 0;
		_movingLeftSplitter = false;
		_movingHorizontalSplitter = false;
		_movingVerticalSplitter = false;
		_mouseOverPart = NO_PART;
		_activatedCanvas = null;

		init();
	}

	/**
	 * Initializes the canvas splitter attributes and register the associated component listeners.
	 */
	protected void init()
	{
		_comp.addFocusListener(new _LyCanvasSplitterFocusListener());
		_comp.addMouseMotionListener(new _LyCanvasSplitterMouseMoveListener());
		_comp.addMouseListener(new _LyCanvasSplitterMouseListener());
		_comp.addKeyListener(new _LyCanvasSplitterKeyListener());

		// Initialize cursors
		_defaultCursor = new Cursor(Cursor.DEFAULT_CURSOR);
		_hResizeCursor = new Cursor(Cursor.W_RESIZE_CURSOR);
		_vResizeCursor = new Cursor(Cursor.N_RESIZE_CURSOR);
		_crossCursor = new Cursor(Cursor.MOVE_CURSOR);

		// Create the scrollbars
		_LyCanvasSplitterScrollbarMouseListener scrollbarMouseListener =
			new _LyCanvasSplitterScrollbarMouseListener();
		_LyCanvasSplitterSelectionListener	    selectionListener =
			new _LyCanvasSplitterSelectionListener();

		_leftSplitter = buildSplitter(true);
		((Adjustable)_leftSplitter).addAdjustmentListener(selectionListener);
		((Adjustable)_leftSplitter).setMinimum(0);
		((Adjustable)_leftSplitter).setUnitIncrement(20);
		_leftSplitter.addMouseListener(scrollbarMouseListener);
		_comp.add(_leftSplitter);

		_rightSplitter = buildSplitter(true);
		((Adjustable)_rightSplitter).addAdjustmentListener(selectionListener);
		((Adjustable)_rightSplitter).setMinimum(0);
		((Adjustable)_rightSplitter).setUnitIncrement(20);
		_rightSplitter.addMouseListener(scrollbarMouseListener);
		_comp.add(_rightSplitter);

		_topSplitter = buildSplitter(false);
		((Adjustable)_topSplitter).addAdjustmentListener(selectionListener);
		((Adjustable)_topSplitter).setMinimum(0);
		((Adjustable)_topSplitter).setUnitIncrement(20);
		_topSplitter.addMouseListener(scrollbarMouseListener);
		_comp.add(_topSplitter);

		_bottomSplitter = buildSplitter(false);
		((Adjustable)_bottomSplitter).addAdjustmentListener(selectionListener);
		((Adjustable)_bottomSplitter).setMinimum(0);
		((Adjustable)_bottomSplitter).setUnitIncrement(20);
		_bottomSplitter.addMouseListener(scrollbarMouseListener);
		_comp.add(_bottomSplitter);

		_leftTopSplitter = buildSplitter(true);
		((Adjustable)_leftTopSplitter).addAdjustmentListener(selectionListener);
		((Adjustable)_leftTopSplitter).setMinimum(0);
		((Adjustable)_leftTopSplitter).setUnitIncrement(20);
		_leftTopSplitter.addMouseListener(scrollbarMouseListener);
		_comp.add(_leftTopSplitter);

		_leftBottomSplitter = buildSplitter(true);
		((Adjustable)_leftBottomSplitter).addAdjustmentListener(selectionListener);
		((Adjustable)_leftBottomSplitter).setMinimum(0);
		((Adjustable)_leftBottomSplitter).setUnitIncrement(20);
		_leftTopSplitter.addMouseListener(scrollbarMouseListener);
		_comp.add(_leftBottomSplitter);
	} // end method init

	/**
	 * Creates a new splitter in the specified orientation.
	 *
	 * @param  horizontal Indicates if the splitter is oriented horizontally (<code>true</code>) or
	 *                    vertically (<code>false</code>).
	 * @return a new splitted component
	 */
	public abstract Component buildSplitter(boolean horizontal);

	/**
	 * Sets the main canvas of this splitter. Main canvas may be splitted into 4 parts.
	 *
	 * @param canvas the main canvas of this splitter
	 */
	public void setMainCanvas(LySplittableCanvas canvas)
	{
		_mainCanvas = canvas;
		canvas.setCanvasSplitter(this);
	}

	/**
	 * Sets the top header canvas of this splitter. Top header canvas may be splitted into 2 parts
	 * (left and right).
	 *
	 * @param canvas the top header canvas of this splitter
	 */
	public void setTopHeaderCanvas(LySplittableCanvas canvas)
	{
		_topHeaderCanvas = canvas;
		_topHeaderHeight = canvas.getTopHeaderHeight();
		canvas.setCanvasSplitter(this);
	}

	/**
	 * Sets the left header canvas of this splitter. Left header canvas may be splitted into 2 parts
	 * (top and bottom).
	 *
	 * @param canvas the left header canvas of this splitter
	 */
	public void setLeftHeaderCanvas(LySplittableCanvas canvas)
	{
		_leftHeaderCanvas = canvas;
		_leftHeaderWidth = canvas.getLeftHeaderWidth();
		canvas.setCanvasSplitter(this);
	}

	/**
	 * Gets the component linked to this manager.
	 *
	 * @return the associated component
	 */
	public Container getComponent()
	{
		return _comp;
	}

	/**
	 * Redraws the component.
	 */
	public void redraw()
	{
		_comp.repaint();
	}

	/**
	 * Returns the size of the main canvas (the total size of the canvas minus the size of its
	 * header).
	 *
	 * @return the size of the main canvas
	 */
	public Dimension getMainViewSize()
	{
		Dimension size = getSize();

		return new Dimension(size.width - _leftHeaderWidth,
			size.height - _topHeaderHeight
			- SPLITTER_THICKNESS);
	}

	/**
	 * Returns vertical offset for current part. Current part may be part which has the mouse over
	 * or drawn part.
	 *
	 * @param  mouseOverPart true if mouse over part must be considered, false if draw part must be
	 *                       considered
	 * @return vertical offset for current part
	 */
	public int getVerticalOffset(boolean mouseOverPart)
	{
		short part = (mouseOverPart ? _mouseOverPart : _drawnPart);
		int   offset = 0;

		if ((part == MAIN_TOP_LEFT_PART) || (part == MAIN_TOP_RIGHT_PART)
				|| (part == LEFT_HEADER_TOP_PART))
			offset = ((Adjustable)_topSplitter).getValue();
		else if ((part == MAIN_BOTTOM_LEFT_PART) || (part == MAIN_BOTTOM_RIGHT_PART)
				|| (part == LEFT_HEADER_BOTTOM_PART))
			offset = ((Adjustable)_bottomSplitter).getValue();

		return offset;
	}

	/**
	 * Returns the size of current part. Current part may be part which has the mouse over or drawn
	 * part.
	 *
	 * @param  mouseOverPart true if mouse over part must be considered, false if draw part must be
	 *                       considered
	 * @return the size of current part
	 */
	public Dimension getCurrentPartSize(boolean mouseOverPart)
	{
		Dimension size = getSize();
		int		  width = size.width, height = size.height;
		short     part = (mouseOverPart ? _mouseOverPart : _drawnPart);

		// If no part is being drawn (or has mouse over) return size of the component
		if (part == NO_PART)
			return size;

		// Compute the ratios of the splitter parts
		float leftRatio = (100F / _horizontalSplitterRatio);
		float rightRatio = (100F / (100F - _horizontalSplitterRatio));
		float topRatio = (100F / _verticalSplitterRatio);
		float bottomRatio = (100F / (100F - _verticalSplitterRatio));
		int   mainLeft = (int)((width - _leftHeaderWidth
					- ((_horizontalSplitterRatio < 100) ? SPLITTER_THICKNESS : 0)) / leftRatio);
		int   mainRight = (int)((width - _leftHeaderWidth - SPLITTER_THICKNESS) / rightRatio);
		int   mainTop = (int)((height - _topHeaderHeight - SPLITTER_THICKNESS) / topRatio);
		int   mainBottom = (int)((height - _topHeaderHeight
					- ((_verticalSplitterRatio > 0) ? SPLITTER_THICKNESS : 1)) / bottomRatio);

		if (part == MAIN_TOP_LEFT_PART)
			return new Dimension(mainLeft, mainTop);

		if (part == MAIN_TOP_RIGHT_PART)
			return new Dimension(mainRight, mainTop);

		if (part == MAIN_BOTTOM_LEFT_PART)
			return new Dimension(mainLeft, mainBottom);

		if (part == MAIN_BOTTOM_RIGHT_PART)
			return new Dimension(mainRight, mainBottom);

		if (part == LEFT_HEADER_TOP_PART)
			return new Dimension(_leftHeaderWidth - SPLITTER_THICKNESS, mainTop);

		if (part == LEFT_HEADER_BOTTOM_PART)
			return new Dimension(_leftHeaderWidth
				- ((_verticalSplitterRatio > 0) ? SPLITTER_THICKNESS : 1), mainBottom);

		if (part == TOP_HEADER_LEFT_PART)
			return new Dimension(mainLeft, _topHeaderHeight);

		if (part == TOP_HEADER_RIGHT_PART)
			return new Dimension(mainRight, _topHeaderHeight);

		return getSize();
	} // end method getCurrentPartSize

	/**
	 * Returns the position of current part. Current part may be part which has the mouse over or
	 * drawn part.
	 *
	 * @param  mouseOverPart true if mouse over part must be considered, false if draw part must be
	 *                       considered
	 * @return the position of current part
	 */
	public Point getCurrentPartPosition(boolean mouseOverPart)
	{
		short part = (mouseOverPart ? _mouseOverPart : _drawnPart);
		Point result = new Point(0, 0);

		if ((part == TOP_HEADER_LEFT_PART) || (part == MAIN_TOP_LEFT_PART)
				|| (part == MAIN_BOTTOM_LEFT_PART))
			result.x = _leftHeaderWidth;
		else if ((part == TOP_HEADER_RIGHT_PART) || (part == MAIN_TOP_RIGHT_PART)
				|| (part == MAIN_BOTTOM_RIGHT_PART))
			result.x = _horizontalSplitterPos + 1;

		if ((part == LEFT_HEADER_TOP_PART) || (part == MAIN_TOP_LEFT_PART)
				|| (part == MAIN_TOP_RIGHT_PART))
			result.y = _topHeaderHeight;
		else if ((part == LEFT_HEADER_BOTTOM_PART) || (part == MAIN_BOTTOM_LEFT_PART)
				|| (part == MAIN_BOTTOM_RIGHT_PART))
			result.y = _verticalSplitterPos;

		return result;
	}

	/**
	 * Returns the width of left header part.
	 *
	 * @return the width of left header part
	 */
	public int getLeftHeaderWidth()
	{
		return _leftHeaderWidth;
	}

	/**
	 * Returns the height of top header part.
	 *
	 * @return the height of top header part
	 */
	public int getTopHeaderHeight()
	{
		return _topHeaderHeight;
	}

	/**
	 * Computes the size of the component, depending on the size of the main canvas.
	 *
	 * @return the preferred size of the component
	 */
	public Dimension getPreferredSize()
	{
		if (_mainCanvas instanceof Component)
		{
			Dimension result = ((Component)_mainCanvas).getPreferredSize();

			result.width += _leftHeaderWidth;
			result.height += _topHeaderHeight;

			return result;
		}

		return _comp.getPreferredSize();
	}

	/**
	 * Indicates which part is being drawn.
	 *
	 * @return the part that is being drawn. The result is one of the following:
	 *
	 *         <ul>
	 *           <li>{@link #LEFT_HEADER_BOTTOM_PART}</li>
	 *           <li>{@link #LEFT_HEADER_TOP_PART}</li>
	 *           <li>{@link #MAIN_BOTTOM_LEFT_PART}</li>
	 *           <li>{@link #MAIN_BOTTOM_RIGHT_PART}</li>
	 *           <li>{@link #MAIN_TOP_LEFT_PART}</li>
	 *           <li>{@link #MAIN_TOP_RIGHT_PART}</li>
	 *           <li>{@link #TOP_HEADER_LEFT_PART}</li>
	 *           <li>{@link #TOP_HEADER_RIGHT_PART}</li>
	 *           <li>{@link #NO_PART}</li>
	 *         </ul>
	 */
	public short getDrawnPart()
	{
		return _drawnPart;
	}

	/**
	 * Returns the canvas concerned by a mouse event. Returns <code>null</code> if no canvas
	 * corresponds.
	 *
	 * @param      e the mouse event
	 * @return     the canvas concerned by the given mouse event
	 * @deprecated Replaced by {@link #getCurrentCanvas()}
	 */
	protected LySplittableCanvas getCurrentCanvas(MouseEvent e)
	{
		return getCurrentCanvas();
	}

	/**
	 * Returns the current canvas. Returns <code>null</code> if no canvas corresponds.
	 *
	 * @return the current canvas
	 */
	protected LySplittableCanvas getCurrentCanvas()
	{

		if ((_mouseOverPart == LEFT_HEADER_TOP_PART) || (_mouseOverPart == LEFT_HEADER_BOTTOM_PART))
			return _leftHeaderCanvas;
		else if ((_mouseOverPart == TOP_HEADER_LEFT_PART)
				|| (_mouseOverPart == TOP_HEADER_RIGHT_PART))
			return _topHeaderCanvas;
		else if ((_mouseOverPart == MAIN_BOTTOM_LEFT_PART)
				|| (_mouseOverPart == MAIN_BOTTOM_RIGHT_PART)
				|| (_mouseOverPart == MAIN_TOP_LEFT_PART)
				|| (_mouseOverPart == MAIN_TOP_RIGHT_PART))
			return _mainCanvas;

		return null;
	}

	/**
	 * Updates the position of event relatively to the canvas position.
	 *
	 * @param  e the mouse event
	 * @return the new mouse mouse event
	 */
	protected MouseEvent transformMouseEvent(MouseEvent e)
	{
		int x = e.getX(), y = e.getY();

		if ((_mouseOverPart == TOP_HEADER_LEFT_PART) || (_mouseOverPart == MAIN_TOP_LEFT_PART)
				|| (_mouseOverPart == MAIN_BOTTOM_LEFT_PART))
			x -= _leftHeaderWidth;
		else if ((_mouseOverPart == TOP_HEADER_RIGHT_PART)
				|| (_mouseOverPart == MAIN_TOP_RIGHT_PART)
				|| (_mouseOverPart == MAIN_BOTTOM_RIGHT_PART))
			x -= _horizontalSplitterPos + 1;

		if ((_mouseOverPart == LEFT_HEADER_TOP_PART) || (_mouseOverPart == MAIN_TOP_LEFT_PART)
				|| (_mouseOverPart == MAIN_TOP_RIGHT_PART))
			y -= _topHeaderHeight;
		else if ((_mouseOverPart == LEFT_HEADER_BOTTOM_PART)
				|| (_mouseOverPart == MAIN_BOTTOM_LEFT_PART)
				|| (_mouseOverPart == MAIN_BOTTOM_RIGHT_PART))
			y -= _verticalSplitterPos;

		return new MouseEvent(e.getComponent(), e.getID(), e.getWhen(), e.getModifiers(), x, y,
			e.getClickCount(), e.isPopupTrigger(), e.getButton());
	}

	/**
	 * Activates a canvas.
	 *
	 * @param canvas the canvas that must be activated
	 */
	public void setActivated(LySplittableCanvas canvas)
	{
		if (_activatedCanvas != canvas)
		{
			_activatedCanvas = canvas;
			_comp.repaint();
		}
	}

	/**
	 * Indicates which canvas is activated.
	 *
	 * @return the activated canvas
	 */
	public LySplittableCanvas getActivatedCanvas()
	{
		return _activatedCanvas;
	}

	/**
	 * Performs an horizontal scrolling on the left or right splitter (depending on the canvas under
	 * the mouse).
	 *
	 * @param scrollX the new scroll position
	 */
	public void scrollX(int scrollX)
	{
		Adjustable scrollable = null;

		if ((_mouseOverPart == MAIN_TOP_LEFT_PART) || (_mouseOverPart == MAIN_BOTTOM_LEFT_PART)
				|| (_mouseOverPart == TOP_HEADER_LEFT_PART))
			scrollable = (Adjustable)_leftSplitter;
		else if ((_mouseOverPart == MAIN_TOP_RIGHT_PART)
				|| (_mouseOverPart == MAIN_BOTTOM_RIGHT_PART)
				|| (_mouseOverPart == TOP_HEADER_RIGHT_PART))
			scrollable = (Adjustable)_rightSplitter;

		if (scrollable != null)
			scrollable.setValue(scrollX);
	}

	/**
	 * Centers the current part on the element currently at the center of the whole canvas.
	 */
	public void centerCurrentPart()
	{
		boolean leftPart = true;

		if ((_mouseOverPart == MAIN_TOP_RIGHT_PART) || (_mouseOverPart == MAIN_BOTTOM_RIGHT_PART)
				|| (_mouseOverPart == TOP_HEADER_RIGHT_PART))
			leftPart = false;

		Adjustable scrollable = (leftPart ? (Adjustable)_leftSplitter : (Adjustable)_rightSplitter);
		Dimension  size = getSize();
		int		   width = size.width - _leftHeaderWidth - SPLITTER_THICKNESS;
		float	   leftRatio = (100F / _horizontalSplitterRatio);
		float	   rightRatio = (100F / (100F - _horizontalSplitterRatio));
		int		   mouseOverWidth = (int)(width / (leftPart ? leftRatio : rightRatio));
		int		   pos = scrollable.getValue();

		scrollX(pos + ((width - mouseOverWidth) / 2));
	}

	/**
	 * Performs a vertical scrolling on the top or bottom splitter (depending on the canvas under
	 * the mouse).
	 *
	 * @param scrollY the new scroll position
	 */
	public void scrollY(int scrollY)
	{
		Adjustable scrollable = null;

		if ((_mouseOverPart == MAIN_TOP_LEFT_PART) || (_mouseOverPart == MAIN_TOP_RIGHT_PART)
				|| (_mouseOverPart == LEFT_HEADER_TOP_PART))
			scrollable = (Adjustable)_topSplitter;
		else if ((_mouseOverPart == MAIN_BOTTOM_LEFT_PART)
				|| (_mouseOverPart == MAIN_BOTTOM_RIGHT_PART)
				|| (_mouseOverPart == LEFT_HEADER_BOTTOM_PART))
			scrollable = (Adjustable)_bottomSplitter;

		if (scrollable != null)
			scrollable.setValue(scrollY);
	}

	/**
	 * Draws the part which has the mouse over in the given graphical context.
	 *
	 * @param gc the graphical context
	 */
	protected void drawMouseOverPart(Graphics gc)
	{
		_drawnPart = _mouseOverPart;

		if (_mouseOverPart == LEFT_HEADER_TOP_PART)
			_leftHeaderCanvas.drawInSplitter(gc, 0, ((Adjustable)_topSplitter).getValue());
		else if (_mouseOverPart == LEFT_HEADER_BOTTOM_PART)
			_leftHeaderCanvas.drawInSplitter(gc, 0, ((Adjustable)_bottomSplitter).getValue());
		else if (_mouseOverPart == TOP_HEADER_LEFT_PART)
			_topHeaderCanvas.drawInSplitter(gc, ((Adjustable)_leftSplitter).getValue(), 0);
		else if (_mouseOverPart == TOP_HEADER_RIGHT_PART)
			_topHeaderCanvas.drawInSplitter(gc, ((Adjustable)_rightSplitter).getValue(), 0);
		else if (_mouseOverPart == MAIN_TOP_LEFT_PART)
		{
			_drawnPart = LEFT_HEADER_TOP_PART;
			_mouseOverPart = LEFT_HEADER_TOP_PART;
			_leftHeaderCanvas.drawInSplitter(gc, 0, ((Adjustable)_topSplitter).getValue());
			_drawnPart = MAIN_TOP_LEFT_PART;
			_mouseOverPart = MAIN_TOP_LEFT_PART;
			_mainCanvas.drawInSplitter(gc, ((Adjustable)_leftSplitter).getValue(),
				((Adjustable)_topSplitter).getValue());
		}
		else if (_mouseOverPart == MAIN_TOP_RIGHT_PART)
		{
			_drawnPart = LEFT_HEADER_TOP_PART;
			_mouseOverPart = LEFT_HEADER_TOP_PART;
			_leftHeaderCanvas.drawInSplitter(gc, 0, ((Adjustable)_topSplitter).getValue());
			_drawnPart = MAIN_TOP_RIGHT_PART;
			_mouseOverPart = MAIN_TOP_RIGHT_PART;
			_mainCanvas.drawInSplitter(gc, ((Adjustable)_rightSplitter).getValue(),
				((Adjustable)_topSplitter).getValue());
		}
		else if (_mouseOverPart == MAIN_BOTTOM_LEFT_PART)
		{
			_drawnPart = LEFT_HEADER_BOTTOM_PART;
			_mouseOverPart = LEFT_HEADER_BOTTOM_PART;
			_leftHeaderCanvas.drawInSplitter(gc, 0, ((Adjustable)_bottomSplitter).getValue());
			_drawnPart = MAIN_BOTTOM_LEFT_PART;
			_mouseOverPart = MAIN_BOTTOM_LEFT_PART;
			_mainCanvas.drawInSplitter(gc, ((Adjustable)_leftSplitter).getValue(),
				((Adjustable)_bottomSplitter).getValue());
		}
		else if (_mouseOverPart == MAIN_BOTTOM_RIGHT_PART)
		{
			_drawnPart = LEFT_HEADER_BOTTOM_PART;
			_mouseOverPart = LEFT_HEADER_BOTTOM_PART;
			_leftHeaderCanvas.drawInSplitter(gc, 0, ((Adjustable)_bottomSplitter).getValue());
			_drawnPart = MAIN_BOTTOM_RIGHT_PART;
			_mouseOverPart = MAIN_BOTTOM_RIGHT_PART;
			_mainCanvas.drawInSplitter(gc, ((Adjustable)_rightSplitter).getValue(),
				((Adjustable)_bottomSplitter).getValue());
		}

		_mouseOverPart = _drawnPart;
		_drawnPart = NO_PART;
	} // end method drawMouseOverPart

	/**
	 * Gets the size of the associated component.
	 *
	 * @return the size of the component.
	 * @see    Component#getSize()
	 */
	public Dimension getSize()
	{
		return _comp.getSize();
	}

	/**
	 * Draws the canvas splitter in this graphics.
	 *
	 * @param g the graphics context to paint the images on
	 */
	public void paintCommon(Graphics g)
	{
		_drawnPart = NO_PART;

		Dimension size = getSize();

		if ((size.width <= 0) || (size.height <= 0))
			return;

		if ((_offscreen == null) || (_offscreenDim.width != size.width)
				|| (_offscreenDim.height != size.height))
		{
			_offscreen = null;
			_offscreen = _comp.createImage(size.width, size.height);
			_offscreenDim = size;
			_lastOffscreen = _comp.createImage(size.width, size.height);
		}

		// Compute dimensions for each part
		_drawnPart = TOP_HEADER_LEFT_PART;
		int x = getCurrentPartSize(false).width + _leftHeaderWidth;

		_drawnPart = LEFT_HEADER_TOP_PART;
		int y = getCurrentPartSize(false).height + _topHeaderHeight;

		_drawnPart = MAIN_TOP_LEFT_PART;
		Dimension leftScrollableSize = _mainCanvas.getScrollableSize();

		_drawnPart = MAIN_TOP_RIGHT_PART;
		Dimension rightScrollableSize = _mainCanvas.getScrollableSize();

		_drawnPart = LEFT_HEADER_TOP_PART;
		Dimension topScrollableSize = _leftHeaderCanvas.getScrollableSize();

		_drawnPart = LEFT_HEADER_BOTTOM_PART;
		Dimension bottomScrollableSize = _leftHeaderCanvas.getScrollableSize();

		// Set the scrollbars
		boolean   displayTopSplitter = (topScrollableSize.height > (y - _topHeaderHeight))
				&& (_verticalSplitterRatio > 0);
		boolean   displayLeftTopSplitter = (topScrollableSize.width
					> (_leftHeaderWidth
						- (displayTopSplitter ? SCROLLBAR_THICKNESS : 0)))
				&& (_verticalSplitterRatio > 0);

		boolean   displayBottomSplitter = (bottomScrollableSize.height > (size.height - y))
				&& (_verticalSplitterRatio < 100);
		boolean   displayLeftBottomSplitter = (bottomScrollableSize.width
					> (_leftHeaderWidth
						- (displayBottomSplitter ? SCROLLBAR_THICKNESS : 0)))
				&& (_verticalSplitterRatio < 100);

		boolean   displayLeftSplitter = (leftScrollableSize.width > (x - _leftHeaderWidth));
		boolean   displayRightSplitter = (rightScrollableSize.width
					> (size.width - x
						- SPLITTER_THICKNESS));

		_leftSplitter.setBounds(_leftHeaderWidth, size.height - SCROLLBAR_THICKNESS,
			x
			- _leftHeaderWidth
			+ 1
			- ((displayLeftSplitter && (_horizontalSplitterRatio == 100)) ? (SPLITTER_THICKNESS * 2)
																		  : 0),
			SCROLLBAR_THICKNESS);
		_leftSplitter.setVisible(displayLeftSplitter);
		((Adjustable)_leftSplitter).setMaximum(leftScrollableSize.width);
		((Adjustable)_leftSplitter).setVisibleAmount(x - _leftHeaderWidth - 1);

		_rightSplitter.setBounds(x + SPLITTER_THICKNESS, size.height - SCROLLBAR_THICKNESS,
			size.width - x - SPLITTER_THICKNESS + 1, SCROLLBAR_THICKNESS);
		_rightSplitter.setVisible(displayRightSplitter);
		((Adjustable)_rightSplitter).setMaximum(rightScrollableSize.width);
		((Adjustable)_rightSplitter).setVisibleAmount(size.width - x - SPLITTER_THICKNESS - 1);

		_topSplitter.setBounds(_leftHeaderWidth - SCROLLBAR_THICKNESS - SPLITTER_THICKNESS + 1,
			_topHeaderHeight, SCROLLBAR_THICKNESS,
			y - _topHeaderHeight + 2
			- (displayLeftTopSplitter ? SCROLLBAR_THICKNESS : 1));
		_topSplitter.setVisible(displayTopSplitter);
		((Adjustable)_topSplitter).setMaximum(topScrollableSize.height
			+ (displayLeftTopSplitter ? SCROLLBAR_THICKNESS : 0));
		((Adjustable)_topSplitter).setVisibleAmount(y - _topHeaderHeight - 1);

		_bottomSplitter.setBounds(_leftHeaderWidth - SCROLLBAR_THICKNESS - SPLITTER_THICKNESS + 1,
			y + SPLITTER_THICKNESS + ((_verticalSplitterRatio == 0) ? SPLITTER_THICKNESS : 0),
			SCROLLBAR_THICKNESS,
			size.height - y - SPLITTER_THICKNESS
			- (displayLeftBottomSplitter ? SCROLLBAR_THICKNESS : 0)
			- ((_verticalSplitterRatio == 0) ? SPLITTER_THICKNESS : 0));
		_bottomSplitter.setVisible(displayBottomSplitter);
		((Adjustable)_bottomSplitter).setMaximum(bottomScrollableSize.height
			+ (displayLeftBottomSplitter ? SCROLLBAR_THICKNESS : 0));
		((Adjustable)_bottomSplitter).setVisibleAmount(size.height - y - 1);

		_leftTopSplitter.setBounds(0, y - SCROLLBAR_THICKNESS + 1,
			_leftHeaderWidth
			- SPLITTER_THICKNESS - (displayTopSplitter ? (SCROLLBAR_THICKNESS - 1) : 0),
			SCROLLBAR_THICKNESS);
		_leftTopSplitter.setVisible(displayLeftTopSplitter);
		((Adjustable)_leftTopSplitter).setMaximum(topScrollableSize.width
			+ (displayTopSplitter ? SCROLLBAR_THICKNESS : 0));
		((Adjustable)_leftTopSplitter).setVisibleAmount(_leftHeaderWidth - 1);

		_leftBottomSplitter.setBounds(0, size.height - SCROLLBAR_THICKNESS,
			_leftHeaderWidth
			- SPLITTER_THICKNESS - (displayBottomSplitter ? (SCROLLBAR_THICKNESS - 1) : 0),
			SCROLLBAR_THICKNESS);
		_leftBottomSplitter.setVisible(displayLeftBottomSplitter);
		((Adjustable)_leftBottomSplitter).setMaximum(bottomScrollableSize.width
			+ (displayBottomSplitter ? SCROLLBAR_THICKNESS : 0));
		((Adjustable)_leftBottomSplitter).setVisibleAmount(_leftHeaderWidth - 1);

		Graphics offscreenGraphics = _offscreen.getGraphics();

		offscreenGraphics.setColor(_comp.getBackground());
		offscreenGraphics.fillRect(0, 0, size.width, size.height);

		// Draw the left header and the main canvas
		Image    offscreenLeft = _comp.createImage(_leftHeaderWidth, size.height);
		Graphics offscreenGraphicsLeft = offscreenLeft.getGraphics();

		// Optimization : draw content only if not moving a splitter.
		// Otherwise, only draw splitter movement and last drawn image.
		boolean  drawContent = (!_movingLeftSplitter && !_movingHorizontalSplitter
					&& !_movingVerticalSplitter);

		if ((_verticalSplitterRatio > 0) && drawContent)
		{
			_drawnPart = LEFT_HEADER_TOP_PART;
			_leftHeaderCanvas.drawInSplitter(offscreenGraphicsLeft,
				((Adjustable)_leftTopSplitter).getValue(), ((Adjustable)_topSplitter).getValue());
			offscreenGraphics.drawImage(offscreenLeft, 0, _topHeaderHeight, _comp);
		}

		Image    offscreenMain = _comp.createImage(size.width, size.height);
		Graphics offscreenGraphicsMain = offscreenMain.getGraphics();

		if ((_horizontalSplitterRatio > 0) && (_verticalSplitterRatio > 0) && drawContent)
		{
			_drawnPart = MAIN_TOP_LEFT_PART;
			_mainCanvas.drawInSplitter(offscreenGraphicsMain,
				((Adjustable)_leftSplitter).getValue(), ((Adjustable)_topSplitter).getValue());
			offscreenGraphics.drawImage(offscreenMain, _leftHeaderWidth, _topHeaderHeight, _comp);
		}

		if ((_horizontalSplitterRatio < 100) && (_verticalSplitterRatio > 0) && drawContent)
		{
			_drawnPart = MAIN_TOP_RIGHT_PART;
			_mainCanvas.drawInSplitter(offscreenGraphicsMain,
				((Adjustable)_rightSplitter).getValue(), ((Adjustable)_topSplitter).getValue());
			offscreenGraphics.drawImage(offscreenMain, x + SPLITTER_THICKNESS, _topHeaderHeight,
				_comp);
		}

		if ((_verticalSplitterRatio < 100) && drawContent)
		{
			_drawnPart = LEFT_HEADER_BOTTOM_PART;
			_leftHeaderCanvas.drawInSplitter(offscreenGraphicsLeft,
				((Adjustable)_leftBottomSplitter).getValue(),
				((Adjustable)_bottomSplitter).getValue());
			offscreenGraphics.drawImage(offscreenLeft, 0,
				y
				+ ((_verticalSplitterRatio > 0) ? SPLITTER_THICKNESS : 0), _comp);
		}

		if ((_horizontalSplitterRatio > 0) && (_verticalSplitterRatio < 100) && drawContent)
		{
			_drawnPart = MAIN_BOTTOM_LEFT_PART;
			_mainCanvas.drawInSplitter(offscreenGraphicsMain,
				((Adjustable)_leftSplitter).getValue(), ((Adjustable)_bottomSplitter).getValue());
			offscreenGraphics.drawImage(offscreenMain, _leftHeaderWidth,
				y
				+ ((_verticalSplitterRatio > 0) ? SPLITTER_THICKNESS : 0), _comp);
		}

		if ((_horizontalSplitterRatio < 100) && (_verticalSplitterRatio < 100) && drawContent)
		{
			_drawnPart = MAIN_BOTTOM_RIGHT_PART;
			_mainCanvas.drawInSplitter(offscreenGraphicsMain,
				((Adjustable)_rightSplitter).getValue(), ((Adjustable)_bottomSplitter).getValue());
			offscreenGraphics.drawImage(offscreenMain, x + SPLITTER_THICKNESS,
				y
				+ ((_verticalSplitterRatio > 0) ? SPLITTER_THICKNESS : 0), _comp);
		}

		// Draw the top header canvas
		Image    offscreenTop = _comp.createImage(size.width, _topHeaderHeight);
		Graphics offscreenGraphicsTop = offscreenTop.getGraphics();

		if ((_horizontalSplitterRatio > 0) && drawContent)
		{
			_drawnPart = TOP_HEADER_LEFT_PART;
			_topHeaderCanvas.drawInSplitter(offscreenGraphicsTop,
				((Adjustable)_leftSplitter).getValue(), 0);
			offscreenGraphics.drawImage(offscreenTop, _leftHeaderWidth, 0, _comp);
		}

		if ((_horizontalSplitterRatio < 100) && drawContent)
		{
			_drawnPart = TOP_HEADER_RIGHT_PART;
			_topHeaderCanvas.drawInSplitter(offscreenGraphicsTop,
				((Adjustable)_rightSplitter).getValue(), 0);
			offscreenGraphics.drawImage(offscreenTop, x + SPLITTER_THICKNESS, 0, _comp);
		}

		// Fill a rectangle in the corners of left headers if needed (if both horizontal
		// and vertical scrollbars are displayed)
		if (displayLeftTopSplitter && displayTopSplitter && drawContent)
		{
			offscreenGraphics.setColor(Color.LIGHT_GRAY);
			offscreenGraphics.fillRect(_topSplitter.getBounds().x, _leftTopSplitter.getBounds().y,
				SCROLLBAR_THICKNESS, SCROLLBAR_THICKNESS);
		}

		if (displayLeftBottomSplitter && displayBottomSplitter && drawContent)
		{
			offscreenGraphics.setColor(Color.LIGHT_GRAY);
			offscreenGraphics.fillRect(_bottomSplitter.getBounds().x,
				_leftBottomSplitter.getBounds().y, SCROLLBAR_THICKNESS, SCROLLBAR_THICKNESS);
		}

		offscreenGraphics.setColor(Color.LIGHT_GRAY);

		// Draw horizontal splitter
		_horizontalSplitterPos = x + (SPLITTER_THICKNESS / 2);

		if (displayLeftSplitter)
		{
			if (_horizontalSplitterRatio == 100)
			{
				offscreenGraphics.fill3DRect(x + 1 - (2 * SPLITTER_THICKNESS),
					size.height
					- SCROLLBAR_THICKNESS, (2 * SPLITTER_THICKNESS) - 1, SCROLLBAR_THICKNESS, true);
				_horizontalSplitterPos -= SPLITTER_THICKNESS;
			}
			else
				offscreenGraphics.fillRect(x + 1, 0, SPLITTER_THICKNESS - 1, size.height);
		}

		// Draw vertical splitter
		_verticalSplitterPos = y + (SPLITTER_THICKNESS / 2);

		if (displayBottomSplitter)
		{
			if (_verticalSplitterRatio == 0)
				offscreenGraphics.fill3DRect(_leftHeaderWidth - SCROLLBAR_THICKNESS
					- SPLITTER_THICKNESS + 1, y + 1, SCROLLBAR_THICKNESS,
					(2 * SPLITTER_THICKNESS) - 1, true);
			else
				offscreenGraphics.fillRect(0, y + 1, size.width, SPLITTER_THICKNESS - 1);
		}

		// Draw left splitter
		offscreenGraphics.fillRect(_leftHeaderWidth - SPLITTER_THICKNESS + 1, 0,
			SPLITTER_THICKNESS - 1, size.height);

		// Draw splitter movement
		if (_movingLeftSplitter)
		{
			offscreenGraphics.drawImage(_lastOffscreen, 0, 0, _comp);
			LySwingUtils.fillShadowRect(offscreenGraphics,
				_movingLeftSplitterX
				- SPLITTER_THICKNESS, 0, SPLITTER_THICKNESS - 1, size.height);
		}

		if (_movingHorizontalSplitter || _movingVerticalSplitter)
			offscreenGraphics.drawImage(_lastOffscreen, 0, 0, _comp);

		if (_movingHorizontalSplitter)
			LySwingUtils.fillShadowRect(offscreenGraphics, _movingHorizontalSplitterX, 0,
				SPLITTER_THICKNESS - 1, size.height);

		if (_movingVerticalSplitter)
			LySwingUtils.fillShadowRect(offscreenGraphics, 0, _movingVerticalSplitterY, size.width,
				SPLITTER_THICKNESS - 1);

		g.drawImage(_offscreen, 0, 0, _comp);

		if (_lastOffscreenGraphics != null)
		{
			_lastOffscreenGraphics.dispose();
			_lastOffscreenGraphics = null;
		}

		// Keep last drawn image (optimization)
		if (drawContent)
		{
			_lastOffscreenGraphics = _lastOffscreen.getGraphics();
			_lastOffscreenGraphics.drawImage(_offscreen, 0, 0, _comp);
		}

		// Call the draw method on the component under the mouse in order to set
		// possible sensitive zones or whatever needed.
		drawMouseOverPart(offscreenGraphics);

		offscreenGraphics.dispose();
		offscreenGraphicsMain.dispose();
		offscreenMain = null;
		offscreenGraphicsTop.dispose();
		offscreenTop = null;
		offscreenGraphicsLeft.dispose();
		offscreenLeft = null;
	} // end method paintCommon

	/* Internal classes for event listeners */

	/**
	 * This class handles mouse move events on this component.
	 */
	class _LyCanvasSplitterMouseMoveListener
		implements MouseMotionListener
	{

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
		 */
		public void mouseDragged(MouseEvent e)
		{
			int		  x = e.getX();
			int		  y = e.getY();
			Dimension dim = getSize();

			// Draw left splitter movement
			if (_movingLeftSplitter)
			{
				int margin = SCROLLBAR_THICKNESS * 3;

				_movingLeftSplitterX = x;

				if (_movingLeftSplitterX < margin)
					_movingLeftSplitterX = margin;
				else if (_movingLeftSplitterX > (dim.width - margin))
					_movingLeftSplitterX = dim.width - margin;

				_comp.repaint();
				return;
			}

			// Draw horizontal splitter movement
			if (_movingHorizontalSplitter)
			{
				_movingHorizontalSplitterX = x;

				if (_movingHorizontalSplitterX < _leftHeaderWidth)
					_movingHorizontalSplitterX = _leftHeaderWidth;
			}

			// Draw vertical splitter movement
			if (_movingVerticalSplitter)
			{
				_movingVerticalSplitterY = y;

				if (_movingVerticalSplitterY < _topHeaderHeight)
					_movingVerticalSplitterY = _topHeaderHeight;
			}

			if (_movingHorizontalSplitter || _movingVerticalSplitter)
			{
				_comp.repaint();
				return;
			}

			// Redirect the event to final canvas
			LySplittableCanvas canvas = getCurrentCanvas(e);
			MouseEvent		   e2 = transformMouseEvent(e);

			if (canvas != null)
			{
				MouseMotionListener[] listeners = canvas.getMouseMotionListeners();

				if ((listeners != null) && (listeners.length > 0))
				{
					int i, n = listeners.length;

					for (i = 0; i < n; i++)
						listeners[i].mouseDragged(e2);
				}
			}
		} // end method mouseDragged

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.MouseMotionListener#mouseMoved(java.awt.event.MouseEvent)
		 */
		public void mouseMoved(MouseEvent e)
		{
			int		  x = e.getX();
			int		  y = e.getY();
			Dimension dim = getSize();

			if ((x > (_leftHeaderWidth - SPLITTER_THICKNESS)) && (x < _leftHeaderWidth))
			{
				_mouseOverPart = NO_PART;
				_comp.setCursor(_hResizeCursor);
				return;
			}
			else if ((x
						> (_horizontalSplitterPos - (SPLITTER_THICKNESS / 2)
							- ((_horizontalSplitterRatio < 100) ? 0 : SPLITTER_THICKNESS)))
					&& (x < (_horizontalSplitterPos + (SPLITTER_THICKNESS / 2))))
			{
				// Check if mouse is over both horizontal and vertical splitters
				if ((y > (_verticalSplitterPos - (SPLITTER_THICKNESS / 2)))
						&& (y < (_verticalSplitterPos + (SPLITTER_THICKNESS / 2))))
				{
					_mouseOverPart = NO_PART;
					_comp.setCursor(_crossCursor);
					return;
				}

				if ((_horizontalSplitterRatio < 100) || (y > (dim.height - SCROLLBAR_THICKNESS)))
				{
					_mouseOverPart = NO_PART;
					_comp.setCursor(_hResizeCursor);
					return;
				}
			}
			else if ((y > (_verticalSplitterPos - (SPLITTER_THICKNESS / 2)))
					&& (y
						< (_verticalSplitterPos + (SPLITTER_THICKNESS / 2)
							+ ((_verticalSplitterRatio > 0) ? 0 : SPLITTER_THICKNESS))))
			{
				if ((_verticalSplitterRatio > 0)
						|| ((x < (_leftHeaderWidth - SPLITTER_THICKNESS))
							&& (x > (_leftHeaderWidth
									- SPLITTER_THICKNESS - SCROLLBAR_THICKNESS))))
				{
					_mouseOverPart = NO_PART;
					_comp.setCursor(_vResizeCursor);
					return;
				}
			}

			// Check which part is under the mouse
			short mouseOverPart = NO_PART;

			if ((x < _leftHeaderWidth) && (y > _topHeaderHeight) && (y < _verticalSplitterPos))
				mouseOverPart = LEFT_HEADER_TOP_PART;
			else if ((x < _leftHeaderWidth) && (y > _verticalSplitterPos))
				mouseOverPart = LEFT_HEADER_BOTTOM_PART;
			else if ((x > _leftHeaderWidth) && (x < _horizontalSplitterPos)
					&& (y < _topHeaderHeight))
				mouseOverPart = TOP_HEADER_LEFT_PART;
			else if ((x > _horizontalSplitterPos) && (y < _topHeaderHeight))
				mouseOverPart = TOP_HEADER_RIGHT_PART;
			else if ((x > _leftHeaderWidth) && (x < _horizontalSplitterPos)
					&& (y > _topHeaderHeight) && (y < _verticalSplitterPos))
				mouseOverPart = MAIN_TOP_LEFT_PART;
			else if ((x > _horizontalSplitterPos) && (y > _topHeaderHeight)
					&& (y < _verticalSplitterPos))
				mouseOverPart = MAIN_TOP_RIGHT_PART;
			else if ((x > _leftHeaderWidth) && (x < _horizontalSplitterPos)
					&& (y > _verticalSplitterPos))
				mouseOverPart = MAIN_BOTTOM_LEFT_PART;
			else if ((x > _horizontalSplitterPos) && (y > _verticalSplitterPos))
				mouseOverPart = MAIN_BOTTOM_RIGHT_PART;

			if ((mouseOverPart != NO_PART) && (mouseOverPart != _mouseOverPart)
					&& (_offscreen != null))
			{
				// Call the draw method on the component under the mouse in order to set
				// possible sensitive zones or whatever needed.
				Graphics offscreenGraphics = _offscreen.getGraphics();

				_mouseOverPart = mouseOverPart;
				drawMouseOverPart(offscreenGraphics);
				offscreenGraphics.dispose();
				_comp.setCursor(_defaultCursor);
			}

			if (mouseOverPart == NO_PART)
				_comp.setCursor(_defaultCursor);

			// Redirect the event to final canvas
			LySplittableCanvas canvas = getCurrentCanvas(e);
			MouseEvent		   e2 = transformMouseEvent(e);

			if (canvas != null)
			{
				MouseMotionListener[] listeners = canvas.getMouseMotionListeners();

				if ((listeners != null) && (listeners.length > 0))
				{
					int i, n = listeners.length;

					for (i = 0; i < n; i++)
						listeners[i].mouseMoved(e2);
				}
			}
		} // end method mouseMoved
	} // end class _LyCanvasSplitterMouseMoveListener


	/**
	 * This class handles mouse events on this component.
	 */
	class _LyCanvasSplitterMouseListener
		implements MouseListener
	{

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
		 */
		public void mouseEntered(MouseEvent e)
		{
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
		 */
		public void mouseExited(MouseEvent e)
		{
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
		 */
		public void mouseClicked(MouseEvent e)
		{
			// Redirect the event to final canvas
			LySplittableCanvas canvas = getCurrentCanvas(e);
			MouseEvent		   e2 = transformMouseEvent(e);

			if (canvas != null)
			{
				MouseListener[] listeners = canvas.getMouseListeners();

				if ((listeners != null) && (listeners.length > 0))
				{
					int i, n = listeners.length;

					for (i = 0; i < n; i++)
						listeners[i].mouseClicked(e2);

					redraw();
				}
			}
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
		 */
		public void mousePressed(MouseEvent e)
		{
			int		  x = e.getX();
			int		  y = e.getY();
			Dimension dim = getSize();

			if ((x > (_leftHeaderWidth - SPLITTER_THICKNESS)) && (x < _leftHeaderWidth))
			{
				_movingLeftSplitter = true;
				_movingLeftSplitterX = _leftHeaderWidth + 1;
				_comp.repaint();
				return;
			}

			if ((x
						> (_horizontalSplitterPos - (SPLITTER_THICKNESS / 2)
							- ((_horizontalSplitterRatio < 100) ? 0 : SPLITTER_THICKNESS)))
					&& (x < (_horizontalSplitterPos + (SPLITTER_THICKNESS / 2))))
			{
				if ((_horizontalSplitterRatio < 100) || (y > (dim.height - SCROLLBAR_THICKNESS)))
				{
					_movingHorizontalSplitter = true;

					_drawnPart = TOP_HEADER_LEFT_PART;
					int x2 = getCurrentPartSize(false).width + _leftHeaderWidth;

					_movingHorizontalSplitterX = x2 + 1;

					if (_horizontalSplitterRatio == 100)
						_movingHorizontalSplitterX -= SPLITTER_THICKNESS;
				}
			}

			if ((y > (_verticalSplitterPos - (SPLITTER_THICKNESS / 2)))
					&& (y
						< (_verticalSplitterPos + (SPLITTER_THICKNESS / 2)
							+ ((_verticalSplitterRatio > 0) ? 0 : SPLITTER_THICKNESS))))
			{
				if ((_verticalSplitterRatio > 0)
						|| ((x < (_leftHeaderWidth - SPLITTER_THICKNESS))
							&& (x > (_leftHeaderWidth
									- SPLITTER_THICKNESS - SCROLLBAR_THICKNESS))))
				{
					_movingVerticalSplitter = true;

					_drawnPart = LEFT_HEADER_TOP_PART;
					int y2 = getCurrentPartSize(false).height + _topHeaderHeight;

					_movingVerticalSplitterY = y2 + 1;
				}
			}

			if (_movingHorizontalSplitter || _movingVerticalSplitter)
			{
				_comp.repaint();
				return;
			}

			// Redirect the event to final canvas
			LySplittableCanvas canvas = getCurrentCanvas(e);
			MouseEvent		   e2 = transformMouseEvent(e);

			if (canvas != null)
			{
				MouseListener[] listeners = canvas.getMouseListeners();

				if ((listeners != null) && (listeners.length > 0))
				{
					int i, n = listeners.length;

					for (i = 0; i < n; i++)
						listeners[i].mousePressed(e2);

					redraw();
				}
			}

			setActivated(canvas);
		} // end method mousePressed

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
		 */
		public void mouseReleased(MouseEvent e)
		{
			if (_movingLeftSplitter)
			{
				_movingLeftSplitter = false;
				_leftHeaderWidth = _movingLeftSplitterX;

				_comp.repaint();
				return;
			}

			boolean movingSplitter = (_movingHorizontalSplitter || _movingVerticalSplitter);

			if (_movingHorizontalSplitter)
			{
				Dimension dim = getSize();

				_movingHorizontalSplitter = false;
				_horizontalSplitterRatio = ((_movingHorizontalSplitterX - _leftHeaderWidth) * 100)
						/ (dim.width - _leftHeaderWidth);

				if (_horizontalSplitterRatio <= 0)
					_horizontalSplitterRatio = 100;

				if (_horizontalSplitterRatio > 100)
					_horizontalSplitterRatio = 100;
			}

			if (_movingVerticalSplitter)
			{
				Dimension dim = getSize();

				_movingVerticalSplitter = false;
				_verticalSplitterRatio = ((_movingVerticalSplitterY - _topHeaderHeight) * 100)
						/ (dim.height - _topHeaderHeight);

				if (_verticalSplitterRatio < 0)
					_verticalSplitterRatio = 0;

				if (_verticalSplitterRatio >= 100)
					_verticalSplitterRatio = 0;
			}

			if (movingSplitter)
			{
				_comp.repaint();
				return;
			}

			// Redirect the event to final canvas
			LySplittableCanvas canvas = getCurrentCanvas(e);
			MouseEvent		   e2 = transformMouseEvent(e);

			if (canvas != null)
			{
				MouseListener[] listeners = canvas.getMouseListeners();

				if ((listeners != null) && (listeners.length > 0))
				{
					int i, n = listeners.length;

					for (i = 0; i < n; i++)
						listeners[i].mouseReleased(e2);

					redraw();
				}
			}
		} // end method mouseReleased
	} // end class _LyCanvasSplitterMouseListener


	/**
	 * This class handles mouse events on this component scroll bar.
	 */
	class _LyCanvasSplitterScrollbarMouseListener
		implements MouseListener
	{

		/**
		 * Sets the cursor to the default cursor.
		 *
		 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
		 */
		public void mouseEntered(MouseEvent e)
		{
			_comp.setCursor(_defaultCursor);
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
		 */
		public void mouseExited(MouseEvent e)
		{
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
		 */
		public void mouseClicked(MouseEvent e)
		{
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
		 */
		public void mousePressed(MouseEvent e)
		{
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
		 */
		public void mouseReleased(MouseEvent e)
		{
		}
	} // end class _LyCanvasSplitterScrollbarMouseListener


	/**
	 * This class handles selection events on this component.
	 */
	class _LyCanvasSplitterSelectionListener
		implements AdjustmentListener
	{

		/**
		 * Redraws the component.
		 *
		 * @see java.awt.event.AdjustmentListener#adjustmentValueChanged(java.awt.event.AdjustmentEvent)
		 */
		public void adjustmentValueChanged(AdjustmentEvent event)
		{
			_comp.repaint();
		}
	}


	/**
	 * This class handles focus events on this component.
	 */
	class _LyCanvasSplitterFocusListener
		implements FocusListener
	{

		/**
		 * Default constructor.
		 */
		_LyCanvasSplitterFocusListener()
		{
		}

		/**
		 * Activates the default canvas.
		 *
		 * @see java.awt.event.FocusListener#focusGained(java.awt.event.FocusEvent)
		 */
		public void focusGained(FocusEvent e)
		{
			setActivated(_activatedCanvas);
		}

		/**
		 * Activates the default canvas.
		 *
		 * @see java.awt.event.FocusListener#focusLost(java.awt.event.FocusEvent)
		 */
		public void focusLost(FocusEvent e)
		{
			setActivated(_activatedCanvas);
		}
	} // end class _LyCanvasSplitterFocusListener


	/**
	 * This class handles key events on this component.
	 */
	class _LyCanvasSplitterKeyListener
		implements KeyListener
	{

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
		 */
		public void keyPressed(KeyEvent e)
		{
			if (_activatedCanvas != null)
			{
				KeyListener[] listeners = _activatedCanvas.getKeyListeners();

				if ((listeners != null) && (listeners.length > 0))
				{
					int i, n = listeners.length;

					for (i = 0; i < n; i++)
						listeners[i].keyPressed(e);

					redraw();
				}
			}
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
		 */
		public void keyTyped(KeyEvent e)
		{
			if (_activatedCanvas != null)
			{
				KeyListener[] listeners = _activatedCanvas.getKeyListeners();

				if ((listeners != null) && (listeners.length > 0))
				{
					int i, n = listeners.length;

					for (i = 0; i < n; i++)
						listeners[i].keyTyped(e);

					redraw();
				}
			}
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
		 */
		public void keyReleased(KeyEvent e)
		{
			if (_activatedCanvas != null)
			{
				KeyListener[] listeners = _activatedCanvas.getKeyListeners();

				if ((listeners != null) && (listeners.length > 0))
				{
					int i, n = listeners.length;

					for (i = 0; i < n; i++)
						listeners[i].keyReleased(e);

					redraw();
				}
			}
		}
	} // end class _LyCanvasSplitterKeyListener
} // end class LyAbstractCanvasSplitterManager
