/*
 * LyAbstractMessageIconManager.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * This class manages a message icon component. This component is composed of a text message and an
 * icon image. An action is triggered when this component is double clicked.
 *
 * @version $Id: LyAbstractMessageIconManager.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public abstract class LyAbstractMessageIconManager
	implements MouseListener
{
	public static final String __VERSION =
		"$Id: LyAbstractMessageIconManager.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Height of the icon image. */
	public final int		   IMAGE_HEIGHT = 18;

	/** Size, in pixels, of the empty border around this message component. */
	public final int		   MESSAGE_GAP = 2;

	/** The current environment. */
	protected LyEnvironment    _environment;

	/** The image of the associated icon. */
	protected Image			   _image;

	/** The text of the message. */
	protected String		   _text;

	/**
	 * The alignment policy. One of the following:
	 *
	 * <ul>
	 *   <li>{@link leon.view.LyMessage#Align.LEFT}</li>
	 *   <li>{@link leon.view.LyMessage#Align.RIGHT}</li>
	 *   <li>{@link leon.view.LyMessage#Align.CENTER}</li>
	 * </ul>
	 */
	protected LyMessage.Align  _align;

	/** The font of this message. */
	protected Font			   _font;

	/** The font metrics associated with the font of this component. */
	protected FontMetrics	   _fontMetrics;

	/** The icon message graphical component. */
	protected Component		   _comp;

	/** The action listeners of the message icon component. */
	protected LyActionListener _listener;

	/**
	 * Creates and initializes a new message icon manager. Sets its attributes to the default
	 * values.
	 *
	 * @param environment the current environment
	 * @param text        the text of the message component
	 * @param image       the image associated to the icon
	 * @param align       the alignment policy. One of the following:
	 *
	 *                    <ul>
	 *                      <li>LyMessage.Align.LEFT</li>
	 *                      <li>LyMessage.Align.RIGHT</li>
	 *                      <li>LyMessage.Align.CENTER</li>
	 *                    </ul>
	 * @param comp        the associated component for this manager
	 * @param listener    the action listener used when double click event is sent.
	 */
	public LyAbstractMessageIconManager(LyEnvironment environment, String text, String image,
        LyMessage.Align align, Component comp, LyActionListener listener)
	{
		_environment = environment;
		_comp = comp;
		_listener = listener;

		setText(text);
		setImage(image);
		setAlign(align);

		_font = (Font)environment.getFont("LY_MESSAGE_FONT");

		Color col = (Color)environment.getColor("LY_MESSAGES_BACKGROUND");

		if (col == null)
			col = (Color)environment.getColor("LY_DEFAULT_BACKGROUND");

		if (col != null)
			_comp.setBackground(col);

		_comp.addMouseListener(this);
	}

	/**
	 * Sets the text of the message component.
	 *
	 * @param text the new text label
	 */
	public void setText(String text)
	{
		String label = _environment.translate(text);

		if (((_text != null) && (!_text.equals(label))) || ((_text == null) && (label != null)))
			_text = label;
	}

	/**
	 * Sets the image icon of the message component.
	 *
	 * @param image the new icon image
	 */
	public void setImage(String image)
	{
		if ((image == null) || (image.length() == 0))
		{
			_image = null;
			return;
		}

		LyString     string = _environment.getString();
		StringBuffer sizeCode = new StringBuffer();

		sizeCode.append(string.VISIBLE_SEP3);
		sizeCode.append(IMAGE_HEIGHT - (2 * MESSAGE_GAP));
		sizeCode.append(string.VISIBLE_SEP2);
		sizeCode.append(IMAGE_HEIGHT - (2 * MESSAGE_GAP));
		sizeCode.append(string.VISIBLE_SEP3_BIS);

		Image img = (Image)_environment.getImage(image, sizeCode.toString());

		if (img != _image)
			_image = img;
	}

	/**
	 * Sets the alignment policy of the message text.
	 *
	 * @param align the new alignemnt policy
	 * @see   #_align
	 */
	public void setAlign(LyMessage.Align align)
	{
		_align = align;
	}

	/**
	 * Gets the component preferred size.
	 *
	 * @return the size of the message icon component depending on the image dimensions and its font
	 *         metrics
	 */
	public Dimension getPreferredSize()
	{
		int w = MESSAGE_GAP, h = 0;

		if (_image != null)
		{
			w = _image.getWidth(null);
			h = _image.getHeight(null) + (2 * MESSAGE_GAP);
		}

		w += MESSAGE_GAP;

		if (_fontMetrics == null)
			_fontMetrics = _comp.getFontMetrics(_font);

		if (_text != null)
			w = w + _fontMetrics.stringWidth(_text) + MESSAGE_GAP;

		h = Math.max(h, _fontMetrics.getHeight() + (2 * MESSAGE_GAP));

		return new Dimension(w, h);
	}

	/**
	 * Does nothing.
	 *
	 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
	 */
	public void mousePressed(MouseEvent e)
	{
	}

	/**
	 * Generates an action event for the associated listener when this icon component is double
	 * clicked.
	 *
	 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
	 */
	public void mouseClicked(MouseEvent e)
	{
		_listener.actionPerformed(new LyActionEvent(_comp));
	}

	/**
	 * Does nothing.
	 *
	 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
	 */
	public void mouseExited(MouseEvent e)
	{
	}

	/**
	 * Does nothing.
	 *
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	public void mouseReleased(MouseEvent e)
	{
	}

	/**
	 * Does nothing.
	 *
	 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
	 */
	public void mouseEntered(MouseEvent e)
	{
	}
} // end class LyAbstractMessageIconManager
