/*
 * LyAbstractTitlePanelManager.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.image.*;
import leon.view.event.*;
import leon.view.win.swing.*;


/**
 * This class manages title panels.
 *
 * @version $Id: LyAbstractTitlePanelManager.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public abstract class LyAbstractTitlePanelManager
{
	public static final String __VERSION =
		"$Id: LyAbstractTitlePanelManager.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Horizontal space between components. */
	public static final int    X_GAP = 5;

	/** Vertical space between components. */
	public static final int    Y_GAP = 1;

	/** Fading used for the background color (use 0 for no fading). */
	public static final int    FADING = 50;

	/** Left label of the titlebar. */
	protected String		   _leftLabel;

	/** Center label of the titlebar. */
	protected String		   _centerLabel;

	/** Right label of the titlebar. */
	protected String		   _rightLabel;

	/** Indicates if the titlebar must be drawn with a fading effet. */
	protected boolean		   _fading;

	/** Rectangle of the right label of the titlebar. */
	protected Rectangle		   _rightLabelRect;

	/** The background image. */
	protected Image			   _bgImage;

	/** The main image. */
	protected Image			   _image;

	/** The main image tooltip. */
	protected String		   _imageToolTip;

	/** The main image rectangle. */
	protected Rectangle		   _imageRect;

	/** The right image. */
	protected Image			   _rightImage;

	/** The right image tooltip. */
	protected String		   _rightImageToolTip;

	/** The right image rectangle. */
	protected Rectangle		   _rightImageRect;

	/** Font for title text. */
	protected Font			   _titleFont;

	/** Color behind title text. */
	protected Color			   _titleBgColor;

	/** Color of title text. */
	protected Color			   _titleColor;

	/** Color behind title image. */
	protected Color			   _imageBgColor;

	/** Color around title image. */
	protected Color			   _imageBorderColor;

	/** Cached font metrics to enhance speed. */
	protected FontMetrics	   _fm;

	/** The title graphical component. */
	protected Component		   _comp;

	/** The action listeners of the title bar component. */
	protected LyActionListener _listener;

	/**
	 * Default constructor.
	 */
	public LyAbstractTitlePanelManager()
	{
		this(null, null, null, null, null, null, null, null, null, null);
	}

	/**
	 * Class constructor of a LyAbstractTitlePanelManager object. Sets the class fields to the
	 * specified values, then initializes those that remained <code>null</code>.
	 *
	 * @param bgImage           the background image
	 * @param image             the main image of the titlebar
	 * @param imageToolTip      the main image tooltip
	 * @param rightImage        the right image of the titlebar
	 * @param rightImageToolTip the right image tooltip
	 * @param leftLabel         the left label on the title
	 * @param centerLabel       the centered label
	 * @param rightLabel        the right label
	 * @param comp              the associated graphical component
	 * @param listener          the action listener
	 */
	public LyAbstractTitlePanelManager(Image bgImage, Image image, String imageToolTip,
		Image rightImage,
		String rightImageToolTip, String leftLabel, String centerLabel, String rightLabel,
		Component comp, LyActionListener listener)
	{
		_bgImage = bgImage;
		_image = image;
		_imageToolTip = imageToolTip;
		_imageRect = new Rectangle();
		_rightImage = rightImage;
		_rightImageToolTip = rightImageToolTip;
		_rightImageRect = new Rectangle();
		_rightLabelRect = new Rectangle();
		_leftLabel = leftLabel;
		_centerLabel = centerLabel;
		_rightLabel = rightLabel;
		_fading = false;
		_listener = listener;
		_comp = comp;

		init();
	}

	/**
	 * Initializes the <code>null</code> title bar labels.
	 */
	protected void init()
	{
		if (_leftLabel == null)
			_leftLabel = "";

		if (_centerLabel == null)
			_centerLabel = "";

		if (_rightLabel == null)
			_rightLabel = "";
	}

	/**
	 * Returns the offscreen image if any.
	 *
	 * @param  offScreenSize size of the offscreen image
	 * @return the offscreen image
	 */
	public abstract Image getImageOffScreen(Dimension offScreenSize);

	/**
	 * Returns the graphics context of the specified offscreen image or the default graphics context
	 * if none is available.
	 *
	 * @param  offScreenImage  the image to get its graphics context
	 * @param  defaultGraphics the default graphics context if the image doesn't have one
	 * @return a graphics context to use
	 */
	public abstract Graphics getTargetGraphics(Image offScreenImage, Graphics defaultGraphics);

	/**
	 * Paints the specified offscreen image to the onscreen graphics context.
	 *
	 * @param offScreenImage   the offscreen image to paint on the onscreen context
	 * @param onScreenGraphics the graphics context to paint the offscreen image on
	 */
	public abstract void paintOnScreen(Image offScreenImage, Graphics onScreenGraphics);

	/**
	 * Draws the title panel in this graphics context.
	 *
	 * @param g the graphics context to paint the images on
	 */
	public void paintCommon(Graphics g)
	{
		// prepare offscreen
		Dimension d1 = _comp.getSize();

		if ((d1.width <= 0) || (d1.height <= 0))
			return;
		
		int		 x_pos, last_x_pos = 0, y_pos, pref_x_pos;
		int		 X_offset = X_GAP; // Leave X_GAP to the left
		int		 X_image_offset = 0;
		int		 X_rightImage_offset = 0;

		String   label;
		Image    offscreen = getImageOffScreen(d1);
		Graphics targetGraphics = getTargetGraphics(offscreen, g);

		targetGraphics.setColor(_titleBgColor);
		targetGraphics.fillRect(0, 0, d1.width, d1.height);

		if (_bgImage != null)
			LySwingUtils.fillBackground(g, d1.width, d1.height, _bgImage);
		// Draw the background with a fading effect
		else if (_fading)
		{
			int red = _titleBgColor.getRed();
			int green = _titleBgColor.getGreen();
			int blue = _titleBgColor.getBlue();
			int w, offset = 1;

			for (w = 0; w < d1.width; w += 10)
			{
				offset = ((FADING * w) / d1.width);

				targetGraphics.setColor(new Color(((red > (255 - offset)) ? 255 : (red + offset)),
						((green > (255 - offset)) ? 255 : (green + offset)),
						((blue > (255 - offset)) ? 255 : (blue + offset))));
				targetGraphics.fillOval(w, ((w * d1.height) / d1.width) - d1.height,
					2 * (d1.width - w), 2 * (d1.height - ((w * d1.height) / d1.width) + d1.height));
			}
		}

		// Draw image
		if (_image != null)
		{
			Dimension d_im = new Dimension();

			d_im.width = _image.getWidth(null);
			d_im.height = _image.getHeight(null);

			X_image_offset += X_GAP;

			if (_imageBgColor != null)
			{
				targetGraphics.setColor(_imageBgColor);
				targetGraphics.fillRect(X_offset, (d1.height - d_im.height) / 2, d_im.width,
					d_im.height);
			}

			if (_imageBorderColor != null)
			{
				targetGraphics.setColor(_imageBorderColor);
				targetGraphics.drawRect(X_offset - 1, ((d1.height - d_im.height) / 2) - 1,
					d_im.width + 2, d_im.height + 2);
			}

			targetGraphics.drawImage(_image, X_offset,
				((d1.height - d_im.height) / 2)
				+ getImageVerticalOffset(), new ImageObserver()
				{

					/*
					 * (non-Javadoc)
					 * @see java.awt.image.ImageObserver#imageUpdate(java.awt.Image, int, int, int, int,
					 * int)
					 */
					public boolean imageUpdate(Image img, int infoflags, int x, int y, int width,
						int height)
					{
						if (infoflags == ImageObserver.ALLBITS)
							_comp.repaint();
						return true;
					}
				});

			/* Rectangle used to manage ToolTip */
			_imageRect.setBounds(X_offset, (d1.height - d_im.height) / 2, d_im.width, d_im.height);

			X_image_offset += d_im.width;
		}

		if (_comp instanceof Container)
		{
			Component[]		components = ((Container)_comp).getComponents();
			Dimension		dim;
			int				i, n = (components == null ? 0 : components.length);
			
			for (i = 0; i < n; i++)
			{
				dim = components[i].getPreferredSize();
				d1.width -= dim.width;
			}
			
			if (n > 0)
				d1.width -= X_GAP;
		}

		if (_rightImage != null)
		{
			Dimension d_im = new Dimension();

			d_im.width = _rightImage.getWidth(null);
			d_im.height = _rightImage.getHeight(null);

			X_rightImage_offset += X_GAP;

			if (_imageBgColor != null)
			{
				targetGraphics.setColor(_imageBgColor);
				targetGraphics.fillRect(d1.width - X_GAP - d_im.width,
					(d1.height - d_im.height) / 2, d_im.width, d_im.height);
			}

			if (_imageBorderColor != null)
			{
				targetGraphics.setColor(_imageBorderColor);
				targetGraphics.drawRect(d1.width - X_GAP - d_im.width - 1,
					((d1.height - d_im.height) / 2) - 1, d_im.width + 2, d_im.height + 2);
			}

			targetGraphics.drawImage(_rightImage, d1.width - X_GAP - d_im.width,
				((d1.height - d_im.height) / 2) + getImageVerticalOffset(), null);

			/* Rectangle used to manage ToolTip */
			_rightImageRect.setBounds(d1.width - X_GAP - d_im.width, (d1.height - d_im.height) / 2,
				d_im.width, d_im.height);

			X_rightImage_offset += d_im.width;
		}

		// draw text
		X_offset += X_GAP;

		// Last drawable position
		x_pos = X_offset + X_image_offset;
		last_x_pos = d1.width - X_GAP - X_rightImage_offset;

		if (_fm == null)
			_fm = _comp.getFontMetrics(_titleFont);

		y_pos = ((d1.height - 1 - _fm.getHeight()) / 2) + _fm.getAscent();

		targetGraphics.setColor(_titleColor);
		targetGraphics.setFont(_titleFont);

		if ((_leftLabel != null) && (_leftLabel.length() > 0))
		{
			// Calculate preferred x pos
			// May substring the title label
			label = getDrawableString(_leftLabel, x_pos, last_x_pos);
			// Draw
			targetGraphics.drawString(label, x_pos, y_pos);
			x_pos += X_offset + _fm.stringWidth(label);
		}
		if ((_centerLabel != null) && (_centerLabel.length() > 0))
		{
			pref_x_pos = (d1.width - (2 * X_offset) - _fm.stringWidth(_centerLabel)) / 2;
			// Should not cover image
			x_pos = Math.max(x_pos, pref_x_pos);
			// May substring the title label
			label = getDrawableString(_centerLabel, x_pos, last_x_pos);
			// Draw
			targetGraphics.drawString(label, x_pos, y_pos);
			x_pos += X_offset + _fm.stringWidth(label);
		}
		if ((_rightLabel != null) && (_rightLabel.length() > 0))
		{
			// Calculate preferred x pos
			pref_x_pos = last_x_pos - _fm.stringWidth(_rightLabel);
			// Should not cover image
			x_pos = Math.max(x_pos, pref_x_pos);
			// May substring the title label
			label = getDrawableString(_rightLabel, x_pos, last_x_pos);
			// Draw
			targetGraphics.drawString(label, x_pos, y_pos);

			_rightLabelRect.setBounds(x_pos, 0, _fm.stringWidth(_rightLabel), _fm.getHeight());
		}

		paintOnScreen(offscreen, g);
	} // end method paintCommon

	/**
	 * Retrieves the drawable part of the specified string between the starting and ending
	 * abscisses. If the string is too large, "..." is added at its end.
	 *
	 * @param  label  the label
	 * @param  xStart start drawing abscisse
	 * @param  xEnd   end drawing abscisse
	 * @return the label that is drawn
	 */
	public String getDrawableString(String label, int xStart, int xEnd)
	{
		if ((label == null) || (xStart >= xEnd))
			return "";

		// cut off chars at the end to make the string fit..
		int    w, width = xEnd - xStart;
		int    l = label.length();
		String sub_str = label;

		w = _fm.stringWidth(label);

		// now make some minor adjustments...
		if (w > width)
		{
			while (w > width)
			{
				l--;
				sub_str = label.substring(0, l);
				w = _fm.stringWidth(sub_str);
			}
		}

		// Add "..." if substring
		if (sub_str.length() < label.length())
		{
			StringBuffer draw_str;
			int			 w_points;
			String		 points;

			points = new String("...");
			w_points = _fm.stringWidth(points);
			w = _fm.stringWidth(sub_str);
			l = sub_str.length();

			while ((l > 0) && ((w + w_points) > width))
			{
				l--;
				sub_str = label.substring(0, l);
				w = _fm.stringWidth(sub_str);
			}
			if (l > 0)
			{
				draw_str = new StringBuffer(sub_str + points);
				sub_str = draw_str.toString();
				return sub_str;
			}
			return "";
		}

		return label;
	} // end method getDrawableString

	/**
	 * Computes minimum size depending on content.
	 *
	 * @return the preferred size of the component
	 * @see    #getPreferredSize()
	 */
	public Dimension getMinimumSize()
	{
		return new Dimension(20, getPreferredSize().height);
	}

	/**
	 * Computes preferred size depending on content.
	 *
	 * @return the preferred size of the component
	 */
	public Dimension getPreferredSize()
	{
		int w = 0, h = 0;

		if (_image != null)
		{
			w = _image.getWidth(null) + X_GAP;
			h = _image.getHeight(null) + (2 * Y_GAP);
		}

		if (_rightImage != null)
		{
			w += _rightImage.getWidth(null) + (2 * X_GAP);
			h = Math.max(h, _rightImage.getHeight(null) + (2 * Y_GAP));
		}

		FontMetrics fm = _comp.getFontMetrics(_titleFont);

		if (_leftLabel != null)
			w = w + fm.stringWidth(_leftLabel) + X_GAP;

		if (_centerLabel != null)
			w = w + fm.stringWidth(_centerLabel) + X_GAP;

		if (_rightLabel != null)
			w = w + fm.stringWidth(_rightLabel) + X_GAP;

		h = Math.max(h, fm.getHeight() + (2 * Y_GAP));

		return (new Dimension(w, h));
	} // end method getPreferredSize

	/**
	 * Gets the current title bar component.
	 *
	 * @return the title bar component
	 */
	public Component getComponent()
	{
		return _comp;
	}

	/**
	 * Sets the left label parameter.
	 *
	 * @param leftLabel new label
	 */
	public void setLeftLabel(String leftLabel)
	{
		_leftLabel = leftLabel;
	}

	/**
	 * Sets the right label parameter.
	 *
	 * @param rightLabel new label
	 */
	public void setRightLabel(String rightLabel)
	{
		_rightLabel = rightLabel;
	}

	/**
	 * Sets the center label parameter.
	 *
	 * @param centerLabel new label
	 */
	public void setCenterLabel(String centerLabel)
	{
		_centerLabel = centerLabel;
	}

	/**
	 * Sets the main image.
	 *
	 * @param image new image
	 */
	public void setImage(Image image)
	{
		_image = image;
	}

	/**
	 * Sets the image tooltip.
	 *
	 * @param toolTip new image tooltip
	 * @see   #getImageToolTip()
	 */
	public void setImageToolTip(String toolTip)
	{
		_imageToolTip = toolTip;
	}

	/**
	 * Sets the right image.
	 *
	 * @param image new image
	 */
	public void setRightImage(Image image)
	{
		_rightImage = image;
	}

	/**
	 * Sets the right image tooltip.
	 *
	 * @param toolTip new right image tooltip
	 * @see   #getRightImageToolTip()
	 */
	public void setRightImageToolTip(String toolTip)
	{
		_rightImageToolTip = toolTip;
	}

	/**
	 * Sets the titlebar font.
	 *
	 * @param font new titleBar font
	 */
	public void setTitleFont(Font font)
	{
		_titleFont = font;
	}

	/**
	 * Sets the titlebar background color.
	 *
	 * @param color the new titlebar background color
	 */
	public void setTitleBackgroundColor(Color color)
	{
		_titleBgColor = color;
	}

	/**
	 * Sets the titlebar main image background color.
	 *
	 * @param color the new image background color
	 */
	public void setImageBackgroundColor(Color color)
	{
		_imageBgColor = color;
	}

	/**
	 * Sets the titlebar main image border color.
	 *
	 * @param color the new image border color
	 */
	public void setImageBorderColor(Color color)
	{
		_imageBorderColor = color;
	}

	/**
	 * Sets the titlebar text color.
	 *
	 * @param color the new titlebar text color
	 */
	public void setTitleColor(Color color)
	{
		_titleColor = color;
	}

	/**
	 * Sets the fading property of the titlebar.
	 *
	 * @param fading Indicates if the titlebar must be drawn with a fading effect.
	 */
	public void setFading(boolean fading)
	{
		_fading = fading;
	}

	/**
	 * Gets the main image rectangle.
	 *
	 * @return the main image rectangle
	 */
	public Rectangle getImageRectangle()
	{
		return _imageRect;
	}

	/**
	 * Gets the right image rectangle.
	 *
	 * @return the right image rectangle
	 */
	public Rectangle getRightImageRectangle()
	{
		return _rightImageRect;
	}

	/**
	 * Gets the right label rectangle.
	 *
	 * @return the right label rectangle
	 */
	public Rectangle getRightLabelRectangle()
	{
		return _rightLabelRect;
	}

	/**
	 * Gets the main image tooltip.
	 *
	 * @return the main image tooltip
	 * @see    #setImageToolTip(String)
	 */
	public String getImageToolTip()
	{
		return _imageToolTip;
	}

	/**
	 * Gets the right image tooltip.
	 *
	 * @return the right image tooltip
	 * @see    #setRightImageToolTip(String)
	 */
	public String getRightImageToolTip()
	{
		return _rightImageToolTip;
	}

	/**
	 * Gets the vertical offset of the image.
	 *
	 * @return 0, for no offset
	 */
	protected int getImageVerticalOffset()
	{
		return 0;
	}
} // end class LyAbstractTitlePanelManager
