/*
 * LyFlowLayout.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.util.*;

import leon.view.*;


/**
 * This class implements a flow layout with a maximum width.
 */
public class LyFlowLayout extends FlowLayout
{
	private static final long  serialVersionUID = -6938054113542181882L;

	public static final String __VERSION =
		"$Id: LyFlowLayout.java 31021 2011-05-16 12:24:11Z nhamon $";

	/* Protected Fields */

	/**
	 * Alignment of the layout. It may have the following values:
	 *
	 * <ul>
	 *   <li>{@link FlowLayout#LEFT}</li>
	 *   <li>{@link FlowLayout#RIGHT}</li>
	 *   <li>{@link FlowLayout#CENTER}</li>
	 *   <li>{@link FlowLayout#LEADING}</li>
	 *   <li>{@link FlowLayout#TRAILING}</li>
	 * </ul>
	 */
	int						   _align;

	/** Horizontal gap, in pixels, of the layout. */
	int						   _hgap;

	/** Vertical gap, in pixels, of the layout. */
	int						   _vgap;

	/** Maximum width of the layout. */
	public int				   _maxWidth;

	/** Indicates if this layout manages an automatic wrapping. */
	public boolean			   _automaticWrap;
	
	/** Current view manager. */
	protected LyViewManager	   _viewManager;

	/**
	 * Construct a class behavior.
	 *
	 * <p>By default, this behaviour is centered, the horizontal and vertical gaps are both set to 5
	 * pixels, the maximum width is set to automatic and there is no automatic wrapping.</p>
	 */
	public LyFlowLayout()
	{
		this(LEFT);
	}

	/**
	 * Construct a class behavior.
	 *
	 * <p>By default, the horizontal and vertical gaps are both set to 5 pixels, the maximum width
	 * is set to automatic and there is no automatic wrapping.</p>
	 *
	 * @param align the alignment policy
	 */
	public LyFlowLayout(int align)
	{
		this(align, 5, 5);
	}

	/**
	 * Construct a class behavior.
	 *
	 * <p>By default, the maximum width is set to automatic and there is no automatic wrapping.</p>
	 *
	 * @param align the alignment policy
	 * @param hgap  the horizontal gap
	 * @param vgap  the vertical gap
	 */
	public LyFlowLayout(int align, int hgap, int vgap)
	{
		this(align, hgap, vgap, -1);
	}

	/**
	 * Construct a class behavior.
	 *
	 * <p>By default, there is no automatic wrapping.</p>
	 *
	 * @param align    the alignment policy
	 * @param hgap     the horizontal gap
	 * @param vgap     the vertical gap
	 * @param maxWidth the maximum width of the layout
	 */
	public LyFlowLayout(int align, int hgap, int vgap, int maxWidth)
	{
		this(align, hgap, vgap, maxWidth, false);
	}

	/**
	 * Construct a class behavior.
	 *
	 * @param align         the alignment policy
	 * @param hgap          the horizontal gap
	 * @param vgap          the vertical gap
	 * @param maxWidth      the maximum width of the layout
	 * @param automaticWrap Indicates if the layout does automatic wrapping.
	 */
	public LyFlowLayout(int align, int hgap, int vgap, int maxWidth, boolean automaticWrap)
	{
		this(align, hgap, vgap, maxWidth, false, null);
	}
	
	/**
	 * Construct a class behavior.
	 *
	 * @param align         the alignment policy
	 * @param hgap          the horizontal gap
	 * @param vgap          the vertical gap
	 * @param maxWidth      the maximum width of the layout
	 * @param automaticWrap Indicates if the layout does automatic wrapping.
	 * @param viewManager   the viewManager instance
	 */
	public LyFlowLayout(int align, int hgap, int vgap, int maxWidth, boolean automaticWrap, LyViewManager viewManager)
	{
		_align = align;
		_hgap = hgap;
		_vgap = vgap;
		_maxWidth = maxWidth;
		_automaticWrap = automaticWrap;
		_viewManager = viewManager;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.LayoutManager#preferredLayoutSize(java.awt.Container)
	 */
	public Dimension preferredLayoutSize(Container target)
	{
		synchronized (target.getTreeLock())
		{
			Dimension dim = new Dimension(0, 0); // current row
			int		  nmembers = target.getComponentCount();
			boolean   firstVisibleComponent = true;
			Insets    insets = target.getInsets();
			int		  totalWidth = 0, totalHeight = 0;
			boolean   jump = false;

			dim.width = insets.left + insets.right + (2 * _hgap);

			for (int i = 0; i < nmembers; i++)
			{
				Component m = target.getComponent(i);

				if (m.isVisible())
				{
					Dimension d = m.getPreferredSize();
					boolean   wasFirstVisibleComponent = firstVisibleComponent;

					dim.height = Math.max(dim.height, d.height);

					if (firstVisibleComponent)
						firstVisibleComponent = false;
					else
						dim.width += _hgap;

					dim.width += d.width;

					if (!wasFirstVisibleComponent && (_maxWidth > 0) && (dim.width >= _maxWidth))
					{
						jump = true;
						dim.width -= d.width;

						totalWidth = Math.max(totalWidth, dim.width);
						totalHeight = totalHeight + dim.height + _vgap;

						dim.width = d.width + insets.left + insets.right + (2 * _hgap);
						dim.height = d.height;
					}
				}
			}

			if (jump)
			{
				totalWidth = Math.max(totalWidth, dim.width);
				totalHeight += dim.height;
			}

			if ((totalWidth == 0) || (totalHeight == 0))
			{
				totalWidth = dim.width;
				totalHeight = dim.height;
			}

			totalHeight += insets.top + insets.bottom + (_vgap * 2);
			
			return new Dimension(totalWidth, totalHeight);
		}
	} // end method preferredLayoutSize

	/**
	 * The minimum size of the layout is the same as the preferred size.
	 *
	 * @see #preferredLayoutSize(Container)
	 * @see java.awt.LayoutManager#minimumLayoutSize(java.awt.Container)
	 */
	public Dimension minimumLayoutSize(Container target)
	{
		return preferredLayoutSize(target);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.LayoutManager#layoutContainer(java.awt.Container)
	 */
	public void layoutContainer(Container target)
	{
		synchronized (target.getTreeLock())
		{
			Insets  insets = target.getInsets();
			int     maxWidth = target.getSize().width - (insets.left + insets.right + (_hgap * 2));
			int     nmembers = target.getComponentCount();
			int     x = 0, y = insets.top + _vgap;
			int     rowh = 0, start = 0;

			boolean ltr = target.getComponentOrientation().isLeftToRight();

			for (int i = 0; i < nmembers; i++)
			{
				Component m = target.getComponent(i);

				if (m.isVisible())
				{
					Dimension d = m.getPreferredSize();

					m.setSize(d.width, d.height);

					if ((x == 0) || ((x + d.width) <= maxWidth))
					{
						if (x > 0)
							x += _hgap;

						x += d.width;
						rowh = Math.max(rowh, d.height);
					}
					else
					{
						moveComponents(target, insets.left + _hgap, y, maxWidth - x, rowh, start, i,
							ltr);
						x = d.width;
						y += _vgap + rowh;
						rowh = d.height;
						start = i;
					}
				}
			}

			moveComponents(target, insets.left + _hgap, y, maxWidth - x, rowh, start, nmembers,
				ltr);

			updateSize(target);
		}
	} // end method layoutContainer

	/**
	 * Updates the size of the specified target if automatic wrapping is enabled.
	 *
	 * @param target the target to update its size
	 */
	public void updateSize(Container target)
	{
		if (_automaticWrap)
		{
			int		  maxTmp = _maxWidth;
			Dimension prefDim = preferredLayoutSize(target);

			if (prefDim.width > target.getSize().width)
			{
				_maxWidth = target.getSize().width;

				target.setSize(preferredLayoutSize(target));

				if ((target.getParent() != null) && (target.getParent().isVisible()))
					target.getParent().doLayout();

				// Check if the height is ok
				if (prefDim.height < target.getSize().height)
				{
					Container		  c = target.getParent();
					Vector<Container> list = new Vector<Container>();

					while (c != null)
					{
						list.addElement(c);

						if (c.getLayout() instanceof BorderLayout)
							break;

						c = c.getParent();
					}

					for (int i = list.size() - 1; i >= 0; i--)
					{
						c = list.elementAt(i);
						c.doLayout();
					}
				}
				
				if (_viewManager != null)
				{
					Runnable	r = new _LyMaxWidthRunnable(maxTmp);
		            
					_viewManager.execGui(r);
				}
				else
					_maxWidth = maxTmp;
			}
		}
	} // end method updateSize

	class _LyMaxWidthRunnable implements Runnable
	{
		int	_width;
		
		_LyMaxWidthRunnable(int width)
		{
			_width = width;
		}
		
		public void run()
		{
			_maxWidth = _width;
		}
	}
	
	/**
	 * Centers the elements in the specified row, if there is any slack.
	 *
	 * @param target   the graphical component that containes the elements to center
	 * @param x        the new x-coordinate center value
	 * @param y        the new y-coordinate center value
	 * @param width    the width of a line in the row, this parameter is not used in this
	 *                 implementation
	 * @param height   the height of a line
	 * @param rowStart the starting row
	 * @param rowEnd   the ending row
	 * @param ltr      Indicates the reading orientation (<code>true</code> for left to right,
	 *                 <code>false</code> for right to left)
	 */
	private void moveComponents(Container target, int x, int y, int width, int height,
		int rowStart, int rowEnd, boolean ltr)
	{
		synchronized (target.getTreeLock())
		{
			switch (_align) {
			case LEFT:
			    x += ltr ? 0 : width;
			    break;
			case CENTER:
			    x += width / 2;
			    break;
			case RIGHT:
			    x += ltr ? width : 0;
			    break;
			case LEADING:
			    break;
			case TRAILING:
			    x += width;
			    break;
			}

			for (int i = rowStart; i < rowEnd; i++)
			{
				Component m = target.getComponent(i);

				if (m.isVisible())
				{
					if (ltr)
						m.setLocation(x, y + ((height - m.getSize().height) / 2));
					else
						m.setLocation(target.getSize().width - x - m.getSize().width,
							y
							+ ((height - m.getSize().height) / 2));

					x += m.getSize().width + _hgap;
				}
			}
		}
	}
} // end class LyFlowLayout
