/*
 * LyJAutoCompletionMultiSelector.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.dnd.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import leon.misc.*;
import leon.view.event.*;


/**
 * This class defines a component containing an editable combo with a set of buttons and a list. It
 * allows to choose some elements of the combo and to put them in the list.
 *
 * @version $Id: LyJAutoCompletionMultiSelector.java 36510 2012-08-17 15:43:06Z jlsaouli $
 * @author  Lyria
 */
public class LyJAutoCompletionMultiSelector extends JPanel
	implements LySelectionListener,
		LyActionListener,
		LyEventTransmitter
{
	private static final long		  serialVersionUID = -3181497342768462525L;

	public static final String		  __VERSION =
		"$Id: LyJAutoCompletionMultiSelector.java 36510 2012-08-17 15:43:06Z jlsaouli $";

	/** The combo containing selectable items. */
	protected LyJSelector			  _selector;

	protected LyJAutoCompletionViewer _autoCompletionViewer;

	/** The list containing selected items. */
	protected LyJListComponent		  _list;

	/** The panel that contains the action button. */
	protected JPanel				  _buttonPanel;

	/** The button that adds selected elements from the combo to the list. */
	protected LyJButton				  _buttonSelect;

	/** The button that removes selected elements from the list. */
	protected LyJButton				  _buttonUnselect;

	/** The button that moves elements from the right list to an upper position. */
	protected LyJButton				  _buttonUp;

	/** The button that moves elements from the right list to a lower position. */
	protected LyJButton				  _buttonDown;

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter     _actionTransmitter;

	/**
	 * Creates an auto-completion multi-selector object with the specified attributes.
	 *
	 * @param environment       the current environment
	 * @param height            the height of this component
	 * @param autoCompletionMax maximum number of values displayed in the auto-completion list
	 * @param ordered           indicates whether this selector is ordered or not
	 */
	public LyJAutoCompletionMultiSelector(LyEnvironment environment, int height,
		int autoCompletionMax, boolean ordered)
	{
		super();

		init(environment, height, autoCompletionMax, ordered);
	}

	/**
	 * Initializes this component. Sets the list components and the associated buttons.
	 *
	 * @param environment       the current environment
	 * @param height            the height of this component
	 * @param autoCompletionMax maximum number of values displayed in the auto-completion list
	 * @param ordered           indicates whether this selector is ordered or not
	 */
	protected void init(LyEnvironment environment, int height, int autoCompletionMax, boolean ordered)
	{
		// FEV 0058 : Try to get default sizes.
		int    defaultWidth, defaultHeight, minWidth, fixedWidth;

		String s;

		s = environment.getResource("LY_MULTIPLE_RELATION_WIDTH");
		defaultWidth = LyTools.intFromString(s);

		if (defaultWidth > 0)
			minWidth = defaultWidth;
		else
		{
			s = environment.getResource("LY_MULTIPLE_RELATION_MIN_WIDTH");
			minWidth = LyTools.intFromString(s);
		}

		if (height <= 0)
		{
			s = environment.getResource("LY_MULTIPLE_RELATION_HEIGHT");
			defaultHeight = LyTools.intFromString(s);
		}
		else
			defaultHeight = height;

		if (minWidth <= 0)
			minWidth = 20;

		if (defaultHeight <= 0)
			defaultHeight = 6;

		// Calculate estimation of size in pixels
		FontMetrics fm = getFontMetrics(getFont());

		fixedWidth = fm.charWidth('0') * minWidth;

		// Create the lists
		_selector = new LyJSelector(null, environment, true);

		LyJButton button = _selector.addButton((Image)environment.getImage("LY_ARROW_DOWN_BIG"),
				"");

		_autoCompletionViewer = new LyJAutoCompletionViewer(_selector.getTextField(), button, new String[0],
				new String[0], new String[0], true, autoCompletionMax, environment);
		_autoCompletionViewer.setActionListener(this);

		_list = new LyJListComponent();
		_list.setCellRenderer(new _LyJCellRenderer());
		_list.setFixedCellWidth(fixedWidth);
		_list.setVisibleRowCount(defaultHeight);
		_selector.addKeyListener(new _LyKeyListener());

		Color selectColor = (Color)environment.getColor("LY_TABLE_SELECT_COLOR");

		if (selectColor != null)
		{
			_list.setSelectionBackground(selectColor);

			// if selection background too dark, selection foreground is white
			int sumColor = selectColor.getRed() + selectColor.getGreen() + selectColor.getBlue();

			if (sumColor < 255)
				_list.setSelectionForeground(Color.white);
		}

		// COMMENTED in view package
		LyActionTransmitter actionTransmitter = _selector.getActionTransmitter();

		if (actionTransmitter != null)
			actionTransmitter.addActionListener(this);

		actionTransmitter = _list.getActionTransmitter();
		if (actionTransmitter != null)
			actionTransmitter.addActionListener(this);

		LySelectionTransmitter selectionTransmitter = _selector.getSelectionTransmitter();

		if (selectionTransmitter != null)
			selectionTransmitter.addSelectionListener(this);

		selectionTransmitter = _list.getSelectionTransmitter();

		if (selectionTransmitter != null)
			selectionTransmitter.addSelectionListener(this);

		// Create the buttons panel
		_buttonPanel = new JPanel();

		Image selectImage = (Image)environment.getImage("LY_MULTI_SELECTOR_SELECT");
		Image unselectImage = (Image)environment.getImage("LY_MULTI_SELECTOR_UNSELECT");
		Image upImage = (Image)environment.getImage("LY_TOOL_MOVE_UP");
		Image downImage = (Image)environment.getImage("LY_TOOL_MOVE_DOWN");

		int   imageWidth = selectImage.getWidth(button);
		int   imageHeight = selectImage.getHeight(button);

		_buttonSelect = new LyJButton(selectImage, null, environment.translate("LY_SELECT"),
				"select");
		_buttonUnselect = new LyJButton(unselectImage, null, environment.translate("LY_UNSELECT"),
				"unselect");
		
		if (ordered)
		{
			_buttonUp = new LyJButton(upImage, null, environment.translate("LY_TOOL_MOVE_UP"), "up");
			_buttonDown = new LyJButton(downImage, null, environment.translate("LY_TOOL_MOVE_UP"),
					"down");
		}
		
		if ((imageWidth > 0) && (imageHeight > 0))
		{
			_buttonSelect.setPreferredSize(new Dimension(imageWidth + 3, imageHeight + 3));
			_buttonUnselect.setPreferredSize(new Dimension(imageWidth + 3, imageHeight + 3));

			if (ordered)
			{
				_buttonUp.setPreferredSize(new Dimension(imageWidth + 3, imageHeight + 3));
				_buttonDown.setPreferredSize(new Dimension(imageWidth + 3, imageHeight + 3));
			}
		}

		_buttonSelect.setBorder(null);
		_buttonUnselect.setBorder(null);

		if (ordered)
		{
			_buttonUp.setBorder(null);
			_buttonDown.setBorder(null);
		}
		
		_buttonSelect.setBackground(null);
		_buttonUnselect.setBackground(null);

		if (ordered)
		{
			_buttonUp.setBackground(null);
			_buttonDown.setBackground(null);
		}
		
		_buttonSelect.setRolloverChangeBackground(true);
		_buttonUnselect.setRolloverChangeBackground(true);

		if (ordered)
		{
			_buttonUp.setRolloverChangeBackground(true);
			_buttonDown.setRolloverChangeBackground(true);
		}
		
		_buttonSelect.setContentAreaFilled(false);
		_buttonUnselect.setContentAreaFilled(false);

		if (ordered)
		{
			_buttonUp.setContentAreaFilled(false);
			_buttonDown.setContentAreaFilled(false);
		}
		
		LyActionTransmitter transmitter;

		transmitter = _buttonSelect.getActionTransmitter();
		if (transmitter != null)
			transmitter.addActionListener(this);

		transmitter = _buttonUnselect.getActionTransmitter();
		if (transmitter != null)
			transmitter.addActionListener(this);

		if (ordered)
		{
			transmitter = _buttonUp.getActionTransmitter();
			if (transmitter != null)
				transmitter.addActionListener(this);
	
			transmitter = _buttonDown.getActionTransmitter();
			if (transmitter != null)
				transmitter.addActionListener(this);
		}
		
		_buttonPanel.setLayout(new BoxLayout(_buttonPanel, BoxLayout.X_AXIS));
		_buttonPanel.setBackground(null);
		_buttonPanel.add(_buttonSelect);
		_buttonPanel.add(_buttonUnselect);

		if (ordered)
		{
			_buttonPanel.add(_buttonUp);
			_buttonPanel.add(_buttonDown);
		}
		
		_buttonPanel.setOpaque(false);
		setOpaque(false);
		setLayout(new BorderLayout(0, 2));

		JPanel northPanel = new JPanel(new BorderLayout(2, 0));

		northPanel.add(_selector, BorderLayout.CENTER);
		northPanel.add(_buttonPanel, BorderLayout.EAST);
		northPanel.setOpaque(false);
		add(northPanel, BorderLayout.NORTH);
		add(new _LyJSpecialScrollPane(_list), BorderLayout.CENTER);

		// Manage drag and drop on the right list
		_list.setDragEnabled(true);
		new DropTarget(_list, new _LyJAutoCompletionMultiSelectorDropListener());

		// Update the buttons
		checkButtons();
	} // end method init

	/**
	 * Returns the auto-completion viewer used by the combo.
	 *
	 * @return the auto-completion viewer instance.
	 */
	public LyJAutoCompletionViewer getAutoCompletionViewer()
	{
		return _autoCompletionViewer;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#getMinimumSize()
	 */
	public Dimension getMinimumSize()
	{
		return getPreferredSize();
	}

	/**
	 * Updates the buttons whether they are enabled or not.
	 */
	public void checkButtons()
	{
		if (isEnabled())
		{
			String selectedValue = _autoCompletionViewer.getSelectedValue();

			_buttonSelect.setEnabled(selectedValue != null);
			_buttonUnselect.setEnabled(_list.getSelectedCount() > 0);

			if (_buttonUp != null)
			{
				_buttonUp.setEnabled((_list.getSelectedCount() > 0)
						&& (_list.getSelectedIndex() > 0));
			}
			
			if (_buttonDown != null)
			{
				int[] selection = _list.getSelectedIndices();

				_buttonDown.setEnabled((_list.getSelectedCount() > 0)
						&& (selection != null)
						&& (selection[selection.length - 1] < (_list.getCount() - 1)));
			}
		}
		else
		{
			_buttonSelect.setEnabled(false);
			_buttonUnselect.setEnabled(false);
			if (_buttonUp != null)
				_buttonUp.setEnabled(false);
			if (_buttonDown != null)
				_buttonDown.setEnabled(false);
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LySelectionListener#selectionPerformed(leon.view.event.LySelectionEvent)
	 */
	public void selectionPerformed(LySelectionEvent selectionEvent)
	{
		checkButtons();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		Object  source = actionEvent.getSource();
		boolean valueChanged = false;

		if (source == _buttonSelect)
		{
			String selectedValue = _autoCompletionViewer.getSelectedValue();

			if (selectedValue != null)
			{
				Vector<String> values = new Vector<String>(1);

				values.addElement(selectedValue);
				selectValues(values);
				valueChanged = true;
			}
		}
		else if (source == _buttonUnselect)
		{
			int		 i, j;
			int[]    indices = _list.getSelectedIndices();
			String[] item;
			String[] displayedShortValues;
			String	 displayedShortValue;
			boolean  wasEmpty = (_autoCompletionViewer.getDisplayedValues().length == 0);

			for (i = indices.length - 1; i >= 0; i--)
			{
				displayedShortValues = _autoCompletionViewer.getDisplayedShortValues();
				item = (String[])_list.getItemAt(indices[i]);
				_list.removeItemAt(indices[i]);
				
				for (j = 0; j < displayedShortValues.length; j++)
				{
					displayedShortValue = displayedShortValues[j];
					
					if (displayedShortValue.compareTo(item[2]) >= 0)
						break;
				}
				
				_autoCompletionViewer.addValue(item[0], item[1], item[2], j);
				
				// If no value was selected in the combo, select the first one.
				if (wasEmpty)
					_autoCompletionViewer.setTextValue(0, false);

				valueChanged = true;
			}
		}
		else if ((source == _buttonUp) || (source == _buttonDown))
		{
			String[] item;
			int[]    selection = _list.getSelectedIndices();

			if (source == _buttonUp)
			{
				for (int i = 0; i < selection.length; i++)
				{
					item = ((String[])_list.getItemAt(selection[i]));
					_list.removeItemAt(selection[i]);
					_list.insertItemAt(item, selection[i] - 1);
					_list.selectItem(item, true);
				}
			}
			else
			{
				for (int i = selection.length - 1; i >= 0; i--)
				{
					item = ((String[])_list.getItemAt(selection[i]));
					_list.removeItemAt(selection[i]);
					_list.insertItemAt(item, selection[i] + 1);
					_list.selectItem(item, true);
				}
			}

			valueChanged = true;
		}

		if (valueChanged)
		{
			// notify
			if (_actionTransmitter != null)
			{
				LyActionEvent event = new LyActionEvent(this, null, getValue(),
						LyActionDemand.TYPE_VALUE_CHANGED);

				_actionTransmitter.processActionEvent(event);
			}
		}

		checkButtons();
	} // end method actionPerformed

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#setEnabled(boolean)
	 */
	public void setEnabled(boolean status)
	{
		super.setEnabled(status);
		_selector.setEnabled(status);

		checkButtons();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#setBackground(java.awt.Color)
	 */
	public void setBackground(Color color)
	{
		if (_selector != null)
			_selector.setBackground(color);

		if (_list != null)
			_list.setBackground(color);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#setFont(java.awt.Font)
	 */
	public void setFont(Font font)
	{
		super.setFont(font);

		if (_selector != null)
			_selector.setFont(font);

		if (_list != null)
			_list.setFont(font);
	}

	/**
	 * Select or unselect a value.
	 *
	 * @param value  the value that must be selected or unselected.
	 * @param select indicates if value must be selected or unselected.
	 */
	public void selectValue(String value, boolean select)
	{
		int			   i, n = _list.getCount();
		Vector<String> alreadySelectedValues = new Vector<String>(n);
		String[]	   item;

		for (i = 0; i < n; i++)
		{
			item = (String[])_list.getItemAt(i);
			alreadySelectedValues.addElement(item[0]);
		}

		if (select && !alreadySelectedValues.contains(value))
		{
			alreadySelectedValues.addElement(value);
			selectValues(alreadySelectedValues);
		}
		else if (!select && alreadySelectedValues.contains(value))
		{
			alreadySelectedValues.removeElement(value);
			selectValues(alreadySelectedValues);
		}
	}

	/**
	 * Select a list of values.
	 *
	 * @param values the list of internal values that must be selected.
	 */
	public void selectValues(Vector<String> values)
	{
		int			   i, n = ((values == null) ? 0 : values.size());
		int			   j, m = _list.getCount();
		Vector<String> alreadySelectedValues = new Vector<String>(m);
		String		   selectedValue, displayedValue, displayedShortValue;
		String[]	   item;

		for (j = 0; j < m; j++)
		{
			item = (String[])_list.getItemAt(j);
			alreadySelectedValues.addElement(item[0]);
		}

		for (i = 0; i < n; i++)
		{
			selectedValue = values.elementAt(i);
			displayedValue = _autoCompletionViewer.getDisplayedValue(selectedValue);
			displayedShortValue = _autoCompletionViewer.getDisplayedShortValue(selectedValue);
			item = new String[] {selectedValue, displayedValue, displayedShortValue};
			
			// Check if value was not already selected
			if (!alreadySelectedValues.contains(selectedValue))
				_list.addItem(item);

			_autoCompletionViewer.removeValue(selectedValue);
		}
	}

	/**
	 * Gets the list of selected elements' id.
	 *
	 * @return the list of selected element ids. These ids are separated by {@link
	 *         LyJListComponent#ITEM_SEP}.
	 */
	public String getValue()
	{
		StringBuffer   sb = null;
		Vector<String> selectedIds = new Vector<String>();
		int			   i, n = _list.getCount();

		for (i = 0; i < n; i++)
			selectedIds.addElement(((String[])_list.getItemAt(i))[0]);

		n = selectedIds.size();

		for (i = 0; i < n; i++)
		{
			if (sb == null)
			{
				sb = new StringBuffer();
				sb.append(selectedIds.elementAt(i));
			}
			else
			{
				sb.append(LyJListComponent.ITEM_SEP);
				sb.append(selectedIds.elementAt(i));
			}
		}
		return ((sb == null) ? null : sb.toString());
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return (_actionTransmitter);
	}

	/**
	 * This multiselector doesn't handle selection events, so it has no selection transmitter.
	 *
	 * @return <code>null</code>
	 * @see    leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		return null;
	}

	/* ****************************************************************************************/

	/**
	 * This class implements the preferred size of its super component.
	 */
	public class _LyJSpecialScrollPane extends JScrollPane
	{
		private static final long serialVersionUID = -7857491982378794672L;

		/** The preferred size of this component. */
		protected Dimension		  _preferredSize;

		/**
		 * Creates a new ScrollPane component.
		 *
		 * @param comp the new component
		 */
		public _LyJSpecialScrollPane(Component comp)
		{
			super(comp);
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.Component#getPreferredSize()
		 */
		public Dimension getPreferredSize()
		{
			if ((_preferredSize == null) || (_preferredSize.width < 10)
					|| (_preferredSize.height < 10))
				_preferredSize = super.getPreferredSize();

			return _preferredSize;
		}
	} // end class _LyJSpecialScrollPane


	/* ********************************************************************************/

	/**
	 * This class implements a customized renderer for list cell.
	 */
	public class _LyJCellRenderer extends DefaultListCellRenderer
	{
		private static final long serialVersionUID = -8395920779745580968L;

		/** The text on the cell. */
		String					  label;

		/** The associated icon of a list cell. */
		Icon					  icon;

		/**
		 * Sole constructor.
		 */
		public _LyJCellRenderer()
		{
			super();
		}

		/*
		 * (non-Javadoc)
		 * @see javax.swing.ListCellRenderer#getListCellRendererComponent(javax.swing.JList,
		 * java.lang.Object, int, boolean, boolean)
		 */
		public Component getListCellRendererComponent(JList list, Object value, int index,
			boolean isSelected, boolean hasFocus)
		{
			JLabel   comp = (JLabel)super.getListCellRendererComponent(list, value, index,
					isSelected, hasFocus);
			String[] values = (String[])value;

			comp.setText(values[2]);
			comp.setToolTipText(values[1]);

			return comp;
		}
	} // end class _LyJCellRenderer


	/* ********************************************************************************/

	/**
	 * This class implements a key listener.
	 */
	public class _LyKeyListener extends KeyAdapter
	{
		/**
		 * Sole constructor.
		 */
		public _LyKeyListener()
		{
			super();
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.KeyAdapter#keyReleased(java.awt.event.KeyEvent)
		 */
		public void keyReleased(KeyEvent e)
		{
			checkButtons();
		}
	} // end class _LyKeyListener


	/**
	 * Provides drap and drop handling.
	 */
	class _LyJAutoCompletionMultiSelectorDropListener
		implements DropTargetListener
	{

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#dragEnter(java.awt.dnd.DropTargetDragEvent)
		 */
		public void dragEnter(DropTargetDragEvent dtde)
		{
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#dragExit(java.awt.dnd.DropTargetEvent)
		 */
		public void dragExit(DropTargetEvent dte)
		{
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#dragOver(java.awt.dnd.DropTargetDragEvent)
		 */
		public void dragOver(DropTargetDragEvent dtde)
		{
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#drop(java.awt.dnd.DropTargetDropEvent)
		 */
		public void drop(DropTargetDropEvent dtde)
		{
			int index = _list.locationToIndex(dtde.getLocation());

			dtde.acceptDrop(DnDConstants.ACTION_COPY_OR_MOVE);

			int[]			 selection = _list.getSelectedIndices();
			String[]		 elem;
			Vector<String[]> elems = new Vector<String[]>();

			// Allow drop only if drop index is outside current selection
			if ((index >= selection[0]) && (index <= selection[selection.length - 1]))
			{
				dtde.getDropTargetContext().dropComplete(true);
				return;
			}

			dtde.getDropTargetContext().dropComplete(true);

			// First remove items from their original positions
			for (int i = selection.length - 1; i >= 0; i--)
			{
				elem = (String[])_list.getItemAt(selection[i]);
				_list.removeItemAt(selection[i]);
				elems.addElement(elem);
			}

			boolean before = (index < selection[0]);

			// Then add items to their target positions
			for (int i = elems.size() - 1; i >= 0; i--)
			{
				elem = elems.elementAt(i);

				int pos = index - i;

				if (before)
					pos = index + (elems.size() - 1) - i;

				_list.insertItemAt(elem, pos);
				_list.selectItem(elem, true);
			}

			checkButtons();
		} // end method drop

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#dropActionChanged(java.awt.dnd.DropTargetDragEvent)
		 */
		public void dropActionChanged(DropTargetDragEvent dtde)
		{
		}
	} // end class _LyJAutoCompletionMultiSelectorDropListener
} // end class LyJAutoCompletionMultiSelector
