/*
 * LyAutoCompletionViewer.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;
import javax.swing.text.*;

import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.win.swing.*;


/**
 * This class implements an auto-completion system for text components.
 *
 * @version $Id: LyJAutoCompletionViewer.java 30106 2011-03-16 15:32:33Z bmichel $
 * @author  Lyria
 */
public class LyJAutoCompletionViewer
	implements LySelectionListener
{
	public static final String				  __VERSION =
		"$Id: LyJAutoCompletionViewer.java 30106 2011-03-16 15:32:33Z bmichel $";

	/** Value used to store empty values in the auto-completion list. */
	public static final String				  EMPTY_VALUE = " ";
	
	/** Timer used to send keyboard events (in milliseconds). */
	public static long						  KEY_PRESSED_TIMER = 300;

	/** */
	public static final String BEAN_KEY_EVENT = "Formula_Editor_Bean_Event_KeyReleased";
	
	/** Text component to be auto-completed. */
	protected JTextComponent					  _text;
	
	/** Button used to open / close auto-completion window. */
	protected JButton						  _button;

	/** List of internal values used for auto-completion. */
	protected String[]						  _internalValues;

	/** List of displayed values used for auto-completion. */
	protected String[]						  _displayedValues;

	/** List of displayed values used for auto-completion. */
	protected String[]						  _displayedShortValues;
	
	/** Indicates whether text value is updated automatically when focus is lost. */
	protected boolean						  _automaticCompletion;

	/** Window containing auto-completion list. */
	protected JWindow						  _popupWindow;
	
	/** Fixed width of the popup. */
	protected int							  _popupWidth;

	/** List containing auto-completion values. */
	protected LyJListComponent				  _list;
	
	/** Index of the selected value from the list of possible values. */
	protected int							  _selectedIndex;
	
	/** List of visible indexes displayed in the graphical list. */
	protected Vector<Integer>				  _visibleIndexes;

	/** Scroll panel containing auto-completion list. */
	protected JScrollPane					  _scrollPane;

	/** Maximum number of values displayed in the auto-completion list. */
	protected int							  _nMax;

	/** Indicates whether listeners on parent components are already registered. */
	protected boolean						  _listenersRegistered;

	/** Listener used to manage graphical events. */
	protected _LyAutoCompletionViewerListener _listener;

	/** Current environment. */
	protected LyEnvironment					  _environment;
	
	/** Listen that must be notified when changing text value. */
	protected LyActionListener				  _actionListener;
	
	/** The time when window was last shown. */
	protected long							  _windowShowTime;
	
	/** Internal value used to check if selection in the list must be notified. */
	protected boolean						  _notifySelection;

	/** Current text for multi-line text. */
	protected String 						  _currentText;
	
	/** indicate whether the text field is multi. */
	protected boolean						  _isMulti;
	
	/** If the text field is multi, indicates the start and end 
	 * index of the current word */
	protected int							  _currentTextStartIndex;
	protected int 							  _currentTextEndIndex;
	
	// indicate whether the autocompletion list is shown when keys are typed
	protected boolean						  _showOnKeyType;
	
	/** Internal timer used to send key events every KEY_PRESSED_TIMER milliseconds. */
	protected LyTimer						  _keyTimer;

	/**
	 * Constructor.
	 *
	 * @param text                text component used for auto-completion
	 * @param button              button component used to open / close auto-completion window
	 * @param internalValues      list of internal values used for auto-completion
	 * @param displayedValues     list of displayed values used for auto-completion
	 * @param displayedValues     list of displayed short values used for auto-completion
	 * @param automaticCompletion indicates whether text value is updated automatically when focus is lost
	 * @param nMax                maximum number of values displayed in the auto-completion list
	 * @param environment         current environment
	 */
	public LyJAutoCompletionViewer(JTextComponent text, JButton button, String[] internalValues, String[] displayedValues,
			String[] displayedShortValues, boolean automaticCompletion, int nMax, LyEnvironment environment)
	{
		_text = text;
		_button = button;
		_internalValues = internalValues;
		_displayedValues = displayedValues;
		_displayedShortValues = displayedShortValues;
		_nMax = nMax;
		_selectedIndex = -1;
		_listenersRegistered = false;
		_environment = environment;
		_windowShowTime = -1;
		_popupWidth = -1;
		_automaticCompletion = automaticCompletion;
		_actionListener = null;
		_visibleIndexes = null;
		_notifySelection = true;
		_currentText = "";
		_currentTextEndIndex = 0;
		_currentTextStartIndex = 0;
		_isMulti = false;
		_showOnKeyType = true;
		_keyTimer = null;

		init();
	}

	/**
	 * Initialization of the component.
	 */
	protected void init()
	{
		// Register listeners
		_listener = new _LyAutoCompletionViewerListener();
		_text.addKeyListener(_listener);
		_text.addFocusListener(_listener);
		
		if (_button != null)
			_button.addActionListener(_listener);
	} // end method init
	
	/**
	 * Set the list of values used for auto-completion.
	 * 
	 * @param internalValues the new list of internal values for auto-completion.
	 * @param displayedValues the new list of displayed values for auto-completion.
	 * @param displayedShortValues the new list of displayed short values for auto-completion.
	 */
	public void setValues(String[] internalValues, String[] displayedValues, String[] displayedShortValues)
	{
		_internalValues = internalValues;
		_displayedValues = displayedValues;
		_displayedShortValues = displayedShortValues;
	}
	
	/**
	 * indicate whether the autocompletion list is shown when keys are typed
	 * 
	 * @param show if true, the completion list is shown when keys are typed
	 */
	public void setShowOnKeyType(boolean show)
	{
		if(_showOnKeyType != show)
			_showOnKeyType = show;
	}

	/**
	 * Set the list of values used for auto-completion.
	 * 
	 * @param nMax max values displayed
	 * @param internalValues the new list of internal values for auto-completion.
	 * @param displayedValues the new list of displayed values for auto-completion.
	 * @param displayedShortValues the new list of displayed short values for auto-completion.
	 */
	public void setValues(int nMax, String[] internalValues, String[] displayedValues, String[] displayedShortValues)
	{
		_nMax = nMax;
		setValues(internalValues, displayedValues, displayedShortValues);
	}
	
	/**
	 * Adds a value to the list of values used for auto-completion.
	 * 
	 * @param internalValue the new internal value to be added.
	 * @param displayedValue the new displayed value to be added.
	 * @param displayedShortValue the new displayed short value to be added.
	 * @param pos the position of the value to be added.
	 */
	public void addValue(String internalValue, String displayedValue, String displayedShortValue, int pos)
	{
		int			i, offset = 0, n = _internalValues.length;
		String[]	newInternalValues = new String[n + 1];
		String[]	newDisplayedValues = new String[n + 1];
		String[]	newDisplayedShortValues = new String[n + 1];
		
		for (i = 0; i < n; i++)
		{
			if (i == pos)
			{
				newInternalValues[i] = internalValue;
				newDisplayedValues[i] = displayedValue;
				newDisplayedShortValues[i] = displayedShortValue;
				offset = 1;
			}
			
			newInternalValues[i + offset] = _internalValues[i];
			newDisplayedValues[i + offset] = _displayedValues[i];
			newDisplayedShortValues[i + offset] = _displayedShortValues[i];
		}
		
		if ((pos < 0) || (pos >= n))
		{
			newInternalValues[n] = internalValue;
			newDisplayedValues[n] = displayedValue;
			newDisplayedShortValues[n] = displayedShortValue;
		}
		
		_internalValues = newInternalValues;
		_displayedValues = newDisplayedValues;
		_displayedShortValues = newDisplayedShortValues;
	}

	/**
	 * Removes a value from the list of values used for auto-completion.
	 * 
	 * @param internalValue the internal value to be removed.
	 */
	public void removeValue(String internalValue)
	{
		int		i, n = _internalValues.length;
		
		for (i = 0; i < n; i++)
		{
			if (_internalValues[i].equals(internalValue))
			{
				removeValueAt(i);
				break;
			}
		}
	}

	/**
	 * Removes a value from the list of values used for auto-completion.
	 * 
	 * @param pos the position of the value to be removed.
	 */
	public void removeValueAt(int pos)
	{
		if ((pos < 0) || (pos >= _internalValues.length))
			return;
		
		int			i, offset = 0, n = _internalValues.length;
		String[]	newInternalValues = new String[n - 1];
		String[]	newDisplayedValues = new String[n - 1];
		String[]	newDisplayedShortValues = new String[n - 1];
		
		for (i = 0; i < n; i++)
		{
			if (i == pos)
				offset = -1;
			else
			{
				newInternalValues[i + offset] = _internalValues[i];
				newDisplayedValues[i + offset] = _displayedValues[i];
				newDisplayedShortValues[i + offset] = _displayedShortValues[i];
			}
		}
		
		_internalValues = newInternalValues;
		_displayedValues = newDisplayedValues;
		_displayedShortValues = newDisplayedShortValues;
		
		if (_selectedIndex == pos)
		{
			if (_selectedIndex > 0)
				setTextValue(_selectedIndex - 1, false);
			else if (_internalValues.length > 0)
				setTextValue(0, false);
			else
			{
				setTextValue(-1, false);
				_text.setText("");
			}
		}
	}
	
	/**
	 * Returns the index of selected value.
	 * 
	 * @return the index of selected value, or -1 if no value was selected.
	 */
	public int getSelectedIndex()
	{
		return _selectedIndex;
	}
	
	/**
	 * Returns the internal selected value if any.
	 * 
	 * @return the internal value that has been selected if any, null otherwise.
	 */
	public String getSelectedValue()
	{
		if (_selectedIndex >= 0)
			return _internalValues[_selectedIndex];
		
		return null;
	}
	
	/**
	 * Returns the displayed value corresponding to a given internal value.
	 * 
	 * @param internalValue the internal value to be used to retrieve displayed value.
	 * @return the displayed value corresponding to the given internal value or null if
	 * no internal value was found for given internal value.
	 */
	public String getDisplayedValue(String internalValue)
	{
		int		i;
		
		for (i = 0; i < _internalValues.length; i++)
		{
			if (_internalValues[i].equals(internalValue))
				return _displayedValues[i];
		}
		
		return null;
	}
	
	/**
	 * Returns the displayed short value corresponding to a given internal value.
	 * 
	 * @param internalValue the internal value to be used to retrieve displayed value.
	 * @return the displayed short value corresponding to the given internal value or null if
	 * no internal value was found for given internal value.
	 */
	public String getDisplayedShortValue(String internalValue)
	{
		int		i;
		
		for (i = 0; i < _internalValues.length; i++)
		{
			if (_internalValues[i].equals(internalValue))
				return _displayedShortValues[i];
		}
		
		return null;
	}

	/**
	 * This method test Component c is member of this class
	 * 
	 * @param c Component to test
	 * @return boolean true if given component is member of current completion viewer
	 */
	public boolean componentIsMine(Component c)
	{
		if ((c != _popupWindow)
				&& (c != _list)
				&& (c != _scrollPane)
				&& (c != _text))
			return false;
		else
			return true;
	}
	
	/**
	 * Returns the list of displayed values for auto-completion.
	 * 
	 * @return the list of displayed values for auto-completion.
	 */
	public String[] getDisplayedValues()
	{
		return _displayedValues;
	}
	
	/**
	 * Returns the list of displayed short values for auto-completion.
	 * 
	 * @return the list of displayed short values for auto-completion.
	 */
	public String[] getDisplayedShortValues()
	{
		return _displayedShortValues;
	}
	
	/**
	 * Returns the text component associated with this auto-completion viewer.
	 * 
	 * @return the text component associated with this auto-completion viewer.
	 */
	public JTextComponent getTextField()
	{
		return _text;
	}
	
	/**
	 * Returns the list component associated with this auto-completion viewer.
	 * 
	 * @return the list component associated with this auto-completion viewer.
	 */
	public LyJListComponent getListField()
	{
		return _list;
	}
	
	/**
	 * Builds the window component.
	 */
	protected void buildWindow()
	{
		// Create auto-completion window
		Color     color = (Color)_environment.getColor("LY_ENABLED_BACKGROUND");
		Color     selectColor = (Color)_environment.getColor("LY_TABLE_SELECT_COLOR");
		Font	  font = (Font)_environment.getFont("LY_DEFAULT_FONT");
		Window    frame = null;
		Container parent = _text.getParent();

		while ((parent != null) && (frame == null))
		{
			if (parent instanceof Window)
				frame = (Window)parent;
			parent = parent.getParent();
		}

		if (frame == null)
			frame = LySwingViewManager.getDefaultFrame();

		_popupWindow = new JWindow(frame);
		_popupWindow.setAlwaysOnTop(true);
		_list = new LyJListComponent();
		_list.setCellRenderer(new _LyAutoCompletionRenderer());
		_scrollPane = new JScrollPane(_list);
		_scrollPane.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
		_popupWindow.getContentPane().add(_scrollPane);
		_list.setFont(font);
		_list.setBackground(color);
		_list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		_list.setSelectionBackground(selectColor);
		_list.setSelectionForeground(color);
		_list.addFocusListener(_listener);
		_list.setFocusable(true);
		_scrollPane.setFocusable(true);

		LySelectionTransmitter selectionTransmitter = _list.getSelectionTransmitter();

		if (selectionTransmitter != null)
			selectionTransmitter.addSelectionListener(this);
	} // end method init

	/**
	 * Updates the values displayed in the list according to the text value.
	 * 
	 * @param fill indicates whether list must be filled with all the values
	 * or only with values matching text value.
	 */
	protected void updateList(boolean fill)
	{
		if (_list == null)
			return;
		
		_notifySelection = false;
		
		String	text;
		
		if(_isMulti)
			text = _currentText;
		else
			text = _text.getText();
		
		String	textLower = ((text == null) ? "" : text.toLowerCase());

		// Send an event to the controller side that may refresh the content of the list
		if (_actionListener != null)
		{
			LyActionEvent	event = new LyActionEvent(this, null, textLower,
					(fill ? LyActionDemand.TYPE_EXPAND : LyActionDemand.TYPE_COMPLETE_VALUE));
			
			_actionListener.actionPerformed(event);
			
			if(!_isMulti)
				_text.setText(text);
		}
		
		int		i, n = _displayedValues.length, count = 0, nItems = _list.getCount();
		String	valueLower;
		boolean	valueSelected = false;

		// Update list content with items matching current text
		for (i = nItems - 1; i >= 0; i--)
			_list.removeItemAt(i);

		if (_visibleIndexes == null)
			_visibleIndexes = new Vector<Integer>(n);
		
		_visibleIndexes.clear();
		
		for (i = 0; (i < n) && ((count < _nMax) || (_nMax < 0)); i++)
		{
			valueLower = ((_displayedValues[i] == null) ? "" : _displayedValues[i].toLowerCase());

			if (fill || valueLower.startsWith(textLower))
			{
				// Add matching value to the list of values
				if ("".equals(_displayedValues[i]))
					_list.addItem(new _LyAutoCompletionStruct(EMPTY_VALUE, EMPTY_VALUE));
				else
					_list.addItem(new _LyAutoCompletionStruct(_internalValues[i], _displayedValues[i]));
				
				// Fill the list of visible indexes
				_visibleIndexes.addElement(new Integer(i));
				
				// Select first value matching the text value
				if (_automaticCompletion && !valueSelected && valueLower.startsWith(textLower))
				{
					_list.selectItem(_list.getItemAt(count), true);
					_list.ensureIndexIsVisible(count);
					valueSelected = true;
				}
				
				count++;
			}
		}
		
		_notifySelection = true;

		if (!valueSelected && !fill)
		{
			// Notify that no value is selected if needed
			int	oldSelectedIndex = _selectedIndex;

			_selectedIndex = -1;
			
			if (((oldSelectedIndex != _selectedIndex) || (_selectedIndex < 0))
					 && (_actionListener != null))
			{
				LyActionEvent	event = null;
				
				if ((_selectedIndex >= 0))
					event = new LyActionEvent(this, null, null, LyActionDemand.TYPE_VALUE_CHANGED);
				
				if (event != null)
					_actionListener.actionPerformed(event);
			}
		}
		
		// Do not display list if nothing to display
		if (count == 0)
			hideWindow();
		else
		{
			// Register listeners on parent components if not already done
			if (!_listenersRegistered)
			{
				Container container = _text.getParent();

				while (container != null)
				{
					container.addMouseListener(_listener);
					container.addComponentListener(_listener);

					if (container instanceof Window)
					{
						((Window)container).addComponentListener(_listener);
						break;
					}

					container = container.getParent();
				}

				_list.addMouseListener(_listener);
				_list.addMouseMotionListener(_listener);
				_listenersRegistered = true;
			}
			
			int oldVisibleRowCount = _list.getVisibleRowCount();
			int visibleRowCount = Math.min(count, 8);

			_list.setVisibleRowCount(visibleRowCount);

			// If visible row count is greater than before, hide the window
			// to ensure a correct repaint of the list.
			if (oldVisibleRowCount < visibleRowCount)
				_popupWindow.setVisible(false);

			setWindowLocation();
			showWindow();
		}


		
	} // end method updateList

	/**
	 * Set the value of the currently typed
	 * 
	 * @param currentText	text value
	 * @param startIndex	start index in the text component
	 * @param endIndex		end index in the text component
	 */
	public void setCurrentText(String currentText, int startIndex, int endIndex)
	{
		_currentText = currentText;
		_currentTextStartIndex = startIndex;
		_currentTextEndIndex = endIndex;
	}
	
	/**
	 * Implementation of the LySelectionListener interface.
	 *
	 * @param selectionEvent description of the selection which has occured
	 */
	public void selectionPerformed(LySelectionEvent selectionEvent)
	{
		if (!_notifySelection)
			return;
		
		// If an item of the list is selected, refresh the text component
		// with the item label and hide the list
		int index = selectionEvent.getIndex(0);

		selectListItem(index, true);
	}

	/**
	 * Selects an item of the list and affect its value to the text component.
	 *
	 * @param index the index of selected item in the list.
	 * @param asynchronous indicates if text component must be set asynchronously.
	 */
	protected void selectListItem(int index, boolean asynchronous)
	{
		if (_list == null)
			return;

		if ((index < 0)
				|| (index >= _list.getCount()))
		{
			setTextValue(-1, true);
		}
		else
		{
			Integer		realIndex = _visibleIndexes.elementAt(index);
			
			if (asynchronous)
			{
				_LyAutoCompletionViewerRunnable		acvr = new _LyAutoCompletionViewerRunnable(realIndex.intValue());
	            LyViewManager						viewManager = (LyViewManager)_environment.getViewManager();

				viewManager.execGui(acvr);
			}
			else
			{
				setTextValue(realIndex.intValue(), true);
				_text.requestFocus();
			}
		}
	}
	
	/**
	 * Checks if the text value corresponds to a given auto-completion value.<br />
	 * This method does nothing if the _automaticCompletion mode is not activated.
	 */
	protected void checkTextValue()
	{
		if (!_automaticCompletion)
			return;
		
		String 	text = _text.getText();
		String	textLower = ((text == null) ? "" : text.toLowerCase()), valueLower;
		int		i, n = _displayedShortValues.length;
		
		for (i = 0; i < n; i++)
		{
			valueLower = ((_displayedShortValues[i] == null) ? "" : _displayedShortValues[i].toLowerCase());
			
			if (valueLower.startsWith(textLower))
			{
				setTextValue(i, true);
				return;
			}
		}
		
		// Cancel any selected value
		setTextValue(-1, true);
	}
	
	/**
	 * Set the popup menu location
	 * 
	 * @param x	x coordinate
	 * @param y	y coordinate
	 */
	public void setLocation(int x, int y)
	{
		Point onScreen = ((JTextArea)_text).getLocationOnScreen();
		_popupWindow.setLocation(onScreen.x+x, onScreen.y+y);
	}
	
	/**
	 * Set if the popup menu is for a multilines text on not
	 * @param isMulti
	 */
	public void setMutli(boolean isMulti)
	{
		_isMulti = isMulti;
	}
	
	/**
	 * Sets the value of the text component.
	 * 
	 * @param index  the index of the new text value to be set.
	 * @param notify indicates if action listener must be notified.
	 */
	public void setTextValue(int index, boolean notify)
	{
		int	oldSelectedIndex = _selectedIndex;
		
		String	displayedValue = null;
		String	tooltip = null;

		if ((index < 0) || (index >= _displayedShortValues.length))
			_selectedIndex = -1;
		else
		{
			displayedValue = _displayedShortValues[index];
			tooltip = _displayedValues[index];
			_selectedIndex = index;
		}
		
		if (displayedValue == EMPTY_VALUE)
			displayedValue = "";
		
		// Set the text value only if a "real" index is given. Otherwise, leave the
		// text as is it.
		if (index >= 0)
		{
			if(_isMulti)
			{
				String 		selectedWord = displayedValue;
				String		text = _text.getText();
				
				String	prefix = text.substring(0, _currentTextStartIndex);
				String	suffix = "";
				
				suffix = text.substring(_currentTextEndIndex);
				
				text = prefix+selectedWord+suffix;
				
				_text.setText(text);
				_text.setCaretPosition((prefix+selectedWord).length());
				_text.setToolTipText(null);
			}
			else
			{
				_text.setText(displayedValue);
				_text.setToolTipText(tooltip);
			}
		}
		else
			_text.setToolTipText(null);
		
		if (notify && (_actionListener != null))
		{
			if ((oldSelectedIndex != _selectedIndex) || (_selectedIndex <= 0))
			{
				LyActionEvent	event;
				
				if (_selectedIndex < 0)
					event = new LyActionEvent(this, null, _text.getText(), LyActionDemand.TYPE_VALUE_CHANGED);
				else
					event = new LyActionEvent(this, null,
							(index < 0 ? null : _internalValues[index]), LyActionDemand.TYPE_VALUE_CHANGED);
				
				_actionListener.actionPerformed(event);
			}
		}
	}
	
	/**
	 * Sets the action listener to be used.
	 * 
	 * @param actionListener the new action listener that must be notified when changing
	 * text value.
	 */
	public void setActionListener(LyActionListener actionListener)
	{
		_actionListener = actionListener;
	}
	
	/**
	 * Shows the popup window.
	 */
	protected void showWindow()
	{
		if ((_text == null)
				|| (_popupWindow == null)
				|| !_text.isVisible()
				|| !_text.isEnabled()
				|| !_text.isEditable()
				|| _popupWindow.isVisible())
			return;
		
		if (_popupWindow == null)
			buildWindow();

		// Be sure popup window is closed
		setWindowLocation();
		_popupWindow.setVisible(true);
		_windowShowTime = System.currentTimeMillis();
	}

	/**
	 * Sets the fixed width of the popup.
	 * 
	 * @param popupWidth the fixed width of the popup.
	 */
	public void setPopupWidth(int popupWidth)
	{
		_popupWidth = popupWidth;
	}
	
	/**
	 * Changes the location of the completion window.
	 */
	protected void setWindowLocation()
	{
		if ((_text == null)
				|| (_popupWindow == null))
			return;

		Point     location = getWindowLocation();
		int		  textWidth = _text.getSize().width;
		int		  buttonWidth = (_button == null ? 1 : _button.getSize().width + 1);
		Dimension preferredSize = _popupWindow.getPreferredSize();
		int		  width = Math.max(textWidth + buttonWidth, preferredSize.width - 15);
		
		if (_popupWidth > 0)
			width = _popupWidth;
		
		_popupWindow.setLocation(location.x, location.y);
		_popupWindow.setSize(width, preferredSize.height);
		_scrollPane.setSize(width, preferredSize.height);
	}

	/**
	 * Determines the completion window location on screen.
	 *
	 * @return the position of the completion window on screen.
	 */
	protected Point getWindowLocation()
	{
		if (_popupWindow == null)
			return null;

		int[]	  screenSize = _environment.getScreenSize(_text);
		Point     location = _text.getLocationOnScreen();
		Dimension size = _popupWindow.getPreferredSize();
		Dimension textSize = _text.getSize();

		location.y += textSize.height - 1;

		Rectangle virtualBounds = new Rectangle();
	    GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
	    GraphicsDevice[] gs = ge.getScreenDevices();
	    GraphicsConfiguration popupGc = _text.getGraphicsConfiguration();
	    
	    GraphicsDevice popupDevice = popupGc.getDevice();
	    for (int j = 0; j < gs.length; j++)
	    { 
	    	GraphicsDevice gd = gs[j];
	    	if (gd.equals(popupDevice))
	    		break;
	    	GraphicsConfiguration[] gc = gd.getConfigurations();
	    	
	    	for (int i=0; i < gc.length; i++)
	    	{
	    		virtualBounds = virtualBounds.union(gc[i].getBounds());
	    	}
	    } 
	      
	    int[]	  virtualScreenSize = new int[] { virtualBounds.width, virtualBounds.height };
	    
	    // compute locationOffset
	    int horizontalOffset = (location.x > virtualScreenSize[0] ? virtualScreenSize[0] : 0);
	    int verticalOffset = (location.y > virtualScreenSize[1] ? virtualScreenSize[1] : 0);
	    if ((horizontalOffset == 0 ) && (location.x > screenSize[0] + screenSize[2]))
	       	horizontalOffset = screenSize[0];
	    
	    if ((verticalOffset == 0 ) && (location.y > screenSize[1] + screenSize[2]))
	    	verticalOffset = screenSize[1];
	    
	    location.x -= horizontalOffset;
	    location.y -= verticalOffset;
		
		// Check if the window can be entirely displayed
		int diffX = screenSize[0] + screenSize[2] - (location.x + size.width);
		int diffY = screenSize[1] + screenSize[3] - (location.y + size.height);
		int newWidth = size.width;
		int newHeight = size.height;

		if (_popupWidth > 0)
			diffX = screenSize[0] + screenSize[2] - (location.x + _popupWidth);
		
		if (diffX < 0)
			location.x += diffX;
		
		if (diffY < 0)
		{
			int diffUpY = location.y - size.height - textSize.height;

			if (diffUpY >= 0)
				location.y -= size.height + textSize.height;
			else
			{
				if (diffUpY < diffY)
					newHeight = screenSize[1] + screenSize[3] - location.y;
				else
				{
					newHeight = location.y - textSize.height;
					location.y -= newHeight + textSize.height;
				}
			}
		}

		if ((newWidth != size.width) || (newHeight != size.height))
			_popupWindow.setSize(newWidth, newHeight);

		location.x += horizontalOffset;
		location.y += verticalOffset;
		return location;
	} // end method getWindowLocation

	/**
	 * Hides the completion window.
	 */
	public void hideWindow()
	{
		// Avoid a flash of the popup window
		if (System.currentTimeMillis() - _windowShowTime < 50)
			return;
		
		if ((_popupWindow != null)
				&& _popupWindow.isVisible())
		{
			// FFT00602/Flandres: Menu droulant d'auto-compltion
			_popupWindow.requestFocusInWindow();
			_popupWindow.setVisible(false);
		}
	}

	/**
	 * Indicates whether the completion window is visible or not.
	 *
	 * @return true if the completion window is visible, false otherwise.
	 */
	public boolean isWindowVisible()
	{
		return ((_popupWindow != null)
				&& _popupWindow.isVisible());
	}

	/* Adapter classes to receive events -------------------------------------------- */

	/**
	 * Listen the parent window : if anything happens, focus is assumed to be lost.
	 */
	class _LyAutoCompletionViewerListener extends KeyAdapter
		implements FocusListener,
			ComponentListener,
			MouseListener,
			MouseMotionListener,
			ActionListener,
			LyTimerClientInterface
	{
		/** Indicates whether the enter key has been pressed before. */
		boolean _enterPressed = false;
		
		/** Time when component got the focus. */
		long	_focusTime = -1;
		
		String	_lastFocusValue = null;

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.KeyAdapter#keyReleased(java.awt.event.KeyEvent)
		 */
		public void keyReleased(KeyEvent e)
		{	
			// FFT01403
			_text.firePropertyChange(BEAN_KEY_EVENT, 0, 1);
			
			int code = e.getKeyCode();

			if (_enterPressed)
			{
				_enterPressed = false;
				e.consume();
				return;
			}

			if ((code == KeyEvent.VK_ESCAPE)
					|| (code == KeyEvent.VK_TAB)
					|| (code == KeyEvent.VK_LEFT)
					|| (code == KeyEvent.VK_RIGHT))
				return;
			// Current window may have been opened while pressing CTRL key : in this
			// case list must not be displayed.
			else if ((code == KeyEvent.VK_CONTROL)
					|| (code == KeyEvent.VK_SHIFT))
				return;
			else if ((((code != KeyEvent.VK_DOWN)
						&& (code != KeyEvent.VK_UP))
					|| !isWindowVisible()) && _showOnKeyType)
			{
				boolean	fill = (code == KeyEvent.VK_DOWN) || (code == KeyEvent.VK_UP);
				
				if ((KEY_PRESSED_TIMER > 0) && !fill)
				{
					synchronized (this)
					{
						// If a timer exists, use a LyTimer instance to send an event
						if (_keyTimer == null)
						{
							_keyTimer = new LyTimer(this, KEY_PRESSED_TIMER);
							_keyTimer.start();
						}
					}
				}
				else
				{
					updateList(fill);
					_text.requestFocus();
				}				
			}
			
			// FFT01403
			_text.firePropertyChange(BEAN_KEY_EVENT, 1, 0);
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.KeyAdapter#keyPressed(java.awt.event.KeyEvent)
		 */
		public void keyPressed(KeyEvent e)
		{			
			if (_popupWindow == null)
				buildWindow();
			
			int code = e.getKeyCode();
			int selectedIndex = (_list == null ? -1 : _list.getSelectedIndex());

			if (code == KeyEvent.VK_ESCAPE)
				return;

			if ((code == KeyEvent.VK_ENTER)
				&& (selectedIndex >= 0)
				&& (selectedIndex < _list.getCount()))
			{
				if (_popupWindow.isVisible())
				{
					// Give the focus to the list in order to ensure that text component
					// will get the focus again and then refresh its internal value.
					_enterPressed = true;
					_list.requestFocus();
					selectListItem(selectedIndex, false);
					hideWindow();
					e.consume();
				}
			}
			else if (code == KeyEvent.VK_DOWN)
			{
				if (_popupWindow.isVisible())
				{
					if ((_list == null) || (_list.getCount() == 0))
					{
						e.consume();
						return;
					}
	
					if (selectedIndex == (_list.getCount() - 1))
						_list.selectItem(_list.getItemAt(_list.getCount() - 1), false);
					else
					{
						if (selectedIndex >= 0)
							_list.selectItem(_list.getItemAt(selectedIndex), false);
	
						selectedIndex++;
						_list.selectItem(_list.getItemAt(selectedIndex), true);
						_list.ensureIndexIsVisible(selectedIndex);
	
						if (_automaticCompletion)
						{
							Integer		realIndex = _visibleIndexes.elementAt(selectedIndex);
							
							setTextValue(realIndex.intValue(), true);
						}
					}
	
					e.consume();
				}
			}
			else if (code == KeyEvent.VK_UP)
			{
				if (_popupWindow.isVisible())
				{
					if ((_list == null) || (_list.getCount() == 0))
					{
						e.consume();
						return;
					}
	
					if (selectedIndex == 0)
						_list.selectItem(_list.getItemAt(0), false);
					else
					{
						if (selectedIndex >= 0)
							_list.selectItem(_list.getItemAt(selectedIndex), false);
	
						if (selectedIndex < 0)
							selectedIndex = _list.getCount() - 1;
						else
							selectedIndex--;
	
						_list.selectItem(_list.getItemAt(selectedIndex), true);
						_list.ensureIndexIsVisible(selectedIndex);
						
						if (_automaticCompletion)
						{
							Integer		realIndex = _visibleIndexes.elementAt(selectedIndex);
							
							setTextValue(realIndex.intValue(), true);
						}
					}
	
					e.consume();
				}
			}
		} // end method keyPressed
		
		/*
		 * (non-Javadoc)
		 * @see java.awt.event.FocusListener#focusGained(java.awt.event.FocusEvent)
		 */
		public void focusGained(FocusEvent e)
		{
			_focusTime = System.currentTimeMillis();
			_lastFocusValue = _text.getText();
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.FocusListener#focusLost(java.awt.event.FocusEvent)
		 */
		public void focusLost(FocusEvent e)
		{
			if (System.currentTimeMillis() - _focusTime < 100)
				return;
			
			if ((e.getOppositeComponent() != _popupWindow)
					&& (e.getOppositeComponent() != _list)
					&& (e.getOppositeComponent() != _scrollPane)
					&& (e.getOppositeComponent() != _text))
			{		
				hideWindow();
				
				if ((_lastFocusValue != null)
						&& !(_lastFocusValue.equals(_text.getText())))
					checkTextValue();
			}
		}
		
		/*
		 * (non-Javadoc)
		 * @see java.awt.event.ComponentListener#componentResized(java.awt.event.ComponentEvent)
		 */
		public void componentResized(ComponentEvent e)
		{
			hideWindow();
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.ComponentListener#componentMoved(java.awt.event.ComponentEvent)
		 */
		public void componentMoved(ComponentEvent e)
		{
			hideWindow();
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.ComponentListener#componentShown(java.awt.event.ComponentEvent)
		 */
		public void componentShown(ComponentEvent e)
		{
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.ComponentListener#componentHidden(java.awt.event.ComponentEvent)
		 */
		public void componentHidden(ComponentEvent e)
		{
			hideWindow();
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
		 */
		public void mouseClicked(MouseEvent e)
		{
			if (e.getComponent() != _list)
				hideWindow();
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
		 */
		public void mousePressed(MouseEvent e)
		{
			if (e.getComponent() != _list)
				hideWindow();
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
		 */
		public void mouseReleased(MouseEvent e)
		{
			if (e.getComponent() != _list)
				hideWindow();
			else
			{
				int selectedIndex = (_list == null ? -1 : _list.getSelectedIndex());
				
				hideWindow();
				_text.requestFocus();
				selectListItem(selectedIndex, true);
			}
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
		 */
		public void mouseEntered(MouseEvent e)
		{
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
		 */
		public void mouseExited(MouseEvent e)
		{
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
		 */
		public void mouseDragged(MouseEvent e)
		{
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.MouseMotionListener#mouseMoved(java.awt.event.MouseEvent)
		 */
		public void mouseMoved(MouseEvent e)
		{
			int		  i, n = (_list == null ? -1 : _list.getCount());
			Rectangle rect;

			for (i = 0; i < n; i++)
			{
				rect = _list.getCellBounds(i, i);

				if (rect.contains(e.getPoint()))
				{
					_list.selectItem(_list.getItemAt(i), true);
					_list.ensureIndexIsVisible(i);
				}
			}
		}
		
		/*
		 * (non-Javadoc)
		 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
		 */
		public void actionPerformed(ActionEvent e)
		{
			if (_popupWindow == null)
				buildWindow();
			
			if (_popupWindow.isVisible())
				hideWindow();
			else
			{
				updateList(true);
				_text.requestFocus();
			}
		}
		
		/**
		 * When a LyTimer ends this method is invoked.
		 * @param timer the timer which expired.
		 */
		public void endTimer(LyTimer timer)
		{
			synchronized (this)
			{
				_LyAutoCompletionViewerTimer runnable = new _LyAutoCompletionViewerTimer();
				
				if (_text.hasFocus() && (_environment != null) && (_environment.getViewManager() != null))
					_environment.getViewManager().execGui(runnable, false);
				
				_keyTimer = null;
			}
		}
	} // end class _LyAutoCompletionViewerListener

	/**
	 * Listen the parent window : if anything happens, focus is assumed to be lost.
	 */
	class _LyAutoCompletionViewerRunnable implements Runnable
	{

		/** Index of the text value to be set on the text component. */
		int _textValueIndex;

		public _LyAutoCompletionViewerRunnable(int textValueIndex)
		{
			_textValueIndex = textValueIndex;
		}
		
		/*
		 * (non-Javadoc)
		 * @see java.lang.Runnable#run()
		 */
		public void run()
		{
			int	caretPosition = (_textValueIndex < _displayedValues.length ? _displayedValues[_textValueIndex].length() : 0);
			
			setTextValue(_textValueIndex, true);
			
			if(_isMulti)
				caretPosition = _currentTextStartIndex + _displayedValues[_textValueIndex].length();
			else
				caretPosition = Math.min(caretPosition, (_text.getText() == null ? 0 : _text.getText().length()));
			
			_text.setCaretPosition(caretPosition);
			_text.requestFocus();
		}
	}

	/**
	 * Runnable used to implement the key timer.
	 */
	class _LyAutoCompletionViewerTimer implements Runnable
	{
		public void run()
		{
			synchronized (this)
			{
				updateList(false);
				_text.requestFocus();
			}
		}
	}

	/**
	 * Structure used to store an auto-completion value.
	 */
	class _LyAutoCompletionStruct
	{
		/** Internal text value of the structure. */
		String _internalValue;

		/** Displayed text value of the structure. */
		String _displayedValue;

		public _LyAutoCompletionStruct(String internalValue, String displayedValue)
		{
			_internalValue = internalValue;
			_displayedValue = displayedValue;
		}
		
		/*
		 * (non-Javadoc)
		 * @see java.lang.Object#toString()
		 */
		public String toString()
		{
			return _displayedValue;
		}
	}
	
	class _LyAutoCompletionRenderer extends DefaultListCellRenderer
	{
		/**
		 * 
		 */
		private static final long serialVersionUID = -559416514211598817L;

		public _LyAutoCompletionRenderer()
		{
			super();
		}
		
		public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus)
		{
	    	super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);
	    	
	    	_LyAutoCompletionStruct	item = (value instanceof _LyAutoCompletionStruct ? (_LyAutoCompletionStruct)value : null);
	        
	        if (item != null)
	        {
	        	if (index >= 0)
	        	{
		        	LyJMultiLineLabel	l = new LyJMultiLineLabel(item._displayedValue, -1, 1, 1);
		        	
		        	if (_popupWidth > 0)
		        		l.setMaxWidth(_popupWidth - 15);
		        	else
		        		l.setMaxWidth(getTextField().getSize().width - 5);
		        	
		        	if (l.getNumLines() > 1)
			        	_list.setFixedHeight(144);
		        	
		        	l.setBackground(getBackground());
		        	l.setForeground(getForeground());
		        	
		        	return l;
	        	}
	        	else
	        		setText(item._displayedValue);
	        }
	    	
	        return this;
		}
	}

	public boolean wasEnterPressed()
	{
		// TODO Auto-generated method stub
		return _listener._enterPressed;
	}
} // end class LyJAutoCompletionViewer
