/*
 * LyJBorderPanel.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;
import javax.swing.border.*;

import leon.view.*;
import leon.view.setinterface.*;
import leon.view.win.swing.*;
import net.miginfocom.swing.*;


/**
 * This class implements a panel who can have severals types of border corresponding to the
 * Leonardi's style.
 *
 * @author  Lyria
 * @version $Id: LyJBorderPanel.java 31060 2011-05-18 10:13:06Z nhamon $
 */
public class LyJBorderPanel extends JPanel
{
	private static final long  serialVersionUID = 5981028955205306386L;

	public static final String __VERSION =
		"$Id: LyJBorderPanel.java 31060 2011-05-18 10:13:06Z nhamon $";

	// Constants

	/** Left alignment. */
	public final static int    LEFT = 0;

	/** Center alignment. */
	public final static int    CENTER = 1;

	/** Right alignment. */
	public final static int    RIGHT = 2;

	/** Default thickness: 10 pixels. */
	public final static int    DEFAULT_THICKNESS = 10;

	/** Default thickness for solid borders: 1 pixel. */
	public final static int    DEFAULT_SOLID_THICKNESS = 1;

	/** Default thickness for raised borders: 2 pixels. */
	public final static int    DEFAULT_RAISED_THICKNESS = 2;

	/** Default thickness for lowered borders: 2 pixels. */
	public final static int    DEFAULT_LOWERED_THICKNESS = 2;

	/** Default thickness for etched-in borders: 10 pixels. */
	public final static int    DEFAULT_IN_THICKNESS = 10;

	/** Default thickness for etched-out borders: 10 pixels. */
	public final static int    DEFAULT_OUT_THICKNESS = 10;

	/** Default gap between border and contained component: 1 pixel. */
	public final static int    DEFAULT_GAP = 1;

	/** Default color: black. Applies to SOLID and etched borders. */
	public final static Color  DEFAULT_COLOR = Color.black;

	/** Default font: TimesRoman, PLAIN, 14. Only applies to etched borders. */
	public final static Font   DEFAULT_FONT = new Font("TimesRoman", Font.PLAIN, 14);

	/** Default alignment: {@link #LEFT}. Only applies to etched borders. */
	public final static int    DEFAULT_ALIGNMENT = LEFT;

	// Data
	private LyArea.Border	   _style;
	private int				   _thickness;
	private int				   _gap;
	private Color			   _color;

	private Font			   _font;
	private String			   _text;
	private int				   _alignment;
	private boolean			   _closable;
	private boolean			   _open;
	private Image			   _openImage;
	private Image			   _closeImage;
	private LyView			   _parentView;

	/**
	 * Default constructor. Makes default border.
	 */
	public LyJBorderPanel()
	{
		super();
		// initialize data
		_style = LyArea.Border.IN;
		_thickness = DEFAULT_THICKNESS;
		_gap = DEFAULT_GAP;
		_color = DEFAULT_COLOR;

		_text = null;
		_font = DEFAULT_FONT;
		_alignment = DEFAULT_ALIGNMENT;
		_closable = false;
		_open = true;
		_openImage = null;
		_closeImage = null;
		_parentView = null;
		updateBorder();

	}

	/**
	 * Class constructor. Makes an etched {@link #IN} border with given text caption.
	 *
	 * @param text the text caption
	 */
	public LyJBorderPanel(String text)
	{
		this();

		_style = LyArea.Border.IN;
		_text = text;
		updateBorder();
	}

	/**
	 * Class constructor of a new JBorder Panel object. Makes {@link #SOLID} border with color and
	 * thickness given.
	 *
	 * @param color     the color for the border
	 * @param thickness the thickness of the border
	 */
	public LyJBorderPanel(Color color, int thickness)
	{
		this();

		_style = LyArea.Border.SOLID;
		_color = color;
		_thickness = thickness;
		updateBorder();
	}

	/**
	 * Class constructor of a new JBorder Panel object. Makes a border of the given style with the
	 * default thickness for that style.
	 *
	 * @param style the style for this border. May be one of the following:
	 *
	 *              <ul>
	 *                <li>{@link #SOLID}</li>
	 *                <li>{@link #RAISED}</li>
	 *                <li>{@link #LOWERED}</li>
	 *                <li>{@link #IN}</li>
	 *                <li>{@link #OUT}</li>
	 *                <li>{@link #NONE}</li>
	 *              </ul>
	 */
	public LyJBorderPanel(LyArea.Border style)
	{
		this();

		// set thickness appropriate to this style
		int thickness;

		switch (style)
		{
		case SOLID:
			thickness = DEFAULT_SOLID_THICKNESS;
			break;
		case RAISED:
			thickness = DEFAULT_RAISED_THICKNESS;
			break;
		case LOWERED:
			thickness = DEFAULT_LOWERED_THICKNESS;
			break;
		case IN:
			thickness = DEFAULT_IN_THICKNESS;
			break;
		case OUT:
			thickness = DEFAULT_OUT_THICKNESS;
			break;
		case NONE:
			thickness = 0;
			break;
		default:
			thickness = 0;
			break;
		}
		_gap = DEFAULT_GAP;
		_style = style;
		_thickness = thickness;
		updateBorder();
	}

	/**
	 * Class constructor of a new JBorder Panel object. Makes border with given style and thickness.
	 *
	 * @param style     the style for this border
	 * @param thickness the thickness for this border
	 * @see   #LyJBorderPanel(int)
	 */
	public LyJBorderPanel(LyArea.Border style, int thickness)
	{
		this();

		_style = style;
		_thickness = thickness;
		updateBorder();
	}

	/**
	 * Sets the style of the border.
	 *
	 * @param  style the new style
	 * @return this LyJBorderPanel object
	 * @see    #getStyle()
	 */
	public LyJBorderPanel setStyle(LyArea.Border style)
	{

		// set the style and re-layout the panel
		_style = style;
		updateBorder();
		// doLayout();
		revalidate();
		repaint(getBounds());

		return this;
	}

	/**
	 * Gets the style of the border.
	 *
	 * @return SOLID, RAISED, LOWERED, IN, OUT, or NONE
	 * @see    #setStyle(int)
	 */
	public LyArea.Border getStyle()
	{
		return _style;
	}

	/**
	 * Sets the border of this component depending on its style value.
	 */
	public void updateBorder()
	{
		int align;

		if (_alignment == CENTER)
			align = TitledBorder.CENTER;
		else if (_alignment == RIGHT)
			align = TitledBorder.RIGHT;
		else
			align = TitledBorder.LEFT;

		if (_closable)
		{
			// Remove every mouse listener
			MouseListener[]			mouseListeners = getMouseListeners();
			MouseMotionListener[]	mouseMotionListeners = getMouseMotionListeners();
			int						i, n1 = (mouseListeners == null ? 0 : mouseListeners.length),
				n2 = (mouseMotionListeners == null ? 0 : mouseMotionListeners.length);
			
			for (i = 0; i < n1; i++)
				removeMouseListener(mouseListeners[i]);
			
			for (i = 0; i < n2; i++)
				removeMouseMotionListener(mouseMotionListeners[i]);
			
			_LyJBorderPanelClosableBorder	border = new _LyJBorderPanelClosableBorder(_thickness, _thickness, _thickness, _thickness);
			
			setBorder(border);
			border.setBorderOpen(_open);
			addMouseMotionListener(border);
			addMouseListener(border);
		}
		else
		{
			switch (_style)
			{
			case IN:
			{
				Border	b = BorderFactory.createCompoundBorder(
							BorderFactory.createEmptyBorder(_thickness, _thickness, _thickness,
								_thickness), BorderFactory.createEtchedBorder(EtchedBorder.LOWERED));
				
				if ((_text != null) && (_text.length() > 0))
					setBorder(BorderFactory.createTitledBorder(
							b, _text, align, TitledBorder.DEFAULT_POSITION, _font, _color));
				else
					setBorder(b);
				
				break;
			}

			case OUT:
			{
				setBorder(BorderFactory.createTitledBorder(
						BorderFactory.createCompoundBorder(
							BorderFactory.createEtchedBorder(EtchedBorder.RAISED),
							BorderFactory.createEmptyBorder(_gap, _gap, _gap, _gap)), _text, align,
						TitledBorder.DEFAULT_POSITION, _font, _color));
				break;
			}

			case SOLID:
			{
				int t = _thickness;

				setBorder(BorderFactory.createCompoundBorder(
						BorderFactory.createMatteBorder(t, t, t, t, _color),
						BorderFactory.createEmptyBorder(_gap, _gap, _gap, _gap)));
				break;
			}

			case RAISED:
			{
				setBorder(BorderFactory.createCompoundBorder(BorderFactory.createRaisedBevelBorder(),
						BorderFactory.createEmptyBorder(_gap, _gap, _gap, _gap)));
				break;
			}

			case LOWERED:
			{
				setBorder(BorderFactory.createCompoundBorder(BorderFactory.createLoweredBevelBorder(),
						BorderFactory.createEmptyBorder(_gap, _gap, _gap, _gap)));
				break;
			}

			case NONE:
				setBorder(BorderFactory.createEmptyBorder(0, 0, 0, 0));
			}
		}
	} // end method updateBorder

	/**
	 * Creates a reusable border for other components.
	 *
	 * @param  style the style of the border
	 * @return a new border for other components
	 */
	public static Border createBorder(LyArea.Border style)
	{
		// initialize data
		int    t = DEFAULT_THICKNESS;
		int    g = DEFAULT_GAP;
		Color  c = DEFAULT_COLOR;

		String txt = null;
		Font   f = DEFAULT_FONT;
		int    a = DEFAULT_ALIGNMENT;

		switch (style)
		{
		case SOLID:
			t = DEFAULT_SOLID_THICKNESS;
			break;
		case RAISED:
			t = DEFAULT_RAISED_THICKNESS;
			break;
		case LOWERED:
			t = DEFAULT_LOWERED_THICKNESS;
			break;
		case IN:
			t = DEFAULT_IN_THICKNESS;
			break;
		case OUT:
			t = DEFAULT_OUT_THICKNESS;
			break;
		case NONE:
			t = 0;
			break;
		default:
			t = 0;
			break;
		}
		g = 0;

		return createBorder(style, t, g, c, f, a, txt);
	} // end method createBorder

	/**
	 * Creates a reusable border for other components.
	 *
	 * @param  style     the style of the border
	 * @param  thickness the thickness
	 * @param  gap       the gap
	 * @param  color     the color of the border
	 * @return a new border for other components
	 */
	public static Border createBorder(LyArea.Border style, int thickness, int gap, Color color)
	{
		return createBorder(style, thickness, gap, color, DEFAULT_FONT, DEFAULT_ALIGNMENT, null);
	}

	/**
	 * This method allows another component to have the same border's style than a LyBorderPanel.
	 *
	 * @param  style     the style of the border
	 * @param  thickness the thickness
	 * @param  gap       the gap
	 * @param  color     the color of the border
	 * @param  font      the font to use
	 * @param  alignment the alignment policy
	 * @param  text      the text of the title
	 * @return a new border for other components
	 */
	public static Border createBorder(LyArea.Border style, int thickness, int gap, Color color, Font font,
		int alignment, String text)
	{
		int align;

		if (alignment == CENTER)
			align = TitledBorder.CENTER;
		else if (alignment == RIGHT)
			align = TitledBorder.RIGHT;
		else
			align = TitledBorder.LEFT;

		switch (style)
		{
		case IN:
			return (BorderFactory.createTitledBorder(
					BorderFactory.createEtchedBorder(EtchedBorder.LOWERED), text, align,
					TitledBorder.DEFAULT_POSITION, font, color));

		case OUT:
			return (BorderFactory.createTitledBorder(
					BorderFactory.createEtchedBorder(EtchedBorder.RAISED), text, align,
					TitledBorder.DEFAULT_POSITION, font, color));

		case SOLID:
		{
			int t = thickness;

			return (BorderFactory.createCompoundBorder(
					BorderFactory.createMatteBorder(t, t, t, t, color),
					BorderFactory.createEmptyBorder(gap, gap, gap, gap)));
		}

		case RAISED:
			return (BorderFactory.createCompoundBorder(BorderFactory.createRaisedBevelBorder(),
					BorderFactory.createEmptyBorder(gap, gap, gap, gap)));

		case LOWERED:
			return (BorderFactory.createCompoundBorder(BorderFactory.createLoweredBevelBorder(),
					BorderFactory.createEmptyBorder(gap, gap, gap, gap)));

		default:
			return (BorderFactory.createEmptyBorder(0, 0, 0, 0));
		}
	} // end method createBorder

	/**
	 * Sets the thickness of the border.
	 *
	 * @param  thickness the new thickness
	 * @return this LyJBorderPanel object
	 * @see    #getThickness()
	 */
	public LyJBorderPanel setThickness(int thickness)
	{
		if (thickness >= 0)
		{
			_thickness = thickness;
			updateBorder();
			// doLayout();
			revalidate();
			repaint();
		}

		return this;
	}

	/**
	 * Gets the thickness of the border.
	 *
	 * @return the thickness used for this border
	 * @see    #setThickness(int)
	 */
	public int getThickness()
	{
		return _thickness;
	}
	
	/**
	 * Sets the closable state of the panel.
	 *
	 * @param  closable the new closable state
	 */
	public void setClosable(boolean closable)
	{
		_closable = closable;
		updateBorder();
	}
	
	/**
	 * Sets the images used to open / close the panel.
	 * Only used when panel is closable.
	 * 
	 * @param openImage  the image to be used to open panel
	 * @param closeImage the image to be used to close panel
	 * @see #setClosable()
	 */
	public void setOpenCloseImages(Image openImage, Image closeImage)
	{
		_openImage = openImage;
		_closeImage = closeImage;
	}
	
	/**
	 * Sets the open state of the panel.
	 * Only used when panel is closable.
	 * 
	 * @param open the new open state
	 * @see #setClosable()
	 */
	public void setOpen(boolean open)
	{
		_open = open;

		Border	border = getBorder();
		
		if (border instanceof _LyJBorderPanelClosableBorder)
			((_LyJBorderPanelClosableBorder)border).setBorderOpen(_open);
	}
	
	/**
	 * Sets the parent LyView instance of this panel.
	 */
	public void setParentView(LyView view)
	{
		_parentView = view;
	}

	/**
	 * Sets the gap between the border and the contained Component.
	 *
	 * @param  gap the new gap, in pixels
	 * @return this LyJBorderPanel object
	 * @see    #getGap()
	 */
	public LyJBorderPanel setGap(int gap)
	{
		if (gap > -1)
		{
			this._gap = gap;
			updateBorder();
			// doLayout();
			revalidate();
			repaint();
		}

		return this;
	}

	/**
	 * Gets the gap between the border and the contained Component.
	 *
	 * @return the gap for this border
	 * @see    #setGap(int)
	 */
	public int getGap()
	{
		return _gap;
	}

	/**
	 * Sets the current color for SOLID borders and the caption text color for etched borders.
	 *
	 * @param  color the new color
	 * @return this LyJBorderPanel object
	 * @see    #getColor()
	 */
	public LyJBorderPanel setColor(Color color)
	{
		this._color = color;
		if ((_style == LyArea.Border.SOLID)
		        || (_style == LyArea.Border.IN)
                || (_style == LyArea.Border.OUT))
			updateBorder();

		return this;
	}

	/**
	 * Gets the current color for SOLID borders and the caption text color for etched borders.
	 *
	 * @return the color used for borders
	 * @see    #setColor(Color)
	 */
	public Color getColor()
	{
		return _color;
	}

	/**
	 * Sets the font. Only applies to etched borders.
	 *
	 * @param  font the font for the etched borders
	 * @return this LyJBorderPanel object
	 * @see    #getTextFont()
	 */
	public LyJBorderPanel setTextFont(Font font)
	{
		// set font
		if (font != null)
		{
			_font = font;
			if ((_style == LyArea.Border.IN) || (_style == LyArea.Border.OUT))
				updateBorder();
		}

		return this;
	}

	/**
	 * Gets the font of the text. Only applies to etched borders.
	 *
	 * @return the font of the text
	 * @see    #setTextFont(Font)
	 */
	public Font getTextFont()
	{
		return _font;
	}

	/**
	 * Sets the text. Only applies to etched borders.
	 *
	 * @param  text the new text
	 * @return this LyJBorderPanel object
	 * @see    #getText()
	 */
	public LyJBorderPanel setText(String text)
	{
		_text = text;
		updateBorder();
		return this;
	}

	/**
	 * Gets the text. Only applies to etched borders.
	 *
	 * @return the text
	 * @see    #setText(String)
	 */
	public String getText()
	{
		return _text;
	}

	/**
	 * Sets the text alignment. Only applies to etched borders.
	 *
	 * @param  alignment the new alignment
	 * @return this LyJBorderPanel object
	 * @see    #getAlignment()
	 */
	public LyJBorderPanel setAlignment(int alignment)
	{
		_alignment = alignment;
		return this;
	}

	/**
	 * Gets the text alignment.
	 *
	 * @return the text alignment
	 * @see    #setAlignment(int)
	 */
	public int getAlignment()
	{
		return _alignment;
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#getPreferredSize()
	 */
	public Dimension getPreferredSize()
	{
		Dimension	result = super.getPreferredSize();

		if (_closable && !_open)
			result.height = 10;
		
		return result;
	}
	
	public Rectangle getBounds()
	{
		Rectangle	result = super.getBounds();

		if (_closable && !_open)
			result.height = 10;
		
		return result;
	}

	/**
	 * Internal class used to draw a closable border.
	 */
	public class _LyJBorderPanelClosableBorder extends EmptyBorder
		implements MouseMotionListener, MouseListener
	{
		public static final int		IMAGE_OFFSET = 2;
		public static final int		TEXT_OFFSET = 7;
		public static final int		SPACING = 2;
		
		private static final long serialVersionUID = 1697271595703373731L;

		private Rectangle				_closeRect;
		private Dimension				_panelOpenDim;
		private ArrayList<Component>	_components;
		private ArrayList<Object>		_componentsInfo;

		public _LyJBorderPanelClosableBorder(int top, int left, int bottom, int right)
		{
			super(top, left, bottom, right);

			_closeRect = null;
			_panelOpenDim = null;
			_components = null;
		}

	    /**
	     * Returns the insets of the border.
	     */
		public Insets getBorderInsets(Component c)
	    {
			Insets		insets = new Insets(2 * SPACING + 1, 2 * SPACING + 1, 2 * SPACING + 1, 2 * SPACING + 1);

			if (c == null || getText() == null || getText().equals(""))
				return insets;
			
	        Font		font = getTextFont();
	        FontMetrics	fm = c.getFontMetrics(font);

			insets.top += fm.getAscent() + fm.getDescent() + SPACING;

			if (_closable && !_open)
				insets.top /= 2;
			
	        return insets;
	    }

		/*
		 * (non-Javadoc)
		 * @see javax.swing.border.Border#getBorderInsets(java.awt.Component)
		 */
		public void paintBorder(Component c, Graphics g, int x, int y, int width, int height)
		{
			String aaText = System.getProperty("swing.aatext");

			// set antialias
			if ("true".equalsIgnoreCase(aaText) && (g instanceof Graphics2D))
				((Graphics2D)g).setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
					RenderingHints.VALUE_TEXT_ANTIALIAS_ON);

	        Font		font = getTextFont();
	        FontMetrics	fm = c.getFontMetrics(font);
	        int         ascent = fm.getAscent();
	        Point		textLoc = new Point(x + TEXT_OFFSET + SPACING, y + SPACING + ascent);
	        Rectangle	borderRect = new Rectangle(x + SPACING, textLoc.y - ascent/2, width - 2 * SPACING - 2, height - textLoc.y + ascent/2 - SPACING - 2);

			// Draw border
			g.setColor(c.getBackground().darker());
			
			if (!_closable || _open)
			{
				g.drawLine(borderRect.x, borderRect.y, borderRect.x, borderRect.y + borderRect.height);
				g.drawLine(borderRect.x, borderRect.y + borderRect.height, borderRect.x + borderRect.width, borderRect.y + borderRect.height);
				g.drawLine(borderRect.x + borderRect.width, borderRect.y + borderRect.height, borderRect.x + borderRect.width, borderRect.y);
			}
			
			g.drawLine(borderRect.x + borderRect.width, borderRect.y, textLoc.x + fm.stringWidth(getText()) + SPACING, borderRect.y);
			g.drawLine(textLoc.x - SPACING, borderRect.y, borderRect.x, borderRect.y);
			
			g.setColor(c.getBackground().brighter());

			if (!_closable || _open)
			{
				g.drawLine(borderRect.x + 1, borderRect.y + 1, borderRect.x + 1, borderRect.y + borderRect.height - 1);
				g.drawLine(borderRect.x, borderRect.y + borderRect.height + 1, borderRect.x + borderRect.width + 1, borderRect.y + borderRect.height + 1);
				g.drawLine(borderRect.x + borderRect.width + 1, borderRect.y + borderRect.height, borderRect.x + borderRect.width + 1, borderRect.y);
			}

			g.drawLine(borderRect.x + borderRect.width - 1, borderRect.y + 1, textLoc.x + fm.stringWidth(getText()) + SPACING, borderRect.y + 1);
			g.drawLine(textLoc.x - SPACING, borderRect.y + 1, borderRect.x + 1, borderRect.y + 1);
			
			// Draw text
			int			imageWidth = _closeImage.getWidth(c);
			int			imageHeight = _closeImage.getHeight(c);
			Shape		oldClip = g.getClip();
			Rectangle	r = new Rectangle(x, y, width - 3 * SPACING - imageWidth - IMAGE_OFFSET, height);

			if (oldClip instanceof Rectangle)
				r = r.intersection((Rectangle)oldClip);
			
			g.setClip(r);
			g.setColor(getForeground());
			g.setFont(getTextFont());
			g.drawString(getText(), textLoc.x, textLoc.y);
			g.setClip(oldClip);
			
			// Draw image
			if (width > imageWidth + 2 * SPACING + IMAGE_OFFSET)
			{
				if (_open)
					g.drawImage(_closeImage, borderRect.x + borderRect.width - imageWidth - IMAGE_OFFSET, y, c);
				else
					g.drawImage(_openImage, borderRect.x + borderRect.width - imageWidth - IMAGE_OFFSET, y, c);
				
				_closeRect = new Rectangle(borderRect.x + borderRect.width - imageWidth - IMAGE_OFFSET, y, imageWidth, imageHeight);
			}
		}

		public void mouseDragged(MouseEvent e)
		{
		}

		public void mouseMoved(MouseEvent e)
		{
			if ((_closeRect != null) && _closeRect.contains(e.getPoint()))
				setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
			else
				setCursor(null);
		}
		
		public void setBorderOpen(boolean open)
		{
			LayoutManager	layoutManager = getLayout();
			
			if (layoutManager instanceof MigLayout)
			{
				MigLayout		layout = (MigLayout)layoutManager;
				
				if (!open)
				{
					_components = new ArrayList<Component>();
					_componentsInfo = new ArrayList<Object>();
					
					int		i, n = getComponentCount();
					
					for (i = n - 1; i >= 0; i--)
					{
						Component	c = getComponent(i);
						Object		layoutInfo = layout.getComponentConstraints(c);
						
						_components.add(c);
						_componentsInfo.add(layoutInfo);
						remove(c);
					}
				}
				else if (_components != null)
				{
					int		i, n = _components.size();
					
					for (i = n - 1; i >= 0; i--)
					{
						_components.add(_components.get(i));
						add(_components.get(i), _componentsInfo.get(i));
					}

					_components.clear();
					_components = null;
					_componentsInfo.clear();
					_componentsInfo = null;
				}
			}
		}

		public void mouseClicked(MouseEvent e)
		{
			if ((_closeRect != null) && _closeRect.contains(e.getPoint()))
			{
				if ((_panelOpenDim == null) && _open)
					_panelOpenDim = getPreferredSize();

				_open = !_open;
				setBorderOpen(_open);
				
				if ((_panelOpenDim == null) && _open)
					_panelOpenDim = getPreferredSize();

				Rectangle		bounds = getBounds();
				int				newHeight = (_open ? _panelOpenDim.height : 20);

				if (_open)
				{
					setPreferredSize(new Dimension(_panelOpenDim.width, newHeight));
					setBounds(bounds.x, bounds.y, _panelOpenDim.width, newHeight);
				}
				else
				{
					setPreferredSize(new Dimension(newHeight, newHeight));
					setBounds(bounds.x, bounds.y, newHeight, newHeight);
				}

				getParent().getParent().getParent().validate();
				getParent().getParent().getParent().doLayout();
				
				if (_parentView != null)
				{
					// If border panel is contained in a splitter, splitter component must perform a
					// layout to ensure a correct disposition of its components.
					Vector<LySplitter>	splitters = new Vector<LySplitter>();

					_parentView.getComponents(LySplitter.class, splitters);
					
					Container			container = null;

					for (LySplitter splitter : splitters)
					{
						LySwingComponent	component = (LySwingComponent)splitter.getInternalView();
						
						container = (Container)component.getComponent();
						
						container.validate();
						container.doLayout();
						container.validate();
						container.doLayout();
					}
				}
				
				repaint();
				
				if (_parentView != null)
				{
					// Reset content only if parent is not a resizable window (otherwise the window
					// may flash too much) or not a dialog with a given width or height
					if ((_parentView instanceof LyFrame) && (((LyFrame)_parentView).isResizable()))
						return;

					if ((_parentView instanceof LyDialog) && (((LyDialog)_parentView).isResizable()))
						return;

					if ((_parentView instanceof LyDialog)
							&& ((((LyDialog)_parentView).getWidth() > 0) || ((((LyDialog)_parentView).getHeight() > 0))))
						return;
					
					if (((LyView)_parentView).isVisible())
					{
						_parentView.getViewManager().execGui(new Runnable() {
							
							public void run()
							{
								Object		internal = _parentView.getInternalView();

								if (internal != null)
									((LyResetContentInterface)internal).resetContent();
							}
						}, false);
					}
					return;
				}
			}
		}

		public void mouseEntered(MouseEvent e)
		{
		}

		public void mouseExited(MouseEvent e)
		{
		}

		public void mousePressed(MouseEvent e)
		{
		}

		public void mouseReleased(MouseEvent e)
		{
		}
	}
} // end class LyJBorderPanel
