/*
 * LyJButton.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.net.*;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.plaf.*;
import javax.swing.plaf.metal.*;
import leon.misc.*;
import leon.view.filters.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;


/**
 * This class implements a simple label with an image.
 *
 * @version $Id: LyJButton.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class LyJButton extends JButton
	implements ActionListener,
		MouseListener,
		LyGetIdInterface,
		LyEventTransmitter,
		PropertyChangeListener
{
	private static final long     serialVersionUID = 2435746627065353703L;

	public static final String    __VERSION =
		"$Id: LyJButton.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The maximum number of characters of this button tooltip. */
	public static int			  TIP_MAX_CHARS = 60;

	/** Color of the focus for the default look and feel. */
	public static Color			  FOCUS_COLOR = null;

	/** Label of the button. */
	protected String			  _label; // the

	/** Font of the button. */
	protected Font				  _font;

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter _actionTransmitter;

	/** Identifier of this button. */
	protected String			  _id;

	/** Indicates if the button actions have to be repeated or not. */
	protected boolean			  _repeatEnabled;
	
	/** Environment of the button. */
	protected LyEnvironment		  _environment;

	/** The button repetitor used with {@link #_repeatEnabled}. */
	protected _LyJButtonRepetitor _repetitor;

	/** Indicates if rollover images are used. */
	protected boolean			  _rollover;

	/** Border of the button. */
	protected Border			  _border;
	
	/** The gray proportion of the image used when mouse is not over the button. */
	protected int				  _grayProp;
	
	/** Indicates if background must be changed when mouse is over the button. */
	protected boolean			  _rolloverChangeBackground;
	
	/** Opener button associated with current one. */
	protected LyJOpenerButton	  _openerButton;

	/**
	 * Workaround for tooltips.
	 *
	 * @see #getToolTipLocation(MouseEvent)
	 */
	protected long				  _lastToolTipDate;

	/**
	 * Indicates if this button component can be painted.
	 *
	 * @deprecated
	 */
	boolean						  _painted = false;

	/**
	 * Creates a new Button component with the specified options. Initializes the class fields to
	 * the specified values, or to the default values if not specified.
	 *
	 * <p>This button is double buffered.</p>
	 *
	 * @param label the label of this button
	 */
	public LyJButton(String label)
	{
		this((Image)null, label, null, null);
	}

	/**
	 * Creates a new Button component with the specified options. Initializes the class fields to
	 * the specified values, or to the default values if not specified.
	 *
	 * <p>This button is double buffered.</p>
	 *
	 * @param image the image to associate to the button
	 * @param label the label of this button
	 */
	public LyJButton(String label, Image image)
	{
		this(image, label, null, null);
	}

	/**
	 * Creates a new Button component with the specified options. Initializes the class fields to
	 * the specified values, or to the default values if not specified.
	 *
	 * <p>This button is double buffered.</p>
	 *
	 * @param image    the image to associate to the button
	 * @param label    the label of this button
	 * @param rollover the rollover property
	 */
	public LyJButton(String label, Image image, boolean rollover)
	{
		this(image, label, null, null, rollover);
	}

	/**
	 * Creates a new Button component with the specified options. Initializes the class fields to
	 * the specified values, or to the default values if not specified.
	 *
	 * <p>This button is double buffered.</p>
	 *
	 * @param image the image to associate to the button
	 */
	public LyJButton(Image image)
	{
		this(image, null, null, null);
	}

	/**
	 * Creates a new Button component with the specified options. Initializes the class fields to
	 * the specified values, or to the default values if not specified.
	 *
	 * <p>This button is double buffered.</p>
	 *
	 * @param image   the image to associate to the button
	 * @param tooltip the tooltip
	 */
	public LyJButton(Image image, String tooltip)
	{
		this(image, null, tooltip, null);
	}

	/**
	 * Creates a new Button component with the specified options. Initializes the class fields to
	 * the specified values, or to the default values if not specified.
	 *
	 * <p>This button is double buffered.</p>
	 *
	 * @param image   the image to associate to the button
	 * @param label   the label of this button
	 * @param tooltip the tooltip
	 * @param id      the identifier string
	 */
	public LyJButton(Image image, String label, String tooltip, String id)
	{
		this(image, label, tooltip, id, true);
	}

	/**
	 * Creates a new Button component with the specified options. Initializes the class fields to
	 * the specified values, or to the default values if not specified.
	 *
	 * <p>This button is double buffered.</p>
	 *
	 * @param url     the image url to associate to the button
	 * @param image   the image to associate to the button
	 * @param label   the label of this button
	 * @param tooltip the tooltip
	 * @param id      the identifier string
	 */
	public LyJButton(URL url, Image image, String label, String tooltip, String id)
	{
		this(null, label, tooltip, id, true);
		
		setImage(url, image);
	}

	/**
	 * Creates a new Button component with the specified options. Initializes the class fields to
	 * the specified values, or to the default values if not specified.
	 *
	 * <p>This button is double buffered.</p>
	 *
	 * @param image    the image to associate to the button
	 * @param label    the label of this button
	 * @param tooltip  the tooltip
	 * @param id       the identifier string
	 * @param rollover the rollover property
	 */
	public LyJButton(Image image, String label, String tooltip, String id, boolean rollover)
	{
		super();
		_id = id;
		_label = label;
		_border = null;
		_rolloverChangeBackground = false;
		_environment = null;
		setModel(new _LyJButtonModel());
		setText(_label);
		setRollover(rollover);
		setImage(image);
		setToolTipText(LyTools.formatToHTML(tooltip, LyJButton.TIP_MAX_CHARS));
		setBackground(null);
		init();

		setDoubleBuffered(true);
		// addPropertyChangeListener(this);
	}

	/**
	 * Registers the action and mouse listeners of this button.
	 */
	public void init()
	{
		_repeatEnabled = false;
		addActionListener(this);
		addMouseListener(this);

		// Usually Swing uses SPACE to validate buttons, lets register ENTER to do the validation
		InputMap  inputMap = getInputMap();
		ActionMap actionMap = getActionMap();
		KeyStroke stroke = KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0);
		String    actionName = "returnPressed";

		inputMap.put(stroke, actionName);

		Action action = new AbstractAction(actionName)
			{
				private static final long serialVersionUID = -3378141864744313522L;

				/*
				 * (non-Javadoc)
				 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
				 */
				public void actionPerformed(ActionEvent e)
				{
					callActionPerformed(e);
				}
			};

		actionMap.put(actionName, action);
		/*
		 * KeyStroke [] keys = getInputMap().allKeys(); System.out.println("KEYS : "+keys.length);
		 * for(int i=0;i<keys.length;i++) { System.out.println("key : "+keys[i]+" -> action :
		 * "+getInputMap().get(keys[i])); } System.out.println();
		 */
	} // end method init

	/**
	 * Sets the rollover property of this boolean.
	 *
	 * <p>Workaround for Swing UI bug where rollover deactivates itself.</p>
	 *
	 * @param roll the rollover property
	 * @see   #setRolloverEnabled(boolean)
	 */
	public void setRollover(boolean roll)
	{
		_rollover = roll;
		setRolloverEnabled(roll);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.swing.AbstractButton#setRolloverEnabled(boolean)
	 */
	public void setRolloverEnabled(boolean roll)
	{
		super.setRolloverEnabled(_rollover);
	}
	
	/**
	 * Indicates if background must be changed when mouse is over the button.
	 * 
	 * @param rollover Indicates if the change background mode is activated.
	 */
	public void setRolloverChangeBackground(boolean rollover)
	{
		_rolloverChangeBackground = rollover;
	}
	
	/**
	 * Sets the gray proportion of the image used when mouse is not over the button.
	 * 
	 * @param grayProp Indicates the new gray proportion.
	 */
	public void setGrayProp(int grayProp)
	{
		_grayProp = grayProp;
	}

	/**
	 * Sets the image displayed by the button. If the button is in rollover mode, the icon is
	 * grayed.
	 *
	 * @param image the new icon image
	 * @see   #getImage()
	 */
	public void setImage(Image image)
	{
		if (image == null)
			return;
		
		ImageIcon trueIcon = new ImageIcon(image);
		ImageIcon grayIcon = new ImageIcon(LyGrayImageFilter.getFilteredImage(image, _grayProp));
		ImageIcon disabledIcon = new ImageIcon(LyDisableImageFilter.getFilteredImage(image));

		if (_rollover)
			setIcon(grayIcon);
		else
			setIcon(trueIcon);

		setRolloverIcon(trueIcon);
		setRolloverSelectedIcon(trueIcon);
		setDisabledIcon(disabledIcon);
		setDisabledSelectedIcon(disabledIcon);
	}

	/**
	 * Sets the image displayed by the button. If the button is in rollover mode, the icon is
	 * grayed.<br />
	 * This method must be used when image is animated.
	 *
	 * @param image the new icon image
	 */
	public void setImage(URL url, Image image)
	{
		if ((url == null) || (image == null))
			return;
		
		ImageIcon	trueIcon = new ImageIcon(url);
		ImageIcon	disabledIcon = new ImageIcon(LyDisableImageFilter.getFilteredImage(image));

		setDisabledIcon(disabledIcon);
		setDisabledSelectedIcon(disabledIcon);
		setIcon(trueIcon);
		setRolloverIcon(trueIcon);
		setRolloverSelectedIcon(trueIcon);
	}

	/**
	 * Gets the image of this button.
	 *
	 * @return the image displayed by this button
	 * @see    #setImage(Image)
	 */
	public Image getImage()
	{
		return ((ImageIcon)getRolloverIcon()).getImage();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#setFont(java.awt.Font)
	 */
	public void setFont(Font font)
	{
		if (font == null)
			return;
		_font = font;
		super.setFont(font);
	}

	/**
	 * Activates or deactivates repeat mode.
	 *
	 * @param repeat Indicates if the repeat mode is activated.
	 */
	public void setRepeatEnabled(boolean repeat)
	{
		_repeatEnabled = repeat;
	}

	/**
	 * Sets the environment of the button.
	 *
	 * @param environment the environment of the button.
	 */
	public void setEnvironment(LyEnvironment environment)
	{
		_environment = environment;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#setEnabled(boolean)
	 */
	public void setEnabled(boolean b)
	{
		if ((!b) && hasFocus())
			transferFocus();
		super.setEnabled(b);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#hasFocus()
	 */
	public boolean hasFocus()
	{
		if (!isEnabled())
			return false;
		return super.hasFocus();
	}

	/**
	 * Performes the specified action. Called when the button has been activated by return.
	 *
	 * @param e an action event
	 */
	public void callActionPerformed(ActionEvent e)
	{
		actionPerformed(e);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent e)
	{
		LyActionEvent actionEvent = new LyActionEvent(this);

		actionEvent.setValue(_id);
		
		if ((e.getModifiers() & InputEvent.CTRL_MASK) != 0)
			actionEvent.setAdditionalValue("CTRL");

		if (_actionTransmitter != null)
		{
			if (_repetitor == null)
				_actionTransmitter.processActionEvent(actionEvent);
			else
			{
				if (!_repetitor.hasSentEvent())
					_actionTransmitter.processActionEvent(actionEvent);
				else
				{
					actionEvent = new LyActionEvent(this, null, _id,
							LyActionDemand.TYPE_VALUE_CHANGED);
					
					_actionTransmitter.processActionEvent(actionEvent);
				}
			}
		}
	}

	/**
	 * Used to manage the repeat mode.
	 *
	 * @see MouseListener#mousePressed(java.awt.event.MouseEvent)
	 */
	public void mousePressed(MouseEvent e)
	{
		if (_repeatEnabled && (e.getModifiers() == InputEvent.BUTTON1_MASK) && isEnabled())
		{
			_repetitor = new _LyJButtonRepetitor(this);
			_repetitor.repeatAction();
		}
	}

	/**
	 * Used to manage the repeat mode.
	 *
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	public void mouseReleased(MouseEvent e)
	{
		if (_rolloverChangeBackground)
			setMouseOverStyle(false, true);

		if (_repetitor != null)
		{
			_repetitor.stopRepeat();
			_repetitor = null;
		}

		((_LyJButtonModel)getModel()).setRollover(false);
		repaint();
	}

	/**
	 * Used to manage the repeat mode.
	 *
	 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
	 */
	public void mouseClicked(MouseEvent e)
	{
		if (_rolloverChangeBackground)
			setMouseOverStyle(false, true);

		if (_repetitor != null)
		{
			_repetitor.stopRepeat();
			_repetitor = null;
		}

		((_LyJButtonModel)getModel()).setRollover(false);
		repaint();
	}

	/**
	 * Sets the mouse over style for current button.<br />
	 * The mouse over style draws a border around the button and fill its area.
	 * Only works with Metal look and feel.
	 * 
	 * @param over indicates if the mouse over style must be set or unset.
	 * @param propagate indicates if the style must be propagated to opener button.
	 */
	protected void setMouseOverStyle(boolean over, boolean propagate)
	{
		// Use mouse over style for Metal look and feel only.
		if (!"Metal".equals(UIManager.getLookAndFeel().getID()))
			return;

		if (over)
		{
			_border = getBorder();			
			setContentAreaFilled(true);
			setBorder(getMouseOverBorder());
			setOpaque(true);
		}
		else
		{
			setContentAreaFilled(false);
			setBorder(_border);
			setOpaque(false);
		}
		
		if (propagate && (_openerButton != null))
		{
			((_LyJButtonModel)_openerButton.getModel()).setRollover(over);
			_openerButton.setMouseOverStyle(over, false);
		}
	}
	
	/**
	 * Sets the opener button associated with current one.
	 * 
	 * @param openerButton the new opener button associated with current one.
	 */
	public void setOpenerButton(LyJOpenerButton openerButton)
	{
		_openerButton = openerButton;
	}
	
	/**
	 * Return the border to be used when mouse over style is set.
	 * 
	 * @return the border for current button.
	 */
	protected Border getMouseOverBorder()
	{
		return BorderFactory.createMatteBorder(1, 1, 1, 1, (Color)UIManager.get("Button.darkShadow"));
	}
	
	/**
	 * Compacts toobars.
	 *
	 * @return smaller insets
	 * @see    JComponent#getInsets()
	 */
	public Insets getInsets()
	{
		if ((_label == null) || (_label.length() == 0))
		{
			if (!isBorderPainted())
				return new Insets(0, 1, 0, 1);
			return new Insets(4, 4, 5, 5);
		}
		return super.getInsets();
	}

	/**
	 * Doesn't do anything.
	 *
	 * @see MouseListener#mouseEntered(java.awt.event.MouseEvent)
	 */
	public void mouseEntered(MouseEvent e)
	{
		if (_rolloverChangeBackground && isEnabled())
			setMouseOverStyle(true, true);
	}

	/**
	 * Used to manage the repeat mode.
	 *
	 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
	 */
	public void mouseExited(MouseEvent e)
	{
		if (_rolloverChangeBackground)
			setMouseOverStyle(false, true);

		if (_repetitor != null)
		{
			_repetitor.stopRepeat();
			_repetitor = null;
		}

		((_LyJButtonModel)getModel()).setRollover(false);
		repaint();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return (_actionTransmitter);
	}

	/**
	 * This button doesn't handle selection events, so it has no selection transmitter.
	 *
	 * @return <code>null</code>
	 * @see    leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		return null;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.getinterface.LyGetIdInterface#getId()
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Prints the content of this button object.
	 *
	 * @return a string representation of this button
	 */
	public String toString()
	{
		String result = "";

		result += getClass().toString();
		result += hashCode() + "; ";
		result += "Rollover = " + isRolloverEnabled() + "; ";
		result += "Icon = " + getIcon() + "; ";
		result += "RollOverIcon = " + getRolloverIcon();
		return result;
	}

	/**
	 * Generates a warning message on the standard output.
	 *
	 * @see java.beans.PropertyChangeListener#propertyChange(java.beans.PropertyChangeEvent)
	 */
	public void propertyChange(PropertyChangeEvent e)
	{
		System.out.println(e.getPropertyName() + " " + e.getOldValue() + " " + e.getNewValue()
			+ "\n");
	}

	/**
	 * public void firePropertyChange(String name, Object oldO, Object newO) { Thread.dumpStack();
	 * super.firePropertyChange(name,oldO,newO); }.
	 */

	/**
	 * .
	 */

	/**
	 * Repeats the action on the button while pressed.
	 */
	class _LyJButtonRepetitor extends Thread
	{

		/** The runnable instance sent to AWT event queue. */
		_LyJButtonEvent	   _buttonEvent;

		/** Indicates the state of this repetitor. */
		boolean			   _running;

		/** Indicates if at least one event has been sent. */
		boolean			   _eventSent;

		/**
		 * Creates a new repetitor thread for the specified button.
		 *
		 * @param button the associated button repeated.
		 */
		public _LyJButtonRepetitor(LyJButton button)
		{
			_buttonEvent = new _LyJButtonEvent(button);
			_eventSent = false;
			
			this.setName("BF AppEngine : Swing button repetitor thread");
		}

		/**
		 * Repeats the action by starting this Thread.
		 */
		public void repeatAction()
		{
			_running = true;
			start();
		}

		/**
		 * Stops this thread.
		 */
		public void stopRepeat()
		{
			_running = false;
			_buttonEvent.stopEvent();
		}

		/**
		 * Gets the status of this Thread.
		 *
		 * @return true if this thread is running, false otherwise.
		 */
		public boolean isRunning()
		{
			return _running;
		}

		/**
		 * Indicates if at least one event has been sent.
		 *
		 * @return true if at least one event has been sent, false otherwise.
		 */
		public boolean hasSentEvent()
		{
			return _eventSent;
		}

		/*
		 * (non-Javadoc)
		 * @see java.lang.Runnable#run()
		 */
		public void run()
		{
			int initRepeater = 200;
			int timeToWait = 240;

			try
			{
				sleep(initRepeater);
			}
			catch (InterruptedException e)
			{
				e.printStackTrace();
			}

			while (_running)
			{
    			if (_running && (_environment != null))
    			{
    				LyViewManager   viewManager = (LyViewManager)_environment.getViewManager();
                    
    				viewManager.execGui(_buttonEvent);
    				_eventSent = true;
    			}

                try
				{
					sleep(timeToWait);

					// Accelerating the movement...
					if (timeToWait > 30)
						timeToWait -= 15;
				}
				catch (InterruptedException e)
				{
					e.printStackTrace();
				}
			}
		} // end method run
	} // end class _LyJButtonRepetitor
    
    class _LyJButtonEvent
        implements Runnable
    {

		/** The button that has this repetitor. */
		LyJButton _button = null;

		/** Indicates the state of this event. */
		boolean			   _running;
        
        public _LyJButtonEvent(LyJButton button)
        {
        	_button = button;
			_running = true;
        }
        
		/**
		 * Stops this event.
		 */
		public void stopEvent()
		{
			_running = false;
		}

        public void run()
        {
			if (_running)
			{
				LyActionEvent actionEvent = new LyActionEvent(_button, null, _id,
						LyActionDemand.TYPE_TIMEOUT);

				LyActionTransmitter transmitter = _button.getActionTransmitter();

				if (transmitter != null)
					transmitter.processActionEvent(actionEvent);
			}
        }
    }

	/**
	 * Deactivate multiple mouse enter on buttons.
	 *
	 * @see javax.swing.JComponent#getToolTipLocation(java.awt.event.MouseEvent)
	 */
	public Point getToolTipLocation(MouseEvent event)
	{
		// When the button is the last in a vertical toolbar, multiple mouseEnter events
		// lock the GUI
		// This was reported against JDK 1.3.1, seems Ok in JDK 1.4

		if (event.getID() == MouseEvent.MOUSE_ENTERED)
		{
			if (_lastToolTipDate > 0)
			{
				if ((System.currentTimeMillis() - _lastToolTipDate) < 400)
					event.translatePoint(-2000, 0);
			}
			_lastToolTipDate = System.currentTimeMillis();
		}

		return super.getToolTipLocation(event);
	}

	/**
	 * Called just before the button does an action. Overloaded to give the focus to the button.
	 *
	 * @see java.awt.Component#processMouseEvent(java.awt.event.MouseEvent)
	 */
	// FFT00447
	protected void processMouseEvent(MouseEvent e)
	{
		if (e.getID() == MouseEvent.MOUSE_PRESSED)
			requestFocus();
		super.processMouseEvent(e);
	}

	/**
	 * .
	 */

	/**
	 * This class is used as ButtonModel for LyJButton. The only change is that when the button is
	 * pressed it takes the focus. It needs to take focus to have textField lose its focus and so
	 * submit its value to the controller.
	 *
	 * @see ButtonModel
	 * @see AbstractButton#setModel(ButtonModel)
	 */
	public class _LyJButtonModel extends DefaultButtonModel
	{
		private static final long serialVersionUID = -2283165098102285155L;

		/**
		 * Default constructor for this button model object.
		 */
		public _LyJButtonModel()
		{
			super();
		}

		/**
		 * Called just before the button does an action. Overloaded to give the focus to the button.
		 *
		 * @see javax.swing.DefaultButtonModel#fireActionPerformed(java.awt.event.ActionEvent)
		 */
		// FFT00447
		protected void fireActionPerformed(ActionEvent e)
		{
			LyJButton.this.requestFocus();
			super.fireActionPerformed(e);
		}

		/*
		 * (non-Javadoc)
		 * @see javax.swing.ButtonModel#setRollover(boolean)
		 */
		public void setRollover(boolean b)
		{
			if ((isRollover() == b) || !this.isEnabled())
				return;
			
			if (b)
				stateMask |= ROLLOVER;
			else
				stateMask &= ~ROLLOVER;

			this.fireStateChanged();
		}
	} // end class _LyJButtonModel

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.swing.AbstractButton#setUI(javax.swing.plaf.ButtonUI)
	 */
	public void setUI(ButtonUI ui)
	{
		if ((FOCUS_COLOR != null) && (ui instanceof MetalButtonUI))
			ui = new MetalButtonUI()
				{

					/*
					 * (non-Javadoc)
					 * @see javax.swing.plaf.metal.MetalButtonUI#getFocusColor()
					 */
					public Color getFocusColor()
					{
						return FOCUS_COLOR;
					}
				};
		super.setUI(ui);
	}
} // end class LyJButton
