/*
 * LyJDateSelector.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;
import java.text.*;
import java.util.*;
import javax.swing.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;


/**
 * This class implements the calendar selection tool for swing viewer.
 *
 * @version $Id: LyJDateSelector.java 29006 2010-12-23 17:44:46Z bmichel $
 * @author  Lyria
 * @see     leon.view.win.swing.LySwingDate
 */
public class LyJDateSelector extends JPanel
	implements MouseListener,
		MouseMotionListener,
		LyGetValueInterface,
		LySetValueInterface,
		LyGetIdInterface,
		LyEventTransmitter
{
	private static final long     serialVersionUID = -4117676075864441090L;

	public static final String    __VERSION =
		"$Id: LyJDateSelector.java 29006 2010-12-23 17:44:46Z bmichel $";

	private static final String   DEFAULT_IDENTIFIER = "_dateSelector_";

	/** Separation space between days in the calendar representation. */
	public static float			  COL_SPACING = 0.75F;

	/** Identifier. */
	protected String			  _id;

	/** Value of the calendar component. */
	protected Calendar			  _currentDate;

	/** Internal calendar used for display only. */
	protected Calendar			  _currentDisplayDate;

	/** The date for today. */
	protected Calendar			  _today;

	/** Export format. */
	protected DateFormat		  _dateFormat;

	/** Title date format. */
	protected DateFormat		  _dateTitleFormat;

	/** Symbols for date ressources. */
	protected DateFormatSymbols   _dateFormatSymbols;

	/** Bold font. */
	protected Font				  _boldFont;

	/** Graphical areas sensitive for navigation. Used for optimization and duplication. */
	protected Rectangle[]		  _navigation;

	// Kept for optimization.
	/** White color. */
	protected Color				  _whiteColor;

	/** Grey color. */
	protected Color				  _greyColor;

	/** Dark grey color. */
	protected Color				  _darkGreyColor;

	/** Red color. */
	protected Color				  _redColor;

	/** Black color. */
	protected Color				  _blackColor;

	/** Yellow color. */
	protected Color				  _yellowColor;

	/** Blue color. */
	protected Color				  _blueColor;

	/** Last mouse position. */
	protected Point				  _mousePosition;

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter _actionTransmitter;

	/**
	 * Default date selector constructor.
	 *
	 * <p>The identifier used in this constructor is {@link #DEFAULT_IDENTIFIER}. No timezone or
	 * locale is used.</p>
	 */
	public LyJDateSelector()
	{
		this(DEFAULT_IDENTIFIER, null, null);
	}

	/**
	 * Date selector constructor.
	 *
	 * <p>The identifier used in this constructor is {@link #DEFAULT_IDENTIFIER}. No timezone is
	 * used.</p>
	 *
	 * @param loc the locale for the week data
	 */
	public LyJDateSelector(Locale loc)
	{
		this(DEFAULT_IDENTIFIER, null, loc);
	}

	/**
	 * Date selector constructor.
	 *
	 * @param id  the date selector identifier
	 * @param tz  the timezone of the date this component represents
	 * @param loc the locale for the week data
	 */
	public LyJDateSelector(String id, TimeZone tz, Locale loc)
	{
		super();

		_id = id;

		init(tz, loc);
	}

	/**
	 * Initializes the date selector with default values.
	 *
	 * @param tz  the timezone to use
	 * @param loc the locale for the week data
	 */
	protected void init(TimeZone tz, Locale loc)
	{
		if (tz == null)
			tz = TimeZone.getDefault();
		if (loc == null)
			loc = Locale.getDefault();

		// TO DO : use LOCALE
		_currentDate = Calendar.getInstance(tz, loc);
		_currentDisplayDate = Calendar.getInstance(tz, loc);
		_today = Calendar.getInstance(tz, loc);

		Date d = new Date();

		_currentDate.setTime(d);
		_currentDisplayDate.setTime(d);
		_today.setTime(d);

		_dateTitleFormat = new SimpleDateFormat("MMM yyyy", loc);

		_dateFormatSymbols = new DateFormatSymbols(loc);

		_navigation = new Rectangle[4];

		// Handle mouse interaction
		addMouseListener(this);
		addMouseMotionListener(this);

		if (_dateFormat == null)
			setDateFormat("dd/MM/yyyy");
	} // end method init

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#getPreferredSize()
	 */
	public Dimension getPreferredSize()
	{
		Graphics g = getGraphics();

		if (g == null)
			return super.getPreferredSize();

		Rectangle bounds = getBoundaries();

		return (new Dimension(bounds.width, bounds.height));
	}

	/**
	 * Returns the date selector boundaries depending on the character width and height.
	 *
	 * @return a Rectangle object with the following properties:
	 *
	 *         <ul>
	 *           <li>the x attribute is the characted width of the current font</li>
	 *           <li>the y attribute is the line height</li>
	 *           <li>the width attribute is the total width (width of the borders, width of the days
	 *             of the week and width of the separator between days)</li>
	 *           <li>the height attribute is the total height of the date selector (height of the
	 *             lines of weeks in the selector)</li>
	 *         </ul>
	 */
	protected Rectangle getBoundaries()
	{
		FontMetrics fm = getFontMetrics(getFont());

		int		    charWidth = fm.stringWidth("10") / 2;
		int		    lineHeight = fm.getHeight() + 3;
		int		    totalWidth = (int)(((2 * 2) // borders
					+ (7 * 2) // days in weeks
					+ (6 * COL_SPACING)) // sep between days
					* charWidth);
		int		    totalHeight = (8 * lineHeight) + 3;

		return (new Rectangle(charWidth, lineHeight, totalWidth, totalHeight));
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#update(java.awt.Graphics)
	 */
	public void update(Graphics g)
	{
		paint(g);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.swing.JComponent#paintComponent(java.awt.Graphics)
	 */
	public void paintComponent(Graphics g)
	{
		super.paintComponent(g);
		
		if (g == null)
			return;

		// Set anti-alias
		((Graphics2D)g).setRenderingHint(RenderingHints.KEY_ANTIALIASING,
			RenderingHints.VALUE_ANTIALIAS_ON);

		Dimension d = getSize();

		// Prepare colors
		if (_whiteColor == null)
			_whiteColor = Color.white;
		if (_greyColor == null)
			_greyColor = new Color(192, 192, 192);
		if (_darkGreyColor == null)
			_darkGreyColor = new Color(128, 128, 128);
		if (_redColor == null)
			_redColor = new Color(200, 0, 0);
		if (_blackColor == null)
			_blackColor = Color.black;
		if (_yellowColor == null)
			_yellowColor = Color.yellow;
		if (_blueColor == null)
			_blueColor = Color.blue;

			// ** Fill in the buffered image **

		// ### retreives Font metrics ###
		// KEEP Synchro with getPreferredSize()
		FontMetrics fm = getFontMetrics(g.getFont());

		Rectangle   bounds = getBoundaries();
		int		    charWidth = bounds.x;
		int		    lineHeight = bounds.y;

		// ### clear content ###
		// setBackground(_whiteColor);
		g.setColor(_whiteColor);
		g.fillRect(0, 0, d.width, d.height);

		// ### draw borders ###
		g.setColor(_greyColor);
		g.fillRect(0, 0, d.width, lineHeight + 2); // title background

		g.setColor(_darkGreyColor);
		g.drawRect(0, 0, d.width - 1, d.height - 1); // border bottom
		g.drawLine(0, 2 + lineHeight, d.width - 2, 2 + lineHeight); // title bottom line
		g.drawLine(2 * charWidth, 2 * lineHeight, d.width - (2 * charWidth), 2 * lineHeight); // line

		// ### draw title ###
		g.setColor(_blackColor);

		Font f = g.getFont();

		if (_boldFont == null)
			_boldFont = new Font(f.getName(), Font.BOLD, f.getSize());

		g.setFont(_boldFont);

		String title = _dateTitleFormat.format(_currentDisplayDate.getTime());
		// center title.
		int    x = (d.width - fm.stringWidth(title)) / 2;

		g.drawString(title, x, lineHeight - fm.getDescent() - 1);

		// ### Draw "marks" next / previous ###
		// check if 'maxdate'
		int WIDTH = 8;
		int BUTTON_OFFSET = 16;

		// Previous year
		{
			int     delta_fix = 1; // even if both polygons are the same size, the first
			// filleld look smaller due to (int) casting.

			Polygon p = new Polygon();

			if (_currentDisplayDate.get(Calendar.YEAR) > 1)
				g.setColor(_blackColor);
			else
				g.setColor(_darkGreyColor);

			p.addPoint(5, lineHeight / 2);
			p.addPoint((int)(5 + (WIDTH / Math.sqrt(2)) + delta_fix), (lineHeight + WIDTH) / 2);
			p.addPoint((int)(5 + (WIDTH / Math.sqrt(2)) + delta_fix), (lineHeight - WIDTH) / 2);
			g.fillPolygon(p);
			g.drawLine((int)(2 + 5 + (WIDTH / Math.sqrt(2)) + delta_fix), (lineHeight + WIDTH) / 2,
				(int)(2 + 5 + (WIDTH / Math.sqrt(2)) + delta_fix), (lineHeight - WIDTH) / 2);
			_navigation[0] = new Rectangle(3, (lineHeight / 2) - (int)(0.75 * WIDTH),
					(int)(1.5 * WIDTH) + 3, (int)(1.5 * WIDTH));
		}

		// Previous month
		{
			int     delta_fix = 1; // even if both polygons are the same size, the first
			// filled look smaller due to (int) casting.

			Polygon p = new Polygon();

			if ((_currentDisplayDate.get(Calendar.YEAR) == 1)
					&& (_currentDisplayDate.get(Calendar.MONTH) == 0))
				g.setColor(_darkGreyColor);
			else
				g.setColor(_blackColor);

			p.addPoint(5 + BUTTON_OFFSET, lineHeight / 2);
			p.addPoint((int)(5 + (WIDTH / Math.sqrt(2)) + delta_fix) + BUTTON_OFFSET,
				(lineHeight + WIDTH) / 2);
			p.addPoint((int)(5 + (WIDTH / Math.sqrt(2)) + delta_fix) + BUTTON_OFFSET,
				(lineHeight - WIDTH) / 2);
			g.fillPolygon(p);

			_navigation[1] = new Rectangle(3 + BUTTON_OFFSET,
					(lineHeight / 2) - (int)(0.75 * WIDTH), (int)(1.5 * WIDTH), (int)(1.5 * WIDTH));
		}

		// next month
		{
			Polygon p = new Polygon();

			g.setColor(_blackColor);

			p.addPoint(d.width - 3 - BUTTON_OFFSET, lineHeight / 2);
			p.addPoint((int)(d.width - 3 - (WIDTH / Math.sqrt(2))) - BUTTON_OFFSET,
				(lineHeight + WIDTH) / 2);
			p.addPoint((int)(d.width - 3 - (WIDTH / Math.sqrt(2))) - BUTTON_OFFSET,
				(lineHeight - WIDTH) / 2);
			g.fillPolygon(p);

			_navigation[2] = new Rectangle((int)(d.width - 1 - (1.5 * WIDTH)) - BUTTON_OFFSET,
					(lineHeight / 2) - (int)(0.75 * WIDTH), (int)(1.5 * WIDTH), (int)(1.5 * WIDTH));
		}
		// next year
		{
			Polygon p = new Polygon();

			g.setColor(_blackColor);

			p.addPoint(d.width - 3, lineHeight / 2);
			p.addPoint((int)(d.width - 3 - (WIDTH / Math.sqrt(2))), (lineHeight + WIDTH) / 2);
			p.addPoint((int)(d.width - 3 - (WIDTH / Math.sqrt(2))), (lineHeight - WIDTH) / 2);
			g.fillPolygon(p);
			g.drawLine((int)(d.width - 3 - (WIDTH / Math.sqrt(2))) - 2, (lineHeight + WIDTH) / 2,
				(int)(d.width - 3 - (WIDTH / Math.sqrt(2))) - 2, (lineHeight - WIDTH) / 2);

			_navigation[3] = new Rectangle((int)(d.width - 1 - (1.5 * WIDTH)),
					(lineHeight / 2)
					- (int)(0.75 * WIDTH), (int)(1.5 * WIDTH), (int)(1.5 * WIDTH));
		}

		// ### Add the week row ###
		String[] days = _dateFormatSymbols.getShortWeekdays();

		// Assume week starts on monday : TO DO, check context
		for (int i = 1; i <= 7; i++)
		{
			String s = days[1 + (i % 7)].substring(0, 1);

			// align header right
			x = (2 * charWidth) - fm.stringWidth(s);
			g.drawString(s, (int)((2 + ((i - 1) * (2 + COL_SPACING))) * charWidth) + x,
				(2
						* lineHeight) - fm.getDescent() - 1); // 2 : left border, 3 : 2 cols with the
			// date, 1 blank between cols.
			/*
			 * DEBUG : show grid off_g.drawLine((2 + (i-ROW_SPACING) * 3) * charWidth, 2 *
			 * lineHeight - 4, (2 + (i-1) * 3) * charWidth, 2 * lineHeight + 10); off_g.drawLine((2 +
			 * (i-ROW_SPACING) * 3+1) * charWidth, 2 * lineHeight - 4, (2 + (i-1) * 3+1) *
			 * charWidth, 2 * lineHeight + 2); off_g.drawLine((2 + (i-ROW_SPACING) * 3+2) *
			 * charWidth, 2 * lineHeight - 4, (2 + (i-1) * 3+2) * charWidth, 2 * lineHeight + 2);
			 */
		}

		g.setFont(f);

		// ### Fill up days ###
		int dayOfMonth = _currentDate.get(Calendar.DAY_OF_MONTH); // keep for the end

		_currentDisplayDate.set(Calendar.DAY_OF_MONTH, 1);

		int first_day = _currentDisplayDate.get(Calendar.DAY_OF_WEEK);

		// _currentDisplayDate.set(Calendar.DAY_OF_MONTH, dayOfMonth); // restore

		// Assume week start on monday : offset.
		first_day = ((first_day + 5) % 7) + 1; // sunday = 1 -> 7 monday = 2 -> 1 ...

		// set labels
		int limit = getDayCount(_currentDisplayDate);
		// if current month & year : highlight 'today'
		int today = -1;

		if ((_today.get(Calendar.MONTH) == _currentDisplayDate.get(Calendar.MONTH))
				&& (_today.get(Calendar.YEAR) == _currentDisplayDate.get(Calendar.YEAR)))
			today = _today.get(Calendar.DAY_OF_MONTH);

		// get ready with labels
		for (int i = 1; i <= 42; i++)
		{
			if ((i >= first_day) && ((i - first_day + 1) <= limit))
			{
				String label = Integer.toString(i - first_day + 1);

				int    row = (i - 1) / 7;
				int    col = 1 + ((i - 1) % 7);

				// draw around selection
				if (((i - first_day + 1) == dayOfMonth)
						&& (_currentDate.get(Calendar.MONTH)
							== _currentDisplayDate.get(Calendar.MONTH))
						&& (_currentDate.get(Calendar.YEAR)
							== _currentDisplayDate.get(Calendar.YEAR)))
				{
					if (today == (i - first_day + 1))
						g.setColor(_blueColor);
					else
						g.setColor(_redColor);

					int radius = Math.max(charWidth + 4, lineHeight / 2);

					g.fillOval((int)((2 + ((col - 1) * (2 + COL_SPACING))) * charWidth) - 4
						+ (charWidth + 4 - radius),
						((2 + row) * lineHeight)
						+ ((lineHeight / 2) - radius) + 1, 2 * radius, 2 * radius);
					g.setColor(_greyColor);
					g.fillOval((int)((2 + ((col - 1) * (2 + COL_SPACING))) * charWidth) - 2
						+ (charWidth + 4 - radius),
						((2 + row) * lineHeight)
						+ ((lineHeight / 2) - radius + 2) + 1, 2 * (radius - 2), 2 * (radius - 2));
					g.setColor(_blackColor);
				}
				else
				{
					if (_mousePosition != null)
					{
						// System.out.println("Pos :"+_mousePosition);
						// check if the mouse is over this date
						if ((_mousePosition.x
									>= ((2 + ((col - 1) * (2 + COL_SPACING))) * charWidth))
								&& (_mousePosition.x
									< ((4 + ((col - 1) * (2 + COL_SPACING)))
										* charWidth))
								&& (_mousePosition.y >= ((2 + row) * lineHeight))
								&& (_mousePosition.y < ((3 + row) * lineHeight)))
						{
							int radius = Math.max(charWidth, (lineHeight - 1) / 2);

							g.setColor(_yellowColor);
							g.fillRect((int)((2 + ((col - 1) * (2 + COL_SPACING))) * charWidth)
								+ (charWidth - radius),
								((2 + row) * lineHeight) + 1
								+ (((lineHeight - 1) / 2) - radius), 2 * radius, 2 * radius);
							g.setColor(_redColor);
							g.drawRect((int)((2 + ((col - 1) * (2 + COL_SPACING))) * charWidth)
								+ (charWidth - radius),
								((2 + row) * lineHeight) + 1
								+ (((lineHeight - 1) / 2) - radius), 2 * radius, 2 * radius);
							g.setColor(_blackColor);
						}
					}
					// if today : draw blue border
					if (today == (i - first_day + 1))
					{
						int radius = Math.max(charWidth, (lineHeight - 1) / 2);

						g.setColor(_blueColor);
						g.drawRect((int)((2 + ((col - 1) * (2 + COL_SPACING))) * charWidth)
							+ (charWidth - radius),
							((2 + row) * lineHeight) + 1
							+ (((lineHeight - 1) / 2) - radius), 2 * radius, 2 * radius);
						g.setColor(_blackColor);
					}
				}
				// align = right
				// x = 2 * charWidth - fm.stringWidth(label);
				// align = middle
				x = charWidth - (fm.stringWidth(label) / 2);

				g.drawString(label, (int)((2 + ((col - 1) * (2 + COL_SPACING))) * charWidth) + x,
					((3 + row) * lineHeight) - fm.getDescent() - 1);
			}
		}

		// draw offscreen image on layout
		// g.drawImage(_offscreen, 0, 0, this);
	} // end method paintComponent

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.getinterface.LyGetValueInterface#getValue()
	 */
	public String getValue()
	{
		return _dateFormat.format(_currentDate.getTime());
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetValueInterface#setValue(java.lang.String)
	 */
	@SuppressWarnings("deprecation")
	public void setValue(String str)
	{
		Date value = null;

		if ((str != null) && (str.length() > 0))
		{
			if (_dateFormat != null)
			{
				try
				{
					value = _dateFormat.parse(str);
				}
				catch (ParseException e)
				{
					System.out.println("[LyJDateSelector] parse error : " + str);
				}
			}
			else
				value = new Date(str);
		}

		if (value == null)
			value = new Date();

		_currentDate.setTime(value);
		_currentDisplayDate.setTime(value);
		repaint();
	} // end method setValue

	public String getId()
	{
		return (_id);
	}

	// **** Mouse Listener interface **
	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
	 */
	public void mousePressed(MouseEvent e)
	{
		// Retreive the day which has been clicked.

		_mousePosition.x = e.getX();
		_mousePosition.y = e.getY();

		// check if a navigation click is requested (next/previous month)
		int navPos = getNavigationArea(_mousePosition.x, _mousePosition.y);

		_currentDisplayDate.set(Calendar.DAY_OF_MONTH, 1);

		if (navPos == 0)
		{
			// Previous year
			if (_currentDisplayDate.get(Calendar.YEAR) > 1)
			{
				_currentDisplayDate.set(Calendar.YEAR, _currentDisplayDate.get(Calendar.YEAR) - 1); // set
				// new
				// date
				repaint();
			}
		}
		else if (navPos == 1)
		{
			// Previous month
			if ((_currentDisplayDate.get(Calendar.YEAR) == 1)
					&& (_currentDisplayDate.get(Calendar.MONTH) == 0))
			{
			}
			else
			{
				_currentDisplayDate.set(Calendar.MONTH,
					_currentDisplayDate.get(Calendar.MONTH) - 1); // set
				// new
				// date
				repaint();
			}
		}
		else if (navPos == 2)
		{
			// next month
			_currentDisplayDate.set(Calendar.MONTH, _currentDisplayDate.get(Calendar.MONTH) + 1); // set
			// new
			// date
			repaint();
		}
		else if (navPos == 3)
		{
			// next year
			_currentDisplayDate.set(Calendar.YEAR, _currentDisplayDate.get(Calendar.YEAR) + 1); // set
			// new
			// date
			repaint();
		}
		else
		{
			int limit = getDayCount(_currentDisplayDate);
			int dayOfMonth = _currentDate.get(Calendar.DAY_OF_MONTH); // keep for the end
			// _currentDisplayDate.set(Calendar.DAY_OF_MONTH, 1);
			int first_day = _currentDisplayDate.get(Calendar.DAY_OF_WEEK);

			// _currentDisplayDate.set(Calendar.DAY_OF_MONTH, dayOfMonth); // restore

			// Assume week start on monday : offset.
			first_day = ((first_day + 5) % 7) + 1; // sunday = 1 -> 7 monday = 2 -> 1 ...

			Rectangle bounds = getBoundaries();
			int		  charWidth = bounds.x;
			int		  lineHeight = bounds.y;

			// get ready with labels
			for (int i = 1; i <= 42; i++)
			{
				if ((i >= first_day) && ((i - first_day + 1) <= limit))
				{
					int row = (i - 1) / 7;
					int col = 1 + ((i - 1) % 7);

					// draw around selection
					// check if the mouse is over this date
					if ((_mousePosition.x >= ((2 + ((col - 1) * (2 + COL_SPACING))) * charWidth))
							&& (_mousePosition.x
								< ((4 + ((col - 1) * (2 + COL_SPACING))) * charWidth))
							&& (_mousePosition.y < ((3 + row) * lineHeight))
							&& (_mousePosition.y >= ((2 + row) * lineHeight)))
					{
						if (((i - first_day + 1) != dayOfMonth)
								|| (_currentDate.get(Calendar.YEAR)
									!= _currentDisplayDate.get(Calendar.YEAR))
								|| (_currentDate.get(Calendar.MONTH)
									!= _currentDisplayDate.get(Calendar.MONTH)))
						{
							_currentDate.set(Calendar.YEAR, _currentDisplayDate.get(Calendar.YEAR));
							_currentDate.set(Calendar.MONTH,
								_currentDisplayDate.get(Calendar.MONTH));
							_currentDate.set(Calendar.DAY_OF_MONTH, i - first_day + 1); // set new
							// date
							_currentDisplayDate.setTime(_currentDate.getTime());

							LyActionEvent actionEvent = new LyActionEvent(this);

							actionEvent.setValue(getId());
							if (_actionTransmitter != null)
								_actionTransmitter.processActionEvent(actionEvent);
							repaint();
						}
						else
						{
							LyActionEvent actionEvent = new LyActionEvent(this);

							actionEvent.setValue(getId());
							if (_actionTransmitter != null)
								_actionTransmitter.processActionEvent(actionEvent);
							repaint();
						}
						break;
					}
				}
			}
		}
	} // end method mousePressed

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
	 */
	public void mouseClicked(MouseEvent e)
	{
		// On double click on title bar, set current date
		_mousePosition.x = e.getX();
		_mousePosition.y = e.getY();

		if (e.getClickCount() == 2)
		{
			// should be out of navigation areas
			int navPos = getNavigationArea(_mousePosition.x, _mousePosition.y);

			if (navPos == -1)
			{
				// should be in titlebar
				Rectangle bounds = getBoundaries();

				if (_mousePosition.y <= bounds.y)
				{
					Date d = new Date();

					_currentDate.setTime(d);
					_currentDisplayDate.setTime(d);
					// repaint();

					LyActionEvent actionEvent = new LyActionEvent(this);

					actionEvent.setValue(null);
					if (_actionTransmitter != null)
						_actionTransmitter.processActionEvent(actionEvent);
				}
			}
		}

	} // end method mouseClicked

	/**
	 * Nothing is done on mouse release.
	 *
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	public void mouseReleased(MouseEvent e)
	{
	}

	/**
	 * Nothing is done when mouse enters a component.
	 *
	 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
	 */
	public void mouseEntered(MouseEvent e)
	{
	}

	/**
	 * Nothing is done when mouse exits a component.
	 *
	 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
	 */
	public void mouseExited(MouseEvent e)
	{
	}

	/**
	 * Nothing is done when mouse is dragged over a component.
	 *
	 * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
	 */
	public void mouseDragged(MouseEvent e)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.MouseMotionListener#mouseMoved(java.awt.event.MouseEvent)
	 */
	public void mouseMoved(MouseEvent e)
	{
		if (_mousePosition == null)
			_mousePosition = new Point(e.getX(), e.getY());
		else
		{
			_mousePosition.x = e.getX();
			_mousePosition.y = e.getY();
		}
		repaint();
	}

	/**
	 * Gets the number of days in the specified month/year.
	 *
	 * @param  calDate the calendar where its days are counted
	 * @return the number of days in the calendar
	 */
	public int getDayCount(Calendar calDate)
	{
		int month = calDate.get(Calendar.MONTH);
		int year = calDate.get(Calendar.YEAR);

		int limit;

		if ((month == 0) || (month == 2) || (month == 4) || (month == 6) || (month == 7)
				|| (month == 9) || (month == 11))
			limit = 31;
		else if ((month == 3) || (month == 5) || (month == 8) || (month == 10))
			limit = 30;
		else
		{
			boolean is_leap;

			is_leap = (((year % 4) == 0) && (((year % 100) != 0) || ((year % 400) == 0)));
			if (is_leap && (month == 1))
				limit = 29;
			else
				limit = 28;
		}

		return limit;
	}

	/**
	 * Checks if the specified (x,y) position is in the navigation area.
	 *
	 * @param  x mouse click x coordinate
	 * @param  y mouse click y coordinate
	 * @return 0 for the left area, 1 for the right or -1 if there is no navigation area at the
	 *         specified position
	 */
	public int getNavigationArea(int x, int y)
	{
		if (_navigation == null)
			return -1;

		for (int i = 0; i < _navigation.length; i++)
		{
			if (_navigation[i] == null)
				continue; // impossible

			if ((x >= _navigation[i].x) && (x <= (_navigation[i].x + _navigation[i].width))
					&& (y >= _navigation[i].y) && (y <= (_navigation[i].y + _navigation[i].height)))
				return i;
		}
		return -1;
	}

	/**
	 * Sets the requested date format.
	 *
	 * @param format new format as defined in the SimpleDateFormat class
	 */
	public void setDateFormat(String format)
	{
		if (format != null)
			_dateFormat = new SimpleDateFormat(format);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return (_actionTransmitter);
	}

	/**
	 * This date selector doesn't handle selection events, so it has no selection transmitter.
	 *
	 * @return <code>null</code>
	 * @see    leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		return (null);
	}

} // end class LyJDateSelector
