/*
 * LyJExpandItem.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import leon.misc.*;
import leon.view.event.*;


/**
 * This class defines an expand item placed in an expand bar.
 *
 * @see LyJExpandBar
 */
public class LyJExpandItem extends JPanel implements MouseListener, MouseMotionListener
{
	private static final long  serialVersionUID = -4730687983924122090L;
	public static final String __VERSION =
		"$Id: LyJExpandItem.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Horizontal gap between components. */
	public static int			X_GAP = 12;

	/** Vertical gap between components. */
	public static int			Y_GAP = 4;

	/** Default cursor. */
	public final static Cursor	DEFAULT_CURSOR;

	/** Hand cursor. */
	public final static Cursor	OVER_CURSOR;

	static
	{
		DEFAULT_CURSOR = new Cursor(Cursor.DEFAULT_CURSOR);
		OVER_CURSOR = new Cursor(Cursor.HAND_CURSOR);
	}

	/** Current environment. */
	protected LyEnvironment		_environment;
	
	/** Font used to draw label. */
	protected Font				_font;
	
	/** Color used to draw label. */
	protected Color				_titleColor;
	
	/** Color used to draw the border when item is opened. */
	protected Color				_borderColor;
	
	/** Color used to draw the content when item is opened. */
	protected Color				_contentColor;
	
	/** Color used to draw the outside of the item. */
	protected Color				_outsideColor;
	
	/** Height of the header part. */
	protected int				_headerHeight;
	
	/** Identifier of the item. */
	protected String			_id;
	
	/** Label of the item. */
	protected String			_label;
	
	/** Image of the item. */
	protected Image				_image;
	
	/** Image used in the upper-left corner. */
	protected Image				_topLeftImage;
	
	/** Image used in the upper-right corner. */
	protected Image				_topRightImage;
	
	/** Image used as the upper background. */
	protected Image				_topImage;
	
	/** Image used for the close button. */
	protected Image				_buttonClose;
	
	/** Image used for the close button with the mouse over. */
	protected Image				_buttonCloseOver;
	
	/** Image used for the open button. */
	protected Image				_buttonOpen;
	
	/** Image used for the open button with the mouse over. */
	protected Image				_buttonOpenOver;
	
	/** Indicates whether the item is opened or closed. */
	protected boolean			_open;
	
	/** Indicates whether the mouse is over the header or not. */
	protected boolean			_mouseOverHeader;
	
	/** Width of the item (used to force all the items of a same bar to have the same width). */
	protected int				_width;

	/** Action listener used to transmit events. */
	protected LyActionListener  _listener;

	/**
	 * Sole constructor.
	 */
	public LyJExpandItem(LyEnvironment environment, String id, String label, String image)
	{
		super();
		
		_environment = environment;
		
		// Load needed images and keep them for optimization.
		_font = (Font)_environment.getFont("LY_EXPANDBAR_TITLE_FONT");
		_titleColor = (Color)_environment.getColor("LY_EXPANDBAR_TITLE_FONT_COLOR");
		_borderColor = (Color)_environment.getColor("LY_EXPANDBAR_BORDER_COLOR");
		_outsideColor = (Color)_environment.getColor("LY_EXPANDBAR_CONTENT_COLOR");
		_contentColor = (Color)_environment.getColor("LY_EXPANDBAR_ITEM_COLOR");
		_topLeftImage = (Image)_environment.getImage("LY_VERTICAL_TOOLBAR_TOP_LEFT");
		_topRightImage = (Image)_environment.getImage("LY_VERTICAL_TOOLBAR_TOP_RIGHT");
		_topImage = (Image)_environment.getImage("LY_VERTICAL_TOOLBAR_TOP");
		_buttonClose = (Image)_environment.getImage("LY_VERTICAL_TOOLBAR_BUTTON_CLOSE");
		_buttonCloseOver = (Image)_environment.getImage("LY_VERTICAL_TOOLBAR_BUTTON_CLOSE_OVER");
		_buttonOpen = (Image)_environment.getImage("LY_VERTICAL_TOOLBAR_BUTTON_OPEN");
		_buttonOpenOver = (Image)_environment.getImage("LY_VERTICAL_TOOLBAR_BUTTON_OPEN_OVER");
		_mouseOverHeader = false;
		_width = -1;
		_id = id;

		setOpaque(false);
		setLabel(label);
		setImage(image);
		
		FontMetrics fm = getFontMetrics(_font);
		int			h = 0;

		if (_topLeftImage != null)
			h = Math.max(h, _topLeftImage.getHeight(this));
		
		if (_image != null)
			h = Math.max(h, _image.getHeight(this) + (2 * Y_GAP));

		if (_topRightImage != null)
			h = Math.max(h, _topRightImage.getHeight(this));

		h = Math.max(h, fm.getHeight() + (2 * Y_GAP));
		_headerHeight = h;

		setBorder(BorderFactory.createEmptyBorder(_headerHeight + Y_GAP, X_GAP, Y_GAP, X_GAP));

		VFlowLayout	layout = new VFlowLayout(VFlowLayout.TOP, 0, 0, false);
		
		layout.setAlignment(FlowLayout.LEFT);
		setLayout(layout);

		addMouseMotionListener(this);
		addMouseListener(this);
	}
	
	/**
	 * Sets the action listener used to transmit events.
	 */
	public void setActionListener(LyActionListener listener)
	{
		_listener = listener;
	}

	/**
	 * Sets the label of the item.
	 * 
	 * @param label the new label of the item.
	 */
	public void setLabel(String label)
	{
		_label = _environment.translate(label);
	}
	
	/**
	 * Sets the image of the item.
	 * 
	 * @param image the new image of the item.
	 */
	public void setImage(String image)
	{
		_image = (Image)_environment.getImage(image);
	}
	
	/**
	 * Sets the state of the item (opened or closed).
	 * 
	 * @param open the new state of the item.
	 */
	public void setOpen(boolean open)
	{
		_open = open;
		
		int			i, n = getComponentCount();
		Component	comp;
		
		for (i = 0; i < n; i++)
		{
			comp = getComponent(i);
			comp.setVisible(open);
		}
	}

	/**
	 * Sets the mouse over parameter or the item.
	 * 
	 * @param over true if the mouse is over the header, false otherwise.
	 */
	protected void setMouseOverHeader(boolean over)
	{
		_mouseOverHeader = over;

		if (over)
			setCursor(LyJExpandItem.OVER_CURSOR);
		else
			setCursor(null);

		repaint();
	}
	
	/**
	 * Sets the width of the item.
	 * 
	 * @param width the new width of the item.
	 */
	public void setWidth(int width)
	{
		_width = width;
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.awt.Component#isFocusTraversable()
	 */
	public boolean isFocusTraversable()
	{
		return true;
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#getPreferredSize()
	 */
	public Dimension getPreferredSize()
	{
		Dimension	result = super.getPreferredSize();
		int			headerWidth = 0;

		if (_topLeftImage != null)
			headerWidth += _topLeftImage.getWidth(this) + X_GAP;
		
		if (_image != null)
			headerWidth += _image.getWidth(this) + X_GAP;
		
		if (_label != null)
		{
			FontMetrics fm = getFontMetrics(_font);

			headerWidth += fm.stringWidth(_label) + X_GAP;
		}
		
		if (_buttonOpen != null)
			headerWidth += _buttonOpen.getWidth(this);

		if (_topRightImage != null)
			headerWidth += _topLeftImage.getWidth(this);

		result.width = Math.max(result.width, headerWidth);
		
		if (!_open)
		{
			setOpen(true);
			result.width = super.getPreferredSize().width;
			setOpen(false);
		}
		
		if (_width > result.width)
			result.width = _width;
		
		if (!_open)
			result.height = _headerHeight;
		
		return result;
	}
	
	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#paintComponent(java.awt.Graphics)
	 */
	public void paintComponent(Graphics g)
	{
		super.paintComponent(g);

		Dimension	dim = getSize();

		if ((dim.width <= 0) || (dim.height <= 0))
			return;

		int		x = 0;
		String	aaText = System.getProperty("swing.aatext");

		// set antialias
		if ("true".equalsIgnoreCase(aaText) && (g instanceof Graphics2D))
			((Graphics2D)g).setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
				RenderingHints.VALUE_TEXT_ANTIALIAS_ON);

		// Draw the content background
		if (_open)
		{
			if (_contentColor != null)
			{
				// If a content color is given, draw background only once and use content color
				// to fill the rest of the content.
				g.setColor(_contentColor);
				g.fillRect(0, _headerHeight - 1, dim.width - 1, dim.height - _headerHeight);
			}

			g.setColor(_borderColor);
			g.drawRect(0, _headerHeight - 1, dim.width - 1, dim.height - _headerHeight);
		}
		
		// Draw the background of the header
		if (_topImage != null)
		{
			int			i, j, imageWidth = _topImage.getWidth(this), imageHeight = _topImage.getHeight(this);
			Rectangle	oldClip = g.getClipBounds();
			Rectangle	newClip = new Rectangle(_topLeftImage.getWidth(this), 0,
					dim.width - _topRightImage.getWidth(this) - _topLeftImage.getWidth(this), _headerHeight);
			
			if (oldClip != null)
				newClip = newClip.intersection(oldClip);
			
			g.setClip(newClip);
			
			for (i = 0; i < (dim.width / imageWidth) + 1; i++)
			{
				for (j = 0; j < (_headerHeight / imageHeight) + 1; j++)
				{
					g.drawImage(_topImage, imageWidth * i, imageHeight * j, this);
				}
			}

			g.setClip(oldClip);
		}
		
		// Draw the upper left corner
		if (_topLeftImage != null)
		{
			g.drawImage(_topLeftImage, 0, 0, this);
			x += _topLeftImage.getWidth(this) + X_GAP;
		}
		
		// Draw the image of the item
		if (_image != null)
		{
			g.drawImage(_image, x, (_headerHeight - _image.getHeight(this)) / 2, this);
			x += _image.getWidth(this) + X_GAP;
		}

		// Draw the label
		if (_label != null)
		{
			FontMetrics fm;

			g.setFont(_font);
			fm = g.getFontMetrics();
			g.setColor(_titleColor);
			g.drawString(_label, x, ((_headerHeight - fm.getHeight()) / 2) + fm.getAscent());
			x += fm.stringWidth(_label) + X_GAP;
		}
		
		// Draw the upper right corner
		if (_topRightImage != null)
			g.drawImage(_topRightImage, dim.width - _topRightImage.getWidth(this), 0, this);

		// Draw the button used to open / close the item
		Image	buttonImage = (_open ? (_mouseOverHeader ? _buttonCloseOver : _buttonClose) : (_mouseOverHeader ? _buttonOpenOver : _buttonOpen));
		
		if (buttonImage != null)
			g.drawImage(buttonImage, dim.width - _topRightImage.getWidth(this) - _buttonOpen.getWidth(this),
					(_headerHeight - _buttonOpen.getHeight(this)) / 2, this);
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
	 */
	public void mouseClicked(MouseEvent e)
	{
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
	 */
	public void mousePressed(MouseEvent e)
	{
		requestFocus();
		
		if (e.getY() <= _headerHeight)
		{
			if (_listener != null)
				_listener.actionPerformed(new LyActionEvent(this, _id, !_open + "", LyActionDemand.TYPE_ACTION));
			
			setOpen(!_open);
			repaint();
		}
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
	 */
	public void mouseEntered(MouseEvent e)
	{
		setMouseOverHeader(false);
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
	 */
	public void mouseExited(MouseEvent e)
	{
		setMouseOverHeader(false);
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	public void mouseReleased(MouseEvent e)
	{
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
	 */
	public void mouseDragged(MouseEvent e)
	{
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseMotionListener#mouseMoved(java.awt.event.MouseEvent)
	 */
	public void mouseMoved(MouseEvent e)
	{
		if ((e.getY() <= _headerHeight)&& !_mouseOverHeader)
			setMouseOverHeader(true);
		else if ((e.getY() > _headerHeight) && _mouseOverHeader)
			setMouseOverHeader(false);
	}
}