/*
 * LyJFileSelector.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.io.*;
import javax.swing.*;
import javax.swing.filechooser.FileFilter;
import leon.misc.*;
import leon.view.*;


/**
 * This class defines a component containing a text field and a button. A click on the button opens
 * a dialog which allows to select a file. The selector may have a type indicating the mime-type of
 * the files that may be chosen (text, image, audio, video...).
 *
 * @version $Id: LyJFileDialog.java 33981 2012-01-31 13:47:29Z bangiole $
 * @author  Lyria
 * @see     LyJSelector
 */
public class LyJFileDialog
{
	public static final String __VERSION =
		"$Id: LyJFileDialog.java 33981 2012-01-31 13:47:29Z bangiole $";

	/** Text type. */
	public final static int    TYPE_TEXT = 0;

	/** Image type. */
	public final static int    TYPE_IMAGE = 1;

	/** Audio type. */
	public final static int    TYPE_AUDIO = 2;

	/** Video type. */
	public final static int    TYPE_VIDEO = 3;

	/** Directory type. */
	public final static int    TYPE_DIRECTORY = 4;

	/** Last opened directory. */
	protected static String    _lastDirectory = null;

	/**
	 * Type of the selector. One of the following:
	 *
	 * <ul>
	 *   <li>{@link #TYPE_TEXT}</li>
	 *   <li>{@link #TYPE_IMAGE}</li>
	 *   <li>{@link #TYPE_AUDIO}</li>
	 *   <li>{@link #TYPE_VIDEO}</li>
	 *   <li>{@link #TYPE_DIRECTORY}</li>
	 * </ul>
	 */
	protected LyFile.Type	   _type;

	/** The parent frame. */
	Frame					   _parent;

	/** The file chooser component. */
	protected JFileChooser     _chooser;

	/**
	 * Creates and initializes a new file dialog.
	 *
	 * @param parent    the parent component
	 * @param title     the title of the dialog
	 * @param type      the type of files to show (see {@link #_type})
	 * @param directory the path to the starting directory
	 * @param filter    a filter for the file names to show
	 */
	public LyJFileDialog(Frame parent, String title, LyFile.Type type, String directory, String filter)
	{
		_parent = parent;
		_chooser = new JFileChooser();

		int typeSelector;

		if (type == LyFile.Type.DIRECTORY)
			typeSelector = JFileChooser.DIRECTORIES_ONLY;
		else
			typeSelector = JFileChooser.FILES_ONLY;

		_chooser.setFileSelectionMode(typeSelector);
		_chooser.setDialogType(JFileChooser.OPEN_DIALOG);

		if (title != null)
			_chooser.setDialogTitle(title);

		_type = type;

		if (directory == null)
			directory = _lastDirectory;

		if (directory != null)
		{
			File dir = new File(directory);

			setDirectory(dir);
		}

		if (filter != null)
			_chooser.setFileFilter(new _FileFilter(filter));
	}

	/**
	 * Shows the file chooser dialog.
	 *
	 * @return the return state of the file chooser on popdown:
	 *
	 *         <ul>
	 *           <li>{@link JFileChooser#CANCEL_OPTION}</li>
	 *           <li>{@link JFileChooser#APPROVE_OPTION}</li>
	 *           <li>{@link JFileChooser#ERROR_OPTION} if an error occurs or the dialog is
	 *             dismissed</li>
	 *         </ul>
	 * @see    JFileChooser#showOpenDialog(java.awt.Component)
	 */
	public int show()
	{
		return _chooser.showOpenDialog(_parent);
	}

	/**
	 * Gets the directory in which the files are selected.
	 *
	 * @return the working directory
	 * @see    #setDirectory(File)
	 */
	public String getDirectory()
	{
		File dir = _chooser.getSelectedFile();

		if (dir.isDirectory())
		{
			_lastDirectory = _chooser.getCurrentDirectory().getPath();
			return dir.getPath();
		}
		_lastDirectory = _chooser.getCurrentDirectory().getPath();

		return _chooser.getCurrentDirectory().getPath();
	}

	/**
	 * Gets the path of the selected file.
	 *
	 * @return the path to the selected file in the dialog
	 * @see    #setFile(File)
	 */
	public String getFile()
	{
		return _chooser.getSelectedFile().getPath();
	}

	/**
	 * Allows to select initial value of the file chooser.
	 *
	 * @param file he initial file or directory
	 * @see   #getFile()
	 * @see   JFileChooser#setSelectedFile(java.io.File)
	 */
	public void setFile(File file)
	{
		_chooser.setSelectedFile(file);
	}

	/**
	 * Sets the current directory on dialog opening.
	 *
	 * @param file a directory or a file in the directory to use
	 * @see   JFileChooser#setCurrentDirectory(java.io.File)
	 * @see   #getDirectory()
	 */
	public void setDirectory(File file)
	{
		_chooser.setCurrentDirectory(file);
	}

	/**
	 * This class defines file filters to not show all the files of a directory.
	 */
	class _FileFilter extends FileFilter
	{

		/** Filter to use. It may be a list of filters separated by space character. */
		String   _filter;

		/** List of atomic file filters. */
		String[] _filters;

		/**
		 * Creates a new file filter.
		 *
		 * @param filter the file filter list separated by space character.
		 */
		_FileFilter(String filter)
		{
			_filter = filter;
			_filters = LyString.stringToList(filter, " ");
		}

		/*
		 * (non-Javadoc)
		 * @see javax.swing.filechooser.FileFilter#accept(java.io.File)
		 */
		public boolean accept(File f)
		{
			if (f.isDirectory())
				return true;

			String name = f.getName();
			int    pos, i, n = _filters.length;

			for (i = 0; i < n; i++)
			{
				pos = _filters[i].indexOf(".");
				if (pos >= 0)
				{
					String filter = _filters[i].substring(pos);
					if (name.endsWith(filter)
							|| name.endsWith(filter.toUpperCase()))
						return true;
				}
			}

			return false;
		}

		/*
		 * (non-Javadoc)
		 * @see javax.swing.filechooser.FileFilter#getDescription()
		 */
		public String getDescription()
		{
			return _filter;
		}
	} // end class _FileFilter
} // end class LyJFileDialog
