/*
 * LyJFileSelector.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import javax.swing.text.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.win.swing.*;
import leon.view.win.swing.text.*;


/**
 * This class defines a component containing a text field and a button. A click on the button opens
 * a dialog which allows to select a file. The selector may have a type indicating the mime-type of
 * the files that may be chosen (text, image, audio, video...).
 *
 * @version $Id: LyJFileSelector.java 34856 2012-03-19 14:05:02Z jlsaouli $
 * @author  Lyria
 * @see     LyJSelector
 */
public class LyJFileSelector extends LyJSelector
{
	private static final long  serialVersionUID = -6345036040096423208L;

	public static final String __VERSION =
		"$Id: LyJFileSelector.java 34856 2012-03-19 14:05:02Z jlsaouli $";

	/** Image width offset for the selector button. */
	public static int		   IMAGE_W_OFFSET = 5;

	/** Image height offset for the selector button. */
	public static int		   IMAGE_H_OFFSET = 6;
    
    /** Default width. */
    public static int          DEFAULT_WIDTH = 40;

	/**
	 * The type of the selector. One of the following:
	 *
	 * <ul>
	 *   <li>{@link LyJFileDialog#TYPE_TEXT}</li>
	 *   <li>{@link LyJFileDialog#TYPE_IMAGE}</li>
	 *   <li>{@link LyJFileDialog#TYPE_AUDIO}</li>
	 *   <li>{@link LyJFileDialog#TYPE_VIDEO}</li>
	 *   <li>{@link LyJFileDialog#TYPE_DIRECTORY}</li>
	 * </ul>
	 */
	protected LyFile.Type	   _type;

	/** The associated button of this file selector. */
	protected LyJButton		   _button;

	/** The file dialog to show to select a file. */
	protected LyJFileDialog    _dialog;

	/** The file dialog title. */
	protected String		   _dialogTitle;

	/** The working directory. */
	protected String		   _directory;

	/** The file filter pattern. */
	protected String		   _filter;

	/**
	 * Creates and initializes a new file selector.
	 *
	 * @param type      the type of the selector
	 * @param button    Not used in this implementation.
	 * @param toolTip   the dialog title
	 * @param image     the image associated to the selector button
	 * @param directory the working directory
	 * @param filter    the file filter to hide in the selector
	 * @param env       the current environment
	 */
	public LyJFileSelector(LyFile.Type type, String button, String toolTip, Image image, String directory,
		String filter, int width, LyEnvironment env)
	{
		super(env, true);

		_type = type;
		_dialogTitle = toolTip;
		_directory = directory;
		_filter = filter;
		_button = addButton(image, toolTip);
        
        if (width <= 0)
        {
            width = env.getIntResource("LY_FILE_SELECTOR_WIDTH");
            if (width <= 0)
                width = DEFAULT_WIDTH;
        }
        
    	_text.setColumns(width);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJSelector#init()
	 */
	protected void init()
	{
		super.init();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		super.actionPerformed(actionEvent);

		if (actionEvent.getSource() == _button)
		{
			if (_dialog == null)
				_dialog = new LyJFileDialog(LySwingViewManager.getRootFrame(this), _dialogTitle,
						_type, _directory, _filter);

			String value = getValue();

			if ((value != null) && (value.length() > 0))
			{
				if ((_directory != null)
						&& !value.toUpperCase().startsWith(_directory.toUpperCase()))
					value = _directory + value;

				File existFile = new File(value);

				if (existFile.exists())
					_dialog.setFile(existFile);
				else if (existFile.isDirectory())
					_dialog.setDirectory(existFile);
			}

			if (_dialog.show() == JFileChooser.APPROVE_OPTION)
			{
				if (_type == LyFile.Type.DIRECTORY)
					setValue(_dialog.getDirectory());
				else
					setValue(_dialog.getFile());

				// FFT00413
				simulateFocusLost();
			}
		}
	} // end method actionPerformed

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJSelector#addButton(java.awt.Image, java.lang.String)
	 */
	public LyJButton addButton(Image image, String tooltip)
	{
		LyJButton button = super.addButton(image, "");

		button.setToolTipText(LyTools.formatToHTML(tooltip, LyJButton.TIP_MAX_CHARS));
		button.setPreferredSize(new Dimension(button.getIcon().getIconWidth() + IMAGE_W_OFFSET,
				button.getIcon().getIconHeight() + IMAGE_H_OFFSET));
		return button;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJSelector#checkButtons()
	 */
	protected void checkButtons()
	{
		_button.setEnabled(isEnabled());
	}

	/**
	 * Regain the keyboard focus, even if it hadn't been lost.
	 */
	public void simulateFocusLost()
	{
		_hasGainedFocus = true;
		focusLost(null);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetValueInterface#setValue(java.lang.String)
	 */
	public void setValue(String str, boolean notify)
	{
		if ((_directory != null) && (str != null)
				&& str.toUpperCase().startsWith(_directory.toUpperCase()))
			str = str.substring(_directory.length());

		// change the value of the text component
		try
		{
			((LyJTextAreaDocument)_text.getDocument()).setString((str == null) ? "" : str);
			if (str != null)
			    _text.setCaretPosition(str.length());
		}
		catch (BadLocationException e)
		{
			System.out.println("SetValue impossible");
		}

		super.setValue(str, notify);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.FocusListener#focusLost(java.awt.event.FocusEvent)
	 */
	public void focusLost(FocusEvent evt)
	{
		if (_directory != null)
		{
			String str = getValue();

			setValue(str);
		}

		super.focusLost(evt);
	}
} // end class LyJFileSelector
