/*
 * LyJFormattedField.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import leon.misc.*;
import leon.view.event.*;
import leon.view.win.swing.text.*;


/**
 * This class implements a TextField that allows formatted input. The format is :
 * '['&nbsp;number&nbsp;'-'&nbsp;number&nbsp;']'&nbsp;(&nbsp;sep&nbsp;'['&nbsp;number&nbsp;'-'&nbsp;number&nbsp;']'&nbsp;)
 * * where variable number is a positive number and sep is a separator String that can be ""
 *
 * @see LyJFormat
 */
public class LyJFormattedField extends LyJTextField
	implements FocusListener,
		KeyListener
{
	private static final long  serialVersionUID = -8751956265197030447L;

	public static final String __VERSION =
		"$Id: LyJFormattedField.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The compiled expression of the format. */
	protected LyJFormat		   _format;

	/** The formatted input component. */
	LyJFormattedInput		   _parent;

	/* *************************** Constructor ********************************/

	/**
	 * Creates a formatted field constructor and initializes it.
	 *
	 * @param  parent the parent component
	 * @param  env    the current environment
	 * @throws LyJBadFormatException is not thrown here
	 * @throws BadLocationException  if the format displayable can't be updates in the associated
	 *                               formatted text document
	 */
	public LyJFormattedField(LyJFormattedInput parent, LyEnvironment env)
		throws LyJBadFormatException,
			BadLocationException
	{
		super(env);
		_format = new LyJFormat();
		_parent = parent;

		initialize();
	}

	/**
	 * Creates a formatted field constructor with the specified format and initializes it.
	 *
	 * @param  format the String expression of the format
	 * @param  parent the parent component
	 * @param  env    the current environment
	 * @throws LyJBadFormatException if the format generates errors during parsing
	 * @throws BadLocationException  if the format displayable can't be updates in the associated
	 *                               formatted text document
	 */
	public LyJFormattedField(String format, LyJFormattedInput parent, LyEnvironment env)
		throws LyJBadFormatException,
			BadLocationException
	{
		super(env);
		_format = new LyJFormat(format);
		_parent = parent;

		initialize();
	}

	/**
	 * Initializes this formatted field.
	 *
	 * @throws BadLocationException if the format displayable can't be updates in the associated
	 *                              formatted text document
	 */
	protected void initialize()
		throws BadLocationException
	{
		LyJFormattedTextDocument doc = new LyJFormattedTextDocument(_format, this);

		doc.setDisplayableString(_format.getDisplayableString(true));
		setDocument(doc);
		addCaretListener(new _LyJCaretListener(_format));
		addFocusListener(this);
		addKeyListener(this);
		setCaret(new _LyJFormattedTextCaret(_format));
	}

	/**
	 * Sets the format for this field.
	 *
	 * @param  format the format that is parsed and set
	 * @throws LyJBadFormatException if the format generates errors during parsing
	 * @see    LyJFormat#setFormatString(String)
	 */
	public void setFormatString(String format)
		throws LyJBadFormatException
	{
		_format.setFormatString(format);
	}

	/**
	 * Invoked when the textField gains the focus. Shows the editing value and resets the caret
	 * position.
	 *
	 * @see java.awt.event.FocusListener#focusGained(java.awt.event.FocusEvent)
	 */
	public void focusGained(FocusEvent e)
	{
		try
		{
			((LyJFormattedTextDocument)getDocument()).setDisplayableString(_format
					.getDisplayableString(true));
		}
		catch (BadLocationException exc)
		{
		}
	}

	/**
	 * Invoked when the textField loses the focus. Shows the editing value and resets the caret
	 * position.
	 *
	 * @see java.awt.event.FocusListener#focusLost(java.awt.event.FocusEvent)
	 */
	public void focusLost(FocusEvent e)
	{
		int dotPos = getCaret().getDot();

		try
		{
			((LyJFormattedTextDocument)getDocument()).setDisplayableString(_format
					.getDisplayableString(false));
			notifyListeners();
		}
		catch (BadLocationException exc)
		{
		}

		getCaret().setDot(dotPos);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
	 */
	public void keyPressed(KeyEvent e)
	{
		if (e.getKeyCode() == KeyEvent.VK_ENTER)
			focusLost(null);
	}

	/**
	 * Nothing is done when a key is released.
	 *
	 * @see java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
	 */
	public void keyReleased(KeyEvent e)
	{
	}

	/**
	 * If the user presses '.', the caret go to the next valid position.
	 *
	 * @see java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
	 */
	public void keyTyped(KeyEvent e)
	{
		if (e.getKeyChar() == '.')
			getCaret().setDot(getNextCaretPosition());
	}

	/**
	 * Gets the next caret position.
	 *
	 * @return the next valid caret position
	 */
	public int getNextCaretPosition()
	{
		int index = _format.getIndexIntervalFromOffset(getCaret().getDot());

		if (index == -1)
			index = _format.getIndexSeparatorFromOffset(getCaret().getDot());

		if (index == (_format.getIntervalsCount() - 1))
			return _format.getInterval(0).getStartOffset();
		return _format.getInterval(index + 1).getStartOffset();
	}

	/**
	 * Gets the text of this text component.
	 *
	 * @return <code>null</code> if the value is not set.
	 * @see    javax.swing.text.JTextComponent#getText()
	 */
	public String getText()
	{
		String text;

		if (_format.isAllEmpty())
			text = null;
		else
			text = super.getText();
		return text;
	}

	/**
	 * Gets the value of this text component.
	 *
	 * @return the current value
	 * @see    #getText()
	 * @see    #setValue(String)
	 */
	public String getValue()
	{
		return getText();
	}

	/**
	 * Set the current value of the text component.
	 *
	 * @param value the value to set
	 * @see   #getValue()
	 */
	public void setValue(String value)
	{
		if ((value == null) || value.equals(""))
		{
			_format.clear();
			value = _format.getDisplayableString(true);
		}
		int dot = getCaret().getDot();

		try
		{
			LyJFormattedTextDocument doc = (LyJFormattedTextDocument)getDocument();

			doc.setDisplayableString(value);
			getCaret().setDot(dot);
		}
		catch (BadLocationException e)
		{
			e.printStackTrace();
		}
	}

	/**
	 * Gets the value at the caret position.
	 *
	 * @return the subvalue at the position of the caret
	 * @see    #setSelectedIntValue(int)
	 * @see    LyJFormat#getIntValue(int)
	 */
	public int getSelectedIntValue()
	{
		int dot = getCaret().getDot();

		return _format.getIntValue(dot);
	}

	/**
	 * Sets the subvalue at the position of the caret.
	 *
	 * @param value the value to set
	 * @see   #getSelectedIntValue()
	 */
	public void setSelectedIntValue(int value)
	{
		int dot = getCaret().getDot();

		_format.setIntValue(dot, value);
		try
		{
			LyJFormattedTextDocument doc = (LyJFormattedTextDocument)getDocument();

			doc.setDisplayableString(_format.getDisplayableString(true));
			getCaret().setDot(dot);
		}
		catch (BadLocationException e)
		{
			e.printStackTrace();
		}
	}

	/**
	 * Gets the minimum value at the caret position.
	 *
	 * @return the minimum value at the position of the caret
	 */
	public int getSelectedMinIntValue()
	{
		int dot = getCaret().getDot();

		return _format.getMinIntValue(dot);
	}

	/**
	 * Gets the maximum value at the caret position.
	 *
	 * @return the maximum value at the position of the caret
	 */
	public int getSelectedMaxIntValue()
	{
		int dot = getCaret().getDot();

		return _format.getMaxIntValue(dot);
	}

	/**
	 * Notify the event listeners.
	 */
	protected void notifyListeners()
	{
		// check if something has changed
		if ((_parent != null) && (_parent.getActionTransmitter() != null))
		{
			LyActionTransmitter transmitter = _parent.getActionTransmitter();

			transmitter.processActionEvent(new LyActionEvent(this, null, getValue(),
					LyActionDemand.TYPE_VALUE_CHANGED));
		}
	}

	/* **************** Caret ******************************/

	/**
	 * A caret that remember its last position.
	 */
	public class _LyJFormattedTextCaret extends DefaultCaret
	{
		private static final long serialVersionUID = 6259348909049277112L;

		/** The mask format. */
		protected LyJFormat		  _format;

		/** The last position of the caret. */
		Point					  _lastCaretPosition;

		/**
		 * Creates a caret for the specified format.
		 *
		 * @param format the mask format
		 */
		public _LyJFormattedTextCaret(LyJFormat format)
		{
			super();
			_format = format;
		}

		/**
		 * Saves the last caret position.
		 *
		 * @see javax.swing.text.Caret#setMagicCaretPosition(java.awt.Point)
		 */
		public void setMagicCaretPosition(Point pt)
		{
			super.setMagicCaretPosition(pt);
			if (pt != null)
				_lastCaretPosition = pt;
		}

		/**
		 * Gets the last caret position.
		 *
		 * @return the last caret position
		 */
		public Point getLastCaretPosition()
		{
			return _lastCaretPosition;
		}

	} // end class _LyJFormattedTextCaret


	/* ********************** LyCaretListener *****************/

	/**
	 * Customizes the navigation in the textField.
	 */
	public class _LyJCaretListener
		implements CaretListener
	{

		/** The mask format. */
		LyJFormat _format;

		/**
		 * Creates a caret listener for the specified format.
		 *
		 * @param format the mask format
		 */
		public _LyJCaretListener(LyJFormat format)
		{
			_format = format;
		}

		/**
		 * Controls that the caret is at a valid position and checks the state of the buttons.
		 *
		 * @see javax.swing.event.CaretListener#caretUpdate(javax.swing.event.CaretEvent)
		 */
		public void caretUpdate(CaretEvent e)
		{
			JTextField component = (JTextField)e.getSource();
			Caret	   caret = ((JTextComponent)e.getSource()).getCaret();
			int		   dot = caret.getDot();

			Point	   pt = ((_LyJFormattedTextCaret)caret).getLastCaretPosition();

			if (pt == null)
				return;
			Position.Bias[] biasRet = new Position.Bias[1];
			int			    lastDot = component.getUI().viewToModel(component, pt, biasRet);

			int			    index = _format.getIndexSeparatorFromOffset(dot);

			if (index != -1)
			{
				if (dot == lastDot) // ce qui arrive quand on place le caret a la souris
					caret.setDot(_format.getInterval(index + 1).getStartOffset());
				else if (dot > lastDot)
					caret.setDot(_format.getInterval(index + 1).getStartOffset());
				else
				{
					// caret.setDot(_format.getInterval(index).getEndOffset());
				}
			}
			_parent.checkButtons();
		} // end method caretUpdate

	} // end class _LyJCaretListener

} // end class LyJFormattedField
