/*
 * LyJFormattedInput.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.text.*;
import leon.misc.*;
import leon.view.event.*;
import leon.view.win.swing.*;
import leon.view.win.swing.text.*;


/**
 * Enables input of formatted text. The format enables input using numeric or fixed strings in the
 * input. between brackets '[' ']'. Example [0-256].[0-256].[0-256].[0-256] for an IP address.
 *
 * @version $Id: LyJFormattedInput.java 34910 2012-03-22 07:50:26Z bangiole $
 * @author  Lyria
 * @see     Panel
 */
public class LyJFormattedInput extends LyJSelector
{
	private static final long  serialVersionUID = 8609147960201885490L;

	public static final String __VERSION =
		"$Id: LyJFormattedInput.java 34910 2012-03-22 07:50:26Z bangiole $";

	/** Constant for default document change state. */
	public static short		   DOCUMENT_CHANGE_DEFAULT = 0;

	/** Constant for key typed document change state. */
	public static short		   DOCUMENT_CHANGE_KEY_TYPED = 1;

	/** Constant for set text document change state. */
	public static short		   DOCUMENT_CHANGE_SET_TEXT = 2;

	/** Image width for the selector button width. */
	public static int		   IMAGE_W_OFFSET = 6;

	/** Image height for the selector button height. */
	public static int		   IMAGE_H_OFFSET = 6;

	/** The empty input field character. */
	public static char		   EMPTY_INPUT_FIELD = '_';

	/** The empty input field for choice elements. */
	public static String	   CHOICE_EMPTY_INPUT_FIELD = ".";

	/** Button for the <b>up</b> action. */
	protected LyJButton		   _up;

	/** Button for the <b>down</b> action. */
	protected LyJButton		   _down;

	/** Initial value of this input. */
	protected String		   _initialValue;

	/** Keeps the last mouse release date to avoid double notification on double click. */
	protected Date			   _lastMouseRelease;

	/** Delay for double click. */
	protected long			   CLICK_DELAY = 300;

	/** The graphic toolkit. */
	protected Toolkit		   _toolkit;

	/** Indicates if the debug traces must be logged. */
	protected boolean		   _showTrace;

	/** Indicates if a formatted document has changed or not. */
	protected short			   _enableDocumentChange;

	/**
	 * Creates a formated input selector object with the specified attributes.
	 *
	 * @param value   the initial value according to the specified format
	 * @param format  the format associated to this input field
	 * @param imgUp   the image for the up button
	 * @param imgDown the image for the down button
	 */
	public LyJFormattedInput(String value, String format, Image imgUp, Image imgDown)
	{
		this(value, format, imgUp, imgDown, null, true);
	}

	/**
	 * Creates a formated input selector object with the specified attributes.
	 *
	 * @param value   the initial value according to the specified format
	 * @param format  the format associated to this input field
	 * @param imgUp   the image for the up button
	 * @param imgDown the image for the down button
	 * @param env     the current environment
	 */
	public LyJFormattedInput(String value, String format, Image imgUp, Image imgDown,
		LyEnvironment env, boolean useArrows)
	{
		super(value, env, useArrows);

		_showTrace = true;
		_up = null;
		_down = null;
		_enableDocumentChange = DOCUMENT_CHANGE_DEFAULT;

		if (_useArrows)
		{
			if (imgUp != null)
				_up = addButton(imgUp, null);

			if (imgDown != null)
				_down = addButton(imgDown, null);

			// Should be done when request focus
			checkButtons();
		}
		_LyFormattedField text = ((_LyFormattedField)_text);

		text.setFormat(format);
		text.setValue(value);

		_initialValue = value;
	}

	protected void init()
	{
		super.init();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		if (actionEvent.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
		{
			if (_actionTransmitter != null)
				_actionTransmitter.processActionEvent(new LyActionEvent(this, null,
					getValue(), LyActionDemand.TYPE_VALUE_CHANGED));

			return;
		}

		_LyFormattedField text = ((_LyFormattedField)_text);

		if (actionEvent.getSource() == _up)
		{
			_LyStructureItem item = text.getCurrentItem();

			if (item != null)
			{
				String value = item.incrementValue();

				if (value != null)
					text.setItemValue(value);
				checkButtons();
			}
		}
		else if (actionEvent.getSource() == _down)
		{
			_LyStructureItem item = text.getCurrentItem();

			if (item != null)
			{
				String value = item.decrementValue();

				if (value != null)
					text.setItemValue(value);
				checkButtons();
			}
		}
		else
			super.actionPerformed(actionEvent);
	} // end method actionPerformed

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJSelector#addButton(java.awt.Image, java.lang.String)
	 */
	public LyJButton addButton(Image image, String tooltip)
	{
		LyJButton button = new LyJButton(image, tooltip)
			{
				private static final long serialVersionUID = -2343355328629947931L;

				/**
				 * This component is never focus traversable.
				 *
				 * @return <code>false</code>
				 * @see    java.awt.Component#isFocusTraversable()
				 */
				public boolean isFocusTraversable()
				{
					return false;
				}

				/*
				 * (non-Javadoc)
				 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
				 */
				public void mousePressed(MouseEvent e)
				{
					_text.requestFocus();
					super.mousePressed(e);
				}
			};

		button.setRequestFocusEnabled(false);
		button.setOpaque(true);
		button.setRollover(false);
		button.setContentAreaFilled(false);

		Icon	icon = new ImageIcon(image);
		int		topBorder = 1;

		if (_buttons.getComponentCount() > 0)
			topBorder = 0;

		button.setIcon(icon);
		button.setPreferredSize(new Dimension(icon.getIconWidth() + IMAGE_W_OFFSET,
				icon.getIconHeight()
				+ IMAGE_H_OFFSET + topBorder));
		button.setHorizontalAlignment(SwingConstants.CENTER);
		button.setVerticalAlignment(SwingConstants.CENTER);
		button.setMargin(new Insets(0, 0, 0, 0));
		button.setRepeatEnabled(true);
		button.setEnvironment(_env);

		button.setBorder(BorderFactory.createMatteBorder(topBorder, 1, 1, 1,
				UIManager.getColor("TextField.darkShadow")));

		LyActionTransmitter transmitter = button.getActionTransmitter();

		if (transmitter != null)
			transmitter.addActionListener(this);

		_buttons.add(button, buildButtonConstraint());

		return button;
	} // end method addButton

	/**
	 * Applies format for this input field.
	 *
	 * @see #getValue()
	 * @see LyJSelector#setValue(String, boolean)
	 */
	public void setValue(String value, boolean notify)
	{
		if ((value == _initialValue)
				|| ((value != null) && (_initialValue != null) && value.equals(_initialValue)))
			return;

		_LyFormattedField text = ((_LyFormattedField)_text);

		int				  pos = text.getCaretPosition();

		text.setValue(value);

		_initialValue = value;

		text._caretPosition = pos;

		try
		{
			text.setCaretPosition(pos);
		}
		catch (IllegalComponentStateException e)
		{
		}

		if (notify && !_idle && (_actionTransmitter != null))
		{
			setIdle(true); // Avoid loop if value is not stable
			_actionTransmitter.processActionEvent(new LyActionEvent(this, null, value,
					LyActionDemand.TYPE_VALUE_CHANGED));
			setIdle(false);
		}
	} // end method setValue

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.getinterface.LyGetValueInterface#getValue()
	 */
	public String getValue()
	{
		_LyFormattedField text = ((_LyFormattedField)_text);

		return text.toDisplayString();

	}

	/**
	 * Updates the buttons.
	 *
	 * @see #checkButtons()
	 * @see LyJSelector#keyPressed(KeyEvent)
	 */
	public void keyPressed(KeyEvent evt)
	{
		checkButtons();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJSelector#checkButtons()
	 */
	protected void checkButtons()
	{
		if (!isEnabled() || !isEditable())
		{
			if (_up != null)
				_up.setEnabled(false);

			if (_down != null)
				_down.setEnabled(false);
			return;
		}

		_LyFormattedField text = ((_LyFormattedField)_text);
		_LyStructureItem  item = text.getCurrentItem();
		int				  selectionCount = text.getSelectionEnd() - text.getSelectionStart();
		_LyStructureItem  itemEnd = text.getStructureItem(text.getCaretPosition()
				+ ((selectionCount > 0) ? (selectionCount - 1) : 0));

		// If there is a non-empty selection over different items, disable buttons
		if ((item == null) || ((selectionCount > 0) && (item != itemEnd)))
		{
			if (_up != null)
				_up.setEnabled(false);

			if (_down != null)
				_down.setEnabled(false);
			return;
		}

		if (_up != null)
		{
			_up.setEnabled(item.isUpButtonEnabled());
			_up.setToolTipText(item.getUpButtonToolTip());
		}

		if (_down != null)
		{
			_down.setEnabled(item.isDownButtonEnabled());
			_down.setToolTipText(item.getDownButtonToolTip());
		}
	} // end method checkButtons

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJSelector#textFieldFactory()
	 */
	public LyJTextField textFieldFactory()
	{
		return new _LyFormattedField(null, null, this, _useArrows);
	}

	/**
	 * Generates a brief sound signal.
	 *
	 * @see Toolkit#beep()
	 */
	public void beep()
	{
		if (_toolkit == null)
			_toolkit = Toolkit.getDefaultToolkit();

		if (_toolkit != null)
			_toolkit.beep();

		_enableDocumentChange = DOCUMENT_CHANGE_KEY_TYPED;
	}

	/**
	 * Transmits the completed status event to the action transmitter.
	 */
	protected void notifyComplete()
	{
		_LyFormattedField text = ((_LyFormattedField)_text);
		String			  result = text.getText();
		String			  format = text.toFormatString();

		if (result == null)
			return;

		if (result.equals(format))
			return;

		if (result.equals(getValue()))
			return;

		LyActionTransmitter transmitter = getActionTransmitter();

		transmitter.processActionEvent(new LyActionEvent(this, null, result,
				LyActionDemand.TYPE_COMPLETE_VALUE));
	}

	/**
	 * Transmits the changed values event to the action transmitter.
	 */
	protected void notifyListeners()
	{
		// check if something has changed
		LyActionTransmitter transmitter = getActionTransmitter();

		transmitter.processActionEvent(new LyActionEvent(this, null, getValue(),
				LyActionDemand.TYPE_VALUE_CHANGED));
	}

	/**
	 * Updates the content of the LyViewComponent from the graphical component.
	 *
	 * @param notify Indicates if the action listeners has to be notified of the update.
	 */
	public void updateViewContent(boolean notify)
	{
		if (notify)
			notifyComplete();

		String value = getValue();

		if (((_initialValue == null) && (value != null))
				|| ((_initialValue != null) && (!_initialValue.equals(value))))
		{
			_initialValue = value; // avoid loops
			if (notify)
				notifyListeners();
		}
	}

	/**
	 * Called when the component loses it's focus.
	 */
	public void focusLost()
	{
		if (isEnabled() && isEditable())
			notifyComplete();
	}

	/**
	 * The minimum value of a structure item is not available.
	 *
	 * @param  item the structure item is not used in this implementation
	 * @return <code>null</code>
	 */
	public String computeMinValue(_LyStructureItem item)
	{
		return null;
	}

	/**
	 * Gets the maximum length of the formatted field text.
	 *
	 * @return the maximum length of this text field
	 */
	public int getMaxLength()
	{
		_LyFormattedField text = (_LyFormattedField)_text;

		return text.getMaxLength();
	}

	/**
	 * Handles formatted documents.
	 */
	public class _LyJFormattedDocument extends LyJTextFieldDocument
	{
		private static final long    serialVersionUID = 4883479603251582306L;

		/** The format that must be respected by this document. */
		protected _LyParsedStructure _structure;

		/** The associated swing text component. */
		protected LyJTextField		 _comp;

		/**
		 * Formatted document constructor for the specified component.
		 *
		 * @param structure the structure to follow
		 * @param comp      the associated text component
		 */
		public _LyJFormattedDocument(_LyParsedStructure structure, LyJTextField comp)
		{
			super(structure.length());
			_structure = structure;
			_comp = comp;
		}

		/**
		 * Controls the user's insertions to allow only well-formed inputs.
		 *
		 * @see javax.swing.text.Document#insertString(int, java.lang.String,
		 *      javax.swing.text.AttributeSet)
		 */
		public void insertString(int offset, String str, AttributeSet a)
			throws BadLocationException
		{
			if (_enableDocumentChange == DOCUMENT_CHANGE_DEFAULT)
			{
				_LyFormattedField	text = ((_LyFormattedField)_text);
				String				oldText = text.getText();

				if (oldText == null)
					oldText = "";

				boolean		oldShowTrace = _showTrace;

				_showTrace = false;

				String		newText = oldText.substring(0, offset) + (str == null ? "" : str) + oldText.substring(offset);
				boolean		match = (text._structure == null ? true : text._structure.setValue(newText));

				_showTrace = oldShowTrace;

				if (match)
				{
					super.insertString(offset, str, a);

					if ((text._structure != null) && (newText.length() > text._structure.length()))
					{
						short	oldState = _enableDocumentChange;

						_enableDocumentChange = DOCUMENT_CHANGE_SET_TEXT;
						text.setText(newText.substring(0, text._structure.length()));
						_enableDocumentChange = oldState;
					}
				}
			}

			if (_enableDocumentChange != DOCUMENT_CHANGE_SET_TEXT)
				_enableDocumentChange = DOCUMENT_CHANGE_DEFAULT;
		}

		/**
		 * Controls the user's deletion to allow only well-formed inputs.
		 *
		 * @see javax.swing.text.Document#remove(int, int)
		 */
		public void remove(int offs, int len)
			throws BadLocationException
		{
			if (_enableDocumentChange == DOCUMENT_CHANGE_DEFAULT)
			{
				_LyFormattedField text = ((_LyFormattedField)_text);

				String	oldText = text.getText();

				if (oldText == null)
					oldText = "";

				boolean		oldShowTrace = _showTrace;

				_showTrace = false;

				String	newText = oldText.substring(0, offs) + oldText.substring(offs + len);
				boolean	match = (text._structure == null ? true : text._structure.setValue(newText));

				_showTrace = oldShowTrace;

				if (match)
				{
					super.remove(offs, len);

					if ((text._structure != null) && (newText.length() < text._structure.length()))
					{
						String	defaultString = text._structure.toString();

						if ((defaultString != null) && (defaultString.length() > newText.length()))
						{
							short	oldState = _enableDocumentChange;

							_enableDocumentChange = DOCUMENT_CHANGE_SET_TEXT;
							text.setText(newText + defaultString.substring(newText.length()));
							_enableDocumentChange = oldState;
						}
					}
				}
			}

			if (_enableDocumentChange != DOCUMENT_CHANGE_SET_TEXT)
				_enableDocumentChange = DOCUMENT_CHANGE_DEFAULT;
		}
	} // end class _LyJFormattedDocument


	/**
	 * Formatted field of a parsed value.
	 */
	public class _LyFormattedField extends LyJTextField
		implements KeyListener,
			MouseListener,
			TextListener,
			FocusListener
	{
		private static final long    serialVersionUID = 4751905209795806177L;

		/** Current format. */
		protected String			 _format;

		/** Initial default value. */
		protected String			 _defaultValue;

		/** Last valid value. */
		protected String			 _lastValue;

		/** Memory representation of format and values. */
		protected _LyParsedStructure _structure;

		/** Parent transmitter to post notifications (simpler implementation than transmitter). */
		protected LyEventTransmitter _parent;

		/** Last caret position when focus is lost, -1 otherwise. */
		protected int				 _caretPosition;

		/** Indicates if we accept arrow keys. */
		protected boolean			 _useArrows;

		/**
		 * Creates a new _LyFormattedField object with default attributes.
		 */
		public _LyFormattedField()
		{
			this(null, null, null, true);
		}

		/**
		 * Creates a new _LyFormattedField object with the specified attributes.
		 *
		 * @param value  the current value. At first, this is the same as the last value.
		 * @param format the format to use
		 * @param parent the parent transmitter
		 */
		public _LyFormattedField(String value, String format, LyEventTransmitter parent,
			boolean useArrows)
		{
			super(_env);

			_format = format;
			_defaultValue = value;
			_lastValue = value;
			_parent = parent;
			_useArrows = useArrows;
			init();
		}

		/**
		 * Initializes this field component and registers its event listeners (keyboard, mouse and
		 * focus).
		 */
		public void init()
		{
			if (_structure != null)
				setColumns(_structure.length());
			else
				_structure = new _LyParsedStructure(_format, _defaultValue);

			setColumns(_structure.length());

			// Register for user interaction
			this.addKeyListener(this);
			this.addMouseListener(this);

			// TODO
			// this.addTextListener(this);

			_LyJFormattedDocument doc = new _LyJFormattedDocument(_structure, this);

			setDocument(doc);

			this.addFocusListener(this);
		}

		/**
		 * Returns the average width of a character for the current font.
		 *
		 * @see javax.swing.JTextField#getColumnWidth()
		 */
		public int getColumnWidth()
		{
			return LySwingUtils.getFontCharWidth(this, getFont());
		}

		/**
		 * Creates the content string according to format and current content. If the structure is
		 * <code>null</code>, initializes it at the current value.
		 *
		 * @param  currentValue the current value of this field
		 * @return true if this field text has been successfuly set, false otherwise
		 * @see    #getValue()
		 */
		public boolean setValue(String currentValue)
		{
			if (_structure == null)
			{
				_structure = new _LyParsedStructure(_format, currentValue);

				String resu = _structure.toString();

				_lastValue = resu;
				setText(resu);
				setNCols(_structure.length());
				setMaxChars(_structure.length());

				// FFT00525
				return true;
			}
			else if (_structure.setValue(currentValue))
			{
				String resu = _structure.toString();

				_lastValue = resu;
				setText(resu);
				// FFT00525
				return true;
			}
			else
			{
				if ((_env != null) && (_env.isDebugLevel(LyEnvironment.DEBUG_WARNING))
						&& _showTrace)
					_env.logWarning("[LyFormattedInput] SetValue denied : " + currentValue);
				return false;
			}
		} // end method setValue

		/**
		 * Nothing is done when a key is typed.
		 *
		 * @see java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
		 */
		public void keyTyped(KeyEvent evt)
		{
		}

		/*
		 * (non-Javadoc)
		 * @see javax.swing.text.JTextComponent#setEnabled(boolean)
		 */
		public void setEnabled(boolean enabled)
		{
			super.setEnabled(enabled);

			if (!enabled)
			{
				String value = _structure.toFormatString();

				_lastValue = value;
				_initialValue = null;

				setText(value);
			}
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
		 */
		public void keyPressed(KeyEvent evt)
		{
			// We do have the focus
			_caretPosition = -1;

			int  key = evt.getKeyCode();
			char keyChar = evt.getKeyChar();
			int  pos = getCaretPosition();

			// check cursor and move caretto nex valid field
			if ((!evt.isShiftDown()) && (key == KeyEvent.VK_LEFT))
			{
				pos = getNextCaretPosition(pos, false);
				setCaretPosition(pos);
				evt.consume();
			}
			else if ((!evt.isShiftDown()) && (key == KeyEvent.VK_RIGHT))
			{
				pos = getNextCaretPosition(pos, true);
				setCaretPosition(pos);
				evt.consume();
			}
			else if ((_useArrows) && (key == KeyEvent.VK_DOWN))
			{
				if (!_text.isEditable() || !_text.isEnabled())
					return;

				_LyStructureItem item = getCurrentItem();

				if (item != null)
				{
					if (!item.isDownButtonEnabled())
					{
						setSelectionStart(getCaretPosition());
						setSelectionEnd(getCaretPosition());
						evt.consume();
						return;
					}

					// The value change may change the position of the structure item
					int    initPos = getStructureItemPos(item); // Pos before change

					String value = item.decrementValue();

					if (value != null)
						setItemValue(value);

					int endPos = getStructureItemPos(item); // Pos after change

					setCaretPosition(_caretPosition - (initPos - endPos));
				}
				else
					_caretPosition = -1;

				setSelectionStart(getCaretPosition());
				setSelectionEnd(getCaretPosition());
				evt.consume();
				return;
			}
			else if ((_useArrows) && (key == KeyEvent.VK_UP))
			{
				if (!_text.isEditable() || !_text.isEnabled())
					return;

				_LyStructureItem item = getCurrentItem();

				if (item != null)
				{
					if (!item.isUpButtonEnabled())
					{
						setSelectionStart(getCaretPosition());
						setSelectionEnd(getCaretPosition());
						evt.consume();
						return;
					}

					// The value change may change the position of the structure item
					int    initPos = getStructureItemPos(item); // Pos before change

					String value = item.incrementValue();

					if (value != null)
						setItemValue(value);

					int endPos = getStructureItemPos(item); // Pos after change

					_caretPosition = _caretPosition - (initPos - endPos);
				}
				else
					_caretPosition = -1;

				setSelectionStart(getCaretPosition());
				setSelectionEnd(getCaretPosition());
				evt.consume();
				return;
			}
			else if ((key != KeyEvent.VK_LEFT) && (key != KeyEvent.VK_RIGHT)
					// (key != KeyEvent.VK_INSERT) &&
					&& (key != KeyEvent.VK_CONTROL) && (key != KeyEvent.VK_SHIFT)
					&& (key != KeyEvent.VK_ALT) && (key != KeyEvent.VK_TAB)
					&& (key != KeyEvent.VK_END) && (key != KeyEvent.VK_HOME))
			{
				int				 start = getSelectionStart();
				int				 end = getSelectionEnd();

				boolean	copyPaste = (evt.isControlDown() && ((key == KeyEvent.VK_V) || (key == KeyEvent.VK_C)));

				if (!copyPaste)
				{
					checkKeyEvent(start, end, keyChar, (key == KeyEvent.VK_DELETE), (key == KeyEvent.VK_BACK_SPACE));
					evt.consume();
				}
			}
		} // end method keyPressed

		protected void checkKeyEvent(int start, int end, char keyChar, boolean delete, boolean backSpace)
		{
			String			 current = getText();
			StringBuffer     value = new StringBuffer(50);
			_LyStructureItem item = getStructureItem(start);
			_LyStructureItem item2 = getStructureItem(end);
			int				 itemStart = getStructureItemPos(start);
			int 			 pos = getCaretPosition();

			// without any selection : INSERT
			if (delete)
			{
				int itemEnd = ((item == null) ? end : (itemStart + item.requiredLength()));

				if (item2 != item)
					itemEnd = end;

				value.append(current.substring(0, start));

				if (start == end)
				{
					// scroll content
					if (item instanceof _LyChoiceStructure)
					{
						item.setValue(null);
						value.append(item.toString());
					}
					else if ((start + 1) < itemEnd)
						value.append(current.substring(start + 1, itemEnd));

					if (!(item instanceof _LyChoiceStructure))
						value.append(LyJFormattedInput.EMPTY_INPUT_FIELD);

					if (itemEnd < current.length())
						value.append(current.substring(itemEnd));
				}
				else
				{
					if ((end < itemEnd) && (current.length() >= itemEnd))
						value.append(current.substring(end, itemEnd));

					int    i, n = end - start;
					String empty =
						(((item instanceof _LyChoiceStructure)
									&& (item2 instanceof _LyChoiceStructure))
							? CHOICE_EMPTY_INPUT_FIELD : (EMPTY_INPUT_FIELD + ""));

					for (i = 0; i < n; i++)
						value.append(empty);
					if (itemEnd < current.length())
						value.append(current.substring(itemEnd));
				}

				// Particular case : the entire text is selected and DEL key pressed
				if ((start == 0) && (end == current.length()))
				{
					value = new StringBuffer();
					pos = itemStart;
				}

				setValue(value.toString());
				setCaretPosition(pos);
			}
			else if (backSpace && (start > 0))
			{
				// If we are at the end (itemstart == -1)
				if ((item == null) && (itemStart == -1))
				{
					item = getStructureItem(start - 1);
					itemStart = getStructureItemPos(start - 1);
					if (item instanceof _LyFixedStructure)
						start = itemStart;
				}

				// Skip backspace : at the begining of current item

				// Propagate to current OR previous item ? Yes if previous item is
				// editable (Case : we are at the end of the editable section)
				if (start == itemStart)
				{
					// We are at the begining of the item : dele occurs on PREVIOUS item;
					do
					{
						item = getStructureItem(start - 1);
						itemStart = getStructureItemPos(start - 1);
						if ((item instanceof _LyFixedStructure) && (start == end))
						{
							start = itemStart;
							end = itemStart;
						}
					}
					while ((item instanceof _LyFixedStructure) && (start == end));

				}

				if (start > itemStart)
				{
					int itemEnd = ((item == null) ? end : (itemStart + item.requiredLength()));
					int i, n = end - start;

					if (item instanceof _LyChoiceStructure)
					{
						item.setValue(null);
						value.append(current.substring(0, itemStart));
						value.append(item.toString());
					}
					else if (start > 0)
					{
						value.append(current.substring(0, start - 1));

						if (end < itemEnd)
							value.append(current.substring(end, itemEnd));

						for (i = 0; i <= n; i++)
							value.append(LyJFormattedInput.EMPTY_INPUT_FIELD);
					}

					if (!(item instanceof _LyChoiceStructure))
						value.append(current.substring(itemEnd));

					setValue(value.toString());
				}

				pos = getNextCaretPosition(pos, false);
				setCaretPosition(pos);
				_caretPosition = pos;
			}
			else if (start == end)
			{
				// Insert
				if ((item != null) && (item.isValid(keyChar)))
				{
					// ** Attempt to INSERT
					if (item.enableInsert(keyChar, start - itemStart))
					{
						int l = item.requiredLength();

						value.append(current.substring(0, start)); // before item and until
						// start
						value.append(keyChar);
						value.append(current.substring(start, itemStart + l - 1)); // skip last
						// char :
						// replaced
						if ((itemStart + l) < current.length())
							value.append(current.substring(itemStart + l));

						// FFT00525
						if (setValue(value.toString()))
						{
							pos = getNextCaretPosition(pos, true);

							// check if the field is full and if it is the same item
							if ((pos == (itemStart + l)) && (item == getStructureItem(pos)))
								pos = getNextCaretPosition(pos, true);
						}
						else
							beep();
						setCaretPosition(pos);
						_caretPosition = pos;
					}
					// ** ITEM may be FULL attempt to REPLACE
					else if (item.enableReplace(keyChar, start - itemStart, 1))
					{
						int l = item.requiredLength();

						value.append(current.substring(0, start)); // before item and until
						// start
						value.append(keyChar);
						value.append(current.substring(start + 1)); // skip char after start :
						// replaced

						// FFT00525
						if (setValue(value.toString()))
						{
							pos = getNextCaretPosition(pos, true);

							// check if the field is full and if it is the same item
							if ((pos == (itemStart + l)) && (item == getStructureItem(pos)))
								pos = getNextCaretPosition(pos, true);
						}
						else
							beep();

						setCaretPosition(pos);
						_caretPosition = pos;
					}
					else
						beep();
				}
				else
					beep();
			}
			else
			{
				// Replace
				if ((item != null)
						&& (item.enableReplace(keyChar, start - itemStart, end - start)))
				{
					if (start > 0)
						value.append(current.substring(0, start));

					value.append(keyChar);

					for (int j = 0; j < (end - start - 1); j++)
						value.append(LyJFormattedInput.EMPTY_INPUT_FIELD);

					value.append(current.substring(end));

					setValue(value.toString());

					pos = getNextCaretPosition(start, true);
					setCaretPosition(pos);
					_caretPosition = pos;
				}
				else
					beep();
			}

			_enableDocumentChange = DOCUMENT_CHANGE_KEY_TYPED;
		}

		/**
		 * Checks the selection upon key release.
		 *
		 * @see java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
		 */
		public void keyReleased(KeyEvent evt)
		{
			checkSelection();
		}

		/* Mouse Listener interface */

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
		 */
		public void mouseClicked(MouseEvent e)
		{
			if (e.getClickCount() == 2)
			{
				// Select clicked component

				setSelectionStart(0);
				setSelectionEnd(_structure.length());
			}
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
		 */
		public void mousePressed(MouseEvent e)
		{
			if (_lastMouseRelease != null)
			{
				// Skip second mouse pressed after mouse release (double clic)
				Date d = new Date();

				if ((d.getTime() - _lastMouseRelease.getTime()) < CLICK_DELAY)
					return;
			}

			_caretPosition = -1;
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
		 */
		public void mouseReleased(MouseEvent e)
		{

			checkSelection();
			if (_lastMouseRelease != null)
			{
				Date d = new Date();
				// Skip second mouse pressed after mouse release (double clic)
				if ((d.getTime() - _lastMouseRelease.getTime()) < CLICK_DELAY)
					return;
			}

			// Check if current caret position is OK
			int				 pos = getCaretPosition();
			_LyStructureItem item = getStructureItem(pos);
			int				 itemStart = getStructureItemPos(pos);

			if ((item != null) && (!item.enableCaret(pos - itemStart)))
			{
				pos = getNextCaretPosition(pos, false);
				setCaretPosition(pos);
			}

			_caretPosition = pos;

			checkButtons();

			_lastMouseRelease = new Date();
		} // end method mouseReleased

		/**
		 * Nothing is done when a mouse enters a component.
		 *
		 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
		 */
		public void mouseEntered(MouseEvent e)
		{
		}

		/**
		 * Nothing is done when a mouse exits a component.
		 *
		 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
		 */
		public void mouseExited(MouseEvent e)
		{
		}

		/**
		 * Ensures that the selection is valid.
		 */
		public void checkSelection()
		{
			// check selection
			int start = getSelectionStart();
			int end = getSelectionEnd();
			int pos = getCaretPosition();

			if (start != end)
			{
				_LyStructureItem item, currentItem;

				// Attempt to get component at pos
				currentItem = getStructureItem(pos);

				if ((currentItem != null) && (!currentItem.enableSelect(0, 0)))
				{
					// Not a selectable item : fetch first selectable
					currentItem = null;
					setSelectionStart(start);
					setSelectionEnd(start);
					return;
				}

				int n = _structure.size();

				// within the same item ?
				int itemStart = 0;
				int itemEnd;

				for (int i = 0; i < n; i++)
				{
					item = _structure.elementAt(i);
					itemEnd = itemStart + item.requiredLength();

					// Wait for currentItem if any
					if ((currentItem != null) && (item == currentItem))
					{
						currentItem = null; // Do not wait any longer

						boolean checkStart = false;
						boolean checkEnd = false;

						// within this item for start :
						if ((start < itemStart)
								|| (start < itemEnd))
							checkStart = true;

						// within this item for end :
						if ((end > itemStart) && (end <= itemEnd))
							checkEnd = true;

						if (checkStart && checkEnd)
						{
							// selection within this item
							if (start < itemStart)
								start = itemStart;

							if (!item.enableSelect(start - itemStart, (start - end)))
							{
								setSelectionStart(start);
								setSelectionEnd(start);
							}

							break;
						}
						else if (checkStart)
						{
							// selection within this item
							if (start < itemStart)
								start = itemStart;

							setSelectionStart(start);

							if (!item.enableSelect(start - itemStart, (start - end)))
								setSelectionEnd(start);
							else
								setSelectionEnd(itemEnd);
							break;
						}
					}
					itemStart = itemEnd; // next
				}
			}
		} // end method checkSelection

		/**
		 * Retreives the structure item at the specified position.
		 *
		 * @param  pos the position of the structure item
		 * @return the structure item at the given position, or <code>null</code> if the item cannot
		 *         be found
		 */
		protected _LyStructureItem getStructureItem(int pos)
		{
			int n = (_structure == null) ? 0 : _structure.size();
			int itemStart = 0;
			int itemEnd;

			for (int i = 0; i < n; i++)
			{
				_LyStructureItem item = _structure.elementAt(i);

				itemEnd = itemStart + item.requiredLength();

				// within this item for start :
				if ((pos >= itemStart) && (pos < itemEnd))
					return item;

				itemStart = itemEnd; // next
			}
			return null;
		}

		/**
		 * Retreives the start position of the structure item at the specified position.
		 *
		 * @param  pos the position of the structure item
		 * @return the structure item start position at the given position, or -1 if the item cannot
		 *         be found
		 */
		protected int getStructureItemPos(int pos)
		{
			int n = _structure.size();
			int itemStart = 0;
			int itemEnd;

			for (int i = 0; i < n; i++)
			{
				_LyStructureItem item = _structure.elementAt(i);

				itemEnd = itemStart + item.requiredLength();

				// within this item for start :
				if ((pos >= itemStart) && (pos < itemEnd))
					return itemStart;

				itemStart = itemEnd; // next
			}
			return -1;
		}

		/**
		 * Retreives the start position of the specified structure item.
		 *
		 * @param  currentItem the structure item
		 * @return the structure item start position at the given position, or -1 if the item cannot
		 *         be found
		 */
		protected int getStructureItemPos(_LyStructureItem currentItem)
		{
			int				 i, n = ((_structure == null) ? 0 : _structure.size());
			int				 itemStart = 0, itemEnd;
			_LyStructureItem item;

			for (i = 0; i < n; i++)
			{
				item = _structure.elementAt(i);
				itemEnd = itemStart + item.requiredLength();

				// within this item for start :
				if (item == currentItem)
					return itemStart;

				itemStart = itemEnd; // next
			}

			return -1;
		}

		/**
		 * Retreives the next stopping caret position on the edition area.
		 *
		 * @param  pos   the start position
		 * @param  right Indicates if the search must be done to the right or to the left.
		 * @return the next caret position in the edition area
		 */
		protected int getNextCaretPosition(int pos, boolean right)
		{
			if (right)
			{
				_LyStructureItem item = getStructureItem(pos);
				int				 itemStart = getStructureItemPos(pos);

				if (item != null)
				{
					if ((!item.enableCaret(pos - itemStart + 1)) // Not a valid pos
							|| (pos == (itemStart + item.requiredLength() - 1))) // Last pos of the
					// input. -> next
					{
						while ((item != null) && (pos >= 0))
						{
							// move caret to next input area
							pos = itemStart + item.requiredLength();
							item = getStructureItem(pos);
							itemStart = getStructureItemPos(pos);

							if ((item != null) && (item.enableCaret(pos - itemStart)))
								return (itemStart);
						}
						return (pos);
					}
				}
				return pos + 1;
			}
			if (pos > 0)
			{
				_LyStructureItem item = getStructureItem(pos - 1);
				int				 itemStart = getStructureItemPos(pos - 1);

				if ((item != null) && (!item.enableCaret(pos - itemStart - 1)))
				{
					while ((item != null) && (pos > 0))
					{
						// move caret to previous input area
						pos = itemStart;
						item = getStructureItem(pos - 1);
						itemStart = getStructureItemPos(pos - 1);
						if ((item != null) && (item.enableCaret(pos - itemStart - 1)))
							return itemStart + item.requiredLength();
					}
					return pos;
				}
				return pos - 1;
			}
			return pos;
		} // end method getNextCaretPosition

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.TextListener#textValueChanged(java.awt.event.TextEvent)
		 */
		public void textValueChanged(TextEvent e)
		{
			if ((_lastValue != null) && (_lastValue.equals(getText())))
				return; // Already validated
			else if (_structure.setValue(getText()))
				_lastValue = getText();
			else
			{
				beep();

				setText(_lastValue);
			}
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.FocusListener#focusGained(java.awt.event.FocusEvent)
		 */
		public void focusGained(FocusEvent e)
		{
			_initialValue = getText();
			_caretPosition = getCaretPosition();
			setCaretPosition(_caretPosition);
		}

		/*
		 * (non-Javadoc)
		 * @see javax.swing.text.JTextComponent#setCaretPosition(int)
		 */
		public void setCaretPosition(int position)
		{
			Document doc = getDocument();

			if (doc != null)
			{
				if (position > doc.getLength())
					position = doc.getLength();
			}

			super.setCaretPosition(position);
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.FocusListener#focusLost(java.awt.event.FocusEvent)
		 */
		public void focusLost(FocusEvent e)
		{
			if (!isEnabled() || !isEditable())
				return;

			_hasGainedFocus = false;

			// keep last caret pos.
			_caretPosition = getCaretPosition();

			LyJFormattedInput.this.focusLost();

			if (_structure != null)
			{
				String value = _structure.toDisplayString();

				if (value == null)
					value = _structure.toString();
				else
					setText(value);

				if (((_initialValue == null) && (value != null))
						|| ((_initialValue != null) && (!_initialValue.equals(value))))
					notifyListeners();
			}
		}

		/**
		 * Transmits the changed values event to the action transmitter.
		 */
		protected void notifyListeners()
		{
			// check if something has changed
			if ((_parent != null) && (_parent.getActionTransmitter() != null))
			{
				LyActionTransmitter transmitter = _parent.getActionTransmitter();

				transmitter.processActionEvent(new LyActionEvent(this, null,
				// getText(),
						getValue(), LyActionDemand.TYPE_VALUE_CHANGED));
			}
		}

		/**
		 * Gets the displayed value of the structure.
		 *
		 * @return the value of this structure item
		 * @see    #setValue(String)
		 * @see    _LyParsedStructure#toDisplayString()
		 */
		protected String getValue()
		{
			return _structure.toDisplayString();
		}

		/*
		 * (non-Javadoc)
		 * @see javax.swing.text.JTextComponent#getCaretPosition()
		 */
		public int getCaretPosition()
		{
			if (_caretPosition >= 0)
				return _caretPosition;
			return super.getCaretPosition();
		}

		/**
		 * Gets the current format.
		 *
		 * @return the format
		 * @see    #setFormat(String)
		 */
		public String getFormat()
		{
			return _format;
		}

		/**
		 * Sets the current format to the specified one, then recreates this structure item.
		 *
		 * @param format the new format of this structure
		 * @see   #getFormat()
		 */
		public void setFormat(String format)
		{
			_format = format;
			_structure = null;

			setValue(_lastValue);
		}

		/**
		 * Gets the structure item at the current caret position.
		 *
		 * @return the structure item at the caret position, or <code>null</code> if there is none
		 */
		protected _LyStructureItem getCurrentItem()
		{
			int				 pos = getCaretPosition();

			_LyStructureItem item = getStructureItem(pos);

			if (item != null)
			{
				// check if set is enabled
				if (!item.enableCaret(0))
				{
					pos = getNextCaretPosition(pos, false);
					item = getStructureItem(pos);
				}
			}
			return item;
		}

		/**
		 * Gets the current structure item value.
		 *
		 * @return the value of the structure item at the current caret position, or -1 if there is
		 *         none
		 * @see    #setItemVal(int)
		 */
		public int getItemVal()
		{
			_LyStructureItem item = getCurrentItem();

			return ((item == null) ? -1 : LyTools.intFromString(item.toDisplayString()));
		}

		/**
		 * Gets the current structure item string.
		 *
		 * @return the displayed string of the structure item at the current caret position, or
		 *         <code>null</code> if there is none
		 * @see    #setItemValue(String)
		 */
		public String getItemString()
		{
			_LyStructureItem item = getCurrentItem();

			return ((item == null) ? null : item.toDisplayString());
		}

		/**
		 * Sets the current structure item value.
		 *
		 * @param val the new value
		 * @see   #getItemVal()
		 */
		public void setItemVal(int val)
		{
			_LyStructureItem item = getCurrentItem();

			if (item != null)
			{
				item.setValue(Integer.toString(val));

				int pos = getCaretPosition();

				setText(_structure.toDisplayString()); // force auto formating
				setCaretPosition(pos);
				notifyListeners();
				this.requestFocus();
			}
		}

		/**
		 * Builds the string of this structure item.
		 *
		 * @return the string formating this structure item
		 * @see    _LyParsedStructure#toFormatString()
		 */
		public String toFormatString()
		{
			return _structure.toFormatString();
		}

		/**
		 * Gets the maximum length of this structure.
		 *
		 * @return the maximum length of this structure, or 10 if the structure is <code>null</code>
		 */
		public int getMaxLength()
		{
			return ((_structure == null) ? 10 : _structure.length());
		}

		/**
		 * Builds the representation string of this structure item.
		 *
		 * @return the string representing this structure item
		 * @see    _LyParsedStructure#toDisplayString()
		 */
		public String toDisplayString()
		{
			return _structure.toDisplayString();
		}

		/**
		 * Sets the current structure item string.
		 *
		 * @param value the new value
		 * @see   #getItemString()
		 */
		public void setItemValue(String value)
		{
			_LyStructureItem item = getCurrentItem();

			if (item != null)
			{
				item.setValue(value);

				int    pos = getCaretPosition();
				String text = _structure.toDisplayString();

				if (text == null)
					text = "";
				setText(text); // force auto formating
				setSelectionStart(pos);
				setSelectionEnd(pos);
				LyJFormattedInput.this.notifyListeners();
				this.requestFocus();
			}
		}
	} // end class _LyFormattedField


	/**
	 * Data structure which holds the required data format and values.
	 */
	protected class _LyParsedStructure
	{
		/** The format. */
		protected String _format;

		/** List of data structures. */
		protected Vector<_LyStructureItem> _structureList;

		/**
		 * Class constructor of a parsed structure object.
		 *
		 * @param format the format of this structure
		 * @param value  the item value
		 */
		public _LyParsedStructure(String format, String value)
		{
			_format = format;

			init();

			setValue(value);
		}

		/**
		 * Initializes the structure list and parses this class format.
		 */
		public void init()
		{
			_structureList = new Vector<_LyStructureItem>();

			parseFormat();
		}

		/**
		 * Sets the new format and populates the list of structures.
		 */
		public void parseFormat()
		{
			String			 remaining = _format;
			int				 n1, n2, n3, n4, n5, n6;
			int				 minValue, maxValue, pitch;
			_LyStructureItem item;
			StringBuffer     buffer;

			while ((remaining != null) && (remaining.length() > 0))
			{
				buffer = new StringBuffer(remaining);
				n1 = LyString.indexOf(buffer, '[', 0);
				n2 = LyString.indexOf(buffer, ']', n1);
				n3 = LyString.indexOf(buffer, '-', n1);
				n5 = LyString.indexOf(buffer, ';', n1);
				n6 = LyString.indexOf(buffer, '|', n1);
				remaining = buffer.toString();

				// we should have two brackets, otherwise, keep string as is
				if ((n1 < 0) || (n2 < 0))
				{
					if ((n1 >= 0) && (_env != null)
							&& (_env.isDebugLevel(LyEnvironment.DEBUG_ERROR)))
						_env.logError("[LyFormattedInput] Possible invalid format : " + _format
							+ " while parsing : " + remaining);

					item = new _LyFixedStructure(remaining);
					_structureList.addElement(item);
					remaining = ""; // done
				}
				else
				{
					if (n1 > 0)
					{
						// fixed string before input
						item = new _LyFixedStructure(remaining.substring(0, n1));
						_structureList.addElement(item);
					}

					if ((n5 > n1) && (n5 < n2))
					{
						// Alphabetic interval
						String allStruct = remaining.substring(n1 + 1, n2);

						item = new _LyAlphabeticIntervalsStructure(allStruct);
					}
					else if ((n6 > n1) && (n6 < n2))
					{
						// choice value
						String allStruct = remaining.substring(n1 + 1, n2);

						item = new _LyChoiceStructure(allStruct);
					}
					// we should have two brackets and a dash inside,
					// otherwise keep string as is until closing dash
					else if ((n3 < 0) || (n3 > n2))
					{
						if ((_env != null) && (_env.isDebugLevel(LyEnvironment.DEBUG_ERROR)))
							_env.logError("[LyFormattedInput] Possible invalid format : " + _format
								+ " while parsing ('-' not found in brackets '[' ']' : "
								+ remaining.substring(n1, n2));

						item = new _LyFixedStructure(remaining.substring(n1, n2 + 1));
						_structureList.addElement(item);
					}
					else
					// alphabetic or numeric
					{
						boolean isNumeric = true;

						if (n3 > n1)
						{
							// two case for now: alphabetic , alphabetic interval, choices, or
							// numeric
							String value = remaining.substring(n1 + 1, n3);

							// first try number
							try
							{
								minValue = Integer.parseInt(value);
							}
							catch (NumberFormatException e)
							{
								isNumeric = false;
							}
						}
						else
							isNumeric = false;

						if (isNumeric)
						{
							String value = remaining.substring(n1 + 1, n3);
							// get min and max
							minValue = LyTools.intFromString(value);
							pitch = 1;

							// A pitch may be specified : ":"
							n4 = remaining.indexOf(":", n3 + 1);

							if ((n4 < 0) || (n4 > n2))
								maxValue = LyTools.intFromString(remaining.substring(n3 + 1, n2));
							else
							{
								maxValue = LyTools.intFromString(remaining.substring(n3 + 1, n4));
								pitch = LyTools.intFromString(remaining.substring(n4 + 1, n2));
							}

							item = new _LyNumericStructure("", minValue, maxValue, pitch);

						}
						else
						{
							String value = remaining.substring(n1 + 1, n3);
							String lowerValue = value;
							String highValue = remaining.substring(n3 + 1, n2);

							char   minChar = lowerValue.charAt(0);
							char   maxChar = highValue.charAt(0);

							item = new _LyAlphabeticStructure((char)0, minChar, maxChar);
						}
					}
					_structureList.addElement(item);
					remaining = remaining.substring(n2 + 1);
				}
			}
		} // end method parseFormat

		/**
		 * Sets the specified value in structure items.
		 *
		 * @param  value the new items value
		 * @return true if the setting has been successful, false otherwise
		 */
		public boolean setValue(String value)
		{
			boolean result = true;
			int     length = length();

			if ((value != null) && (value.equals("")))
				value = null;

			int				 i, n = _structureList.size();
			int				 l, pos = 0;
			_LyStructureItem item;
			String			 itemVal;
			boolean			 tmpResult = false;

			for (i = 0; i < n; i++)
			{
				item = _structureList.elementAt(i);
				l = item.requiredLength();
				if (value == null)
					itemVal = "";
				else
				{
					if ((pos + l) >= value.length())
						l = value.length() - pos;

					itemVal = value.substring(pos, pos + l);
				}

				if (item instanceof _LyChoiceStructure)
				{
					if (itemVal.length() == 0)
						result = result & item.setValue(itemVal);
					else
					{
						tmpResult = false;
						while ((l > 0) && !tmpResult)
						{
							tmpResult = item.setValue(itemVal);
							if (tmpResult)
							{
								l = item.toString().length();
								break;
							}
							l--;
							if (l > 0)
								itemVal = value.substring(pos, pos + l);
						}

						result = result & tmpResult;
					}
				}
				else if ((item instanceof _LyNumericStructure) && (itemVal.length() >= 1))
				{
					// a number may be written as 1, an should be 01, or 001 ...
					// Try to guess
					StringBuffer sb = new StringBuffer(l);
					int			 c = 0;

					for (c = 0; c < l; c++)
					{
						if (Character.isDigit(itemVal.charAt(c))
								|| (LyJFormattedInput.EMPTY_INPUT_FIELD == itemVal.charAt(c)))
							sb.append(itemVal.charAt(c));
						else
						{
							// Non numeric character ! Break here
							// Change position in the structure and Break
							l = c;
							result = false;
						}
					}

					for (c = l; c < item.requiredLength(); c++)
						sb.insert(0, "0"); // Complete with '0' the numeric value

					// Change itemVal
					itemVal = sb.toString();

					result = result & item.setValue(itemVal);
				}
				else if (!(item instanceof _LyFixedStructure))
					result = result & item.setValue(itemVal);

				pos += l;
			}

			if (!result && (_env != null) && _showTrace
					&& (_env.isDebugLevel(LyEnvironment.DEBUG_WARNING)))
				_env.logWarning("[LyFormattedInput] Do not know how to set : " + value
					+ " - length : " + length);

			return result;
		} // end method setValue

		/**
		 * Computes the lenght of current input.
		 *
		 * @return the length
		 */
		public int length()
		{
			int l = 0;

			for (int i = 0; i < _structureList.size(); i++)
			{
				_LyStructureItem item = _structureList.elementAt(i);

				l += item.requiredLength();
			}

			return l;
		}

		/**
		 * Returns the structure item at the specified index.
		 *
		 * @param  idx index of the item
		 * @return the element at specified index, <code>null</code> if not found
		 */
		public _LyStructureItem elementAt(int idx)
		{
			if ((idx >= 0) && (idx < _structureList.size()))
				return _structureList.elementAt(idx);

			return null; // not found
		}

		/**
		 * Returns the number of structure elements in this object.
		 *
		 * @return the number of elements
		 */
		public int size()
		{
			return _structureList.size();
		}

		/**
		 * Builds the string representation of the content and its value.
		 *
		 * @return the string to display
		 */
		public String toString()
		{
			StringBuffer sb = new StringBuffer(50);

			for (int i = 0; i < _structureList.size(); i++)
			{
				_LyStructureItem item = _structureList.elementAt(i);

				sb.append(item.toString());
			}

			return sb.toString();
		}

		/**
		 * Builds the string representation of the format.
		 *
		 * @return the format string
		 */
		public String toFormatString()
		{
			StringBuffer     buffer = new StringBuffer(50);
			int				 i, n = _structureList.size();
			_LyStructureItem item;

			for (i = 0; i < n; i++)
			{
				item = _structureList.elementAt(i);
				buffer.append(item.toFormatString());
			}

			return buffer.toString();
		}

		/**
		 * Builds the string representation of the content and value. For every empty value, uses
		 * the minimum available.
		 *
		 * @return the string to display
		 */
		public String toDisplayString()
		{
			StringBuffer     buffer = new StringBuffer(50);
			boolean			 isNull = true;
			int				 i, n = _structureList.size();
			_LyStructureItem item;
			String			 temp;

			for (i = 0; i < n; i++)
			{
				item = _structureList.elementAt(i);
				temp = item.toDisplayString();

				if (temp == null)
				{
					temp = item.toMinString();

					// for (j = 0, l = item.requiredLength(); j < l; j++)
					// temp += item.getMin();

				}
				else if ((temp != null) && !(item instanceof _LyFixedStructure))
					isNull = false;

				buffer.append(temp);
			}

			return (isNull ? null : buffer.toString());
		} // end method toDisplayString
	} // end class _LyParsedStructure

	/**
	 * Data structure which holds the required data format and values.
	 */
	protected interface _LyStructureItem
	{
		String toString();

		/**
		 * Builds the string representation of the format.
		 *
		 * @return the format string
		 */
		String toFormatString();

		/**
		 * Builds the string representation of the content and value. For every empty value, uses
		 * the minimum available.
		 *
		 * @return the string to display
		 */
		String toDisplayString();

		/**
		 * Returns a string representing the minimum value of the item.
		 *
		 * @return the minimum string
		 */
		String toMinString();

		/**
		 * Indicates if the caret can be at the specified position.
		 *
		 * @param  pos the requested caret position
		 * @return <code>true</code> if the caret can be placed at the specified position, <code>
		 *         false</code> otherwise
		 */
		boolean enableCaret(int pos);

		/**
		 * Indicates if the specified character can be inserted at the given position.
		 *
		 * @param  c   the character to insert
		 * @param  pos the position of the insertion
		 * @return <code>true</code> if the character can be inserted, <code>false</code> otherwise
		 */
		boolean enableInsert(char c, int pos);

		/**
		 * Indicates if the specified character needs optional zeros to be added at the specified
		 * position.
		 *
		 * @param  c   the numbered characted
		 * @param  pos the position
		 * @return <code>true</code> if the character needs zeros at the specified position, <code>
		 *         false</code> otherwise
		 */
		boolean needZero(char c, int pos);

		/**
		 * Indicates if the specified selection (designated by its position and its length) can be
		 * replaced by the specified character.
		 *
		 * @param  c      the character replacing the selection
		 * @param  pos    the starting position of the selection
		 * @param  length the length of the selection
		 * @return <code>true</code> if the character can replace the selection, <code>false</code>
		 *         otherwise
		 */
		boolean enableReplace(char c, int pos, int length);

		/**
		 * Indicates if the specified selection (designated by its position and its length) can be
		 * done.
		 *
		 * @param  pos    the starting position of the selection
		 * @param  length the length of the selection
		 * @return <code>true</code> if the selection can be done, <code>false</code> otherwise
		 */
		boolean enableSelect(int pos, int length);

		/**
		 * Sets the specified value in structure items.
		 *
		 * @param  value the new items value
		 * @return true if the setting has been successful, false otherwise
		 */
		boolean setValue(String value);

		/**
		 * Gets the length necessary to validate this field.
		 *
		 * @return the number of characters required for this field
		 */
		int requiredLength();

		// public int getMin();

		// public int getMax();

		// public int getPitch();

		/**
		 * Returns the following position of the caret, from the specified caret position.
		 *
		 * @param  currentPos the current caret position
		 * @param  next       Indicates if the caret position is the next or the previous.
		 * @return the new caret position, -1 if not in this component
		 */
		int getCaretPosition(int currentPos, boolean next);

		/**
		 * Checks if the given character is valid for the structure.
		 *
		 * @param  character the character to check
		 * @return true if the character is valid
		 */
		boolean isValid(char character);

		/**
		 * Increments the current value of the structure.
		 *
		 * @return the new value
		 */
		String incrementValue();

		/**
		 * Decrements the current value of the structure.
		 *
		 * @return the new value
		 */
		String decrementValue();

		/**
		 * Checks if the up button is enabled.
		 *
		 * @return true if the button is enabled
		 */
		boolean isUpButtonEnabled();

		/**
		 * Checks if the down button is enabled.
		 *
		 * @return true if the button is enabled
		 */
		boolean isDownButtonEnabled();

		/**
		 * Gets the tooltip for the up button.
		 *
		 * @return the tooltip, may be <code>null</code>
		 */
		String getUpButtonToolTip();

		/**
		 * Gets the tooltip for the down button.
		 *
		 * @return the tooltip, may be <code>null</code>
		 */
		String getDownButtonToolTip();

	}

	/**
	 * Data structure which holds the required data format and values.
	 */
	protected class _LyFixedStructure
		implements _LyStructureItem
	{

		/** The current internal value. */
		protected String _value;

		/**
		 * Creates a fixed item structure. This item cannot be modified.
		 *
		 * @param value the value of the structure item
		 */
		public _LyFixedStructure(String value)
		{
			_value = LyString.strip(value, LyString.MODIFIER);
		}

		public String toString()
		{
			return _value;
		}

		/**
		 * Disables caret.
		 *
		 * @param  pos not used in this implementation
		 * @return <code>false</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#enableCaret(int)
		 */
		public boolean enableCaret(int pos)
		{
			return false;
		}

		/**
		 * Disables the need of zeros.
		 *
		 * @param  c   not used in this implementation
		 * @param  pos not used in this implementation
		 * @return <code>false</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#needZero(char, int)
		 */
		public boolean needZero(char c, int pos)
		{
			return false;
		}

		/**
		 * Disables insertion.
		 *
		 * @param  c   not used in this implementation
		 * @param  pos not used in this implementation
		 * @return <code>false</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#enableInsert(char,
		 *         int)
		 */
		public boolean enableInsert(char c, int pos)
		{
			return false;
		}

		/**
		 * Disables replacement.
		 *
		 * @param  c      not used in this implementation
		 * @param  pos    not used in this implementation
		 * @param  length not used in this implementation
		 * @return <code>false</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#enableReplace(char,
		 *         int, int)
		 */
		public boolean enableReplace(char c, int pos, int length)
		{
			return false;
		}

		/**
		 * Disables selection.
		 *
		 * @param  pos    not used in this implementation
		 * @param  length not used in this implementation
		 * @return <code>false</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#enableSelect(int,
		 *         int)
		 */
		public boolean enableSelect(int pos, int length)
		{
			return false;
		}

		/**
		 * Gets the required length of this fixed field.
		 *
		 * @return the length of the fixed field value.
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#requiredLength()
		 */
		public int requiredLength()
		{
			return _value.length();
		}

		/**
		 * Disables the modification of its value. This method does nothing but specifies it
		 * terminated correctly.
		 *
		 * @param  value not used in this implementation
		 * @return <code>true</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#setValue(java.lang.String)
		 */
		public boolean setValue(String value)
		{
			// NOTHING TO DO
			return true;
		}

		public String toDisplayString()
		{
			return toString();
		}

		public String toFormatString()
		{
			return toString();
		}

		/**
		 * Doesn't have minimum values.
		 *
		 * @return -1
		 */
		public int getMin()
		{
			return -1;
		}

		/**
		 * Doesn't have maximum values.
		 *
		 * @return -1
		 */
		public int getMax()
		{
			return -1;
		}

		/**
		 * Doesn't have pitch values.
		 *
		 * @return 1
		 */
		public int getPitch()
		{
			return 1;
		}

		/**
		 * Disables validation of other values.
		 *
		 * @return <code>false</code>
		 * @param  character not used in this implementation
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#isValid(char)
		 */
		public boolean isValid(char character)
		{
			return false;
		}

		/**
		 * Disables incrementations.
		 *
		 * @return <code>null</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#incrementValue()
		 */
		public String incrementValue()
		{
			return null;
		}

		/**
		 * Disables decrementations.
		 *
		 * @return <code>null</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#decrementValue()
		 */
		public String decrementValue()
		{
			return null;
		}

		/**
		 * Doesn't have enabled buttons.
		 *
		 * @return <code>false</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#isUpButtonEnabled()
		 */
		public boolean isUpButtonEnabled()
		{
			return false;
		}

		/**
		 * Doesn't have enabled buttons.
		 *
		 * @return <code>false</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#isDownButtonEnabled()
		 */
		public boolean isDownButtonEnabled()
		{
			return false;
		}

		/**
		 * Doesn't have button tooltips.
		 *
		 * @return <code>null</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#getUpButtonToolTip()
		 */
		public String getUpButtonToolTip()
		{
			return null;
		}

		/**
		 * Doesn't have button tooltips.
		 *
		 * @return <code>null</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#getDownButtonToolTip()
		 */
		public String getDownButtonToolTip()
		{
			return null;
		}

		/**
		 * Doesn't need caret positions.
		 *
		 * @param  currentPos not used in this implementation
		 * @param  next       not used in this implementation
		 * @return -1
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#getCaretPosition(int,
		 *         boolean)
		 */
		public int getCaretPosition(int currentPos, boolean next)
		{
			return -1; // always next/previous component
		}

		public String toMinString()
		{
			return toString();
		}
	} // end class _LyFixedStructure


	/**
	 * Data structure which holds a numeric input.
	 */
	protected class _LyNumericStructure
		implements _LyStructureItem
	{

		/** Current internal value. */
		protected String _value;

		/** Minimum value. */
		protected int    _min;

		/** Maximum value. */
		protected int    _max;

		/** Length of the field value. */
		protected int    _length;

		/** Increment. */
		protected int    _pitch;

		/**
		 * Creates a new numeric structure item with the specified attributes.
		 *
		 * @param value    the current value
		 * @param minValue the minimum value
		 * @param maxValue the maximum value
		 * @param pitch    the pitch between the incremented and decremented values
		 */
		public _LyNumericStructure(String value, int minValue, int maxValue, int pitch)
		{
			_value = value;
			_min = minValue;
			_max = maxValue;
			_pitch = pitch;
			_length = 1;

			int current = _max;

			while ((current / 10) > 0)
			{
				current = current / 10;
				_length++;
			}
		}

		public boolean enableCaret(int pos)
		{
			return ((pos >= 0) && (pos <= requiredLength()));
		}

		public boolean enableInsert(char c, int pos)
		{
			if ((_value.trim().length() == requiredLength()) || !Character.isDigit(c))
				return false;

			int    realPos = _text.getCaretPosition();
			String current = _text.getText();
			int    start = ((_LyFormattedField)_text).getStructureItemPos(realPos);
			int    emptyIndex = current.substring(start + pos, start + requiredLength()).indexOf(
					LyJFormattedInput.EMPTY_INPUT_FIELD);

			if (emptyIndex == -1)
				return false;

			// check if value is enabled
			StringBuffer buffer = new StringBuffer(requiredLength());

			if (pos < _value.length())
			{
				buffer.append(_value.substring(0, pos)).append(c);
				buffer.append(_value.substring(pos));
			}
			else
				buffer.append(_value).append(c);

			int val = LyTools.intFromString(buffer.toString().trim());

			if (pos < (requiredLength() - 1))
				return true;

			if ((val >= _min) && (val <= _max))
				return true;
			return false;
		} // end method enableInsert

		public boolean needZero(char c, int pos)
		{
			if (pos != 0)
				return false;

			int value = LyTools.intFromString(c + "");

			for (int i = 0; i < (_length - 1); i++)
				value *= 10;

			return (value > _max);
		}

		public boolean enableReplace(char c, int pos, int length)
		{
			// check if digit
			if (((pos < 0) || (pos >= requiredLength())) || !Character.isDigit(c))
				return false;

			// check if value is enabled
			StringBuffer buffer = new StringBuffer(requiredLength());

			if ((pos > 0) && (_value.length() >= pos))
				buffer.append(_value.substring(0, pos));
			buffer.append(c);
			if ((pos + length) < _value.length())
				buffer.append(_value.substring(pos + length));

			int val = LyTools.intFromString(buffer.toString().trim());

			if (((val >= _min) && (val <= _max)) || (pos < (requiredLength() - 1)))
				return true;
			return false;
		}

		/**
		 * Enables the selection at any position in the field.
		 *
		 * @param  pos    not used in this implementation
		 * @param  length not used in this implementation
		 * @return <code>true</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#enableSelect(int,
		 *         int)
		 */
		public boolean enableSelect(int pos, int length)
		{
			return true;
		}

		public int requiredLength()
		{
			return _length;
		}

		public boolean setValue(String value)
		{
			value = value.replace(LyJFormattedInput.EMPTY_INPUT_FIELD, ' ');

			int val = -1;

			try
			{
				String s = value.trim();

				if (s.length() > 0)
					val = Integer.parseInt(s);
				else
					val = 0;
			}
			catch (NumberFormatException e)
			{
				if ((_env != null) && _showTrace
						&& (_env.isDebugLevel(LyEnvironment.DEBUG_WARNING)))
					_env.logWarning(
						"[LyFormattedInput] Warning : Invalid integer value in formatted input : "
						+ value);
				return false;
			}

			if ((value.trim().length() == 0) || ((val >= _min) && (val <= _max))
					|| (value.trim().length() < requiredLength()))
			{
				_value = value;
				return true;
			}
			return false;
		} // end method setValue

		public String toString()
		{
			if (_value.length() == _length)
				return (_value.replace(' ', LyJFormattedInput.EMPTY_INPUT_FIELD));

			StringBuffer buffer = new StringBuffer(_length);
			int			 i;

			for (i = _value.length(); i < _length; i++)
				buffer.append(LyJFormattedInput.EMPTY_INPUT_FIELD);

			return buffer.toString();
		}

		public String toFormatString()
		{
			String format = "";

			for (int i = 0; i < _length; i++)
				format += LyJFormattedInput.EMPTY_INPUT_FIELD;

			return format;
		}

		public String toDisplayString()
		{
			String value = _value.trim();

			// if no value return null
			if (value.length() == 0)
				return null;

			value = value.replace(' ', '0');

			int val = LyTools.intFromString(value);

			// Case value is smaller than final value length
			if (value.length() < _length)
			{
				while (value.length() < _length)
				{
					val = LyTools.intFromString(value);
					// if numeric value is correct, add "0" before for good length
					if ((val >= _min) && (val <= _max))
					{
						// Special case for date inputs
						if (LyJFormattedInput.this instanceof LyJFormattedDateInput)
						{
							if (val < 10)
								value = "200" + val;
							else if (val < 30)
								value = "20" + val;
							else if (val < 100)
								value = "19" + val;
							else
								value = "0" + val;
						}
						else
						{
							while (value.length() < _length)
								value = "0" + value;
						}

						return value;
					}
					// add one "0" after
					value = value + "0";
				}

				// case last round length = _length
				val = LyTools.intFromString(value);
			}

			// if it is smaller than min, good value is min
			if (val < _min)
			{
				String temp = "" + _min;

				while (temp.length() < _length)
					temp = "0" + temp;
				return temp;
			}
			// if it is greater than max, good value is max
			else if (val > _max)
				return "" + _max;

			// compute value is correct.
			return value;
		} // end method toDisplayString

		/**
		 * Gets the minimum value of this field.
		 *
		 * @return the minimum value
		 */
		public int getMin()
		{
			return _min;
		}

		/**
		 * Gets the maximum value of this field.
		 *
		 * @return the maximum value
		 */
		public int getMax()
		{
			return _max;
		}

		/**
		 * Gets the increment value of this field.
		 *
		 * @return the increment value
		 */
		public int getPitch()
		{
			return _pitch;
		}

		/**
		 * Checks if the character is a digit.
		 *
		 * @see leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#isValid(char)
		 */
		public boolean isValid(char character)
		{
			return Character.isDigit(character);
		}

		public String incrementValue()
		{
			int itemValue = ((_value == null) ? -1 : LyTools.intFromString(_value));

			if ((_value == null) || (_value.length() == 0))
				itemValue = _min;
			else
				itemValue = Math.min(itemValue + _pitch, _max);
			_value = "" + itemValue;
			return _value;
		}

		public String decrementValue()
		{
			int itemValue = ((_value == null) ? -1 : LyTools.intFromString(_value));

			if ((_value == null) || (_value.length() == 0))
				itemValue = _max;
			else
				itemValue = Math.max(itemValue - _pitch, _min);
			_value = "" + itemValue;
			return _value;
		}

		public boolean isUpButtonEnabled()
		{

			if ((_value == null) || (_value.length() == 0))
				return true;

			int itemVal = LyTools.intFromString(_value);

			if (itemVal >= _max)
				return false;
			return true;
		}

		public boolean isDownButtonEnabled()
		{
			if ((_value == null) || (_value.length() == 0))
				return false;

			int itemVal = LyTools.intFromString(_value);

			if (itemVal <= _min)
				return false;
			return true;
		}

		public String getUpButtonToolTip()
		{
			if (_max >= 0)
				return Integer.toString(_max);
			return null;

		}

		public String getDownButtonToolTip()
		{
			if (_min >= 0)
				return Integer.toString(_min);
			return null;

		}

		public int getCaretPosition(int currentPos, boolean next)
		{
			if (next)
			{
				if (currentPos >= _length)
				{
					return -1; // next component
				}
				if ((currentPos + 1) >= _length)
				{
					return -1;
				}

				return currentPos + 1;
			}
			if (currentPos == 0)
				return -1; // previous comp

			return currentPos - 1;
		}

		public String toMinString()
		{
			StringBuffer buffer = new StringBuffer(_length);
			String		 value = _value.trim();
			int			 i;

			if (value.length() == 0)
			{
				String computedValue = LyJFormattedInput.this.computeMinValue(this);

				if (computedValue == null)
					// Use min value
					value = "" + _min;
				else
					value = computedValue;

				for (i = value.length(); i < _length; i++)
					buffer.append('0');

				buffer.append(value);
				return buffer.toString();
			}

			return toDisplayString();
		}
	} // end class _LyNumericStructure


	/**
	 * Data structure which holds an alphabetic input.
	 */
	protected class _LyAlphabeticStructure
		implements _LyStructureItem
	{

		/** Current internal value. */
		protected String _value;

		/** The internal character. */
		protected char   _internal;

		/** Minimum value. */
		protected char   _minValue;

		/** Maximum value. */
		protected char   _maxValue;

		/** Indicates if this item is lower case or upper case. */
		boolean			 _isLowerCase;

		/** Indicates if this item allows both lower and upper case or not. */
		boolean			 _allowBothCase;

		/**
		 * Creates a new alphabetic structure with the specified attributes.
		 *
		 * @param value    the current value
		 * @param minValue the minimum value
		 * @param maxValue the maximum value
		 */
		public _LyAlphabeticStructure(char value, char minValue, char maxValue)
		{
			// _value = value;
			_internal = value;
			_minValue = minValue;
			_maxValue = maxValue;

			// Do some check..
			boolean isMinLowerCase = Character.isLowerCase(minValue);
			boolean isMaxLowerCase = Character.isLowerCase(maxValue);

			if (isMinLowerCase && isMaxLowerCase) // only lowercase
			{
				_isLowerCase = true;
				_allowBothCase = false;
			}
			else if (!isMinLowerCase && !isMaxLowerCase) // only uppercase
			{
				_isLowerCase = false;
				_allowBothCase = false;
			}
			else
				_allowBothCase = true;

			if (_allowBothCase)
			{
				if (isMinLowerCase)
					throw new IllegalStateException("BF AppEngine : Invalid format");

				if (!isMaxLowerCase)
					throw new IllegalStateException("BF AppEngine : Invalid format");
			}
			else
			{
				if (minValue > maxValue)
					throw new IllegalStateException("BF AppEngine : Invalid format");
			}
		}

		public String toString()
		{
			if (_internal > 0)
				return (new Character(_internal)).toString();
			return Character.toString(LyJFormattedInput.EMPTY_INPUT_FIELD);
		}

		public String toDisplayString()
		{
			if (_internal > 0)
				return new Character(_internal).toString();
			return null;
		}

		public String toFormatString()
		{
			return Character.toString(LyJFormattedInput.EMPTY_INPUT_FIELD);
		}

		/**
		 * Enables the caret at every position.
		 *
		 * @param  pos not used in this implementation
		 * @return <code>true</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#enableCaret(int)
		 */
		public boolean enableCaret(int pos)
		{
			return true;
		}

		/**
		 * Enables insertions at every position.
		 *
		 * @param  c   not used in this implementation
		 * @param  pos not used in this implementation
		 * @return <code>true</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#enableInsert(char,
		 *         int)
		 */
		public boolean enableInsert(char c, int pos)
		{
			return true;
		}

		/**
		 * Disables the need of zeros.
		 *
		 * @param  c   not used in this implementation
		 * @param  pos not used in this implementation
		 * @return <code>false</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#needZero(char, int)
		 */
		public boolean needZero(char c, int pos)
		{
			return false;
		}

		/**
		 * Enables replacement.
		 *
		 * @param  c      not used in this implementation
		 * @param  pos    not used in this implementation
		 * @param  length not used in this implementation
		 * @return <code>true</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#enableReplace(char,
		 *         int, int)
		 */
		public boolean enableReplace(char c, int pos, int length)
		{
			return true;
		}

		/**
		 * Enables selection.
		 *
		 * @param  pos    not used in this implementation
		 * @param  length not used in this implementation
		 * @return <code>true</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#enableSelect(int,
		 *         int)
		 */
		public boolean enableSelect(int pos, int length)
		{
			return true;
		}

		public boolean setValue(String value)
		{
			value = value.replace(LyJFormattedInput.EMPTY_INPUT_FIELD, ' ');
			value = value.trim();
			if (value.length() > 1)
				return false;

			if (value.length() == 0)
			{
				_internal = (char)0;
				return true;
			}

			char character = value.charAt(0);

			if ((character >= _minValue) && (character <= _maxValue))
			{
				_internal = character;
				_value = (new Character(_internal)).toString();
				return true;
			}

			// try inverting case
			if (Character.isLowerCase(character))
				character = Character.toUpperCase(character);
			else
				character = Character.toLowerCase(character);

			if ((character >= _minValue) && (character <= _maxValue))
			{
				_internal = character;
				_value = (new Character(_internal)).toString();
				return true;
			}

			return false;
		} // end method setValue

		/**
		 * Gets the length of an alphabetic structure item, which is always 1 character.
		 *
		 * @return 1
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#requiredLength()
		 */
		public int requiredLength()
		{
			return 1;
		}

		public boolean isValid(char character)
		{
			char value = character;

			if (!Character.isLetter(value))
				return false;

			if ((character >= _minValue) && (character <= _maxValue))
				return true;

			// Try to invert the case of the character, to check i it is valid...
			if (Character.isLowerCase(character))
				value = Character.toUpperCase(character);
			else
				value = Character.toLowerCase(character);

			if ((value >= _minValue) && (value <= _maxValue))
				return true;

			return false;
		}

		public String incrementValue()
		{
			if ((_value == null) || (_value.length() == 0))
			{
				_value = (new Character(_minValue)).toString();
				_internal = _minValue;
				return _value;
			}

			char value = _value.charAt(0);

			if (value >= _maxValue)
				return _value;

			value++;

			if (!Character.isLetter(value))
			{
				do
					value++;
				while (!Character.isLetter(value));
			}

			_internal = value;
			_value = (new Character(value)).toString();

			return _value;
		} // end method incrementValue

		public String decrementValue()
		{
			if ((_value == null) || (_value.length() == 0))
			{
				_value = (new Character(_maxValue)).toString();
				_internal = _maxValue;
				return _value;
			}

			char value = _value.charAt(0);

			if (value <= _minValue)
				return _value;

			value--;

			if (!Character.isLetter(value))
			{
				do
					value--;
				while (!Character.isLetter(value));
			}

			_internal = value;
			_value = (new Character(value)).toString();

			return _value;
		} // end method decrementValue

		public boolean isUpButtonEnabled()
		{
			if ((_value == null) || (_value.length() == 0))
				return true;

			char value = _value.charAt(0);

			if (value < _maxValue)
				return true;

			return false;
		}

		public boolean isDownButtonEnabled()
		{
			if ((_value == null) || (_value.length() == 0))
				return false;

			char value = _value.charAt(0);

			if (value > _minValue)
				return true;

			return false;
		}

		public String getUpButtonToolTip()
		{
			return (new Character(_maxValue)).toString();
		}

		public String getDownButtonToolTip()
		{
			return new Character(_minValue).toString();
		}

		public int getCaretPosition(int currentPos, boolean next)
		{
			if (next && (currentPos == 0))
				return -1; // next component

			if (next && (currentPos == 1)) // should not happen
				return -1;

			if (!next && (currentPos == 0)) // leaving
				return -1;

			return 0;
		}

		public String toMinString()
		{
			if (_internal > 0)
				return toDisplayString();
			return Character.toString(_minValue);
		}
	} // end class _LyAlphabeticStructure


	/**
	 * Data structure which holds a choice input.
	 */
	protected class _LyChoiceStructure
		implements _LyStructureItem
	{
		/** All possible values. */
		String[] _choices;

		/** The current value as a string. */
		String   _value;

		/** The maximum length of the field. */
		int		 _length;

		/** The current value as an index in the vector. */
		int		 _currentIndex;

		/** Stores all possible values as LOWER CASE (optimizes searchs). */
		String[] _choices_lower_case;

		/**
		 * Creates a choice structure from the specified list values.
		 *
		 * @param values a list of values separated by '|'
		 */
		public _LyChoiceStructure(String values)
		{
			if ((values == null) || (values.trim().length() == 0))
				throw new IllegalStateException("BF AppEngine : a choice format must not be null");
			values = values.trim();
			if (values.endsWith("|"))
			{
				if ((_env != null) && _showTrace
						&& (_env.isDebugLevel(LyEnvironment.DEBUG_WARNING)))
					_env.logWarning("[LyFormattedInput] The given format '"
						+ values
						+ "' has at least one empty choice value. Those values will be ignored...");
				values = values.substring(0, values.length() - 1);
			}

			if (values.indexOf(CHOICE_EMPTY_INPUT_FIELD) >= 0)
			{
				if ((_env != null) && _showTrace
						&& (_env.isDebugLevel(LyEnvironment.DEBUG_WARNING)))
					_env.logWarning("[LyFormattedInput] The given format '" + values
						+ "' contains the invalid character '" + CHOICE_EMPTY_INPUT_FIELD
						+ "'. This character will be removed.");
				values = LyString.replace(values, CHOICE_EMPTY_INPUT_FIELD, "");
			}

			String[] choices = LyString.stringToList(values, '|');

			// Compute max length
			_currentIndex = -1;
			_value = null;
			_length = 0;

			int i, current, nMax = 0;

			for (i = 0; i < choices.length; i++)
			{
				if (choices[i].trim().length() != 0)
					nMax++;
				_length = Math.max(_length, choices[i].trim().length());
			}

			_choices = new String[nMax];
			current = 0;
			for (i = 0; i < choices.length; i++)
			{
				if (choices[i].trim().length() != 0)
				{
					_choices[current] = choices[i].trim();
					current++;
				}
			}

			if (choices.length != nMax)
			{
				if ((_env != null) && _showTrace
						&& (_env.isDebugLevel(LyEnvironment.DEBUG_WARNING)))
					_env.logWarning("[LyFormattedInput] The given format '" + values
						+ "' has at least one "
						+ "empty choice value. Those values will be ignored...");
			}
		}

		public String toString()
		{
			if ((_value == null) || (_value.length() == 0))
			{
				String result = "";

				for (int i = 0; i < _length; i++)
					result += CHOICE_EMPTY_INPUT_FIELD;

				return result;
			}
			return _value;
		}

		public String toDisplayString()
		{
			if (_currentIndex >= 0)
				return _value;
			else if (_value == null)
				return null;
			else
			{
				String startsWith = null;
				String value = _value.trim();

				value = LyString.replace(value, CHOICE_EMPTY_INPUT_FIELD, "");
				for (int i = 0; i < _choices.length; i++)
				{
					if (_choices[i].equals(value))
					{
						_currentIndex = i;
						_value = _choices[i];
						return _value;
					}
					else if ((startsWith == null) && (_choices[i].startsWith(value)))
						startsWith = _choices[i];
				}
				if (startsWith != null)
					return startsWith;
			}
			return _value;
		} // end method toDisplayString

		/**
		 * Enables caret.
		 *
		 * @param  pos not used in this implementation
		 * @return <code>true</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#enableCaret(int)
		 */
		public boolean enableCaret(int pos)
		{
			return true;
		}

		/**
		 * Disables insertion.
		 *
		 * @param  c   not used in this implementation
		 * @param  pos not used in this implementation
		 * @return <code>false</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#enableInsert(char,
		 *         int)
		 */
		public boolean enableInsert(char c, int pos)
		{
			return false;
		}

		/**
		 * Disables the need of zeros.
		 *
		 * @param  c   not used in this implementation
		 * @param  pos not used in this implementation
		 * @return <code>false</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#needZero(char, int)
		 */
		public boolean needZero(char c, int pos)
		{
			return false;
		}

		/**
		 * Enables replacement.
		 *
		 * @param  c      not used in this implementation
		 * @param  pos    not used in this implementation
		 * @param  length not used in this implementation
		 * @return <code>true</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#enableReplace(char,
		 *         int, int)
		 */
		public boolean enableReplace(char c, int pos, int length)
		{
			return true;
		}

		/**
		 * Enables selection.
		 *
		 * @param  pos    not used in this implementation
		 * @param  length not used in this implementation
		 * @return <code>true</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#enableSelect(int,
		 *         int)
		 */
		public boolean enableSelect(int pos, int length)
		{
			return true;
		}

		public boolean setValue(String value)
		{
			boolean result = setValueInternal(value, true);

			if ((!result) && (value != null))
			{
				String trimmedValue = value.trim();

				trimmedValue = LyString.replace(trimmedValue, CHOICE_EMPTY_INPUT_FIELD, "");

				if (!value.equals(trimmedValue))
					result = setValueInternal(trimmedValue, true);

				// Retry ignore case
				if (!result)
					result = setValueInternal(value, false);
				if ((!result) && (!value.equals(trimmedValue)))
					result = setValueInternal(trimmedValue, false);
			}

			return result;
		}

		/**
		 * Called by setValue with "real" value then the trimmed value.
		 *
		 * @param  value         the string value for this choice
		 * @param  caseSensitive Indicates if we are case sensitive or not.
		 * @return true if the value is set correctly
		 */
		protected boolean setValueInternal(String value, boolean caseSensitive)
		{
			// check if the given value is valid
			if ((value == null) || (value.length() == 0))
			{
				_value = null;
				_currentIndex = -1;
				return true;
			}

			String[] choices_compare;

			if (caseSensitive)
				choices_compare = _choices;
			else
			{
				// No case sensitive : compare with _choice_lower_case and build list if first time
				value = value.toLowerCase();

				if (_choices_lower_case == null)
				{
					_choices_lower_case = new String[_choices.length];

					for (int i = 0; i < _choices.length; i++)
						_choices_lower_case[i] = _choices[i].toLowerCase();
				}
				choices_compare = _choices_lower_case;
			}

			boolean found = false;

			// Check if the value is complete or is the beginning of a choice
			for (int i = 0; (i < _choices.length) && !found; i++)
			{
				if (choices_compare[i].equals(value))
				{
					_value = _choices[i];
					_currentIndex = i;
					found = true;
				}
				else if (choices_compare[i].startsWith(value))
				{
					_value = value;
					_currentIndex = -1;
				}
			}

			if (found && (_currentIndex != -1)) // only one value found
				return true;

			if (_currentIndex == -1)
			{
				// try to complete with max number of character
				// First build a vector of all possible strings
				Vector<String> validChoice = new Vector<String>();

				for (int i = 0; i < _choices.length; i++)
				{
					if (choices_compare[i].startsWith(value))
						validChoice.addElement(_choices[i]);
				}

				if (validChoice.size() == 1)
					return setValue(validChoice.elementAt(0));
				else if (validChoice.size() > 1)
				{
					// Try to find the common part of all strings
					String commonPart = LyString.getCommonPart(validChoice);

					if (!value.equals(commonPart))
						return setValueInternal(commonPart, true);
				}
				else
					return false;
			}

			if ((_currentIndex == -1) && (_value != null))
			{
				// complete string to max length
				for (int j = _value.length(); j < _length; j++)
					_value += CHOICE_EMPTY_INPUT_FIELD;

				return true;
			}

			_value = null;
			_currentIndex = -1;
			return false;
		} // end method setValueInternal

		public int requiredLength()
		{
			return _length;
		}

		public String toFormatString()
		{
			String format = "";

			for (int i = 0; i < _length; i++)
				format += CHOICE_EMPTY_INPUT_FIELD;

			return format;
		}

		public boolean isValid(char character)
		{
			String value = ((_value == null) ? "" : _value);

			value = LyString.replace(value, CHOICE_EMPTY_INPUT_FIELD, "");
			value = value + character;

			for (int i = 0; i < _choices.length; i++)
			{
				if (_choices[i].startsWith(value))
					return true;
			}

			return false;
		}

		public String incrementValue()
		{
			if (_currentIndex < (_choices.length - 1))
				_currentIndex++;

			_value = _choices[_currentIndex];
			return _value;
		}

		public String decrementValue()
		{
			if (_currentIndex > 0)
				_currentIndex--;
			else
				_currentIndex = 0;

			_value = _choices[_currentIndex];
			return _value;
		}

		public boolean isUpButtonEnabled()
		{
			if (_currentIndex < (_choices.length - 1))
				return true;

			return false;
		}

		public boolean isDownButtonEnabled()
		{
			if (_currentIndex > 0)
				return true;
			return false;
		}

		/**
		 * Doesn't have button tooltips.
		 *
		 * @return <code>null</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#getUpButtonToolTip()
		 */
		public String getUpButtonToolTip()
		{
			return null;
		}

		/**
		 * Doesn't have button tooltips.
		 *
		 * @return <code>null</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#getDownButtonToolTip()
		 */
		public String getDownButtonToolTip()
		{
			return null;
		}

		public int getCaretPosition(int currentPos, boolean next)
		{
			String value = _value;

			if (next && (currentPos < 0))
				return 0;

			if ((_value == null) || (_value.length() == 0))
				value = "";

			value = LyString.replace(value, CHOICE_EMPTY_INPUT_FIELD, "");

			if (next)
			{
				if (_currentIndex >= 0)
					return -1;
				if (currentPos >= value.length())
					return -1; // Next comp

				return value.length();
			}
			if (currentPos == 0)
				return -1;

			return 0;
		}

		public String toMinString()
		{
			if ((_value == null) || (_value.length() == 0))
				return _choices[0];
			return toDisplayString();

		}
	} // end class _LyChoiceStructure


	/**
	 * This structure implements a list of possible values for a character. The corresponding format
	 * is [a-z;A-Z;...]
	 */
	protected class _LyAlphabeticIntervalsStructure
		implements _LyStructureItem
	{

		/** All the availables characters. */
		Vector<Character> _availableChar;

		/** The current value as a string. */
		String _value;

		/** The current value as an index in the vector. */
		int    _position;

		/**
		 * Creates an alphabetical interval structure.
		 *
		 * @param intervals all intervals separeted by ';'
		 */
		public _LyAlphabeticIntervalsStructure(String intervals)
		{
			_availableChar = new Vector<Character>();

			String[] values = LyString.stringToList(intervals, ';');
			String   temp;
			int		 n;
			String   min, max;
			char     minValue, maxValue, current;

			for (int i = 0; i < values.length; i++)
			{
				temp = values[i];
				n = temp.indexOf("-");

				if (n < 0)
					throw new IllegalStateException("BF AppEngine : Invalid Format " + intervals);

				min = temp.substring(0, n);
				max = temp.substring(n + 1);

				if ((min == null) || (min.length() != 1) || (max == null) || (max.length() != 1))
					throw new IllegalStateException("BF AppEngine : Invalid Format " + intervals);

				minValue = min.charAt(0);
				maxValue = max.charAt(0);

				if (minValue > maxValue)
					throw new IllegalStateException("BF AppEngine : Invalid Format " + intervals);

				current = minValue;
				do
				{
					_availableChar.addElement(new Character(current));
					current++;
				}
				while (current <= maxValue);
			}

			_value = "";
			_position = -1;
		}

		public String toString()
		{
			if (_position >= 0)
				return _availableChar.elementAt(_position).toString();
			return "" + LyJFormattedInput.EMPTY_INPUT_FIELD;
		}

		public String toDisplayString()
		{
			if (_position >= 0)
				return _availableChar.elementAt(_position).toString();
			return null;
		}

		/**
		 * Enables caret.
		 *
		 * @param  pos not used in this implementation
		 * @return <code>true</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#enableCaret(int)
		 */
		public boolean enableCaret(int pos)
		{
			return true;
		}

		/**
		 * Disables the need of zeros.
		 *
		 * @param  c   not used in this implementation
		 * @param  pos not used in this implementation
		 * @return <code>false</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#needZero(char, int)
		 */
		public boolean needZero(char c, int pos)
		{
			return false;
		}

		/**
		 * Enables insertion.
		 *
		 * @param  c   not used in this implementation
		 * @param  pos not used in this implementation
		 * @return <code>true</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#enableInsert(char,
		 *         int)
		 */
		public boolean enableInsert(char c, int pos)
		{
			return true;
		}

		/**
		 * Enables replacement.
		 *
		 * @param  c      not used in this implementation
		 * @param  pos    not used in this implementation
		 * @param  length not used in this implementation
		 * @return <code>true</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#enableReplace(char,
		 *         int, int)
		 */
		public boolean enableReplace(char c, int pos, int length)
		{
			return true;
		}

		/**
		 * Enables selection.
		 *
		 * @param  pos    not used in this implementation
		 * @param  length not used in this implementation
		 * @return <code>true</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#enableSelect(int,
		 *         int)
		 */
		public boolean enableSelect(int pos, int length)
		{
			return true;
		}

		public boolean setValue(String value)
		{
			if (value == null)
			{
				_position = -1;
				return true;
			}

			value = value.replace(LyJFormattedInput.EMPTY_INPUT_FIELD, ' ');
			value = value.trim();
			if (value.length() > 1)
				return false;

			if (value.length() == 0)
			{
				_position = -1;
				_value = null;
				return true;
			}

			char	  character = value.charAt(0);
			Character valueChar = new Character(character);

			if (_availableChar.contains(valueChar))
			{
				_position = _availableChar.indexOf(valueChar);
				_value = valueChar.toString();
				return true;
			}

			if (Character.isLowerCase(character))
				valueChar = new Character(Character.toUpperCase(character));
			else
				valueChar = new Character(Character.toLowerCase(character));

			if (_availableChar.contains(valueChar))
			{
				_position = _availableChar.indexOf(valueChar);
				_value = valueChar.toString();
				return true;
			}

			_position = -1;
			_value = null;
			return false;
		} // end method setValue

		/**
		 * The length of an alphabetical interval.
		 *
		 * @return 1
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#requiredLength()
		 */
		public int requiredLength()
		{
			return 1;
		}

		public int getCaretPosition(int currentPos, boolean next)
		{
			if (next && (currentPos == 0))
			{
				return -1;
			} // next component

			if (next && (currentPos == 1)) // should not happen
			{
				return -1;
			}

			if (!next && (currentPos == 0)) // leaving
			{
				return -1;
			}

			return 0;
		}

		public boolean isValid(char character)
		{
			if (!Character.isLetter(character) && !Character.isDigit(character))
				return false;

			Character valueChar = new Character(character);

			if (_availableChar.contains(valueChar))
				return true;

			if (Character.isLowerCase(character))
				valueChar = new Character(Character.toUpperCase(character));
			else
				valueChar = new Character(Character.toLowerCase(character));

			if (_availableChar.contains(valueChar))
				return true;

			return false;
		}

		public String incrementValue()
		{
			if (_position < 0)
				_position = 0;
			else if (_position < (_availableChar.size() - 1))
				_position++;

			_value = _availableChar.elementAt(_position).toString();

			return _value;

		}

		public String decrementValue()
		{
			if (_position < 0)
				_position = 0;
			else if (_position > 0)
				_position--;

			_value = _availableChar.elementAt(_position).toString();

			return _value;
		}

		public boolean isUpButtonEnabled()
		{
			if (_position < (_availableChar.size() - 1))
				return true;

			return false;
		}

		public boolean isDownButtonEnabled()
		{
			if (_position > 0)
				return true;
			return false;
		}

		/**
		 * Doesn't have button tooltips.
		 *
		 * @return <code>null</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#getUpButtonToolTip()
		 */
		public String getUpButtonToolTip()
		{
			return null;
		}

		/**
		 * Doesn't have button tooltips.
		 *
		 * @return <code>null</code>
		 * @see    leon.view.win.swing.component.LyJFormattedInput._LyStructureItem#getDownButtonToolTip()
		 */
		public String getDownButtonToolTip()
		{
			return null;
		}

		public String toMinString()
		{
			if (_position < 0)

				return _availableChar.elementAt(0).toString();
			return toDisplayString();
		}

		public String toFormatString()
		{
			return Character.toString(LyJFormattedInput.EMPTY_INPUT_FIELD);
		}
	} // end class _LyAlphabeticIntervalsStructure
} // end class LyJFormattedInput
