/*
 * LyHRefComponent.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.border.*;
import leon.view.filters.*;
import leon.view.event.*;
import leon.view.setinterface.*;


/**
 * This class implements a simple label with an image.
 *
 * @version $Id: LyJHRefComponent.java 29564 2011-02-10 15:43:35Z nhamon $
 * @author  Lyria
 */
public class LyJHRefComponent extends JLabel
	implements MouseMotionListener,
		MouseListener,
		LyEventTransmitter,
		LySetLabelInterface,
		FocusListener,
		KeyListener
{
	private static final long     serialVersionUID = -76086121350558759L;

	public static final String    __VERSION =
		"$Id: LyJHRefComponent.java 29564 2011-02-10 15:43:35Z nhamon $";

	/** Image is grayed and may be highlighted when mouse is over. */
	public static boolean		  USE_GRAYED_IMAGE = true;

	/** Separates highlighting of image and label. */
	public static boolean		  SEPARATE_HIGHLIGHT = false;

	/** Mouse cursor when over HRef links. */
	public static Cursor		  LY_OVER_CURSOR = new Cursor(Cursor.HAND_CURSOR);

	/** Minimum height, in pixels. */
	public int					  MIN_HEIGHT = 22;

	/** The image. */
	protected Image				  _img;

	/** The gray image. */
	protected Image				  _grayImg;

	/** Default Color. */
	protected Color				  _colorDefault;

	/** Disabled Color. */
	protected Color				  _colorDisable;

	/** Font highlight. */
	protected Color				  _colorHighlight;

	/** Font visited. */
	protected Color				  _colorVisited;

	/** Current MouseStatus. */
	protected boolean			  _highlight_image;

	/** Current MouseStatus. */
	protected boolean			  _highlight_label;

	/** Sensitive image area. */
	protected Rectangle			  _bounds_image;

	/** Sensitive label area. */
	protected Rectangle			  _bounds_label;

	/** Visited reference. */
	protected boolean			  _visited;

	/** Indicates if href is sensitive or not. */
	protected boolean			  _sensitive;

	/** Indicates if href is idle or not. */
	protected boolean			  _idle;

	/** Indicates if href has the focus or not. */
	protected boolean			  _hasFocus;

	/** User data. */
	protected Object			  _userData;

	/** List of listeners for action events. */
	protected Vector<?>			  _actionListeners;

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter _actionTransmitter;

	/** The font metrics of the associated font. */
	protected FontMetrics		  _fm;

	/** The last image on this href component. */
	Image						  _lastImage;

	private Icon				  _grayIcon;

	private Icon				  _icon;

	/** Minimum width of the component. */
	protected int				  _minimumWidth;

	/**
	 * Sole constructor. Creates a default working href component.
	 */
	public LyJHRefComponent()
	{
		super();
		_visited = false;
		_sensitive = true;
		_minimumWidth = -1;
	}

	/**
	 * Creates a href component with the specified attributes.
	 *
	 * @param label     the text label
	 * @param image     the icon image
	 * @param font      the font of this href component
	 * @param visited   Indicates if this href component is visited.
	 * @param sensitive Indicates if this href component is sensitive.
	 * @param userData  the user datas
	 */
	public LyJHRefComponent(String label, Image image, Font font, boolean visited,
		boolean sensitive, Object userData)
	{
		this();
		setFont(font);
		setText(label);
		setImage(image);
		setIcon(_icon);

		_visited = visited;
		_sensitive = sensitive;
		_userData = userData;

		setToolTipText(label);

		init();
	}

	/**
	 * Returns user data.
	 *
	 * @return the user data
	 * @see    #setUserData(Object)
	 */
	public Object getUserData()
	{
		return _userData;
	}

	/**
	 * This component can always gain focus.
	 *
	 * @return <code>true</code>
	 * @see    java.awt.Component#isFocusTraversable()
	 */
	public boolean isFocusTraversable()
	{
		return true;
	}

	/**
	 * Initializes this component and registers its event listeners.
	 */
	protected void init()
	{
		// Register to the ToolTipManager
		if (_img != null)
		{
			ToolTipManager ttm = ToolTipManager.sharedInstance();

			ttm.registerComponent(this);
		}

		// changed by mouseEvents
		_highlight_image = false;
		_highlight_label = false;

		addMouseListener(this);
		addMouseMotionListener(this);
		addFocusListener(this);
		addKeyListener(this);
	}

	/**
	 * Highlights the text.
	 *
	 * @see javax.swing.JComponent#paintComponent(java.awt.Graphics)
	 */
	public void paintComponent(Graphics g)
	{
		if (_idle)
			return;

		super.paintComponent(g);

		Dimension d = getSize();
		int		  hLabel = _fm.getHeight();
		int		  wLabel = (getText() != null) ? (_fm.stringWidth(getText())) : (0);
		int		  hImg = (_img != null) ? (_img.getHeight(this)) : (0);
		int		  wImg = (_img != null) ? (_img.getWidth(this)) : (0);
		Border    border = getBorder();
		int		  x = (wImg == 0) ? (0) : (getIconTextGap() + wImg);
		int		  borderWidth = 0;

		if (border instanceof EmptyBorder)
			borderWidth += ((EmptyBorder)border).getBorderInsets().left;

		x += borderWidth;
		_bounds_label = new Rectangle(x, (d.height - hLabel) / 2, wLabel, hLabel);
		_bounds_image = new Rectangle(borderWidth, (d.height - hImg) / 2, wImg, hImg);

		if (_highlight_label)
			g.drawLine(x, ((d.height + hLabel) / 2) - 1, x + wLabel, ((d.height + hLabel) / 2) - 1);
	}

	/* Implementation of MouseMotionListener */

	/**
	 * Does nothing.
	 *
	 * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
	 */
	public void mouseDragged(MouseEvent e)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.MouseMotionListener#mouseMoved(java.awt.event.MouseEvent)
	 */
	public void mouseMoved(MouseEvent e)
	{
		if (!_sensitive)
			return;

		int x = e.getX();
		int y = e.getY();

		if (SEPARATE_HIGHLIGHT)
		{
			if (isInLabel(x, y))
				highlightLabel(true);
			else
				highlightLabel(false);
			if (isInImage(x, y))
				highlightImage(true);
			else
				highlightImage(false);
		}
		else
		{
			if (isInLabel(x, y) || isInImage(x, y))
			{
				highlightLabel(true);
				highlightImage(true);
			}
			else
			{
				highlightLabel(false);
				highlightImage(false);
			}
		}
	} // end method mouseMoved

	/* Implementation of MouseListener */

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
	 */
	public void mousePressed(MouseEvent e)
	{
		if ((e.getModifiers() & InputEvent.BUTTON3_MASK) > 0)
		{
			// Popup menu on the href component
			if (_actionTransmitter != null)
			{
				LyActionEvent actionEvent = new LyActionEvent(this, null, null,
						LyMenuDemand.TYPE_POPUP);
				actionEvent.setParameter(new int[] {e.getX(), e.getY()});

				_actionTransmitter.processActionEvent(actionEvent);
			}
		}
		if ((e.getModifiers() & InputEvent.BUTTON1_MASK) > 0)
			activateLink();
	}

	/**
	 * Does nothing.
	 *
	 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
	 */
	public void mouseClicked(MouseEvent e)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
	 */
	public void mouseExited(MouseEvent e)
	{
		highlightLabel(false);
		highlightImage(false);
	}

	/**
	 * Does nothing.
	 *
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	public void mouseReleased(MouseEvent e)
	{
	}

	/**
	 * Does nothing.
	 *
	 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
	 */
	public void mouseEntered(MouseEvent e)
	{
	}

	/**
	 * Does this action when link has been activate.
	 */
	public void activateLink()
	{
		// If not sensitive, do nothing (no link)
		if ( /* (_highlight_label || _highlight_image) && */(_sensitive))
		{
			// execute actions
			if (_actionTransmitter != null)
			{
				LyActionEvent actionEvent = new LyActionEvent(this, null, null,
						LyActionDemand.TYPE_HREF);

				_actionTransmitter.processActionEvent(actionEvent);
			}
		}
	}

	/**
	 * Highligts the text label depending on the status and the mouse position, and sets the
	 * corresponding cursor.
	 *
	 * @param status the new status
	 */
	public void highlightLabel(boolean status)
	{
		// Changed ?
		if (_highlight_label != status)
		{
			_highlight_label = status;

			if ((status) && (_sensitive))
			{
				setCursor(LY_OVER_CURSOR);
				setForeground(_colorHighlight);
			}
			else
			{
				setCursor(Cursor.getDefaultCursor());

				if (_sensitive)
					setForeground((_visited) ? (_colorVisited) : (_colorDefault));
				else
					setForeground(_colorDisable);
			}

			repaint();
		}
	}

	/**
	 * Uses the highlighted or grayed images on the component depending on the status.
	 *
	 * @param status the new status
	 */
	public void highlightImage(boolean status)
	{
		// Changed ?
		if (_highlight_image != status)
		{
			_highlight_image = status;

			if ((status) && (_sensitive))
			{
				setCursor(LY_OVER_CURSOR);
				setIcon(_icon);
			}
			else
			{
				setCursor(Cursor.getDefaultCursor());
				setIcon(_grayIcon);
			}

			repaint();
		}
	}

	/**
	 * Checks if the mouse is over the text.
	 *
	 * @param  x the x-coordinate of the mouse position
	 * @param  y the y-coordinate of the mouse position
	 * @return <code>true</code> if the specified position is in the text label area
	 */
	protected boolean isInLabel(int x, int y)
	{
		if (_bounds_label == null)
			return false;
		return ((x >= _bounds_label.x) && (x <= (_bounds_label.x + _bounds_label.width))
				&& (y >= _bounds_label.y) && (y <= (_bounds_label.y + _bounds_label.height)));
	}

	/**
	 * Checks if the mouse is over the image.
	 *
	 * @param  x the x-coordinate of the mouse position
	 * @param  y the y-coordinate of the mouse position
	 * @return <code>true</code> if the specified position is in the image area
	 */
	protected boolean isInImage(int x, int y)
	{
		if (_bounds_image == null)
			return false;
		return ((x >= _bounds_image.x) && (x <= (_bounds_image.x + _bounds_image.width))
				&& (y >= _bounds_image.y) && (y <= (_bounds_image.y + _bounds_image.height)));
	}

	/**
	 * Refreshes the component when focus is gained.
	 *
	 * @see java.awt.event.FocusListener#focusGained(java.awt.event.FocusEvent)
	 */
	public void focusGained(FocusEvent e)
	{
		repaint();
	}

	/**
	 * Refreshes the component when focus is lost.
	 *
	 * @see java.awt.event.FocusListener#focusLost(java.awt.event.FocusEvent)
	 */
	public void focusLost(FocusEvent e)
	{
		repaint();
	}

	/**
	 * Does nothing when a key is pressed.
	 *
	 * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
	 */
	public void keyPressed(KeyEvent e)
	{
	}

	/**
	 * Does nothing when a key is released.
	 *
	 * @see java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
	 */
	public void keyReleased(KeyEvent e)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
	 */
	public void keyTyped(KeyEvent e)
	{
		if (e.getKeyChar() == ' ')
			activateLink();
		else
		{
			if (_img != null)
				_lastImage = _img;

			if (_img == null)
				setImage(_lastImage);
			else
				setImage(null);

		}

	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return (_actionTransmitter);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		return (null);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
	 */
	public void setLabel(String label)
	{
		setText(label);
		checkResize();
	}

	/**
	 * Sets the user data.
	 *
	 * @param userData the new user data
	 * @see   #getUserData()
	 */
	public void setUserData(Object userData)
	{
		_userData = userData;
	}

	/**
	 * Sets the borders depending on the component's idle mode, then refreshes it.
	 */
	protected void checkResize()
	{
		if (_idle)
			setBorder(BorderFactory.createEmptyBorder(0, 0, 0, 0));
		else
			setBorder(BorderFactory.createEmptyBorder(0, 3, 0, 3));

		repaint();
	}

	/**
	 * Set the minimum width of this component.
	 * 
	 * @param minimumWidth the new minimum width of the component (a negative value means that
	 * there is no minimum width).
	 */
	public void setMinimumWidth(int minimumWidth)
	{
		_minimumWidth = minimumWidth;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#getPreferredSize()
	 */
	public Dimension getPreferredSize()
	{
		Dimension dim = super.getPreferredSize();

		dim.width = Math.max(dim.width, _minimumWidth);
		dim.height = Math.max(dim.height, MIN_HEIGHT);

		return dim;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#addNotify()
	 */
	public void addNotify()
	{
		highlightLabel(true);
		highlightImage(true);
		highlightLabel(false);
		highlightImage(false);
		super.addNotify();
	}

	/**
	 * Sets the disabled color.
	 *
	 * @param color the new disabled color
	 */
	public void setDisableColor(Color color)
	{
		_colorDisable = color;
	}

	/**
	 * Sets the default color.
	 *
	 * @param color the new default color
	 */
	public void setDefaultColor(Color color)
	{
		_colorDefault = color;
	}

	/**
	 * Sets the highlight color.
	 *
	 * @param color the new highlight color
	 */
	public void setHighlightColor(Color color)
	{
		_colorHighlight = color;
	}

	/**
	 * Sets the visited color.
	 *
	 * @param color the new visited color
	 */
	public void setVisitedColor(Color color)
	{
		_colorVisited = color;
	}

	/**
	 * Sets the font and gets its associated font metrics.
	 *
	 * @see java.awt.Component#setFont(java.awt.Font)
	 */
	public void setFont(Font f)
	{
		super.setFont(f);
		_fm = getFontMetrics(f);
	}

	/**
	 * Sets the new image and initializes the icon image with the same value.
	 *
	 * @param img the new image
	 */
	public void setImage(Image img)
	{
		if (img != null)
		{
			_img = img;
			_icon = new ImageIcon(_img);
			// prepare grayed Image
			// if (USE_GRAYED_IMAGE)
			{
				_grayImg = LyGrayImageFilter.getFilteredImage(_img);
				_grayIcon = new ImageIcon(_grayImg);
				setIcon(_grayIcon);
			}
		}
		else
		{
			_img = null;
			_grayImg = null;
			_icon = null;
			_grayIcon = null;
			setIcon(null);
		}
		checkResize();
	}

	/**
	 * Sets the sensitivity.
	 *
	 * @param sensitive Indicates if the component is sensitive or not.
	 */
	public void setSensitive(boolean sensitive)
	{
		boolean wasSensitive = _sensitive;

		_sensitive = sensitive;

		highlightLabel(false);

		if (sensitive && !wasSensitive)
		{
			setForeground(_colorDefault);
			repaint();
		}
		else if (!sensitive && wasSensitive)
		{
			setForeground(_colorDisable);
			repaint();
		}
	}

	/**
	 * Sets the visited status.
	 *
	 * @param visited Indicates if the component is visited or not.
	 */
	public void setVisited(boolean visited)
	{
		boolean wasVisited = _visited;

		_visited = visited;

		highlightLabel(false);

		if (visited && !wasVisited)
		{
			setForeground(_colorVisited);
			repaint();
		}
		else if (!visited && wasVisited)
		{
			setForeground(_colorDefault);
			repaint();
		}
	}

	/**
	 * Sets the idle mode.
	 *
	 * @param status the new idle status
	 */
	public void setIdle(boolean status)
	{
		_idle = status;
		checkResize();
	}
} // end class LyJHRefComponent
