/*
 * LyJHRefContainer.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import javax.swing.*;
import leon.misc.*;


/**
 * This class defines a tree component in a panel.
 *
 * @version $Id: LyJHRefContainer.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     LyJTreeComponent
 */
public class LyJHRefContainer extends JScrollPane
{
	private static final long  serialVersionUID = -4561020500260193908L;

	public static final String __VERSION =
		"$Id: LyJHRefContainer.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Minimum width of the component. */
	public static int		   MIN_WIDTH = 80;

	/** Minimum height of the component. */
	public static int		   MIN_HEIGHT = 75;

	/** Internal panel containing href elements. */
	protected JPanel		   _internalPanel;

	/** Desired component height (in lines). */
	protected int			   _contentHeight;

	/**
	 * Creates a default href container.
	 *
	 * @param environment the current environment
	 */
	public LyJHRefContainer(LyEnvironment environment)
	{
		super(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
			ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);

		init(environment);
	}

	/**
	 * Initializes this component.
	 *
	 * @param environment the current environment
	 */
	public void init(LyEnvironment environment)
	{
		String s;

		_internalPanel = new ScrollablePanel();
		_internalPanel.setLayout(new VFlowLayout(FlowLayout.LEFT, 0, 0, true));
		_internalPanel.setBorder(null);

		// Retrieve desired component height from the resources of the application
		s = environment.getResource("LY_MULTIPLE_RELATION_CONSULT_HEIGHT");
		_contentHeight = LyTools.intFromString(s);

		// Default height is 3 lines
		if (_contentHeight <= 0)
			_contentHeight = 3;

		setViewportView(_internalPanel);
	}

	/**
	 * Changes the background color of this component and its internal panel.
	 *
	 * @param color the new background color
	 */
	public void setBackgroundColor(Color color)
	{
		super.setBackground(color);
		getViewport().setBackground(color);
		_internalPanel.setBackground(color);
	}

	/**
	 * Changes the desired component height (in lines).
	 *
	 * @param height the new desired component height (in lines).
	 */
	public void setContentHeight(int height)
	{
		_contentHeight = height;
	}
	
	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#getPreferredSize()
	 */
	public Dimension getPreferredSize()
	{
		Dimension d;

		if (_internalPanel == null)
			d = super.getPreferredSize();
		else
		{
			int n = _internalPanel.getComponentCount();

			if (n == 0)
			{
				if (_contentHeight <= 0)
					d = _internalPanel.getPreferredSize();
				else
					d = new Dimension(_internalPanel.getPreferredSize().width, 22 * _contentHeight);
			}
			else
			{
				// Preferred is _contentHeight items
				int i, w = 0, h = 0;

				// Keep height for the _content_height first items
				for (i = 0; i < _contentHeight; i++)
				{
					Component c = _internalPanel.getComponent(Math.min(i, n - 1));
					Dimension dc = c.getPreferredSize();

					w = Math.max(w, dc.width);
					h += dc.height;
				}

				// Keep width for all others
				for (i = _contentHeight; i < n; i++)
				{
					Component c = _internalPanel.getComponent(Math.min(i, n - 1));
					Dimension dc = c.getPreferredSize();

					w = Math.max(w, dc.width);
				}

				// Scrollbar will appear : keep enough space
				w += 15;
				d = new Dimension(w, h);
			}
		}

		// Workaround for scrollbar bug
		d = new Dimension(Math.max(MIN_WIDTH, d.width + 4), Math.max(MIN_HEIGHT, d.height + 4));

		return d;
	} // end method getPreferredSize

	/**
	 * Returns a dimension with {@link #MIN_WIDTH} and {@link #MIN_HEIGHT}.
	 *
	 * @return the minimum default size
	 */
	public Dimension getMinimumSize()
	{
		return new Dimension(MIN_WIDTH, MIN_HEIGHT);
	}

	/**
	 * Adds a href component to this HRef box.
	 *
	 * @param hRefComp the href component to add
	 * @param pos      the position to add the new href component
	 */
	public void add(LyJHRefComponent hRefComp, int pos)
	{
		if ((pos < 0) || (pos >= getHRefComponentCount()))
			_internalPanel.add(hRefComp);
		else
			_internalPanel.add(hRefComp, pos);

		validate();
	}

	/**
	 * Removes the specified component from the HRef box.
	 *
	 * @param hRefComp the component to remove
	 */
	public void remove(LyJHRefComponent hRefComp)
	{
		_internalPanel.remove(hRefComp);

		validate();
	}

	/**
	 * Gets the component at the specified index.
	 *
	 * @param  index the index of the component to retrieve
	 * @return the designated component
	 */
	public LyJHRefComponent getHRefComponent(int index)
	{
		return (LyJHRefComponent)_internalPanel.getComponent(index);
	}

	/**
	 * Gets the component count.
	 *
	 * @return the number of components in this container
	 */
	public int getHRefComponentCount()
	{
		return _internalPanel.getComponentCount();
	}

	/* ********************************** Internal Classes ************************************/

	/**
	 * This class implements a customized scrollpane for a panel of hrefs.
	 */
	public class ScrollablePanel extends JPanel
		implements Scrollable
	{
		private static final long serialVersionUID = -5491121826829183036L;

		/**
		 * Sole constructor.
		 */
		public ScrollablePanel()
		{
			super();
		}

		/*
		 * (non-Javadoc)
		 * @see javax.swing.Scrollable#getPreferredScrollableViewportSize()
		 */
		public Dimension getPreferredScrollableViewportSize()
		{
			return this.getPreferredSize();
		}

		/*
		 * (non-Javadoc)
		 * @see javax.swing.Scrollable#getScrollableBlockIncrement(java.awt.Rectangle, int, int)
		 */
		public int getScrollableBlockIncrement(Rectangle visibleRect, int orientation,
			int direction)
		{
			if (orientation == SwingConstants.HORIZONTAL)
				return visibleRect.width;
			return visibleRect.height;
		}

		/**
		 * Doesn't force the height of this scrollable object to match its viewport.
		 *
		 * @return <code>false</code>
		 * @see    javax.swing.Scrollable#getScrollableTracksViewportHeight()
		 */
		public boolean getScrollableTracksViewportHeight()
		{
			return false;
		}

		/**
		 * Forces the width of this scrollable object to match its viewport.
		 *
		 * @return <code>true</code>
		 * @see    javax.swing.Scrollable#getScrollableTracksViewportWidth()
		 */
		public boolean getScrollableTracksViewportWidth()
		{
			return true;
		}

		/*
		 * (non-Javadoc)
		 * @see javax.swing.Scrollable#getScrollableUnitIncrement(java.awt.Rectangle, int, int)
		 */
		public int getScrollableUnitIncrement(Rectangle visibleRect, int orientation, int direction)
		{
			if (orientation == SwingConstants.HORIZONTAL)
				return 20;
			if (this.getComponentCount() > 0)
				return this.getComponent(0).getSize().height;
			return 0;
		}
	} // end class ScrollablePanel
} // end class LyJHRefContainer
