/*
 * LyJIconJLabel.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.border.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * This class implements the icon component for swing viewer.
 *
 * @version $Id: LyJIconJLabel.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     leon.view.LyIcon
 */
public class LyJIconJLabel extends JLabel implements MouseListener
{
	private static final long  serialVersionUID = -2441528597821462726L;

	public static final String __VERSION =
		"$Id: LyJIconJLabel.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Maximum number of characters for the tooltips. */
	public static int		   TIP_MAX_CHARS = 60;

	/** Default display width, in pixels. */
	public static int		   DEFAULT_WIDTH = 20;

	/** Default display height, in pixels. */
	public static int		   DEFAULT_HEIGHT = 20;

	/** Default gap, in pixels. */
	public static int		   DEFAULT_GAP = 2;

	/** Default cursor. */
	public static Cursor	   DEFAULT_CURSOR = new Cursor(Cursor.DEFAULT_CURSOR);
	
	/** Mouse cursor when over sensitive icons. */
	public static Cursor	   OVER_CURSOR = new Cursor(Cursor.HAND_CURSOR);

	/** Original image. */
	protected Image			   _image;

	/** Resized image. */
	protected Image			   _resizedImage;

	/** Total width to display. May be greater than {@link #_image} width. */
	protected int			   _width;

	/** Total height to display. May be greater than {@link #_image} height. */
	protected int			   _height;
	
	/** Indicates whether this icon is sensitive and may be edited. */
	protected boolean		   _sensitive;

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter _actionTransmitter;

	/**
	 * Type of the border. One of the following:
	 *
	 * <ul>
	 *   <li>{@link #LyIcon.Border.NONE}</li>
	 *   <li>{@link #LyIcon.Border.SINGLE}</li>
	 *   <li>{@link #LyIcon.Border.DOUBLE}</li>
	 *   <li>{@link #LyIcon.Border.RAISED}</li>
	 *   <li>{@link #LyIcon.Border.LOWERED}</li>
	 * </ul>
	 */
	protected LyIcon.Border	   _borderType;

	/** Vertical gap. */
	protected int			   _vGap;

	/** Horizontal gap. */
	protected int			   _hGap;

	/** Associated border. */
	protected Border		   _border;

	/**
	 * Creates a default icon label component for the specified image. Initializes the class
	 * attributes to the default values.
	 *
	 * @param image the associated icon image
	 */
	public LyJIconJLabel(Image image)
	{
		this(image, null, DEFAULT_WIDTH, DEFAULT_HEIGHT);
	}

	/**
	 * Creates an icon label component with the specified attributes. This component has no border.
	 *
	 * @param image  the icon image
	 * @param tip    the tooltip text
	 * @param width  the width of the displayed image
	 * @param height the height of the displayed image
	 */
	public LyJIconJLabel(Image image, String tip, int width, int height)
	{
		this(image, tip, width, height, LyIcon.Border.NONE);
	}

	/**
	 * Creates an icon label component with the specified attributes.
	 *
	 * @param image  the icon image
	 * @param tip    the tooltip text
	 * @param width  the width of the displayed image
	 * @param height the height of the displayed image
	 * @param border the border type. Same as {@link #_borderType}.
	 */
	public LyJIconJLabel(Image image, String tip, int width, int height,
        LyIcon.Border border)
	{
		super();

		_image = image;
		if (height <= 0)
			_height = image.getHeight(null);
		else
			_height = height;

		if (width <= 0)
			_width = image.getWidth(null);
		else
			_width = width;

		String newTip;

		newTip = LyTools.formatToHTML(tip, LyJIconJLabel.TIP_MAX_CHARS);
		setToolTipText(newTip);

        setBorderType(border);
		setImage(_image);
		_sensitive = false;
		addMouseListener(this);
	}

	/**
	 * Creates an icon label component with the specified attributes and no tooltip.
	 *
	 * @param image  the icon image
	 * @param width  the width of the displayed image
	 * @param height the height of the displayed image
	 * @param border the border type. Same as _borderType.
	 */
	public LyJIconJLabel(Image image, int width, int height, LyIcon.Border border)
	{
		super();

		_image = image;

		if ((height <= 0) && (image != null))
			_height = image.getHeight(null);
		else
			_height = height;

		if ((width <= 0) && (image != null))
			_width = image.getWidth(null);
		else
			_width = width;

		setImage(_image);
		setBorderType(border);
	}

	/**
	 * Rescales the icon image if needed.
	 */
	protected void computeImage()
	{
		if (_image == null)
		{
			_resizedImage = null;

			return;
		}

		int    w = _image.getWidth(null), h = _image.getHeight(null);

		int    width = _width;
		int    height = _height;

		// if image size is bigger than the global size
		//      we try to keep the aspect ratio
		double imageRatio = (double)w / (double)h;
		if ((w < width) && (h < height))
		{
			double widthRatio = (double)w / (double)width;
			double heightRatio = (double)h / (double)height;

			if (widthRatio >= heightRatio)
			{
				w = width;
				h = (int)(w / imageRatio);
			}
			else
			{
				h = height;
				w = (int)(h * imageRatio);
			}
			_resizedImage = _image.getScaledInstance(w, h, Image.SCALE_SMOOTH);
		}
		else if ((w > width) && (h > height))
		{
			double widthRatio = (double)w / (double)width;
			double heightRatio = (double)h / (double)height;

			if (widthRatio >= heightRatio)
			{
				w = width;
				h = (int)(w / imageRatio);
			}
			else
			{
				h = height;
				w = (int)(h * imageRatio);
			}

			_resizedImage = _image.getScaledInstance(w, h, Image.SCALE_SMOOTH);
		}
		else if (w > width)
		{
			w = width;
			h = (int)(w / imageRatio);

			_resizedImage = _image.getScaledInstance(w, h, Image.SCALE_SMOOTH);
		}
		else if (h > height)
		{
			h = height;
			w = (int)(h * imageRatio);

			_resizedImage = _image.getScaledInstance(w, h, Image.SCALE_SMOOTH);
		}
		else
			// NOTHING TO DO
			_resizedImage = _image;
	} // end method computeImage

	/**
	 * Computes the vertical and horizontal gaps depending on this component's preferred size.
	 */
	// FFT00560
	public void computeGap()
	{
		Dimension dim = getPreferredSize();

		if (dim.width < _width)
			_vGap = (_width - dim.width) / 2;
		else
			_vGap = DEFAULT_GAP;

		if (dim.height < _height)
			_hGap = (_height - dim.height) / 2;
		else
			_hGap = DEFAULT_GAP;

		setBorderType(_borderType);
	}

	/**
	 * Sets the type of the border.
	 *
	 * @param borderType the type of the border. One of the following:
	 *
	 *                   <ul>
	 *                     <li>LyIcon.Border.NONE</li>
	 *                     <li>LyIcon.Border.SINGLE</li>
	 *                     <li>LyIcon.Border.DOUBLE</li>
	 *                     <li>LyIcon.Border.RAISED</li>
	 *                     <li>LyIcon.Border.LOWERED</li>
	 *                   </ul>
	 * @see   #getBorderType()
	 */
	public void setBorderType(LyIcon.Border borderType)
	{
		if (_resizedImage == null)
		{
			setBorder(null);
			return;
		}
        
        if (borderType == null)
            borderType = LyIcon.Border.NONE;

		_borderType = borderType;

		switch (_borderType)
		{
		case SINGLE:
		{
			_border = BorderFactory.createCompoundBorder(BorderFactory.createLineBorder(
						Color.black), BorderFactory.createEmptyBorder(_hGap, _vGap, _hGap, _vGap));
			break;
		}

		case DOUBLE:
		{
			_border = BorderFactory.createCompoundBorder(BorderFactory.createLineBorder(
						Color.black),
					BorderFactory.createCompoundBorder(BorderFactory.createLineBorder(Color.black),
						BorderFactory.createEmptyBorder(_hGap, _vGap, _hGap, _vGap)));
			break;
		}

		case RAISED:
		{
			_border = BorderFactory.createCompoundBorder(BorderFactory.createBevelBorder(
						BevelBorder.RAISED, Color.lightGray.brighter(), Color.lightGray.darker()),
					BorderFactory.createEmptyBorder(_hGap, _vGap, _hGap, _vGap));
			break;
		}

		case LOWERED:
		{
			_border = BorderFactory.createCompoundBorder(BorderFactory.createBevelBorder(
						BevelBorder.LOWERED, Color.lightGray.brighter(), Color.lightGray.darker()),
					BorderFactory.createEmptyBorder(_hGap, _vGap, _hGap, _vGap));
			break;
		}

		case NONE:
			_border = BorderFactory.createEmptyBorder(_hGap, _vGap, _hGap, _vGap);
		}
		setBorder(_border);
	} // end method setBorderType

	/**
	 * Gets the type of the border.
	 *
	 * @return the type of the border
	 * @see    #setBorderType(int)
	 */
	public LyIcon.Border getBorderType()
	{
		return _borderType;
	}

	/**
	 * Sets the displayed image.
	 *
	 * @param image the icon image
	 */
	public void setImage(Image image)
	{
		_image = image;
		
		if (_image != null)
		{
			if (_height <= 0)
				_height = _image.getHeight(null);

			if (_width <= 0)
				_width = _image.getWidth(null);
		}
		
		computeImage();

		ImageIcon icon = null;

		if (_resizedImage != null)
			icon = new ImageIcon(_resizedImage);

		setIcon(icon);

		computeGap();
	}
	
	/**
	 * Sets or unsets the sensitive mode on this icon.
	 * 
	 * @param sensitive true if icon is sensitive (may be changed by clicking on it), false otherwise.
	 */
	public void setSensitive(boolean sensitive)
	{
		_sensitive = sensitive;
		setCursor(_sensitive ? OVER_CURSOR : DEFAULT_CURSOR);
	}
	
	/**
	 * Sets the width and height of the icon.
	 * 
	 * @param width new width value.
	 * @param height new height value.
	 */
	public void setDimension(int width, int height)
	{
		_width = width;
		_height = height;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return (_actionTransmitter);
	}

    /**
     * Invoked when the mouse button has been clicked (pressed
     * and released) on a component.
     */
    public void mouseClicked(MouseEvent e)
    {
		if (_sensitive && (_actionTransmitter != null))
		{
			LyActionEvent	actionEvent = null;
			
			if (e.getModifiers() == InputEvent.BUTTON3_MASK)
			{
				Point		  pos = e.getPoint();

				actionEvent = new LyActionEvent(this, null, null, LyMenuDemand.TYPE_POPUP);
				actionEvent.setParameter(new int[] { pos.x, pos.y });
			}
			else
				actionEvent = new LyActionEvent(this);
			
			_actionTransmitter.processActionEvent(actionEvent);
		}
    }

    /**
     * Invoked when a mouse button has been pressed on a component.
     */
    public void mousePressed(MouseEvent e)
    {
    }

    /**
     * Invoked when a mouse button has been released on a component.
     */
    public void mouseReleased(MouseEvent e)
    {
    }

    /**
     * Invoked when the mouse enters a component.
     */
    public void mouseEntered(MouseEvent e)
    {
    }

    /**
     * Invoked when the mouse exits a component.
     */
    public void mouseExited(MouseEvent e)
    {
    }
} // end class LyJIconJLabel
