/*
 * LyJListComponent.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import leon.view.event.*;
import leon.view.getinterface.*;


/**
 * This class implements a list of data.
 *
 * @see JList
 * @see LyJMultiSelector
 */
public class LyJListComponent extends JList
	implements LySelectionListener,
		LyActionListener,
		LyEventTransmitter
{
	private static final long		    serialVersionUID = -650230096697989857L;

	public static final String		    __VERSION =
		"$Id: LyJListComponent.java 28997 2010-12-23 10:17:15Z bmichel $";

	/** Separator character. */
	public static final char		    ITEM_SEP = '\1';

	/** The data contained by this list. */
	protected Vector<Object>			_content;

	/** The {@link ListSelectionListener} that receives selection events from the list. */
	protected _LyJListSelectionListener _selectionListener;

	/** The {@link MouseListener} that receives mouse events from the list. */
	protected _LyJListMouseListener     _mouseListener;

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter	    _actionTransmitter;

	/** Provides notification events on component's selections. */
	protected LySelectionTransmitter    _selectionTransmitter;
	
	/** Fixed height that must be used to display component. */
	protected int						_fixedHeight;

	/**
	 * Sole constructor. Initializes the list component and registers its event listeners.
	 */
	public LyJListComponent()
	{
		super();
		_content = new Vector<Object>();
		setModel(new _LyJListModel());
		_selectionListener = new _LyJListSelectionListener();
		_mouseListener = new _LyJListMouseListener();
		_fixedHeight = -1;
		addLyJListSelectionListener();
		addLyJListMouseListener();
		setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		setSelectionModel(new _LyMultiSelectionModel());
	}

	/**
	 * Registers the selection events.
	 *
	 * @see #removeLyJListSelectionListener()
	 */
	public void addLyJListSelectionListener()
	{
		addListSelectionListener(_selectionListener);
	}

	/**
	 * Registers the mouse events.
	 *
	 * @see #removeLyJListMouseListener()
	 */
	public void addLyJListMouseListener()
	{
		addMouseListener(_mouseListener);
	}

	/**
	 * Unregisters the selection events.
	 *
	 * @see #addLyJListSelectionListener()
	 */
	public void removeLyJListSelectionListener()
	{
		removeListSelectionListener(_selectionListener);
	}

	/**
	 * Unregisters the mouse events.
	 *
	 * @see #addLyJListMouseListener()
	 */
	public void removeLyJListMouseListener()
	{
		removeMouseListener(_mouseListener);
	}

	/**
	 * Selects or deselects the specified list item.
	 *
	 * @param item   the item to select
	 * @param select Indicates if the specified item element has to be selected or deselected.
	 */
	public void selectItem(Object item, boolean select)
	{
		int index = indexOf(item);

		removeLyJListSelectionListener();

		if (select)
			addSelectionInterval(index, index);
		else
			removeSelectionInterval(index, index);

		addLyJListSelectionListener();
	}

	/**
	 * Resets the content of the list to the specified list content.
	 *
	 * @param items the new content of the list
	 */
	public void setItems(Vector<Object> items)
	{
		if (items == null)
			items = new Vector<Object>();
        
		int oldSize = _content.size();
		int newSize = items.size();

		_content = items;
		getLyJListModel().listUpdated(oldSize, newSize);
	}
    
    public Dimension getPreferredSize() 
    {
    	synchronized (_content) 
    	{
    		return super.getPreferredSize();	
		}
    }
	/**
	 * Gets the scrollable viewport to have more space in width.
	 *
	 * @see javax.swing.Scrollable#getPreferredScrollableViewportSize()
	 */
	public Dimension getPreferredScrollableViewportSize()
	{
		Dimension size = super.getPreferredScrollableViewportSize();

		return new Dimension(size.width + 20, (_fixedHeight > 0 ? _fixedHeight : size.height));
	}
	
	/**
	 * Sets the fixed height of this component.
	 * 
	 * @param fixedHeight the fixed height of this component in pixels.
	 */
	public void setFixedHeight(int fixedHeight)
	{
		_fixedHeight = fixedHeight;
	}

	/**
	 * Adds an item at the end of this list.
	 *
	 * @param item the item to add
	 */
	public void addItem(Object item)
	{
		_content.addElement(item);
		getLyJListModel().itemAdded(_content.size() - 1);
	}

	/**
	 * Inserts an item in this list at the specified position.
	 *
	 * @param item  the item to add
	 * @param index the index where to add
	 */
	public void insertItemAt(Object item, int index)
	{
		_content.insertElementAt(item, index);
		getLyJListModel().itemAdded(index);
	}

	/**
	 * Removes an item from this list.
	 *
	 * @param item the item to remove from the list
	 */
	public void removeItem(Object item)
	{
		int index = _content.indexOf(item);

		_content.removeElement(item);
		getLyJListModel().itemRemoved(index);
	}

	/**
	 * Removes an item from this list.
	 *
	 * @param index the index of the item to remove
	 */
	public void removeItemAt(int index)
	{
		_content.removeElementAt(index);

		getLyJListModel().itemRemoved(index);
	}

	/**
	 * Gets this list data model.
	 *
	 * @return the data model of the list
	 */
	public _LyJListModel getLyJListModel()
	{
		return (_LyJListModel)getModel();
	}

	/**
	 * Gets the position in this list of the specified item.
	 *
	 * @param  item the item
	 * @return the index of an item
	 */
	public int indexOf(Object item)
	{
		return _content.indexOf(item);
	}

	/**
	 * Gets the number of items.
	 *
	 * @return the count of items
	 */
	public int getCount()
	{
		return getModel().getSize();
	}

	/**
	 * Gets the number of selected items it this list.
	 *
	 * @return the count of selected items
	 */
	public int getSelectedCount()
	{
		return getSelectedIndices().length;
	}

	/**
	 * Gets the item at the specified position.
	 *
	 * @param  index the position index of the object to retrieve
	 * @return the item at position index
	 */
	public Object getItemAt(int index)
	{
		return getModel().getElementAt(index);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent e)
	{
		if (_actionTransmitter != null)
			_actionTransmitter.processActionEvent(e);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LySelectionListener#selectionPerformed(leon.view.event.LySelectionEvent)
	 */
	public void selectionPerformed(LySelectionEvent e)
	{
		if (_selectionTransmitter != null)
			_selectionTransmitter.processSelectionEvent(e);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();
		return (_actionTransmitter);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		if (_selectionTransmitter == null)
			_selectionTransmitter = new LySimpleSelectionTransmitter();
		return (_selectionTransmitter);
	}

	/* *******************************************************************************/

	/**
	 * This class implements the access to the data to display in the list.
	 */
	public class _LyJListModel extends AbstractListModel
	{
		private static final long serialVersionUID = 1496406309412483516L;

		/**
		 * Sole constructor.
		 */
		public _LyJListModel()
		{
			super();
		}

		/*
		 * (non-Javadoc)
		 * @see javax.swing.ListModel#getElementAt(int)
		 */
		public Object getElementAt(int index)
		{
			if (index >= _content.size())
			{
				return null;
			}
			return _content.elementAt(index);
		}

		/*
		 * (non-Javadoc)
		 * @see javax.swing.ListModel#getSize()
		 */
		public int getSize()
		{
			if (_content == null)
				return 0;
			return _content.size();
		}

		/**
		 * Notifies the LyListComponent that a row has been added.
		 *
		 * @param index the index of the new row
		 */
		public void itemAdded(int index)
		{
			fireIntervalAdded(this, index, index);
		}

		/**
		 * Notifies the LyListComponent that a row has been removed.
		 *
		 * @param index the index of the new row
		 */
		public void itemRemoved(int index)
		{
			fireIntervalRemoved(this, index, index);
		}

		/**
		 * Notifies the LyListComponent that the data has totally changed.
		 *
		 * @param oldSize the old size of the list
		 * @param newSize the new size of the list
		 */
		public void listUpdated(int oldSize, int newSize)
		{
			if (oldSize > 0)
				fireIntervalRemoved(this, 0, oldSize - 1);
			if (newSize > 0)
				fireIntervalAdded(this, 0, newSize - 1);
		}

	} // end class _LyJListModel


	/* **************************************************************************************/

	/**
	 * This class manages the selection with a mouse drag.
	 */
	class _LyMultiSelectionModel extends DefaultListSelectionModel
	{
		private static final long serialVersionUID = 8896244959677984122L;

		/** Indicates that the gesture has started. */
		boolean					  gestureStarted = false;

		/**
		 * This property allows listeners to update only when a change has been finalized, rather
		 * than always handling all of the intermediate values.
		 */
		boolean					  _adjusting = false;

		/** Starting value of the selection. */
		int						  startIndex = 0;

		/*
		 * (non-Javadoc)
		 * @see javax.swing.ListSelectionModel#setSelectionInterval(int, int)
		 */
		public void setSelectionInterval(int index0, int index1)
		{
			if (!_adjusting)
				super.setSelectionInterval(index0, index1);
			else
			{
				if (gestureStarted)
					super.setSelectionInterval(startIndex, index1);
				else
				{
					super.setSelectionInterval(index0, index1);

					gestureStarted = true;
					startIndex = index0;
				}
			}
		}

		/*
		 * (non-Javadoc)
		 * @see javax.swing.ListSelectionModel#setValueIsAdjusting(boolean)
		 */
		public void setValueIsAdjusting(boolean isAdjusting)
		{
			_adjusting = isAdjusting;
			if (isAdjusting == false)
				gestureStarted = false;
		}
	} // end class _LyMultiSelectionModel


	/**
	 * This class manages the selection events from the view to the model.
	 */
	public class _LyJListSelectionListener
		implements ListSelectionListener
	{

		/*
		 * (non-Javadoc)
		 * @see javax.swing.event.ListSelectionListener#valueChanged(javax.swing.event.ListSelectionEvent)
		 */
		public void valueChanged(ListSelectionEvent e)
		{
			Vector<Object> selectedIds;

			if (!e.getValueIsAdjusting())
			{
				selectedIds = getSelectedIds();
				if (_selectionTransmitter != null)
				{
					LySelectionEvent select = new LySelectionEvent(this, selectedIds, true);

					select.addIndex(getMinSelectionIndex());
					select.setSelectionSize(selectedIds.size());

					selectionPerformed(select);
				}
			}
		}

		/**
		 * Gets the ids of the selected objects.
		 *
		 * @return the list of ids of all the selected objects in this list
		 */
		public Vector<Object> getSelectedIds()
		{
			Vector<Object>	   selectedIds = new Vector<Object>();
			ListSelectionModel selection = getSelectionModel();
			int				   i, n = getCount();

			for (i = 0; i < n; i++)
			{
				if (selection.isSelectedIndex(i))
					selectedIds.add(getItemAt(i));
			}
			return selectedIds;
		}

	} // end class _LyJListSelectionListener


	/* *************************************************************************************/

	/**
	 * This class manages the mouse events that occur on the table.
	 */
	public class _LyJListMouseListener
		implements MouseListener
	{

		/**
		 * Does nothing.
		 *
		 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
		 */
		public void mouseEntered(MouseEvent e)
		{
		}

		/**
		 * Does nothing.
		 *
		 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
		 */
		public void mouseExited(MouseEvent e)
		{
		}

		/**
		 * Does nothing.
		 *
		 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
		 */
		public void mousePressed(MouseEvent e)
		{
		}

		/**
		 * Does nothing.
		 *
		 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
		 */
		public void mouseReleased(MouseEvent e)
		{
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
		 */
		public void mouseClicked(MouseEvent e)
		{
			Point p = e.getPoint();
			int   index = locationToIndex(p);

			// left double click
			if ((e.getClickCount() == 2) && ((e.getModifiers() & InputEvent.BUTTON1_MASK) > 0)
					&& (index != -1))
			{
				if (_actionTransmitter != null)
				{
					Object item = getItemAt(index);

					if (!(item instanceof LyGetIdInterface))
						return;
					LyActionEvent actionEvent = new LyActionEvent(e.getSource());

					actionEvent.setValue(((LyGetIdInterface)item).getId());
					actionEvent.addIndex(index);
					actionPerformed(actionEvent);
				}
			}
		}

	} // end class _LyJListMouseListener

} // end class LyJListComponent
