/*
 * LyJMessageWindow.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import leon.misc.*;
import leon.view.*;
import leon.view.setinterface.*;

import java.awt.*;
import java.awt.event.*;

import java.util.*;

import javax.swing.*;


/**
 * Default message window.
 *
 * @version $Id: LyJMessageWindow.java 35072 2012-03-28 12:27:27Z bangiole $
 * @author  Lyria
 */
public class LyJMessageWindow extends JDialog implements WindowListener, ActionListener,
    KeyListener, LyFreeInterface
{
    private static final long serialVersionUID = -150801547091870737L;

    public static final String __VERSION =
        "$Id: LyJMessageWindow.java 35072 2012-03-28 12:27:27Z bangiole $";

    /** Dialog command bar content. */
    protected BitSet _commands;

    /** Text canvas. */
    protected Component _text;

    /** Action listeners. */
    protected Vector<ActionListener> _actionListeners;

    /** Closing status. */
    protected int _commandResult;

    /** Keep association between buttons and code. */
    protected Object[] _assoc;

    /** The current environment. */
    protected LyEnvironment _environment;

    /** Client data freely set by opener. */
    protected Object _clientData;

    /** Inside panel. */
    protected JPanel _panel;

    /** The background color. */
    protected Color _bgColor;

    /**
     * Creates a message window object with the specified attributes.
     *
     * @param frame       the parent frame
     * @param title       the title of this message window
     * @param commands    the setted dialog command bar commands
     * @param modal       Indicates if this window is modal.
     * @param bgColor     the background color
     * @param environment the current environment
     */
    public LyJMessageWindow(Frame frame, String title, BitSet commands, boolean modal,
        Color bgColor, LyEnvironment environment)
    {
        super(frame, title, modal);

        setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
        _environment = environment;
        _commands = commands;
        _panel = null;
        _bgColor = bgColor;

        if ((_commands == null) || (_commands.size() == 0))
        {
            _commands = new BitSet();
            _commands.set(LyViewManager.COMMAND_OK);
        }

        initialize();
    }

    /**
     * Creates a message window object with the specified attributes.
     *
     * @param dialog      the parent dialog
     * @param title       the title of this message window
     * @param commands    the setted dialog command bar commands
     * @param modal       Indicates if this window is modal.
     * @param bgColor     the background color
     * @param environment the current environment
     */
    public LyJMessageWindow(Dialog dialog, String title, BitSet commands, boolean modal,
        Color bgColor, LyEnvironment environment)
    {
        super(dialog, title, modal);

        setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
        _environment = environment;
        _commands = commands;
        _panel = null;
        _bgColor = bgColor;

        if ((_commands == null) || (_commands.size() == 0))
        {
            _commands = new BitSet();
            _commands.set(LyViewManager.COMMAND_OK);
        }

        initialize();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LyFreeInterface#free()
     */
    public void free()
    {

        if (isDisplayable())
        {
            setVisible(false);
            dispose();
        }
    }

    /**
     * Initilizes this message window, creates the correct command buttons and registers its event
     * listeners.
     */
    protected void initialize()
    {
        _panel = new JPanel(new BorderLayout());

        // Create command bar
        JPanel commandBar = new JPanel();

        commandBar.setLayout(new FlowLayout());

        // Build text area
        _text = buildTextComponent();

        // If an image is given as a background, change the content panel
        Image image = (Image) _environment.getImage("LY_MESSAGE_BACKGROUND_IMAGE", null, false);

        _panel.setOpaque(false);
        ((LyJMultiLineLabel) _text).setOpaque(false);
        commandBar.setOpaque(false);

        if (image != null)
        {
            LyJBackgroundPanel contentPane = new LyJBackgroundPanel(image);

            setContentPane(contentPane);
            getContentPane().setBackground(Color.YELLOW);
        }

        getContentPane().setLayout(new FlowLayout(FlowLayout.CENTER, 7, 0));

        JRootPane rootPane = getRootPane();

        _assoc = new Object[LyViewManager.COMMAND_HELP];

        if (_commands.get(LyViewManager.COMMAND_OK))
            addButton(commandBar, LyViewManager.COMMAND_OK, "LY_OK");

        if (_commands.get(LyViewManager.COMMAND_YES))
            addButton(commandBar, LyViewManager.COMMAND_YES, "LY_YES");

        if (_commands.get(LyViewManager.COMMAND_ALL))
            addButton(commandBar, LyViewManager.COMMAND_ALL, "LY_ALL");

        if (_commands.get(LyViewManager.COMMAND_NONE))
            addButton(commandBar, LyViewManager.COMMAND_NONE, "LY_NONE");

        if (_commands.get(LyViewManager.COMMAND_NO))
            addButton(commandBar, LyViewManager.COMMAND_NO, "LY_NO");

        if (_commands.get(LyViewManager.COMMAND_CANCEL))
            addButton(commandBar, LyViewManager.COMMAND_CANCEL, "LY_CANCEL");

        if (_commands.get(LyViewManager.COMMAND_HELP))
            addButton(commandBar, LyViewManager.COMMAND_HELP, "LY_HELP");

        if (_bgColor == null)
            _bgColor = (Color) _environment.getColor("LY_MESSAGE_BACKGROUND");

        if (_bgColor == null)
            _bgColor = (Color) _environment.getColor("LY_DEFAULT_BACKGROUND");

        if (_bgColor != null)
        {
            commandBar.setBackground(_bgColor);
            getContentPane().setBackground(_bgColor);
        }

        _panel.add("Center", _text);
        _panel.add("South", commandBar);

        if (commandBar.getComponentCount() > 0)
        {
            Component button = commandBar.getComponent(0);

            if (button instanceof JButton)
                rootPane.setDefaultButton((JButton) button);
        }

        getContentPane().add(_panel);

        addWindowListener(this);

        setResizable(false);

        _commandResult = 0;
    } // end method initialize

    /**
     * Adds a button at the end of the specified command bar.
     *
     * @param  commandBar the command bar to add the button
     * @param  type       the type of this button
     * @param  label      the button label
     * @return the added button
     */
    protected JButton addButton(JPanel commandBar, short type, String label)
    {
        Image   image = (Image) _environment.getImage(label, null, false);
        boolean rollover = true;
        Font    font = (Font) _environment.getFont("LY_COMMAND_FONT");
        JButton b;

        label = _environment.translate(label);

        if (_environment.testResource("LY_USE_IMAGE_FOR_COMMAND"))
            b = new LyJButton(null, image, rollover);
        else
            b = new LyJButton(label, image, rollover);

        b.setToolTipText(label);
        b.setFont(font);
        b.setDefaultCapable(true);
        b.setBackground(null);
        commandBar.add(b);
        b.addActionListener(this);
        b.addKeyListener(this);
        _assoc[type] = b;

        return b;
    }

    /**
     * Builds a centered text component.
     *
     * @return this text component
     */
    protected Component buildTextComponent()
    {
        LyJMultiLineLabel text = new LyJMultiLineLabel(getTitle(), 70);

        text.setAlignment(LyJMultiLineLabel.CENTER);
        if (_bgColor != null)
            text.setBackground(_bgColor);
        return text;
    }

    /**
     * Gets the closing status.
     *
     * @return the COMMAND code
     */
    public int getCommandResult()
    {
        return _commandResult;
    }

    /**
     * Adds an action listener to this component.
     *
     * @param listener the action listener to add
     */
    public void addActionListener(ActionListener listener)
    {

        if (_actionListeners == null)
            _actionListeners = new Vector<ActionListener>();

        _actionListeners.addElement(listener);
    }

    /**
     * Sets the message shown in the view.
     *
     * @param message the new window message
     */
    public void setMessage(String message)
    {
        ((LyJMultiLineLabel) _text).setLabel(message);
    }

    /**
     * Sets the icon shown in the view.
     *
     * @param icon the new window icon
     */
    public void setIcon(String icon)
    {

        if (icon != null)
        {
            JLabel iconLabel = new JLabel(new ImageIcon((Image) _environment.getImage(icon)));

            iconLabel.setBackground(null);
            _panel.add("West", iconLabel);
        }
    }

    /**
     * (non-Javadoc).
     *
     * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
     */
    public void actionPerformed(ActionEvent actionEvent)
    {
        JButton b = (JButton) actionEvent.getSource();

        actionPerformed(b);
    }

    /**
     * Invoked when a button is pressed.
     *
     * @param b the invoking button
     */
    protected void actionPerformed(JButton b)
    {
        // Change closing status
        _commandResult = 0;

        int i, n = ((_assoc == null) ? 0 : _assoc.length);

        for (i = 0; (i < n) && (_commandResult == 0); i++)
        {

            if (b == _assoc[i])
                _commandResult = i;
        }

        close(b.getText());
    }

    /**
     * Does nothing.
     *
     * @see java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
     */
    public void keyTyped(KeyEvent evt)
    {
    }

    /**
     * (non-Javadoc).
     *
     * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
     */
    public void keyPressed(KeyEvent evt)
    {
        int key = evt.getKeyCode();

        if (key == KeyEvent.VK_ENTER)
        {
            JButton b = (JButton) evt.getSource();

            actionPerformed(b);
        }
    }

    /**
     * Does nothing.
     *
     * @see java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
     */
    public void keyReleased(KeyEvent evt)
    {
    }

    /**
     * Sets the visible property of this window.
     *
     * @param status    Indicates if this window is shown or hid.
     * @param newThread Displays this dialog window starting from another thread.
     */
    public void setVisible(boolean status, boolean newThread)
    {

        if (status && !isShowing())
        {
            pack();
            setLocation();

            if (!isModal())
            {
                toFront();
                super.setVisible(true);
            }
            else
            {

                // Showing a dialog window from AWT-Event-Queue crashes the JVM
                // so we display it starting from another thread
                if (newThread)
                    new Thread()
                        {

                            /*
                             * (non-Javadoc)
                             * @see java.lang.Runnable#run()
                             */
                            public void run()
                            {
                                showDialog();
                            }
                        }.start();
                else
                    super.setVisible(true);
            }
        }
        else
            super.setVisible(status);
    } // end method setVisible

    /**
     * Shows this message window.
     */
    void showDialog()
    {
        super.setVisible(true);
    }

    /**
     * Does nothing.
     *
     * @see java.awt.event.WindowListener#windowOpened(java.awt.event.WindowEvent)
     */
    public void windowOpened(WindowEvent e)
    {
        // nothing to do
    }

    /**
     * Calls the closing method for this window.
     *
     * @see WindowListener#windowClosing
     */
    public void windowClosing(WindowEvent e)
    {
        close(null);
    }

    /**
     * Does nothing.
     *
     * @see java.awt.event.WindowListener#windowClosed(java.awt.event.WindowEvent)
     */
    public void windowClosed(WindowEvent e)
    {
        // nothing to do
    }

    /**
     * Does nothing.
     *
     * @see java.awt.event.WindowListener#windowIconified(java.awt.event.WindowEvent)
     */
    public void windowIconified(WindowEvent e)
    {
        // nothing to do
    }

    /**
     * Does nothing.
     *
     * @see java.awt.event.WindowListener#windowDeiconified(java.awt.event.WindowEvent)
     */
    public void windowDeiconified(WindowEvent e)
    {
        // nothing to do
    }

    /**
     * Does nothing.
     *
     * @see java.awt.event.WindowListener#windowActivated(java.awt.event.WindowEvent)
     */
    public void windowActivated(WindowEvent e)
    {
        // nothing to do
    }

    /**
     * Does nothing.
     *
     * @see java.awt.event.WindowListener#windowDeactivated(java.awt.event.WindowEvent)
     */
    public void windowDeactivated(WindowEvent e)
    {
        // nothing to do
    }

    /**
     * Closes this current window then disposes it.
     *
     * @param text a command associated with this window action event
     */
    protected void close(String text)
    {
        setVisible(false);

        if (_actionListeners != null)
        {
            ActionEvent e = new ActionEvent(this, hashCode(), text);

            int i, n = _actionListeners.size();

            for (i = 0; i < n; i++)
            {
                ActionListener listener = _actionListeners.elementAt(i);

                listener.actionPerformed(e);

            }
        }

        dispose();
    }

    /**
     * Moves the component at the default location on the screen.
     */
    protected void setLocation()
    {
        // FFT00435/Flandres
        Dimension size = getSize();
        int[]     screenSize = _environment.getScreenSize(this);

        setLocation(((screenSize[0] - size.width) / 2) + screenSize[2],
            ((screenSize[1] - size.height) / 2) + screenSize[3]);
    }

    /**
     * Gets the client data freely set by opener.
     *
     * @return the client data
     * @see    #setClientData(Object)
     */
    public Object getClientData()
    {
        return _clientData;
    }

    /**
     * Sets the client data.
     *
     * @param clientData the new client data object
     * @see   #getClientData()
     */
    public void setClientData(Object clientData)
    {
        _clientData = clientData;
    }
} // end class LyJMessageWindow
