/*
 * LyJMultiLineLabel.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.util.*;
import javax.swing.*;

import leon.misc.*;


/**
 * This example is from the book _Java in a Nutshell_ by David Flanagan.<br>
 * Written by David Flanagan.<br>
 * Copyright (c) 1998 O'Reilly & Associates.<br>
 * You may study, use, modify, and distribute this example for any purpose. This example is provided
 * WITHOUT WARRANTY either expressed or implied.
 *
 * <p>Tilman 05/07/96: added a maxChars parameter to specify a maximum line length. If any of the
 * lines is longer, it will be cut at the last space character before maxChars, until it fits into
 * maxChars. Set maxChars to -1 if you don't care.</p>
 *
 * <p>Van 07/30/96: added the removal of backslash-r from the multi-line string that might have been
 * added by strings in Windows (which expect newlines to be indicated by backslash-r backslash-n)
 * </p>
 *
 * @version $Id: LyJMultiLineLabel.java 28422 2010-11-03 12:03:38Z ehelmste $
 */

public class LyJMultiLineLabel extends JPanel
{
	private static final long  serialVersionUID = -5123248100633332112L;

	public static final String __VERSION =
		"$Id: LyJMultiLineLabel.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Left alignment constant. */
	public static final int    LEFT = Label.LEFT;

	/** Center alignment constant. */
	public static final int    CENTER = Label.CENTER;

	/** Right alignment constant. */
	public static final int    RIGHT = Label.RIGHT;

	/** The label of this component. */
	protected String		   _label;

	/** The lines of text to display. */
	protected Vector<String>   _lines;

	/** The number of lines. */
	protected int			   _num_lines;

	/** The maximum width of lines (in characters). */
	protected int			   _maxChars;

	/** The maximum width of lines (in pixels). */
	protected int			   _maxWidth;

	/** Left and right margins. */
	protected int			   _margin_width;

	/** Top and bottom margins. */
	protected int			   _margin_height;

	/** Total height of the font. */
	protected int			   _line_height;

	/** Font height above baseline. */
	protected int			   _line_ascent;

	/** How wide each line is. */
	protected int[]			   _line_widths;

	/** The width of the widest line. */
	protected int			   _max_width;

	/** The alignment of the text. */
	protected int			   _alignment = LEFT;

	/** Number of columns. Use -1 for automatic calculation. */
	protected int			   _nCols;

	/** Number of rows. Use -1 for automatic calculation. */
	protected int			   _nRows;

	/** Width of a single character. Use -1 for automatic calculation. */
	protected int			   _charWidth = -1;

	/** The tooltip text. */
	protected String		   _tooltip = null;

	/** The biggest font used to compute the size of this label. */
	protected Font			   _bigFont;

	/** Minimum width of the component. */
	protected int			   _minimumWidth;
	
	/**
	 * Breaks a specified label up into an array of lines. It uses the StringTokenizer utility
	 * class.
	 *
	 * @param label the new multiline label
	 */
	protected void newLabel(String label)
	{
		if (label == null)
			label = "";

		_label = label;

		String[]	list = LyString.stringToList(label, '\n');

		_lines = new Vector<String>();
        
		for (int i = 0; (list != null) && (i < list.length); i++)
		{
			String line = list[i];

			if (_maxWidth > 0)
			{
				FontMetrics		fm = getFontMetrics(getFont());
				
				if (fm.stringWidth(line) <= _maxWidth)
					_lines.addElement(line);
				else
				{
					while (line.length() > 0)
					{
						int			pos = 1;
						String		lineTmp = "";
						int			offset = line.indexOf(' ');
						
						while ((pos <= line.length()) && (fm.stringWidth(line.substring(0, pos)) <= _maxWidth))
						{
							lineTmp = line.substring(0, pos);
							pos++;
						}
						
						offset = lineTmp.lastIndexOf(' ');
						
						if ((offset < 0) && !line.equals(lineTmp))
							offset = line.indexOf(' ');
						
						if ((offset < 0) || line.equals(lineTmp))
						{
							_lines.addElement(line);
							break;
						}
						else
							_lines.addElement(line.substring(0, offset));

						line = line.substring(offset + 1);
					}
				}
			}
			else
			{
				if ((_maxChars == -1) || (line.length() <= _maxChars))
					_lines.addElement(line);
				else
				{
					while (line.length() > _maxChars)
					{
						int offset = line.lastIndexOf(' ', _maxChars);

						if (offset == -1)
						{
							// didn't get one within maxChars!
							offset = line.indexOf(' ');
							if (offset == -1)
								break;
						}
						_lines.addElement(line.substring(0, offset));
						line = line.substring(offset + 1);
					}
					_lines.addElement(line);
				}
			}
		}
		_num_lines = _lines.size();
		_line_widths = new int[_num_lines];

		// Register to the ToolTipManager
		ToolTipManager ttm = ToolTipManager.sharedInstance();

		ttm.registerComponent(this);
	} // end method newLabel

	/**
	 * Figures out how the font is, and how wide each line of the label is, and how wide the widest
	 * line is.
	 */
	protected void measure()
	{
		if (getFont() == null)
			return;

		Font font = getFont();

		if (_bigFont != null) // use the biggest font to compute the size
			font = _bigFont;

		FontMetrics fm = getFontMetrics(font);

		// If we don't have font metrics yet, just return.
		if (fm == null)
			return;

		_line_height = fm.getHeight();
		_line_ascent = fm.getAscent();
		_max_width = 0;
		for (int i = 0; i < _num_lines; i++)
		{
			_line_widths[i] = fm.stringWidth(_lines.elementAt(i));
			if (_line_widths[i] > _max_width)
				_max_width = _line_widths[i];
		}

		_charWidth = computeCharWidth(fm);

	} // end method measure

	/**
	 * Creates a new multiline label object from the specified attributes. Breaks the label up into
	 * separate lines, and saves the other info.
	 *
	 * @param label         the multiline label
	 * @param maxChars      the maximum number of characters per line
	 * @param margin_width  the left and right margins
	 * @param margin_height the top and bottom margins
	 * @param alignment     the alignment of the text
	 */
	public LyJMultiLineLabel(String label, int maxChars, int margin_width, int margin_height,
		int alignment)
	{
		_maxChars = maxChars;
		_nCols = -1;
		_nRows = -1;
		_minimumWidth = -1;
		_maxWidth = -1;
		newLabel(label);
		_margin_width = margin_width;
		_margin_height = margin_height;
		_alignment = alignment;
		_bigFont = null;
	}

	/**
	 * Creates a new multiline label object from the specified attributes. Breaks the label up into
	 * separate lines, and saves the other info. By default, the alignment is on the left.
	 *
	 * @param label         the multiline label
	 * @param maxChars      the maximum number of characters per line
	 * @param margin_width  the left and right margins
	 * @param margin_height the top and bottom margins
	 */
	public LyJMultiLineLabel(String label, int maxChars, int margin_width, int margin_height)
	{
		this(label, maxChars, margin_width, margin_height, LEFT);
	}

	/**
	 * Creates a new multiline label object from the specified attributes. Breaks the label up into
	 * separate lines, and saves the other info. By default, the horizontal and vertical margins are
	 * 20 pixels.
	 *
	 * @param label     the multiline label
	 * @param maxChars  the maximum number of characters per line
	 * @param alignment the alignment of the text
	 */
	public LyJMultiLineLabel(String label, int maxChars, int alignment)
	{
		this(label, maxChars, 10, 10, alignment);
	}

	/**
	 * Creates a new multiline label object from the specified attributes. Breaks the label up into
	 * separate lines, and saves the other info. By default, the horizontal and vertical margins are
	 * 20 pixels and the alignment is on the left.
	 *
	 * @param label    the multiline label
	 * @param maxChars the maximum number of characters per line
	 */
	public LyJMultiLineLabel(String label, int maxChars)
	{
		this(label, maxChars, 10, 10, LEFT);
	}

	/**
	 * Creates a new multiline label object from the specified attributes. Breaks the label up into
	 * separate lines, and saves the other info. By default, the width of this label is the maximum
	 * available, its horizontal and vertical margins are 20 pixels and the alignment is on the
	 * left.
	 *
	 * @param label the multiline label
	 */
	public LyJMultiLineLabel(String label)
	{
		this(label, -1, 10, 10, LEFT);
	}

	/**
	 * Creates a new multiline label object from the specified attributes. Breaks the label up into
	 * separate lines, and saves the other info. By default, the label is "MultiLineLabel", the
	 * width of this label is the maximum available, its horizontal and vertical margins are 20
	 * pixels and the alignment is on the left.
	 */
	public LyJMultiLineLabel()
	{
		this("MultiLineLabel");
	}

	/**
	 * Sets the label of this component. This label is broken in multiple lines and it is measured
	 * again.
	 *
	 * @param label the new label
	 * @see   #getLabel()
	 */
	public void setLabel(String label)
	{
		if (label == null)
			label = "";
		newLabel(label);
		measure();
		repaint();
	}

	/**
	 * Gets the label of this component.
	 *
	 * @return the label
	 * @see    #setLabel(String)
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Sets the biggest font for this component.
	 *
	 * @param f the font that is the biggest for the measurement computations
	 */
	public void setBigFont(Font f)
	{
		_bigFont = f;

	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#setFont(java.awt.Font)
	 */
	public void setFont(Font f)
	{
		super.setFont(f);
		_charWidth = -1;
		measure();
		repaint();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#setForeground(java.awt.Color)
	 */
	public void setForeground(Color c)
	{
		super.setForeground(c);
		repaint();
	}

	/**
	 * Sets the alignment policy of this label.
	 *
	 * @param a the alignment value. One of the following:
	 *
	 *          <ul>
	 *            <li>{@link #LEFT}</li>
	 *            <li>{@link #RIGHT}</li>
	 *            <li>{@link #CENTER}</li>
	 *          </ul>
	 * @see   #getAlignment()
	 */
	public void setAlignment(int a)
	{
		_alignment = a;
		repaint();
	}

	/**
	 * Sets the horizontal margins of this component.
	 *
	 * @param mw the margin width, in pixels
	 * @see   #getMarginWidth()
	 */
	public void setMarginWidth(int mw)
	{
		_margin_width = mw;
		repaint();
	}

	/**
	 * Sets the vertical margins of this component.
	 *
	 * @param mh the margin height, in pixels
	 * @see   #getMarginHeight()
	 */
	public void setMarginHeight(int mh)
	{
		_margin_height = mh;
		repaint();
	}

	/**
	 * Gets the alignment policy of this component.
	 *
	 * @return the alignment policy
	 * @see    #setAlignment(int)
	 */
	public int getAlignment()
	{
		return _alignment;
	}

	/**
	 * Gets the horizontal margin for this component.
	 *
	 * @return the margin width, in pixels
	 * @see    #setMarginWidth(int)
	 */
	public int getMarginWidth()
	{
		return _margin_width;
	}

	/**
	 * Gets the vertical margin for this component.
	 *
	 * @return the margin height, in pixels
	 * @see    #setMarginHeight(int)
	 */
	public int getMarginHeight()
	{
		return _margin_height;
	}

	/**
	 * Invoked after our Canvas is first created but before it can actually be displayed. After this
	 * superclass' {@link JComponent#addNotify()}, this component font metrics is set and the {@link
	 * #measure()} method can be called to figure out how big the label is.
	 *
	 * @see java.awt.Component#addNotify()
	 */
	public void addNotify()
	{
		super.addNotify();
		measure();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#getPreferredSize()
	 */
	public Dimension getPreferredSize()
	{
		int w, h;

		// Auto calculation or specified ?
		if (_nCols < 0)
			w = _max_width + (2 * _margin_width);
		else
		{
			// Ensure we know the char width
			if (_charWidth < 0)
			{
				FontMetrics fm = getFontMetrics(getFont());

				_charWidth = computeCharWidth(fm);
			}
			w = (_nCols * _charWidth) + (2 * _margin_width);
		}

		int rows;

		if (_nRows > 0)
			rows = _nRows;
		else if (_num_lines > 0)
			rows = _num_lines;
		else
			rows = 1;

		h = (rows * _line_height) + (2 * _margin_height);
		
		if (_minimumWidth > 0)
			w = Math.max(w, _minimumWidth);
		
		return new Dimension(w, h);
	} // end method getPreferredSize

	/**
	 * Set the minimum width of this component.
	 * 
	 * @param minimumWidth the new minimum width of the component (a negative value means that
	 * there is no minimum width).
	 */
	public void setMinimumWidth(int minimumWidth)
	{
		_minimumWidth = minimumWidth;
	}
	
	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#getMinimumSize()
	 */
	public Dimension getMinimumSize()
	{
		return getPreferredSize();
	}
	
	/**
	 * Returns the number of lines used to display label.
	 */
	public int getNumLines()
	{
		return _num_lines;
	}

	/**
	 * Draws the label (applets use the same method). The margins and alignment are handled here,
	 * and the color and fonts are set by the superclass in the Graphics context object passed.
	 *
	 * @see javax.swing.JComponent#paintComponent(java.awt.Graphics)
	 */
	public void paintComponent(Graphics g)
	{
		super.paintComponent(g);

		String aaText = System.getProperty("swing.aatext");

		// set antialias
		if ("true".equalsIgnoreCase(aaText) && (g instanceof Graphics2D))
			((Graphics2D)g).setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
				RenderingHints.VALUE_TEXT_ANTIALIAS_ON);

		int		  x, y;
		Dimension d = getSize();

		y = _line_ascent + ((d.height - (_num_lines * _line_height)) / 2);

		g.setFont(getFont());

		for (int i = 0; i < _num_lines; i++, y += _line_height)
		{
			int width = _line_widths[i];

			if (_bigFont != null)
			{
				// Compute offset
				FontMetrics fm = getFontMetrics(getFont());

				width = fm.stringWidth(_lines.elementAt(i));
			}

			switch (_alignment)
			{
			case LEFT:
				x = _margin_width;
				break;
			case CENTER:
			default:
				x = (d.width - width) / 2;
				break;
			case RIGHT:
				x = d.width - _margin_width - width;
				break;
			}
			g.drawString(_lines.elementAt(i), x, y);
		}
	} // end method paintComponent

	/**
	 * Sets the number of columns, or -1 if automatic calculation.
	 *
	 * @param nCols the number of columns
	 */
	public void setNCols(int nCols)
	{
		_nCols = nCols;
	}

	/**
	 * Sets the number of rows, or -1 if automatic calculation.
	 *
	 * @param nRows the number of rows
	 */
	public void setNRows(int nRows)
	{
		_nRows = nRows;
	}
	
	/**
	 * Sets the maximum number of characters per line
	 * @param maxWidth the maximum number of characters per line
	 */
	public void setMaxWidth(int maxWidth)
	{
		_maxWidth = maxWidth;
		newLabel(getLabel());
		measure();
		repaint();
	}

	/**
	 * Computes the average character width for the string "0123456789" for the specified font
	 * metrics.
	 *
	 * @param  fm the font metrics to compute its character width
	 * @return the width of the characters in this metrics
	 */
	protected int computeCharWidth(FontMetrics fm)
	{
		return (fm.stringWidth("0123456789") / 10);
	}

} // end class LyJMultiLineLabel
