/*
 * LyJMultiSelector.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.dnd.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.border.*;

import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * This class defines a component containing two lists and a set of buttons. It allows to choose
 * some elements of the first list and to put them in the second one.
 *
 * @author  Lyria
 */
public class LyJMultiSelector extends JPanel
	implements LySelectionListener,
		LyActionListener,
		LyEventTransmitter
{
	private static final long     serialVersionUID = -3181497546768462525L;

	public static final String    __VERSION = "$Id: LyJMultiSelector.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The left list. */
	protected LyJListComponent    _list1;

	/** The right list. */
	protected LyJListComponent    _list2;

	/** The center panel that contains the action buttons. */
	protected JPanel			  _centerButtonPanel;

	/** The right panel that contains the action buttons. */
	protected JPanel			  _rightButtonPanel;

	/** The button that selects all elements to the right list. */
	protected LyJButton			  _buttonRightAll;

	/** The button that selects the current element to the right list. */
	protected LyJButton			  _buttonRight;

	/** The button that unselects the current element and returns it to the left list. */
	protected LyJButton			  _buttonLeft;

	/** The button that unselects all elements to the left list. */
	protected LyJButton			  _buttonLeftAll;

	/** The button that moves elements from the right list to an upper position. */
	protected LyJButton			  _buttonUp;

	/** The button that moves elements from the right list to a lower position. */
	protected LyJButton			  _buttonDown;

	/** The minimum number of elements value. */
	protected int				  _nmin;

	/** The maximum number of elements value. */
	protected int				  _nmax;

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter _actionTransmitter;
	
	/** Idle mode on the component. */
	protected boolean			  _idle;
	
	/** List that last sent a drag event. */
	protected LyJListComponent	  _dragList;

	/** Indicates if the left list has a header label **/ 
	protected boolean			  _hasLeftHeader = false;

	/** Indicates if the right list has a header label **/
	protected boolean			  _hasRightHeader = false;

	/**
	 * Creates a multiselector object with the specified attributes.
	 *
	 * @param environment the current environment
	 * @param height      the height of this component
	 */
	public LyJMultiSelector(LyEnvironment environment, int height)
	{
		this(environment, height, true);
	}
	
	/**
	 * Creates a multiselector object with the specified attributes.
	 *
	 * @param environment the current environment
	 * @param height      the height of this component
	 * @param ordered     indicates whether this selector is ordered or not
	 */
	public LyJMultiSelector(LyEnvironment environment, int height, boolean ordered)
	{
		super();

		init(environment, height, ordered);
	}

	/**
	 * Initializes this component. Sets the list components and the associated buttons.
	 *
	 * @param environment the current environment
	 * @param height      the height of this component
	 */
	protected void init(LyEnvironment environment, int height)
	{
		init(environment, height, true);
	}

	/**
	 * Initializes this component. Sets the list components and the associated buttons.
	 *
	 * @param environment the current environment
	 * @param height      the height of this component
	 * @param ordered     indicates whether this selector is ordered or not
	 */
	protected void init(LyEnvironment environment, int height, boolean ordered)
	{
		// FEV 0058 : Try to get default sizes.
		int    defaultWidth, defaultHeight, minWidth, fixedWidth;

		String s;

		s = environment.getResource("LY_MULTIPLE_RELATION_WIDTH");
		defaultWidth = LyTools.intFromString(s);

		if (defaultWidth > 0)
			minWidth = defaultWidth;
		else
		{
			s = environment.getResource("LY_MULTIPLE_RELATION_MIN_WIDTH");
			minWidth = LyTools.intFromString(s);
		}

		if (height <= 0)
		{
			s = environment.getResource("LY_MULTIPLE_RELATION_HEIGHT");
			defaultHeight = LyTools.intFromString(s);
		}
		else
			defaultHeight = height;

		if (minWidth <= 0)
			minWidth = 20;

		if (defaultHeight <= 0)
			defaultHeight = 6;

		// Calculate estimation of size in pixels
		FontMetrics fm = getFontMetrics(getFont());

		fixedWidth = fm.charWidth('0') * minWidth;

		// Create the lists
		_list1 = new LyJListComponent();

		_list1.setCellRenderer(new _LyJCellRenderer());

		_list2 = new LyJListComponent();

		_list2.setCellRenderer(new _LyJCellRenderer());

		_list1.setFixedCellWidth(fixedWidth);
		_list1.setVisibleRowCount(defaultHeight);
		_list1.setBackground(Color.white);
		_list2.setFixedCellWidth(fixedWidth);
		_list2.setVisibleRowCount(defaultHeight);
		_list2.setBackground(Color.white);

		Color selectColor = (Color)environment.getColor("LY_TABLE_SELECT_COLOR");

		if (selectColor != null)
		{
			_list1.setSelectionBackground(selectColor);
			_list2.setSelectionBackground(selectColor);

			// if selection background too dark, selection foreground is white
			int sumColor = selectColor.getRed() + selectColor.getGreen() + selectColor.getBlue();

			if (sumColor < 255)
			{
				_list1.setSelectionForeground(Color.white);
				_list2.setSelectionForeground(Color.white);
			}
		}

		// COMMENTED in view package
		LyActionTransmitter actionTransmitter = _list1.getActionTransmitter();

		if (actionTransmitter != null)
			actionTransmitter.addActionListener(this);

		actionTransmitter = _list2.getActionTransmitter();
		if (actionTransmitter != null)
			actionTransmitter.addActionListener(this);

		LySelectionTransmitter selectionTransmitter = _list1.getSelectionTransmitter();

		if (selectionTransmitter != null)
			selectionTransmitter.addSelectionListener(this);

		selectionTransmitter = _list2.getSelectionTransmitter();

		if (selectionTransmitter != null)
			selectionTransmitter.addSelectionListener(this);

		// Initialize min and max
		_nmin = -1;
		_nmax = -1;

		// Create the buttons panels
		_centerButtonPanel = new JPanel();
		_rightButtonPanel = new JPanel();

		_buttonRightAll = new LyJButton((Image)environment.getImage("LY_RIGHT_ALL"), null,
				environment.translate("LY_SELECT_ALL"), "right_all");
		_buttonRight = new LyJButton((Image)environment.getImage("LY_RIGHT"), null,
				environment.translate("LY_SELECT"), "right");
		_buttonLeft = new LyJButton((Image)environment.getImage("LY_LEFT"), null,
				environment.translate("LY_UNSELECT"), "left_all");
		_buttonLeftAll = new LyJButton((Image)environment.getImage("LY_LEFT_ALL"), null,
				environment.translate("LY_UNSELECT_ALL"), "left_all");
		
		if (ordered)
		{
			_buttonUp = new LyJButton((Image)environment.getImage("LY_TOOL_MOVE_UP"), null,
					environment.translate("LY_TOOL_MOVE_UP"), "up");
			_buttonDown = new LyJButton((Image)environment.getImage("LY_TOOL_MOVE_DOWN"), null,
					environment.translate("LY_TOOL_MOVE_DOWN"), "down");
		}
		
		_buttonRightAll.setContentAreaFilled(false);
		_buttonRight.setContentAreaFilled(false);
		_buttonLeft.setContentAreaFilled(false);
		_buttonLeftAll.setContentAreaFilled(false);

		if (ordered)
		{
			_buttonUp.setContentAreaFilled(false);
			_buttonDown.setContentAreaFilled(false);
		}
		
		_buttonRightAll.setBackground(null);
		_buttonRight.setBackground(null);
		_buttonLeft.setBackground(null);
		_buttonLeftAll.setBackground(null);

		if (ordered)
		{
			_buttonUp.setBackground(null);
			_buttonDown.setBackground(null);
		}
		
		Insets buttonInsets = new Insets(5, 15, 5, 15);
		Border border = new EmptyBorder(buttonInsets);

		_buttonRightAll.setBorder(border);
		_buttonRight.setBorder(border);
		_buttonLeft.setBorder(border);
		_buttonLeftAll.setBorder(border);

		if (ordered)
		{
			_buttonUp.setBorder(border);
			_buttonDown.setBorder(border);
		}
		
		_buttonRightAll.setRolloverChangeBackground(true);
		_buttonRight.setRolloverChangeBackground(true);
		_buttonLeft.setRolloverChangeBackground(true);
		_buttonLeftAll.setRolloverChangeBackground(true);

		if (ordered)
		{
			_buttonUp.setRolloverChangeBackground(true);
			_buttonDown.setRolloverChangeBackground(true);
		}
		
		LyActionTransmitter transmitter;

		transmitter = _buttonRightAll.getActionTransmitter();
		if (transmitter != null)
			transmitter.addActionListener(this);

		transmitter = _buttonRight.getActionTransmitter();
		if (transmitter != null)
			transmitter.addActionListener(this);

		transmitter = _buttonLeft.getActionTransmitter();
		if (transmitter != null)
			transmitter.addActionListener(this);

		transmitter = _buttonLeftAll.getActionTransmitter();
		if (transmitter != null)
			transmitter.addActionListener(this);

		if (ordered)
		{
			transmitter = _buttonUp.getActionTransmitter();
			if (transmitter != null)
				transmitter.addActionListener(this);
	
			transmitter = _buttonDown.getActionTransmitter();
			if (transmitter != null)
				transmitter.addActionListener(this);
		}
		
		// Add the buttons to the panel with a box layout
		_centerButtonPanel.setLayout(new BoxLayout(_centerButtonPanel, BoxLayout.Y_AXIS));
		_centerButtonPanel.setBackground(null);

		_centerButtonPanel.add(_buttonRightAll);
		_centerButtonPanel.add(_buttonRight);
		_centerButtonPanel.add(_buttonLeft);
		_centerButtonPanel.add(_buttonLeftAll);

		_centerButtonPanel.setOpaque(false);
		
		_rightButtonPanel.setLayout(new BoxLayout(_rightButtonPanel, BoxLayout.Y_AXIS));
		_rightButtonPanel.setBackground(null);

		if (ordered)
		{
			_rightButtonPanel.add(_buttonUp);
			_rightButtonPanel.add(_buttonDown);
		}

		_rightButtonPanel.setOpaque(false);
		
		setOpaque(false);
		setLayout(new GridBagLayout());
		
		GridBagConstraints	gbc = new GridBagConstraints();
		
		// Add list headers if needed
		String leftHeader = environment.translate("LY_MULTISELECTOR_LEFT_HEADER");
		String rightHeader = environment.translate("LY_MULTISELECTOR_RIGHT_HEADER");
		Font   font = (Font)environment.getFont("LY_DEFAULT_FONT");

		if ((leftHeader != null) && (leftHeader.length() > 0))
		{
			JPanel leftPanel = new JPanel(new BorderLayout());
			JLabel label = new JLabel(leftHeader);

			border = new EtchedBorder();
			label.setBorder(border);
			label.setFont(font);
			leftPanel.add(label, BorderLayout.NORTH);
			leftPanel.add(new _LyJSpecialScrollPane(_list1), BorderLayout.CENTER);

			add(leftPanel, gbc);
			_hasLeftHeader = true;
		}
		else
			add(new _LyJSpecialScrollPane(_list1), gbc);

		add(_centerButtonPanel, gbc);

		if ((rightHeader != null) && (rightHeader.length() > 0))
		{
			JPanel rightPanel = new JPanel(new BorderLayout());
			JLabel label = new JLabel(rightHeader);

			border = new EtchedBorder();
			label.setBorder(border);
			label.setFont(font);
			rightPanel.add(label, BorderLayout.NORTH);
			rightPanel.add(new _LyJSpecialScrollPane(_list2), BorderLayout.CENTER);

			add(rightPanel, gbc);
			_hasRightHeader = true;
		}
		else
			add(new _LyJSpecialScrollPane(_list2), gbc);

		gbc.anchor = GridBagConstraints.NORTH;
		add(_rightButtonPanel, gbc);

		addKeyListener(new _LyJMultiSelectorKeyListener());
		_idle = false;

		// Manage drag and drop on the two lists
		_list1.setDragEnabled(true);
		_list2.setDragEnabled(true);
		
		DragSource	dragSource = new DragSource();
		
		dragSource.createDefaultDragGestureRecognizer(_list1, DnDConstants.ACTION_COPY_OR_MOVE,
				new _LyJMultiSelectorDragListener());
		dragSource.createDefaultDragGestureRecognizer(_list2, DnDConstants.ACTION_COPY_OR_MOVE,
				new _LyJMultiSelectorDragListener());
		new DropTarget(_list1, new _LyJMultiSelectorDropListener());
		new DropTarget(_list2, new _LyJMultiSelectorDropListener());
		
		((_LyJSpecialScrollPane)_list1.getParent().getParent()).setFixedWidth(fixedWidth);
		((_LyJSpecialScrollPane)_list2.getParent().getParent()).setFixedWidth(fixedWidth);
		
		// Update the buttons
		checkButtons();
	} // end method init

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#getMinimumSize()
	 */
	public Dimension getMinimumSize()
	{
		return getPreferredSize();
	}

	/**
	 * Gets the maximum size between the two lists. Used to ensure that the 2 lists have the same
	 * size.
	 *
	 * @return the list preferred size
	 */
	public Dimension getListPreferredSize()
	{
		Dimension size1 = ((_LyJSpecialScrollPane)_list1.getParent().getParent())
				.getSuperPreferredSize();
		Dimension size2 = ((_LyJSpecialScrollPane)_list2.getParent().getParent())
				.getSuperPreferredSize();

		return new Dimension(Math.max(size1.width, size2.width),
			Math.max(size1.height, size2.height));
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#getPreferredSize()
	 */
	public Dimension getPreferredSize()
	{
		Dimension list1Size;
		if (_hasLeftHeader)
			list1Size= _list1.getParent().getParent().getParent().getPreferredSize();
		else
			list1Size= _list1.getParent().getParent().getPreferredSize();
		Dimension list2Size;
		if (_hasRightHeader)
			list2Size = _list2.getParent().getParent().getParent().getPreferredSize();
		else
			list2Size = _list2.getParent().getParent().getPreferredSize();
		Dimension centerButtonsSize = _centerButtonPanel.getPreferredSize();
		Dimension rightButtonsSize = _rightButtonPanel.getPreferredSize();

		int		  width = list1Size.width + centerButtonsSize.width + list2Size.width + rightButtonsSize.width;
		int		  height = Math.max(list1Size.height,
				Math.max(centerButtonsSize.height, list2Size.height));

		return new Dimension(width, height);
	}

	/**
	 * Updates the buttons whether they are enabled or not.
	 */
	public void checkButtons()
	{
		if (_idle)
			return;
		
		if (isEnabled())
		{
			_buttonRightAll.setEnabled((_list1.getCount() > 0)
					&& ((_nmax < 0) || ((_list1.getCount() + _list2.getCount()) <= _nmax)));

			_buttonLeftAll.setEnabled((_list2.getCount() > 0) && (_nmin <= 0));

			_buttonRight.setEnabled((_list1.getSelectedCount() > 0)
					&& ((_nmax < 0) || ((_list2.getCount() + _list1.getSelectedCount()) <= _nmax)));

			_buttonLeft.setEnabled((_list2.getSelectedCount() > 0)
					&& ((_nmin < 0) || ((_list2.getCount() - _list2.getSelectedCount()) >= _nmin)));

			if (_buttonUp != null)
				_buttonUp.setEnabled((_list2.getSelectedCount() > 0) && (_list2.getSelectedIndex() > 0));

			int[]	selection = _list2.getSelectedIndices();
			
			if (_buttonDown != null)
				_buttonDown.setEnabled((_list2.getSelectedCount() > 0)
						&& (selection != null) && (selection[selection.length - 1] < _list2.getCount() - 1));
		}
		else
		{
			_buttonRightAll.setEnabled(false);
			_buttonLeftAll.setEnabled(false);
			_buttonRight.setEnabled(false);
			_buttonLeft.setEnabled(false);
		}
	}

	/**
	 * Sets the content of this selector. The left list is set with the content of the
	 * unSelectedLinks argument and the right list is set with the selectedLinks content.
	 *
	 * @param unSelectedLinks list of unselected links for the left list
	 * @param selectedLinks   lift of selected links for the right list
	 */
	public void set(Vector<LyLink> unSelectedLinks, Vector<LyLink> selectedLinks)
	{
		_list1.setItems(new Vector<Object>(unSelectedLinks));
		_list2.setItems(new Vector<Object>(selectedLinks));
		checkButtons();
	}

	/**
	 * Adds the specified link to the right list and removes it from the left list.
	 *
	 * @param link the link to select
	 */
	public void select(LyLink link)
	{
		if (link == null)
			return;

		int idx = _list1.indexOf(link);

		if (idx != -1)
		{
			if ((_nmax < 0) || (_list2.getCount() < _nmax))
			{
				_list1.removeItemAt(idx);
				_list2.addItem(link);
				checkButtons();
			}
			else
				_list1.setSelectedValue(link, false);
		}
	}

	/**
	 * Removes the specified link from the right list and adds it to the left list.
	 *
	 * @param link the link to unselect
	 */
	public void unselect(LyLink link)
	{
		if (link == null)
			return;

		int idx = _list2.indexOf(link);

		if (idx != -1)
		{
			if ((_nmin < 0) || (_list2.getCount() > _nmin))
			{
				_list2.removeItemAt(idx);
				_list1.addItem(link);
				checkButtons();
			}
			else
				_list2.setSelectedValue(link, false);
		}
	}

	/**
	 * Adds the specified link at the specified position. If the link is selected it is added to the
	 * right list, else to the left one.
	 *
	 * @param link  the link to add
	 * @param index the position of the link to add
	 */
	public void addItemAt(LyLink link, int index)
	{
		if (link == null)
			return;

		boolean			 selected = link.isSelected();

		LyJListComponent list = selected ? _list2 : _list1;

		if (selected)
			index -= _list1.getCount();

		if (index < 0)
			list.addItem(link);
		else
			list.insertItemAt(link, index);

		checkButtons();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LySelectionListener#selectionPerformed(leon.view.event.LySelectionEvent)
	 */
	public void selectionPerformed(LySelectionEvent selectionEvent)
	{
		checkButtons();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		LyJListComponent list_from = null, list_to = null;
		boolean			 all = false;
		Object			 source = actionEvent.getSource();
		boolean			 valueChanged = false;

		if ((source == _buttonRight) || (source == _buttonRightAll))
		{
			list_from = _list1;
			list_to = _list2;
			valueChanged = true;
		}
		else if ((source == _buttonLeft) || (source == _buttonLeftAll))
		{
			list_from = _list2;
			list_to = _list1;
			valueChanged = true;
		}
		else if ((source == _list1) || (source == _list2))
		{
			// notify
			if ((_actionTransmitter != null)
					&& (actionEvent.getType() == LyActionDemand.TYPE_ACTION))
			{
				if (actionEvent.getValue() == null)
				{
					// Fetch selected object name
					String value =
						((LyLink)((LyJListComponent)source).getItemAt(actionEvent.getIndex(0)))
							.getId();

					actionEvent.setValue(value);
				}
				_actionTransmitter.processActionEvent(actionEvent);
			}
			return;
		}
		else if ((source == _buttonUp) || (source == _buttonDown))
		{
			LyLink	elem;
			int[]	selection = _list2.getSelectedIndices();
			
			_idle = true;
			
			if (source == _buttonUp)
			{
				for (int i = 0; i < selection.length; i++)
				{
					elem = ((LyLink)_list2.getItemAt(selection[i]));
					_list2.removeItemAt(selection[i]);
					_list2.insertItemAt(elem, selection[i] - 1);
					_list2.selectItem(elem, true);
				}
			}
			else
			{
				for (int i = selection.length - 1; i >= 0; i--)
				{
					elem = ((LyLink)_list2.getItemAt(selection[i]));
					_list2.removeItemAt(selection[i]);
					_list2.insertItemAt(elem, selection[i] + 1);
					_list2.selectItem(elem, true);
				}
			}
			
			valueChanged = true;
			_idle = false;
		}
		else
			return;

		if ((source == _buttonRightAll) || (source == _buttonLeftAll))
			all = true;

		if ((list_from != null) && (list_to != null))
			switchSelection(list_from, list_to, all, -1);

		if (valueChanged)
		{
			// notify
			if (_actionTransmitter != null)
			{
				LyActionEvent event = new LyActionEvent(this, null, getValue(),
						LyActionDemand.TYPE_VALUE_CHANGED);

				_actionTransmitter.processActionEvent(event);
			}
		}

		checkButtons();
	} // end method actionPerformed
	
	/**
	 * Switch elements from a list to the other one.
	 * 
	 * @param listFrom list from which elements should be removed.
	 * @param listTo list to which elements should be added.
	 * @param all indicates whether all elements or only selected elements must be moved.
	 * @param forcedIndex if positive, indicates the position where elements must be added in
	 * the target list.
	 */
	protected void switchSelection(LyJListComponent listFrom, LyJListComponent listTo, boolean all, int forcedIndex)
	{
		int    ns = (all ? listFrom.getCount() : listFrom.getSelectedCount());
		int    nt = listTo.getCount();
		int    i = 0, n = listFrom.getCount(), idx;
		LyLink elem, elem2;

		while ((ns > 0) && (i < n))
		{
			if (all || listFrom.isSelectedIndex(i))
			{
				elem = ((LyLink)listFrom.getItemAt(i));
				listFrom.removeItemAt(i);

				ns--;

				if (forcedIndex >= 0)
					idx = forcedIndex;
				else
				{
					if (listTo == _list2)
						idx = listTo.getMinSelectionIndex();
					else
					{
						// Try to retrieve original position
						for (idx = 0; idx < nt; idx++)
						{
							elem2 = ((LyLink)listTo.getItemAt(idx));
							if (elem.getLabel().compareTo(elem2.getLabel()) < 0)
								break;
						}
					}
				}

				if ((idx >= 0) && (idx < nt))
					listTo.insertItemAt(elem, idx);
				else
					listTo.addItem(elem);
				
				if (listTo == _list2)
					listTo.selectItem(elem, false);

				nt++;
			}
			else
				i++;
		}
	}
	
	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#setEnabled(boolean)
	 */
	public void setEnabled(boolean status)
	{
		super.setEnabled(status);

        _list1.setDragEnabled(status);
        _list2.setDragEnabled(status);
        
		checkButtons();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#setBackground(java.awt.Color)
	 */
	public void setBackground(Color color)
	{
		if (_list1 != null)
			_list1.setBackground(color);

		if (_list2 != null)
			_list2.setBackground(color);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#setFont(java.awt.Font)
	 */
	public void setFont(Font font)
	{
		super.setFont(font);

		if (_list1 != null)
			_list1.setFont(font);

		if (_list2 != null)
			_list2.setFont(font);
	}

	/**
	 * Returns the list associated with the given multiselector number. If no multiselector is
	 * defined for the specified number, then returns <code>null</code>.
	 *
	 * @param  i index of the listbox: 0 for the first one, 1 for the second one
	 * @return the corresponding listbox or <code>null</code> if not found
	 */
	public LyJListComponent getListComponent(int i)
	{
		if (i == 0)
			return _list1;
		if (i == 1)
			return _list2;
		return null;
	}

	/**
	 * Gets the list of selected elements' id.
	 *
	 * @return the list of selected element ids. These ids are separated by {@link
	 *         LyJListComponent#ITEM_SEP}.
	 */
	public String getValue()
	{
		StringBuffer sb = null;
		Vector<String>		 selectedIds = new Vector<String>();
		int			 i, n = _list2.getCount();

		for (i = 0; i < n; i++)
			selectedIds.addElement(((LyLink)_list2.getItemAt(i)).getId());

		n = selectedIds.size();

		for (i = 0; i < n; i++)
		{
			if (sb == null)
			{
				sb = new StringBuffer();
				sb.append(selectedIds.elementAt(i));
			}
			else
			{
				sb.append(LyJListComponent.ITEM_SEP);
				sb.append(selectedIds.elementAt(i));
			}
		}
		return ((sb == null) ? null : sb.toString());
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return (_actionTransmitter);
	}

	/**
	 * This multiselector doesn't handle selection events, so it has no selection transmitter.
	 *
	 * @return <code>null</code>
	 * @see    leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		return null;
	}

	/**
	 * Adds the specified key listener to both the selected and unselected lists.
	 *
	 * @see java.awt.Component#addKeyListener(java.awt.event.KeyListener)
	 */
	public void addKeyListener(KeyListener listener)
	{
		_list1.addKeyListener(listener);
		_list2.addKeyListener(listener);
	}

	/* ****************************************************************************************/

	/**
	 * This class implements the preferred size of its super component.
	 */
	public class _LyJSpecialScrollPane extends JScrollPane
	{
		private static final long serialVersionUID = -7857491982378794672L;

		/** The preferred size of this component. */
		protected Dimension		  _preferredSize;
		
		/** Fixed width of this component. */
		protected int			  _fixedWidth;

		/**
		 * Creates a new ScrollPane component.
		 *
		 * @param comp the new component
		 */
		public _LyJSpecialScrollPane(Component comp)
		{
			super(comp);
		}

		/**
		 * Gets the preferred size of the parent component if this component doesn't have one, or if
		 * it is too small.
		 *
		 * @return the preferred size of this component
		 */
		public Dimension getSuperPreferredSize()
		{
			if ((_preferredSize == null) || (_preferredSize.width < 10)
					|| (_preferredSize.height < 10))
				_preferredSize = super.getPreferredSize();
			return _preferredSize;
		}

		/**
		 * Get the minimum size for this component.
		 * This implementation returns the preferred size
		 */
		public Dimension getMinimumSize()
		{
			return super.getPreferredSize();
		}

		/**
		 * Sets the fixed width of this component.
		 * 
		 * @param width fixed width of this component.
		 */
		public void setFixedWidth(int width)
		{
			_fixedWidth = width;
		}
		
		/*
		 * (non-Javadoc)
		 * @see java.awt.Component#getPreferredSize()
		 */
		public Dimension getPreferredSize()
		{
			if (_fixedWidth > 0)
				return new Dimension(_fixedWidth + 40, getListPreferredSize().height);
			
			return getListPreferredSize();
		}
	} // end class _LyJSpecialScrollPane


	/* ********************************************************************************/

	/**
	 * This class implements a customized renderer for list cell.
	 */
	public class _LyJCellRenderer extends DefaultListCellRenderer
	{
		private static final long serialVersionUID = -8395920779745580968L;

		/** The text on the cell. */
		String					  label;

		/** The associated icon of a list cell. */
		Icon					  icon;

		/**
		 * Sole constructor.
		 */
		public _LyJCellRenderer()
		{
			super();
		}

		/*
		 * (non-Javadoc)
		 * @see javax.swing.ListCellRenderer#getListCellRendererComponent(javax.swing.JList,
		 * java.lang.Object, int, boolean, boolean)
		 */
		public Component getListCellRendererComponent(JList list, Object value, int index,
			boolean isSelected, boolean hasFocus)
		{
			JLabel		  comp = (JLabel)super.getListCellRendererComponent(list, value, index,
					isSelected, hasFocus);
			if (value == null)
				return comp;
			
			LyLink		  link = (LyLink)value;
			
			LyEnvironment environment = link.getEnvironment();

			if (link.getShortLabel() != null)
				label = environment.translate(link.getShortLabel());
			else
				label = environment.translate(link.getLabel());
			
			if (link.getImage() != null)
				icon = new ImageIcon((Image)environment.getImage(link.getImage()));
			else
				icon = null;
			
			comp.setText(label);
			comp.setIcon(icon);
			comp.setToolTipText(environment.translate(link.getLabel()));
			
			return comp;
		}
	} // end class _LyJCellRenderer


	/**
	 * Multiselector key listener.
	 */
	public class _LyJMultiSelectorKeyListener
		implements KeyListener
	{

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
		 */
		public void keyTyped(KeyEvent event)
		{
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
		 */
		public void keyPressed(KeyEvent event)
		{
			int key = event.getKeyCode();

			// If left or right arrow, simulate a clic on left or right button
			if (key == KeyEvent.VK_LEFT)
			{
				LyActionEvent actionEvent = new LyActionEvent(_buttonLeft);

				actionPerformed(actionEvent);
			}
			else if (key == KeyEvent.VK_RIGHT)
			{
				LyActionEvent actionEvent = new LyActionEvent(_buttonRight);

				actionPerformed(actionEvent);
			}
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
		 */
		public void keyReleased(KeyEvent event)
		{
		}
	} // end class _LyJMultiSelectorKeyListener

	/**
	 * Provides drap and drop handling.
	 */
	class _LyJMultiSelectorDragListener
		implements DragGestureListener
	{
		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DragGestureListener#dragGestureRecognized(java.awt.dnd.DragGestureEvent)
		 */
	     public void dragGestureRecognized(DragGestureEvent dge)
	     {
	    	 _dragList = (LyJListComponent)dge.getComponent();
	     }
	}

	/**
	 * Provides drap and drop handling.
	 */
	class _LyJMultiSelectorDropListener
		implements DropTargetListener
	{

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#dragEnter(java.awt.dnd.DropTargetDragEvent)
		 */
		public void dragEnter(DropTargetDragEvent dtde)
		{
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#dragExit(java.awt.dnd.DropTargetEvent)
		 */
		public void dragExit(DropTargetEvent dte)
		{
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#dragOver(java.awt.dnd.DropTargetDragEvent)
		 */
		public void dragOver(DropTargetDragEvent dtde)
		{
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#drop(java.awt.dnd.DropTargetDropEvent)
		 */
		public void drop(DropTargetDropEvent dtde)
		{
            if (!isEnabled())
                return;
            
			DropTarget			dropTarget = (DropTarget)dtde.getSource();
			LyJListComponent	dropList = (LyJListComponent)dropTarget.getComponent();
			int					index = dropList.locationToIndex(dtde.getLocation());
			
			dtde.acceptDrop(DnDConstants.ACTION_COPY_OR_MOVE);
			
			if (_dragList != dropList)
			{
				switchSelection(_dragList, dropList, false, index);
			}
			else
			{
				int[]	selection = dropList.getSelectedIndices();
				LyLink	elem;
				Vector<LyLink>	elems = new Vector<LyLink>();
				
				// Allow drop only if drop index is outside current selection
				if ((index >= selection[0]) && (index <= selection[selection.length - 1]))
				{
					dtde.getDropTargetContext().dropComplete(true);
					return;
				}

				dtde.getDropTargetContext().dropComplete(true);
				
				// First remove items from their original positions
				for (int i = selection.length - 1; i >= 0; i--)
				{
					elem = (LyLink)dropList.getItemAt(selection[i]);
					dropList.removeItemAt(selection[i]);
					elems.addElement(elem);
				}
				
				boolean		before = (index < selection[0]);
				
				// Then add items to their target positions
				for (int i = elems.size() - 1; i >= 0; i--)
				{
					elem = elems.elementAt(i);
					
					int		pos = index - i;
					
					if (before)
						pos = index + (elems.size() - 1) - i;
					
					dropList.insertItemAt(elem, pos);
					dropList.selectItem(elem, true);
				}
			}
			
			// notify
			if (_actionTransmitter != null)
			{
				LyActionEvent event = new LyActionEvent(this, null, getValue(),
						LyActionDemand.TYPE_VALUE_CHANGED);

				_actionTransmitter.processActionEvent(event);
			}

			checkButtons();
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#dropActionChanged(java.awt.dnd.DropTargetDragEvent)
		 */
		public void dropActionChanged(DropTargetDragEvent dtde)
		{
		}
	} // end class _LyJMultiSelectorDropListener
} // end class LyJMultiSelector
