/*
 * LyJNumberSelector.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;
import java.text.*;

import javax.swing.*;
import javax.swing.text.*;
import leon.misc.*;
import leon.view.event.*;
import leon.view.win.swing.text.*;


/**
 * This class defines a component containing a text field and two buttons. It allows to select a
 * number which may be an integer or a float.
 *
 * @version $Id: LyJNumberSelector.java 34860 2012-03-19 14:34:58Z jlsaouli $
 * @author  Lyria
 * @see     LyJSelector
 */
public class LyJNumberSelector extends LyJSelector
{
	private static final long  serialVersionUID = -4426355989070414135L;

	public static final String __VERSION =
		"$Id: LyJNumberSelector.java 34860 2012-03-19 14:34:58Z jlsaouli $";

	/** Image width of the selector button. */
	public static int		   IMAGE_W_OFFSET = 6;

	/** Image height of the selector button. */
	public static int		   IMAGE_H_OFFSET = 6;

	/** Increasing button. */
	protected LyJButton		   _up;

	/** Decreasing button. */
	protected LyJButton		   _down;

	/** Value of this number component. */
	protected double		   _value;

	/** Minimum value of this button. */
	protected double		   _min;

	/** Maximum value of this button. */
	protected double		   _max;

	/** Increment and decrement unit value. */
	protected double		   _pitch;

	/** Number of decimals for this component. */
	protected int			   _decimals;

	/** Units showed for this component. */
	protected String		   _unit;

	/** Indicates that the value must be shown after initialization. */
	protected boolean		   _showInitValue;

	/** Keeps the values before modifying them. */
	protected String		   _oldString;

	/** Decimal character separator. */
	protected char			   _decimalSeparator;
	
	/** Group character separator. */
	protected char			   _groupSeparator;

	/** Indicates that this number can have any value under the default one. */
	protected boolean		   _noMin;

	/** Indicates that this number can have any value beyond the default one. */
	protected boolean		   _noMax;

	/**
	 * Creates and initializes a number selector for the specified parameters.
	 *
	 * @param value         the default value
	 * @param min           the minimum value
	 * @param max           the maximum value
	 * @param pitch         the increment and decrement unit value
	 * @param decimals      the number of decimals
	 * @param unit          the unit used
	 * @param showInitValue Indicates that the specified value is show after creation
	 * @param env           the current environment
	 */
	public LyJNumberSelector(double value, double min, double max, double pitch, int decimals,
		String unit, boolean showInitValue, LyEnvironment env, boolean useArrows)
	{
		super(null, env, useArrows);
		_text.setDocument(new LyJNumberDocument(0, min, max, _decimalSeparator, decimals)
			{
				private static final long serialVersionUID = -5425947719261980563L;

				/*
				 * (non-Javadoc)
				 * @see javax.swing.text.Document#insertString(int, java.lang.String,
				 * javax.swing.text.AttributeSet)
				 */
				public void insertString(int offset, String str, AttributeSet a)
					throws BadLocationException
				{
					super.insertString(offset, str, a);
					LyJNumberSelector.this.setValue(getValue());
					LyJNumberSelector.this.checkButtons();
				}

				/*
				 * (non-Javadoc)
				 * @see javax.swing.text.Document#remove(int, int)
				 */
				public void remove(int offset, int length)
					throws BadLocationException
				{
					super.remove(offset, length);
					LyJNumberSelector.this.setValue(getValue());
					LyJNumberSelector.this.checkButtons();
				}

			});

		_noMin = false;
		_noMax = false;

		_value = value;
		_min = min;
		_max = max;

		if (_min == Double.NEGATIVE_INFINITY)
			_noMin = true;

		if (_max == Double.POSITIVE_INFINITY)
			_noMax = true;

		_pitch = LyTools.getPitch(pitch, decimals);
		_decimals = decimals;
		_showInitValue = showInitValue;
		_unit = unit;

		endInit(env);
	}

	/**
	 * Finishes the initialization of this component.
	 *
	 * @param environment the current environment
	 */
	public void endInit(LyEnvironment environment)
	{
		_unit = environment.translate(_unit);

		DecimalFormatSymbols format = new DecimalFormatSymbols(environment.getLocale());

		_decimalSeparator = format.getDecimalSeparator();
		_groupSeparator = format.getGroupingSeparator();
		((LyJNumberDocument)_text.getDocument()).setDecimalSeparator(_decimalSeparator);

		if (_useArrows)
		{
			_up = addButton((Image)environment.getImage("LY_ARROW_UP"), null);
			_down = addButton((Image)environment.getImage("LY_ARROW_DOWN"), null);
		}
		if ((_unit != null) && (_unit.length() != 0))
		{
			JLabel label = new JLabel("  " + _unit);

			// FFT01121
			label.setFont((Font)environment.getFont("LY_DEFAULT_FONT"));
			label.setForeground((Color)environment.getColor("LY_LABEL_NORMAL_STATE"));
			setComponent(label);
		}

		// Compute the content
		resetValue(environment);

		if (_showInitValue)
			setValue(_value);

	} // end method endInit

	/**
	 * Computes the length of the field depending on minimum and maximum values.
	 *
	 * @return the requested size (length) of the field
	 */
	protected int computeSize()
	{
		int    size = 0;
		double var = 1;
		int    i = 0;

		if (!_noMax && !_noMin)
		{
			// Get the max between _min and _max (abs values)
			double max = Math.max(Math.abs(_max), Math.abs(_min));

			if (max > Long.MAX_VALUE)
				max = Long.MAX_VALUE;

			for (i = 0; var <= max; i++)
				var *= 10;

			size = i + 1 + ((_decimals == 0) ? 0 : ((_decimals < 0) ? 10 : (_decimals + 1)));
		}
		else if (_decimals == 0)
			size = 10;
		else
			size = 20;

		return size;
	}

	/**
	 * Creates a new image button with the specified tip message.
	 *
	 * @param  image   the image for the button
	 * @param  tooltip the tip message being shown
	 * @return a reference to the button.
	 * @see    leon.view.win.swing.component.LyJSelector#addButton(java.awt.Image, java.lang.String)
	 */
	public LyJButton addButton(Image image, String tooltip)
	{
		LyJButton button = new LyJButton(image, tooltip)
			{
				private static final long serialVersionUID = -1457219099433880361L;

				/**
				 * This button cannot gain focus.
				 *
				 * @return <code>false</code>
				 * @see    java.awt.Component#isFocusTraversable()
				 */
				public boolean isFocusTraversable()
				{
					return false;
				}

				/*
				 * (non-Javadoc)
				 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
				 */
				public void mousePressed(MouseEvent e)
				{
					_text.requestFocus();
					super.mousePressed(e);
				}
			};

		button.setRequestFocusEnabled(false);
		button.setOpaque(true);
		button.setRollover(false);
		button.setContentAreaFilled(false);

		Icon	icon = new ImageIcon(image);
		int		topBorder = 1;
		
		if (_buttons.getComponentCount() > 0)
			topBorder = 0;

		button.setIcon(icon);
		button.setPreferredSize(new Dimension(icon.getIconWidth() + IMAGE_W_OFFSET,
				icon.getIconHeight()
				+ IMAGE_H_OFFSET + topBorder));
		button.setHorizontalAlignment(SwingConstants.CENTER);
		button.setVerticalAlignment(SwingConstants.CENTER);
		button.setMargin(new Insets(0, 0, 0, 0));
		button.setRepeatEnabled(true);
		button.setEnvironment(_env);

		LyActionTransmitter transmitter = button.getActionTransmitter();

		if (transmitter != null)
			transmitter.addActionListener(this);
		
		button.setBorder(BorderFactory.createMatteBorder(topBorder, 1, 1, 1,
				UIManager.getColor("TextField.darkShadow")));

		_buttons.add(button, buildButtonConstraint());

		return button;
	} // end method addButton

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJSelector#checkButtons()
	 */
	protected void checkButtons()
	{
		if (!_useArrows)
			return;

		if (!isEnabled() || !isEditable())
		{
			_up.setEnabled(false);
			_down.setEnabled(false);
			return;
		}

		if (_noMax)
			_up.setEnabled(true);
		else
		{
			if (_value >= _max)
				_up.setEnabled(false);
			else
				_up.setEnabled(true);
		}

		if (_noMin)
			_down.setEnabled(true);
		else
		{
			if (_value <= _min)
				_down.setEnabled(false);
			else
				_down.setEnabled(true);
		}
	} // end method checkButtons

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		if (actionEvent.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
		{
			if (_actionTransmitter != null)
				_actionTransmitter.processActionEvent(new LyActionEvent(this, null,
					getValue(), LyActionDemand.TYPE_VALUE_CHANGED));
			
			return;
		}
		
		if (actionEvent.getType() == LyActionDemand.TYPE_TIMEOUT)
			setIdle(true);

		if (actionEvent.getSource() == _up)
		{
			// if we didn't had focus, store the last value
			_lastValue = getValue();
			// set the new value
			addPitch(true);
		}
		else if (actionEvent.getSource() == _down)
		{
			// if we didn't had focus, store the last value
			_lastValue = getValue();
			// set the new value
			addPitch(false);
		}
		else
			return;

		if (actionEvent.getType() == LyActionDemand.TYPE_TIMEOUT)
			setIdle(false);
	}

	/**
	 * Increases or decreases the component's value by the pitch.
	 *
	 * @param add Indicates if the pitch should be added or removed.
	 */
	protected void addPitch(boolean add)
	{
		String value = getValue();

		if ((value == null) || (value.length() == 0))
		{
			// first time we press a bouton,
			// if 0 is in interval, it's the first value,
			if ((0 <= _max) && (0 >= _min))
				setValue(0.0);
			else if (Math.abs(_max) < Math.abs(_min))
				setValue(_max);
			else
				setValue(_min);
			return;
		}

		double new_value;

		if (_decimals == 0)
		{
			long old_value = LyTools.longFromString(value);

			new_value = (add ? (old_value + _pitch) : (old_value - _pitch));
		}
		else
		{
			// Workaround for java bug : values are multiplicated by precision
			// to avoid errors on operations
			int i, precision = 1;
			int decimals = ((_decimals <= 0) ? 6 : _decimals);

			for (i = 0; i < decimals; i++)
				precision = precision * 10;

			double old_value = LyTools.doubleFromString(value), val;

			if (add)
				val = old_value + _pitch;
			else
				val = old_value - _pitch;
			val = Math.round(val * precision);
			new_value = val / precision;
		}

		setValue(new_value);
	} // end method addPitch

	/**
	 * Sets this component's new value. If the value is greater than the maximum allowed, then sets
	 * the value to the max value. Same for the minimum value.
	 *
	 * @param value the new value
	 * @see   #getValue()
	 */
	protected void setValue(double value)
	{
		if ((value >= _max) && !_noMax)
		{
			value = _max;
			setIdle(false); // button will be disabled NOTIFY
		}
		else if ((value <= _min) && !_noMin) // FFT00083
		{
			value = _min;
			setIdle(false); // button will be disabled NOTIFY
		}

		String new_value;

		if (_decimals == 0)
			new_value = String.valueOf((long)value);
		else
		{
			new_value = String.valueOf(value);
			new_value = LyTools.checkDecimals(new_value, _decimals);
		}
		
		if (new_value != null)
			_value = LyTools.doubleFromString(new_value);
		
	    if (_decimals != 0)
    		// store the old string
    		new_value = new_value.replace('.', _decimalSeparator);

		// FFT00558
		setValue(new_value);

		_oldString = new_value;

		// change the value of the text component
		try
		{
			// Check the number of decimals
			if ((_decimals > 0) && (new_value != null))
			{
				int pos = new_value.indexOf(_decimalSeparator);
				int decimalsNb = new_value.length() - pos - 1;

				if ((pos > 0) && (decimalsNb < _decimals))
				{
					int i, n = _decimals - decimalsNb;

					for (i = 0; i < n; i++)
						new_value = new_value + '0';
				}
			}

			LyJNumberDocument document = (LyJNumberDocument)_text.getDocument();

			document.setString((new_value == null) ? "" : new_value);
		}
		catch (BadLocationException e)
		{
			System.err.println("LyJNumberSelector.setValue(double) impossible");
		}

		// set the focus to the text component
		_text.requestFocus();

		checkButtons();
	} // end method setValue

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetValueInterface#setValue(java.lang.String)
	 */
	public void setValue(String str, boolean notify)
	{
		if (str != null)
			str = str.replace(_decimalSeparator, '.');

		_value = LyTools.doubleFromString(str);

		if (str != null)
		{
			str = str.replace('.', _decimalSeparator);
			
			str = LyTools.removeChar(str, _groupSeparator);
		}

		_oldString = str;

		super.setValue(str, notify);
		checkButtons();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.getinterface.LyGetValueInterface#getValue()
	 */
	public String getValue()
	{
		String value = super.getValue();

		String result = value.replace(_decimalSeparator, '.');

		return result;
	}

	/**
	 * Resets the text value if there is one and computes a new size for the text component. This
	 * method is used when min, max or decimals are changed.
	 *
	 * @param environment the current environment
	 */
	private void resetValue(LyEnvironment environment)
	{
		int size = computeSize();
		int currentSize = _text.getColumns();

		if (size != currentSize)
		{
			// When size is changed, the value is reset.
			setMaxChars(size);
			_text.invalidate();
			((JComponent)(_text.getParent())).revalidate();
		}

		// Reset the value.
		String val = getValue();

		if ((val != null) && (val.length() != 0))
			setValue(_value);

		// Compute the tips.
		if (_useArrows)
		{
			String min_str = null;
			String max_str = null;

			if (!_noMin)
			{
				if (_decimals == 0)
					min_str = String.valueOf((long)_min).replace('.', _decimalSeparator);
				else
					min_str = String.valueOf(_min).replace('.', _decimalSeparator);
			}
			else
				min_str = environment.translate("LY_NO_MIN_VALUE");

			String down_tooltip = environment.getString("LY_MIN_VALUE",
					new String[] { min_str, _unit });

			_down.setToolTipText(down_tooltip);

			if (!_noMax)
			{
				if (_decimals == 0)
					max_str = String.valueOf((long)_max).replace('.', _decimalSeparator);
				else
					max_str = String.valueOf(_max).replace('.', _decimalSeparator);
			}
			else
				max_str = environment.translate("LY_NO_MAX_VALUE");

			String up_tooltip = environment.getString("LY_MAX_VALUE",
					new String[] { max_str, _unit });

			_up.setToolTipText(up_tooltip);
		}
	} // end method resetValue

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJSelector#valueDiffersFrom(java.lang.String)
	 */
	protected boolean valueDiffersFrom(String value)
	{
		// Don't make any difference between '.' and the decimal separator
		String current = getValue().replace('.', _decimalSeparator);

		value = value.replace('.', _decimalSeparator);

		return (!current.equals(value));
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
	 */
	public void keyPressed(KeyEvent evt)
	{
		if (!_text.isEnabled() || !_text.isEditable())
			return;

		super.keyPressed(evt);

		int keyCode = evt.getKeyCode();

		if (_useArrows)
		{
			if (keyCode == KeyEvent.VK_UP)
				addPitch(true);
			else if (keyCode == KeyEvent.VK_DOWN)
				addPitch(false);
			checkButtons();
		}
	}

	/**
	 * Sets the format. This method is not implemented yet.
	 *
	 * @param environment the current environment
	 * @param format      the new value of format
	 */
	public void setFormat(LyEnvironment environment, String format)
	{
		// Not used yet.
	}

	/**
	 * Sets the minimum value. If this value is equal to {@link Double#NEGATIVE_INFINITY}, the
	 * {@link #_noMin} flag is set automatically.
	 *
	 * @param environment the current environment
	 * @param vMin        the new value of the minimum
	 */
	public void setMin(LyEnvironment environment, double vMin)
	{
		_min = vMin;
		_noMin = (_min == Double.NEGATIVE_INFINITY);
		resetValue(environment);
	}

	/**
	 * Sets the maximum value. If this value is equal to {@link Double#POSITIVE_INFINITY}, the
	 * {@link #_noMax} flag is set automatically.
	 *
	 * @param environment the current environment
	 * @param vMax        the new value of the maximum
	 */
	public void setMax(LyEnvironment environment, double vMax)
	{
		_max = vMax;
		_noMax = (_max == Double.POSITIVE_INFINITY);
		resetValue(environment);
	}

	/**
	 * Sets the pitch.
	 *
	 * @param environment the current environment, not used in this implementation
	 * @param pitch       the new pitch value
	 */
	public void setPitch(LyEnvironment environment, double pitch)
	{
		_pitch = pitch;
	}

	/**
	 * Sets the number of decimals.
	 *
	 * @param environment the current environment
	 * @param decimals    the new number of decimals
	 */
	public void setDecimals(LyEnvironment environment, int decimals)
	{
		_decimals = decimals;
		((LyJNumberDocument)_text.getDocument()).setDecimals(decimals);
		resetValue(environment);
	}
    
    /**
     * (non-Javadoc).
     * 
     * @see javax.swing.JComponent#getMinimumSize()
     */
    public Dimension getMinimumSize()
    {
        Dimension result = super.getMinimumSize();

        if (_width > 0)
            return new Dimension(_width, result.height);

        int     minWidth = (_buttons == null ? 0 : _buttons.getPreferredSize().width);
        
        return new Dimension(20 + minWidth, result.height);
    }
} // end class LyJNumberSelector
