/*
 * LyJOpenerButton.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;

import javax.swing.*;
import javax.swing.border.*;


/**
 * This class implements a toggle button that open a popupWindow.
 *
 * @version $Id: LyJOpenerButton.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class LyJOpenerButton extends LyJButton
{
	private static final long  serialVersionUID = 837269806937370379L;

	public static final String __VERSION =
		"$Id: LyJOpenerButton.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** An associated button of this toggle button. */
	protected LyJButton		   _neighbor;

	/**
	 * The toggle button orientation. {@link LyJToolBarComponent#HORIZONTAL} or {@link
	 * LyJToolBarComponent#VERTICAL}.
	 */
	protected int			   _orientation;

	/**
	 * Creates a new horizontal toggle button component with the specified image and initializes it.
	 *
	 * @param image the image of the toggle button
	 */
	public LyJOpenerButton(Image image)
	{
		super(image);
		setOrientation(SwingConstants.HORIZONTAL);
	}

	/**
	 * Initializes the button. Removes its margins and sets its center alignment.
	 */
	public void init()
	{
		super.init();
		setMargin(new Insets(0, 0, 0, 0));
		setHorizontalAlignment(CENTER);
		setRequestFocusEnabled(false);
	}

	/**
	 * This component cannot gain focus.
	 *
	 * @return <code>false</code>
	 * @see    java.awt.Component#isFocusTraversable()
	 */
	public boolean isFocusTraversable()
	{
		return false;
	}

	/**
	 * Gets the associated button.
	 *
	 * @return the button that is associated to this toggle button
	 */
	public LyJButton getButton()
	{
		return _neighbor;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#getMaximumSize()
	 */
	public Dimension getMaximumSize()
	{
		return getPreferredSize();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#getMinimumSize()
	 */
	public Dimension getMinimumSize()
	{
		return getPreferredSize();
	}

	/**
	 * Computes the preferred size of this component. This size depends of the neighbor button.
	 *
	 * @return the preferred size of the button
	 * @see    java.awt.Component#getPreferredSize()
	 */
	public Dimension getPreferredSize()
	{
		if (_orientation == SwingConstants.HORIZONTAL)
		{
			Dimension dim = super.getPreferredSize();

			return new Dimension(dim.width,
				(_neighbor != null) ? (_neighbor.getPreferredSize().height) : (dim.height));
		}
		Dimension dim = super.getPreferredSize();

		return new Dimension((_neighbor != null) ? (_neighbor.getPreferredSize().width)
												 : (dim.width), dim.height);
	}

	/**
	 * Sets the neighbor button.
	 *
	 * @param but the associated button
	 */
	public void setNeighbor(LyJButton but)
	{
		_neighbor = but;
	}

	/**
	 * Sets the mouse over style for current button.<br />
	 * The mouse over style draws a border around the button and fill its area.
	 * Only works with Metal look and feel.
	 * 
	 * @param over indicates if the mouse over style must be set or unset.
	 * @param propagate indicates if the style must be propagated to opener button.
	 */
	protected void setMouseOverStyle(boolean over, boolean propagate)
	{
		super.setMouseOverStyle(over, false);
		
		if (propagate && (_neighbor != null))
		{
			((_LyJButtonModel)_neighbor.getModel()).setRollover(over);
			_neighbor.setMouseOverStyle(over, false);
		}
	}
	
	/**
	 * Return the border to be used when mouse over style is set.
	 * 
	 * @return the border for current button.
	 */
	protected Border getMouseOverBorder()
	{
		return BorderFactory.createMatteBorder(1, 0, 1, 1, (Color)UIManager.get("Button.darkShadow"));
	}

	/**
	 * Sets the orientation of this component.
	 *
	 * @param orientation the orientation of this toggle button
	 * @see   #getOrientation()
	 */
	public void setOrientation(int orientation)
	{
		_orientation = orientation;
	}

	/**
	 * Gets the orientation of this component.
	 *
	 * @return the toggle button orientation
	 * @see    #setOrientation(int)
	 */
	public int getOrientation()
	{
		return _orientation;
	}

} // end class LyJOpenerButton
