/*
 * LyJPopupSelector.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import leon.misc.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.view.win.swing.*;


/**
 * This class implements a popup selection between a combo (choice) and a graphical component.
 *
 * @version $Id: LyJPopupSelector.java 28997 2010-12-23 10:17:15Z bmichel $
 * @author  Lyria
 */
public class LyJPopupSelector
	implements MouseListener,
		FocusListener,
		LyActionListener,
		ComponentListener
{
	public static final String __VERSION =
		"$Id: LyJPopupSelector.java 28997 2010-12-23 10:17:15Z bmichel $";

	/** Left alignment. */
	public final static int    ALIGN_LEFT = 0;

	/** Right alignment. */
	public final static int    ALIGN_RIGHT = 1;

	/** Top alignment. */
	public final static int    ALIGN_TOP = 2;

	/** Bottom alignment. */
	public final static int    ALIGN_BOTTOM = 3;

	/** The combo box element. */
	protected Component		   _choice;

	/** The associated popup window. */
	protected Component		   _popup;

	/**
	 * The alignment of this popup selector. One of the following:
	 *
	 * <ul>
	 *   <li>{@link #ALIGN_LEFT}</li>
	 *   <li>{@link #ALIGN_RIGHT}</li>
	 *   <li>{@link #ALIGN_TOP}</li>
	 *   <li>{@link #ALIGN_BOTTOM}</li>
	 * </ul>
	 */
	protected int			   _align;

	/** A popup window. */
	protected JWindow		   _tipWindow;

	/** Date when the focus is lost. */
	protected Date			   _lostFocusDate;

	/** List of popup items that need to have event listeners registered. */
	protected Vector<Component>		   _registeredPopupItems;

	/** The environment. */
	protected LyEnvironment    _environment;

	/**
	 * Creates a new popup selector with the specified attributes.
	 *
	 * @param choice      the component which ensures the display
	 * @param popup       the component that edit the value of the combo. Raised when opening the
	 *                    combo.
	 * @param environment the current environment
	 */
	public LyJPopupSelector(Component choice, Component popup, LyEnvironment environment)
	{
		this(choice, popup, ALIGN_RIGHT, environment);
	}

	/**
	 * Creates a new popup selector with the specified attributes.
	 *
	 * @param choice      the component which ensures the display
	 * @param popup       component to edit the value of the combo. Raised when opening the combo.
	 * @param align       specify the alignment (right or left)
	 * @param environment the current environment
	 */
	public LyJPopupSelector(Component choice, Component popup, int align, LyEnvironment environment)
	{
		_choice = choice;
		_popup = popup;
		_align = align;
		_environment = environment;

		init();
	}

	/**
	 * Initializes this component. Registers its mouse action listeners.
	 */
	protected void init()
	{
		// check if the _choice item is an action transmitter,
		// otherwise register for mouse clicks
		if (_choice instanceof LyEventTransmitter)
		{
			LyActionTransmitter transmitter = ((LyEventTransmitter)_choice).getActionTransmitter();

			if (transmitter != null)
				transmitter.addActionListener(this);
		}
		else
			_choice.addMouseListener(this);
	}

	/**
	 * Nothing is done.
	 *
	 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
	 */
	public void mouseClicked(MouseEvent e)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
	 */
	public void mousePressed(MouseEvent e)
	{
		if (e.getSource() == _choice)
		{
			// Swap visibility of the popup
			if (_popup.isShowing())
				hideTipWindow();
			else
			{
				if (!userClosingRequest())
					showTipWindow();
			}
		}
	}

	/**
	 * Nothing is done.
	 *
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	public void mouseReleased(MouseEvent e)
	{
	}

	/**
	 * Nothing is done.
	 *
	 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
	 */
	public void mouseEntered(MouseEvent e)
	{
	}

	/**
	 * Nothing is done.
	 *
	 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
	 */
	public void mouseExited(MouseEvent e)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		if ((_registeredPopupItems != null)
				&& (_registeredPopupItems.contains(actionEvent.getSource())))
		{
			// ## Update the content of the _choice ##
			String  value = null;
			boolean ready = false; // indicates if we managed to get the value

			Object  c = actionEvent.getSource();

			if (c instanceof LyGetValueInterface)
			{
				value = ((LyGetValueInterface)c).getValue();
				ready = true;
			}

			if (ready && (_choice instanceof LySetValueInterface))
				((LySetValueInterface)_choice).setValue(value);

			hideTipWindow();
			hideParentTipWindow();
		}
		else if (actionEvent.getType() == LyActionDemand.TYPE_ACTION)
		{
			// ## Swap visibility of the popup ##
			if ((_tipWindow != null) && _tipWindow.isShowing())
			{
				hideTipWindow();
				hideParentTipWindow();
			}
			else
			{
				if (!userClosingRequest())
					showTipWindow();
			}

			actionEvent.consume();
		}
	} // end method actionPerformed

	/**
	 * Sets the popup window and raises it.
	 */
	protected void showTipWindow()
	{
		Color background = (Color)_environment.getColor("LY_DEFAULT_BACKGROUND");

		if ((_choice == null) || (_popup == null) || (!_choice.isShowing()))
			return;

		if (_popup instanceof JPopupMenu)
		{
			JPopupMenu	popupMenu = (JPopupMenu)_popup;
			Point		location = getTipWindowLocation();
			
			popupMenu.show(_choice, location.x, location.y);
			
			return;
		}
		
		// be sure popup is closed
		hideTipWindow();

		// Be sure the value is Ok
		if ((_choice instanceof LyGetValueInterface) && (_popup instanceof LySetValueInterface))
		{
			String value = ((LyGetValueInterface)_choice).getValue();

			((LySetValueInterface)_popup).setValue(value);
		}

		Window    frame = null;

		Container parent = _choice.getParent();

		while ((parent != null) && (frame == null))
		{
			if (parent instanceof Window)
				frame = (Window)parent;
			parent = parent.getParent();
		}

		if (frame == null)
			frame = LySwingViewManager.getDefaultFrame();

		frame.addComponentListener(this);

		_tipWindow = new _LyPopupWindow(frame, this);

		_tipWindow.addWindowListener(new _LyPopupWindowAdapter());
		_tipWindow.setSize(200, 20);
		_tipWindow.setBackground(background);
		_tipWindow.getContentPane().setBackground(background);
		_tipWindow.addMouseListener(this);
		_tipWindow.addFocusListener(this);
		_popup.addFocusListener(this);

		// Register for action events sent by the popup window
		Vector<Component> v = new Vector<Component>();

		v.addElement(_popup);
		registerEvents(v); // Find at any depth and recurse
		
		// else : meaning of handling mouse clics ? TODO ?

		_popup.setBackground(background);
		_tipWindow.getContentPane().add("Center", _popup);

		_tipWindow.pack();

		Point location = getTipWindowLocation();

		_tipWindow.setLocation(location.x, location.y);

		_tipWindow.setVisible(true);
		_tipWindow.requestFocus();

		_popup.requestFocus();
	} // end method showTipWindow

	/**
	 * Registers action events for every component in the specified vector.
	 *
	 * @param  v the list of components that need to be registered
	 * @return <code>true</code> if one of the component of the vectors was registered, <code>
	 *         false</code> otherwise. The sub-components are checked and registered if the ones on
	 *         the specified vector cannot be registered, until no more data is available, or a
	 *         component has been registered.
	 */
	protected boolean registerEvents(Vector<Component> v)
	{
		// Check current level
		int     i, n = v.size();
		boolean found = false;

		if (n == 0)
			return false; // No more data

		for (i = 0; i < n; i++)
		{
			Component c = v.elementAt(i);

			if (c instanceof LyJOpenerButton)
				c = ((LyJOpenerButton)c).getButton();

			if (c instanceof LyEventTransmitter)
			{
				LyActionTransmitter transmitter = ((LyEventTransmitter)c).getActionTransmitter();

				if (transmitter != null)
				{
					transmitter.addActionListener(this);

					if (_registeredPopupItems == null)
						_registeredPopupItems = new Vector<Component>();
					_registeredPopupItems.addElement(c);
				}

				found = true;
			}
		}

		if (found)
			return true;

		// If not found : seek one level deeper
		Vector<Component> v1 = new Vector<Component>();

		for (i = 0; i < n; i++)
		{
			Object c = v.elementAt(i);

			if (c instanceof Container)
			{
				Container c1 = (Container)c;
				int		  i1, n1 = c1.getComponentCount();

				for (i1 = 0; i1 < n1; i1++)
					v1.addElement(c1.getComponent(i1));
			}
		}

		return registerEvents(v1);
	} // end method registerEvents

	/**
	 * Determines the popup window location on screen.
	 *
	 * @return the top left position of the popup window on the screen, depending on its alignment
	 */
	protected Point getTipWindowLocation()
	{
		Dimension size = null;
		Point     location = null;
		
		if (_popup instanceof JPopupMenu)
		{
			size = _popup.getPreferredSize();
			location = new Point(-_choice.getSize().width, 0);
		}
		else
		{
			if (_tipWindow == null)
				return null;

			size = _tipWindow.getPreferredSize();
			location = _choice.getLocationOnScreen();
		}
	
		// Get the current screenSize
		Window    frame = null;
		Container parent = _choice.getParent();

		while ((parent != null) && (frame == null))
		{
			if (parent instanceof Window)
				frame = (Window)parent;
			
			parent = parent.getParent();
		}

		if (frame == null)
			frame = LySwingViewManager.getDefaultFrame();
		
		int[]	  screenSize = _environment.getScreenSize(frame);
		Dimension screenDim = new Dimension(screenSize[0], screenSize[1]);
		Dimension choiceSize = _choice.getSize();
	    
	    // compute locationOffset
	    int horizontalOffset = screenSize[2];
	    int verticalOffset = screenSize[3];
	    
	    location.x -= horizontalOffset;
	    location.y -= verticalOffset;
	    
		if (_align == ALIGN_RIGHT)
		{
			location.x += choiceSize.width - size.width;
			location.y += choiceSize.height;
		}
		else if (_align == ALIGN_LEFT)
			location.y += choiceSize.height;
		else if (_align == ALIGN_TOP)
			location.x += choiceSize.width;
		else if (_align == ALIGN_BOTTOM)
		{
			location.x += choiceSize.width;
			location.y += choiceSize.height - size.height;
		}

		// Check if the window can be entirely displayed
		int diffX = screenDim.width - (location.x + size.width);
		int diffY = screenDim.height - (location.y + size.height);

		if ((diffX < 0) && ((location.x - size.width + choiceSize.width) >= 0))
			location.x += diffX;//size.width - choiceSize.width;
		if (location.x < 0) // window is partially displayed
			location.x = 0;
		if ((diffY < 0) && ((location.y - size.height - choiceSize.height) >= 0))
			location.y -= size.height + choiceSize.height;
		if (location.y < 0)
			location.y = 0;
		
		location.x += horizontalOffset;
		location.y += verticalOffset;
		
		return location;
	} // end method getTipWindowLocation

	/**
	 * Hides the popup window, then disposes it.
	 */
	public void hideTipWindow()
	{
		Vector<Component> popupItems = null;

		if (_tipWindow != null)
		{
			JWindow tipWindow = _tipWindow;

			_tipWindow = null;
			tipWindow.setVisible(false);

			tipWindow.removeMouseListener(this);
			tipWindow.removeFocusListener(this);
			_popup.removeFocusListener(this);

			if (_registeredPopupItems != null)
			{
				@SuppressWarnings("unchecked")
				Vector<Component> temp = (Vector<Component>)_registeredPopupItems.clone();
				popupItems = temp; // To avoid dead lock in
				// threads
			}

			if (popupItems != null)
			{
				int i, n = popupItems.size();

				for (i = 0; i < n; i++)
				{
					Object			    c = popupItems.elementAt(i);

					LyActionTransmitter transmitter = ((LyEventTransmitter)c)
							.getActionTransmitter();

					transmitter.removeActionListener(this);
				}
			}

			_registeredPopupItems = null;
			tipWindow.setSize(0, 0);
			tipWindow.dispose();
		}
	} // end method hideTipWindow

	/**
	 * Hides the parent window if it is a popup (cascading closure of popups). Closes all its parent
	 * popups recursively.
	 */
	public void hideParentTipWindow()
	{
		Window			 parentRootWindow = LySwingViewManager.getRootWindow(_choice);
		LyJPopupSelector selector = null;

		// If parent is a popup retrieve its popup selector and hide it
		if (parentRootWindow instanceof _LyPopupWindow)
			selector = ((_LyPopupWindow)parentRootWindow).getSelector();
		else if (parentRootWindow instanceof _LyPopupDialog)
			selector = ((_LyPopupDialog)parentRootWindow).getSelector();

		if (selector != null)
		{
			selector.hideTipWindow();
			// Recurse on parent
			selector.hideParentTipWindow();
		}
	}

	// ** Focus listener interface implementation
	/**
	 * Nothing is done.
	 *
	 * @see java.awt.event.FocusListener#focusGained(java.awt.event.FocusEvent)
	 */
	public void focusGained(FocusEvent e)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.FocusListener#focusLost(java.awt.event.FocusEvent)
	 */
	public void focusLost(FocusEvent e)
	{
		// If popup window is opened with popup button, tip window may
		// lost focus while popup gains it. In this case we must not
		// hide tip window.
		if ((e.getComponent() == _tipWindow) && (e.getOppositeComponent() == _popup))
			return;
		
		// If focus is lost due to an action on a registered component, tip window will
		// be hidden by the actionPerformed() method.
		if ((_registeredPopupItems != null) && (_registeredPopupItems.contains(e.getOppositeComponent())))
			return;

		// Patch JDK1.4 : hide window asynchronously to avoid discarding of events
		// on components of the window
		new Thread()
			{

				/*
				 * (non-Javadoc)
				 * @see java.lang.Runnable#run()
				 */
				public void run()
				{
					try
					{
						sleep(200);
					}
					catch (Exception e)
					{
					}
					hideTipWindow();
				}
			}.start();

		_lostFocusDate = new Date();
	} // end method focusLost

	/**
	 * Indicates wether closing is due to user request on the {@link #_choice} component.
	 *
	 * @return the closing status
	 */
	protected boolean userClosingRequest()
	{
		// The focus is lost when "normally" closing the
		// popup in an action performed
		if (_lostFocusDate != null)
		{
			Date d = new Date();

			if ((d.getTime() - _lostFocusDate.getTime()) < 500)
			{
				_lostFocusDate = null;
				return true;
			}
			_lostFocusDate = null;
		}

		return false;
	}

	/**
	 * Hides the popup window.
	 *
	 * @see #hideTipWindow()
	 * @see java.awt.event.ComponentListener#componentResized(java.awt.event.ComponentEvent)
	 */
	public void componentResized(ComponentEvent arg0)
	{
		hideTipWindow();
	}

	/**
	 * Hides the popup window.
	 *
	 * @see #hideTipWindow()
	 * @see java.awt.event.ComponentListener#componentMoved(java.awt.event.ComponentEvent)
	 */
	public void componentMoved(ComponentEvent arg0)
	{
		hideTipWindow();
	}

	/**
	 * Nothing is done.
	 *
	 * @see java.awt.event.ComponentListener#componentShown(java.awt.event.ComponentEvent)
	 */
	public void componentShown(ComponentEvent arg0)
	{
	}

	/**
	 * Hides the popup window.
	 *
	 * @see #hideTipWindow()
	 * @see java.awt.event.ComponentListener#componentHidden(java.awt.event.ComponentEvent)
	 */
	public void componentHidden(ComponentEvent arg0)
	{
		hideTipWindow();
	}

	/**
	 * Handles window events.
	 */
	class _LyPopupWindowAdapter extends WindowAdapter
	{

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.WindowListener#windowClosing(java.awt.event.WindowEvent)
		 */
		public void windowClosing(WindowEvent e)
		{
			// Should never happen : possible when using XTERM
			hideTipWindow();
			_lostFocusDate = new Date();
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.WindowListener#windowDeactivated(java.awt.event.WindowEvent)
		 */
		public void windowDeactivated(WindowEvent e)
		{
			hideTipWindow();
			_lostFocusDate = new Date();
		}
	}


	/**
	 * Handles the associated popup window selector.
	 */
	class _LyPopupWindow extends JWindow
	{
		private static final long  serialVersionUID = 1087556344253466488L;

		/** The associated popup selector. */
		protected LyJPopupSelector _selector;

		/**
		 * Creates a popup window for the specified window and associated to the specified popup
		 * selector.
		 *
		 * @param window   the parent window of this popup
		 * @param selector the associated popup selector
		 */
		public _LyPopupWindow(Window window, LyJPopupSelector selector)
		{
			super(window);

			_selector = selector;
		}

		/**
		 * Gets the popup selector of this component.
		 *
		 * @return the popup selector
		 */
		public LyJPopupSelector getSelector()
		{
			return _selector;
		}
	} // end class _LyPopupWindow


	/**
	 * Handles the associated popup dialog selector.
	 */
	class _LyPopupDialog extends JDialog
	{
		private static final long  serialVersionUID = 2886918627258650944L;

		/** The associated popup selector. */
		protected LyJPopupSelector _selector;

		/**
		 * Creates a popup dialog for the specified frame and associated to the specified popup
		 * selector.
		 *
		 * @param frame    the parent dialog of this popup
		 * @param selector the associated popup selector
		 */
		public _LyPopupDialog(Frame frame, LyJPopupSelector selector)
		{
			super(frame);

			_selector = selector;
		}

		/**
		 * Gets the popup selector of this component.
		 *
		 * @return the popup selector
		 */
		public LyJPopupSelector getSelector()
		{
			return _selector;
		}
	} // end class _LyPopupDialog
} // end class LyJPopupSelector
