/*
 * LyJSelector.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.text.*;
import leon.misc.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.view.win.swing.*;
import leon.view.win.swing.text.*;


/**
 * This class defines a component containing a text field and a set of buttons.
 *
 * @version $Id: LyJSelector.java 34856 2012-03-19 14:05:02Z jlsaouli $
 * @author  Lyria
 * @see     TextField
 */
public class LyJSelector extends JPanel
	implements FocusListener,
		KeyListener,
		LyGetValueInterface,
		LySetValueInterface,
		LyActionListener,
		LyEventTransmitter,
		LySetEditableInterface,
		LySetIdleInterface,
		LyIsIdleInterface
{
	private static final long     serialVersionUID = -6399735449862159316L;

	public static final String    __VERSION =
		"$Id: LyJSelector.java 34856 2012-03-19 14:05:02Z jlsaouli $";

	/** The text of this component. */
	protected LyJTextField		  _text;

	/** The panel containing this selector's buttons. */
	protected JPanel			  _buttons;

	/** The maximum width of this component's text. Use -1 for automatic calculation. */
	protected int				  _maxChars;

	/** Keeps the last value of this component. */
	protected String			  _lastValue;

	/** Idle mode. When in idle mode, the changed notifications are note send upon changes. */
	protected boolean			  _idle;

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter _actionTransmitter;

	/** The enabled color. */
	protected Color				  _enabledColor;

	/** The disabled color. */
	protected Color				  _disabledColor;

	/** The environment. */
	protected LyEnvironment		  _env;

	/** Indicates that this component has gained the keyboard focus or not. */
	protected boolean			  _hasGainedFocus;

	/** Indicates that this component is editable or not. */
	protected boolean			  _editable;

	/** The width of this component. Use -1 for automatic calculation. */
	protected int				  _width;

	/** Indicates if this component display arrows button and accept arrow key entry. */
	protected boolean			  _useArrows;
	
	// if false the button will never be enable
	protected boolean			  _autorizeEnabledButtons;

	/**
	 * Creates a new selector component with the specified text value.
	 *
	 * @param value the text value of this component
	 * @param env   the current environment
	 */
	protected LyJSelector(LyEnvironment env, boolean useArrows)
	{
		super();
		
		_useArrows = useArrows;
		_env = env;
		_editable = true;
		init();
	}
	
	/**
	 * Creates a new selector component with the specified text value.
	 *
	 * @param value the text value of this component
	 * @param env   the current environment
	 */
	public LyJSelector(String value, LyEnvironment env, boolean useArrows)
	{
		this(env, useArrows);

		if (value != null)
			setValue(value);
	}

	/**
	 * Initializes the attributes of this component.
	 */
	protected void init()
	{
		_maxChars = -1;
		_width = -1;

		_text = textFieldFactory();

		_text.addKeyListener(this);

		_hasGainedFocus = false;

		_idle = false;
		
		_autorizeEnabledButtons = true;

		JPanel centerPanel = new JPanel(new BorderLayout());

		centerPanel.add(_text, "Center");

		if (_useArrows)
		{
			_buttons = new JPanel();
			_buttons.setLayout(new GridBagLayout());
			_buttons.setOpaque(false);
			centerPanel.add(_buttons, "East");
		}

		// Add the components
		setLayout(new BorderLayout());
		add(centerPanel, "Center");

		_text.addFocusListener(this);
		_lastValue = getValue();

		_enabledColor = null;
		_disabledColor = null;
		
		_text.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createMatteBorder(1, 1, 1, (_useArrows ? 0 : 1), UIManager.getColor("TextField.darkShadow")),
				BorderFactory.createEmptyBorder(1, 2, 1, 2)));

		setBackColor(null);
	} // end method init

	/**
	 * Creates a button with the specified attributes and registers the action listeners then adds
	 * it to this component's buttons panel.
	 *
	 * @param  image the image to associate to the button
	 * @param  label the label of this button
	 * @return the newly created button
	 */
	public LyJButton addButton(Image image, String label)
	{
		LyJButton		    button = new LyJButton(image, label, null, "selector_button")
			{
				private static final long serialVersionUID = 714474171529528225L;

				/**
				 * This component cannot gain focus.
				 *
				 * @return <code>false</code>
				 * @see    java.awt.Component#isFocusTraversable()
				 */
				public boolean isFocusTraversable()
				{
					return false;
				}
			};

		button.setRequestFocusEnabled(false);
		button.setOpaque(true);
		button.setRollover(false);
		button.setContentAreaFilled(false);

		LyActionTransmitter transmitter = button.getActionTransmitter();

		if (transmitter != null)
			transmitter.addActionListener(this);

		int		topBorder = 1;
		
		if (_buttons.getComponentCount() > 0)
			topBorder = 0;
		
		button.setBorder(BorderFactory.createMatteBorder(topBorder, 1, 1, 1,
				UIManager.getColor("TextField.darkShadow")));

		_buttons.add(button, buildButtonConstraint());

		return button;
	}
	
	/**
	 * Builds the constraints to be used for a button.
	 * 
	 * @return a grid bag constraint instance.
	 */
	protected GridBagConstraints buildButtonConstraint()
	{
		GridBagConstraints	constraints = new GridBagConstraints();
		
		constraints.fill = GridBagConstraints.BOTH;
		constraints.weighty = 1.0;
		constraints.weightx = 1.0;
		constraints.gridwidth = GridBagConstraints.REMAINDER;
		
		return constraints;
	}

	/**
	 * Sets the enabled color.
	 *
	 * @param color the new enabled color
	 */
	public void setEnabledColor(Color color)
	{
		_enabledColor = color;
	}

	/**
	 * Sets the disabled color.
	 *
	 * @param color the new disabled color
	 */
	public void setDisabledColor(Color color)
	{
		_disabledColor = color;
	}

	/**
	 * Adds the specified component as the "East" position of this component.
	 *
	 * @param comp the component to add
	 */
	public void setComponent(Component comp)
	{
		add(comp, "East");
	}

	/**
	 * Sets the background color.
	 *
	 * @param color the new background color
	 */
	public void setBackColor(Color color)
	{
		super.setBackground(color);
		int i, n = getComponentCount();

		for (i = 0; i < n; i++)
			getComponent(i).setBackground(color);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		if (_actionTransmitter != null)
			_actionTransmitter.processActionEvent(actionEvent);
	}

	/**
	 * Nothing is done.
	 *
	 * @see java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
	 */
	public void keyTyped(KeyEvent evt)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
	 */
	public void keyPressed(KeyEvent evt)
	{
		int key = evt.getKeyCode();

		// If enter, simulate a focusLost
		if (key == KeyEvent.VK_ENTER)
			focusLost(null);
	}

	/**
	 * Nothing is done.
	 *
	 * @see java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
	 */
	public void keyReleased(KeyEvent evt)
	{
	}

	/**
	 * Sets the maximum number of characters of this component's text area.
	 *
	 * @param maxChars the maximum width of the text
	 */
	public void setMaxChars(int maxChars)
	{
		_maxChars = maxChars;

		_text.setColumns(maxChars);
		((LyJTextFieldDocument)_text.getDocument()).setMaxChars(maxChars);
	}

	/**
	 * Sets the number of text columns.
	 *
	 * @param nCols the number of text columns
	 */
	public void setNCols(int nCols)
	{
		_text.setColumns(nCols);
	}

	/**
	 * Sets the width of this component.
	 *
	 * @param width the width of this component. Use -1 for automatic calculation.
	 */
	public void setWidth(int width)
	{
		_width = width;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#getPreferredSize()
	 */
	public Dimension getPreferredSize()
	{
		Dimension result = super.getPreferredSize();

		if (_width > 0)
			return new Dimension(_width, result.height);
		return result;
	}

	/**
	 * Gets the text field of this component.
	 *
	 * @return the text field
	 */
	public LyJTextField getTextField()
	{
		return _text;
	}

	/**
	 * Gets the buttons panel of this component.
	 *
	 * @return the buttons panel
	 */
	public JPanel getButtons()
	{
		return _buttons;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.getinterface.LyGetValueInterface#getValue()
	 */
	public String getValue()
	{
		Document doc = _text.getDocument();
		String   value;

		try
		{
			value = doc.getText(0, doc.getLength());
		}
		catch (BadLocationException e)
		{
			value = _text.getText();
		}

		return value;

	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetValueInterface#setValue(java.lang.String)
	 */
	public final void setValue(String str)
	{
		setValue(str, true);
	}

	/**
	 * Changes the value of the text component.
	 *
	 * @param str    the new value
	 * @param notify Indicates that the action transmitter must be notified of this change.
	 * @see   #getValue()
	 */
	public void setValue(String str, boolean notify)
	{
		String set = str;

		if (set == null)
			set = "";

		if ((_maxChars > 0) && (set.length() > _maxChars))
			set = str.substring(0, _maxChars);

		if (valueDiffersFrom(set))
		{
			// change the value of the text component
			try
			{
				((LyJTextAreaDocument)_text.getDocument()).setString((str == null) ? "" : str);
			}
			catch (BadLocationException e)
			{ /* _date.getEnvironment().logError("SetValue impossible"); */
			}

			_lastValue = getValue();

			if (notify && !_idle)
			{
				if (_actionTransmitter != null)
					_actionTransmitter.processActionEvent(new LyActionEvent(this, null,
					// set,
							getValue(), LyActionDemand.TYPE_VALUE_CHANGED));
			}
		}
	} // end method setValue

	/**
	 * Checks that the specified value is different than the one of this component.
	 *
	 * @param  value the value to test this component's value
	 * @return <code>true</code> if the value differs, <code>false</code> otherwise
	 */
	protected boolean valueDiffersFrom(String value)
	{
		return (!getValue().equals(value));
	}

	/**
	 * Autorize the button to be enabled
	 * 
	 * @param enable if false the button will never be enable
	 */
	public synchronized void setAutorizeEnabledButtons(boolean enable)
	{
		if(enable != _autorizeEnabledButtons)
		{
			_autorizeEnabledButtons = enable;
			checkButtons();
		}
	}
	
	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#setEnabled(boolean)
	 */
	public synchronized void setEnabled(boolean enable)
	{
		super.setEnabled(enable);

		_text.setEnabled(enable);
		if (_buttons != null)
			_buttons.setEnabled(enable && isEditable());

		Color bgColor;

		if (enable)
			bgColor = _enabledColor;
		else
			bgColor = _disabledColor;

		setBackground(bgColor);

		checkButtons();
	}

	/**
	 * Updates the buttons. Doesn't do anything in this implementation.
	 */
	protected void checkButtons()
	{
		int			i, n = _buttons.getComponentCount();
		Component	component;
		
		for (i = 0; i < n; i++)
		{
			component = _buttons.getComponent(i);
			component.setEnabled(isEnabled() && isEditable() && _autorizeEnabledButtons);
		}
	}

	/**
	 * Creates and initializes a new TextField instance.
	 *
	 * @return the new text field
	 */
	public LyJTextField textFieldFactory()
	{
		LyJTextField text = new LyJTextField(_env)
			{
				private static final long serialVersionUID = 4510201261234596706L;

				/*
				 * (non-Javadoc)
				 * @see javax.swing.JTextField#getColumnWidth()
				 */
				public int getColumnWidth()
				{
					return LySwingUtils.getFontCharWidth(this, getFont());
				}

				/*
				 * (non-Javadoc)
				 * @see java.awt.Component#getMaximumSize()
				 */
				public Dimension getMaximumSize()
				{
					return this.getPreferredSize();
				}
			};

		text.setDocument(new LyJTextFieldDocument(0)
			{
				private static final long serialVersionUID = -2789472984860384L;

				/*
				 * (non-Javadoc)
				 * @see javax.swing.text.Document#insertString(int, java.lang.String,
				 * javax.swing.text.AttributeSet)
				 */
				public void insertString(int offset, String str, AttributeSet a)
					throws BadLocationException
				{
					super.insertString(offset, str, a);
					checkButtons();
				}
			});

		return text;
	} // end method textFieldFactory

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.swing.JComponent#requestFocus()
	 */
	public void requestFocus()
	{
		_text.requestFocus();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.FocusListener#focusGained(java.awt.event.FocusEvent)
	 */
	public void focusGained(FocusEvent e)
	{
		_lastValue = getValue();
		_hasGainedFocus = true;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.FocusListener#focusLost(java.awt.event.FocusEvent)
	 */
	public void focusLost(FocusEvent e)
	{
		if (!_hasGainedFocus)
			return;

		_hasGainedFocus = false;

		String value = getValue();
		// check if something has changed

		if ((!_idle)
				&& (((_lastValue == null) && (value != null))
					|| ((_lastValue != null) && (!_lastValue.equals(value)))))
		{
			if (_actionTransmitter != null)
				_actionTransmitter.processActionEvent(new LyActionEvent(this, null, value,
						LyActionDemand.TYPE_VALUE_CHANGED));
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetIdleInterface#setIdle(boolean)
	 */
	public void setIdle(boolean status)
	{
		_idle = status;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.getinterface.LyIsIdleInterface#isIdle()
	 */
	public boolean isIdle()
	{
		return _idle;
	}

	/**
	 * Checks if this component is editable or not.
	 *
	 * @return true if this component is editable, false otherwise
	 * @see    #setEditable(boolean)
	 */
	public boolean isEditable()
	{
		return _editable;
	}

	/**
	 * Sets the editable status of the field item.
	 *
	 * @see leon.view.setinterface.LySetEditableInterface#setEditable(boolean)
	 * @see #isEditable()
	 */
	public void setEditable(boolean editable)
	{
		if (_text != null)
		{
			_text.setEditable(editable);
			if (_buttons != null)
				_buttons.setEnabled(editable && isEnabled());

			Color bgColor;

			if (editable && _text.isEnabled())
				bgColor = _enabledColor;
			else
				bgColor = _disabledColor;

			setBackground(bgColor);
		}

		_editable = editable;
		checkButtons();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#setBackground(java.awt.Color)
	 */
	public void setBackground(Color color)
	{
		if (_text != null)
			_text.setBackground(color);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#setFont(java.awt.Font)
	 */
	public void setFont(Font font)
	{
		if (_text != null)
			_text.setFont(font);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return (_actionTransmitter);
	}

	/**
	 * This selector doesn't handle selection events, so it has no selection transmitter.
	 *
	 * @return <code>null</code>
	 * @see    leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		return null;
	}

	/**
	 * Adds a key listener on the text field.
	 *
	 * @see java.awt.Component#addKeyListener(java.awt.event.KeyListener)
	 */
	public void addKeyListener(KeyListener listener)
	{
		_text.addKeyListener(listener);
	}
} // end class LyJSelector
