/*
 * LyJSeparator.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import javax.swing.*;


/**
 * This class defines a line which acts as a graphical separator.
 *
 * @version $Id: LyJSeparator.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     JPanel
 */
public class LyJSeparator extends JSeparator
{
	private static final long  serialVersionUID = 8623334159329129082L;

	public static final String __VERSION =
		"$Id: LyJSeparator.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Menu style. */
	public static final int    MENU_STYLE = 0;

	/** ToolBar style. */
	public static final int    TOOLBAR_STYLE = 1;

	/** Minimum length of the separator, in pixels. */
	public static final int    MIN_LENGTH = 22;

	/** Style of the separator. {@link #MENU_STYLE} or {@link #TOOLBAR_STYLE}. */
	protected int			   _style;

	/** Invisible component used for left/top gap. */
	protected Component		   _beforeGap;

	/** Invisible component used for right/bottom gap. */
	protected Component		   _afterGap;

	/**
	 * Creates a separator object with the specified attributes. No gap is used here.
	 *
	 * @param align           the orientation. {@link #HORIZONTAL} or {@link #VERTICAL}.
	 * @param style           the style. MENU_STYLE or TOOLBAR_STYLE.
	 * @param foregroundColor the foreground color
	 */
	public LyJSeparator(int align, int style, Color foregroundColor)
	{
		this(align, style, foregroundColor, 0);
	}

	/**
	 * Creates a separator object with the specified attributes.
	 *
	 * @param align           the orientation. HORIZONTAL or VERTICAL.
	 * @param style           the style, not used here
	 * @param foregroundColor the foreground color
	 * @param gap             the gap around this separator
	 */
	public LyJSeparator(int align, int style, Color foregroundColor, int gap)
	{
		super();
		init();

		setOrientation(align);
		setForeground(foregroundColor);

		int	width = gap, height = gap;
		
		if (getOrientation() == SwingConstants.HORIZONTAL)
			width = MIN_LENGTH;
		else
			height = MIN_LENGTH;
		
		_beforeGap = Box.createRigidArea(new Dimension(width, height));
		_afterGap = Box.createRigidArea(new Dimension(width, height));
	}

	/**
	 * Initializes this component. Nothing is done here.
	 */
	protected void init()
	{
	}

	/**
	 * Gets the gap before this component. Used only for TOOLBAR_STYLE.
	 *
	 * @return the gap between the previous component and this separator
	 * @see    #setBeforeGap(int)
	 */
	public Component getBeforeGap()
	{
		return _beforeGap;
	}

	/**
	 * Gets the gap after this component. Used only for TOOLBAR_STYLE.
	 *
	 * @return the gap between the next component and this separator
	 * @see    #setAfterGap(int)
	 */
	public Component getAfterGap()
	{
		return _afterGap;
	}

	/**
	 * Sets the gap between the previous component and this separator.
	 *
	 * @param gap the new gap value
	 * @see   #getBeforeGap()
	 */
	public void setBeforeGap(int gap)
	{
		if (getOrientation() == SwingConstants.HORIZONTAL)
			_beforeGap.setSize(new Dimension(MIN_LENGTH, gap));
		else
			_beforeGap.setSize(new Dimension(gap, MIN_LENGTH));
	}

	/**
	 * Sets the gap between the next component and this separator.
	 *
	 * @param gap the new gap value
	 * @see   #getAfterGap()
	 */
	public void setAfterGap(int gap)
	{
		if (getOrientation() == SwingConstants.HORIZONTAL)
			_afterGap.setSize(new Dimension(MIN_LENGTH, gap));
		else
			_afterGap.setSize(new Dimension(gap, MIN_LENGTH));
	}

	/**
	 * Gets the size to have a nice looking size and to auto-adapt with orientation change (for a
	 * floatable toolbar).
	 *
	 * @see java.awt.Component#getPreferredSize()
	 */
	public Dimension getPreferredSize()
	{
		Dimension dim = super.getPreferredSize();

		if (getOrientation() == SwingConstants.VERTICAL)
			// FFT00453 when toolbar is horizontal the layout is a LyFlowLayout
			return new Dimension(dim.width, MIN_LENGTH);
		// when toolbar is vertical the layout is a BoxLayout
		return dim;
	}

	/**
	 * Gets the size to have a nice looking size and to auto-adapt with orientation change (for a
	 * floatable toolbar).
	 *
	 * @see java.awt.Component#getMaximumSize()
	 */
	public Dimension getMaximumSize()
	{
		Dimension dim = getPreferredSize();

		if (getOrientation() == SwingConstants.VERTICAL)
			return new Dimension(dim.width, Integer.MAX_VALUE);
		return new Dimension(Integer.MAX_VALUE, dim.height);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#getMinimumSize()
	 */
	public Dimension getMinimumSize()
	{
		return getPreferredSize();
	}

} // end class LyJSeparator
