/*
 * LyJSplitterComponent.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.util.*;
import javax.swing.*;


/**
 * This class implements a container who can layout severals components separed by a mobile bar.
 *
 * @see leon.view.win.swing.LySwingSplitter
 */
public class LyJSplitterComponent extends JPanel
{
	private static final long	    serialVersionUID = 3588253611533774071L;

	public static final String	    __VERSION =
		"$Id: LyJSplitterComponent.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Orientation of the splitter. {@link JSplitPane#HORIZONTAL_SPLIT} or {@link
	 * JSplitPane#VERTICAL_SPLIT}.
	 */
	protected int				    _orientation;

	/** Array of absolute sizes coming from the data source. */
	protected int[]				    _sizes;

	/** Array of relative sizes computed in this class. */
	protected double[]			    _relativeSizes;

	/** Vector of all split pane used in this Component. */
	protected Vector<_LyJSplitPane> _splitPanes;

	/** Size of the divider. */
	protected int				    _dividerSize;

	/**
	 * Creates a new splitter component with the specified attributes.
	 *
	 * @param orientation orientation of the splitter. JSplitPane.HORIZONTAL or JSplitPane.VERTICAL.
	 * @param components  the components to display
	 * @param sizes       the initial sizes of the components
	 */
	public LyJSplitterComponent(int orientation, Component[] components, int[] sizes)
	{
		super();
        
		setLayout(new BorderLayout());
		_orientation = orientation;
		_sizes = sizes;
		_splitPanes = new Vector<_LyJSplitPane>(1);

		// By default, in Metal Look & Feel, size is 10 pixels.
		// By default, in System Windows Look & Feel, size is 7 pixels.
		// it sets 9 pixels minimum
		_dividerSize = UIManager.getInt("SplitPane.dividerSize");

		if (_dividerSize < 9)
			_dividerSize = 9;

		computeInitialSizes();
		addAll(components);
	}

	/**
	 * Adds all the Components from the specified array.
	 *
	 * @param components the array of all the components which have to de added
	 */
	public void addAll(Component[] components)
	{
		// the number of components >= 2
		int length = components.length;

		if (length == 1)
		{
			add(components[0], "Center");
			return;
		}

		_LyJSplitPane splitter = new _LyJSplitPane(_orientation, true, components[length - 2],
				components[length - 1], length - 2);

		// sets divider size FFT00998
		splitter.setDividerSize(_dividerSize);

		_splitPanes.add(0, splitter);

		for (int i = length - 3; i >= 0; i--)
		{
			splitter = new _LyJSplitPane(_orientation, true, components[i], splitter, i);

			// sets divider size FFT00998
			splitter.setDividerSize(_dividerSize);

			_splitPanes.add(0, splitter);
		}
		add(splitter, "Center");
	} // end method addAll

	/**
	 * Computes the relative sizes from the absolute sizes. The relative sizes are used in {@link
	 * _LyJSplitPane#setDividerLocation(double)}.
	 */
	public void computeInitialSizes()
	{
		int		 i, n = _sizes.length;
		double[] normalSizes = normalizeSizes();

		_relativeSizes = new double[n - 1];
		double sum = 0;

		for (i = 0; i < (n - 1); i++)
		{
			sum += normalSizes[n - 1 - i];
			_relativeSizes[n - 2 - i] = normalSizes[n - 2 - i] / (normalSizes[n - 2 - i] + sum);
		}
	}

	/**
	 * Normalizes the array of absolute sizes.
	 *
	 * @return an array with values set between 0 and 1
	 */
	public double[] normalizeSizes()
	{
		double[] result = new double[_sizes.length];
		int		 max = max(_sizes);

		for (int i = 0; i < result.length; i++)
			result[i] = _sizes[i] / ((double)max);
		return result;
	}

	/**
	 * Checks if the minimum size of internal components is enough.
	 *
	 * @param d1 first component dimensions
	 * @param d2 second component dimensions
	 */
	public void checkMinimumSize(Dimension d1, Dimension d2)
	{
		if (_splitPanes.size() == 1)
		{
			_LyJSplitPane panel = _splitPanes.elementAt(0);

			panel.checkMinimumSize(d1, d2);
		}
	}

	/**
	 * Gets the maximum value of the specified array.
	 *
	 * @param  array the array containing the values to test
	 * @return the max value from an array of int
	 */
	public int max(int[] array)
	{
		int result = 0;

		for (int i = 0; i < array.length; i++)
			if (array[i] > result)
				result = array[i];
		return result;
	}

	/**
	 * Gets the minimum size of this splitter component.
	 *
	 * @return the minimum dimensions as a splitter can be hidden
	 * @see    java.awt.Component#getMinimumSize()
	 */
	public Dimension getMinimumSize()
	{
		return new Dimension(0, 0);
	}

	/**
	 * Sets the resizable porperty of ths component.
	 *
	 * @param enabled Indicates that this component is resizable.
	 */
	public void setResizable(boolean enabled)
	{
		int i, n = _splitPanes.size();

		for (i = 0; i < n; i++)
		{
			((JSplitPane)_splitPanes.get(i)).setEnabled(enabled);
			((JSplitPane)_splitPanes.get(i)).setOneTouchExpandable(enabled);
		}
	}

	/**
	 * Gets the size of the splitter, if available.
	 *
	 * @return the size of the splitter or 0 if unknown
	 */
	public int getSplitterSize()
	{
		if (_splitPanes.size() == 1)
		{
			_LyJSplitPane panel = _splitPanes.elementAt(0);

			return panel.getDividerSize();
		}
		return 0;
	}

	/* ************************* Internal Class ******************************/

	/**
	 * This class implements a panel divided in two parts by a divider bar.
	 */
	public class _LyJSplitPane extends JSplitPane
	{
		private static final long serialVersionUID = -2175911670927638405L;

		/** The position of this SplitPane in the LySplitterComponent. */
		protected int			  _pos;
        
		/** The proportional location. */
		protected double		  _proportionalLocation;

		/**
		 * Creates a new SplitPane component with the specified attributes.
		 *
		 * @param o     the orientation. JSplitPane.HORIZONTAL_SPLIT or JSplitPane.VERTICAL_SPLIT.
		 * @param exp   Indicates that this splitter component must redraw itself when changing, or
		 *              wait for the splitting action to terminate before redrawing.
		 * @param left  the left or top
		 * @param right the right or bottom component
		 * @param pos   the position,that this SplitPane have in the LySplitterComponent
		 * @see   JSplitPane
		 */
		public _LyJSplitPane(int o, boolean exp, Component left, Component right, int pos)
		{
			super(o, exp, left, right);
			
			_pos = pos;
			_proportionalLocation = _relativeSizes[_pos];
			
			setOneTouchExpandable(true);
			setBorder(null);
			setDividerLocation(_proportionalLocation);
	        setResizeWeight(_proportionalLocation);
			setContinuousLayout(false);
		}

		/**
		 * Sets proportional location to a nested split panel.
		 *
		 * @see java.awt.Component#paint(java.awt.Graphics)
		 */
		public void paint(Graphics g)
		{
			if ((_proportionalLocation >= 0) && (_proportionalLocation <= 1))
			{
				setDividerLocation(_proportionalLocation);
				_proportionalLocation = -1;
			}

			super.paint(g);
		}

		/**
		 * Checks if the minimum size of internal components is enough.
		 *
		 * @param d1 first component dimensions
		 * @param d2 second component dimensions
		 */
		public void checkMinimumSize(Dimension d1, Dimension d2)
		{
			Dimension d = this.getSize();
			boolean   update = false;

			// Check left component
			if (d1 != null)
			{
				if (getOrientation() == VERTICAL_SPLIT)
				{
					if (d1.height > getDividerLocation())
					{
						// Not enough on left !
						_proportionalLocation = (d1.height) / ((double)d.height);
						update = true;
					}
				}
				else
				{
					if (d1.width > getDividerLocation())
					{
						_proportionalLocation = (d1.width) / ((double)d.width);
						update = true;
					}
				}
			}
			// Check right component
			if ((!update) && (d2 != null))
			{
				if (getOrientation() == VERTICAL_SPLIT)
				{
					if (d2.height > (d.height - getDividerLocation() - getDividerSize()))
					{
						// Not enough on right!
						_proportionalLocation = 1D - (d.height / ((double)d.height - d2.height - getDividerSize()));
						update = true;
					}
				}
				else
				{
					if (d2.width > (d.width - getDividerLocation() - getDividerSize()))
					{
						this._proportionalLocation = 1D - ((d2.width + getDividerSize()) / ((double)d.width));
						update = true;
					}
				}
			}

			// Need to force update on width / height
			if (update)
			{
				if (_proportionalLocation >= 1)
					_proportionalLocation = 0.999;
				else if (this._proportionalLocation <= 0)
					_proportionalLocation = 0.001;

				setDividerLocation(_proportionalLocation);
			}
		} // end method checkMinimumSize

		/**
		 * Gets the position of the SplitPane.
		 *
		 * @return its position
		 */
		public int getPosition()
		{
			return _pos;
		}
	} // end class _LyJSplitPane
} // end class LyJSplitterComponent
