/*
 * LyJTableComponent.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.datatransfer.*;
import java.awt.dnd.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;
import javax.swing.text.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.win.swing.*;
import leon.view.win.swing.table.*;
import leon.view.win.swing.table.LyJCellRenderer.*;


/**
 * This class implements graphical component that display the data in a table view.
 *
 * @version $Id: LyJTableComponent.java 35415 2012-04-30 15:55:03Z nhamon $
 * @author  Lyria
 * @see     LySwingTable
 * @see     JTable
 */
public class LyJTableComponent extends JTable
	implements LySelectionListener,
		LyActionListener,
		LyEventTransmitter,
		LyJCustomizedScrollable
{
	private static final long			   serialVersionUID = -1282156200406570201L;

	public static final String			   __VERSION =
		"$Id: LyJTableComponent.java 35415 2012-04-30 15:55:03Z nhamon $";

	/** The maximum width of a line in a tooltip. */
	public static int					   TIP_MAX_CHARS = 60;

	/** The mouseListener of this table. */
	public MouseListener				   _mouseListener;

	/** The selection listener of this table. */
	protected ListSelectionListener		   _selectionListener;

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter		   _actionTransmitter;

	/** Provides notification events on component's selections. */
	protected LySelectionTransmitter	   _selectionTransmitter;

	/** Image for sort down. */
	protected Image						   _sortDownImage;

	/** Image for sort up. */
	protected Image						   _sortUpImage;

	/** Icon for sort down. */
	protected Icon						   _sortDownIcon;

	/** Icon for sort up. */
	protected Icon						   _sortUpIcon;

	/** The renderer of the table's cells. */
	protected TableCellRenderer			   _cellRenderer;

	/** Minimum height. */
	protected int						   _minHeight;

	/** Displayed height. */
	protected int						   _displayedHeight;

	/** Width of the margin displayed on the right of the table. */
	protected int						   _rightMargin;

	/** List of heights for all the cells contained in the table. */
	protected Hashtable<String, Integer>[] _cellHeights;

	/** Indicates that drag and drop is allowed for this component. */
	protected boolean					   _dragNDrop;

	/** The drag source for drag and drop. */
	protected DragSource				   _dragSource;

	/** Indicates that height must be computed for each row of the table or not. */
	protected boolean					   _computeHeight;

	/** Indicates that this table has at least one editable cell. */
	protected boolean					   _hasEditableCells;

	/** Indicates the minimum number of characters used to compute table width. */
	protected int						   _minChars;

	/**
	 * Preferred scrollable viewport size (kept for optimization and to avoir resize of tables when
	 * a min height is given).
	 */
	protected Dimension					   _preferredScrollableSize;

	/**
	 * Creates a table object that don't have drag and drop and initializes it.
	 *
	 * @param downImage         the sort down image
	 * @param upImage           the sort up image
	 * @param mouseListener     the mouse listener of this table
	 * @param selectionListener the selection listener of this table
	 */
	public LyJTableComponent(Image downImage, Image upImage, MouseListener mouseListener,
		ListSelectionListener selectionListener)
	{
		super();

		_sortDownImage = downImage;
		_sortUpImage = upImage;
		_sortDownIcon = new ImageIcon(downImage);
		_sortUpIcon = new ImageIcon(upImage);
		_mouseListener = mouseListener;
		_selectionListener = selectionListener;
		_rightMargin = 100;
		_computeHeight = false;
		_hasEditableCells = false;
		_minChars = -1;
		_preferredScrollableSize = null;

		initialize();
	}

	/**
	 * Creates a table object with the specified attributes and initializes it.
	 *
	 * @param downImage         the sort down image
	 * @param upImage           the sort up image
	 * @param mouseListener     the mouse listener of this table
	 * @param selectionListener the selection listener of this table
	 * @param dragNDrop         Indicates that drag and drop is allowed on this component.
	 * @deprecated
	 */
	public LyJTableComponent(Image downImage, Image upImage, MouseListener mouseListener,
		ListSelectionListener selectionListener, boolean dragNDrop)
	{
		this(downImage, upImage, mouseListener, selectionListener);
	}

	/**
	 * Initializes this table component.
	 */
	public void initialize()
	{
		setAutoResizeMode(AUTO_RESIZE_OFF);
		setCellSelectionEnabled(false);
		setColumnSelectionAllowed(false);
		setRowSelectionAllowed(true);
		addLyJTableMouseListener();
		addLyJTableSelectionListener();
		setDoubleBuffered(true);
	}

	/**
	 * Initialize drag and drop or table component.
	 */
	public void setDragAndDrop()
	{
		if (_dragNDrop)
			return;
		
		_dragSource = new DragSource();
		_dragSource.createDefaultDragGestureRecognizer(this, DnDConstants.ACTION_COPY_OR_MOVE,
			new _LyTableCanvasDragListener());
		new DropTarget(this, new _LyTableCanvasDropListener());
		new DropTarget(((JScrollPane)getParent().getParent()).getViewport(), new _LyTableCanvasDropListener());
		_dragNDrop = true;
	}
	
	/**
	 * Checks that the cell is still in edition. Editable status of a cell may change.
	 *
	 * @see javax.swing.JTable#isEditing()
	 */
	public boolean isEditing()
	{
		boolean parentEdit = super.isEditing();

		if (parentEdit)
		{
			int r = getEditingRow();
			int c = getEditingColumn();

			return (getTableModel().isCellEditable(r, c));
		}
		return parentEdit;
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.JTable#getDefaultEditor(java.lang.Class)
	 */
	public TableCellEditor getDefaultEditor(Class<?> columnClass)
	{
		// if (LySwingFormField.class == columnClass)
		// XXX TO IMPROVE
		return new LyCellEditor();
		// else
		// return super.getDefaultEditor(columnClass);
	}

	/**
	 * Gets the image for sort down.
	 *
	 * @return the sort down image
	 */
	public Image getSortDownImage()
	{
		return _sortDownImage;
	}

	/**
	 * Gets the image for sort up.
	 *
	 * @return the sort up image
	 */
	public Image getSortUpImage()
	{
		return _sortUpImage;
	}

	/**
	 * Gets the icon for sort down.
	 *
	 * @return the sort down icon
	 */
	public Icon getSortDownIcon()
	{
		return _sortDownIcon;
	}

	/**
	 * Gets the icon for sort up.
	 *
	 * @return the sort up icon
	 */
	public Icon getSortUpIcon()
	{
		return _sortUpIcon;
	}

	/**
	 * Registers mouse events for this table and its header.
	 *
	 * @see #removeLyJTableMouseListener()
	 */
	public void addLyJTableMouseListener()
	{
		if (_mouseListener != null)
		{
			super.addMouseListener(_mouseListener);

			getTableHeader().addMouseListener(_mouseListener);
		}
	}

	/**
	 * Registers selection events for this table.
	 *
	 * @see #removeLyJTableSelectionListener()
	 */
	public void addLyJTableSelectionListener()
	{
		if (_selectionListener != null)
			getSelectionModel().addListSelectionListener(_selectionListener);
	}

	/**
	 * Unregisters mouse events for this table and its header.
	 *
	 * @see #addLyJTableMouseListener()
	 */
	public void removeLyJTableMouseListener()
	{
		super.removeMouseListener(_mouseListener);
		getTableHeader().removeMouseListener(_mouseListener);
	}

	/**
	 * Unregisters selection events for this table.
	 *
	 * @see #addLyJTableSelectionListener()
	 */
	public void removeLyJTableSelectionListener()
	{
		getSelectionModel().removeListSelectionListener(_selectionListener);
	}

	/**
	 * Gets the mouse events listener of this table.
	 *
	 * @return the mouse listener
	 */
	public MouseListener getLyJTableMouseListener()
	{
		return _mouseListener;
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.Scrollable#getPreferredScrollableViewportSize()
	 */
	public Dimension getPreferredScrollableViewportSize()
	{
		if ((_minHeight > 0) && (_preferredScrollableSize != null))
			return _preferredScrollableSize;

		Dimension dim = getCustomizedScrollablePreferredSize();
		int		  defaultHeight = dim.height;

		dim.width += _rightMargin;
		
		if (_displayedHeight > 0)
			dim.height = _displayedHeight * getRowHeight();
		else
			dim.height = Math.max(_minHeight * getRowHeight(), dim.height);

		int		  displayedRows = dim.height / getRowHeight();
		
		// If height is lower than default height, add extra space to avoid
		// horizontal scrollbars.
		if ((dim.height < defaultHeight)
				|| (LyJTableModel.EMPTY_ROW_COUNT < displayedRows))
		{
			JScrollBar	vBar = ((JScrollPane)getParent().getParent()).getVerticalScrollBar();
			
			if (vBar != null)
				dim.width += vBar.getPreferredSize().getWidth();
		}

		if (_minChars > 0)
			dim.width = Math.max(dim.width,
					_minChars * LySwingUtils.getFontCharWidth(this, getFont()));

		_preferredScrollableSize = dim;

		return dim;
	}

	/**
	 * Set the minimum number of char ro display.
	 *
	 * @param chars the minimum number of char to display
	 */
	public void setMinChars(int chars)
	{
		_minChars = chars;
	}

	/**
	 * Sets the right margin.
	 *
	 * @param margin the new right margin
	 */
	public void setRightMargin(int margin)
	{
		_rightMargin = margin;
	}

	/**
	 * Sets the minimum height.
	 *
	 * @param height the new minimum height
	 */
	public void setMinHeight(int height)
	{
		_minHeight = height;
	}

	/**
	 * Sets the displayed height.
	 *
	 * @param height the new displayed height
	 */
	public void setDisplayedHeight(int height)
	{
		_displayedHeight = height;
	}

	/**
	 * Doesn't consider the last cosmetic column in layout computing and ignores the empty rows.
	 *
	 * @see java.awt.Component#getSize()
	 */
	public Dimension getSize()
	{
		Dimension   dim = super.getSize();
		TableColumn lastColumn = getColumn(LyJTableModel.EMPTY_CELL_STATUS_VALUE);
		int		    emptyRowsHeight = getTableModel().getEmptyRowCount() * getRowHeight();

		if (lastColumn == null)
			return new Dimension(dim.width, dim.height - emptyRowsHeight);
		int width = dim.width - lastColumn.getWidth();

		return new Dimension(width, dim.height - emptyRowsHeight);
	}

	/**
	 * Determines the part of the table that must activate scrollbars apparition.
	 *
	 * @return the size of the data part of the table
	 * @see    leon.view.win.swing.table.LyJCustomizedScrollable#getCustomizedScrollablePreferredSize()
	 */
	public Dimension getCustomizedScrollablePreferredSize()
	{
		TableColumn lastColumn = getColumn(LyJTableModel.EMPTY_CELL_STATUS_VALUE);
		int		    lastWidth = (lastColumn != null) ? (lastColumn.getWidth()) : 0;
		int		    emptyRowsHeight = 0;
		int		    n = getTableModel().getEmptyRowCount();

		emptyRowsHeight = getRowHeight() * n;

		return (new Dimension(getPreferredSize().width - lastWidth,
				getPreferredSize().height
				- emptyRowsHeight));
	}

	/**
	 * Gets this table data model.
	 *
	 * @return the data model
	 */
	public LyJTableModel getTableModel()
	{
		return (LyJTableModel)getModel();
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.JTable#createDefaultColumnsFromModel()
	 */
	public void createDefaultColumnsFromModel()
	{
		TableModel m = getModel();

		if (m != null)
		{
			// Remove any current columns
			TableColumnModel cm = getColumnModel();

			while (cm.getColumnCount() > 0)
				cm.removeColumn(cm.getColumn(0));

			// Create new columns from the data model info
			for (int i = 0; i < m.getColumnCount(); i++)
			{
				LyJTableColumn newColumn = new LyJTableColumn(i);

				addColumn(newColumn);
			}
		}
	}

	/**
	 * Scrolls this table to show the row at the specified index.
	 *
	 * @param row the index of the row
	 */
	public void scrollToRow(int row)
	{
		JScrollBar vBar;

		try
		{
			vBar = ((JScrollPane)getParent().getParent()).getVerticalScrollBar();
		}
		catch (Exception e)
		{
			return;
		}
		Object		  lock = null;

		LyJTableModel tableModel = (LyJTableModel)getModel();

		if (tableModel != null)
			lock = tableModel.getLock();
		else
			lock = getTreeLock();
		synchronized (lock)
		{
			int vBarValue = vBar.getValue();
			int i, totalHeight = 0, nextTotalHeight = 0;

			for (i = 0; i < row; i++)
				totalHeight += getRowHeight(i);

			nextTotalHeight = totalHeight + getRowHeight(row);

			// Scroll up if needed
			if (totalHeight < vBarValue)
				vBar.setValue(totalHeight);

			// Scroll down if needed
			if (nextTotalHeight > (getVisibleRect().height + vBarValue))
			{
				// If scroll must be greater than scrollbar maximum value, revalidate
				// scroll pane in order to ensure that scrollbar will be updated.
				if (nextTotalHeight > vBar.getMaximum())
				{
					((JScrollPane)getParent().getParent()).invalidate();
					((JScrollPane)getParent().getParent()).validate();
					((JScrollPane)getParent().getParent()).doLayout();
				}

				vBar.setValue(nextTotalHeight - getVisibleRect().height);
			}
		}
	} // end method scrollToRow

	/**
	 * Scrolls this table to show the column at the specified index.
	 *
	 * @param column the index of the column
	 */
	public void scrollToColumn(int column)
	{
		JScrollBar hBar;

		try
		{
			hBar = ((JScrollPane)getParent().getParent()).getHorizontalScrollBar();
		}
		catch (Exception e)
		{
			return;
		}

		synchronized (getTreeLock())
		{
			int hBarValue = hBar.getValue();
			int i, totalWidth = 0, nextTotalWidth = 0;

			for (i = 0; i < column; i++)
				totalWidth += getColumnModel().getColumn(i).getPreferredWidth();

			nextTotalWidth = totalWidth + getColumnModel().getColumn(column).getWidth();

			// Scroll left if needed
			if (totalWidth < hBarValue)
				hBar.setValue(totalWidth);

			// Scroll down if needed
			if (nextTotalWidth > (getVisibleRect().width + hBarValue))
			{
				// If scroll must be greater than scrollbar maximum value, revalidate
				// scroll pane in order to ensure that scrollbar will be updated.
				if (nextTotalWidth > hBar.getMaximum())
				{
					((JScrollPane)getParent().getParent()).invalidate();
					((JScrollPane)getParent().getParent()).validate();
					((JScrollPane)getParent().getParent()).doLayout();
				}

				hBar.setValue(nextTotalWidth - getVisibleRect().width);
			}
		}
	} // end method scrollToColumn

	/*
	 * (non-Javadoc)
	 * @see javax.swing.JTable#getCellRenderer(int, int)
	 */
	public TableCellRenderer getCellRenderer(int row, int column)
	{
		if (_cellRenderer != null)
			return _cellRenderer;
		return super.getCellRenderer(row, column);
	}

	/**
	 * Sets the cell renderer of this table.
	 *
	 * @param cellRenderer the cell renderer
	 * @see   #getCellRenderer(int, int)
	 */
	public void setCellRenderer(TableCellRenderer cellRenderer)
	{
		_cellRenderer = cellRenderer;
	}

	/**
	 * Sets the header identifiers.
	 *
	 * @param names the array of columns identifiers
	 */
	public void setHeaderNames(String[] names)
	{
		((DefaultTableModel)getModel()).setColumnIdentifiers(names);
	}

	/**
	 * Indicates whether table has at least one editable cell.
	 *
	 * @return true if table has at least one editable cell, false otherwise.
	 */
	public boolean hasEditableCells()
	{
		return _hasEditableCells;
	}

	/**
	 * Sets the value of the parameter that indicates if table has at least one editable cell.
	 *
	 * @param hasEditableCells the new parameter status.
	 */
	public void setEditableCells(boolean hasEditableCells)
	{
		_hasEditableCells = hasEditableCells;
	}

	/**
	 * Edit the next editable cell of this table.
	 *
	 * @param next indicates if next or previous cell must be edited.
	 */
	public void editNextCell(boolean next)
	{
		int     editedRow = getEditingRow();
		int     editedColumn = getEditingColumn();
		int     nbRows = getRowCount();
		int     nbColumns = getColumnCount();
		int     row = editedRow;
		int     step = (next ? 1 : -1);
		int     column = editedColumn + step;
		boolean stop = false;

		for (; !stop && ((next && (row < nbRows)) || (!next && (row >= 0))); row = row + step)
		{
			for (; !stop && ((next && (column < nbColumns)) || (!next && (column >= 0)));
					column = column + step)
			{
				if (isCellEditable(row, column))
				{
					editCellAt(row, column);

					Component comp = getComponent(1);

					comp = getFocusableComponent(comp);

					if ((comp != null) && comp.isEnabled())
					{
						comp.requestFocusInWindow();
						stop = true;
					}
				}
			}

			column = (next ? 0 : (nbColumns - 1));
		}
	} // end method editNextCell

	/**
	 * Returns the first focusable component from the given component. Focusable component are
	 * instances of JTextComponent or JComboBox.
	 *
	 * @param  component the root component used to retrieve focusable component.
	 * @return the first focusable component from the given component.
	 */
	protected Component getFocusableComponent(Component component)
	{
		if (component instanceof Container)
		{
			Container   container = (Container)component;
			Component[] subComponents = container.getComponents();
			Component   subComponent;
			Component   focusableComponent;
			int		    i;

			for (i = 0; i < subComponents.length; i++)
			{
				subComponent = subComponents[i];
				focusableComponent = getFocusableComponent(subComponent);

				if (focusableComponent != null)
					return focusableComponent;
			}
		}

		if ((component instanceof JTextComponent)
				|| (component instanceof JComboBox)
				|| (component instanceof JCheckBox))
			return component;

		return null;
	}

	/**
	 * Sets the header background color.
	 *
	 * @param color the new background color for the header
	 */
	public void setHeaderColor(Color color)
	{
		getTableHeader().setBackground(color);
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent e)
	{
		if (_actionTransmitter != null)
			_actionTransmitter.processActionEvent(e);
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.event.LySelectionListener#selectionPerformed(leon.view.event.LySelectionEvent)
	 */
	public void selectionPerformed(LySelectionEvent e)
	{
		if (_selectionTransmitter != null)
			_selectionTransmitter.processSelectionEvent(e);
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();
		return _actionTransmitter;
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		if (_selectionTransmitter == null)
			_selectionTransmitter = new LySimpleSelectionTransmitter();
		return _selectionTransmitter;
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.JTable#createDefaultTableHeader()
	 */
	protected JTableHeader createDefaultTableHeader()
	{
		return (new _LyJTableHeader());
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#getToolTipText(java.awt.event.MouseEvent)
	 */
	public String getToolTipText(MouseEvent e)
	{
		String		   tip = null;
		java.awt.Point p = e.getPoint();
		int			   rowIndex = rowAtPoint(p);
		int			   colIndex = columnAtPoint(p);

		LyJTableModel  tableModel = getTableModel();
		Object		   comp = tableModel.getValueAt(rowIndex, colIndex);

		if (comp instanceof LySwingCell)
		{
			tip = ((LySwingCell)comp).getTooltip();
			tip = LyTools.formatToHTML(tip, LyJTableComponent.TIP_MAX_CHARS);
		}

		return tip;
	}

	/**
	 * Specifies that the height must be computed for each row of the table.
	 *
	 * @param compute Indicates that height must be computed for each row of the table.
	 */
	public void setComputeHeight(boolean compute)
	{
		_computeHeight = compute;
	}

	/**
	 * Indicates that height must be computed for each row of the table.
	 *
	 * @return <code>true</code> if height must be computed for each row of the table, <code>
	 *         false</code> otherwise
	 */
	public boolean isComputeHeight()
	{
		return _computeHeight;
	}

	/**
	 * Automatically computes the heights of rows.
	 */
	public void cacheHeight()
	{
		TableModel		  m = getModel();
		int				  r, rowCount = m.getRowCount();
		int				  c, columnCount = m.getColumnCount();
		TableCellRenderer renderer;
		Component		  comp;

		@SuppressWarnings("unchecked")
		Hashtable<String, Integer>[] temp = (Hashtable<String, Integer>[])new Hashtable[rowCount];
		_cellHeights = temp;

		for (r = 0; r < rowCount; r++)
		{
			_cellHeights[r] = new Hashtable<String, Integer>();

			for (c = 0; c < columnCount; c++)
			{
				renderer = getCellRenderer(r, c);
				
				comp = renderer.getTableCellRendererComponent(this, m.getValueAt(r, c), false,
						false, r, c);

				if (comp instanceof _LyJMultiCell)
				{
					_LyJMultiCell cell = (_LyJMultiCell)comp;

					cell.computeCellHeights(_cellHeights[r], cell.getCell(), null, this, getFont());
				}
			}
		}
	} // end method cacheHeight

	/**
	 * Computes the height for a given row and a given level.
	 *
	 * @param  row   index of the row to be computed
	 * @param  level level to be computed
	 * @return the height of the given row and level
	 */
	public int computeLevelHeight(int row, String level)
	{
		if ((row < 0) || (_cellHeights == null) || (row >= _cellHeights.length))
			return -1;

		if (level == null)
			level = "";

		int		  height = 0;
		Hashtable<String, Integer> cellHeights = _cellHeights[row];

		if (cellHeights == null)
			return -1;

		Integer     		computedHeightValue = (Integer)cellHeights.get(level);
		int		    		computedHeight = 0;
		Enumeration<String>	levels = cellHeights.keys();
		String	    		subLevel, subLevelTmp;
		int		    		pos;

		if (computedHeightValue != null)
			computedHeight = computedHeightValue.intValue();

		while (levels.hasMoreElements())
		{
			subLevel = (String)levels.nextElement();
			pos = subLevel.lastIndexOf(LyTools.PIPE_SEP);

			if (pos >= 0)
			{
				subLevelTmp = subLevel.substring(0, pos);

				if (level.equals(subLevelTmp))
					height += computeLevelHeight(row, subLevel);
			}
		}

		return Math.max(height, computedHeight);
	} // end method computeLevelHeight

	public void paint(Graphics g)
	{
		synchronized (getTreeLock())
		{
			super.paint(g);
			
	        Rectangle clip = g.getClipBounds();
	    	Point upperLeft = clip.getLocation();
	    	Point lowerRight = new Point(clip.x + clip.width - 1, clip.y + clip.height - 1);
            int rMin = rowAtPoint(upperLeft);
            int rMax = rowAtPoint(lowerRight);
            
            if (rMin == -1)
            	rMin = 0;

            if (rMax == -1)
            	rMax = getRowCount()-1;
            
            if (rMax >= getTableModel().getDataRowCount())
            	rMax = getTableModel().getDataRowCount() - 1;
            
            for (int rowIndex = rMin; rowIndex <= rMax; rowIndex++)
            {
            	LySwingRow	swingRow = getTableModel().getRowAt(rowIndex);
            	LyRow		row = (LyRow)swingRow.getViewComponent();
            	
            	if ((row.getRows() != null) && (row.getChildCount() == 1))
            	{
    				String aaText = System.getProperty("swing.aatext");
    				Font   defaultFont = getFont();
    				Color  defaultColor = getBackground();
    				Color  defaultLabelColor = getForeground();

    				// set antialias
    				if ("true".equalsIgnoreCase(aaText) && (g instanceof Graphics2D))
    					((Graphics2D)g).setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
    						RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
					
    				LyEnvironment	environment = row.getEnvironment();
    				String			label = (row.getLabel() == null ? null : environment.translate(row.getLabel()));
					Font			font = (row.getFont() != null ? (Font)environment.getFont(row.getFont()) : null);
					Color			color = (Color)environment.getColor(row.getColor());
					Color			labelColor = (Color)environment.getColor(row.getFontColor());
					Image			arrow = (row.isOpen() ? swingRow.getOpenRowImage() : swingRow.getClosedRowImage());

					if (font == null)
						font = defaultFont;
					
					if (color == null)
						color = defaultColor;
					
					if (labelColor == null)
						labelColor = defaultLabelColor;
					
    				Rectangle			rect = getCellRect(rowIndex, 0, false);
					FontMetrics			fm = getFontMetrics(font);
					int					offset = LyJCellRenderer.HIERARCHY_OFFSET_WIDTH * swingRow.getDepth();
					
					g.setFont(font);
					g.setColor(color);
					g.fillRect(offset, rect.y, getSize().width, rect.height);
					g.drawImage(arrow, offset + 1, rect.y + (rect.height - arrow.getHeight(this)) / 2, this);
					
					if (label != null)
					{
						g.setColor(labelColor);
						g.drawString(label, offset + LyJCellRenderer.HIERARCHY_OFFSET_WIDTH, rect.y + (rect.height - fm.getHeight()) / 2 + fm.getAscent());
					}
            	}
            }
		}
	}

	/**
	 * Handles table headers.
	 */
	public class _LyJTableHeader extends JTableHeader implements MouseListener
	{
		private static final long			   serialVersionUID = 3889010625956956055L;

		/** List of the header groups. */
		protected Vector<_LyJTableHeaderGroup> _headerGroups;

		/** The depth of the header groups. */
		protected int						   _depth;

		/** Height of a header group. */
		protected int						   _groupHeight;
		
		/** Indicates whether mouse is in the header. */
		protected boolean					   _mouseIn;

		/**
		 * Creates a new header object.
		 */
		public _LyJTableHeader()
		{
			super();

			_headerGroups = null;
			_depth = 0;
			_groupHeight = 0;
			_mouseIn = false;
			addMouseListener(this);
		}

		/*
		 * (non-Javadoc)
		 * @see javax.swing.JComponent#getToolTipText(java.awt.event.MouseEvent)
		 */
		public String getToolTipText(MouseEvent event)
		{
			// If the table contains header groups, check if the mouse is over a group
			if (_headerGroups != null)
			{
				int					 i, n = _headerGroups.size();
				_LyJTableHeaderGroup group;
				Component			 comp;
				Rectangle			 r;

				for (i = 0; i < n; i++)
				{
					group = _headerGroups.elementAt(i);
					comp = group.getComponent();
					r = comp.getBounds();

					if (r.contains(event.getPoint()))
						return group.getTooltip();
				}
			}

			/* activate tooltips */
			// FFT01507
			int			   idx = getTable().getColumnModel().getColumnIndexAtX(event.getX());
			LyJTableColumn col = (LyJTableColumn)(getTable().getColumnModel().getColumn(idx));

			return ((col == null) ? null : col.getTooltip());
		} // end method getToolTipText

		/**
		 * Adds a component representing a header group.
		 *
		 * @param comp     the component
		 * @param startCol the starting column for the header group
		 * @param endCol   the ending column for the header group
		 * @param depth    the depth of the header group
		 * @param tooltip  the tooltip of the header group
		 * @see   _LyJTableHeaderGroup
		 */
		public void addHeaderGroup(Component comp, int startCol, int endCol, int depth,
			String tooltip)
		{
			add(comp);

			Dimension d = comp.getPreferredSize();

			_depth = Math.max(_depth, depth);
			_groupHeight = Math.max(_groupHeight, d.height);

			comp.setBounds(0, d.height * (depth - 1), d.width, d.height);

			_LyJTableHeaderGroup group = new _LyJTableHeaderGroup(comp, startCol, endCol, depth,
					tooltip);

			if (_headerGroups == null)
				_headerGroups = new Vector<_LyJTableHeaderGroup>();

			_headerGroups.addElement(group);

			updateHeaderGroupsPosition();
		}

		/**
		 * Updates the header groups positions.
		 *
		 * @see #updateHeaderGroupsPosition()
		 * @see javax.swing.event.TableColumnModelListener#columnMarginChanged(javax.swing.event.ChangeEvent)
		 */
		public void columnMarginChanged(ChangeEvent e)
		{
			super.columnMarginChanged(e);

			updateHeaderGroupsPosition();
			
			if (_mouseIn)
			{
				LyJTableColumnModel	cm = (LyJTableColumnModel)getColumnModel();
				LySwingTable		swingTable = cm.getSwingTable();
				int					i, count = cm.getColumnCount();
				Rectangle			rect;
				StringBuffer		buffer = new StringBuffer();
				int					charWidth = LySwingUtils.getFontCharWidth(this, getFont());
				
				for (i = 0; i < count; i++)
				{
					if ((i == 0) && swingTable.hasStatusColumn())
						continue;

					if (buffer.length() > 0)
						buffer.append("##");
	
					rect = getHeaderRect(i);
					buffer.append(((float)rect.width) / charWidth);
				}
				
				swingTable.setNeedResizeColumns(false);
				
				LyActionEvent actionEvent = new LyActionEvent(this, "resizeColumn", buffer.toString(), LyActionDemand.TYPE_ACTION);

				actionPerformed(actionEvent);
			}
		}

		/**
		 * Updates the header groups positions.
		 */
		protected void updateHeaderGroupsPosition()
		{
			if ((_headerGroups == null) || (_headerGroups.size() == 0))
				return;

			int					 i, n = _headerGroups.size();
			_LyJTableHeaderGroup group;
			Component			 comp;
			Rectangle			 r;
			int					 startColumn, endColumn, startX, endX;

			// For each group, compute its new x position and width
			for (i = 0; i < n; i++)
			{
				group = _headerGroups.elementAt(i);
				comp = group.getComponent();
				r = comp.getBounds();
				startColumn = group.getStartColumn();
				endColumn = group.getEndColumn();
				startX = getHeaderRect(startColumn).x;
				endX = getHeaderRect(endColumn).x;
				comp.setBounds(startX, r.y, endX - startX, r.height);
			}
		}

		/**
		 * Gets the preferred size. Takes into account the header groups.
		 */
		public Dimension getPreferredSize()
		{
			synchronized (getTreeLock())
			{
				if (_depth == 0)
					return super.getPreferredSize();
				return new Dimension(super.getPreferredSize().width, _groupHeight * (_depth + 1));
			}
		}

		public void mouseClicked(MouseEvent e)
		{
		}

		public void mousePressed(MouseEvent e)
		{
		}

		public void mouseReleased(MouseEvent e)
		{
		}

		public void mouseEntered(MouseEvent e)
		{
			_mouseIn = true;
		}

		public void mouseExited(MouseEvent e)
		{
			_mouseIn = false;
		}
	} // end class _LyJTableHeader


	/**
	 * Defines a header group.
	 */
	class _LyJTableHeaderGroup
	{

		/** The header group. */
		Component _comp;

		/** The starting column. */
		int		  _startCol;

		/** The ending column. */
		int		  _endCol;

		/** The depth of this header group. */
		int		  _depth;

		/** The tooltip of this header group. */
		String    _tooltip;

		/**
		 * Creates a new header group with the specified attributes.
		 *
		 * @param comp     the component
		 * @param startCol the starting column for the header group
		 * @param endCol   the ending column for the header group
		 * @param depth    the depth of the header group
		 * @param tooltip  the tooltip of the header group
		 */
		_LyJTableHeaderGroup(Component comp, int startCol, int endCol, int depth, String tooltip)
		{
			_comp = comp;
			_startCol = startCol;
			_endCol = endCol;
			_depth = depth;
			_tooltip = tooltip;
		}

		/**
		 * Gets the header group component.
		 *
		 * @return the header group
		 */
		public Component getComponent()
		{
			return _comp;
		}

		/**
		 * Gets the starting column index.
		 *
		 * @return the starting column
		 */
		public int getStartColumn()
		{
			return _startCol;
		}

		/**
		 * Gets the ending column index.
		 *
		 * @return the ending column
		 */
		public int getEndColumn()
		{
			return _endCol;
		}

		/**
		 * Gets the depth of this table group.
		 *
		 * @return the depth of this table group
		 */
		public int getDepth()
		{
			return _depth;
		}

		/**
		 * Gets the tooltip of this table group.
		 *
		 * @return the tooltip
		 */
		public String getTooltip()
		{
			return _tooltip;
		}
	} // end class _LyJTableHeaderGroup


	/**
	 * Provides drag handling as well as drag gestures.
	 */
	class _LyTableCanvasDragListener
		implements DragGestureListener,
			DragSourceListener
	{

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DragSourceListener#dragDropEnd(java.awt.dnd.DragSourceDropEvent)
		 */
		public void dragDropEnd(DragSourceDropEvent dsde)
		{
			LyActionEvent actionEvent = new LyActionEvent(this, null,
					String.valueOf(dsde.getDropAction()), LyDragDropDemand.TYPE_DRAG_FINISHED);

			actionPerformed(actionEvent);
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.dnd.DragSourceListener#dragEnter(java.awt.dnd.DragSourceDragEvent)
		 */
		public void dragEnter(DragSourceDragEvent dsde)
		{
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.dnd.DragSourceListener#dragExit(java.awt.dnd.DragSourceEvent)
		 */
		public void dragExit(DragSourceEvent dse)
		{
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.dnd.DragSourceListener#dragOver(java.awt.dnd.DragSourceDragEvent)
		 */
		public void dragOver(DragSourceDragEvent dsde)
		{
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.dnd.DragSourceListener#dropActionChanged(java.awt.dnd.DragSourceDragEvent)
		 */
		public void dropActionChanged(DragSourceDragEvent dsde)
		{
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DragGestureListener#dragGestureRecognized(java.awt.dnd.DragGestureEvent)
		 */
		public void dragGestureRecognized(DragGestureEvent dge)
		{
			java.awt.Point p = dge.getDragOrigin();
			int			   rowIndex = rowAtPoint(p);
			int			   colIndex = columnAtPoint(p);

			if ((rowIndex >= 0) && (rowIndex < getTableModel().getDataRowCount()))
			{
				Transferable t = new StringSelection(rowIndex + " " + colIndex);

				try
				{
				    _dragSource.startDrag(dge, DragSource.DefaultCopyDrop, t, this);
				}
				catch (Exception e)
				{
				    // If anything occurred, ignore drag
				    return;
				}

				Vector<String>   selectedIds = ((LyJTableSelectionListener)_selectionListener)
						.getSelectedIds();
				LySelectionEvent select = new LySelectionEvent(this, selectedIds, true);

				select.addIndex(rowIndex);
				select.setSelectionSize(selectedIds.size());
				select.setExtend(true);

				selectionPerformed(select);

				LyActionEvent actionEvent = new LyActionEvent(this, null, null,
						LyDragDropDemand.TYPE_DRAG_START);

				actionEvent.addIndex(rowIndex);
				actionEvent.addIndex(colIndex);
				actionPerformed(actionEvent);
			}
		} // end method dragGestureRecognized
	} // end class _LyTableCanvasDragListener


	/**
	 * Provides drap and drop handling.
	 */
	class _LyTableCanvasDropListener
		implements DropTargetListener
	{

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#dragEnter(java.awt.dnd.DropTargetDragEvent)
		 */
		public void dragEnter(DropTargetDragEvent dtde)
		{
			LyActionEvent actionEvent = new LyActionEvent(this, null, null,
					LyDragDropDemand.TYPE_DRAG_ENTER);

			actionPerformed(actionEvent);
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#dragExit(java.awt.dnd.DropTargetEvent)
		 */
		public void dragExit(DropTargetEvent dte)
		{
			LyActionEvent actionEvent = new LyActionEvent(this, null, null,
					LyDragDropDemand.TYPE_DRAG_LEAVE);

			actionPerformed(actionEvent);
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#dragOver(java.awt.dnd.DropTargetDragEvent)
		 */
		public void dragOver(DropTargetDragEvent dtde)
		{
			LyActionEvent actionEvent = new LyActionEvent(this, null, null,
					LyDragDropDemand.TYPE_DRAG_OVER);

			actionEvent.setParameter(dtde);
			actionPerformed(actionEvent);
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#drop(java.awt.dnd.DropTargetDropEvent)
		 */
		public void drop(DropTargetDropEvent dtde)
		{
			java.awt.Point p = dtde.getLocation();
			int			   rowIndex = rowAtPoint(p);
			int			   colIndex = columnAtPoint(p);

			dtde.acceptDrop(DnDConstants.ACTION_COPY_OR_MOVE);
			dtde.getDropTargetContext().dropComplete(true);

			LyActionEvent actionEvent = new LyActionEvent(this, null, null,
					LyDragDropDemand.TYPE_DROP);

			actionEvent.addIndex(rowIndex);
			actionEvent.addIndex(colIndex);
			actionPerformed(actionEvent);
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#dropActionChanged(java.awt.dnd.DropTargetDragEvent)
		 */
		public void dropActionChanged(DropTargetDragEvent dtde)
		{
			LyActionEvent actionEvent = new LyActionEvent(this, null, null,
					LyDragDropDemand.TYPE_DRAG_OPERATION_CHANGED);

			actionPerformed(actionEvent);
		}
	} // end class _LyTableCanvasDropListener
} // end class LyJTableComponent
