/*
 * LyJTextArea.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.text.*;
import leon.misc.*;
import leon.view.*;


/**
 * This class manages a text area with a popup menu.
 *
 * @version $Id: LyJTextArea.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class LyJTextArea extends JTextArea
{
	private static final long  serialVersionUID = -3138009511979249486L;

	public static final String __VERSION =
		"$Id: LyJTextArea.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The cancel string. */
	public static String	   CANCEL;

	/** The cut string. */
	public static String	   CUT;

	/** The copy string. */
	public static String	   COPY;

	/** The paste string. */
	public static String	   PASTE;

	/** The delete string. */
	public static String	   DELETE;

	/** The select all string. */
	public static String	   SELECT_ALL;

	/** The cancel item. */
	protected JMenuItem		   _cancelAction;

	/** The cut item. */
	protected JMenuItem		   _cutAction;

	/** The copy item. */
	protected JMenuItem		   _copyAction;

	/** The paste item. */
	protected JMenuItem		   _pasteAction;

	/** The delete item. */
	protected JMenuItem		   _deleteAction;

	/** The select all item. */
	protected JMenuItem		   _selectAllAction;

	/** The menu. */
	protected JPopupMenu	   _menu;
	
	/** The environment of the area. */
	protected LyEnvironment	   _environment;

	/**
	 * Creates a default text area with an empty text area.
	 *
	 * @param env the current environment to access the resources
	 */
	public LyJTextArea(LyEnvironment env)
	{
		super();
		init(env);
	}

	/**
	 * Creates a text area with the specified attributes.
	 *
	 * @param doc     the document model that rules input
	 * @param text    the initial text of the text field
	 * @param rows    the number of rows of this text field
	 * @param columns the number of columns of this text field
	 * @param env     the current environment to access the resources
	 */
	public LyJTextArea(Document doc, String text, int rows, int columns, LyEnvironment env)
	{
		super(doc, text, rows, columns);
		setText(text);
		init(env);
	}

	/**
	 * Creates a text area with the specified attributes. By default, the text area is empty and the
	 * number of rows and columns is 0.
	 *
	 * @param doc the document model that rules input
	 * @param env the current environment to access the resources
	 */
	protected LyJTextArea(Document doc, LyEnvironment env)
	{
		super(doc);
		init(env);
	}

	/**
	 * Creates a text area with the specified attributes. By default, the text area is empty and the
	 * default document model is used.
	 *
	 * @param rows    the number of rows of this text field
	 * @param columns the number of columns of this text field
	 * @param env     the current environment to access the resources
	 */
	public LyJTextArea(int rows, int columns, LyEnvironment env)
	{
		super(rows, columns);
		init(env);
	}

	/**
	 * Creates a text area with the specified attributes. By default, the default document model is
	 * used and the number of rows and columns is 0.
	 *
	 * @param text the initial text of the text field
	 * @param env  the current environment to access the resources
	 */
	public LyJTextArea(String text, LyEnvironment env)
	{
		super(text);
		init(env);
	}

	/**
	 * Creates a text area with the specified attributes. The default document model is used.
	 *
	 * @param text    the initial text of the text field
	 * @param rows    the number of rows of this text field
	 * @param columns the number of columns of this text field
	 * @param env     the current environment to access the resources
	 */
	public LyJTextArea(String text, int rows, int columns, LyEnvironment env)
	{
		super(text, rows, columns);
		init(env);
	}

	/**
	 * Initializes this component and its attributes in the specified environment.
	 *
	 * @param env the current environment
	 */
	public void init(LyEnvironment env)
	{
		CANCEL = env.translate("LY_ITEM_CANCEL");
		CUT = env.translate("LY_ITEM_CUT");
		COPY = env.translate("LY_ITEM_COPY");
		PASTE = env.translate("LY_ITEM_PASTE");
		DELETE = env.translate("LY_ITEM_DELETE");
		SELECT_ALL = env.translate("LY_ITEM_SELECT_ALL");
		_menu = createPopupMenu();
		addMouseListener(new _LyPopupInvoker());
		_environment = env;
		
		// FFT00974
		Insets insets = getMargin();

		insets = new Insets(insets.top, insets.left + 1, insets.bottom, insets.right + 1);
		setMargin(insets);
	}

	/**
	 * Executes the paste action if this text area is editable.
	 *
	 * @see javax.swing.text.JTextComponent#paste()
	 */
	public void paste()
	{
		if (isEditable() && (_environment != null))
		{
			requestFocus();
			
			_LyPasteInvoker	pasteInvoker = new _LyPasteInvoker();
            LyViewManager	viewManager = (LyViewManager)_environment.getViewManager();
            
			viewManager.execGui(pasteInvoker);
		}
	}

	/**
	 * Call the paste method of the super class.
	 */
	public void superPaste()
	{
		super.paste();
	}

	/**
	 * Executes the delete action on the selected text.
	 */
	public void delete()
	{
		String text = getText();
		String selectedText = getSelectedText();

		if ((text == null) || (selectedText == null))
			return;
		int    offset = text.indexOf(selectedText);
		String newText = text.substring(0, offset) + text.substring(offset + selectedText.length());

		setText(newText);
	}

	/**
	 * Nothing is done for the undo action.
	 */
	public void undo()
	{
	}

	/**
	 * Creates the editing popup menu and initializes its actions.
	 *
	 * @return the popup menu of this text area
	 */
	public JPopupMenu createPopupMenu()
	{
		JPopupMenu menu = new LyJPopupMenu();

		_cancelAction = new JMenuItem(CANCEL);
		_cancelAction.addActionListener(new ActionListener()
			{

				/*
				 * (non-Javadoc)
				 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
				 */
				public void actionPerformed(ActionEvent e)
				{
					undo();
				}
			});
		_cutAction = new JMenuItem(CUT);
		_cutAction.addActionListener(new ActionListener()
			{

				/*
				 * (non-Javadoc)
				 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
				 */
				public void actionPerformed(ActionEvent e)
				{
					cut();
				}
			});
		_copyAction = new JMenuItem(COPY);
		_copyAction.addActionListener(new ActionListener()
			{

				/*
				 * (non-Javadoc)
				 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
				 */
				public void actionPerformed(ActionEvent e)
				{
					copy();
				}
			});
		_pasteAction = new JMenuItem(PASTE);
		_pasteAction.addActionListener(new ActionListener()
			{

				/*
				 * (non-Javadoc)
				 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
				 */
				public void actionPerformed(ActionEvent e)
				{
					paste();
				}
			});
		_deleteAction = new JMenuItem(DELETE);
		_deleteAction.addActionListener(new ActionListener()
			{

				/*
				 * (non-Javadoc)
				 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
				 */
				public void actionPerformed(ActionEvent e)
				{
					delete();
				}
			});
		_selectAllAction = new JMenuItem(SELECT_ALL);
		_selectAllAction.addActionListener(new ActionListener()
			{

				/*
				 * (non-Javadoc)
				 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
				 */
				public void actionPerformed(ActionEvent e)
				{
					selectAll();
				}
			});

		menu.add(_cancelAction);
		menu.addSeparator();
		menu.add(_cutAction);
		menu.add(_copyAction);
		menu.add(_pasteAction);
		menu.add(_deleteAction);
		menu.addSeparator();
		menu.add(_selectAllAction);

		return menu;
	} // end method createPopupMenu

	/**
	 * Gets the popup menu of this text area.
	 *
	 * @return the popup menu
	 */
	public JPopupMenu getMenu()
	{
		return _menu;
	}

	/**
	 * Shows the popup menu at the specified (x,y) position in the specified component.
	 *
	 * @param comp the invoking component
	 * @param x    the x-coordinate to show the popup menu
	 * @param y    the y-coordinate to show the popup menu
	 */
	public void showMenu(Component comp, int x, int y)
	{
		boolean copy = (getSelectedText() != null);
		boolean paste = (isEditable());
		boolean cut = copy && paste;
		boolean delete = cut;
		boolean selectAll = (getText() != "");
		boolean cancel = false;

		_cancelAction.setEnabled(cancel);
		_cutAction.setEnabled(cut);
		_copyAction.setEnabled(copy);
		_pasteAction.setEnabled(paste);
		_deleteAction.setEnabled(delete);
		_selectAllAction.setEnabled(selectAll);

		getMenu().show(comp, x, y);
	}

	/* *************************** Internal Classes ************************************/

	/**
	 * Defines the mouse event listeners for this popup menu.
	 */
	public class _LyPopupInvoker
		implements MouseListener
	{

		/**
		 * Sole constructor.
		 */
		public _LyPopupInvoker()
		{
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
		 */
		public void mousePressed(MouseEvent e)
		{
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
		 */
		public void mouseReleased(MouseEvent e)
		{
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
		 */
		public void mouseClicked(MouseEvent e)
		{
			if (e.getModifiers() == InputEvent.BUTTON3_MASK)
			{
				requestFocus();
				showMenu(LyJTextArea.this, e.getX(), e.getY());
			}
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
		 */
		public void mouseExited(MouseEvent e)
		{
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
		 */
		public void mouseEntered(MouseEvent e)
		{
		}
	} // end class _LyPopupInvoker
	
	/**
	 * Implements the paste action for this popup menu.
	 */
	class _LyPasteInvoker implements Runnable
	{
		/**
		 * Sole constructor.
		 */
		_LyPasteInvoker()
		{
			super();
		}
		
		/*
		 * (non-Javadoc)
		 * @see java.lang.Thread#run()
		 */
		public void run()
		{
			int		end = LyJTextArea.this.getSelectionEnd();
			String	text = getText();
			int		textLength = (text == null ? 0 : text.length());

			LyJTextArea.this.superPaste();

			// Retrieve next caret position
			String	newText = getText();
			int		newTextLength = (newText == null ? 0 : newText.length());
			int		diff = newTextLength - textLength;
			int		newPosition = Math.max(0, Math.min(end + diff, newTextLength));
			
			setCaretPosition(newPosition);
		}
	} // end class _LyPasteInvoker
} // end class LyJTextArea
