/*
 * LyJTitlePanel.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import leon.misc.*;
import leon.view.event.*;
import leon.view.getinterface.*;


/**
 * Defines a title panel for the swing view.
 *
 * @version $Id: LyJTitlePanel.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     Panel
 */
public class LyJTitlePanel extends JPanel
	implements LyGetManagerInterface
{
	private static final long	    serialVersionUID = 4300016588576740345L;

	public static final String	    __VERSION =
		"$Id: LyJTitlePanel.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The common source code for Swing management. */
	protected _LyJTitlePanelManager _manager;

	/** The maximum size of a tooltip line. */
	public static int			    TIP_MAX_CHARS = 60;

	   // Default min width
    public static int               DEFAULT_MIN_WIDTH = 400;
    
	/**
	 * Creates a new title panel with the specified attributes.
	 *
	 * @param bgImage           the background image
	 * @param image             the main image
	 * @param imageToolTip      the toolTip of the main image
	 * @param rightImage        the right image
	 * @param rightImageToolTip the toolTip of the right image
	 * @param leftLabel         the left label
	 * @param centerLabel       the center label
	 * @param rightLabel        the right label
	 * @param listener          the action listener
	 */
	public LyJTitlePanel(Image bgImage, Image image, String imageToolTip, Image rightImage,
		String rightImageToolTip, String leftLabel, String centerLabel, String rightLabel,
		LyActionListener listener)
	{
		super();
		_manager = new _LyJTitlePanelManager(bgImage, image, imageToolTip, rightImage,
				rightImageToolTip, leftLabel, centerLabel, rightLabel, this, listener);
		ToolTipManager.sharedInstance().registerComponent(this);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.swing.JComponent#paintComponent(java.awt.Graphics)
	 */
	public void paintComponent(Graphics g)
	{
		super.paintComponent(g);

		String aaText = System.getProperty("swing.aatext");

		// set antialias
		if ("true".equalsIgnoreCase(aaText) && (g instanceof Graphics2D))
			((Graphics2D)g).setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
				RenderingHints.VALUE_TEXT_ANTIALIAS_ON);

		_manager.paintCommon(g);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#getMinimumSize()
	 */
	public Dimension getMinimumSize()
	{
		return _manager.getMinimumSize();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#getPreferredSize()
	 */
	public Dimension getPreferredSize()
	{
		return _manager.getPreferredSize();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.getinterface.LyGetManagerInterface#getManager()
	 */
	public Object getManager()
	{
		return _manager;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.swing.JComponent#getToolTipText(java.awt.event.MouseEvent)
	 */
	public String getToolTipText(MouseEvent e)
	{
		Point     p = e.getPoint();
		Rectangle rect = _manager.getImageRectangle();
		Rectangle rightRect = _manager.getRightImageRectangle();

		if (rect.contains(p))
			return LyTools.formatToHTML(_manager.getImageToolTip(), LyJTitlePanel.TIP_MAX_CHARS);
		else if (rightRect.contains(p))
			return LyTools.formatToHTML(_manager.getRightImageToolTip(),
				LyJTitlePanel.TIP_MAX_CHARS);
		else
			return null;
	}

	/**
	 * This title panel can gain focus.
	 *
	 * @return <code>true</code>
	 * @see    java.awt.Component#isFocusable()
	 */
	public boolean isFocusable()
	{
		return false;
	}

	/* ******************************************************************************************/

	/**
	 * Implements the LyAbstractTitlePanelManager to manage offscreen image.
	 *
	 * @see LyAbstractTitlePanelManager
	 */
	public class _LyJTitlePanelManager extends LyAbstractTitlePanelManager
		implements MouseListener
	{

		/**
		 * Sole constructor. (For invocation by subclass constructors, typically implicit.)
		 */
		public _LyJTitlePanelManager()
		{
			super();
		}

		/**
		 * Creates a title panel manager and adds it mouse listeners.
		 *
		 * @param bgImage           the background image
		 * @param image             the main image of the titlebar
		 * @param imageToolTip      the main image tooltip
		 * @param rightImage        the right image of the titlebar
		 * @param rightImageToolTip the right image tooltip
		 * @param leftLabel         the left label on the title
		 * @param centerLabel       the centered label
		 * @param rightLabel        the right label
		 * @param comp              the associated graphical component
		 * @param listener          the action listener
		 */
		public _LyJTitlePanelManager(Image bgImage, Image image, String imageToolTip,
			Image rightImage,
			String rightImageToolTip, String leftLabel, String centerLabel, String rightLabel,
			Component comp, LyActionListener listener)
		{
			super(bgImage, image, imageToolTip, rightImage, rightImageToolTip, leftLabel,
				centerLabel, rightLabel, comp, listener);

			comp.addMouseListener(this);
		}

		/**
		 * This canvas manager doesn't have offscreen images.
		 *
		 * @param  offScreenSize the size of the image to return
		 * @return <code>null</code>
		 * @see    LyAbstractTitlePanelManager#getImageOffScreen(Dimension)
		 */
		public Image getImageOffScreen(Dimension offScreenSize)
		{
			return null;
		}

		/**
		 * This canvas manager doesn't have offscreen images. There is no vertical offset.
		 *
		 * @return <code>-1</code>
		 * @see    LyAbstractTitlePanelManager#getImageVerticalOffset()
		 */
		protected int getImageVerticalOffset()
		{
			return -1;
		}

		/**
		 * This canvas manager doesn't have offscreen images.
		 *
		 * @param  offScreenImage  the offscreen image
		 * @param  defaultGraphics the default graphics context to use
		 * @return the specified defaut graphics
		 * @see    LyAbstractTitlePanelManager#getTargetGraphics(Image, Graphics)
		 */
		public Graphics getTargetGraphics(Image offScreenImage, Graphics defaultGraphics)
		{
			return defaultGraphics;
		}

		/**
		 * This canvas manager doesn't have offscreen images. Nothing is done.
		 *
		 * @see LyAbstractTitlePanelManager#paintOnScreen(Image, Graphics)
		 */
		public void paintOnScreen(Image offScreenImage, Graphics onScreenGraphics)
		{
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
		 */
		public void mousePressed(MouseEvent e)
		{
			Point p = e.getPoint();

			if (e.getButton() != 3)
				return;
			Rectangle rect = getImageRectangle();

			if (rect.contains(p))
			{
				LyActionEvent event = new LyActionEvent(this, "leftImage", null,
						LyMenuDemand.TYPE_POPUP);

				event.setParameter(new int[] { p.x, p.y });
				_listener.actionPerformed(event);
			}
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
		 */
		public void mouseClicked(MouseEvent e)
		{
			Point     p = e.getPoint();
			Rectangle rect = getImageRectangle();

			if ((rect.contains(p)) && (e.getClickCount() == 2))
				_listener.actionPerformed(new LyActionEvent(_comp, "leftImage", null,
						LyActionDemand.TYPE_ACTION));

			rect = getRightLabelRectangle();
			if ((rect.contains(p)) && (e.getClickCount() == 2))
				_listener.actionPerformed(new LyActionEvent(_comp, "rightLabel", null,
						LyActionDemand.TYPE_ACTION));

		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
		 */
		public void mouseExited(MouseEvent e)
		{
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
		 */
		public void mouseReleased(MouseEvent e)
		{
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
		 */
		public void mouseEntered(MouseEvent e)
		{
		}

		/*
		 * (non-Javadoc)
		 * @see leon.view.win.swing.component.LyAbstractTitlePanelManager#getPreferredSize()
		 */
		public Dimension getPreferredSize()
		{
			int          w = 0, h = 0;
			
			if (_image != null)
			{
				w = _image.getWidth(null) + X_GAP;
				h = _image.getHeight(null) + (4 * Y_GAP);
			}

			if (_rightImage != null)
			{
				w += _rightImage.getWidth(null) + (2 * X_GAP);
				h = Math.max(h, _rightImage.getHeight(null) + (4 * Y_GAP));
			}

			FontMetrics fm = _comp.getFontMetrics(_titleFont);

			if (_leftLabel != null)
				w = w + fm.stringWidth(_leftLabel) + X_GAP;

			if (_centerLabel != null)
				w = w + fm.stringWidth(_centerLabel) + X_GAP;

			if (_rightLabel != null)
				w = w + fm.stringWidth(_rightLabel) + X_GAP;

			h = Math.max(h, fm.getHeight() + (4 * Y_GAP));

			if (_comp instanceof Container)
			{
				Component[]		components = ((Container)_comp).getComponents();
				Dimension		dim;
				int				i, n = (components == null ? 0 : components.length);
				
				for (i = 0; i < n; i++)
				{
					dim = components[i].getPreferredSize();
					w += dim.width + (2 * X_GAP);
					h = Math.max(h, dim.height + (4 * Y_GAP));
				}
			}

			return new Dimension(Math.min(w, DEFAULT_MIN_WIDTH), h);
		} // end method getPreferredSize
	} // end class _LyJTitlePanelManager

} // end class LyJTitlePanel
