/*
 * LyJToggleButton.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import leon.misc.*;
import leon.view.filters.*;
import leon.view.event.*;
import leon.view.getinterface.*;


/**
 * This class implements a simple label with an image.
 *
 * @version $Id: LyJToggleButton.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     leon.view.win.swing.component.LyJIconJLabel
 */
public class LyJToggleButton extends JToggleButton
	implements ActionListener,
		LyGetIdInterface,
		LyEventTransmitter
{
	private static final long     serialVersionUID = 182152707748582059L;

	public static final String    __VERSION =
		"$Id: LyJToggleButton.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The maximum size of a tooltip line. */
	public static int			  TIP_MAX_CHARS = 60;

	/** The label of this button. */
	protected String			  _label;

	/** The font used to display text. */
	protected Font				  _font;

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter _actionTransmitter;

	/** The id of this button. */
	protected String			  _id;

	/**
	 * Creates a toggle button with the specified icon image.
	 *
	 * @param image the image of this button
	 */
	public LyJToggleButton(Image image)
	{
		this(image, null, null, null);
	}

	/**
	 * Creates a toggle button with the specified icon image and tooltip.
	 *
	 * @param image   the image of this button
	 * @param tooltip the tooltip text of this button
	 */
	public LyJToggleButton(Image image, String tooltip)
	{
		this(image, null, tooltip, null);
	}

	/**
	 * Creates a toggle button with the specified attributes.
	 *
	 * @param image   the image of this button
	 * @param label   the label of this button
	 * @param tooltip the tooltip text of this button
	 * @param id      the id of this button
	 */
	public LyJToggleButton(Image image, String label, String tooltip, String id)
	{
		super();
		_id = id;
		_label = label;
		setText(_label);
		setRolloverEnabled(true);
		setImage(image);
		setToolTipText(LyTools.formatToHTML(tooltip, LyJToggleButton.TIP_MAX_CHARS));
		addActionListener(this);
	}

	/**
	 * Sets the icon displayed by the button.
	 *
	 * @param image the new icon image of this toggle button
	 * @see   #getImage()
	 */
	public void setImage(Image image)
	{
		ImageIcon trueIcon = new ImageIcon(image);
		ImageIcon grayIcon = new ImageIcon(LyGrayImageFilter.getFilteredImage(image));

		setRolloverIcon(trueIcon);
		setRolloverSelectedIcon(trueIcon);
		setIcon(grayIcon);
		setText(_label);
	}

	/**
	 * Gets the icon image of this button.
	 *
	 * @return the image displayed by the button
	 * @see    #setImage(Image)
	 */
	public Image getImage()
	{
		return ((ImageIcon)getRolloverIcon()).getImage();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#setFont(java.awt.Font)
	 */
	public void setFont(Font font)
	{
		if (font == null)
			return;
		_font = font;
		super.setFont(font);
	}

	/**
	 * This button cannot grab focus.
	 *
	 * @return <code>false</code>
	 * @see    java.awt.Component#isFocusTraversable()
	 */
	public boolean isFocusTraversable()
	{
		return false;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent e)
	{
		LyActionEvent actionEvent = new LyActionEvent(this);

		actionEvent.setValue(_id);

		if (_actionTransmitter != null)
			_actionTransmitter.processActionEvent(actionEvent);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return (_actionTransmitter);
	}

	/**
	 * This toggle button doesn't handle selection events, so it has no selection transmitter.
	 *
	 * @return <code>null</code>
	 * @see    leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		return null;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.getinterface.LyGetIdInterface#getId()
	 */
	public String getId()
	{
		return _id;
	}

} // end class LyJToggleButton
