/*
 * LyJToolBarComponent.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.border.*;

import leon.view.*;
import leon.view.win.swing.*;


/**
 * This class implements a floatable toolbar.
 *
 * @version $Id: LyJToolBarComponent.java 31021 2011-05-16 12:24:11Z nhamon $
 * @author  Lyria
 * @see     JToolBar
 */
public class LyJToolBarComponent extends JToolBar implements MouseListener
{
	private static final long  serialVersionUID = -3819085758053815967L;

	public static final String __VERSION =
		"$Id: LyJToolBarComponent.java 31021 2011-05-16 12:24:11Z nhamon $";

	/** Enables the floatability of toolbars. */
	public static boolean	   TOOLBAR_FLOATABILITY = false;

	/** Default gap, in pixels, between the button and the border. */
	public static final int    DEFAULT_GAP = 5;

	/** The maximum width used in wrap mode. */
	protected int			   _maxWidth;

	/** The wrap mode property. */
	protected boolean		   _wrap;

	/** The align mode property. */
	protected int			   _align;

	/** Locks the toolbar orientation to avoid orientation change when floating. */
	// FFT1141 SWING - Toolbar and popup
	protected boolean		   _orientationLock;

	/** The image used as a background. */
	protected Image			   _bgImage;

	/**
	 * Creates a default toolbar component.
	 */
	public LyJToolBarComponent()
	{
		this(null, null, 0, false);
	}

	/**
	 * Creates a new toolbar component with the specified attributes.
	 *
	 * @param name     the name of the toolbar
	 * @param border   the border of the toolbar
	 * @param maxWidth the maximum width for wrap mode
	 * @param wrap     the wrap mode property
	 */
	public LyJToolBarComponent(String name, Border border, int maxWidth, boolean wrap)
	{
		this(name, border, maxWidth, wrap, LyFlowLayout.LEFT);
	}

	/**
	 * Creates a new toolbar component with the specified attributes.
	 *
	 * @param name     the name of the toolbar
	 * @param border   the border of the toolbar
	 * @param maxWidth the maximum width for wrap mode
	 * @param wrap     the wrap mode property
	 * @param align    the align mode property
	 */
	public LyJToolBarComponent(String name, Border border, int maxWidth, boolean wrap, int align)
	{
		super(name);

		_align = align;
		_maxWidth = maxWidth;
		_wrap = wrap;
		_orientationLock = false;

		setFloatable(false);
		// border = BorderFactory.createCompoundBorder(getBorder(),border);
		setBorder(border);

		setDoubleBuffered(true);
		
		addMouseListener(this);
	}

	/**
	 * Locks the orientation to avoid change when floating.
	 *
	 * @param status Indicates that the orientation can change or not.
	 * @see   JToolBar#setOrientation(int)
	 */
	public void setOrientationLock(boolean status, LyViewManager viewManager)
	{
		_orientationLock = status;

		if ((status) && (getOrientation() == HORIZONTAL))
			setLayout(new LyFlowLayout(_align, 0, 0, _maxWidth, _wrap, viewManager));
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.swing.JToolBar#setOrientation(int)
	 */
	public void setOrientation(int o)
	{
		if (_orientationLock)
			return;
		Component[] children = getComponents();

		for (int i = 0; i < children.length; i++)
		{
			// This child is a separator
			if (children[i] instanceof LyJSeparator)
			{
				if (o == HORIZONTAL)
					((LyJSeparator)children[i]).setOrientation(SwingConstants.VERTICAL);
				else
					((LyJSeparator)children[i]).setOrientation(SwingConstants.HORIZONTAL);
			}
			// This child is an opener button
			else if (children[i] instanceof LyJOpenerButton)
				((LyJOpenerButton)children[i]).setOrientation(o);
			// This child is another toolbar
			else if (children[i] instanceof LyJToolBarComponent)
				((LyJToolBarComponent)children[i]).setOrientation(o);
		}
		super.setOrientation(o);
	}

	/**
	 * Adds a default separator. The default gap is used before and after this separator.
	 *
	 * @param  sep the separator to add
	 * @return the separator
	 */
	public Component addSeparator(LyJSeparator sep)
	{
		return addSeparator(sep, DEFAULT_GAP, DEFAULT_GAP);
	}

	/**
	 * Adds a separator with the specified before and after gaps.
	 *
	 * @param  sep       the new separator to add
	 * @param  beforeGap the size, in pixels, of the gap before the specified separator
	 * @param  afterGap  the size, in pixels, of the gap after the specified separator
	 * @return the separator with its before and after gaps set
	 */
	public Component addSeparator(LyJSeparator sep, int beforeGap, int afterGap)
	{
		sep.setBeforeGap(beforeGap);
		sep.setAfterGap(afterGap);
		add(sep.getBeforeGap());
		add(sep);
		add(sep.getAfterGap());
		return sep;
	}

	/**
	 * Set the image used as a background for the toolbar component.
	 *
	 * @param image the reference of the image used as a background
	 */
	public void setImageBackground(Image image)
	{
		_bgImage = image;
	}

	/**
	 * Sets the toolbar unfloatable when its parent is already a ToolBar.
	 *
	 * @see java.awt.Component#addNotify()
	 */
	public void addNotify()
	{
		super.addNotify();
		if (getParent().getLayout() instanceof BorderLayout)
			setFloatable(TOOLBAR_FLOATABILITY);
		else
			setFloatable(false);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#getMaximumSize()
	 */
	public Dimension getMaximumSize()
	{
		return getPreferredSize();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#getMinimumSize()
	 */
	public Dimension getMinimumSize()
	{
		if (getOrientation() == HORIZONTAL)
			return new Dimension(0, super.getMinimumSize().height);
		else
			return new Dimension(super.getMinimumSize().width, 20);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.swing.JComponent#paintComponent(Graphics)
	 */
	protected void paintComponent(Graphics g)
	{
		super.paintComponent(g);

		if (_bgImage != null)
			LySwingUtils.fillBackground(g, getWidth(), getHeight(), _bgImage);
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
	 */
	public void mouseClicked(MouseEvent e)
	{
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
	 */
	public void mousePressed(MouseEvent e)
	{
		requestFocus();
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
	 */
	public void mouseEntered(MouseEvent e)
	{
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
	 */
	public void mouseExited(MouseEvent e)
	{
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	public void mouseReleased(MouseEvent e)
	{
	}
} // end class LyJToolBarComponent
