/*
 * LyJTreeComponent.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.awt.datatransfer.*;
import java.awt.dnd.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import leon.misc.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.graphic.LyGraphicalObject;
import leon.view.setinterface.*;
import leon.view.win.swing.*;


/**
 * This class implements a graphical component that display the data in a tree view.
 *
 * @version $Id: LyJTreeComponent.java 35659 2012-05-31 14:42:25Z nhamon $
 * @author  Lyria
 * @see     LySwingTree
 * @see     JTree
 */
public class LyJTreeComponent extends JTree
	implements MouseMotionListener,
		MouseListener,
		KeyListener,
		TreeSelectionListener,
		TreeExpansionListener,
		LySelectionListener,
		LyActionListener,
		LyEventTransmitter,
		LySetIdleInterface,
		LyIsIdleInterface,
		LySplittableCanvas
{
	private static final long				  serialVersionUID = 3550319682133087502L;

	public static final String				  __VERSION =
		"$Id: LyJTreeComponent.java 35659 2012-05-31 14:42:25Z nhamon $";

	/** Cursor for mouse over sensitive node. */
	public static Cursor					  LY_OVER_CURSOR = new Cursor(Cursor.HAND_CURSOR);

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter			  _actionTransmitter;

	/** Provides notification events on component's selections. */
	protected LySelectionTransmitter		  _selectionTransmitter;

	/** Sensitive property. */
	protected boolean						  _isSensitive;

	/** The index of the highlighted row. */
	protected int							  _highLightRow;

	/** Text of the highlighted row. */
	protected String						  _highLightText;

	/** The offset between the text and the icon of the highlighted row. */
	protected int							  _textOffset;

	/** Font metrics used to caculate the length of the text of the highlighted row. */
	protected FontMetrics					  _fm;

	/** Idle status. */
	protected boolean						  _idle;

	/** Keep for multi selection only. */
	protected Vector<Object>				  _selectedPath;

	/** Indicates that only leaves may be selected. */
	protected boolean						  _selectLeavesOnly;

	/** Indicates that default icons must be used for nodes if none is defined. */
	protected boolean						  _displayDefaultIcons;

	/** Indicates that the tree is in a popup. */
	protected boolean						  _inPopup;

	/** Minimum width for the canvas. */
	protected int							  _minWidth;

	/** Number of rows to display. */
	protected int							  _nRows;

	/** Height factor. */
	protected float							  _heightFactor;

	/** Initial font. */
	protected Font							  _initialFont;

	/** Initial row height. */
	protected int							  _initialRowHeight;

	/** The canvas splitter used to represent the tree (may be <code>null</code>). */
	protected LyAbstractCanvasSplitterManager _canvasSplitter;

	/** The drag source for drag and drop. */
	protected DragSource					  _dragSource;

	/** Indicates that checkbox mode is activated. */
	protected boolean						  _checkboxMode;

	/** Fixed width for this tree. */
	protected int							  _fixedWidth;

	/** Indicates that all the nodes of the tree must be opened. */
	protected boolean						  _forceOpen;

	/** Indicates if selections in the tree must be notified. */
	protected boolean						  _notifySelections;

	/** Popup menu displayed in the tree. */
	protected JPopupMenu					  _popupMenu;

	/**
	 * Creates a new tree component. This tree component cannot have drag and drop actions and its
	 * elements are not opened.
	 *
	 * @param root the data this tree displays
	 */
	public LyJTreeComponent(LyJTreeElement root)
	{
		this(root, false);
	}

	/**
	 * Creates a new tree component. This tree component's nodes are not opened.
	 *
	 * @param root      the data this tree displays
	 * @param dragNDrop Indicates that drag and drop is allowed on this component.
	 */
	public LyJTreeComponent(LyJTreeElement root, boolean dragNDrop)
	{
		this(root, dragNDrop, false);
	}

	/**
	 * Creates a tree component. Initializes it and registers its event listeners (mouse, tree...).
	 *
	 * @param root      the data this tree displays
	 * @param dragNDrop Indicates that drag and drop is allowed on this component.
	 * @param forceOpen Indicates that the nodes must be opened.
	 */
	public LyJTreeComponent(LyJTreeElement root, boolean dragNDrop, boolean forceOpen)
	{
		super(root.getMutableTreeNode());

		_forceOpen = forceOpen;

		// Set the border
		setBorder(BorderFactory.createEmptyBorder(3, 3, 3, 3));

		// Set lines drawed style
		putClientProperty("JTree.lineStyle", "Angled");

		// Double clic expand/collapse nodes
		setToggleClickCount(2);

		if (dragNDrop)
		{
			_dragSource = new DragSource();
			_dragSource.createDefaultDragGestureRecognizer(this, DnDConstants.ACTION_COPY_OR_MOVE,
				new _LyTreeCanvasDragListener());
			new DropTarget(this, new _LyTreeCanvasDropListener());
		}

		// Register mouse events
		addMouseMotionListener(this);
		addMouseListener(this);

		// Register key events
		addKeyListener(this);

		// Register selection events
		addTreeSelectionListener(this);

		// Register expand/collapse events
		addTreeExpansionListener(this);

		// Set a customized renderer
		setCellRenderer(new _LyJTreeCellRenderer());

		// Set auto-calculate row height
		setRowHeight(0);

		// Set the content of the tree
		updateTree(root);

		// Set double-buffering
		setDoubleBuffered(true);
		_isSensitive = false;
		_highLightRow = -1;
		setFont(getFont());
		_highLightText = "";
		_textOffset = -1;
		_idle = false;
		_selectLeavesOnly = false;
		_displayDefaultIcons = true;
		_inPopup = false;
		_minWidth = 0;
		_nRows = -1;
		_heightFactor = 1;
		_initialFont = getFont();
		_initialRowHeight = getRowHeight();
		_notifySelections = true;
		_popupMenu = null;

		ToolTipManager.sharedInstance().registerComponent(this);

		_checkboxMode = false;
		_fixedWidth = -1;
	}

	/**
	 * Sets the checkbox mode.
	 *
	 * @param checkboxMode the checkbox mode
	 */
	public void setCheckboxMode(boolean checkboxMode)
	{
		_checkboxMode = checkboxMode;
		
		if (_checkboxMode)
			setToggleClickCount(-1);
	}

	/**
	 * Sets the sensitive mode.
	 *
	 * @param sensitive Indicates that the tree is sensitive.
	 */
	public void setSensitive(boolean sensitive)
	{
		_isSensitive = sensitive;
	}

	/**
	 * Sets the popup mode.
	 *
	 * @param popup Indicates that the tree is in a popup.
	 */
	public void setPopupMode(boolean popup)
	{
		_inPopup = popup;
	}

	/**
	 * Sets the force open mode.
	 *
	 * @param forceOpen Indicates that all the nodes of this tree must be opened.
	 */
	public void setForceOpen(boolean forceOpen)
	{
		_forceOpen = forceOpen;
	}

	/**
	 * Sets the minimum width of this tree component.
	 *
	 * @param minWidth the new minimum width
	 */
	public void setMinWidth(int minWidth)
	{
		_minWidth = minWidth;
	}

	/**
	 * Changes the method for computing height of the tree. Tries to display (approximatively) nRows
	 * in the tree.
	 *
	 * @param nRows the number of row to display. Use -1 for standard behavior.
	 */
	public void setNRows(int nRows)
	{
		_nRows = nRows;
	}

	/**
	 * Sets the height factor of this tree.
	 *
	 * @param factor the factor to use to compute the height of the rows.
	 * @see   #getHeightFactor()
	 */
	public void setHeightFactor(float factor)
	{
		_LyJTreeCellRenderer renderer = (_LyJTreeCellRenderer)getCellRenderer();

		_heightFactor = factor;

		if (_initialRowHeight <= 0)
		{
			Rectangle bounds = getRowBounds(getRowCount() - 1);

			_initialRowHeight = bounds.height;
		}

		if (factor < 1)
		{
			Font font = getFont();

			font = new Font(_initialFont.getName(), _initialFont.getStyle(),
					(int)(_initialFont.getSize() * factor));

			setFont(font);
			renderer.setDisplayImages(false);
			setRowHeight((int)(_initialRowHeight * factor));
		}
		else
		{
			setFont(_initialFont);
			renderer.setDisplayImages(true);
			setRowHeight(0);
		}
	} // end method setHeightFactor

	/**
	 * Gets the height factor.
	 *
	 * @return the factor used to compute the heights of the rows
	 * @see    #setHeightFactor(float)
	 */
	public float getHeightFactor()
	{
		return _heightFactor;
	}

	/**
	 * Centers the tree vertically on the specified element.
	 *
	 * @param path the node to center the element
	 */
	public void centerOnElement(TreePath path)
	{
		Rectangle rect = getPathBounds(path);
		Rectangle visibleRect = getVisibleRect();

		scrollRectToVisible(new Rectangle(0, 0, 0, 0));
		scrollRectToVisible(new Rectangle(0, rect.y + (visibleRect.height / 2), 0,
				rect.height / 2));
	}

	/**
	 * Changes the fixed width of the tree.
	 *
	 * @param fixedWidth the new fixed width of the tree
	 */
	public void setFixedWidth(int fixedWidth)
	{
		_fixedWidth = fixedWidth;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.swing.Scrollable#getPreferredScrollableViewportSize()
	 */
	public Dimension getPreferredScrollableViewportSize()
	{
		Dimension size = super.getPreferredSize();

		if (_forceOpen)
			return super.getPreferredScrollableViewportSize();

		if (size.width < _minWidth)
			size.width = _minWidth;

		if (_nRows > 0)
		{
			if (getRowCount() > 0)
				size.height = getRowBounds(0).x + (getRowBounds(0).height * _nRows);
			else
				size.height = 15 * _nRows;
		}
		else
			size.height = Math.max(LySwingTree.MINIMUM_SIZE.height, size.height);

		if (_fixedWidth < 0)
			size.width = Math.max(LySwingTree.MINIMUM_SIZE.width, size.width);
		else
			size.width = _fixedWidth;

		return size;
	}

	/**
	 * Sets the state of all the nodes. This method does not notify the data model.
	 *
	 * @param node the root of the sub-tree that is updated
	 */
	public void updateTree(LyJTreeElement node)
	{
		removeTreeSelectionListener(this);
		removeTreeExpansionListener(this);

		updateTreeSlave(node);

		addTreeSelectionListener(this);
		addTreeExpansionListener(this);
	}

	/**
	 * Sets the font of this component and gets the font metrics when it changes.
	 *
	 * @see java.awt.Component#setFont(java.awt.Font)
	 */
	public void setFont(Font font)
	{
		super.setFont(font);
		_fm = getFontMetrics(font);
	}

	/**
	 * Paints the component with the highlighted row.
	 *
	 * @see javax.swing.JComponent#paintComponent(java.awt.Graphics)
	 */
	public void paintComponent(Graphics g)
	{
		synchronized (getTreeLock())
		{
			if (g == null)
				return;

			super.paintComponent(g);

			Rectangle bounds = getRowBounds(_highLightRow);

			if (bounds == null)
				return;
			int hLabel = _fm.getHeight();
			int wLabel = _fm.stringWidth(_highLightText);

			g.setColor(Color.blue);
			g.drawLine(bounds.x + _textOffset, bounds.y + ((bounds.height + hLabel) / 2) - 1,
				bounds.x
				+ _textOffset + wLabel - 1, bounds.y + ((bounds.height + hLabel) / 2) - 1);
		}
	}

	/**
	 * Sets the state (open/selected) of all the nodes of the tree recursively.
	 *
	 * @param node the root of the sub-tree that is updated
	 * @see   #updateTree(LyJTreeElement)
	 */
	protected void updateTreeSlave(LyJTreeElement node)
	{
		TreePath						  path = node.getTreePath();
		LyJTreeElementImpl._LyNodeContent content = node.getContent();

		node.setTree(this);

		if (content == null)
			return;

		if (content.isSelected())
			addSelectionPath(path);

		int i, n = node.getMutableTreeNode().getChildCount();

		for (i = 0; i < n; i++)
		{
			DefaultMutableTreeNode mutableTreeNode = (DefaultMutableTreeNode)node
					.getMutableTreeNode().getChildAt(i);

			updateTreeSlave(((LyJTreeElement)mutableTreeNode.getUserObject()));
		}

		if (content.isOpen() || _forceOpen)
			expandPath(path);
		else
			collapsePath(path);
	}

	/**
	 * Sets the select leaves only mode.
	 *
	 * @param selectLeavesOnly Indicates that the select leaves only mode must be set or unset.
	 */
	public void setSelectLeavesOnly(boolean selectLeavesOnly)
	{
		_selectLeavesOnly = selectLeavesOnly;
	}

	/**
	 * Indicates if selection events are notified.
	 *
	 * @return true if selection events are notified, false otherwise.
	 */
	public boolean isNotifySelections()
	{
		return _notifySelections;
	}

	/**
	 * Sets the notify selections mode.
	 *
	 * @param notify Indicates whether that selection events are notify or not.
	 */
	public void setNotifySelections(boolean notify)
	{
		_notifySelections = notify;
	}

	/**
	 * Sets the display default icons mode.
	 *
	 * @param displayDefaultIcons Indicates that the default display icons mode must be set or
	 *                            unset.
	 */
	public void setDisplayDefaultIcons(boolean displayDefaultIcons)
	{
		_displayDefaultIcons = displayDefaultIcons;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#processMouseEvent(java.awt.event.MouseEvent)
	 */
	protected void processMouseEvent(MouseEvent e)
	{
		if (_checkboxMode)
		{
			int id = e.getID();

			switch (id)
			{
			case MouseEvent.MOUSE_PRESSED:
				if (e.getClickCount() == 1)
				{
					int modifiers = e.getModifiers();

					modifiers = modifiers | InputEvent.CTRL_MASK;

					e = new MouseEvent((Component)e.getSource(), id, e.getWhen(), modifiers,
							e.getX(), e.getY(), e.getClickCount(), e.isPopupTrigger());
				}
				else if (e.getClickCount() >= 2)
				{
					return;
				}
				break;
			}
		}

		super.processMouseEvent(e);
	}

	/**
	 * Called when the selection on the tree has changed.
	 *
	 * @see javax.swing.event.TreeSelectionListener#valueChanged(javax.swing.event.TreeSelectionEvent)
	 */
	public void valueChanged(TreeSelectionEvent e)
	{
		TreePath[] paths = e.getPaths();

		// FFT 1207 : multi-select
		if (_selectedPath == null)
			_selectedPath = new Vector<Object>();

		for (int i = 0; i < paths.length; i++)
		{
			TreePath path = paths[i];

			if (e.isAddedPath(path))
			{
				Object  swingNode = path.getLastPathComponent();
				boolean add = true;

				// If select leaves only, remove all the nodes that have children
				if ((swingNode instanceof DefaultMutableTreeNode) && _selectLeavesOnly)
				{
					DefaultMutableTreeNode treeNode = (DefaultMutableTreeNode)swingNode;

					if (treeNode.getChildCount() > 0)
						add = false;
				}

				if (add)
					_selectedPath.add(path.getLastPathComponent());
			}
			else
				_selectedPath.remove(path.getLastPathComponent());
		}

		if (_notifySelections)
		{
			LySelectionEvent selectionEvent = new LySelectionEvent(e.getSource(), _selectedPath,
					true);

			selectionPerformed(selectionEvent);

			if (_inPopup
					&& (getSelectionModel().getSelectionMode()
						== TreeSelectionModel.SINGLE_TREE_SELECTION))
			{
				Window parentWindow = getParentWindow();

				if (parentWindow != null)
					parentWindow.setVisible(false);
			}
		}
	} // end method valueChanged

	/**
	 * Returns the parent window of the tree component.
	 *
	 * @return the parent window of the tree component, or <code>null</code> if there is none
	 */
	protected Window getParentWindow()
	{
		Component parent = getParent();

		while (!(parent instanceof Window) && (parent != null))
			parent = parent.getParent();

		return (Window)parent;
	}

	/**
	 * Gets all selected valuesi without the nodes that have children when the leaves only select
	 * mode is set.
	 *
	 * @see javax.swing.JTree#getSelectionPaths()
	 */
	public TreePath[] getSelectionPaths()
	{
		TreePath[] paths = super.getSelectionPaths();
		Vector<TreePath>     selectedPaths = new Vector<TreePath>();

		if ((paths != null) && (paths.length > 0))
		{
			for (int i = 0; i < paths.length; i++)
			{
				TreePath path = paths[i];
				Object   swingNode = path.getLastPathComponent();
				boolean  add = true;

				// If select leaves only, remove all the nodes that have children
				if ((swingNode instanceof DefaultMutableTreeNode) && _selectLeavesOnly)
				{
					DefaultMutableTreeNode treeNode = (DefaultMutableTreeNode)swingNode;

					if (treeNode.getChildCount() > 0)
						add = false;
				}

				if (add)
					selectedPaths.addElement(path);
			}
		}

		paths = new TreePath[selectedPaths.size()];

		for (int i = 0; i < selectedPaths.size(); i++)
			paths[i] = selectedPaths.elementAt(i);

		return paths;
	} // end method getSelectionPaths

	/**
	 * Underlines the text under the mouse cursor and changes the cursor to a hand.
	 *
	 * @see java.awt.event.MouseMotionListener#mouseMoved(java.awt.event.MouseEvent)
	 */
	public void mouseMoved(MouseEvent e)
	{
		if (!_isSensitive)
			return;
		int oldHighLightRow = _highLightRow;

		_highLightRow = getRowForLocation(e.getX(), e.getY());

		if (oldHighLightRow != _highLightRow)
		{
			Graphics	g = getGraphics();
			Polygon		bounds = null;
			Rectangle	rowBounds = null, popupBounds = null;

			if ((_popupMenu != null)
					&& _popupMenu.isVisible())
			{
				Point	popupLocation = _popupMenu.getLocationOnScreen();
				Point	treeLocation = getLocationOnScreen();
				Point	popupTreeLocation = new Point(popupLocation.x - treeLocation.x, popupLocation.y - treeLocation.y);

				popupBounds = new Rectangle(popupTreeLocation.x, popupTreeLocation.y,
						_popupMenu.getWidth(), _popupMenu.getHeight());
			}

			if (_highLightRow == -1)
			{
				rowBounds = getRowBounds(oldHighLightRow);

				if (popupBounds != null)
					bounds = getRepaintPolygon(rowBounds, popupBounds);

				setCursor(Cursor.getDefaultCursor());
			}
			else
			{
				rowBounds = getRowBounds(_highLightRow);

				if (popupBounds != null)
				{
					bounds = getRepaintPolygon(rowBounds, popupBounds);
					g.setClip(bounds);
				}
				else
					g.setClip(rowBounds);

				paint(g);

				rowBounds = getRowBounds(oldHighLightRow);

				if (popupBounds != null)
					bounds = getRepaintPolygon(rowBounds, popupBounds);

				setCursor(LY_OVER_CURSOR);
			}

			if (bounds != null)
			{
				g.setClip(bounds);
				paint(g);
			}
			else
			{
				g.setClip(rowBounds);
				paint(g);
			}

			if ((_popupMenu != null)
					&& _popupMenu.isVisible())
				_popupMenu.repaint();
		}
	} // end method mouseMoved

	/**
	 * Computes the polygon that must be used to repaint tree part when a popup menu is displayed.
	 *
	 * @param rowBounds the bounds of the row under the mouse.
	 * @param popupBounds the bounds of the popup menu.
	 * @return the polygon that must be repaint (corresponds to the popup menu bounds removed to
	 * row bounds).
	 */
	protected Polygon getRepaintPolygon(Rectangle rowBounds, Rectangle popupBounds)
	{
		if ((rowBounds == null) || (popupBounds == null))
			return null;

		Rectangle	intersection = rowBounds.intersection(popupBounds);
		Polygon		result;
		int[]		xPoints, yPoints;

		if ((intersection.x == rowBounds.x)
				&& (intersection.y == rowBounds.y))
		{
			xPoints = new int[] {intersection.x + intersection.width, rowBounds.x + rowBounds.width, rowBounds.x + rowBounds.width,
					rowBounds.x, rowBounds.x, intersection.x + intersection.width};
			yPoints = new int[] {intersection.y, intersection.y, rowBounds.y + rowBounds.height,
					rowBounds.y + rowBounds.height, intersection.y + intersection.height, intersection.y + intersection.height};
		}
		else if ((intersection.x + intersection.width == rowBounds.x + rowBounds.width)
				&& (intersection.y == rowBounds.y))
		{
			xPoints = new int[] {intersection.x, intersection.x, intersection.x + intersection.width,
					intersection.x + intersection.width, rowBounds.x, rowBounds.x};
			yPoints = new int[] {intersection.y, intersection.y + intersection.height, intersection.y + intersection.height,
					rowBounds.y + rowBounds.height, rowBounds.y + rowBounds.height, rowBounds.y};
		}
		else if ((intersection.x + intersection.width == rowBounds.x + rowBounds.width)
				&& (intersection.y + intersection.height == rowBounds.y + rowBounds.height))
		{
			xPoints = new int[] {intersection.x + intersection.width, intersection.x, intersection.x,
					rowBounds.x, rowBounds.x, rowBounds.x + rowBounds.width};
			yPoints = new int[] {intersection.y, intersection.y, intersection.y + intersection.height,
					rowBounds.y + rowBounds.height, rowBounds.y, rowBounds.y};
		}
		else
		{
			xPoints = new int[] {rowBounds.x, rowBounds.x + rowBounds.width, rowBounds.x + rowBounds.width,
					intersection.x + intersection.width, intersection.x + intersection.width, rowBounds.x};
			yPoints = new int[] {rowBounds.y, rowBounds.y, rowBounds.y + rowBounds.height,
					rowBounds.y + rowBounds.height, intersection.y, intersection.y};
		}

		result = new Polygon(xPoints, yPoints, 6);

		return result;
	}

	/**
	 * Registers the popup menu displayed on the tree component.
	 *
	 * @param popupMenu the popup menu instance displayed on the tree component.
	 */
	public void setPopupMenu(JPopupMenu popupMenu)
	{
		_popupMenu = popupMenu;
	}

	/**
	 * Nothing is done.
	 *
	 * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
	 */
	public void mouseDragged(MouseEvent e)
	{
	}

	/**
	 * Nothing is done.
	 *
	 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
	 */
	public void mouseEntered(MouseEvent e)
	{
	}

	/**
	 * Nothing is done.
	 *
	 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
	 */
	public void mouseExited(MouseEvent e)
	{
	}

	/**
	 * Nothing is done.
	 *
	 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
	 */
	public void mousePressed(MouseEvent e)
	{
	}

	/**
	 * Nothing is done.
	 *
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	public void mouseReleased(MouseEvent e)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
	 */
	public void mouseClicked(MouseEvent e)
	{
		TreePath path = getPathForLocation(e.getX(), e.getY());
		Object   swingNode = (path != null) ? (path.getLastPathComponent()) : (null);

		if (swingNode instanceof DefaultMutableTreeNode)
			swingNode = ((DefaultMutableTreeNode)swingNode).getUserObject();

		// A right click occurs
		if (e.getModifiers() == InputEvent.BUTTON3_MASK)
		{
			if (!isPathSelected(path) && (path != null))
			{
				if (_checkboxMode)
					addSelectionPath(path);
				else
					setSelectionPath(path);
			}

			if ((swingNode != null) && (swingNode instanceof LyJTreeElement))
			{
				LyActionEvent actionEvent = new LyActionEvent(this, null,
						((LyJTreeElement)swingNode).getId(), LyMenuDemand.TYPE_POPUP);
				Point		  pos = e.getPoint();

				actionEvent.setParameter(new int[] { pos.x, pos.y });
				actionPerformed(actionEvent);
			}
		}

		// A left double click occurs
		if ((e.getClickCount() == 2)
				&& (e.getModifiers() == MouseEvent.BUTTON1_MASK))
		{
			if ((swingNode != null) && (swingNode instanceof LyJTreeElement))
			{
				LyJTreeElement treeElement = (LyJTreeElement)swingNode;
				Rectangle	   validityRect = treeElement.getValidityRectangle();
				Rectangle	   bounds = getPathBounds(path);

				if ((validityRect != null) && (bounds != null))
				{
					Rectangle validityBounds = new Rectangle(bounds.x + validityRect.x,
							bounds.y
							+ validityRect.y, validityRect.width, validityRect.height);

					if (validityBounds.contains(e.getPoint()))
					{
						treeElement.changeValidity();
						e.consume();
						repaint();
						return;
					}
				}

				if (treeElement.getMutableTreeNode().isLeaf() || (getToggleClickCount() < 0))
				{
					LyActionEvent actionEvent = new LyActionEvent(this);
					String		  key = ((LyJTreeElement)swingNode).getId();

					actionEvent.setValue(key);
					actionPerformed(actionEvent);
				}
				else
				{
					// NOTHING : the node is autmatically collaspe/expand
					// see getToggleClickCount
				}
			}
		}
	} // end method mouseClicked

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.swing.event.TreeExpansionListener#treeCollapsed(javax.swing.event.TreeExpansionEvent)
	 */
	public void treeCollapsed(TreeExpansionEvent event)
	{
		TreePath			   path = event.getPath();
		DefaultMutableTreeNode mutableTreeNode = (DefaultMutableTreeNode)path
				.getLastPathComponent();
		LyJTreeElement		   node = ((LyJTreeElement)mutableTreeNode.getUserObject());

		node.treeCollapsed(event);

		String		  key = node.getId();
		Boolean		  bool = Boolean.FALSE;
		LyActionEvent actionEvent = new LyActionEvent(this, null, null, LyActionDemand.TYPE_EXPAND);
		StringBuffer  buf = new StringBuffer();

		buf.append(key).append(LyDemand.SEPARATOR).append(bool.toString());
		actionEvent.setValue(buf.toString());
		actionPerformed(actionEvent);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.swing.event.TreeExpansionListener#treeExpanded(javax.swing.event.TreeExpansionEvent)
	 */
	public void treeExpanded(TreeExpansionEvent event)
	{
		TreePath			   path = event.getPath();
		DefaultMutableTreeNode mutableTreeNode = (DefaultMutableTreeNode)path
				.getLastPathComponent();
		LyJTreeElement		   node = ((LyJTreeElement)mutableTreeNode.getUserObject());

		node.treeExpanded(event);

		String		  key = node.getId();
		Boolean		  bool = Boolean.TRUE;
		LyActionEvent actionEvent = new LyActionEvent(this, null, null, LyActionDemand.TYPE_EXPAND);
		StringBuffer  buf = new StringBuffer();

		buf.append(key).append(LyDemand.SEPARATOR).append(bool.toString());
		actionEvent.setValue(buf.toString());
		actionPerformed(actionEvent);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		if (_actionTransmitter != null)
			_actionTransmitter.processActionEvent(actionEvent);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LySelectionListener#selectionPerformed(leon.view.event.LySelectionEvent)
	 */
	public void selectionPerformed(LySelectionEvent selectionEvent)
	{
		if (_selectionTransmitter != null)
			_selectionTransmitter.processSelectionEvent(selectionEvent);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return (_actionTransmitter);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		if (_selectionTransmitter == null)
			_selectionTransmitter = new LySimpleSelectionTransmitter();

		return (_selectionTransmitter);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetIdleInterface#setIdle(boolean)
	 */
	public void setIdle(boolean status)
	{
		_idle = status;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.getinterface.LyIsIdleInterface#isIdle()
	 */
	public boolean isIdle()
	{
		return _idle;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LySplittableCanvas#drawInSplitter(java.awt.Graphics, int,
	 *      int)
	 */
	public void drawInSplitter(Graphics gc, int xOffset, int yOffset)
	{
		Dimension dim = _canvasSplitter.getCurrentPartSize(false);
		Rectangle clip = gc.getClipBounds();
		Dimension scrollDim = getScrollableSize();
		Dimension max = new Dimension(Math.max(dim.width, scrollDim.width),
				Math.max(dim.height, scrollDim.height));
		Image     i = _canvasSplitter.getComponent().createImage(scrollDim.width, scrollDim.height);
		Graphics  g = i.getGraphics();

		g.setClip(0, 0, max.width, max.height);
		g.setColor(getBackground());
		g.fillRect(0, 0, max.width, max.height);
		paintComponent(g);

		gc.setClip(0, 0, dim.width, dim.height);
		gc.setColor(getBackground());
		gc.fillRect(0, 0, max.width, max.height);
		gc.drawImage(i, 0, 0, dim.width, dim.height, xOffset, yOffset, xOffset + dim.width,
			yOffset
			+ dim.height, this);
		gc.setClip(clip);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#getWidth()
	 */
	public int getWidth()
	{
		if (_canvasSplitter != null)
		{
			Dimension dim = _canvasSplitter.getCurrentPartSize(false);

			return dim.width;
		}

		return super.getWidth();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#getHeight()
	 */
	public int getHeight()
	{
		if (_canvasSplitter != null)
		{
			Dimension dim = _canvasSplitter.getCurrentPartSize(false);

			return dim.height;
		}

		return super.getHeight();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LySplittableCanvas#setCanvasSplitter(leon.view.win.swing.component.LyAbstractCanvasSplitterManager)
	 */
	public void setCanvasSplitter(LyAbstractCanvasSplitterManager splitter)
	{
		_canvasSplitter = splitter;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LySplittableCanvas#getScrollableSize()
	 */
	public Dimension getScrollableSize()
	{
		int		  i, nRows = getRowCount();
		int		  width = 0;
		Rectangle bounds = new Rectangle(0, 0);

		for (i = 0; i < nRows; i++)
		{
			bounds = getRowBounds(nRows - 1);

			if (width < (bounds.x + bounds.width))
				width = bounds.x + bounds.width;
		}

		return new Dimension(width, bounds.y + bounds.height);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LySplittableCanvas#getTopHeaderHeight()
	 */
	public int getTopHeaderHeight()
	{
		return getPreferredSize().height;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LySplittableCanvas#getLeftHeaderWidth()
	 */
	public int getLeftHeaderWidth()
	{
		return getPreferredSize().width;
	}

	/**
	 * Nothing is done.
	 *
	 * @see java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
	 */
	public void keyTyped(KeyEvent evt)
	{
	}

	/**
	 * Nothing is done.
	 *
	 * @see java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
	 */
	public void keyReleased(KeyEvent evt)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
	 */
	public void keyPressed(KeyEvent evt)
	{
		int			code = evt.getKeyCode();
		TreePath	path = getSelectionPath();

		if (path != null)
		{
			LyActionEvent	actionEvent = null;
			Object   		swingNode = (path != null) ? (path.getLastPathComponent()) : (null);

			if (swingNode instanceof DefaultMutableTreeNode)
				swingNode = ((DefaultMutableTreeNode)swingNode).getUserObject();

			if ((swingNode != null) && (swingNode instanceof LyJTreeElement))
			{
				switch (code)
				{
				case KeyEvent.VK_DELETE:
					actionEvent = new LyActionEvent(this, null, null,
							LyActionDemand.TYPE_DELETE);
					break;
				case KeyEvent.VK_ENTER:
					actionEvent = new LyActionEvent(this, null, null,
							LyActionDemand.TYPE_ACTION);
					break;
				}

				if (actionEvent != null)
				{
					actionEvent.setValue(((LyJTreeElement)swingNode).getId());
					actionPerformed(actionEvent);
				}
			}
		}
	}

	/**
	 * Displays the tree in text mode on the user console. Used for debug.
	 */

	protected void printOnScreenTreeState()
	{
		DefaultMutableTreeNode mutableTreeNode = (DefaultMutableTreeNode)getModel().getRoot();
		LyJTreeElement		   root = ((LyJTreeElement)mutableTreeNode.getUserObject());

		Enumeration<?>		   enumList = root.getMutableTreeNode().breadthFirstEnumeration();
		LyJTreeElement		   node;

		while (enumList.hasMoreElements())
		{
			node = ((LyJTreeElement)enumList.nextElement());

			System.out.println(node.getContent().getLabel() + " " + node.getContent().isSelected()
				+ " " + node.getContent().isOpen() + " " + " ***** " + node.getLabel() + " "
				+ node.isSelected() + " " + node.isOpen());
		}
		System.out.println("\n\n");
	}

	/* ************************* Internal Class ********************************************/

	/**
	 * This class implements the way each node is displayed switching its node description.
	 *
	 * @see TreeCellRenderer
	 * @see JTree
	 */
	public class _LyJTreeCellRenderer extends DefaultTreeCellRenderer
	{
		private static final long serialVersionUID = -2197120734755756764L;

		/** Gap between images, in pixels. */
		public static final int   IMAGE_GAP = 2;

		/** List of additional images. */
		protected Image[]		  _images;

		/** Index of validity image in the additional images list. */
		protected int			  _validityImageIndex;

		/** Instance of LyJTreeElement associated with this cell renderer. */
		protected LyJTreeElement  _treeElement;

		/** Indicates that the images must be displayed. */
		protected boolean		  _displayImages;

		/**
		 * Default constructor of a tree cell renderer.
		 */
		public _LyJTreeCellRenderer()
		{
			super();

			_images = null;
			_validityImageIndex = -1;
			_treeElement = null;
			_displayImages = true;

			if (!_displayDefaultIcons)
			{
				setIcon(null);
				setLeafIcon(null);
				setOpenIcon(null);
				setClosedIcon(null);
			}
		}

		/*
		 * (non-Javadoc)
		 * @see javax.swing.tree.TreeCellRenderer#getTreeCellRendererComponent(javax.swing.JTree,
		 * java.lang.Object, boolean, boolean, boolean, int, boolean)
		 */
		public Component getTreeCellRendererComponent(JTree tree, Object value, boolean sel,
			boolean expanded, boolean leaf, int row, boolean hasFocus)
		{
			// Call super to configure background color and others properties
			// Comp equals 'this'
			JLabel comp = (JLabel)super.getTreeCellRendererComponent(tree, value, sel, expanded,
					leaf, row, hasFocus);

			if (value instanceof DefaultMutableTreeNode)
				value = ((DefaultMutableTreeNode)value).getUserObject();

			if (value instanceof LyJTreeElement)
			{
				if (value instanceof LyGraphicalObject)
				{
					if (((LyGraphicalObject)value).isFreed())
						return null;
				}
				_treeElement = (LyJTreeElement)value;

				LyJTreeElementImpl._LyNodeContent node = _treeElement.getContent();

				if (node == null)
					return comp;

				Image		  image = node.getImage();
				String		  label = node.getLabel();
				String		  tooltip = node.getTooltip();
				LyEnvironment environment = _treeElement.getEnvironment();
				Vector<Image> images = new Vector<Image>();

				if (_treeElement.getColor() != null)
					setBackgroundNonSelectionColor(_treeElement.getColor());
				else
					setBackgroundNonSelectionColor(tree.getBackground());

				if (selected && (_treeElement.getSelectedFontColor() != null))
					setForeground(_treeElement.getSelectedFontColor());
				else if (_treeElement.getFontColor() != null)
					setForeground(_treeElement.getFontColor());

				if (_treeElement.getSelectedColor() != null)
					setBackgroundSelectionColor(_treeElement.getSelectedColor());

				if (_treeElement.getFont() != null)
				{
					Font font = _treeElement.getFont();

					font = new Font(font.getName(), font.getStyle(),
							(int)(font.getSize() * _heightFactor));

					comp.setFont(font);
				}
				else
					comp.setFont(tree.getFont());

				// Retrieve image for checkbox (checkbox selection mode only)
				if ((_treeElement.getSelectedCheckboxImage() != null)
						|| (_treeElement.getEmptyCheckboxImage() != null))
				{
					setBackgroundSelectionColor(tree.getBackground());
					setForeground(_treeElement.getFontColor());

					if (sel)
						images.addElement(_treeElement.getSelectedCheckboxImage());
					else
						images.addElement(_treeElement.getEmptyCheckboxImage());
				}

				if (environment != null)
				{
					label = environment.translate(label);
					tooltip = environment.translate(tooltip);
				}

				setToolTipText(tooltip);

				// Retrieve main image of the node
				if (image != null)
					images.addElement(image);

				// Retrieve additional images
				Vector<Image> additionalImages = _treeElement.getAdditionalImages();

				_validityImageIndex = _treeElement.getValidityImageIndex();

				// If tree contains validity images, do not toggle on double-click
				// (change validity instead).
				if (_validityImageIndex >= 0)
					setToggleClickCount(-1);

				_validityImageIndex += images.size() - 1;

				if (additionalImages != null)
					images.addAll(additionalImages);

				if ((images.size() > 0) && (_displayImages))
				{
					setIcon(new ImageIcon(images.elementAt(0)));
					images.removeElementAt(0);
				}

				_images = new Image[images.size()];
				images.toArray(_images);

				comp.setText(label);

				if ((_isSensitive) && (_highLightRow == row))
					highlight(comp);
			}

			Dimension imagesSize = getImagesSize();

			setIconTextGap(imagesSize.width + 4);

			return comp;
		} // end method getTreeCellRendererComponent

		/**
		 * Sets the display images mode.
		 *
		 * @param display Indicates that the images must be displayed
		 */
		public void setDisplayImages(boolean display)
		{
			_displayImages = display;
		}

		/**
		 * Gets the preferred size of this tree component. Gives more space between rows.
		 *
		 * @see java.awt.Component#getPreferredSize()
		 */
		public Dimension getPreferredSize()
		{
			if ((getIcon() == null) && ((_images == null) || (_images.length == 0))
					&& (getText() == null))
				return new Dimension(11, 11);

			Dimension imagesSize = getImagesSize();
			int		  width = super.getPreferredSize().width;
			int		  height = super.getPreferredSize().height;

			return new Dimension(width, Math.max(height, imagesSize.height) + 3);
		}

		/**
		 * Draws the component and its additional images.
		 *
		 * @see javax.swing.JComponent#paintComponent(java.awt.Graphics)
		 */
		public void paintComponent(Graphics g)
		{
			super.paintComponent(g);

			Dimension d = getSize();

			if ((getIcon() == null) && ((_images == null) || (_images.length == 0))
					&& (getText() == null))
			{
				g.setColor(Color.black);
				g.fillRect(1, 1, 9, 9);
			}

			// Draw additional images
			if ((_images != null) && (_displayImages))
			{
				int i, n = _images.length;
				int offset = ((getIcon() == null) ? 0 : (getIcon().getIconWidth() + IMAGE_GAP));

				for (i = 0; i < n; i++)
				{
					if ((_treeElement != null) && (_validityImageIndex == i))
						_treeElement.setValidityRectangle(new Rectangle(offset,
								(d.height - _images[i].getHeight(this)) / 2,
								_images[i].getWidth(this), _images[i].getHeight(this)));

					g.drawImage(_images[i], offset, (d.height - _images[i].getHeight(this)) / 2,
						this);

					offset += _images[i].getWidth(this) + IMAGE_GAP;
				}
			}
		} // end method paintComponent

		/**
		 * Returns the dimension of additional images.
		 *
		 * @return the dimension of additional images
		 */
		protected Dimension getImagesSize()
		{
			if (_images == null)
				return new Dimension(0, 0);

			int   i, n = _images.length;
			int   width = 0, height = 0;
			Image image;

			for (i = 0; i < n; i++)
			{
				image = _images[i];

				width += IMAGE_GAP + image.getWidth(this);
				height = Math.max(height, image.getHeight(this));
			}

			return new Dimension(width, height);
		}

		/**
		 * Highlights the specified label.
		 *
		 * @param comp the label to highlight
		 */
		public void highlight(JLabel comp)
		{
			comp.setForeground(Color.blue);
			_highLightText = comp.getText();
			_textOffset = comp.getIcon().getIconWidth() + comp.getIconTextGap();
		}
	} // end class _LyJTreeCellRenderer


	/**
	 * Provides drag handling as well as drag gestures.
	 */
	class _LyTreeCanvasDragListener
		implements DragGestureListener,
			DragSourceListener
	{

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DragSourceListener#dragDropEnd(java.awt.dnd.DragSourceDropEvent)
		 */
		public void dragDropEnd(DragSourceDropEvent dsde)
		{
			LyActionEvent actionEvent = new LyActionEvent(this, null,
					String.valueOf(dsde.getDropAction()), LyDragDropDemand.TYPE_DRAG_FINISHED);

			actionPerformed(actionEvent);
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.dnd.DragSourceListener#dragEnter(java.awt.dnd.DragSourceDragEvent)
		 */
		public void dragEnter(DragSourceDragEvent dsde)
		{
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.dnd.DragSourceListener#dragExit(java.awt.dnd.DragSourceEvent)
		 */
		public void dragExit(DragSourceEvent dse)
		{
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.dnd.DragSourceListener#dragOver(java.awt.dnd.DragSourceDragEvent)
		 */
		public void dragOver(DragSourceDragEvent dsde)
		{
		}

		/**
		 * Nothing is done.
		 *
		 * @see java.awt.dnd.DragSourceListener#dropActionChanged(java.awt.dnd.DragSourceDragEvent)
		 */
		public void dropActionChanged(DragSourceDragEvent dsde)
		{
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DragGestureListener#dragGestureRecognized(java.awt.dnd.DragGestureEvent)
		 */
		public void dragGestureRecognized(DragGestureEvent dge)
		{
			TreePath path = getPathForLocation(dge.getDragOrigin().x, dge.getDragOrigin().y);
			Object   swingNode = ((path != null) ? path.getLastPathComponent() : null);

			if (swingNode instanceof DefaultMutableTreeNode)
				swingNode = ((DefaultMutableTreeNode)swingNode).getUserObject();

			if ((swingNode != null) && (swingNode instanceof LyJTreeElement))
			{
				Transferable t = new StringSelection(((LyJTreeElement)swingNode).getId());

				_dragSource.startDrag(dge, DragSource.DefaultCopyDrop, t, this);

				LyActionEvent actionEvent = new LyActionEvent(this, null,
						((LyJTreeElement)swingNode).getId(), LyDragDropDemand.TYPE_DRAG_START);

				actionPerformed(actionEvent);
			}
		}
	} // end class _LyTreeCanvasDragListener


	/**
	 * Provides drap and drop handling.
	 */
	class _LyTreeCanvasDropListener
		implements DropTargetListener
	{

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#dragEnter(java.awt.dnd.DropTargetDragEvent)
		 */
		public void dragEnter(DropTargetDragEvent dtde)
		{
			LyActionEvent actionEvent = new LyActionEvent(this, null, null,
					LyDragDropDemand.TYPE_DRAG_ENTER);

			actionPerformed(actionEvent);
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#dragExit(java.awt.dnd.DropTargetEvent)
		 */
		public void dragExit(DropTargetEvent dte)
		{
			LyActionEvent actionEvent = new LyActionEvent(this, null, null,
					LyDragDropDemand.TYPE_DRAG_LEAVE);

			actionPerformed(actionEvent);
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#dragOver(java.awt.dnd.DropTargetDragEvent)
		 */
		public void dragOver(DropTargetDragEvent dtde)
		{
			LyActionEvent actionEvent = new LyActionEvent(this, null, null,
					LyDragDropDemand.TYPE_DRAG_OVER);

			actionEvent.setParameter(dtde);
			actionPerformed(actionEvent);
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#drop(java.awt.dnd.DropTargetDropEvent)
		 */
		public void drop(DropTargetDropEvent dtde)
		{
			TreePath path = getPathForLocation(dtde.getLocation().x, dtde.getLocation().y);
			Object   swingNode = ((path != null) ? path.getLastPathComponent() : null);

			if (swingNode instanceof DefaultMutableTreeNode)
				swingNode = ((DefaultMutableTreeNode)swingNode).getUserObject();

			if ((swingNode != null) && (swingNode instanceof LyJTreeElement))
			{
				dtde.acceptDrop(DnDConstants.ACTION_COPY_OR_MOVE);
				dtde.getDropTargetContext().dropComplete(true);

				LyActionEvent actionEvent = new LyActionEvent(this, null,
						((LyJTreeElement)swingNode).getId(), LyDragDropDemand.TYPE_DROP);

				actionPerformed(actionEvent);
			}
			else
				dtde.rejectDrop();
		}

		/*
		 * (non-Javadoc)
		 * @see java.awt.dnd.DropTargetListener#dropActionChanged(java.awt.dnd.DropTargetDragEvent)
		 */
		public void dropActionChanged(DropTargetDragEvent dtde)
		{
			LyActionEvent actionEvent = new LyActionEvent(this, null, null,
					LyDragDropDemand.TYPE_DRAG_OPERATION_CHANGED);

			actionPerformed(actionEvent);
		}
	} // end class _LyTreeCanvasDropListener
} // end class LyJTreeComponent
