/*
 * LyJTreeElement.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.util.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import leon.misc.*;
import leon.view.win.swing.component.LyJTreeElementImpl.*;


/**
 * This interface defines an object than can be drawn in a tree.
 *
 * @version $Id: LyJTreeElement.java 29006 2010-12-23 17:44:46Z bmichel $
 * @author  Lyria
 * @see     LyTree
 */
public interface LyJTreeElement
{
	String __VERSION = "$Id: LyJTreeElement.java 29006 2010-12-23 17:44:46Z bmichel $";

	/**
	 * Gets the mutable tree node object related to the current node.
	 *
	 * @return the mutable tree node of this tree element
	 */
	DefaultMutableTreeNode getMutableTreeNode();

	/**
	 * Gets the path to go to the mutable tree node.
	 *
	 * @return the tree path of this tree element
	 */
	TreePath getTreePath();

	/**
	 * Gets this node content.
	 *
	 * @return the content of this node
	 */
	_LyNodeContent getContent();

	/**
	 * Sets this tree node component.
	 *
	 * @param tree the new tree component
	 * @see   #getTree()
	 */
	void setTree(LyJTreeComponent tree);

	/**
	 * Gets this tree node component.
	 *
	 * @return this tree component
	 * @see    #setTree(LyJTreeComponent)
	 */
	LyJTreeComponent getTree();

	/**
	 * Gets this tree node's environment.
	 *
	 * @return the tree node environment
	 */
	LyEnvironment getEnvironment();

	/**
	 * Gets this tree node identifier.
	 *
	 * @return the id of this tree node
	 */
	String getId();

	/**
	 * Gets this tree node label.
	 *
	 * @return the label of this tree node
	 */
	String getLabel();

	/**
	 * Indicates that this tree node is selected.
	 *
	 * @return <code>true</code> if this tree node is selected, <code>false</code> otherwise
	 */
	boolean isSelected();

	/**
	 * Indicates that this tree node is opened.
	 *
	 * @return <code>true</code> if this tree node is opened, <code>false</code> otherwise
	 */
	boolean isOpen();

	/**
	 * Gets the height factor of this tree node.
	 *
	 * @return the height factor
	 */
	int getHeightFactor();

	/**
	 * Called whenever an item in the tree has been expanded.
	 *
	 * @param event the event generated by the expansion
	 */
	void treeExpanded(TreeExpansionEvent event);

	/**
	 * Called whenever an item in the tree has been collapsed.
	 *
	 * @param event the event generated by the collapse
	 */
	void treeCollapsed(TreeExpansionEvent event);

	/**
	 * Gets the image for the selected checkbox.
	 *
	 * @return the selected checkbox image
	 */
	Image getSelectedCheckboxImage();

	/**
	 * Gets the image for the empty checkbox.
	 *
	 * @return the empty checkbox image
	 */
	Image getEmptyCheckboxImage();

	/**
	 * Gets the list of the additional images of this tree node.
	 *
	 * @return the list of the additional images
	 */
	Vector<Image> getAdditionalImages();

	/**
	 * Gets the index of the validity image in this tree element.
	 *
	 * @return the index of the validity image, or -1 if there is none
	 */
	int getValidityImageIndex();

	/**
	 * Sets the validity rectangle of this tree node.
	 *
	 * @param rect the rectangle where this tree node is valid
	 * @see   #getValidityRectangle()
	 */
	void setValidityRectangle(Rectangle rect);

	/**
	 * Gets the validity rectangle of this tree node.
	 *
	 * @return the rectangle where this tree node is valid
	 * @see    #setValidityRectangle(Rectangle)
	 */
	Rectangle getValidityRectangle();

	/**
	 * Gets the background color of this tree node.
	 *
	 * @return the background color
	 */
	Color getColor();

	/**
	 * Gets the background color of this tree node when selected.
	 *
	 * @return the background color when selected
	 */
	Color getSelectedColor();

	/**
	 * Gets the foreground color of this tree node.
	 *
	 * @return the foreground color
	 */
	Color getFontColor();

	/**
	 * Gets the font color of this tree node when selected.
	 *
	 * @return the font color when selected
	 */
	Color getSelectedFontColor();

	/**
	 * Gets the font of this tree node.
	 *
	 * @return the font
	 */
	Font getFont();

	/**
	 * Changes the validity of the node in a three step mode: toggles the validity of the node and
	 * restores its default if this is a parent node on the third click.
	 */
	void changeValidity();
}