/*
 * LyJTreeElementEntryImpl.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.util.*;
import leon.misc.*;
import leon.view.*;;


/**
 * Implementation of a LyJTreeElement for a Dictionnary.
 *
 * @version $Id: LyJTreeElementEntryImpl.java 28997 2010-12-23 10:17:15Z bmichel $
 * @author  Lyria
 * @see     LyJTreeElement
 * @see     LyDictionary
 */
public class LyJTreeElementEntryImpl extends LyJTreeElementImpl
{
	public static final String __VERSION =
		"$Id: LyJTreeElementEntryImpl.java 28997 2010-12-23 10:17:15Z bmichel $";

	/** The LyEntry associated with this component. */
	protected LyEntry		   _entry;

	/**
	 * Creates a default tree element.
	 */
	public LyJTreeElementEntryImpl()
	{
		super(null);
		_entry = null;
	}

	/**
	 * Creates a tree element for a class entry.
	 *
	 * @param entry the class entry
	 */
	public LyJTreeElementEntryImpl(LyEntry entry)
	{
		this();
		_entry = entry;
	}

	/**
	 * Gets the associated view component of this tree element class entry.
	 *
	 * @return the view component
	 */
	public LyComponent getViewComponent()
	{
		return _entry;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#isSelected()
	 */
	public boolean isSelected()
	{
		return _entry.isChecked();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getLabel()
	 */
	public String getLabel()
	{
		return ((_entry == null) ? null : _entry.getLabel());
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getEnvironment()
	 */
	public LyEnvironment getEnvironment()
	{
		return _entry.getEnvironment();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#isOpen()
	 */
	public boolean isOpen()
	{
		return _entry.isOpen();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getId()
	 */
	public String getId()
	{
		return _entry.getId();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElementImpl#updateValidityFromView()
	 */
	public void updateValidityFromView()
	{
		if (_entry == null)
			return;

		// Update children validity status
		if (_entry.getEntryCount() > 0)
		{
			LyTreeNode           currentNode;
			LyTreeNode.Validity     validity = _entry.getValidity();

			_treeValidityStep = 1;
			_treeValidityBackup = new Hashtable<LyComponent, LyTreeNode.Validity>();
            
			backupChildrenValidity(_entry);

			if ((validity == LyTreeNode.Validity.VALID)
				|| (validity == LyTreeNode.Validity.INVALID))
			{
				// Set all to invalid
				for (Enumeration<LyComponent> e = _treeValidityBackup.keys(); e.hasMoreElements();)
				{
					currentNode = (LyTreeNode)e.nextElement();
					currentNode.setValidity(validity);
				}
			}
		}

		// Now recompute the hierarchy
		computeTreeNodeValidity(getRootEntry(_entry));
	} // end method updateValidityFromView

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#changeValidity()
	 */
	public void changeValidity()
	{
		if (_entry == null)
			return;

		// behavior only if we have children
		if (_entry.getEntryCount() == 0)
		{
			// Just swap validity status
			if (_entry.getValidity() == LyTreeNode.Validity.VALID)
				_entry.setValidity(LyTreeNode.Validity.INVALID);
			else
				_entry.setValidity(LyTreeNode.Validity.VALID);
		}
		else
		{
			LyEntry currentNode;

			// Me must be able to 'undo' the process, so Keep all in memory backup
			_treeValidityStep++; // Next stage
			if (_treeValidityStep < 3)
			{
				// Backup them all
				if (_treeValidityStep == 1)
				{
					_treeValidityBackup = new Hashtable<LyComponent, LyTreeNode.Validity>();
					backupChildrenValidity(_entry);
				}
				// now swap their status :
				if (_entry.getValidity().equals(LyTreeNode.Validity.VALID))
				{
					// Set all to invalid
					for (Enumeration<LyComponent> e = _treeValidityBackup.keys(); e.hasMoreElements();)
					{
						currentNode = (LyEntry)e.nextElement();
						currentNode.setValidity(LyTreeNode.Validity.INVALID);
					}
				}
				else
				{
					// Set all to valid
					for (Enumeration<LyComponent> e = _treeValidityBackup.keys(); e.hasMoreElements();)
					{
						currentNode = (LyEntry)e.nextElement();
						currentNode.setValidity(LyTreeNode.Validity.VALID);
					}
				}
			}
			else
			{
				// Restore original values
				_treeValidityStep = 0;
                
				// Set all to valid
				for (Enumeration<LyComponent> e = _treeValidityBackup.keys(); e.hasMoreElements();)
				{
					currentNode = (LyEntry)e.nextElement();
					currentNode.setValidity(_treeValidityBackup.get(currentNode));
				}
			}
		}

		// Now recompute the hierarchy
		computeTreeNodeValidity(getRootEntry(_entry));
	} // end method changeValidity

	/**
	 * Gets the top level entry of the specified tree node.
	 *
	 * @param  treeNode the tree element to find its root tree entry
	 * @return the top level root tree entry
	 */
	public LyEntry getRootEntry(LyEntry treeNode)
	{
		while ((treeNode != null) && (treeNode.getParent() instanceof LyEntry))
			treeNode = (LyEntry)treeNode.getParent();

		return treeNode;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElementImpl#computeTreeNodeValidity(leon.view.LyComponent)
	 */
	public void computeTreeNodeValidity(LyComponent node)
	{
		LyEntry entry = (LyEntry)node;

		if (entry == null)
			return;

		int n = entry.getEntryCount();

		if (n == 0)
		{
			// Sets default value to valid.
			if (entry.getValidity() == null)
			{
				if (entry.isChecked())
					entry.setValidity(LyTreeNode.Validity.VALID);
				else
					entry.setValidity(LyTreeNode.Validity.INVALID);
			}
			return;
		} // Nothing to do on leaves

		if (entry.getValidity() == null)
			entry.setValidity(LyTreeNode.Validity.VALID);

		boolean validFound = false;
		boolean invalidFound = false;
		LyEntry currentNode;

		for (int i = 0; i < n; i++)
		{
			currentNode = entry.getEntry(i);
			computeTreeNodeValidity(currentNode); // Recurse deeper

			// We try to compute state at this level as long as partially is not already detected
			if ((!validFound) || (!invalidFound))
			{
				if ((!validFound)
						&& (currentNode.getValidity().equals(LyTreeNode.Validity.VALID)))
					validFound = true;
				else if ((!invalidFound)
						&& (currentNode.getValidity().equals(LyTreeNode.Validity.INVALID)))
					invalidFound = true;
				else if (currentNode.getValidity().equals(LyTreeNode.Validity.PARTIALLY))
				{
					validFound = true;
					invalidFound = true;
				}
			}
		}
		// Partially valid is detected
		if (validFound && invalidFound)
		{
			if (!entry.getValidity().equals(LyTreeNode.Validity.PARTIALLY))
				entry.setValidity(LyTreeNode.Validity.PARTIALLY);
		}
		else if (validFound)
		{
			if (!entry.getValidity().equals(LyTreeNode.Validity.VALID))
				entry.setValidity(LyTreeNode.Validity.VALID);
		}
		else if (invalidFound)
		{
			if (!entry.getValidity().equals(LyTreeNode.Validity.INVALID))
				entry.setValidity(LyTreeNode.Validity.INVALID);
		}
	} // end method computeTreeNodeValidity

	/**
	 * Backups the children validity status of the specified node.
	 *
	 * @param node the node to backup
	 */
	public void backupChildrenValidity(LyEntry node)
	{
		if (node == null)
			return;

		int i, n = node.getEntryCount();

		// Sets default value to valid.
		if (node.getValidity() == null)
		{
			if (n == 0)
				node.setValidity(node.isChecked() ? LyTreeNode.Validity.VALID
												  : LyTreeNode.Validity.INVALID);
			else
				node.setValidity(LyTreeNode.Validity.VALID);
		}

		for (i = 0; i < n; i++)
			backupChildrenValidity(node.getEntry(i)); // Recurse

		if (n == 0)
			_treeValidityBackup.put(node, node.getValidity());
	}

} // end class LyJTreeElementEntryImpl
