/*
 * LyJTreeElementImpl.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;
import java.util.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import leon.misc.*;
import leon.view.*;


/**
 * This class implements the Tree Node object (as described in the dtd) for swing viewer.
 *
 * @version $Id: LyJTreeElementImpl.java 28997 2010-12-23 10:17:15Z bmichel $
 * @author  Lyria
 * @see     LyTreeNode
 */
public class LyJTreeElementImpl
	implements LyJTreeElement
{
	public static final String							  __VERSION =
		"$Id: LyJTreeElementImpl.java 28997 2010-12-23 10:17:15Z bmichel $";

	/** The object contained by this node. */
	protected _LyNodeContent							  _treeNodeContent;

	/** Used to notify the displayed tree that a node has changed. */
	protected LyJTreeComponent							  _treeComp;

	/** Mutable tree node object related to the current node. */
	protected _LyMutableTreeNode						  _mutableTreeNode;

	/** The tree node. */
	protected LyTreeNode								  _treeNode;

	/** Height factor of the current node. */
	protected int										  _heightFactor;

	/** The empty checkbox image. */
	protected Image										  _emptyCheckboxImage;

	/** The selected checkbox image. */
	protected Image										  _selectedCheckboxImage;

	/** The list of additional images. */
	protected Vector<Image>								  _additionalImages;

	/** The validity image. */
	protected Image										  _validityImage;

	/** The validity rectangle. */
	protected Rectangle									  _validityRect;

	/** Backup all node validity for undo. */
	// Backup all node validity because on step 3 we have to undo
	protected Hashtable<LyComponent, LyTreeNode.Validity> _treeValidityBackup;

	/**
	 * Step is 0 when no change, then 1 (toggles to valid or invalid), then 2 (re-toggle), then 3
	 * (undo) before returning to 0.
	 */
	protected int										  _treeValidityStep;

	/** Indicates that the validity is being changed by a gui action. */
	protected boolean									  _guiValidityChange;

	/**
	 * Creates a tree element from the specified tree node.
	 *
	 * @param node the data source
	 */
	public LyJTreeElementImpl(LyTreeNode node)
	{
		_treeNode = node;
		_emptyCheckboxImage = null;
		_selectedCheckboxImage = null;
		_additionalImages = null;
		_validityImage = null;
		_validityRect = null;
		_treeValidityBackup = null;
		_treeValidityStep = 0;
		_guiValidityChange = false;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getMutableTreeNode()
	 */
	public DefaultMutableTreeNode getMutableTreeNode()
	{
		return _mutableTreeNode;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getContent()
	 */
	public _LyNodeContent getContent()
	{
		return _mutableTreeNode.getNodeContent();
	}

	/**
	 * Set the height factor.
	 *
	 * @param factor the new height factor
	 * @see   #getHeightFactor()
	 */
	public void setHeightFactor(int factor)
	{
		_heightFactor = factor;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getHeightFactor()
	 */
	public int getHeightFactor()
	{
		return _heightFactor;
	}

	/**
	 * Sets the empty checkbox image.
	 *
	 * @param image the new empty checkbox image
	 * @see   #getEmptyCheckboxImage()
	 */
	public void setEmptyCheckboxImage(Image image)
	{
		_emptyCheckboxImage = image;
	}

	/**
	 * Sets the selected checkbox image.
	 *
	 * @param image the new selected checkbox image
	 * @see   #getSelectedCheckboxImage()
	 */
	public void setSelectedCheckboxImage(Image image)
	{
		_selectedCheckboxImage = image;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getSelectedCheckboxImage()
	 */
	public Image getSelectedCheckboxImage()
	{
		return _selectedCheckboxImage;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getEmptyCheckboxImage()
	 */
	public Image getEmptyCheckboxImage()
	{
		return _emptyCheckboxImage;
	}

	/**
	 * Adds the specified image to the list of additional images.
	 *
	 * @param image the new image to add
	 * @see   #removeAdditionalImage(Image)
	 */
	public void addAdditionalImage(Image image)
	{
		if (_additionalImages == null)
			_additionalImages = new Vector<Image>(1);

		if (image != null)
			_additionalImages.addElement(image);
	}

	/**
	 * Removes the specified image from the additional images list.
	 *
	 * @param image the image to remove
	 * @see   #addAdditionalImage(Image)
	 */
	public void removeAdditionalImage(Image image)
	{
		if ((_additionalImages == null) || (image == null))
			return;

		_additionalImages.removeElement(image);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getAdditionalImages()
	 */
	public Vector<Image> getAdditionalImages()
	{
		return _additionalImages;
	}

	/**
	 * Sets the validity image.
	 *
	 * @param image the new validity image
	 */
	public void setValidityImage(Image image)
	{
		if (_validityImage != null)
		{
			int idx = _additionalImages.indexOf(_validityImage);

			if (idx >= 0)
			{
				// Replace
				_additionalImages.remove(_validityImage);
				_additionalImages.insertElementAt(image, idx);
			}
			else
				addAdditionalImage(image);
		}
		else
			addAdditionalImage(image);
		_validityImage = image;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getValidityImageIndex()
	 */
	public int getValidityImageIndex()
	{
		if ((_validityImage != null) && (_additionalImages != null))
			return _additionalImages.indexOf(_validityImage);
		return -1;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#setValidityRectangle(java.awt.Rectangle)
	 */
	public void setValidityRectangle(Rectangle rect)
	{
		_validityRect = rect;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getValidityRectangle()
	 */
	public Rectangle getValidityRectangle()
	{
		return _validityRect;
	}

	/**
	 * Updates the validity from the view component.
	 */
	public void updateValidityFromView()
	{
		if (_guiValidityChange)
			return;

		// Update children validity status
		if (_treeNode.getTreeNodeCount() > 0)
		{
			LyTreeNode		    currentNode;
			LyTreeNode.Validity validity = _treeNode.getValidity();

			_treeValidityStep = 0;
			_treeValidityBackup = new Hashtable<LyComponent, LyTreeNode.Validity>();
            
			backupChildrenValidity(_treeNode);

			if (validity.equals(LyTreeNode.Validity.VALID)
					|| validity.equals(LyTreeNode.Validity.INVALID))
			{
				// Set all to invalid
				for (Enumeration<LyComponent> e = _treeValidityBackup.keys(); e.hasMoreElements();)
				{
					currentNode = (LyTreeNode)e.nextElement();
					currentNode.setValidity(validity);
				}
			}
		}

		// Now recompute the hierarchy
		computeTreeNodeValidity(getRootTreeNode(_treeNode));
	} // end method updateValidityFromView

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#changeValidity()
	 */
	public void changeValidity()
	{
		if (_treeNode == null)
			return;

		_guiValidityChange = true;

		// behavior differ depending on children
		if (_treeNode.getTreeNodeCount() == 0)
		{
			// Just swap validity status
			if (_treeNode.getValidity().equals(LyTreeNode.Validity.VALID))
				_treeNode.setValidity(LyTreeNode.Validity.INVALID);
			else
				_treeNode.setValidity(LyTreeNode.Validity.VALID);
		}
		else
		{
			LyTreeNode currentNode;

			// Me must be able to 'undo' the process, so Keep all in memory backup
			_treeValidityStep++; // Next stage

			if (_treeValidityStep < 3)
			{
				// Backup them all
				if (_treeValidityStep == 1)
				{
					// If current node is valid or invalid (indicating that all its children have the
					// same validity), do not use a three state mode but only switch between valid
					// and invalid modes.
					if (_treeNode.getValidity().equals(LyTreeNode.Validity.VALID)
							|| _treeNode.getValidity().equals(LyTreeNode.Validity.INVALID))
						_treeValidityStep++;

					_treeValidityBackup = new Hashtable<LyComponent, LyTreeNode.Validity>();
					backupChildrenValidity(_treeNode);
				}

				// now swap their status :
				if (_treeNode.getValidity().equals(LyTreeNode.Validity.VALID))
				{
					// Set all to invalid
					for (Enumeration<LyComponent> e = _treeValidityBackup.keys(); e.hasMoreElements();)
					{
						currentNode = (LyTreeNode)e.nextElement();
						currentNode.setValidity(LyTreeNode.Validity.INVALID);
					}
				}
				else
				{
					// Set all to valid
					for (Enumeration<LyComponent> e = _treeValidityBackup.keys(); e.hasMoreElements();)
					{
						currentNode = (LyTreeNode)e.nextElement();
						currentNode.setValidity(LyTreeNode.Validity.VALID);
					}
				}
			}
			else
			{
				// Restore original values
				_treeValidityStep = 0;
				// Set all to valid
				for (Enumeration<LyComponent> e = _treeValidityBackup.keys(); e.hasMoreElements();)
				{
					currentNode = (LyTreeNode)e.nextElement();
					currentNode.setValidity(_treeValidityBackup.get(currentNode));
				}
			}
		}

		// Now recompute the hierarchy
		computeTreeNodeValidity(getRootTreeNode(_treeNode));

		_guiValidityChange = false;
	} // end method changeValidity

	/**
	 * Sets the expandable state of the tree element.
	 */
	public void setExpandable(boolean expandable)
	{
		_mutableTreeNode.setExpandable(expandable);
	}
	
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getColor()
	 */
	public Color getColor()
	{
		LyEnvironment env = getEnvironment();

		if ((_treeNode != null) && (_treeNode.getColor() != null) && (env != null))
			return (Color)env.getColor(_treeNode.getColor());
		return null;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getSelectedColor()
	 */
	public Color getSelectedColor()
	{
		LyEnvironment env = getEnvironment();

		if (env != null)
			return (Color)env.getColor("LY_TREENODE_SELECTED_BG_COLOR");
		return null;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getFontColor()
	 */
	public Color getFontColor()
	{
		LyEnvironment env = getEnvironment();

		if ((_treeNode != null) && (_treeNode.getFontColor() != null) && (env != null))
			return (Color)env.getColor(_treeNode.getFontColor());
		else if (env != null)
			return (Color)env.getColor("LY_TREENODE_FONT_COLOR");
		return null;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getSelectedFontColor()
	 */
	public Color getSelectedFontColor()
	{
		LyEnvironment env = getEnvironment();

		if (env != null)
			return (Color)env.getColor("LY_TREENODE_SELECTED_FONT_COLOR");
		return null;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getFont()
	 */
	public Font getFont()
	{
		LyEnvironment env = getEnvironment();

		if ((_treeNode != null) && (_treeNode.getFont() != null) && (env != null))
			return (Font)env.getFont(_treeNode.getFont());
		return null;
	}

	/**
	 * Gets the top level tree node of the specified tree element.
	 *
	 * @param  treeNode the tree element to find its root tree node
	 * @return the top level root tree node
	 */
	public LyTreeNode getRootTreeNode(LyTreeNode treeNode)
	{
		while ((treeNode != null) && (treeNode.getParent() instanceof LyTreeNode))
			treeNode = (LyTreeNode)treeNode.getParent();

		return treeNode;
	}

	/**
	 * Backups the children validity status of the specified node.
	 *
	 * @param treeNode the node to backup
	 */
	public void backupChildrenValidity(LyTreeNode treeNode)
	{
		if (treeNode == null)
			return;

		int i, n = treeNode.getTreeNodeCount();

		for (i = 0; i < n; i++)
			backupChildrenValidity(treeNode.getTreeNode(i)); // Recurse

		if (n == 0)
			_treeValidityBackup.put(treeNode, treeNode.getValidity());
	}

	/**
	 * Clears subtree validity backup infos.
	 */
	public void clearValidityBackupInfos()
	{
		_treeValidityBackup = null;
		_treeValidityStep = 0;
	}

	/**
	 * Computes the subtree validity.
	 *
	 * @param node the current node which may be a LySwingTreeNode or a LySwingEntry (dictionnary).
	 */
	public void computeTreeNodeValidity(LyComponent node)
	{
		LyTreeNode treeNode = (LyTreeNode)node;

		if (treeNode == null)
			return;

		int n = treeNode.getTreeNodeCount();

		// Sets default value to valid.
		if (treeNode.getValidity() == null)
			treeNode.setValidity(LyTreeNode.Validity.VALID);

		if (n == 0)
			return; // Nothing to do on leaves

		boolean    validFound = false;
		boolean    invalidFound = false;
		LyTreeNode currentNode;

		for (int i = 0; i < n; i++)
		{
			currentNode = treeNode.getTreeNode(i);
			computeTreeNodeValidity(currentNode); // Recurse deeper

			// We try to compute state at this level as long as partially is not already detected
			if ((!validFound) || (!invalidFound))
			{
				if ((!validFound)
						&& (currentNode.getValidity().equals(LyTreeNode.Validity.VALID)))
					validFound = true;
				else if ((!invalidFound)
						&& (currentNode.getValidity().equals(LyTreeNode.Validity.INVALID)))
					invalidFound = true;
				else if (currentNode.getValidity().equals(LyTreeNode.Validity.PARTIALLY))
				{
					validFound = true;
					invalidFound = true;
				}
			}
		}
		// Partially valid is detected
		if (validFound && invalidFound)
		{
			if (!treeNode.getValidity().equals(LyTreeNode.Validity.PARTIALLY))
				treeNode.setValidity(LyTreeNode.Validity.PARTIALLY);
		}
		else if (validFound)
		{
			if (!treeNode.getValidity().equals(LyTreeNode.Validity.VALID))
				treeNode.setValidity(LyTreeNode.Validity.VALID);
		}
		else if (invalidFound)
		{
			if (!treeNode.getValidity().equals(LyTreeNode.Validity.INVALID))
				treeNode.setValidity(LyTreeNode.Validity.INVALID);
		}
	} // end method computeTreeNodeValidity

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getEnvironment()
	 */
	public LyEnvironment getEnvironment()
	{
		return _treeNode.getEnvironment();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getId()
	 */
	public String getId()
	{
		return _treeNode.getId();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getLabel()
	 */
	public String getLabel()
	{
		return _treeNode.getLabel();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getTreePath()
	 */
	public TreePath getTreePath()
	{
		TreeNode[] nodesPath = _mutableTreeNode.getPath();
		TreePath   path = new TreePath(nodesPath);

		return path;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#isOpen()
	 */
	public boolean isOpen()
	{
		return _treeNode.isOpen();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#isSelected()
	 */
	public boolean isSelected()
	{
		return _treeNode.isSelected();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#setTree(leon.view.win.swing.component.LyJTreeComponent)
	 */
	public void setTree(LyJTreeComponent tree)
	{
		_treeComp = tree;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.component.LyJTreeElement#getTree()
	 */
	public LyJTreeComponent getTree()
	{
		return _treeComp;
	}

	/**
	 * Sets the mutable tree node of this tree element.
	 *
	 * @param mutableTreeNode the new mutable tree node
	 * @see   #getMutableTreeNode()
	 */
	public void setMutableTreeNode(_LyMutableTreeNode mutableTreeNode)
	{
		_mutableTreeNode = mutableTreeNode;
	}

	/**
	 * Nothing is done when the tree is expanded.
	 *
	 * @param event the <code>TreeExpansionEvent</code>, not used in this implementation
	 * @see   javax.swing.event.TreeExpansionListener#treeExpanded(javax.swing.event.TreeExpansionEvent)
	 */
	public void treeExpanded(TreeExpansionEvent event)
	{
	}

	/**
	 * Nothing is done when the tree is collapsed.
	 *
	 * @param event the <code>TreeExpansionEvent</code>, not used in this implementation
	 * @see   javax.swing.event.TreeExpansionListener#treeCollapsed(javax.swing.event.TreeExpansionEvent)
	 */
	public void treeCollapsed(TreeExpansionEvent event)
	{
	}

	/* ****************************** Internal Classes *******************************/

	/**
	 * This class implements the informations that are useful to correctly display a node.
	 */
	public static class _LyNodeContent
	{

		/** The image associated with this node. */
		Image   _image; // cached image to increase speed

		/** The label of this node. */
		String  _label;

		/** The tooltip text of this node. */
		String  _tooltip;

		/** Indicates that this node is selected or not. */
		boolean _selected;

		/** Indicates that this node is opened or not. */
		boolean _open;

		/**
		 * Creates a node with the specified informations. The tooltip text of this node content is
		 * not set.
		 *
		 * @param image    the image of this node
		 * @param label    the text label
		 * @param selected Indicates that this node is selected.
		 * @param open     Indicates that this node is opened.
		 */
		public _LyNodeContent(Image image, String label, boolean selected, boolean open)
		{
			this(image, label, null, selected, open);
		}

		/**
		 * Creates a node with the specified informations.
		 *
		 * @param image    the image of this node
		 * @param label    the text label
		 * @param tooltip  the tooltip text
		 * @param selected Indicates that this node is selected.
		 * @param open     Indicates that this node is opened.
		 */
		public _LyNodeContent(Image image, String label, String tooltip, boolean selected,
			boolean open)
		{
			_image = image;
			_label = label;
			_tooltip = tooltip;
			_selected = selected;
			_open = open;
		}

		/**
		 * Gets the image of this node.
		 *
		 * @return the image
		 * @see    #setImage(Image)
		 */
		public Image getImage()
		{
			return _image;
		}

		/**
		 * Sets the image of this node.
		 *
		 * @param im the new image
		 * @see   #getImage()
		 */
		public void setImage(Image im)
		{
			_image = im;
		}

		/**
		 * Gets the text label of this node.
		 *
		 * @return the label
		 * @see    #setLabel(String)
		 */
		public String getLabel()
		{
			return _label;
		}

		/**
		 * Sets the text label of this node.
		 *
		 * @param label the new text label
		 * @see   #getLabel()
		 */
		public void setLabel(String label)
		{
			_label = label;
		}

		/**
		 * Gets the tooltip text of this node.
		 *
		 * @return the tooltip
		 * @see    #setTooltip(String)
		 */
		public String getTooltip()
		{
			return _tooltip;
		}

		/**
		 * Sets the tooltip text of this node.
		 *
		 * @param tooltip the new tooltip
		 * @see   #getTooltip()
		 */
		public void setTooltip(String tooltip)
		{
			_tooltip = tooltip;
		}

		/**
		 * Indicates that this node is selected.
		 *
		 * @return <code>true</code> if this node is selected, <code>false</code> otherwise
		 * @see    #setSelected(boolean)
		 */
		public boolean isSelected()
		{
			return _selected;
		}

		/**
		 * Sets the selected status of this node.
		 *
		 * @param selected Indicates that this node is selected or not.
		 * @see   #isSelected()
		 */
		public void setSelected(boolean selected)
		{
			_selected = selected;
		}

		/**
		 * Indicates that this node is opened.
		 *
		 * @return <code>true</code> if this node is opened, <code>false</code> otherwise
		 * @see    #setOpen(boolean)
		 */
		public boolean isOpen()
		{
			return _open;
		}

		/**
		 * Sets the opened status of this node.
		 *
		 * @param open Indicates that this node is opened or not.
		 * @see   #isOpen()
		 */
		public void setOpen(boolean open)
		{
			_open = open;
		}

		/*
		 * (non-Javadoc)
		 * @see java.lang.Object#toString()
		 */
		public String toString()
		{
			return getLabel();
		}
	} // end class _LyNodeContent


	/**
	 * Defines a mutable tree node for this tree element.
	 */
	public static class _LyMutableTreeNode extends DefaultMutableTreeNode
	{
		private static final long serialVersionUID = 1759834296275224468L;

		/** The node content. */
		protected _LyNodeContent  _nodeContent;

		/** The associated tree element. */
		protected LyJTreeElement  _treeElement;
		
		/** Expandable state of the tree element. */
		protected boolean		  _expandable = false;

		/**
		 * Sets the content of the node.
		 *
		 * @param content the new content of this node
		 * @see   #getNodeContent()
		 */
		public void setNodeContent(_LyNodeContent content)
		{
			_nodeContent = content;
		}

		/**
		 * Gets the content of the node.
		 *
		 * @return the content of the node
		 * @see    #setNodeContent
		 */
		public _LyNodeContent getNodeContent()
		{
			return _nodeContent;
		}

		/**
		 * Sets the tree element of this node.
		 *
		 * @param treeElement the new tree element
		 * @see   #getTreeElement()
		 */
		public void setTreeElement(LyJTreeElement treeElement)
		{
			_treeElement = treeElement;
		}

		/**
		 * Gets the tree element of this node.
		 *
		 * @return the tree element
		 * @see    #setTreeElement(LyJTreeElement)
		 */
		public LyJTreeElement getTreeElement()
		{
			return _treeElement;
		}
		
		/**
		 * Sets the expandable state of the tree element.
		 */
		public void setExpandable(boolean expandable)
		{
			_expandable = expandable;
		}
		
		/*
		 * (non-Javadoc)
		 * @see javax.swing.tree.DefaultMutableTreeNode#isLeaf()
		 */
		public boolean isLeaf()
		{
			return (_expandable ? false : super.isLeaf());
		}

		/*
		 * (non-Javadoc)
		 * @see java.lang.Object#toString()
		 */
		public String toString()
		{
			return _nodeContent.toString();
		}
	} // end class _LyMutableTreeNode
} // end class LyJTreeElementImpl
