/*
 * VFlowLayout.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.component;

import java.awt.*;


/**
 * VFlowLayout is similair to FlowLayout except it lays out components vertically.
 *
 * <p>Extends FlowLayout because it mimics much of the behavior of the FlowLayout class, except
 * vertically. An additional feature is that you can specify a fill to edge flag, which causes the
 * VFlowLayout manager to resize all components to expand to the column width.</p>
 *
 * <p>Warning: This causes problems when the main panel has less space that it needs and it seems to
 * prohibit multi-column output.</p>
 *
 * @version $Id: VFlowLayout.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class VFlowLayout extends FlowLayout
{
	private static final long  serialVersionUID = -5196218847796136295L;

	public static final String __VERSION =
		"$Id: VFlowLayout.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The top alignment. */
	public static final int    TOP = 0;

	/** The middle alignment. */
	public static final int    MIDDLE = 1;

	/** The bottom alignment. */
	public static final int    BOTTOM = 2;

	/** The alignment policy. {@link #TOP}, {@link #MIDDLE} or {@link #BOTTOM}. */
	int						   align;

	/** The horizontal gap, in pixels. */
	int						   hgap;

	/** The vertical gap, in pixels. */
	int						   vgap;

	/** Indicates that this layout fills the empty space or not. */
	boolean					   fill;

	/**
	 * Constructs a new VFlowLayout with a middle alignemnt, and the fill to edge flag set.
	 */
	public VFlowLayout()
	{
		this(MIDDLE, 5, 5, true);
	}

	/**
	 * Constructs a new VFlowLayout with a middle alignemnt.
	 *
	 * @param fill the fill to edge flag
	 */
	public VFlowLayout(boolean fill)
	{
		this(MIDDLE, 5, 5, fill);
	}

	/**
	 * Constructs a new VFlowLayout with a middle alignemnt.
	 *
	 * @param align the alignment value
	 */
	public VFlowLayout(int align)
	{
		this(align, 5, 5, true);
	}

	/**
	 * Constructs a new VFlowLayout.
	 *
	 * @param align the alignment value
	 * @param fill  the fill to edge flag
	 */
	public VFlowLayout(int align, boolean fill)
	{
		this(align, 5, 5, fill);
	}

	/**
	 * Constructs a new VFlowLayout.
	 *
	 * @param align the alignment value
	 * @param hgap  the horizontal gap variable
	 * @param vgap  the vertical gap variable
	 * @param fill  the fill to edge flag
	 */
	public VFlowLayout(int align, int hgap, int vgap, boolean fill)
	{
		this.align = align;
		this.hgap = hgap;
		this.vgap = vgap;
		this.fill = fill;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.LayoutManager#preferredLayoutSize(java.awt.Container)
	 */
	public Dimension preferredLayoutSize(Container target)
	{
		Dimension tarsiz = new Dimension(0, 0);

		for (int i = 0; i < target.getComponentCount(); i++)
		{
			Component m = target.getComponent(i);

			if (m.isVisible())
			{
				Dimension d = m.getPreferredSize();

				tarsiz.width = Math.max(tarsiz.width, d.width);
				if (i > 0)
					tarsiz.height += vgap;
				tarsiz.height += d.height;
			}
		}
		Insets insets = target.getInsets();

		tarsiz.width += insets.left + insets.right + (hgap * 2);
		tarsiz.height += insets.top + insets.bottom + (vgap * 2);
		return tarsiz;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.LayoutManager#minimumLayoutSize(java.awt.Container)
	 */
	public Dimension minimumLayoutSize(Container target)
	{
		Dimension tarsiz = new Dimension(0, 0);

		for (int i = 0; i < target.getComponentCount(); i++)
		{
			Component m = target.getComponent(i);

			if (m.isVisible())
			{
				Dimension d = m.getMinimumSize();

				tarsiz.width = Math.max(tarsiz.width, d.width);
				if (i > 0)
					tarsiz.height += vgap;
				tarsiz.height += d.height;
			}
		}
		Insets insets = target.getInsets();

		tarsiz.width += insets.left + insets.right + (hgap * 2);
		tarsiz.height += insets.top + insets.bottom + (vgap * 2);
		return tarsiz;
	}

	/**
	 * Places the components defined by first to last within the target container using the bounds
	 * box defined.
	 *
	 * @param target the container
	 * @param x      the x coordinate of the area
	 * @param y      the y coordinate of the area
	 * @param width  the width of the area
	 * @param height the height of the area
	 * @param first  the first component of the container to place
	 * @param last   the last component of the container to place
	 */
	private void placethem(Container target, int x, int y, int width, int height, int first,
		int last)
	{
		if (align == VFlowLayout.MIDDLE)
			y += height / 2;
		if (align == VFlowLayout.BOTTOM)
			y += height;

		for (int i = first; i < last; i++)
		{
			Component m = target.getComponent(i);
			Dimension md = m.getSize();

			if (m.isVisible())
			{
				int px = x;

				if (getAlignment() == CENTER)
					px += (width - md.width) / 2;
				else if (getAlignment() == RIGHT)
					px += width - md.width;

				m.setLocation(px, y);
				y += vgap + md.height;
			}
		}
	}

	/**
	 * Lays out the container.
	 *
	 * @param target the container to lay out.
	 * @see   java.awt.LayoutManager#layoutContainer(java.awt.Container)
	 */
	public void layoutContainer(Container target)
	{
		Insets insets = target.getInsets();
		int    maxheight = target.getSize().height - (insets.top + insets.bottom + (vgap * 2));
		int    maxwidth = target.getSize().width - (insets.left + insets.right + (hgap * 2));
		int    numcomp = target.getComponentCount();
		int    x = insets.left + hgap, y = 0;
		int    colw = 0, start = 0;

		for (int i = 0; i < numcomp; i++)
		{
			Component m = target.getComponent(i);

			if (m.isVisible())
			{
				Dimension d = m.getPreferredSize();

				if (this.fill)
				{
					m.setSize(maxwidth, d.height);
					d.width = maxwidth;
				}
				else
					m.setSize(d.width, d.height);

				if ((y == 0) || ((y + d.height) <= maxheight))
				{
					if (y > 0)
						y += vgap;
					y += d.height;
					colw = Math.max(colw, d.width);
				}
				else
				{
					placethem(target, x, insets.top + vgap, colw, maxheight - y, start, i);
					y = d.height;
					x += hgap + colw;
					colw = d.width;
					start = i;
				}
			}
		}
		placethem(target, x, insets.top + vgap, colw, maxheight - y, start, numcomp);
	} // end method layoutContainer
} // end class VFlowLayout
