/*
 * LyJHeaderRenderer.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.table;

import java.awt.*;
import java.awt.image.*;

import javax.swing.*;
import javax.swing.table.*;
import leon.view.*;
import leon.view.win.swing.*;
import leon.view.win.swing.component.*;


/**
 * This class configures the style of a header cell during display.
 *
 * @version $Id: LyJHeaderRenderer.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class LyJHeaderRenderer extends DefaultTableCellRenderer
{
	private static final long  serialVersionUID = 6054535008928093995L;

	public static final String __VERSION =
		"$Id: LyJHeaderRenderer.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The border used for a header cell. */
	LyJSpecialCellBorder	   _headerCellBorder;

	/** The font of the header text. */
	Font					   _font;

	/** The background color of the header. */
	Color					   _color;

	/** The foreground color of the header. */
	Color					   _foreground;

	/** Indicates that the header is hierarchical. */
	boolean					   _hierarchical;

	/**
	 * Creates a header renderer for the specified attributes.
	 *
	 * @param font         the font of the header
	 * @param color        the background color
	 * @param foreground   the foreground color
	 * @param hierarchical Indicates that this header is hierarchical.
	 */
	public LyJHeaderRenderer(Font font, Color color, Color foreground, boolean hierarchical)
	{
		super();
		_headerCellBorder = new LyJSpecialCellBorder(3, 3, 3, 3);
		_font = font;
		_color = color;
		_foreground = foreground;
		_hierarchical = hierarchical;
	}

	/**
	 * Gets the cell renderer for this header.
	 *
	 * @see javax.swing.table.TableCellRenderer#getTableCellRendererComponent(javax.swing.JTable,
	 *      java.lang.Object, boolean, boolean, int, int)
	 */
	public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected,
		boolean hasFocus, int row, int column)
	{
		// Call the super method to correctly configure display parameters
		// Comp is equal to 'this'
		JLabel			  comp = (JLabel)super.getTableCellRendererComponent(table, "", isSelected,
				hasFocus, row, column);
		LyJTableComponent tableComp = (LyJTableComponent)table;

		setOpaque(true);

		if (!getBackground().equals(_color))
			setBackground(_color);

		if (!getForeground().equals(_foreground))
			setForeground(_foreground);

		// empty header
		if (value instanceof String)
		{
			setBorder(_headerCellBorder);
			setIcon(null);
			setText("");
		}
		// normal header
		else if (value instanceof LySwingHeader)
		{
			LySwingHeader header = (LySwingHeader)value;

			if (!getFont().equals(_font))
				setFont(_font);

			if ((header.isSortable()) && (header.getSortableHeaderFont() != null))
				setFont(header.getSortableHeaderFont());

			setBorder(_headerCellBorder);

			// sorted column
			if (header.getSort() >= 0)
			{
				FontMetrics fm = getFontMetrics(getFont());

				setIcon((header.isIncreasing()) ? (tableComp.getSortUpIcon())
												: (tableComp.getSortDownIcon()));

				// avalaible space
				// FFT01118
				int    width = tableComp.getTableHeader().getHeaderRect(column).getSize().width - 3;

				int    spaceCount;
				int    sortNumber = (header.getSort() + 1);
				String space = "", clipString = "...";

				// width of space character
				int    spaceWidth = fm.stringWidth(" ");

				Image	headerImage = header.getImage();
				int		headerImageWidth = (headerImage == null ? 0 : headerImage.getWidth(this) + 1);

				// width of label to display
				String	initialHeaderLabel = header.getLabel();
				
				if (initialHeaderLabel == null)
					initialHeaderLabel = "";
				
				int		textWidth = fm.stringWidth(initialHeaderLabel);

				// final label to display
				String	headerLabel;

				// width of sort number
				int    sortWidth = fm.stringWidth("" + sortNumber) + 2;

				// icon gap: icon and sor number
				int    gap = (10 + sortWidth);

				// available space for label
				int    availTextWidth = width - gap - 5 - headerImageWidth;

				// enough space to display label
				if (availTextWidth > textWidth)
				{
					//add space characters to a good sort number position
					if (header.getAlign() == SwingConstants.RIGHT)
						spaceCount = (sortWidth + 10) / (spaceWidth);
					else if (header.getAlign() == SwingConstants.CENTER)
						spaceCount = ((width - textWidth - sortWidth - 5)/2) / (spaceWidth);
					else
						spaceCount = (width - textWidth - sortWidth - headerImageWidth) / (spaceWidth);

					for (int i = 0; i < spaceCount; i++)
						space += " ";

					headerLabel = initialHeaderLabel + space + sortNumber;
				}
				else
				{
					// no place nothing
					if (availTextWidth <= 0)
					{
						headerLabel = "";
						setIcon(null);
					}
					else
					{
						int    nChars;
						int    totalWidth = fm.stringWidth(clipString);
						String text = header.getLabel();

						if (text == null)
							text = "";

						// enough place for a clipString
						if (totalWidth < availTextWidth)
						{
							for (nChars = 0; nChars < text.length(); nChars++)
							{
								totalWidth += fm.charWidth(text.charAt(nChars));
								if (totalWidth > availTextWidth)
									break;
							}
							text = text.substring(0, nChars) + clipString;
						}
						else
						// not enough place for clipString
						{
							text = clipString;
							while (text.length() > 0)
							{
								text = text.substring(0, text.length() - 1);
								totalWidth = fm.stringWidth(text);
								if (totalWidth < availTextWidth)
									break;
							}
						}

						// add space characters to a good sort number position
						totalWidth = fm.stringWidth(text) + sortWidth;
						spaceCount = (width - totalWidth - headerImageWidth) / (spaceWidth);
						for (int i = 0; i < spaceCount; i++)
							space += " ";

						headerLabel = text + space + sortNumber;
					}

				}
				
				setText(headerLabel);
				setIconTextGap(-gap);
				setHorizontalAlignment(header.getAlign());
				setHorizontalTextPosition(JLabel.LEADING);

				if (headerImage != null)
				{
					Image			sortImage = (header.isIncreasing() ? tableComp.getSortUpImage() : tableComp.getSortDownImage());
					int				imageHeight = Math.max(headerImage.getHeight(this), sortImage.getHeight(this));
					
					if (width > 0)
					{
						BufferedImage	image = new BufferedImage(width, imageHeight, BufferedImage.TYPE_INT_ARGB);
						
						if (header.getAlign() == SwingConstants.RIGHT)
							image.getGraphics().drawImage(headerImage, width - sortImage.getWidth(this) - fm.stringWidth(headerLabel) - 2 - sortWidth,
									(imageHeight - headerImage.getHeight(this)) / 2, this);
						else if (header.getAlign() == SwingConstants.CENTER)
						{
							if (headerLabel.length() == 0)
								image.getGraphics().drawImage(headerImage, (width - headerImage.getWidth(this)) / 2,
										(imageHeight - headerImage.getHeight(this)) / 2, this);
							else
								image.getGraphics().drawImage(headerImage, (width - fm.stringWidth(headerLabel)) - headerImage.getWidth(this) - 4,
										(imageHeight - headerImage.getHeight(this)) / 2, this);
						}
						else
							image.getGraphics().drawImage(headerImage, 0, (imageHeight - headerImage.getHeight(this)) / 2, this);
							
						if (width > headerImageWidth + sortWidth + 15)
							image.getGraphics().drawImage(sortImage, width - sortImage.getWidth(this) - 5 - sortWidth,
									(imageHeight - sortImage.getHeight(this)) / 2, this);
						
						setIcon(new ImageIcon(image));
						
						if (header.getAlign() == SwingConstants.RIGHT)
							setIconTextGap(-fm.stringWidth(headerLabel) - 4);
						else if (header.getAlign() == SwingConstants.CENTER)
							setIconTextGap(-fm.stringWidth(headerLabel) - 4);
						else
							setIconTextGap(-width + gap + headerImage.getWidth(this) - 10 - sortWidth);
						
						setHorizontalTextPosition(JLabel.TRAILING);
					}
				}
			}
			else
			{
				Image	image = header.getImage();
				
				setIcon(image == null ? null : new ImageIcon(image));
				setText(header.getLabel());
				setIconTextGap(0);
				setHorizontalAlignment(header.getAlign());
				setHorizontalTextPosition(JLabel.TRAILING);
			}
			
			if (_hierarchical)
				setVerticalAlignment(SwingConstants.BOTTOM);
		}
		// normal headers
		else if (value instanceof LyHeaders)
		{
			LyHeaders headers = (LyHeaders)value;

			if (!getFont().equals(_font))
				setFont(_font);

			setBorder(_headerCellBorder);
			setIcon(null);
			setText(headers.getLabel());
			setHorizontalAlignment(SwingConstants.LEFT);
			setHorizontalTextPosition(SwingConstants.TRAILING);
		}
		// Unreachable
		else
			System.out.println("LyJHeaderRenderer error : unrecognized value");
		return comp;
	} // end method getTableCellRendererComponent
} // end class LyJHeaderRenderer
