/*
 * LyJTableFooter.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.table;

import java.awt.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import javax.swing.table.*;
import leon.view.win.swing.*;


/**
 * This class implements a row of cells that can display some text.
 *
 * <p>There is a cell for each column of the parent table. This row is displayed at the bottom of
 * the parent table and doesn't move with the vertical scrollbar</p>
 *
 * @version $Id: LyJTableFooter.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class LyJTableFooter extends JPanel
	implements LayoutManager,
		TableColumnModelListener
{
	private static final long  serialVersionUID = -5006241448316619877L;

	public static final String __VERSION =
		"$Id: LyJTableFooter.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The height of a this footer. */
	public int				   HEIGHT;

	/** The parent table. */
	protected JTable		   _tableComp;

	/** The columnModel of the table. */
	protected TableColumnModel _columnModel;

	/**
	 * Creates a footer for the specified table, initializes and sets its content.
	 *
	 * @param table the parent table
	 */
	public LyJTableFooter(JTable table)
	{
		super();
		_tableComp = table;
		_columnModel = _tableComp.getColumnModel();
		_columnModel.addColumnModelListener(this);
		setLayout(this);
		setContent();
		setForeground(Color.black);
	}

	/**
	 * Fills the row content.
	 */
	public void setContent()
	{
		int			   i, n = _columnModel.getColumnCount();
		LyJTableColumn column = null;
		Object		   value = null;
		int			   align;

		for (i = 0; i < n; i++)
		{
			column = (LyJTableColumn)_columnModel.getColumn(i);
			value = column.getFooterValue();
			align = column.getFooterAlign();

			add(createFooterCell(value, align), i);
		}
	}

	/**
	 * Creates the cell containing the footer element of the specified object.
	 *
	 * @param  object the source data that contain the text
	 * @return the component to display in a cell
	 * @see    #createFooterCell(Object, int)
	 */
	protected Component createFooterCell(Object object)
	{
		return createFooterCell(object, SwingConstants.LEFT);
	}

	/**
	 * Creates the cell containing the footer element of the specified object, with the given
	 * alignment.
	 *
	 * @param  object the source data that containing the text
	 * @param  align  the alignment of the cell
	 * @return the component to display in a cell
	 */
	protected Component createFooterCell(Object object, int align)
	{
		String text;
		Border border = new LyJFooterCellBorder(3, 3, 3, 3);

		if (object instanceof String)
			text = " ";
		else if (object instanceof LySwingFooter)
			text = ((LySwingFooter)object).getLabel();
		else
			text = "ERROR";

		JLabel comp = new JLabel(text);

		comp.setHorizontalAlignment(align);
		comp.setOpaque(true);
		comp.setBorder(border);
		return comp;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#setFont(java.awt.Font)
	 */
	public void setFont(Font font)
	{
		super.setFont(font);
		int i, n = getComponentCount();

		for (i = 0; i < n; i++)
			getComponent(i).setFont(font);
		HEIGHT = getFontMetrics(font).getHeight() + 3;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#setBackground(java.awt.Color)
	 */
	public void setBackground(Color color)
	{
		super.setBackground(color);
		int i, n = getComponentCount();

		for (i = 0; i < n; i++)
			getComponent(i).setBackground(color);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.Component#setForeground(java.awt.Color)
	 */
	public void setForeground(Color color)
	{
		super.setForeground(color);
		int i, n = getComponentCount();

		for (i = 0; i < n; i++)
			getComponent(i).setForeground(color);
	}

	/**
	 * Sets the text to display in the footer at the specified index.
	 *
	 * @param index the index of the footer to update
	 * @param text  the text to display in the footer
	 */
	public void setFooter(int index, String text)
	{
		JLabel label = (JLabel)getComponent(index);

		if (label != null)
			label.setText(text);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.swing.event.TableColumnModelListener#columnAdded(javax.swing.event.TableColumnModelEvent)
	 */
	public void columnAdded(TableColumnModelEvent e)
	{
		add(createFooterCell(
				((LyJTableColumn)_columnModel.getColumn(e.getToIndex())).getFooterValue()),
			e.getToIndex());
		revalidate();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.swing.event.TableColumnModelListener#columnMarginChanged(javax.swing.event.ChangeEvent)
	 */
	public void columnMarginChanged(ChangeEvent e)
	{
		revalidate();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.swing.event.TableColumnModelListener#columnMoved(javax.swing.event.TableColumnModelEvent)
	 */
	public void columnMoved(TableColumnModelEvent e)
	{
		JLabel labelTo = (JLabel)getComponent(e.getToIndex());
		JLabel labelFrom = (JLabel)getComponent(e.getFromIndex());
		String tmp = labelTo.getText();

		labelTo.setText(labelFrom.getText());
		labelFrom.setText(tmp);
		revalidate();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.swing.event.TableColumnModelListener#columnRemoved(javax.swing.event.TableColumnModelEvent)
	 */
	public void columnRemoved(TableColumnModelEvent e)
	{
		remove(e.getFromIndex());
		revalidate();
	}

	/**
	 * Nothing is done.
	 *
	 * @see javax.swing.event.TableColumnModelListener#columnSelectionChanged(javax.swing.event.ListSelectionEvent)
	 */
	public void columnSelectionChanged(ListSelectionEvent e)
	{
	}

	/**
	 * Nothing is done.
	 *
	 * @see java.awt.LayoutManager#addLayoutComponent(java.lang.String, java.awt.Component)
	 */
	public void addLayoutComponent(String name, Component comp)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.LayoutManager#layoutContainer(java.awt.Container)
	 */
	public void layoutContainer(Container parent)
	{
		int n = parent.getComponentCount();
		int m = _columnModel.getColumnCount();

		if (n != m)
			// This component isn't ready yet so recall layoutContainer
			// Normally unreachable
			System.out.println("LyJTableFooter layoutContainer componentCount = " + n
				+ " , columnCount = " + m);
		// layoutContainer(parent);
		else
		{
			int x = 0;

			for (int i = 0; i < n; i++)
			{
				parent.getComponent(i).setBounds(x, 0, _columnModel.getColumn(i).getWidth(),
					HEIGHT);
				x += _columnModel.getColumn(i).getWidth();
			}
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.LayoutManager#minimumLayoutSize(java.awt.Container)
	 */
	public Dimension minimumLayoutSize(Container parent)
	{
		int i, n = _columnModel.getColumnCount();
		int height, width;

		if (n == 0)
			return new Dimension(0, 0);
		height = HEIGHT;
		width = 0;
		for (i = 0; i < n; i++)
			width += _columnModel.getColumn(i).getMinWidth();
		Insets insets = parent.getInsets();

		return new Dimension(width + insets.left + insets.right,
			height + insets.top
			+ insets.bottom);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.LayoutManager#preferredLayoutSize(java.awt.Container)
	 */
	public Dimension preferredLayoutSize(Container parent)
	{
		int i, n = _columnModel.getColumnCount();
		int height, width;

		if (n == 0)
			return new Dimension(0, 0);
		height = HEIGHT;
		width = 0;
		for (i = 0; i < n; i++)
			width += _columnModel.getColumn(i).getPreferredWidth();
		Insets insets = parent.getInsets();

		return new Dimension(width + insets.left + insets.right,
			height + insets.top
			+ insets.bottom);
	}

	/**
	 * Nothing is done.
	 *
	 * @see java.awt.LayoutManager#removeLayoutComponent(java.awt.Component)
	 */
	public void removeLayoutComponent(Component comp)
	{
	}

} // end class LyJTableFooter
