/*
 * LyJTableModel.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.table;

import java.awt.*;
import javax.swing.table.*;
import leon.view.win.swing.*;


/**
 * This class is the link between the leon data source and the table component. It is managing the
 * population of the table.
 *
 * @version $Id: LyJTableModel.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class LyJTableModel extends AbstractTableModel
{
	private static final long  serialVersionUID = -4227655738943387917L;

	public static final String __VERSION =
		"$Id: LyJTableModel.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The empty cell. */
	public static final String EMPTY_CELL_VALUE = "EMPTY_CELL_VALUE ";

	/** The empty status cell. */
	public static final String EMPTY_CELL_STATUS_VALUE = "EMPTY_CELL_STATUS_VALUE";

	/** The number of empty rows. */
	public static int		   EMPTY_ROW_COUNT = 50;

	/** The associated table. */
	protected LySwingTable     _swingTable;

	/**
	 * Creates a table model for the specified table.
	 *
	 * @param swingTable the associated swing table
	 */
	public LyJTableModel(LySwingTable swingTable)
	{
		super();
		_swingTable = swingTable;
	}

	/**
	 * Counts one more empty column for a better visual aspect.
	 *
	 * @see javax.swing.table.TableModel#getColumnCount()
	 */
	public int getColumnCount()
	{
		return _swingTable.getHeaders().size() + 1;
	}

	/**
	 * Counts some empty rows for a better visual aspect.
	 *
	 * @see javax.swing.table.TableModel#getRowCount()
	 */
	public int getRowCount()
	{
		//return getDataRowCount() + getEmptyRowCount();
		int dataRowCount = getDataRowCount(); 

		return Math.max(dataRowCount, EMPTY_ROW_COUNT); // idem but faster than above

	}
	/**
	 * Gets the value of the cell at the specified row and column indices. If the cell at (rowIndex,
	 * columnIndex) doesn't truly exists, returns a flag that will be used by the renderer.
	 *
	 * @see #setValueAt(Object, int, int)
	 * @see javax.swing.table.TableModel#getValueAt(int, int)
	 */
	/*public Object getValueAt(int rowIndex, int columnIndex)
	{
		Object value = getInternalValueAt(rowIndex, columnIndex);
		
		if ((value == EMPTY_CELL_VALUE) || (value == EMPTY_CELL_STATUS_VALUE))
			return "";
		else if (value instanceof LySwingCell)
		{
			return ((LySwingCell)value).getLabel();
		}
		return value;
	}*/
	
	/**
	 * Gets the value of the cell at the specified row and column indices. If the cell at (rowIndex,
	 * columnIndex) doesn't truly exists, returns a flag that will be used by the renderer.
	 *
	 * @see #setValueAt(Object, int, int)
	 * @see javax.swing.table.TableModel#getValueAt(int, int)
	 */
	public Object getValueAt(int rowIndex, int columnIndex)
	{
		try
		{
			// EMPTY COLUMN
			if (columnIndex >= getDataColumnCount())
				return EMPTY_CELL_VALUE;
			// TRUE DATA COLUMN
			else if (rowIndex < getDataRowCount())
			{ // TRUE DATA ROW

				// TRUE DATA STATUS CELL
				if (getColumnHeaderValue(columnIndex) instanceof String)
					return getRowImage(rowIndex);
				// TRUE DATA NORMAL CELL
				LySwingRow	swingRow = _swingTable.getRow(rowIndex);
				
				if (swingRow == null)
					return EMPTY_CELL_VALUE;
				
				return swingRow.getCell(columnIndex);
			}
			else
			// EMPTY ROW
			{
				// EMPTY STATUS CELL
				if (getColumnHeaderValue(columnIndex) instanceof String)
					return EMPTY_CELL_STATUS_VALUE;
				// EMPTY NORMAL CELL
				return EMPTY_CELL_VALUE;
			}
		}
		catch (RuntimeException e)
		{
			return EMPTY_CELL_VALUE;
		}
	} // end method getValueAt

	/**
	 * Gets the text label of the specified column.
	 *
	 * @see javax.swing.table.TableModel#getColumnName(int)
	 */
	public String getColumnName(int i)
	{
		if (i >= getDataColumnCount())
			return EMPTY_CELL_STATUS_VALUE;
		else if (!_swingTable.hasHeaders())
		{
			String buf = "_";

			buf += i;
			return buf;
		}
		else
			return _swingTable.getHeader(i).getLabel();
	}

	/**
	 * Gets the class of the datas in the specified column.
	 *
	 * @see javax.swing.table.TableModel#getColumnClass(int)
	 */
	public Class<?> getColumnClass(int i)
	{
		if (i >= getDataColumnCount())
			return String.class;
		else if (getColumnHeaderValue(i) instanceof String)
			return Image.class;
		else
		{
			Object content = getValueAt(0, i);

			if (content instanceof LySwingFormField)
				return LySwingFormField.class;
			if (content instanceof LySwingToolBar)
				return LySwingToolBar.class;
			return LySwingCell.class;
		}
	}

	/**
	 * Nothing is done.
	 *
	 * @see #getValueAt(int, int)
	 * @see javax.swing.table.TableModel#setValueAt(java.lang.Object, int, int)
	 */
	public void setValueAt(Object value, int rowIndex, int columnIndex)
	{
	}

	/**
	 * Gets the header label of the specified column.
	 *
	 * @param  i the index of the column
	 * @return the object used to display the text of the column header at position i
	 */
	public Object getColumnHeaderValue(int i)
	{
		if ((i >= getDataColumnCount()) || !_swingTable.hasHeaders())
			return getColumnName(i);
		return _swingTable.getHeader(i);
	}

	/**
	 * Gets the graphical row at the specified index.
	 *
	 * @param  idx the index of the row to retrieve
	 * @return the data source of the row at position i
	 */
	public LySwingRow getRowAt(int idx)
	{
		return _swingTable.getRow(idx);
	}

	/**
	 * Gets the default number of empty rows.
	 *
	 * @return the number of empty rows
	 */
	public int getEmptyRowCount()
	{
		int dataRowCount = getDataRowCount();
		
		return Math.max(0, EMPTY_ROW_COUNT - dataRowCount);
	}

	/**
	 * Gets the value object of the specified footer column.
	 *
	 * @param  i the index of the column
	 * @return the object used to display the text of the column footer at position i
	 */
	public Object getColumnFooterValue(int i)
	{
		if (i >= getDataColumnCount())
			return getColumnName(i);
		return _swingTable.getFooter(i);
	}

	/**
	 * Gets the number of rows with values.
	 *
	 * @return the number of not empty rows
	 */
	public int getDataRowCount()
	{
		return _swingTable.getDataRowCount();
	}

	/**
	 * Gets the number of columns with value.
	 *
	 * @return the number of not empty column
	 */
	public int getDataColumnCount()
	{
		return _swingTable.getHeaders().size();
	}

	/**
	 * Gets the image of the specified row.
	 *
	 * @param  index the index of the row
	 * @return the image of the specified row
	 */
	public Image getRowImage(int index)
	{
		if (index >= getDataRowCount())
			return null;
		return _swingTable.getRow(index).getImage();
	}

	/**
	 * Gets the tooltip of the specified row.
	 *
	 * @param  index the index of the row
	 * @return the tooltip to display
	 */
	public String getRowToolTip(int index)
	{
		if (index >= getDataRowCount())
			return null;
		return _swingTable.getRow(index).getToolTip();
	}

	/**
	 * Gets the identifier of the specified row.
	 *
	 * @param  index the index of the row
	 * @return the id of this row
	 */
	public String getRowId(int index)
	{
		if (index >= getDataRowCount())
			return null;
		return _swingTable.getRow(index).getId();
	}

	/**
	 * Gets the editable state of the cell at the specified position.
	 *
	 * @param  row    the row of the cell
	 * @param  column the column of the cell
	 * @return <code>true</code> if the specified cell is an editable field component, <code>
	 *         false</code> otherwise
	 * @see    javax.swing.table.TableModel#isCellEditable(int, int)
	 */
	public boolean isCellEditable(int row, int column)
	{
		return (getValueAt(row, column) instanceof LySwingFormField
			|| getValueAt(row, column) instanceof LySwingToolBar);
	}

	/**
	 * Get a lock on the table to prevent concurrent access on rows
	 * return a valid lock for synchronized()
	 */
	public Object getLock()
	{
		return _swingTable.getLock();
	}
} // end class LyJTableModel
