/*
 * LyJTableMouseListener.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.table;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.table.*;

import leon.view.*;
import leon.view.event.*;
import leon.view.win.swing.*;
import leon.view.win.swing.component.*;


/**
 * This class manages the mouse events that occur on the table.
 */
public class LyJTableMouseListener extends MouseAdapter
{
	public static final String __VERSION =
		"$Id: LyJTableMouseListener.java 33376 2011-12-01 15:57:34Z nhamon $";

	/** The table component. */
	protected LySwingTable     _swingTable;

	/**
	 * Creates a new table mouse listener for the specified table.
	 *
	 * @param swingTable the associated table
	 */
	public LyJTableMouseListener(LySwingTable swingTable)
	{
		_swingTable = swingTable;
	}

	/**
	 * Indicates that the column at the specified index is sortable.
	 *
	 * @param  idx the index of the column
	 * @return <code>true</code> if the column at index idx is sortable, else <code>false</code>
	 */
	protected boolean isSortable(int idx)
	{
		if (idx >= (_swingTable.getTableModel()).getDataColumnCount())
			return false;

		LySwingHeader header = _swingTable.getHeader(idx);

		return ((header != null) && (header.isSortable()));
	}

	/**
	 * Return the identifiers of the selected rows of the table component.
	 *
	 * @return the ids of the selected rows
	 */
	public Vector<String> getSelectedIds()
	{
		Vector<String>	   selectedIds = new Vector<String>();
		ListSelectionModel selection = _swingTable.getTableComponent().getSelectionModel();
		int				   i, n = (_swingTable.getTableModel()).getDataRowCount();

		for (i = 0; i < n; i++)
		{
			if (selection.isSelectedIndex(i))
				selectedIds.add(_swingTable.getRow(i).getId());
		}

		return selectedIds;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
	 */
	
	public void mouseClicked(MouseEvent e)
	{
		// Event occurs on header
		if (e.getSource() instanceof JTableHeader)
		{
			if ((e.getClickCount() == 2)
					&& ((e.getModifiers() & InputEvent.BUTTON1_MASK) > 0))
			{
				// Adjust column width
				int 		colIndex = _swingTable.getTableComponent().columnAtPoint(new Point(e.getX(), 0));
				Rectangle	rect = _swingTable.getTableComponent().getTableHeader().getHeaderRect(colIndex);
				boolean		isOnLimit = false;
				
				if ((e.getX() >= (rect.x + rect.width - 3)) && (e.getX() <= (rect.x + rect.width)))
					isOnLimit = true;
				else if ((e.getX() >= rect.x) && (e.getX() <= (rect.x + 3)))
				{
					isOnLimit = true;
					colIndex--;
				}
				
				if (isOnLimit && (colIndex >= 0))
				{
					TableColumnModel columnModel = _swingTable.getTableComponent().getColumnModel();
					LyJTableColumn   column;
					int				 preferredWidth;

					column = (LyJTableColumn)columnModel.getColumn(colIndex);

					// Compute and set the preferred width of this column
					preferredWidth = _swingTable.computeColumnPreferredWidth(_swingTable.getTableComponent(), colIndex,
							_swingTable.getTableComponent().getFont(), -1, false);
					column.setPreferredWidth(preferredWidth);
				}
				
				return;
			}
			
			if (((e.getModifiers() & InputEvent.BUTTON1_MASK) > 0)
					|| ((e.getModifiers() & InputEvent.BUTTON2_MASK) > 0))
			{
				if (_swingTable.getTableComponent().getActionTransmitter() != null)
				{
					int 		column = _swingTable.getTableComponent().columnAtPoint(new Point(e.getX(), 0));
					int 		idx = _swingTable.getTableComponent().convertColumnIndexToModel(column);
					Rectangle	rect = _swingTable.getTableComponent().getTableHeader().getHeaderRect(column);

					if (((e.getX() >= (rect.x + rect.width - 3)) && (e.getX() <= (rect.x + rect.width)))
							|| ((e.getX() >= rect.x) && (e.getX() <= (rect.x + 3))))
						return;
					
					if (!isSortable(idx))
						return;

					LyActionEvent actionEvent;

					if (!e.isControlDown())
						actionEvent = new LyActionEvent(this, null, null, LyActionDemand.TYPE_SORT);
					else
						actionEvent = new LyActionEvent(this, null, null,
								LyActionDemand.TYPE_EXTEND_SORT);

					actionEvent.addIndex(idx);
					_swingTable.getTableComponent().actionPerformed(actionEvent);

					if (_swingTable.getTableComponent().getTableHeader() != null)
					    _swingTable.getTableComponent().getTableHeader().repaint();

					if (_swingTable.hasStatusColumn())
						_swingTable.getStatusHeader().updateHeader();
				}
			}
			else if ((e.getModifiers() & InputEvent.BUTTON3_MASK) > 0)
			{
				// Right click on header
				if (_swingTable.getTableComponent().getActionTransmitter() != null)
				{
					JTableHeader  header = (JTableHeader)e.getSource();
					Point		  relativePosition = header.getLocation();
					int			  column = _swingTable.getTableComponent().columnAtPoint(new Point(
								e.getX(), 0));
					int			  idx = _swingTable.getTableComponent().convertColumnIndexToModel(
							column);
					Point		  p = e.getPoint();
					short		  type = (e.isControlDown() ? LyMenuDemand.TYPE_EXTEND_POPUP
															: LyMenuDemand.TYPE_POPUP);
					LyActionEvent actionEvent = new LyActionEvent(this, "_headers", null, type);

					actionEvent.addIndex(idx);
					actionEvent.setParameter(
						new int[] { p.x + relativePosition.x, p.y + relativePosition.y });
					_swingTable.getTableComponent().actionPerformed(actionEvent);
				}
			}
		}
		// Event occurs on table
		else if (e.getSource() instanceof LyJTableComponent)
		{
			Point p = e.getPoint();
			int   rowIndex = _swingTable.getTableComponent().rowAtPoint(p);

			if (rowIndex >= _swingTable.getTableModel().getDataRowCount())
			{
				if (((e.getModifiers() & InputEvent.BUTTON3_MASK) <= 0)
						&& (e.getClickCount() < 2)) // not a right click nor a double-click
					return;
					// FFT1140 right clic is allowed on empty lines
			}
			
			LySwingRow	swingRow = _swingTable.getRow(rowIndex);
			LyRow		row = (swingRow == null ? null : (LyRow)swingRow.getViewComponent());
			
			int column = _swingTable.getTableComponent().columnAtPoint(new Point(e.getX(), 0));
			int columnIndex = _swingTable.getTableComponent().convertColumnIndexToModel(column);

			// Left simple click
			if ((e.getClickCount() == 1) && ((e.getModifiers() & InputEvent.BUTTON1_MASK) > 0))
			{
				Hashtable<Integer, Integer> cursors = _swingTable.getRowCursors();
				int		  firstColumn = (_swingTable.hasStatusColumn() ? 1 : 0);
				
				if (!e.isControlDown() && !e.isShiftDown())
				{
					ListSelectionModel	selectionModel = _swingTable.getTableComponent().getSelectionModel();

					if (selectionModel != null)
					{
						selectionModel.setAnchorSelectionIndex(rowIndex);
						selectionModel.setLeadSelectionIndex(rowIndex);
					}
				}
				
				// Check if the click occured on a cursor to open / close a row
				if ((cursors != null) && (cursors.get(new Integer(rowIndex)) != null)
						&& (columnIndex == firstColumn))
				{
					Integer offsetValue = (Integer)cursors.get(new Integer(rowIndex));
					int     offset = offsetValue.intValue();

					if ((row != null)
							&& (p.x >= offset)
							&& (p.x <= (offset + LyJCellRenderer.HIERARCHY_OFFSET_WIDTH)))
					{
						row.setOpen(!row.isOpen());
						_swingTable.resetRowCursors();

						String		  item = swingRow.getId();
						LyActionEvent actionEvent = new LyActionEvent(this, item, row.isOpen() + "",
								LyActionDemand.TYPE_OPEN);

						actionEvent.addIndex(rowIndex);
						_swingTable.getTableComponent().actionPerformed(actionEvent);
					}
				}

				if ((row != null) && !row.isSensitive())
					return;

				Vector<String> selectedIds = getSelectedIds();

				if (_swingTable.getTableComponent().getSelectionTransmitter() != null)
				{
					LySelectionEvent select = new LySelectionEvent(this, selectedIds, true);

					select.addIndex(_swingTable.getTableComponent().rowAtPoint(e.getPoint()));
					select.setSelectionSize(selectedIds.size());
					select.setSubIndex(columnIndex);
					select.setPosition(new int[] {p.x, p.y});

					if (e.isControlDown() || e.isShiftDown())
						select.setExtend(true);

					_swingTable.getTableComponent().selectionPerformed(select);
				}
			}
			// Left double click
			else if ((e.getClickCount() == 2) && ((e.getModifiers() & InputEvent.BUTTON1_MASK) > 0))
			{
				if ((row != null) && !row.isSensitive())
					return;

				if (_swingTable.getTableComponent().getActionTransmitter() != null)
				{
					if (!_swingTable.isStatusColumn(columnIndex))
					{
						String			item = null;
						LyActionEvent	actionEvent = new LyActionEvent(this);

						if (rowIndex < (_swingTable.getTableModel()).getDataRowCount())
							item = _swingTable.getRow(rowIndex).getId();
						
						actionEvent.setValue(item);
						actionEvent.addIndex(rowIndex);
						actionEvent.setSubIndex(columnIndex);
						_swingTable.getTableComponent().actionPerformed(actionEvent);
					}
				}
			}
			// Right click
			else if ((e.getModifiers() & InputEvent.BUTTON3_MASK) > 0)
			{
				if ((row != null) && !row.isSensitive())
					return;

				if (_swingTable.getTableComponent().getActionTransmitter() != null)
				{
					String item = null;

					if (rowIndex >= (_swingTable.getTableModel()).getDataRowCount())
					{
						// FFT1140: SWING - Popup menu
						// POPUP outside the rows
						item = null;
						rowIndex = -1;
					}
					else if (rowIndex >= 0)
					{
						item = _swingTable.getRow(rowIndex).getId();

						if (!_swingTable.getTableComponent().isRowSelected(rowIndex))
							_swingTable.getTableComponent().setRowSelectionInterval(rowIndex,
								rowIndex);
					}

					short		  		type = (e.isControlDown() ? LyMenuDemand.TYPE_EXTEND_POPUP
							: LyMenuDemand.TYPE_POPUP);
					LyActionEvent		actionEvent = new LyActionEvent(this, null, item, type);
					LyJTableComponent 	table = (LyJTableComponent)e.getSource();
					Point				relativePosition = new Point(0, 0);
					
					if (e.isControlDown())
						relativePosition = table.getLocation();

					actionEvent.addIndex(rowIndex);
					actionEvent.setParameter(new int[] { p.x + relativePosition.x, p.y + relativePosition.y});
					actionEvent.setSubIndex(columnIndex);
					_swingTable.getTableComponent().actionPerformed(actionEvent);
				}
			}
		}
		// Event occurs on status header
		else if (e.getSource() instanceof LyJTableStatusHeader)
		{
			if (((e.getModifiers() & InputEvent.BUTTON1_MASK) > 0)
					|| ((e.getModifiers() & InputEvent.BUTTON2_MASK) > 0))
			{
				if (_swingTable.getTableComponent().getActionTransmitter() != null)
				{
					// if status not sortable : discard event
					if (!isSortable(0))
						return;

					LyActionEvent actionEvent;

					if (!e.isControlDown())
						actionEvent = new LyActionEvent(_swingTable.getTableComponent(), null, null,
								LyActionDemand.TYPE_SORT);
					else
						actionEvent = new LyActionEvent(_swingTable.getTableComponent(), null, null,
								LyActionDemand.TYPE_EXTEND_SORT);

					actionEvent.addIndex(0);
					_swingTable.getTableComponent().actionPerformed(actionEvent);
					_swingTable.getTableComponent().getTableHeader().repaint();
					_swingTable.getStatusHeader().updateHeader();
				}
			}
		}
		// Click occurs on empty zone when LyJTableModel.EMPTY_ROW_COUNT is lower than number of rows
		else if (e.getSource() instanceof JViewport)
		{
			if (_swingTable.getTableComponent().getSelectionTransmitter() != null)
			{
				Vector<Object>		selectedIds = new Vector<Object>();
				LySelectionEvent	select = new LySelectionEvent(this, selectedIds, true);

				select.addIndex(_swingTable.getTableComponent().rowAtPoint(e.getPoint()));
				select.setSelectionSize(selectedIds.size());
				_swingTable.getTableComponent().selectionPerformed(select);
			}
		}
	} // end method mouseClicked
} // end class LyJTableMouseListener
