/*
 * LyJTableScrollPane.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.table;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;


/**
 * This scrollPane can manage a column footer component that acts as the column header but at the
 * bottom of the main scrolling component.
 *
 * @version $Id: LyJTableScrollPane.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class LyJTableScrollPane extends JScrollPane
	implements ScrollPaneConstants,
		AdjustmentListener
{
	private static final long  serialVersionUID = -494717647164715834L;

	public static final String __VERSION =
		"$Id: LyJTableScrollPane.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The identifier of the column footer. */
	public static final String COLUMN_FOOTER = "COLUMN_FOOTER";

	/** The viewport of the footer component. */
	JViewport				   columnFooter;
	
	/** Layout used to display table. */
	LyJTableScrollPaneLayout   _layout;

	/**
	 * Creates a <code>JScrollPane</code> that displays this table component in a viewport whose
	 * view position can be controlled with a pair of scrollbars.
	 *
	 * <p>The scrollbar policies specifies when the scrollbars are displayed. For example, if <code>
	 * vsbPolicy</code> is <code>VERTICAL_SCROLLBAR_AS_NEEDED</code> then the vertical scrollbar
	 * only appears if the view doesn't fit vertically. The available policy settings are listed at
	 * {@link JScrollPane#setVerticalScrollBarPolicy(int)} and {@link
	 * JScrollPane#setHorizontalScrollBarPolicy(int)}.</p>
	 *
	 * @see   JScrollPane#setViewportView(java.awt.Component)
	 * @param view      the component to display in the scrollpanes viewport
	 * @param vsbPolicy an integer that specifies the vertical scrollbar policy
	 * @param hsbPolicy an integer that specifies the horizontal scrollbar policy
	 */
	public LyJTableScrollPane(Component view, int vsbPolicy, int hsbPolicy)
	{
		_layout = new LyJTableScrollPaneLayout();
		setLayout(_layout);
		setVerticalScrollBarPolicy(vsbPolicy);
		setHorizontalScrollBarPolicy(hsbPolicy);
		setViewport(createViewport());
		setVerticalScrollBar(createVerticalScrollBar());
		setHorizontalScrollBar(createHorizontalScrollBar());
		if (view != null)
			setViewportView(view);
		setOpaque(true);
		// updateUI();
		getHorizontalScrollBar().addAdjustmentListener(this);
	}

	/**
	 * Creates a <code>JScrollPane</code> that displays the contents of the specified component,
	 * where both horizontal and vertical scrollbars appear whenever the component's contents are
	 * larger than the view.
	 *
	 * @see   JScrollPane#setViewportView(java.awt.Component)
	 * @param view the component to display in the scrollpane's viewport
	 */
	public LyJTableScrollPane(Component view)
	{
		this(view, VERTICAL_SCROLLBAR_AS_NEEDED, HORIZONTAL_SCROLLBAR_AS_NEEDED);
	}

	/**
	 * Creates an empty (no viewport view) <code>JScrollPane</code> with specified scrollbar
	 * policies. The available policy settings are listed at JScrollPane#setVerticalScrollBarPolicy
	 * and JScrollPane#setHorizontalScrollBarPolicy(int).
	 *
	 * @see   JScrollPane#setViewportView(java.awt.Component)
	 * @param vsbPolicy an integer that specifies the vertical scrollbar policy
	 * @param hsbPolicy an integer that specifies the horizontal scrollbar policy
	 */
	public LyJTableScrollPane(int vsbPolicy, int hsbPolicy)
	{
		this(null, vsbPolicy, hsbPolicy);
	}

	/**
	 * Creates an empty (no viewport view) <code>JScrollPane</code> where both horizontal and
	 * vertical scrollbars appear when needed.
	 */
	public LyJTableScrollPane()
	{
		this(null, VERTICAL_SCROLLBAR_AS_NEEDED, HORIZONTAL_SCROLLBAR_AS_NEEDED);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.swing.JScrollPane#getViewportBorderBounds()
	 */
	public Rectangle getViewportBorderBounds()
	{
		Rectangle borderR = super.getViewportBorderBounds();

		return borderR;
	}

	/**
	 * Returns the column footer.
	 *
	 * @return the <code>columnFooter</code> property
	 * @see    #setColumnFooter(JViewport)
	 */
	public JViewport getColumnFooter()
	{
		return columnFooter;
	}

	/**
	 * Removes the old columnFooter, if it exists. If the new columnFooter isn't <code>null</code>,
	 * synchronizes the x coordinate of its viewPosition with the viewport (if there is one) and
	 * then adds it to the scrollpane.
	 *
	 * <p>Most applications will find it more convenient to use {@link
	 * #setColumnFooterView(Component)} to add a column footer component and its viewport to the
	 * scrollpane.</p>
	 *
	 * @see   #getColumnFooter()
	 * @see   #setColumnFooterView(Component)
	 * @param columnFooter the new viewport to use
	 * @beaninfo
	 *   bound: true description: The column footer child for this scrollpane attribute:
	 *          visualUpdate true
	 */
	public void setColumnFooter(JViewport columnFooter)
	{
		JViewport old = getColumnFooter();

		this.columnFooter = columnFooter;
		if (columnFooter != null)
			add(columnFooter, COLUMN_FOOTER);
		else if (old != null)
			remove(old);
		firePropertyChange("columnFooter", old, columnFooter);

		revalidate();
		repaint();
	}

	/**
	 * Creates a column-footer viewport if necessary, sets its view, and then adds the column-footer
	 * viewport to the scrollpane. For example:
	 *
	 * <pre>
	 * JScrollPane scrollpane = new JScrollPane();
	 * scrollpane.setViewportView(myBigComponentToScroll);
	 * scrollpane.setColumnFooterView(myBigComponentsColumnHeader);
	 * </pre>
	 *
	 * @see   #setColumnFooter(JViewport)
	 * @see   JViewport#setView(java.awt.Component)
	 * @param view the component to display as the column footer
	 */
	public void setColumnFooterView(Component view)
	{
		if (getColumnFooter() == null)
			setColumnFooter(createViewport());
		getColumnFooter().setView(view);
	}

	/**
	 * Scrolls the footer viewport.
	 *
	 * @see java.awt.event.AdjustmentListener#adjustmentValueChanged(java.awt.event.AdjustmentEvent)
	 */
	public void adjustmentValueChanged(AdjustmentEvent e)
	{
		if (columnFooter != null)
		{
			Point p = columnFooter.getViewPosition();

			p.x = viewport.getViewPosition().x;
			columnFooter.setViewPosition(p);
		}
	}
	
	/**
	 * Set the screen dimension used to compute preferred size of the table.
	 * 
	 * @param screenSize the screen dimension.
	 */
	public void setScreenSize(int[] screenSize)
	{
		_layout.setScreenSize(screenSize);
	}
	
	/**
	 * Overriden to use table locks when painting
	 */
    protected void paintChildren(Graphics g) 
    {
    /*	Object lock = null;
    	JTable table = (JTable)getViewport().getView();
		if (table != null)
		{
			LyJTableModel     tableModel = (LyJTableModel)table.getModel();
			
			if (tableModel != null)
				lock = tableModel.getLock();
		}
		if (lock == null)
			lock = getTreeLock();
    	synchronized (lock)
    	{*/
    		super.paintChildren(g);
    	/*}*/
    }
} // end class LyJTableScrollPane
