/*
 * LyJTableScrollPaneLayout.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.table;

import java.awt.*;
import javax.swing.*;
import javax.swing.border.*;

import leon.view.win.swing.*;


/**
 * This class lays out a LyJTableScrollPane. It manages a column footer component.
 *
 * @version $Id: LyJTableScrollPaneLayout.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class LyJTableScrollPaneLayout extends ScrollPaneLayout
{
	private static final long  serialVersionUID = -8714090527583789108L;

	public static final String __VERSION =
		"$Id: LyJTableScrollPaneLayout.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The id of the column footer. */
	public static final String COLUMN_FOOTER = "COLUMN_FOOTER";

	/**
	 * Maximum height of the table preferred size.
	 * @deprecated no longer used.
	 */
	public static int		   MAX_PREFERRED_HEIGHT = 350;

	/** The footer viewport. */
	JViewport				   colFoot;

	/**
	 * Within a dialog window, do not change the preferred size workaround for a bug : when the
	 * internal table is resized, the view size is fixed and lyout was using the minimum size.
	 */
	protected Dimension		   _preferredSize;

	/** Screen dimension used to compute preferred size. */
	protected int[]			   _screenSize;

	/**
	 * This method must be called after setting a JScrollPanes layout manager. It initializes all of
	 * the internal fields that are ordinarily set by addLayoutComponent().<br>
	 * For example:
	 *
	 * <pre>
	 *      ScrollPaneLayout mySPLayout = new ScrollPanelLayout() {
	 *          public void layoutContainer(Container p) {
	 *              super.layoutContainer(p);
	 *              // do some extra work here ...
	 *          }
	 *      };
	 *      scrollpane.setLayout(mySPLayout):
	 *      mySPLayout.syncWithScrollPane(scrollpane);
	 * </pre>
	 *
	 * <p>This method is overriden to set the column footer to the given JScrollPane column footer
	 * if it is a {@link LyJTableScrollPane}.</p>
	 */
	public void syncWithScrollPane(JScrollPane sp)
	{
		super.syncWithScrollPane(sp);
		if (sp instanceof LyJTableScrollPane)
			colFoot = ((LyJTableScrollPane)sp).getColumnFooter();
	}

	/**
	 * Adds the specified component to the layout. The layout is identified using one of:
	 *
	 * <ul>
	 *   <li>{@link javax.swing.ScrollPaneConstants#VIEWPORT}</li>
	 *   <li>{@link javax.swing.ScrollPaneConstants#VERTICAL_SCROLLBAR}</li>
	 *   <li>{@link javax.swing.ScrollPaneConstants#HORIZONTAL_SCROLLBAR}</li>
	 *   <li>{@link javax.swing.ScrollPaneConstants#ROW_HEADER}</li>
	 *   <li>{@link javax.swing.ScrollPaneConstants#COLUMN_HEADER}</li>
	 *   <li>{@link javax.swing.ScrollPaneConstants#LOWER_LEFT_CORNER}</li>
	 *   <li>{@link javax.swing.ScrollPaneConstants#LOWER_RIGHT_CORNER}</li>
	 *   <li>{@link javax.swing.ScrollPaneConstants#UPPER_LEFT_CORNER}</li>
	 *   <li>{@link javax.swing.ScrollPaneConstants#UPPER_RIGHT_CORNER}</li>
	 *   <li>{@link #COLUMN_FOOTER}</li>
	 * </ul>
	 *
	 * Overriden to add COLUMN_FOOTER treatment.
	 *
	 * @param s the component identifier
	 * @param c the the component to be added
	 */
	public void addLayoutComponent(String s, Component c)
	{
		if (s.equals(COLUMN_FOOTER))
			colFoot = (JViewport)addSingletonComponent(colFoot, c);
		else
			super.addLayoutComponent(s, c);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.LayoutManager#removeLayoutComponent(java.awt.Component)
	 */
	public void removeLayoutComponent(Component c)
	{
		if (c == colFoot)
			colFoot = null;
		else
			super.removeLayoutComponent(c);
	}

	/**
	 * Gets the JViewport object that is the column footer.
	 *
	 * @return the JViewport object that is the column footer
	 */
	public JViewport getColumnFooter()
	{
		return colFoot;
	}
	
	/**
	 * Set the screen dimension used to compute preferred size of the table.
	 * 
	 * @param screenSize the screen dimension.
	 */
	public void setScreenSize(int[] screenSize)
	{
		_screenSize = screenSize;
	}

	/**
	 * The preferred size of a ScrollPane is the size of the insets, plus the preferred size of the
	 * viewport, plus the preferred size of the visible headers, plus the preferred size of the
	 * scrollbars that will appear given the current view and the current scrollbar displayPolicies.
	 * <br>
	 * Overriden to manage our own prefered size.
	 *
	 * @param  parent the Container that will be laid out
	 * @return a Dimension object specifying the preferred size of the viewport and any scrollbars.
	 * @see    ViewportLayout
	 * @see    LayoutManager
	 */
	public Dimension preferredLayoutSize(Container parent)
	{
		if (_preferredSize != null)
			return _preferredSize;

		Dimension preferredSize = super.preferredLayoutSize(parent);
		int		  prefWidth = preferredSize.width;
		int		  prefHeight = preferredSize.height;

		if ((colFoot != null) && colFoot.isVisible())
			prefHeight += colFoot.getPreferredSize().height;

		preferredSize = new Dimension(prefWidth, prefHeight);
		
		boolean isDialog = false;

		while (parent != null)
		{
			parent = parent.getParent();
			
			if (parent instanceof JDialog)
				isDialog = true;
		}
		
		if ((_screenSize != null)
				&& (_screenSize.length >= 2))
		{
			int	ratio = (isDialog ? LySwingDialog.MAX_SIZE_RATIO : LySwingFrame.MAX_SIZE_RATIO);

			if (preferredSize.height > _screenSize[1] * ratio)
				preferredSize.height = _screenSize[1] * ratio;
		}

		if ((prefWidth > 0) && (prefHeight > 0))
			_preferredSize = preferredSize;
		
		return preferredSize;
	} // end method preferredLayoutSize

	/**
	 * Lays out the scrollpane. The positioning of components depends on the following constraints:
	 *
	 * <ul>
	 *   <li>The row header, if present and visible, gets its preferred height and the viewports
	 *     width.</li>
	 *   <li>The column header, if present and visible, gets its preferred width and the viewports
	 *     height.</li>
	 *   <li>If a vertical scrollbar is needed, i.e. if the viewports extent height is smaller than
	 *     its view height or if the displayPolicy is ALWAYS, it's treated like the row header wrt
	 *     it's dimensions and it's made visible.</li>
	 *   <li>If a horizontal scrollbar is needed it's treated like the column header (and see the
	 *     vertical scrollbar item).</li>
	 *   <li>If the scrollpane has a non-null viewportBorder, then space is allocated for that.</li>
	 *   <li>The viewport gets the space available after accounting for the previous
	 *     constraints.</li>
	 *   <li>The corner components, if provided, are aligned with the ends of the scrollbars and
	 *     headers. If there's a vertical scrollbar the right corners appear, if there's a
	 *     horizontal scrollbar the lower corners appear, a row header gets left corners and a
	 *     column header gets upper corners.</li>
	 * </ul>
	 *
	 * @param parent the Container to lay out
	 */
	public void layoutContainer(Container parent)
	{
		/*
		 * Sync the (now obsolete) policy fields with the JScrollPane.
		 */
		JScrollPane scrollPane = (JScrollPane)parent;

		synchronized(parent.getTreeLock())
		{
		vsbPolicy = scrollPane.getVerticalScrollBarPolicy();
		hsbPolicy = scrollPane.getHorizontalScrollBarPolicy();

		if (scrollPane instanceof LyJTableScrollPane)
		{
			Rectangle availR = scrollPane.getBounds();

			availR.x = availR.y = 0;

			Insets insets = parent.getInsets();

			availR.x = insets.left;
			availR.y = insets.top;
			availR.width -= insets.left + insets.right;
			availR.height -= insets.top + insets.bottom;

			/*
			 * Get the scrollPane's orientation.
			 */
			boolean   leftToRight = scrollPane.getComponentOrientation().isLeftToRight();

			/*
			 * If there's a visible column header remove the space it needs from the top of availR.
			 * The column header is treated as if it were fixed height, arbitrary width.
			 */

			Rectangle colHeadR = new Rectangle(0, availR.y, 0, 0);

			if ((colHead != null) && (colHead.isVisible()))
			{
				int colHeadHeight = colHead.getPreferredSize().height;

				colHeadR.height = colHeadHeight;
				availR.y += colHeadHeight;
				availR.height -= colHeadHeight;
			}

			/*
			 * If there's a visible column footer remove the space it needs from the top of availR.
			 * The column footer is treated as if it were fixed height, arbitrary width.
			 */

			Rectangle colFootR = new Rectangle(0, 0, 0, 0);

			if ((colFoot != null) && (colFoot.isVisible()))
			{
				int colFootHeight = colFoot.getPreferredSize().height;

				colFootR.height = colFootHeight;
				availR.height -= colFootHeight;
			}

			/*
			 * If there's a visible row header remove the space it needs from the left or right of
			 * availR. The row header is treated as if it were fixed width, arbitrary height.
			 */

			Rectangle rowHeadR = new Rectangle(0, 0, 0, 0);

			if ((rowHead != null) && (rowHead.isVisible()))
			{
				int rowHeadWidth = rowHead.getPreferredSize().width;

				rowHeadR.width = rowHeadWidth;
				availR.width -= rowHeadWidth;
				if (leftToRight)
				{
					rowHeadR.x = availR.x;
					availR.x += rowHeadWidth;
				}
				else
					rowHeadR.x = availR.x + availR.width;
			}

			/*
			 * If there's a JScrollPane.viewportBorder, remove the space it occupies for availR.
			 */

			Border viewportBorder = scrollPane.getViewportBorder();
			Insets vpbInsets;

			if (viewportBorder != null)
			{
				vpbInsets = viewportBorder.getBorderInsets(parent);
				availR.x += vpbInsets.left;
				availR.y += vpbInsets.top;
				availR.width -= vpbInsets.left + vpbInsets.right;
				availR.height -= vpbInsets.top + vpbInsets.bottom;
			}
			else
				vpbInsets = new Insets(0, 0, 0, 0);

			/*
			 * At this point availR is the space available for the viewport and scrollbars. rowHeadR
			 * is correct except for its height and y and colHeadR is correct except for its width
			 * and x. Once we're through computing the dimensions of these three parts we can go
			 * back and set the dimensions of rowHeadR.height, rowHeadR.y, colHeadR.width,
			 * colHeadR.x and the bounds for the corners. We'll decide about putting up scrollbars
			 * by comparing the viewport views preferred size with the viewports extent size
			 * (generally just its size). Using the preferredSize is reasonable because layout
			 * proceeds top down - so we expect the viewport to be layed out next. And we assume
			 * that the viewports layout manager will give the view it's preferred size. One
			 * exception to this is when the view implements Scrollable and
			 * Scrollable.getViewTracksViewport{Width,Height} methods return true. If the view is
			 * tracking the viewports width we don't bother with a horizontal scrollbar, similar if
			 * view.getViewTracksViewport(Height) is true we don't bother with a vertical scrollbar.
			 */

			Component  view = (viewport != null) ? viewport.getView() : null;

			Dimension  viewPrefSize = getCustomizedScrollablePreferredSize(view);

			Dimension  extentSize = (viewport != null)
				? viewport.toViewCoordinates(availR.getSize()) : new Dimension(0, 0);

			boolean    viewTracksViewportWidth = false;
			boolean    viewTracksViewportHeight = false;
			Scrollable sv;

			if (view instanceof Scrollable)
			{
				sv = (Scrollable)view;
				viewTracksViewportWidth = sv.getScrollableTracksViewportWidth();
				viewTracksViewportHeight = sv.getScrollableTracksViewportHeight();
			}
			else
				sv = null;

			/*
			 * If there's a vertical scrollbar and we need one, allocate space for it (we'll make it
			 * visible later). A vertical scrollbar is considered to be fixed width, arbitrary
			 * height.
			 */

			Rectangle vsbR = new Rectangle(0, availR.y - vpbInsets.top, 0, 0);

			boolean   vsbNeeded;

			if (vsbPolicy == VERTICAL_SCROLLBAR_ALWAYS)
				vsbNeeded = true;
			else if (vsbPolicy == VERTICAL_SCROLLBAR_NEVER)
				vsbNeeded = false;
			else
			{ // vsbPolicy == VERTICAL_SCROLLBAR_AS_NEEDED
				vsbNeeded = !viewTracksViewportHeight && (viewPrefSize.height > extentSize.height);
			}

			if ((vsb != null) && vsbNeeded)
			{
				adjustForVSB(true, availR, vsbR, vpbInsets, leftToRight);
				extentSize = viewport.toViewCoordinates(availR.getSize());
			}

			/*
			 * If there's a horizontal scrollbar and we need one, allocate space for it (we'll make
			 * it visible later). A horizontal scrollbar is considered to be fixed height, arbitrary
			 * width.
			 */

			Rectangle hsbR = new Rectangle(availR.x - vpbInsets.left, 0, 0, 0);
			boolean   hsbNeeded;

			if (hsbPolicy == HORIZONTAL_SCROLLBAR_ALWAYS)
				hsbNeeded = true;
			else if (hsbPolicy == HORIZONTAL_SCROLLBAR_NEVER)
				hsbNeeded = false;
			else
			{ // hsbPolicy == HORIZONTAL_SCROLLBAR_AS_NEEDED
				hsbNeeded = !viewTracksViewportWidth && (viewPrefSize.width > extentSize.width);
			}

			if ((hsb != null) && hsbNeeded)
			{
				adjustForHSB(true, availR, hsbR, vpbInsets, colFootR);

				/*
				 * If we added the horizontal scrollbar then we've implicitly reduced the vertical
				 * space available to the viewport. As a consequence we may have to add the vertical
				 * scrollbar, if that hasn't been done so already. Ofcourse we don't bother with any
				 * of this if the vsbPolicy is NEVER.
				 */
				if ((vsb != null) && !vsbNeeded && (vsbPolicy != VERTICAL_SCROLLBAR_NEVER))
				{

					extentSize = viewport.toViewCoordinates(availR.getSize());
					vsbNeeded = viewPrefSize.height > extentSize.height;

					if (vsbNeeded)
						adjustForVSB(true, availR, vsbR, vpbInsets, leftToRight);
				}
			}

			/*
			 * Set the size of the viewport first, and then recheck the Scrollable methods. Some
			 * components base their return values for the Scrollable methods on the size of the
			 * Viewport, so that if we don't ask after resetting the bounds we may have gotten the
			 * wrong answer.
			 */

			if (viewport != null)
			{
				viewport.setBounds(availR);
				if (sv != null)
				{
					extentSize = viewport.toViewCoordinates(availR.getSize());

					boolean oldHSBNeeded = hsbNeeded;
					boolean oldVSBNeeded = vsbNeeded;

					viewTracksViewportWidth = sv.getScrollableTracksViewportWidth();
					viewTracksViewportHeight = sv.getScrollableTracksViewportHeight();

					if ((vsb != null) && (vsbPolicy == VERTICAL_SCROLLBAR_AS_NEEDED))
					{
						boolean newVSBNeeded = (!viewTracksViewportHeight)
								&& (viewPrefSize.height > extentSize.height);

						if (newVSBNeeded != vsbNeeded)
						{
							vsbNeeded = newVSBNeeded;
							adjustForVSB(vsbNeeded, availR, vsbR, vpbInsets, leftToRight);
							extentSize = viewport.toViewCoordinates(availR.getSize());
						}
					}
					if ((hsb != null) && (hsbPolicy == HORIZONTAL_SCROLLBAR_AS_NEEDED))
					{
						boolean newHSBbNeeded = (!viewTracksViewportWidth)
								&& (viewPrefSize.width > extentSize.width);

						if (newHSBbNeeded != hsbNeeded)
						{
							hsbNeeded = newHSBbNeeded;
							adjustForHSB(hsbNeeded, availR, hsbR, vpbInsets, colFootR);
							if ((vsb != null) && !vsbNeeded
									&& (vsbPolicy != VERTICAL_SCROLLBAR_NEVER))
							{
								extentSize = viewport.toViewCoordinates(availR.getSize());
								vsbNeeded = viewPrefSize.height > extentSize.height;

								if (vsbNeeded)
									adjustForVSB(true, availR, vsbR, vpbInsets, leftToRight);
							}
						}
					}
					if ((oldHSBNeeded != hsbNeeded) || (oldVSBNeeded != vsbNeeded))
					{
						viewport.setBounds(availR);

						// You could argue that we should recheck the
						// Scrollable methods again until they stop changing,
						// but they might never stop changing, so we stop here
						// and don't do any additional checks.
					}
				}
			}

			/*
			 * We now have the final size of the viewport: availR. Now fixup the header and
			 * scrollbar widths/heights.
			 */
			vsbR.height = availR.height + vpbInsets.top + vpbInsets.bottom;
			hsbR.width = availR.width + vpbInsets.left + vpbInsets.right;

			rowHeadR.height = availR.height + vpbInsets.top + vpbInsets.bottom;
			rowHeadR.y = availR.y - vpbInsets.top;

			colHeadR.width = availR.width + vpbInsets.left + vpbInsets.right;
			colHeadR.x = availR.x - vpbInsets.left;

			colFootR.width = availR.width + vpbInsets.left + vpbInsets.right;
			colFootR.x = availR.x - vpbInsets.left;

			colFootR.y += availR.y + availR.height;

			/*
			 * Set the bounds of the remaining components. The scrollbars are made invisible if
			 * they're not needed.
			 */

			if (rowHead != null)
				rowHead.setBounds(rowHeadR);

			if (colHead != null)
				colHead.setBounds(colHeadR);

			if (colFoot != null)
				colFoot.setBounds(colFootR);

			if (vsb != null)
			{
				if (vsbNeeded)
				{
					vsb.setVisible(true);
					vsb.setBounds(vsbR);
				}
				else
					vsb.setVisible(false);
			}

			if (hsb != null)
			{
				if (hsbNeeded)
				{
					hsb.setVisible(true);
					hsb.setBounds(hsbR);
				}
				else
					hsb.setVisible(false);
			}

			if (lowerLeft != null)
				lowerLeft.setBounds(leftToRight ? rowHeadR.x : vsbR.x, colFootR.y,
					leftToRight ? rowHeadR.width : vsbR.width, hsbR.height + colFootR.height);

			if (lowerRight != null)
				lowerRight.setBounds(leftToRight ? vsbR.x : rowHeadR.x, colFootR.y,
					leftToRight ? vsbR.width : rowHeadR.width, hsbR.height + colFootR.height);

			if (upperLeft != null)
				upperLeft.setBounds(leftToRight ? rowHeadR.x : vsbR.x, colHeadR.y,
					leftToRight ? rowHeadR.width : vsbR.width, colHeadR.height);

			if (upperRight != null)
				upperRight.setBounds(leftToRight ? vsbR.x : rowHeadR.x, colHeadR.y,
					leftToRight ? vsbR.width : rowHeadR.width, colHeadR.height);
		}
		else
			super.layoutContainer(parent);
		}
	} // end method layoutContainer

	/**
	 * Adjusts the given Rectangle <code>available</code> based on if the vertical scrollbar (vsb)
	 * is needed (<code>wantsVSB</code>). The location of the vsb is updated in <code>vsbR</code>,
	 * and the viewport border insets (<code>vpbInsets</code>) are used to offset the vsb.
	 *
	 * @param wantsVSB    indicates if a vertical scrollbar is needed
	 * @param available   the available rectangle
	 * @param vsbR        the rectangle that containd the updated location of the vertical scrollbar
	 * @param vpbInsets   the viewport border inserts
	 * @param leftToRight the reading way
	 */
	private void adjustForVSB(boolean wantsVSB, Rectangle available, Rectangle vsbR,
		Insets vpbInsets, boolean leftToRight)
	{
		int vsbWidth = vsb.getPreferredSize().width;

		if (wantsVSB)
		{
			available.width -= vsbWidth;
			vsbR.width = vsbWidth;

			if (leftToRight)
				vsbR.x = available.x + available.width + vpbInsets.right;
			else
			{
				vsbR.x = available.x - vpbInsets.left;
				available.x += vsbWidth;
			}
		}
		else
			available.width += vsbWidth;
	}

	/**
	 * Adjusts the Rectangle <code>available</code> based on if the horizontal scrollbar is needed
	 * (<code>wantsHSB</code>). The location of the hsb is updated in <code>hsbR</code>, and the
	 * viewport border insets (<code>vpbInsets</code>) are used to offset the hsb.
	 *
	 * @param wantsHSB  indicates if an horizontal scrollbar is needed
	 * @param available the available rectangle
	 * @param hsbR      the horizontal scrollbar
	 * @param vpbInsets the viewport border inserts
	 * @param colFootR  the column footer rectangle
	 */
	private void adjustForHSB(boolean wantsHSB, Rectangle available, Rectangle hsbR,
		Insets vpbInsets, Rectangle colFootR)
	{
		int hsbHeight = hsb.getPreferredSize().height;

		if (wantsHSB)
		{
			available.height -= hsbHeight;
			hsbR.y = available.y + available.height + vpbInsets.bottom + colFootR.height;
			hsbR.height = hsbHeight;
		}
		else
			available.height += hsbHeight;
	}

	/**
	 * Overload to have a customized scrollable preferred size.
	 *
	 * @param  view the view
	 * @return the prefered size if he view has one, (0,0) otherwise
	 */
	public Dimension getCustomizedScrollablePreferredSize(Component view)
	{
		if (view instanceof LyJCustomizedScrollable)
			return ((LyJCustomizedScrollable)view).getCustomizedScrollablePreferredSize();
		return ((view != null) ? view.getPreferredSize() : new Dimension(0, 0));
	}

	/**
	 * The UI resource version of ScrollPaneLayout.
	 */
	public static class UIResource extends LyJTableScrollPaneLayout
		implements javax.swing.plaf.UIResource
	{
		private static final long serialVersionUID = -408128469738251373L;
	}

} // end class LyJTableScrollPaneLayout
