/*
 * LyJTableStatusColumn.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.table;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import leon.misc.*;
import leon.view.event.*;
import leon.view.win.swing.component.*;


/**
 * This class implements a column of cells that can display an image. There is a cell for each row
 * of its associated LyTableComp. This column is displayed at the left side of the parent table. It
 * doesn't move with the horizontal scrollbar.
 *
 * @version $Id: LyJTableStatusColumn.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class LyJTableStatusColumn extends JPanel
	implements LayoutManager,
		TableModelListener,
		ActionListener
{
	private static final long   serialVersionUID = 7991609258716180243L;

	public static final String  __version =
		"$Id: LyJTableStatusColumn.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The width of a cell. */
	public static final int     WIDTH = 22;

	/** The minimum height of a cell. */
	public static final int     HEIGHT = 16;

	/** The maximum width of a line in a tooltip. */
	public static int		    TIP_MAX_CHARS = 60;

	/** The parent table. */
	protected LyJTableComponent _tableComp;

	/** The data model of the parent table. */
	protected LyJTableModel     _model;

	/** The component to display as header of the column. */
	protected Component		    _statusHeader;

	/** The component to display as footer of the column. */
	protected Component		    _statusFooter;

	/** The font used to draw text. */
	protected Font			    _font;

	/** The background color. */
	protected Color			    _color;

	/** Indicates if there is at least one image in the status column. */
	protected boolean		    _hasImage;

	/**
	 * Constructor. Calls the super implementation and initializes table font, background color
	 * table component, model, layout. Then it updates the content and add a mouse listener on the
	 * table.
	 *
	 * @param table the parent table of this column
	 * @param font  the table font
	 * @param color the table background color
	 */
	public LyJTableStatusColumn(JTable table, Font font, Color color)
	{
		super();

		_hasImage = false;
		_font = font;
		_color = color;
		setBackground(_color);
		_tableComp = (LyJTableComponent)table;
		_model = (LyJTableModel)_tableComp.getModel();
		setLayout(this);
		updateContent();
		_model.addTableModelListener(this);
	}

	/**
	 * ReFill columns.
	 */
	public void updateContent()
	{
		int rowCount = _model.getRowCount();
		int compCount = getComponentCount();
		int dataRowCount = _model.getDataRowCount();
		int i;

		if (rowCount > compCount)
		{
			// refresh existing status
			for (i = 0; i < compCount; i++)
				update(i);

			// add new
			for (i = compCount; i < rowCount; i++)
			{
				if (i < dataRowCount)
					add(createStatusColumnCell(_model.getRowImage(i), _model.getRowToolTip(i)), i);
				else
					add(createStatusColumnCell(null, null), i);
			}
		}
		else
		{
			for (i = compCount - 1; i >= rowCount; i--)
				remove(i);

			// refresh existing status
			for (i = 0; i < rowCount; i++)
				update(i);
		}
	} // end method updateContent

	/**
	 * Indicates that the status column has at least one image.
	 */
	public void setHasImage()
	{
		_hasImage = true;
	}
	
	/**
	 * Update the image to display in the row.
	 * 
	 * @param row the number of the row to update
	 */
	public void update(int row)
	{
		Image  image = _model.getRowImage(row);
		String tooltip = _model.getRowToolTip(row);
		Icon   icon = ((image != null) ? (new ImageIcon(image)) : null);
		
		if (icon != null)
			_hasImage = true;
		
		if (row >= getComponentCount())
			return;

		Component		 component = getComponent(row);
		JButton			 button;
		
		if (component instanceof JToolBar)
		{
			JToolBar	toolbar = (JToolBar)component;
			
			button = (JButton)toolbar.getComponent(0);
		}
		else
			button = (JButton)component;

		// If not already registered as listener
		ActionListener[] listeners = button.getActionListeners();
		int				 i, n = ((listeners == null) ? 0 : listeners.length);

		for (i = 0; i < n; i++)
			if (listeners[i] == this)
				break;
		if ((i == n) // Not found
				&& ((icon != null) || (tooltip != null)))
			button.addActionListener(this);
		else if ((i < n) // Found
				&& ((icon == null) && (tooltip == null)))
			button.removeActionListener(this);

		button.setIcon(icon);
		button.setRequestFocusEnabled(icon != null);
		button.setToolTipText(LyTools.formatToHTML(tooltip, LyJTableStatusColumn.TIP_MAX_CHARS));
	} // end method update

	/**
	 * Creates a component to display in a cell.
	 *
	 * @param  image   the image to display in this cell
	 * @param  tooltip the tooltip to display in the cell
	 * @return the created component
	 */
	public Component createStatusColumnCell(Image image, String tooltip)
	{
		Icon icon = null;

		if (image != null)
		{
			icon = new ImageIcon(image);
			_hasImage = true;
		}

		JButton comp = new JButton(icon)
			{
				private static final long serialVersionUID = 3799838938381989166L;

				/**
				 * This button can gain focus.
				 *
				 * @return <code>true</code>
				 * @see    java.awt.Component#isFocusable()
				 */
				public boolean isFocusable()
				{
					return (getIcon() != null);
				}
			};
			
		LyJToolBarComponent	toolbar = new LyJToolBarComponent();
		
		comp.setRequestFocusEnabled(icon != null);
		comp.setBorder(new LyJSpecialCellBorder(1, 1, 1, 1, false, true));
		comp.setBackground(_color);
		comp.setBorderPainted(true);
		comp.setOpaque(true);
		comp.setToolTipText(LyTools.formatToHTML(tooltip, LyJTableStatusColumn.TIP_MAX_CHARS));

		if ((image != null) || (tooltip != null))
			comp.addActionListener(this);

		comp.setText("");
		
		toolbar.setLayout(new BorderLayout(0, 0));
		toolbar.add(comp, BorderLayout.CENTER);
		
		return toolbar;
	} // end method createStatusColumnCell

	/**
	 * Gets the component used as header for this column.
	 *
	 * @return the header
	 */
	public LyJTableStatusHeader getHeader()
	{
		if (_statusHeader == null)
			_statusHeader = new LyJTableStatusHeader(_font, _color, _model, _tableComp);

		return (LyJTableStatusHeader)_statusHeader;
	}

	/**
	 * Gets the component used as footer for this column.
	 *
	 * @return the footer
	 */
	public Component getFooter()
	{
		if (_statusFooter == null)
		{
			_statusFooter = createStatusColumnCell(null, null);
			
			Component	borderComponent = _statusFooter;
			
			if (borderComponent instanceof JToolBar)
				borderComponent = ((JToolBar)_statusFooter).getComponent(0);
			
			((JComponent)borderComponent).setBorder(new LyJFooterCellBorder());
		}

		return _statusFooter;
	}

	/**
	 * Called when an update of the layout is needed. Calculates the minimum size dimensions for the
	 * specified container, given the components it contains.
	 *
	 * @param parent the component to be laid out
	 * @see   #preferredLayoutSize(Container)
	 */
	public void layoutContainer(Container parent)
	{
		int n = parent.getComponentCount();
		int m = _model.getRowCount();

		if (n != m)
		{
			updateContent();
			layoutContainer(parent);
		}
		else
		{
			int i, height = HEIGHT, y = 0;

			for (i = 0; i < n; i++)
			{
				height = Math.max(HEIGHT,
						Math.max(getComponent(i).getPreferredSize().height,
							_tableComp.getRowHeight(i)));
				_tableComp.setRowHeight(i, height);
			}

			for (i = 0; i < n; i++)
			{
				getComponent(i).setBounds(0, y, WIDTH, _tableComp.getRowHeight(i));
				y += _tableComp.getRowHeight(i);
			}
		}
	} // end method layoutContainer

	/**
	 * Calculates the minimum size dimensions for the specified container, given the components it
	 * contains.
	 *
	 * @param  parent the component to be laid out
	 * @return the minimum size of this container.
	 * @see    #preferredLayoutSize(Container)
	 */
	public Dimension minimumLayoutSize(Container parent)
	{
		return new Dimension(preferredLayoutSize(parent).width, 10);
	}

	/**
	 * Calculates the preferred size dimensions for the specified container, given the components it
	 * contains.
	 *
	 * @param  parent the container to be laid out
	 * @return the preferred size of this container.
	 * @see    #minimumLayoutSize
	 */
	public Dimension preferredLayoutSize(Container parent)
	{
		int n = _model.getRowCount();

		if (n == 0)
			return new Dimension(0, 0);

		if (!_hasImage)
			return new Dimension(0, 0);

		int i, height = 0, width;

		for (i = 0; i < n; i++)
			height += _tableComp.getRowHeight(i);

		Insets insets = parent.getInsets();

		width = WIDTH;
		return new Dimension(width + insets.left + insets.right,
			height + insets.top
			+ insets.bottom);
	}

	/**
	 * Called when a component has been added to this container.<br>
	 * This method does nothing.
	 */
	public void addLayoutComponent(String constraint, Component comp)
	{
	}

	/**
	 * Called when a component has been removed from this container.<br>
	 * This method does nothing.
	 */
	public void removeLayoutComponent(Component comp)
	{
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#setEnabled(boolean)
	 */
	public void setEnabled(boolean enabled)
	{
		super.setEnabled(enabled);
		
		int			i, n = getComponentCount();
		Component	comp;
		
		for (i = 0; i < n; i++)
		{
			comp = getComponent(i);
			comp.setEnabled(enabled);
		}
	}

	/**
	 * Called when the model of the parent table has changed.<br>
	 * This fine grain notification tells listeners the exact range of cells, rows, or columns that
	 * changed.
	 */
	public void tableChanged(TableModelEvent e)
	{
		int type = e.getType();
		int row = e.getFirstRow();

		switch (type)
		{
		case TableModelEvent.INSERT:
		{
			add(createStatusColumnCell(_model.getRowImage(row), _model.getRowToolTip(row)), row);
			break;
		}
		case TableModelEvent.DELETE:
		{
			remove(row);
			break;
		}
		case TableModelEvent.UPDATE:
		{
			update(row);
			break;
		}
		default:
		{
		}
		}
	}

	/**
	 * Retrieves the id of cell that emits the action listener and send event to controller.
	 */
	public void actionPerformed(ActionEvent e)
	{
		if (_tableComp.getActionTransmitter() == null)
			return;

		Component   comp = (Component)e.getSource();
		Component[] comps = getComponents();
		int		    i, nComps = comps.length, rowIndex = -1;

		for (i = 0; (i < nComps) && (rowIndex == -1); i++)
		{
			if (comp == comps[i]
			            || (comps[i] instanceof JToolBar
			            		&& ((JToolBar)comps[i]).getComponentCount() > 0
								&& comp == ((JToolBar)comps[i]).getComponent(0)))
				rowIndex = i;
		}

		String		  item = _model.getRowId(rowIndex);
		LyActionEvent actionEvent = new LyActionEvent(this, null, item, LyActionDemand.TYPE_STATUS);

		actionEvent.addIndex(rowIndex);

		_tableComp.removeEditor();
		_tableComp.actionPerformed(actionEvent);
	}
} // end class LyJTableStatusColumn
