/*
 * LyJNumberDocument.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.text;

import javax.swing.text.*;
import leon.misc.*;


/**
 * This class manages the content of a number field.
 *
 * @version $Id: LyJNumberDocument.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class LyJNumberDocument extends LyJTextFieldDocument
{
	private static final long  serialVersionUID = 5914029521814761005L;

	public static final String __VERSION =
		"$Id: LyJNumberDocument.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The maximum value allowed for this number field. */
	protected double		   _max;

	/** The minimum value allowed for this number field. */
	protected double		   _min;

	/** The char used for decimal separation. */
	protected char			   _decimalSeparator;

	/** The number of allowed decimals. */
	protected int			   _decimals;

	/**
	 * Constructor. Calls the super implementation and initializes class attributes with the given
	 * values.
	 *
	 * @param maxChars         the maximum number of chars
	 * @param min              the minimum value
	 * @param max              the maximum value
	 * @param decimalSeparator the character used for decimal separation
	 * @param decimals         the number of allowed decimals
	 */
	public LyJNumberDocument(int maxChars, double min, double max, char decimalSeparator,
		int decimals)
	{
		super(maxChars);
		_max = max;
		_min = min;
		_decimals = decimals;
		_decimalSeparator = decimalSeparator;
	}

	/**
	 * Inserts a string in the number field.<br>
	 * Overriden to filter input.
	 *
	 * @param  offset the position of this insert
	 * @param  str    the string to insert
	 * @param  a      the attributes to associate with the inserted content. This may be null if
	 *                there are no attributes
	 * @throws BadLocationException if the value of this field cannot be retrieved
	 * @see    javax.swing.text.Document#insertString(int, java.lang.String,
	 *         javax.swing.text.AttributeSet)
	 */
	public void insertString(int offset, String str, AttributeSet a)
		throws BadLocationException
	{

		String value = getText(0, getLength());

		if (str.length() == 1) // The user types a char
		{
			char keyChar = str.charAt(0);

			if ((keyChar == '-'))
			{
				if (_min >= 0)
				{
					beep();
					return;
				}

				// Check that there is only one '-' and that its position is
				// first
				int pos = value.indexOf('-');

				// If no '-' return
				if (pos < 0)
				{
					if (offset == 0)
						super.insertString(offset, str, a);
					else
					{
						beep();
						return;
					}
				}
				else
				{
					beep();
					return;
				}
			}
			else if ((keyChar == _decimalSeparator) || (keyChar == '.'))
			{
				int pos = value.indexOf(_decimalSeparator);

				// Check that there is only one decimal separator
				// FFT00994
				if (_decimals == 0)
				{
					beep();
					return;
				}

				// FFT00994
				if (pos >= 0)
				{
					beep();
					return;
				}

				// Check that the number of decimals is still correct
				// FFT00994
				if ((_decimals > 0) && ((value.length() - offset) > _decimals))
				{
					beep();
					return;
				}

				if ((keyChar == '.') && (_decimalSeparator != '.'))
					str = _decimalSeparator + "";

				super.insertString(offset, str, a);
			}
			else if (Character.isDigit(keyChar))
			{
				// Check if trying to put the digit before the '-'
				int pos = value.indexOf('-');

				if (offset <= pos)
				{
					beep();
					return;
				}

				// Check the number of decimals
				pos = value.indexOf(_decimalSeparator);

				// FFT00562
				if ((pos >= 0) && (offset > pos) && ((value.length() - pos - 1) == _decimals))
				{
					beep();
					return;
				}

				// Check if value always between min & max
				double val = LyTools.doubleFromString(value);
				double newVal = LyTools.doubleFromString(insertStringInString(offset, value, str));

				if ((val < _min) || (val > _max))
					super.insertString(offset, str, a);
				else if ((newVal >= _min) && (newVal <= _max))
					super.insertString(offset, str, a);
				else
				{
					beep();
					return;
				}
			}
			else
			{
				beep();
				return;
			}
		}
		else
		// The user pushes a button or a controller sets the value -> no control
		// needed
		{
			String newValue = insertStringInString(offset, value, str);
			double newVal = LyTools.doubleFromString(newValue);

			if ((newVal >= _min) && (newVal <= _max))
				super.insertString(offset, str, a);
			else
			{
			}
		}
	} // end method insertString

	/**
	 * Sets the minimum value.
	 *
	 * @param min the new minimum value
	 */
	public void setMin(double min)
	{
		_min = min;
	}

	/**
	 * Set the maximum value.
	 *
	 * @param max the new maximum value
	 */
	public void setMax(double max)
	{
		_max = max;
	}

	/**
	 * Set the number of decimals.
	 *
	 * @param decimals the new number of decmals
	 */
	public void setDecimals(int decimals)
	{
		_decimals = decimals;
	}

	/**
	 * Set the character used for decimal separation.
	 *
	 * @param decimalSeparator the new decimal separator
	 */
	public void setDecimalSeparator(char decimalSeparator)
	{
		_decimalSeparator = decimalSeparator;
	}

	/**
	 * Inserts a String into another one.
	 *
	 * @param  offset the position of this insert
	 * @param  target the target string of this insert
	 * @param  toAdd  the string to add
	 * @return the result of this insert
	 */
	protected String insertStringInString(int offset, String target, String toAdd)
	{
		String begin = target.substring(0, offset);
		String end = target.substring(offset);
		String result = begin + toAdd + end;

		return result;
	}

} // end class LyJNumberDocument
