/*
 * LyJTextAreaDocument.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing.text;

import java.awt.*;
import java.util.*;
import javax.swing.text.*;
import leon.misc.*;


/**
 * This class manages the content of a multilined text area.
 *
 * @version $Id: LyJTextAreaDocument.java 34698 2012-03-15 09:56:52Z jlsaouli $
 * @author  Lyria
 */
public class LyJTextAreaDocument extends PlainDocument
{
	private static final long  serialVersionUID = -1520133564828706206L;

	public static final String __VERSION =
		"$Id: LyJTextAreaDocument.java 34698 2012-03-15 09:56:52Z jlsaouli $";

	/** The maximum number of characters of this text area. */
	protected int			   _maxChars;

	/** List of valid characters sets. */
	
	protected Collection<String>	_patterns;

	/** The application environment. */
	protected LyEnvironment    _environment;

	/**
	 * Constructor. Calls the super implementation and initialize the maximum number of char with
	 * the given number and sets the list of patterns to <code>null</code>.
	 *
	 * @param maxChars The maximum number of char of this text area
	 */
	public LyJTextAreaDocument(int maxChars)
	{
		super();
		_maxChars = maxChars;
		_patterns = null;
	}

	/**
	 * Sets the list of available patterns.
	 *
	 * @param patterns the new list og patterns
	 */
	
	public void setPatterns(Collection<String> patterns)
	{
		_patterns = patterns;
	}

	/**
	 * Sets the application environment.
	 *
	 * @param environment the application environment
	 */
	public void setEnvironment(LyEnvironment environment)
	{
		_environment = environment;
	}

	/**
	 * Checks if the given string matches patterns of this text area.
	 *
	 * @param  str the string that must be checked
	 * @return null if the string does not match the pattern, otherwise return the string with
	 *         possible modifications
	 */
	
	protected String verifyPatterns(String str)
	{
		if ((_patterns == null) || (str == null))
			return ((str == null) ? "" : str);

		String   charType, c;
		boolean  cont = true;
		Iterator<String> it;
		// Each character of the value must match at least one charType
		for (int i = 0; (i < str.length()) && (cont); i++)
		{
			c = "" + str.charAt(i);
			cont = false;
			it = _patterns.iterator();
			while ((it.hasNext()) && (!cont))
			{
				charType = it.next();

				// If only upper characters allowed, transform
				if ((charType != null) && (charType.indexOf("[A-Z]") >= 0))
				{
					c = c.toUpperCase();
					str = str.substring(0, i) + c + str.substring(i + 1);
				}

				cont |= LyString.matchFormat(_environment, charType, c);
			}
		}

		return (cont ? str : null);
	} // end method verifyPatterns

	/**
	 * Insert a string in the text area.<br>
	 * Overriden to filter input.
	 *
	 * @param  offset the position of this insert
	 * @param  str    the string to insert
	 * @param  a      the attributes to associate with the inserted content. This may be null if
	 *                there are no attributes
	 * @throws BadLocationException if the value of this field cannot be found
	 * @see    javax.swing.text.Document#insertString(int, java.lang.String,
	 *         javax.swing.text.AttributeSet)
	 */
	public void insertString(int offset, String str, AttributeSet a)
		throws BadLocationException
	{
		if ((str = verifyPatterns(str)) == null)
		{
			beep();
			return;
		}

		String value = getText(0, getLength());
		int    length = str.length();

		if (_maxChars > 0)
		{
			if ((value == null) && (length <= _maxChars))
				super.insertString(offset, str, a);
			else if ((value != null) && (value.length() + length) <= _maxChars)
				super.insertString(offset, str, a);
			else if ((value != null)&& (value.length() + length) > _maxChars)
			{
				if (offset < _maxChars)
				{
					int n = (value.length() + length) - _maxChars;

					super.insertString(offset, str, a);
					remove(_maxChars, n);
				}
				else
					beep();
			}
			else
				beep();
		}
		else
			super.insertString(offset, str, a);

	} // end method insertString

	/**
	 * Sets the content of this document without any control.
	 *
	 * @param  str the new content
	 * @throws BadLocationException
	 */
	public void setString(String str)
		throws BadLocationException
	{
		super.remove(0, getLength());
		super.insertString(0, str, null);
	}

	/**
	 * Emit a beep.
	 */
	public void beep()
	{
		Toolkit.getDefaultToolkit().beep();
	}

	/**
	 * Sets the maximum number of characters of this area.
	 *
	 * @param maxChars the new maximum
	 */
	public void setMaxChars(int maxChars)
	{
		_maxChars = maxChars;
	}

} // end class LyJTextAreaDocument
