/*
 * LySqlRequestEncoder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.sql;

import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.info.LyFilter.*;
import leon.info.binding.*;
import leon.info.daemon.*;
import leon.misc.*;
import leon.peer.*;
import leon.peer.union.*;
import leon.peer.util.*;

import java.io.*;

import java.sql.*;

import java.util.*;


/**
 * This class is used to send requests to a SQL database connection and to receive its answers.
 *
 * @version $Id: LySqlRequestEncoder.java 41837 2012-12-21 12:58:52Z bangiole $
 * @author  Lyria
 * @see     LyRequestEncoder
 * @see     LySqlConnection
 * @see     LySqlDataProvider
 */
public class LySqlRequestEncoder extends LyBlobRequestEncoder implements LyJoinRequestEncoder
{
    private static final String PERF_SQL_CREATE_OBJECT = "SQL-createObject";
    private static final String PERF_SQL_DELETE_OBJECTS = "SQL-deleteObjects";
    private static final String PERF_SQL_SET_OBJECT = "SQL-setObject";
    private static final String PERF_SQL_SEND_RESPONSE = "SQL-sendResponse";
    private static final String PERF_SQL_GET_OBJECTS = "SQL-getObjects";
    private static final String PERF_SQL_ENCODE_REQUEST = "SQL-encodeRequest";
    private static final String PERF_SQL_SEND_QUERY = "SQL-sendQuery";
    private static final String PERF_SQL_GET_OBJECT = "SQL-getObject";
    private static final String PERF_SQL_DECODE_VALUES = "SQL-decodeValues";
    private static final String PERF_SQL_GET_SIZE = "SQL-getSize";

    public static final String __VERSION =
        "$Id: LySqlRequestEncoder.java 41837 2012-12-21 12:58:52Z bangiole $";
    public static final String DOMAIN_FILTER_RELATION = "LyFilterRelationDomain";

    protected static int       _tableAliasCounter = 0;
    public static final char   CHAR_QUOTE = '\'';
    public static final String STRING_QUOTE = "'";
    public static final String STRING_TWO_QUOTES = "''";

    protected static final String UNION_SELECT_FILTER = "$$UNION_SELECT_FILTER$$";
    protected static final int    TYPE_DATE = 0;
    protected static final int    TYPE_ARRAY = 1;
    protected static final int    TYPE_STRING = 2;
    protected static final int    TYPE_BOOLEAN = 3;
    protected static final int    TYPE_NUMBER = 4;

    /**
     * Strings corresponding to supported conditions. Be carefull, theses strings must be sorted
     * like LyFilterExpression conditions.
     */
    static String[] _conditionNames =
        {
            "=", "<", "<=", ">", ">=", "contains", null, // No condition for starts with
            null, // No condition for ends with
        };

    /**
     * Strings corresponding to supported conditions. Be carefull, theses strings must be sorted
     * like LyFilterExpression conditions.
     */
    static String[] _notConditionNames =
        {
            "<>", ">=", ">", "<=", "<", null, // No inverted condition for contains
            null, // No inverted condition for starts with
            null, // No inverted condition for ends with
        };

    /** Current blobs/clobs being used by the request. */
    protected static HashMap<Thread, ArrayList<Object[]>> _usedBlobsList;

    // quoted for table names and columns : modified in init method of subclasses
    protected String QUOTE_CHAR_LEFT;
    // quoted for table names and columns : modified in inti method of subclasses
    protected String QUOTE_CHAR_RIGHT;

    /**  */
    protected Vector<LyField> _fields;

    /**  */
    protected String _valueSep;

    /**  */
    protected boolean _trim;

    /** List of peer classes used to send join requests for a LyUnionRequestEncoder. */
    protected Collection<LyClass> _joinPeerClasses;

    /** Using escape sequence or null if no escape is used. */
    protected String _escapeSequence = null;

    /**
     * This class is used to keep the encoding context of a request when going into multiple
     * encoders.
     */
    protected LySqlRequestInfo _sqlRequestInfo;

    /**
     * Constructor.
     *
     * @param provider
     * @param dataSource
     * @param valueSep
     */
    public LySqlRequestEncoder(LyDataProvider provider, LyDataSource dataSource, String valueSep)
    {
        super(provider, dataSource);
        _valueSep = valueSep;
    }

    /**
     * Initialisation of the data of the instance.
     */
    protected void init()
    {
        super.init();
        QUOTE_CHAR_LEFT = "";
        QUOTE_CHAR_RIGHT = "";
        _fields = null;
        _valueSep = null;
        _trim = false;
        _joinPeerClasses = null;
    }

    /**
     */
    public void setTrim(boolean trim)
    {
        _trim = trim;
    }

    /**
     * Set the request encoder response. Inherited method of LyRequestEncoder to set properties to
     * the response.
     *
     * @param the response
     */
    public void setResponse(LyResponse response)
    {
        _response = response;

        // Set properties of the response.
        _response.setSorted(true);
        _response.setFiltered(true);
        _response.setCounted(true);
        _response.setPartial(true);
        _response.setDistinctValues(true);
        _response.setPartialNotification(true);
    }

    /**
     * Get an object.
     *
     * @param  objectId object identifier
     * @param  fields   the fields to load (if null, load all)
     * @return true if the request was correctly sent to the external data provider.
     */
    public boolean getObject(LyObjectId objectId, Vector<LyField> fields)
    {
        LyEnvironment environment = getEnvironment();
        long          before = System.currentTimeMillis();

        if (environment.getPerfLogger() != null)
            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
                PERF_SQL_GET_OBJECT, "objectId=" + objectId.getFullString());

        // Set the response in running mode
        _response.setStatus(LyResponse.Status.RUNNING);

        if (environment.getPerfLogger() != null)
            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
                PERF_SQL_ENCODE_REQUEST, null);

        // Create the SQL request
        String request_text = encodeSelectRequest(null, objectId, fields, null, null, -1);

        if (environment.getPerfLogger() != null)
            environment.getPerfLogger().addEndEvent(_provider.getApplication(),
                PERF_SQL_ENCODE_REQUEST, null);

        // Get a valid connection.
        LySqlConnection connection = (LySqlConnection) _provider.getConnection();

        if (connection == null)
        {
            //      No available connection
            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
                environment.logOut("No avalaible connection.");

            _dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
            return false;
        }

        try
        {
	        if (environment.getPerfLogger() != null)
	            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
	                PERF_SQL_SEND_QUERY, null);

	        // Send the request to the connection
	        ResultSet rs = connection.sendQueryRequest(request_text);

	        if (environment.getPerfLogger() != null)
	            environment.getPerfLogger().addEndEvent(_provider.getApplication(), PERF_SQL_SEND_QUERY,
	                null);

	        if (rs == null)
	        {
	            // An error occured while sending the request to the database
	            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
	                environment.logOut("An error occured while sending the request \"" + request_text
	                    + "\" to the database.");

	            _response.setError(connection.getError());

	            if ((connection != null)
	            	&& connection.isReserved())
	            {
	                _provider.releaseConnection(connection);
	                connection = null;
	            }

	            _dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);

	            if (environment.getPerfLogger() != null)
	                environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                    PERF_SQL_GET_OBJECT, LyPerfLogger.ERROR_STRING);

	            return false;
	        }

	        if (environment.getPerfLogger() != null)
	            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
	                PERF_SQL_DECODE_VALUES, null);

	        // Decode the values
	        Vector<LyRowObject> objects;

	        try
	        {
	            objects = decodeSelectValues(rs, _fields);
	        }
	        catch (Exception e)
	        {
	            environment.writeException(e);

	            try
	            {
	                rs.close();
	            }
	            catch (SQLException sqle)
	            {

	                if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
	                        && environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
	                    environment.writeException(sqle);
	            }

	            _provider.releaseConnection(connection);
	            connection = null;

	            _response.setError(e.getMessage());
	            _dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);

	            if (environment.getPerfLogger() != null)
	            {
	                environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                    PERF_SQL_DECODE_VALUES, "error");
	                environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                    PERF_SQL_GET_OBJECT, "error");
	            }

	            return false;
	        }

	        if (environment.getPerfLogger() != null)
	            environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                PERF_SQL_DECODE_VALUES, null);

	        // Release the connection.
	        try
	        {
	            rs.close();
	        }
	        catch (SQLException sqle)
	        {

	            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
	                    && environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
	                environment.writeException(sqle);
	        }

	        _provider.releaseConnection(connection);
	        connection = null;

	        int no = objects.size();

	        if (environment.getPerfLogger() != null)
	            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
	                PERF_SQL_SEND_RESPONSE, null);

	        if (no == 1)
	        {
	            // Add the object to the response
	            LyRowObject object = objects.elementAt(0);

	            _dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, object, null);
	        }
	        else if (no == 0)
	            // If no object was found, nothing to do
	            _dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, null, null);
	        else if (no > 1)
	        {
	            // More than one object was found : error
	            environment.logWarning("More than one object with id :" + objectId + " were found.");

	            _response.setError(environment.getMessage("LY_ER_GET_OBJECT", objectId));
	            _dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
	        }

	        if (environment.getPerfLogger() != null)
	        {
	            environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                PERF_SQL_SEND_RESPONSE, null);
	            environment.getPerfLogger().addEndEvent(_provider.getApplication(), PERF_SQL_GET_OBJECT,
	                null);
	        }

	        long after = System.currentTimeMillis();

	        if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
	                && environment.isDebugLevel(LyEnvironment.DEBUG_PERF))
	        {
	            long delay = after - before;

	            environment.logOut("Delay = " + delay);

	            if (delay > ((LySqlDataProvider) _provider).getDelay())
	            {
	                environment.logOut("Request delay exceeded = " + request_text);
	                Thread.dumpStack();
	            }
	        }
        }
        finally
        {
            if ((connection != null)
            	&& connection.isReserved())
            {
                environment.logOut("[CONNECTION] connection " + connection + " was not released, freeing it.");
                if (environment.isDebugLevel(LyEnvironment.DEBUG_TRANSACTION))
                    Thread.dumpStack();
            }
        }

        return true;
    } // end method getObject

    /**
     * Get a set of objects with a join request.
     *
     * @param  encoder the asking union encoder.
     * @param  fields  the fields to load (if null, load all)
     * @param  sort    indicates sorting criterion, or null if default sort is used.
     * @param  filter  indicates filtering criterion, or null if there is no filter.
     * @param  n       maximum number of objects to get.
     * @return true if the request was correctly sent to the external data provider.
     */
    public Collection<LyValueSet> getObjects(LyUnionRequestEncoder encoder, Vector<LyField> fields,
        LySort sort, LyFilter filter, int n)
    {
        // WARNING : for this kind of request there is no _response object !!!
        LyEnvironment environment = getEnvironment();
        long          before = System.currentTimeMillis();

        _joinPeerClasses = encoder.getPeerClasses();
        _classBinding = encoder.getClassBinding();

        if (filter != null)
            filter = filter.getPeerFilter();

        // Create the SQL request
        String request_text = encodeSelectRequest(encoder, null, fields, sort, filter, n);

        // Get a valid connection.
        LySqlConnection connection = (LySqlConnection) _provider.getConnection();

        if (connection == null)
        {
            //      No available connection
            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
                environment.logOut("No avalaible connection.");

            return null;
        }

        Collection<LyValueSet> values = null;

        try
        {
	        // Send the request to the connection
	        ResultSet rs = connection.sendQueryRequest(request_text, n);

	        if (rs == null)
	        {
	            // An error occurred while sending the request to the database
	            if ((connection != null)
	            	&& connection.isReserved())
	            {
	                _provider.releaseConnection(connection);
	                connection = null;
	            }

	            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
	                environment.logOut("An error occured while sending the request \"" + request_text
	                    + "\" to the database.");

	            return null;
	        }

	        // Decode the values
	        try
	        {
	            values = decodeSelectValues(encoder, rs, _fields);
	        }
	        catch (Exception e)
	        {
	            environment.writeException(e);

	            try
	            {
	                rs.close();
	            }
	            catch (SQLException sqle)
	            {

	                if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
	                        && environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
	                    environment.writeException(sqle);
	            }

	            _provider.releaseConnection(connection);
	            connection = null;
	            return null;
	        }

	        // Release the connection.
	        try
	        {
	            rs.close();
	        }
	        catch (SQLException sqle)
	        {

	            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
	                    && environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
	                environment.writeException(sqle);
	        }

	        _provider.releaseConnection(connection);
	        connection = null;

	        long after = System.currentTimeMillis();

	        if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
	                && environment.isDebugLevel(LyEnvironment.DEBUG_PERF))
	        {
	            long delay = after - before;

	            environment.logOut("Delay = " + delay);

	            if (delay > ((LySqlDataProvider) _provider).getDelay())
	            {
	                environment.logOut("Request delay exceeded = " + request_text);
	                Thread.dumpStack();
	            }
	        }
        }
        finally
        {
            if ((connection != null)
            	&& connection.isReserved())
            {
                environment.logOut("[CONNECTION] connection " + connection + " was not released, freeing it.");
                if (environment.isDebugLevel(LyEnvironment.DEBUG_TRANSACTION))
                    Thread.dumpStack();
            }
        }

        return values;
    } // end method getObjects

    /**
     * Get a set of objects.
     *
     * @param  fields the fields to load (if null, load all)
     * @param  sort   indicates sorting criterion, or null if default sort is used.
     * @param  filter indicates filtering criterion, or null if there is no filter.
     * @param  n      maximum number of objects to get.
     * @return true if the request was correctly sent to the external data provider.
     */
    public boolean getObjects(Vector<LyField> fields, LySort sort, LyFilter filter, int n)
    {
        LyEnvironment environment = getEnvironment();
        long          before = System.currentTimeMillis();

        if (environment.getPerfLogger() != null)
            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
                PERF_SQL_GET_OBJECTS, "class=" + _class.getName());
        // Set the response in running mode
        _response.setStatus(LyResponse.Status.RUNNING);

        if (environment.getPerfLogger() != null)
            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
                PERF_SQL_ENCODE_REQUEST, null);

        if (filter != null)
            filter = filter.getPeerFilter();

        int		maxObjectsToNotify = 0;
        int		maxPeerLimit = LyPageWorkSpace.MAX_PEER_LIMIT;

        if ((n > maxPeerLimit)
        	&& !_response.isPaginated()
        	&& _response.isPartialNotification())
        {
            maxObjectsToNotify = n / maxPeerLimit;
            n = n % maxPeerLimit;
        }

        // Create the SQL request
        // for debug
        getRequestInfo().getCurrentTask().addInfo("* BEGIN getObjects() filter=" + filter + " sort="
            + sort + " n=" + n);

        String request_text = encodeSelectRequest(null, null, fields, sort, filter, n);

        if (environment.getPerfLogger() != null)
            environment.getPerfLogger().addEndEvent(_provider.getApplication(),
                PERF_SQL_ENCODE_REQUEST, null);

        // Get a valid connection.
        LySqlConnection connection = (LySqlConnection) _provider.getConnection();

        if (connection == null)
        {
            //      No available connection
            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
                environment.logOut("No avalaible connection.");

            _dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);

            if (environment.getPerfLogger() != null)
                environment.getPerfLogger().addEndEvent(_provider.getApplication(),
                    PERF_SQL_GET_OBJECTS, LyPerfLogger.ERROR_STRING);

            return false;
        }

        try
        {
	        getRequestInfo().getCurrentTask().addInfo("* END getObjects() : send Query");
	        getRequestInfo().getCurrentTask().addSql(request_text);

	        if (environment.getPerfLogger() != null)
	            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
	                PERF_SQL_SEND_QUERY, null);

	        // Send the request to the connection
	        ResultSet rs = connection.sendQueryRequest(request_text, n);

	        if (environment.getPerfLogger() != null)
	            environment.getPerfLogger().addEndEvent(_provider.getApplication(), PERF_SQL_SEND_QUERY,
	                null);

	        if (rs == null)
	        {

	            // An error occured while sending the request to the database
	            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
	                environment.logOut("An error occured while sending the request \"" + request_text
	                    + "\" to the database.");

	            _response.setError(connection.getError());

	            if ((connection != null)
	            	&& connection.isReserved())
	            {
	                _provider.releaseConnection(connection);
	                connection = null;
	            }

	            _dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);

	            if (environment.getPerfLogger() != null)
	                environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                    PERF_SQL_GET_OBJECTS, LyPerfLogger.ERROR_STRING);

	            return false;
	        }

	        // Decode the values
	        Vector<LyRowObject> objects;

	        try
	        {

	            if (environment.getPerfLogger() != null)
	                environment.getPerfLogger().addStartEvent(_provider.getApplication(),
	                    PERF_SQL_DECODE_VALUES, null);

	            objects = decodeSelectValues(rs, _fields);

	            if (environment.getPerfLogger() != null)
	                environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                    PERF_SQL_DECODE_VALUES, null);
	        }
	        catch (Exception e)
	        {
	            environment.writeException(e);

	            try
	            {
	                rs.close();
	            }
	            catch (SQLException sqle)
	            {

	                if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
	                        && environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
	                    environment.writeException(sqle);
	            }

	            _provider.releaseConnection(connection);
	            connection = null;

	            _response.setError(e.getMessage());
	            _dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);

	            if (environment.getPerfLogger() != null)
	            {
	                environment.getPerfLogger().addEndEvent(_provider.getApplication()
	                    .getCurrentSession(), PERF_SQL_DECODE_VALUES, LyPerfLogger.ERROR_STRING);
	                environment.getPerfLogger().addEndEvent(_provider.getApplication()
	                    .getCurrentSession(), PERF_SQL_GET_OBJECTS, LyPerfLogger.ERROR_STRING);
	            }

	            return false;
	        }

	        // Release the connection.
	        try
	        {
	            rs.close();
	        }
	        catch (SQLException sqle)
	        {

	            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
	                    && environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
	                environment.writeException(sqle);
	        }

	        _provider.releaseConnection(connection);
	        connection = null;

	        int         i, start = 0;
	        int         end = objects.size();
	        LyRowObject object;

	        if (maxObjectsToNotify > 0)
	            start = end - maxObjectsToNotify;
	        else if (maxObjectsToNotify < 0)
	            end = -maxObjectsToNotify;

	        if (start < 0)
	            start = 0;

	        if (end > objects.size())
	            end = objects.size();

	        if (environment.getPerfLogger() != null)
	            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
	                PERF_SQL_SEND_RESPONSE, null);

	        for (i = start; i < end; i++)
	        {
	            object = objects.elementAt(i);
	            _dataSource.receiveResponse(_response, LyResponse.Status.RUNNING, object, null);

	            // Be sure that results are really sorted when no sort is specified
	            if ((i > 0) && (sort == null) && _response.isSorted()
	                    && (_class.getObjectId(object).compare(
	                            _class.getObjectId(objects.elementAt(i - 1))) < 0))
	                _response.setSorted(false);
	        }

	        _dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, null, null);

	        if (environment.getPerfLogger() != null)
	        {
	            environment.getPerfLogger().addEndEvent(_provider.getApplication().getCurrentSession(),
	                PERF_SQL_SEND_RESPONSE, null);

	            environment.getPerfLogger().addEndEvent(_provider.getApplication().getCurrentSession(),
	                PERF_SQL_GET_OBJECTS, null);
	        }

	        long after = System.currentTimeMillis();

	        if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
	                && environment.isDebugLevel(LyEnvironment.DEBUG_PERF))
	        {
	            long delay = after - before;

	            environment.logOut("Delay = " + delay);

	            if (delay > ((LySqlDataProvider) _provider).getDelay())
	            {
	                environment.logOut("Request delay exceeded = " + request_text);
	                Thread.dumpStack();
	            }
	        }
        }
        finally
        {
            if ((connection != null)
            	&& connection.isReserved())
            {
                environment.logOut("[CONNECTION] connection " + connection + " was not released, freeing it.");
                if (environment.isDebugLevel(LyEnvironment.DEBUG_TRANSACTION))
                    Thread.dumpStack();
            }
        }

        return true;
    } // end method getObjects

    /**
     * Return the number of objects from the data provider with a join request.
     *
     * @param  encoder The union encoder demanding the number of the objects.
     * @param  filter  indicates filtering criterion, or null if there is no filter.
     * @return The count of objects matching the filter or null if there is an error.
     */
    public Integer getSize(LyUnionRequestEncoder encoder, LyFilter filter)
    {
        LyEnvironment environment = getEnvironment();
        long          before = System.currentTimeMillis();

        // WARNING : for this kind of request there is no _response object !!!
        _joinPeerClasses = encoder.getPeerClasses();
        _classBinding = encoder.getClassBinding();

        if (filter != null)
            filter = filter.getPeerFilter();

        // Create the SQL request
        String request_text = encodeCountRequest(encoder, filter);

        // Get a valid connection.
        LySqlConnection connection = (LySqlConnection) _provider.getConnection();

        if (connection == null)
        {
            //      No available connection
            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
                environment.logOut("No avalaible connection.");

            return null;
        }

        Integer size = null;

        try
        {
	        // Send the request to the connection
	        ResultSet rs = connection.sendQueryRequest(request_text);

	        if (rs == null)
	        {
	            // An error occurred while sending the request to the database
	            if ((connection != null)
	            	&& connection.isReserved())
	            {
	                _provider.releaseConnection(connection);
		            connection = null;
		        }

	            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
	                environment.logOut("An error occured while sending the request \"" + request_text
	                    + "\" to the database.");

	            return null;
	        }

	        // Decode the value
	        try
	        {
	            size = decodeCountValues(rs);
	        }
	        catch (Exception e)
	        {

	            try
	            {
	                rs.close();
	            }
	            catch (SQLException sqle)
	            {

	                if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
	                        && environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
	                    environment.writeException(sqle);
	            }

	            _provider.releaseConnection(connection);
	            connection = null;

	            environment.writeException(e);
	            return null;
	        }

	        // Release the connection.
	        try
	        {
	            rs.close();
	        }
	        catch (SQLException sqle)
	        {

	            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
	                    && environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
	                environment.writeException(sqle);
	        }

	        _provider.releaseConnection(connection);
	        connection = null;

	        long after = System.currentTimeMillis();

	        if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
	                && environment.isDebugLevel(LyEnvironment.DEBUG_PERF))
	        {
	            long delay = after - before;

	            environment.logOut("Delay = " + delay);

	            if (delay > ((LySqlDataProvider) _provider).getDelay())
	            {
	                environment.logOut("Request delay exceeded = " + request_text);
	                Thread.dumpStack();
	            }
	        }
        }
        finally
        {
            if ((connection != null)
            	&& connection.isReserved())
            {
                environment.logOut("[CONNECTION] connection " + connection + " was not released, freeing it.");
                if (environment.isDebugLevel(LyEnvironment.DEBUG_TRANSACTION))
                    Thread.dumpStack();
            }
        }

        return size;
    } // end method getSize

    /**
     * Return the number of objects from the data provider .
     *
     * @param  filter indicates filtering criterion, or null if there is no filter.
     * @return true if the request was correctly sent to the external data provider.
     */
    public boolean getSize(LyFilter filter)
    {
        LyEnvironment environment = getEnvironment();
        long          before = System.currentTimeMillis();

        if (environment.getPerfLogger() != null)
            environment.getPerfLogger().addStartEvent(_provider.getApplication(), PERF_SQL_GET_SIZE,
                "class=" + _class.getName());

        // Set the response in running mode
        _response.setStatus(LyResponse.Status.RUNNING);

        if (environment.getPerfLogger() != null)
            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
                PERF_SQL_ENCODE_REQUEST, null);

        if (filter != null)
            filter = filter.getPeerFilter();

        // Create the SQL request
        String request_text = encodeCountRequest(null, filter);

        if (environment.getPerfLogger() != null)
            environment.getPerfLogger().addEndEvent(_provider.getApplication(),
                PERF_SQL_ENCODE_REQUEST, null);

        // Get a valid connection.
        LySqlConnection connection = (LySqlConnection) _provider.getConnection();

        if (connection == null)
        {
            //      No available connection
            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
                environment.logOut("No avalaible connection.");

            _dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);

            if (environment.getPerfLogger() != null)
                environment.getPerfLogger().addEndEvent(_provider.getApplication(),
                    PERF_SQL_GET_SIZE, LyPerfLogger.ERROR_STRING);

            return false;
        }

        try
        {
	        if (environment.getPerfLogger() != null)
	            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
	                PERF_SQL_SEND_QUERY, null);

	        // Send the request to the connection
	        ResultSet rs = connection.sendQueryRequest(request_text);

	        if (environment.getPerfLogger() != null)
	            environment.getPerfLogger().addEndEvent(_provider.getApplication(), PERF_SQL_SEND_QUERY,
	                null);

	        if (rs == null)
	        {

	            // An error occured while sending the request to the database
	            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
	                environment.logOut("An error occured while sending the request \"" + request_text
	                    + "\" to the database.");

	            _response.setError(connection.getError());

	            if ((connection != null)
	            	&& connection.isReserved())
	            {
	                _provider.releaseConnection(connection);
	                connection = null;
	            }

	            _dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);

	            if (environment.getPerfLogger() != null)
	                environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                    PERF_SQL_GET_SIZE, LyPerfLogger.ERROR_STRING);

	            return false;
	        }

	        // Decode the value
	        Integer size;

	        try
	        {

	            if (environment.getPerfLogger() != null)
	                environment.getPerfLogger().addStartEvent(_provider.getApplication(),
	                    PERF_SQL_DECODE_VALUES, null);

	            size = decodeCountValues(rs);

	            if (environment.getPerfLogger() != null)
	                environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                    PERF_SQL_DECODE_VALUES, null);

	        }
	        catch (Exception e)
	        {
	            _response.setError(e.getMessage());

	            try
	            {
	                rs.close();
	            }
	            catch (SQLException sqle)
	            {

	                if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
	                        && environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
	                    environment.writeException(sqle);
	            }

	            _provider.releaseConnection(connection);
	            connection = null;

	            _dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);

	            if (environment.getPerfLogger() != null)
	            {
	                environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                    PERF_SQL_DECODE_VALUES, LyPerfLogger.ERROR_STRING);
	                environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                    PERF_SQL_GET_SIZE, LyPerfLogger.ERROR_STRING);
	            }

	            return false;
	        }

	        // Release the connection.
	        try
	        {
	            rs.close();
	        }
	        catch (SQLException sqle)
	        {

	            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
	                    && environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
	                environment.writeException(sqle);
	        }

	        _provider.releaseConnection(connection);
	        connection = null;

	        if (environment.getPerfLogger() != null)
	            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
	                PERF_SQL_SEND_RESPONSE, null);

	        // Add the size returned to the response
	        _dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, size, null);

	        if (environment.getPerfLogger() != null)
	            environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                PERF_SQL_SEND_RESPONSE, null);

	        if (environment.getPerfLogger() != null)
	            environment.getPerfLogger().addEndEvent(_provider.getApplication(), PERF_SQL_GET_SIZE,
	                null);

	        long after = System.currentTimeMillis();

	        if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
	                && environment.isDebugLevel(LyEnvironment.DEBUG_PERF))
	        {
	            long delay = after - before;

	            environment.logOut("Delay = " + delay);

	            if (delay > ((LySqlDataProvider) _provider).getDelay())
	            {
	                environment.logOut("Request delay exceeded = " + request_text);
	                Thread.dumpStack();
	            }
	        }
        }
        finally
        {
            if ((connection != null)
            	&& connection.isReserved())
            {
                environment.logOut("[CONNECTION] connection " + connection + " was not released, freeing it.");
                if (environment.isDebugLevel(LyEnvironment.DEBUG_TRANSACTION))
                    Thread.dumpStack();
            }
        }

        return true;
    } // end method getSize

    /**
     * Methods for deletion.*
     */
    /**
     * To delete an object.
     *
     * @param  object the object to delete
     * @return true if the request was correctly sent to the external data provider.
     */
    public boolean deleteObject(Object object)
    {
        Vector<Object> objects = new Vector<Object>(1);

        objects.addElement(object);
        return deleteObjects(objects);
    }

    /**
     * To delete an object.
     *
     * @param  object the object to delete
     * @return true if the request was correctly sent to the external data provider.
     */
    public boolean deleteObjects(Vector<Object> objects)
    {
        LyEnvironment environment = getEnvironment();
        long          before = System.currentTimeMillis();

        if (environment.getPerfLogger() != null)
        {
            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
                PERF_SQL_DELETE_OBJECTS, "class=" + _class.getName());
            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
                PERF_SQL_ENCODE_REQUEST, null);
        }

        // Set the response in running mode
        _response.setStatus(LyResponse.Status.RUNNING);

        // Create the SQL request
        String request_text = encodeDeleteRequest(objects);

        if (environment.getPerfLogger() != null)
            environment.getPerfLogger().addEndEvent(_provider.getApplication(),
                PERF_SQL_ENCODE_REQUEST, null);

        // Get a valid connection
        LySqlConnection connection = (LySqlConnection) _provider.getConnection();

        if (connection == null)
        {
            //      No available connection
            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
                environment.logOut("No avalaible connection.");

            _dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);

            if (environment.getPerfLogger() != null)
                environment.getPerfLogger().addEndEvent(_provider.getApplication(),
                    PERF_SQL_DELETE_OBJECTS, LyPerfLogger.ERROR_STRING);

            return false;
        }

        try
        {
	        if (environment.getPerfLogger() != null)
	            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
	                PERF_SQL_SEND_QUERY, null);

	        // Send the request to the connection
	        connection.setHookForUpdate(!_dataSource.isRollBack());

	        int nb_lines = connection.sendUpdateRequest(request_text);

	        if (environment.getPerfLogger() != null)
	            environment.getPerfLogger().addEndEvent(_provider.getApplication(), PERF_SQL_SEND_QUERY,
	                null);

	        if (nb_lines >= 0)
	        {
	            int    i, n = objects.size();
	            Object object;

	            // Release the connection.
	            _provider.releaseConnection(connection);
	            connection = null;

	            if (environment.getPerfLogger() != null)
	                environment.getPerfLogger().addStartEvent(_provider.getApplication(),
	                    PERF_SQL_SEND_RESPONSE, null);

	            for (i = 0; i < n; i++)
	            {
	                object = objects.elementAt(i);
	                _dataSource.receiveResponse(_response, LyResponse.Status.RUNNING, object, null);
	            }

	            _dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, null, null);

	            if (environment.getPerfLogger() != null)
	                environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                    PERF_SQL_SEND_RESPONSE, null);
	        }
	        else
	        {

	            // An error occurred while sending the request to the database
	            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
	                environment.logOut("An error occured while sending the request \"" + request_text
	                    + "\" to the database.");

	            _response.setError(connection.getError());

	            // Release the connection.
	            if ((connection != null)
	            	&& connection.isReserved())
	            {
	                _provider.releaseConnection(connection);
	                connection = null;
	            }

	            _dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);

	            if (environment.getPerfLogger() != null)
	                environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                    PERF_SQL_DELETE_OBJECTS, LyPerfLogger.ERROR_STRING);

	            return false;
	        }

	        if (environment.getPerfLogger() != null)
	            environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                PERF_SQL_DELETE_OBJECTS, null);

	        long after = System.currentTimeMillis();

	        if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
	                && environment.isDebugLevel(LyEnvironment.DEBUG_PERF))
	        {
	            long delay = after - before;

	            environment.logOut("Delay = " + delay);

	            if (delay > ((LySqlDataProvider) _provider).getDelay())
	            {
	                environment.logOut("Request delay exceeded = " + request_text);
	                Thread.dumpStack();
	            }
	        }
	    }
	    finally
	    {
	        if ((connection != null)
	        	&& connection.isReserved())
	        {
	            environment.logOut("[CONNECTION] connection " + connection + " was not released, freeing it.");
	            if (environment.isDebugLevel(LyEnvironment.DEBUG_TRANSACTION))
	                Thread.dumpStack();
	        }
	    }

        return true;
    } // end method deleteObjects

    /**
     * Methods for creation.*
     */
    /**
     * To create a new object.
     *
     * @param  values the values of the new object
     * @return true if the request was correctly sent to the external data provider.
     */
    public boolean createObject(Object object)
    {
        LyEnvironment environment = getEnvironment();
        // Set the response in running mode
        _response.setStatus(LyResponse.Status.RUNNING);

        if (environment.getPerfLogger() != null)
        {
            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
                PERF_SQL_CREATE_OBJECT, "class=" + _class.getName());
            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
                PERF_SQL_ENCODE_REQUEST, null);
        }

        // Create the SQL request
        String request_text = encodeCreateRequest(object);

        if (environment.getPerfLogger() != null)
            environment.getPerfLogger().addEndEvent(_provider.getApplication(),
                PERF_SQL_ENCODE_REQUEST, null);

        // Get a valid connection
        LySqlConnection connection = (LySqlConnection) _provider.getConnection();

        if (connection == null)
        {
            //      No available connection
            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
                environment.logOut("No avalaible connection.");

            ((LyRowObject) object).cancel();
            _dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);

            if (environment.getPerfLogger() != null)
                environment.getPerfLogger().addEndEvent(_provider.getApplication(),
                    PERF_SQL_CREATE_OBJECT, LyPerfLogger.ERROR_STRING);

            return false;
        }

        try
        {
	        if (environment.getPerfLogger() != null)
	            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
	                PERF_SQL_SEND_QUERY, null);
	        // Send the request to the connection
	        connection.setHookForUpdate(!_dataSource.isRollBack());

	        int nb_lines = connection.sendUpdateRequest(request_text);

	        if (environment.getPerfLogger() != null)
	            environment.getPerfLogger().addEndEvent(_provider.getApplication(), PERF_SQL_SEND_QUERY,
	                null);

	        if (nb_lines >= 0)
	        {

	            if (environment.getPerfLogger() != null)
	                environment.getPerfLogger().addStartEvent(_provider.getApplication(),
	                    PERF_SQL_DECODE_VALUES, null);

	            // The identifier and some values may be provided by the database.
	            completeObject(object, connection);

	            // Release the connection.
	            _provider.releaseConnection(connection);
	            connection = null;

	            ((LyRowObject) object).validate(null);

	            if (environment.getPerfLogger() != null)
	            {
	                environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                    PERF_SQL_DECODE_VALUES, null);

	                environment.getPerfLogger().addStartEvent(_provider.getApplication(),
	                    PERF_SQL_SEND_RESPONSE, null);
	            }

	            _dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, object, null);

	            if (environment.getPerfLogger() != null)
	                environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                    PERF_SQL_SEND_RESPONSE, null);
	        }
	        else
	        {
	            // An error occured while sending the request to the database
	            _response.setError(connection.getError());

	            // Release the connection.
	            if ((connection != null)
	            	&& connection.isReserved())
	            {
	                _provider.releaseConnection(connection);
	                connection = null;
	            }

	            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
	                environment.logOut("An error occured while sending the request \"" + request_text
	                    + "\" to the database.");

	            ((LyRowObject) object).cancel();
	            _dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);

	            if (environment.getPerfLogger() != null)
	                environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                    PERF_SQL_CREATE_OBJECT, LyPerfLogger.ERROR_STRING);

	            return false;
	        }

	        if (environment.getPerfLogger() != null)
	            environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                PERF_SQL_CREATE_OBJECT, null);
        }
        finally
        {
            if ((connection != null)
            	&& connection.isReserved())
            {
                environment.logOut("[CONNECTION] connection " + connection + " was not released, freeing it.");
                if (environment.isDebugLevel(LyEnvironment.DEBUG_TRANSACTION))
                    Thread.dumpStack();
            }
        }

        return true;
    } // end method createObject

    /**
     * Methods for modification.*
     */
    /**
     * To set an object.
     *
     * @param  object the object to set
     * @param  values the new values of the object
     * @return true if the request was correctly sent to the external data provider.
     */
    public boolean setObject(Object object)
    {
        LyEnvironment environment = getEnvironment();
        long          before = System.currentTimeMillis();

        if (environment.getPerfLogger() != null)
        {
            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
                PERF_SQL_SET_OBJECT, "class=" + _class.getName());
            environment.getPerfLogger().addStartEvent(_provider.getApplication(),
                PERF_SQL_ENCODE_REQUEST, null);
        }
        // Set the response in running mode
        _response.setStatus(LyResponse.Status.RUNNING);

        // Create the SQL request
        String request_text = encodeSetRequest(object);

        if (environment.getPerfLogger() != null)
            environment.getPerfLogger().addEndEvent(_provider.getApplication(),
                PERF_SQL_ENCODE_REQUEST, null);

        if (request_text != null)
        {
            // Get a valid connection
            LySqlConnection connection = (LySqlConnection) _provider.getConnection();

            if (connection == null)
            {
                //      No available connection
                if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
                    environment.logOut("No avalaible connection.");

                ((LyRowObject) object).cancel();
                _dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);

                if (environment.getPerfLogger() != null)
                    environment.getPerfLogger().addEndEvent(_provider.getApplication(),
                        PERF_SQL_SET_OBJECT, LyPerfLogger.ERROR_STRING);

                return false;
            }

            try
            {
	            // Send the request to the connection
	            if (environment.getPerfLogger() != null)
	                environment.getPerfLogger().addStartEvent(_provider.getApplication(),
	                    PERF_SQL_SEND_QUERY, null);

	            connection.setHookForUpdate(!_dataSource.isRollBack());

	            int nb_lines = connection.sendUpdateRequest(request_text);
	            int nb_lines_min = getUpdateMinimumLines(object);

	            if (environment.getPerfLogger() != null)
	                environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                    PERF_SQL_SEND_QUERY, null);

	            if (nb_lines >= nb_lines_min)
	            {
	                // Release the connection.
	                _provider.releaseConnection(connection);
	                connection = null;

	                if (environment.getPerfLogger() != null)
	                    environment.getPerfLogger().addStartEvent(_provider.getApplication(),
	                        PERF_SQL_DECODE_VALUES, null);

	                LyValueSet oldValues = ((LyRowObject) object).validate(null);

	                if (environment.getPerfLogger() != null)
	                {
	                    environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                        PERF_SQL_DECODE_VALUES, null);
	                    environment.getPerfLogger().addStartEvent(_provider.getApplication(),
	                        PERF_SQL_SEND_RESPONSE, null);
	                }

	                _dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, object,
	                    oldValues);

	                if (environment.getPerfLogger() != null)
	                    environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                        PERF_SQL_SEND_RESPONSE, null);
	            }
	            else
	            {
	                _response.setError(connection.getError());

	                // Release the connection.
	                if ((connection != null)
	                	&& connection.isReserved())
	                {
	                    _provider.releaseConnection(connection);
	                    connection = null;
	                }

	                // An error occured while sending the request to the database
	                if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
	                    environment.logOut("An error occured while sending the request \""
	                        + request_text + "\" to the database.");

	                ((LyRowObject) object).cancel();
	                _dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);

	                if (environment.getPerfLogger() != null)
	                    environment.getPerfLogger().addEndEvent(_provider.getApplication(),
	                        PERF_SQL_SET_OBJECT, LyPerfLogger.ERROR_STRING);

	                return false;
	            }
            }
            finally
            {
                if ((connection != null)
                	&& connection.isReserved())
                {
                    environment.logOut("[CONNECTION] connection " + connection + " was not released, freeing it.");
                    if (environment.isDebugLevel(LyEnvironment.DEBUG_TRANSACTION))
                        Thread.dumpStack();
                }
            }
        }
        else
            // The object has no modified values.
            _dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, object, null);

        if (environment.getPerfLogger() != null)
            environment.getPerfLogger().addEndEvent(_provider.getApplication(), PERF_SQL_SET_OBJECT,
                null);

        long after = System.currentTimeMillis();

        if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
                && environment.isDebugLevel(LyEnvironment.DEBUG_PERF))
        {
            long delay = after - before;

            environment.logOut("Delay = " + delay);

            if (delay > ((LySqlDataProvider) _provider).getDelay())
            {
                environment.logOut("Request delay exceeded = " + request_text);
                Thread.dumpStack();
            }
        }

        return true;
    } // end method setObject

    /**
     * Returns the minimum number of lines which should be updated when a the setObject()
     * method is called. If this minimum is not reached, the response sent to the datasource
     * will be in error.
     * Default implementation returns 0.
     *
     * @param object the object being set
     * @return the minimum number of lines which should be updated (0 by default)
     */
    protected int getUpdateMinimumLines(Object object)
    {
        return 0;
    }

    /**
     * To set a set of objects.
     *
     * @param  objects the objects to set
     * @param  values  the new values of the objects
     * @return true if the request was correctly sent to the external data provider.
     */
    public boolean setObjects(Vector<Object> objects)
    {
        LyEnvironment environment = getEnvironment();

        throw new LyUnsupportedOperationException("BF AppEngine : "
            + environment.getMessage("LY_ER_UNSUPPORTED_OPERATION", "LySqlRequestEncoder",
                "setObjects"));
    }

    /**
     * Methods for encoding requests.*
     */

    /**
     * Returns the table name of the specified field parameter. This method is used when we build a
     * join request for an union.
     */
    protected String getTableName(LyField field)
    {

        if (_joinPeerClasses == null)
            return null;

        Iterator<LyClass> it = _joinPeerClasses.iterator();
        LyClass           peerClass;
        Vector<LyField>   fields;

        while (it.hasNext())
        {
            peerClass = it.next();
            fields = new Vector<LyField>(peerClass.getFields());

            if (fields.contains(field))
                return getTableName(peerClass);
        }

        return null;
    }

    /**
     * Returns an alias for the specified field to encode a select request.
     */
    protected String getAlias(String tableName, LyField field)
    {
        return null;
    }

    /**
     * Encodes a select-type request.
     *
     * @param  encoder  an union encoder for join requests (may be null)
     * @param  objectId object identifier (may be null).
     * @param  fields   the fields to load (if null, load all)
     * @param  sort     indicates sorting criterion (may be null).
     * @param  filter   indicates filtering criterion (may be null).
     * @param  n        maximum number of objects to get (-1 if no limit).
     * @return the generated request.
     */
    protected String encodeSelectRequest(LyUnionRequestEncoder encoder, LyObjectId objectId,
        Vector<LyField> fields, LySort sort, LyFilter filter, int limit)
    {
        LyColumnField    field;
        StringBuffer     request;
        int              i, n;
        String           filterCondition, idCondition, classCondition, unionCondition, str;
        String[][]       relationSortConditions;
        boolean          and = false, needDistinct;
        LyFilter         classFilter = ((_classBinding == null) ? null : _classBinding.getFilter());
        Vector<LyFilter> unionFilters = null;
        LyEnvironment    environment = getEnvironment();

        if (classFilter != null)
            classFilter = classFilter.getPeerFilter();

        getRequestInfo().clearCurrentUsedClasses();

        if (_classBinding instanceof LyUnionBinding)
            unionFilters = getFilters((LyUnionBinding) _classBinding);

        request = new StringBuffer();

        needDistinct = isDistinctNeeded(filter) || isDistinctNeeded(classFilter);

        // Search for the fields to put in the request.
        if ((fields == null) || (fields.size() == 0))
        {
            _fields = new Vector<LyField>(_class.getFields());

            // Remove CLOB or BLOB
            if (needDistinct)
            {

                for (i = _fields.size() - 1; i >= 0; i--)
                {
                    field = (LyColumnField) _fields.elementAt(i);

                    if (isBlob(field) || isClob(field))
                        _fields.remove(field);
                }
            }
        }
        else
            _fields = fields;

        if (unionFilters != null)
        {

            for (i = 0; (!needDistinct) && (i < unionFilters.size()); i++)
                needDistinct = isDistinctNeeded(unionFilters.elementAt(i));
        }

        // Do not use distinct when we have blob/clob fields : FFT01219
        for (i = _fields.size() - 1; needDistinct && (i >= 0); i--)
        {
            field = (LyColumnField) _fields.elementAt(i);

            if (isBlob(field) || isClob(field))
                needDistinct = false;
        }

        LyInfoList<LyClassInfo> targetClasses = new LyInfoList<LyClassInfo>(false);
        getRequestInfo().getCurrentTask().addInfo("BEGIN encodeSelectRequest, isDistinctNeeded:"
            + needDistinct);

        // Retrieve the target classes for filters. If filters contain filter(s)
        // on relations, the target tables must be added to the 'FROM' clause.
        getTargetClasses(classFilter, targetClasses);
        getTargetClasses(filter, targetClasses);

        if (unionFilters != null)
        {

            // Check into union filters if some additional classes are required
            for (i = 0; i < unionFilters.size(); i++)
                getTargetClasses(unionFilters.elementAt(i), targetClasses);
        }

        // retrieve the target classes for the relation sort condition
        LyInfoList<LyClassInfo> sortClasses = new LyInfoList<LyClassInfo>(true);

        getTargetClasses(sort, sortClasses);

        LyInfoList<LyClassInfo> filteredSortClasses = new LyInfoList<LyClassInfo>(false);

        filteredSortClasses.addList(sortClasses);

        boolean hasMultiple = false;

        if (filteredSortClasses.getSize() != sortClasses.getSize())
            hasMultiple = true;

        if (!hasMultiple)
            targetClasses.addList(sortClasses);

        // If there is at least one target table, we must prefix the 'SELECT'
        // clause parameters with the name of the current table.
        boolean writeTableName = (targetClasses.getSize() > 0) && (_joinPeerClasses == null);
        String  tableName = (writeTableName ? getTableName(_class) : null);

        getRequestInfo().addUsedClass(tableName);

        // now compute the table aliases
        // A First pass to compute all requested aliases in the sqlRequestInfo
        // ** analysis begins **
        LyClassInfo   classInfo;
        String        targetTableName;
        String        tableAlias;
        List<LyClass> peerClasses = getPeerClasses(_class);

        for (Iterator<LyClass> it = peerClasses.iterator(); it.hasNext();)
        {
            targetTableName = getTableName(it.next());
            tableAlias = computeTableAlias(targetTableName);

            // add the table alias for the current logical class (null)
            addTableAlias(null, targetTableName, tableAlias);
        }

        // Now we must encode the conditions. (still analysis)
        getRequestInfo().getCurrentTask().addInfo("[BEGIN Analysis] encodeSelectRequest");
        encodeCondition(filter, tableName, null); // filterCondition
        encodeCondition(objectId); // idCondition
        encodeCondition(encoder, null); // unionCondition
        encodeCondition(classFilter, tableName, null); // classCondition

        Map<LyField, String> aliases = new HashMap<LyField, String>();

        encodeRelationSortCondition(sort, hasMultiple, aliases); // relationSortConditions
        // ** analysis ends ** : lock and do it for real **
        getRequestInfo().getCurrentTask().addInfo("[END Analysis] encodeSelectRequest");

        String[][] relationSortConditionsTemp = encodeRelationSortCondition(sort, hasMultiple,
                aliases);

        if ((_joinPeerClasses == null) && (relationSortConditionsTemp != null)
                && (relationSortConditionsTemp.length > 0))
            // W4 FFT00757 : we shall use aliases because of the left join
            getRequestInfo().setForceUseAlias(true);

        getRequestInfo().setLockAliasCount(true);

        getRequestInfo().getCurrentTask().addInfo(
            "[BEGIN alias locked -> Encode] encodeSelectRequest");
        getRequestInfo().getCurrentTask().addSql(
            "--- [BEGIN alias locked -> Encode] encodeSelectRequest ---");
        str = beginEncodeSelectRequest(limit, needDistinct);
        request.append(str);
        getRequestInfo().getCurrentTask().addSql(str);


        // Now we must encode the conditions. ** same as analysis but store the results **
        filterCondition = encodeCondition(filter, tableName, null);
        idCondition = encodeCondition(objectId);
        unionCondition = encodeCondition(encoder, null);
        classCondition = encodeCondition(classFilter, tableName, null);

        aliases = new HashMap<LyField, String>();

        relationSortConditions = encodeRelationSortCondition(sort, hasMultiple, aliases);


        /* // XXX: Commented for junit tests to pass
        if ((!writeTableName)
        &&(relationSortConditions != null)
        && (relationSortConditions.length > 0))
        {
        writeTableName = true;
        tableName = getTableName(_class);
        }
         */

        getRequestInfo().setLockAliasCount(false);

        // FFT02616 : check if actually being used for performance
        HashMap<String, String> aliasMapUsed = getRequestInfo().getAllAliases();
        for (Iterator<String> iter = aliasMapUsed.keySet().iterator(); iter.hasNext();)
        {
            tableAlias = iter.next();

            if (aliasMapUsed.size() > 0)
            {
                boolean used= false;

                if (!used && (idCondition != null) && (idCondition.indexOf(tableAlias)>=0))
                    used = true;
                if (!used && (classCondition != null) && (classCondition.indexOf(tableAlias)>=0))
                    used = true;
                if (!used && (filterCondition != null) && (filterCondition.indexOf(tableAlias)>=0))
                    used = true;
                if (!used && (unionCondition != null) && (unionCondition.indexOf(tableAlias)>=0))
                    used = true;
                for (int i1=0; (!used) && (relationSortConditions != null) && i1<relationSortConditions.length; i1++)
                {
                    if (relationSortConditions[i1][1].indexOf(tableAlias)>0)
                        used = true;
                }
                if (!used)
                    getRequestInfo().skipUnusedAliases(tableAlias);
            }
        }

        // check if we have an alias forced on tablename
        if (!writeTableName)
        {
            String t = getTableName(_class);
            String alias = getTableAlias(null, t);

            if ((t != null) && (!t.equals(alias)))
            {
                tableName = t;
                writeTableName = true;
            }
        }

        n = _fields.size();

        HashMap<LyField, String> fieldAliases = new HashMap<LyField, String>();
        Vector<String> existingFields = new Vector<String>(n);

        for (i = 0; i < n; i++)
        {
            field = (LyColumnField) _fields.elementAt(i);

            if (i != 0)
                request.append(", ");

            if (writeTableName)
                str = tableName;
            else
                str = getTableName(field);

            existingFields.add(field.getId());

            if (
                !appendBlobFieldId(field, request,
                        ((str == null) ? null : getTableAlias(null, str))))
            {
                String id = "";

                if (str != null)
                    id = QUOTE_CHAR_LEFT + getTableAlias(null, str) + QUOTE_CHAR_RIGHT + ".";

                id = id + QUOTE_CHAR_LEFT + field.getId() + QUOTE_CHAR_RIGHT;

                if (needDistinct && (field.getValueClass() == java.util.Date.class))
                {

                    if (getRequestInfo().isSQLFormatDateInOrderByClauseEmpty())
                    {
                        encodeSort(sort, tableName, fieldAliases); // dummy, just to detect date formats

                        if (getRequestInfo().isSQLFormatDateInOrderByClauseEmpty())
                            getRequestInfo().setSQLFormatDateInOrderByClause("NOTHING", "NOTHING");
                    }

                    if (getRequestInfo().getSQLFormatDateInOrderByClause(id) != null)
                        id = getRequestInfo().getSQLFormatDateInOrderByClause(id);
                }

                request.append(id);
            }

            getRequestInfo().addUsedClass(str);

            str = getAlias(str, field);

            if (str != null)
            {
                request.append(" " + str);
                fieldAliases.put(field, str);
                // add  also the xxxxxx
            }
        }

        String otherFields = completeSelectedFields(existingFields, sort, needDistinct, tableName);

        if (otherFields != null)
        {
            request.append(", ");
            request.append(otherFields);
        }

        // add the column for the relation sort if needed
        encodeNameRelationColumns(request, sort);

        request.append(" from ");

        // IN fact all used aliases -- table names should be added in the from clause
        // Store what we have from former code below (compatibility) then complete afterwards.
        HashMap<String, String> foundAliases = new HashMap<String, String>();
        Vector<String>          foundTableNames = new Vector<String>();
        boolean                 optimize = true; //getEnvironment().testResource("LY_OPTIMIZE_SQL");
        ArrayList<String>       usedTablesNames = getRequestInfo().getCurrentUsedClasses();
        boolean                 firstElementAdded = false;

        if (_joinPeerClasses == null)
        {
            targetTableName = getTableName(_class);
            tableAlias = getTableAlias(null, targetTableName);

            if (!getRequestInfo().isIgnoreInSelectFrom(tableName, tableAlias))
            {
                request.append(QUOTE_CHAR_LEFT).append(targetTableName).append(QUOTE_CHAR_RIGHT);
                firstElementAdded = true;

                if (!tableAlias.equals(targetTableName))
                {
                    request.append(" ").append(QUOTE_CHAR_LEFT).append(tableAlias).append(
                        QUOTE_CHAR_RIGHT);
                    foundAliases.put(tableAlias, targetTableName);
                }

                foundTableNames.addElement(targetTableName);
            }

            if ((relationSortConditions != null) && (relationSortConditions.length > 0))
            {
                String previousAlias = null;

                for (int k = 0; k < relationSortConditions.length; k++)
                {
                    String targetTableNameLJ = relationSortConditions[k][0];

                    if (hasMultiple)
                        tableAlias = relationSortConditions[k][2];
                    else
                    {

                        if ((_classBinding != null) && (_classBinding.getBinding() != null)
                                && _classBinding.getBinding().equals(targetTableNameLJ))
                            tableAlias = getTableAliasForSelfTargetRelationClass(null, null, null,
                                    targetTableNameLJ);
                        else
                            tableAlias = getTableAlias(null, targetTableNameLJ);
                    }

                    if ((tableAlias != null) && !tableAlias.equals(targetTableNameLJ)
                            && tableAlias.equals(previousAlias))
                    {
                        request.append(" and ");
                        request.append(relationSortConditions[k][1]);
                    }
                    else
                    {
                        request.append(" left join ");
                        request.append(QUOTE_CHAR_LEFT).append(targetTableNameLJ).append(
                            QUOTE_CHAR_RIGHT);

                        if (!tableAlias.equals(targetTableNameLJ))
                        {
                            request.append(" ").append(tableAlias);
                            foundAliases.put(tableAlias, targetTableNameLJ);
                            previousAlias = tableAlias;
                        }
                        else
                            previousAlias = null;

                        foundTableNames.addElement(targetTableNameLJ);
                        request.append(" on ");
                        request.append(relationSortConditions[k][1]);
                    }
                }
            }
        }
        else
        {
            boolean first = true;
            LyClass peerClass;

            for (Iterator<LyClass> it = _joinPeerClasses.iterator(); it.hasNext();)
            {
                peerClass = it.next();
                targetTableName = getTableName(peerClass);
                tableAlias = getTableAlias(null, targetTableName);

                if (foundTableNames.contains(targetTableName) && tableAlias.equals(targetTableName))
                    continue;

                if (optimize && !usedTablesNames.contains(targetTableName))
                    continue;

                if (!foundTableNames.contains(targetTableName))
                    foundTableNames.addElement(targetTableName);

                if (!getRequestInfo().isIgnoreInSelectFrom(targetTableName, tableAlias))
                {

                    if (first)
                    {
                        request.append(QUOTE_CHAR_LEFT).append(targetTableName).append(
                            QUOTE_CHAR_RIGHT);

                        if (!tableAlias.equals(targetTableName))
                        {
                            request.append(" ").append(tableAlias);
                            foundAliases.put(tableAlias, targetTableName);
                        }

                        first = false;
                    }
                    else
                    {
                        request.append(", ").append(QUOTE_CHAR_LEFT).append(targetTableName).append(
                            QUOTE_CHAR_RIGHT);

                        if (!tableAlias.equals(targetTableName))
                        {
                            request.append(" ").append(tableAlias);
                            foundAliases.put(tableAlias, targetTableName);
                        }
                    }

                    firstElementAdded = true;
                }

                if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
                    environment.logOut(" > add target : " + targetTableName + " alias : "
                        + tableAlias);

                if ((relationSortConditions != null) && (relationSortConditions.length > 0))
                {

                    for (int k = 0; k < relationSortConditions.length; k++)
                    {

                        // check if it applies on targetTableName
                        if ((!targetTableName.equals(relationSortConditions[k][3]))
                                || (!tableAlias.equals(relationSortConditions[k][4])))
                            continue;

                        String targetTableNameLJ = relationSortConditions[k][0];

                        if (hasMultiple)
                            tableAlias = relationSortConditions[k][2];
                        else
                            tableAlias = getTableAlias(null, targetTableNameLJ);

                        request.append(" left join ");
                        request.append(QUOTE_CHAR_LEFT).append(targetTableNameLJ).append(
                            QUOTE_CHAR_RIGHT);

                        if (!tableAlias.equals(targetTableNameLJ))
                        {
                            request.append(" ").append(tableAlias);
                            foundAliases.put(tableAlias, targetTableNameLJ);
                        }

                        foundTableNames.addElement(targetTableNameLJ);
                        request.append(" on ");
                        request.append(relationSortConditions[k][1]);
                    }
                }

            }
        }

        HashMap<String, String> allAliases = getRequestInfo().getAllAliases();

        for (i = 0; i < targetClasses.getSize(); i++)
        {
            classInfo = targetClasses.getInfo(i);
            targetTableName = classInfo.getBinding().getBinding();

            if ((targetTableName == null) && (classInfo.getBinding() instanceof LyUnionBinding))
            {
                LyClass peerClass = classInfo.getPeerClass();

                peerClasses = getPeerClasses(peerClass);

                for (Iterator<LyClass> it = peerClasses.iterator(); it.hasNext();)
                {
                    peerClass = it.next();
                    targetTableName = getTableName(peerClass);

                    if (!targetTableName.equals(tableName))
                    {

                        if (optimize && !usedTablesNames.contains(targetTableName))
                            continue;

                        tableAlias = getTableAlias(classInfo, targetTableName);

                        // avoid duplication of table aliases
                        if (foundTableNames.contains(tableAlias)
                                || (foundTableNames.contains(targetTableName)
                                    && tableAlias.equals(targetTableName)))
                            continue;
                        if (allAliases.containsValue(targetTableName))
                        {
                            if (!foundTableNames.contains(targetTableName))
                                foundTableNames.addElement(targetTableName);
                            continue;
                        }

                        if (!getRequestInfo().isIgnoreInSelectFrom(targetTableName, tableAlias))
                        {

                            if (!foundTableNames.contains(targetTableName))
                            {
                                foundTableNames.addElement(targetTableName);
                                foundTableNames.addElement(tableAlias);
                            }

                            if (firstElementAdded)
                                request.append(", ");
                            else
                                request.append(" ");

                            request.append(QUOTE_CHAR_LEFT).append(targetTableName).append(
                                QUOTE_CHAR_RIGHT);
                            firstElementAdded = true;

                            if (!tableAlias.equals(targetTableName))
                            {
                                request.append(" ").append(tableAlias);
                                foundAliases.put(tableAlias, targetTableName);
                            }

                            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
                                environment.logOut(" > add target : " + targetTableName
                                    + " alias : " + tableAlias);
                        }
                    }
                }
            }
            else if (!targetTableName.equals(tableName))
            {
                tableAlias = getTableAlias(classInfo, targetTableName);

                // avoid duplication of table aliases
                if (foundTableNames.contains(tableAlias)
                        || (foundTableNames.contains(targetTableName)
                            && tableAlias.equals(targetTableName))
                        || (foundAliases.get(tableAlias) != null))
                    continue;

                if (optimize && !usedTablesNames.contains(targetTableName))
                    continue;

                if (!getRequestInfo().isIgnoreInSelectFrom(targetTableName, tableAlias))
                {

                    if (!foundTableNames.contains(targetTableName))
                        foundTableNames.addElement(targetTableName);

                    if (firstElementAdded)
                        request.append(", ");
                    else
                        request.append(" ");

                    request.append(QUOTE_CHAR_LEFT).append(targetTableName).append(
                        QUOTE_CHAR_RIGHT);
                    firstElementAdded = true;

                    if (!tableAlias.equals(targetTableName))
                    {
                        request.append(" ").append(tableAlias);
                        foundAliases.put(tableAlias, targetTableName);
                    }

                    if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
                        environment.logOut(" > add target : " + targetTableName + " alias : "
                            + tableAlias);
                }
            }
        }

        for (i = 0; i < usedTablesNames.size(); i++)
        {
            targetTableName = usedTablesNames.get(i);

            if (targetTableName == null)
                continue;

            if (!targetTableName.equals(tableName))
            {

                if (foundTableNames.contains(targetTableName))
                    continue;

                foundTableNames.addElement(targetTableName);

                if (firstElementAdded)
                    request.append(", ");
                else
                    request.append(" ");

                request.append(QUOTE_CHAR_LEFT).append(targetTableName).append(QUOTE_CHAR_RIGHT);
                firstElementAdded = true;

                if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
                    environment.logOut(" > add target : " + targetTableName);
            }
        }

        // here we hare, we shall also add all remaining aliases
        HashMap<String, String> aliasMap = getRequestInfo().getAllAliases();

        for (Iterator<String> iter = aliasMap.keySet().iterator(); iter.hasNext();)
        {
            tableAlias = iter.next();

            if (foundAliases.get(tableAlias) == null)
            {
                targetTableName = aliasMap.get(tableAlias);

                if (!getRequestInfo().isIgnoreInSelectFrom(targetTableName, tableAlias))
                {

                    if (!foundTableNames.contains(targetTableName))
                        foundTableNames.addElement(targetTableName);

                    if (firstElementAdded)
                        request.append(", ");
                    else
                        request.append(" ");

                    request.append(QUOTE_CHAR_LEFT).append(targetTableName).append(
                        QUOTE_CHAR_RIGHT);
                    firstElementAdded = true;

                    request.append(" ").append(tableAlias);
                }
            }
        }

        if (usesUnionSelect() && (getRequestInfo().getUnionSelectInfos(0) != null))
        {
            int      index = 0;
            Object[] unionSelectInfo = getRequestInfo().getUnionSelectInfos(index);
            String   childFilter;
            String   beginRequest = request.toString();
            int      k = beginRequest.lastIndexOf("select ");

            if (k > 0)
                beginRequest = beginRequest.substring(k);

            //** Force create request
            request = new StringBuffer();
            str = beginEncodeSelectRequest(-1, needDistinct); // do *NOT* use limit here
            request.append(str);

            n = _fields.size();

            for (i = 0; i < n; i++)
            {
                field = (LyColumnField) _fields.elementAt(i);

                if (i != 0)
                    request.append(", ");

                if (writeTableName)
                    str = tableName;
                else
                    str = getTableName(field);

                if (
                    !appendBlobFieldId(field, request,
                            ((str == null) ? null : getTableAlias(null, str))))
                {
                    str = fieldAliases.get(field);

                    if (str != null)
                        request.append(QUOTE_CHAR_LEFT).append(str).append(QUOTE_CHAR_RIGHT);
                    else
                        request.append(QUOTE_CHAR_LEFT).append(field.getId()).append(
                            QUOTE_CHAR_RIGHT);
                }
            }

            request.append(" from (");

            while (unionSelectInfo != null)
            {

                if (index > 0)
                    request.append(" UNION ");

                request.append(beginRequest);

                @SuppressWarnings("unchecked")
                ArrayList<String> childTargetClasses = (ArrayList<String>) unionSelectInfo[0];

                childFilter = (String) unionSelectInfo[1];

                //** add extra classes
                for (i = 0; i < childTargetClasses.size(); i++)
                {
                    targetTableName = childTargetClasses.get(i);

                    if (!foundTableNames.contains(targetTableName))
                    {
                        tableAlias = getTableAlias(null, targetTableName);

                        // avoid duplication of table aliases
                        if (!foundTableNames.contains(tableAlias))
                        {

                            if (!getRequestInfo().isIgnoreInSelectFrom(targetTableName, tableAlias))
                            {
                                request.append(", ").append(QUOTE_CHAR_LEFT).append(targetTableName)
                                .append(QUOTE_CHAR_RIGHT);

                                if (!tableAlias.equals(targetTableName))
                                    request.append(" ").append(tableAlias);
                            }
                        }
                    }
                }
                // now add filter condition
                and = false;

                if (idCondition != null)
                {
                    request.append(" where (");
                    request.append(LyString.replace(idCondition, UNION_SELECT_FILTER, childFilter));
                    request.append(")");
                    and = true;
                }

                if (classCondition != null)
                {

                    if (and)
                        request.append(" and ");
                    else
                        request.append(" where ");

                    request.append(LyString.replace(classCondition, UNION_SELECT_FILTER,
                            childFilter));
                    and = true;
                }

                if (filterCondition != null)
                {

                    if (and)
                        request.append(" and ");
                    else
                        request.append(" where ");

                    request.append(LyString.replace(filterCondition, UNION_SELECT_FILTER,
                            childFilter));
                    and = true;
                }

                if (unionCondition != null)
                {

                    if (and)
                        request.append(" and ");
                    else
                        request.append(" where ");

                    request.append(LyString.replace(unionCondition, UNION_SELECT_FILTER,
                            childFilter));
                    and = true;
                }

                index++;
                unionSelectInfo = getRequestInfo().getUnionSelectInfos(index);
            }

            request.append(") "); // end of the UNION select

            if (limit > 0)
            {
                String limitStr = encodeLimitCondition(limit);

                if (limitStr != null)
                    request.append(" where ").append(limitStr);
            }

            fieldAliases.putAll(aliases);
            str = endEncodeSelectRequest(sort, -1, null, fieldAliases);

            if (str != null)
                request.append(str);
            else
            {

                if (sort != null)
                    _response.setSorted(false);

                if (limit > 0)
                    _response.setCounted(false);
            }

            getRequestInfo().clearCurrentUsedClasses();
        }
        else
        {

            // DO AS USUAL without UNION select
            if (idCondition != null)
            {
                request.append(" where (");
                request.append(idCondition);
                request.append(")");
                and = true;
            }

            if (classCondition != null)
            {

                if (and)
                    request.append(" and ");
                else
                    request.append(" where ");

                request.append(classCondition);
                and = true;
            }

            if (filterCondition != null)
            {

                if (and)
                    request.append(" and ");
                else
                    request.append(" where ");

                request.append(filterCondition);
                and = true;
            }

            if (unionCondition != null)
            {

                if (and)
                    request.append(" and ");
                else
                    request.append(" where ");

                request.append(unionCondition);
                and = true;
            }

            str = endEncodeSelectRequest(sort, limit, tableName, aliases);

            if (str != null)
                request.append(str);
            else
            {

                if (sort != null)
                    _response.setSorted(false);

                if (limit > 0)
                    _response.setCounted(false);
            }

            getRequestInfo().clearCurrentUsedClasses();
        }

        getRequestInfo().getCurrentTask().addInfo("END encodeSelectRequest");
        getRequestInfo().getCurrentTask().addSql(request.toString());

        return request.toString();
    } // end method encodeSelectRequest


    protected boolean needCompleteSelectedField(LySort sort)
    {
    	if (sort == null)
    		return false;
    	int             i, n;
        LySortCriterion criterion;
        String          addedField;

        n = sort.getCriterionCount();
        LyFieldInfo fieldInfo;

        for (i = 0; i < n; i++)
        {
        	criterion = sort.getCriterion(i);
        	fieldInfo = criterion.getFieldInfo();

        	if (!(fieldInfo instanceof LyTextFieldInfo))
        		continue;

        	if (((LyTextFieldInfo)fieldInfo).isMultilingual())
        		return true;
        }

        return false;
    }

    /**
     * Used to add a list of fields to a select request. May be overridden for specific encoders.
     *
     * @param  sort         The sort of the request
     * @param  needDistinct Indicates if a need distinct clause is in the select request.
     * @param  tableName    The table name
     * @return A string with a list of comma separated fields to add to the select request or null
     *         if there is no fields to add.
     */
    public String completeSelectedFields(Vector<String> existingFields, LySort sort, boolean needDistinct, String tableName)
    {
    	boolean needComplete = needCompleteSelectedField(sort);

        if (!needComplete && (!needDistinct || (sort == null)))
            return null;

        StringBuffer    result = null;
        int             i, n;
        LySortCriterion criterion;
        String          addedField;
        LyFieldInfo     fieldInfo;
        n = sort.getCriterionCount();

        for (i = 0; i < n; i++)
        {
            criterion = sort.getCriterion(i);
            fieldInfo = criterion.getFieldInfo();
            // just add multilingual field if !needDistinct
            if (!needDistinct)
            {
            	if (!(fieldInfo instanceof LyTextFieldInfo))
            		continue;
            	if (!((LyTextFieldInfo)fieldInfo).isMultilingual())
            		continue;
            }
            addedField = completeSelectedFields(existingFields, criterion, tableName);

            if (addedField != null)
            {

                if (result == null)
                    result = new StringBuffer("");
                else
                    result.append(", ");

                result.append(addedField);
            }
        }

        if (result != null)
            return result.toString();
        else
            return null;
    } // end method completeSelectedFields

    /**
     * Used to add a list of fields to a select request. May be ovverriden for specific encoders.
     *
     * @param  criterion The sort criterion of the request
     * @param  tableName The table name
     * @return A string with a list of comma separated fields to add to the select request or null
     *         if there is no fields to add.
     */
    public String completeSelectedFields(Vector<String> existingFields, LySortCriterion criterion, String tableName)
    {

        if (criterion == null)
            return null;

        LyFieldInfo fieldInfo = criterion.getFieldInfo();

        if (fieldInfo.getBinding() == null)
            return null;

        // Case of sort on a relation field info
        if (fieldInfo instanceof LyRelationFieldInfo)
        {
            // Get the name fields
            LyRelationFieldInfo     relation = (LyRelationFieldInfo) fieldInfo;
            LyInfoList<LyFieldInfo> fields = getNameFieldInfos(relation);
            LyClass                 peerClass;
            LySqlRequestEncoder     encoder;
            StringBuffer            result = null;
            String                  relatedTableName;
            String                  condition;
            LySortCriterion         relatedCriterion;

            if (fields == null)
                return null;

            // Get the 1st target class
            LyClassInfo classInfo = relation.getClassInfo(0);

            for (int i = 0; i < fields.getSize(); i++)
            {
                fieldInfo = fields.getInfo(i);

                // Get the table name related to this field
                peerClass = getPeerClass(fieldInfo, classInfo.getBinding());
                relatedTableName = peerClass.getName();

                // Get encoder for the this field
                encoder = getRequestEncoder(getDataSource(peerClass), false);
                encoder.setClassBinding(classInfo.getBinding());

                // Encode condition
                relatedCriterion = new LySortCriterion(fieldInfo, criterion.isIncreasing());
                condition = encoder.completeSelectedFields(existingFields, relatedCriterion, relatedTableName);

                if (condition != null)
                {

                    if (result == null)
                        result = new StringBuffer();
                    else
                        result.append(", ");

                    result.append(condition);
                }
            }

            if (result != null)
                return result.toString();
            else
                return null;
        }
        else if ((fieldInfo instanceof LyTextFieldInfo)
                && ((LyTextFieldInfo) fieldInfo).isMultilingual())
        {
            LyField       field = _classBinding.getPeerField(fieldInfo);
            LyApplication application = _provider.getApplication();
            LyDataSession currentSession = application.getCurrentSession();
            String        lang = currentSession.getEnvironment().getLanguage();

            return getSelectedLangField(lang, field, tableName, true);
        }

        if (!supportCaseSensitive() || !(fieldInfo instanceof LyTextFieldInfo))
            return null;

        LyField[]    peerFields = _classBinding.getPeerFields(fieldInfo);
        String       id;
        StringBuffer result = null;
        int          i, n = ((peerFields == null) ? 0 : peerFields.length);

        for (i = 0; i < n; i++)
        {
            id = peerFields[i].getId();

            if ((existingFields != null)
                    && existingFields.contains(id))
                continue;

            if (_class.getField(id) == null)
                continue;

            if (existingFields != null)
                existingFields.add(id);

            if (tableName == null)
                tableName = getTableName(peerFields[i]);

            if (tableName != null)
                id = QUOTE_CHAR_LEFT + getTableAlias(null, tableName) + QUOTE_CHAR_RIGHT + "."
                    + QUOTE_CHAR_LEFT + id + QUOTE_CHAR_RIGHT;
            else
                id = QUOTE_CHAR_LEFT + id + QUOTE_CHAR_RIGHT;

            if (result == null)
                result = new StringBuffer();
            else
                result.append(", ");

            result.append(id);
        }

        return ((result == null) ? null : result.toString());
    } // end method completeSelectedFields

    /**
     * Look for filters inside a LyUnionBinding.
     *
     * @param  unionBinding current binding
     * @return the filter list if any is found, null otherwise
     */
    public Vector<LyFilter> getFilters(LyUnionBinding unionBinding)
    {
        Vector<LyFilter> v, result = null;
        LyFilter         filter;
        int              i;
        LyClassBinding   binding1 = unionBinding.getClassBinding1();
        LyClassBinding   binding2 = unionBinding.getClassBinding2();

        if (binding1 instanceof LyUnionBinding)
        {
            // Recurse into sub-unionBindings
            v = getFilters((LyUnionBinding) binding1);

            if (v != null)
            {

                for (i = 0; i < v.size(); i++)
                {

                    if (result == null)
                        result = new Vector<LyFilter>();

                    result.addElement(v.elementAt(i));
                }
            }
        }
        else
        {
            filter = binding1.getFilter();

            if (filter != null)
            {

                if (result == null)
                    result = new Vector<LyFilter>();

                result.addElement(filter);
            }
        }

        // *** Same for second binding
        if (binding2 instanceof LyUnionBinding)
        {
            // Recurse into sub-unionBindings
            v = getFilters((LyUnionBinding) binding2);

            if (v != null)
            {

                for (i = 0; i < v.size(); i++)
                {

                    if (result == null)
                        result = new Vector<LyFilter>();

                    result.addElement(v.elementAt(i));
                }
            }
        }
        else
        {
            filter = binding2.getFilter();

            if (filter != null)
            {

                if (result == null)
                    result = new Vector<LyFilter>();

                result.addElement(filter);
            }
        }

        return result;
    } // end method getFilters

    /**
     * Search if the distinct clause is needed for the specified filter.
     *
     * @param  filter the filter of the the request
     * @return true if the encoding request may have a distinct clause, else false.
     */
    protected boolean isDistinctNeeded(LyFilter filter)
    {

        if (filter == null)
            return false;

        int             i, n;
        LyFilterElement element;
        boolean         result = false;

        n = filter.getSize();

        if (n == 0)
            return result;

        for (i = 0; (i < n) && (!result); i++)
        {
            element = filter.getElement(i);

            if (element instanceof LyFilter)
                result = isDistinctNeeded((LyFilter) element);
            else if (element instanceof LyFilterExpression)
            {
                LyFieldInfo fieldInfo = ((LyFilterExpression) element).getFieldInfo();

                if ((fieldInfo instanceof LyRelationFieldInfo)
                        && (((LyRelationFieldInfo) fieldInfo).getDaemon() != null))
                    result = true; // FFT02614
            }
            else if (element instanceof LyFilterRelation)
                result = true;
            else if (element instanceof LyJoinFilter)
                result = true;
        }

        return result;
    }

    /**
     * Computes the target classes for the select count request.
     *
     * @param classFilter class filter element to compute.
     * @param filter      current filter element to compute.
     * @param classInfos  list of class infos filled by this method.
     */
    protected void getCountTargetClasses(LyFilterElement classFilter, LyFilterElement filter,
        LyInfoList<LyClassInfo> classInfos)
    {
        // Retrieve the target classes for filters. If filters contain filter(s)
        // on relations, the target tables must be added to the 'FROM' clause.
        getTargetClasses(classFilter, classInfos);
        getTargetClasses(filter, classInfos);
    }

    /**
     * Computes the target classes of a filter.
     *
     * @param element    filter element to compute. If instance of relation filter a call to
     *                   getPossibleClasses() method is performed to retrieve possible classes for
     *                   this filter.
     * @param classInfos list of class infos filled by this method.
     */
    protected void getTargetClasses(LyFilterElement element, LyInfoList<LyClassInfo> classInfos)
    {
        getTargetClasses(element, classInfos, true);
    }

    /**
     * Computes the target classes of a filter.
     *
     * @param element        filter element to compute. If instance of relation filter a call to
     *                       getPossibleClasses() method is performed to retrieve possible classes
     *                       for this filter.
     * @param classInfos     list of class infos filled by this method.
     * @param useJoinClasses indicates if join classes shall be added to the list
     */
    protected void getTargetClasses(LyFilterElement element, LyInfoList<LyClassInfo> classInfos,
        boolean useJoinClasses)
    {

        if (element == null)
            return;

        int i, n;

        // If filter instanceof LyFilter, browse its sub-elements and call
        // this method recursively
        if (element instanceof LyFilter)
        {
            LyFilter filter = (LyFilter) element;

            n = filter.getSize();

            if (n == 0)
                return;

            for (i = 0; i < n; i++)
                getTargetClasses(filter.getElement(i), classInfos, useJoinClasses);
        }
        // If filter instance of LyFilterExpression and field of the expression
        // is a relation with a join daemon, add the join class
        else if (element instanceof LyFilterExpression)
        {
            LyFieldInfo fieldInfo = ((LyFilterExpression) element).getFieldInfo();

            if ((useJoinClasses) && (fieldInfo instanceof LyRelationFieldInfo))
            {
                LyRelationFieldInfo relation = (LyRelationFieldInfo) fieldInfo;
                LyInfoDaemon        daemon = relation.getDaemon();

                // If relation has a join daemon, add the join class to the list
                if ((daemon != null) && (daemon instanceof LyJoinDaemon))
                {
                    LyJoinDaemon joinDaemon = (LyJoinDaemon) daemon;

                    // Will be encoded with a select XXX where (select count(*) XXX = 0)
                    if (((LyFilterExpression) element).getValue() != null)
                    {
                        LyClassInfo joinClass = joinDaemon.getJoinClass();

                        classInfos.addInfo(joinClass);
                    }
                }
            }
        }
        // If filter instanceof LyFilterRelation, retrieve possible classes
        else if (element instanceof LyFilterRelation)
        {
            LyFilterRelation        filterRelation = (LyFilterRelation) element;
            LyRelationFieldInfo     relation = filterRelation.getRelationFieldInfo();
            LyInfoDaemon            daemon = relation.getDaemon();
            LyInfoList<LyClassInfo> relationClasses = filterRelation.getPossibleClasses(false);
            LyClassInfo             classInfo;

            // The join classes are always used for filter relations,
            // therefore the useJoinClasses boolean is no more used.
            if ((daemon != null) && (daemon instanceof LyJoinDaemon))
            {
                LyJoinDaemon joinDaemon = (LyJoinDaemon) daemon;
                LyClassInfo  joinClass = joinDaemon.getJoinClass();

                classInfos.addInfo(joinClass);
            }

            // add only classes that have a binding defined (abstract or concrete)
            LyInfoList<LyClassInfo> classes = new LyInfoList<LyClassInfo>();

            n = relationClasses.getSize();

            for (i = 0; i < n; i++)
            {
                classInfo = relationClasses.getInfo(i);

                if (classInfo.getBinding() != null)
                    classes.addInfo(classInfo);
                else if (classInfo.isAbstract())
                {
                    LyInfoList<LyClassInfo> subClasses = classInfo.getSubClasses(true);
                    LyClassInfo             subClass;

                    for (int j = 0; j < subClasses.getSize(); j++)
                    {
                        subClass = subClasses.getInfo(j);

                        if (!subClass.isAbstract())
                            classes.addInfo(subClass);
                    }
                }
            }

            LyClass      peerClass, peerClass2;
            LyDataSource dataSource;

            n = classes.getSize();

            for (i = 0; i < n; i++)
            {
                classInfo = classes.getInfo(i);
                peerClass = classInfo.getBinding().getPeerClass();

                if (peerClass != null)
                {
                    dataSource = getDataSource(peerClass);

                    if (_provider != dataSource.getDataProvider())
                    {
                        ArrayList<LyClass> c = getPeerClasses(peerClass);

                        for (int p = 0; p < c.size(); p++)
                        {
                            peerClass2 = c.get(p);
                            dataSource = getDataSource(peerClass2);

                            if (_provider == dataSource.getDataProvider())
                                classInfos.addInfo(classInfo);
                        }

                        continue;
                    }
                }

                classInfos.addInfo(classInfo);
            }

            getTargetClasses(filterRelation.getFilter(), classInfos, useJoinClasses);
        }
        // If filter instanceof LyJoinFilter, retrieve possible classes
        else if (element instanceof LyJoinFilter)
        {
            LyJoinFilter join = (LyJoinFilter) element;
            LyClassInfo  sourceClassInfo = join.getSourceClassInfo();
            LyFilter     filter = join.getSourceFilter();

            classInfos.addInfo(sourceClassInfo);
            getTargetClasses(filter, classInfos, useJoinClasses);
        }
    } // end method getTargetClasses

    /**
     * Computes the target classes of a sort : search for criterions on relations that may generate
     * a join clause.
     *
     * @param sort       current sort of the select request.
     * @param classInfos list of class infos filled by this method.
     */
    protected void getTargetClasses(LySort sort, LyInfoList<LyClassInfo> classInfos)
    {

        if (sort == null)
            return;

        LyInfoList<LyRelationFieldInfo> relations = getNameRelationSort(sort);

        if (relations == null)
            return;

        LyRelationFieldInfo relation;

        for (int i = 0; i < relations.getSize(); i++)
        {
            relation = relations.getInfo(i);
            classInfos.addList(relation.getClassInfos());
        }
    }

    /**
     */
    protected String beginEncodeSelectRequest(int limit, boolean distinct)
    {

        if (distinct)
            return "select distinct ";
        else
            return "select ";
    }

    /**
     * @return the table name prefix for the SQL query encoding
     */
    protected String getTableName(LyClass peerClass)
    {

        if (peerClass instanceof LyUnionClass)
            // Should not happen : the request is considering the union as a dingle class
            // supported only for FFT002415 and this very special join table (join as only one id field when target class has two with an union binding)
            // default can just use the first part of the union (just a guess)
            return getTableName(((LyUnionClass) peerClass).getClass1());

        return peerClass.getName();
    }

    /**
     */
    protected String endEncodeSelectRequest(LySort sort, int limit, String tableName,
        Map<LyField, String> sortAliases)
    {

        if (sort != null)
            return encodeSort(sort, tableName, sortAliases);

        return null;
    }

    /**
     */
    protected String encodeSort(LySort sort, String tableName, Map<LyField, String> aliases)
    {

        if (sort == null)
            return null;

        StringBuffer    result = null;
        int             i, n;
        LySortCriterion criterion;
        String          condition, alias;

        n = sort.getCriterionCount();

        for (i = 0; i < n; i++)
        {
            criterion = sort.getCriterion(i);
            alias = ((aliases != null) ? aliases.get(criterion.getFieldInfo()) : null);
            condition = encodeSort(criterion, tableName, alias, aliases);

            if (condition != null)
            {

                if (result == null)
                    result = new StringBuffer(" order by ");
                else
                    result.append(", ");

                result.append(condition);
            }
        }

        if (result != null)
            return result.toString();
        else
            return null;
    } // end method encodeSort

    /**
     */
    protected String encodeSort(LySortCriterion criterion, String tableName, String alias,
        Map<LyField, String> aliases)
    {
        if (criterion == null)
            return null;

        LyFieldInfo fieldInfo = criterion.getFieldInfo();

        if ((criterion.getJavaName() != null)
        	|| (fieldInfo.getBinding() == null))
        {
            if (_response != null)
                _response.setSorted(false);

            return null;
        }

        /* // XXX: Commented for junit tests to pass
        if ((fieldInfo instanceof LyRelationFieldInfo)
        && !fieldInfo.hasMark(LyFieldInfo.MARK_HIDDEN)
        && ((LyRelationFieldInfo)fieldInfo).getClassInfo(0).isNameDifferentOfId())
         */
        if (fieldInfo instanceof LyRelationFieldInfo)
        {
            String condition = encodeNameRelationSort((LyRelationFieldInfo) fieldInfo,
                    criterion.isIncreasing(), alias, aliases);

            if (condition != null)
                return condition;
        }

        LyField[]    peerFields = _classBinding.getPeerFields(fieldInfo);
        String       id, sortField;
        StringBuffer result = null;
        int          i, n = ((peerFields == null) ? 0 : peerFields.length);
        boolean      caseSensitive = false;

        if (!supportCaseSensitive() || !(fieldInfo instanceof LyTextFieldInfo))
            caseSensitive = true;

        for (i = 0; i < n; i++)
        {
            id = peerFields[i].getId();

            if (_class.getField(id) == null)
                continue;

            if (tableName == null)
                tableName = getTableName(peerFields[i]);

            // If field is BLOB or CLOB don't try to sort on it...
            if (isBlob((LyColumnField) peerFields[i]) || isClob((LyColumnField) peerFields[i]))
                continue;

            String quotedId = QUOTE_CHAR_LEFT + id + QUOTE_CHAR_RIGHT;

            if (tableName != null)
                quotedId = QUOTE_CHAR_LEFT
                    + ((alias != null) ? alias : getTableAlias(null, tableName)) + QUOTE_CHAR_RIGHT
                    + "." + QUOTE_CHAR_LEFT + id + QUOTE_CHAR_RIGHT;

            if ((alias == null) && (aliases != null) && (aliases.get(peerFields[i]) != null))
                quotedId = QUOTE_CHAR_LEFT + aliases.get(peerFields[i]) + QUOTE_CHAR_LEFT;

            getRequestInfo().addUsedClass(tableName);

            if (result == null)
                result = new StringBuffer();
            else
                result.append(", ");

            if (!caseSensitive)
                result.append("lower(");

            if ((peerFields[i].getValueClass() == java.util.Date.class)
                    && (fieldInfo instanceof LyDateFieldInfo))
                //FFT03603
                sortField = getSQLFormatDate(quotedId, (LyDateFieldInfo) fieldInfo);
            else if ((fieldInfo instanceof LyTextFieldInfo)
                    && ((LyTextFieldInfo) fieldInfo).isMultilingual())
                sortField = getSortLangField(quotedId, (LyTextFieldInfo) fieldInfo);
            else
                sortField = quotedId;

            result.append(sortField);

            if (!caseSensitive)
                result.append(")");

            // FFT Fl 529  : sort criteria depends on optional or not (perfs).
            boolean optional = (!fieldInfo.hasMark(LyFieldInfo.MARK_NOT_NULL))
                && (fieldInfo.hasMark(LyFieldInfo.MARK_OPTIONAL));

            result.append(getSqlSortCriterion(criterion.isIncreasing(), optional));
        }

        return ((result == null) ? null : result.toString());
    } // end method encodeSort

    /**
     * String force formating date for order by using FORMAT(quotedId, 'YYYY-MM-DD HH:MM:SS')
     * but syntax is specific depending on database vendor.
     * If not overridden this method does nothing and returns the quotedId to keep backward compatibility.
     * @param quotedId id of the date field (using aliases, table names and quotes if needed)
     * @param peerField peer field related to this column
     */
    public String getSQLFormatDate(String quotedId, LyDateFieldInfo fieldInfo)
    {
        //FFT03603
        return quotedId;
    }

    /**
     * Complete the sort criteria for ascending or descending SQL syntax
     * @param increasing true if we expect an increasing sort
     * @param optional is true if the field is optional.
     */
    protected String getSqlSortCriterion(boolean increasing, boolean optional)
    {

        if (increasing)
            return " ASC";
        else
            return " DESC";
    }

    protected String encodeNameRelationSort(LyRelationFieldInfo relation, boolean increasing,
        String alias, Map<LyField, String> aliases)
    {
        // get the name fields
        LyInfoList<LyFieldInfo> fields = getNameFieldInfos(relation);

        if (fields == null)
            return null;

        // get a sort name on the classes of the relation
        LySort sort = new LySort();

        // build a sort for the name fields
        LyFieldInfo fieldInfo;

        for (int i = 0; i < fields.getSize(); i++)
        {
            fieldInfo = fields.getInfo(i);
            sort.addCriterion(fieldInfo, increasing);
        }

        // get the 1st target class
        LyClassInfo classInfo = relation.getClassInfo(0);

        // encode the sort
        StringBuffer        result = null;
        LySortCriterion     criterion;
        String              condition;
        String              tableName;
        LyClass             peerClass;
        LySqlRequestEncoder encoder;
        int                 n = sort.getCriterionCount();

        for (int i = 0; i < n; i++)
        {
            // get the criterion
            criterion = sort.getCriterion(i);

            // get the field
            fieldInfo = criterion.getFieldInfo();

            // get the table name related to this field
            peerClass = getPeerClass(fieldInfo, classInfo.getBinding());
            tableName = peerClass.getName();

            // get encoder for the this field
            encoder = getRequestEncoder(getDataSource(peerClass), false);

            if (encoder == null)
                continue;

            encoder.setClassBinding(classInfo.getBinding());

            boolean status = getRequestInfo()._lockAliasCount;

            boolean isSelfTargetRelationClass = false; // same class pointed relation

            if ((encoder.getClassBinding() != null) && (getClassBinding() != null)
                    && (encoder.getClassBinding().getBinding() != null)
                    && encoder.getClassBinding().getBinding().equals(
                        getClassBinding().getBinding()))
                isSelfTargetRelationClass = true;

            if (isSelfTargetRelationClass) // should apply all the time, not only for self pointed relations ? Keept as is for compatibility
                getRequestInfo().setLockAliasCount(true); // temporary allow adding aliases

            LySqlRequestInfo requestInfo = setChildRequestInfo(encoder,
                    LySqlRequestInfo.Join.SUM_ALIAS_COUNTS);

            if (isSelfTargetRelationClass)
                getRequestInfo().setLockAliasCount(status); // restore

            requestInfo.setAliasContext(null, classInfo, relation, isSelfTargetRelationClass);

            // encode condition
            if (tableName == null)
                condition = null;
            else
                condition = encoder.encodeSort(criterion, tableName, alias, aliases);

            if (condition != null)
            {
                // add the table name
                requestInfo.addUsedClass(tableName);

                if (result == null)
                    result = new StringBuffer();
                else
                    result.append(", ");

                result.append(condition);
            }
        }

        if (result != null)
            return result.toString();
        else
            return null;
    } // end method encodeNameRelationSort

    /**
     * Encode the column names in the select request for relationnal sorts.<br>
     * This feature is required for some database (ORACLE).
     *
     * @param request current request beeing built.
     * @param sort    the current sort.
     */
    protected void encodeNameRelationColumns(StringBuffer request, LySort sort)
    {
    }

    /**
     * Get the 'NAME' logical fields of the classes that are target of the given relation.<br>
     * Are taken into account only non-local fields that are common to each target class.
     *
     * @param  relation a logical relation field info.
     * @return a list of fields.
     */
    protected LyInfoList<LyFieldInfo> getNameFieldInfos(LyRelationFieldInfo relation)
    {

        // we do not perform a relation sort on multiple relations.
        if (relation.isMultiple())
            return null;

        //  get common name fields of target classes
        LyInfoList<LyClassInfo> classes = relation.getClassInfos();
        LyInfoList<LyFieldInfo> fields = new LyInfoList<LyFieldInfo>();
        LyClassInfo             classInfo;
        LyFieldInfo             fieldInfo;
        int                     i, j, n = classes.getSize();
        LyClass                 peerClass;
        LyDataSource            dataSource;
        LyDataProvider          dataProvider;

        // get name fields of target classes (if those are different from id)
        for (i = 0; i < n; i++)
        {
            classInfo = classes.getInfo(i);

            if (classInfo.isVirtual())
                classInfo = classInfo.getRealClass();

            // ensure class is loaded
            classInfo.load();

            peerClass = classInfo.getPeerClass();

            if (peerClass == null)
                continue;

            // If the target classInfo is not on the same provider,
            // We can't do anything, so ignore it.
            dataSource = peerClass.getDataSource();
            dataProvider = dataSource.getDataProvider();

            if (_provider != dataProvider)
                continue;

            if (classInfo.isLocal())
                continue;

            if (classInfo.isNameDifferentOfId())
                fields.addList(classInfo.getFieldInfos(LyFieldInfo.MARK_NAME));
        }

        // get the number of 'name' fields found.
        int nf = fields.getSize();

        if (nf == 0)
            return null;

        // remove local fields and those that are not common to each class
        for (i = (nf - 1); i >= 0; i--)
        {
            fieldInfo = fields.getInfo(i);

            if (fieldInfo.getBinding() == null)
            {
                fields.removeInfo(fieldInfo);
                continue;
            }

            if (fieldInfo instanceof LyRelationFieldInfo)
            {
                LyInfoList<LyFieldInfo> nameFields = getNameFieldInfos((LyRelationFieldInfo)
                        fieldInfo);

                // remove relations that may provoque multiple relation sort
                // relation sort is enabled only to 1 level.
                if (nameFields != null)
                {
                    fields.removeInfo(fieldInfo);
                    continue;
                }
            }

            for (j = 0; j < n; j++)
            {
                classInfo = classes.getInfo(j);

                if (!classInfo.contains(fieldInfo))
                {
                    fields.removeInfo(fieldInfo);
                    break;
                }
            }
        }

        // return remaining fields
        if (fields.getSize() == 0)
            return null;
        else
            return fields;
    } // end method getNameFieldInfos

    /**
     * Tells if the given sort has a criterion on a relation field that requires a join clause to
     * sort on NAME fields of the target classes.
     *
     * @param  sort current sort.
     * @return a list of logical relations.
     */
    protected LyInfoList<LyRelationFieldInfo> getNameRelationSort(LySort sort)
    {
        // test fields of the sort
        LyInfoList<LyFieldInfo> fields = sort.getFieldInfos();

        if ((fields == null) || (fields.getSize() == 0))
            return null;

        LyFieldInfo                     fieldInfo;
        LyInfoList<LyRelationFieldInfo> result = new LyInfoList<LyRelationFieldInfo>(false);

        for (int i = 0; i < fields.getSize(); i++)
        {
            fieldInfo = fields.getInfo(i);

            if (!(fieldInfo instanceof LyRelationFieldInfo) || (fieldInfo.getBinding() == null))
                continue;

            if (getNameFieldInfos((LyRelationFieldInfo) fieldInfo) != null)
                result.addInfo((LyRelationFieldInfo) fieldInfo);
        }

        return result;
    }

    /**
     * Gives the list of 'NAME' columns on which a relational sort is performed.
     *
     * @param  sort current sort of a select request.
     * @return a list of fieldInfo or null.
     */
    protected LyInfoList<LyFieldInfo> getNameRelationColumns(LySort sort)
    {
        // get relations
        LyInfoList<LyRelationFieldInfo> relations = getNameRelationSort(sort);

        if (relations == null)
            return null;

        LyInfoList<LyFieldInfo> names;
        LyInfoList<LyFieldInfo> result = null;

        for (int i = 0; i < relations.getSize(); i++)
        {
            names = getNameFieldInfos(relations.getInfo(i));

            if (names != null)
            {

                if (result == null)
                    result = new LyInfoList<LyFieldInfo>();

                result.addList(names);
            }
        }

        return result;
    }

    /**
     * Compute the peer class of a logical field that belongs to a given logical class.
     *
     * @param  fieldInfo    logical field.
     * @param  classBinding the binding of a logical class
     * @return a peer class or null if not found.
     */
    protected LyClass getPeerClass(LyFieldInfo fieldInfo, LyClassBinding classBinding)
    {

        if (classBinding == null)
            return null;

        // get the peer field
        LyField[] peerFields = classBinding.getPeerFields(fieldInfo);

        if (peerFields == null)
            return null;

        // get all the peer classes of the current class
        LyClass       peerClass, result = null;
        List<LyClass> peerClasses = getPeerClasses(classBinding.getPeerClass());

        for (Iterator<LyClass> i = peerClasses.iterator(); i.hasNext();)
        {
            peerClass = i.next();

            if (peerClass.getField(peerFields[0].getId()) != null)
            {
                result = peerClass;
                break;
            }
        }

        return result;
    }

    /**
     */
    protected String getCondition(LyFilterExpression.Condition condition, boolean inverted)
    {

        if (!inverted)
            return _conditionNames[condition.ordinal()];
        else
            return _notConditionNames[condition.ordinal()];
    }

    /**
     * To encode the condition to send join requests when managing unions.
     */
    protected String encodeCondition(LyUnionRequestEncoder encoder, LyClassInfo unionClassInfo)
    {

        if (encoder == null)
            return null;

        String         result = null;
        LyUnionBinding unionBinding = (LyUnionBinding) encoder.getClassBinding();

        result = encodeCondition(unionBinding, unionClassInfo, false);
        return result;
    }

    /**
     * To encode the condition to send join requests when managing unions.
     */
    protected String encodeCondition(LyUnionBinding unionBinding, LyClassInfo unionClassInfo,
        boolean optimize)
    {

        if (unionBinding == null)
            return null;

        getRequestInfo().getCurrentTask().addInfo("BEGIN encodeCondition (LyUnionBinding)");

        LyClassBinding binding1, binding2;
        String         condition1, condition2, condition3;
        Vector<String> values1, values2;
        String         str1, str2;
        StringBuffer   buffer, buffer2;
        LyClass        cls;
        LyField        field;
        LyClassBinding binding;
        String         tableName1, tableName2;
        LyFilter       filter;

        binding1 = unionBinding.getClassBinding1();
        binding2 = unionBinding.getClassBinding2();

        condition1 = null;
        condition2 = null;
        condition3 = null;

        if (binding1 instanceof LyUnionBinding)
            condition1 = encodeCondition((LyUnionBinding) binding1, unionClassInfo, optimize);
        else
        {
            filter = binding1.getFilter();

            if (filter != null)
                condition1 = encodeCondition(filter, null, null);
        }

        if (binding2 instanceof LyUnionBinding)
            condition2 = encodeCondition((LyUnionBinding) binding2, unionClassInfo, optimize);
        else
        {
            filter = binding2.getFilter();

            if (filter != null)
                condition2 = encodeCondition(filter, null, null);
        }

        if (!(binding1 instanceof LyUnionBinding) || !(binding2 instanceof LyUnionBinding))
        {

            values1 = unionBinding.getValues1();
            values2 = unionBinding.getValues2();
            buffer = null;

            for (int i = 0; i < values1.size(); i++)
            {
                str1 = values1.elementAt(i);
                binding = unionBinding.getClassBinding1();
                cls = binding.getPeerClass();
                field = cls.getField(str1);
                tableName1 = getTableName(field);

                buffer2 = new StringBuffer();

                buffer2.append("(");
                buffer2.append(QUOTE_CHAR_LEFT);
                buffer2.append(getTableAlias(unionClassInfo, tableName1));
                buffer2.append(QUOTE_CHAR_RIGHT);
                buffer2.append(".");
                buffer2.append(QUOTE_CHAR_LEFT);
                buffer2.append(field.getId());
                buffer2.append(QUOTE_CHAR_RIGHT);

                buffer2.append("=");

                str2 = values2.elementAt(i);
                binding = unionBinding.getClassBinding2();
                cls = binding.getPeerClass();
                field = cls.getField(str2);
                tableName2 = getTableName(field);

                buffer2.append(QUOTE_CHAR_LEFT);
                buffer2.append(getTableAlias(unionClassInfo, tableName2));
                buffer2.append(QUOTE_CHAR_RIGHT);
                buffer2.append(".");
                buffer2.append(QUOTE_CHAR_LEFT);
                buffer2.append(field.getId());
                buffer2.append(QUOTE_CHAR_RIGHT);
                buffer2.append(")");

                if (optimize)
                {
                    // We add the join condition only when it makes sense : the union is being used before
                    ArrayList<String> userClasses;

                    if (getRequestInfo()._lockAliasCount)
                        userClasses = getRequestInfo().getRootParent().getCurrentUsedClasses();
                    else
                        userClasses = getRequestInfo().getCurrentUsedClasses();

                    if (!userClasses.contains(tableName1) || !userClasses.contains(tableName2))
                        continue; // SKIPPED
                }

                getRequestInfo().addUsedClass(tableName1);
                getRequestInfo().addUsedClass(tableName2);

                if (buffer == null)
                    buffer = new StringBuffer();
                else
                    buffer.append(" and ");

                buffer.append(buffer2.toString());
            }

            if (buffer != null)
                condition3 = buffer.toString();
        }

        buffer = new StringBuffer();
        boolean needAnd = false;;

        if (condition1 != null)
        {
            buffer.append(condition1);
            needAnd = true;
        }

        if (condition2 != null)
        {
            if (needAnd)
                buffer.append(" and ");
            buffer.append(condition2);
            needAnd = true;
        }

        if (condition3 != null)
        {
            if (needAnd)
                buffer.append(" and ");
            buffer.append(condition3);
        }

        String ret = ((buffer == null) ? null : buffer.toString());
        getRequestInfo().getCurrentTask().addInfo("END encodeCondition (LyUnionBinding)");
        getRequestInfo().getCurrentTask().addSql(ret);

        return ret;
    } // end method encodeCondition

    /**
     */
    protected String encodeCondition(LyFilter filter, String tableName, LyClassInfo classInfo)
    {

        if (filter == null)
            return null;

        StringBuffer     result = null;
        int              i, n;
        LyFilter.Type    type;
        LyFilterElement  element;
        String           oper;
        String           elem;
        boolean          addParenthesis = false;
        LySqlRequestInfo info = getRequestInfo();

        // If filter is a negation, use inverted filter to encode condition
        if (filter.isNot())
            filter = filter.getInvertedFilter();

        n = filter.getSize();

        if (n == 0)
            return null;

        type = filter.getType();

        if (type == LyFilter.Type.AND)
            oper = " and ";
        else if (type == LyFilter.Type.OR)
        {
            oper = " or ";

            if (usesUnionSelect())
                info.setDetectingUnionSelect(true);
        }
        else
            return null;

        getRequestInfo().getCurrentTask().addInfo("BEGIN encodeCondition (LyFilter) : n=" + n);

        for (i = 0; i < n; i++)
        {
            element = filter.getElement(i);
            elem = null;

            if (element instanceof LyFilter)
                elem = encodeCondition((LyFilter) element, tableName, classInfo);
            else if (element instanceof LyFilterExpression)
                elem = encodeCondition((LyFilterExpression) element, tableName, classInfo);
            else if (element instanceof LyFilterRelation)
                elem = encodeCondition((LyFilterRelation) element, tableName, classInfo);
            else if (element instanceof LyJoinFilter)
                elem = encodeCondition((LyJoinFilter) element, tableName);

            if (elem == null)
            {

                if (_response != null)
                    _response.setFiltered(false);
            }
            else
            {

                if (result != null)
                {
                    result.append(oper);
                    addParenthesis = true;
                }
                else
                    result = new StringBuffer();

                result.append("(");
                result.append(elem);
                result.append(")");

                if (getRequestInfo().isDetectingUnionSelect())
                    getRequestInfo().addUnionSelectInfos(elem);
            }
        }

        if (info.isDetectingUnionSelect())
        {
            info.setDetectingUnionSelect(false);

            if (info.getUnionSelectInfos(0) != null)
                // *** Union Select is active, put a parameter for the parent (will be a variable)
                return UNION_SELECT_FILTER;
        }

        if (result != null)
        {
            getRequestInfo().getCurrentTask().addInfo("END encodeCondition (LyFilter)");
            getRequestInfo().getCurrentTask().addSql(result.toString());

            if (addParenthesis)
                return "(" + result.toString() + ")";
            else
                return result.toString();
        }
        else
            return null;
    } // end method encodeCondition

    /**
     */
    protected String encodeCondition(LyFilterRelation relation, String tableName,
        LyClassInfo classInfoSource)
    {
        LyRelationFieldInfo relationFieldInfo = relation.getRelationFieldInfo();
        getRequestInfo().getCurrentTask().addInfo(
            "begin encodeCondition (LyFilterRelation) relation=" + relationFieldInfo.getId()
            + " tablename=" + tableName + " classInfoSource=" + classInfoSource);

        StringBuffer            result = null, result2 = null;
        LyInfoList<LyClassInfo> classInfos = relation.getPossibleClasses(false);
        LyClassInfo             classInfo, subClass;
        LyFilter                relationFilter = relation.getFilter();
        int                     i, j, n;
        LyClass                 peerClass;
        LyDataSource            dataSource;
        LySqlRequestEncoder     encoder;
        String                  targetTableName, elem, classFilterCondition;
        Vector<String>          existingWhereStatements = new Vector<String>(); // Avoid duplicate where A or A clauses...

        // add only classes that have a binding defined (abstract or concrete)
        LyInfoList<LyClassInfo> classes = new LyInfoList<LyClassInfo>(), subClasses;

        for (i = 0, n = classInfos.getSize(); i < n; i++)
        {
            classInfo = classInfos.getInfo(i);

            if (classInfo.getBinding() != null)
                classes.addInfo(classInfo);
            else if (classInfo.isAbstract())
            {
                subClasses = classInfo.getSubClasses(true);

                for (j = 0; j < subClasses.getSize(); j++)
                {
                    subClass = subClasses.getInfo(j);

                    if (!subClass.isAbstract())
                        classes.addInfo(subClass);
                }
            }
        }

        for (i = 0, n = classes.getSize(); i < n; i++)
        {
            classInfo = classes.getInfo(i);

            if (classInfo.getBinding() == null)
                continue;

            peerClass = classInfo.getBinding().getPeerClass();
            dataSource = getDataSource(peerClass);
            encoder = getRequestEncoder(dataSource);

            targetTableName = classInfo.getBinding().getBinding();

            if ((targetTableName == null) && (classInfo.getBinding() instanceof LyUnionBinding))
            {
                //define the encoder for the union class
                encoder.setJoinPeerClasses(getPeerClasses(peerClass));
                encoder.setClassBinding(classInfo.getBinding());
                encoder._class = peerClass;

                LySqlRequestEncoder encoderUnion = encoder;
                //                              LySqlRequestInfo        requestInfoUnion = setChildRequestInfo(encoderUnion);

                //                              requestInfoUnion.setTargetClassInfo(null);
                // all classes are on the same level
                LySqlRequestInfo requestInfo;
                LySqlRequestInfo unionRequestInfo = setChildRequestInfo(encoderUnion,
                        LySqlRequestInfo.Join.NEW_ALIAS_ONLY);

                unionRequestInfo.setAliasContext(null, classInfo, null);

                // encode filter
                String filterCondition = encoderUnion.encodeCondition(relationFilter, null,
                        classInfo);

                // encode join condition
                ArrayList<LyBinding> bindings = getPeerBindings(classInfo.getBinding());
                LyClassBinding       binding;
                String               joinCondition = null;

                for (j = 0; j < bindings.size(); j++)
                {
                    binding = (LyClassBinding) bindings.get(j);
                    peerClass = binding.getPeerClass();
                    dataSource = getDataSource(peerClass);
                    encoder = getRequestEncoder(dataSource);

                    targetTableName = getTableName(peerClass);

                    if (encoder == null)
                        continue;

                    requestInfo = setChildRequestInfo(encoder,
                            LySqlRequestInfo.Join.SUM_ALIAS_COUNTS);
                    requestInfo.setAliasContext(null, classInfo, null);
                    encoder.setClassBinding(classInfo.getBinding());

                    // Encode relation between tables
                    joinCondition = encodeJoinCondition(relationFieldInfo, binding, tableName,
                            classInfo);
                    clearChildRequestInfo(encoder);

                    if (joinCondition != null)
                        break;
                }

                // encode union
                boolean optimize = true; //getEnvironment().testResource("LY_OPTIMIZE_SQL");
                String  unionCondition = encoderUnion.encodeCondition((LyUnionBinding) classInfo
                        .getBinding(), classInfo, optimize);

                clearChildRequestInfo(encoderUnion);

                // encode whole relation filter
                if ((joinCondition != null) || (filterCondition != null)
                        || (unionCondition != null))
                {
                    result2 = new StringBuffer();

                    if ((joinCondition != null) && (joinCondition.length() > 0))
                    {
                        result2.append("(");
                        result2.append(joinCondition);
                        result2.append(")");
                    }

                    if ((filterCondition != null) && (filterCondition.length() > 0))
                    {

                        if (result2.length() > 0)
                            result2.append(" and ");

                        result2.append("(");
                        result2.append(filterCondition.toString());
                        result2.append(")");
                    }

                    if ((unionCondition != null) && (unionCondition.length() > 0))
                    {

                        if (result2.length() > 0)
                            result2.append(" and ");

                        result2.append("(");
                        result2.append(unionCondition);
                        result2.append(")");
                    }

                    if (result2.length() > 0)
                    {

                        if (!existingWhereStatements.contains(result2.toString()))
                        {

                            if (result == null)
                                result = new StringBuffer();
                            else
                                result.append(" or ");

                            result.append("(").append(result2).append(")");
                            existingWhereStatements.addElement(result2.toString());
                        }
                    }
                }

                continue;
            }

            if (encoder == null)
                continue;

            LyClassBinding classInfoBinding = classInfo.getBinding();

            encoder.setClassBinding(classInfoBinding);

            LySqlRequestInfo requestInfo = setChildRequestInfo(encoder,
                    LySqlRequestInfo.Join.SUM_ALIAS_COUNTS);

            // Encode relation between tables
            LyRelationFieldInfo fieldInfo = relation.getRelationFieldInfo();
            LyInfoDaemon        daemon = fieldInfo.getDaemon();

            // FFT Structis 291 : filter on current class + relation filter on current class : musr encode with aliases
            if (getRequestInfo().getCurrentUsedClasses().contains(targetTableName)
                    && ((tableName == null) || (tableName.equals(targetTableName))) // FFT00723 Flandres
                    && (daemon == null)) // to avoid side effects, already works with the daemon
            {
                // MUST use alias
                String alias1 = encoder.getRequestInfo().getTableAlias(classInfo, fieldInfo,
                        targetTableName);

                if (targetTableName.equals(alias1))
                {
                    alias1 = computeTableAlias(targetTableName);
                    encoder.getRequestInfo().addTableAlias(null, classInfo, fieldInfo,
                        targetTableName, alias1, true, false);
                }
            }

            String joinCondition = encodeJoinCondition(fieldInfo, classInfo.getBinding(), tableName,
                    classInfo);

            if (joinCondition != null)
            {
                // If the daemon is a LyJoinDaemon, three tables are used
                String domain = null;

                if (daemon instanceof LyJoinDaemon)
                {
                    LyJoinDaemon        joinDaemon = (LyJoinDaemon) daemon;
                    LyClassInfo         joinClassInfo = joinDaemon.getJoinClass();
                    LyRelationFieldInfo targetRelation = joinDaemon.getTargetRelation();

                    // If the daemon is used twice, must use aliases on the daemon class (Fl FFT504)
                    String joinTableName = joinClassInfo.getBinding().getBinding();

                    if (getRequestInfo().getCurrentUsedClasses().contains(joinTableName))
                    {
                        //                                               MUST use alias
                        String alias1 = encoder.getRequestInfo().getTableAlias(joinClassInfo,
                                targetRelation, joinTableName);

                        if (joinTableName.equals(alias1))
                        {
                            domain = LySqlRequestEncoder.DOMAIN_FILTER_RELATION;
                            encoder.getRequestInfo().copyAliasFromDomain(encoder.getRequestInfo(),
                                encoder.getRequestInfo().getDomain(), domain);
                            encoder.getRequestInfo().setAliasDomain(domain);
                            alias1 = encoder.getRequestInfo().getTableAlias(joinClassInfo,
                                    targetRelation, joinTableName);

                            if (joinTableName.equals(alias1)
                                    && (!encoder.getRequestInfo()._lockAliasCount))
                            {
                                alias1 = computeTableAlias(joinTableName);
                                encoder.getRequestInfo().addTableAlias(domain, joinClassInfo,
                                    targetRelation, joinTableName, alias1, true, false);
                            }
                        }
                    }

                    classInfo = joinClassInfo;
                    requestInfo.setAliasContext(domain, classInfo, targetRelation);
                }
                else
                    requestInfo.setAliasContext(domain, classInfo, fieldInfo);

                elem = encoder.encodeCondition(relationFilter, targetTableName, classInfo);
                encoder.getRequestInfo().setAliasDomain(null);

                if (elem != null)
                {
                    classFilterCondition = null;

                    if ((classInfoBinding != null) && (classInfoBinding.getFilter() != null))
                        classFilterCondition = encoder.encodeCondition(classInfoBinding.getFilter(),
                                targetTableName, classInfo);

                    result2 = new StringBuffer();
                    result2.append("((");
                    result2.append(joinCondition);
                    result2.append(") and (");
                    result2.append(elem);

                    if (classFilterCondition != null)
                    {
                        result2.append(") and (");
                        result2.append(classFilterCondition);
                    }

                    result2.append("))");

                    if (!existingWhereStatements.contains(result2.toString()))
                    {

                        if (result == null)
                            result = new StringBuffer();
                        else
                            result.append(" or ");

                        result.append(result2);
                        existingWhereStatements.addElement(result2.toString());
                    }
                }
            }

            requestInfo.setAliasContext(null, null, null);
            clearChildRequestInfo(encoder);
        }

        String ret = ((result == null) ? null : result.toString());
        getRequestInfo().getCurrentTask().addInfo("end encodeCondition (LyFilterRelation) relation="
            + relationFieldInfo.getId() + " tablename=" + tableName + " classInfoSource="
            + classInfoSource);
        getRequestInfo().getCurrentTask().addSql(ret);

        return ret;
    } // end method encodeCondition

    /**
     */
    protected String encodeCondition(LyJoinFilter join, String tableName)
    {
        getRequestInfo().getCurrentTask().addInfo("BEGIN encodeCondition (LyJoinFilter) " + join);

        LyFilter       filter = join.getSourceFilter();
        LyFieldInfo    source = join.getSourceFieldInfo();
        LyFieldInfo    target = join.getTargetFieldInfo();
        LyClassInfo    classInfo = join.getSourceClassInfo();
        LyClassBinding sourceBinding = classInfo.getBinding();
        LyClass        peerSource = sourceBinding.getPeerClass();
        String         sourceTable = getTableName(peerSource);
        String         filterCondition = null;
        StringBuffer   condition = new StringBuffer();

        if (filter != null)
        {
            // Encode source condition for filter
            LyDataSource        dataSource = getDataSource(peerSource);
            LySqlRequestEncoder encoder = getRequestEncoder(dataSource);

            encoder.setClassBinding(sourceBinding);

            filterCondition = encoder.encodeCondition(filter, sourceTable, null);

            if (filterCondition != null)
            {
                condition.append(filterCondition);
                condition.append(" and ");
            }
        }

        // Encode join condition.
        LyField sourceField, targetField;
        String  idSource, idTarget;

        LyField[] sourceFields, targetFields;
        sourceFields = sourceBinding.getPeerFields(source);
        targetFields = _classBinding.getPeerFields(target);

        if (sourceFields.length != targetFields.length)
            throw new IllegalStateException("Unable to create join condition because " + source
                + " and " + target + " don't have the same number of peer fields");

        int i, n = sourceFields.length;

        if (filterCondition != null)
            condition.append("(");

        for (i = 0; i < n; i++)
        {

            if (i > 0)
                condition.append(" and ");

            if (n > 1)
                condition.append("(");

            sourceField = sourceFields[i];
            idSource = sourceField.getId();
            targetField = targetFields[i];
            idTarget = targetField.getId();

            if (tableName == null)
                tableName = getTableName(targetField);

            if (tableName != null)
                idTarget = QUOTE_CHAR_LEFT + tableName + QUOTE_CHAR_RIGHT + "." + QUOTE_CHAR_LEFT
                    + idTarget + QUOTE_CHAR_RIGHT;

            getRequestInfo().addUsedClass(tableName);

            idSource = QUOTE_CHAR_LEFT + sourceTable + QUOTE_CHAR_RIGHT + "." + QUOTE_CHAR_LEFT
                + idSource + QUOTE_CHAR_RIGHT;
            getRequestInfo().addUsedClass(sourceTable);

            condition.append(idTarget);
            condition.append("=");
            condition.append(idSource);

            if (n > 1)
                condition.append(")");
        }

        if (filterCondition != null)
            condition.append(")");

        getRequestInfo().getCurrentTask().addInfo("END encodeCondition (LyJoinFilter) " + join);
        getRequestInfo().getCurrentTask().addSql(condition.toString());

        return condition.toString();
    } // end method encodeCondition

    /**
     * Encodes a join condition between two tables.
     *
     * @param  fieldInfo       The relation field info that represents the link between the two
     *                         classes.
     * @param  classBinding    Binding of the target class info.
     * @param  tableName       name of the table from which the relation starts.
     * @param  targetClassInfo the target logical class of the relation field.(may be null)
     * @return the encoded condition.
     */
    protected String encodeJoinCondition(LyRelationFieldInfo fieldInfo, LyClassBinding classBinding,
        String tableName, LyClassInfo targetClassInfo)
    {
        LyInfoDaemon daemon = fieldInfo.getDaemon();

        if (daemon != null)
            return encodeDaemonJoinCondition(fieldInfo, classBinding, tableName, targetClassInfo);

        LyFieldBinding fieldBinding = fieldInfo.getBinding();

        if (fieldBinding == null)
            return null;

        getRequestInfo().getCurrentTask().addInfo("enter encodeJoinCondition relation="
            + fieldInfo.getId() + " tablename=" + tableName + " targetClassInfo="
            + targetClassInfo);

        LyField[]       arrayPeerField;
        LyField         peerField;
        StringBuffer    joinBuffer = new StringBuffer();
        LyClass         peerClass = classBinding.getPeerClass();
        String          targetTableName = getTableName(peerClass);
        Vector<LyField> idFields = new Vector<LyField>(peerClass.getIdFields());
        boolean         convertToString = fieldInfo.isMultiple();
        String          idFieldsString = null;
        boolean         isBindings = false;

        if (fieldBinding instanceof LyFieldsBinding)
        {
            arrayPeerField = _classBinding.getPeerFields(fieldInfo);
            isBindings = true;
        }
        else
        {
            arrayPeerField = new LyField[1];
            arrayPeerField[0] = _classBinding.getPeerField(fieldInfo);
            idFieldsString = encodeIdFieldsForJoin(idFields, targetTableName, targetClassInfo,
                    fieldInfo, convertToString);
        }

        getRequestInfo().addUsedClass(tableName);

        //if the number of idFields of peerClass is greater than the number of peerFields of classBinding
        //then this correction doesn't work
        for (int i = 0; i < arrayPeerField.length; i++)
        {
            peerField = arrayPeerField[i];

            if (i != 0)
                joinBuffer.append(" and ");

            if (peerField == null)
                return null;

            String peerFieldId = peerField.getId();

            if (isBindings)
            {
                Vector<LyField> fields = new Vector<LyField>();

                fields.add(idFields.elementAt(i));
                idFieldsString = encodeIdFieldsForJoin(fields, targetTableName, targetClassInfo,
                        fieldInfo, convertToString);
            }

            if (tableName == null)
                tableName = getTableName(peerField);

            // Case of only one value in the first table
            joinBuffer.append("(");

            if (convertToString)
                joinBuffer.append(getToStringFunction() + "(");

            getRequestInfo().addUsedClass(tableName);

            joinBuffer.append(QUOTE_CHAR_LEFT).append(getTableAlias(targetClassInfo, tableName))
            .append(QUOTE_CHAR_RIGHT).append(".").append(QUOTE_CHAR_LEFT).append(peerFieldId)
            .append(QUOTE_CHAR_RIGHT);

            if (convertToString)
            {
                joinBuffer.append(")");
                joinBuffer.append(" like (");
            }
            else
                joinBuffer.append("=(");

            joinBuffer.append(idFieldsString);
            joinBuffer.append("))");

            if (fieldInfo.isMultiple())
            {
                getRequestInfo().addUsedClass(tableName);

                // Case of multiple values in the first table and the value of the
                // target table is the first one
                joinBuffer.append(" or (");
                joinBuffer.append(QUOTE_CHAR_LEFT).append(getTableAlias(targetClassInfo, tableName))
                .append(QUOTE_CHAR_RIGHT).append(".").append(QUOTE_CHAR_LEFT).append(peerFieldId)
                .append(QUOTE_CHAR_RIGHT);
                joinBuffer.append(" like (");
                joinBuffer.append(idFieldsString);
                joinBuffer.append(" + ");
                joinBuffer.append("'").append(_valueSep).append("%'");
                joinBuffer.append("))");

                // Case of multiple values in the first table and the value of the
                // target table is the last one
                joinBuffer.append(" or (");
                joinBuffer.append(QUOTE_CHAR_LEFT).append(getTableAlias(targetClassInfo, tableName))
                .append(QUOTE_CHAR_RIGHT);
                joinBuffer.append(".").append(QUOTE_CHAR_LEFT).append(peerFieldId).append(
                    QUOTE_CHAR_RIGHT);
                joinBuffer.append(" like ('%").append(_valueSep).append("'");
                joinBuffer.append(" + ");
                joinBuffer.append(idFieldsString);
                joinBuffer.append("))");

                // Case of multiple values in the first table and the value of the
                // target table is one in the middle
                joinBuffer.append(" or (");
                joinBuffer.append(QUOTE_CHAR_LEFT).append(getTableAlias(targetClassInfo, tableName))
                .append(QUOTE_CHAR_RIGHT);
                joinBuffer.append(".").append(QUOTE_CHAR_LEFT).append(peerFieldId).append(
                    QUOTE_CHAR_RIGHT);
                joinBuffer.append(" like ('%").append(_valueSep).append("'");
                joinBuffer.append(" + ");
                joinBuffer.append(idFieldsString);
                joinBuffer.append(" + ");
                joinBuffer.append("'").append(_valueSep).append("%'");
                joinBuffer.append("))");
            }
        }

        getRequestInfo().getCurrentTask().addInfo("exit encodeJoinCondition");
        getRequestInfo().getCurrentTask().addSql(joinBuffer.toString());
        return joinBuffer.toString();
    } // end method encodeJoinCondition

    /**
     * Encodes a join condition on a relation with a daemon.
     *
     * @param  fieldInfo       The relation field info on the source class.
     * @param  classBinding    Binding of the target class info.
     * @param  tableName       name of the table from which the relation starts.
     * @param  targetClassInfo the target logical class of the relation field.(may be null)
     * @return the encoded condition.
     */
    protected String encodeDaemonJoinCondition(LyRelationFieldInfo fieldInfo,
        LyClassBinding classBinding, String tableName, LyClassInfo targetClassInfo)
    {
        StringBuffer joinBuffer = new StringBuffer();
        LyInfoDaemon daemon = fieldInfo.getDaemon();

        // If the daemon is a LyJoinDaemon, three tables are used
        if (daemon instanceof LyJoinDaemon)
        {
            getRequestInfo().getCurrentTask().addInfo(
                "begin encodeDaemonJoinCondition LyJoinDaemon");

            LyJoinDaemon        joinDaemon = (LyJoinDaemon) daemon;
            LyClassInfo         joinClassInfo = joinDaemon.getJoinClass();
            LyClassBinding      joinClassBinding = joinClassInfo.getBinding();
            String              joinTableName = joinClassBinding.getBinding();
            LyRelationFieldInfo sourceRelation = joinDaemon.getSourceRelation();
            LyRelationFieldInfo targetRelation = joinDaemon.getTargetRelation();
            LyClass             joinPeerClass = joinClassInfo.getBinding().getPeerClass();
            LyDataSource        dataSource = getDataSource(joinPeerClass);
            LySqlRequestEncoder encoder = getRequestEncoder(dataSource);

            if (encoder == null)
                return null;

            encoder.setClassBinding(joinClassInfo.getBinding());

            StringBuffer     joinBuffer1 = new StringBuffer();
            StringBuffer     joinBuffer2 = new StringBuffer();
            LySqlRequestInfo requestInfo = setChildRequestInfo(encoder,
                    LySqlRequestInfo.Join.SUM_ALIAS_COUNTS);

            requestInfo.setAliasContext(getRequestInfo().getDomain(), null, null);

            {
                String joinTableAlias = encoder.computeTableAlias(joinTableName);

                // add the table alias for the current logical class (null)
                encoder.addTableAlias(null, joinTableName, joinTableAlias);
            }

            // Flandres FFT00438 case of a join between between the same table : use aliases
            if ((_classBinding != null) && (_classBinding.getBinding() != null)
                    && classBinding.getBinding().equals(_classBinding.getBinding()))
            {
                // one child request info  for EACH join conditions
                String sharedTableName = classBinding.getBinding();

                String alias1 = requestInfo.getTableAlias(null, null, sharedTableName);
                String alias2 = requestInfo.getTableAlias(joinClassInfo, targetRelation,
                        sharedTableName);

                if (!alias1.equals(sharedTableName))
                    // do not compute source alias : this is current alias
                    requestInfo.addTableAlias(null, joinClassInfo, sourceRelation, sharedTableName,
                        alias1, true, false);

                if (alias2.equals(sharedTableName))
                {
                    alias2 = computeTableAlias(sharedTableName);
                    requestInfo.addTableAlias(null, joinClassInfo, targetRelation, sharedTableName,
                        alias2, true, false);
                }

                joinBuffer1.append(encoder.encodeJoinCondition(sourceRelation, _classBinding,
                        joinTableName, joinClassInfo));

                joinBuffer2.append(encoder.encodeJoinCondition(targetRelation, classBinding,
                        joinTableName, joinClassInfo));
            }
            else
            {
                // one child request info  for both join conditions
                joinBuffer1.append(encoder.encodeJoinCondition(sourceRelation, _classBinding,
                        joinTableName, joinClassInfo));
                joinBuffer2.append(encoder.encodeJoinCondition(targetRelation, classBinding,
                        joinTableName, joinClassInfo));
            }

            clearChildRequestInfo(encoder);

            joinBuffer.append("(");
            joinBuffer.append(joinBuffer1);
            joinBuffer.append(") and (");
            joinBuffer.append(joinBuffer2);
            joinBuffer.append(")");
        }
        // If the daemon is a LyReverseDaemon, only two tables are used
        else if (daemon instanceof LyReverseDaemon)
        {
            getRequestInfo().getCurrentTask().addInfo("encodeDaemonJoinCondition LyReverseDaemon");

            LyReverseDaemon     reverseDaemon = (LyReverseDaemon) daemon;
            String              targetTableName = classBinding.getBinding();
            LyRelationFieldInfo reverseRelation = reverseDaemon.getReverseRelation();
            LyClass             targetPeerClass = classBinding.getPeerClass();
            LyDataSource        dataSource = getDataSource(targetPeerClass);
            LySqlRequestEncoder encoder = getRequestEncoder(dataSource);

            if (encoder == null)
                return null;

            encoder.setClassBinding(classBinding);

            LySqlRequestInfo requestInfo = setChildRequestInfo(encoder,
                    LySqlRequestInfo.Join.SUM_ALIAS_COUNTS);

            requestInfo.setAliasContext(null, null, null);

            String joinCondition = encoder.encodeJoinCondition(reverseRelation, _classBinding,
                    targetTableName, targetClassInfo);

            clearChildRequestInfo(encoder);

            if (joinCondition == null)
                return null;
            else
                joinBuffer.append(joinCondition);
        }

        String ret = ((joinBuffer.length() > 0) ? joinBuffer.toString() : null);
        getRequestInfo().getCurrentTask().addInfo("end encodeDaemonJoinCondition LyJoinDaemon");
        getRequestInfo().getCurrentTask().addSql(ret);

        return ret;
    } // end method encodeDaemonJoinCondition

    /**
     * Encodes the sequence of id fields that must be used for the target table of a join condition.
     *
     * @param  fieldInfo       The relation field info that represents the link between the two
     *                         classes.
     * @param  targetClassInfo Target class info pointed by the relation.
     * @param  fieldInfo       The relation field info that represents the link between the two
     *                         classes.
     * @param  tableName       name of the table from which the relation starts.
     * @return the encoded condition.
     */
    protected String encodeIdFieldsForJoin(Vector<LyField> idFields, String tableName,
        LyClassInfo targetClassInfo, LyRelationFieldInfo fieldInfo, boolean convertToString)
    {
        StringBuffer  buffer = new StringBuffer();
        LyColumnField field;
        int           i, n = idFields.size();
        String        str;

        for (i = 0; i < n; i++)
        {
            str = tableName;
            field = (LyColumnField) idFields.elementAt(i);

            if (str == null)
                str = getTableName(field);

            if (str == null)
            {
                _joinPeerClasses.addAll(getPeerClasses(targetClassInfo.getPeerClass()));
                str = getTableName(field);
            }

            if (i != 0)
            {
                buffer.append(" + '");
                buffer.append(LyObjectId.FIELD_SEP);
                buffer.append("' + ");
            }

            if (convertToString)
                buffer.append(getToStringFunction() + "(");

            getRequestInfo().addUsedClass(str);

            /* // XXX: Commented for junit tests to pass
            buffer.append(getTableAlias(targetClassInfo, str));
             */
            String targetClassInfoAlias;

            if ((_classBinding != null) && (_classBinding.getBinding() != null)
                    && _classBinding.getBinding().equals(str))
                targetClassInfoAlias = getTableAliasForSelfTargetRelationClass(null,
                        targetClassInfo, fieldInfo, str);
            else
                targetClassInfoAlias = getRequestInfo().getTableAlias(targetClassInfo, fieldInfo,
                        str);

            buffer.append(QUOTE_CHAR_LEFT).append(targetClassInfoAlias).append(QUOTE_CHAR_RIGHT);
            buffer.append(".");
            buffer.append(QUOTE_CHAR_LEFT).append(field.getId()).append(QUOTE_CHAR_RIGHT);

            if (convertToString)
                buffer.append(")");
        }

        getRequestInfo().getCurrentTask().addInfo("encodeIdFieldsForJoin : targetClassInfo="
            + ((targetClassInfo == null) ? "null" : targetClassInfo.getId()) + " fieldInfo="
            + ((fieldInfo == null) ? "null" : fieldInfo.getId()));
        getRequestInfo().getCurrentTask().addSql(buffer.toString());

        return buffer.toString();
    } // end method encodeIdFieldsForJoin


    /**
     * Search for the table alias of a given table for a given logical class class and target class on the case of a self (class
     * and target class identical) pointing relation
     * @param  classInfo        target logical class
     * @param  fieldInfo        optional parameter to search the alias name in the context of this
     *                          fieldInfo
     * @param  tableName        table name
     * @return the table alias if found, the table name otherwise.
     */
    public String getTableAliasForSelfTargetRelationClass(String domain, LyClassInfo classInfo,
        LyFieldInfo fieldInfo, String tableName)
    {
        // FFT02864 self pointing relation for join : we already assume to come from a relational fieldInfo
        String tableAlias = getRequestInfo().getTableAlias(domain, classInfo, null /*fieldInfo*/,
                tableName, true, true);

        // we MUST use an alias and a new one in that case
        if (tableAlias.equals(tableName))
        {
            tableAlias = computeTableAlias(tableName);

            addTableAlias(classInfo, tableName, tableAlias, true, true);

            if ((classInfo != null)
                    && tableName.equals(
                        getRequestInfo().getTableAlias(domain, null, null, tableName, true, true)))
                // store as new default alias for self class
                addTableAlias(null, tableName, tableAlias, false, true);
        }

        return tableAlias;
    }

    /**
     * Indicates the name of the function that must be used to convert a expression to a string.
     *
     * @return the name of the function
     */
    protected String getToStringFunction()
    {
        return "";
    }

    /**
     */
    protected String encodeCondition(LyFilterExpression expression, String tableName,
        LyClassInfo classInfo)
    {

        if (expression == null)
            return null;

        getRequestInfo().getCurrentTask().addInfo("BEGIN encodeCondition (LyFilterExpression) : "
            + expression + " tableName=" + tableName);

        LyFieldInfo    fieldInfo = expression.getFieldInfo();
        LyFieldBinding fieldBinding = fieldInfo.getBinding();

        // Encode relation between tables
        if (fieldInfo instanceof LyRelationFieldInfo)
        {
            LyRelationFieldInfo relationFieldInfo = (LyRelationFieldInfo) fieldInfo;
            LyInfoDaemon        daemon = relationFieldInfo.getDaemon();

            // If daemon is a join daemon, encode join between tables and encode
            // condition on the join table
            if (daemon instanceof LyJoinDaemon)
            {
                LyJoinDaemon        joinDaemon = (LyJoinDaemon) daemon;
                LyClassInfo         joinClassInfo;
                LyRelationFieldInfo sourceRelation, targetRelation;

                joinClassInfo = joinDaemon.getJoinClass();
                sourceRelation = joinDaemon.getSourceRelation();
                targetRelation = joinDaemon.getTargetRelation();


                return encodeJoinCondition(expression, joinClassInfo, sourceRelation,
                        targetRelation, tableName, classInfo);
            }
            else if (daemon instanceof LyReverseDaemon)
                return encodeReverseDaemonCondition(expression, (LyReverseDaemon) daemon,
                        relationFieldInfo, tableName, classInfo);
        }
        else if (fieldInfo instanceof LyArrayFieldInfo)
        {
            // If fieldInfo is an array of relations or structs we must encode
            // a join.
            LyArrayFieldInfo arrayFieldInfo = (LyArrayFieldInfo) fieldInfo;
            LyInfoDaemon     daemon = arrayFieldInfo.getDaemon();
            LyFieldInfo      internFieldInfo = arrayFieldInfo.getFieldInfo();

            // If daemon is an array daemon, encode join between tables and encode
            // condition on the join table
            if ((internFieldInfo instanceof LyRelationFieldInfo)
                    && (daemon instanceof LyArrayDaemon))
            {
                LyArrayDaemon       arrayDaemon = (LyArrayDaemon) daemon;
                LyClassInfo         arrayClassInfo = arrayDaemon.getArrayClassInfo();
                LyRelationFieldInfo sourceRelation = (LyRelationFieldInfo) arrayClassInfo
                    .getFieldInfo(0);
                LyRelationFieldInfo targetRelation = (LyRelationFieldInfo) arrayClassInfo
                    .getFieldInfo(1);

                return encodeJoinCondition(expression, arrayClassInfo, sourceRelation,
                        targetRelation, tableName, classInfo);
            }
        }

        if (fieldBinding == null)
            return null;

        if (fieldBinding instanceof LyFieldsBinding)
        {
            LyField[]  peerFields = _classBinding.getPeerFields(fieldInfo);
            int        i, n = ((peerFields == null) ? 0 : peerFields.length);
            Class<?>[] classes = new Class[n];
            boolean    added;

            for (i = 0; i < n; i++)
                classes[i] = peerFields[i].getValueClass();

            Object       value = expression.getValue();
            Object[]     fieldValues = fieldInfo.getPeerValues(value, classes, null);
            StringBuffer result = new StringBuffer();
            String       id;

            result.append("((");

            for (i = 0, added = false; i < n; i++)
            {
                id = peerFields[i].getId();

                if (_class.getField(id) == null)
                    continue;

                if (added)
                    if (expression.isInverted())
                        result.append(") or (");
                    else
                        result.append(") and (");

                added = true;
                result.append(encodeCondition(expression, peerFields[i],
                        ((fieldValues != null) ? fieldValues[i] : null), tableName, classInfo));
            }

            result.append("))");
            getRequestInfo().getCurrentTask().addInfo("END encodeCondition (LyFilterExpression)");
            getRequestInfo().getCurrentTask().addSql(result.toString());

            return result.toString();
        }
        else
        {
            LyField peerField = _classBinding.getPeerField(fieldInfo);

            if (peerField == null)
                return null; // May occur on filter relations on union class

            String id = peerField.getId();

            if (_class.getField(id) == null)
                return null;

            Object value = expression.getValue();

            getRequestInfo().getCurrentTask().addInfo("END encodeCondition (LyFilterExpression)");
            return encodeCondition(expression, peerField, value, tableName, classInfo);
        }
    } // end method encodeCondition

    /**
     */
    protected String encodeJoinCondition(LyFilterExpression expression, LyClassInfo joinClassInfo,
        LyRelationFieldInfo sourceRelation, LyRelationFieldInfo targetRelation, String tableName,
        LyClassInfo classInfo)
    {
        // Create a request encoder for the join class
        LyClassBinding      joinClassBinding = joinClassInfo.getBinding();
        String              joinTableName = joinClassBinding.getBinding();
        LyClass             joinPeerClass = joinClassInfo.getBinding().getPeerClass();
        LyDataSource        dataSource = getDataSource(joinPeerClass);
        Object              value = expression.getValue();
        LySqlRequestEncoder encoder = getRequestEncoder(dataSource);
        LyFilterExpression  beginExpression = expression;

        if (encoder == null)
            return null;

        StringBuffer result = new StringBuffer();

        encoder.setClassBinding(joinClassInfo.getBinding());

        LySqlRequestInfo requestInfo = setChildRequestInfo(encoder,
                LySqlRequestInfo.Join.SUM_ALIAS_COUNTS);


        requestInfo.setAliasContext(getRequestInfo().getDomain(), null, null);

        if (value == null)
        {
            // Add the join clause : we need the used class list due to this condition
            // Keep current
            String condition = encoder.encodeJoinCondition(sourceRelation, _classBinding,
                    joinTableName, joinClassInfo);

            // Restore
            ArrayList<String> newUsedClassesList = requestInfo.getCurrentUsedClasses();

            requestInfo.clearCurrentUsedClasses();

            // When the relation is null (empty) thid is a count() condition on the target objects
            // Should be like : where (select count(*) from JOIN_TABLE where REL_KEYWORD_OBJECT.OBJECT_ID_FK=GENERIC_EVENT.GENERIC_EVENT_ID)=0)
            if (condition != null)
            {
                result.append("(");
                result.append("(SELECT COUNT(*) from ");

                boolean found = false;
                String  tableNameCurrent;

                for (int i = 0; i < newUsedClassesList.size(); i++)
                {
                    tableNameCurrent = newUsedClassesList.get(i);

                    if (!tableNameCurrent.equals(joinTableName))
                    {
                        requestInfo.addUsedClass(tableNameCurrent); // This one should be kept
                        continue; // Skip this, we already add on the top level clause
                    }

                    if (found)
                        result.append(',');

                    result.append(tableNameCurrent);
                    found = true;
                }

                result.append(" where ").append(condition);

                if (expression.isInverted())
                    result.append(")>0");
                else
                    result.append(")=0");

                result.append(")");
            }
        }
        else
        {

            if ((encoder.getRequestInfo().getDomain() != null)
                    && (encoder.getRequestInfo()._lockAliasCount)
                    && (encoder.getRequestInfo().getDomain().equals(
                            LySqlRequestEncoder.DOMAIN_FILTER_RELATION)))
            {
                //                               workaround for FFT504 encoding TABLE_A --join rel--> TABLE_A --join rel--> TABLE_A
                String alias = encoder.getRequestInfo().getTableAlias(joinClassInfo, targetRelation,
                        tableName);

                if (!tableName.equals(alias))
                {
                    encoder.getRequestInfo()._lockAliasCount = false; // temporary allow adding aliases
                    encoder.getRequestInfo().addTableAlias(null, joinClassInfo, sourceRelation,
                        tableName, alias, true, false);
                    encoder.getRequestInfo()._lockAliasCount = true; // restore
                }

                alias = encoder.getRequestInfo().getTableAlias(joinClassInfo, targetRelation,
                        joinTableName);

                if (!joinTableName.equals(alias))
                {
                    encoder.getRequestInfo()._lockAliasCount = false; // temporary allow adding aliases
                    encoder.getRequestInfo().addTableAlias(null, joinClassInfo, null, joinTableName,
                        alias, true, false);
                    encoder.getRequestInfo()._lockAliasCount = true; // restore
                }
            }

            // Add the join clause
            String condition = encoder.encodeJoinCondition(sourceRelation, _classBinding,
                    joinTableName, joinClassInfo);

            // Filtering condition now applies on the target relation of the
            // join class.
            expression = expression.cloneFilterExpression();
            expression.setFieldInfo(targetRelation);
            beginExpression = expression;

            LyFilter filterOnTargetRelation = null;

            if (!targetRelation.isMultiple())
            {

                if ((expression.getCondition() == LyFilterExpression.Condition.CONTAINS)
                        || (expression.getCondition() == LyFilterExpression.Condition.EQUALS))
                {
                    // May convert a multiple value to a single one...
                    Object   expValue = expression.getInitValue();
                    Object[] expValueArray;

                    if (!(expValue instanceof Object[]))
                        expValueArray = new Object[] { expValue };
                    else
                        expValueArray = (Object[]) expValue;

                    if (expValueArray.length > 1)
                        filterOnTargetRelation = new LyFilter(LyFilter.Type.OR);

                    for (int i = 0; i < expValueArray.length; i++)
                    {
                        expression = new LyFilterExpression(targetRelation, expression.isInverted(),
                                LyFilterExpression.Condition.EQUALS, expValueArray[i],
                                expression.getModifiers());

                        if (filterOnTargetRelation != null)
                            filterOnTargetRelation.addElement(expression);

                        value = expression.getValue();
                    }
                }
            }

            LyFieldBinding fieldBinding = targetRelation.getBinding();
            String         fieldCondition = null;

            if (fieldBinding instanceof LyFieldsBinding)
            {
                LyField[]  peerFields = joinClassBinding.getPeerFields(targetRelation);
                int        i, n = ((peerFields == null) ? 0 : peerFields.length);
                Class<?>[] classes = new Class[n];

                for (i = 0; i < n; i++)
                    classes[i] = peerFields[i].getValueClass();

                Object[] values;

                if (value instanceof Object[])
                    values = (Object[]) value;
                else
                {
                    values = new Object[1];
                    values[0] = value;
                }

                StringBuffer buffer = new StringBuffer();
                Object[]     fieldValues;
                String       id;
                int          v, nv = values.length;
                boolean      added;

                if (nv > 1)
                    buffer.append("(");

                for (v = 0; v < nv; v++)
                {
                    fieldValues = targetRelation.getPeerValues(values[v], classes, null);

                    if (v > 0)
                        buffer.append(") or (");

                    buffer.append("(");

                    for (i = 0, added = false; i < n; i++)
                    {
                        id = peerFields[i].getId();

                        if (joinPeerClass.getField(id) == null)
                            continue;

                        if (added)
                            buffer.append(") and (");

                        added = true;
                        buffer.append(encodeCondition(expression, peerFields[i],
                                ((fieldValues != null) ? fieldValues[i] : null), joinTableName,
                                classInfo));
                    }

                    buffer.append(")");
                }

                if (nv > 1)
                    buffer.append(")");

                fieldCondition = buffer.toString();
            }
            else
            {
                LyField peerField = joinClassBinding.getPeerField(targetRelation);

                if (filterOnTargetRelation != null)
                    fieldCondition = encoder.encodeCondition(filterOnTargetRelation, joinTableName,
                            classInfo);
                else
                    fieldCondition = encoder.encodeCondition(expression, peerField, value,
                            joinTableName, classInfo);
            }

            if (expression.isInverted())
            {

                // Add a second clause to encode equality cases
                if (filterOnTargetRelation != null)
                {

                    // FFT02614 CASE 1 : contains or equals on multiple field values in the filter
                    for (int i = 0; i < filterOnTargetRelation.getSize(); i++)
                    {
                        String oppositeSelectCondition = encodeOppositeSelectForJoinCondition(
                                joinClassBinding, true, true, joinTableName,
                                filterOnTargetRelation.getElement(i),
                                dataSource, encoder, sourceRelation, targetRelation, joinTableName,
                                joinClassInfo);

                        if (i > 0)
                            result.append(" or");

                        result.append(" ").append(oppositeSelectCondition);
                    }

                    if (beginExpression.getCondition() == LyFilterExpression.Condition.EQUALS)
                    {
                        LyFilter filter1 = new LyFilter(Type.AND);

                        for (int i1 = 0; i1 < filterOnTargetRelation.getSize(); i1++)
                        {
                            LyFilterExpression filterExpression = (LyFilterExpression)
                                filterOnTargetRelation.getElement(i1);

                            /*LyFilterExpression revertedExpression = new LyFilterExpression(targetRelation,
                                            !filterExpression.isInverted(), LyFilterExpression.Condition.EQUALS, filterExpression.getInitValue(), filterExpression.getModifiers());
                             */
                            filter1.addElement(filterExpression);
                        }

                        String oppositeSelectCondition = encodeOppositeSelectForJoinCondition(
                                joinClassBinding, false, false, joinTableName, filter1, dataSource,
                                encoder, sourceRelation, targetRelation, joinTableName,
                                joinClassInfo);

                        result.append(" or ").append(oppositeSelectCondition);
                    }
                }
                else
                {

                    if (beginExpression.getCondition() == LyFilterExpression.Condition.CONTAINS)
                    {
                        // FFT02598  in that case the condition <> on the join is *NOT* sufficient for multiple relations
                        // ! contains obj1 means that if related to obj1 and obj2 , it should be removed
                        // similar to EQUALS when NOT inverted
                        String oppositeSelectCondition = encodeOppositeSelectForJoinCondition(
                                joinClassBinding, true, true, joinTableName, beginExpression,
                                dataSource, encoder, sourceRelation, targetRelation, joinTableName,
                                joinClassInfo);
                        result.append(oppositeSelectCondition);
                    }
                    else if (beginExpression.getCondition()
                            == LyFilterExpression.Condition.CONTAINS)
                    {
                        // Similar issue as FFT02598 just above
                        String oppositeSelectCondition = encodeOppositeSelectForJoinCondition(
                                joinClassBinding, true, true, joinTableName, beginExpression,
                                dataSource, encoder, sourceRelation, targetRelation, joinTableName,
                                joinClassInfo);
                        result.append(oppositeSelectCondition);
                    }
                }
            }
            else
            {

                // Add a second clause to encode equality cases
                if (filterOnTargetRelation != null)
                {

                    // FFT02614 CASE 1 : contains or equals on multiple field vqlues in the filter
                    for (int i = 0; i < filterOnTargetRelation.getSize(); i++)
                    {
                        String oppositeSelectCondition = encodeOppositeSelectForJoinCondition(
                                joinClassBinding, false, false, joinTableName,
                                filterOnTargetRelation.getElement(i),
                                dataSource, encoder, sourceRelation, targetRelation, joinTableName,
                                joinClassInfo);

                        if (i > 0)
                            result.append(" and");

                        result.append(" ").append(oppositeSelectCondition);
                    }

                    if (beginExpression.getCondition() == LyFilterExpression.Condition.EQUALS)
                    {
                        String oppositeSelectCondition = encodeOppositeSelectForJoinCondition(
                                joinClassBinding, true, true, joinTableName, filterOnTargetRelation,
                                dataSource, encoder, sourceRelation, targetRelation, joinTableName,
                                joinClassInfo);

                        result.append(" and ").append(oppositeSelectCondition);
                    }
                }
                else
                {

                    if (condition != null)
                    {
                        result.append("(");
                        result.append(condition);
                        result.append(")");
                    }

                    if (fieldCondition != null)
                    {
                        result.append(" and (");
                        result.append(fieldCondition);
                        result.append(")");
                    }

                    if ((!expression.isInverted())
                            && (beginExpression.getCondition()
                                == LyFilterExpression.Condition.EQUALS))
                    {
                        String oppositeSelectCondition = encodeOppositeSelectForJoinCondition(
                                joinClassBinding, true, true, joinTableName, beginExpression,
                                dataSource, encoder, sourceRelation, targetRelation, joinTableName,
                                joinClassInfo);

                        result.append(" and ").append(oppositeSelectCondition);
                    }
                }
            }
        }

        clearChildRequestInfo(encoder);

        return ((result.length() > 0) ? result.toString() : null);
    } // end method encodeJoinCondition

    protected String encodeOppositeSelectForJoinCondition(LyClassBinding joinClassBinding,
        boolean notExists, boolean revertExpression, String joinTableName,
        LyFilterElement filterElement, LyDataSource dataSource, LySqlRequestEncoder encoder,
        LyRelationFieldInfo sourceRelation, LyRelationFieldInfo targetRelation, String tableName,
        LyClassInfo joinClassInfo)
    {
        StringBuffer result = new StringBuffer();

        LyField[]       peerFields = joinClassBinding.getPeerFields(sourceRelation);
        int             i, n = ((peerFields == null) ? 0 : peerFields.length);
        Vector<LyField> sourceidFields = new Vector<LyField>();
        StringBuffer    idFieldsCondition = new StringBuffer();

        for (i = 0; i < n; i++)
        {
            sourceidFields.add(peerFields[i]);

            if (i > 0)
                idFieldsCondition.append(", ");

            if (joinTableName != null)
                idFieldsCondition.append(joinTableName).append('.');

            idFieldsCondition.append(peerFields[i].getId());
        }

        // use the exact reverted expression

        LyFilter oppositeFilter = new LyFilter();

        if (revertExpression)
        {

            if (filterElement instanceof LyFilter)
            {
                LyFilter filter1 = (LyFilter) filterElement;
                oppositeFilter = new LyFilter((filter1.getType() == Type.AND) ? Type.OR : Type.AND);

                for (int i1 = 0; i1 < filter1.getSize(); i1++)
                {
                    LyFilterExpression filterExpression = (LyFilterExpression) filter1.getElement(
                            i1);
                    LyFilterExpression revertedExpression = new LyFilterExpression(targetRelation,
                            !filterExpression.isInverted(), LyFilterExpression.Condition.EQUALS,
                            filterExpression.getInitValue(), filterExpression.getModifiers());

                    oppositeFilter.addElement(revertedExpression);
                }
            }
            else
            {
                LyFilterExpression filterExpression = (LyFilterExpression) filterElement;
                LyFilterExpression revertedExpression = new LyFilterExpression(targetRelation,
                        !filterExpression.isInverted(), LyFilterExpression.Condition.EQUALS,
                        filterExpression.getInitValue(), filterExpression.getModifiers());

                oppositeFilter.addElement(revertedExpression);
            }
        }
        else
        {

            if (filterElement instanceof LyFilter)
                oppositeFilter = (LyFilter) filterElement;
            else
                oppositeFilter.addElement(filterElement);
        }

        LySqlRequestEncoder subEncoder = encoder.getRequestEncoder(dataSource);
        LySqlRequestInfo    subRequestInfo = encoder.setChildRequestInfo(subEncoder,
                LySqlRequestInfo.Join.NOTHING);

        subEncoder.setClassBinding(encoder.getClassBinding());

        if ((encoder.getRequestInfo().getDomain() != null)
                && (encoder.getRequestInfo()._lockAliasCount)
                && (encoder.getRequestInfo().getDomain().equals(
                        LySqlRequestEncoder.DOMAIN_FILTER_RELATION)))
        {
            //                       workaround for FFT504 encoding TABLE_A --join rel--> TABLE_A --join rel--> TABLE_A and equals condition
            String alias = encoder.getRequestInfo().getTableAlias(joinClassInfo, targetRelation,
                    tableName);

            if (!tableName.equals(alias))
            {
                subEncoder.addTableAlias(null, tableName, alias); // add twice to force usage
                subEncoder.addTableAlias(null, tableName, alias);
                subRequestInfo.setIgnoreInSelectFrom(tableName, alias, true); // from parent select (re-use)
            }
        }

        String tableAlias = computeTableAlias(joinTableName);

        // Force the use of aliases : add alias !
        subEncoder.addTableAlias(null, joinTableName, tableAlias);

        String oppositeSelect = subEncoder.encodeSelectRequest(null, null, sourceidFields, null,
                oppositeFilter, -1);
        String condition2 = subEncoder.encodeJoinCondition(sourceRelation, _classBinding,
                joinTableName, null);
        int    n1 = oppositeSelect.indexOf("where (") + "where (".length();

        //result.append(" and (").append(idFieldsCondition).append(") not in (").append(oppositeSelect).append(")");
        if (notExists)
            result.append("not ");

        result.append("exists (").append(oppositeSelect.substring(0, n1)).append(condition2).append(
            " and ").append(oppositeSelect.substring(n1)).append(")");
        encoder.clearChildRequestInfo(subEncoder);

        boolean removeUsedClass = notExists;
        if (filterElement instanceof LyFilterExpression)
        {
            LyFilterExpression exp = (LyFilterExpression)filterElement;
            if (exp.getCondition().equals(LyFilterExpression.Condition.EQUALS) && !exp.isInverted())
                removeUsedClass = false;
        }
        if (removeUsedClass)
        	getRequestInfo().removeUsedClass(encoder.getTableName(encoder.getDataClass()));

        return result.toString();
    }

    /**
     * Encode a filter expression that applies on a reverse relation.
     *
     * @param  expression        a filter expression
     * @param  reverseDaemon     the reverse daemon
     * @param  relationFieldInfo the reverse relation
     * @param  tableName         the current table name
     * @return null for this implementation
     */
    protected String encodeReverseDaemonCondition(LyFilterExpression expression,
        LyReverseDaemon reverseDaemon, LyRelationFieldInfo relationFieldInfo, String tableName,
        LyClassInfo classInfo)
    {
        return null;
    }

    /**
     * Encode a JOIN clause for the tables that are used to sort on name for relations criterions.
     *
     * @param  sort current sort.
     * @return an array of strings (one for each relation condition): the first string is the name
     *         of the related table and the second one is the join clause.
     */
    protected String[][] encodeRelationSortCondition(LySort sort, boolean multipleClasses,
        Map<LyField, String> aliases)
    {

        if (sort == null)
            return null;

        LyInfoList<LyRelationFieldInfo> relations = getNameRelationSort(sort);

        if ((relations == null) || (relations.getSize() == 0))
            return null;

        // encode 1 condition for each relation
        int                     i, j, n = relations.getSize();
        String[][]              conditions = new String[n][5];
        LyRelationFieldInfo     relation;
        LyClassInfo             classInfo;
        String                  tableName;
        String                  unionCondition;
        LyClass                 peerClass;
        LyClassBinding          classBinding;
        StringBuffer            sb;
        Vector<String>          unions;
        LySqlRequestEncoder     encoder;
        String                  alias = null;
        LyInfoList<LyClassInfo> classes;
        LySqlRequestInfo        requestInfo;

        for (i = 0; i < n; i++)
        {
            unions = null;
            relation = relations.getInfo(i);

            // get table name of the relation
            peerClass = getPeerClass(relation, _classBinding);
            tableName = peerClass.getName();

            // we have to get the id
            classInfo = relation.getClassInfo(0);
            classBinding = classInfo.getBinding();

            if (multipleClasses)
            {
                classes = new LyInfoList<LyClassInfo>();
                classes.addInfo(classInfo);
                computeTableAliases(classes);

                alias = getTableAlias(classInfo, classBinding.getBinding());
            }

            conditions[i][0] = classBinding.getBinding();
            conditions[i][1] = encodeJoinCondition(relation, classBinding, tableName, classInfo);
            conditions[i][2] = alias;
            conditions[i][3] = tableName; // source table name
            conditions[i][4] = getTableAlias(classInfo, tableName);
            ; // source table alias
            aliases.put(relation, alias);

            // check if we have union bindings then add some union clauses
            for (j = 0; j < relation.getClassInfoCount(); j++)
            {
                classInfo = relation.getClassInfo(j);
                classBinding = classInfo.getBinding();

                if (classBinding instanceof LyUnionBinding)
                {
                    encoder = getRequestEncoder(getDataSource(classBinding.getPeerClass()), false);
                    encoder.setClassBinding(classBinding);
                    encoder.setJoinPeerClasses(getPeerClasses(classBinding.getPeerClass()));
                    requestInfo = setChildRequestInfo(encoder,
                            LySqlRequestInfo.Join.NEW_ALIAS_ONLY);


                    requestInfo.setAliasContext(null, classInfo, null);
                    unionCondition = encoder.encodeCondition((LyUnionBinding) classBinding,
                            classInfo, true);
                    clearChildRequestInfo(encoder);

                    if ((unionCondition != null) && (unionCondition.trim().length() > 0))
                    {

                        if (unions == null)
                            unions = new Vector<String>();

                        unions.add(unionCondition);
                    }
                }
            }

            // we may have to add union clauses
            if (unions != null)
            {
                sb = new StringBuffer();
                sb.append('(').append(conditions[i][1]).append(" and ");

                if (unions.size() > 1)
                    sb.append('(');

                for (j = 0; j < unions.size(); j++)
                {

                    if (j > 0)
                        sb.append(" or ");

                    sb.append(unions.get(j));
                }

                if (unions.size() > 1)
                    sb.append(')');

                sb.append(')');

                conditions[i][1] = sb.toString();
            }
        }

        return conditions;
    } // end method encodeRelationSortCondition

    /**
     * Retrieve the related data source of a given peer class.
     *
     * @param  peerClass a peer class.
     * @return a data source.
     */
    protected LyDataSource getDataSource(LyClass peerClass)
    {
        LyDataSource  dataSource = null;
        LyApplication application = _provider.getApplication();

        try
        {
            dataSource = application.getDataSource(peerClass.getId());
        }
        catch (IllegalStateException e)
        {

            if (peerClass instanceof LyUnionClass)
            {
                LyUnionClass   unionClass = (LyUnionClass) peerClass;
                LyUnionBinding binding = unionClass.getUnionBinding();

                dataSource = getDataSource(binding.getClassBinding1().getPeerClass());

                if (dataSource == null)
                    dataSource = getDataSource(binding.getClassBinding2().getPeerClass());
            }
        }

        return dataSource;
    }

    /**
     * Builds a request encoder for specified data source. If data source has not the same provider
     * as the current one, no encoder is created.
     *
     * @param  dataSource            Data source used to create the request encoder.
     * @param  assignJoinPeerClasses true if the current joinPeerClasses must be assigned to the
     *                               returned encoder
     * @return the created request encoder, or null if data provider for specified data source is
     *         not the same as the current one.
     */
    protected LySqlRequestEncoder getRequestEncoder(LyDataSource dataSource,
        boolean assignJoinPeerClasses)
    {
        LyClass             peerClass = dataSource.getDataClass();
        Collection<LyClass> peerClasses = getPeerClasses(peerClass);

        for (Iterator<LyClass> it = peerClasses.iterator(); it.hasNext();)
        {
            peerClass = it.next();
            dataSource = peerClass.getDataSource();

            LyDataProvider dataProvider = dataSource.getDataProvider();

            // If data provider of the target class is not the same as current class
            // filter is ignored and will be applied after loading (in memory).
            if (_provider != dataProvider)
            {
                LyEnvironment environment = getEnvironment();

                if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
                    environment.logOut("Trying to create an sql command on classes "
                        + "managed by different data providers. " + "Filter on class \""
                        + getTableName(_class) + "\" is ignored in the sql encoder.");

                return null;
            }
        }

        LySqlRequestEncoder encoder = (LySqlRequestEncoder) _provider.getRequestEncoder(dataSource);

        if (assignJoinPeerClasses && (_joinPeerClasses != null))
            encoder.setJoinPeerClasses(_joinPeerClasses);

        return encoder;
    } // end method getRequestEncoder

    protected ArrayList<LyClass> getPeerClasses(LyClass cls)
    {
        ArrayList<LyClass> result = new ArrayList<LyClass>();

        if (cls instanceof LyUnionClass)
        {
            LyUnionBinding      unionBinding = ((LyUnionClass) cls).getUnionBinding();
            LyClassBinding      cls1Binding = unionBinding.getClassBinding1();
            LyClass             cls1 = cls1Binding.getPeerClass();
            Collection<LyClass> result1 = getPeerClasses(cls1);

            result.addAll(result1);

            LyClassBinding      cls2Binding = unionBinding.getClassBinding2();
            LyClass             cls2 = cls2Binding.getPeerClass();
            Collection<LyClass> result2 = getPeerClasses(cls2);

            result.addAll(result2);
        }
        else
            result.add(cls);

        return result;
    }

    protected ArrayList<LyBinding> getPeerBindings(LyBinding binding)
    {
        ArrayList<LyBinding> result = new ArrayList<LyBinding>();

        if (binding instanceof LyUnionBinding)
        {
            LyUnionBinding        unionBinding = (LyUnionBinding) binding;
            LyClassBinding        cls1Binding = unionBinding.getClassBinding1();
            Collection<LyBinding> result1 = getPeerBindings(cls1Binding);

            result.addAll(result1);

            LyClassBinding cls2Binding = unionBinding.getClassBinding2();

            Collection<LyBinding> result2 = getPeerBindings(cls2Binding);

            result.addAll(result2);
        }
        else
            result.add(binding);

        return result;
    }

    /**
     * Builds a request encoder for specified data source. If data source has not the same provider
     * as the current one, no encoder is created.
     *
     * @param  dataSource Data source used to create the request encoder.
     * @return the created request encoder, or null if data provider for specified data source is
     *         not the same as the current one.
     */
    protected LySqlRequestEncoder getRequestEncoder(LyDataSource dataSource)
    {
        return getRequestEncoder(dataSource, true);
    }

    protected String encodeCondition(LyFilterExpression expression, LyField peerField, Object value,
        String tableName, LyClassInfo classInfo)
    {
        getRequestInfo().getCurrentTask().addInfo("BEGIN encodeCondition (LyFilterExpression 2) : "
            + expression);

        LyFieldInfo                  fieldInfo = expression.getFieldInfo();
        LyFilterExpression.Condition condition = expression.getCondition();
        boolean                      inverted = expression.isInverted();
        String                       id = peerField.getId();
        StringBuffer                 result = null;
        String                       val;
        boolean                      caseSensitive = expression.isCaseSensitive();
        LyEnvironment                environment = getEnvironment();
        boolean                      ignoreAccent = ((environment != null)
                && environment.testResource("LY_IGNORE_TEXT_ACCENT"));

        // If case-sensitive is not supported by database, case-sensitive is always true
        if (!supportCaseSensitive())
            caseSensitive = true;

        // Manage mode with no accent
        if (ignoreAccent && (fieldInfo instanceof LyTextFieldInfo) && (value instanceof String))
            value = LyString.getSimpleString((String) value);

        // generating values for the request
        // case of a multiple RELATION
        if ((fieldInfo instanceof LyRelationFieldInfo) && (value instanceof Object[]))
        {
            LyRelationFieldInfo relation = (LyRelationFieldInfo) fieldInfo;

            val = "";

            if (relation.isOrdered())
                val = getSqlValue((LyColumnField) peerField, value, TYPE_ARRAY);
        }
        // case of a multiple CHOICE
        else if (value instanceof java.util.BitSet)
        {
            if (condition == LyFilterExpression.Condition.CONTAINS)
                val = "";
            else
                val = fieldInfo.getPeerValue(value, peerField.getValueClass(), null).toString();
        }
        // case of a DATE
        else if (value instanceof java.util.Date)
            val = getSqlValue((LyColumnField) peerField, value, TYPE_DATE);
        // case of a BOOLEAN
        else if (value instanceof Boolean)
            val = getSqlValue((LyColumnField) peerField, value, TYPE_BOOLEAN);
        // case of a NUMBER
        else if ((value instanceof java.lang.Number) && !(fieldInfo instanceof LyChoiceFieldInfo))
            val = getSqlValue((LyColumnField) peerField, value, TYPE_NUMBER);
        // common case
        else
        {
            val = (String) fieldInfo.getPeerValue(value, String.class, null);
            val = getSqlValue((LyColumnField) peerField, val, TYPE_STRING);
        }

        // If table name must be used, add it before field id
        if (tableName == null)
            tableName = getTableName(peerField);

        if (tableName == null)
            tableName = getTableName(_class);

        // case of multilingual text and filtering on one lang only
        if ((fieldInfo instanceof LyTextFieldInfo)
        	&& ((LyTextFieldInfo) fieldInfo).isMultilingual()
        	&& !expression.hasModifier(LyFilterExpression.MOD_MULTILINGUAL))
        {
            LyApplication application = _provider.getApplication();
            LyDataSession currentSession = application.getCurrentSession();
            String        lang = currentSession.getEnvironment().getLanguage();

            id = getSelectedLangField(lang, peerField, tableName, false);
            if (tableName != null)
            	getRequestInfo().addUsedClass(tableName);
        }
        else if (tableName != null)
        {
            id = QUOTE_CHAR_LEFT + getTableAlias(classInfo, tableName) + QUOTE_CHAR_RIGHT + "."
                + QUOTE_CHAR_LEFT + id + QUOTE_CHAR_RIGHT;
            getRequestInfo().addUsedClass(tableName);
        }

        // generating the request
        if ((val == null) && (condition == LyFilterExpression.Condition.EQUALS))
        {
            result = new StringBuffer();
            result.append(id);

            if (inverted)
                result.append(" is not null");
            else
                result.append(" is null");
        }
		else if ((val == null) && inverted
				&& ((condition == LyFilterExpression.Condition.CONTAINS)
						|| (condition == LyFilterExpression.Condition.STARTS_WITH)
						|| (condition == LyFilterExpression.Condition.ENDS_WITH)))
		{
			// Does not contain null, does not start with null, does not end with null :
			// do not return any element !
			result = new StringBuffer();
			result.append(id);
			result.append(" is not null and ");
			result.append(id);
			result.append(" is null");
		}
        else if (val != null)
        {
            String valStrEncoded;

            result = new StringBuffer();

            // case of a STARTS_WITH request
            if (condition == LyFilterExpression.Condition.STARTS_WITH)
            {
                //case of invertion
                if (inverted)
                    result.append("not(");

                if (!caseSensitive)
                    result.append("lower(");

                result.append(id);

                if (!caseSensitive)
                    result.append(")");

                result.append(" like '");

                if (!caseSensitive)
                    valStrEncoded = value.toString().toLowerCase();
                else
                    valStrEncoded = value.toString();

                valStrEncoded = encodeLikeCondition(valStrEncoded);
                result.append(valStrEncoded);
                result.append("%'");

                if (_escapeSequence != null)
                    result.append(_escapeSequence); // Using ESCAPE '<getEscapeCharacter>' in like conditions

                _escapeSequence = null;

				// case of invertion
				if (inverted)
				{
					result.append(")");
					result.append(" or ");
					result.append(id);
					result.append(" is null ");
				}
			}
			// case of a ENDS_WITH request
            else if (condition == LyFilterExpression.Condition.ENDS_WITH)
            {
                //case of invertion
                if (inverted)
                    result.append("not(");

                if (!caseSensitive)
                    result.append("lower(");

                result.append(id);

                if (!caseSensitive)
                    result.append(")");

                result.append(" like '%");

                if (!caseSensitive)
                    valStrEncoded = value.toString().toLowerCase();
                else
                    valStrEncoded = value.toString();

                valStrEncoded = encodeLikeCondition(valStrEncoded);
                result.append(valStrEncoded);
                result.append("'");

                if (_escapeSequence != null)
                    result.append(_escapeSequence); // Using ESCAPE '<getEscapeCharacter>' in like conditions

                _escapeSequence = null;

				// case of invertion
				if (inverted)
				{
					result.append(")");
					result.append(" or ");
					result.append(id);
					result.append(" is null ");
				}
			}
			// case of a CONTAINS request
            else if (condition == LyFilterExpression.Condition.CONTAINS)
            {

                //case of invertion
                if (inverted)
                    result.append("not(");

                // case of a multiple RELATION
                if ((fieldInfo instanceof LyRelationFieldInfo) && (value instanceof Object[]))
                {
                    LyRelationFieldInfo relation = (LyRelationFieldInfo) fieldInfo;

                    // case of an ordered RELATION
                    if (relation.isOrdered())
                    {
                        result.append(id).append(" like '%").append(encodeLikeCondition(val))
                        .append("%'");

                        if (_escapeSequence != null)
                            result.append(_escapeSequence); // Using ESCAPE '<getEscapeCharacter>' in like conditions

                        _escapeSequence = null;
                    }
                    else if (value != null)
                    { // non ordered relation : test every value

                        Object[] values = (Object[]) value;
                        String   s;

                        for (int k = 0; k < values.length; k++)
                        {
                            s = values[k].toString();
                            s = encodeLikeCondition(s);

                            result.append("(");

                            // case of string a the first place
                            result.append("(").append(id).append(" like '").append(s).append(
                                _valueSep).append("%'");

                            if (_escapeSequence != null)
                                result.append(_escapeSequence); // Using ESCAPE '<getEscapeCharacter>' in like conditions

                            _escapeSequence = null;
                            result.append(") or ");

                            // case of string in the middle of the value
                            result.append("(").append(id).append(" like '%").append(_valueSep)
                            .append(s).append(_valueSep).append("%'");

                            if (_escapeSequence != null)
                                result.append(_escapeSequence); // Using ESCAPE '<getEscapeCharacter>' in like conditions

                            _escapeSequence = null;
                            result.append(") or ");

                            // case of string at the last place
                            result.append("(").append(id).append(" like '%").append(_valueSep)
                            .append(s).append("'");

                            if (_escapeSequence != null)
                                result.append(_escapeSequence); // Using ESCAPE '<getEscapeCharacter>' in like conditions

                            _escapeSequence = null;
                            result.append(") or ");

                            // case of a unique value inside the column
                            result.append("(").append(id).append(" like '").append(s).append("'");

                            if (_escapeSequence != null)
                                result.append(_escapeSequence); // Using ESCAPE '<getEscapeCharacter>' in like conditions

                            _escapeSequence = null;

                            result.append("))");

                            if (k != (values.length - 1))
                                result.append(" and ");
                        }
                    }
                }
                // case of a multiple CHOICE
                else if (value instanceof java.util.BitSet)
                {
                    // collecting bits of the bitset
                    BitSet             bs = (BitSet) value;
                    ArrayList<Integer> bits = new ArrayList<Integer>();
                    LyChoiceItem       item;
                    short              s;
                    int                k, n = bs.size();

                    for (k = 0; k < n; k++)
                        if (bs.get(k))
                        {
                            item = ((LyChoiceFieldInfo) fieldInfo).getChoiceItem((short) k);
                            s = item.getShortValue();
                            bits.add(new Integer(s));
                        }

                    // generating the request
                    int bit;

                    for (Iterator<Integer> it = bits.iterator(); it.hasNext();)
                    {
                        bit = it.next().intValue();
                        bit = (1 << bit);
                        result.append(encodeBitAnd(id, Integer.toString(bit)));

                        if (it.hasNext())
                            result.append(" and ");
                    }
                }
                else if (fieldInfo instanceof LyTextFieldInfo)
                {

                    if (!caseSensitive)
                        result.append("lower(");

                    result.append(id);

                    if (!caseSensitive)
                        result.append(")");

                    result.append(" like '%");

                    if (!caseSensitive)
                        valStrEncoded = value.toString().toLowerCase();
                    else
                        valStrEncoded = value.toString();

                    valStrEncoded = encodeLikeCondition(valStrEncoded);
                    result.append(valStrEncoded);
                    result.append("%'");

                    if (_escapeSequence != null)
                        result.append(_escapeSequence); // Using ESCAPE '<getEscapeCharacter>' in like conditions

                    _escapeSequence = null;
                }
                else
                {
                    // relation contains on Id has no sense : use equals
                    // Common case
                    result.append(id);
                    result.append(getCondition(LyFilterExpression.Condition.EQUALS, inverted));
                    result.append(val);
                    getEnvironment().logWarning(
                        "LySqlRequestEncoder : received a CONTAINS condition on a physical value. RDBMS can not have multiple value on a field. Convert CONTAINS into EQUALS : "
                        + result);
                }

                // case of invertion
				if (inverted)
				{
					result.append(")");
					result.append(" or ");
					result.append(id);
					result.append(" is null ");
				}
			}
			// case of an EQUALS request on a multiple RELATION
            else if ((condition == LyFilterExpression.Condition.EQUALS)
                    && (fieldInfo instanceof LyRelationFieldInfo) && (value instanceof Object[]))
            {
                LyRelationFieldInfo relation = (LyRelationFieldInfo) fieldInfo;

                // case of an ordered RELATION
                if (relation.isOrdered())
                {
                    result.append(id);
                    result.append(getCondition(condition, inverted));
                    result.append(val);
                }
                // non ordered RELATION : test every value
                else if (value != null)
                {
                    int      length = 0;
                    String   s;
                    Object[] values = (Object[]) value;

                    //case of invertion
                    if (inverted)
                        result.append("not(");

                    for (int k = 0; k < values.length; k++)
                    {
                        s = values[k].toString();
                        result.append("(").append(id).append(" like '%").append(s).append("%')");
                        result.append(" and ");
                        // compute the total length of the value
                        length += s.length();
                    }
                    // add the number of value separators for the total length of the value
                    length += (values.length - 1);

                    // adding a test on the length of the value
                    result.append("(LENGTH(").append(id).append(") = ").append(length).append(")");

                    // case of invertion
                    if (inverted)
                        result.append(")");
                }
            }
            else if (fieldInfo instanceof LyTextFieldInfo)
            {

                if (!caseSensitive)
                    result.append("lower(");

                result.append(id);

                if (!caseSensitive)
                    result.append(")");

                result.append(getCondition(condition, inverted));

                if (!caseSensitive)
                    result.append(val.toString().toLowerCase());
				else
					result.append(val);

				if (inverted)
				{
					result.append(" or ");
					result.append(id);
					result.append(" is null ");
				}
			}
			else if ((condition == LyFilterExpression.Condition.EQUALS)
					&& (fieldInfo instanceof LyDateFieldInfo)
                    && (((LyDateFieldInfo) fieldInfo).getType() == LyDateFieldInfo.Type.DATE))
            {
                result.append("(");
                result.append(id);
                result.append(getCondition(LyFilterExpression.Condition.GREATER_OR_EQUALS,
                        inverted));
                result.append(val);

                if (inverted)
                    result.append(") or (");
                else
                    result.append(") and (");

                result.append(id);
                result.append(getCondition(LyFilterExpression.Condition.LOWER_THAN, inverted));
                val = getSqlValue((LyColumnField) peerField,
                        new java.util.Date(
                            ((java.util.Date) value).getTime() + (24 * 60 * 60 * 1000)), TYPE_DATE);
                result.append(val);
                result.append(")");
            }
            // Common case
            else
            {
                result.append(id);
                result.append(getCondition(condition, inverted));
                result.append(val);
            }
        }

        String ret = ((result == null) ? null : result.toString());
        getRequestInfo().getCurrentTask().addInfo("END encodeCondition (LyFilterExpression 2)");
        getRequestInfo().getCurrentTask().addSql(ret);

        return (ret);
    } // end method encodeCondition

    /**
     * Encodes binary operation.
     * @param param1 First param.
     * @param param2 Second param.
     */
    protected String encodeBitAnd(String param1, String param2)
    {
        return ("((" + param1 + " & " + param2 + ") != 0)");
    }

    /**
     */
    protected String encodeCondition(LyObjectId objectId)
    {

        if (objectId == null)
            return null;

        getRequestInfo().getCurrentTask().addInfo("BEGIN encodeCondition (LyObjectId)");

        LyEnvironment             environment = getEnvironment();
        StringBuffer              condition;
        Vector<? extends LyField> idFields;
        String                    val;
        String[]                  values;
        int                       i, n, p;
        LyColumnField             field;

        idFields = _class.getIdFields();
        n = idFields.size();

        if (n > 1)
        {
            values = objectId.toStrings();

            if (n != values.length)
                values = LyString.stringToList(objectId.toString(), LyObjectId.OLD_FIELD_SEP);
        }
        else
        {
            values = new String[1];
            values[0] = objectId.toString();
        }

        p = values.length;

        if (n != p)
            throw new IllegalStateException("BF AppEngine : "
                + environment.getMessage("LY_ER_CONDITION_OBJECT_ID", objectId.toName(),
                    _class.getId()));

        condition = new StringBuffer();

        for (i = 0; i < n; i++)
        {
            field = (LyColumnField) idFields.elementAt(i);
            val = getSqlValue(field, values[i], TYPE_STRING);

            if (i != 0)
                condition.append(" and ");

            if (n != 1)
                condition.append("(");

            condition.append(QUOTE_CHAR_LEFT).append(field.getId()).append(QUOTE_CHAR_RIGHT);

            // generating the request
            if ((val == null) || (val.length() == 0))
                condition.append(" is null");
            else
            {
                condition.append("=");
                condition.append(val);
            }

            if (n != 1)
                condition.append(")");
        }

        getRequestInfo().getCurrentTask().addInfo("END encodeCondition (LyObjectId)");
        getRequestInfo().getCurrentTask().addSql(condition.toString());

        return condition.toString();
    } // end method encodeCondition

    /**
     * Encodes a count-type request.
     *
     * @param  encoder to send join count requests (may be null).
     * @param  filter  indicates filtering criterion (may be null).
     * @return the generated request.
     */
    protected String encodeCountRequest(LyUnionRequestEncoder encoder, LyFilter filter)
    {
        StringBuffer            request;
        String                  filterCondition;
        String                  classCondition;
        String                  unionCondition;
        boolean                 and = false;
        LyInfoList<LyClassInfo> targetClasses = new LyInfoList<LyClassInfo>(false);
        LyFilter                classFilter = ((_classBinding == null) ? null
                                                                       : _classBinding.getFilter());
        Vector<LyFilter>        unionFilters = null;

        if (classFilter != null)
            classFilter = classFilter.getPeerFilter();

        getRequestInfo().clearCurrentUsedClasses();
        request = new StringBuffer();
        getCountTargetClasses(classFilter, filter, targetClasses);

        if (_classBinding instanceof LyUnionBinding)
            unionFilters = getFilters((LyUnionBinding) _classBinding);

        if (unionFilters != null)
        {

            // Check into union filters if some additional classes are required
            for (int i = 0; i < unionFilters.size(); i++)
                getTargetClasses(unionFilters.elementAt(i), targetClasses);
        }

        // now compute the table aliases
        LyClassInfo   classInfo;
        String        targetTableName;
        String        tableAlias;
        List<LyClass> peerClasses = getPeerClasses(_class);

        for (Iterator<LyClass> it = peerClasses.iterator(); it.hasNext();)
        {
            targetTableName = getTableName(it.next());
            tableAlias = computeTableAlias(targetTableName);

            // add the table alias for the current logical class (null)
            addTableAlias(null, targetTableName, tableAlias);
        }

        // If there is at least one target table, we must prefix the 'SELECT'
        // clause parameters with the name of the current table.
        boolean writeTableName = (targetClasses.getSize() > 0) && (_joinPeerClasses == null);
        String  tableName = (writeTableName ? getTableName(_class) : null);

        getRequestInfo().clearCurrentUsedClasses();
        getRequestInfo().addUsedClass(tableName);

        LyEnvironment environment = getEnvironment();

        if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
            environment.logOut("* select count " + getTableName(_class) + " _joinPeerClasses : "
                + ((_joinPeerClasses == null) ? 0 : _joinPeerClasses.size()) + " targetClasses "
                + ((targetClasses == null) ? 0 : targetClasses.getSize()));

        //               ** analysis starts : just to detect aliases **
        getRequestInfo().setLockAliasCount(false);
        encodeCondition(classFilter, tableName, null); //classCondition
        encodeCondition(filter, tableName, null); //filterCondition
        unionCondition = encodeCondition(encoder, null); //unionCondition
        //               ** analysis ends : do it for real **
        getRequestInfo().setLockAliasCount(true);
        //               Begin select count encoding
        request.append(encodeSelectCountBegin(filter, writeTableName));

        classCondition = encodeCondition(classFilter, tableName, null);
        filterCondition = encodeCondition(filter, tableName, null);
        unionCondition = encodeCondition(encoder, null);

        // FFT02616 : check if actually being used for performance
        getRequestInfo().setLockAliasCount(false);
        HashMap<String, String> aliasMapUsed = getRequestInfo().getAllAliases();
        for (Iterator<String> iter = aliasMapUsed.keySet().iterator(); iter.hasNext();)
        {
            tableAlias = iter.next();

            if (aliasMapUsed.size() > 0)
            {
                boolean used= false;

                if (!used && (classCondition != null) && (classCondition.indexOf(tableAlias)>=0))
                    used = true;
                if (!used && (filterCondition != null) && (filterCondition.indexOf(tableAlias)>=0))
                    used = true;
                if (!used && (unionCondition != null) && (unionCondition.indexOf(tableAlias)>=0))
                    used = true;
                if (!used)
                    getRequestInfo().skipUnusedAliases(tableAlias);
            }
        }

        // IN fact all used aliases -- table names should be added in the from clause
        // Store what we have from former code below (compatibility) then complete afterwards.
        HashMap<String, String> foundAliases = new HashMap<String, String>();
        Vector<String>          foundTableNames = new Vector<String>(); // Avoid duplicated
        boolean                 optimize = true; //getEnvironment().testResource("LY_OPTIMIZE_SQL");
        ArrayList<String>       usedTablesNames = getRequestInfo().getCurrentUsedClasses();

        if (_joinPeerClasses == null)
        {
            targetTableName = getTableName(_class);
            tableAlias = getTableAlias(null, targetTableName);

            request.append(QUOTE_CHAR_LEFT).append(targetTableName).append(QUOTE_CHAR_RIGHT);

            if (!tableAlias.equals(targetTableName))
            {
                request.append(" ").append(tableAlias);
                foundAliases.put(tableAlias, targetTableName);
            }

            foundTableNames.addElement(targetTableName);
        }
        else
        {
            boolean first = true;
            LyClass peerClass;

            for (Iterator<LyClass> it = _joinPeerClasses.iterator(); it.hasNext();)
            {
                peerClass = it.next();
                targetTableName = getTableName(peerClass);
                tableAlias = getTableAlias(null, targetTableName);

                if (foundTableNames.contains(targetTableName) && tableAlias.equals(targetTableName))
                    continue;

                if ((optimize) && (!usedTablesNames.contains(targetTableName)))
                    continue;

                if (!foundTableNames.contains(targetTableName))
                    foundTableNames.addElement(targetTableName);

                if (first)
                {
                    request.append(QUOTE_CHAR_LEFT).append(targetTableName).append(
                        QUOTE_CHAR_RIGHT);

                    if (!tableAlias.equals(targetTableName))
                    {
                        request.append(" ").append(tableAlias);
                        foundAliases.put(tableAlias, targetTableName);
                    }

                    first = false;
                }
                else
                {
                    request.append(", ").append(QUOTE_CHAR_LEFT).append(targetTableName).append(
                        QUOTE_CHAR_RIGHT);

                    if (!tableAlias.equals(targetTableName))
                    {
                        request.append(" ").append(tableAlias);
                        foundAliases.put(tableAlias, targetTableName);
                    }
                }

                if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
                    environment.logOut(" > add peer : " + targetTableName + " alias : "
                        + tableAlias);
            }
        }

        for (int i = 0; i < targetClasses.getSize(); i++)
        {
            classInfo = targetClasses.getInfo(i);
            targetTableName = classInfo.getBinding().getBinding();

            if ((targetTableName == null) && (classInfo.getPeerClass() instanceof LyUnionClass))
            {
                LyClass peerClass = classInfo.getPeerClass();

                peerClasses = getPeerClasses(peerClass);

                for (Iterator<LyClass> it = peerClasses.iterator(); it.hasNext();)
                {
                    peerClass = it.next();
                    targetTableName = getTableName(peerClass);

                    if (!targetTableName.equals(tableName))
                    {

                        if ((optimize) && (!usedTablesNames.contains(targetTableName)))
                            continue;

                        tableAlias = getTableAlias(classInfo, targetTableName);

                        // avoid duplication of table aliases
                        if (foundTableNames.contains(tableAlias)
                                || (foundTableNames.contains(targetTableName)
                                    && tableAlias.equals(targetTableName)))
                            continue;

                        if (!foundTableNames.contains(targetTableName))
                        {
                            foundTableNames.addElement(targetTableName);
                            foundTableNames.addElement(tableAlias);
                        }

                        request.append(", ").append(QUOTE_CHAR_LEFT).append(targetTableName).append(
                            QUOTE_CHAR_RIGHT);

                        if (!tableAlias.equals(targetTableName))
                        {
                            request.append(" ").append(tableAlias);
                            foundAliases.put(tableAlias, targetTableName);
                        }

                        if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
                            environment.logOut(" > add target : " + targetTableName + " alias : "
                                + tableAlias);
                    }
                }
            }
            else if ((targetTableName != null)
            	&& !targetTableName.equals(tableName))
            {
                tableAlias = getTableAlias(classInfo, targetTableName);

                if (foundTableNames.contains(targetTableName) && tableAlias.equals(targetTableName))
                    continue;

                if ((optimize) && (!usedTablesNames.contains(targetTableName)))
                    continue;

                if (!foundTableNames.contains(targetTableName))
                    foundTableNames.addElement(targetTableName);

                request.append(", ").append(QUOTE_CHAR_LEFT).append(targetTableName).append(
                    QUOTE_CHAR_RIGHT);

                if (!tableAlias.equals(targetTableName))
                {
                    request.append(" ").append(tableAlias);
                    foundAliases.put(tableAlias, targetTableName);
                }

                if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
                    environment.logOut(" > add target : " + targetTableName + " alias : "
                        + tableAlias);
            }
        }

        for (int i = 0; i < usedTablesNames.size(); i++)
        {
            targetTableName = usedTablesNames.get(i);

            if (targetTableName == null)
                continue;

            if (!targetTableName.equals(tableName))
            {

                if (foundTableNames.contains(targetTableName))
                    continue;

                foundTableNames.addElement(targetTableName);

                request.append(", ").append(QUOTE_CHAR_LEFT).append(targetTableName).append(
                    QUOTE_CHAR_RIGHT);

                if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
                    environment.logOut(" > add target : " + targetTableName);
            }
        }

        // here we hare, we shall also add all remaining aliases
        HashMap<String, String> aliasMap = getRequestInfo().getAllAliases();

        for (Iterator<String> iter = aliasMap.keySet().iterator(); iter.hasNext();)
        {
            tableAlias = iter.next();

            if (foundAliases.get(tableAlias) == null)
            {
                targetTableName = aliasMap.get(tableAlias);

                if (!foundTableNames.contains(targetTableName))
                    foundTableNames.addElement(targetTableName);

                request.append(", ").append(QUOTE_CHAR_LEFT).append(targetTableName).append(
                    QUOTE_CHAR_RIGHT);
                request.append(" ").append(tableAlias);
            }
        }

        if (usesUnionSelect() && (getRequestInfo().getUnionSelectInfos(0) != null))
        {
            int      index = 0;
            Object[] unionSelectInfo = getRequestInfo().getUnionSelectInfos(index);
            String   childFilter;
            String   beginRequest = request.toString();
            int      k = beginRequest.lastIndexOf("select ");

            if (k > 0)
                beginRequest = beginRequest.substring(k);

            while (unionSelectInfo != null)
            {

                if (index > 0)
                    request.append(" UNION ").append(beginRequest);

                @SuppressWarnings("unchecked")
                ArrayList<String> childTargetClasses = (ArrayList<String>) unionSelectInfo[0];
                childFilter = (String) unionSelectInfo[1];

                //** add extra classes
                for (int i = 0; i < childTargetClasses.size(); i++)
                {
                    targetTableName = childTargetClasses.get(i);

                    if (!foundTableNames.contains(targetTableName))
                    {
                        tableAlias = getTableAlias(null, targetTableName);

                        // avoid duplication of table aliases
                        if (!foundTableNames.contains(tableAlias))
                        {
                            request.append(", ").append(QUOTE_CHAR_LEFT).append(targetTableName)
                            .append(QUOTE_CHAR_RIGHT);

                            if (!tableAlias.equals(targetTableName))
                                request.append(" ").append(tableAlias);
                        }
                    }
                }
                // now add filter condition
                // do as usual without union select
                and = false;

                if (classCondition != null)
                {
                    request.append(" where ");
                    request.append(LyString.replace(classCondition, UNION_SELECT_FILTER,
                            childFilter));
                    and = true;
                }

                if (filterCondition != null)
                {

                    if (and)
                        request.append(" and ");
                    else
                        request.append(" where ");

                    request.append(LyString.replace(filterCondition, UNION_SELECT_FILTER,
                            childFilter));
                    and = true;
                }

                if (unionCondition != null)
                {

                    if (and)
                        request.append(" and ");
                    else
                        request.append(" where ");

                    request.append(LyString.replace(unionCondition, UNION_SELECT_FILTER,
                            childFilter));
                    and = true;
                }

                index++;
                unionSelectInfo = getRequestInfo().getUnionSelectInfos(index);
            }

            getRequestInfo().clearCurrentUsedClasses();
            request.append(completeEncodeSelectCount());
        }
        else
        {
            String optReq = null;

            if ((foundTableNames.size() == 1) && (classCondition == null)
                    && (filterCondition == null) && (unionCondition == null))
                optReq = encodeFastGetSize(foundTableNames.get(0).toString());

            if (optReq == null)
            {

                // do as usual without union select
                if (classCondition != null)
                {
                    request.append(" where ");
                    request.append(classCondition);
                    and = true;
                }

                if (filterCondition != null)
                {

                    if (and)
                        request.append(" and ");
                    else
                        request.append(" where ");

                    request.append(filterCondition);
                    and = true;
                }

                if (unionCondition != null)
                {

                    if (and)
                        request.append(" and ");
                    else
                        request.append(" where ");

                    request.append(unionCondition);
                    and = true;
                }

                request.append(completeEncodeSelectCount());
            }
            else
                request = new StringBuffer(optReq);

            getRequestInfo().clearCurrentUsedClasses();
        }

        return request.toString();
    } // end method encodeCountRequest

    /**
     * Utility internal method (may change based on database).
     *
     * @see #encodeSelectCountEnd()
     */
    protected String encodeSelectCountBegin(LyFilter filter, boolean writeTableName)
    {
        StringBuffer sb = new StringBuffer();
        String       tableName = (writeTableName ? getTableName(_class) : null);

        // Test the filter needs a distinct on its key
        Vector<? extends LyField> idFields = _class.getIdFields();

        // note : the count(disctinct column) works with only one column
        if (idFields.size() == 1) //&& isDistinctOnKeyNeeded(filter))
        {
            sb.append("select count(distinct ");

            // add the name of the first key column
            LyColumnField field = (LyColumnField) idFields.elementAt(0);

            if (writeTableName)
                sb.append(QUOTE_CHAR_LEFT).append(getTableAlias(null, tableName)).append(
                    QUOTE_CHAR_RIGHT).append(".");

            sb.append(QUOTE_CHAR_LEFT).append(field.getId()).append(QUOTE_CHAR_RIGHT).append(
                ") from ");
        }
        else
            sb.append("select distinct count(*) from ");

        return sb.toString();
    }

    /**
     * Utility internal method (may change based on database).
     *
     * @see #encodeSelectCountBegin()
     */
    protected String completeEncodeSelectCount()
    {
        return "";
    }

    /**
     * Encodes a delete-type request.
     *
     * @param  objects list of objects to delete.
     * @return the generated request.
     */
    protected String encodeDeleteRequest(Vector<Object> objects)
    {
        StringBuffer request, condition;
        String       filterCondition;

        getRequestInfo().clearCurrentUsedClasses();

        request = new StringBuffer();

        request.append("delete from ");
        request.append(QUOTE_CHAR_LEFT).append(getTableName(_class)).append(QUOTE_CHAR_RIGHT);

        int         i, n;
        LyRowObject object;
        LyObjectId  objectId;

        n = objects.size();
        condition = new StringBuffer();

        for (i = 0; i < n; i++)
        {
            object = (LyRowObject) objects.elementAt(i);
            objectId = _class.getObjectId(object);

            if ((objectId != null) && (objectId.toString() != null)
                    && ((objectId.toString().indexOf(",,") >= 0)
                        || (objectId.toString().startsWith(","))
                        || (objectId.toString().endsWith(","))) && (_class instanceof LyPeerClass))
            {
                // TODO : workaround
                LyEnvironment environment = getEnvironment();

                objectId = ((LyPeerClass) _class).getObjectIdNoCache(object);
                environment.logTrace("Invalid ObjectId in delete request , recomputed to "
                    + objectId);
            }

            filterCondition = encodeCondition(objectId);

            if (i != 0)
                condition.append(" or ");

            if (n != 1)
                condition.append("(");

            condition.append(filterCondition);

            if (n != 1)
                condition.append(")");
        }

        request.append(" where (");
        request.append(condition.toString());
        request.append(")");

        getRequestInfo().clearCurrentUsedClasses();

        return request.toString();
    } // end method encodeDeleteRequest

    /**
     * Encodes a create-type request.
     *
     * @param  values list of values used for creation.
     * @return the generated request.
     */
    protected String encodeCreateRequest(Object object)
    {
        getRequestInfo().clearCurrentUsedClasses();

        LyColumnField field;
        Object        val;
        StringBuffer  request, values, fields;
        int           i, n;
        LyValueSet    valueSet;
        LyValue       value;

        valueSet = ((LyRowObject) object).getNewValues();

        request = new StringBuffer();
        values = new StringBuffer();
        fields = new StringBuffer();

        request.append("insert into ");
        request.append(QUOTE_CHAR_LEFT).append(getTableName(_class)).append(QUOTE_CHAR_RIGHT);

        n = ((valueSet == null) ? 0 : valueSet.getSize());

        if (n == 0)
        {
            valueSet = new LyValueSet();

            LyRowObject  rowObj = ((LyRowObject) object);
            LyTableClass tClass = rowObj.getTableClass();

            int fc = tClass.getFieldCount();

            for (int f = 0; (f < fc) && (valueSet.size() == 0); f++)
            {
                LyColumnField cField = (LyColumnField) tClass.getField(f);

                if (!tClass.getIdFields().contains(cField))
                    valueSet.addValue(cField, rowObj.getValue(cField));
            }

            n = ((valueSet == null) ? 0 : valueSet.getSize());
        }

        fields.append(" (");
        values.append("(");

        for (i = 0; i < n; i++)
        {
            value = valueSet.getFieldValue(i);

            if (i != 0)
            {
                fields.append(", ");
                values.append(", ");
            }

            field = (LyColumnField) value.getField();

            if (appendBlobFieldId(field, fields, null))
            {
                val = value.getValue();

                if (val == null)
                    val = "null";
                else
                    val = getSqlValue(field, LyFileTools.getFileName(val.toString()), TYPE_STRING);

                values.append(val);
                values.append(", ");
            }
            else
                fields.append(QUOTE_CHAR_LEFT).append(field.getId()).append(QUOTE_CHAR_RIGHT);

            val = value.getValue();

            if (val == null)
                values.append("null");
            else
            {

                if (isBlob(field) || isClob(field))
                {
                    Object[] blobInfo = getBlobInputStream(field, val);

                    addCurrentBlob(field, blobInfo[0], (Integer) blobInfo[1]);

                    if (blobInfo[2] != null)
                        // new file name
                        ((LyRowObject) object).setValue(field, blobInfo[2]);

                    val = "?"; // prepared Statement
                }
                else if (val instanceof java.util.Date)
                    val = getSqlValue(field, val, TYPE_DATE);
                else if (val instanceof Object[])
                    val = getSqlValue(field, val, TYPE_ARRAY);
                else if (val instanceof Boolean)
                    val = getSqlValue(field, val, TYPE_BOOLEAN);
                else
                    val = getSqlValue(field, val.toString(), TYPE_STRING);

                values.append(val);
            }
        }

        completeCreateRequest(valueSet, fields, values);

        fields.append(")");
        values.append(")");

        request.append(fields.toString());
        request.append(" values ");
        request.append(values.toString());

        getRequestInfo().clearCurrentUsedClasses();

        return request.toString();
    } // end method encodeCreateRequest


    /**
     */
    protected void completeCreateRequest(LyValueSet valueSet, StringBuffer fields,
        StringBuffer values)
    {
        return;
    }

    /**
     * Encodes a set-type request.
     *
     * @param  object the object to set.
     * @return the generated request.
     */
    protected String encodeSetRequest(Object object)
    {
        LyColumnField field;
        Object        val;
        StringBuffer  request, values;
        int           i, n;
        String        idCondition, str;
        LyValueSet    valueSet;
        LyValue       value;
        LyObjectId    objectId;
        LyFilter      classFilter = ((_classBinding == null) ? null : _classBinding.getFilter());

        getRequestInfo().clearCurrentUsedClasses();

        valueSet = ((LyRowObject) object).getNewValues();

        if (valueSet == null)
            return null;

        request = new StringBuffer();
        values = new StringBuffer();

        request.append("update ");
        request.append(QUOTE_CHAR_LEFT).append(getTableName(_class)).append(QUOTE_CHAR_RIGHT);
        request.append(" set ");

        n = valueSet.getSize();

        for (i = 0; i < n; i++)
        {
            value = valueSet.getFieldValue(i);

            if (i != 0)
                values.append(", ");

            field = (LyColumnField) value.getField();

            if ((isBlob(field) /* || isClob(field)*/) && (field.getId().indexOf(BLOB_SEP) > 0))
            {
                // Specific case for BLOBS and CLOBS stored in two separate columns in database,
                // first column is the name, second column is the LOB
                String id = field.getId();
                int    n1 = id.indexOf(BLOB_SEP);

                values.append(QUOTE_CHAR_LEFT).append(id.substring(0, n1)).append(QUOTE_CHAR_RIGHT);
                values.append("=");
                val = value.getValue();

                if (val == null)
                    values.append("null");
                else
                {
                    str = getSqlValue(field, LyFileTools.getFileName(val.toString()), TYPE_STRING);
                    values.append(str);
                }

                values.append(", ");
                values.append(id.substring(n1 + BLOB_SEP.length()));
            }
            else
                values.append(QUOTE_CHAR_LEFT).append(field.getId()).append(QUOTE_CHAR_RIGHT);

            values.append("=");
            val = value.getValue();

            if (val == null)
                values.append("null");
            else
            {

                if (isBlob(field) || isClob(field))
                {
                    Object[] blobInfo = getBlobInputStream(field, val);

                    addCurrentBlob(field, blobInfo[0], (Integer) blobInfo[1]);

                    if (blobInfo[2] != null)
                        // new file name
                        ((LyRowObject) object).setValue(field, blobInfo[2]);

                    str = "?"; // prepared Statement
                }
                else if (val instanceof java.util.Date)
                    str = getSqlValue(field, val, TYPE_DATE);
                else if (val instanceof Object[])
                    str = getSqlValue(field, val, TYPE_ARRAY);
                else if (val instanceof Boolean)
                    str = getSqlValue(field, val, TYPE_BOOLEAN);
                else
                    str = getSqlValue(field, val.toString(), TYPE_STRING);

                values.append(str);
            }
        }

        request.append(values.toString());

        // Retrieve the target classes for filters. If filters contain filter(s)
        // on relations, the target tables must be added to the 'FROM' clause.
        LyInfoList<LyClassInfo> targetClasses = new LyInfoList<LyClassInfo>(false);

        getTargetClasses(classFilter, targetClasses);

        // FFT 01487
        Vector<LyFilter> unionFilters = null;

        if (_classBinding instanceof LyUnionBinding)
            unionFilters = getFilters((LyUnionBinding) _classBinding);

        if (unionFilters != null)
        {

            // Check into union filters if some additional classes are required
            for (i = 0; i < unionFilters.size(); i++)
                getTargetClasses(unionFilters.elementAt(i), targetClasses);
        }

        objectId = _class.getObjectId(object);
        idCondition = encodeCondition(objectId);

        if (idCondition != null)
        {
            request.append(" where (");
            request.append(idCondition);
            request.append(")");
        }

        getRequestInfo().clearCurrentUsedClasses();

        return request.toString();
    } // end method encodeSetRequest

    /**
     */
    protected String getSqlValue(LyColumnField field, Object value, int type)
    {
        LySqlSpecificEncoder encoder;

        encoder = ((LySqlDataProvider) _provider).getSpecificEncoder(field);

        if (encoder != null)
            return encoder.getSqlValue(field, value);

        if (type == TYPE_DATE)
            return getSqlValue(field, (java.util.Date) value);
        else if (type == TYPE_ARRAY)
            return getSqlValue(field, (Object[]) value);
        else if (type == TYPE_STRING)
            return getSqlValue(field, (String) value);
        else if (type == TYPE_BOOLEAN)
            return getSqlValue(field, (Boolean) value);
        else if (value != null)
            return value.toString();

        return null;
    }

    /**
     */
    protected String getSqlValue(LyColumnField field, java.util.Date value)
    {

        if (value == null)
            return null;

        Class<?> valueClass = field.getValueClass();

        if (valueClass == java.util.Date.class)
        {
            StringBuffer  result;
            long          val;
            Time          time;
            java.sql.Date sqlDate;

            val = value.getTime();

            time = new Time(val);
            sqlDate = new java.sql.Date(val);

            result = new StringBuffer();
            result.append(CHAR_QUOTE);
            result.append(sqlDate.toString());
            result.append(" ");
            result.append(time.toString());
            result.append(CHAR_QUOTE);

            return result.toString();
        }
        else
            return getSqlValue(field, value.toString());
    } // end method getSqlValue

    /**
     * Encode multivalue String[] or Integer[] with separators.
     */
    protected String getSqlValue(LyColumnField field, Object[] values)
    {

        if (values == null)
            return null;

        StringBuffer sb = new StringBuffer(200);

        int i, n = values.length;

        for (i = 0; i < n; i++)
        {

            if (i > 0)
                sb.append(_valueSep);

            if (values[i] != null)
                sb.append(values[i].toString());
        }

        return (getSqlValue(field, sb.toString()));
    }

    /**
     */
    protected String getSqlValue(LyColumnField field, String value)
    {

        if (value == null)
            return null;

        Class<?>     valueClass = field.getValueClass();
        String       str;
        StringBuffer result;

        result = new StringBuffer();

        if (((valueClass == String.class) || (valueClass == String[].class)// FFT01374 relation on a choice field which is id field : may be a short (peer) but a string
                    || (((valueClass == Short.class) || (valueClass == String[].class))
                        && (LyTools.intFromString(value) == 0) && (!"0".equals(value)))))
        {
            // We must replace the char ' by ''
            str = LyString.replace(value, STRING_QUOTE, STRING_TWO_QUOTES);
            result.append(CHAR_QUOTE);

            // If we must trim values, it is done now...
            if (_trim && (str != null))
                str = str.trim();

            result.append(str);
            result.append(CHAR_QUOTE);
        }
        else if (valueClass == java.util.Date.class)
        {
            java.util.Date date;
            long           val;
            Time           time;
            java.sql.Date  sqlDate;

            date = LyDateFormatter.fromString(value);
            val = date.getTime();

            time = new Time(val);
            sqlDate = new java.sql.Date(val);

            result.append(CHAR_QUOTE);
            result.append(sqlDate.toString());
            result.append(" ");
            result.append(time.toString());
            result.append(CHAR_QUOTE);
        }
        else
            result.append(value);

        return result.toString();
    } // end method getSqlValue

    /**
     * Convert a boolean value to an SQL value.<br>
     * This generic implementation returns '0' for false and '1' for true.
     */
    protected String getSqlValue(LyColumnField field, Boolean value)
    {
        if (value == null)
            return null;

        Class<?> valueClass = field.getValueClass();

        if ((valueClass == Boolean.class) || (valueClass == Short.class))
        {
            if (value.booleanValue())
                return "1";
            else
                return "0";
        }
        else
            return getSqlValue(field, value.toString());
    }


    /**
     * This method is used after an insert request to complete the object if some values or the
     * identifier is set by the database.
     */
    protected void completeObject(Object object, LySqlConnection connection)
    {
    }

    /**
     * Methods for decoding results.*
     */
    /**
     * Decodes a select-type join request.
     *
     * @param  encoder the union encoder for managing the results.
     * @param  rs      the ResultSet to decode.
     * @param  fields  the fields that were loaded (if null every field of the class)
     * @return the objects decoded from the ResultSet.
     */
    protected Collection<LyValueSet> decodeSelectValues(LyUnionRequestEncoder encoder, ResultSet rs,
        Vector<LyField> fields) throws LySqlUnsupportedConversionException, SQLException
    {
        Collection<LyValueSet> values = new ArrayList<LyValueSet>();
        LyColumnField          field;
        Object                 value;
        LyValueSet             valueSet;
        int                    i, n = fields.size();

        // Reading the results.
        while (rs.next())
        {
            valueSet = new LyValueSet();

            // Fields are in the right order.
            for (i = 0; i < n; i++)
            {
                field = (LyColumnField) fields.elementAt(i);
                value = getPeerValue(rs, field, i + 1, field.getValueClass());
                valueSet.addValue(field, value);
            }

            values.add(valueSet);
        }

        return values;
    }

    /**
     * Decodes a select-type request.
     *
     * @param  rs     the ResultSet to decode.
     * @param  fields the fields that were loaded (if null every field of the class)
     * @return the objects decoded from the ResultSet.
     */
    protected Vector<LyRowObject> decodeSelectValues(ResultSet rs, Vector<LyField> fields)
        throws LySqlUnsupportedConversionException, SQLException
    {
        Vector<LyRowObject> objects = new Vector<LyRowObject>();
        LyRowObject         object;
        LyColumnField       field;
        int                 i, n = fields.size(), blobSepOffset = 0; //offset is when a blob value is stored on 2 fields
        Object[]            values;
        int[]               indexes = new int[n];
        int                 nf = ((LyTableClass) _class).getFieldCount();

        for (i = 0; i < n; i++)
        {
            field = (LyColumnField) fields.elementAt(i);
            indexes[i] = ((LyTableClass) _class).getIndex(field);
        }

        // Reading the results.
        while (rs.next())
        {
            values = new Object[nf];

            if (n < nf)
            {

                for (i = 0; i < nf; i++)
                    values[i] = LyPeerClass.VALUE_NOT_LOADED;
            }

            // Fields are in the right order.
            for (i = 0, blobSepOffset = 0; i < n; i++)
            {
                field = (LyColumnField) fields.elementAt(i);
                values[indexes[i]] = getPeerValue(rs, field, i + 1 + blobSepOffset,
                        field.getValueClass());

                if (field.getId().indexOf(BLOB_SEP) > 0)
                    blobSepOffset++;
            }

            object = new LyRowObject((LyTableClass) _class, values);
            objects.addElement(object);
        }

        return objects;
    } // end method decodeSelectValues

    /**
     * To decode a value given by the database.
     */
    protected Object getPeerValue(ResultSet rs, LyColumnField field, int index, Class<?> valueClass)
        throws LySqlUnsupportedConversionException, SQLException
    {
        LySqlSpecificEncoder encoder;

        encoder = ((LySqlDataProvider) _provider).getSpecificEncoder(field);

        if (encoder != null)
            return encoder.getPeerValue(rs, field, index, valueClass);

        Object value;
        String str;

        if (valueClass == String.class)
        {

            if (isBlob(field) || isClob(field))
                str = getPeerValueForBlob(rs, field, index, valueClass);
            else
                str = rs.getString(index);

            if ((str != null) && (_trim))
                str = str.trim();

            value = str;
        }
        else if (valueClass == String[].class)
        {
            str = rs.getString(index);

            if ((str != null) && (_trim))
                str = str.trim();

            value = LyString.stringToList(str, LyTools.charFromString(_valueSep));
        }
        else if (valueClass == Short.class)
            value = new Short(rs.getShort(index));
        else if (valueClass == java.util.Date.class)
        {
            int type = ((LySqlColumnField) field).getDateType();

            if (type == java.sql.Types.TIMESTAMP)
            {
                String format = ((LySqlColumnField) field).getFormat();

                if (format != null)
                {
                    value = rs.getTimestamp(index);

                    if (value != null)
                        value = LyDateFormatter.buildDate(((java.sql.Timestamp) value).getTime(),
                                format);
                }
                else
                {
                    str = rs.getString(index);

                    if (str == null)
                        value = null;
                    else if (str.length() >= 19)
                    {

                        // Optimisation : hard-decoding of date string
                        // Standard format is : YYYY-MM-DD HH:MM:SS.N
                        try
                        {
                            value = LyDateFormatter.buildDate(Integer.parseInt(str.substring(0, 4))
                                    - 1900, Integer.parseInt(str.substring(5, 7)) - 1,
                                    Integer.parseInt(str.substring(8, 10)),
                                    Integer.parseInt(str.substring(11, 13)),
                                    Integer.parseInt(str.substring(14, 16)),
                                    Integer.parseInt(str.substring(17, 19)));
                        }
                        catch (Exception e)
                        {
                            value = rs.getTimestamp(index);

                            if (value != null)
                                value = LyDateFormatter.buildDate(((java.sql.Timestamp) value)
                                        .getTime(), null);
                        }
                    }
                    else if ((str.length() == 8) && (str.charAt(2) == ':')
                            && (str.charAt(5) == ':'))
                    {

                        // Optimisation : hard-decoding of time string
                        // Standard format is : HH:MM:SS
                        try
                        {
                            value = LyDateFormatter.buildDate(0, 0, 0,
                                    Integer.parseInt(str.substring(0, 2)),
                                    Integer.parseInt(str.substring(3, 5)),
                                    Integer.parseInt(str.substring(6, 8)));
                        }
                        catch (Exception e)
                        {
                            value = rs.getTimestamp(index);

                            if (value != null)
                                value = LyDateFormatter.buildDate(((java.sql.Timestamp) value)
                                        .getTime(), "hh:mm:ss");
                        }
                    }
                    else if ((str.length() == 10) && (str.charAt(4) == '-')
                            && (str.charAt(7) == '-'))
                    {

                        // Optimisation : hard-decoding of date string
                        // Format is : YYYY-MM-DD
                        try
                        {
                            value = LyDateFormatter.buildDate(Integer.parseInt(str.substring(0, 4))
                                    - 1900, Integer.parseInt(str.substring(5, 7)) - 1,
                                    Integer.parseInt(str.substring(8, 10)), 0, 0, 0);
                        }
                        catch (Exception e)
                        {
                            value = rs.getTimestamp(index);

                            if (value != null)
                                value = LyDateFormatter.buildDate(((java.sql.Timestamp) value)
                                        .getTime(), "yyyy:MM:dd");
                        }
                    }
                    else
                    {
                        value = rs.getTimestamp(index);

                        if (value != null)
                            value = LyDateFormatter.buildDate(((java.sql.Timestamp) value)
                                    .getTime(), null);
                    }
                }
            }
            else if (type == java.sql.Types.DATE)
            {
                value = rs.getDate(index);

                if (value != null)
                    value = LyDateFormatter.buildDate(((java.sql.Date) value).getTime(),
                            "yyyy:MM:dd");
            }
            else if (type == java.sql.Types.TIME)
            {
                value = rs.getTime(index);

                if (value != null)
                    value = LyDateFormatter.buildDate(((java.sql.Time) value).getTime(),
                            "hh:mm:ss");
            }
            else
            {
                LyEnvironment environment = getEnvironment();
                throw new LySqlUnsupportedConversionException("BF AppEngine : "
                    + environment.getMessage("LY_ER_UNSUPPORTED_CONVERSION", field.getId(),
                        valueClass.getName()));
            }
        }
        else if (valueClass == Integer.class)
            value = new Integer(rs.getInt(index));
        else if (valueClass == Long.class)
            value = new Long(rs.getLong(index));
        else if (valueClass == Float.class)
            value = new Float(rs.getFloat(index));
        else if (valueClass == Double.class)
            value = new Double(rs.getDouble(index));
        else if (valueClass == Boolean.class)
            value = Boolean.valueOf(rs.getBoolean(index));
        else
        {
            rs.close();

            LyEnvironment environment = getEnvironment();
            throw new LySqlUnsupportedConversionException("BF AppEngine : "
                + environment.getMessage("LY_ER_UNSUPPORTED_CONVERSION", field.getId(),
                    valueClass.getName()));
        }

        if (rs.wasNull())
            value = null;

        return value;
    } // end method getPeerValue

    protected String getPeerValueForBlob(ResultSet rs, LyColumnField field, int index,
        Class<?> valueClass) throws LySqlUnsupportedConversionException, SQLException
    {
        String str = null;

        if (isBlob(field))
        {
            int         sepIndex = field.getId().indexOf(BLOB_SEP);
            InputStream iStream = null;
            String      fileNameStr = null;
            int         i, blobLength, nBytesRead = 0;
            File        f;
            Blob        blob;

            // 2 possibilities : the file name is in the blob itself, or in another column in the database (sepIndex > 0)
            if (sepIndex > 0)
            {
                fileNameStr = rs.getString(index);
                // two values in two fields : filename then BLOB
                blob = rs.getBlob(index + 1);
            }
            else
                // one BLOB only or filename in the blob
                blob = rs.getBlob(index);



            // FFT00378 Flandres : blob null
            if (blob == null)
                blobLength = 0;
            else
            {
                iStream = blob.getBinaryStream();
                blobLength = (int) blob.length();
            }

            if (iStream == null)
                str = null;
            else
            {
                try
                {
                    byte[]              fileNameBytes = (LySqlRequestEncoder.BLOB_SEP).getBytes("UTF-8");
                    PushbackInputStream pbIStream = new PushbackInputStream(iStream,
                            fileNameBytes.length);

                    // do we have a value or a file or a filename + file ?
                    if (fileNameStr == null)
                    {
                        int     l = fileNameBytes.length;
                        int[]   c = new int[l];
                        boolean foundSep = false;

                        for (i = 0; i < l; i++)
                            c[i] = pbIStream.read();

                        nBytesRead = nBytesRead + l;

                        for (i = 0; i < l; i++)
                        {

                            if ((((byte) c[i]) != fileNameBytes[i]) || (c[i] < 0))
                                break; // Error : BLOB_SEP not found
                        }

                        if (i >= l)
                            foundSep = true;
                        else
                        {

                            // backward compatibility (non  UTF-8)
                            if ((c[0] == BLOB_SEP.charAt(0)) && (c[1] == BLOB_SEP.charAt(1)))
                            {
                                foundSep = true;

                                for (i = l - 1; i >= 2; i--)
                                    pbIStream.unread(c[i]);

                                l = 2;
                                fileNameBytes[0] = (byte) c[0];
                                fileNameBytes[1] = (byte) c[1];
                            }
                        }

                        if (foundSep)
                        {
                            // case : BLOB_SEP+filename+BLOB_SEP+<binary content>
                            foundSep = false; // waiting for next sep

                            //StringBuffer fileName = new StringBuffer();
                            List<Byte> fileNameResultBytes = new ArrayList<Byte>();

                            for (i = 0; i < l; i++)
                                c[i] = pbIStream.read();

                            nBytesRead = nBytesRead + l;

                            while (!foundSep && (c[0] >= 0))
                            {

                                for (i = 0; i < l; i++)
                                {

                                    if ((((byte) c[i]) != fileNameBytes[i]) || (c[i] < 0))
                                        break; // BLOB_SEP not found yet
                                }

                                if (i >= l)
                                    foundSep = true;
                                else
                                {
                                    fileNameResultBytes.add(new Byte((byte) c[0]));

                                    //fileName.append((char)c[0]);
                                    for (i = 0; i < (l - 1); i++)
                                        c[i] = c[i + 1];

                                    c[l - 1] = pbIStream.read();
                                    nBytesRead = nBytesRead + 1;
                                }
                            }

                            byte[] fileNameResultBytesArray = new byte[fileNameResultBytes.size()];

                            for (int idx = 0; idx < fileNameResultBytes.size(); idx++)
                                fileNameResultBytesArray[idx] = fileNameResultBytes.get(idx);

                            fileNameStr = new String(fileNameResultBytesArray, "UTF-8"); //fileName.toString();
                        }
                        else
                        {

                            // directly the value
                            for (i = l - 1; i >= 0; i--)
                                pbIStream.unread(c[i]);

                            nBytesRead = nBytesRead - l;
                        }
                    }

                    byte[] data;
                    int    newBytesRead;

                    if (fileNameStr == null)
                    {
                        ByteArrayOutputStream bOut = new ByteArrayOutputStream();

                        while (nBytesRead < blobLength)
                        {

                            if ((blobLength - nBytesRead) < MAX_BLOB_BUFFER)
                                data = new byte[blobLength - nBytesRead];
                            else
                                data = new byte[MAX_BLOB_BUFFER];

                            newBytesRead = pbIStream.read(data);
                            bOut.write(data, 0, newBytesRead);
                            nBytesRead = nBytesRead + newBytesRead;
                        }

                        str = bOut.toString();
                    }
                    else
                    {
                        f = createTempfile(fileNameStr, null);
                        nBytesRead = writeDataToFile(pbIStream, f, nBytesRead, blobLength);
                        return f.getAbsolutePath();
                    }

                    pbIStream.close();
                }
                catch (IOException e)
                {
                    getEnvironment().writeException(e);
                    str = null;
                }
            }
        }
        else if (isClob(field))
        {
            Reader iStream = rs.getCharacterStream(index);

            if (iStream == null)
                str = null;
            else
            {

                char[] data;
                int    nBytesRead = 0;
                int    blobLength, newBytesRead;
                Clob   clob;

                try
                {
                    clob = rs.getClob(index);

                    if (clob == null)
                        str = null; // FFT00378 Flandres
                    else
                    {
                        blobLength = (int) clob.length();

                        StringWriter bOut = new StringWriter();

                        while (nBytesRead < blobLength)
                        {

                            if ((blobLength - nBytesRead) < MAX_BLOB_BUFFER)
                                data = new char[blobLength - nBytesRead];
                            else
                                data = new char[MAX_BLOB_BUFFER];

                            newBytesRead = iStream.read(data);
                            bOut.write(data, 0, newBytesRead);
                            nBytesRead = nBytesRead + newBytesRead;
                        }

                        str = bOut.toString();
                    }
                }
                catch (IOException e)
                {
                    getEnvironment().writeException(e);
                    str = null;
                }
            }
        }

        return str;
    } // end method getPeerValueForBlob


    /**
     * Decodes a count-type request.
     *
     * @param  rs the ResultSet to decode.
     * @return the result of the count request as an Integer.
     */
    protected Integer decodeCountValues(ResultSet rs) throws SQLException
    {
        int n;

        rs.next();
        n = rs.getInt(1);

        return new Integer(n);
    }

    /**
     * Returns a boolean that indicates if case sensitive is supported. Default implementation
     * return true.
     *
     * @return True if supported, false otherwise.
     */
    public boolean supportCaseSensitive()
    {
        return true;
    }

    /**
     * To set the list of peer classes corresponding to the current encoding join request.
     */
    protected void setJoinPeerClasses(Collection<LyClass> joinPeerClasses)
    {
        _joinPeerClasses = joinPeerClasses;
    }

    /**
     * Return a new instance of the specified class. The encoder must build the object with the same
     * method as it loads the object itself.
     *
     * @param  peerClass the physical class of the created object.
     * @return a new instance of of a physical object of the specified class.
     */
    public Object buildObject(LyUnionRequestEncoder encoder, LyClass peerClass)
    {
        return new LyRowObject((LyTableClass) peerClass, new LyValueSet());
    }

    /**
     * Reset current blobs used in the values for insert or update.
     */
    protected static void clearCurrentBlobs()
    {

        if (_usedBlobsList == null)
            return;

        _usedBlobsList.remove(Thread.currentThread());
    }

    /**
     * Add a new entry in the current blobs used in the values for insert or update.
     *
     * @param field  the field mapped on a BLOB in DB
     * @param value  value specified for this blob : must be an InputStream or a Reader
     * @param length number of bytes to read from the stream
     */
    protected static void addCurrentBlob(LyColumnField field, Object value, Integer length)
    {

        if ((!(value instanceof Reader)) && (!(value instanceof InputStream)))
            throw new IllegalArgumentException("Value for Blob/Clob " + field.getId()
                + " must be a Reader or an InputStream. Value=" + value);

        if (_usedBlobsList == null)
            _usedBlobsList = new HashMap<Thread, ArrayList<Object[]>>();

        // Use an array list to set the blobs : ORDER IS IMPORTANT.
        ArrayList<Object[]> blobList = _usedBlobsList.get(Thread.currentThread());

        if (blobList == null)
        {
            blobList = new ArrayList<Object[]>();
            _usedBlobsList.put(Thread.currentThread(), blobList);
        }

        blobList.add(new Object[] { field, value, length });
    }

    /**
     * fetch an existing entry in the current blobs used in the values for insert or update.
     *
     * @param  index index of the plob (position in the value list, starts at 0)
     * @return a couple (field : the field mapped on a BLOB in DB, value : value specified for this
     *         blob)
     * @see    addCurrentBlob
     */
    protected static Object[] getCurrentBlob(int index)
    {

        if (_usedBlobsList == null)
            return null;

        // Use an array list to set the blobs : ORDER IS IMPORTANT.
        ArrayList<Object[]> blobList = _usedBlobsList.get(Thread.currentThread());

        if ((blobList == null) || (blobList.size() <= index))
            return null;

        return blobList.get(index);
    }

    /**
     * Compute the aliases of the tables related to a given set of classes.
     *
     * @param classes a set of logical classes.
     */
    protected void computeTableAliases(LyInfoList<LyClassInfo> classes)
    {
        LyClassInfo classInfo;
        LyBinding   binding;
        String      targetTableName;
        String      tableAlias;

        for (int k = 0; k < classes.getSize(); k++)
        {
            classInfo = classes.getInfo(k);
            binding = classInfo.getBinding();
            targetTableName = (binding == null) ? null : binding.getBinding();

            if ((targetTableName == null) && (classInfo.getPeerClass() instanceof LyUnionClass))
            {
                LyClass            peerClass = classInfo.getPeerClass();
                ArrayList<LyClass> peerClasses = getPeerClasses(peerClass);

                for (Iterator<LyClass> it = peerClasses.iterator(); it.hasNext();)
                {
                    peerClass = it.next();
                    targetTableName = getTableName(peerClass);
                    tableAlias = computeTableAlias(targetTableName);

                    addTableAlias(classInfo, targetTableName, tableAlias);
                }
            }
            else
            {
                tableAlias = computeTableAlias(targetTableName);
                addTableAlias(classInfo, targetTableName, tableAlias);
            }
        }
    } // end method computeTableAliases

    /**
     * @return a boolean that tells if this encoder uses aliases with table.
     */
    protected boolean usesTableAliases()
    {
        return true;
    }

    /**
     * Check if this encore accepts union select to optimize extended filter on multiple union
     * classes.
     *
     * @return a boolean that tells if this encoder uses the union select optimization.
     */
    protected boolean usesUnionSelect()
    {
        return false;
    }

    /**
     * Compute an alias for a table.<br>
     * This alias is used for tables when multiple occurences of a table are used in the same
     * request but in different context (for different logical classes).
     *
     * @param  tableName a table name
     * @return an alias
     */
    protected String computeTableAlias(String tableName)
    {

        if (tableName == null)
            return null;

        if (!usesTableAliases())
            return tableName;

        // count the number of previous aliases on the same table
        int count = _tableAliasCounter++;

        return "LY_" + count;
    }

    /**
     * Store a table alias for a given logical class and a table name.
     *
     * @param classInfo target logical class
     * @param tableName table name
     * @param alias     table alias
     */
    protected void addTableAlias(LyClassInfo classInfo, String tableName, String alias)
    {
        addTableAlias(classInfo, tableName, alias, true, false);
    }

    /**
     * Store a table alias for a given logical class and a table name.
     *
     * @param classInfo target logical class
     * @param tableName table name
     * @param alias     table alias
     * @param counted   indicates if this alias is COUNTED in the used alias or not. Excepted for
     *                  internal use the counted status should be true.
     */
    protected void addTableAlias(LyClassInfo classInfo, String tableName, String alias,
        boolean counted, boolean isTargetRelationClass)
    {

        if ((alias == null) || (tableName == null) || alias.equals(tableName))
            return;

        getRequestInfo().addTableAlias(null, classInfo, null, tableName, alias, counted,
            isTargetRelationClass);
    } // end method addTableAlias

    /**
     * Search for the table alias of a given table for a given logical class.
     *
     * @param  classInfo target logical class
     * @param  tableName table name
     * @return the table alias if found, the table name otherwise.
     */
    protected String getTableAlias(LyClassInfo classInfo, String tableName)
    {
        String resu = null;

        if (!usesTableAliases())
            resu = tableName;

        resu = getRequestInfo().getTableAlias(classInfo, null, tableName);

        return resu;
    } // end method getTableAlias

    /**
     * Recursive method that check if a filter or one of its sub-filter has a simple filter
     * expression. It ignores relation filters.
     */
    protected boolean hasFilterExpression(LyFilter filter)
    {
        LyFilterElement element;

        for (int i = 0; i < filter.getSize(); i++)
        {
            element = filter.getElement(i);

            if (element instanceof LyFilter)
            {

                if (hasFilterExpression((LyFilter) element))
                    return true;
            }
            else if (element instanceof LyFilterExpression)
                return true;
        }

        return false;
    }

    /**
     * Recursive method that check if a filter or one of its sub-filter has a relation filter.
     */
    protected boolean hasFilterRelation(LyFilter filter)
    {
        LyFilterElement element;

        for (int i = 0; i < filter.getSize(); i++)
        {
            element = filter.getElement(i);

            if (element instanceof LyFilter)
            {

                if (hasFilterRelation((LyFilter) element))
                    return true;
            }
            else if (element instanceof LyFilterRelation)
                return true;
        }

        return false;
    }

    /**
     * Escape a like condition in request.
     *
     * @author nbo
     * @param  condition the condition to use in like <condition> request
     * @return the encoded value
     * @see    getEscapeCharacter()
     */
    protected String encodeLikeCondition(String condition)
    {

        if (condition == null)
            return null;

        int          i, n = condition.length();
        StringBuffer sb = new StringBuffer(n);
        char         c;

        for (i = 0; i < n; i++)
        {
            c = condition.charAt(i);

            if (c == '\'')
                sb.append(c); // Double quote
            else if ((c == '_') || (c == '%'))
            {
                sb.append(getEscapeCharacter()); // Escape sequence

                if (_escapeSequence == null)
                    _escapeSequence = " ESCAPE '" + getEscapeCharacter() + "'"; // Start with whitespace
            }

            sb.append(c);
        }

        return sb.toString();
    }

    /**
     * Escape character used in a like condition in request.
     *
     * @author nbo
     * @return the escape character
     */
    protected char getEscapeCharacter()
    {
        return '^';
    }

    /**
     * Append the id of the blob field (if needed) to the specified buffer. Specific case for BLOBS
     * and CLOBS stored in two separate columns in database first column is the name, second column
     * is the LOB separated by BLOB_SEP
     *
     * @param  field  the field to add
     * @param  buffer the target string buffer
     * @return true if added, false otherwise
     */
    protected boolean appendBlobFieldId(LyColumnField field, StringBuffer buffer, String tableName)
    {

        if ((isBlob(field) /* || isClob(field)*/) && (field.getId().indexOf(BLOB_SEP) > 0))
        {
            // Specific case for BLOBS and CLOBS stored in two separate columns in database,
            // first column is the name, second column is the LOB
            String id = field.getId();
            int    n1 = id.indexOf(BLOB_SEP);

            if (tableName != null)
                buffer.append(QUOTE_CHAR_LEFT).append(tableName).append(QUOTE_CHAR_RIGHT).append(
                    ".");

            buffer.append(QUOTE_CHAR_LEFT).append(id.substring(0, n1)).append(QUOTE_CHAR_RIGHT);
            buffer.append(", ");

            if (tableName != null)
                buffer.append(QUOTE_CHAR_LEFT).append(tableName).append(QUOTE_CHAR_RIGHT).append(
                    ".");

            buffer.append(QUOTE_CHAR_LEFT).append(id.substring(n1 + BLOB_SEP.length())).append(
                QUOTE_CHAR_RIGHT);
            return true;
        }

        return false;

    }

    /**
     * Get or create the request info : this is the context for aliases and encoding info to
     * generate the request.
     *
     * @return the requestInfo
     */
    public LySqlRequestInfo getRequestInfo()
    {

        if (_sqlRequestInfo == null)
            _sqlRequestInfo = new LySqlRequestInfo(_class);

        return _sqlRequestInfo;
    }

    /**
     * Accessor (write) the request info : this method is for internal use and should be used with
     * care (unpredictable aliases may occur in SQL Requests).
     *
     * @param sqlRequestInfo the requestInfo
     */
    protected void setRequestInfo(LySqlRequestInfo sqlRequestInfo)
    {
        _sqlRequestInfo = sqlRequestInfo;
    }

    /**
     * Transfer the request info to the specified encoder the specified encoders's request info
     * becomes a new child of current context.
     *
     * @param encoder   target encoder
     * @param childType the type of child, one of the TYPE_xxx constants. parameter is ignored when
     *                  parent is null.
     */
    public LySqlRequestInfo setChildRequestInfo(LySqlRequestEncoder encoder,
        LySqlRequestInfo.Join childType)
    {
        LySqlRequestInfo targetRequestInfo = encoder.getRequestInfo();

        if (targetRequestInfo.getParent() == getRequestInfo())
            return targetRequestInfo; // already set

        if (targetRequestInfo.getParent() != null)
            throw new IllegalStateException(
                "Parent request info already defined : current is for class '" + _class
                + "' and child is '" + encoder._class + "' and concurrent parent is "
                + targetRequestInfo.getParent().getDataClass());

        targetRequestInfo.setParent(getRequestInfo(), childType);
        return targetRequestInfo;
    }

    /**
     * We are noi longer the parent of the specified encoder's request info the specified encoders's
     * request info becomes a new child of current context.
     *
     * @param encoder target encoder
     */
    public LySqlRequestInfo clearChildRequestInfo(LySqlRequestEncoder encoder)
    {
        LySqlRequestInfo targetRequestInfo = encoder.getRequestInfo();

        if (targetRequestInfo.getParent() != getRequestInfo())
            throw new IllegalStateException(
                "Parent request info is *not* this : current is for class '" + _class
                + "' and child is '" + encoder._class + "' and concurrent parent is "
                + targetRequestInfo.getParent());

        // *** We may need to report aliases to the higher level ?
        targetRequestInfo.setAliasContext(null, null, null);
        targetRequestInfo.setParent(null, LySqlRequestInfo.Join.NOTHING);

        return targetRequestInfo;
    }

    /**
     * encode the limit condition.
     *
     * @param limit max numbut of rows fetched
     */
    public String encodeLimitCondition(int limit)
    {
        return null;
    }

    /**
     * May propose a fast way to fetch the size of the table (DB specific)
     * @param tableName the name of the table
     * @return the request or null to use generic SQL select count ...
     */
    protected String encodeFastGetSize(String tableName)
    {
        return null;
    }

    /**
     * Indicates if the connector accepts the storage of additional info in the blob (like file name)
     */
    protected boolean allowAdditionalInfoInBlob()
    {
        return true;
    }

    /**
     * Gets selected field for a multilingual field.
     * @param lang          Selected lang.
     * @param field         Selected field.
     * @param tableName     Table name
     * @return Selected lang field.
     */
    protected String getSelectedLangField(String lang, LyField field, String tableName, boolean as)
    {
        if (lang == null)
            return null;

        String locate = getLocateFunction();
        String substring = getSubstringFunction();

        if ((locate == null) || (substring == null))
            return null;

        String id = field.getId();
        String quotedId;

        quotedId = QUOTE_CHAR_LEFT;

        if (tableName != null)
            quotedId += getTableAlias(null, tableName) + QUOTE_CHAR_RIGHT + "." + QUOTE_CHAR_LEFT;

        quotedId += id + QUOTE_CHAR_RIGHT;

        String startLang = LyTextFieldInfo.OPEN_START_LANG + lang
            + LyTextFieldInfo.CLOSE_START_LANG;
        String endLang = LyTextFieldInfo.OPEN_END_LANG + lang + LyTextFieldInfo.CLOSE_END_LANG;

        String	selectedLangField = "IF(" + locate + "('" + startLang + "'," + quotedId + ")>0," + substring + "("
            + quotedId + "," + locate + "('" + startLang + "'," + quotedId + ")+"
            + startLang.length() + "," + locate + "('" + endLang + "', " + quotedId + ")-" + locate
            + "('" + startLang + "'," + quotedId + ") - " + startLang.length() + ")," + quotedId + ")";

        if (as)
        	selectedLangField += " as " + id + "_" + lang.toUpperCase();
        return selectedLangField;
    }

    /**
     * Sort condition for multilingual text and current language.
     */
    protected String getSortLangField(String quotedId, LyTextFieldInfo textFieldInfo)
    {

        if (!textFieldInfo.isMultilingual() || (getLocateFunction() == null)
                || (getSubstringFunction() == null))
            return quotedId;

        LyApplication application = _provider.getApplication();
        LyDataSession currentSession = application.getCurrentSession();

        if (currentSession != null)
        {
            String  currentLang = currentSession.getEnvironment().getLanguage();
            LyField field = _classBinding.getPeerField(textFieldInfo);

            return field.getId() + '_' + currentLang.toUpperCase();
        }
        else
            return quotedId;
    }

    /**
     * Gets locate function.
     */
    protected String getLocateFunction()
    {
        return null;
    }

    /**
     * Gets substring function.
     */
    protected String getSubstringFunction()
    {
        return "SUBSTRING";
    }
} // end class LySqlRequestEncoder
